
#
# PyGLoo
#
# https://github.com/benos-9k/PyGLoo
#
# Python GL runtime function loader
# Interaction with ctypes is necessary to use loaded functions.
# Currently for Windows and Linux.
#
# Designed to circumvent Maya being shit.
#
# Title: PyGLoo (Python + GLEW, but unrelated to GLEW itself)
#
# @author Ben Allen
#

import os
import platform
import ctypes

__version__ = '0.1.1'

# class to contain the loaded function pointers
class Context(object):
	def __init__(self):
		self._begin = False
	# }
# }

print 'PyGLoo: initializing GL library...'

# function for making function types
APIFUNCTYPE = None
if os.name == 'nt':
	APIFUNCTYPE = ctypes.WINFUNCTYPE
	print 'PyGLoo: using stdcall on win32'
elif os.name == 'posix':
	APIFUNCTYPE = ctypes.CFUNCTYPE
	print 'PyGLoo: using cdecl on posix'
else:
	APIFUNCTYPE = ctypes.CFUNCTYPE
	print 'PyGLoo: Warning: unknown OS, guessing cdecl'
# }

# function for making pointer types
POINTER = ctypes.POINTER

# load the GL dll and function-loading-function
_libgl = None
_getProcAddressImpl = None
if os.name == 'nt':
	print 'PyGLoo: loading opengl32.dll on win32'
	_libgl = ctypes.windll.LoadLibrary('opengl32')
	_getProcAddressImpl = _libgl.wglGetProcAddress
elif os.name == 'posix':
	print 'PyGLoo: loading libGL.so on posix'
	_libgl = ctypes.cdll.LoadLibrary('libGL.so')
	_getProcAddressImpl = _libgl.glXGetProcAddress
else:
	print 'PyGLoo: Warning: unknown OS, guessing libGL.so'
	_libgl = ctypes.cdll.LoadLibrary('libGL.so')
	_getProcAddressImpl = _libgl.glXGetProcAddress
# }
_getProcAddressImpl.restype = APIFUNCTYPE(None)
_getProcAddressImpl.argtypes = (ctypes.c_char_p,)

def _getProcAddress(name):
	'Wrapper for (wgl/glX)GetProcAddress that searches directly in the DLL if needed'
	f = _getProcAddressImpl(name)
	if not f:
		# entry point not found, this is typical of GL 1.x functions on Windows
		try:
			f = _libgl.__getattr__(name)
		except AttributeError:
			# not here either
			pass
		# }
	# }
	return f
# }

print 'PyGLoo: initialized GL library'

print 'PyGLoo: initializing type system...'

# mappings for fixed bitwidth integral types
# only really needed for defining GLintptr
_int_types = [ctypes.c_byte, ctypes.c_short, ctypes.c_int, ctypes.c_long, ctypes.c_longlong]
_uint_types = [ctypes.c_ubyte, ctypes.c_ushort, ctypes.c_uint, ctypes.c_ulong, ctypes.c_ulonglong]
_ints_by_size = dict({(ctypes.sizeof(T), T) for T in _int_types})
_uints_by_size = dict({(ctypes.sizeof(T), T) for T in _uint_types})

# primary GL types
GLenum = _uints_by_size[4]
GLboolean = _uints_by_size[1]
GLbitfield = _uints_by_size[4]
GLvoid = None
GLbyte = _ints_by_size[1]
GLshort = _ints_by_size[2]
GLint = _ints_by_size[4]
GLclampx = GLint
GLubyte = _uints_by_size[1]
GLushort = _uints_by_size[2]
GLuint = _uints_by_size[4]
GLsizei = GLint
GLfloat = ctypes.c_float
GLclampf = GLfloat
GLdouble = ctypes.c_double
GLclampd = GLdouble
GLeglImageOES = ctypes.c_void_p
GLchar = ctypes.c_char
GLcharARB = GLchar
GLhandleARB = GLuint
if platform.system() == 'Darwin':
	# ugh, mac os
	GLhandleARB = ctypes.c_void_p
# }
GLhalf = GLushort
GLhalfARB = GLhalf
GLfixed = GLint
GLintptr = _ints_by_size[ctypes.sizeof(ctypes.c_void_p)]
GLsizeiptr = GLintptr
GLint64 = _ints_by_size[8]
GLuint64 = _uints_by_size[8]
GLintptrARB = GLintptr
GLsizeiptrARB = GLsizeiptr
GLint64EXT = GLint64
GLuint64EXT = GLuint64

# this should work, right?
# struct __GLsync *
GLsync = ctypes.c_void_p

# i don't think these are easily supportable
# struct _cl_context
# struct _cl_event

# callbacks
GLDEBUGPROC = APIFUNCTYPE(GLvoid, GLenum, GLenum, GLuint, GLenum, GLsizei, ctypes.POINTER(GLchar), ctypes.c_void_p)
GLDEBUGPROCARB = GLDEBUGPROC
GLDEBUGPROCKHR = GLDEBUGPROC

# vendor extensions
GLDEBUGPROCAMD = APIFUNCTYPE(GLvoid, GLuint, GLenum, GLenum, GLsizei, ctypes.POINTER(GLchar), ctypes.c_void_p)
GLhalfNV = GLushort
GLvdpauSurfaceNV = GLintptr

def c_array(c_type, data):
	'Helper function for making and populating ctypes arrays'
	data = list(data)
	return (c_type * len(data))(*data)
# }

class GLError(Exception):
	def __init__(self, msg):
		Exception.__init__(self, msg)
	# }
# }

def _errcheck(result, func, args):
	gl = func.gl
	# calling glGetError() between glBegin() and glEnd() is forbidden
	if func.__name__ == 'glEnd':
		gl._begin = False
	if func.__name__ == 'glBegin':
		gl._begin = True
	if gl._begin:
		return result
	# }
	err = gl.glGetError()
	if err != 0:
		raise GLError('GL error {err} after call to function {func}; args: {args}'.format(err=err, func=func.__name__, args=repr(args)))
	# }
	return result
# }

print 'PyGLoo: initialized type system'

print 'PyGLoo: initializing enum definitions...'

# enum definitions
GL_CURRENT_BIT = 0x00000001
GL_POINT_BIT = 0x00000002
GL_LINE_BIT = 0x00000004
GL_POLYGON_BIT = 0x00000008
GL_POLYGON_STIPPLE_BIT = 0x00000010
GL_PIXEL_MODE_BIT = 0x00000020
GL_LIGHTING_BIT = 0x00000040
GL_FOG_BIT = 0x00000080
GL_DEPTH_BUFFER_BIT = 0x00000100
GL_ACCUM_BUFFER_BIT = 0x00000200
GL_STENCIL_BUFFER_BIT = 0x00000400
GL_VIEWPORT_BIT = 0x00000800
GL_TRANSFORM_BIT = 0x00001000
GL_ENABLE_BIT = 0x00002000
GL_COLOR_BUFFER_BIT = 0x00004000
GL_HINT_BIT = 0x00008000
GL_EVAL_BIT = 0x00010000
GL_LIST_BIT = 0x00020000
GL_TEXTURE_BIT = 0x00040000
GL_SCISSOR_BIT = 0x00080000
GL_MULTISAMPLE_BIT = 0x20000000
GL_MULTISAMPLE_BIT_ARB = 0x20000000
GL_MULTISAMPLE_BIT_EXT = 0x20000000
GL_MULTISAMPLE_BIT_3DFX = 0x20000000
GL_ALL_ATTRIB_BITS = 0xFFFFFFFF
GL_COVERAGE_BUFFER_BIT_NV = 0x00008000
GL_CLIENT_PIXEL_STORE_BIT = 0x00000001
GL_CLIENT_VERTEX_ARRAY_BIT = 0x00000002
GL_CLIENT_ALL_ATTRIB_BITS = 0xFFFFFFFF
GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT = 0x00000001
GL_CONTEXT_FLAG_DEBUG_BIT = 0x00000002
GL_CONTEXT_FLAG_DEBUG_BIT_KHR = 0x00000002
GL_CONTEXT_FLAG_ROBUST_ACCESS_BIT_ARB = 0x00000004
GL_CONTEXT_FLAG_ROBUST_ACCESS_BIT = 0x00000004
GL_CONTEXT_CORE_PROFILE_BIT = 0x00000001
GL_CONTEXT_COMPATIBILITY_PROFILE_BIT = 0x00000002
GL_MAP_READ_BIT = 0x0001
GL_MAP_READ_BIT_EXT = 0x0001
GL_MAP_WRITE_BIT = 0x0002
GL_MAP_WRITE_BIT_EXT = 0x0002
GL_MAP_INVALIDATE_RANGE_BIT = 0x0004
GL_MAP_INVALIDATE_RANGE_BIT_EXT = 0x0004
GL_MAP_INVALIDATE_BUFFER_BIT = 0x0008
GL_MAP_INVALIDATE_BUFFER_BIT_EXT = 0x0008
GL_MAP_FLUSH_EXPLICIT_BIT = 0x0010
GL_MAP_FLUSH_EXPLICIT_BIT_EXT = 0x0010
GL_MAP_UNSYNCHRONIZED_BIT = 0x0020
GL_MAP_UNSYNCHRONIZED_BIT_EXT = 0x0020
GL_MAP_PERSISTENT_BIT = 0x0040
GL_MAP_COHERENT_BIT = 0x0080
GL_DYNAMIC_STORAGE_BIT = 0x0100
GL_CLIENT_STORAGE_BIT = 0x0200
GL_SPARSE_STORAGE_BIT_ARB = 0x0400
GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT = 0x00000001
GL_VERTEX_ATTRIB_ARRAY_BARRIER_BIT_EXT = 0x00000001
GL_ELEMENT_ARRAY_BARRIER_BIT = 0x00000002
GL_ELEMENT_ARRAY_BARRIER_BIT_EXT = 0x00000002
GL_UNIFORM_BARRIER_BIT = 0x00000004
GL_UNIFORM_BARRIER_BIT_EXT = 0x00000004
GL_TEXTURE_FETCH_BARRIER_BIT = 0x00000008
GL_TEXTURE_FETCH_BARRIER_BIT_EXT = 0x00000008
GL_SHADER_GLOBAL_ACCESS_BARRIER_BIT_NV = 0x00000010
GL_SHADER_IMAGE_ACCESS_BARRIER_BIT = 0x00000020
GL_SHADER_IMAGE_ACCESS_BARRIER_BIT_EXT = 0x00000020
GL_COMMAND_BARRIER_BIT = 0x00000040
GL_COMMAND_BARRIER_BIT_EXT = 0x00000040
GL_PIXEL_BUFFER_BARRIER_BIT = 0x00000080
GL_PIXEL_BUFFER_BARRIER_BIT_EXT = 0x00000080
GL_TEXTURE_UPDATE_BARRIER_BIT = 0x00000100
GL_TEXTURE_UPDATE_BARRIER_BIT_EXT = 0x00000100
GL_BUFFER_UPDATE_BARRIER_BIT = 0x00000200
GL_BUFFER_UPDATE_BARRIER_BIT_EXT = 0x00000200
GL_FRAMEBUFFER_BARRIER_BIT = 0x00000400
GL_FRAMEBUFFER_BARRIER_BIT_EXT = 0x00000400
GL_TRANSFORM_FEEDBACK_BARRIER_BIT = 0x00000800
GL_TRANSFORM_FEEDBACK_BARRIER_BIT_EXT = 0x00000800
GL_ATOMIC_COUNTER_BARRIER_BIT = 0x00001000
GL_ATOMIC_COUNTER_BARRIER_BIT_EXT = 0x00001000
GL_SHADER_STORAGE_BARRIER_BIT = 0x00002000
GL_CLIENT_MAPPED_BUFFER_BARRIER_BIT = 0x00004000
GL_QUERY_BUFFER_BARRIER_BIT = 0x00008000
GL_ALL_BARRIER_BITS = 0xFFFFFFFF
GL_ALL_BARRIER_BITS_EXT = 0xFFFFFFFF
GL_QUERY_DEPTH_PASS_EVENT_BIT_AMD = 0x00000001
GL_QUERY_DEPTH_FAIL_EVENT_BIT_AMD = 0x00000002
GL_QUERY_STENCIL_FAIL_EVENT_BIT_AMD = 0x00000004
GL_QUERY_DEPTH_BOUNDS_FAIL_EVENT_BIT_AMD = 0x00000008
GL_QUERY_ALL_EVENT_BITS_AMD = 0xFFFFFFFF
GL_SYNC_FLUSH_COMMANDS_BIT = 0x00000001
GL_SYNC_FLUSH_COMMANDS_BIT_APPLE = 0x00000001
GL_VERTEX_SHADER_BIT = 0x00000001
GL_VERTEX_SHADER_BIT_EXT = 0x00000001
GL_FRAGMENT_SHADER_BIT = 0x00000002
GL_FRAGMENT_SHADER_BIT_EXT = 0x00000002
GL_GEOMETRY_SHADER_BIT = 0x00000004
GL_GEOMETRY_SHADER_BIT_EXT = 0x00000004
GL_TESS_CONTROL_SHADER_BIT = 0x00000008
GL_TESS_CONTROL_SHADER_BIT_EXT = 0x00000008
GL_TESS_EVALUATION_SHADER_BIT = 0x00000010
GL_TESS_EVALUATION_SHADER_BIT_EXT = 0x00000010
GL_COMPUTE_SHADER_BIT = 0x00000020
GL_ALL_SHADER_BITS = 0xFFFFFFFF
GL_ALL_SHADER_BITS_EXT = 0xFFFFFFFF
GL_TEXTURE_STORAGE_SPARSE_BIT_AMD = 0x00000001
GL_RED_BIT_ATI = 0x00000001
GL_GREEN_BIT_ATI = 0x00000002
GL_BLUE_BIT_ATI = 0x00000004
GL_2X_BIT_ATI = 0x00000001
GL_4X_BIT_ATI = 0x00000002
GL_8X_BIT_ATI = 0x00000004
GL_HALF_BIT_ATI = 0x00000008
GL_QUARTER_BIT_ATI = 0x00000010
GL_EIGHTH_BIT_ATI = 0x00000020
GL_SATURATE_BIT_ATI = 0x00000040
GL_COMP_BIT_ATI = 0x00000002
GL_NEGATE_BIT_ATI = 0x00000004
GL_BIAS_BIT_ATI = 0x00000008
GL_TRACE_OPERATIONS_BIT_MESA = 0x0001
GL_TRACE_PRIMITIVES_BIT_MESA = 0x0002
GL_TRACE_ARRAYS_BIT_MESA = 0x0004
GL_TRACE_TEXTURES_BIT_MESA = 0x0008
GL_TRACE_PIXELS_BIT_MESA = 0x0010
GL_TRACE_ERRORS_BIT_MESA = 0x0020
GL_TRACE_ALL_BITS_MESA = 0xFFFF
GL_BOLD_BIT_NV = 0x01
GL_ITALIC_BIT_NV = 0x02
GL_GLYPH_WIDTH_BIT_NV = 0x01
GL_GLYPH_HEIGHT_BIT_NV = 0x02
GL_GLYPH_HORIZONTAL_BEARING_X_BIT_NV = 0x04
GL_GLYPH_HORIZONTAL_BEARING_Y_BIT_NV = 0x08
GL_GLYPH_HORIZONTAL_BEARING_ADVANCE_BIT_NV = 0x10
GL_GLYPH_VERTICAL_BEARING_X_BIT_NV = 0x20
GL_GLYPH_VERTICAL_BEARING_Y_BIT_NV = 0x40
GL_GLYPH_VERTICAL_BEARING_ADVANCE_BIT_NV = 0x80
GL_GLYPH_HAS_KERNING_BIT_NV = 0x100
GL_FONT_X_MIN_BOUNDS_BIT_NV = 0x00010000
GL_FONT_Y_MIN_BOUNDS_BIT_NV = 0x00020000
GL_FONT_X_MAX_BOUNDS_BIT_NV = 0x00040000
GL_FONT_Y_MAX_BOUNDS_BIT_NV = 0x00080000
GL_FONT_UNITS_PER_EM_BIT_NV = 0x00100000
GL_FONT_ASCENDER_BIT_NV = 0x00200000
GL_FONT_DESCENDER_BIT_NV = 0x00400000
GL_FONT_HEIGHT_BIT_NV = 0x00800000
GL_FONT_MAX_ADVANCE_WIDTH_BIT_NV = 0x01000000
GL_FONT_MAX_ADVANCE_HEIGHT_BIT_NV = 0x02000000
GL_FONT_UNDERLINE_POSITION_BIT_NV = 0x04000000
GL_FONT_UNDERLINE_THICKNESS_BIT_NV = 0x08000000
GL_FONT_HAS_KERNING_BIT_NV = 0x10000000
GL_FONT_NUM_GLYPH_INDICES_BIT_NV = 0x20000000
GL_PERFQUERY_SINGLE_CONTEXT_INTEL = 0x00000000
GL_PERFQUERY_GLOBAL_CONTEXT_INTEL = 0x00000001
GL_VERTEX23_BIT_PGI = 0x00000004
GL_VERTEX4_BIT_PGI = 0x00000008
GL_COLOR3_BIT_PGI = 0x00010000
GL_COLOR4_BIT_PGI = 0x00020000
GL_EDGEFLAG_BIT_PGI = 0x00040000
GL_INDEX_BIT_PGI = 0x00080000
GL_MAT_AMBIENT_BIT_PGI = 0x00100000
GL_MAT_AMBIENT_AND_DIFFUSE_BIT_PGI = 0x00200000
GL_MAT_DIFFUSE_BIT_PGI = 0x00400000
GL_MAT_EMISSION_BIT_PGI = 0x00800000
GL_MAT_COLOR_INDEXES_BIT_PGI = 0x01000000
GL_MAT_SHININESS_BIT_PGI = 0x02000000
GL_MAT_SPECULAR_BIT_PGI = 0x04000000
GL_NORMAL_BIT_PGI = 0x08000000
GL_TEXCOORD1_BIT_PGI = 0x10000000
GL_TEXCOORD2_BIT_PGI = 0x20000000
GL_TEXCOORD3_BIT_PGI = 0x40000000
GL_TEXCOORD4_BIT_PGI = 0x80000000
GL_COLOR_BUFFER_BIT0_QCOM = 0x00000001
GL_COLOR_BUFFER_BIT1_QCOM = 0x00000002
GL_COLOR_BUFFER_BIT2_QCOM = 0x00000004
GL_COLOR_BUFFER_BIT3_QCOM = 0x00000008
GL_COLOR_BUFFER_BIT4_QCOM = 0x00000010
GL_COLOR_BUFFER_BIT5_QCOM = 0x00000020
GL_COLOR_BUFFER_BIT6_QCOM = 0x00000040
GL_COLOR_BUFFER_BIT7_QCOM = 0x00000080
GL_DEPTH_BUFFER_BIT0_QCOM = 0x00000100
GL_DEPTH_BUFFER_BIT1_QCOM = 0x00000200
GL_DEPTH_BUFFER_BIT2_QCOM = 0x00000400
GL_DEPTH_BUFFER_BIT3_QCOM = 0x00000800
GL_DEPTH_BUFFER_BIT4_QCOM = 0x00001000
GL_DEPTH_BUFFER_BIT5_QCOM = 0x00002000
GL_DEPTH_BUFFER_BIT6_QCOM = 0x00004000
GL_DEPTH_BUFFER_BIT7_QCOM = 0x00008000
GL_STENCIL_BUFFER_BIT0_QCOM = 0x00010000
GL_STENCIL_BUFFER_BIT1_QCOM = 0x00020000
GL_STENCIL_BUFFER_BIT2_QCOM = 0x00040000
GL_STENCIL_BUFFER_BIT3_QCOM = 0x00080000
GL_STENCIL_BUFFER_BIT4_QCOM = 0x00100000
GL_STENCIL_BUFFER_BIT5_QCOM = 0x00200000
GL_STENCIL_BUFFER_BIT6_QCOM = 0x00400000
GL_STENCIL_BUFFER_BIT7_QCOM = 0x00800000
GL_MULTISAMPLE_BUFFER_BIT0_QCOM = 0x01000000
GL_MULTISAMPLE_BUFFER_BIT1_QCOM = 0x02000000
GL_MULTISAMPLE_BUFFER_BIT2_QCOM = 0x04000000
GL_MULTISAMPLE_BUFFER_BIT3_QCOM = 0x08000000
GL_MULTISAMPLE_BUFFER_BIT4_QCOM = 0x10000000
GL_MULTISAMPLE_BUFFER_BIT5_QCOM = 0x20000000
GL_MULTISAMPLE_BUFFER_BIT6_QCOM = 0x40000000
GL_MULTISAMPLE_BUFFER_BIT7_QCOM = 0x80000000
GL_TEXTURE_DEFORMATION_BIT_SGIX = 0x00000001
GL_GEOMETRY_DEFORMATION_BIT_SGIX = 0x00000002
GL_RESTART_SUN = 0x0001
GL_REPLACE_MIDDLE_SUN = 0x0002
GL_REPLACE_OLDEST_SUN = 0x0003
GL_LAYOUT_DEFAULT_INTEL = 0
GL_LAYOUT_LINEAR_INTEL = 1
GL_LAYOUT_LINEAR_CPU_CACHED_INTEL = 2
GL_NEXT_BUFFER_NV = -2
GL_SKIP_COMPONENTS4_NV = -3
GL_SKIP_COMPONENTS3_NV = -4
GL_SKIP_COMPONENTS2_NV = -5
GL_SKIP_COMPONENTS1_NV = -6
GL_CLOSE_PATH_NV = 0x00
GL_MOVE_TO_NV = 0x02
GL_RELATIVE_MOVE_TO_NV = 0x03
GL_LINE_TO_NV = 0x04
GL_RELATIVE_LINE_TO_NV = 0x05
GL_HORIZONTAL_LINE_TO_NV = 0x06
GL_RELATIVE_HORIZONTAL_LINE_TO_NV = 0x07
GL_VERTICAL_LINE_TO_NV = 0x08
GL_RELATIVE_VERTICAL_LINE_TO_NV = 0x09
GL_QUADRATIC_CURVE_TO_NV = 0x0A
GL_RELATIVE_QUADRATIC_CURVE_TO_NV = 0x0B
GL_CUBIC_CURVE_TO_NV = 0x0C
GL_RELATIVE_CUBIC_CURVE_TO_NV = 0x0D
GL_SMOOTH_QUADRATIC_CURVE_TO_NV = 0x0E
GL_RELATIVE_SMOOTH_QUADRATIC_CURVE_TO_NV = 0x0F
GL_SMOOTH_CUBIC_CURVE_TO_NV = 0x10
GL_RELATIVE_SMOOTH_CUBIC_CURVE_TO_NV = 0x11
GL_SMALL_CCW_ARC_TO_NV = 0x12
GL_RELATIVE_SMALL_CCW_ARC_TO_NV = 0x13
GL_SMALL_CW_ARC_TO_NV = 0x14
GL_RELATIVE_SMALL_CW_ARC_TO_NV = 0x15
GL_LARGE_CCW_ARC_TO_NV = 0x16
GL_RELATIVE_LARGE_CCW_ARC_TO_NV = 0x17
GL_LARGE_CW_ARC_TO_NV = 0x18
GL_RELATIVE_LARGE_CW_ARC_TO_NV = 0x19
GL_CONIC_CURVE_TO_NV = 0x1A
GL_RELATIVE_CONIC_CURVE_TO_NV = 0x1B
GL_SHARED_EDGE_NV = 0xC0
GL_ROUNDED_RECT_NV = 0xE8
GL_RELATIVE_ROUNDED_RECT_NV = 0xE9
GL_ROUNDED_RECT2_NV = 0xEA
GL_RELATIVE_ROUNDED_RECT2_NV = 0xEB
GL_ROUNDED_RECT4_NV = 0xEC
GL_RELATIVE_ROUNDED_RECT4_NV = 0xED
GL_ROUNDED_RECT8_NV = 0xEE
GL_RELATIVE_ROUNDED_RECT8_NV = 0xEF
GL_RESTART_PATH_NV = 0xF0
GL_DUP_FIRST_CUBIC_CURVE_TO_NV = 0xF2
GL_DUP_LAST_CUBIC_CURVE_TO_NV = 0xF4
GL_RECT_NV = 0xF6
GL_RELATIVE_RECT_NV = 0xF7
GL_CIRCULAR_CCW_ARC_TO_NV = 0xF8
GL_CIRCULAR_CW_ARC_TO_NV = 0xFA
GL_CIRCULAR_TANGENT_ARC_TO_NV = 0xFC
GL_ARC_TO_NV = 0xFE
GL_RELATIVE_ARC_TO_NV = 0xFF
GL_FALSE = 0
GL_NO_ERROR = 0
GL_ZERO = 0
GL_NONE = 0
GL_NONE_OES = 0
GL_TRUE = 1
GL_ONE = 1
GL_INVALID_INDEX = 0xFFFFFFFF
GL_TIMEOUT_IGNORED = 0xFFFFFFFFFFFFFFFF
GL_TIMEOUT_IGNORED_APPLE = 0xFFFFFFFFFFFFFFFF
GL_VERSION_ES_CL_1_0 = 1
GL_VERSION_ES_CM_1_1 = 1
GL_VERSION_ES_CL_1_1 = 1
GL_POINTS = 0x0000
GL_LINES = 0x0001
GL_LINE_LOOP = 0x0002
GL_LINE_STRIP = 0x0003
GL_TRIANGLES = 0x0004
GL_TRIANGLE_STRIP = 0x0005
GL_TRIANGLE_FAN = 0x0006
GL_QUADS = 0x0007
GL_QUADS_EXT = 0x0007
GL_QUAD_STRIP = 0x0008
GL_POLYGON = 0x0009
GL_LINES_ADJACENCY = 0x000A
GL_LINES_ADJACENCY_ARB = 0x000A
GL_LINES_ADJACENCY_EXT = 0x000A
GL_LINE_STRIP_ADJACENCY = 0x000B
GL_LINE_STRIP_ADJACENCY_ARB = 0x000B
GL_LINE_STRIP_ADJACENCY_EXT = 0x000B
GL_TRIANGLES_ADJACENCY = 0x000C
GL_TRIANGLES_ADJACENCY_ARB = 0x000C
GL_TRIANGLES_ADJACENCY_EXT = 0x000C
GL_TRIANGLE_STRIP_ADJACENCY = 0x000D
GL_TRIANGLE_STRIP_ADJACENCY_ARB = 0x000D
GL_TRIANGLE_STRIP_ADJACENCY_EXT = 0x000D
GL_PATCHES = 0x000E
GL_PATCHES_EXT = 0x000E
GL_ACCUM = 0x0100
GL_LOAD = 0x0101
GL_RETURN = 0x0102
GL_MULT = 0x0103
GL_ADD = 0x0104
GL_NEVER = 0x0200
GL_LESS = 0x0201
GL_EQUAL = 0x0202
GL_LEQUAL = 0x0203
GL_GREATER = 0x0204
GL_NOTEQUAL = 0x0205
GL_GEQUAL = 0x0206
GL_ALWAYS = 0x0207
GL_SRC_COLOR = 0x0300
GL_ONE_MINUS_SRC_COLOR = 0x0301
GL_SRC_ALPHA = 0x0302
GL_ONE_MINUS_SRC_ALPHA = 0x0303
GL_DST_ALPHA = 0x0304
GL_ONE_MINUS_DST_ALPHA = 0x0305
GL_DST_COLOR = 0x0306
GL_ONE_MINUS_DST_COLOR = 0x0307
GL_SRC_ALPHA_SATURATE = 0x0308
GL_FRONT_LEFT = 0x0400
GL_FRONT_RIGHT = 0x0401
GL_BACK_LEFT = 0x0402
GL_BACK_RIGHT = 0x0403
GL_FRONT = 0x0404
GL_BACK = 0x0405
GL_LEFT = 0x0406
GL_RIGHT = 0x0407
GL_FRONT_AND_BACK = 0x0408
GL_AUX0 = 0x0409
GL_AUX1 = 0x040A
GL_AUX2 = 0x040B
GL_AUX3 = 0x040C
GL_INVALID_ENUM = 0x0500
GL_INVALID_VALUE = 0x0501
GL_INVALID_OPERATION = 0x0502
GL_STACK_OVERFLOW = 0x0503
GL_STACK_OVERFLOW_KHR = 0x0503
GL_STACK_UNDERFLOW = 0x0504
GL_STACK_UNDERFLOW_KHR = 0x0504
GL_OUT_OF_MEMORY = 0x0505
GL_INVALID_FRAMEBUFFER_OPERATION = 0x0506
GL_INVALID_FRAMEBUFFER_OPERATION_EXT = 0x0506
GL_INVALID_FRAMEBUFFER_OPERATION_OES = 0x0506
GL_CONTEXT_LOST = 0x0507
GL_CONTEXT_LOST_KHR = 0x0507
GL_2D = 0x0600
GL_3D = 0x0601
GL_3D_COLOR = 0x0602
GL_3D_COLOR_TEXTURE = 0x0603
GL_4D_COLOR_TEXTURE = 0x0604
GL_PASS_THROUGH_TOKEN = 0x0700
GL_POINT_TOKEN = 0x0701
GL_LINE_TOKEN = 0x0702
GL_POLYGON_TOKEN = 0x0703
GL_BITMAP_TOKEN = 0x0704
GL_DRAW_PIXEL_TOKEN = 0x0705
GL_COPY_PIXEL_TOKEN = 0x0706
GL_LINE_RESET_TOKEN = 0x0707
GL_EXP = 0x0800
GL_EXP2 = 0x0801
GL_CW = 0x0900
GL_CCW = 0x0901
GL_COEFF = 0x0A00
GL_ORDER = 0x0A01
GL_DOMAIN = 0x0A02
GL_CURRENT_COLOR = 0x0B00
GL_CURRENT_INDEX = 0x0B01
GL_CURRENT_NORMAL = 0x0B02
GL_CURRENT_TEXTURE_COORDS = 0x0B03
GL_CURRENT_RASTER_COLOR = 0x0B04
GL_CURRENT_RASTER_INDEX = 0x0B05
GL_CURRENT_RASTER_TEXTURE_COORDS = 0x0B06
GL_CURRENT_RASTER_POSITION = 0x0B07
GL_CURRENT_RASTER_POSITION_VALID = 0x0B08
GL_CURRENT_RASTER_DISTANCE = 0x0B09
GL_POINT_SMOOTH = 0x0B10
GL_POINT_SIZE = 0x0B11
GL_POINT_SIZE_RANGE = 0x0B12
GL_SMOOTH_POINT_SIZE_RANGE = 0x0B12
GL_POINT_SIZE_GRANULARITY = 0x0B13
GL_SMOOTH_POINT_SIZE_GRANULARITY = 0x0B13
GL_LINE_SMOOTH = 0x0B20
GL_LINE_WIDTH = 0x0B21
GL_LINE_WIDTH_RANGE = 0x0B22
GL_SMOOTH_LINE_WIDTH_RANGE = 0x0B22
GL_LINE_WIDTH_GRANULARITY = 0x0B23
GL_SMOOTH_LINE_WIDTH_GRANULARITY = 0x0B23
GL_LINE_STIPPLE = 0x0B24
GL_LINE_STIPPLE_PATTERN = 0x0B25
GL_LINE_STIPPLE_REPEAT = 0x0B26
GL_LIST_MODE = 0x0B30
GL_MAX_LIST_NESTING = 0x0B31
GL_LIST_BASE = 0x0B32
GL_LIST_INDEX = 0x0B33
GL_POLYGON_MODE = 0x0B40
GL_POLYGON_SMOOTH = 0x0B41
GL_POLYGON_STIPPLE = 0x0B42
GL_EDGE_FLAG = 0x0B43
GL_CULL_FACE = 0x0B44
GL_CULL_FACE_MODE = 0x0B45
GL_FRONT_FACE = 0x0B46
GL_LIGHTING = 0x0B50
GL_LIGHT_MODEL_LOCAL_VIEWER = 0x0B51
GL_LIGHT_MODEL_TWO_SIDE = 0x0B52
GL_LIGHT_MODEL_AMBIENT = 0x0B53
GL_SHADE_MODEL = 0x0B54
GL_COLOR_MATERIAL_FACE = 0x0B55
GL_COLOR_MATERIAL_PARAMETER = 0x0B56
GL_COLOR_MATERIAL = 0x0B57
GL_FOG = 0x0B60
GL_FOG_INDEX = 0x0B61
GL_FOG_DENSITY = 0x0B62
GL_FOG_START = 0x0B63
GL_FOG_END = 0x0B64
GL_FOG_MODE = 0x0B65
GL_FOG_COLOR = 0x0B66
GL_DEPTH_RANGE = 0x0B70
GL_DEPTH_TEST = 0x0B71
GL_DEPTH_WRITEMASK = 0x0B72
GL_DEPTH_CLEAR_VALUE = 0x0B73
GL_DEPTH_FUNC = 0x0B74
GL_ACCUM_CLEAR_VALUE = 0x0B80
GL_STENCIL_TEST = 0x0B90
GL_STENCIL_CLEAR_VALUE = 0x0B91
GL_STENCIL_FUNC = 0x0B92
GL_STENCIL_VALUE_MASK = 0x0B93
GL_STENCIL_FAIL = 0x0B94
GL_STENCIL_PASS_DEPTH_FAIL = 0x0B95
GL_STENCIL_PASS_DEPTH_PASS = 0x0B96
GL_STENCIL_REF = 0x0B97
GL_STENCIL_WRITEMASK = 0x0B98
GL_MATRIX_MODE = 0x0BA0
GL_NORMALIZE = 0x0BA1
GL_VIEWPORT = 0x0BA2
GL_MODELVIEW_STACK_DEPTH = 0x0BA3
GL_MODELVIEW0_STACK_DEPTH_EXT = 0x0BA3
GL_PATH_MODELVIEW_STACK_DEPTH_NV = 0x0BA3
GL_PROJECTION_STACK_DEPTH = 0x0BA4
GL_PATH_PROJECTION_STACK_DEPTH_NV = 0x0BA4
GL_TEXTURE_STACK_DEPTH = 0x0BA5
GL_MODELVIEW_MATRIX = 0x0BA6
GL_MODELVIEW0_MATRIX_EXT = 0x0BA6
GL_PATH_MODELVIEW_MATRIX_NV = 0x0BA6
GL_PROJECTION_MATRIX = 0x0BA7
GL_PATH_PROJECTION_MATRIX_NV = 0x0BA7
GL_TEXTURE_MATRIX = 0x0BA8
GL_ATTRIB_STACK_DEPTH = 0x0BB0
GL_CLIENT_ATTRIB_STACK_DEPTH = 0x0BB1
GL_ALPHA_TEST = 0x0BC0
GL_ALPHA_TEST_QCOM = 0x0BC0
GL_ALPHA_TEST_FUNC = 0x0BC1
GL_ALPHA_TEST_FUNC_QCOM = 0x0BC1
GL_ALPHA_TEST_REF = 0x0BC2
GL_ALPHA_TEST_REF_QCOM = 0x0BC2
GL_DITHER = 0x0BD0
GL_BLEND_DST = 0x0BE0
GL_BLEND_SRC = 0x0BE1
GL_BLEND = 0x0BE2
GL_LOGIC_OP_MODE = 0x0BF0
GL_INDEX_LOGIC_OP = 0x0BF1
GL_LOGIC_OP = 0x0BF1
GL_COLOR_LOGIC_OP = 0x0BF2
GL_AUX_BUFFERS = 0x0C00
GL_DRAW_BUFFER = 0x0C01
GL_DRAW_BUFFER_EXT = 0x0C01
GL_READ_BUFFER = 0x0C02
GL_READ_BUFFER_EXT = 0x0C02
GL_READ_BUFFER_NV = 0x0C02
GL_SCISSOR_BOX = 0x0C10
GL_SCISSOR_TEST = 0x0C11
GL_INDEX_CLEAR_VALUE = 0x0C20
GL_INDEX_WRITEMASK = 0x0C21
GL_COLOR_CLEAR_VALUE = 0x0C22
GL_COLOR_WRITEMASK = 0x0C23
GL_INDEX_MODE = 0x0C30
GL_RGBA_MODE = 0x0C31
GL_DOUBLEBUFFER = 0x0C32
GL_STEREO = 0x0C33
GL_RENDER_MODE = 0x0C40
GL_PERSPECTIVE_CORRECTION_HINT = 0x0C50
GL_POINT_SMOOTH_HINT = 0x0C51
GL_LINE_SMOOTH_HINT = 0x0C52
GL_POLYGON_SMOOTH_HINT = 0x0C53
GL_FOG_HINT = 0x0C54
GL_TEXTURE_GEN_S = 0x0C60
GL_TEXTURE_GEN_T = 0x0C61
GL_TEXTURE_GEN_R = 0x0C62
GL_TEXTURE_GEN_Q = 0x0C63
GL_PIXEL_MAP_I_TO_I = 0x0C70
GL_PIXEL_MAP_S_TO_S = 0x0C71
GL_PIXEL_MAP_I_TO_R = 0x0C72
GL_PIXEL_MAP_I_TO_G = 0x0C73
GL_PIXEL_MAP_I_TO_B = 0x0C74
GL_PIXEL_MAP_I_TO_A = 0x0C75
GL_PIXEL_MAP_R_TO_R = 0x0C76
GL_PIXEL_MAP_G_TO_G = 0x0C77
GL_PIXEL_MAP_B_TO_B = 0x0C78
GL_PIXEL_MAP_A_TO_A = 0x0C79
GL_PIXEL_MAP_I_TO_I_SIZE = 0x0CB0
GL_PIXEL_MAP_S_TO_S_SIZE = 0x0CB1
GL_PIXEL_MAP_I_TO_R_SIZE = 0x0CB2
GL_PIXEL_MAP_I_TO_G_SIZE = 0x0CB3
GL_PIXEL_MAP_I_TO_B_SIZE = 0x0CB4
GL_PIXEL_MAP_I_TO_A_SIZE = 0x0CB5
GL_PIXEL_MAP_R_TO_R_SIZE = 0x0CB6
GL_PIXEL_MAP_G_TO_G_SIZE = 0x0CB7
GL_PIXEL_MAP_B_TO_B_SIZE = 0x0CB8
GL_PIXEL_MAP_A_TO_A_SIZE = 0x0CB9
GL_UNPACK_SWAP_BYTES = 0x0CF0
GL_UNPACK_LSB_FIRST = 0x0CF1
GL_UNPACK_ROW_LENGTH = 0x0CF2
GL_UNPACK_ROW_LENGTH_EXT = 0x0CF2
GL_UNPACK_SKIP_ROWS = 0x0CF3
GL_UNPACK_SKIP_ROWS_EXT = 0x0CF3
GL_UNPACK_SKIP_PIXELS = 0x0CF4
GL_UNPACK_SKIP_PIXELS_EXT = 0x0CF4
GL_UNPACK_ALIGNMENT = 0x0CF5
GL_PACK_SWAP_BYTES = 0x0D00
GL_PACK_LSB_FIRST = 0x0D01
GL_PACK_ROW_LENGTH = 0x0D02
GL_PACK_SKIP_ROWS = 0x0D03
GL_PACK_SKIP_PIXELS = 0x0D04
GL_PACK_ALIGNMENT = 0x0D05
GL_MAP_COLOR = 0x0D10
GL_MAP_STENCIL = 0x0D11
GL_INDEX_SHIFT = 0x0D12
GL_INDEX_OFFSET = 0x0D13
GL_RED_SCALE = 0x0D14
GL_RED_BIAS = 0x0D15
GL_ZOOM_X = 0x0D16
GL_ZOOM_Y = 0x0D17
GL_GREEN_SCALE = 0x0D18
GL_GREEN_BIAS = 0x0D19
GL_BLUE_SCALE = 0x0D1A
GL_BLUE_BIAS = 0x0D1B
GL_ALPHA_SCALE = 0x0D1C
GL_ALPHA_BIAS = 0x0D1D
GL_DEPTH_SCALE = 0x0D1E
GL_DEPTH_BIAS = 0x0D1F
GL_MAX_EVAL_ORDER = 0x0D30
GL_MAX_LIGHTS = 0x0D31
GL_MAX_CLIP_PLANES = 0x0D32
GL_MAX_CLIP_PLANES_IMG = 0x0D32
GL_MAX_CLIP_DISTANCES = 0x0D32
GL_MAX_CLIP_DISTANCES_APPLE = 0x0D32
GL_MAX_TEXTURE_SIZE = 0x0D33
GL_MAX_PIXEL_MAP_TABLE = 0x0D34
GL_MAX_ATTRIB_STACK_DEPTH = 0x0D35
GL_MAX_MODELVIEW_STACK_DEPTH = 0x0D36
GL_PATH_MAX_MODELVIEW_STACK_DEPTH_NV = 0x0D36
GL_MAX_NAME_STACK_DEPTH = 0x0D37
GL_MAX_PROJECTION_STACK_DEPTH = 0x0D38
GL_PATH_MAX_PROJECTION_STACK_DEPTH_NV = 0x0D38
GL_MAX_TEXTURE_STACK_DEPTH = 0x0D39
GL_MAX_VIEWPORT_DIMS = 0x0D3A
GL_MAX_CLIENT_ATTRIB_STACK_DEPTH = 0x0D3B
GL_SUBPIXEL_BITS = 0x0D50
GL_INDEX_BITS = 0x0D51
GL_RED_BITS = 0x0D52
GL_GREEN_BITS = 0x0D53
GL_BLUE_BITS = 0x0D54
GL_ALPHA_BITS = 0x0D55
GL_DEPTH_BITS = 0x0D56
GL_STENCIL_BITS = 0x0D57
GL_ACCUM_RED_BITS = 0x0D58
GL_ACCUM_GREEN_BITS = 0x0D59
GL_ACCUM_BLUE_BITS = 0x0D5A
GL_ACCUM_ALPHA_BITS = 0x0D5B
GL_NAME_STACK_DEPTH = 0x0D70
GL_AUTO_NORMAL = 0x0D80
GL_MAP1_COLOR_4 = 0x0D90
GL_MAP1_INDEX = 0x0D91
GL_MAP1_NORMAL = 0x0D92
GL_MAP1_TEXTURE_COORD_1 = 0x0D93
GL_MAP1_TEXTURE_COORD_2 = 0x0D94
GL_MAP1_TEXTURE_COORD_3 = 0x0D95
GL_MAP1_TEXTURE_COORD_4 = 0x0D96
GL_MAP1_VERTEX_3 = 0x0D97
GL_MAP1_VERTEX_4 = 0x0D98
GL_MAP2_COLOR_4 = 0x0DB0
GL_MAP2_INDEX = 0x0DB1
GL_MAP2_NORMAL = 0x0DB2
GL_MAP2_TEXTURE_COORD_1 = 0x0DB3
GL_MAP2_TEXTURE_COORD_2 = 0x0DB4
GL_MAP2_TEXTURE_COORD_3 = 0x0DB5
GL_MAP2_TEXTURE_COORD_4 = 0x0DB6
GL_MAP2_VERTEX_3 = 0x0DB7
GL_MAP2_VERTEX_4 = 0x0DB8
GL_MAP1_GRID_DOMAIN = 0x0DD0
GL_MAP1_GRID_SEGMENTS = 0x0DD1
GL_MAP2_GRID_DOMAIN = 0x0DD2
GL_MAP2_GRID_SEGMENTS = 0x0DD3
GL_TEXTURE_1D = 0x0DE0
GL_TEXTURE_2D = 0x0DE1
GL_FEEDBACK_BUFFER_POINTER = 0x0DF0
GL_FEEDBACK_BUFFER_SIZE = 0x0DF1
GL_FEEDBACK_BUFFER_TYPE = 0x0DF2
GL_SELECTION_BUFFER_POINTER = 0x0DF3
GL_SELECTION_BUFFER_SIZE = 0x0DF4
GL_TEXTURE_WIDTH = 0x1000
GL_TEXTURE_HEIGHT = 0x1001
GL_TEXTURE_INTERNAL_FORMAT = 0x1003
GL_TEXTURE_COMPONENTS = 0x1003
GL_TEXTURE_BORDER_COLOR = 0x1004
GL_TEXTURE_BORDER_COLOR_EXT = 0x1004
GL_TEXTURE_BORDER_COLOR_NV = 0x1004
GL_TEXTURE_BORDER = 0x1005
GL_TEXTURE_TARGET = 0x1006
GL_DONT_CARE = 0x1100
GL_FASTEST = 0x1101
GL_NICEST = 0x1102
GL_AMBIENT = 0x1200
GL_DIFFUSE = 0x1201
GL_SPECULAR = 0x1202
GL_POSITION = 0x1203
GL_SPOT_DIRECTION = 0x1204
GL_SPOT_EXPONENT = 0x1205
GL_SPOT_CUTOFF = 0x1206
GL_CONSTANT_ATTENUATION = 0x1207
GL_LINEAR_ATTENUATION = 0x1208
GL_QUADRATIC_ATTENUATION = 0x1209
GL_COMPILE = 0x1300
GL_COMPILE_AND_EXECUTE = 0x1301
GL_BYTE = 0x1400
GL_UNSIGNED_BYTE = 0x1401
GL_SHORT = 0x1402
GL_UNSIGNED_SHORT = 0x1403
GL_INT = 0x1404
GL_UNSIGNED_INT = 0x1405
GL_FLOAT = 0x1406
GL_2_BYTES = 0x1407
GL_2_BYTES_NV = 0x1407
GL_3_BYTES = 0x1408
GL_3_BYTES_NV = 0x1408
GL_4_BYTES = 0x1409
GL_4_BYTES_NV = 0x1409
GL_DOUBLE = 0x140A
GL_DOUBLE_EXT = 0x140A
GL_HALF_FLOAT = 0x140B
GL_HALF_FLOAT_ARB = 0x140B
GL_HALF_FLOAT_NV = 0x140B
GL_HALF_APPLE = 0x140B
GL_FIXED = 0x140C
GL_FIXED_OES = 0x140C
GL_INT64_NV = 0x140E
GL_UNSIGNED_INT64_ARB = 0x140F
GL_UNSIGNED_INT64_NV = 0x140F
GL_CLEAR = 0x1500
GL_AND = 0x1501
GL_AND_REVERSE = 0x1502
GL_COPY = 0x1503
GL_AND_INVERTED = 0x1504
GL_NOOP = 0x1505
GL_XOR = 0x1506
GL_XOR_NV = 0x1506
GL_OR = 0x1507
GL_NOR = 0x1508
GL_EQUIV = 0x1509
GL_INVERT = 0x150A
GL_OR_REVERSE = 0x150B
GL_COPY_INVERTED = 0x150C
GL_OR_INVERTED = 0x150D
GL_NAND = 0x150E
GL_SET = 0x150F
GL_EMISSION = 0x1600
GL_SHININESS = 0x1601
GL_AMBIENT_AND_DIFFUSE = 0x1602
GL_COLOR_INDEXES = 0x1603
GL_MODELVIEW = 0x1700
GL_MODELVIEW0_ARB = 0x1700
GL_MODELVIEW0_EXT = 0x1700
GL_PATH_MODELVIEW_NV = 0x1700
GL_PROJECTION = 0x1701
GL_PATH_PROJECTION_NV = 0x1701
GL_TEXTURE = 0x1702
GL_COLOR = 0x1800
GL_COLOR_EXT = 0x1800
GL_DEPTH = 0x1801
GL_DEPTH_EXT = 0x1801
GL_STENCIL = 0x1802
GL_STENCIL_EXT = 0x1802
GL_COLOR_INDEX = 0x1900
GL_STENCIL_INDEX = 0x1901
GL_STENCIL_INDEX_OES = 0x1901
GL_DEPTH_COMPONENT = 0x1902
GL_RED = 0x1903
GL_RED_EXT = 0x1903
GL_RED_NV = 0x1903
GL_GREEN = 0x1904
GL_GREEN_NV = 0x1904
GL_BLUE = 0x1905
GL_BLUE_NV = 0x1905
GL_ALPHA = 0x1906
GL_RGB = 0x1907
GL_RGBA = 0x1908
GL_LUMINANCE = 0x1909
GL_LUMINANCE_ALPHA = 0x190A
GL_BITMAP = 0x1A00
GL_POINT = 0x1B00
GL_LINE = 0x1B01
GL_FILL = 0x1B02
GL_RENDER = 0x1C00
GL_FEEDBACK = 0x1C01
GL_SELECT = 0x1C02
GL_FLAT = 0x1D00
GL_SMOOTH = 0x1D01
GL_KEEP = 0x1E00
GL_REPLACE = 0x1E01
GL_INCR = 0x1E02
GL_DECR = 0x1E03
GL_VENDOR = 0x1F00
GL_RENDERER = 0x1F01
GL_VERSION = 0x1F02
GL_EXTENSIONS = 0x1F03
GL_S = 0x2000
GL_T = 0x2001
GL_R = 0x2002
GL_Q = 0x2003
GL_MODULATE = 0x2100
GL_DECAL = 0x2101
GL_TEXTURE_ENV_MODE = 0x2200
GL_TEXTURE_ENV_COLOR = 0x2201
GL_TEXTURE_ENV = 0x2300
GL_EYE_LINEAR = 0x2400
GL_EYE_LINEAR_NV = 0x2400
GL_OBJECT_LINEAR = 0x2401
GL_OBJECT_LINEAR_NV = 0x2401
GL_SPHERE_MAP = 0x2402
GL_TEXTURE_GEN_MODE = 0x2500
GL_TEXTURE_GEN_MODE_OES = 0x2500
GL_OBJECT_PLANE = 0x2501
GL_EYE_PLANE = 0x2502
GL_NEAREST = 0x2600
GL_LINEAR = 0x2601
GL_NEAREST_MIPMAP_NEAREST = 0x2700
GL_LINEAR_MIPMAP_NEAREST = 0x2701
GL_NEAREST_MIPMAP_LINEAR = 0x2702
GL_LINEAR_MIPMAP_LINEAR = 0x2703
GL_TEXTURE_MAG_FILTER = 0x2800
GL_TEXTURE_MIN_FILTER = 0x2801
GL_TEXTURE_WRAP_S = 0x2802
GL_TEXTURE_WRAP_T = 0x2803
GL_CLAMP = 0x2900
GL_REPEAT = 0x2901
GL_POLYGON_OFFSET_UNITS = 0x2A00
GL_POLYGON_OFFSET_POINT = 0x2A01
GL_POLYGON_OFFSET_LINE = 0x2A02
GL_R3_G3_B2 = 0x2A10
GL_V2F = 0x2A20
GL_V3F = 0x2A21
GL_C4UB_V2F = 0x2A22
GL_C4UB_V3F = 0x2A23
GL_C3F_V3F = 0x2A24
GL_N3F_V3F = 0x2A25
GL_C4F_N3F_V3F = 0x2A26
GL_T2F_V3F = 0x2A27
GL_T4F_V4F = 0x2A28
GL_T2F_C4UB_V3F = 0x2A29
GL_T2F_C3F_V3F = 0x2A2A
GL_T2F_N3F_V3F = 0x2A2B
GL_T2F_C4F_N3F_V3F = 0x2A2C
GL_T4F_C4F_N3F_V4F = 0x2A2D
GL_CLIP_PLANE0 = 0x3000
GL_CLIP_PLANE0_IMG = 0x3000
GL_CLIP_DISTANCE0 = 0x3000
GL_CLIP_DISTANCE0_APPLE = 0x3000
GL_CLIP_PLANE1 = 0x3001
GL_CLIP_PLANE1_IMG = 0x3001
GL_CLIP_DISTANCE1 = 0x3001
GL_CLIP_DISTANCE1_APPLE = 0x3001
GL_CLIP_PLANE2 = 0x3002
GL_CLIP_PLANE2_IMG = 0x3002
GL_CLIP_DISTANCE2 = 0x3002
GL_CLIP_DISTANCE2_APPLE = 0x3002
GL_CLIP_PLANE3 = 0x3003
GL_CLIP_PLANE3_IMG = 0x3003
GL_CLIP_DISTANCE3 = 0x3003
GL_CLIP_DISTANCE3_APPLE = 0x3003
GL_CLIP_PLANE4 = 0x3004
GL_CLIP_PLANE4_IMG = 0x3004
GL_CLIP_DISTANCE4 = 0x3004
GL_CLIP_DISTANCE4_APPLE = 0x3004
GL_CLIP_PLANE5 = 0x3005
GL_CLIP_PLANE5_IMG = 0x3005
GL_CLIP_DISTANCE5 = 0x3005
GL_CLIP_DISTANCE5_APPLE = 0x3005
GL_CLIP_DISTANCE6 = 0x3006
GL_CLIP_DISTANCE6_APPLE = 0x3006
GL_CLIP_DISTANCE7 = 0x3007
GL_CLIP_DISTANCE7_APPLE = 0x3007
GL_LIGHT0 = 0x4000
GL_LIGHT1 = 0x4001
GL_LIGHT2 = 0x4002
GL_LIGHT3 = 0x4003
GL_LIGHT4 = 0x4004
GL_LIGHT5 = 0x4005
GL_LIGHT6 = 0x4006
GL_LIGHT7 = 0x4007
GL_ABGR_EXT = 0x8000
GL_CONSTANT_COLOR = 0x8001
GL_CONSTANT_COLOR_EXT = 0x8001
GL_ONE_MINUS_CONSTANT_COLOR = 0x8002
GL_ONE_MINUS_CONSTANT_COLOR_EXT = 0x8002
GL_CONSTANT_ALPHA = 0x8003
GL_CONSTANT_ALPHA_EXT = 0x8003
GL_ONE_MINUS_CONSTANT_ALPHA = 0x8004
GL_ONE_MINUS_CONSTANT_ALPHA_EXT = 0x8004
GL_BLEND_COLOR = 0x8005
GL_BLEND_COLOR_EXT = 0x8005
GL_FUNC_ADD = 0x8006
GL_FUNC_ADD_EXT = 0x8006
GL_FUNC_ADD_OES = 0x8006
GL_MIN = 0x8007
GL_MIN_EXT = 0x8007
GL_MAX = 0x8008
GL_MAX_EXT = 0x8008
GL_BLEND_EQUATION = 0x8009
GL_BLEND_EQUATION_EXT = 0x8009
GL_BLEND_EQUATION_OES = 0x8009
GL_BLEND_EQUATION_RGB = 0x8009
GL_BLEND_EQUATION_RGB_EXT = 0x8009
GL_BLEND_EQUATION_RGB_OES = 0x8009
GL_FUNC_SUBTRACT = 0x800A
GL_FUNC_SUBTRACT_EXT = 0x800A
GL_FUNC_SUBTRACT_OES = 0x800A
GL_FUNC_REVERSE_SUBTRACT = 0x800B
GL_FUNC_REVERSE_SUBTRACT_EXT = 0x800B
GL_FUNC_REVERSE_SUBTRACT_OES = 0x800B
GL_CMYK_EXT = 0x800C
GL_CMYKA_EXT = 0x800D
GL_PACK_CMYK_HINT_EXT = 0x800E
GL_UNPACK_CMYK_HINT_EXT = 0x800F
GL_CONVOLUTION_1D = 0x8010
GL_CONVOLUTION_1D_EXT = 0x8010
GL_CONVOLUTION_2D = 0x8011
GL_CONVOLUTION_2D_EXT = 0x8011
GL_SEPARABLE_2D = 0x8012
GL_SEPARABLE_2D_EXT = 0x8012
GL_CONVOLUTION_BORDER_MODE = 0x8013
GL_CONVOLUTION_BORDER_MODE_EXT = 0x8013
GL_CONVOLUTION_FILTER_SCALE = 0x8014
GL_CONVOLUTION_FILTER_SCALE_EXT = 0x8014
GL_CONVOLUTION_FILTER_BIAS = 0x8015
GL_CONVOLUTION_FILTER_BIAS_EXT = 0x8015
GL_REDUCE = 0x8016
GL_REDUCE_EXT = 0x8016
GL_CONVOLUTION_FORMAT = 0x8017
GL_CONVOLUTION_FORMAT_EXT = 0x8017
GL_CONVOLUTION_WIDTH = 0x8018
GL_CONVOLUTION_WIDTH_EXT = 0x8018
GL_CONVOLUTION_HEIGHT = 0x8019
GL_CONVOLUTION_HEIGHT_EXT = 0x8019
GL_MAX_CONVOLUTION_WIDTH = 0x801A
GL_MAX_CONVOLUTION_WIDTH_EXT = 0x801A
GL_MAX_CONVOLUTION_HEIGHT = 0x801B
GL_MAX_CONVOLUTION_HEIGHT_EXT = 0x801B
GL_POST_CONVOLUTION_RED_SCALE = 0x801C
GL_POST_CONVOLUTION_RED_SCALE_EXT = 0x801C
GL_POST_CONVOLUTION_GREEN_SCALE = 0x801D
GL_POST_CONVOLUTION_GREEN_SCALE_EXT = 0x801D
GL_POST_CONVOLUTION_BLUE_SCALE = 0x801E
GL_POST_CONVOLUTION_BLUE_SCALE_EXT = 0x801E
GL_POST_CONVOLUTION_ALPHA_SCALE = 0x801F
GL_POST_CONVOLUTION_ALPHA_SCALE_EXT = 0x801F
GL_POST_CONVOLUTION_RED_BIAS = 0x8020
GL_POST_CONVOLUTION_RED_BIAS_EXT = 0x8020
GL_POST_CONVOLUTION_GREEN_BIAS = 0x8021
GL_POST_CONVOLUTION_GREEN_BIAS_EXT = 0x8021
GL_POST_CONVOLUTION_BLUE_BIAS = 0x8022
GL_POST_CONVOLUTION_BLUE_BIAS_EXT = 0x8022
GL_POST_CONVOLUTION_ALPHA_BIAS = 0x8023
GL_POST_CONVOLUTION_ALPHA_BIAS_EXT = 0x8023
GL_HISTOGRAM = 0x8024
GL_HISTOGRAM_EXT = 0x8024
GL_PROXY_HISTOGRAM = 0x8025
GL_PROXY_HISTOGRAM_EXT = 0x8025
GL_HISTOGRAM_WIDTH = 0x8026
GL_HISTOGRAM_WIDTH_EXT = 0x8026
GL_HISTOGRAM_FORMAT = 0x8027
GL_HISTOGRAM_FORMAT_EXT = 0x8027
GL_HISTOGRAM_RED_SIZE = 0x8028
GL_HISTOGRAM_RED_SIZE_EXT = 0x8028
GL_HISTOGRAM_GREEN_SIZE = 0x8029
GL_HISTOGRAM_GREEN_SIZE_EXT = 0x8029
GL_HISTOGRAM_BLUE_SIZE = 0x802A
GL_HISTOGRAM_BLUE_SIZE_EXT = 0x802A
GL_HISTOGRAM_ALPHA_SIZE = 0x802B
GL_HISTOGRAM_ALPHA_SIZE_EXT = 0x802B
GL_HISTOGRAM_LUMINANCE_SIZE = 0x802C
GL_HISTOGRAM_LUMINANCE_SIZE_EXT = 0x802C
GL_HISTOGRAM_SINK = 0x802D
GL_HISTOGRAM_SINK_EXT = 0x802D
GL_MINMAX = 0x802E
GL_MINMAX_EXT = 0x802E
GL_MINMAX_FORMAT = 0x802F
GL_MINMAX_FORMAT_EXT = 0x802F
GL_MINMAX_SINK = 0x8030
GL_MINMAX_SINK_EXT = 0x8030
GL_TABLE_TOO_LARGE_EXT = 0x8031
GL_TABLE_TOO_LARGE = 0x8031
GL_UNSIGNED_BYTE_3_3_2 = 0x8032
GL_UNSIGNED_BYTE_3_3_2_EXT = 0x8032
GL_UNSIGNED_SHORT_4_4_4_4 = 0x8033
GL_UNSIGNED_SHORT_4_4_4_4_EXT = 0x8033
GL_UNSIGNED_SHORT_5_5_5_1 = 0x8034
GL_UNSIGNED_SHORT_5_5_5_1_EXT = 0x8034
GL_UNSIGNED_INT_8_8_8_8 = 0x8035
GL_UNSIGNED_INT_8_8_8_8_EXT = 0x8035
GL_UNSIGNED_INT_10_10_10_2 = 0x8036
GL_UNSIGNED_INT_10_10_10_2_EXT = 0x8036
GL_POLYGON_OFFSET_EXT = 0x8037
GL_POLYGON_OFFSET_FILL = 0x8037
GL_POLYGON_OFFSET_FACTOR = 0x8038
GL_POLYGON_OFFSET_FACTOR_EXT = 0x8038
GL_POLYGON_OFFSET_BIAS_EXT = 0x8039
GL_RESCALE_NORMAL = 0x803A
GL_RESCALE_NORMAL_EXT = 0x803A
GL_ALPHA4 = 0x803B
GL_ALPHA4_EXT = 0x803B
GL_ALPHA8 = 0x803C
GL_ALPHA8_EXT = 0x803C
GL_ALPHA8_OES = 0x803C
GL_ALPHA12 = 0x803D
GL_ALPHA12_EXT = 0x803D
GL_ALPHA16 = 0x803E
GL_ALPHA16_EXT = 0x803E
GL_LUMINANCE4 = 0x803F
GL_LUMINANCE4_EXT = 0x803F
GL_LUMINANCE8 = 0x8040
GL_LUMINANCE8_EXT = 0x8040
GL_LUMINANCE8_OES = 0x8040
GL_LUMINANCE12 = 0x8041
GL_LUMINANCE12_EXT = 0x8041
GL_LUMINANCE16 = 0x8042
GL_LUMINANCE16_EXT = 0x8042
GL_LUMINANCE4_ALPHA4 = 0x8043
GL_LUMINANCE4_ALPHA4_EXT = 0x8043
GL_LUMINANCE4_ALPHA4_OES = 0x8043
GL_LUMINANCE6_ALPHA2 = 0x8044
GL_LUMINANCE6_ALPHA2_EXT = 0x8044
GL_LUMINANCE8_ALPHA8 = 0x8045
GL_LUMINANCE8_ALPHA8_EXT = 0x8045
GL_LUMINANCE8_ALPHA8_OES = 0x8045
GL_LUMINANCE12_ALPHA4 = 0x8046
GL_LUMINANCE12_ALPHA4_EXT = 0x8046
GL_LUMINANCE12_ALPHA12 = 0x8047
GL_LUMINANCE12_ALPHA12_EXT = 0x8047
GL_LUMINANCE16_ALPHA16 = 0x8048
GL_LUMINANCE16_ALPHA16_EXT = 0x8048
GL_INTENSITY = 0x8049
GL_INTENSITY_EXT = 0x8049
GL_INTENSITY4 = 0x804A
GL_INTENSITY4_EXT = 0x804A
GL_INTENSITY8 = 0x804B
GL_INTENSITY8_EXT = 0x804B
GL_INTENSITY12 = 0x804C
GL_INTENSITY12_EXT = 0x804C
GL_INTENSITY16 = 0x804D
GL_INTENSITY16_EXT = 0x804D
GL_RGB2_EXT = 0x804E
GL_RGB4 = 0x804F
GL_RGB4_EXT = 0x804F
GL_RGB5 = 0x8050
GL_RGB5_EXT = 0x8050
GL_RGB8 = 0x8051
GL_RGB8_EXT = 0x8051
GL_RGB8_OES = 0x8051
GL_RGB10 = 0x8052
GL_RGB10_EXT = 0x8052
GL_RGB12 = 0x8053
GL_RGB12_EXT = 0x8053
GL_RGB16 = 0x8054
GL_RGB16_EXT = 0x8054
GL_RGBA2 = 0x8055
GL_RGBA2_EXT = 0x8055
GL_RGBA4 = 0x8056
GL_RGBA4_EXT = 0x8056
GL_RGBA4_OES = 0x8056
GL_RGB5_A1 = 0x8057
GL_RGB5_A1_EXT = 0x8057
GL_RGB5_A1_OES = 0x8057
GL_RGBA8 = 0x8058
GL_RGBA8_EXT = 0x8058
GL_RGBA8_OES = 0x8058
GL_RGB10_A2 = 0x8059
GL_RGB10_A2_EXT = 0x8059
GL_RGBA12 = 0x805A
GL_RGBA12_EXT = 0x805A
GL_RGBA16 = 0x805B
GL_RGBA16_EXT = 0x805B
GL_TEXTURE_RED_SIZE = 0x805C
GL_TEXTURE_RED_SIZE_EXT = 0x805C
GL_TEXTURE_GREEN_SIZE = 0x805D
GL_TEXTURE_GREEN_SIZE_EXT = 0x805D
GL_TEXTURE_BLUE_SIZE = 0x805E
GL_TEXTURE_BLUE_SIZE_EXT = 0x805E
GL_TEXTURE_ALPHA_SIZE = 0x805F
GL_TEXTURE_ALPHA_SIZE_EXT = 0x805F
GL_TEXTURE_LUMINANCE_SIZE = 0x8060
GL_TEXTURE_LUMINANCE_SIZE_EXT = 0x8060
GL_TEXTURE_INTENSITY_SIZE = 0x8061
GL_TEXTURE_INTENSITY_SIZE_EXT = 0x8061
GL_REPLACE_EXT = 0x8062
GL_PROXY_TEXTURE_1D = 0x8063
GL_PROXY_TEXTURE_1D_EXT = 0x8063
GL_PROXY_TEXTURE_2D = 0x8064
GL_PROXY_TEXTURE_2D_EXT = 0x8064
GL_TEXTURE_TOO_LARGE_EXT = 0x8065
GL_TEXTURE_PRIORITY = 0x8066
GL_TEXTURE_PRIORITY_EXT = 0x8066
GL_TEXTURE_RESIDENT = 0x8067
GL_TEXTURE_RESIDENT_EXT = 0x8067
GL_TEXTURE_1D_BINDING_EXT = 0x8068
GL_TEXTURE_BINDING_1D = 0x8068
GL_TEXTURE_2D_BINDING_EXT = 0x8069
GL_TEXTURE_BINDING_2D = 0x8069
GL_TEXTURE_3D_BINDING_EXT = 0x806A
GL_TEXTURE_3D_BINDING_OES = 0x806A
GL_TEXTURE_BINDING_3D = 0x806A
GL_TEXTURE_BINDING_3D_OES = 0x806A
GL_PACK_SKIP_IMAGES = 0x806B
GL_PACK_SKIP_IMAGES_EXT = 0x806B
GL_PACK_IMAGE_HEIGHT = 0x806C
GL_PACK_IMAGE_HEIGHT_EXT = 0x806C
GL_UNPACK_SKIP_IMAGES = 0x806D
GL_UNPACK_SKIP_IMAGES_EXT = 0x806D
GL_UNPACK_IMAGE_HEIGHT = 0x806E
GL_UNPACK_IMAGE_HEIGHT_EXT = 0x806E
GL_TEXTURE_3D = 0x806F
GL_TEXTURE_3D_EXT = 0x806F
GL_TEXTURE_3D_OES = 0x806F
GL_PROXY_TEXTURE_3D = 0x8070
GL_PROXY_TEXTURE_3D_EXT = 0x8070
GL_TEXTURE_DEPTH = 0x8071
GL_TEXTURE_DEPTH_EXT = 0x8071
GL_TEXTURE_WRAP_R = 0x8072
GL_TEXTURE_WRAP_R_EXT = 0x8072
GL_TEXTURE_WRAP_R_OES = 0x8072
GL_MAX_3D_TEXTURE_SIZE = 0x8073
GL_MAX_3D_TEXTURE_SIZE_EXT = 0x8073
GL_MAX_3D_TEXTURE_SIZE_OES = 0x8073
GL_VERTEX_ARRAY = 0x8074
GL_VERTEX_ARRAY_EXT = 0x8074
GL_VERTEX_ARRAY_KHR = 0x8074
GL_NORMAL_ARRAY = 0x8075
GL_NORMAL_ARRAY_EXT = 0x8075
GL_COLOR_ARRAY = 0x8076
GL_COLOR_ARRAY_EXT = 0x8076
GL_INDEX_ARRAY = 0x8077
GL_INDEX_ARRAY_EXT = 0x8077
GL_TEXTURE_COORD_ARRAY = 0x8078
GL_TEXTURE_COORD_ARRAY_EXT = 0x8078
GL_EDGE_FLAG_ARRAY = 0x8079
GL_EDGE_FLAG_ARRAY_EXT = 0x8079
GL_VERTEX_ARRAY_SIZE = 0x807A
GL_VERTEX_ARRAY_SIZE_EXT = 0x807A
GL_VERTEX_ARRAY_TYPE = 0x807B
GL_VERTEX_ARRAY_TYPE_EXT = 0x807B
GL_VERTEX_ARRAY_STRIDE = 0x807C
GL_VERTEX_ARRAY_STRIDE_EXT = 0x807C
GL_VERTEX_ARRAY_COUNT_EXT = 0x807D
GL_NORMAL_ARRAY_TYPE = 0x807E
GL_NORMAL_ARRAY_TYPE_EXT = 0x807E
GL_NORMAL_ARRAY_STRIDE = 0x807F
GL_NORMAL_ARRAY_STRIDE_EXT = 0x807F
GL_NORMAL_ARRAY_COUNT_EXT = 0x8080
GL_COLOR_ARRAY_SIZE = 0x8081
GL_COLOR_ARRAY_SIZE_EXT = 0x8081
GL_COLOR_ARRAY_TYPE = 0x8082
GL_COLOR_ARRAY_TYPE_EXT = 0x8082
GL_COLOR_ARRAY_STRIDE = 0x8083
GL_COLOR_ARRAY_STRIDE_EXT = 0x8083
GL_COLOR_ARRAY_COUNT_EXT = 0x8084
GL_INDEX_ARRAY_TYPE = 0x8085
GL_INDEX_ARRAY_TYPE_EXT = 0x8085
GL_INDEX_ARRAY_STRIDE = 0x8086
GL_INDEX_ARRAY_STRIDE_EXT = 0x8086
GL_INDEX_ARRAY_COUNT_EXT = 0x8087
GL_TEXTURE_COORD_ARRAY_SIZE = 0x8088
GL_TEXTURE_COORD_ARRAY_SIZE_EXT = 0x8088
GL_TEXTURE_COORD_ARRAY_TYPE = 0x8089
GL_TEXTURE_COORD_ARRAY_TYPE_EXT = 0x8089
GL_TEXTURE_COORD_ARRAY_STRIDE = 0x808A
GL_TEXTURE_COORD_ARRAY_STRIDE_EXT = 0x808A
GL_TEXTURE_COORD_ARRAY_COUNT_EXT = 0x808B
GL_EDGE_FLAG_ARRAY_STRIDE = 0x808C
GL_EDGE_FLAG_ARRAY_STRIDE_EXT = 0x808C
GL_EDGE_FLAG_ARRAY_COUNT_EXT = 0x808D
GL_VERTEX_ARRAY_POINTER = 0x808E
GL_VERTEX_ARRAY_POINTER_EXT = 0x808E
GL_NORMAL_ARRAY_POINTER = 0x808F
GL_NORMAL_ARRAY_POINTER_EXT = 0x808F
GL_COLOR_ARRAY_POINTER = 0x8090
GL_COLOR_ARRAY_POINTER_EXT = 0x8090
GL_INDEX_ARRAY_POINTER = 0x8091
GL_INDEX_ARRAY_POINTER_EXT = 0x8091
GL_TEXTURE_COORD_ARRAY_POINTER = 0x8092
GL_TEXTURE_COORD_ARRAY_POINTER_EXT = 0x8092
GL_EDGE_FLAG_ARRAY_POINTER = 0x8093
GL_EDGE_FLAG_ARRAY_POINTER_EXT = 0x8093
GL_INTERLACE_SGIX = 0x8094
GL_DETAIL_TEXTURE_2D_SGIS = 0x8095
GL_DETAIL_TEXTURE_2D_BINDING_SGIS = 0x8096
GL_LINEAR_DETAIL_SGIS = 0x8097
GL_LINEAR_DETAIL_ALPHA_SGIS = 0x8098
GL_LINEAR_DETAIL_COLOR_SGIS = 0x8099
GL_DETAIL_TEXTURE_LEVEL_SGIS = 0x809A
GL_DETAIL_TEXTURE_MODE_SGIS = 0x809B
GL_DETAIL_TEXTURE_FUNC_POINTS_SGIS = 0x809C
GL_MULTISAMPLE = 0x809D
GL_MULTISAMPLE_ARB = 0x809D
GL_MULTISAMPLE_EXT = 0x809D
GL_MULTISAMPLE_SGIS = 0x809D
GL_SAMPLE_ALPHA_TO_COVERAGE = 0x809E
GL_SAMPLE_ALPHA_TO_COVERAGE_ARB = 0x809E
GL_SAMPLE_ALPHA_TO_MASK_EXT = 0x809E
GL_SAMPLE_ALPHA_TO_MASK_SGIS = 0x809E
GL_SAMPLE_ALPHA_TO_ONE = 0x809F
GL_SAMPLE_ALPHA_TO_ONE_ARB = 0x809F
GL_SAMPLE_ALPHA_TO_ONE_EXT = 0x809F
GL_SAMPLE_ALPHA_TO_ONE_SGIS = 0x809F
GL_SAMPLE_COVERAGE = 0x80A0
GL_SAMPLE_COVERAGE_ARB = 0x80A0
GL_SAMPLE_MASK_EXT = 0x80A0
GL_SAMPLE_MASK_SGIS = 0x80A0
GL_1PASS_EXT = 0x80A1
GL_1PASS_SGIS = 0x80A1
GL_2PASS_0_EXT = 0x80A2
GL_2PASS_0_SGIS = 0x80A2
GL_2PASS_1_EXT = 0x80A3
GL_2PASS_1_SGIS = 0x80A3
GL_4PASS_0_EXT = 0x80A4
GL_4PASS_0_SGIS = 0x80A4
GL_4PASS_1_EXT = 0x80A5
GL_4PASS_1_SGIS = 0x80A5
GL_4PASS_2_EXT = 0x80A6
GL_4PASS_2_SGIS = 0x80A6
GL_4PASS_3_EXT = 0x80A7
GL_4PASS_3_SGIS = 0x80A7
GL_SAMPLE_BUFFERS = 0x80A8
GL_SAMPLE_BUFFERS_ARB = 0x80A8
GL_SAMPLE_BUFFERS_EXT = 0x80A8
GL_SAMPLE_BUFFERS_SGIS = 0x80A8
GL_SAMPLES = 0x80A9
GL_SAMPLES_ARB = 0x80A9
GL_SAMPLES_EXT = 0x80A9
GL_SAMPLES_SGIS = 0x80A9
GL_SAMPLE_COVERAGE_VALUE = 0x80AA
GL_SAMPLE_COVERAGE_VALUE_ARB = 0x80AA
GL_SAMPLE_MASK_VALUE_EXT = 0x80AA
GL_SAMPLE_MASK_VALUE_SGIS = 0x80AA
GL_SAMPLE_COVERAGE_INVERT = 0x80AB
GL_SAMPLE_COVERAGE_INVERT_ARB = 0x80AB
GL_SAMPLE_MASK_INVERT_EXT = 0x80AB
GL_SAMPLE_MASK_INVERT_SGIS = 0x80AB
GL_SAMPLE_PATTERN_EXT = 0x80AC
GL_SAMPLE_PATTERN_SGIS = 0x80AC
GL_LINEAR_SHARPEN_SGIS = 0x80AD
GL_LINEAR_SHARPEN_ALPHA_SGIS = 0x80AE
GL_LINEAR_SHARPEN_COLOR_SGIS = 0x80AF
GL_SHARPEN_TEXTURE_FUNC_POINTS_SGIS = 0x80B0
GL_COLOR_MATRIX = 0x80B1
GL_COLOR_MATRIX_SGI = 0x80B1
GL_COLOR_MATRIX_STACK_DEPTH = 0x80B2
GL_COLOR_MATRIX_STACK_DEPTH_SGI = 0x80B2
GL_MAX_COLOR_MATRIX_STACK_DEPTH = 0x80B3
GL_MAX_COLOR_MATRIX_STACK_DEPTH_SGI = 0x80B3
GL_POST_COLOR_MATRIX_RED_SCALE = 0x80B4
GL_POST_COLOR_MATRIX_RED_SCALE_SGI = 0x80B4
GL_POST_COLOR_MATRIX_GREEN_SCALE = 0x80B5
GL_POST_COLOR_MATRIX_GREEN_SCALE_SGI = 0x80B5
GL_POST_COLOR_MATRIX_BLUE_SCALE = 0x80B6
GL_POST_COLOR_MATRIX_BLUE_SCALE_SGI = 0x80B6
GL_POST_COLOR_MATRIX_ALPHA_SCALE = 0x80B7
GL_POST_COLOR_MATRIX_ALPHA_SCALE_SGI = 0x80B7
GL_POST_COLOR_MATRIX_RED_BIAS = 0x80B8
GL_POST_COLOR_MATRIX_RED_BIAS_SGI = 0x80B8
GL_POST_COLOR_MATRIX_GREEN_BIAS = 0x80B9
GL_POST_COLOR_MATRIX_GREEN_BIAS_SGI = 0x80B9
GL_POST_COLOR_MATRIX_BLUE_BIAS = 0x80BA
GL_POST_COLOR_MATRIX_BLUE_BIAS_SGI = 0x80BA
GL_POST_COLOR_MATRIX_ALPHA_BIAS = 0x80BB
GL_POST_COLOR_MATRIX_ALPHA_BIAS_SGI = 0x80BB
GL_TEXTURE_COLOR_TABLE_SGI = 0x80BC
GL_PROXY_TEXTURE_COLOR_TABLE_SGI = 0x80BD
GL_TEXTURE_ENV_BIAS_SGIX = 0x80BE
GL_SHADOW_AMBIENT_SGIX = 0x80BF
GL_TEXTURE_COMPARE_FAIL_VALUE_ARB = 0x80BF
GL_BLEND_DST_RGB = 0x80C8
GL_BLEND_DST_RGB_EXT = 0x80C8
GL_BLEND_DST_RGB_OES = 0x80C8
GL_BLEND_SRC_RGB = 0x80C9
GL_BLEND_SRC_RGB_EXT = 0x80C9
GL_BLEND_SRC_RGB_OES = 0x80C9
GL_BLEND_DST_ALPHA = 0x80CA
GL_BLEND_DST_ALPHA_EXT = 0x80CA
GL_BLEND_DST_ALPHA_OES = 0x80CA
GL_BLEND_SRC_ALPHA = 0x80CB
GL_BLEND_SRC_ALPHA_EXT = 0x80CB
GL_BLEND_SRC_ALPHA_OES = 0x80CB
GL_422_EXT = 0x80CC
GL_422_REV_EXT = 0x80CD
GL_422_AVERAGE_EXT = 0x80CE
GL_422_REV_AVERAGE_EXT = 0x80CF
GL_COLOR_TABLE = 0x80D0
GL_COLOR_TABLE_SGI = 0x80D0
GL_POST_CONVOLUTION_COLOR_TABLE = 0x80D1
GL_POST_CONVOLUTION_COLOR_TABLE_SGI = 0x80D1
GL_POST_COLOR_MATRIX_COLOR_TABLE = 0x80D2
GL_POST_COLOR_MATRIX_COLOR_TABLE_SGI = 0x80D2
GL_PROXY_COLOR_TABLE = 0x80D3
GL_PROXY_COLOR_TABLE_SGI = 0x80D3
GL_PROXY_POST_CONVOLUTION_COLOR_TABLE = 0x80D4
GL_PROXY_POST_CONVOLUTION_COLOR_TABLE_SGI = 0x80D4
GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE = 0x80D5
GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE_SGI = 0x80D5
GL_COLOR_TABLE_SCALE = 0x80D6
GL_COLOR_TABLE_SCALE_SGI = 0x80D6
GL_COLOR_TABLE_BIAS = 0x80D7
GL_COLOR_TABLE_BIAS_SGI = 0x80D7
GL_COLOR_TABLE_FORMAT = 0x80D8
GL_COLOR_TABLE_FORMAT_SGI = 0x80D8
GL_COLOR_TABLE_WIDTH = 0x80D9
GL_COLOR_TABLE_WIDTH_SGI = 0x80D9
GL_COLOR_TABLE_RED_SIZE = 0x80DA
GL_COLOR_TABLE_RED_SIZE_SGI = 0x80DA
GL_COLOR_TABLE_GREEN_SIZE = 0x80DB
GL_COLOR_TABLE_GREEN_SIZE_SGI = 0x80DB
GL_COLOR_TABLE_BLUE_SIZE = 0x80DC
GL_COLOR_TABLE_BLUE_SIZE_SGI = 0x80DC
GL_COLOR_TABLE_ALPHA_SIZE = 0x80DD
GL_COLOR_TABLE_ALPHA_SIZE_SGI = 0x80DD
GL_COLOR_TABLE_LUMINANCE_SIZE = 0x80DE
GL_COLOR_TABLE_LUMINANCE_SIZE_SGI = 0x80DE
GL_COLOR_TABLE_INTENSITY_SIZE = 0x80DF
GL_COLOR_TABLE_INTENSITY_SIZE_SGI = 0x80DF
GL_BGR = 0x80E0
GL_BGR_EXT = 0x80E0
GL_BGRA = 0x80E1
GL_BGRA_EXT = 0x80E1
GL_BGRA_IMG = 0x80E1
GL_COLOR_INDEX1_EXT = 0x80E2
GL_COLOR_INDEX2_EXT = 0x80E3
GL_COLOR_INDEX4_EXT = 0x80E4
GL_COLOR_INDEX8_EXT = 0x80E5
GL_COLOR_INDEX12_EXT = 0x80E6
GL_COLOR_INDEX16_EXT = 0x80E7
GL_MAX_ELEMENTS_VERTICES = 0x80E8
GL_MAX_ELEMENTS_VERTICES_EXT = 0x80E8
GL_MAX_ELEMENTS_INDICES = 0x80E9
GL_MAX_ELEMENTS_INDICES_EXT = 0x80E9
GL_PHONG_WIN = 0x80EA
GL_PHONG_HINT_WIN = 0x80EB
GL_FOG_SPECULAR_TEXTURE_WIN = 0x80EC
GL_TEXTURE_INDEX_SIZE_EXT = 0x80ED
GL_PARAMETER_BUFFER_ARB = 0x80EE
GL_PARAMETER_BUFFER_BINDING_ARB = 0x80EF
GL_CLIP_VOLUME_CLIPPING_HINT_EXT = 0x80F0
GL_DUAL_ALPHA4_SGIS = 0x8110
GL_DUAL_ALPHA8_SGIS = 0x8111
GL_DUAL_ALPHA12_SGIS = 0x8112
GL_DUAL_ALPHA16_SGIS = 0x8113
GL_DUAL_LUMINANCE4_SGIS = 0x8114
GL_DUAL_LUMINANCE8_SGIS = 0x8115
GL_DUAL_LUMINANCE12_SGIS = 0x8116
GL_DUAL_LUMINANCE16_SGIS = 0x8117
GL_DUAL_INTENSITY4_SGIS = 0x8118
GL_DUAL_INTENSITY8_SGIS = 0x8119
GL_DUAL_INTENSITY12_SGIS = 0x811A
GL_DUAL_INTENSITY16_SGIS = 0x811B
GL_DUAL_LUMINANCE_ALPHA4_SGIS = 0x811C
GL_DUAL_LUMINANCE_ALPHA8_SGIS = 0x811D
GL_QUAD_ALPHA4_SGIS = 0x811E
GL_QUAD_ALPHA8_SGIS = 0x811F
GL_QUAD_LUMINANCE4_SGIS = 0x8120
GL_QUAD_LUMINANCE8_SGIS = 0x8121
GL_QUAD_INTENSITY4_SGIS = 0x8122
GL_QUAD_INTENSITY8_SGIS = 0x8123
GL_DUAL_TEXTURE_SELECT_SGIS = 0x8124
GL_QUAD_TEXTURE_SELECT_SGIS = 0x8125
GL_POINT_SIZE_MIN = 0x8126
GL_POINT_SIZE_MIN_ARB = 0x8126
GL_POINT_SIZE_MIN_EXT = 0x8126
GL_POINT_SIZE_MIN_SGIS = 0x8126
GL_POINT_SIZE_MAX = 0x8127
GL_POINT_SIZE_MAX_ARB = 0x8127
GL_POINT_SIZE_MAX_EXT = 0x8127
GL_POINT_SIZE_MAX_SGIS = 0x8127
GL_POINT_FADE_THRESHOLD_SIZE = 0x8128
GL_POINT_FADE_THRESHOLD_SIZE_ARB = 0x8128
GL_POINT_FADE_THRESHOLD_SIZE_EXT = 0x8128
GL_POINT_FADE_THRESHOLD_SIZE_SGIS = 0x8128
GL_DISTANCE_ATTENUATION_EXT = 0x8129
GL_DISTANCE_ATTENUATION_SGIS = 0x8129
GL_POINT_DISTANCE_ATTENUATION = 0x8129
GL_POINT_DISTANCE_ATTENUATION_ARB = 0x8129
GL_FOG_FUNC_SGIS = 0x812A
GL_FOG_FUNC_POINTS_SGIS = 0x812B
GL_MAX_FOG_FUNC_POINTS_SGIS = 0x812C
GL_CLAMP_TO_BORDER = 0x812D
GL_CLAMP_TO_BORDER_ARB = 0x812D
GL_CLAMP_TO_BORDER_EXT = 0x812D
GL_CLAMP_TO_BORDER_NV = 0x812D
GL_CLAMP_TO_BORDER_SGIS = 0x812D
GL_TEXTURE_MULTI_BUFFER_HINT_SGIX = 0x812E
GL_CLAMP_TO_EDGE = 0x812F
GL_CLAMP_TO_EDGE_SGIS = 0x812F
GL_PACK_SKIP_VOLUMES_SGIS = 0x8130
GL_PACK_IMAGE_DEPTH_SGIS = 0x8131
GL_UNPACK_SKIP_VOLUMES_SGIS = 0x8132
GL_UNPACK_IMAGE_DEPTH_SGIS = 0x8133
GL_TEXTURE_4D_SGIS = 0x8134
GL_PROXY_TEXTURE_4D_SGIS = 0x8135
GL_TEXTURE_4DSIZE_SGIS = 0x8136
GL_TEXTURE_WRAP_Q_SGIS = 0x8137
GL_MAX_4D_TEXTURE_SIZE_SGIS = 0x8138
GL_PIXEL_TEX_GEN_SGIX = 0x8139
GL_TEXTURE_MIN_LOD = 0x813A
GL_TEXTURE_MIN_LOD_SGIS = 0x813A
GL_TEXTURE_MAX_LOD = 0x813B
GL_TEXTURE_MAX_LOD_SGIS = 0x813B
GL_TEXTURE_BASE_LEVEL = 0x813C
GL_TEXTURE_BASE_LEVEL_SGIS = 0x813C
GL_TEXTURE_MAX_LEVEL = 0x813D
GL_TEXTURE_MAX_LEVEL_APPLE = 0x813D
GL_TEXTURE_MAX_LEVEL_SGIS = 0x813D
GL_PIXEL_TILE_BEST_ALIGNMENT_SGIX = 0x813E
GL_PIXEL_TILE_CACHE_INCREMENT_SGIX = 0x813F
GL_PIXEL_TILE_WIDTH_SGIX = 0x8140
GL_PIXEL_TILE_HEIGHT_SGIX = 0x8141
GL_PIXEL_TILE_GRID_WIDTH_SGIX = 0x8142
GL_PIXEL_TILE_GRID_HEIGHT_SGIX = 0x8143
GL_PIXEL_TILE_GRID_DEPTH_SGIX = 0x8144
GL_PIXEL_TILE_CACHE_SIZE_SGIX = 0x8145
GL_FILTER4_SGIS = 0x8146
GL_TEXTURE_FILTER4_SIZE_SGIS = 0x8147
GL_SPRITE_SGIX = 0x8148
GL_SPRITE_MODE_SGIX = 0x8149
GL_SPRITE_AXIS_SGIX = 0x814A
GL_SPRITE_TRANSLATION_SGIX = 0x814B
GL_SPRITE_AXIAL_SGIX = 0x814C
GL_SPRITE_OBJECT_ALIGNED_SGIX = 0x814D
GL_SPRITE_EYE_ALIGNED_SGIX = 0x814E
GL_TEXTURE_4D_BINDING_SGIS = 0x814F
GL_IGNORE_BORDER_HP = 0x8150
GL_CONSTANT_BORDER = 0x8151
GL_CONSTANT_BORDER_HP = 0x8151
GL_REPLICATE_BORDER = 0x8153
GL_REPLICATE_BORDER_HP = 0x8153
GL_CONVOLUTION_BORDER_COLOR = 0x8154
GL_CONVOLUTION_BORDER_COLOR_HP = 0x8154
GL_IMAGE_SCALE_X_HP = 0x8155
GL_IMAGE_SCALE_Y_HP = 0x8156
GL_IMAGE_TRANSLATE_X_HP = 0x8157
GL_IMAGE_TRANSLATE_Y_HP = 0x8158
GL_IMAGE_ROTATE_ANGLE_HP = 0x8159
GL_IMAGE_ROTATE_ORIGIN_X_HP = 0x815A
GL_IMAGE_ROTATE_ORIGIN_Y_HP = 0x815B
GL_IMAGE_MAG_FILTER_HP = 0x815C
GL_IMAGE_MIN_FILTER_HP = 0x815D
GL_IMAGE_CUBIC_WEIGHT_HP = 0x815E
GL_CUBIC_HP = 0x815F
GL_AVERAGE_HP = 0x8160
GL_IMAGE_TRANSFORM_2D_HP = 0x8161
GL_POST_IMAGE_TRANSFORM_COLOR_TABLE_HP = 0x8162
GL_PROXY_POST_IMAGE_TRANSFORM_COLOR_TABLE_HP = 0x8163
GL_OCCLUSION_TEST_HP = 0x8165
GL_OCCLUSION_TEST_RESULT_HP = 0x8166
GL_TEXTURE_LIGHTING_MODE_HP = 0x8167
GL_TEXTURE_POST_SPECULAR_HP = 0x8168
GL_TEXTURE_PRE_SPECULAR_HP = 0x8169
GL_LINEAR_CLIPMAP_LINEAR_SGIX = 0x8170
GL_TEXTURE_CLIPMAP_CENTER_SGIX = 0x8171
GL_TEXTURE_CLIPMAP_FRAME_SGIX = 0x8172
GL_TEXTURE_CLIPMAP_OFFSET_SGIX = 0x8173
GL_TEXTURE_CLIPMAP_VIRTUAL_DEPTH_SGIX = 0x8174
GL_TEXTURE_CLIPMAP_LOD_OFFSET_SGIX = 0x8175
GL_TEXTURE_CLIPMAP_DEPTH_SGIX = 0x8176
GL_MAX_CLIPMAP_DEPTH_SGIX = 0x8177
GL_MAX_CLIPMAP_VIRTUAL_DEPTH_SGIX = 0x8178
GL_POST_TEXTURE_FILTER_BIAS_SGIX = 0x8179
GL_POST_TEXTURE_FILTER_SCALE_SGIX = 0x817A
GL_POST_TEXTURE_FILTER_BIAS_RANGE_SGIX = 0x817B
GL_POST_TEXTURE_FILTER_SCALE_RANGE_SGIX = 0x817C
GL_REFERENCE_PLANE_SGIX = 0x817D
GL_REFERENCE_PLANE_EQUATION_SGIX = 0x817E
GL_IR_INSTRUMENT1_SGIX = 0x817F
GL_INSTRUMENT_BUFFER_POINTER_SGIX = 0x8180
GL_INSTRUMENT_MEASUREMENTS_SGIX = 0x8181
GL_LIST_PRIORITY_SGIX = 0x8182
GL_CALLIGRAPHIC_FRAGMENT_SGIX = 0x8183
GL_PIXEL_TEX_GEN_Q_CEILING_SGIX = 0x8184
GL_PIXEL_TEX_GEN_Q_ROUND_SGIX = 0x8185
GL_PIXEL_TEX_GEN_Q_FLOOR_SGIX = 0x8186
GL_PIXEL_TEX_GEN_ALPHA_REPLACE_SGIX = 0x8187
GL_PIXEL_TEX_GEN_ALPHA_NO_REPLACE_SGIX = 0x8188
GL_PIXEL_TEX_GEN_ALPHA_LS_SGIX = 0x8189
GL_PIXEL_TEX_GEN_ALPHA_MS_SGIX = 0x818A
GL_FRAMEZOOM_SGIX = 0x818B
GL_FRAMEZOOM_FACTOR_SGIX = 0x818C
GL_MAX_FRAMEZOOM_FACTOR_SGIX = 0x818D
GL_TEXTURE_LOD_BIAS_S_SGIX = 0x818E
GL_TEXTURE_LOD_BIAS_T_SGIX = 0x818F
GL_TEXTURE_LOD_BIAS_R_SGIX = 0x8190
GL_GENERATE_MIPMAP = 0x8191
GL_GENERATE_MIPMAP_SGIS = 0x8191
GL_GENERATE_MIPMAP_HINT = 0x8192
GL_GENERATE_MIPMAP_HINT_SGIS = 0x8192
GL_GEOMETRY_DEFORMATION_SGIX = 0x8194
GL_TEXTURE_DEFORMATION_SGIX = 0x8195
GL_DEFORMATIONS_MASK_SGIX = 0x8196
GL_MAX_DEFORMATION_ORDER_SGIX = 0x8197
GL_FOG_OFFSET_SGIX = 0x8198
GL_FOG_OFFSET_VALUE_SGIX = 0x8199
GL_TEXTURE_COMPARE_SGIX = 0x819A
GL_TEXTURE_COMPARE_OPERATOR_SGIX = 0x819B
GL_TEXTURE_LEQUAL_R_SGIX = 0x819C
GL_TEXTURE_GEQUAL_R_SGIX = 0x819D
GL_DEPTH_COMPONENT16 = 0x81A5
GL_DEPTH_COMPONENT16_ARB = 0x81A5
GL_DEPTH_COMPONENT16_OES = 0x81A5
GL_DEPTH_COMPONENT16_SGIX = 0x81A5
GL_DEPTH_COMPONENT24 = 0x81A6
GL_DEPTH_COMPONENT24_ARB = 0x81A6
GL_DEPTH_COMPONENT24_OES = 0x81A6
GL_DEPTH_COMPONENT24_SGIX = 0x81A6
GL_DEPTH_COMPONENT32 = 0x81A7
GL_DEPTH_COMPONENT32_ARB = 0x81A7
GL_DEPTH_COMPONENT32_OES = 0x81A7
GL_DEPTH_COMPONENT32_SGIX = 0x81A7
GL_ARRAY_ELEMENT_LOCK_FIRST_EXT = 0x81A8
GL_ARRAY_ELEMENT_LOCK_COUNT_EXT = 0x81A9
GL_CULL_VERTEX_EXT = 0x81AA
GL_CULL_VERTEX_EYE_POSITION_EXT = 0x81AB
GL_CULL_VERTEX_OBJECT_POSITION_EXT = 0x81AC
GL_IUI_V2F_EXT = 0x81AD
GL_IUI_V3F_EXT = 0x81AE
GL_IUI_N3F_V2F_EXT = 0x81AF
GL_IUI_N3F_V3F_EXT = 0x81B0
GL_T2F_IUI_V2F_EXT = 0x81B1
GL_T2F_IUI_V3F_EXT = 0x81B2
GL_T2F_IUI_N3F_V2F_EXT = 0x81B3
GL_T2F_IUI_N3F_V3F_EXT = 0x81B4
GL_INDEX_TEST_EXT = 0x81B5
GL_INDEX_TEST_FUNC_EXT = 0x81B6
GL_INDEX_TEST_REF_EXT = 0x81B7
GL_INDEX_MATERIAL_EXT = 0x81B8
GL_INDEX_MATERIAL_PARAMETER_EXT = 0x81B9
GL_INDEX_MATERIAL_FACE_EXT = 0x81BA
GL_YCRCB_422_SGIX = 0x81BB
GL_YCRCB_444_SGIX = 0x81BC
GL_WRAP_BORDER_SUN = 0x81D4
GL_UNPACK_CONSTANT_DATA_SUNX = 0x81D5
GL_TEXTURE_CONSTANT_DATA_SUNX = 0x81D6
GL_TRIANGLE_LIST_SUN = 0x81D7
GL_REPLACEMENT_CODE_SUN = 0x81D8
GL_GLOBAL_ALPHA_SUN = 0x81D9
GL_GLOBAL_ALPHA_FACTOR_SUN = 0x81DA
GL_TEXTURE_COLOR_WRITEMASK_SGIS = 0x81EF
GL_EYE_DISTANCE_TO_POINT_SGIS = 0x81F0
GL_OBJECT_DISTANCE_TO_POINT_SGIS = 0x81F1
GL_EYE_DISTANCE_TO_LINE_SGIS = 0x81F2
GL_OBJECT_DISTANCE_TO_LINE_SGIS = 0x81F3
GL_EYE_POINT_SGIS = 0x81F4
GL_OBJECT_POINT_SGIS = 0x81F5
GL_EYE_LINE_SGIS = 0x81F6
GL_OBJECT_LINE_SGIS = 0x81F7
GL_LIGHT_MODEL_COLOR_CONTROL = 0x81F8
GL_LIGHT_MODEL_COLOR_CONTROL_EXT = 0x81F8
GL_SINGLE_COLOR = 0x81F9
GL_SINGLE_COLOR_EXT = 0x81F9
GL_SEPARATE_SPECULAR_COLOR = 0x81FA
GL_SEPARATE_SPECULAR_COLOR_EXT = 0x81FA
GL_SHARED_TEXTURE_PALETTE_EXT = 0x81FB
GL_TEXT_FRAGMENT_SHADER_ATI = 0x8200
GL_FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING = 0x8210
GL_FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING_EXT = 0x8210
GL_FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE = 0x8211
GL_FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE_EXT = 0x8211
GL_FRAMEBUFFER_ATTACHMENT_RED_SIZE = 0x8212
GL_FRAMEBUFFER_ATTACHMENT_GREEN_SIZE = 0x8213
GL_FRAMEBUFFER_ATTACHMENT_BLUE_SIZE = 0x8214
GL_FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE = 0x8215
GL_FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE = 0x8216
GL_FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE = 0x8217
GL_FRAMEBUFFER_DEFAULT = 0x8218
GL_FRAMEBUFFER_UNDEFINED = 0x8219
GL_FRAMEBUFFER_UNDEFINED_OES = 0x8219
GL_DEPTH_STENCIL_ATTACHMENT = 0x821A
GL_MAJOR_VERSION = 0x821B
GL_MINOR_VERSION = 0x821C
GL_NUM_EXTENSIONS = 0x821D
GL_CONTEXT_FLAGS = 0x821E
GL_BUFFER_IMMUTABLE_STORAGE = 0x821F
GL_BUFFER_STORAGE_FLAGS = 0x8220
GL_PRIMITIVE_RESTART_FOR_PATCHES_SUPPORTED = 0x8221
GL_INDEX = 0x8222
GL_COMPRESSED_RED = 0x8225
GL_COMPRESSED_RG = 0x8226
GL_RG = 0x8227
GL_RG_EXT = 0x8227
GL_RG_INTEGER = 0x8228
GL_R8 = 0x8229
GL_R8_EXT = 0x8229
GL_R16 = 0x822A
GL_RG8 = 0x822B
GL_RG8_EXT = 0x822B
GL_RG16 = 0x822C
GL_R16F = 0x822D
GL_R16F_EXT = 0x822D
GL_R32F = 0x822E
GL_R32F_EXT = 0x822E
GL_RG16F = 0x822F
GL_RG16F_EXT = 0x822F
GL_RG32F = 0x8230
GL_RG32F_EXT = 0x8230
GL_R8I = 0x8231
GL_R8UI = 0x8232
GL_R16I = 0x8233
GL_R16UI = 0x8234
GL_R32I = 0x8235
GL_R32UI = 0x8236
GL_RG8I = 0x8237
GL_RG8UI = 0x8238
GL_RG16I = 0x8239
GL_RG16UI = 0x823A
GL_RG32I = 0x823B
GL_RG32UI = 0x823C
GL_SYNC_CL_EVENT_ARB = 0x8240
GL_SYNC_CL_EVENT_COMPLETE_ARB = 0x8241
GL_DEBUG_OUTPUT_SYNCHRONOUS = 0x8242
GL_DEBUG_OUTPUT_SYNCHRONOUS_ARB = 0x8242
GL_DEBUG_OUTPUT_SYNCHRONOUS_KHR = 0x8242
GL_DEBUG_NEXT_LOGGED_MESSAGE_LENGTH = 0x8243
GL_DEBUG_NEXT_LOGGED_MESSAGE_LENGTH_ARB = 0x8243
GL_DEBUG_NEXT_LOGGED_MESSAGE_LENGTH_KHR = 0x8243
GL_DEBUG_CALLBACK_FUNCTION = 0x8244
GL_DEBUG_CALLBACK_FUNCTION_ARB = 0x8244
GL_DEBUG_CALLBACK_FUNCTION_KHR = 0x8244
GL_DEBUG_CALLBACK_USER_PARAM = 0x8245
GL_DEBUG_CALLBACK_USER_PARAM_ARB = 0x8245
GL_DEBUG_CALLBACK_USER_PARAM_KHR = 0x8245
GL_DEBUG_SOURCE_API = 0x8246
GL_DEBUG_SOURCE_API_ARB = 0x8246
GL_DEBUG_SOURCE_API_KHR = 0x8246
GL_DEBUG_SOURCE_WINDOW_SYSTEM = 0x8247
GL_DEBUG_SOURCE_WINDOW_SYSTEM_ARB = 0x8247
GL_DEBUG_SOURCE_WINDOW_SYSTEM_KHR = 0x8247
GL_DEBUG_SOURCE_SHADER_COMPILER = 0x8248
GL_DEBUG_SOURCE_SHADER_COMPILER_ARB = 0x8248
GL_DEBUG_SOURCE_SHADER_COMPILER_KHR = 0x8248
GL_DEBUG_SOURCE_THIRD_PARTY = 0x8249
GL_DEBUG_SOURCE_THIRD_PARTY_ARB = 0x8249
GL_DEBUG_SOURCE_THIRD_PARTY_KHR = 0x8249
GL_DEBUG_SOURCE_APPLICATION = 0x824A
GL_DEBUG_SOURCE_APPLICATION_ARB = 0x824A
GL_DEBUG_SOURCE_APPLICATION_KHR = 0x824A
GL_DEBUG_SOURCE_OTHER = 0x824B
GL_DEBUG_SOURCE_OTHER_ARB = 0x824B
GL_DEBUG_SOURCE_OTHER_KHR = 0x824B
GL_DEBUG_TYPE_ERROR = 0x824C
GL_DEBUG_TYPE_ERROR_ARB = 0x824C
GL_DEBUG_TYPE_ERROR_KHR = 0x824C
GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR = 0x824D
GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_ARB = 0x824D
GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR_KHR = 0x824D
GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR = 0x824E
GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_ARB = 0x824E
GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR_KHR = 0x824E
GL_DEBUG_TYPE_PORTABILITY = 0x824F
GL_DEBUG_TYPE_PORTABILITY_ARB = 0x824F
GL_DEBUG_TYPE_PORTABILITY_KHR = 0x824F
GL_DEBUG_TYPE_PERFORMANCE = 0x8250
GL_DEBUG_TYPE_PERFORMANCE_ARB = 0x8250
GL_DEBUG_TYPE_PERFORMANCE_KHR = 0x8250
GL_DEBUG_TYPE_OTHER = 0x8251
GL_DEBUG_TYPE_OTHER_ARB = 0x8251
GL_DEBUG_TYPE_OTHER_KHR = 0x8251
GL_LOSE_CONTEXT_ON_RESET = 0x8252
GL_LOSE_CONTEXT_ON_RESET_ARB = 0x8252
GL_LOSE_CONTEXT_ON_RESET_EXT = 0x8252
GL_LOSE_CONTEXT_ON_RESET_KHR = 0x8252
GL_GUILTY_CONTEXT_RESET = 0x8253
GL_GUILTY_CONTEXT_RESET_ARB = 0x8253
GL_GUILTY_CONTEXT_RESET_EXT = 0x8253
GL_GUILTY_CONTEXT_RESET_KHR = 0x8253
GL_INNOCENT_CONTEXT_RESET = 0x8254
GL_INNOCENT_CONTEXT_RESET_ARB = 0x8254
GL_INNOCENT_CONTEXT_RESET_EXT = 0x8254
GL_INNOCENT_CONTEXT_RESET_KHR = 0x8254
GL_UNKNOWN_CONTEXT_RESET = 0x8255
GL_UNKNOWN_CONTEXT_RESET_ARB = 0x8255
GL_UNKNOWN_CONTEXT_RESET_EXT = 0x8255
GL_UNKNOWN_CONTEXT_RESET_KHR = 0x8255
GL_RESET_NOTIFICATION_STRATEGY = 0x8256
GL_RESET_NOTIFICATION_STRATEGY_ARB = 0x8256
GL_RESET_NOTIFICATION_STRATEGY_EXT = 0x8256
GL_RESET_NOTIFICATION_STRATEGY_KHR = 0x8256
GL_PROGRAM_BINARY_RETRIEVABLE_HINT = 0x8257
GL_PROGRAM_SEPARABLE = 0x8258
GL_PROGRAM_SEPARABLE_EXT = 0x8258
GL_ACTIVE_PROGRAM = 0x8259
GL_ACTIVE_PROGRAM_EXT = 0x8259
GL_PROGRAM_PIPELINE_BINDING = 0x825A
GL_PROGRAM_PIPELINE_BINDING_EXT = 0x825A
GL_MAX_VIEWPORTS = 0x825B
GL_VIEWPORT_SUBPIXEL_BITS = 0x825C
GL_VIEWPORT_BOUNDS_RANGE = 0x825D
GL_LAYER_PROVOKING_VERTEX = 0x825E
GL_LAYER_PROVOKING_VERTEX_EXT = 0x825E
GL_VIEWPORT_INDEX_PROVOKING_VERTEX = 0x825F
GL_UNDEFINED_VERTEX = 0x8260
GL_UNDEFINED_VERTEX_EXT = 0x8260
GL_NO_RESET_NOTIFICATION = 0x8261
GL_NO_RESET_NOTIFICATION_ARB = 0x8261
GL_NO_RESET_NOTIFICATION_EXT = 0x8261
GL_NO_RESET_NOTIFICATION_KHR = 0x8261
GL_MAX_COMPUTE_SHARED_MEMORY_SIZE = 0x8262
GL_MAX_COMPUTE_UNIFORM_COMPONENTS = 0x8263
GL_MAX_COMPUTE_ATOMIC_COUNTER_BUFFERS = 0x8264
GL_MAX_COMPUTE_ATOMIC_COUNTERS = 0x8265
GL_MAX_COMBINED_COMPUTE_UNIFORM_COMPONENTS = 0x8266
GL_COMPUTE_WORK_GROUP_SIZE = 0x8267
GL_DEBUG_TYPE_MARKER = 0x8268
GL_DEBUG_TYPE_MARKER_KHR = 0x8268
GL_DEBUG_TYPE_PUSH_GROUP = 0x8269
GL_DEBUG_TYPE_PUSH_GROUP_KHR = 0x8269
GL_DEBUG_TYPE_POP_GROUP = 0x826A
GL_DEBUG_TYPE_POP_GROUP_KHR = 0x826A
GL_DEBUG_SEVERITY_NOTIFICATION = 0x826B
GL_DEBUG_SEVERITY_NOTIFICATION_KHR = 0x826B
GL_MAX_DEBUG_GROUP_STACK_DEPTH = 0x826C
GL_MAX_DEBUG_GROUP_STACK_DEPTH_KHR = 0x826C
GL_DEBUG_GROUP_STACK_DEPTH = 0x826D
GL_DEBUG_GROUP_STACK_DEPTH_KHR = 0x826D
GL_MAX_UNIFORM_LOCATIONS = 0x826E
GL_INTERNALFORMAT_SUPPORTED = 0x826F
GL_INTERNALFORMAT_PREFERRED = 0x8270
GL_INTERNALFORMAT_RED_SIZE = 0x8271
GL_INTERNALFORMAT_GREEN_SIZE = 0x8272
GL_INTERNALFORMAT_BLUE_SIZE = 0x8273
GL_INTERNALFORMAT_ALPHA_SIZE = 0x8274
GL_INTERNALFORMAT_DEPTH_SIZE = 0x8275
GL_INTERNALFORMAT_STENCIL_SIZE = 0x8276
GL_INTERNALFORMAT_SHARED_SIZE = 0x8277
GL_INTERNALFORMAT_RED_TYPE = 0x8278
GL_INTERNALFORMAT_GREEN_TYPE = 0x8279
GL_INTERNALFORMAT_BLUE_TYPE = 0x827A
GL_INTERNALFORMAT_ALPHA_TYPE = 0x827B
GL_INTERNALFORMAT_DEPTH_TYPE = 0x827C
GL_INTERNALFORMAT_STENCIL_TYPE = 0x827D
GL_MAX_WIDTH = 0x827E
GL_MAX_HEIGHT = 0x827F
GL_MAX_DEPTH = 0x8280
GL_MAX_LAYERS = 0x8281
GL_MAX_COMBINED_DIMENSIONS = 0x8282
GL_COLOR_COMPONENTS = 0x8283
GL_DEPTH_COMPONENTS = 0x8284
GL_STENCIL_COMPONENTS = 0x8285
GL_COLOR_RENDERABLE = 0x8286
GL_DEPTH_RENDERABLE = 0x8287
GL_STENCIL_RENDERABLE = 0x8288
GL_FRAMEBUFFER_RENDERABLE = 0x8289
GL_FRAMEBUFFER_RENDERABLE_LAYERED = 0x828A
GL_FRAMEBUFFER_BLEND = 0x828B
GL_READ_PIXELS = 0x828C
GL_READ_PIXELS_FORMAT = 0x828D
GL_READ_PIXELS_TYPE = 0x828E
GL_TEXTURE_IMAGE_FORMAT = 0x828F
GL_TEXTURE_IMAGE_TYPE = 0x8290
GL_GET_TEXTURE_IMAGE_FORMAT = 0x8291
GL_GET_TEXTURE_IMAGE_TYPE = 0x8292
GL_MIPMAP = 0x8293
GL_MANUAL_GENERATE_MIPMAP = 0x8294
GL_AUTO_GENERATE_MIPMAP = 0x8295
GL_COLOR_ENCODING = 0x8296
GL_SRGB_READ = 0x8297
GL_SRGB_WRITE = 0x8298
GL_SRGB_DECODE_ARB = 0x8299
GL_FILTER = 0x829A
GL_VERTEX_TEXTURE = 0x829B
GL_TESS_CONTROL_TEXTURE = 0x829C
GL_TESS_EVALUATION_TEXTURE = 0x829D
GL_GEOMETRY_TEXTURE = 0x829E
GL_FRAGMENT_TEXTURE = 0x829F
GL_COMPUTE_TEXTURE = 0x82A0
GL_TEXTURE_SHADOW = 0x82A1
GL_TEXTURE_GATHER = 0x82A2
GL_TEXTURE_GATHER_SHADOW = 0x82A3
GL_SHADER_IMAGE_LOAD = 0x82A4
GL_SHADER_IMAGE_STORE = 0x82A5
GL_SHADER_IMAGE_ATOMIC = 0x82A6
GL_IMAGE_TEXEL_SIZE = 0x82A7
GL_IMAGE_COMPATIBILITY_CLASS = 0x82A8
GL_IMAGE_PIXEL_FORMAT = 0x82A9
GL_IMAGE_PIXEL_TYPE = 0x82AA
GL_SIMULTANEOUS_TEXTURE_AND_DEPTH_TEST = 0x82AC
GL_SIMULTANEOUS_TEXTURE_AND_STENCIL_TEST = 0x82AD
GL_SIMULTANEOUS_TEXTURE_AND_DEPTH_WRITE = 0x82AE
GL_SIMULTANEOUS_TEXTURE_AND_STENCIL_WRITE = 0x82AF
GL_TEXTURE_COMPRESSED_BLOCK_WIDTH = 0x82B1
GL_TEXTURE_COMPRESSED_BLOCK_HEIGHT = 0x82B2
GL_TEXTURE_COMPRESSED_BLOCK_SIZE = 0x82B3
GL_CLEAR_BUFFER = 0x82B4
GL_TEXTURE_VIEW = 0x82B5
GL_VIEW_COMPATIBILITY_CLASS = 0x82B6
GL_FULL_SUPPORT = 0x82B7
GL_CAVEAT_SUPPORT = 0x82B8
GL_IMAGE_CLASS_4_X_32 = 0x82B9
GL_IMAGE_CLASS_2_X_32 = 0x82BA
GL_IMAGE_CLASS_1_X_32 = 0x82BB
GL_IMAGE_CLASS_4_X_16 = 0x82BC
GL_IMAGE_CLASS_2_X_16 = 0x82BD
GL_IMAGE_CLASS_1_X_16 = 0x82BE
GL_IMAGE_CLASS_4_X_8 = 0x82BF
GL_IMAGE_CLASS_2_X_8 = 0x82C0
GL_IMAGE_CLASS_1_X_8 = 0x82C1
GL_IMAGE_CLASS_11_11_10 = 0x82C2
GL_IMAGE_CLASS_10_10_10_2 = 0x82C3
GL_VIEW_CLASS_128_BITS = 0x82C4
GL_VIEW_CLASS_96_BITS = 0x82C5
GL_VIEW_CLASS_64_BITS = 0x82C6
GL_VIEW_CLASS_48_BITS = 0x82C7
GL_VIEW_CLASS_32_BITS = 0x82C8
GL_VIEW_CLASS_24_BITS = 0x82C9
GL_VIEW_CLASS_16_BITS = 0x82CA
GL_VIEW_CLASS_8_BITS = 0x82CB
GL_VIEW_CLASS_S3TC_DXT1_RGB = 0x82CC
GL_VIEW_CLASS_S3TC_DXT1_RGBA = 0x82CD
GL_VIEW_CLASS_S3TC_DXT3_RGBA = 0x82CE
GL_VIEW_CLASS_S3TC_DXT5_RGBA = 0x82CF
GL_VIEW_CLASS_RGTC1_RED = 0x82D0
GL_VIEW_CLASS_RGTC2_RG = 0x82D1
GL_VIEW_CLASS_BPTC_UNORM = 0x82D2
GL_VIEW_CLASS_BPTC_FLOAT = 0x82D3
GL_VERTEX_ATTRIB_BINDING = 0x82D4
GL_VERTEX_ATTRIB_RELATIVE_OFFSET = 0x82D5
GL_VERTEX_BINDING_DIVISOR = 0x82D6
GL_VERTEX_BINDING_OFFSET = 0x82D7
GL_VERTEX_BINDING_STRIDE = 0x82D8
GL_MAX_VERTEX_ATTRIB_RELATIVE_OFFSET = 0x82D9
GL_MAX_VERTEX_ATTRIB_BINDINGS = 0x82DA
GL_TEXTURE_VIEW_MIN_LEVEL = 0x82DB
GL_TEXTURE_VIEW_MIN_LEVEL_EXT = 0x82DB
GL_TEXTURE_VIEW_NUM_LEVELS = 0x82DC
GL_TEXTURE_VIEW_NUM_LEVELS_EXT = 0x82DC
GL_TEXTURE_VIEW_MIN_LAYER = 0x82DD
GL_TEXTURE_VIEW_MIN_LAYER_EXT = 0x82DD
GL_TEXTURE_VIEW_NUM_LAYERS = 0x82DE
GL_TEXTURE_VIEW_NUM_LAYERS_EXT = 0x82DE
GL_TEXTURE_IMMUTABLE_LEVELS = 0x82DF
GL_BUFFER = 0x82E0
GL_BUFFER_KHR = 0x82E0
GL_SHADER = 0x82E1
GL_SHADER_KHR = 0x82E1
GL_PROGRAM = 0x82E2
GL_PROGRAM_KHR = 0x82E2
GL_QUERY = 0x82E3
GL_QUERY_KHR = 0x82E3
GL_PROGRAM_PIPELINE = 0x82E4
GL_MAX_VERTEX_ATTRIB_STRIDE = 0x82E5
GL_SAMPLER = 0x82E6
GL_SAMPLER_KHR = 0x82E6
GL_DISPLAY_LIST = 0x82E7
GL_MAX_LABEL_LENGTH = 0x82E8
GL_MAX_LABEL_LENGTH_KHR = 0x82E8
GL_NUM_SHADING_LANGUAGE_VERSIONS = 0x82E9
GL_QUERY_TARGET = 0x82EA
GL_TEXTURE_BINDING = 0x82EB
GL_TRANSFORM_FEEDBACK_OVERFLOW_ARB = 0x82EC
GL_TRANSFORM_FEEDBACK_STREAM_OVERFLOW_ARB = 0x82ED
GL_VERTICES_SUBMITTED_ARB = 0x82EE
GL_PRIMITIVES_SUBMITTED_ARB = 0x82EF
GL_VERTEX_SHADER_INVOCATIONS_ARB = 0x82F0
GL_TESS_CONTROL_SHADER_PATCHES_ARB = 0x82F1
GL_TESS_EVALUATION_SHADER_INVOCATIONS_ARB = 0x82F2
GL_GEOMETRY_SHADER_PRIMITIVES_EMITTED_ARB = 0x82F3
GL_FRAGMENT_SHADER_INVOCATIONS_ARB = 0x82F4
GL_COMPUTE_SHADER_INVOCATIONS_ARB = 0x82F5
GL_CLIPPING_INPUT_PRIMITIVES_ARB = 0x82F6
GL_CLIPPING_OUTPUT_PRIMITIVES_ARB = 0x82F7
GL_SPARSE_BUFFER_PAGE_SIZE_ARB = 0x82F8
GL_MAX_CULL_DISTANCES = 0x82F9
GL_MAX_COMBINED_CLIP_AND_CULL_DISTANCES = 0x82FA
GL_CONTEXT_RELEASE_BEHAVIOR = 0x82FB
GL_CONTEXT_RELEASE_BEHAVIOR_KHR = 0x82FB
GL_CONTEXT_RELEASE_BEHAVIOR_FLUSH = 0x82FC
GL_CONTEXT_RELEASE_BEHAVIOR_FLUSH_KHR = 0x82FC
GL_DEPTH_PASS_INSTRUMENT_SGIX = 0x8310
GL_DEPTH_PASS_INSTRUMENT_COUNTERS_SGIX = 0x8311
GL_DEPTH_PASS_INSTRUMENT_MAX_SGIX = 0x8312
GL_FRAGMENTS_INSTRUMENT_SGIX = 0x8313
GL_FRAGMENTS_INSTRUMENT_COUNTERS_SGIX = 0x8314
GL_FRAGMENTS_INSTRUMENT_MAX_SGIX = 0x8315
GL_CONVOLUTION_HINT_SGIX = 0x8316
GL_YCRCB_SGIX = 0x8318
GL_YCRCBA_SGIX = 0x8319
GL_UNPACK_COMPRESSED_SIZE_SGIX = 0x831A
GL_PACK_MAX_COMPRESSED_SIZE_SGIX = 0x831B
GL_PACK_COMPRESSED_SIZE_SGIX = 0x831C
GL_SLIM8U_SGIX = 0x831D
GL_SLIM10U_SGIX = 0x831E
GL_SLIM12S_SGIX = 0x831F
GL_ALPHA_MIN_SGIX = 0x8320
GL_ALPHA_MAX_SGIX = 0x8321
GL_SCALEBIAS_HINT_SGIX = 0x8322
GL_ASYNC_MARKER_SGIX = 0x8329
GL_PIXEL_TEX_GEN_MODE_SGIX = 0x832B
GL_ASYNC_HISTOGRAM_SGIX = 0x832C
GL_MAX_ASYNC_HISTOGRAM_SGIX = 0x832D
GL_PIXEL_TRANSFORM_2D_EXT = 0x8330
GL_PIXEL_MAG_FILTER_EXT = 0x8331
GL_PIXEL_MIN_FILTER_EXT = 0x8332
GL_PIXEL_CUBIC_WEIGHT_EXT = 0x8333
GL_CUBIC_EXT = 0x8334
GL_AVERAGE_EXT = 0x8335
GL_PIXEL_TRANSFORM_2D_STACK_DEPTH_EXT = 0x8336
GL_MAX_PIXEL_TRANSFORM_2D_STACK_DEPTH_EXT = 0x8337
GL_PIXEL_TRANSFORM_2D_MATRIX_EXT = 0x8338
GL_FRAGMENT_MATERIAL_EXT = 0x8349
GL_FRAGMENT_NORMAL_EXT = 0x834A
GL_FRAGMENT_COLOR_EXT = 0x834C
GL_ATTENUATION_EXT = 0x834D
GL_SHADOW_ATTENUATION_EXT = 0x834E
GL_TEXTURE_APPLICATION_MODE_EXT = 0x834F
GL_TEXTURE_LIGHT_EXT = 0x8350
GL_TEXTURE_MATERIAL_FACE_EXT = 0x8351
GL_TEXTURE_MATERIAL_PARAMETER_EXT = 0x8352
GL_PIXEL_TEXTURE_SGIS = 0x8353
GL_PIXEL_FRAGMENT_RGB_SOURCE_SGIS = 0x8354
GL_PIXEL_FRAGMENT_ALPHA_SOURCE_SGIS = 0x8355
GL_PIXEL_GROUP_COLOR_SGIS = 0x8356
GL_LINE_QUALITY_HINT_SGIX = 0x835B
GL_ASYNC_TEX_IMAGE_SGIX = 0x835C
GL_ASYNC_DRAW_PIXELS_SGIX = 0x835D
GL_ASYNC_READ_PIXELS_SGIX = 0x835E
GL_MAX_ASYNC_TEX_IMAGE_SGIX = 0x835F
GL_MAX_ASYNC_DRAW_PIXELS_SGIX = 0x8360
GL_MAX_ASYNC_READ_PIXELS_SGIX = 0x8361
GL_UNSIGNED_BYTE_2_3_3_REV = 0x8362
GL_UNSIGNED_BYTE_2_3_3_REV_EXT = 0x8362
GL_UNSIGNED_SHORT_5_6_5 = 0x8363
GL_UNSIGNED_SHORT_5_6_5_EXT = 0x8363
GL_UNSIGNED_SHORT_5_6_5_REV = 0x8364
GL_UNSIGNED_SHORT_5_6_5_REV_EXT = 0x8364
GL_UNSIGNED_SHORT_4_4_4_4_REV = 0x8365
GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT = 0x8365
GL_UNSIGNED_SHORT_4_4_4_4_REV_IMG = 0x8365
GL_UNSIGNED_SHORT_1_5_5_5_REV = 0x8366
GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT = 0x8366
GL_UNSIGNED_INT_8_8_8_8_REV = 0x8367
GL_UNSIGNED_INT_8_8_8_8_REV_EXT = 0x8367
GL_UNSIGNED_INT_2_10_10_10_REV = 0x8368
GL_UNSIGNED_INT_2_10_10_10_REV_EXT = 0x8368
GL_TEXTURE_MAX_CLAMP_S_SGIX = 0x8369
GL_TEXTURE_MAX_CLAMP_T_SGIX = 0x836A
GL_TEXTURE_MAX_CLAMP_R_SGIX = 0x836B
GL_MIRRORED_REPEAT = 0x8370
GL_MIRRORED_REPEAT_ARB = 0x8370
GL_MIRRORED_REPEAT_IBM = 0x8370
GL_MIRRORED_REPEAT_OES = 0x8370
GL_RGB_S3TC = 0x83A0
GL_RGB4_S3TC = 0x83A1
GL_RGBA_S3TC = 0x83A2
GL_RGBA4_S3TC = 0x83A3
GL_RGBA_DXT5_S3TC = 0x83A4
GL_RGBA4_DXT5_S3TC = 0x83A5
GL_VERTEX_PRECLIP_SGIX = 0x83EE
GL_VERTEX_PRECLIP_HINT_SGIX = 0x83EF
GL_COMPRESSED_RGB_S3TC_DXT1_EXT = 0x83F0
GL_COMPRESSED_RGBA_S3TC_DXT1_EXT = 0x83F1
GL_COMPRESSED_RGBA_S3TC_DXT3_ANGLE = 0x83F2
GL_COMPRESSED_RGBA_S3TC_DXT3_EXT = 0x83F2
GL_COMPRESSED_RGBA_S3TC_DXT5_ANGLE = 0x83F3
GL_COMPRESSED_RGBA_S3TC_DXT5_EXT = 0x83F3
GL_PARALLEL_ARRAYS_INTEL = 0x83F4
GL_VERTEX_ARRAY_PARALLEL_POINTERS_INTEL = 0x83F5
GL_NORMAL_ARRAY_PARALLEL_POINTERS_INTEL = 0x83F6
GL_COLOR_ARRAY_PARALLEL_POINTERS_INTEL = 0x83F7
GL_TEXTURE_COORD_ARRAY_PARALLEL_POINTERS_INTEL = 0x83F8
GL_PERFQUERY_DONOT_FLUSH_INTEL = 0x83F9
GL_PERFQUERY_FLUSH_INTEL = 0x83FA
GL_PERFQUERY_WAIT_INTEL = 0x83FB
GL_TEXTURE_MEMORY_LAYOUT_INTEL = 0x83FF
GL_FRAGMENT_LIGHTING_SGIX = 0x8400
GL_FRAGMENT_COLOR_MATERIAL_SGIX = 0x8401
GL_FRAGMENT_COLOR_MATERIAL_FACE_SGIX = 0x8402
GL_FRAGMENT_COLOR_MATERIAL_PARAMETER_SGIX = 0x8403
GL_MAX_FRAGMENT_LIGHTS_SGIX = 0x8404
GL_MAX_ACTIVE_LIGHTS_SGIX = 0x8405
GL_CURRENT_RASTER_NORMAL_SGIX = 0x8406
GL_LIGHT_ENV_MODE_SGIX = 0x8407
GL_FRAGMENT_LIGHT_MODEL_LOCAL_VIEWER_SGIX = 0x8408
GL_FRAGMENT_LIGHT_MODEL_TWO_SIDE_SGIX = 0x8409
GL_FRAGMENT_LIGHT_MODEL_AMBIENT_SGIX = 0x840A
GL_FRAGMENT_LIGHT_MODEL_NORMAL_INTERPOLATION_SGIX = 0x840B
GL_FRAGMENT_LIGHT0_SGIX = 0x840C
GL_FRAGMENT_LIGHT1_SGIX = 0x840D
GL_FRAGMENT_LIGHT2_SGIX = 0x840E
GL_FRAGMENT_LIGHT3_SGIX = 0x840F
GL_FRAGMENT_LIGHT4_SGIX = 0x8410
GL_FRAGMENT_LIGHT5_SGIX = 0x8411
GL_FRAGMENT_LIGHT6_SGIX = 0x8412
GL_FRAGMENT_LIGHT7_SGIX = 0x8413
GL_PACK_RESAMPLE_SGIX = 0x842C
GL_UNPACK_RESAMPLE_SGIX = 0x842D
GL_RESAMPLE_REPLICATE_SGIX = 0x842E
GL_RESAMPLE_ZERO_FILL_SGIX = 0x842F
GL_RESAMPLE_DECIMATE_SGIX = 0x8430
GL_TANGENT_ARRAY_EXT = 0x8439
GL_BINORMAL_ARRAY_EXT = 0x843A
GL_CURRENT_TANGENT_EXT = 0x843B
GL_CURRENT_BINORMAL_EXT = 0x843C
GL_TANGENT_ARRAY_TYPE_EXT = 0x843E
GL_TANGENT_ARRAY_STRIDE_EXT = 0x843F
GL_BINORMAL_ARRAY_TYPE_EXT = 0x8440
GL_BINORMAL_ARRAY_STRIDE_EXT = 0x8441
GL_TANGENT_ARRAY_POINTER_EXT = 0x8442
GL_BINORMAL_ARRAY_POINTER_EXT = 0x8443
GL_MAP1_TANGENT_EXT = 0x8444
GL_MAP2_TANGENT_EXT = 0x8445
GL_MAP1_BINORMAL_EXT = 0x8446
GL_MAP2_BINORMAL_EXT = 0x8447
GL_NEAREST_CLIPMAP_NEAREST_SGIX = 0x844D
GL_NEAREST_CLIPMAP_LINEAR_SGIX = 0x844E
GL_LINEAR_CLIPMAP_NEAREST_SGIX = 0x844F
GL_FOG_COORDINATE_SOURCE = 0x8450
GL_FOG_COORDINATE_SOURCE_EXT = 0x8450
GL_FOG_COORD_SRC = 0x8450
GL_FOG_COORDINATE = 0x8451
GL_FOG_COORD = 0x8451
GL_FOG_COORDINATE_EXT = 0x8451
GL_FRAGMENT_DEPTH = 0x8452
GL_FRAGMENT_DEPTH_EXT = 0x8452
GL_CURRENT_FOG_COORDINATE = 0x8453
GL_CURRENT_FOG_COORD = 0x8453
GL_CURRENT_FOG_COORDINATE_EXT = 0x8453
GL_FOG_COORDINATE_ARRAY_TYPE = 0x8454
GL_FOG_COORDINATE_ARRAY_TYPE_EXT = 0x8454
GL_FOG_COORD_ARRAY_TYPE = 0x8454
GL_FOG_COORDINATE_ARRAY_STRIDE = 0x8455
GL_FOG_COORDINATE_ARRAY_STRIDE_EXT = 0x8455
GL_FOG_COORD_ARRAY_STRIDE = 0x8455
GL_FOG_COORDINATE_ARRAY_POINTER = 0x8456
GL_FOG_COORDINATE_ARRAY_POINTER_EXT = 0x8456
GL_FOG_COORD_ARRAY_POINTER = 0x8456
GL_FOG_COORDINATE_ARRAY = 0x8457
GL_FOG_COORDINATE_ARRAY_EXT = 0x8457
GL_FOG_COORD_ARRAY = 0x8457
GL_COLOR_SUM = 0x8458
GL_COLOR_SUM_ARB = 0x8458
GL_COLOR_SUM_EXT = 0x8458
GL_CURRENT_SECONDARY_COLOR = 0x8459
GL_CURRENT_SECONDARY_COLOR_EXT = 0x8459
GL_SECONDARY_COLOR_ARRAY_SIZE = 0x845A
GL_SECONDARY_COLOR_ARRAY_SIZE_EXT = 0x845A
GL_SECONDARY_COLOR_ARRAY_TYPE = 0x845B
GL_SECONDARY_COLOR_ARRAY_TYPE_EXT = 0x845B
GL_SECONDARY_COLOR_ARRAY_STRIDE = 0x845C
GL_SECONDARY_COLOR_ARRAY_STRIDE_EXT = 0x845C
GL_SECONDARY_COLOR_ARRAY_POINTER = 0x845D
GL_SECONDARY_COLOR_ARRAY_POINTER_EXT = 0x845D
GL_SECONDARY_COLOR_ARRAY = 0x845E
GL_SECONDARY_COLOR_ARRAY_EXT = 0x845E
GL_CURRENT_RASTER_SECONDARY_COLOR = 0x845F
GL_ALIASED_POINT_SIZE_RANGE = 0x846D
GL_ALIASED_LINE_WIDTH_RANGE = 0x846E
GL_SCREEN_COORDINATES_REND = 0x8490
GL_INVERTED_SCREEN_W_REND = 0x8491
GL_TEXTURE0 = 0x84C0
GL_TEXTURE0_ARB = 0x84C0
GL_TEXTURE1 = 0x84C1
GL_TEXTURE1_ARB = 0x84C1
GL_TEXTURE2 = 0x84C2
GL_TEXTURE2_ARB = 0x84C2
GL_TEXTURE3 = 0x84C3
GL_TEXTURE3_ARB = 0x84C3
GL_TEXTURE4 = 0x84C4
GL_TEXTURE4_ARB = 0x84C4
GL_TEXTURE5 = 0x84C5
GL_TEXTURE5_ARB = 0x84C5
GL_TEXTURE6 = 0x84C6
GL_TEXTURE6_ARB = 0x84C6
GL_TEXTURE7 = 0x84C7
GL_TEXTURE7_ARB = 0x84C7
GL_TEXTURE8 = 0x84C8
GL_TEXTURE8_ARB = 0x84C8
GL_TEXTURE9 = 0x84C9
GL_TEXTURE9_ARB = 0x84C9
GL_TEXTURE10 = 0x84CA
GL_TEXTURE10_ARB = 0x84CA
GL_TEXTURE11 = 0x84CB
GL_TEXTURE11_ARB = 0x84CB
GL_TEXTURE12 = 0x84CC
GL_TEXTURE12_ARB = 0x84CC
GL_TEXTURE13 = 0x84CD
GL_TEXTURE13_ARB = 0x84CD
GL_TEXTURE14 = 0x84CE
GL_TEXTURE14_ARB = 0x84CE
GL_TEXTURE15 = 0x84CF
GL_TEXTURE15_ARB = 0x84CF
GL_TEXTURE16 = 0x84D0
GL_TEXTURE16_ARB = 0x84D0
GL_TEXTURE17 = 0x84D1
GL_TEXTURE17_ARB = 0x84D1
GL_TEXTURE18 = 0x84D2
GL_TEXTURE18_ARB = 0x84D2
GL_TEXTURE19 = 0x84D3
GL_TEXTURE19_ARB = 0x84D3
GL_TEXTURE20 = 0x84D4
GL_TEXTURE20_ARB = 0x84D4
GL_TEXTURE21 = 0x84D5
GL_TEXTURE21_ARB = 0x84D5
GL_TEXTURE22 = 0x84D6
GL_TEXTURE22_ARB = 0x84D6
GL_TEXTURE23 = 0x84D7
GL_TEXTURE23_ARB = 0x84D7
GL_TEXTURE24 = 0x84D8
GL_TEXTURE24_ARB = 0x84D8
GL_TEXTURE25 = 0x84D9
GL_TEXTURE25_ARB = 0x84D9
GL_TEXTURE26 = 0x84DA
GL_TEXTURE26_ARB = 0x84DA
GL_TEXTURE27 = 0x84DB
GL_TEXTURE27_ARB = 0x84DB
GL_TEXTURE28 = 0x84DC
GL_TEXTURE28_ARB = 0x84DC
GL_TEXTURE29 = 0x84DD
GL_TEXTURE29_ARB = 0x84DD
GL_TEXTURE30 = 0x84DE
GL_TEXTURE30_ARB = 0x84DE
GL_TEXTURE31 = 0x84DF
GL_TEXTURE31_ARB = 0x84DF
GL_ACTIVE_TEXTURE = 0x84E0
GL_ACTIVE_TEXTURE_ARB = 0x84E0
GL_CLIENT_ACTIVE_TEXTURE = 0x84E1
GL_CLIENT_ACTIVE_TEXTURE_ARB = 0x84E1
GL_MAX_TEXTURE_UNITS = 0x84E2
GL_MAX_TEXTURE_UNITS_ARB = 0x84E2
GL_TRANSPOSE_MODELVIEW_MATRIX = 0x84E3
GL_TRANSPOSE_MODELVIEW_MATRIX_ARB = 0x84E3
GL_PATH_TRANSPOSE_MODELVIEW_MATRIX_NV = 0x84E3
GL_TRANSPOSE_PROJECTION_MATRIX = 0x84E4
GL_TRANSPOSE_PROJECTION_MATRIX_ARB = 0x84E4
GL_PATH_TRANSPOSE_PROJECTION_MATRIX_NV = 0x84E4
GL_TRANSPOSE_TEXTURE_MATRIX = 0x84E5
GL_TRANSPOSE_TEXTURE_MATRIX_ARB = 0x84E5
GL_TRANSPOSE_COLOR_MATRIX = 0x84E6
GL_TRANSPOSE_COLOR_MATRIX_ARB = 0x84E6
GL_SUBTRACT = 0x84E7
GL_SUBTRACT_ARB = 0x84E7
GL_MAX_RENDERBUFFER_SIZE = 0x84E8
GL_MAX_RENDERBUFFER_SIZE_EXT = 0x84E8
GL_MAX_RENDERBUFFER_SIZE_OES = 0x84E8
GL_COMPRESSED_ALPHA = 0x84E9
GL_COMPRESSED_ALPHA_ARB = 0x84E9
GL_COMPRESSED_LUMINANCE = 0x84EA
GL_COMPRESSED_LUMINANCE_ARB = 0x84EA
GL_COMPRESSED_LUMINANCE_ALPHA = 0x84EB
GL_COMPRESSED_LUMINANCE_ALPHA_ARB = 0x84EB
GL_COMPRESSED_INTENSITY = 0x84EC
GL_COMPRESSED_INTENSITY_ARB = 0x84EC
GL_COMPRESSED_RGB = 0x84ED
GL_COMPRESSED_RGB_ARB = 0x84ED
GL_COMPRESSED_RGBA = 0x84EE
GL_COMPRESSED_RGBA_ARB = 0x84EE
GL_TEXTURE_COMPRESSION_HINT = 0x84EF
GL_TEXTURE_COMPRESSION_HINT_ARB = 0x84EF
GL_UNIFORM_BLOCK_REFERENCED_BY_TESS_CONTROL_SHADER = 0x84F0
GL_UNIFORM_BLOCK_REFERENCED_BY_TESS_EVALUATION_SHADER = 0x84F1
GL_ALL_COMPLETED_NV = 0x84F2
GL_FENCE_STATUS_NV = 0x84F3
GL_FENCE_CONDITION_NV = 0x84F4
GL_TEXTURE_RECTANGLE = 0x84F5
GL_TEXTURE_RECTANGLE_ARB = 0x84F5
GL_TEXTURE_RECTANGLE_NV = 0x84F5
GL_TEXTURE_BINDING_RECTANGLE = 0x84F6
GL_TEXTURE_BINDING_RECTANGLE_ARB = 0x84F6
GL_TEXTURE_BINDING_RECTANGLE_NV = 0x84F6
GL_PROXY_TEXTURE_RECTANGLE = 0x84F7
GL_PROXY_TEXTURE_RECTANGLE_ARB = 0x84F7
GL_PROXY_TEXTURE_RECTANGLE_NV = 0x84F7
GL_MAX_RECTANGLE_TEXTURE_SIZE = 0x84F8
GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB = 0x84F8
GL_MAX_RECTANGLE_TEXTURE_SIZE_NV = 0x84F8
GL_DEPTH_STENCIL = 0x84F9
GL_DEPTH_STENCIL_EXT = 0x84F9
GL_DEPTH_STENCIL_NV = 0x84F9
GL_DEPTH_STENCIL_OES = 0x84F9
GL_UNSIGNED_INT_24_8 = 0x84FA
GL_UNSIGNED_INT_24_8_EXT = 0x84FA
GL_UNSIGNED_INT_24_8_NV = 0x84FA
GL_UNSIGNED_INT_24_8_OES = 0x84FA
GL_MAX_TEXTURE_LOD_BIAS = 0x84FD
GL_MAX_TEXTURE_LOD_BIAS_EXT = 0x84FD
GL_TEXTURE_MAX_ANISOTROPY_EXT = 0x84FE
GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT = 0x84FF
GL_TEXTURE_FILTER_CONTROL = 0x8500
GL_TEXTURE_FILTER_CONTROL_EXT = 0x8500
GL_TEXTURE_LOD_BIAS = 0x8501
GL_TEXTURE_LOD_BIAS_EXT = 0x8501
GL_MODELVIEW1_STACK_DEPTH_EXT = 0x8502
GL_COMBINE4_NV = 0x8503
GL_MAX_SHININESS_NV = 0x8504
GL_MAX_SPOT_EXPONENT_NV = 0x8505
GL_MODELVIEW1_MATRIX_EXT = 0x8506
GL_INCR_WRAP = 0x8507
GL_INCR_WRAP_EXT = 0x8507
GL_INCR_WRAP_OES = 0x8507
GL_DECR_WRAP = 0x8508
GL_DECR_WRAP_EXT = 0x8508
GL_DECR_WRAP_OES = 0x8508
GL_VERTEX_WEIGHTING_EXT = 0x8509
GL_MODELVIEW1_ARB = 0x850A
GL_MODELVIEW1_EXT = 0x850A
GL_CURRENT_VERTEX_WEIGHT_EXT = 0x850B
GL_VERTEX_WEIGHT_ARRAY_EXT = 0x850C
GL_VERTEX_WEIGHT_ARRAY_SIZE_EXT = 0x850D
GL_VERTEX_WEIGHT_ARRAY_TYPE_EXT = 0x850E
GL_VERTEX_WEIGHT_ARRAY_STRIDE_EXT = 0x850F
GL_VERTEX_WEIGHT_ARRAY_POINTER_EXT = 0x8510
GL_NORMAL_MAP = 0x8511
GL_NORMAL_MAP_ARB = 0x8511
GL_NORMAL_MAP_EXT = 0x8511
GL_NORMAL_MAP_NV = 0x8511
GL_NORMAL_MAP_OES = 0x8511
GL_REFLECTION_MAP = 0x8512
GL_REFLECTION_MAP_ARB = 0x8512
GL_REFLECTION_MAP_EXT = 0x8512
GL_REFLECTION_MAP_NV = 0x8512
GL_REFLECTION_MAP_OES = 0x8512
GL_TEXTURE_CUBE_MAP = 0x8513
GL_TEXTURE_CUBE_MAP_ARB = 0x8513
GL_TEXTURE_CUBE_MAP_EXT = 0x8513
GL_TEXTURE_CUBE_MAP_OES = 0x8513
GL_TEXTURE_BINDING_CUBE_MAP = 0x8514
GL_TEXTURE_BINDING_CUBE_MAP_ARB = 0x8514
GL_TEXTURE_BINDING_CUBE_MAP_EXT = 0x8514
GL_TEXTURE_BINDING_CUBE_MAP_OES = 0x8514
GL_TEXTURE_CUBE_MAP_POSITIVE_X = 0x8515
GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB = 0x8515
GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT = 0x8515
GL_TEXTURE_CUBE_MAP_POSITIVE_X_OES = 0x8515
GL_TEXTURE_CUBE_MAP_NEGATIVE_X = 0x8516
GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB = 0x8516
GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT = 0x8516
GL_TEXTURE_CUBE_MAP_NEGATIVE_X_OES = 0x8516
GL_TEXTURE_CUBE_MAP_POSITIVE_Y = 0x8517
GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB = 0x8517
GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT = 0x8517
GL_TEXTURE_CUBE_MAP_POSITIVE_Y_OES = 0x8517
GL_TEXTURE_CUBE_MAP_NEGATIVE_Y = 0x8518
GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB = 0x8518
GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT = 0x8518
GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_OES = 0x8518
GL_TEXTURE_CUBE_MAP_POSITIVE_Z = 0x8519
GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB = 0x8519
GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT = 0x8519
GL_TEXTURE_CUBE_MAP_POSITIVE_Z_OES = 0x8519
GL_TEXTURE_CUBE_MAP_NEGATIVE_Z = 0x851A
GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB = 0x851A
GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT = 0x851A
GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_OES = 0x851A
GL_PROXY_TEXTURE_CUBE_MAP = 0x851B
GL_PROXY_TEXTURE_CUBE_MAP_ARB = 0x851B
GL_PROXY_TEXTURE_CUBE_MAP_EXT = 0x851B
GL_MAX_CUBE_MAP_TEXTURE_SIZE = 0x851C
GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB = 0x851C
GL_MAX_CUBE_MAP_TEXTURE_SIZE_EXT = 0x851C
GL_MAX_CUBE_MAP_TEXTURE_SIZE_OES = 0x851C
GL_VERTEX_ARRAY_RANGE_APPLE = 0x851D
GL_VERTEX_ARRAY_RANGE_NV = 0x851D
GL_VERTEX_ARRAY_RANGE_LENGTH_APPLE = 0x851E
GL_VERTEX_ARRAY_RANGE_LENGTH_NV = 0x851E
GL_VERTEX_ARRAY_RANGE_VALID_NV = 0x851F
GL_VERTEX_ARRAY_STORAGE_HINT_APPLE = 0x851F
GL_MAX_VERTEX_ARRAY_RANGE_ELEMENT_NV = 0x8520
GL_VERTEX_ARRAY_RANGE_POINTER_APPLE = 0x8521
GL_VERTEX_ARRAY_RANGE_POINTER_NV = 0x8521
GL_REGISTER_COMBINERS_NV = 0x8522
GL_VARIABLE_A_NV = 0x8523
GL_VARIABLE_B_NV = 0x8524
GL_VARIABLE_C_NV = 0x8525
GL_VARIABLE_D_NV = 0x8526
GL_VARIABLE_E_NV = 0x8527
GL_VARIABLE_F_NV = 0x8528
GL_VARIABLE_G_NV = 0x8529
GL_CONSTANT_COLOR0_NV = 0x852A
GL_CONSTANT_COLOR1_NV = 0x852B
GL_PRIMARY_COLOR_NV = 0x852C
GL_SECONDARY_COLOR_NV = 0x852D
GL_SPARE0_NV = 0x852E
GL_SPARE1_NV = 0x852F
GL_DISCARD_NV = 0x8530
GL_E_TIMES_F_NV = 0x8531
GL_SPARE0_PLUS_SECONDARY_COLOR_NV = 0x8532
GL_VERTEX_ARRAY_RANGE_WITHOUT_FLUSH_NV = 0x8533
GL_MULTISAMPLE_FILTER_HINT_NV = 0x8534
GL_PER_STAGE_CONSTANTS_NV = 0x8535
GL_UNSIGNED_IDENTITY_NV = 0x8536
GL_UNSIGNED_INVERT_NV = 0x8537
GL_EXPAND_NORMAL_NV = 0x8538
GL_EXPAND_NEGATE_NV = 0x8539
GL_HALF_BIAS_NORMAL_NV = 0x853A
GL_HALF_BIAS_NEGATE_NV = 0x853B
GL_SIGNED_IDENTITY_NV = 0x853C
GL_SIGNED_NEGATE_NV = 0x853D
GL_SCALE_BY_TWO_NV = 0x853E
GL_SCALE_BY_FOUR_NV = 0x853F
GL_SCALE_BY_ONE_HALF_NV = 0x8540
GL_BIAS_BY_NEGATIVE_ONE_HALF_NV = 0x8541
GL_COMBINER_INPUT_NV = 0x8542
GL_COMBINER_MAPPING_NV = 0x8543
GL_COMBINER_COMPONENT_USAGE_NV = 0x8544
GL_COMBINER_AB_DOT_PRODUCT_NV = 0x8545
GL_COMBINER_CD_DOT_PRODUCT_NV = 0x8546
GL_COMBINER_MUX_SUM_NV = 0x8547
GL_COMBINER_SCALE_NV = 0x8548
GL_COMBINER_BIAS_NV = 0x8549
GL_COMBINER_AB_OUTPUT_NV = 0x854A
GL_COMBINER_CD_OUTPUT_NV = 0x854B
GL_COMBINER_SUM_OUTPUT_NV = 0x854C
GL_MAX_GENERAL_COMBINERS_NV = 0x854D
GL_NUM_GENERAL_COMBINERS_NV = 0x854E
GL_COLOR_SUM_CLAMP_NV = 0x854F
GL_COMBINER0_NV = 0x8550
GL_COMBINER1_NV = 0x8551
GL_COMBINER2_NV = 0x8552
GL_COMBINER3_NV = 0x8553
GL_COMBINER4_NV = 0x8554
GL_COMBINER5_NV = 0x8555
GL_COMBINER6_NV = 0x8556
GL_COMBINER7_NV = 0x8557
GL_PRIMITIVE_RESTART_NV = 0x8558
GL_PRIMITIVE_RESTART_INDEX_NV = 0x8559
GL_FOG_DISTANCE_MODE_NV = 0x855A
GL_EYE_RADIAL_NV = 0x855B
GL_EYE_PLANE_ABSOLUTE_NV = 0x855C
GL_EMBOSS_LIGHT_NV = 0x855D
GL_EMBOSS_CONSTANT_NV = 0x855E
GL_EMBOSS_MAP_NV = 0x855F
GL_RED_MIN_CLAMP_INGR = 0x8560
GL_GREEN_MIN_CLAMP_INGR = 0x8561
GL_BLUE_MIN_CLAMP_INGR = 0x8562
GL_ALPHA_MIN_CLAMP_INGR = 0x8563
GL_RED_MAX_CLAMP_INGR = 0x8564
GL_GREEN_MAX_CLAMP_INGR = 0x8565
GL_BLUE_MAX_CLAMP_INGR = 0x8566
GL_ALPHA_MAX_CLAMP_INGR = 0x8567
GL_INTERLACE_READ_INGR = 0x8568
GL_COMBINE = 0x8570
GL_COMBINE_ARB = 0x8570
GL_COMBINE_EXT = 0x8570
GL_COMBINE_RGB = 0x8571
GL_COMBINE_RGB_ARB = 0x8571
GL_COMBINE_RGB_EXT = 0x8571
GL_COMBINE_ALPHA = 0x8572
GL_COMBINE_ALPHA_ARB = 0x8572
GL_COMBINE_ALPHA_EXT = 0x8572
GL_RGB_SCALE = 0x8573
GL_RGB_SCALE_ARB = 0x8573
GL_RGB_SCALE_EXT = 0x8573
GL_ADD_SIGNED = 0x8574
GL_ADD_SIGNED_ARB = 0x8574
GL_ADD_SIGNED_EXT = 0x8574
GL_INTERPOLATE = 0x8575
GL_INTERPOLATE_ARB = 0x8575
GL_INTERPOLATE_EXT = 0x8575
GL_CONSTANT = 0x8576
GL_CONSTANT_ARB = 0x8576
GL_CONSTANT_EXT = 0x8576
GL_CONSTANT_NV = 0x8576
GL_PRIMARY_COLOR = 0x8577
GL_PRIMARY_COLOR_ARB = 0x8577
GL_PRIMARY_COLOR_EXT = 0x8577
GL_PREVIOUS = 0x8578
GL_PREVIOUS_ARB = 0x8578
GL_PREVIOUS_EXT = 0x8578
GL_SOURCE0_RGB = 0x8580
GL_SOURCE0_RGB_ARB = 0x8580
GL_SOURCE0_RGB_EXT = 0x8580
GL_SRC0_RGB = 0x8580
GL_SOURCE1_RGB = 0x8581
GL_SOURCE1_RGB_ARB = 0x8581
GL_SOURCE1_RGB_EXT = 0x8581
GL_SRC1_RGB = 0x8581
GL_SOURCE2_RGB = 0x8582
GL_SOURCE2_RGB_ARB = 0x8582
GL_SOURCE2_RGB_EXT = 0x8582
GL_SRC2_RGB = 0x8582
GL_SOURCE3_RGB_NV = 0x8583
GL_SOURCE0_ALPHA = 0x8588
GL_SOURCE0_ALPHA_ARB = 0x8588
GL_SOURCE0_ALPHA_EXT = 0x8588
GL_SRC0_ALPHA = 0x8588
GL_SOURCE1_ALPHA = 0x8589
GL_SOURCE1_ALPHA_ARB = 0x8589
GL_SOURCE1_ALPHA_EXT = 0x8589
GL_SRC1_ALPHA = 0x8589
GL_SOURCE2_ALPHA = 0x858A
GL_SOURCE2_ALPHA_ARB = 0x858A
GL_SOURCE2_ALPHA_EXT = 0x858A
GL_SRC2_ALPHA = 0x858A
GL_SOURCE3_ALPHA_NV = 0x858B
GL_OPERAND0_RGB = 0x8590
GL_OPERAND0_RGB_ARB = 0x8590
GL_OPERAND0_RGB_EXT = 0x8590
GL_OPERAND1_RGB = 0x8591
GL_OPERAND1_RGB_ARB = 0x8591
GL_OPERAND1_RGB_EXT = 0x8591
GL_OPERAND2_RGB = 0x8592
GL_OPERAND2_RGB_ARB = 0x8592
GL_OPERAND2_RGB_EXT = 0x8592
GL_OPERAND3_RGB_NV = 0x8593
GL_OPERAND0_ALPHA = 0x8598
GL_OPERAND0_ALPHA_ARB = 0x8598
GL_OPERAND0_ALPHA_EXT = 0x8598
GL_OPERAND1_ALPHA = 0x8599
GL_OPERAND1_ALPHA_ARB = 0x8599
GL_OPERAND1_ALPHA_EXT = 0x8599
GL_OPERAND2_ALPHA = 0x859A
GL_OPERAND2_ALPHA_ARB = 0x859A
GL_OPERAND2_ALPHA_EXT = 0x859A
GL_OPERAND3_ALPHA_NV = 0x859B
GL_PACK_SUBSAMPLE_RATE_SGIX = 0x85A0
GL_UNPACK_SUBSAMPLE_RATE_SGIX = 0x85A1
GL_PIXEL_SUBSAMPLE_4444_SGIX = 0x85A2
GL_PIXEL_SUBSAMPLE_2424_SGIX = 0x85A3
GL_PIXEL_SUBSAMPLE_4242_SGIX = 0x85A4
GL_PERTURB_EXT = 0x85AE
GL_TEXTURE_NORMAL_EXT = 0x85AF
GL_LIGHT_MODEL_SPECULAR_VECTOR_APPLE = 0x85B0
GL_TRANSFORM_HINT_APPLE = 0x85B1
GL_UNPACK_CLIENT_STORAGE_APPLE = 0x85B2
GL_BUFFER_OBJECT_APPLE = 0x85B3
GL_STORAGE_CLIENT_APPLE = 0x85B4
GL_VERTEX_ARRAY_BINDING = 0x85B5
GL_VERTEX_ARRAY_BINDING_APPLE = 0x85B5
GL_VERTEX_ARRAY_BINDING_OES = 0x85B5
GL_TEXTURE_RANGE_LENGTH_APPLE = 0x85B7
GL_TEXTURE_RANGE_POINTER_APPLE = 0x85B8
GL_YCBCR_422_APPLE = 0x85B9
GL_UNSIGNED_SHORT_8_8_APPLE = 0x85BA
GL_UNSIGNED_SHORT_8_8_MESA = 0x85BA
GL_UNSIGNED_SHORT_8_8_REV_APPLE = 0x85BB
GL_UNSIGNED_SHORT_8_8_REV_MESA = 0x85BB
GL_TEXTURE_STORAGE_HINT_APPLE = 0x85BC
GL_STORAGE_PRIVATE_APPLE = 0x85BD
GL_STORAGE_CACHED_APPLE = 0x85BE
GL_STORAGE_SHARED_APPLE = 0x85BF
GL_REPLACEMENT_CODE_ARRAY_SUN = 0x85C0
GL_REPLACEMENT_CODE_ARRAY_TYPE_SUN = 0x85C1
GL_REPLACEMENT_CODE_ARRAY_STRIDE_SUN = 0x85C2
GL_REPLACEMENT_CODE_ARRAY_POINTER_SUN = 0x85C3
GL_R1UI_V3F_SUN = 0x85C4
GL_R1UI_C4UB_V3F_SUN = 0x85C5
GL_R1UI_C3F_V3F_SUN = 0x85C6
GL_R1UI_N3F_V3F_SUN = 0x85C7
GL_R1UI_C4F_N3F_V3F_SUN = 0x85C8
GL_R1UI_T2F_V3F_SUN = 0x85C9
GL_R1UI_T2F_N3F_V3F_SUN = 0x85CA
GL_R1UI_T2F_C4F_N3F_V3F_SUN = 0x85CB
GL_SLICE_ACCUM_SUN = 0x85CC
GL_QUAD_MESH_SUN = 0x8614
GL_TRIANGLE_MESH_SUN = 0x8615
GL_VERTEX_PROGRAM_ARB = 0x8620
GL_VERTEX_PROGRAM_NV = 0x8620
GL_VERTEX_STATE_PROGRAM_NV = 0x8621
GL_VERTEX_ATTRIB_ARRAY_ENABLED = 0x8622
GL_VERTEX_ATTRIB_ARRAY_ENABLED_ARB = 0x8622
GL_ATTRIB_ARRAY_SIZE_NV = 0x8623
GL_VERTEX_ATTRIB_ARRAY_SIZE = 0x8623
GL_VERTEX_ATTRIB_ARRAY_SIZE_ARB = 0x8623
GL_ATTRIB_ARRAY_STRIDE_NV = 0x8624
GL_VERTEX_ATTRIB_ARRAY_STRIDE = 0x8624
GL_VERTEX_ATTRIB_ARRAY_STRIDE_ARB = 0x8624
GL_ATTRIB_ARRAY_TYPE_NV = 0x8625
GL_VERTEX_ATTRIB_ARRAY_TYPE = 0x8625
GL_VERTEX_ATTRIB_ARRAY_TYPE_ARB = 0x8625
GL_CURRENT_ATTRIB_NV = 0x8626
GL_CURRENT_VERTEX_ATTRIB = 0x8626
GL_CURRENT_VERTEX_ATTRIB_ARB = 0x8626
GL_PROGRAM_LENGTH_ARB = 0x8627
GL_PROGRAM_LENGTH_NV = 0x8627
GL_PROGRAM_STRING_ARB = 0x8628
GL_PROGRAM_STRING_NV = 0x8628
GL_MODELVIEW_PROJECTION_NV = 0x8629
GL_IDENTITY_NV = 0x862A
GL_INVERSE_NV = 0x862B
GL_TRANSPOSE_NV = 0x862C
GL_INVERSE_TRANSPOSE_NV = 0x862D
GL_MAX_PROGRAM_MATRIX_STACK_DEPTH_ARB = 0x862E
GL_MAX_TRACK_MATRIX_STACK_DEPTH_NV = 0x862E
GL_MAX_PROGRAM_MATRICES_ARB = 0x862F
GL_MAX_TRACK_MATRICES_NV = 0x862F
GL_MATRIX0_NV = 0x8630
GL_MATRIX1_NV = 0x8631
GL_MATRIX2_NV = 0x8632
GL_MATRIX3_NV = 0x8633
GL_MATRIX4_NV = 0x8634
GL_MATRIX5_NV = 0x8635
GL_MATRIX6_NV = 0x8636
GL_MATRIX7_NV = 0x8637
GL_CURRENT_MATRIX_STACK_DEPTH_ARB = 0x8640
GL_CURRENT_MATRIX_STACK_DEPTH_NV = 0x8640
GL_CURRENT_MATRIX_ARB = 0x8641
GL_CURRENT_MATRIX_NV = 0x8641
GL_VERTEX_PROGRAM_POINT_SIZE = 0x8642
GL_VERTEX_PROGRAM_POINT_SIZE_ARB = 0x8642
GL_VERTEX_PROGRAM_POINT_SIZE_NV = 0x8642
GL_PROGRAM_POINT_SIZE = 0x8642
GL_PROGRAM_POINT_SIZE_ARB = 0x8642
GL_PROGRAM_POINT_SIZE_EXT = 0x8642
GL_VERTEX_PROGRAM_TWO_SIDE = 0x8643
GL_VERTEX_PROGRAM_TWO_SIDE_ARB = 0x8643
GL_VERTEX_PROGRAM_TWO_SIDE_NV = 0x8643
GL_PROGRAM_PARAMETER_NV = 0x8644
GL_ATTRIB_ARRAY_POINTER_NV = 0x8645
GL_VERTEX_ATTRIB_ARRAY_POINTER = 0x8645
GL_VERTEX_ATTRIB_ARRAY_POINTER_ARB = 0x8645
GL_PROGRAM_TARGET_NV = 0x8646
GL_PROGRAM_RESIDENT_NV = 0x8647
GL_TRACK_MATRIX_NV = 0x8648
GL_TRACK_MATRIX_TRANSFORM_NV = 0x8649
GL_VERTEX_PROGRAM_BINDING_NV = 0x864A
GL_PROGRAM_ERROR_POSITION_ARB = 0x864B
GL_PROGRAM_ERROR_POSITION_NV = 0x864B
GL_OFFSET_TEXTURE_RECTANGLE_NV = 0x864C
GL_OFFSET_TEXTURE_RECTANGLE_SCALE_NV = 0x864D
GL_DOT_PRODUCT_TEXTURE_RECTANGLE_NV = 0x864E
GL_DEPTH_CLAMP = 0x864F
GL_DEPTH_CLAMP_NV = 0x864F
GL_VERTEX_ATTRIB_ARRAY0_NV = 0x8650
GL_VERTEX_ATTRIB_ARRAY1_NV = 0x8651
GL_VERTEX_ATTRIB_ARRAY2_NV = 0x8652
GL_VERTEX_ATTRIB_ARRAY3_NV = 0x8653
GL_VERTEX_ATTRIB_ARRAY4_NV = 0x8654
GL_VERTEX_ATTRIB_ARRAY5_NV = 0x8655
GL_VERTEX_ATTRIB_ARRAY6_NV = 0x8656
GL_VERTEX_ATTRIB_ARRAY7_NV = 0x8657
GL_VERTEX_ATTRIB_ARRAY8_NV = 0x8658
GL_VERTEX_ATTRIB_ARRAY9_NV = 0x8659
GL_VERTEX_ATTRIB_ARRAY10_NV = 0x865A
GL_VERTEX_ATTRIB_ARRAY11_NV = 0x865B
GL_VERTEX_ATTRIB_ARRAY12_NV = 0x865C
GL_VERTEX_ATTRIB_ARRAY13_NV = 0x865D
GL_VERTEX_ATTRIB_ARRAY14_NV = 0x865E
GL_VERTEX_ATTRIB_ARRAY15_NV = 0x865F
GL_MAP1_VERTEX_ATTRIB0_4_NV = 0x8660
GL_MAP1_VERTEX_ATTRIB1_4_NV = 0x8661
GL_MAP1_VERTEX_ATTRIB2_4_NV = 0x8662
GL_MAP1_VERTEX_ATTRIB3_4_NV = 0x8663
GL_MAP1_VERTEX_ATTRIB4_4_NV = 0x8664
GL_MAP1_VERTEX_ATTRIB5_4_NV = 0x8665
GL_MAP1_VERTEX_ATTRIB6_4_NV = 0x8666
GL_MAP1_VERTEX_ATTRIB7_4_NV = 0x8667
GL_MAP1_VERTEX_ATTRIB8_4_NV = 0x8668
GL_MAP1_VERTEX_ATTRIB9_4_NV = 0x8669
GL_MAP1_VERTEX_ATTRIB10_4_NV = 0x866A
GL_MAP1_VERTEX_ATTRIB11_4_NV = 0x866B
GL_MAP1_VERTEX_ATTRIB12_4_NV = 0x866C
GL_MAP1_VERTEX_ATTRIB13_4_NV = 0x866D
GL_MAP1_VERTEX_ATTRIB14_4_NV = 0x866E
GL_MAP1_VERTEX_ATTRIB15_4_NV = 0x866F
GL_MAP2_VERTEX_ATTRIB0_4_NV = 0x8670
GL_MAP2_VERTEX_ATTRIB1_4_NV = 0x8671
GL_MAP2_VERTEX_ATTRIB2_4_NV = 0x8672
GL_MAP2_VERTEX_ATTRIB3_4_NV = 0x8673
GL_MAP2_VERTEX_ATTRIB4_4_NV = 0x8674
GL_MAP2_VERTEX_ATTRIB5_4_NV = 0x8675
GL_MAP2_VERTEX_ATTRIB6_4_NV = 0x8676
GL_MAP2_VERTEX_ATTRIB7_4_NV = 0x8677
GL_PROGRAM_BINDING_ARB = 0x8677
GL_MAP2_VERTEX_ATTRIB8_4_NV = 0x8678
GL_MAP2_VERTEX_ATTRIB9_4_NV = 0x8679
GL_MAP2_VERTEX_ATTRIB10_4_NV = 0x867A
GL_MAP2_VERTEX_ATTRIB11_4_NV = 0x867B
GL_MAP2_VERTEX_ATTRIB12_4_NV = 0x867C
GL_MAP2_VERTEX_ATTRIB13_4_NV = 0x867D
GL_MAP2_VERTEX_ATTRIB14_4_NV = 0x867E
GL_MAP2_VERTEX_ATTRIB15_4_NV = 0x867F
GL_TEXTURE_COMPRESSED_IMAGE_SIZE = 0x86A0
GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB = 0x86A0
GL_TEXTURE_COMPRESSED = 0x86A1
GL_TEXTURE_COMPRESSED_ARB = 0x86A1
GL_NUM_COMPRESSED_TEXTURE_FORMATS = 0x86A2
GL_NUM_COMPRESSED_TEXTURE_FORMATS_ARB = 0x86A2
GL_COMPRESSED_TEXTURE_FORMATS = 0x86A3
GL_COMPRESSED_TEXTURE_FORMATS_ARB = 0x86A3
GL_MAX_VERTEX_UNITS_ARB = 0x86A4
GL_MAX_VERTEX_UNITS_OES = 0x86A4
GL_ACTIVE_VERTEX_UNITS_ARB = 0x86A5
GL_WEIGHT_SUM_UNITY_ARB = 0x86A6
GL_VERTEX_BLEND_ARB = 0x86A7
GL_CURRENT_WEIGHT_ARB = 0x86A8
GL_WEIGHT_ARRAY_TYPE_ARB = 0x86A9
GL_WEIGHT_ARRAY_TYPE_OES = 0x86A9
GL_WEIGHT_ARRAY_STRIDE_ARB = 0x86AA
GL_WEIGHT_ARRAY_STRIDE_OES = 0x86AA
GL_WEIGHT_ARRAY_SIZE_ARB = 0x86AB
GL_WEIGHT_ARRAY_SIZE_OES = 0x86AB
GL_WEIGHT_ARRAY_POINTER_ARB = 0x86AC
GL_WEIGHT_ARRAY_POINTER_OES = 0x86AC
GL_WEIGHT_ARRAY_ARB = 0x86AD
GL_WEIGHT_ARRAY_OES = 0x86AD
GL_DOT3_RGB = 0x86AE
GL_DOT3_RGB_ARB = 0x86AE
GL_DOT3_RGBA = 0x86AF
GL_DOT3_RGBA_ARB = 0x86AF
GL_DOT3_RGBA_IMG = 0x86AF
GL_COMPRESSED_RGB_FXT1_3DFX = 0x86B0
GL_COMPRESSED_RGBA_FXT1_3DFX = 0x86B1
GL_MULTISAMPLE_3DFX = 0x86B2
GL_SAMPLE_BUFFERS_3DFX = 0x86B3
GL_SAMPLES_3DFX = 0x86B4
GL_EVAL_2D_NV = 0x86C0
GL_EVAL_TRIANGULAR_2D_NV = 0x86C1
GL_MAP_TESSELLATION_NV = 0x86C2
GL_MAP_ATTRIB_U_ORDER_NV = 0x86C3
GL_MAP_ATTRIB_V_ORDER_NV = 0x86C4
GL_EVAL_FRACTIONAL_TESSELLATION_NV = 0x86C5
GL_EVAL_VERTEX_ATTRIB0_NV = 0x86C6
GL_EVAL_VERTEX_ATTRIB1_NV = 0x86C7
GL_EVAL_VERTEX_ATTRIB2_NV = 0x86C8
GL_EVAL_VERTEX_ATTRIB3_NV = 0x86C9
GL_EVAL_VERTEX_ATTRIB4_NV = 0x86CA
GL_EVAL_VERTEX_ATTRIB5_NV = 0x86CB
GL_EVAL_VERTEX_ATTRIB6_NV = 0x86CC
GL_EVAL_VERTEX_ATTRIB7_NV = 0x86CD
GL_EVAL_VERTEX_ATTRIB8_NV = 0x86CE
GL_EVAL_VERTEX_ATTRIB9_NV = 0x86CF
GL_EVAL_VERTEX_ATTRIB10_NV = 0x86D0
GL_EVAL_VERTEX_ATTRIB11_NV = 0x86D1
GL_EVAL_VERTEX_ATTRIB12_NV = 0x86D2
GL_EVAL_VERTEX_ATTRIB13_NV = 0x86D3
GL_EVAL_VERTEX_ATTRIB14_NV = 0x86D4
GL_EVAL_VERTEX_ATTRIB15_NV = 0x86D5
GL_MAX_MAP_TESSELLATION_NV = 0x86D6
GL_MAX_RATIONAL_EVAL_ORDER_NV = 0x86D7
GL_MAX_PROGRAM_PATCH_ATTRIBS_NV = 0x86D8
GL_RGBA_UNSIGNED_DOT_PRODUCT_MAPPING_NV = 0x86D9
GL_UNSIGNED_INT_S8_S8_8_8_NV = 0x86DA
GL_UNSIGNED_INT_8_8_S8_S8_REV_NV = 0x86DB
GL_DSDT_MAG_INTENSITY_NV = 0x86DC
GL_SHADER_CONSISTENT_NV = 0x86DD
GL_TEXTURE_SHADER_NV = 0x86DE
GL_SHADER_OPERATION_NV = 0x86DF
GL_CULL_MODES_NV = 0x86E0
GL_OFFSET_TEXTURE_MATRIX_NV = 0x86E1
GL_OFFSET_TEXTURE_2D_MATRIX_NV = 0x86E1
GL_OFFSET_TEXTURE_SCALE_NV = 0x86E2
GL_OFFSET_TEXTURE_2D_SCALE_NV = 0x86E2
GL_OFFSET_TEXTURE_BIAS_NV = 0x86E3
GL_OFFSET_TEXTURE_2D_BIAS_NV = 0x86E3
GL_PREVIOUS_TEXTURE_INPUT_NV = 0x86E4
GL_CONST_EYE_NV = 0x86E5
GL_PASS_THROUGH_NV = 0x86E6
GL_CULL_FRAGMENT_NV = 0x86E7
GL_OFFSET_TEXTURE_2D_NV = 0x86E8
GL_DEPENDENT_AR_TEXTURE_2D_NV = 0x86E9
GL_DEPENDENT_GB_TEXTURE_2D_NV = 0x86EA
GL_SURFACE_STATE_NV = 0x86EB
GL_DOT_PRODUCT_NV = 0x86EC
GL_DOT_PRODUCT_DEPTH_REPLACE_NV = 0x86ED
GL_DOT_PRODUCT_TEXTURE_2D_NV = 0x86EE
GL_DOT_PRODUCT_TEXTURE_3D_NV = 0x86EF
GL_DOT_PRODUCT_TEXTURE_CUBE_MAP_NV = 0x86F0
GL_DOT_PRODUCT_DIFFUSE_CUBE_MAP_NV = 0x86F1
GL_DOT_PRODUCT_REFLECT_CUBE_MAP_NV = 0x86F2
GL_DOT_PRODUCT_CONST_EYE_REFLECT_CUBE_MAP_NV = 0x86F3
GL_HILO_NV = 0x86F4
GL_DSDT_NV = 0x86F5
GL_DSDT_MAG_NV = 0x86F6
GL_DSDT_MAG_VIB_NV = 0x86F7
GL_HILO16_NV = 0x86F8
GL_SIGNED_HILO_NV = 0x86F9
GL_SIGNED_HILO16_NV = 0x86FA
GL_SIGNED_RGBA_NV = 0x86FB
GL_SIGNED_RGBA8_NV = 0x86FC
GL_SURFACE_REGISTERED_NV = 0x86FD
GL_SIGNED_RGB_NV = 0x86FE
GL_SIGNED_RGB8_NV = 0x86FF
GL_SURFACE_MAPPED_NV = 0x8700
GL_SIGNED_LUMINANCE_NV = 0x8701
GL_SIGNED_LUMINANCE8_NV = 0x8702
GL_SIGNED_LUMINANCE_ALPHA_NV = 0x8703
GL_SIGNED_LUMINANCE8_ALPHA8_NV = 0x8704
GL_SIGNED_ALPHA_NV = 0x8705
GL_SIGNED_ALPHA8_NV = 0x8706
GL_SIGNED_INTENSITY_NV = 0x8707
GL_SIGNED_INTENSITY8_NV = 0x8708
GL_DSDT8_NV = 0x8709
GL_DSDT8_MAG8_NV = 0x870A
GL_DSDT8_MAG8_INTENSITY8_NV = 0x870B
GL_SIGNED_RGB_UNSIGNED_ALPHA_NV = 0x870C
GL_SIGNED_RGB8_UNSIGNED_ALPHA8_NV = 0x870D
GL_HI_SCALE_NV = 0x870E
GL_LO_SCALE_NV = 0x870F
GL_DS_SCALE_NV = 0x8710
GL_DT_SCALE_NV = 0x8711
GL_MAGNITUDE_SCALE_NV = 0x8712
GL_VIBRANCE_SCALE_NV = 0x8713
GL_HI_BIAS_NV = 0x8714
GL_LO_BIAS_NV = 0x8715
GL_DS_BIAS_NV = 0x8716
GL_DT_BIAS_NV = 0x8717
GL_MAGNITUDE_BIAS_NV = 0x8718
GL_VIBRANCE_BIAS_NV = 0x8719
GL_TEXTURE_BORDER_VALUES_NV = 0x871A
GL_TEXTURE_HI_SIZE_NV = 0x871B
GL_TEXTURE_LO_SIZE_NV = 0x871C
GL_TEXTURE_DS_SIZE_NV = 0x871D
GL_TEXTURE_DT_SIZE_NV = 0x871E
GL_TEXTURE_MAG_SIZE_NV = 0x871F
GL_MODELVIEW2_ARB = 0x8722
GL_MODELVIEW3_ARB = 0x8723
GL_MODELVIEW4_ARB = 0x8724
GL_MODELVIEW5_ARB = 0x8725
GL_MODELVIEW6_ARB = 0x8726
GL_MODELVIEW7_ARB = 0x8727
GL_MODELVIEW8_ARB = 0x8728
GL_MODELVIEW9_ARB = 0x8729
GL_MODELVIEW10_ARB = 0x872A
GL_MODELVIEW11_ARB = 0x872B
GL_MODELVIEW12_ARB = 0x872C
GL_MODELVIEW13_ARB = 0x872D
GL_MODELVIEW14_ARB = 0x872E
GL_MODELVIEW15_ARB = 0x872F
GL_MODELVIEW16_ARB = 0x8730
GL_MODELVIEW17_ARB = 0x8731
GL_MODELVIEW18_ARB = 0x8732
GL_MODELVIEW19_ARB = 0x8733
GL_MODELVIEW20_ARB = 0x8734
GL_MODELVIEW21_ARB = 0x8735
GL_MODELVIEW22_ARB = 0x8736
GL_MODELVIEW23_ARB = 0x8737
GL_MODELVIEW24_ARB = 0x8738
GL_MODELVIEW25_ARB = 0x8739
GL_MODELVIEW26_ARB = 0x873A
GL_MODELVIEW27_ARB = 0x873B
GL_MODELVIEW28_ARB = 0x873C
GL_MODELVIEW29_ARB = 0x873D
GL_MODELVIEW30_ARB = 0x873E
GL_MODELVIEW31_ARB = 0x873F
GL_DOT3_RGB_EXT = 0x8740
GL_Z400_BINARY_AMD = 0x8740
GL_DOT3_RGBA_EXT = 0x8741
GL_PROGRAM_BINARY_LENGTH_OES = 0x8741
GL_PROGRAM_BINARY_LENGTH = 0x8741
GL_MIRROR_CLAMP_ATI = 0x8742
GL_MIRROR_CLAMP_EXT = 0x8742
GL_MIRROR_CLAMP_TO_EDGE = 0x8743
GL_MIRROR_CLAMP_TO_EDGE_ATI = 0x8743
GL_MIRROR_CLAMP_TO_EDGE_EXT = 0x8743
GL_MODULATE_ADD_ATI = 0x8744
GL_MODULATE_SIGNED_ADD_ATI = 0x8745
GL_MODULATE_SUBTRACT_ATI = 0x8746
GL_SET_AMD = 0x874A
GL_REPLACE_VALUE_AMD = 0x874B
GL_STENCIL_OP_VALUE_AMD = 0x874C
GL_STENCIL_BACK_OP_VALUE_AMD = 0x874D
GL_VERTEX_ATTRIB_ARRAY_LONG = 0x874E
GL_OCCLUSION_QUERY_EVENT_MASK_AMD = 0x874F
GL_DEPTH_STENCIL_MESA = 0x8750
GL_UNSIGNED_INT_24_8_MESA = 0x8751
GL_UNSIGNED_INT_8_24_REV_MESA = 0x8752
GL_UNSIGNED_SHORT_15_1_MESA = 0x8753
GL_UNSIGNED_SHORT_1_15_REV_MESA = 0x8754
GL_TRACE_MASK_MESA = 0x8755
GL_TRACE_NAME_MESA = 0x8756
GL_YCBCR_MESA = 0x8757
GL_PACK_INVERT_MESA = 0x8758
GL_DEBUG_OBJECT_MESA = 0x8759
GL_TEXTURE_1D_STACK_MESAX = 0x8759
GL_DEBUG_PRINT_MESA = 0x875A
GL_TEXTURE_2D_STACK_MESAX = 0x875A
GL_DEBUG_ASSERT_MESA = 0x875B
GL_PROXY_TEXTURE_1D_STACK_MESAX = 0x875B
GL_PROXY_TEXTURE_2D_STACK_MESAX = 0x875C
GL_TEXTURE_1D_STACK_BINDING_MESAX = 0x875D
GL_TEXTURE_2D_STACK_BINDING_MESAX = 0x875E
GL_STATIC_ATI = 0x8760
GL_DYNAMIC_ATI = 0x8761
GL_PRESERVE_ATI = 0x8762
GL_DISCARD_ATI = 0x8763
GL_BUFFER_SIZE = 0x8764
GL_BUFFER_SIZE_ARB = 0x8764
GL_OBJECT_BUFFER_SIZE_ATI = 0x8764
GL_BUFFER_USAGE = 0x8765
GL_BUFFER_USAGE_ARB = 0x8765
GL_OBJECT_BUFFER_USAGE_ATI = 0x8765
GL_ARRAY_OBJECT_BUFFER_ATI = 0x8766
GL_ARRAY_OBJECT_OFFSET_ATI = 0x8767
GL_ELEMENT_ARRAY_ATI = 0x8768
GL_ELEMENT_ARRAY_TYPE_ATI = 0x8769
GL_ELEMENT_ARRAY_POINTER_ATI = 0x876A
GL_MAX_VERTEX_STREAMS_ATI = 0x876B
GL_VERTEX_STREAM0_ATI = 0x876C
GL_VERTEX_STREAM1_ATI = 0x876D
GL_VERTEX_STREAM2_ATI = 0x876E
GL_VERTEX_STREAM3_ATI = 0x876F
GL_VERTEX_STREAM4_ATI = 0x8770
GL_VERTEX_STREAM5_ATI = 0x8771
GL_VERTEX_STREAM6_ATI = 0x8772
GL_VERTEX_STREAM7_ATI = 0x8773
GL_VERTEX_SOURCE_ATI = 0x8774
GL_BUMP_ROT_MATRIX_ATI = 0x8775
GL_BUMP_ROT_MATRIX_SIZE_ATI = 0x8776
GL_BUMP_NUM_TEX_UNITS_ATI = 0x8777
GL_BUMP_TEX_UNITS_ATI = 0x8778
GL_DUDV_ATI = 0x8779
GL_DU8DV8_ATI = 0x877A
GL_BUMP_ENVMAP_ATI = 0x877B
GL_BUMP_TARGET_ATI = 0x877C
GL_VERTEX_SHADER_EXT = 0x8780
GL_VERTEX_SHADER_BINDING_EXT = 0x8781
GL_OP_INDEX_EXT = 0x8782
GL_OP_NEGATE_EXT = 0x8783
GL_OP_DOT3_EXT = 0x8784
GL_OP_DOT4_EXT = 0x8785
GL_OP_MUL_EXT = 0x8786
GL_OP_ADD_EXT = 0x8787
GL_OP_MADD_EXT = 0x8788
GL_OP_FRAC_EXT = 0x8789
GL_OP_MAX_EXT = 0x878A
GL_OP_MIN_EXT = 0x878B
GL_OP_SET_GE_EXT = 0x878C
GL_OP_SET_LT_EXT = 0x878D
GL_OP_CLAMP_EXT = 0x878E
GL_OP_FLOOR_EXT = 0x878F
GL_OP_ROUND_EXT = 0x8790
GL_OP_EXP_BASE_2_EXT = 0x8791
GL_OP_LOG_BASE_2_EXT = 0x8792
GL_OP_POWER_EXT = 0x8793
GL_OP_RECIP_EXT = 0x8794
GL_OP_RECIP_SQRT_EXT = 0x8795
GL_OP_SUB_EXT = 0x8796
GL_OP_CROSS_PRODUCT_EXT = 0x8797
GL_OP_MULTIPLY_MATRIX_EXT = 0x8798
GL_OP_MOV_EXT = 0x8799
GL_OUTPUT_VERTEX_EXT = 0x879A
GL_OUTPUT_COLOR0_EXT = 0x879B
GL_OUTPUT_COLOR1_EXT = 0x879C
GL_OUTPUT_TEXTURE_COORD0_EXT = 0x879D
GL_OUTPUT_TEXTURE_COORD1_EXT = 0x879E
GL_OUTPUT_TEXTURE_COORD2_EXT = 0x879F
GL_OUTPUT_TEXTURE_COORD3_EXT = 0x87A0
GL_OUTPUT_TEXTURE_COORD4_EXT = 0x87A1
GL_OUTPUT_TEXTURE_COORD5_EXT = 0x87A2
GL_OUTPUT_TEXTURE_COORD6_EXT = 0x87A3
GL_OUTPUT_TEXTURE_COORD7_EXT = 0x87A4
GL_OUTPUT_TEXTURE_COORD8_EXT = 0x87A5
GL_OUTPUT_TEXTURE_COORD9_EXT = 0x87A6
GL_OUTPUT_TEXTURE_COORD10_EXT = 0x87A7
GL_OUTPUT_TEXTURE_COORD11_EXT = 0x87A8
GL_OUTPUT_TEXTURE_COORD12_EXT = 0x87A9
GL_OUTPUT_TEXTURE_COORD13_EXT = 0x87AA
GL_OUTPUT_TEXTURE_COORD14_EXT = 0x87AB
GL_OUTPUT_TEXTURE_COORD15_EXT = 0x87AC
GL_OUTPUT_TEXTURE_COORD16_EXT = 0x87AD
GL_OUTPUT_TEXTURE_COORD17_EXT = 0x87AE
GL_OUTPUT_TEXTURE_COORD18_EXT = 0x87AF
GL_OUTPUT_TEXTURE_COORD19_EXT = 0x87B0
GL_OUTPUT_TEXTURE_COORD20_EXT = 0x87B1
GL_OUTPUT_TEXTURE_COORD21_EXT = 0x87B2
GL_OUTPUT_TEXTURE_COORD22_EXT = 0x87B3
GL_OUTPUT_TEXTURE_COORD23_EXT = 0x87B4
GL_OUTPUT_TEXTURE_COORD24_EXT = 0x87B5
GL_OUTPUT_TEXTURE_COORD25_EXT = 0x87B6
GL_OUTPUT_TEXTURE_COORD26_EXT = 0x87B7
GL_OUTPUT_TEXTURE_COORD27_EXT = 0x87B8
GL_OUTPUT_TEXTURE_COORD28_EXT = 0x87B9
GL_OUTPUT_TEXTURE_COORD29_EXT = 0x87BA
GL_OUTPUT_TEXTURE_COORD30_EXT = 0x87BB
GL_OUTPUT_TEXTURE_COORD31_EXT = 0x87BC
GL_OUTPUT_FOG_EXT = 0x87BD
GL_SCALAR_EXT = 0x87BE
GL_VECTOR_EXT = 0x87BF
GL_MATRIX_EXT = 0x87C0
GL_VARIANT_EXT = 0x87C1
GL_INVARIANT_EXT = 0x87C2
GL_LOCAL_CONSTANT_EXT = 0x87C3
GL_LOCAL_EXT = 0x87C4
GL_MAX_VERTEX_SHADER_INSTRUCTIONS_EXT = 0x87C5
GL_MAX_VERTEX_SHADER_VARIANTS_EXT = 0x87C6
GL_MAX_VERTEX_SHADER_INVARIANTS_EXT = 0x87C7
GL_MAX_VERTEX_SHADER_LOCAL_CONSTANTS_EXT = 0x87C8
GL_MAX_VERTEX_SHADER_LOCALS_EXT = 0x87C9
GL_MAX_OPTIMIZED_VERTEX_SHADER_INSTRUCTIONS_EXT = 0x87CA
GL_MAX_OPTIMIZED_VERTEX_SHADER_VARIANTS_EXT = 0x87CB
GL_MAX_OPTIMIZED_VERTEX_SHADER_LOCAL_CONSTANTS_EXT = 0x87CC
GL_MAX_OPTIMIZED_VERTEX_SHADER_INVARIANTS_EXT = 0x87CD
GL_MAX_OPTIMIZED_VERTEX_SHADER_LOCALS_EXT = 0x87CE
GL_VERTEX_SHADER_INSTRUCTIONS_EXT = 0x87CF
GL_VERTEX_SHADER_VARIANTS_EXT = 0x87D0
GL_VERTEX_SHADER_INVARIANTS_EXT = 0x87D1
GL_VERTEX_SHADER_LOCAL_CONSTANTS_EXT = 0x87D2
GL_VERTEX_SHADER_LOCALS_EXT = 0x87D3
GL_VERTEX_SHADER_OPTIMIZED_EXT = 0x87D4
GL_X_EXT = 0x87D5
GL_Y_EXT = 0x87D6
GL_Z_EXT = 0x87D7
GL_W_EXT = 0x87D8
GL_NEGATIVE_X_EXT = 0x87D9
GL_NEGATIVE_Y_EXT = 0x87DA
GL_NEGATIVE_Z_EXT = 0x87DB
GL_NEGATIVE_W_EXT = 0x87DC
GL_ZERO_EXT = 0x87DD
GL_ONE_EXT = 0x87DE
GL_NEGATIVE_ONE_EXT = 0x87DF
GL_NORMALIZED_RANGE_EXT = 0x87E0
GL_FULL_RANGE_EXT = 0x87E1
GL_CURRENT_VERTEX_EXT = 0x87E2
GL_MVP_MATRIX_EXT = 0x87E3
GL_VARIANT_VALUE_EXT = 0x87E4
GL_VARIANT_DATATYPE_EXT = 0x87E5
GL_VARIANT_ARRAY_STRIDE_EXT = 0x87E6
GL_VARIANT_ARRAY_TYPE_EXT = 0x87E7
GL_VARIANT_ARRAY_EXT = 0x87E8
GL_VARIANT_ARRAY_POINTER_EXT = 0x87E9
GL_INVARIANT_VALUE_EXT = 0x87EA
GL_INVARIANT_DATATYPE_EXT = 0x87EB
GL_LOCAL_CONSTANT_VALUE_EXT = 0x87EC
GL_LOCAL_CONSTANT_DATATYPE_EXT = 0x87ED
GL_ATC_RGBA_INTERPOLATED_ALPHA_AMD = 0x87EE
GL_PN_TRIANGLES_ATI = 0x87F0
GL_MAX_PN_TRIANGLES_TESSELATION_LEVEL_ATI = 0x87F1
GL_PN_TRIANGLES_POINT_MODE_ATI = 0x87F2
GL_PN_TRIANGLES_NORMAL_MODE_ATI = 0x87F3
GL_PN_TRIANGLES_TESSELATION_LEVEL_ATI = 0x87F4
GL_PN_TRIANGLES_POINT_MODE_LINEAR_ATI = 0x87F5
GL_PN_TRIANGLES_POINT_MODE_CUBIC_ATI = 0x87F6
GL_PN_TRIANGLES_NORMAL_MODE_LINEAR_ATI = 0x87F7
GL_PN_TRIANGLES_NORMAL_MODE_QUADRATIC_ATI = 0x87F8
GL_3DC_X_AMD = 0x87F9
GL_3DC_XY_AMD = 0x87FA
GL_VBO_FREE_MEMORY_ATI = 0x87FB
GL_TEXTURE_FREE_MEMORY_ATI = 0x87FC
GL_RENDERBUFFER_FREE_MEMORY_ATI = 0x87FD
GL_NUM_PROGRAM_BINARY_FORMATS = 0x87FE
GL_NUM_PROGRAM_BINARY_FORMATS_OES = 0x87FE
GL_PROGRAM_BINARY_FORMATS = 0x87FF
GL_PROGRAM_BINARY_FORMATS_OES = 0x87FF
GL_STENCIL_BACK_FUNC = 0x8800
GL_STENCIL_BACK_FUNC_ATI = 0x8800
GL_STENCIL_BACK_FAIL = 0x8801
GL_STENCIL_BACK_FAIL_ATI = 0x8801
GL_STENCIL_BACK_PASS_DEPTH_FAIL = 0x8802
GL_STENCIL_BACK_PASS_DEPTH_FAIL_ATI = 0x8802
GL_STENCIL_BACK_PASS_DEPTH_PASS = 0x8803
GL_STENCIL_BACK_PASS_DEPTH_PASS_ATI = 0x8803
GL_FRAGMENT_PROGRAM_ARB = 0x8804
GL_PROGRAM_ALU_INSTRUCTIONS_ARB = 0x8805
GL_PROGRAM_TEX_INSTRUCTIONS_ARB = 0x8806
GL_PROGRAM_TEX_INDIRECTIONS_ARB = 0x8807
GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB = 0x8808
GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB = 0x8809
GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB = 0x880A
GL_MAX_PROGRAM_ALU_INSTRUCTIONS_ARB = 0x880B
GL_MAX_PROGRAM_TEX_INSTRUCTIONS_ARB = 0x880C
GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB = 0x880D
GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB = 0x880E
GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB = 0x880F
GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB = 0x8810
GL_RGBA32F = 0x8814
GL_RGBA32F_ARB = 0x8814
GL_RGBA32F_EXT = 0x8814
GL_RGBA_FLOAT32_APPLE = 0x8814
GL_RGBA_FLOAT32_ATI = 0x8814
GL_RGB32F = 0x8815
GL_RGB32F_ARB = 0x8815
GL_RGB32F_EXT = 0x8815
GL_RGB_FLOAT32_APPLE = 0x8815
GL_RGB_FLOAT32_ATI = 0x8815
GL_ALPHA32F_ARB = 0x8816
GL_ALPHA32F_EXT = 0x8816
GL_ALPHA_FLOAT32_APPLE = 0x8816
GL_ALPHA_FLOAT32_ATI = 0x8816
GL_INTENSITY32F_ARB = 0x8817
GL_INTENSITY_FLOAT32_APPLE = 0x8817
GL_INTENSITY_FLOAT32_ATI = 0x8817
GL_LUMINANCE32F_ARB = 0x8818
GL_LUMINANCE32F_EXT = 0x8818
GL_LUMINANCE_FLOAT32_APPLE = 0x8818
GL_LUMINANCE_FLOAT32_ATI = 0x8818
GL_LUMINANCE_ALPHA32F_ARB = 0x8819
GL_LUMINANCE_ALPHA32F_EXT = 0x8819
GL_LUMINANCE_ALPHA_FLOAT32_APPLE = 0x8819
GL_LUMINANCE_ALPHA_FLOAT32_ATI = 0x8819
GL_RGBA16F = 0x881A
GL_RGBA16F_ARB = 0x881A
GL_RGBA16F_EXT = 0x881A
GL_RGBA_FLOAT16_APPLE = 0x881A
GL_RGBA_FLOAT16_ATI = 0x881A
GL_RGB16F = 0x881B
GL_RGB16F_ARB = 0x881B
GL_RGB16F_EXT = 0x881B
GL_RGB_FLOAT16_APPLE = 0x881B
GL_RGB_FLOAT16_ATI = 0x881B
GL_ALPHA16F_ARB = 0x881C
GL_ALPHA16F_EXT = 0x881C
GL_ALPHA_FLOAT16_APPLE = 0x881C
GL_ALPHA_FLOAT16_ATI = 0x881C
GL_INTENSITY16F_ARB = 0x881D
GL_INTENSITY_FLOAT16_APPLE = 0x881D
GL_INTENSITY_FLOAT16_ATI = 0x881D
GL_LUMINANCE16F_ARB = 0x881E
GL_LUMINANCE16F_EXT = 0x881E
GL_LUMINANCE_FLOAT16_APPLE = 0x881E
GL_LUMINANCE_FLOAT16_ATI = 0x881E
GL_LUMINANCE_ALPHA16F_ARB = 0x881F
GL_LUMINANCE_ALPHA16F_EXT = 0x881F
GL_LUMINANCE_ALPHA_FLOAT16_APPLE = 0x881F
GL_LUMINANCE_ALPHA_FLOAT16_ATI = 0x881F
GL_RGBA_FLOAT_MODE_ARB = 0x8820
GL_RGBA_FLOAT_MODE_ATI = 0x8820
GL_WRITEONLY_RENDERING_QCOM = 0x8823
GL_MAX_DRAW_BUFFERS = 0x8824
GL_MAX_DRAW_BUFFERS_ARB = 0x8824
GL_MAX_DRAW_BUFFERS_ATI = 0x8824
GL_MAX_DRAW_BUFFERS_EXT = 0x8824
GL_MAX_DRAW_BUFFERS_NV = 0x8824
GL_DRAW_BUFFER0 = 0x8825
GL_DRAW_BUFFER0_ARB = 0x8825
GL_DRAW_BUFFER0_ATI = 0x8825
GL_DRAW_BUFFER0_EXT = 0x8825
GL_DRAW_BUFFER0_NV = 0x8825
GL_DRAW_BUFFER1 = 0x8826
GL_DRAW_BUFFER1_ARB = 0x8826
GL_DRAW_BUFFER1_ATI = 0x8826
GL_DRAW_BUFFER1_EXT = 0x8826
GL_DRAW_BUFFER1_NV = 0x8826
GL_DRAW_BUFFER2 = 0x8827
GL_DRAW_BUFFER2_ARB = 0x8827
GL_DRAW_BUFFER2_ATI = 0x8827
GL_DRAW_BUFFER2_EXT = 0x8827
GL_DRAW_BUFFER2_NV = 0x8827
GL_DRAW_BUFFER3 = 0x8828
GL_DRAW_BUFFER3_ARB = 0x8828
GL_DRAW_BUFFER3_ATI = 0x8828
GL_DRAW_BUFFER3_EXT = 0x8828
GL_DRAW_BUFFER3_NV = 0x8828
GL_DRAW_BUFFER4 = 0x8829
GL_DRAW_BUFFER4_ARB = 0x8829
GL_DRAW_BUFFER4_ATI = 0x8829
GL_DRAW_BUFFER4_EXT = 0x8829
GL_DRAW_BUFFER4_NV = 0x8829
GL_DRAW_BUFFER5 = 0x882A
GL_DRAW_BUFFER5_ARB = 0x882A
GL_DRAW_BUFFER5_ATI = 0x882A
GL_DRAW_BUFFER5_EXT = 0x882A
GL_DRAW_BUFFER5_NV = 0x882A
GL_DRAW_BUFFER6 = 0x882B
GL_DRAW_BUFFER6_ARB = 0x882B
GL_DRAW_BUFFER6_ATI = 0x882B
GL_DRAW_BUFFER6_EXT = 0x882B
GL_DRAW_BUFFER6_NV = 0x882B
GL_DRAW_BUFFER7 = 0x882C
GL_DRAW_BUFFER7_ARB = 0x882C
GL_DRAW_BUFFER7_ATI = 0x882C
GL_DRAW_BUFFER7_EXT = 0x882C
GL_DRAW_BUFFER7_NV = 0x882C
GL_DRAW_BUFFER8 = 0x882D
GL_DRAW_BUFFER8_ARB = 0x882D
GL_DRAW_BUFFER8_ATI = 0x882D
GL_DRAW_BUFFER8_EXT = 0x882D
GL_DRAW_BUFFER8_NV = 0x882D
GL_DRAW_BUFFER9 = 0x882E
GL_DRAW_BUFFER9_ARB = 0x882E
GL_DRAW_BUFFER9_ATI = 0x882E
GL_DRAW_BUFFER9_EXT = 0x882E
GL_DRAW_BUFFER9_NV = 0x882E
GL_DRAW_BUFFER10 = 0x882F
GL_DRAW_BUFFER10_ARB = 0x882F
GL_DRAW_BUFFER10_ATI = 0x882F
GL_DRAW_BUFFER10_EXT = 0x882F
GL_DRAW_BUFFER10_NV = 0x882F
GL_DRAW_BUFFER11 = 0x8830
GL_DRAW_BUFFER11_ARB = 0x8830
GL_DRAW_BUFFER11_ATI = 0x8830
GL_DRAW_BUFFER11_EXT = 0x8830
GL_DRAW_BUFFER11_NV = 0x8830
GL_DRAW_BUFFER12 = 0x8831
GL_DRAW_BUFFER12_ARB = 0x8831
GL_DRAW_BUFFER12_ATI = 0x8831
GL_DRAW_BUFFER12_EXT = 0x8831
GL_DRAW_BUFFER12_NV = 0x8831
GL_DRAW_BUFFER13 = 0x8832
GL_DRAW_BUFFER13_ARB = 0x8832
GL_DRAW_BUFFER13_ATI = 0x8832
GL_DRAW_BUFFER13_EXT = 0x8832
GL_DRAW_BUFFER13_NV = 0x8832
GL_DRAW_BUFFER14 = 0x8833
GL_DRAW_BUFFER14_ARB = 0x8833
GL_DRAW_BUFFER14_ATI = 0x8833
GL_DRAW_BUFFER14_EXT = 0x8833
GL_DRAW_BUFFER14_NV = 0x8833
GL_DRAW_BUFFER15 = 0x8834
GL_DRAW_BUFFER15_ARB = 0x8834
GL_DRAW_BUFFER15_ATI = 0x8834
GL_DRAW_BUFFER15_EXT = 0x8834
GL_DRAW_BUFFER15_NV = 0x8834
GL_COLOR_CLEAR_UNCLAMPED_VALUE_ATI = 0x8835
GL_COMPRESSED_LUMINANCE_ALPHA_3DC_ATI = 0x8837
GL_BLEND_EQUATION_ALPHA = 0x883D
GL_BLEND_EQUATION_ALPHA_EXT = 0x883D
GL_BLEND_EQUATION_ALPHA_OES = 0x883D
GL_SUBSAMPLE_DISTANCE_AMD = 0x883F
GL_MATRIX_PALETTE_ARB = 0x8840
GL_MATRIX_PALETTE_OES = 0x8840
GL_MAX_MATRIX_PALETTE_STACK_DEPTH_ARB = 0x8841
GL_MAX_PALETTE_MATRICES_ARB = 0x8842
GL_MAX_PALETTE_MATRICES_OES = 0x8842
GL_CURRENT_PALETTE_MATRIX_ARB = 0x8843
GL_CURRENT_PALETTE_MATRIX_OES = 0x8843
GL_MATRIX_INDEX_ARRAY_ARB = 0x8844
GL_MATRIX_INDEX_ARRAY_OES = 0x8844
GL_CURRENT_MATRIX_INDEX_ARB = 0x8845
GL_MATRIX_INDEX_ARRAY_SIZE_ARB = 0x8846
GL_MATRIX_INDEX_ARRAY_SIZE_OES = 0x8846
GL_MATRIX_INDEX_ARRAY_TYPE_ARB = 0x8847
GL_MATRIX_INDEX_ARRAY_TYPE_OES = 0x8847
GL_MATRIX_INDEX_ARRAY_STRIDE_ARB = 0x8848
GL_MATRIX_INDEX_ARRAY_STRIDE_OES = 0x8848
GL_MATRIX_INDEX_ARRAY_POINTER_ARB = 0x8849
GL_MATRIX_INDEX_ARRAY_POINTER_OES = 0x8849
GL_TEXTURE_DEPTH_SIZE = 0x884A
GL_TEXTURE_DEPTH_SIZE_ARB = 0x884A
GL_DEPTH_TEXTURE_MODE = 0x884B
GL_DEPTH_TEXTURE_MODE_ARB = 0x884B
GL_TEXTURE_COMPARE_MODE = 0x884C
GL_TEXTURE_COMPARE_MODE_ARB = 0x884C
GL_TEXTURE_COMPARE_MODE_EXT = 0x884C
GL_TEXTURE_COMPARE_FUNC = 0x884D
GL_TEXTURE_COMPARE_FUNC_ARB = 0x884D
GL_TEXTURE_COMPARE_FUNC_EXT = 0x884D
GL_COMPARE_R_TO_TEXTURE = 0x884E
GL_COMPARE_R_TO_TEXTURE_ARB = 0x884E
GL_COMPARE_REF_DEPTH_TO_TEXTURE_EXT = 0x884E
GL_COMPARE_REF_TO_TEXTURE = 0x884E
GL_COMPARE_REF_TO_TEXTURE_EXT = 0x884E
GL_TEXTURE_CUBE_MAP_SEAMLESS = 0x884F
GL_OFFSET_PROJECTIVE_TEXTURE_2D_NV = 0x8850
GL_OFFSET_PROJECTIVE_TEXTURE_2D_SCALE_NV = 0x8851
GL_OFFSET_PROJECTIVE_TEXTURE_RECTANGLE_NV = 0x8852
GL_OFFSET_PROJECTIVE_TEXTURE_RECTANGLE_SCALE_NV = 0x8853
GL_OFFSET_HILO_TEXTURE_2D_NV = 0x8854
GL_OFFSET_HILO_TEXTURE_RECTANGLE_NV = 0x8855
GL_OFFSET_HILO_PROJECTIVE_TEXTURE_2D_NV = 0x8856
GL_OFFSET_HILO_PROJECTIVE_TEXTURE_RECTANGLE_NV = 0x8857
GL_DEPENDENT_HILO_TEXTURE_2D_NV = 0x8858
GL_DEPENDENT_RGB_TEXTURE_3D_NV = 0x8859
GL_DEPENDENT_RGB_TEXTURE_CUBE_MAP_NV = 0x885A
GL_DOT_PRODUCT_PASS_THROUGH_NV = 0x885B
GL_DOT_PRODUCT_TEXTURE_1D_NV = 0x885C
GL_DOT_PRODUCT_AFFINE_DEPTH_REPLACE_NV = 0x885D
GL_HILO8_NV = 0x885E
GL_SIGNED_HILO8_NV = 0x885F
GL_FORCE_BLUE_TO_ONE_NV = 0x8860
GL_POINT_SPRITE = 0x8861
GL_POINT_SPRITE_ARB = 0x8861
GL_POINT_SPRITE_NV = 0x8861
GL_POINT_SPRITE_OES = 0x8861
GL_COORD_REPLACE = 0x8862
GL_COORD_REPLACE_ARB = 0x8862
GL_COORD_REPLACE_NV = 0x8862
GL_COORD_REPLACE_OES = 0x8862
GL_POINT_SPRITE_R_MODE_NV = 0x8863
GL_PIXEL_COUNTER_BITS_NV = 0x8864
GL_QUERY_COUNTER_BITS = 0x8864
GL_QUERY_COUNTER_BITS_ARB = 0x8864
GL_QUERY_COUNTER_BITS_EXT = 0x8864
GL_CURRENT_OCCLUSION_QUERY_ID_NV = 0x8865
GL_CURRENT_QUERY = 0x8865
GL_CURRENT_QUERY_ARB = 0x8865
GL_CURRENT_QUERY_EXT = 0x8865
GL_PIXEL_COUNT_NV = 0x8866
GL_QUERY_RESULT = 0x8866
GL_QUERY_RESULT_ARB = 0x8866
GL_QUERY_RESULT_EXT = 0x8866
GL_PIXEL_COUNT_AVAILABLE_NV = 0x8867
GL_QUERY_RESULT_AVAILABLE = 0x8867
GL_QUERY_RESULT_AVAILABLE_ARB = 0x8867
GL_QUERY_RESULT_AVAILABLE_EXT = 0x8867
GL_MAX_FRAGMENT_PROGRAM_LOCAL_PARAMETERS_NV = 0x8868
GL_MAX_VERTEX_ATTRIBS = 0x8869
GL_MAX_VERTEX_ATTRIBS_ARB = 0x8869
GL_VERTEX_ATTRIB_ARRAY_NORMALIZED = 0x886A
GL_VERTEX_ATTRIB_ARRAY_NORMALIZED_ARB = 0x886A
GL_MAX_TESS_CONTROL_INPUT_COMPONENTS = 0x886C
GL_MAX_TESS_CONTROL_INPUT_COMPONENTS_EXT = 0x886C
GL_MAX_TESS_EVALUATION_INPUT_COMPONENTS = 0x886D
GL_MAX_TESS_EVALUATION_INPUT_COMPONENTS_EXT = 0x886D
GL_DEPTH_STENCIL_TO_RGBA_NV = 0x886E
GL_DEPTH_STENCIL_TO_BGRA_NV = 0x886F
GL_FRAGMENT_PROGRAM_NV = 0x8870
GL_MAX_TEXTURE_COORDS = 0x8871
GL_MAX_TEXTURE_COORDS_ARB = 0x8871
GL_MAX_TEXTURE_COORDS_NV = 0x8871
GL_MAX_TEXTURE_IMAGE_UNITS = 0x8872
GL_MAX_TEXTURE_IMAGE_UNITS_ARB = 0x8872
GL_MAX_TEXTURE_IMAGE_UNITS_NV = 0x8872
GL_FRAGMENT_PROGRAM_BINDING_NV = 0x8873
GL_PROGRAM_ERROR_STRING_ARB = 0x8874
GL_PROGRAM_ERROR_STRING_NV = 0x8874
GL_PROGRAM_FORMAT_ASCII_ARB = 0x8875
GL_PROGRAM_FORMAT_ARB = 0x8876
GL_WRITE_PIXEL_DATA_RANGE_NV = 0x8878
GL_READ_PIXEL_DATA_RANGE_NV = 0x8879
GL_WRITE_PIXEL_DATA_RANGE_LENGTH_NV = 0x887A
GL_READ_PIXEL_DATA_RANGE_LENGTH_NV = 0x887B
GL_WRITE_PIXEL_DATA_RANGE_POINTER_NV = 0x887C
GL_READ_PIXEL_DATA_RANGE_POINTER_NV = 0x887D
GL_GEOMETRY_SHADER_INVOCATIONS = 0x887F
GL_GEOMETRY_SHADER_INVOCATIONS_EXT = 0x887F
GL_FLOAT_R_NV = 0x8880
GL_FLOAT_RG_NV = 0x8881
GL_FLOAT_RGB_NV = 0x8882
GL_FLOAT_RGBA_NV = 0x8883
GL_FLOAT_R16_NV = 0x8884
GL_FLOAT_R32_NV = 0x8885
GL_FLOAT_RG16_NV = 0x8886
GL_FLOAT_RG32_NV = 0x8887
GL_FLOAT_RGB16_NV = 0x8888
GL_FLOAT_RGB32_NV = 0x8889
GL_FLOAT_RGBA16_NV = 0x888A
GL_FLOAT_RGBA32_NV = 0x888B
GL_TEXTURE_FLOAT_COMPONENTS_NV = 0x888C
GL_FLOAT_CLEAR_COLOR_VALUE_NV = 0x888D
GL_FLOAT_RGBA_MODE_NV = 0x888E
GL_TEXTURE_UNSIGNED_REMAP_MODE_NV = 0x888F
GL_DEPTH_BOUNDS_TEST_EXT = 0x8890
GL_DEPTH_BOUNDS_EXT = 0x8891
GL_ARRAY_BUFFER = 0x8892
GL_ARRAY_BUFFER_ARB = 0x8892
GL_ELEMENT_ARRAY_BUFFER = 0x8893
GL_ELEMENT_ARRAY_BUFFER_ARB = 0x8893
GL_ARRAY_BUFFER_BINDING = 0x8894
GL_ARRAY_BUFFER_BINDING_ARB = 0x8894
GL_ELEMENT_ARRAY_BUFFER_BINDING = 0x8895
GL_ELEMENT_ARRAY_BUFFER_BINDING_ARB = 0x8895
GL_VERTEX_ARRAY_BUFFER_BINDING = 0x8896
GL_VERTEX_ARRAY_BUFFER_BINDING_ARB = 0x8896
GL_NORMAL_ARRAY_BUFFER_BINDING = 0x8897
GL_NORMAL_ARRAY_BUFFER_BINDING_ARB = 0x8897
GL_COLOR_ARRAY_BUFFER_BINDING = 0x8898
GL_COLOR_ARRAY_BUFFER_BINDING_ARB = 0x8898
GL_INDEX_ARRAY_BUFFER_BINDING = 0x8899
GL_INDEX_ARRAY_BUFFER_BINDING_ARB = 0x8899
GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING = 0x889A
GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING_ARB = 0x889A
GL_EDGE_FLAG_ARRAY_BUFFER_BINDING = 0x889B
GL_EDGE_FLAG_ARRAY_BUFFER_BINDING_ARB = 0x889B
GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING = 0x889C
GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING_ARB = 0x889C
GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING_ARB = 0x889D
GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING = 0x889D
GL_FOG_COORD_ARRAY_BUFFER_BINDING = 0x889D
GL_WEIGHT_ARRAY_BUFFER_BINDING = 0x889E
GL_WEIGHT_ARRAY_BUFFER_BINDING_ARB = 0x889E
GL_WEIGHT_ARRAY_BUFFER_BINDING_OES = 0x889E
GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING = 0x889F
GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING_ARB = 0x889F
GL_PROGRAM_INSTRUCTIONS_ARB = 0x88A0
GL_MAX_PROGRAM_INSTRUCTIONS_ARB = 0x88A1
GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB = 0x88A2
GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB = 0x88A3
GL_PROGRAM_TEMPORARIES_ARB = 0x88A4
GL_MAX_PROGRAM_TEMPORARIES_ARB = 0x88A5
GL_PROGRAM_NATIVE_TEMPORARIES_ARB = 0x88A6
GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB = 0x88A7
GL_PROGRAM_PARAMETERS_ARB = 0x88A8
GL_MAX_PROGRAM_PARAMETERS_ARB = 0x88A9
GL_PROGRAM_NATIVE_PARAMETERS_ARB = 0x88AA
GL_MAX_PROGRAM_NATIVE_PARAMETERS_ARB = 0x88AB
GL_PROGRAM_ATTRIBS_ARB = 0x88AC
GL_MAX_PROGRAM_ATTRIBS_ARB = 0x88AD
GL_PROGRAM_NATIVE_ATTRIBS_ARB = 0x88AE
GL_MAX_PROGRAM_NATIVE_ATTRIBS_ARB = 0x88AF
GL_PROGRAM_ADDRESS_REGISTERS_ARB = 0x88B0
GL_MAX_PROGRAM_ADDRESS_REGISTERS_ARB = 0x88B1
GL_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB = 0x88B2
GL_MAX_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB = 0x88B3
GL_MAX_PROGRAM_LOCAL_PARAMETERS_ARB = 0x88B4
GL_MAX_PROGRAM_ENV_PARAMETERS_ARB = 0x88B5
GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB = 0x88B6
GL_TRANSPOSE_CURRENT_MATRIX_ARB = 0x88B7
GL_READ_ONLY = 0x88B8
GL_READ_ONLY_ARB = 0x88B8
GL_WRITE_ONLY = 0x88B9
GL_WRITE_ONLY_ARB = 0x88B9
GL_WRITE_ONLY_OES = 0x88B9
GL_READ_WRITE = 0x88BA
GL_READ_WRITE_ARB = 0x88BA
GL_BUFFER_ACCESS = 0x88BB
GL_BUFFER_ACCESS_ARB = 0x88BB
GL_BUFFER_ACCESS_OES = 0x88BB
GL_BUFFER_MAPPED = 0x88BC
GL_BUFFER_MAPPED_ARB = 0x88BC
GL_BUFFER_MAPPED_OES = 0x88BC
GL_BUFFER_MAP_POINTER = 0x88BD
GL_BUFFER_MAP_POINTER_ARB = 0x88BD
GL_BUFFER_MAP_POINTER_OES = 0x88BD
GL_WRITE_DISCARD_NV = 0x88BE
GL_TIME_ELAPSED = 0x88BF
GL_TIME_ELAPSED_EXT = 0x88BF
GL_MATRIX0_ARB = 0x88C0
GL_MATRIX1_ARB = 0x88C1
GL_MATRIX2_ARB = 0x88C2
GL_MATRIX3_ARB = 0x88C3
GL_MATRIX4_ARB = 0x88C4
GL_MATRIX5_ARB = 0x88C5
GL_MATRIX6_ARB = 0x88C6
GL_MATRIX7_ARB = 0x88C7
GL_MATRIX8_ARB = 0x88C8
GL_MATRIX9_ARB = 0x88C9
GL_MATRIX10_ARB = 0x88CA
GL_MATRIX11_ARB = 0x88CB
GL_MATRIX12_ARB = 0x88CC
GL_MATRIX13_ARB = 0x88CD
GL_MATRIX14_ARB = 0x88CE
GL_MATRIX15_ARB = 0x88CF
GL_MATRIX16_ARB = 0x88D0
GL_MATRIX17_ARB = 0x88D1
GL_MATRIX18_ARB = 0x88D2
GL_MATRIX19_ARB = 0x88D3
GL_MATRIX20_ARB = 0x88D4
GL_MATRIX21_ARB = 0x88D5
GL_MATRIX22_ARB = 0x88D6
GL_MATRIX23_ARB = 0x88D7
GL_MATRIX24_ARB = 0x88D8
GL_MATRIX25_ARB = 0x88D9
GL_MATRIX26_ARB = 0x88DA
GL_MATRIX27_ARB = 0x88DB
GL_MATRIX28_ARB = 0x88DC
GL_MATRIX29_ARB = 0x88DD
GL_MATRIX30_ARB = 0x88DE
GL_MATRIX31_ARB = 0x88DF
GL_STREAM_DRAW = 0x88E0
GL_STREAM_DRAW_ARB = 0x88E0
GL_STREAM_READ = 0x88E1
GL_STREAM_READ_ARB = 0x88E1
GL_STREAM_COPY = 0x88E2
GL_STREAM_COPY_ARB = 0x88E2
GL_STATIC_DRAW = 0x88E4
GL_STATIC_DRAW_ARB = 0x88E4
GL_STATIC_READ = 0x88E5
GL_STATIC_READ_ARB = 0x88E5
GL_STATIC_COPY = 0x88E6
GL_STATIC_COPY_ARB = 0x88E6
GL_DYNAMIC_DRAW = 0x88E8
GL_DYNAMIC_DRAW_ARB = 0x88E8
GL_DYNAMIC_READ = 0x88E9
GL_DYNAMIC_READ_ARB = 0x88E9
GL_DYNAMIC_COPY = 0x88EA
GL_DYNAMIC_COPY_ARB = 0x88EA
GL_PIXEL_PACK_BUFFER = 0x88EB
GL_PIXEL_PACK_BUFFER_ARB = 0x88EB
GL_PIXEL_PACK_BUFFER_EXT = 0x88EB
GL_PIXEL_UNPACK_BUFFER = 0x88EC
GL_PIXEL_UNPACK_BUFFER_ARB = 0x88EC
GL_PIXEL_UNPACK_BUFFER_EXT = 0x88EC
GL_PIXEL_PACK_BUFFER_BINDING = 0x88ED
GL_PIXEL_PACK_BUFFER_BINDING_ARB = 0x88ED
GL_PIXEL_PACK_BUFFER_BINDING_EXT = 0x88ED
GL_ETC1_SRGB8_NV = 0x88EE
GL_PIXEL_UNPACK_BUFFER_BINDING = 0x88EF
GL_PIXEL_UNPACK_BUFFER_BINDING_ARB = 0x88EF
GL_PIXEL_UNPACK_BUFFER_BINDING_EXT = 0x88EF
GL_DEPTH24_STENCIL8 = 0x88F0
GL_DEPTH24_STENCIL8_EXT = 0x88F0
GL_DEPTH24_STENCIL8_OES = 0x88F0
GL_TEXTURE_STENCIL_SIZE = 0x88F1
GL_TEXTURE_STENCIL_SIZE_EXT = 0x88F1
GL_STENCIL_TAG_BITS_EXT = 0x88F2
GL_STENCIL_CLEAR_TAG_VALUE_EXT = 0x88F3
GL_MAX_PROGRAM_EXEC_INSTRUCTIONS_NV = 0x88F4
GL_MAX_PROGRAM_CALL_DEPTH_NV = 0x88F5
GL_MAX_PROGRAM_IF_DEPTH_NV = 0x88F6
GL_MAX_PROGRAM_LOOP_DEPTH_NV = 0x88F7
GL_MAX_PROGRAM_LOOP_COUNT_NV = 0x88F8
GL_SRC1_COLOR = 0x88F9
GL_ONE_MINUS_SRC1_COLOR = 0x88FA
GL_ONE_MINUS_SRC1_ALPHA = 0x88FB
GL_MAX_DUAL_SOURCE_DRAW_BUFFERS = 0x88FC
GL_VERTEX_ATTRIB_ARRAY_INTEGER = 0x88FD
GL_VERTEX_ATTRIB_ARRAY_INTEGER_EXT = 0x88FD
GL_VERTEX_ATTRIB_ARRAY_INTEGER_NV = 0x88FD
GL_VERTEX_ATTRIB_ARRAY_DIVISOR = 0x88FE
GL_VERTEX_ATTRIB_ARRAY_DIVISOR_ANGLE = 0x88FE
GL_VERTEX_ATTRIB_ARRAY_DIVISOR_ARB = 0x88FE
GL_VERTEX_ATTRIB_ARRAY_DIVISOR_EXT = 0x88FE
GL_VERTEX_ATTRIB_ARRAY_DIVISOR_NV = 0x88FE
GL_MAX_ARRAY_TEXTURE_LAYERS = 0x88FF
GL_MAX_ARRAY_TEXTURE_LAYERS_EXT = 0x88FF
GL_MIN_PROGRAM_TEXEL_OFFSET = 0x8904
GL_MIN_PROGRAM_TEXEL_OFFSET_EXT = 0x8904
GL_MIN_PROGRAM_TEXEL_OFFSET_NV = 0x8904
GL_MAX_PROGRAM_TEXEL_OFFSET = 0x8905
GL_MAX_PROGRAM_TEXEL_OFFSET_EXT = 0x8905
GL_MAX_PROGRAM_TEXEL_OFFSET_NV = 0x8905
GL_PROGRAM_ATTRIB_COMPONENTS_NV = 0x8906
GL_PROGRAM_RESULT_COMPONENTS_NV = 0x8907
GL_MAX_PROGRAM_ATTRIB_COMPONENTS_NV = 0x8908
GL_MAX_PROGRAM_RESULT_COMPONENTS_NV = 0x8909
GL_STENCIL_TEST_TWO_SIDE_EXT = 0x8910
GL_ACTIVE_STENCIL_FACE_EXT = 0x8911
GL_MIRROR_CLAMP_TO_BORDER_EXT = 0x8912
GL_SAMPLES_PASSED = 0x8914
GL_SAMPLES_PASSED_ARB = 0x8914
GL_GEOMETRY_VERTICES_OUT = 0x8916
GL_GEOMETRY_LINKED_VERTICES_OUT_EXT = 0x8916
GL_GEOMETRY_INPUT_TYPE = 0x8917
GL_GEOMETRY_LINKED_INPUT_TYPE_EXT = 0x8917
GL_GEOMETRY_OUTPUT_TYPE = 0x8918
GL_GEOMETRY_LINKED_OUTPUT_TYPE_EXT = 0x8918
GL_SAMPLER_BINDING = 0x8919
GL_CLAMP_VERTEX_COLOR = 0x891A
GL_CLAMP_VERTEX_COLOR_ARB = 0x891A
GL_CLAMP_FRAGMENT_COLOR = 0x891B
GL_CLAMP_FRAGMENT_COLOR_ARB = 0x891B
GL_CLAMP_READ_COLOR = 0x891C
GL_CLAMP_READ_COLOR_ARB = 0x891C
GL_FIXED_ONLY = 0x891D
GL_FIXED_ONLY_ARB = 0x891D
GL_TESS_CONTROL_PROGRAM_NV = 0x891E
GL_TESS_EVALUATION_PROGRAM_NV = 0x891F
GL_FRAGMENT_SHADER_ATI = 0x8920
GL_REG_0_ATI = 0x8921
GL_REG_1_ATI = 0x8922
GL_REG_2_ATI = 0x8923
GL_REG_3_ATI = 0x8924
GL_REG_4_ATI = 0x8925
GL_REG_5_ATI = 0x8926
GL_REG_6_ATI = 0x8927
GL_REG_7_ATI = 0x8928
GL_REG_8_ATI = 0x8929
GL_REG_9_ATI = 0x892A
GL_REG_10_ATI = 0x892B
GL_REG_11_ATI = 0x892C
GL_REG_12_ATI = 0x892D
GL_REG_13_ATI = 0x892E
GL_REG_14_ATI = 0x892F
GL_REG_15_ATI = 0x8930
GL_REG_16_ATI = 0x8931
GL_REG_17_ATI = 0x8932
GL_REG_18_ATI = 0x8933
GL_REG_19_ATI = 0x8934
GL_REG_20_ATI = 0x8935
GL_REG_21_ATI = 0x8936
GL_REG_22_ATI = 0x8937
GL_REG_23_ATI = 0x8938
GL_REG_24_ATI = 0x8939
GL_REG_25_ATI = 0x893A
GL_REG_26_ATI = 0x893B
GL_REG_27_ATI = 0x893C
GL_REG_28_ATI = 0x893D
GL_REG_29_ATI = 0x893E
GL_REG_30_ATI = 0x893F
GL_REG_31_ATI = 0x8940
GL_CON_0_ATI = 0x8941
GL_CON_1_ATI = 0x8942
GL_CON_2_ATI = 0x8943
GL_CON_3_ATI = 0x8944
GL_CON_4_ATI = 0x8945
GL_CON_5_ATI = 0x8946
GL_CON_6_ATI = 0x8947
GL_CON_7_ATI = 0x8948
GL_CON_8_ATI = 0x8949
GL_CON_9_ATI = 0x894A
GL_CON_10_ATI = 0x894B
GL_CON_11_ATI = 0x894C
GL_CON_12_ATI = 0x894D
GL_CON_13_ATI = 0x894E
GL_CON_14_ATI = 0x894F
GL_CON_15_ATI = 0x8950
GL_CON_16_ATI = 0x8951
GL_CON_17_ATI = 0x8952
GL_CON_18_ATI = 0x8953
GL_CON_19_ATI = 0x8954
GL_CON_20_ATI = 0x8955
GL_CON_21_ATI = 0x8956
GL_CON_22_ATI = 0x8957
GL_CON_23_ATI = 0x8958
GL_CON_24_ATI = 0x8959
GL_CON_25_ATI = 0x895A
GL_CON_26_ATI = 0x895B
GL_CON_27_ATI = 0x895C
GL_CON_28_ATI = 0x895D
GL_CON_29_ATI = 0x895E
GL_CON_30_ATI = 0x895F
GL_CON_31_ATI = 0x8960
GL_MOV_ATI = 0x8961
GL_ADD_ATI = 0x8963
GL_MUL_ATI = 0x8964
GL_SUB_ATI = 0x8965
GL_DOT3_ATI = 0x8966
GL_DOT4_ATI = 0x8967
GL_MAD_ATI = 0x8968
GL_LERP_ATI = 0x8969
GL_CND_ATI = 0x896A
GL_CND0_ATI = 0x896B
GL_DOT2_ADD_ATI = 0x896C
GL_SECONDARY_INTERPOLATOR_ATI = 0x896D
GL_NUM_FRAGMENT_REGISTERS_ATI = 0x896E
GL_NUM_FRAGMENT_CONSTANTS_ATI = 0x896F
GL_NUM_PASSES_ATI = 0x8970
GL_NUM_INSTRUCTIONS_PER_PASS_ATI = 0x8971
GL_NUM_INSTRUCTIONS_TOTAL_ATI = 0x8972
GL_NUM_INPUT_INTERPOLATOR_COMPONENTS_ATI = 0x8973
GL_NUM_LOOPBACK_COMPONENTS_ATI = 0x8974
GL_COLOR_ALPHA_PAIRING_ATI = 0x8975
GL_SWIZZLE_STR_ATI = 0x8976
GL_SWIZZLE_STQ_ATI = 0x8977
GL_SWIZZLE_STR_DR_ATI = 0x8978
GL_SWIZZLE_STQ_DQ_ATI = 0x8979
GL_SWIZZLE_STRQ_ATI = 0x897A
GL_SWIZZLE_STRQ_DQ_ATI = 0x897B
GL_INTERLACE_OML = 0x8980
GL_INTERLACE_READ_OML = 0x8981
GL_FORMAT_SUBSAMPLE_24_24_OML = 0x8982
GL_FORMAT_SUBSAMPLE_244_244_OML = 0x8983
GL_PACK_RESAMPLE_OML = 0x8984
GL_UNPACK_RESAMPLE_OML = 0x8985
GL_RESAMPLE_REPLICATE_OML = 0x8986
GL_RESAMPLE_ZERO_FILL_OML = 0x8987
GL_RESAMPLE_AVERAGE_OML = 0x8988
GL_RESAMPLE_DECIMATE_OML = 0x8989
GL_POINT_SIZE_ARRAY_TYPE_OES = 0x898A
GL_POINT_SIZE_ARRAY_STRIDE_OES = 0x898B
GL_POINT_SIZE_ARRAY_POINTER_OES = 0x898C
GL_MODELVIEW_MATRIX_FLOAT_AS_INT_BITS_OES = 0x898D
GL_PROJECTION_MATRIX_FLOAT_AS_INT_BITS_OES = 0x898E
GL_TEXTURE_MATRIX_FLOAT_AS_INT_BITS_OES = 0x898F
GL_VERTEX_ATTRIB_MAP1_APPLE = 0x8A00
GL_VERTEX_ATTRIB_MAP2_APPLE = 0x8A01
GL_VERTEX_ATTRIB_MAP1_SIZE_APPLE = 0x8A02
GL_VERTEX_ATTRIB_MAP1_COEFF_APPLE = 0x8A03
GL_VERTEX_ATTRIB_MAP1_ORDER_APPLE = 0x8A04
GL_VERTEX_ATTRIB_MAP1_DOMAIN_APPLE = 0x8A05
GL_VERTEX_ATTRIB_MAP2_SIZE_APPLE = 0x8A06
GL_VERTEX_ATTRIB_MAP2_COEFF_APPLE = 0x8A07
GL_VERTEX_ATTRIB_MAP2_ORDER_APPLE = 0x8A08
GL_VERTEX_ATTRIB_MAP2_DOMAIN_APPLE = 0x8A09
GL_DRAW_PIXELS_APPLE = 0x8A0A
GL_FENCE_APPLE = 0x8A0B
GL_ELEMENT_ARRAY_APPLE = 0x8A0C
GL_ELEMENT_ARRAY_TYPE_APPLE = 0x8A0D
GL_ELEMENT_ARRAY_POINTER_APPLE = 0x8A0E
GL_COLOR_FLOAT_APPLE = 0x8A0F
GL_UNIFORM_BUFFER = 0x8A11
GL_BUFFER_SERIALIZED_MODIFY_APPLE = 0x8A12
GL_BUFFER_FLUSHING_UNMAP_APPLE = 0x8A13
GL_AUX_DEPTH_STENCIL_APPLE = 0x8A14
GL_PACK_ROW_BYTES_APPLE = 0x8A15
GL_UNPACK_ROW_BYTES_APPLE = 0x8A16
GL_RELEASED_APPLE = 0x8A19
GL_VOLATILE_APPLE = 0x8A1A
GL_RETAINED_APPLE = 0x8A1B
GL_UNDEFINED_APPLE = 0x8A1C
GL_PURGEABLE_APPLE = 0x8A1D
GL_RGB_422_APPLE = 0x8A1F
GL_UNIFORM_BUFFER_BINDING = 0x8A28
GL_UNIFORM_BUFFER_START = 0x8A29
GL_UNIFORM_BUFFER_SIZE = 0x8A2A
GL_MAX_VERTEX_UNIFORM_BLOCKS = 0x8A2B
GL_MAX_GEOMETRY_UNIFORM_BLOCKS = 0x8A2C
GL_MAX_GEOMETRY_UNIFORM_BLOCKS_EXT = 0x8A2C
GL_MAX_FRAGMENT_UNIFORM_BLOCKS = 0x8A2D
GL_MAX_COMBINED_UNIFORM_BLOCKS = 0x8A2E
GL_MAX_UNIFORM_BUFFER_BINDINGS = 0x8A2F
GL_MAX_UNIFORM_BLOCK_SIZE = 0x8A30
GL_MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS = 0x8A31
GL_MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS = 0x8A32
GL_MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS_EXT = 0x8A32
GL_MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS = 0x8A33
GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT = 0x8A34
GL_ACTIVE_UNIFORM_BLOCK_MAX_NAME_LENGTH = 0x8A35
GL_ACTIVE_UNIFORM_BLOCKS = 0x8A36
GL_UNIFORM_TYPE = 0x8A37
GL_UNIFORM_SIZE = 0x8A38
GL_UNIFORM_NAME_LENGTH = 0x8A39
GL_UNIFORM_BLOCK_INDEX = 0x8A3A
GL_UNIFORM_OFFSET = 0x8A3B
GL_UNIFORM_ARRAY_STRIDE = 0x8A3C
GL_UNIFORM_MATRIX_STRIDE = 0x8A3D
GL_UNIFORM_IS_ROW_MAJOR = 0x8A3E
GL_UNIFORM_BLOCK_BINDING = 0x8A3F
GL_UNIFORM_BLOCK_DATA_SIZE = 0x8A40
GL_UNIFORM_BLOCK_NAME_LENGTH = 0x8A41
GL_UNIFORM_BLOCK_ACTIVE_UNIFORMS = 0x8A42
GL_UNIFORM_BLOCK_ACTIVE_UNIFORM_INDICES = 0x8A43
GL_UNIFORM_BLOCK_REFERENCED_BY_VERTEX_SHADER = 0x8A44
GL_UNIFORM_BLOCK_REFERENCED_BY_GEOMETRY_SHADER = 0x8A45
GL_UNIFORM_BLOCK_REFERENCED_BY_FRAGMENT_SHADER = 0x8A46
GL_TEXTURE_SRGB_DECODE_EXT = 0x8A48
GL_DECODE_EXT = 0x8A49
GL_SKIP_DECODE_EXT = 0x8A4A
GL_PROGRAM_PIPELINE_OBJECT_EXT = 0x8A4F
GL_RGB_RAW_422_APPLE = 0x8A51
GL_FRAGMENT_SHADER_DISCARDS_SAMPLES_EXT = 0x8A52
GL_SYNC_OBJECT_APPLE = 0x8A53
GL_COMPRESSED_SRGB_PVRTC_2BPPV1_EXT = 0x8A54
GL_COMPRESSED_SRGB_PVRTC_4BPPV1_EXT = 0x8A55
GL_COMPRESSED_SRGB_ALPHA_PVRTC_2BPPV1_EXT = 0x8A56
GL_COMPRESSED_SRGB_ALPHA_PVRTC_4BPPV1_EXT = 0x8A57
GL_FRAGMENT_SHADER = 0x8B30
GL_FRAGMENT_SHADER_ARB = 0x8B30
GL_VERTEX_SHADER = 0x8B31
GL_VERTEX_SHADER_ARB = 0x8B31
GL_PROGRAM_OBJECT_ARB = 0x8B40
GL_PROGRAM_OBJECT_EXT = 0x8B40
GL_SHADER_OBJECT_ARB = 0x8B48
GL_SHADER_OBJECT_EXT = 0x8B48
GL_MAX_FRAGMENT_UNIFORM_COMPONENTS = 0x8B49
GL_MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB = 0x8B49
GL_MAX_VERTEX_UNIFORM_COMPONENTS = 0x8B4A
GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB = 0x8B4A
GL_MAX_VARYING_FLOATS = 0x8B4B
GL_MAX_VARYING_COMPONENTS = 0x8B4B
GL_MAX_VARYING_COMPONENTS_EXT = 0x8B4B
GL_MAX_VARYING_FLOATS_ARB = 0x8B4B
GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS = 0x8B4C
GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB = 0x8B4C
GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS = 0x8B4D
GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS_ARB = 0x8B4D
GL_OBJECT_TYPE_ARB = 0x8B4E
GL_SHADER_TYPE = 0x8B4F
GL_OBJECT_SUBTYPE_ARB = 0x8B4F
GL_FLOAT_VEC2 = 0x8B50
GL_FLOAT_VEC2_ARB = 0x8B50
GL_FLOAT_VEC3 = 0x8B51
GL_FLOAT_VEC3_ARB = 0x8B51
GL_FLOAT_VEC4 = 0x8B52
GL_FLOAT_VEC4_ARB = 0x8B52
GL_INT_VEC2 = 0x8B53
GL_INT_VEC2_ARB = 0x8B53
GL_INT_VEC3 = 0x8B54
GL_INT_VEC3_ARB = 0x8B54
GL_INT_VEC4 = 0x8B55
GL_INT_VEC4_ARB = 0x8B55
GL_BOOL = 0x8B56
GL_BOOL_ARB = 0x8B56
GL_BOOL_VEC2 = 0x8B57
GL_BOOL_VEC2_ARB = 0x8B57
GL_BOOL_VEC3 = 0x8B58
GL_BOOL_VEC3_ARB = 0x8B58
GL_BOOL_VEC4 = 0x8B59
GL_BOOL_VEC4_ARB = 0x8B59
GL_FLOAT_MAT2 = 0x8B5A
GL_FLOAT_MAT2_ARB = 0x8B5A
GL_FLOAT_MAT3 = 0x8B5B
GL_FLOAT_MAT3_ARB = 0x8B5B
GL_FLOAT_MAT4 = 0x8B5C
GL_FLOAT_MAT4_ARB = 0x8B5C
GL_SAMPLER_1D = 0x8B5D
GL_SAMPLER_1D_ARB = 0x8B5D
GL_SAMPLER_2D = 0x8B5E
GL_SAMPLER_2D_ARB = 0x8B5E
GL_SAMPLER_3D = 0x8B5F
GL_SAMPLER_3D_ARB = 0x8B5F
GL_SAMPLER_3D_OES = 0x8B5F
GL_SAMPLER_CUBE = 0x8B60
GL_SAMPLER_CUBE_ARB = 0x8B60
GL_SAMPLER_1D_SHADOW = 0x8B61
GL_SAMPLER_1D_SHADOW_ARB = 0x8B61
GL_SAMPLER_2D_SHADOW = 0x8B62
GL_SAMPLER_2D_SHADOW_ARB = 0x8B62
GL_SAMPLER_2D_SHADOW_EXT = 0x8B62
GL_SAMPLER_2D_RECT = 0x8B63
GL_SAMPLER_2D_RECT_ARB = 0x8B63
GL_SAMPLER_2D_RECT_SHADOW = 0x8B64
GL_SAMPLER_2D_RECT_SHADOW_ARB = 0x8B64
GL_FLOAT_MAT2x3 = 0x8B65
GL_FLOAT_MAT2x3_NV = 0x8B65
GL_FLOAT_MAT2x4 = 0x8B66
GL_FLOAT_MAT2x4_NV = 0x8B66
GL_FLOAT_MAT3x2 = 0x8B67
GL_FLOAT_MAT3x2_NV = 0x8B67
GL_FLOAT_MAT3x4 = 0x8B68
GL_FLOAT_MAT3x4_NV = 0x8B68
GL_FLOAT_MAT4x2 = 0x8B69
GL_FLOAT_MAT4x2_NV = 0x8B69
GL_FLOAT_MAT4x3 = 0x8B6A
GL_FLOAT_MAT4x3_NV = 0x8B6A
GL_DELETE_STATUS = 0x8B80
GL_OBJECT_DELETE_STATUS_ARB = 0x8B80
GL_COMPILE_STATUS = 0x8B81
GL_OBJECT_COMPILE_STATUS_ARB = 0x8B81
GL_LINK_STATUS = 0x8B82
GL_OBJECT_LINK_STATUS_ARB = 0x8B82
GL_VALIDATE_STATUS = 0x8B83
GL_OBJECT_VALIDATE_STATUS_ARB = 0x8B83
GL_INFO_LOG_LENGTH = 0x8B84
GL_OBJECT_INFO_LOG_LENGTH_ARB = 0x8B84
GL_ATTACHED_SHADERS = 0x8B85
GL_OBJECT_ATTACHED_OBJECTS_ARB = 0x8B85
GL_ACTIVE_UNIFORMS = 0x8B86
GL_OBJECT_ACTIVE_UNIFORMS_ARB = 0x8B86
GL_ACTIVE_UNIFORM_MAX_LENGTH = 0x8B87
GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB = 0x8B87
GL_SHADER_SOURCE_LENGTH = 0x8B88
GL_OBJECT_SHADER_SOURCE_LENGTH_ARB = 0x8B88
GL_ACTIVE_ATTRIBUTES = 0x8B89
GL_OBJECT_ACTIVE_ATTRIBUTES_ARB = 0x8B89
GL_ACTIVE_ATTRIBUTE_MAX_LENGTH = 0x8B8A
GL_OBJECT_ACTIVE_ATTRIBUTE_MAX_LENGTH_ARB = 0x8B8A
GL_FRAGMENT_SHADER_DERIVATIVE_HINT = 0x8B8B
GL_FRAGMENT_SHADER_DERIVATIVE_HINT_ARB = 0x8B8B
GL_FRAGMENT_SHADER_DERIVATIVE_HINT_OES = 0x8B8B
GL_SHADING_LANGUAGE_VERSION = 0x8B8C
GL_SHADING_LANGUAGE_VERSION_ARB = 0x8B8C
GL_CURRENT_PROGRAM = 0x8B8D
GL_ACTIVE_PROGRAM_EXT = 0x8B8D
GL_PALETTE4_RGB8_OES = 0x8B90
GL_PALETTE4_RGBA8_OES = 0x8B91
GL_PALETTE4_R5_G6_B5_OES = 0x8B92
GL_PALETTE4_RGBA4_OES = 0x8B93
GL_PALETTE4_RGB5_A1_OES = 0x8B94
GL_PALETTE8_RGB8_OES = 0x8B95
GL_PALETTE8_RGBA8_OES = 0x8B96
GL_PALETTE8_R5_G6_B5_OES = 0x8B97
GL_PALETTE8_RGBA4_OES = 0x8B98
GL_PALETTE8_RGB5_A1_OES = 0x8B99
GL_IMPLEMENTATION_COLOR_READ_TYPE = 0x8B9A
GL_IMPLEMENTATION_COLOR_READ_TYPE_OES = 0x8B9A
GL_IMPLEMENTATION_COLOR_READ_FORMAT = 0x8B9B
GL_IMPLEMENTATION_COLOR_READ_FORMAT_OES = 0x8B9B
GL_POINT_SIZE_ARRAY_OES = 0x8B9C
GL_TEXTURE_CROP_RECT_OES = 0x8B9D
GL_MATRIX_INDEX_ARRAY_BUFFER_BINDING_OES = 0x8B9E
GL_POINT_SIZE_ARRAY_BUFFER_BINDING_OES = 0x8B9F
GL_FRAGMENT_PROGRAM_POSITION_MESA = 0x8BB0
GL_FRAGMENT_PROGRAM_CALLBACK_MESA = 0x8BB1
GL_FRAGMENT_PROGRAM_CALLBACK_FUNC_MESA = 0x8BB2
GL_FRAGMENT_PROGRAM_CALLBACK_DATA_MESA = 0x8BB3
GL_VERTEX_PROGRAM_POSITION_MESA = 0x8BB4
GL_VERTEX_PROGRAM_CALLBACK_MESA = 0x8BB5
GL_VERTEX_PROGRAM_CALLBACK_FUNC_MESA = 0x8BB6
GL_VERTEX_PROGRAM_CALLBACK_DATA_MESA = 0x8BB7
GL_COUNTER_TYPE_AMD = 0x8BC0
GL_COUNTER_RANGE_AMD = 0x8BC1
GL_UNSIGNED_INT64_AMD = 0x8BC2
GL_PERCENTAGE_AMD = 0x8BC3
GL_PERFMON_RESULT_AVAILABLE_AMD = 0x8BC4
GL_PERFMON_RESULT_SIZE_AMD = 0x8BC5
GL_PERFMON_RESULT_AMD = 0x8BC6
GL_TEXTURE_WIDTH_QCOM = 0x8BD2
GL_TEXTURE_HEIGHT_QCOM = 0x8BD3
GL_TEXTURE_DEPTH_QCOM = 0x8BD4
GL_TEXTURE_INTERNAL_FORMAT_QCOM = 0x8BD5
GL_TEXTURE_FORMAT_QCOM = 0x8BD6
GL_TEXTURE_TYPE_QCOM = 0x8BD7
GL_TEXTURE_IMAGE_VALID_QCOM = 0x8BD8
GL_TEXTURE_NUM_LEVELS_QCOM = 0x8BD9
GL_TEXTURE_TARGET_QCOM = 0x8BDA
GL_TEXTURE_OBJECT_VALID_QCOM = 0x8BDB
GL_STATE_RESTORE = 0x8BDC
GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG = 0x8C00
GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG = 0x8C01
GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG = 0x8C02
GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG = 0x8C03
GL_MODULATE_COLOR_IMG = 0x8C04
GL_RECIP_ADD_SIGNED_ALPHA_IMG = 0x8C05
GL_TEXTURE_ALPHA_MODULATE_IMG = 0x8C06
GL_FACTOR_ALPHA_MODULATE_IMG = 0x8C07
GL_FRAGMENT_ALPHA_MODULATE_IMG = 0x8C08
GL_ADD_BLEND_IMG = 0x8C09
GL_SGX_BINARY_IMG = 0x8C0A
GL_TEXTURE_RED_TYPE = 0x8C10
GL_TEXTURE_RED_TYPE_ARB = 0x8C10
GL_TEXTURE_GREEN_TYPE = 0x8C11
GL_TEXTURE_GREEN_TYPE_ARB = 0x8C11
GL_TEXTURE_BLUE_TYPE = 0x8C12
GL_TEXTURE_BLUE_TYPE_ARB = 0x8C12
GL_TEXTURE_ALPHA_TYPE = 0x8C13
GL_TEXTURE_ALPHA_TYPE_ARB = 0x8C13
GL_TEXTURE_LUMINANCE_TYPE = 0x8C14
GL_TEXTURE_LUMINANCE_TYPE_ARB = 0x8C14
GL_TEXTURE_INTENSITY_TYPE = 0x8C15
GL_TEXTURE_INTENSITY_TYPE_ARB = 0x8C15
GL_TEXTURE_DEPTH_TYPE = 0x8C16
GL_TEXTURE_DEPTH_TYPE_ARB = 0x8C16
GL_UNSIGNED_NORMALIZED = 0x8C17
GL_UNSIGNED_NORMALIZED_ARB = 0x8C17
GL_UNSIGNED_NORMALIZED_EXT = 0x8C17
GL_TEXTURE_1D_ARRAY = 0x8C18
GL_TEXTURE_1D_ARRAY_EXT = 0x8C18
GL_PROXY_TEXTURE_1D_ARRAY = 0x8C19
GL_PROXY_TEXTURE_1D_ARRAY_EXT = 0x8C19
GL_TEXTURE_2D_ARRAY = 0x8C1A
GL_TEXTURE_2D_ARRAY_EXT = 0x8C1A
GL_PROXY_TEXTURE_2D_ARRAY = 0x8C1B
GL_PROXY_TEXTURE_2D_ARRAY_EXT = 0x8C1B
GL_TEXTURE_BINDING_1D_ARRAY = 0x8C1C
GL_TEXTURE_BINDING_1D_ARRAY_EXT = 0x8C1C
GL_TEXTURE_BINDING_2D_ARRAY = 0x8C1D
GL_TEXTURE_BINDING_2D_ARRAY_EXT = 0x8C1D
GL_GEOMETRY_PROGRAM_NV = 0x8C26
GL_MAX_PROGRAM_OUTPUT_VERTICES_NV = 0x8C27
GL_MAX_PROGRAM_TOTAL_OUTPUT_COMPONENTS_NV = 0x8C28
GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS = 0x8C29
GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS_ARB = 0x8C29
GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS_EXT = 0x8C29
GL_TEXTURE_BUFFER = 0x8C2A
GL_TEXTURE_BUFFER_ARB = 0x8C2A
GL_TEXTURE_BUFFER_EXT = 0x8C2A
GL_TEXTURE_BUFFER_BINDING = 0x8C2A
GL_TEXTURE_BUFFER_BINDING_EXT = 0x8C2A
GL_MAX_TEXTURE_BUFFER_SIZE = 0x8C2B
GL_MAX_TEXTURE_BUFFER_SIZE_ARB = 0x8C2B
GL_MAX_TEXTURE_BUFFER_SIZE_EXT = 0x8C2B
GL_TEXTURE_BINDING_BUFFER = 0x8C2C
GL_TEXTURE_BINDING_BUFFER_ARB = 0x8C2C
GL_TEXTURE_BINDING_BUFFER_EXT = 0x8C2C
GL_TEXTURE_BUFFER_DATA_STORE_BINDING = 0x8C2D
GL_TEXTURE_BUFFER_DATA_STORE_BINDING_ARB = 0x8C2D
GL_TEXTURE_BUFFER_DATA_STORE_BINDING_EXT = 0x8C2D
GL_TEXTURE_BUFFER_FORMAT_ARB = 0x8C2E
GL_TEXTURE_BUFFER_FORMAT_EXT = 0x8C2E
GL_ANY_SAMPLES_PASSED = 0x8C2F
GL_ANY_SAMPLES_PASSED_EXT = 0x8C2F
GL_SAMPLE_SHADING = 0x8C36
GL_SAMPLE_SHADING_ARB = 0x8C36
GL_SAMPLE_SHADING_OES = 0x8C36
GL_MIN_SAMPLE_SHADING_VALUE = 0x8C37
GL_MIN_SAMPLE_SHADING_VALUE_ARB = 0x8C37
GL_MIN_SAMPLE_SHADING_VALUE_OES = 0x8C37
GL_R11F_G11F_B10F = 0x8C3A
GL_R11F_G11F_B10F_APPLE = 0x8C3A
GL_R11F_G11F_B10F_EXT = 0x8C3A
GL_UNSIGNED_INT_10F_11F_11F_REV = 0x8C3B
GL_UNSIGNED_INT_10F_11F_11F_REV_APPLE = 0x8C3B
GL_UNSIGNED_INT_10F_11F_11F_REV_EXT = 0x8C3B
GL_RGBA_SIGNED_COMPONENTS_EXT = 0x8C3C
GL_RGB9_E5 = 0x8C3D
GL_RGB9_E5_APPLE = 0x8C3D
GL_RGB9_E5_EXT = 0x8C3D
GL_UNSIGNED_INT_5_9_9_9_REV = 0x8C3E
GL_UNSIGNED_INT_5_9_9_9_REV_APPLE = 0x8C3E
GL_UNSIGNED_INT_5_9_9_9_REV_EXT = 0x8C3E
GL_TEXTURE_SHARED_SIZE = 0x8C3F
GL_TEXTURE_SHARED_SIZE_EXT = 0x8C3F
GL_SRGB = 0x8C40
GL_SRGB_EXT = 0x8C40
GL_SRGB8 = 0x8C41
GL_SRGB8_EXT = 0x8C41
GL_SRGB8_NV = 0x8C41
GL_SRGB_ALPHA = 0x8C42
GL_SRGB_ALPHA_EXT = 0x8C42
GL_SRGB8_ALPHA8 = 0x8C43
GL_SRGB8_ALPHA8_EXT = 0x8C43
GL_SLUMINANCE_ALPHA = 0x8C44
GL_SLUMINANCE_ALPHA_EXT = 0x8C44
GL_SLUMINANCE_ALPHA_NV = 0x8C44
GL_SLUMINANCE8_ALPHA8 = 0x8C45
GL_SLUMINANCE8_ALPHA8_EXT = 0x8C45
GL_SLUMINANCE8_ALPHA8_NV = 0x8C45
GL_SLUMINANCE = 0x8C46
GL_SLUMINANCE_EXT = 0x8C46
GL_SLUMINANCE_NV = 0x8C46
GL_SLUMINANCE8 = 0x8C47
GL_SLUMINANCE8_EXT = 0x8C47
GL_SLUMINANCE8_NV = 0x8C47
GL_COMPRESSED_SRGB = 0x8C48
GL_COMPRESSED_SRGB_EXT = 0x8C48
GL_COMPRESSED_SRGB_ALPHA = 0x8C49
GL_COMPRESSED_SRGB_ALPHA_EXT = 0x8C49
GL_COMPRESSED_SLUMINANCE = 0x8C4A
GL_COMPRESSED_SLUMINANCE_EXT = 0x8C4A
GL_COMPRESSED_SLUMINANCE_ALPHA = 0x8C4B
GL_COMPRESSED_SLUMINANCE_ALPHA_EXT = 0x8C4B
GL_COMPRESSED_SRGB_S3TC_DXT1_EXT = 0x8C4C
GL_COMPRESSED_SRGB_S3TC_DXT1_NV = 0x8C4C
GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT1_EXT = 0x8C4D
GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT1_NV = 0x8C4D
GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT3_EXT = 0x8C4E
GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT3_NV = 0x8C4E
GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT5_EXT = 0x8C4F
GL_COMPRESSED_SRGB_ALPHA_S3TC_DXT5_NV = 0x8C4F
GL_COMPRESSED_LUMINANCE_LATC1_EXT = 0x8C70
GL_COMPRESSED_SIGNED_LUMINANCE_LATC1_EXT = 0x8C71
GL_COMPRESSED_LUMINANCE_ALPHA_LATC2_EXT = 0x8C72
GL_COMPRESSED_SIGNED_LUMINANCE_ALPHA_LATC2_EXT = 0x8C73
GL_TESS_CONTROL_PROGRAM_PARAMETER_BUFFER_NV = 0x8C74
GL_TESS_EVALUATION_PROGRAM_PARAMETER_BUFFER_NV = 0x8C75
GL_TRANSFORM_FEEDBACK_VARYING_MAX_LENGTH = 0x8C76
GL_TRANSFORM_FEEDBACK_VARYING_MAX_LENGTH_EXT = 0x8C76
GL_BACK_PRIMARY_COLOR_NV = 0x8C77
GL_BACK_SECONDARY_COLOR_NV = 0x8C78
GL_TEXTURE_COORD_NV = 0x8C79
GL_CLIP_DISTANCE_NV = 0x8C7A
GL_VERTEX_ID_NV = 0x8C7B
GL_PRIMITIVE_ID_NV = 0x8C7C
GL_GENERIC_ATTRIB_NV = 0x8C7D
GL_TRANSFORM_FEEDBACK_ATTRIBS_NV = 0x8C7E
GL_TRANSFORM_FEEDBACK_BUFFER_MODE = 0x8C7F
GL_TRANSFORM_FEEDBACK_BUFFER_MODE_EXT = 0x8C7F
GL_TRANSFORM_FEEDBACK_BUFFER_MODE_NV = 0x8C7F
GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS = 0x8C80
GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS_EXT = 0x8C80
GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_COMPONENTS_NV = 0x8C80
GL_ACTIVE_VARYINGS_NV = 0x8C81
GL_ACTIVE_VARYING_MAX_LENGTH_NV = 0x8C82
GL_TRANSFORM_FEEDBACK_VARYINGS = 0x8C83
GL_TRANSFORM_FEEDBACK_VARYINGS_EXT = 0x8C83
GL_TRANSFORM_FEEDBACK_VARYINGS_NV = 0x8C83
GL_TRANSFORM_FEEDBACK_BUFFER_START = 0x8C84
GL_TRANSFORM_FEEDBACK_BUFFER_START_EXT = 0x8C84
GL_TRANSFORM_FEEDBACK_BUFFER_START_NV = 0x8C84
GL_TRANSFORM_FEEDBACK_BUFFER_SIZE = 0x8C85
GL_TRANSFORM_FEEDBACK_BUFFER_SIZE_EXT = 0x8C85
GL_TRANSFORM_FEEDBACK_BUFFER_SIZE_NV = 0x8C85
GL_TRANSFORM_FEEDBACK_RECORD_NV = 0x8C86
GL_PRIMITIVES_GENERATED = 0x8C87
GL_PRIMITIVES_GENERATED_EXT = 0x8C87
GL_PRIMITIVES_GENERATED_NV = 0x8C87
GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN = 0x8C88
GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_EXT = 0x8C88
GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV = 0x8C88
GL_RASTERIZER_DISCARD = 0x8C89
GL_RASTERIZER_DISCARD_EXT = 0x8C89
GL_RASTERIZER_DISCARD_NV = 0x8C89
GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS = 0x8C8A
GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS_EXT = 0x8C8A
GL_MAX_TRANSFORM_FEEDBACK_INTERLEAVED_COMPONENTS_NV = 0x8C8A
GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS = 0x8C8B
GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS_EXT = 0x8C8B
GL_MAX_TRANSFORM_FEEDBACK_SEPARATE_ATTRIBS_NV = 0x8C8B
GL_INTERLEAVED_ATTRIBS = 0x8C8C
GL_INTERLEAVED_ATTRIBS_EXT = 0x8C8C
GL_INTERLEAVED_ATTRIBS_NV = 0x8C8C
GL_SEPARATE_ATTRIBS = 0x8C8D
GL_SEPARATE_ATTRIBS_EXT = 0x8C8D
GL_SEPARATE_ATTRIBS_NV = 0x8C8D
GL_TRANSFORM_FEEDBACK_BUFFER = 0x8C8E
GL_TRANSFORM_FEEDBACK_BUFFER_EXT = 0x8C8E
GL_TRANSFORM_FEEDBACK_BUFFER_NV = 0x8C8E
GL_TRANSFORM_FEEDBACK_BUFFER_BINDING = 0x8C8F
GL_TRANSFORM_FEEDBACK_BUFFER_BINDING_EXT = 0x8C8F
GL_TRANSFORM_FEEDBACK_BUFFER_BINDING_NV = 0x8C8F
GL_ATC_RGB_AMD = 0x8C92
GL_ATC_RGBA_EXPLICIT_ALPHA_AMD = 0x8C93
GL_POINT_SPRITE_COORD_ORIGIN = 0x8CA0
GL_LOWER_LEFT = 0x8CA1
GL_UPPER_LEFT = 0x8CA2
GL_STENCIL_BACK_REF = 0x8CA3
GL_STENCIL_BACK_VALUE_MASK = 0x8CA4
GL_STENCIL_BACK_WRITEMASK = 0x8CA5
GL_DRAW_FRAMEBUFFER_BINDING = 0x8CA6
GL_DRAW_FRAMEBUFFER_BINDING_ANGLE = 0x8CA6
GL_DRAW_FRAMEBUFFER_BINDING_APPLE = 0x8CA6
GL_DRAW_FRAMEBUFFER_BINDING_EXT = 0x8CA6
GL_DRAW_FRAMEBUFFER_BINDING_NV = 0x8CA6
GL_FRAMEBUFFER_BINDING = 0x8CA6
GL_FRAMEBUFFER_BINDING_ANGLE = 0x8CA6
GL_FRAMEBUFFER_BINDING_EXT = 0x8CA6
GL_FRAMEBUFFER_BINDING_OES = 0x8CA6
GL_RENDERBUFFER_BINDING = 0x8CA7
GL_RENDERBUFFER_BINDING_ANGLE = 0x8CA7
GL_RENDERBUFFER_BINDING_EXT = 0x8CA7
GL_RENDERBUFFER_BINDING_OES = 0x8CA7
GL_READ_FRAMEBUFFER = 0x8CA8
GL_READ_FRAMEBUFFER_ANGLE = 0x8CA8
GL_READ_FRAMEBUFFER_APPLE = 0x8CA8
GL_READ_FRAMEBUFFER_EXT = 0x8CA8
GL_READ_FRAMEBUFFER_NV = 0x8CA8
GL_DRAW_FRAMEBUFFER = 0x8CA9
GL_DRAW_FRAMEBUFFER_ANGLE = 0x8CA9
GL_DRAW_FRAMEBUFFER_APPLE = 0x8CA9
GL_DRAW_FRAMEBUFFER_EXT = 0x8CA9
GL_DRAW_FRAMEBUFFER_NV = 0x8CA9
GL_READ_FRAMEBUFFER_BINDING = 0x8CAA
GL_READ_FRAMEBUFFER_BINDING_ANGLE = 0x8CAA
GL_READ_FRAMEBUFFER_BINDING_APPLE = 0x8CAA
GL_READ_FRAMEBUFFER_BINDING_EXT = 0x8CAA
GL_READ_FRAMEBUFFER_BINDING_NV = 0x8CAA
GL_RENDERBUFFER_COVERAGE_SAMPLES_NV = 0x8CAB
GL_RENDERBUFFER_SAMPLES = 0x8CAB
GL_RENDERBUFFER_SAMPLES_ANGLE = 0x8CAB
GL_RENDERBUFFER_SAMPLES_APPLE = 0x8CAB
GL_RENDERBUFFER_SAMPLES_EXT = 0x8CAB
GL_RENDERBUFFER_SAMPLES_NV = 0x8CAB
GL_DEPTH_COMPONENT32F = 0x8CAC
GL_DEPTH32F_STENCIL8 = 0x8CAD
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE = 0x8CD0
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE_EXT = 0x8CD0
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE_OES = 0x8CD0
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME = 0x8CD1
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME_EXT = 0x8CD1
GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME_OES = 0x8CD1
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL = 0x8CD2
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL_EXT = 0x8CD2
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL_OES = 0x8CD2
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE = 0x8CD3
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE_EXT = 0x8CD3
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE_OES = 0x8CD3
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_3D_ZOFFSET_EXT = 0x8CD4
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_3D_ZOFFSET_OES = 0x8CD4
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER = 0x8CD4
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER_EXT = 0x8CD4
GL_FRAMEBUFFER_COMPLETE = 0x8CD5
GL_FRAMEBUFFER_COMPLETE_EXT = 0x8CD5
GL_FRAMEBUFFER_COMPLETE_OES = 0x8CD5
GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT = 0x8CD6
GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT = 0x8CD6
GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_OES = 0x8CD6
GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT = 0x8CD7
GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT = 0x8CD7
GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_OES = 0x8CD7
GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS = 0x8CD9
GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT = 0x8CD9
GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_OES = 0x8CD9
GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT = 0x8CDA
GL_FRAMEBUFFER_INCOMPLETE_FORMATS_OES = 0x8CDA
GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER = 0x8CDB
GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT = 0x8CDB
GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_OES = 0x8CDB
GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER = 0x8CDC
GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT = 0x8CDC
GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_OES = 0x8CDC
GL_FRAMEBUFFER_UNSUPPORTED = 0x8CDD
GL_FRAMEBUFFER_UNSUPPORTED_EXT = 0x8CDD
GL_FRAMEBUFFER_UNSUPPORTED_OES = 0x8CDD
GL_MAX_COLOR_ATTACHMENTS = 0x8CDF
GL_MAX_COLOR_ATTACHMENTS_EXT = 0x8CDF
GL_MAX_COLOR_ATTACHMENTS_NV = 0x8CDF
GL_COLOR_ATTACHMENT0 = 0x8CE0
GL_COLOR_ATTACHMENT0_EXT = 0x8CE0
GL_COLOR_ATTACHMENT0_NV = 0x8CE0
GL_COLOR_ATTACHMENT0_OES = 0x8CE0
GL_COLOR_ATTACHMENT1 = 0x8CE1
GL_COLOR_ATTACHMENT1_EXT = 0x8CE1
GL_COLOR_ATTACHMENT1_NV = 0x8CE1
GL_COLOR_ATTACHMENT2 = 0x8CE2
GL_COLOR_ATTACHMENT2_EXT = 0x8CE2
GL_COLOR_ATTACHMENT2_NV = 0x8CE2
GL_COLOR_ATTACHMENT3 = 0x8CE3
GL_COLOR_ATTACHMENT3_EXT = 0x8CE3
GL_COLOR_ATTACHMENT3_NV = 0x8CE3
GL_COLOR_ATTACHMENT4 = 0x8CE4
GL_COLOR_ATTACHMENT4_EXT = 0x8CE4
GL_COLOR_ATTACHMENT4_NV = 0x8CE4
GL_COLOR_ATTACHMENT5 = 0x8CE5
GL_COLOR_ATTACHMENT5_EXT = 0x8CE5
GL_COLOR_ATTACHMENT5_NV = 0x8CE5
GL_COLOR_ATTACHMENT6 = 0x8CE6
GL_COLOR_ATTACHMENT6_EXT = 0x8CE6
GL_COLOR_ATTACHMENT6_NV = 0x8CE6
GL_COLOR_ATTACHMENT7 = 0x8CE7
GL_COLOR_ATTACHMENT7_EXT = 0x8CE7
GL_COLOR_ATTACHMENT7_NV = 0x8CE7
GL_COLOR_ATTACHMENT8 = 0x8CE8
GL_COLOR_ATTACHMENT8_EXT = 0x8CE8
GL_COLOR_ATTACHMENT8_NV = 0x8CE8
GL_COLOR_ATTACHMENT9 = 0x8CE9
GL_COLOR_ATTACHMENT9_EXT = 0x8CE9
GL_COLOR_ATTACHMENT9_NV = 0x8CE9
GL_COLOR_ATTACHMENT10 = 0x8CEA
GL_COLOR_ATTACHMENT10_EXT = 0x8CEA
GL_COLOR_ATTACHMENT10_NV = 0x8CEA
GL_COLOR_ATTACHMENT11 = 0x8CEB
GL_COLOR_ATTACHMENT11_EXT = 0x8CEB
GL_COLOR_ATTACHMENT11_NV = 0x8CEB
GL_COLOR_ATTACHMENT12 = 0x8CEC
GL_COLOR_ATTACHMENT12_EXT = 0x8CEC
GL_COLOR_ATTACHMENT12_NV = 0x8CEC
GL_COLOR_ATTACHMENT13 = 0x8CED
GL_COLOR_ATTACHMENT13_EXT = 0x8CED
GL_COLOR_ATTACHMENT13_NV = 0x8CED
GL_COLOR_ATTACHMENT14 = 0x8CEE
GL_COLOR_ATTACHMENT14_EXT = 0x8CEE
GL_COLOR_ATTACHMENT14_NV = 0x8CEE
GL_COLOR_ATTACHMENT15 = 0x8CEF
GL_COLOR_ATTACHMENT15_EXT = 0x8CEF
GL_COLOR_ATTACHMENT15_NV = 0x8CEF
GL_DEPTH_ATTACHMENT = 0x8D00
GL_DEPTH_ATTACHMENT_EXT = 0x8D00
GL_DEPTH_ATTACHMENT_OES = 0x8D00
GL_STENCIL_ATTACHMENT = 0x8D20
GL_STENCIL_ATTACHMENT_EXT = 0x8D20
GL_STENCIL_ATTACHMENT_OES = 0x8D20
GL_FRAMEBUFFER = 0x8D40
GL_FRAMEBUFFER_EXT = 0x8D40
GL_FRAMEBUFFER_OES = 0x8D40
GL_RENDERBUFFER = 0x8D41
GL_RENDERBUFFER_EXT = 0x8D41
GL_RENDERBUFFER_OES = 0x8D41
GL_RENDERBUFFER_WIDTH = 0x8D42
GL_RENDERBUFFER_WIDTH_EXT = 0x8D42
GL_RENDERBUFFER_WIDTH_OES = 0x8D42
GL_RENDERBUFFER_HEIGHT = 0x8D43
GL_RENDERBUFFER_HEIGHT_EXT = 0x8D43
GL_RENDERBUFFER_HEIGHT_OES = 0x8D43
GL_RENDERBUFFER_INTERNAL_FORMAT = 0x8D44
GL_RENDERBUFFER_INTERNAL_FORMAT_EXT = 0x8D44
GL_RENDERBUFFER_INTERNAL_FORMAT_OES = 0x8D44
GL_STENCIL_INDEX1 = 0x8D46
GL_STENCIL_INDEX1_EXT = 0x8D46
GL_STENCIL_INDEX1_OES = 0x8D46
GL_STENCIL_INDEX4 = 0x8D47
GL_STENCIL_INDEX4_EXT = 0x8D47
GL_STENCIL_INDEX4_OES = 0x8D47
GL_STENCIL_INDEX8 = 0x8D48
GL_STENCIL_INDEX8_EXT = 0x8D48
GL_STENCIL_INDEX8_OES = 0x8D48
GL_STENCIL_INDEX16 = 0x8D49
GL_STENCIL_INDEX16_EXT = 0x8D49
GL_RENDERBUFFER_RED_SIZE = 0x8D50
GL_RENDERBUFFER_RED_SIZE_EXT = 0x8D50
GL_RENDERBUFFER_RED_SIZE_OES = 0x8D50
GL_RENDERBUFFER_GREEN_SIZE = 0x8D51
GL_RENDERBUFFER_GREEN_SIZE_EXT = 0x8D51
GL_RENDERBUFFER_GREEN_SIZE_OES = 0x8D51
GL_RENDERBUFFER_BLUE_SIZE = 0x8D52
GL_RENDERBUFFER_BLUE_SIZE_EXT = 0x8D52
GL_RENDERBUFFER_BLUE_SIZE_OES = 0x8D52
GL_RENDERBUFFER_ALPHA_SIZE = 0x8D53
GL_RENDERBUFFER_ALPHA_SIZE_EXT = 0x8D53
GL_RENDERBUFFER_ALPHA_SIZE_OES = 0x8D53
GL_RENDERBUFFER_DEPTH_SIZE = 0x8D54
GL_RENDERBUFFER_DEPTH_SIZE_EXT = 0x8D54
GL_RENDERBUFFER_DEPTH_SIZE_OES = 0x8D54
GL_RENDERBUFFER_STENCIL_SIZE = 0x8D55
GL_RENDERBUFFER_STENCIL_SIZE_EXT = 0x8D55
GL_RENDERBUFFER_STENCIL_SIZE_OES = 0x8D55
GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE = 0x8D56
GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_ANGLE = 0x8D56
GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_APPLE = 0x8D56
GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_EXT = 0x8D56
GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_NV = 0x8D56
GL_MAX_SAMPLES = 0x8D57
GL_MAX_SAMPLES_ANGLE = 0x8D57
GL_MAX_SAMPLES_APPLE = 0x8D57
GL_MAX_SAMPLES_EXT = 0x8D57
GL_MAX_SAMPLES_NV = 0x8D57
GL_TEXTURE_GEN_STR_OES = 0x8D60
GL_HALF_FLOAT_OES = 0x8D61
GL_RGB565_OES = 0x8D62
GL_RGB565 = 0x8D62
GL_ETC1_RGB8_OES = 0x8D64
GL_TEXTURE_EXTERNAL_OES = 0x8D65
GL_SAMPLER_EXTERNAL_OES = 0x8D66
GL_TEXTURE_BINDING_EXTERNAL_OES = 0x8D67
GL_REQUIRED_TEXTURE_IMAGE_UNITS_OES = 0x8D68
GL_PRIMITIVE_RESTART_FIXED_INDEX = 0x8D69
GL_ANY_SAMPLES_PASSED_CONSERVATIVE = 0x8D6A
GL_ANY_SAMPLES_PASSED_CONSERVATIVE_EXT = 0x8D6A
GL_MAX_ELEMENT_INDEX = 0x8D6B
GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_SAMPLES_EXT = 0x8D6C
GL_RGBA32UI = 0x8D70
GL_RGBA32UI_EXT = 0x8D70
GL_RGB32UI = 0x8D71
GL_RGB32UI_EXT = 0x8D71
GL_ALPHA32UI_EXT = 0x8D72
GL_INTENSITY32UI_EXT = 0x8D73
GL_LUMINANCE32UI_EXT = 0x8D74
GL_LUMINANCE_ALPHA32UI_EXT = 0x8D75
GL_RGBA16UI = 0x8D76
GL_RGBA16UI_EXT = 0x8D76
GL_RGB16UI = 0x8D77
GL_RGB16UI_EXT = 0x8D77
GL_ALPHA16UI_EXT = 0x8D78
GL_INTENSITY16UI_EXT = 0x8D79
GL_LUMINANCE16UI_EXT = 0x8D7A
GL_LUMINANCE_ALPHA16UI_EXT = 0x8D7B
GL_RGBA8UI = 0x8D7C
GL_RGBA8UI_EXT = 0x8D7C
GL_RGB8UI = 0x8D7D
GL_RGB8UI_EXT = 0x8D7D
GL_ALPHA8UI_EXT = 0x8D7E
GL_INTENSITY8UI_EXT = 0x8D7F
GL_LUMINANCE8UI_EXT = 0x8D80
GL_LUMINANCE_ALPHA8UI_EXT = 0x8D81
GL_RGBA32I = 0x8D82
GL_RGBA32I_EXT = 0x8D82
GL_RGB32I = 0x8D83
GL_RGB32I_EXT = 0x8D83
GL_ALPHA32I_EXT = 0x8D84
GL_INTENSITY32I_EXT = 0x8D85
GL_LUMINANCE32I_EXT = 0x8D86
GL_LUMINANCE_ALPHA32I_EXT = 0x8D87
GL_RGBA16I = 0x8D88
GL_RGBA16I_EXT = 0x8D88
GL_RGB16I = 0x8D89
GL_RGB16I_EXT = 0x8D89
GL_ALPHA16I_EXT = 0x8D8A
GL_INTENSITY16I_EXT = 0x8D8B
GL_LUMINANCE16I_EXT = 0x8D8C
GL_LUMINANCE_ALPHA16I_EXT = 0x8D8D
GL_RGBA8I = 0x8D8E
GL_RGBA8I_EXT = 0x8D8E
GL_RGB8I = 0x8D8F
GL_RGB8I_EXT = 0x8D8F
GL_ALPHA8I_EXT = 0x8D90
GL_INTENSITY8I_EXT = 0x8D91
GL_LUMINANCE8I_EXT = 0x8D92
GL_LUMINANCE_ALPHA8I_EXT = 0x8D93
GL_RED_INTEGER = 0x8D94
GL_RED_INTEGER_EXT = 0x8D94
GL_GREEN_INTEGER = 0x8D95
GL_GREEN_INTEGER_EXT = 0x8D95
GL_BLUE_INTEGER = 0x8D96
GL_BLUE_INTEGER_EXT = 0x8D96
GL_ALPHA_INTEGER = 0x8D97
GL_ALPHA_INTEGER_EXT = 0x8D97
GL_RGB_INTEGER = 0x8D98
GL_RGB_INTEGER_EXT = 0x8D98
GL_RGBA_INTEGER = 0x8D99
GL_RGBA_INTEGER_EXT = 0x8D99
GL_BGR_INTEGER = 0x8D9A
GL_BGR_INTEGER_EXT = 0x8D9A
GL_BGRA_INTEGER = 0x8D9B
GL_BGRA_INTEGER_EXT = 0x8D9B
GL_LUMINANCE_INTEGER_EXT = 0x8D9C
GL_LUMINANCE_ALPHA_INTEGER_EXT = 0x8D9D
GL_RGBA_INTEGER_MODE_EXT = 0x8D9E
GL_INT_2_10_10_10_REV = 0x8D9F
GL_MAX_PROGRAM_PARAMETER_BUFFER_BINDINGS_NV = 0x8DA0
GL_MAX_PROGRAM_PARAMETER_BUFFER_SIZE_NV = 0x8DA1
GL_VERTEX_PROGRAM_PARAMETER_BUFFER_NV = 0x8DA2
GL_GEOMETRY_PROGRAM_PARAMETER_BUFFER_NV = 0x8DA3
GL_FRAGMENT_PROGRAM_PARAMETER_BUFFER_NV = 0x8DA4
GL_MAX_PROGRAM_GENERIC_ATTRIBS_NV = 0x8DA5
GL_MAX_PROGRAM_GENERIC_RESULTS_NV = 0x8DA6
GL_FRAMEBUFFER_ATTACHMENT_LAYERED = 0x8DA7
GL_FRAMEBUFFER_ATTACHMENT_LAYERED_ARB = 0x8DA7
GL_FRAMEBUFFER_ATTACHMENT_LAYERED_EXT = 0x8DA7
GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS = 0x8DA8
GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_ARB = 0x8DA8
GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_EXT = 0x8DA8
GL_FRAMEBUFFER_INCOMPLETE_LAYER_COUNT_ARB = 0x8DA9
GL_FRAMEBUFFER_INCOMPLETE_LAYER_COUNT_EXT = 0x8DA9
GL_LAYER_NV = 0x8DAA
GL_DEPTH_COMPONENT32F_NV = 0x8DAB
GL_DEPTH32F_STENCIL8_NV = 0x8DAC
GL_FLOAT_32_UNSIGNED_INT_24_8_REV = 0x8DAD
GL_FLOAT_32_UNSIGNED_INT_24_8_REV_NV = 0x8DAD
GL_SHADER_INCLUDE_ARB = 0x8DAE
GL_DEPTH_BUFFER_FLOAT_MODE_NV = 0x8DAF
GL_FRAMEBUFFER_SRGB = 0x8DB9
GL_FRAMEBUFFER_SRGB_EXT = 0x8DB9
GL_FRAMEBUFFER_SRGB_CAPABLE_EXT = 0x8DBA
GL_COMPRESSED_RED_RGTC1 = 0x8DBB
GL_COMPRESSED_RED_RGTC1_EXT = 0x8DBB
GL_COMPRESSED_SIGNED_RED_RGTC1 = 0x8DBC
GL_COMPRESSED_SIGNED_RED_RGTC1_EXT = 0x8DBC
GL_COMPRESSED_RED_GREEN_RGTC2_EXT = 0x8DBD
GL_COMPRESSED_RG_RGTC2 = 0x8DBD
GL_COMPRESSED_SIGNED_RED_GREEN_RGTC2_EXT = 0x8DBE
GL_COMPRESSED_SIGNED_RG_RGTC2 = 0x8DBE
GL_SAMPLER_1D_ARRAY = 0x8DC0
GL_SAMPLER_1D_ARRAY_EXT = 0x8DC0
GL_SAMPLER_2D_ARRAY = 0x8DC1
GL_SAMPLER_2D_ARRAY_EXT = 0x8DC1
GL_SAMPLER_BUFFER = 0x8DC2
GL_SAMPLER_BUFFER_EXT = 0x8DC2
GL_SAMPLER_1D_ARRAY_SHADOW = 0x8DC3
GL_SAMPLER_1D_ARRAY_SHADOW_EXT = 0x8DC3
GL_SAMPLER_2D_ARRAY_SHADOW = 0x8DC4
GL_SAMPLER_2D_ARRAY_SHADOW_EXT = 0x8DC4
GL_SAMPLER_2D_ARRAY_SHADOW_NV = 0x8DC4
GL_SAMPLER_CUBE_SHADOW = 0x8DC5
GL_SAMPLER_CUBE_SHADOW_EXT = 0x8DC5
GL_SAMPLER_CUBE_SHADOW_NV = 0x8DC5
GL_UNSIGNED_INT_VEC2 = 0x8DC6
GL_UNSIGNED_INT_VEC2_EXT = 0x8DC6
GL_UNSIGNED_INT_VEC3 = 0x8DC7
GL_UNSIGNED_INT_VEC3_EXT = 0x8DC7
GL_UNSIGNED_INT_VEC4 = 0x8DC8
GL_UNSIGNED_INT_VEC4_EXT = 0x8DC8
GL_INT_SAMPLER_1D = 0x8DC9
GL_INT_SAMPLER_1D_EXT = 0x8DC9
GL_INT_SAMPLER_2D = 0x8DCA
GL_INT_SAMPLER_2D_EXT = 0x8DCA
GL_INT_SAMPLER_3D = 0x8DCB
GL_INT_SAMPLER_3D_EXT = 0x8DCB
GL_INT_SAMPLER_CUBE = 0x8DCC
GL_INT_SAMPLER_CUBE_EXT = 0x8DCC
GL_INT_SAMPLER_2D_RECT = 0x8DCD
GL_INT_SAMPLER_2D_RECT_EXT = 0x8DCD
GL_INT_SAMPLER_1D_ARRAY = 0x8DCE
GL_INT_SAMPLER_1D_ARRAY_EXT = 0x8DCE
GL_INT_SAMPLER_2D_ARRAY = 0x8DCF
GL_INT_SAMPLER_2D_ARRAY_EXT = 0x8DCF
GL_INT_SAMPLER_BUFFER = 0x8DD0
GL_INT_SAMPLER_BUFFER_EXT = 0x8DD0
GL_UNSIGNED_INT_SAMPLER_1D = 0x8DD1
GL_UNSIGNED_INT_SAMPLER_1D_EXT = 0x8DD1
GL_UNSIGNED_INT_SAMPLER_2D = 0x8DD2
GL_UNSIGNED_INT_SAMPLER_2D_EXT = 0x8DD2
GL_UNSIGNED_INT_SAMPLER_3D = 0x8DD3
GL_UNSIGNED_INT_SAMPLER_3D_EXT = 0x8DD3
GL_UNSIGNED_INT_SAMPLER_CUBE = 0x8DD4
GL_UNSIGNED_INT_SAMPLER_CUBE_EXT = 0x8DD4
GL_UNSIGNED_INT_SAMPLER_2D_RECT = 0x8DD5
GL_UNSIGNED_INT_SAMPLER_2D_RECT_EXT = 0x8DD5
GL_UNSIGNED_INT_SAMPLER_1D_ARRAY = 0x8DD6
GL_UNSIGNED_INT_SAMPLER_1D_ARRAY_EXT = 0x8DD6
GL_UNSIGNED_INT_SAMPLER_2D_ARRAY = 0x8DD7
GL_UNSIGNED_INT_SAMPLER_2D_ARRAY_EXT = 0x8DD7
GL_UNSIGNED_INT_SAMPLER_BUFFER = 0x8DD8
GL_UNSIGNED_INT_SAMPLER_BUFFER_EXT = 0x8DD8
GL_GEOMETRY_SHADER = 0x8DD9
GL_GEOMETRY_SHADER_ARB = 0x8DD9
GL_GEOMETRY_SHADER_EXT = 0x8DD9
GL_GEOMETRY_VERTICES_OUT_ARB = 0x8DDA
GL_GEOMETRY_VERTICES_OUT_EXT = 0x8DDA
GL_GEOMETRY_INPUT_TYPE_ARB = 0x8DDB
GL_GEOMETRY_INPUT_TYPE_EXT = 0x8DDB
GL_GEOMETRY_OUTPUT_TYPE_ARB = 0x8DDC
GL_GEOMETRY_OUTPUT_TYPE_EXT = 0x8DDC
GL_MAX_GEOMETRY_VARYING_COMPONENTS_ARB = 0x8DDD
GL_MAX_GEOMETRY_VARYING_COMPONENTS_EXT = 0x8DDD
GL_MAX_VERTEX_VARYING_COMPONENTS_ARB = 0x8DDE
GL_MAX_VERTEX_VARYING_COMPONENTS_EXT = 0x8DDE
GL_MAX_GEOMETRY_UNIFORM_COMPONENTS = 0x8DDF
GL_MAX_GEOMETRY_UNIFORM_COMPONENTS_ARB = 0x8DDF
GL_MAX_GEOMETRY_UNIFORM_COMPONENTS_EXT = 0x8DDF
GL_MAX_GEOMETRY_OUTPUT_VERTICES = 0x8DE0
GL_MAX_GEOMETRY_OUTPUT_VERTICES_ARB = 0x8DE0
GL_MAX_GEOMETRY_OUTPUT_VERTICES_EXT = 0x8DE0
GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS = 0x8DE1
GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS_ARB = 0x8DE1
GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS_EXT = 0x8DE1
GL_MAX_VERTEX_BINDABLE_UNIFORMS_EXT = 0x8DE2
GL_MAX_FRAGMENT_BINDABLE_UNIFORMS_EXT = 0x8DE3
GL_MAX_GEOMETRY_BINDABLE_UNIFORMS_EXT = 0x8DE4
GL_ACTIVE_SUBROUTINES = 0x8DE5
GL_ACTIVE_SUBROUTINE_UNIFORMS = 0x8DE6
GL_MAX_SUBROUTINES = 0x8DE7
GL_MAX_SUBROUTINE_UNIFORM_LOCATIONS = 0x8DE8
GL_NAMED_STRING_LENGTH_ARB = 0x8DE9
GL_NAMED_STRING_TYPE_ARB = 0x8DEA
GL_MAX_BINDABLE_UNIFORM_SIZE_EXT = 0x8DED
GL_UNIFORM_BUFFER_EXT = 0x8DEE
GL_UNIFORM_BUFFER_BINDING_EXT = 0x8DEF
GL_LOW_FLOAT = 0x8DF0
GL_MEDIUM_FLOAT = 0x8DF1
GL_HIGH_FLOAT = 0x8DF2
GL_LOW_INT = 0x8DF3
GL_MEDIUM_INT = 0x8DF4
GL_HIGH_INT = 0x8DF5
GL_UNSIGNED_INT_10_10_10_2_OES = 0x8DF6
GL_INT_10_10_10_2_OES = 0x8DF7
GL_SHADER_BINARY_FORMATS = 0x8DF8
GL_NUM_SHADER_BINARY_FORMATS = 0x8DF9
GL_SHADER_COMPILER = 0x8DFA
GL_MAX_VERTEX_UNIFORM_VECTORS = 0x8DFB
GL_MAX_VARYING_VECTORS = 0x8DFC
GL_MAX_FRAGMENT_UNIFORM_VECTORS = 0x8DFD
GL_RENDERBUFFER_COLOR_SAMPLES_NV = 0x8E10
GL_MAX_MULTISAMPLE_COVERAGE_MODES_NV = 0x8E11
GL_MULTISAMPLE_COVERAGE_MODES_NV = 0x8E12
GL_QUERY_WAIT = 0x8E13
GL_QUERY_WAIT_NV = 0x8E13
GL_QUERY_NO_WAIT = 0x8E14
GL_QUERY_NO_WAIT_NV = 0x8E14
GL_QUERY_BY_REGION_WAIT = 0x8E15
GL_QUERY_BY_REGION_WAIT_NV = 0x8E15
GL_QUERY_BY_REGION_NO_WAIT = 0x8E16
GL_QUERY_BY_REGION_NO_WAIT_NV = 0x8E16
GL_QUERY_WAIT_INVERTED = 0x8E17
GL_QUERY_NO_WAIT_INVERTED = 0x8E18
GL_QUERY_BY_REGION_WAIT_INVERTED = 0x8E19
GL_QUERY_BY_REGION_NO_WAIT_INVERTED = 0x8E1A
GL_POLYGON_OFFSET_CLAMP_EXT = 0x8E1B
GL_MAX_COMBINED_TESS_CONTROL_UNIFORM_COMPONENTS = 0x8E1E
GL_MAX_COMBINED_TESS_CONTROL_UNIFORM_COMPONENTS_EXT = 0x8E1E
GL_MAX_COMBINED_TESS_EVALUATION_UNIFORM_COMPONENTS = 0x8E1F
GL_MAX_COMBINED_TESS_EVALUATION_UNIFORM_COMPONENTS_EXT = 0x8E1F
GL_COLOR_SAMPLES_NV = 0x8E20
GL_TRANSFORM_FEEDBACK = 0x8E22
GL_TRANSFORM_FEEDBACK_NV = 0x8E22
GL_TRANSFORM_FEEDBACK_BUFFER_PAUSED = 0x8E23
GL_TRANSFORM_FEEDBACK_PAUSED = 0x8E23
GL_TRANSFORM_FEEDBACK_BUFFER_PAUSED_NV = 0x8E23
GL_TRANSFORM_FEEDBACK_BUFFER_ACTIVE = 0x8E24
GL_TRANSFORM_FEEDBACK_ACTIVE = 0x8E24
GL_TRANSFORM_FEEDBACK_BUFFER_ACTIVE_NV = 0x8E24
GL_TRANSFORM_FEEDBACK_BINDING = 0x8E25
GL_TRANSFORM_FEEDBACK_BINDING_NV = 0x8E25
GL_FRAME_NV = 0x8E26
GL_FIELDS_NV = 0x8E27
GL_CURRENT_TIME_NV = 0x8E28
GL_TIMESTAMP = 0x8E28
GL_TIMESTAMP_EXT = 0x8E28
GL_NUM_FILL_STREAMS_NV = 0x8E29
GL_PRESENT_TIME_NV = 0x8E2A
GL_PRESENT_DURATION_NV = 0x8E2B
GL_DEPTH_COMPONENT16_NONLINEAR_NV = 0x8E2C
GL_PROGRAM_MATRIX_EXT = 0x8E2D
GL_TRANSPOSE_PROGRAM_MATRIX_EXT = 0x8E2E
GL_PROGRAM_MATRIX_STACK_DEPTH_EXT = 0x8E2F
GL_TEXTURE_SWIZZLE_R = 0x8E42
GL_TEXTURE_SWIZZLE_R_EXT = 0x8E42
GL_TEXTURE_SWIZZLE_G = 0x8E43
GL_TEXTURE_SWIZZLE_G_EXT = 0x8E43
GL_TEXTURE_SWIZZLE_B = 0x8E44
GL_TEXTURE_SWIZZLE_B_EXT = 0x8E44
GL_TEXTURE_SWIZZLE_A = 0x8E45
GL_TEXTURE_SWIZZLE_A_EXT = 0x8E45
GL_TEXTURE_SWIZZLE_RGBA = 0x8E46
GL_TEXTURE_SWIZZLE_RGBA_EXT = 0x8E46
GL_ACTIVE_SUBROUTINE_UNIFORM_LOCATIONS = 0x8E47
GL_ACTIVE_SUBROUTINE_MAX_LENGTH = 0x8E48
GL_ACTIVE_SUBROUTINE_UNIFORM_MAX_LENGTH = 0x8E49
GL_NUM_COMPATIBLE_SUBROUTINES = 0x8E4A
GL_COMPATIBLE_SUBROUTINES = 0x8E4B
GL_QUADS_FOLLOW_PROVOKING_VERTEX_CONVENTION = 0x8E4C
GL_QUADS_FOLLOW_PROVOKING_VERTEX_CONVENTION_EXT = 0x8E4C
GL_FIRST_VERTEX_CONVENTION = 0x8E4D
GL_FIRST_VERTEX_CONVENTION_EXT = 0x8E4D
GL_LAST_VERTEX_CONVENTION = 0x8E4E
GL_LAST_VERTEX_CONVENTION_EXT = 0x8E4E
GL_PROVOKING_VERTEX = 0x8E4F
GL_PROVOKING_VERTEX_EXT = 0x8E4F
GL_SAMPLE_POSITION = 0x8E50
GL_SAMPLE_POSITION_NV = 0x8E50
GL_SAMPLE_LOCATION_NV = 0x8E50
GL_SAMPLE_MASK = 0x8E51
GL_SAMPLE_MASK_NV = 0x8E51
GL_SAMPLE_MASK_VALUE = 0x8E52
GL_SAMPLE_MASK_VALUE_NV = 0x8E52
GL_TEXTURE_BINDING_RENDERBUFFER_NV = 0x8E53
GL_TEXTURE_RENDERBUFFER_DATA_STORE_BINDING_NV = 0x8E54
GL_TEXTURE_RENDERBUFFER_NV = 0x8E55
GL_SAMPLER_RENDERBUFFER_NV = 0x8E56
GL_INT_SAMPLER_RENDERBUFFER_NV = 0x8E57
GL_UNSIGNED_INT_SAMPLER_RENDERBUFFER_NV = 0x8E58
GL_MAX_SAMPLE_MASK_WORDS = 0x8E59
GL_MAX_SAMPLE_MASK_WORDS_NV = 0x8E59
GL_MAX_GEOMETRY_PROGRAM_INVOCATIONS_NV = 0x8E5A
GL_MAX_GEOMETRY_SHADER_INVOCATIONS = 0x8E5A
GL_MAX_GEOMETRY_SHADER_INVOCATIONS_EXT = 0x8E5A
GL_MIN_FRAGMENT_INTERPOLATION_OFFSET = 0x8E5B
GL_MIN_FRAGMENT_INTERPOLATION_OFFSET_OES = 0x8E5B
GL_MIN_FRAGMENT_INTERPOLATION_OFFSET_NV = 0x8E5B
GL_MAX_FRAGMENT_INTERPOLATION_OFFSET = 0x8E5C
GL_MAX_FRAGMENT_INTERPOLATION_OFFSET_OES = 0x8E5C
GL_MAX_FRAGMENT_INTERPOLATION_OFFSET_NV = 0x8E5C
GL_FRAGMENT_INTERPOLATION_OFFSET_BITS = 0x8E5D
GL_FRAGMENT_INTERPOLATION_OFFSET_BITS_OES = 0x8E5D
GL_FRAGMENT_PROGRAM_INTERPOLATION_OFFSET_BITS_NV = 0x8E5D
GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET = 0x8E5E
GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET_ARB = 0x8E5E
GL_MIN_PROGRAM_TEXTURE_GATHER_OFFSET_NV = 0x8E5E
GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET = 0x8E5F
GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET_ARB = 0x8E5F
GL_MAX_PROGRAM_TEXTURE_GATHER_OFFSET_NV = 0x8E5F
GL_MAX_TRANSFORM_FEEDBACK_BUFFERS = 0x8E70
GL_MAX_VERTEX_STREAMS = 0x8E71
GL_PATCH_VERTICES = 0x8E72
GL_PATCH_VERTICES_EXT = 0x8E72
GL_PATCH_DEFAULT_INNER_LEVEL = 0x8E73
GL_PATCH_DEFAULT_INNER_LEVEL_EXT = 0x8E73
GL_PATCH_DEFAULT_OUTER_LEVEL = 0x8E74
GL_PATCH_DEFAULT_OUTER_LEVEL_EXT = 0x8E74
GL_TESS_CONTROL_OUTPUT_VERTICES = 0x8E75
GL_TESS_CONTROL_OUTPUT_VERTICES_EXT = 0x8E75
GL_TESS_GEN_MODE = 0x8E76
GL_TESS_GEN_MODE_EXT = 0x8E76
GL_TESS_GEN_SPACING = 0x8E77
GL_TESS_GEN_SPACING_EXT = 0x8E77
GL_TESS_GEN_VERTEX_ORDER = 0x8E78
GL_TESS_GEN_VERTEX_ORDER_EXT = 0x8E78
GL_TESS_GEN_POINT_MODE = 0x8E79
GL_TESS_GEN_POINT_MODE_EXT = 0x8E79
GL_ISOLINES = 0x8E7A
GL_ISOLINES_EXT = 0x8E7A
GL_FRACTIONAL_ODD = 0x8E7B
GL_FRACTIONAL_ODD_EXT = 0x8E7B
GL_FRACTIONAL_EVEN = 0x8E7C
GL_FRACTIONAL_EVEN_EXT = 0x8E7C
GL_MAX_PATCH_VERTICES = 0x8E7D
GL_MAX_PATCH_VERTICES_EXT = 0x8E7D
GL_MAX_TESS_GEN_LEVEL = 0x8E7E
GL_MAX_TESS_GEN_LEVEL_EXT = 0x8E7E
GL_MAX_TESS_CONTROL_UNIFORM_COMPONENTS = 0x8E7F
GL_MAX_TESS_CONTROL_UNIFORM_COMPONENTS_EXT = 0x8E7F
GL_MAX_TESS_EVALUATION_UNIFORM_COMPONENTS = 0x8E80
GL_MAX_TESS_EVALUATION_UNIFORM_COMPONENTS_EXT = 0x8E80
GL_MAX_TESS_CONTROL_TEXTURE_IMAGE_UNITS = 0x8E81
GL_MAX_TESS_CONTROL_TEXTURE_IMAGE_UNITS_EXT = 0x8E81
GL_MAX_TESS_EVALUATION_TEXTURE_IMAGE_UNITS = 0x8E82
GL_MAX_TESS_EVALUATION_TEXTURE_IMAGE_UNITS_EXT = 0x8E82
GL_MAX_TESS_CONTROL_OUTPUT_COMPONENTS = 0x8E83
GL_MAX_TESS_CONTROL_OUTPUT_COMPONENTS_EXT = 0x8E83
GL_MAX_TESS_PATCH_COMPONENTS = 0x8E84
GL_MAX_TESS_PATCH_COMPONENTS_EXT = 0x8E84
GL_MAX_TESS_CONTROL_TOTAL_OUTPUT_COMPONENTS = 0x8E85
GL_MAX_TESS_CONTROL_TOTAL_OUTPUT_COMPONENTS_EXT = 0x8E85
GL_MAX_TESS_EVALUATION_OUTPUT_COMPONENTS = 0x8E86
GL_MAX_TESS_EVALUATION_OUTPUT_COMPONENTS_EXT = 0x8E86
GL_TESS_EVALUATION_SHADER = 0x8E87
GL_TESS_EVALUATION_SHADER_EXT = 0x8E87
GL_TESS_CONTROL_SHADER = 0x8E88
GL_TESS_CONTROL_SHADER_EXT = 0x8E88
GL_MAX_TESS_CONTROL_UNIFORM_BLOCKS = 0x8E89
GL_MAX_TESS_CONTROL_UNIFORM_BLOCKS_EXT = 0x8E89
GL_MAX_TESS_EVALUATION_UNIFORM_BLOCKS = 0x8E8A
GL_MAX_TESS_EVALUATION_UNIFORM_BLOCKS_EXT = 0x8E8A
GL_COMPRESSED_RGBA_BPTC_UNORM = 0x8E8C
GL_COMPRESSED_RGBA_BPTC_UNORM_ARB = 0x8E8C
GL_COMPRESSED_SRGB_ALPHA_BPTC_UNORM = 0x8E8D
GL_COMPRESSED_SRGB_ALPHA_BPTC_UNORM_ARB = 0x8E8D
GL_COMPRESSED_RGB_BPTC_SIGNED_FLOAT = 0x8E8E
GL_COMPRESSED_RGB_BPTC_SIGNED_FLOAT_ARB = 0x8E8E
GL_COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT = 0x8E8F
GL_COMPRESSED_RGB_BPTC_UNSIGNED_FLOAT_ARB = 0x8E8F
GL_COVERAGE_COMPONENT_NV = 0x8ED0
GL_COVERAGE_COMPONENT4_NV = 0x8ED1
GL_COVERAGE_ATTACHMENT_NV = 0x8ED2
GL_COVERAGE_BUFFERS_NV = 0x8ED3
GL_COVERAGE_SAMPLES_NV = 0x8ED4
GL_COVERAGE_ALL_FRAGMENTS_NV = 0x8ED5
GL_COVERAGE_EDGE_FRAGMENTS_NV = 0x8ED6
GL_COVERAGE_AUTOMATIC_NV = 0x8ED7
GL_BUFFER_GPU_ADDRESS_NV = 0x8F1D
GL_VERTEX_ATTRIB_ARRAY_UNIFIED_NV = 0x8F1E
GL_ELEMENT_ARRAY_UNIFIED_NV = 0x8F1F
GL_VERTEX_ATTRIB_ARRAY_ADDRESS_NV = 0x8F20
GL_VERTEX_ARRAY_ADDRESS_NV = 0x8F21
GL_NORMAL_ARRAY_ADDRESS_NV = 0x8F22
GL_COLOR_ARRAY_ADDRESS_NV = 0x8F23
GL_INDEX_ARRAY_ADDRESS_NV = 0x8F24
GL_TEXTURE_COORD_ARRAY_ADDRESS_NV = 0x8F25
GL_EDGE_FLAG_ARRAY_ADDRESS_NV = 0x8F26
GL_SECONDARY_COLOR_ARRAY_ADDRESS_NV = 0x8F27
GL_FOG_COORD_ARRAY_ADDRESS_NV = 0x8F28
GL_ELEMENT_ARRAY_ADDRESS_NV = 0x8F29
GL_VERTEX_ATTRIB_ARRAY_LENGTH_NV = 0x8F2A
GL_VERTEX_ARRAY_LENGTH_NV = 0x8F2B
GL_NORMAL_ARRAY_LENGTH_NV = 0x8F2C
GL_COLOR_ARRAY_LENGTH_NV = 0x8F2D
GL_INDEX_ARRAY_LENGTH_NV = 0x8F2E
GL_TEXTURE_COORD_ARRAY_LENGTH_NV = 0x8F2F
GL_EDGE_FLAG_ARRAY_LENGTH_NV = 0x8F30
GL_SECONDARY_COLOR_ARRAY_LENGTH_NV = 0x8F31
GL_FOG_COORD_ARRAY_LENGTH_NV = 0x8F32
GL_ELEMENT_ARRAY_LENGTH_NV = 0x8F33
GL_GPU_ADDRESS_NV = 0x8F34
GL_MAX_SHADER_BUFFER_ADDRESS_NV = 0x8F35
GL_COPY_READ_BUFFER = 0x8F36
GL_COPY_READ_BUFFER_NV = 0x8F36
GL_COPY_READ_BUFFER_BINDING = 0x8F36
GL_COPY_WRITE_BUFFER = 0x8F37
GL_COPY_WRITE_BUFFER_NV = 0x8F37
GL_COPY_WRITE_BUFFER_BINDING = 0x8F37
GL_MAX_IMAGE_UNITS = 0x8F38
GL_MAX_IMAGE_UNITS_EXT = 0x8F38
GL_MAX_COMBINED_IMAGE_UNITS_AND_FRAGMENT_OUTPUTS = 0x8F39
GL_MAX_COMBINED_IMAGE_UNITS_AND_FRAGMENT_OUTPUTS_EXT = 0x8F39
GL_MAX_COMBINED_SHADER_OUTPUT_RESOURCES = 0x8F39
GL_IMAGE_BINDING_NAME = 0x8F3A
GL_IMAGE_BINDING_NAME_EXT = 0x8F3A
GL_IMAGE_BINDING_LEVEL = 0x8F3B
GL_IMAGE_BINDING_LEVEL_EXT = 0x8F3B
GL_IMAGE_BINDING_LAYERED = 0x8F3C
GL_IMAGE_BINDING_LAYERED_EXT = 0x8F3C
GL_IMAGE_BINDING_LAYER = 0x8F3D
GL_IMAGE_BINDING_LAYER_EXT = 0x8F3D
GL_IMAGE_BINDING_ACCESS = 0x8F3E
GL_IMAGE_BINDING_ACCESS_EXT = 0x8F3E
GL_DRAW_INDIRECT_BUFFER = 0x8F3F
GL_DRAW_INDIRECT_UNIFIED_NV = 0x8F40
GL_DRAW_INDIRECT_ADDRESS_NV = 0x8F41
GL_DRAW_INDIRECT_LENGTH_NV = 0x8F42
GL_DRAW_INDIRECT_BUFFER_BINDING = 0x8F43
GL_MAX_PROGRAM_SUBROUTINE_PARAMETERS_NV = 0x8F44
GL_MAX_PROGRAM_SUBROUTINE_NUM_NV = 0x8F45
GL_DOUBLE_MAT2 = 0x8F46
GL_DOUBLE_MAT2_EXT = 0x8F46
GL_DOUBLE_MAT3 = 0x8F47
GL_DOUBLE_MAT3_EXT = 0x8F47
GL_DOUBLE_MAT4 = 0x8F48
GL_DOUBLE_MAT4_EXT = 0x8F48
GL_DOUBLE_MAT2x3 = 0x8F49
GL_DOUBLE_MAT2x3_EXT = 0x8F49
GL_DOUBLE_MAT2x4 = 0x8F4A
GL_DOUBLE_MAT2x4_EXT = 0x8F4A
GL_DOUBLE_MAT3x2 = 0x8F4B
GL_DOUBLE_MAT3x2_EXT = 0x8F4B
GL_DOUBLE_MAT3x4 = 0x8F4C
GL_DOUBLE_MAT3x4_EXT = 0x8F4C
GL_DOUBLE_MAT4x2 = 0x8F4D
GL_DOUBLE_MAT4x2_EXT = 0x8F4D
GL_DOUBLE_MAT4x3 = 0x8F4E
GL_DOUBLE_MAT4x3_EXT = 0x8F4E
GL_VERTEX_BINDING_BUFFER = 0x8F4F
GL_MALI_SHADER_BINARY_ARM = 0x8F60
GL_MALI_PROGRAM_BINARY_ARM = 0x8F61
GL_MAX_SHADER_PIXEL_LOCAL_STORAGE_FAST_SIZE_EXT = 0x8F63
GL_SHADER_PIXEL_LOCAL_STORAGE_EXT = 0x8F64
GL_FETCH_PER_SAMPLE_ARM = 0x8F65
GL_FRAGMENT_SHADER_FRAMEBUFFER_FETCH_MRT_ARM = 0x8F66
GL_MAX_SHADER_PIXEL_LOCAL_STORAGE_SIZE_EXT = 0x8F67
GL_RED_SNORM = 0x8F90
GL_RG_SNORM = 0x8F91
GL_RGB_SNORM = 0x8F92
GL_RGBA_SNORM = 0x8F93
GL_R8_SNORM = 0x8F94
GL_RG8_SNORM = 0x8F95
GL_RGB8_SNORM = 0x8F96
GL_RGBA8_SNORM = 0x8F97
GL_R16_SNORM = 0x8F98
GL_RG16_SNORM = 0x8F99
GL_RGB16_SNORM = 0x8F9A
GL_RGBA16_SNORM = 0x8F9B
GL_SIGNED_NORMALIZED = 0x8F9C
GL_PRIMITIVE_RESTART = 0x8F9D
GL_PRIMITIVE_RESTART_INDEX = 0x8F9E
GL_MAX_PROGRAM_TEXTURE_GATHER_COMPONENTS_ARB = 0x8F9F
GL_PERFMON_GLOBAL_MODE_QCOM = 0x8FA0
GL_BINNING_CONTROL_HINT_QCOM = 0x8FB0
GL_CPU_OPTIMIZED_QCOM = 0x8FB1
GL_GPU_OPTIMIZED_QCOM = 0x8FB2
GL_RENDER_DIRECT_TO_FRAMEBUFFER_QCOM = 0x8FB3
GL_GPU_DISJOINT_EXT = 0x8FBB
GL_SHADER_BINARY_VIV = 0x8FC4
GL_INT8_NV = 0x8FE0
GL_INT8_VEC2_NV = 0x8FE1
GL_INT8_VEC3_NV = 0x8FE2
GL_INT8_VEC4_NV = 0x8FE3
GL_INT16_NV = 0x8FE4
GL_INT16_VEC2_NV = 0x8FE5
GL_INT16_VEC3_NV = 0x8FE6
GL_INT16_VEC4_NV = 0x8FE7
GL_INT64_VEC2_NV = 0x8FE9
GL_INT64_VEC3_NV = 0x8FEA
GL_INT64_VEC4_NV = 0x8FEB
GL_UNSIGNED_INT8_NV = 0x8FEC
GL_UNSIGNED_INT8_VEC2_NV = 0x8FED
GL_UNSIGNED_INT8_VEC3_NV = 0x8FEE
GL_UNSIGNED_INT8_VEC4_NV = 0x8FEF
GL_UNSIGNED_INT16_NV = 0x8FF0
GL_UNSIGNED_INT16_VEC2_NV = 0x8FF1
GL_UNSIGNED_INT16_VEC3_NV = 0x8FF2
GL_UNSIGNED_INT16_VEC4_NV = 0x8FF3
GL_UNSIGNED_INT64_VEC2_NV = 0x8FF5
GL_UNSIGNED_INT64_VEC3_NV = 0x8FF6
GL_UNSIGNED_INT64_VEC4_NV = 0x8FF7
GL_FLOAT16_NV = 0x8FF8
GL_FLOAT16_VEC2_NV = 0x8FF9
GL_FLOAT16_VEC3_NV = 0x8FFA
GL_FLOAT16_VEC4_NV = 0x8FFB
GL_DOUBLE_VEC2 = 0x8FFC
GL_DOUBLE_VEC2_EXT = 0x8FFC
GL_DOUBLE_VEC3 = 0x8FFD
GL_DOUBLE_VEC3_EXT = 0x8FFD
GL_DOUBLE_VEC4 = 0x8FFE
GL_DOUBLE_VEC4_EXT = 0x8FFE
GL_SAMPLER_BUFFER_AMD = 0x9001
GL_INT_SAMPLER_BUFFER_AMD = 0x9002
GL_UNSIGNED_INT_SAMPLER_BUFFER_AMD = 0x9003
GL_TESSELLATION_MODE_AMD = 0x9004
GL_TESSELLATION_FACTOR_AMD = 0x9005
GL_DISCRETE_AMD = 0x9006
GL_CONTINUOUS_AMD = 0x9007
GL_TEXTURE_CUBE_MAP_ARRAY = 0x9009
GL_TEXTURE_CUBE_MAP_ARRAY_ARB = 0x9009
GL_TEXTURE_CUBE_MAP_ARRAY_EXT = 0x9009
GL_TEXTURE_BINDING_CUBE_MAP_ARRAY = 0x900A
GL_TEXTURE_BINDING_CUBE_MAP_ARRAY_ARB = 0x900A
GL_TEXTURE_BINDING_CUBE_MAP_ARRAY_EXT = 0x900A
GL_PROXY_TEXTURE_CUBE_MAP_ARRAY = 0x900B
GL_PROXY_TEXTURE_CUBE_MAP_ARRAY_ARB = 0x900B
GL_SAMPLER_CUBE_MAP_ARRAY = 0x900C
GL_SAMPLER_CUBE_MAP_ARRAY_ARB = 0x900C
GL_SAMPLER_CUBE_MAP_ARRAY_EXT = 0x900C
GL_SAMPLER_CUBE_MAP_ARRAY_SHADOW = 0x900D
GL_SAMPLER_CUBE_MAP_ARRAY_SHADOW_ARB = 0x900D
GL_SAMPLER_CUBE_MAP_ARRAY_SHADOW_EXT = 0x900D
GL_INT_SAMPLER_CUBE_MAP_ARRAY = 0x900E
GL_INT_SAMPLER_CUBE_MAP_ARRAY_ARB = 0x900E
GL_INT_SAMPLER_CUBE_MAP_ARRAY_EXT = 0x900E
GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY = 0x900F
GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY_ARB = 0x900F
GL_UNSIGNED_INT_SAMPLER_CUBE_MAP_ARRAY_EXT = 0x900F
GL_ALPHA_SNORM = 0x9010
GL_LUMINANCE_SNORM = 0x9011
GL_LUMINANCE_ALPHA_SNORM = 0x9012
GL_INTENSITY_SNORM = 0x9013
GL_ALPHA8_SNORM = 0x9014
GL_LUMINANCE8_SNORM = 0x9015
GL_LUMINANCE8_ALPHA8_SNORM = 0x9016
GL_INTENSITY8_SNORM = 0x9017
GL_ALPHA16_SNORM = 0x9018
GL_LUMINANCE16_SNORM = 0x9019
GL_LUMINANCE16_ALPHA16_SNORM = 0x901A
GL_INTENSITY16_SNORM = 0x901B
GL_FACTOR_MIN_AMD = 0x901C
GL_FACTOR_MAX_AMD = 0x901D
GL_DEPTH_CLAMP_NEAR_AMD = 0x901E
GL_DEPTH_CLAMP_FAR_AMD = 0x901F
GL_VIDEO_BUFFER_NV = 0x9020
GL_VIDEO_BUFFER_BINDING_NV = 0x9021
GL_FIELD_UPPER_NV = 0x9022
GL_FIELD_LOWER_NV = 0x9023
GL_NUM_VIDEO_CAPTURE_STREAMS_NV = 0x9024
GL_NEXT_VIDEO_CAPTURE_BUFFER_STATUS_NV = 0x9025
GL_VIDEO_CAPTURE_TO_422_SUPPORTED_NV = 0x9026
GL_LAST_VIDEO_CAPTURE_STATUS_NV = 0x9027
GL_VIDEO_BUFFER_PITCH_NV = 0x9028
GL_VIDEO_COLOR_CONVERSION_MATRIX_NV = 0x9029
GL_VIDEO_COLOR_CONVERSION_MAX_NV = 0x902A
GL_VIDEO_COLOR_CONVERSION_MIN_NV = 0x902B
GL_VIDEO_COLOR_CONVERSION_OFFSET_NV = 0x902C
GL_VIDEO_BUFFER_INTERNAL_FORMAT_NV = 0x902D
GL_PARTIAL_SUCCESS_NV = 0x902E
GL_SUCCESS_NV = 0x902F
GL_FAILURE_NV = 0x9030
GL_YCBYCR8_422_NV = 0x9031
GL_YCBAYCR8A_4224_NV = 0x9032
GL_Z6Y10Z6CB10Z6Y10Z6CR10_422_NV = 0x9033
GL_Z6Y10Z6CB10Z6A10Z6Y10Z6CR10Z6A10_4224_NV = 0x9034
GL_Z4Y12Z4CB12Z4Y12Z4CR12_422_NV = 0x9035
GL_Z4Y12Z4CB12Z4A12Z4Y12Z4CR12Z4A12_4224_NV = 0x9036
GL_Z4Y12Z4CB12Z4CR12_444_NV = 0x9037
GL_VIDEO_CAPTURE_FRAME_WIDTH_NV = 0x9038
GL_VIDEO_CAPTURE_FRAME_HEIGHT_NV = 0x9039
GL_VIDEO_CAPTURE_FIELD_UPPER_HEIGHT_NV = 0x903A
GL_VIDEO_CAPTURE_FIELD_LOWER_HEIGHT_NV = 0x903B
GL_VIDEO_CAPTURE_SURFACE_ORIGIN_NV = 0x903C
GL_TEXTURE_COVERAGE_SAMPLES_NV = 0x9045
GL_TEXTURE_COLOR_SAMPLES_NV = 0x9046
GL_GPU_MEMORY_INFO_DEDICATED_VIDMEM_NVX = 0x9047
GL_GPU_MEMORY_INFO_TOTAL_AVAILABLE_MEMORY_NVX = 0x9048
GL_GPU_MEMORY_INFO_CURRENT_AVAILABLE_VIDMEM_NVX = 0x9049
GL_GPU_MEMORY_INFO_EVICTION_COUNT_NVX = 0x904A
GL_GPU_MEMORY_INFO_EVICTED_MEMORY_NVX = 0x904B
GL_IMAGE_1D = 0x904C
GL_IMAGE_1D_EXT = 0x904C
GL_IMAGE_2D = 0x904D
GL_IMAGE_2D_EXT = 0x904D
GL_IMAGE_3D = 0x904E
GL_IMAGE_3D_EXT = 0x904E
GL_IMAGE_2D_RECT = 0x904F
GL_IMAGE_2D_RECT_EXT = 0x904F
GL_IMAGE_CUBE = 0x9050
GL_IMAGE_CUBE_EXT = 0x9050
GL_IMAGE_BUFFER = 0x9051
GL_IMAGE_BUFFER_EXT = 0x9051
GL_IMAGE_1D_ARRAY = 0x9052
GL_IMAGE_1D_ARRAY_EXT = 0x9052
GL_IMAGE_2D_ARRAY = 0x9053
GL_IMAGE_2D_ARRAY_EXT = 0x9053
GL_IMAGE_CUBE_MAP_ARRAY = 0x9054
GL_IMAGE_CUBE_MAP_ARRAY_EXT = 0x9054
GL_IMAGE_2D_MULTISAMPLE = 0x9055
GL_IMAGE_2D_MULTISAMPLE_EXT = 0x9055
GL_IMAGE_2D_MULTISAMPLE_ARRAY = 0x9056
GL_IMAGE_2D_MULTISAMPLE_ARRAY_EXT = 0x9056
GL_INT_IMAGE_1D = 0x9057
GL_INT_IMAGE_1D_EXT = 0x9057
GL_INT_IMAGE_2D = 0x9058
GL_INT_IMAGE_2D_EXT = 0x9058
GL_INT_IMAGE_3D = 0x9059
GL_INT_IMAGE_3D_EXT = 0x9059
GL_INT_IMAGE_2D_RECT = 0x905A
GL_INT_IMAGE_2D_RECT_EXT = 0x905A
GL_INT_IMAGE_CUBE = 0x905B
GL_INT_IMAGE_CUBE_EXT = 0x905B
GL_INT_IMAGE_BUFFER = 0x905C
GL_INT_IMAGE_BUFFER_EXT = 0x905C
GL_INT_IMAGE_1D_ARRAY = 0x905D
GL_INT_IMAGE_1D_ARRAY_EXT = 0x905D
GL_INT_IMAGE_2D_ARRAY = 0x905E
GL_INT_IMAGE_2D_ARRAY_EXT = 0x905E
GL_INT_IMAGE_CUBE_MAP_ARRAY = 0x905F
GL_INT_IMAGE_CUBE_MAP_ARRAY_EXT = 0x905F
GL_INT_IMAGE_2D_MULTISAMPLE = 0x9060
GL_INT_IMAGE_2D_MULTISAMPLE_EXT = 0x9060
GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY = 0x9061
GL_INT_IMAGE_2D_MULTISAMPLE_ARRAY_EXT = 0x9061
GL_UNSIGNED_INT_IMAGE_1D = 0x9062
GL_UNSIGNED_INT_IMAGE_1D_EXT = 0x9062
GL_UNSIGNED_INT_IMAGE_2D = 0x9063
GL_UNSIGNED_INT_IMAGE_2D_EXT = 0x9063
GL_UNSIGNED_INT_IMAGE_3D = 0x9064
GL_UNSIGNED_INT_IMAGE_3D_EXT = 0x9064
GL_UNSIGNED_INT_IMAGE_2D_RECT = 0x9065
GL_UNSIGNED_INT_IMAGE_2D_RECT_EXT = 0x9065
GL_UNSIGNED_INT_IMAGE_CUBE = 0x9066
GL_UNSIGNED_INT_IMAGE_CUBE_EXT = 0x9066
GL_UNSIGNED_INT_IMAGE_BUFFER = 0x9067
GL_UNSIGNED_INT_IMAGE_BUFFER_EXT = 0x9067
GL_UNSIGNED_INT_IMAGE_1D_ARRAY = 0x9068
GL_UNSIGNED_INT_IMAGE_1D_ARRAY_EXT = 0x9068
GL_UNSIGNED_INT_IMAGE_2D_ARRAY = 0x9069
GL_UNSIGNED_INT_IMAGE_2D_ARRAY_EXT = 0x9069
GL_UNSIGNED_INT_IMAGE_CUBE_MAP_ARRAY = 0x906A
GL_UNSIGNED_INT_IMAGE_CUBE_MAP_ARRAY_EXT = 0x906A
GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE = 0x906B
GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_EXT = 0x906B
GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY = 0x906C
GL_UNSIGNED_INT_IMAGE_2D_MULTISAMPLE_ARRAY_EXT = 0x906C
GL_MAX_IMAGE_SAMPLES = 0x906D
GL_MAX_IMAGE_SAMPLES_EXT = 0x906D
GL_IMAGE_BINDING_FORMAT = 0x906E
GL_IMAGE_BINDING_FORMAT_EXT = 0x906E
GL_RGB10_A2UI = 0x906F
GL_PATH_FORMAT_SVG_NV = 0x9070
GL_PATH_FORMAT_PS_NV = 0x9071
GL_STANDARD_FONT_NAME_NV = 0x9072
GL_SYSTEM_FONT_NAME_NV = 0x9073
GL_FILE_NAME_NV = 0x9074
GL_PATH_STROKE_WIDTH_NV = 0x9075
GL_PATH_END_CAPS_NV = 0x9076
GL_PATH_INITIAL_END_CAP_NV = 0x9077
GL_PATH_TERMINAL_END_CAP_NV = 0x9078
GL_PATH_JOIN_STYLE_NV = 0x9079
GL_PATH_MITER_LIMIT_NV = 0x907A
GL_PATH_DASH_CAPS_NV = 0x907B
GL_PATH_INITIAL_DASH_CAP_NV = 0x907C
GL_PATH_TERMINAL_DASH_CAP_NV = 0x907D
GL_PATH_DASH_OFFSET_NV = 0x907E
GL_PATH_CLIENT_LENGTH_NV = 0x907F
GL_PATH_FILL_MODE_NV = 0x9080
GL_PATH_FILL_MASK_NV = 0x9081
GL_PATH_FILL_COVER_MODE_NV = 0x9082
GL_PATH_STROKE_COVER_MODE_NV = 0x9083
GL_PATH_STROKE_MASK_NV = 0x9084
GL_COUNT_UP_NV = 0x9088
GL_COUNT_DOWN_NV = 0x9089
GL_PATH_OBJECT_BOUNDING_BOX_NV = 0x908A
GL_CONVEX_HULL_NV = 0x908B
GL_BOUNDING_BOX_NV = 0x908D
GL_TRANSLATE_X_NV = 0x908E
GL_TRANSLATE_Y_NV = 0x908F
GL_TRANSLATE_2D_NV = 0x9090
GL_TRANSLATE_3D_NV = 0x9091
GL_AFFINE_2D_NV = 0x9092
GL_AFFINE_3D_NV = 0x9094
GL_TRANSPOSE_AFFINE_2D_NV = 0x9096
GL_TRANSPOSE_AFFINE_3D_NV = 0x9098
GL_UTF8_NV = 0x909A
GL_UTF16_NV = 0x909B
GL_BOUNDING_BOX_OF_BOUNDING_BOXES_NV = 0x909C
GL_PATH_COMMAND_COUNT_NV = 0x909D
GL_PATH_COORD_COUNT_NV = 0x909E
GL_PATH_DASH_ARRAY_COUNT_NV = 0x909F
GL_PATH_COMPUTED_LENGTH_NV = 0x90A0
GL_PATH_FILL_BOUNDING_BOX_NV = 0x90A1
GL_PATH_STROKE_BOUNDING_BOX_NV = 0x90A2
GL_SQUARE_NV = 0x90A3
GL_ROUND_NV = 0x90A4
GL_TRIANGULAR_NV = 0x90A5
GL_BEVEL_NV = 0x90A6
GL_MITER_REVERT_NV = 0x90A7
GL_MITER_TRUNCATE_NV = 0x90A8
GL_SKIP_MISSING_GLYPH_NV = 0x90A9
GL_USE_MISSING_GLYPH_NV = 0x90AA
GL_PATH_ERROR_POSITION_NV = 0x90AB
GL_PATH_FOG_GEN_MODE_NV = 0x90AC
GL_ACCUM_ADJACENT_PAIRS_NV = 0x90AD
GL_ADJACENT_PAIRS_NV = 0x90AE
GL_FIRST_TO_REST_NV = 0x90AF
GL_PATH_GEN_MODE_NV = 0x90B0
GL_PATH_GEN_COEFF_NV = 0x90B1
GL_PATH_GEN_COLOR_FORMAT_NV = 0x90B2
GL_PATH_GEN_COMPONENTS_NV = 0x90B3
GL_PATH_DASH_OFFSET_RESET_NV = 0x90B4
GL_MOVE_TO_RESETS_NV = 0x90B5
GL_MOVE_TO_CONTINUES_NV = 0x90B6
GL_PATH_STENCIL_FUNC_NV = 0x90B7
GL_PATH_STENCIL_REF_NV = 0x90B8
GL_PATH_STENCIL_VALUE_MASK_NV = 0x90B9
GL_SCALED_RESOLVE_FASTEST_EXT = 0x90BA
GL_SCALED_RESOLVE_NICEST_EXT = 0x90BB
GL_MIN_MAP_BUFFER_ALIGNMENT = 0x90BC
GL_PATH_STENCIL_DEPTH_OFFSET_FACTOR_NV = 0x90BD
GL_PATH_STENCIL_DEPTH_OFFSET_UNITS_NV = 0x90BE
GL_PATH_COVER_DEPTH_FUNC_NV = 0x90BF
GL_IMAGE_FORMAT_COMPATIBILITY_TYPE = 0x90C7
GL_IMAGE_FORMAT_COMPATIBILITY_BY_SIZE = 0x90C8
GL_IMAGE_FORMAT_COMPATIBILITY_BY_CLASS = 0x90C9
GL_MAX_VERTEX_IMAGE_UNIFORMS = 0x90CA
GL_MAX_TESS_CONTROL_IMAGE_UNIFORMS = 0x90CB
GL_MAX_TESS_CONTROL_IMAGE_UNIFORMS_EXT = 0x90CB
GL_MAX_TESS_EVALUATION_IMAGE_UNIFORMS = 0x90CC
GL_MAX_TESS_EVALUATION_IMAGE_UNIFORMS_EXT = 0x90CC
GL_MAX_GEOMETRY_IMAGE_UNIFORMS = 0x90CD
GL_MAX_GEOMETRY_IMAGE_UNIFORMS_EXT = 0x90CD
GL_MAX_FRAGMENT_IMAGE_UNIFORMS = 0x90CE
GL_MAX_COMBINED_IMAGE_UNIFORMS = 0x90CF
GL_MAX_DEEP_3D_TEXTURE_WIDTH_HEIGHT_NV = 0x90D0
GL_MAX_DEEP_3D_TEXTURE_DEPTH_NV = 0x90D1
GL_SHADER_STORAGE_BUFFER = 0x90D2
GL_SHADER_STORAGE_BUFFER_BINDING = 0x90D3
GL_SHADER_STORAGE_BUFFER_START = 0x90D4
GL_SHADER_STORAGE_BUFFER_SIZE = 0x90D5
GL_MAX_VERTEX_SHADER_STORAGE_BLOCKS = 0x90D6
GL_MAX_GEOMETRY_SHADER_STORAGE_BLOCKS = 0x90D7
GL_MAX_GEOMETRY_SHADER_STORAGE_BLOCKS_EXT = 0x90D7
GL_MAX_TESS_CONTROL_SHADER_STORAGE_BLOCKS = 0x90D8
GL_MAX_TESS_CONTROL_SHADER_STORAGE_BLOCKS_EXT = 0x90D8
GL_MAX_TESS_EVALUATION_SHADER_STORAGE_BLOCKS = 0x90D9
GL_MAX_TESS_EVALUATION_SHADER_STORAGE_BLOCKS_EXT = 0x90D9
GL_MAX_FRAGMENT_SHADER_STORAGE_BLOCKS = 0x90DA
GL_MAX_COMPUTE_SHADER_STORAGE_BLOCKS = 0x90DB
GL_MAX_COMBINED_SHADER_STORAGE_BLOCKS = 0x90DC
GL_MAX_SHADER_STORAGE_BUFFER_BINDINGS = 0x90DD
GL_MAX_SHADER_STORAGE_BLOCK_SIZE = 0x90DE
GL_SHADER_STORAGE_BUFFER_OFFSET_ALIGNMENT = 0x90DF
GL_SYNC_X11_FENCE_EXT = 0x90E1
GL_DEPTH_STENCIL_TEXTURE_MODE = 0x90EA
GL_MAX_COMPUTE_WORK_GROUP_INVOCATIONS = 0x90EB
GL_MAX_COMPUTE_FIXED_GROUP_INVOCATIONS_ARB = 0x90EB
GL_UNIFORM_BLOCK_REFERENCED_BY_COMPUTE_SHADER = 0x90EC
GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_COMPUTE_SHADER = 0x90ED
GL_DISPATCH_INDIRECT_BUFFER = 0x90EE
GL_DISPATCH_INDIRECT_BUFFER_BINDING = 0x90EF
GL_COLOR_ATTACHMENT_EXT = 0x90F0
GL_MULTIVIEW_EXT = 0x90F1
GL_MAX_MULTIVIEW_BUFFERS_EXT = 0x90F2
GL_CONTEXT_ROBUST_ACCESS = 0x90F3
GL_CONTEXT_ROBUST_ACCESS_EXT = 0x90F3
GL_CONTEXT_ROBUST_ACCESS_KHR = 0x90F3
GL_COMPUTE_PROGRAM_NV = 0x90FB
GL_COMPUTE_PROGRAM_PARAMETER_BUFFER_NV = 0x90FC
GL_TEXTURE_2D_MULTISAMPLE = 0x9100
GL_PROXY_TEXTURE_2D_MULTISAMPLE = 0x9101
GL_TEXTURE_2D_MULTISAMPLE_ARRAY = 0x9102
GL_TEXTURE_2D_MULTISAMPLE_ARRAY_OES = 0x9102
GL_PROXY_TEXTURE_2D_MULTISAMPLE_ARRAY = 0x9103
GL_TEXTURE_BINDING_2D_MULTISAMPLE = 0x9104
GL_TEXTURE_BINDING_2D_MULTISAMPLE_ARRAY = 0x9105
GL_TEXTURE_BINDING_2D_MULTISAMPLE_ARRAY_OES = 0x9105
GL_TEXTURE_SAMPLES = 0x9106
GL_TEXTURE_FIXED_SAMPLE_LOCATIONS = 0x9107
GL_SAMPLER_2D_MULTISAMPLE = 0x9108
GL_INT_SAMPLER_2D_MULTISAMPLE = 0x9109
GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE = 0x910A
GL_SAMPLER_2D_MULTISAMPLE_ARRAY = 0x910B
GL_SAMPLER_2D_MULTISAMPLE_ARRAY_OES = 0x910B
GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY = 0x910C
GL_INT_SAMPLER_2D_MULTISAMPLE_ARRAY_OES = 0x910C
GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY = 0x910D
GL_UNSIGNED_INT_SAMPLER_2D_MULTISAMPLE_ARRAY_OES = 0x910D
GL_MAX_COLOR_TEXTURE_SAMPLES = 0x910E
GL_MAX_DEPTH_TEXTURE_SAMPLES = 0x910F
GL_MAX_INTEGER_SAMPLES = 0x9110
GL_MAX_SERVER_WAIT_TIMEOUT = 0x9111
GL_MAX_SERVER_WAIT_TIMEOUT_APPLE = 0x9111
GL_OBJECT_TYPE = 0x9112
GL_OBJECT_TYPE_APPLE = 0x9112
GL_SYNC_CONDITION = 0x9113
GL_SYNC_CONDITION_APPLE = 0x9113
GL_SYNC_STATUS = 0x9114
GL_SYNC_STATUS_APPLE = 0x9114
GL_SYNC_FLAGS = 0x9115
GL_SYNC_FLAGS_APPLE = 0x9115
GL_SYNC_FENCE = 0x9116
GL_SYNC_FENCE_APPLE = 0x9116
GL_SYNC_GPU_COMMANDS_COMPLETE = 0x9117
GL_SYNC_GPU_COMMANDS_COMPLETE_APPLE = 0x9117
GL_UNSIGNALED = 0x9118
GL_UNSIGNALED_APPLE = 0x9118
GL_SIGNALED = 0x9119
GL_SIGNALED_APPLE = 0x9119
GL_ALREADY_SIGNALED = 0x911A
GL_ALREADY_SIGNALED_APPLE = 0x911A
GL_TIMEOUT_EXPIRED = 0x911B
GL_TIMEOUT_EXPIRED_APPLE = 0x911B
GL_CONDITION_SATISFIED = 0x911C
GL_CONDITION_SATISFIED_APPLE = 0x911C
GL_WAIT_FAILED = 0x911D
GL_WAIT_FAILED_APPLE = 0x911D
GL_BUFFER_ACCESS_FLAGS = 0x911F
GL_BUFFER_MAP_LENGTH = 0x9120
GL_BUFFER_MAP_OFFSET = 0x9121
GL_MAX_VERTEX_OUTPUT_COMPONENTS = 0x9122
GL_MAX_GEOMETRY_INPUT_COMPONENTS = 0x9123
GL_MAX_GEOMETRY_INPUT_COMPONENTS_EXT = 0x9123
GL_MAX_GEOMETRY_OUTPUT_COMPONENTS = 0x9124
GL_MAX_GEOMETRY_OUTPUT_COMPONENTS_EXT = 0x9124
GL_MAX_FRAGMENT_INPUT_COMPONENTS = 0x9125
GL_CONTEXT_PROFILE_MASK = 0x9126
GL_UNPACK_COMPRESSED_BLOCK_WIDTH = 0x9127
GL_UNPACK_COMPRESSED_BLOCK_HEIGHT = 0x9128
GL_UNPACK_COMPRESSED_BLOCK_DEPTH = 0x9129
GL_UNPACK_COMPRESSED_BLOCK_SIZE = 0x912A
GL_PACK_COMPRESSED_BLOCK_WIDTH = 0x912B
GL_PACK_COMPRESSED_BLOCK_HEIGHT = 0x912C
GL_PACK_COMPRESSED_BLOCK_DEPTH = 0x912D
GL_PACK_COMPRESSED_BLOCK_SIZE = 0x912E
GL_TEXTURE_IMMUTABLE_FORMAT = 0x912F
GL_TEXTURE_IMMUTABLE_FORMAT_EXT = 0x912F
GL_SGX_PROGRAM_BINARY_IMG = 0x9130
GL_RENDERBUFFER_SAMPLES_IMG = 0x9133
GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_IMG = 0x9134
GL_MAX_SAMPLES_IMG = 0x9135
GL_TEXTURE_SAMPLES_IMG = 0x9136
GL_COMPRESSED_RGBA_PVRTC_2BPPV2_IMG = 0x9137
GL_COMPRESSED_RGBA_PVRTC_4BPPV2_IMG = 0x9138
GL_MAX_DEBUG_MESSAGE_LENGTH = 0x9143
GL_MAX_DEBUG_MESSAGE_LENGTH_AMD = 0x9143
GL_MAX_DEBUG_MESSAGE_LENGTH_ARB = 0x9143
GL_MAX_DEBUG_MESSAGE_LENGTH_KHR = 0x9143
GL_MAX_DEBUG_LOGGED_MESSAGES = 0x9144
GL_MAX_DEBUG_LOGGED_MESSAGES_AMD = 0x9144
GL_MAX_DEBUG_LOGGED_MESSAGES_ARB = 0x9144
GL_MAX_DEBUG_LOGGED_MESSAGES_KHR = 0x9144
GL_DEBUG_LOGGED_MESSAGES = 0x9145
GL_DEBUG_LOGGED_MESSAGES_AMD = 0x9145
GL_DEBUG_LOGGED_MESSAGES_ARB = 0x9145
GL_DEBUG_LOGGED_MESSAGES_KHR = 0x9145
GL_DEBUG_SEVERITY_HIGH = 0x9146
GL_DEBUG_SEVERITY_HIGH_AMD = 0x9146
GL_DEBUG_SEVERITY_HIGH_ARB = 0x9146
GL_DEBUG_SEVERITY_HIGH_KHR = 0x9146
GL_DEBUG_SEVERITY_MEDIUM = 0x9147
GL_DEBUG_SEVERITY_MEDIUM_AMD = 0x9147
GL_DEBUG_SEVERITY_MEDIUM_ARB = 0x9147
GL_DEBUG_SEVERITY_MEDIUM_KHR = 0x9147
GL_DEBUG_SEVERITY_LOW = 0x9148
GL_DEBUG_SEVERITY_LOW_AMD = 0x9148
GL_DEBUG_SEVERITY_LOW_ARB = 0x9148
GL_DEBUG_SEVERITY_LOW_KHR = 0x9148
GL_DEBUG_CATEGORY_API_ERROR_AMD = 0x9149
GL_DEBUG_CATEGORY_WINDOW_SYSTEM_AMD = 0x914A
GL_DEBUG_CATEGORY_DEPRECATION_AMD = 0x914B
GL_DEBUG_CATEGORY_UNDEFINED_BEHAVIOR_AMD = 0x914C
GL_DEBUG_CATEGORY_PERFORMANCE_AMD = 0x914D
GL_DEBUG_CATEGORY_SHADER_COMPILER_AMD = 0x914E
GL_DEBUG_CATEGORY_APPLICATION_AMD = 0x914F
GL_DEBUG_CATEGORY_OTHER_AMD = 0x9150
GL_BUFFER_OBJECT_EXT = 0x9151
GL_DATA_BUFFER_AMD = 0x9151
GL_PERFORMANCE_MONITOR_AMD = 0x9152
GL_QUERY_OBJECT_AMD = 0x9153
GL_QUERY_OBJECT_EXT = 0x9153
GL_VERTEX_ARRAY_OBJECT_AMD = 0x9154
GL_VERTEX_ARRAY_OBJECT_EXT = 0x9154
GL_SAMPLER_OBJECT_AMD = 0x9155
GL_EXTERNAL_VIRTUAL_MEMORY_BUFFER_AMD = 0x9160
GL_QUERY_BUFFER = 0x9192
GL_QUERY_BUFFER_AMD = 0x9192
GL_QUERY_BUFFER_BINDING = 0x9193
GL_QUERY_BUFFER_BINDING_AMD = 0x9193
GL_QUERY_RESULT_NO_WAIT = 0x9194
GL_QUERY_RESULT_NO_WAIT_AMD = 0x9194
GL_VIRTUAL_PAGE_SIZE_X_ARB = 0x9195
GL_VIRTUAL_PAGE_SIZE_X_AMD = 0x9195
GL_VIRTUAL_PAGE_SIZE_Y_ARB = 0x9196
GL_VIRTUAL_PAGE_SIZE_Y_AMD = 0x9196
GL_VIRTUAL_PAGE_SIZE_Z_ARB = 0x9197
GL_VIRTUAL_PAGE_SIZE_Z_AMD = 0x9197
GL_MAX_SPARSE_TEXTURE_SIZE_ARB = 0x9198
GL_MAX_SPARSE_TEXTURE_SIZE_AMD = 0x9198
GL_MAX_SPARSE_3D_TEXTURE_SIZE_ARB = 0x9199
GL_MAX_SPARSE_3D_TEXTURE_SIZE_AMD = 0x9199
GL_MAX_SPARSE_ARRAY_TEXTURE_LAYERS_ARB = 0x919A
GL_MAX_SPARSE_ARRAY_TEXTURE_LAYERS = 0x919A
GL_MIN_SPARSE_LEVEL_AMD = 0x919B
GL_MIN_LOD_WARNING_AMD = 0x919C
GL_TEXTURE_BUFFER_OFFSET = 0x919D
GL_TEXTURE_BUFFER_OFFSET_EXT = 0x919D
GL_TEXTURE_BUFFER_SIZE = 0x919E
GL_TEXTURE_BUFFER_SIZE_EXT = 0x919E
GL_TEXTURE_BUFFER_OFFSET_ALIGNMENT = 0x919F
GL_TEXTURE_BUFFER_OFFSET_ALIGNMENT_EXT = 0x919F
GL_STREAM_RASTERIZATION_AMD = 0x91A0
GL_VERTEX_ELEMENT_SWIZZLE_AMD = 0x91A4
GL_VERTEX_ID_SWIZZLE_AMD = 0x91A5
GL_TEXTURE_SPARSE_ARB = 0x91A6
GL_VIRTUAL_PAGE_SIZE_INDEX_ARB = 0x91A7
GL_NUM_VIRTUAL_PAGE_SIZES_ARB = 0x91A8
GL_SPARSE_TEXTURE_FULL_ARRAY_CUBE_MIPMAPS_ARB = 0x91A9
GL_NUM_SPARSE_LEVELS_ARB = 0x91AA
GL_COMPUTE_SHADER = 0x91B9
GL_MAX_COMPUTE_UNIFORM_BLOCKS = 0x91BB
GL_MAX_COMPUTE_TEXTURE_IMAGE_UNITS = 0x91BC
GL_MAX_COMPUTE_IMAGE_UNIFORMS = 0x91BD
GL_MAX_COMPUTE_WORK_GROUP_COUNT = 0x91BE
GL_MAX_COMPUTE_WORK_GROUP_SIZE = 0x91BF
GL_MAX_COMPUTE_FIXED_GROUP_SIZE_ARB = 0x91BF
GL_UNPACK_FLIP_Y_WEBGL = 0x9240
GL_UNPACK_PREMULTIPLY_ALPHA_WEBGL = 0x9241
GL_CONTEXT_LOST_WEBGL = 0x9242
GL_UNPACK_COLORSPACE_CONVERSION_WEBGL = 0x9243
GL_BROWSER_DEFAULT_WEBGL = 0x9244
GL_SHADER_BINARY_DMP = 0x9250
GL_SMAPHS30_PROGRAM_BINARY_DMP = 0x9251
GL_SMAPHS_PROGRAM_BINARY_DMP = 0x9252
GL_DMP_PROGRAM_BINARY_DMP = 0x9253
GL_GCCSO_SHADER_BINARY_FJ = 0x9260
GL_COMPRESSED_R11_EAC = 0x9270
GL_COMPRESSED_R11_EAC_OES = 0x9270
GL_COMPRESSED_SIGNED_R11_EAC = 0x9271
GL_COMPRESSED_SIGNED_R11_EAC_OES = 0x9271
GL_COMPRESSED_RG11_EAC = 0x9272
GL_COMPRESSED_RG11_EAC_OES = 0x9272
GL_COMPRESSED_SIGNED_RG11_EAC = 0x9273
GL_COMPRESSED_SIGNED_RG11_EAC_OES = 0x9273
GL_COMPRESSED_RGB8_ETC2 = 0x9274
GL_COMPRESSED_RGB8_ETC2_OES = 0x9274
GL_COMPRESSED_SRGB8_ETC2 = 0x9275
GL_COMPRESSED_SRGB8_ETC2_OES = 0x9275
GL_COMPRESSED_RGB8_PUNCHTHROUGH_ALPHA1_ETC2 = 0x9276
GL_COMPRESSED_RGB8_PUNCHTHROUGH_ALPHA1_ETC2_OES = 0x9276
GL_COMPRESSED_SRGB8_PUNCHTHROUGH_ALPHA1_ETC2 = 0x9277
GL_COMPRESSED_SRGB8_PUNCHTHROUGH_ALPHA1_ETC2_OES = 0x9277
GL_COMPRESSED_RGBA8_ETC2_EAC = 0x9278
GL_COMPRESSED_RGBA8_ETC2_EAC_OES = 0x9278
GL_COMPRESSED_SRGB8_ALPHA8_ETC2_EAC = 0x9279
GL_COMPRESSED_SRGB8_ALPHA8_ETC2_EAC_OES = 0x9279
GL_BLEND_PREMULTIPLIED_SRC_NV = 0x9280
GL_BLEND_OVERLAP_NV = 0x9281
GL_UNCORRELATED_NV = 0x9282
GL_DISJOINT_NV = 0x9283
GL_CONJOINT_NV = 0x9284
GL_BLEND_ADVANCED_COHERENT_KHR = 0x9285
GL_BLEND_ADVANCED_COHERENT_NV = 0x9285
GL_SRC_NV = 0x9286
GL_DST_NV = 0x9287
GL_SRC_OVER_NV = 0x9288
GL_DST_OVER_NV = 0x9289
GL_SRC_IN_NV = 0x928A
GL_DST_IN_NV = 0x928B
GL_SRC_OUT_NV = 0x928C
GL_DST_OUT_NV = 0x928D
GL_SRC_ATOP_NV = 0x928E
GL_DST_ATOP_NV = 0x928F
GL_PLUS_NV = 0x9291
GL_PLUS_DARKER_NV = 0x9292
GL_MULTIPLY_KHR = 0x9294
GL_MULTIPLY_NV = 0x9294
GL_SCREEN_KHR = 0x9295
GL_SCREEN_NV = 0x9295
GL_OVERLAY_KHR = 0x9296
GL_OVERLAY_NV = 0x9296
GL_DARKEN_KHR = 0x9297
GL_DARKEN_NV = 0x9297
GL_LIGHTEN_KHR = 0x9298
GL_LIGHTEN_NV = 0x9298
GL_COLORDODGE_KHR = 0x9299
GL_COLORDODGE_NV = 0x9299
GL_COLORBURN_KHR = 0x929A
GL_COLORBURN_NV = 0x929A
GL_HARDLIGHT_KHR = 0x929B
GL_HARDLIGHT_NV = 0x929B
GL_SOFTLIGHT_KHR = 0x929C
GL_SOFTLIGHT_NV = 0x929C
GL_DIFFERENCE_KHR = 0x929E
GL_DIFFERENCE_NV = 0x929E
GL_MINUS_NV = 0x929F
GL_EXCLUSION_KHR = 0x92A0
GL_EXCLUSION_NV = 0x92A0
GL_CONTRAST_NV = 0x92A1
GL_INVERT_RGB_NV = 0x92A3
GL_LINEARDODGE_NV = 0x92A4
GL_LINEARBURN_NV = 0x92A5
GL_VIVIDLIGHT_NV = 0x92A6
GL_LINEARLIGHT_NV = 0x92A7
GL_PINLIGHT_NV = 0x92A8
GL_HARDMIX_NV = 0x92A9
GL_HSL_HUE_KHR = 0x92AD
GL_HSL_HUE_NV = 0x92AD
GL_HSL_SATURATION_KHR = 0x92AE
GL_HSL_SATURATION_NV = 0x92AE
GL_HSL_COLOR_KHR = 0x92AF
GL_HSL_COLOR_NV = 0x92AF
GL_HSL_LUMINOSITY_KHR = 0x92B0
GL_HSL_LUMINOSITY_NV = 0x92B0
GL_PLUS_CLAMPED_NV = 0x92B1
GL_PLUS_CLAMPED_ALPHA_NV = 0x92B2
GL_MINUS_CLAMPED_NV = 0x92B3
GL_INVERT_OVG_NV = 0x92B4
GL_PRIMITIVE_BOUNDING_BOX_EXT = 0x92BE
GL_ATOMIC_COUNTER_BUFFER = 0x92C0
GL_ATOMIC_COUNTER_BUFFER_BINDING = 0x92C1
GL_ATOMIC_COUNTER_BUFFER_START = 0x92C2
GL_ATOMIC_COUNTER_BUFFER_SIZE = 0x92C3
GL_ATOMIC_COUNTER_BUFFER_DATA_SIZE = 0x92C4
GL_ATOMIC_COUNTER_BUFFER_ACTIVE_ATOMIC_COUNTERS = 0x92C5
GL_ATOMIC_COUNTER_BUFFER_ACTIVE_ATOMIC_COUNTER_INDICES = 0x92C6
GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_VERTEX_SHADER = 0x92C7
GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_TESS_CONTROL_SHADER = 0x92C8
GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_TESS_EVALUATION_SHADER = 0x92C9
GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_GEOMETRY_SHADER = 0x92CA
GL_ATOMIC_COUNTER_BUFFER_REFERENCED_BY_FRAGMENT_SHADER = 0x92CB
GL_MAX_VERTEX_ATOMIC_COUNTER_BUFFERS = 0x92CC
GL_MAX_TESS_CONTROL_ATOMIC_COUNTER_BUFFERS = 0x92CD
GL_MAX_TESS_CONTROL_ATOMIC_COUNTER_BUFFERS_EXT = 0x92CD
GL_MAX_TESS_EVALUATION_ATOMIC_COUNTER_BUFFERS = 0x92CE
GL_MAX_TESS_EVALUATION_ATOMIC_COUNTER_BUFFERS_EXT = 0x92CE
GL_MAX_GEOMETRY_ATOMIC_COUNTER_BUFFERS = 0x92CF
GL_MAX_GEOMETRY_ATOMIC_COUNTER_BUFFERS_EXT = 0x92CF
GL_MAX_FRAGMENT_ATOMIC_COUNTER_BUFFERS = 0x92D0
GL_MAX_COMBINED_ATOMIC_COUNTER_BUFFERS = 0x92D1
GL_MAX_VERTEX_ATOMIC_COUNTERS = 0x92D2
GL_MAX_TESS_CONTROL_ATOMIC_COUNTERS = 0x92D3
GL_MAX_TESS_CONTROL_ATOMIC_COUNTERS_EXT = 0x92D3
GL_MAX_TESS_EVALUATION_ATOMIC_COUNTERS = 0x92D4
GL_MAX_TESS_EVALUATION_ATOMIC_COUNTERS_EXT = 0x92D4
GL_MAX_GEOMETRY_ATOMIC_COUNTERS = 0x92D5
GL_MAX_GEOMETRY_ATOMIC_COUNTERS_EXT = 0x92D5
GL_MAX_FRAGMENT_ATOMIC_COUNTERS = 0x92D6
GL_MAX_COMBINED_ATOMIC_COUNTERS = 0x92D7
GL_MAX_ATOMIC_COUNTER_BUFFER_SIZE = 0x92D8
GL_ACTIVE_ATOMIC_COUNTER_BUFFERS = 0x92D9
GL_UNIFORM_ATOMIC_COUNTER_BUFFER_INDEX = 0x92DA
GL_UNSIGNED_INT_ATOMIC_COUNTER = 0x92DB
GL_MAX_ATOMIC_COUNTER_BUFFER_BINDINGS = 0x92DC
GL_FRAGMENT_COVERAGE_TO_COLOR_NV = 0x92DD
GL_FRAGMENT_COVERAGE_COLOR_NV = 0x92DE
GL_DEBUG_OUTPUT = 0x92E0
GL_DEBUG_OUTPUT_KHR = 0x92E0
GL_UNIFORM = 0x92E1
GL_UNIFORM_BLOCK = 0x92E2
GL_PROGRAM_INPUT = 0x92E3
GL_PROGRAM_OUTPUT = 0x92E4
GL_BUFFER_VARIABLE = 0x92E5
GL_SHADER_STORAGE_BLOCK = 0x92E6
GL_IS_PER_PATCH = 0x92E7
GL_IS_PER_PATCH_EXT = 0x92E7
GL_VERTEX_SUBROUTINE = 0x92E8
GL_TESS_CONTROL_SUBROUTINE = 0x92E9
GL_TESS_EVALUATION_SUBROUTINE = 0x92EA
GL_GEOMETRY_SUBROUTINE = 0x92EB
GL_FRAGMENT_SUBROUTINE = 0x92EC
GL_COMPUTE_SUBROUTINE = 0x92ED
GL_VERTEX_SUBROUTINE_UNIFORM = 0x92EE
GL_TESS_CONTROL_SUBROUTINE_UNIFORM = 0x92EF
GL_TESS_EVALUATION_SUBROUTINE_UNIFORM = 0x92F0
GL_GEOMETRY_SUBROUTINE_UNIFORM = 0x92F1
GL_FRAGMENT_SUBROUTINE_UNIFORM = 0x92F2
GL_COMPUTE_SUBROUTINE_UNIFORM = 0x92F3
GL_TRANSFORM_FEEDBACK_VARYING = 0x92F4
GL_ACTIVE_RESOURCES = 0x92F5
GL_MAX_NAME_LENGTH = 0x92F6
GL_MAX_NUM_ACTIVE_VARIABLES = 0x92F7
GL_MAX_NUM_COMPATIBLE_SUBROUTINES = 0x92F8
GL_NAME_LENGTH = 0x92F9
GL_TYPE = 0x92FA
GL_ARRAY_SIZE = 0x92FB
GL_OFFSET = 0x92FC
GL_BLOCK_INDEX = 0x92FD
GL_ARRAY_STRIDE = 0x92FE
GL_MATRIX_STRIDE = 0x92FF
GL_IS_ROW_MAJOR = 0x9300
GL_ATOMIC_COUNTER_BUFFER_INDEX = 0x9301
GL_BUFFER_BINDING = 0x9302
GL_BUFFER_DATA_SIZE = 0x9303
GL_NUM_ACTIVE_VARIABLES = 0x9304
GL_ACTIVE_VARIABLES = 0x9305
GL_REFERENCED_BY_VERTEX_SHADER = 0x9306
GL_REFERENCED_BY_TESS_CONTROL_SHADER = 0x9307
GL_REFERENCED_BY_TESS_CONTROL_SHADER_EXT = 0x9307
GL_REFERENCED_BY_TESS_EVALUATION_SHADER = 0x9308
GL_REFERENCED_BY_TESS_EVALUATION_SHADER_EXT = 0x9308
GL_REFERENCED_BY_GEOMETRY_SHADER = 0x9309
GL_REFERENCED_BY_GEOMETRY_SHADER_EXT = 0x9309
GL_REFERENCED_BY_FRAGMENT_SHADER = 0x930A
GL_REFERENCED_BY_COMPUTE_SHADER = 0x930B
GL_TOP_LEVEL_ARRAY_SIZE = 0x930C
GL_TOP_LEVEL_ARRAY_STRIDE = 0x930D
GL_LOCATION = 0x930E
GL_LOCATION_INDEX = 0x930F
GL_FRAMEBUFFER_DEFAULT_WIDTH = 0x9310
GL_FRAMEBUFFER_DEFAULT_HEIGHT = 0x9311
GL_FRAMEBUFFER_DEFAULT_LAYERS = 0x9312
GL_FRAMEBUFFER_DEFAULT_LAYERS_EXT = 0x9312
GL_FRAMEBUFFER_DEFAULT_SAMPLES = 0x9313
GL_FRAMEBUFFER_DEFAULT_FIXED_SAMPLE_LOCATIONS = 0x9314
GL_MAX_FRAMEBUFFER_WIDTH = 0x9315
GL_MAX_FRAMEBUFFER_HEIGHT = 0x9316
GL_MAX_FRAMEBUFFER_LAYERS = 0x9317
GL_MAX_FRAMEBUFFER_LAYERS_EXT = 0x9317
GL_MAX_FRAMEBUFFER_SAMPLES = 0x9318
GL_RASTER_MULTISAMPLE_EXT = 0x9327
GL_RASTER_SAMPLES_EXT = 0x9328
GL_MAX_RASTER_SAMPLES_EXT = 0x9329
GL_RASTER_FIXED_SAMPLE_LOCATIONS_EXT = 0x932A
GL_MULTISAMPLE_RASTERIZATION_ALLOWED_EXT = 0x932B
GL_EFFECTIVE_RASTER_SAMPLES_EXT = 0x932C
GL_DEPTH_SAMPLES_NV = 0x932D
GL_STENCIL_SAMPLES_NV = 0x932E
GL_MIXED_DEPTH_SAMPLES_SUPPORTED_NV = 0x932F
GL_MIXED_STENCIL_SAMPLES_SUPPORTED_NV = 0x9330
GL_COVERAGE_MODULATION_TABLE_NV = 0x9331
GL_COVERAGE_MODULATION_NV = 0x9332
GL_COVERAGE_MODULATION_TABLE_SIZE_NV = 0x9333
GL_WARP_SIZE_NV = 0x9339
GL_WARPS_PER_SM_NV = 0x933A
GL_SM_COUNT_NV = 0x933B
GL_FILL_RECTANGLE_NV = 0x933C
GL_SAMPLE_LOCATION_SUBPIXEL_BITS_NV = 0x933D
GL_SAMPLE_LOCATION_PIXEL_GRID_WIDTH_NV = 0x933E
GL_SAMPLE_LOCATION_PIXEL_GRID_HEIGHT_NV = 0x933F
GL_PROGRAMMABLE_SAMPLE_LOCATION_TABLE_SIZE_NV = 0x9340
GL_PROGRAMMABLE_SAMPLE_LOCATION_NV = 0x9341
GL_FRAMEBUFFER_PROGRAMMABLE_SAMPLE_LOCATIONS_NV = 0x9342
GL_FRAMEBUFFER_SAMPLE_LOCATION_PIXEL_GRID_NV = 0x9343
GL_MAX_COMPUTE_VARIABLE_GROUP_INVOCATIONS_ARB = 0x9344
GL_MAX_COMPUTE_VARIABLE_GROUP_SIZE_ARB = 0x9345
GL_CONSERVATIVE_RASTERIZATION_NV = 0x9346
GL_SUBPIXEL_PRECISION_BIAS_X_BITS_NV = 0x9347
GL_SUBPIXEL_PRECISION_BIAS_Y_BITS_NV = 0x9348
GL_MAX_SUBPIXEL_PRECISION_BIAS_BITS_NV = 0x9349
GL_LOCATION_COMPONENT = 0x934A
GL_TRANSFORM_FEEDBACK_BUFFER_INDEX = 0x934B
GL_TRANSFORM_FEEDBACK_BUFFER_STRIDE = 0x934C
GL_CLIP_ORIGIN = 0x935C
GL_CLIP_DEPTH_MODE = 0x935D
GL_NEGATIVE_ONE_TO_ONE = 0x935E
GL_ZERO_TO_ONE = 0x935F
GL_CLEAR_TEXTURE = 0x9365
GL_FONT_GLYPHS_AVAILABLE_NV = 0x9368
GL_FONT_TARGET_UNAVAILABLE_NV = 0x9369
GL_FONT_UNAVAILABLE_NV = 0x936A
GL_FONT_UNINTELLIGIBLE_NV = 0x936B
GL_STANDARD_FONT_FORMAT_NV = 0x936C
GL_FRAGMENT_INPUT_NV = 0x936D
GL_UNIFORM_BUFFER_UNIFIED_NV = 0x936E
GL_UNIFORM_BUFFER_ADDRESS_NV = 0x936F
GL_UNIFORM_BUFFER_LENGTH_NV = 0x9370
GL_NUM_SAMPLE_COUNTS = 0x9380
GL_TRANSLATED_SHADER_SOURCE_LENGTH_ANGLE = 0x93A0
GL_BGRA8_EXT = 0x93A1
GL_TEXTURE_USAGE_ANGLE = 0x93A2
GL_FRAMEBUFFER_ATTACHMENT_ANGLE = 0x93A3
GL_PACK_REVERSE_ROW_ORDER_ANGLE = 0x93A4
GL_PROGRAM_BINARY_ANGLE = 0x93A6
GL_COMPRESSED_RGBA_ASTC_4x4_KHR = 0x93B0
GL_COMPRESSED_RGBA_ASTC_5x4_KHR = 0x93B1
GL_COMPRESSED_RGBA_ASTC_5x5_KHR = 0x93B2
GL_COMPRESSED_RGBA_ASTC_6x5_KHR = 0x93B3
GL_COMPRESSED_RGBA_ASTC_6x6_KHR = 0x93B4
GL_COMPRESSED_RGBA_ASTC_8x5_KHR = 0x93B5
GL_COMPRESSED_RGBA_ASTC_8x6_KHR = 0x93B6
GL_COMPRESSED_RGBA_ASTC_8x8_KHR = 0x93B7
GL_COMPRESSED_RGBA_ASTC_10x5_KHR = 0x93B8
GL_COMPRESSED_RGBA_ASTC_10x6_KHR = 0x93B9
GL_COMPRESSED_RGBA_ASTC_10x8_KHR = 0x93BA
GL_COMPRESSED_RGBA_ASTC_10x10_KHR = 0x93BB
GL_COMPRESSED_RGBA_ASTC_12x10_KHR = 0x93BC
GL_COMPRESSED_RGBA_ASTC_12x12_KHR = 0x93BD
GL_COMPRESSED_RGBA_ASTC_3x3x3_OES = 0x93C0
GL_COMPRESSED_RGBA_ASTC_4x3x3_OES = 0x93C1
GL_COMPRESSED_RGBA_ASTC_4x4x3_OES = 0x93C2
GL_COMPRESSED_RGBA_ASTC_4x4x4_OES = 0x93C3
GL_COMPRESSED_RGBA_ASTC_5x4x4_OES = 0x93C4
GL_COMPRESSED_RGBA_ASTC_5x5x4_OES = 0x93C5
GL_COMPRESSED_RGBA_ASTC_5x5x5_OES = 0x93C6
GL_COMPRESSED_RGBA_ASTC_6x5x5_OES = 0x93C7
GL_COMPRESSED_RGBA_ASTC_6x6x5_OES = 0x93C8
GL_COMPRESSED_RGBA_ASTC_6x6x6_OES = 0x93C9
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_4x4_KHR = 0x93D0
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x4_KHR = 0x93D1
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x5_KHR = 0x93D2
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x5_KHR = 0x93D3
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x6_KHR = 0x93D4
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_8x5_KHR = 0x93D5
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_8x6_KHR = 0x93D6
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_8x8_KHR = 0x93D7
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x5_KHR = 0x93D8
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x6_KHR = 0x93D9
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x8_KHR = 0x93DA
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_10x10_KHR = 0x93DB
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_12x10_KHR = 0x93DC
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_12x12_KHR = 0x93DD
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_3x3x3_OES = 0x93E0
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_4x3x3_OES = 0x93E1
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_4x4x3_OES = 0x93E2
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_4x4x4_OES = 0x93E3
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x4x4_OES = 0x93E4
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x5x4_OES = 0x93E5
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_5x5x5_OES = 0x93E6
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x5x5_OES = 0x93E7
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x6x5_OES = 0x93E8
GL_COMPRESSED_SRGB8_ALPHA8_ASTC_6x6x6_OES = 0x93E9
GL_COMPRESSED_SRGB_ALPHA_PVRTC_2BPPV2_IMG = 0x93F0
GL_COMPRESSED_SRGB_ALPHA_PVRTC_4BPPV2_IMG = 0x93F1
GL_PERFQUERY_COUNTER_EVENT_INTEL = 0x94F0
GL_PERFQUERY_COUNTER_DURATION_NORM_INTEL = 0x94F1
GL_PERFQUERY_COUNTER_DURATION_RAW_INTEL = 0x94F2
GL_PERFQUERY_COUNTER_THROUGHPUT_INTEL = 0x94F3
GL_PERFQUERY_COUNTER_RAW_INTEL = 0x94F4
GL_PERFQUERY_COUNTER_TIMESTAMP_INTEL = 0x94F5
GL_PERFQUERY_COUNTER_DATA_UINT32_INTEL = 0x94F8
GL_PERFQUERY_COUNTER_DATA_UINT64_INTEL = 0x94F9
GL_PERFQUERY_COUNTER_DATA_FLOAT_INTEL = 0x94FA
GL_PERFQUERY_COUNTER_DATA_DOUBLE_INTEL = 0x94FB
GL_PERFQUERY_COUNTER_DATA_BOOL32_INTEL = 0x94FC
GL_PERFQUERY_QUERY_NAME_LENGTH_MAX_INTEL = 0x94FD
GL_PERFQUERY_COUNTER_NAME_LENGTH_MAX_INTEL = 0x94FE
GL_PERFQUERY_COUNTER_DESC_LENGTH_MAX_INTEL = 0x94FF
GL_PERFQUERY_GPA_EXTENDED_COUNTERS_INTEL = 0x9500
GL_RASTER_POSITION_UNCLIPPED_IBM = 0x19262
GL_CULL_VERTEX_IBM = 103050
GL_ALL_STATIC_DATA_IBM = 103060
GL_STATIC_VERTEX_ARRAY_IBM = 103061
GL_VERTEX_ARRAY_LIST_IBM = 103070
GL_NORMAL_ARRAY_LIST_IBM = 103071
GL_COLOR_ARRAY_LIST_IBM = 103072
GL_INDEX_ARRAY_LIST_IBM = 103073
GL_TEXTURE_COORD_ARRAY_LIST_IBM = 103074
GL_EDGE_FLAG_ARRAY_LIST_IBM = 103075
GL_FOG_COORDINATE_ARRAY_LIST_IBM = 103076
GL_SECONDARY_COLOR_ARRAY_LIST_IBM = 103077
GL_VERTEX_ARRAY_LIST_STRIDE_IBM = 103080
GL_NORMAL_ARRAY_LIST_STRIDE_IBM = 103081
GL_COLOR_ARRAY_LIST_STRIDE_IBM = 103082
GL_INDEX_ARRAY_LIST_STRIDE_IBM = 103083
GL_TEXTURE_COORD_ARRAY_LIST_STRIDE_IBM = 103084
GL_EDGE_FLAG_ARRAY_LIST_STRIDE_IBM = 103085
GL_FOG_COORDINATE_ARRAY_LIST_STRIDE_IBM = 103086
GL_SECONDARY_COLOR_ARRAY_LIST_STRIDE_IBM = 103087
GL_PREFER_DOUBLEBUFFER_HINT_PGI = 0x1A1F8
GL_CONSERVE_MEMORY_HINT_PGI = 0x1A1FD
GL_RECLAIM_MEMORY_HINT_PGI = 0x1A1FE
GL_NATIVE_GRAPHICS_HANDLE_PGI = 0x1A202
GL_NATIVE_GRAPHICS_BEGIN_HINT_PGI = 0x1A203
GL_NATIVE_GRAPHICS_END_HINT_PGI = 0x1A204
GL_ALWAYS_FAST_HINT_PGI = 0x1A20C
GL_ALWAYS_SOFT_HINT_PGI = 0x1A20D
GL_ALLOW_DRAW_OBJ_HINT_PGI = 0x1A20E
GL_ALLOW_DRAW_WIN_HINT_PGI = 0x1A20F
GL_ALLOW_DRAW_FRG_HINT_PGI = 0x1A210
GL_ALLOW_DRAW_MEM_HINT_PGI = 0x1A211
GL_STRICT_DEPTHFUNC_HINT_PGI = 0x1A216
GL_STRICT_LIGHTING_HINT_PGI = 0x1A217
GL_STRICT_SCISSOR_HINT_PGI = 0x1A218
GL_FULL_STIPPLE_HINT_PGI = 0x1A219
GL_CLIP_NEAR_HINT_PGI = 0x1A220
GL_CLIP_FAR_HINT_PGI = 0x1A221
GL_WIDE_LINE_HINT_PGI = 0x1A222
GL_BACK_NORMALS_HINT_PGI = 0x1A223
GL_VERTEX_DATA_HINT_PGI = 0x1A22A
GL_VERTEX_CONSISTENT_HINT_PGI = 0x1A22B
GL_MATERIAL_SIDE_HINT_PGI = 0x1A22C
GL_MAX_VERTEX_HINT_PGI = 0x1A22D

print 'PyGLoo: initialized enum definitions'

# API entry points
def init():
	print 'PyGLoo: initializing API entry points...'
	gl = Context()
	
	gl.glAccum = _getProcAddress('glAccum')
	gl.glAccum.gl = gl
	gl.glAccum.__name__ = 'glAccum'
	gl.glAccum.errcheck = _errcheck
	gl.glAccum.restype = GLvoid
	gl.glAccum.argtypes = (GLenum, GLfloat,)
	gl.glAccumxOES = _getProcAddress('glAccumxOES')
	gl.glAccumxOES.gl = gl
	gl.glAccumxOES.__name__ = 'glAccumxOES'
	gl.glAccumxOES.errcheck = _errcheck
	gl.glAccumxOES.restype = GLvoid
	gl.glAccumxOES.argtypes = (GLenum, GLfixed,)
	gl.glActiveProgramEXT = _getProcAddress('glActiveProgramEXT')
	gl.glActiveProgramEXT.gl = gl
	gl.glActiveProgramEXT.__name__ = 'glActiveProgramEXT'
	gl.glActiveProgramEXT.errcheck = _errcheck
	gl.glActiveProgramEXT.restype = GLvoid
	gl.glActiveProgramEXT.argtypes = (GLuint,)
	gl.glActiveShaderProgram = _getProcAddress('glActiveShaderProgram')
	gl.glActiveShaderProgram.gl = gl
	gl.glActiveShaderProgram.__name__ = 'glActiveShaderProgram'
	gl.glActiveShaderProgram.errcheck = _errcheck
	gl.glActiveShaderProgram.restype = GLvoid
	gl.glActiveShaderProgram.argtypes = (GLuint, GLuint,)
	gl.glActiveShaderProgramEXT = _getProcAddress('glActiveShaderProgramEXT')
	gl.glActiveShaderProgramEXT.gl = gl
	gl.glActiveShaderProgramEXT.__name__ = 'glActiveShaderProgramEXT'
	gl.glActiveShaderProgramEXT.errcheck = _errcheck
	gl.glActiveShaderProgramEXT.restype = GLvoid
	gl.glActiveShaderProgramEXT.argtypes = (GLuint, GLuint,)
	gl.glActiveStencilFaceEXT = _getProcAddress('glActiveStencilFaceEXT')
	gl.glActiveStencilFaceEXT.gl = gl
	gl.glActiveStencilFaceEXT.__name__ = 'glActiveStencilFaceEXT'
	gl.glActiveStencilFaceEXT.errcheck = _errcheck
	gl.glActiveStencilFaceEXT.restype = GLvoid
	gl.glActiveStencilFaceEXT.argtypes = (GLenum,)
	gl.glActiveTexture = _getProcAddress('glActiveTexture')
	gl.glActiveTexture.gl = gl
	gl.glActiveTexture.__name__ = 'glActiveTexture'
	gl.glActiveTexture.errcheck = _errcheck
	gl.glActiveTexture.restype = GLvoid
	gl.glActiveTexture.argtypes = (GLenum,)
	gl.glActiveTextureARB = _getProcAddress('glActiveTextureARB')
	gl.glActiveTextureARB.gl = gl
	gl.glActiveTextureARB.__name__ = 'glActiveTextureARB'
	gl.glActiveTextureARB.errcheck = _errcheck
	gl.glActiveTextureARB.restype = GLvoid
	gl.glActiveTextureARB.argtypes = (GLenum,)
	gl.glActiveVaryingNV = _getProcAddress('glActiveVaryingNV')
	gl.glActiveVaryingNV.gl = gl
	gl.glActiveVaryingNV.__name__ = 'glActiveVaryingNV'
	gl.glActiveVaryingNV.errcheck = _errcheck
	gl.glActiveVaryingNV.restype = GLvoid
	gl.glActiveVaryingNV.argtypes = (GLuint, POINTER(GLchar),)
	gl.glAlphaFragmentOp1ATI = _getProcAddress('glAlphaFragmentOp1ATI')
	gl.glAlphaFragmentOp1ATI.gl = gl
	gl.glAlphaFragmentOp1ATI.__name__ = 'glAlphaFragmentOp1ATI'
	gl.glAlphaFragmentOp1ATI.errcheck = _errcheck
	gl.glAlphaFragmentOp1ATI.restype = GLvoid
	gl.glAlphaFragmentOp1ATI.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glAlphaFragmentOp2ATI = _getProcAddress('glAlphaFragmentOp2ATI')
	gl.glAlphaFragmentOp2ATI.gl = gl
	gl.glAlphaFragmentOp2ATI.__name__ = 'glAlphaFragmentOp2ATI'
	gl.glAlphaFragmentOp2ATI.errcheck = _errcheck
	gl.glAlphaFragmentOp2ATI.restype = GLvoid
	gl.glAlphaFragmentOp2ATI.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glAlphaFragmentOp3ATI = _getProcAddress('glAlphaFragmentOp3ATI')
	gl.glAlphaFragmentOp3ATI.gl = gl
	gl.glAlphaFragmentOp3ATI.__name__ = 'glAlphaFragmentOp3ATI'
	gl.glAlphaFragmentOp3ATI.errcheck = _errcheck
	gl.glAlphaFragmentOp3ATI.restype = GLvoid
	gl.glAlphaFragmentOp3ATI.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glAlphaFunc = _getProcAddress('glAlphaFunc')
	gl.glAlphaFunc.gl = gl
	gl.glAlphaFunc.__name__ = 'glAlphaFunc'
	gl.glAlphaFunc.errcheck = _errcheck
	gl.glAlphaFunc.restype = GLvoid
	gl.glAlphaFunc.argtypes = (GLenum, GLfloat,)
	gl.glAlphaFuncQCOM = _getProcAddress('glAlphaFuncQCOM')
	gl.glAlphaFuncQCOM.gl = gl
	gl.glAlphaFuncQCOM.__name__ = 'glAlphaFuncQCOM'
	gl.glAlphaFuncQCOM.errcheck = _errcheck
	gl.glAlphaFuncQCOM.restype = GLvoid
	gl.glAlphaFuncQCOM.argtypes = (GLenum, GLclampf,)
	gl.glAlphaFuncx = _getProcAddress('glAlphaFuncx')
	gl.glAlphaFuncx.gl = gl
	gl.glAlphaFuncx.__name__ = 'glAlphaFuncx'
	gl.glAlphaFuncx.errcheck = _errcheck
	gl.glAlphaFuncx.restype = GLvoid
	gl.glAlphaFuncx.argtypes = (GLenum, GLfixed,)
	gl.glAlphaFuncxOES = _getProcAddress('glAlphaFuncxOES')
	gl.glAlphaFuncxOES.gl = gl
	gl.glAlphaFuncxOES.__name__ = 'glAlphaFuncxOES'
	gl.glAlphaFuncxOES.errcheck = _errcheck
	gl.glAlphaFuncxOES.restype = GLvoid
	gl.glAlphaFuncxOES.argtypes = (GLenum, GLfixed,)
	gl.glApplyTextureEXT = _getProcAddress('glApplyTextureEXT')
	gl.glApplyTextureEXT.gl = gl
	gl.glApplyTextureEXT.__name__ = 'glApplyTextureEXT'
	gl.glApplyTextureEXT.errcheck = _errcheck
	gl.glApplyTextureEXT.restype = GLvoid
	gl.glApplyTextureEXT.argtypes = (GLenum,)
	gl.glAreProgramsResidentNV = _getProcAddress('glAreProgramsResidentNV')
	gl.glAreProgramsResidentNV.gl = gl
	gl.glAreProgramsResidentNV.__name__ = 'glAreProgramsResidentNV'
	gl.glAreProgramsResidentNV.errcheck = _errcheck
	gl.glAreProgramsResidentNV.restype = GLboolean
	gl.glAreProgramsResidentNV.argtypes = (GLsizei, POINTER(GLuint), POINTER(GLboolean),)
	gl.glAreTexturesResident = _getProcAddress('glAreTexturesResident')
	gl.glAreTexturesResident.gl = gl
	gl.glAreTexturesResident.__name__ = 'glAreTexturesResident'
	gl.glAreTexturesResident.errcheck = _errcheck
	gl.glAreTexturesResident.restype = GLboolean
	gl.glAreTexturesResident.argtypes = (GLsizei, POINTER(GLuint), POINTER(GLboolean),)
	gl.glAreTexturesResidentEXT = _getProcAddress('glAreTexturesResidentEXT')
	gl.glAreTexturesResidentEXT.gl = gl
	gl.glAreTexturesResidentEXT.__name__ = 'glAreTexturesResidentEXT'
	gl.glAreTexturesResidentEXT.errcheck = _errcheck
	gl.glAreTexturesResidentEXT.restype = GLboolean
	gl.glAreTexturesResidentEXT.argtypes = (GLsizei, POINTER(GLuint), POINTER(GLboolean),)
	gl.glArrayElement = _getProcAddress('glArrayElement')
	gl.glArrayElement.gl = gl
	gl.glArrayElement.__name__ = 'glArrayElement'
	gl.glArrayElement.errcheck = _errcheck
	gl.glArrayElement.restype = GLvoid
	gl.glArrayElement.argtypes = (GLint,)
	gl.glArrayElementEXT = _getProcAddress('glArrayElementEXT')
	gl.glArrayElementEXT.gl = gl
	gl.glArrayElementEXT.__name__ = 'glArrayElementEXT'
	gl.glArrayElementEXT.errcheck = _errcheck
	gl.glArrayElementEXT.restype = GLvoid
	gl.glArrayElementEXT.argtypes = (GLint,)
	gl.glArrayObjectATI = _getProcAddress('glArrayObjectATI')
	gl.glArrayObjectATI.gl = gl
	gl.glArrayObjectATI.__name__ = 'glArrayObjectATI'
	gl.glArrayObjectATI.errcheck = _errcheck
	gl.glArrayObjectATI.restype = GLvoid
	gl.glArrayObjectATI.argtypes = (GLenum, GLint, GLenum, GLsizei, GLuint, GLuint,)
	gl.glAsyncMarkerSGIX = _getProcAddress('glAsyncMarkerSGIX')
	gl.glAsyncMarkerSGIX.gl = gl
	gl.glAsyncMarkerSGIX.__name__ = 'glAsyncMarkerSGIX'
	gl.glAsyncMarkerSGIX.errcheck = _errcheck
	gl.glAsyncMarkerSGIX.restype = GLvoid
	gl.glAsyncMarkerSGIX.argtypes = (GLuint,)
	gl.glAttachObjectARB = _getProcAddress('glAttachObjectARB')
	gl.glAttachObjectARB.gl = gl
	gl.glAttachObjectARB.__name__ = 'glAttachObjectARB'
	gl.glAttachObjectARB.errcheck = _errcheck
	gl.glAttachObjectARB.restype = GLvoid
	gl.glAttachObjectARB.argtypes = (GLhandleARB, GLhandleARB,)
	gl.glAttachShader = _getProcAddress('glAttachShader')
	gl.glAttachShader.gl = gl
	gl.glAttachShader.__name__ = 'glAttachShader'
	gl.glAttachShader.errcheck = _errcheck
	gl.glAttachShader.restype = GLvoid
	gl.glAttachShader.argtypes = (GLuint, GLuint,)
	gl.glBegin = _getProcAddress('glBegin')
	gl.glBegin.gl = gl
	gl.glBegin.__name__ = 'glBegin'
	gl.glBegin.errcheck = _errcheck
	gl.glBegin.restype = GLvoid
	gl.glBegin.argtypes = (GLenum,)
	gl.glBeginConditionalRender = _getProcAddress('glBeginConditionalRender')
	gl.glBeginConditionalRender.gl = gl
	gl.glBeginConditionalRender.__name__ = 'glBeginConditionalRender'
	gl.glBeginConditionalRender.errcheck = _errcheck
	gl.glBeginConditionalRender.restype = GLvoid
	gl.glBeginConditionalRender.argtypes = (GLuint, GLenum,)
	gl.glBeginConditionalRenderNV = _getProcAddress('glBeginConditionalRenderNV')
	gl.glBeginConditionalRenderNV.gl = gl
	gl.glBeginConditionalRenderNV.__name__ = 'glBeginConditionalRenderNV'
	gl.glBeginConditionalRenderNV.errcheck = _errcheck
	gl.glBeginConditionalRenderNV.restype = GLvoid
	gl.glBeginConditionalRenderNV.argtypes = (GLuint, GLenum,)
	gl.glBeginConditionalRenderNVX = _getProcAddress('glBeginConditionalRenderNVX')
	gl.glBeginConditionalRenderNVX.gl = gl
	gl.glBeginConditionalRenderNVX.__name__ = 'glBeginConditionalRenderNVX'
	gl.glBeginConditionalRenderNVX.errcheck = _errcheck
	gl.glBeginConditionalRenderNVX.restype = GLvoid
	gl.glBeginConditionalRenderNVX.argtypes = (GLuint,)
	gl.glBeginFragmentShaderATI = _getProcAddress('glBeginFragmentShaderATI')
	gl.glBeginFragmentShaderATI.gl = gl
	gl.glBeginFragmentShaderATI.__name__ = 'glBeginFragmentShaderATI'
	gl.glBeginFragmentShaderATI.errcheck = _errcheck
	gl.glBeginFragmentShaderATI.restype = GLvoid
	gl.glBeginFragmentShaderATI.argtypes = ()
	gl.glBeginOcclusionQueryNV = _getProcAddress('glBeginOcclusionQueryNV')
	gl.glBeginOcclusionQueryNV.gl = gl
	gl.glBeginOcclusionQueryNV.__name__ = 'glBeginOcclusionQueryNV'
	gl.glBeginOcclusionQueryNV.errcheck = _errcheck
	gl.glBeginOcclusionQueryNV.restype = GLvoid
	gl.glBeginOcclusionQueryNV.argtypes = (GLuint,)
	gl.glBeginPerfMonitorAMD = _getProcAddress('glBeginPerfMonitorAMD')
	gl.glBeginPerfMonitorAMD.gl = gl
	gl.glBeginPerfMonitorAMD.__name__ = 'glBeginPerfMonitorAMD'
	gl.glBeginPerfMonitorAMD.errcheck = _errcheck
	gl.glBeginPerfMonitorAMD.restype = GLvoid
	gl.glBeginPerfMonitorAMD.argtypes = (GLuint,)
	gl.glBeginPerfQueryINTEL = _getProcAddress('glBeginPerfQueryINTEL')
	gl.glBeginPerfQueryINTEL.gl = gl
	gl.glBeginPerfQueryINTEL.__name__ = 'glBeginPerfQueryINTEL'
	gl.glBeginPerfQueryINTEL.errcheck = _errcheck
	gl.glBeginPerfQueryINTEL.restype = GLvoid
	gl.glBeginPerfQueryINTEL.argtypes = (GLuint,)
	gl.glBeginQuery = _getProcAddress('glBeginQuery')
	gl.glBeginQuery.gl = gl
	gl.glBeginQuery.__name__ = 'glBeginQuery'
	gl.glBeginQuery.errcheck = _errcheck
	gl.glBeginQuery.restype = GLvoid
	gl.glBeginQuery.argtypes = (GLenum, GLuint,)
	gl.glBeginQueryARB = _getProcAddress('glBeginQueryARB')
	gl.glBeginQueryARB.gl = gl
	gl.glBeginQueryARB.__name__ = 'glBeginQueryARB'
	gl.glBeginQueryARB.errcheck = _errcheck
	gl.glBeginQueryARB.restype = GLvoid
	gl.glBeginQueryARB.argtypes = (GLenum, GLuint,)
	gl.glBeginQueryEXT = _getProcAddress('glBeginQueryEXT')
	gl.glBeginQueryEXT.gl = gl
	gl.glBeginQueryEXT.__name__ = 'glBeginQueryEXT'
	gl.glBeginQueryEXT.errcheck = _errcheck
	gl.glBeginQueryEXT.restype = GLvoid
	gl.glBeginQueryEXT.argtypes = (GLenum, GLuint,)
	gl.glBeginQueryIndexed = _getProcAddress('glBeginQueryIndexed')
	gl.glBeginQueryIndexed.gl = gl
	gl.glBeginQueryIndexed.__name__ = 'glBeginQueryIndexed'
	gl.glBeginQueryIndexed.errcheck = _errcheck
	gl.glBeginQueryIndexed.restype = GLvoid
	gl.glBeginQueryIndexed.argtypes = (GLenum, GLuint, GLuint,)
	gl.glBeginTransformFeedback = _getProcAddress('glBeginTransformFeedback')
	gl.glBeginTransformFeedback.gl = gl
	gl.glBeginTransformFeedback.__name__ = 'glBeginTransformFeedback'
	gl.glBeginTransformFeedback.errcheck = _errcheck
	gl.glBeginTransformFeedback.restype = GLvoid
	gl.glBeginTransformFeedback.argtypes = (GLenum,)
	gl.glBeginTransformFeedbackEXT = _getProcAddress('glBeginTransformFeedbackEXT')
	gl.glBeginTransformFeedbackEXT.gl = gl
	gl.glBeginTransformFeedbackEXT.__name__ = 'glBeginTransformFeedbackEXT'
	gl.glBeginTransformFeedbackEXT.errcheck = _errcheck
	gl.glBeginTransformFeedbackEXT.restype = GLvoid
	gl.glBeginTransformFeedbackEXT.argtypes = (GLenum,)
	gl.glBeginTransformFeedbackNV = _getProcAddress('glBeginTransformFeedbackNV')
	gl.glBeginTransformFeedbackNV.gl = gl
	gl.glBeginTransformFeedbackNV.__name__ = 'glBeginTransformFeedbackNV'
	gl.glBeginTransformFeedbackNV.errcheck = _errcheck
	gl.glBeginTransformFeedbackNV.restype = GLvoid
	gl.glBeginTransformFeedbackNV.argtypes = (GLenum,)
	gl.glBeginVertexShaderEXT = _getProcAddress('glBeginVertexShaderEXT')
	gl.glBeginVertexShaderEXT.gl = gl
	gl.glBeginVertexShaderEXT.__name__ = 'glBeginVertexShaderEXT'
	gl.glBeginVertexShaderEXT.errcheck = _errcheck
	gl.glBeginVertexShaderEXT.restype = GLvoid
	gl.glBeginVertexShaderEXT.argtypes = ()
	gl.glBeginVideoCaptureNV = _getProcAddress('glBeginVideoCaptureNV')
	gl.glBeginVideoCaptureNV.gl = gl
	gl.glBeginVideoCaptureNV.__name__ = 'glBeginVideoCaptureNV'
	gl.glBeginVideoCaptureNV.errcheck = _errcheck
	gl.glBeginVideoCaptureNV.restype = GLvoid
	gl.glBeginVideoCaptureNV.argtypes = (GLuint,)
	gl.glBindAttribLocation = _getProcAddress('glBindAttribLocation')
	gl.glBindAttribLocation.gl = gl
	gl.glBindAttribLocation.__name__ = 'glBindAttribLocation'
	gl.glBindAttribLocation.errcheck = _errcheck
	gl.glBindAttribLocation.restype = GLvoid
	gl.glBindAttribLocation.argtypes = (GLuint, GLuint, POINTER(GLchar),)
	gl.glBindAttribLocationARB = _getProcAddress('glBindAttribLocationARB')
	gl.glBindAttribLocationARB.gl = gl
	gl.glBindAttribLocationARB.__name__ = 'glBindAttribLocationARB'
	gl.glBindAttribLocationARB.errcheck = _errcheck
	gl.glBindAttribLocationARB.restype = GLvoid
	gl.glBindAttribLocationARB.argtypes = (GLhandleARB, GLuint, POINTER(GLcharARB),)
	gl.glBindBuffer = _getProcAddress('glBindBuffer')
	gl.glBindBuffer.gl = gl
	gl.glBindBuffer.__name__ = 'glBindBuffer'
	gl.glBindBuffer.errcheck = _errcheck
	gl.glBindBuffer.restype = GLvoid
	gl.glBindBuffer.argtypes = (GLenum, GLuint,)
	gl.glBindBufferARB = _getProcAddress('glBindBufferARB')
	gl.glBindBufferARB.gl = gl
	gl.glBindBufferARB.__name__ = 'glBindBufferARB'
	gl.glBindBufferARB.errcheck = _errcheck
	gl.glBindBufferARB.restype = GLvoid
	gl.glBindBufferARB.argtypes = (GLenum, GLuint,)
	gl.glBindBufferBase = _getProcAddress('glBindBufferBase')
	gl.glBindBufferBase.gl = gl
	gl.glBindBufferBase.__name__ = 'glBindBufferBase'
	gl.glBindBufferBase.errcheck = _errcheck
	gl.glBindBufferBase.restype = GLvoid
	gl.glBindBufferBase.argtypes = (GLenum, GLuint, GLuint,)
	gl.glBindBufferBaseEXT = _getProcAddress('glBindBufferBaseEXT')
	gl.glBindBufferBaseEXT.gl = gl
	gl.glBindBufferBaseEXT.__name__ = 'glBindBufferBaseEXT'
	gl.glBindBufferBaseEXT.errcheck = _errcheck
	gl.glBindBufferBaseEXT.restype = GLvoid
	gl.glBindBufferBaseEXT.argtypes = (GLenum, GLuint, GLuint,)
	gl.glBindBufferBaseNV = _getProcAddress('glBindBufferBaseNV')
	gl.glBindBufferBaseNV.gl = gl
	gl.glBindBufferBaseNV.__name__ = 'glBindBufferBaseNV'
	gl.glBindBufferBaseNV.errcheck = _errcheck
	gl.glBindBufferBaseNV.restype = GLvoid
	gl.glBindBufferBaseNV.argtypes = (GLenum, GLuint, GLuint,)
	gl.glBindBufferOffsetEXT = _getProcAddress('glBindBufferOffsetEXT')
	gl.glBindBufferOffsetEXT.gl = gl
	gl.glBindBufferOffsetEXT.__name__ = 'glBindBufferOffsetEXT'
	gl.glBindBufferOffsetEXT.errcheck = _errcheck
	gl.glBindBufferOffsetEXT.restype = GLvoid
	gl.glBindBufferOffsetEXT.argtypes = (GLenum, GLuint, GLuint, GLintptr,)
	gl.glBindBufferOffsetNV = _getProcAddress('glBindBufferOffsetNV')
	gl.glBindBufferOffsetNV.gl = gl
	gl.glBindBufferOffsetNV.__name__ = 'glBindBufferOffsetNV'
	gl.glBindBufferOffsetNV.errcheck = _errcheck
	gl.glBindBufferOffsetNV.restype = GLvoid
	gl.glBindBufferOffsetNV.argtypes = (GLenum, GLuint, GLuint, GLintptr,)
	gl.glBindBufferRange = _getProcAddress('glBindBufferRange')
	gl.glBindBufferRange.gl = gl
	gl.glBindBufferRange.__name__ = 'glBindBufferRange'
	gl.glBindBufferRange.errcheck = _errcheck
	gl.glBindBufferRange.restype = GLvoid
	gl.glBindBufferRange.argtypes = (GLenum, GLuint, GLuint, GLintptr, GLsizeiptr,)
	gl.glBindBufferRangeEXT = _getProcAddress('glBindBufferRangeEXT')
	gl.glBindBufferRangeEXT.gl = gl
	gl.glBindBufferRangeEXT.__name__ = 'glBindBufferRangeEXT'
	gl.glBindBufferRangeEXT.errcheck = _errcheck
	gl.glBindBufferRangeEXT.restype = GLvoid
	gl.glBindBufferRangeEXT.argtypes = (GLenum, GLuint, GLuint, GLintptr, GLsizeiptr,)
	gl.glBindBufferRangeNV = _getProcAddress('glBindBufferRangeNV')
	gl.glBindBufferRangeNV.gl = gl
	gl.glBindBufferRangeNV.__name__ = 'glBindBufferRangeNV'
	gl.glBindBufferRangeNV.errcheck = _errcheck
	gl.glBindBufferRangeNV.restype = GLvoid
	gl.glBindBufferRangeNV.argtypes = (GLenum, GLuint, GLuint, GLintptr, GLsizeiptr,)
	gl.glBindBuffersBase = _getProcAddress('glBindBuffersBase')
	gl.glBindBuffersBase.gl = gl
	gl.glBindBuffersBase.__name__ = 'glBindBuffersBase'
	gl.glBindBuffersBase.errcheck = _errcheck
	gl.glBindBuffersBase.restype = GLvoid
	gl.glBindBuffersBase.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLuint),)
	gl.glBindBuffersRange = _getProcAddress('glBindBuffersRange')
	gl.glBindBuffersRange.gl = gl
	gl.glBindBuffersRange.__name__ = 'glBindBuffersRange'
	gl.glBindBuffersRange.errcheck = _errcheck
	gl.glBindBuffersRange.restype = GLvoid
	gl.glBindBuffersRange.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLuint), POINTER(GLintptr), POINTER(GLsizeiptr),)
	gl.glBindFragDataLocation = _getProcAddress('glBindFragDataLocation')
	gl.glBindFragDataLocation.gl = gl
	gl.glBindFragDataLocation.__name__ = 'glBindFragDataLocation'
	gl.glBindFragDataLocation.errcheck = _errcheck
	gl.glBindFragDataLocation.restype = GLvoid
	gl.glBindFragDataLocation.argtypes = (GLuint, GLuint, POINTER(GLchar),)
	gl.glBindFragDataLocationEXT = _getProcAddress('glBindFragDataLocationEXT')
	gl.glBindFragDataLocationEXT.gl = gl
	gl.glBindFragDataLocationEXT.__name__ = 'glBindFragDataLocationEXT'
	gl.glBindFragDataLocationEXT.errcheck = _errcheck
	gl.glBindFragDataLocationEXT.restype = GLvoid
	gl.glBindFragDataLocationEXT.argtypes = (GLuint, GLuint, POINTER(GLchar),)
	gl.glBindFragDataLocationIndexed = _getProcAddress('glBindFragDataLocationIndexed')
	gl.glBindFragDataLocationIndexed.gl = gl
	gl.glBindFragDataLocationIndexed.__name__ = 'glBindFragDataLocationIndexed'
	gl.glBindFragDataLocationIndexed.errcheck = _errcheck
	gl.glBindFragDataLocationIndexed.restype = GLvoid
	gl.glBindFragDataLocationIndexed.argtypes = (GLuint, GLuint, GLuint, POINTER(GLchar),)
	gl.glBindFragmentShaderATI = _getProcAddress('glBindFragmentShaderATI')
	gl.glBindFragmentShaderATI.gl = gl
	gl.glBindFragmentShaderATI.__name__ = 'glBindFragmentShaderATI'
	gl.glBindFragmentShaderATI.errcheck = _errcheck
	gl.glBindFragmentShaderATI.restype = GLvoid
	gl.glBindFragmentShaderATI.argtypes = (GLuint,)
	gl.glBindFramebuffer = _getProcAddress('glBindFramebuffer')
	gl.glBindFramebuffer.gl = gl
	gl.glBindFramebuffer.__name__ = 'glBindFramebuffer'
	gl.glBindFramebuffer.errcheck = _errcheck
	gl.glBindFramebuffer.restype = GLvoid
	gl.glBindFramebuffer.argtypes = (GLenum, GLuint,)
	gl.glBindFramebufferEXT = _getProcAddress('glBindFramebufferEXT')
	gl.glBindFramebufferEXT.gl = gl
	gl.glBindFramebufferEXT.__name__ = 'glBindFramebufferEXT'
	gl.glBindFramebufferEXT.errcheck = _errcheck
	gl.glBindFramebufferEXT.restype = GLvoid
	gl.glBindFramebufferEXT.argtypes = (GLenum, GLuint,)
	gl.glBindFramebufferOES = _getProcAddress('glBindFramebufferOES')
	gl.glBindFramebufferOES.gl = gl
	gl.glBindFramebufferOES.__name__ = 'glBindFramebufferOES'
	gl.glBindFramebufferOES.errcheck = _errcheck
	gl.glBindFramebufferOES.restype = GLvoid
	gl.glBindFramebufferOES.argtypes = (GLenum, GLuint,)
	gl.glBindImageTexture = _getProcAddress('glBindImageTexture')
	gl.glBindImageTexture.gl = gl
	gl.glBindImageTexture.__name__ = 'glBindImageTexture'
	gl.glBindImageTexture.errcheck = _errcheck
	gl.glBindImageTexture.restype = GLvoid
	gl.glBindImageTexture.argtypes = (GLuint, GLuint, GLint, GLboolean, GLint, GLenum, GLenum,)
	gl.glBindImageTextureEXT = _getProcAddress('glBindImageTextureEXT')
	gl.glBindImageTextureEXT.gl = gl
	gl.glBindImageTextureEXT.__name__ = 'glBindImageTextureEXT'
	gl.glBindImageTextureEXT.errcheck = _errcheck
	gl.glBindImageTextureEXT.restype = GLvoid
	gl.glBindImageTextureEXT.argtypes = (GLuint, GLuint, GLint, GLboolean, GLint, GLenum, GLint,)
	gl.glBindImageTextures = _getProcAddress('glBindImageTextures')
	gl.glBindImageTextures.gl = gl
	gl.glBindImageTextures.__name__ = 'glBindImageTextures'
	gl.glBindImageTextures.errcheck = _errcheck
	gl.glBindImageTextures.restype = GLvoid
	gl.glBindImageTextures.argtypes = (GLuint, GLsizei, POINTER(GLuint),)
	gl.glBindLightParameterEXT = _getProcAddress('glBindLightParameterEXT')
	gl.glBindLightParameterEXT.gl = gl
	gl.glBindLightParameterEXT.__name__ = 'glBindLightParameterEXT'
	gl.glBindLightParameterEXT.errcheck = _errcheck
	gl.glBindLightParameterEXT.restype = GLuint
	gl.glBindLightParameterEXT.argtypes = (GLenum, GLenum,)
	gl.glBindMaterialParameterEXT = _getProcAddress('glBindMaterialParameterEXT')
	gl.glBindMaterialParameterEXT.gl = gl
	gl.glBindMaterialParameterEXT.__name__ = 'glBindMaterialParameterEXT'
	gl.glBindMaterialParameterEXT.errcheck = _errcheck
	gl.glBindMaterialParameterEXT.restype = GLuint
	gl.glBindMaterialParameterEXT.argtypes = (GLenum, GLenum,)
	gl.glBindMultiTextureEXT = _getProcAddress('glBindMultiTextureEXT')
	gl.glBindMultiTextureEXT.gl = gl
	gl.glBindMultiTextureEXT.__name__ = 'glBindMultiTextureEXT'
	gl.glBindMultiTextureEXT.errcheck = _errcheck
	gl.glBindMultiTextureEXT.restype = GLvoid
	gl.glBindMultiTextureEXT.argtypes = (GLenum, GLenum, GLuint,)
	gl.glBindParameterEXT = _getProcAddress('glBindParameterEXT')
	gl.glBindParameterEXT.gl = gl
	gl.glBindParameterEXT.__name__ = 'glBindParameterEXT'
	gl.glBindParameterEXT.errcheck = _errcheck
	gl.glBindParameterEXT.restype = GLuint
	gl.glBindParameterEXT.argtypes = (GLenum,)
	gl.glBindProgramARB = _getProcAddress('glBindProgramARB')
	gl.glBindProgramARB.gl = gl
	gl.glBindProgramARB.__name__ = 'glBindProgramARB'
	gl.glBindProgramARB.errcheck = _errcheck
	gl.glBindProgramARB.restype = GLvoid
	gl.glBindProgramARB.argtypes = (GLenum, GLuint,)
	gl.glBindProgramNV = _getProcAddress('glBindProgramNV')
	gl.glBindProgramNV.gl = gl
	gl.glBindProgramNV.__name__ = 'glBindProgramNV'
	gl.glBindProgramNV.errcheck = _errcheck
	gl.glBindProgramNV.restype = GLvoid
	gl.glBindProgramNV.argtypes = (GLenum, GLuint,)
	gl.glBindProgramPipeline = _getProcAddress('glBindProgramPipeline')
	gl.glBindProgramPipeline.gl = gl
	gl.glBindProgramPipeline.__name__ = 'glBindProgramPipeline'
	gl.glBindProgramPipeline.errcheck = _errcheck
	gl.glBindProgramPipeline.restype = GLvoid
	gl.glBindProgramPipeline.argtypes = (GLuint,)
	gl.glBindProgramPipelineEXT = _getProcAddress('glBindProgramPipelineEXT')
	gl.glBindProgramPipelineEXT.gl = gl
	gl.glBindProgramPipelineEXT.__name__ = 'glBindProgramPipelineEXT'
	gl.glBindProgramPipelineEXT.errcheck = _errcheck
	gl.glBindProgramPipelineEXT.restype = GLvoid
	gl.glBindProgramPipelineEXT.argtypes = (GLuint,)
	gl.glBindRenderbuffer = _getProcAddress('glBindRenderbuffer')
	gl.glBindRenderbuffer.gl = gl
	gl.glBindRenderbuffer.__name__ = 'glBindRenderbuffer'
	gl.glBindRenderbuffer.errcheck = _errcheck
	gl.glBindRenderbuffer.restype = GLvoid
	gl.glBindRenderbuffer.argtypes = (GLenum, GLuint,)
	gl.glBindRenderbufferEXT = _getProcAddress('glBindRenderbufferEXT')
	gl.glBindRenderbufferEXT.gl = gl
	gl.glBindRenderbufferEXT.__name__ = 'glBindRenderbufferEXT'
	gl.glBindRenderbufferEXT.errcheck = _errcheck
	gl.glBindRenderbufferEXT.restype = GLvoid
	gl.glBindRenderbufferEXT.argtypes = (GLenum, GLuint,)
	gl.glBindRenderbufferOES = _getProcAddress('glBindRenderbufferOES')
	gl.glBindRenderbufferOES.gl = gl
	gl.glBindRenderbufferOES.__name__ = 'glBindRenderbufferOES'
	gl.glBindRenderbufferOES.errcheck = _errcheck
	gl.glBindRenderbufferOES.restype = GLvoid
	gl.glBindRenderbufferOES.argtypes = (GLenum, GLuint,)
	gl.glBindSampler = _getProcAddress('glBindSampler')
	gl.glBindSampler.gl = gl
	gl.glBindSampler.__name__ = 'glBindSampler'
	gl.glBindSampler.errcheck = _errcheck
	gl.glBindSampler.restype = GLvoid
	gl.glBindSampler.argtypes = (GLuint, GLuint,)
	gl.glBindSamplers = _getProcAddress('glBindSamplers')
	gl.glBindSamplers.gl = gl
	gl.glBindSamplers.__name__ = 'glBindSamplers'
	gl.glBindSamplers.errcheck = _errcheck
	gl.glBindSamplers.restype = GLvoid
	gl.glBindSamplers.argtypes = (GLuint, GLsizei, POINTER(GLuint),)
	gl.glBindTexGenParameterEXT = _getProcAddress('glBindTexGenParameterEXT')
	gl.glBindTexGenParameterEXT.gl = gl
	gl.glBindTexGenParameterEXT.__name__ = 'glBindTexGenParameterEXT'
	gl.glBindTexGenParameterEXT.errcheck = _errcheck
	gl.glBindTexGenParameterEXT.restype = GLuint
	gl.glBindTexGenParameterEXT.argtypes = (GLenum, GLenum, GLenum,)
	gl.glBindTexture = _getProcAddress('glBindTexture')
	gl.glBindTexture.gl = gl
	gl.glBindTexture.__name__ = 'glBindTexture'
	gl.glBindTexture.errcheck = _errcheck
	gl.glBindTexture.restype = GLvoid
	gl.glBindTexture.argtypes = (GLenum, GLuint,)
	gl.glBindTextureEXT = _getProcAddress('glBindTextureEXT')
	gl.glBindTextureEXT.gl = gl
	gl.glBindTextureEXT.__name__ = 'glBindTextureEXT'
	gl.glBindTextureEXT.errcheck = _errcheck
	gl.glBindTextureEXT.restype = GLvoid
	gl.glBindTextureEXT.argtypes = (GLenum, GLuint,)
	gl.glBindTextureUnit = _getProcAddress('glBindTextureUnit')
	gl.glBindTextureUnit.gl = gl
	gl.glBindTextureUnit.__name__ = 'glBindTextureUnit'
	gl.glBindTextureUnit.errcheck = _errcheck
	gl.glBindTextureUnit.restype = GLvoid
	gl.glBindTextureUnit.argtypes = (GLuint, GLuint,)
	gl.glBindTextureUnitParameterEXT = _getProcAddress('glBindTextureUnitParameterEXT')
	gl.glBindTextureUnitParameterEXT.gl = gl
	gl.glBindTextureUnitParameterEXT.__name__ = 'glBindTextureUnitParameterEXT'
	gl.glBindTextureUnitParameterEXT.errcheck = _errcheck
	gl.glBindTextureUnitParameterEXT.restype = GLuint
	gl.glBindTextureUnitParameterEXT.argtypes = (GLenum, GLenum,)
	gl.glBindTextures = _getProcAddress('glBindTextures')
	gl.glBindTextures.gl = gl
	gl.glBindTextures.__name__ = 'glBindTextures'
	gl.glBindTextures.errcheck = _errcheck
	gl.glBindTextures.restype = GLvoid
	gl.glBindTextures.argtypes = (GLuint, GLsizei, POINTER(GLuint),)
	gl.glBindTransformFeedback = _getProcAddress('glBindTransformFeedback')
	gl.glBindTransformFeedback.gl = gl
	gl.glBindTransformFeedback.__name__ = 'glBindTransformFeedback'
	gl.glBindTransformFeedback.errcheck = _errcheck
	gl.glBindTransformFeedback.restype = GLvoid
	gl.glBindTransformFeedback.argtypes = (GLenum, GLuint,)
	gl.glBindTransformFeedbackNV = _getProcAddress('glBindTransformFeedbackNV')
	gl.glBindTransformFeedbackNV.gl = gl
	gl.glBindTransformFeedbackNV.__name__ = 'glBindTransformFeedbackNV'
	gl.glBindTransformFeedbackNV.errcheck = _errcheck
	gl.glBindTransformFeedbackNV.restype = GLvoid
	gl.glBindTransformFeedbackNV.argtypes = (GLenum, GLuint,)
	gl.glBindVertexArray = _getProcAddress('glBindVertexArray')
	gl.glBindVertexArray.gl = gl
	gl.glBindVertexArray.__name__ = 'glBindVertexArray'
	gl.glBindVertexArray.errcheck = _errcheck
	gl.glBindVertexArray.restype = GLvoid
	gl.glBindVertexArray.argtypes = (GLuint,)
	gl.glBindVertexArrayAPPLE = _getProcAddress('glBindVertexArrayAPPLE')
	gl.glBindVertexArrayAPPLE.gl = gl
	gl.glBindVertexArrayAPPLE.__name__ = 'glBindVertexArrayAPPLE'
	gl.glBindVertexArrayAPPLE.errcheck = _errcheck
	gl.glBindVertexArrayAPPLE.restype = GLvoid
	gl.glBindVertexArrayAPPLE.argtypes = (GLuint,)
	gl.glBindVertexArrayOES = _getProcAddress('glBindVertexArrayOES')
	gl.glBindVertexArrayOES.gl = gl
	gl.glBindVertexArrayOES.__name__ = 'glBindVertexArrayOES'
	gl.glBindVertexArrayOES.errcheck = _errcheck
	gl.glBindVertexArrayOES.restype = GLvoid
	gl.glBindVertexArrayOES.argtypes = (GLuint,)
	gl.glBindVertexBuffer = _getProcAddress('glBindVertexBuffer')
	gl.glBindVertexBuffer.gl = gl
	gl.glBindVertexBuffer.__name__ = 'glBindVertexBuffer'
	gl.glBindVertexBuffer.errcheck = _errcheck
	gl.glBindVertexBuffer.restype = GLvoid
	gl.glBindVertexBuffer.argtypes = (GLuint, GLuint, GLintptr, GLsizei,)
	gl.glBindVertexBuffers = _getProcAddress('glBindVertexBuffers')
	gl.glBindVertexBuffers.gl = gl
	gl.glBindVertexBuffers.__name__ = 'glBindVertexBuffers'
	gl.glBindVertexBuffers.errcheck = _errcheck
	gl.glBindVertexBuffers.restype = GLvoid
	gl.glBindVertexBuffers.argtypes = (GLuint, GLsizei, POINTER(GLuint), POINTER(GLintptr), POINTER(GLsizei),)
	gl.glBindVertexShaderEXT = _getProcAddress('glBindVertexShaderEXT')
	gl.glBindVertexShaderEXT.gl = gl
	gl.glBindVertexShaderEXT.__name__ = 'glBindVertexShaderEXT'
	gl.glBindVertexShaderEXT.errcheck = _errcheck
	gl.glBindVertexShaderEXT.restype = GLvoid
	gl.glBindVertexShaderEXT.argtypes = (GLuint,)
	gl.glBindVideoCaptureStreamBufferNV = _getProcAddress('glBindVideoCaptureStreamBufferNV')
	gl.glBindVideoCaptureStreamBufferNV.gl = gl
	gl.glBindVideoCaptureStreamBufferNV.__name__ = 'glBindVideoCaptureStreamBufferNV'
	gl.glBindVideoCaptureStreamBufferNV.errcheck = _errcheck
	gl.glBindVideoCaptureStreamBufferNV.restype = GLvoid
	gl.glBindVideoCaptureStreamBufferNV.argtypes = (GLuint, GLuint, GLenum, GLintptrARB,)
	gl.glBindVideoCaptureStreamTextureNV = _getProcAddress('glBindVideoCaptureStreamTextureNV')
	gl.glBindVideoCaptureStreamTextureNV.gl = gl
	gl.glBindVideoCaptureStreamTextureNV.__name__ = 'glBindVideoCaptureStreamTextureNV'
	gl.glBindVideoCaptureStreamTextureNV.errcheck = _errcheck
	gl.glBindVideoCaptureStreamTextureNV.restype = GLvoid
	gl.glBindVideoCaptureStreamTextureNV.argtypes = (GLuint, GLuint, GLenum, GLenum, GLuint,)
	gl.glBinormal3bEXT = _getProcAddress('glBinormal3bEXT')
	gl.glBinormal3bEXT.gl = gl
	gl.glBinormal3bEXT.__name__ = 'glBinormal3bEXT'
	gl.glBinormal3bEXT.errcheck = _errcheck
	gl.glBinormal3bEXT.restype = GLvoid
	gl.glBinormal3bEXT.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glBinormal3bvEXT = _getProcAddress('glBinormal3bvEXT')
	gl.glBinormal3bvEXT.gl = gl
	gl.glBinormal3bvEXT.__name__ = 'glBinormal3bvEXT'
	gl.glBinormal3bvEXT.errcheck = _errcheck
	gl.glBinormal3bvEXT.restype = GLvoid
	gl.glBinormal3bvEXT.argtypes = (POINTER(GLbyte),)
	gl.glBinormal3dEXT = _getProcAddress('glBinormal3dEXT')
	gl.glBinormal3dEXT.gl = gl
	gl.glBinormal3dEXT.__name__ = 'glBinormal3dEXT'
	gl.glBinormal3dEXT.errcheck = _errcheck
	gl.glBinormal3dEXT.restype = GLvoid
	gl.glBinormal3dEXT.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glBinormal3dvEXT = _getProcAddress('glBinormal3dvEXT')
	gl.glBinormal3dvEXT.gl = gl
	gl.glBinormal3dvEXT.__name__ = 'glBinormal3dvEXT'
	gl.glBinormal3dvEXT.errcheck = _errcheck
	gl.glBinormal3dvEXT.restype = GLvoid
	gl.glBinormal3dvEXT.argtypes = (POINTER(GLdouble),)
	gl.glBinormal3fEXT = _getProcAddress('glBinormal3fEXT')
	gl.glBinormal3fEXT.gl = gl
	gl.glBinormal3fEXT.__name__ = 'glBinormal3fEXT'
	gl.glBinormal3fEXT.errcheck = _errcheck
	gl.glBinormal3fEXT.restype = GLvoid
	gl.glBinormal3fEXT.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glBinormal3fvEXT = _getProcAddress('glBinormal3fvEXT')
	gl.glBinormal3fvEXT.gl = gl
	gl.glBinormal3fvEXT.__name__ = 'glBinormal3fvEXT'
	gl.glBinormal3fvEXT.errcheck = _errcheck
	gl.glBinormal3fvEXT.restype = GLvoid
	gl.glBinormal3fvEXT.argtypes = (POINTER(GLfloat),)
	gl.glBinormal3iEXT = _getProcAddress('glBinormal3iEXT')
	gl.glBinormal3iEXT.gl = gl
	gl.glBinormal3iEXT.__name__ = 'glBinormal3iEXT'
	gl.glBinormal3iEXT.errcheck = _errcheck
	gl.glBinormal3iEXT.restype = GLvoid
	gl.glBinormal3iEXT.argtypes = (GLint, GLint, GLint,)
	gl.glBinormal3ivEXT = _getProcAddress('glBinormal3ivEXT')
	gl.glBinormal3ivEXT.gl = gl
	gl.glBinormal3ivEXT.__name__ = 'glBinormal3ivEXT'
	gl.glBinormal3ivEXT.errcheck = _errcheck
	gl.glBinormal3ivEXT.restype = GLvoid
	gl.glBinormal3ivEXT.argtypes = (POINTER(GLint),)
	gl.glBinormal3sEXT = _getProcAddress('glBinormal3sEXT')
	gl.glBinormal3sEXT.gl = gl
	gl.glBinormal3sEXT.__name__ = 'glBinormal3sEXT'
	gl.glBinormal3sEXT.errcheck = _errcheck
	gl.glBinormal3sEXT.restype = GLvoid
	gl.glBinormal3sEXT.argtypes = (GLshort, GLshort, GLshort,)
	gl.glBinormal3svEXT = _getProcAddress('glBinormal3svEXT')
	gl.glBinormal3svEXT.gl = gl
	gl.glBinormal3svEXT.__name__ = 'glBinormal3svEXT'
	gl.glBinormal3svEXT.errcheck = _errcheck
	gl.glBinormal3svEXT.restype = GLvoid
	gl.glBinormal3svEXT.argtypes = (POINTER(GLshort),)
	gl.glBinormalPointerEXT = _getProcAddress('glBinormalPointerEXT')
	gl.glBinormalPointerEXT.gl = gl
	gl.glBinormalPointerEXT.__name__ = 'glBinormalPointerEXT'
	gl.glBinormalPointerEXT.errcheck = _errcheck
	gl.glBinormalPointerEXT.restype = GLvoid
	gl.glBinormalPointerEXT.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glBitmap = _getProcAddress('glBitmap')
	gl.glBitmap.gl = gl
	gl.glBitmap.__name__ = 'glBitmap'
	gl.glBitmap.errcheck = _errcheck
	gl.glBitmap.restype = GLvoid
	gl.glBitmap.argtypes = (GLsizei, GLsizei, GLfloat, GLfloat, GLfloat, GLfloat, POINTER(GLubyte),)
	gl.glBitmapxOES = _getProcAddress('glBitmapxOES')
	gl.glBitmapxOES.gl = gl
	gl.glBitmapxOES.__name__ = 'glBitmapxOES'
	gl.glBitmapxOES.errcheck = _errcheck
	gl.glBitmapxOES.restype = GLvoid
	gl.glBitmapxOES.argtypes = (GLsizei, GLsizei, GLfixed, GLfixed, GLfixed, GLfixed, POINTER(GLubyte),)
	gl.glBlendBarrierKHR = _getProcAddress('glBlendBarrierKHR')
	gl.glBlendBarrierKHR.gl = gl
	gl.glBlendBarrierKHR.__name__ = 'glBlendBarrierKHR'
	gl.glBlendBarrierKHR.errcheck = _errcheck
	gl.glBlendBarrierKHR.restype = GLvoid
	gl.glBlendBarrierKHR.argtypes = ()
	gl.glBlendBarrierNV = _getProcAddress('glBlendBarrierNV')
	gl.glBlendBarrierNV.gl = gl
	gl.glBlendBarrierNV.__name__ = 'glBlendBarrierNV'
	gl.glBlendBarrierNV.errcheck = _errcheck
	gl.glBlendBarrierNV.restype = GLvoid
	gl.glBlendBarrierNV.argtypes = ()
	gl.glBlendColor = _getProcAddress('glBlendColor')
	gl.glBlendColor.gl = gl
	gl.glBlendColor.__name__ = 'glBlendColor'
	gl.glBlendColor.errcheck = _errcheck
	gl.glBlendColor.restype = GLvoid
	gl.glBlendColor.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glBlendColorEXT = _getProcAddress('glBlendColorEXT')
	gl.glBlendColorEXT.gl = gl
	gl.glBlendColorEXT.__name__ = 'glBlendColorEXT'
	gl.glBlendColorEXT.errcheck = _errcheck
	gl.glBlendColorEXT.restype = GLvoid
	gl.glBlendColorEXT.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glBlendColorxOES = _getProcAddress('glBlendColorxOES')
	gl.glBlendColorxOES.gl = gl
	gl.glBlendColorxOES.__name__ = 'glBlendColorxOES'
	gl.glBlendColorxOES.errcheck = _errcheck
	gl.glBlendColorxOES.restype = GLvoid
	gl.glBlendColorxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glBlendEquation = _getProcAddress('glBlendEquation')
	gl.glBlendEquation.gl = gl
	gl.glBlendEquation.__name__ = 'glBlendEquation'
	gl.glBlendEquation.errcheck = _errcheck
	gl.glBlendEquation.restype = GLvoid
	gl.glBlendEquation.argtypes = (GLenum,)
	gl.glBlendEquationEXT = _getProcAddress('glBlendEquationEXT')
	gl.glBlendEquationEXT.gl = gl
	gl.glBlendEquationEXT.__name__ = 'glBlendEquationEXT'
	gl.glBlendEquationEXT.errcheck = _errcheck
	gl.glBlendEquationEXT.restype = GLvoid
	gl.glBlendEquationEXT.argtypes = (GLenum,)
	gl.glBlendEquationIndexedAMD = _getProcAddress('glBlendEquationIndexedAMD')
	gl.glBlendEquationIndexedAMD.gl = gl
	gl.glBlendEquationIndexedAMD.__name__ = 'glBlendEquationIndexedAMD'
	gl.glBlendEquationIndexedAMD.errcheck = _errcheck
	gl.glBlendEquationIndexedAMD.restype = GLvoid
	gl.glBlendEquationIndexedAMD.argtypes = (GLuint, GLenum,)
	gl.glBlendEquationOES = _getProcAddress('glBlendEquationOES')
	gl.glBlendEquationOES.gl = gl
	gl.glBlendEquationOES.__name__ = 'glBlendEquationOES'
	gl.glBlendEquationOES.errcheck = _errcheck
	gl.glBlendEquationOES.restype = GLvoid
	gl.glBlendEquationOES.argtypes = (GLenum,)
	gl.glBlendEquationSeparate = _getProcAddress('glBlendEquationSeparate')
	gl.glBlendEquationSeparate.gl = gl
	gl.glBlendEquationSeparate.__name__ = 'glBlendEquationSeparate'
	gl.glBlendEquationSeparate.errcheck = _errcheck
	gl.glBlendEquationSeparate.restype = GLvoid
	gl.glBlendEquationSeparate.argtypes = (GLenum, GLenum,)
	gl.glBlendEquationSeparateEXT = _getProcAddress('glBlendEquationSeparateEXT')
	gl.glBlendEquationSeparateEXT.gl = gl
	gl.glBlendEquationSeparateEXT.__name__ = 'glBlendEquationSeparateEXT'
	gl.glBlendEquationSeparateEXT.errcheck = _errcheck
	gl.glBlendEquationSeparateEXT.restype = GLvoid
	gl.glBlendEquationSeparateEXT.argtypes = (GLenum, GLenum,)
	gl.glBlendEquationSeparateIndexedAMD = _getProcAddress('glBlendEquationSeparateIndexedAMD')
	gl.glBlendEquationSeparateIndexedAMD.gl = gl
	gl.glBlendEquationSeparateIndexedAMD.__name__ = 'glBlendEquationSeparateIndexedAMD'
	gl.glBlendEquationSeparateIndexedAMD.errcheck = _errcheck
	gl.glBlendEquationSeparateIndexedAMD.restype = GLvoid
	gl.glBlendEquationSeparateIndexedAMD.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendEquationSeparateOES = _getProcAddress('glBlendEquationSeparateOES')
	gl.glBlendEquationSeparateOES.gl = gl
	gl.glBlendEquationSeparateOES.__name__ = 'glBlendEquationSeparateOES'
	gl.glBlendEquationSeparateOES.errcheck = _errcheck
	gl.glBlendEquationSeparateOES.restype = GLvoid
	gl.glBlendEquationSeparateOES.argtypes = (GLenum, GLenum,)
	gl.glBlendEquationSeparatei = _getProcAddress('glBlendEquationSeparatei')
	gl.glBlendEquationSeparatei.gl = gl
	gl.glBlendEquationSeparatei.__name__ = 'glBlendEquationSeparatei'
	gl.glBlendEquationSeparatei.errcheck = _errcheck
	gl.glBlendEquationSeparatei.restype = GLvoid
	gl.glBlendEquationSeparatei.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendEquationSeparateiARB = _getProcAddress('glBlendEquationSeparateiARB')
	gl.glBlendEquationSeparateiARB.gl = gl
	gl.glBlendEquationSeparateiARB.__name__ = 'glBlendEquationSeparateiARB'
	gl.glBlendEquationSeparateiARB.errcheck = _errcheck
	gl.glBlendEquationSeparateiARB.restype = GLvoid
	gl.glBlendEquationSeparateiARB.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendEquationSeparateiEXT = _getProcAddress('glBlendEquationSeparateiEXT')
	gl.glBlendEquationSeparateiEXT.gl = gl
	gl.glBlendEquationSeparateiEXT.__name__ = 'glBlendEquationSeparateiEXT'
	gl.glBlendEquationSeparateiEXT.errcheck = _errcheck
	gl.glBlendEquationSeparateiEXT.restype = GLvoid
	gl.glBlendEquationSeparateiEXT.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendEquationi = _getProcAddress('glBlendEquationi')
	gl.glBlendEquationi.gl = gl
	gl.glBlendEquationi.__name__ = 'glBlendEquationi'
	gl.glBlendEquationi.errcheck = _errcheck
	gl.glBlendEquationi.restype = GLvoid
	gl.glBlendEquationi.argtypes = (GLuint, GLenum,)
	gl.glBlendEquationiARB = _getProcAddress('glBlendEquationiARB')
	gl.glBlendEquationiARB.gl = gl
	gl.glBlendEquationiARB.__name__ = 'glBlendEquationiARB'
	gl.glBlendEquationiARB.errcheck = _errcheck
	gl.glBlendEquationiARB.restype = GLvoid
	gl.glBlendEquationiARB.argtypes = (GLuint, GLenum,)
	gl.glBlendEquationiEXT = _getProcAddress('glBlendEquationiEXT')
	gl.glBlendEquationiEXT.gl = gl
	gl.glBlendEquationiEXT.__name__ = 'glBlendEquationiEXT'
	gl.glBlendEquationiEXT.errcheck = _errcheck
	gl.glBlendEquationiEXT.restype = GLvoid
	gl.glBlendEquationiEXT.argtypes = (GLuint, GLenum,)
	gl.glBlendFunc = _getProcAddress('glBlendFunc')
	gl.glBlendFunc.gl = gl
	gl.glBlendFunc.__name__ = 'glBlendFunc'
	gl.glBlendFunc.errcheck = _errcheck
	gl.glBlendFunc.restype = GLvoid
	gl.glBlendFunc.argtypes = (GLenum, GLenum,)
	gl.glBlendFuncIndexedAMD = _getProcAddress('glBlendFuncIndexedAMD')
	gl.glBlendFuncIndexedAMD.gl = gl
	gl.glBlendFuncIndexedAMD.__name__ = 'glBlendFuncIndexedAMD'
	gl.glBlendFuncIndexedAMD.errcheck = _errcheck
	gl.glBlendFuncIndexedAMD.restype = GLvoid
	gl.glBlendFuncIndexedAMD.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendFuncSeparate = _getProcAddress('glBlendFuncSeparate')
	gl.glBlendFuncSeparate.gl = gl
	gl.glBlendFuncSeparate.__name__ = 'glBlendFuncSeparate'
	gl.glBlendFuncSeparate.errcheck = _errcheck
	gl.glBlendFuncSeparate.restype = GLvoid
	gl.glBlendFuncSeparate.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparateEXT = _getProcAddress('glBlendFuncSeparateEXT')
	gl.glBlendFuncSeparateEXT.gl = gl
	gl.glBlendFuncSeparateEXT.__name__ = 'glBlendFuncSeparateEXT'
	gl.glBlendFuncSeparateEXT.errcheck = _errcheck
	gl.glBlendFuncSeparateEXT.restype = GLvoid
	gl.glBlendFuncSeparateEXT.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparateINGR = _getProcAddress('glBlendFuncSeparateINGR')
	gl.glBlendFuncSeparateINGR.gl = gl
	gl.glBlendFuncSeparateINGR.__name__ = 'glBlendFuncSeparateINGR'
	gl.glBlendFuncSeparateINGR.errcheck = _errcheck
	gl.glBlendFuncSeparateINGR.restype = GLvoid
	gl.glBlendFuncSeparateINGR.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparateIndexedAMD = _getProcAddress('glBlendFuncSeparateIndexedAMD')
	gl.glBlendFuncSeparateIndexedAMD.gl = gl
	gl.glBlendFuncSeparateIndexedAMD.__name__ = 'glBlendFuncSeparateIndexedAMD'
	gl.glBlendFuncSeparateIndexedAMD.errcheck = _errcheck
	gl.glBlendFuncSeparateIndexedAMD.restype = GLvoid
	gl.glBlendFuncSeparateIndexedAMD.argtypes = (GLuint, GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparateOES = _getProcAddress('glBlendFuncSeparateOES')
	gl.glBlendFuncSeparateOES.gl = gl
	gl.glBlendFuncSeparateOES.__name__ = 'glBlendFuncSeparateOES'
	gl.glBlendFuncSeparateOES.errcheck = _errcheck
	gl.glBlendFuncSeparateOES.restype = GLvoid
	gl.glBlendFuncSeparateOES.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparatei = _getProcAddress('glBlendFuncSeparatei')
	gl.glBlendFuncSeparatei.gl = gl
	gl.glBlendFuncSeparatei.__name__ = 'glBlendFuncSeparatei'
	gl.glBlendFuncSeparatei.errcheck = _errcheck
	gl.glBlendFuncSeparatei.restype = GLvoid
	gl.glBlendFuncSeparatei.argtypes = (GLuint, GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparateiARB = _getProcAddress('glBlendFuncSeparateiARB')
	gl.glBlendFuncSeparateiARB.gl = gl
	gl.glBlendFuncSeparateiARB.__name__ = 'glBlendFuncSeparateiARB'
	gl.glBlendFuncSeparateiARB.errcheck = _errcheck
	gl.glBlendFuncSeparateiARB.restype = GLvoid
	gl.glBlendFuncSeparateiARB.argtypes = (GLuint, GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFuncSeparateiEXT = _getProcAddress('glBlendFuncSeparateiEXT')
	gl.glBlendFuncSeparateiEXT.gl = gl
	gl.glBlendFuncSeparateiEXT.__name__ = 'glBlendFuncSeparateiEXT'
	gl.glBlendFuncSeparateiEXT.errcheck = _errcheck
	gl.glBlendFuncSeparateiEXT.restype = GLvoid
	gl.glBlendFuncSeparateiEXT.argtypes = (GLuint, GLenum, GLenum, GLenum, GLenum,)
	gl.glBlendFunci = _getProcAddress('glBlendFunci')
	gl.glBlendFunci.gl = gl
	gl.glBlendFunci.__name__ = 'glBlendFunci'
	gl.glBlendFunci.errcheck = _errcheck
	gl.glBlendFunci.restype = GLvoid
	gl.glBlendFunci.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendFunciARB = _getProcAddress('glBlendFunciARB')
	gl.glBlendFunciARB.gl = gl
	gl.glBlendFunciARB.__name__ = 'glBlendFunciARB'
	gl.glBlendFunciARB.errcheck = _errcheck
	gl.glBlendFunciARB.restype = GLvoid
	gl.glBlendFunciARB.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendFunciEXT = _getProcAddress('glBlendFunciEXT')
	gl.glBlendFunciEXT.gl = gl
	gl.glBlendFunciEXT.__name__ = 'glBlendFunciEXT'
	gl.glBlendFunciEXT.errcheck = _errcheck
	gl.glBlendFunciEXT.restype = GLvoid
	gl.glBlendFunciEXT.argtypes = (GLuint, GLenum, GLenum,)
	gl.glBlendParameteriNV = _getProcAddress('glBlendParameteriNV')
	gl.glBlendParameteriNV.gl = gl
	gl.glBlendParameteriNV.__name__ = 'glBlendParameteriNV'
	gl.glBlendParameteriNV.errcheck = _errcheck
	gl.glBlendParameteriNV.restype = GLvoid
	gl.glBlendParameteriNV.argtypes = (GLenum, GLint,)
	gl.glBlitFramebuffer = _getProcAddress('glBlitFramebuffer')
	gl.glBlitFramebuffer.gl = gl
	gl.glBlitFramebuffer.__name__ = 'glBlitFramebuffer'
	gl.glBlitFramebuffer.errcheck = _errcheck
	gl.glBlitFramebuffer.restype = GLvoid
	gl.glBlitFramebuffer.argtypes = (GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLbitfield, GLenum,)
	gl.glBlitFramebufferANGLE = _getProcAddress('glBlitFramebufferANGLE')
	gl.glBlitFramebufferANGLE.gl = gl
	gl.glBlitFramebufferANGLE.__name__ = 'glBlitFramebufferANGLE'
	gl.glBlitFramebufferANGLE.errcheck = _errcheck
	gl.glBlitFramebufferANGLE.restype = GLvoid
	gl.glBlitFramebufferANGLE.argtypes = (GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLbitfield, GLenum,)
	gl.glBlitFramebufferEXT = _getProcAddress('glBlitFramebufferEXT')
	gl.glBlitFramebufferEXT.gl = gl
	gl.glBlitFramebufferEXT.__name__ = 'glBlitFramebufferEXT'
	gl.glBlitFramebufferEXT.errcheck = _errcheck
	gl.glBlitFramebufferEXT.restype = GLvoid
	gl.glBlitFramebufferEXT.argtypes = (GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLbitfield, GLenum,)
	gl.glBlitFramebufferNV = _getProcAddress('glBlitFramebufferNV')
	gl.glBlitFramebufferNV.gl = gl
	gl.glBlitFramebufferNV.__name__ = 'glBlitFramebufferNV'
	gl.glBlitFramebufferNV.errcheck = _errcheck
	gl.glBlitFramebufferNV.restype = GLvoid
	gl.glBlitFramebufferNV.argtypes = (GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLbitfield, GLenum,)
	gl.glBlitNamedFramebuffer = _getProcAddress('glBlitNamedFramebuffer')
	gl.glBlitNamedFramebuffer.gl = gl
	gl.glBlitNamedFramebuffer.__name__ = 'glBlitNamedFramebuffer'
	gl.glBlitNamedFramebuffer.errcheck = _errcheck
	gl.glBlitNamedFramebuffer.restype = GLvoid
	gl.glBlitNamedFramebuffer.argtypes = (GLuint, GLuint, GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLint, GLbitfield, GLenum,)
	gl.glBufferAddressRangeNV = _getProcAddress('glBufferAddressRangeNV')
	gl.glBufferAddressRangeNV.gl = gl
	gl.glBufferAddressRangeNV.__name__ = 'glBufferAddressRangeNV'
	gl.glBufferAddressRangeNV.errcheck = _errcheck
	gl.glBufferAddressRangeNV.restype = GLvoid
	gl.glBufferAddressRangeNV.argtypes = (GLenum, GLuint, GLuint64EXT, GLsizeiptr,)
	gl.glBufferData = _getProcAddress('glBufferData')
	gl.glBufferData.gl = gl
	gl.glBufferData.__name__ = 'glBufferData'
	gl.glBufferData.errcheck = _errcheck
	gl.glBufferData.restype = GLvoid
	gl.glBufferData.argtypes = (GLenum, GLsizeiptr, POINTER(GLvoid), GLenum,)
	gl.glBufferDataARB = _getProcAddress('glBufferDataARB')
	gl.glBufferDataARB.gl = gl
	gl.glBufferDataARB.__name__ = 'glBufferDataARB'
	gl.glBufferDataARB.errcheck = _errcheck
	gl.glBufferDataARB.restype = GLvoid
	gl.glBufferDataARB.argtypes = (GLenum, GLsizeiptrARB, POINTER(GLvoid), GLenum,)
	gl.glBufferPageCommitmentARB = _getProcAddress('glBufferPageCommitmentARB')
	gl.glBufferPageCommitmentARB.gl = gl
	gl.glBufferPageCommitmentARB.__name__ = 'glBufferPageCommitmentARB'
	gl.glBufferPageCommitmentARB.errcheck = _errcheck
	gl.glBufferPageCommitmentARB.restype = GLvoid
	gl.glBufferPageCommitmentARB.argtypes = (GLenum, GLintptr, GLsizeiptr, GLboolean,)
	gl.glBufferParameteriAPPLE = _getProcAddress('glBufferParameteriAPPLE')
	gl.glBufferParameteriAPPLE.gl = gl
	gl.glBufferParameteriAPPLE.__name__ = 'glBufferParameteriAPPLE'
	gl.glBufferParameteriAPPLE.errcheck = _errcheck
	gl.glBufferParameteriAPPLE.restype = GLvoid
	gl.glBufferParameteriAPPLE.argtypes = (GLenum, GLenum, GLint,)
	gl.glBufferStorage = _getProcAddress('glBufferStorage')
	gl.glBufferStorage.gl = gl
	gl.glBufferStorage.__name__ = 'glBufferStorage'
	gl.glBufferStorage.errcheck = _errcheck
	gl.glBufferStorage.restype = GLvoid
	gl.glBufferStorage.argtypes = (GLenum, GLsizeiptr, POINTER(GLvoid), GLbitfield,)
	gl.glBufferSubData = _getProcAddress('glBufferSubData')
	gl.glBufferSubData.gl = gl
	gl.glBufferSubData.__name__ = 'glBufferSubData'
	gl.glBufferSubData.errcheck = _errcheck
	gl.glBufferSubData.restype = GLvoid
	gl.glBufferSubData.argtypes = (GLenum, GLintptr, GLsizeiptr, POINTER(GLvoid),)
	gl.glBufferSubDataARB = _getProcAddress('glBufferSubDataARB')
	gl.glBufferSubDataARB.gl = gl
	gl.glBufferSubDataARB.__name__ = 'glBufferSubDataARB'
	gl.glBufferSubDataARB.errcheck = _errcheck
	gl.glBufferSubDataARB.restype = GLvoid
	gl.glBufferSubDataARB.argtypes = (GLenum, GLintptrARB, GLsizeiptrARB, POINTER(GLvoid),)
	gl.glCallList = _getProcAddress('glCallList')
	gl.glCallList.gl = gl
	gl.glCallList.__name__ = 'glCallList'
	gl.glCallList.errcheck = _errcheck
	gl.glCallList.restype = GLvoid
	gl.glCallList.argtypes = (GLuint,)
	gl.glCallLists = _getProcAddress('glCallLists')
	gl.glCallLists.gl = gl
	gl.glCallLists.__name__ = 'glCallLists'
	gl.glCallLists.errcheck = _errcheck
	gl.glCallLists.restype = GLvoid
	gl.glCallLists.argtypes = (GLsizei, GLenum, POINTER(GLvoid),)
	gl.glCheckFramebufferStatus = _getProcAddress('glCheckFramebufferStatus')
	gl.glCheckFramebufferStatus.gl = gl
	gl.glCheckFramebufferStatus.__name__ = 'glCheckFramebufferStatus'
	gl.glCheckFramebufferStatus.errcheck = _errcheck
	gl.glCheckFramebufferStatus.restype = GLenum
	gl.glCheckFramebufferStatus.argtypes = (GLenum,)
	gl.glCheckFramebufferStatusEXT = _getProcAddress('glCheckFramebufferStatusEXT')
	gl.glCheckFramebufferStatusEXT.gl = gl
	gl.glCheckFramebufferStatusEXT.__name__ = 'glCheckFramebufferStatusEXT'
	gl.glCheckFramebufferStatusEXT.errcheck = _errcheck
	gl.glCheckFramebufferStatusEXT.restype = GLenum
	gl.glCheckFramebufferStatusEXT.argtypes = (GLenum,)
	gl.glCheckFramebufferStatusOES = _getProcAddress('glCheckFramebufferStatusOES')
	gl.glCheckFramebufferStatusOES.gl = gl
	gl.glCheckFramebufferStatusOES.__name__ = 'glCheckFramebufferStatusOES'
	gl.glCheckFramebufferStatusOES.errcheck = _errcheck
	gl.glCheckFramebufferStatusOES.restype = GLenum
	gl.glCheckFramebufferStatusOES.argtypes = (GLenum,)
	gl.glCheckNamedFramebufferStatus = _getProcAddress('glCheckNamedFramebufferStatus')
	gl.glCheckNamedFramebufferStatus.gl = gl
	gl.glCheckNamedFramebufferStatus.__name__ = 'glCheckNamedFramebufferStatus'
	gl.glCheckNamedFramebufferStatus.errcheck = _errcheck
	gl.glCheckNamedFramebufferStatus.restype = GLenum
	gl.glCheckNamedFramebufferStatus.argtypes = (GLuint, GLenum,)
	gl.glCheckNamedFramebufferStatusEXT = _getProcAddress('glCheckNamedFramebufferStatusEXT')
	gl.glCheckNamedFramebufferStatusEXT.gl = gl
	gl.glCheckNamedFramebufferStatusEXT.__name__ = 'glCheckNamedFramebufferStatusEXT'
	gl.glCheckNamedFramebufferStatusEXT.errcheck = _errcheck
	gl.glCheckNamedFramebufferStatusEXT.restype = GLenum
	gl.glCheckNamedFramebufferStatusEXT.argtypes = (GLuint, GLenum,)
	gl.glClampColor = _getProcAddress('glClampColor')
	gl.glClampColor.gl = gl
	gl.glClampColor.__name__ = 'glClampColor'
	gl.glClampColor.errcheck = _errcheck
	gl.glClampColor.restype = GLvoid
	gl.glClampColor.argtypes = (GLenum, GLenum,)
	gl.glClampColorARB = _getProcAddress('glClampColorARB')
	gl.glClampColorARB.gl = gl
	gl.glClampColorARB.__name__ = 'glClampColorARB'
	gl.glClampColorARB.errcheck = _errcheck
	gl.glClampColorARB.restype = GLvoid
	gl.glClampColorARB.argtypes = (GLenum, GLenum,)
	gl.glClear = _getProcAddress('glClear')
	gl.glClear.gl = gl
	gl.glClear.__name__ = 'glClear'
	gl.glClear.errcheck = _errcheck
	gl.glClear.restype = GLvoid
	gl.glClear.argtypes = (GLbitfield,)
	gl.glClearAccum = _getProcAddress('glClearAccum')
	gl.glClearAccum.gl = gl
	gl.glClearAccum.__name__ = 'glClearAccum'
	gl.glClearAccum.errcheck = _errcheck
	gl.glClearAccum.restype = GLvoid
	gl.glClearAccum.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glClearAccumxOES = _getProcAddress('glClearAccumxOES')
	gl.glClearAccumxOES.gl = gl
	gl.glClearAccumxOES.__name__ = 'glClearAccumxOES'
	gl.glClearAccumxOES.errcheck = _errcheck
	gl.glClearAccumxOES.restype = GLvoid
	gl.glClearAccumxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glClearBufferData = _getProcAddress('glClearBufferData')
	gl.glClearBufferData.gl = gl
	gl.glClearBufferData.__name__ = 'glClearBufferData'
	gl.glClearBufferData.errcheck = _errcheck
	gl.glClearBufferData.restype = GLvoid
	gl.glClearBufferData.argtypes = (GLenum, GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearBufferSubData = _getProcAddress('glClearBufferSubData')
	gl.glClearBufferSubData.gl = gl
	gl.glClearBufferSubData.__name__ = 'glClearBufferSubData'
	gl.glClearBufferSubData.errcheck = _errcheck
	gl.glClearBufferSubData.restype = GLvoid
	gl.glClearBufferSubData.argtypes = (GLenum, GLenum, GLintptr, GLsizeiptr, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearBufferfi = _getProcAddress('glClearBufferfi')
	gl.glClearBufferfi.gl = gl
	gl.glClearBufferfi.__name__ = 'glClearBufferfi'
	gl.glClearBufferfi.errcheck = _errcheck
	gl.glClearBufferfi.restype = GLvoid
	gl.glClearBufferfi.argtypes = (GLenum, GLint, GLfloat, GLint,)
	gl.glClearBufferfv = _getProcAddress('glClearBufferfv')
	gl.glClearBufferfv.gl = gl
	gl.glClearBufferfv.__name__ = 'glClearBufferfv'
	gl.glClearBufferfv.errcheck = _errcheck
	gl.glClearBufferfv.restype = GLvoid
	gl.glClearBufferfv.argtypes = (GLenum, GLint, POINTER(GLfloat),)
	gl.glClearBufferiv = _getProcAddress('glClearBufferiv')
	gl.glClearBufferiv.gl = gl
	gl.glClearBufferiv.__name__ = 'glClearBufferiv'
	gl.glClearBufferiv.errcheck = _errcheck
	gl.glClearBufferiv.restype = GLvoid
	gl.glClearBufferiv.argtypes = (GLenum, GLint, POINTER(GLint),)
	gl.glClearBufferuiv = _getProcAddress('glClearBufferuiv')
	gl.glClearBufferuiv.gl = gl
	gl.glClearBufferuiv.__name__ = 'glClearBufferuiv'
	gl.glClearBufferuiv.errcheck = _errcheck
	gl.glClearBufferuiv.restype = GLvoid
	gl.glClearBufferuiv.argtypes = (GLenum, GLint, POINTER(GLuint),)
	gl.glClearColor = _getProcAddress('glClearColor')
	gl.glClearColor.gl = gl
	gl.glClearColor.__name__ = 'glClearColor'
	gl.glClearColor.errcheck = _errcheck
	gl.glClearColor.restype = GLvoid
	gl.glClearColor.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glClearColorIiEXT = _getProcAddress('glClearColorIiEXT')
	gl.glClearColorIiEXT.gl = gl
	gl.glClearColorIiEXT.__name__ = 'glClearColorIiEXT'
	gl.glClearColorIiEXT.errcheck = _errcheck
	gl.glClearColorIiEXT.restype = GLvoid
	gl.glClearColorIiEXT.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glClearColorIuiEXT = _getProcAddress('glClearColorIuiEXT')
	gl.glClearColorIuiEXT.gl = gl
	gl.glClearColorIuiEXT.__name__ = 'glClearColorIuiEXT'
	gl.glClearColorIuiEXT.errcheck = _errcheck
	gl.glClearColorIuiEXT.restype = GLvoid
	gl.glClearColorIuiEXT.argtypes = (GLuint, GLuint, GLuint, GLuint,)
	gl.glClearColorx = _getProcAddress('glClearColorx')
	gl.glClearColorx.gl = gl
	gl.glClearColorx.__name__ = 'glClearColorx'
	gl.glClearColorx.errcheck = _errcheck
	gl.glClearColorx.restype = GLvoid
	gl.glClearColorx.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glClearColorxOES = _getProcAddress('glClearColorxOES')
	gl.glClearColorxOES.gl = gl
	gl.glClearColorxOES.__name__ = 'glClearColorxOES'
	gl.glClearColorxOES.errcheck = _errcheck
	gl.glClearColorxOES.restype = GLvoid
	gl.glClearColorxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glClearDepth = _getProcAddress('glClearDepth')
	gl.glClearDepth.gl = gl
	gl.glClearDepth.__name__ = 'glClearDepth'
	gl.glClearDepth.errcheck = _errcheck
	gl.glClearDepth.restype = GLvoid
	gl.glClearDepth.argtypes = (GLdouble,)
	gl.glClearDepthdNV = _getProcAddress('glClearDepthdNV')
	gl.glClearDepthdNV.gl = gl
	gl.glClearDepthdNV.__name__ = 'glClearDepthdNV'
	gl.glClearDepthdNV.errcheck = _errcheck
	gl.glClearDepthdNV.restype = GLvoid
	gl.glClearDepthdNV.argtypes = (GLdouble,)
	gl.glClearDepthf = _getProcAddress('glClearDepthf')
	gl.glClearDepthf.gl = gl
	gl.glClearDepthf.__name__ = 'glClearDepthf'
	gl.glClearDepthf.errcheck = _errcheck
	gl.glClearDepthf.restype = GLvoid
	gl.glClearDepthf.argtypes = (GLfloat,)
	gl.glClearDepthfOES = _getProcAddress('glClearDepthfOES')
	gl.glClearDepthfOES.gl = gl
	gl.glClearDepthfOES.__name__ = 'glClearDepthfOES'
	gl.glClearDepthfOES.errcheck = _errcheck
	gl.glClearDepthfOES.restype = GLvoid
	gl.glClearDepthfOES.argtypes = (GLclampf,)
	gl.glClearDepthx = _getProcAddress('glClearDepthx')
	gl.glClearDepthx.gl = gl
	gl.glClearDepthx.__name__ = 'glClearDepthx'
	gl.glClearDepthx.errcheck = _errcheck
	gl.glClearDepthx.restype = GLvoid
	gl.glClearDepthx.argtypes = (GLfixed,)
	gl.glClearDepthxOES = _getProcAddress('glClearDepthxOES')
	gl.glClearDepthxOES.gl = gl
	gl.glClearDepthxOES.__name__ = 'glClearDepthxOES'
	gl.glClearDepthxOES.errcheck = _errcheck
	gl.glClearDepthxOES.restype = GLvoid
	gl.glClearDepthxOES.argtypes = (GLfixed,)
	gl.glClearIndex = _getProcAddress('glClearIndex')
	gl.glClearIndex.gl = gl
	gl.glClearIndex.__name__ = 'glClearIndex'
	gl.glClearIndex.errcheck = _errcheck
	gl.glClearIndex.restype = GLvoid
	gl.glClearIndex.argtypes = (GLfloat,)
	gl.glClearNamedBufferData = _getProcAddress('glClearNamedBufferData')
	gl.glClearNamedBufferData.gl = gl
	gl.glClearNamedBufferData.__name__ = 'glClearNamedBufferData'
	gl.glClearNamedBufferData.errcheck = _errcheck
	gl.glClearNamedBufferData.restype = GLvoid
	gl.glClearNamedBufferData.argtypes = (GLuint, GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearNamedBufferDataEXT = _getProcAddress('glClearNamedBufferDataEXT')
	gl.glClearNamedBufferDataEXT.gl = gl
	gl.glClearNamedBufferDataEXT.__name__ = 'glClearNamedBufferDataEXT'
	gl.glClearNamedBufferDataEXT.errcheck = _errcheck
	gl.glClearNamedBufferDataEXT.restype = GLvoid
	gl.glClearNamedBufferDataEXT.argtypes = (GLuint, GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearNamedBufferSubData = _getProcAddress('glClearNamedBufferSubData')
	gl.glClearNamedBufferSubData.gl = gl
	gl.glClearNamedBufferSubData.__name__ = 'glClearNamedBufferSubData'
	gl.glClearNamedBufferSubData.errcheck = _errcheck
	gl.glClearNamedBufferSubData.restype = GLvoid
	gl.glClearNamedBufferSubData.argtypes = (GLuint, GLenum, GLintptr, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearNamedBufferSubDataEXT = _getProcAddress('glClearNamedBufferSubDataEXT')
	gl.glClearNamedBufferSubDataEXT.gl = gl
	gl.glClearNamedBufferSubDataEXT.__name__ = 'glClearNamedBufferSubDataEXT'
	gl.glClearNamedBufferSubDataEXT.errcheck = _errcheck
	gl.glClearNamedBufferSubDataEXT.restype = GLvoid
	gl.glClearNamedBufferSubDataEXT.argtypes = (GLuint, GLenum, GLsizeiptr, GLsizeiptr, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearNamedFramebufferfi = _getProcAddress('glClearNamedFramebufferfi')
	gl.glClearNamedFramebufferfi.gl = gl
	gl.glClearNamedFramebufferfi.__name__ = 'glClearNamedFramebufferfi'
	gl.glClearNamedFramebufferfi.errcheck = _errcheck
	gl.glClearNamedFramebufferfi.restype = GLvoid
	gl.glClearNamedFramebufferfi.argtypes = (GLuint, GLenum, GLfloat, GLint,)
	gl.glClearNamedFramebufferfv = _getProcAddress('glClearNamedFramebufferfv')
	gl.glClearNamedFramebufferfv.gl = gl
	gl.glClearNamedFramebufferfv.__name__ = 'glClearNamedFramebufferfv'
	gl.glClearNamedFramebufferfv.errcheck = _errcheck
	gl.glClearNamedFramebufferfv.restype = GLvoid
	gl.glClearNamedFramebufferfv.argtypes = (GLuint, GLenum, GLint, POINTER(GLfloat),)
	gl.glClearNamedFramebufferiv = _getProcAddress('glClearNamedFramebufferiv')
	gl.glClearNamedFramebufferiv.gl = gl
	gl.glClearNamedFramebufferiv.__name__ = 'glClearNamedFramebufferiv'
	gl.glClearNamedFramebufferiv.errcheck = _errcheck
	gl.glClearNamedFramebufferiv.restype = GLvoid
	gl.glClearNamedFramebufferiv.argtypes = (GLuint, GLenum, GLint, POINTER(GLint),)
	gl.glClearNamedFramebufferuiv = _getProcAddress('glClearNamedFramebufferuiv')
	gl.glClearNamedFramebufferuiv.gl = gl
	gl.glClearNamedFramebufferuiv.__name__ = 'glClearNamedFramebufferuiv'
	gl.glClearNamedFramebufferuiv.errcheck = _errcheck
	gl.glClearNamedFramebufferuiv.restype = GLvoid
	gl.glClearNamedFramebufferuiv.argtypes = (GLuint, GLenum, GLint, POINTER(GLuint),)
	gl.glClearStencil = _getProcAddress('glClearStencil')
	gl.glClearStencil.gl = gl
	gl.glClearStencil.__name__ = 'glClearStencil'
	gl.glClearStencil.errcheck = _errcheck
	gl.glClearStencil.restype = GLvoid
	gl.glClearStencil.argtypes = (GLint,)
	gl.glClearTexImage = _getProcAddress('glClearTexImage')
	gl.glClearTexImage.gl = gl
	gl.glClearTexImage.__name__ = 'glClearTexImage'
	gl.glClearTexImage.errcheck = _errcheck
	gl.glClearTexImage.restype = GLvoid
	gl.glClearTexImage.argtypes = (GLuint, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClearTexSubImage = _getProcAddress('glClearTexSubImage')
	gl.glClearTexSubImage.gl = gl
	gl.glClearTexSubImage.__name__ = 'glClearTexSubImage'
	gl.glClearTexSubImage.errcheck = _errcheck
	gl.glClearTexSubImage.restype = GLvoid
	gl.glClearTexSubImage.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glClientActiveTexture = _getProcAddress('glClientActiveTexture')
	gl.glClientActiveTexture.gl = gl
	gl.glClientActiveTexture.__name__ = 'glClientActiveTexture'
	gl.glClientActiveTexture.errcheck = _errcheck
	gl.glClientActiveTexture.restype = GLvoid
	gl.glClientActiveTexture.argtypes = (GLenum,)
	gl.glClientActiveTextureARB = _getProcAddress('glClientActiveTextureARB')
	gl.glClientActiveTextureARB.gl = gl
	gl.glClientActiveTextureARB.__name__ = 'glClientActiveTextureARB'
	gl.glClientActiveTextureARB.errcheck = _errcheck
	gl.glClientActiveTextureARB.restype = GLvoid
	gl.glClientActiveTextureARB.argtypes = (GLenum,)
	gl.glClientActiveVertexStreamATI = _getProcAddress('glClientActiveVertexStreamATI')
	gl.glClientActiveVertexStreamATI.gl = gl
	gl.glClientActiveVertexStreamATI.__name__ = 'glClientActiveVertexStreamATI'
	gl.glClientActiveVertexStreamATI.errcheck = _errcheck
	gl.glClientActiveVertexStreamATI.restype = GLvoid
	gl.glClientActiveVertexStreamATI.argtypes = (GLenum,)
	gl.glClientAttribDefaultEXT = _getProcAddress('glClientAttribDefaultEXT')
	gl.glClientAttribDefaultEXT.gl = gl
	gl.glClientAttribDefaultEXT.__name__ = 'glClientAttribDefaultEXT'
	gl.glClientAttribDefaultEXT.errcheck = _errcheck
	gl.glClientAttribDefaultEXT.restype = GLvoid
	gl.glClientAttribDefaultEXT.argtypes = (GLbitfield,)
	gl.glClientWaitSync = _getProcAddress('glClientWaitSync')
	gl.glClientWaitSync.gl = gl
	gl.glClientWaitSync.__name__ = 'glClientWaitSync'
	gl.glClientWaitSync.errcheck = _errcheck
	gl.glClientWaitSync.restype = GLenum
	gl.glClientWaitSync.argtypes = (GLsync, GLbitfield, GLuint64,)
	gl.glClientWaitSyncAPPLE = _getProcAddress('glClientWaitSyncAPPLE')
	gl.glClientWaitSyncAPPLE.gl = gl
	gl.glClientWaitSyncAPPLE.__name__ = 'glClientWaitSyncAPPLE'
	gl.glClientWaitSyncAPPLE.errcheck = _errcheck
	gl.glClientWaitSyncAPPLE.restype = GLenum
	gl.glClientWaitSyncAPPLE.argtypes = (GLsync, GLbitfield, GLuint64,)
	gl.glClipControl = _getProcAddress('glClipControl')
	gl.glClipControl.gl = gl
	gl.glClipControl.__name__ = 'glClipControl'
	gl.glClipControl.errcheck = _errcheck
	gl.glClipControl.restype = GLvoid
	gl.glClipControl.argtypes = (GLenum, GLenum,)
	gl.glClipPlane = _getProcAddress('glClipPlane')
	gl.glClipPlane.gl = gl
	gl.glClipPlane.__name__ = 'glClipPlane'
	gl.glClipPlane.errcheck = _errcheck
	gl.glClipPlane.restype = GLvoid
	gl.glClipPlane.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glClipPlanef = _getProcAddress('glClipPlanef')
	gl.glClipPlanef.gl = gl
	gl.glClipPlanef.__name__ = 'glClipPlanef'
	gl.glClipPlanef.errcheck = _errcheck
	gl.glClipPlanef.restype = GLvoid
	gl.glClipPlanef.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glClipPlanefIMG = _getProcAddress('glClipPlanefIMG')
	gl.glClipPlanefIMG.gl = gl
	gl.glClipPlanefIMG.__name__ = 'glClipPlanefIMG'
	gl.glClipPlanefIMG.errcheck = _errcheck
	gl.glClipPlanefIMG.restype = GLvoid
	gl.glClipPlanefIMG.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glClipPlanefOES = _getProcAddress('glClipPlanefOES')
	gl.glClipPlanefOES.gl = gl
	gl.glClipPlanefOES.__name__ = 'glClipPlanefOES'
	gl.glClipPlanefOES.errcheck = _errcheck
	gl.glClipPlanefOES.restype = GLvoid
	gl.glClipPlanefOES.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glClipPlanex = _getProcAddress('glClipPlanex')
	gl.glClipPlanex.gl = gl
	gl.glClipPlanex.__name__ = 'glClipPlanex'
	gl.glClipPlanex.errcheck = _errcheck
	gl.glClipPlanex.restype = GLvoid
	gl.glClipPlanex.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glClipPlanexIMG = _getProcAddress('glClipPlanexIMG')
	gl.glClipPlanexIMG.gl = gl
	gl.glClipPlanexIMG.__name__ = 'glClipPlanexIMG'
	gl.glClipPlanexIMG.errcheck = _errcheck
	gl.glClipPlanexIMG.restype = GLvoid
	gl.glClipPlanexIMG.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glClipPlanexOES = _getProcAddress('glClipPlanexOES')
	gl.glClipPlanexOES.gl = gl
	gl.glClipPlanexOES.__name__ = 'glClipPlanexOES'
	gl.glClipPlanexOES.errcheck = _errcheck
	gl.glClipPlanexOES.restype = GLvoid
	gl.glClipPlanexOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glColor3b = _getProcAddress('glColor3b')
	gl.glColor3b.gl = gl
	gl.glColor3b.__name__ = 'glColor3b'
	gl.glColor3b.errcheck = _errcheck
	gl.glColor3b.restype = GLvoid
	gl.glColor3b.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glColor3bv = _getProcAddress('glColor3bv')
	gl.glColor3bv.gl = gl
	gl.glColor3bv.__name__ = 'glColor3bv'
	gl.glColor3bv.errcheck = _errcheck
	gl.glColor3bv.restype = GLvoid
	gl.glColor3bv.argtypes = (POINTER(GLbyte),)
	gl.glColor3d = _getProcAddress('glColor3d')
	gl.glColor3d.gl = gl
	gl.glColor3d.__name__ = 'glColor3d'
	gl.glColor3d.errcheck = _errcheck
	gl.glColor3d.restype = GLvoid
	gl.glColor3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glColor3dv = _getProcAddress('glColor3dv')
	gl.glColor3dv.gl = gl
	gl.glColor3dv.__name__ = 'glColor3dv'
	gl.glColor3dv.errcheck = _errcheck
	gl.glColor3dv.restype = GLvoid
	gl.glColor3dv.argtypes = (POINTER(GLdouble),)
	gl.glColor3f = _getProcAddress('glColor3f')
	gl.glColor3f.gl = gl
	gl.glColor3f.__name__ = 'glColor3f'
	gl.glColor3f.errcheck = _errcheck
	gl.glColor3f.restype = GLvoid
	gl.glColor3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glColor3fVertex3fSUN = _getProcAddress('glColor3fVertex3fSUN')
	gl.glColor3fVertex3fSUN.gl = gl
	gl.glColor3fVertex3fSUN.__name__ = 'glColor3fVertex3fSUN'
	gl.glColor3fVertex3fSUN.errcheck = _errcheck
	gl.glColor3fVertex3fSUN.restype = GLvoid
	gl.glColor3fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glColor3fVertex3fvSUN = _getProcAddress('glColor3fVertex3fvSUN')
	gl.glColor3fVertex3fvSUN.gl = gl
	gl.glColor3fVertex3fvSUN.__name__ = 'glColor3fVertex3fvSUN'
	gl.glColor3fVertex3fvSUN.errcheck = _errcheck
	gl.glColor3fVertex3fvSUN.restype = GLvoid
	gl.glColor3fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat),)
	gl.glColor3fv = _getProcAddress('glColor3fv')
	gl.glColor3fv.gl = gl
	gl.glColor3fv.__name__ = 'glColor3fv'
	gl.glColor3fv.errcheck = _errcheck
	gl.glColor3fv.restype = GLvoid
	gl.glColor3fv.argtypes = (POINTER(GLfloat),)
	gl.glColor3hNV = _getProcAddress('glColor3hNV')
	gl.glColor3hNV.gl = gl
	gl.glColor3hNV.__name__ = 'glColor3hNV'
	gl.glColor3hNV.errcheck = _errcheck
	gl.glColor3hNV.restype = GLvoid
	gl.glColor3hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glColor3hvNV = _getProcAddress('glColor3hvNV')
	gl.glColor3hvNV.gl = gl
	gl.glColor3hvNV.__name__ = 'glColor3hvNV'
	gl.glColor3hvNV.errcheck = _errcheck
	gl.glColor3hvNV.restype = GLvoid
	gl.glColor3hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glColor3i = _getProcAddress('glColor3i')
	gl.glColor3i.gl = gl
	gl.glColor3i.__name__ = 'glColor3i'
	gl.glColor3i.errcheck = _errcheck
	gl.glColor3i.restype = GLvoid
	gl.glColor3i.argtypes = (GLint, GLint, GLint,)
	gl.glColor3iv = _getProcAddress('glColor3iv')
	gl.glColor3iv.gl = gl
	gl.glColor3iv.__name__ = 'glColor3iv'
	gl.glColor3iv.errcheck = _errcheck
	gl.glColor3iv.restype = GLvoid
	gl.glColor3iv.argtypes = (POINTER(GLint),)
	gl.glColor3s = _getProcAddress('glColor3s')
	gl.glColor3s.gl = gl
	gl.glColor3s.__name__ = 'glColor3s'
	gl.glColor3s.errcheck = _errcheck
	gl.glColor3s.restype = GLvoid
	gl.glColor3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glColor3sv = _getProcAddress('glColor3sv')
	gl.glColor3sv.gl = gl
	gl.glColor3sv.__name__ = 'glColor3sv'
	gl.glColor3sv.errcheck = _errcheck
	gl.glColor3sv.restype = GLvoid
	gl.glColor3sv.argtypes = (POINTER(GLshort),)
	gl.glColor3ub = _getProcAddress('glColor3ub')
	gl.glColor3ub.gl = gl
	gl.glColor3ub.__name__ = 'glColor3ub'
	gl.glColor3ub.errcheck = _errcheck
	gl.glColor3ub.restype = GLvoid
	gl.glColor3ub.argtypes = (GLubyte, GLubyte, GLubyte,)
	gl.glColor3ubv = _getProcAddress('glColor3ubv')
	gl.glColor3ubv.gl = gl
	gl.glColor3ubv.__name__ = 'glColor3ubv'
	gl.glColor3ubv.errcheck = _errcheck
	gl.glColor3ubv.restype = GLvoid
	gl.glColor3ubv.argtypes = (POINTER(GLubyte),)
	gl.glColor3ui = _getProcAddress('glColor3ui')
	gl.glColor3ui.gl = gl
	gl.glColor3ui.__name__ = 'glColor3ui'
	gl.glColor3ui.errcheck = _errcheck
	gl.glColor3ui.restype = GLvoid
	gl.glColor3ui.argtypes = (GLuint, GLuint, GLuint,)
	gl.glColor3uiv = _getProcAddress('glColor3uiv')
	gl.glColor3uiv.gl = gl
	gl.glColor3uiv.__name__ = 'glColor3uiv'
	gl.glColor3uiv.errcheck = _errcheck
	gl.glColor3uiv.restype = GLvoid
	gl.glColor3uiv.argtypes = (POINTER(GLuint),)
	gl.glColor3us = _getProcAddress('glColor3us')
	gl.glColor3us.gl = gl
	gl.glColor3us.__name__ = 'glColor3us'
	gl.glColor3us.errcheck = _errcheck
	gl.glColor3us.restype = GLvoid
	gl.glColor3us.argtypes = (GLushort, GLushort, GLushort,)
	gl.glColor3usv = _getProcAddress('glColor3usv')
	gl.glColor3usv.gl = gl
	gl.glColor3usv.__name__ = 'glColor3usv'
	gl.glColor3usv.errcheck = _errcheck
	gl.glColor3usv.restype = GLvoid
	gl.glColor3usv.argtypes = (POINTER(GLushort),)
	gl.glColor3xOES = _getProcAddress('glColor3xOES')
	gl.glColor3xOES.gl = gl
	gl.glColor3xOES.__name__ = 'glColor3xOES'
	gl.glColor3xOES.errcheck = _errcheck
	gl.glColor3xOES.restype = GLvoid
	gl.glColor3xOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glColor3xvOES = _getProcAddress('glColor3xvOES')
	gl.glColor3xvOES.gl = gl
	gl.glColor3xvOES.__name__ = 'glColor3xvOES'
	gl.glColor3xvOES.errcheck = _errcheck
	gl.glColor3xvOES.restype = GLvoid
	gl.glColor3xvOES.argtypes = (POINTER(GLfixed),)
	gl.glColor4b = _getProcAddress('glColor4b')
	gl.glColor4b.gl = gl
	gl.glColor4b.__name__ = 'glColor4b'
	gl.glColor4b.errcheck = _errcheck
	gl.glColor4b.restype = GLvoid
	gl.glColor4b.argtypes = (GLbyte, GLbyte, GLbyte, GLbyte,)
	gl.glColor4bv = _getProcAddress('glColor4bv')
	gl.glColor4bv.gl = gl
	gl.glColor4bv.__name__ = 'glColor4bv'
	gl.glColor4bv.errcheck = _errcheck
	gl.glColor4bv.restype = GLvoid
	gl.glColor4bv.argtypes = (POINTER(GLbyte),)
	gl.glColor4d = _getProcAddress('glColor4d')
	gl.glColor4d.gl = gl
	gl.glColor4d.__name__ = 'glColor4d'
	gl.glColor4d.errcheck = _errcheck
	gl.glColor4d.restype = GLvoid
	gl.glColor4d.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glColor4dv = _getProcAddress('glColor4dv')
	gl.glColor4dv.gl = gl
	gl.glColor4dv.__name__ = 'glColor4dv'
	gl.glColor4dv.errcheck = _errcheck
	gl.glColor4dv.restype = GLvoid
	gl.glColor4dv.argtypes = (POINTER(GLdouble),)
	gl.glColor4f = _getProcAddress('glColor4f')
	gl.glColor4f.gl = gl
	gl.glColor4f.__name__ = 'glColor4f'
	gl.glColor4f.errcheck = _errcheck
	gl.glColor4f.restype = GLvoid
	gl.glColor4f.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glColor4fNormal3fVertex3fSUN = _getProcAddress('glColor4fNormal3fVertex3fSUN')
	gl.glColor4fNormal3fVertex3fSUN.gl = gl
	gl.glColor4fNormal3fVertex3fSUN.__name__ = 'glColor4fNormal3fVertex3fSUN'
	gl.glColor4fNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glColor4fNormal3fVertex3fSUN.restype = GLvoid
	gl.glColor4fNormal3fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glColor4fNormal3fVertex3fvSUN = _getProcAddress('glColor4fNormal3fVertex3fvSUN')
	gl.glColor4fNormal3fVertex3fvSUN.gl = gl
	gl.glColor4fNormal3fVertex3fvSUN.__name__ = 'glColor4fNormal3fVertex3fvSUN'
	gl.glColor4fNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glColor4fNormal3fVertex3fvSUN.restype = GLvoid
	gl.glColor4fNormal3fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glColor4fv = _getProcAddress('glColor4fv')
	gl.glColor4fv.gl = gl
	gl.glColor4fv.__name__ = 'glColor4fv'
	gl.glColor4fv.errcheck = _errcheck
	gl.glColor4fv.restype = GLvoid
	gl.glColor4fv.argtypes = (POINTER(GLfloat),)
	gl.glColor4hNV = _getProcAddress('glColor4hNV')
	gl.glColor4hNV.gl = gl
	gl.glColor4hNV.__name__ = 'glColor4hNV'
	gl.glColor4hNV.errcheck = _errcheck
	gl.glColor4hNV.restype = GLvoid
	gl.glColor4hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glColor4hvNV = _getProcAddress('glColor4hvNV')
	gl.glColor4hvNV.gl = gl
	gl.glColor4hvNV.__name__ = 'glColor4hvNV'
	gl.glColor4hvNV.errcheck = _errcheck
	gl.glColor4hvNV.restype = GLvoid
	gl.glColor4hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glColor4i = _getProcAddress('glColor4i')
	gl.glColor4i.gl = gl
	gl.glColor4i.__name__ = 'glColor4i'
	gl.glColor4i.errcheck = _errcheck
	gl.glColor4i.restype = GLvoid
	gl.glColor4i.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glColor4iv = _getProcAddress('glColor4iv')
	gl.glColor4iv.gl = gl
	gl.glColor4iv.__name__ = 'glColor4iv'
	gl.glColor4iv.errcheck = _errcheck
	gl.glColor4iv.restype = GLvoid
	gl.glColor4iv.argtypes = (POINTER(GLint),)
	gl.glColor4s = _getProcAddress('glColor4s')
	gl.glColor4s.gl = gl
	gl.glColor4s.__name__ = 'glColor4s'
	gl.glColor4s.errcheck = _errcheck
	gl.glColor4s.restype = GLvoid
	gl.glColor4s.argtypes = (GLshort, GLshort, GLshort, GLshort,)
	gl.glColor4sv = _getProcAddress('glColor4sv')
	gl.glColor4sv.gl = gl
	gl.glColor4sv.__name__ = 'glColor4sv'
	gl.glColor4sv.errcheck = _errcheck
	gl.glColor4sv.restype = GLvoid
	gl.glColor4sv.argtypes = (POINTER(GLshort),)
	gl.glColor4ub = _getProcAddress('glColor4ub')
	gl.glColor4ub.gl = gl
	gl.glColor4ub.__name__ = 'glColor4ub'
	gl.glColor4ub.errcheck = _errcheck
	gl.glColor4ub.restype = GLvoid
	gl.glColor4ub.argtypes = (GLubyte, GLubyte, GLubyte, GLubyte,)
	gl.glColor4ubVertex2fSUN = _getProcAddress('glColor4ubVertex2fSUN')
	gl.glColor4ubVertex2fSUN.gl = gl
	gl.glColor4ubVertex2fSUN.__name__ = 'glColor4ubVertex2fSUN'
	gl.glColor4ubVertex2fSUN.errcheck = _errcheck
	gl.glColor4ubVertex2fSUN.restype = GLvoid
	gl.glColor4ubVertex2fSUN.argtypes = (GLubyte, GLubyte, GLubyte, GLubyte, GLfloat, GLfloat,)
	gl.glColor4ubVertex2fvSUN = _getProcAddress('glColor4ubVertex2fvSUN')
	gl.glColor4ubVertex2fvSUN.gl = gl
	gl.glColor4ubVertex2fvSUN.__name__ = 'glColor4ubVertex2fvSUN'
	gl.glColor4ubVertex2fvSUN.errcheck = _errcheck
	gl.glColor4ubVertex2fvSUN.restype = GLvoid
	gl.glColor4ubVertex2fvSUN.argtypes = (POINTER(GLubyte), POINTER(GLfloat),)
	gl.glColor4ubVertex3fSUN = _getProcAddress('glColor4ubVertex3fSUN')
	gl.glColor4ubVertex3fSUN.gl = gl
	gl.glColor4ubVertex3fSUN.__name__ = 'glColor4ubVertex3fSUN'
	gl.glColor4ubVertex3fSUN.errcheck = _errcheck
	gl.glColor4ubVertex3fSUN.restype = GLvoid
	gl.glColor4ubVertex3fSUN.argtypes = (GLubyte, GLubyte, GLubyte, GLubyte, GLfloat, GLfloat, GLfloat,)
	gl.glColor4ubVertex3fvSUN = _getProcAddress('glColor4ubVertex3fvSUN')
	gl.glColor4ubVertex3fvSUN.gl = gl
	gl.glColor4ubVertex3fvSUN.__name__ = 'glColor4ubVertex3fvSUN'
	gl.glColor4ubVertex3fvSUN.errcheck = _errcheck
	gl.glColor4ubVertex3fvSUN.restype = GLvoid
	gl.glColor4ubVertex3fvSUN.argtypes = (POINTER(GLubyte), POINTER(GLfloat),)
	gl.glColor4ubv = _getProcAddress('glColor4ubv')
	gl.glColor4ubv.gl = gl
	gl.glColor4ubv.__name__ = 'glColor4ubv'
	gl.glColor4ubv.errcheck = _errcheck
	gl.glColor4ubv.restype = GLvoid
	gl.glColor4ubv.argtypes = (POINTER(GLubyte),)
	gl.glColor4ui = _getProcAddress('glColor4ui')
	gl.glColor4ui.gl = gl
	gl.glColor4ui.__name__ = 'glColor4ui'
	gl.glColor4ui.errcheck = _errcheck
	gl.glColor4ui.restype = GLvoid
	gl.glColor4ui.argtypes = (GLuint, GLuint, GLuint, GLuint,)
	gl.glColor4uiv = _getProcAddress('glColor4uiv')
	gl.glColor4uiv.gl = gl
	gl.glColor4uiv.__name__ = 'glColor4uiv'
	gl.glColor4uiv.errcheck = _errcheck
	gl.glColor4uiv.restype = GLvoid
	gl.glColor4uiv.argtypes = (POINTER(GLuint),)
	gl.glColor4us = _getProcAddress('glColor4us')
	gl.glColor4us.gl = gl
	gl.glColor4us.__name__ = 'glColor4us'
	gl.glColor4us.errcheck = _errcheck
	gl.glColor4us.restype = GLvoid
	gl.glColor4us.argtypes = (GLushort, GLushort, GLushort, GLushort,)
	gl.glColor4usv = _getProcAddress('glColor4usv')
	gl.glColor4usv.gl = gl
	gl.glColor4usv.__name__ = 'glColor4usv'
	gl.glColor4usv.errcheck = _errcheck
	gl.glColor4usv.restype = GLvoid
	gl.glColor4usv.argtypes = (POINTER(GLushort),)
	gl.glColor4x = _getProcAddress('glColor4x')
	gl.glColor4x.gl = gl
	gl.glColor4x.__name__ = 'glColor4x'
	gl.glColor4x.errcheck = _errcheck
	gl.glColor4x.restype = GLvoid
	gl.glColor4x.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glColor4xOES = _getProcAddress('glColor4xOES')
	gl.glColor4xOES.gl = gl
	gl.glColor4xOES.__name__ = 'glColor4xOES'
	gl.glColor4xOES.errcheck = _errcheck
	gl.glColor4xOES.restype = GLvoid
	gl.glColor4xOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glColor4xvOES = _getProcAddress('glColor4xvOES')
	gl.glColor4xvOES.gl = gl
	gl.glColor4xvOES.__name__ = 'glColor4xvOES'
	gl.glColor4xvOES.errcheck = _errcheck
	gl.glColor4xvOES.restype = GLvoid
	gl.glColor4xvOES.argtypes = (POINTER(GLfixed),)
	gl.glColorFormatNV = _getProcAddress('glColorFormatNV')
	gl.glColorFormatNV.gl = gl
	gl.glColorFormatNV.__name__ = 'glColorFormatNV'
	gl.glColorFormatNV.errcheck = _errcheck
	gl.glColorFormatNV.restype = GLvoid
	gl.glColorFormatNV.argtypes = (GLint, GLenum, GLsizei,)
	gl.glColorFragmentOp1ATI = _getProcAddress('glColorFragmentOp1ATI')
	gl.glColorFragmentOp1ATI.gl = gl
	gl.glColorFragmentOp1ATI.__name__ = 'glColorFragmentOp1ATI'
	gl.glColorFragmentOp1ATI.errcheck = _errcheck
	gl.glColorFragmentOp1ATI.restype = GLvoid
	gl.glColorFragmentOp1ATI.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glColorFragmentOp2ATI = _getProcAddress('glColorFragmentOp2ATI')
	gl.glColorFragmentOp2ATI.gl = gl
	gl.glColorFragmentOp2ATI.__name__ = 'glColorFragmentOp2ATI'
	gl.glColorFragmentOp2ATI.errcheck = _errcheck
	gl.glColorFragmentOp2ATI.restype = GLvoid
	gl.glColorFragmentOp2ATI.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glColorFragmentOp3ATI = _getProcAddress('glColorFragmentOp3ATI')
	gl.glColorFragmentOp3ATI.gl = gl
	gl.glColorFragmentOp3ATI.__name__ = 'glColorFragmentOp3ATI'
	gl.glColorFragmentOp3ATI.errcheck = _errcheck
	gl.glColorFragmentOp3ATI.restype = GLvoid
	gl.glColorFragmentOp3ATI.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glColorMask = _getProcAddress('glColorMask')
	gl.glColorMask.gl = gl
	gl.glColorMask.__name__ = 'glColorMask'
	gl.glColorMask.errcheck = _errcheck
	gl.glColorMask.restype = GLvoid
	gl.glColorMask.argtypes = (GLboolean, GLboolean, GLboolean, GLboolean,)
	gl.glColorMaskIndexedEXT = _getProcAddress('glColorMaskIndexedEXT')
	gl.glColorMaskIndexedEXT.gl = gl
	gl.glColorMaskIndexedEXT.__name__ = 'glColorMaskIndexedEXT'
	gl.glColorMaskIndexedEXT.errcheck = _errcheck
	gl.glColorMaskIndexedEXT.restype = GLvoid
	gl.glColorMaskIndexedEXT.argtypes = (GLuint, GLboolean, GLboolean, GLboolean, GLboolean,)
	gl.glColorMaski = _getProcAddress('glColorMaski')
	gl.glColorMaski.gl = gl
	gl.glColorMaski.__name__ = 'glColorMaski'
	gl.glColorMaski.errcheck = _errcheck
	gl.glColorMaski.restype = GLvoid
	gl.glColorMaski.argtypes = (GLuint, GLboolean, GLboolean, GLboolean, GLboolean,)
	gl.glColorMaskiEXT = _getProcAddress('glColorMaskiEXT')
	gl.glColorMaskiEXT.gl = gl
	gl.glColorMaskiEXT.__name__ = 'glColorMaskiEXT'
	gl.glColorMaskiEXT.errcheck = _errcheck
	gl.glColorMaskiEXT.restype = GLvoid
	gl.glColorMaskiEXT.argtypes = (GLuint, GLboolean, GLboolean, GLboolean, GLboolean,)
	gl.glColorMaterial = _getProcAddress('glColorMaterial')
	gl.glColorMaterial.gl = gl
	gl.glColorMaterial.__name__ = 'glColorMaterial'
	gl.glColorMaterial.errcheck = _errcheck
	gl.glColorMaterial.restype = GLvoid
	gl.glColorMaterial.argtypes = (GLenum, GLenum,)
	gl.glColorP3ui = _getProcAddress('glColorP3ui')
	gl.glColorP3ui.gl = gl
	gl.glColorP3ui.__name__ = 'glColorP3ui'
	gl.glColorP3ui.errcheck = _errcheck
	gl.glColorP3ui.restype = GLvoid
	gl.glColorP3ui.argtypes = (GLenum, GLuint,)
	gl.glColorP3uiv = _getProcAddress('glColorP3uiv')
	gl.glColorP3uiv.gl = gl
	gl.glColorP3uiv.__name__ = 'glColorP3uiv'
	gl.glColorP3uiv.errcheck = _errcheck
	gl.glColorP3uiv.restype = GLvoid
	gl.glColorP3uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glColorP4ui = _getProcAddress('glColorP4ui')
	gl.glColorP4ui.gl = gl
	gl.glColorP4ui.__name__ = 'glColorP4ui'
	gl.glColorP4ui.errcheck = _errcheck
	gl.glColorP4ui.restype = GLvoid
	gl.glColorP4ui.argtypes = (GLenum, GLuint,)
	gl.glColorP4uiv = _getProcAddress('glColorP4uiv')
	gl.glColorP4uiv.gl = gl
	gl.glColorP4uiv.__name__ = 'glColorP4uiv'
	gl.glColorP4uiv.errcheck = _errcheck
	gl.glColorP4uiv.restype = GLvoid
	gl.glColorP4uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glColorPointer = _getProcAddress('glColorPointer')
	gl.glColorPointer.gl = gl
	gl.glColorPointer.__name__ = 'glColorPointer'
	gl.glColorPointer.errcheck = _errcheck
	gl.glColorPointer.restype = GLvoid
	gl.glColorPointer.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glColorPointerEXT = _getProcAddress('glColorPointerEXT')
	gl.glColorPointerEXT.gl = gl
	gl.glColorPointerEXT.__name__ = 'glColorPointerEXT'
	gl.glColorPointerEXT.errcheck = _errcheck
	gl.glColorPointerEXT.restype = GLvoid
	gl.glColorPointerEXT.argtypes = (GLint, GLenum, GLsizei, GLsizei, POINTER(GLvoid),)
	gl.glColorPointerListIBM = _getProcAddress('glColorPointerListIBM')
	gl.glColorPointerListIBM.gl = gl
	gl.glColorPointerListIBM.__name__ = 'glColorPointerListIBM'
	gl.glColorPointerListIBM.errcheck = _errcheck
	gl.glColorPointerListIBM.restype = GLvoid
	gl.glColorPointerListIBM.argtypes = (GLint, GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glColorPointervINTEL = _getProcAddress('glColorPointervINTEL')
	gl.glColorPointervINTEL.gl = gl
	gl.glColorPointervINTEL.__name__ = 'glColorPointervINTEL'
	gl.glColorPointervINTEL.errcheck = _errcheck
	gl.glColorPointervINTEL.restype = GLvoid
	gl.glColorPointervINTEL.argtypes = (GLint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glColorSubTable = _getProcAddress('glColorSubTable')
	gl.glColorSubTable.gl = gl
	gl.glColorSubTable.__name__ = 'glColorSubTable'
	gl.glColorSubTable.errcheck = _errcheck
	gl.glColorSubTable.restype = GLvoid
	gl.glColorSubTable.argtypes = (GLenum, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glColorSubTableEXT = _getProcAddress('glColorSubTableEXT')
	gl.glColorSubTableEXT.gl = gl
	gl.glColorSubTableEXT.__name__ = 'glColorSubTableEXT'
	gl.glColorSubTableEXT.errcheck = _errcheck
	gl.glColorSubTableEXT.restype = GLvoid
	gl.glColorSubTableEXT.argtypes = (GLenum, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glColorTable = _getProcAddress('glColorTable')
	gl.glColorTable.gl = gl
	gl.glColorTable.__name__ = 'glColorTable'
	gl.glColorTable.errcheck = _errcheck
	gl.glColorTable.restype = GLvoid
	gl.glColorTable.argtypes = (GLenum, GLenum, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glColorTableEXT = _getProcAddress('glColorTableEXT')
	gl.glColorTableEXT.gl = gl
	gl.glColorTableEXT.__name__ = 'glColorTableEXT'
	gl.glColorTableEXT.errcheck = _errcheck
	gl.glColorTableEXT.restype = GLvoid
	gl.glColorTableEXT.argtypes = (GLenum, GLenum, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glColorTableParameterfv = _getProcAddress('glColorTableParameterfv')
	gl.glColorTableParameterfv.gl = gl
	gl.glColorTableParameterfv.__name__ = 'glColorTableParameterfv'
	gl.glColorTableParameterfv.errcheck = _errcheck
	gl.glColorTableParameterfv.restype = GLvoid
	gl.glColorTableParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glColorTableParameterfvSGI = _getProcAddress('glColorTableParameterfvSGI')
	gl.glColorTableParameterfvSGI.gl = gl
	gl.glColorTableParameterfvSGI.__name__ = 'glColorTableParameterfvSGI'
	gl.glColorTableParameterfvSGI.errcheck = _errcheck
	gl.glColorTableParameterfvSGI.restype = GLvoid
	gl.glColorTableParameterfvSGI.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glColorTableParameteriv = _getProcAddress('glColorTableParameteriv')
	gl.glColorTableParameteriv.gl = gl
	gl.glColorTableParameteriv.__name__ = 'glColorTableParameteriv'
	gl.glColorTableParameteriv.errcheck = _errcheck
	gl.glColorTableParameteriv.restype = GLvoid
	gl.glColorTableParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glColorTableParameterivSGI = _getProcAddress('glColorTableParameterivSGI')
	gl.glColorTableParameterivSGI.gl = gl
	gl.glColorTableParameterivSGI.__name__ = 'glColorTableParameterivSGI'
	gl.glColorTableParameterivSGI.errcheck = _errcheck
	gl.glColorTableParameterivSGI.restype = GLvoid
	gl.glColorTableParameterivSGI.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glColorTableSGI = _getProcAddress('glColorTableSGI')
	gl.glColorTableSGI.gl = gl
	gl.glColorTableSGI.__name__ = 'glColorTableSGI'
	gl.glColorTableSGI.errcheck = _errcheck
	gl.glColorTableSGI.restype = GLvoid
	gl.glColorTableSGI.argtypes = (GLenum, GLenum, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glCombinerInputNV = _getProcAddress('glCombinerInputNV')
	gl.glCombinerInputNV.gl = gl
	gl.glCombinerInputNV.__name__ = 'glCombinerInputNV'
	gl.glCombinerInputNV.errcheck = _errcheck
	gl.glCombinerInputNV.restype = GLvoid
	gl.glCombinerInputNV.argtypes = (GLenum, GLenum, GLenum, GLenum, GLenum, GLenum,)
	gl.glCombinerOutputNV = _getProcAddress('glCombinerOutputNV')
	gl.glCombinerOutputNV.gl = gl
	gl.glCombinerOutputNV.__name__ = 'glCombinerOutputNV'
	gl.glCombinerOutputNV.errcheck = _errcheck
	gl.glCombinerOutputNV.restype = GLvoid
	gl.glCombinerOutputNV.argtypes = (GLenum, GLenum, GLenum, GLenum, GLenum, GLenum, GLenum, GLboolean, GLboolean, GLboolean,)
	gl.glCombinerParameterfNV = _getProcAddress('glCombinerParameterfNV')
	gl.glCombinerParameterfNV.gl = gl
	gl.glCombinerParameterfNV.__name__ = 'glCombinerParameterfNV'
	gl.glCombinerParameterfNV.errcheck = _errcheck
	gl.glCombinerParameterfNV.restype = GLvoid
	gl.glCombinerParameterfNV.argtypes = (GLenum, GLfloat,)
	gl.glCombinerParameterfvNV = _getProcAddress('glCombinerParameterfvNV')
	gl.glCombinerParameterfvNV.gl = gl
	gl.glCombinerParameterfvNV.__name__ = 'glCombinerParameterfvNV'
	gl.glCombinerParameterfvNV.errcheck = _errcheck
	gl.glCombinerParameterfvNV.restype = GLvoid
	gl.glCombinerParameterfvNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glCombinerParameteriNV = _getProcAddress('glCombinerParameteriNV')
	gl.glCombinerParameteriNV.gl = gl
	gl.glCombinerParameteriNV.__name__ = 'glCombinerParameteriNV'
	gl.glCombinerParameteriNV.errcheck = _errcheck
	gl.glCombinerParameteriNV.restype = GLvoid
	gl.glCombinerParameteriNV.argtypes = (GLenum, GLint,)
	gl.glCombinerParameterivNV = _getProcAddress('glCombinerParameterivNV')
	gl.glCombinerParameterivNV.gl = gl
	gl.glCombinerParameterivNV.__name__ = 'glCombinerParameterivNV'
	gl.glCombinerParameterivNV.errcheck = _errcheck
	gl.glCombinerParameterivNV.restype = GLvoid
	gl.glCombinerParameterivNV.argtypes = (GLenum, POINTER(GLint),)
	gl.glCombinerStageParameterfvNV = _getProcAddress('glCombinerStageParameterfvNV')
	gl.glCombinerStageParameterfvNV.gl = gl
	gl.glCombinerStageParameterfvNV.__name__ = 'glCombinerStageParameterfvNV'
	gl.glCombinerStageParameterfvNV.errcheck = _errcheck
	gl.glCombinerStageParameterfvNV.restype = GLvoid
	gl.glCombinerStageParameterfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glCompileShader = _getProcAddress('glCompileShader')
	gl.glCompileShader.gl = gl
	gl.glCompileShader.__name__ = 'glCompileShader'
	gl.glCompileShader.errcheck = _errcheck
	gl.glCompileShader.restype = GLvoid
	gl.glCompileShader.argtypes = (GLuint,)
	gl.glCompileShaderARB = _getProcAddress('glCompileShaderARB')
	gl.glCompileShaderARB.gl = gl
	gl.glCompileShaderARB.__name__ = 'glCompileShaderARB'
	gl.glCompileShaderARB.errcheck = _errcheck
	gl.glCompileShaderARB.restype = GLvoid
	gl.glCompileShaderARB.argtypes = (GLhandleARB,)
	gl.glCompileShaderIncludeARB = _getProcAddress('glCompileShaderIncludeARB')
	gl.glCompileShaderIncludeARB.gl = gl
	gl.glCompileShaderIncludeARB.__name__ = 'glCompileShaderIncludeARB'
	gl.glCompileShaderIncludeARB.errcheck = _errcheck
	gl.glCompileShaderIncludeARB.restype = GLvoid
	gl.glCompileShaderIncludeARB.argtypes = (GLuint, GLsizei, POINTER(POINTER(GLchar)), POINTER(GLint),)
	gl.glCompressedMultiTexImage1DEXT = _getProcAddress('glCompressedMultiTexImage1DEXT')
	gl.glCompressedMultiTexImage1DEXT.gl = gl
	gl.glCompressedMultiTexImage1DEXT.__name__ = 'glCompressedMultiTexImage1DEXT'
	gl.glCompressedMultiTexImage1DEXT.errcheck = _errcheck
	gl.glCompressedMultiTexImage1DEXT.restype = GLvoid
	gl.glCompressedMultiTexImage1DEXT.argtypes = (GLenum, GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedMultiTexImage2DEXT = _getProcAddress('glCompressedMultiTexImage2DEXT')
	gl.glCompressedMultiTexImage2DEXT.gl = gl
	gl.glCompressedMultiTexImage2DEXT.__name__ = 'glCompressedMultiTexImage2DEXT'
	gl.glCompressedMultiTexImage2DEXT.errcheck = _errcheck
	gl.glCompressedMultiTexImage2DEXT.restype = GLvoid
	gl.glCompressedMultiTexImage2DEXT.argtypes = (GLenum, GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedMultiTexImage3DEXT = _getProcAddress('glCompressedMultiTexImage3DEXT')
	gl.glCompressedMultiTexImage3DEXT.gl = gl
	gl.glCompressedMultiTexImage3DEXT.__name__ = 'glCompressedMultiTexImage3DEXT'
	gl.glCompressedMultiTexImage3DEXT.errcheck = _errcheck
	gl.glCompressedMultiTexImage3DEXT.restype = GLvoid
	gl.glCompressedMultiTexImage3DEXT.argtypes = (GLenum, GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedMultiTexSubImage1DEXT = _getProcAddress('glCompressedMultiTexSubImage1DEXT')
	gl.glCompressedMultiTexSubImage1DEXT.gl = gl
	gl.glCompressedMultiTexSubImage1DEXT.__name__ = 'glCompressedMultiTexSubImage1DEXT'
	gl.glCompressedMultiTexSubImage1DEXT.errcheck = _errcheck
	gl.glCompressedMultiTexSubImage1DEXT.restype = GLvoid
	gl.glCompressedMultiTexSubImage1DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedMultiTexSubImage2DEXT = _getProcAddress('glCompressedMultiTexSubImage2DEXT')
	gl.glCompressedMultiTexSubImage2DEXT.gl = gl
	gl.glCompressedMultiTexSubImage2DEXT.__name__ = 'glCompressedMultiTexSubImage2DEXT'
	gl.glCompressedMultiTexSubImage2DEXT.errcheck = _errcheck
	gl.glCompressedMultiTexSubImage2DEXT.restype = GLvoid
	gl.glCompressedMultiTexSubImage2DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedMultiTexSubImage3DEXT = _getProcAddress('glCompressedMultiTexSubImage3DEXT')
	gl.glCompressedMultiTexSubImage3DEXT.gl = gl
	gl.glCompressedMultiTexSubImage3DEXT.__name__ = 'glCompressedMultiTexSubImage3DEXT'
	gl.glCompressedMultiTexSubImage3DEXT.errcheck = _errcheck
	gl.glCompressedMultiTexSubImage3DEXT.restype = GLvoid
	gl.glCompressedMultiTexSubImage3DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage1D = _getProcAddress('glCompressedTexImage1D')
	gl.glCompressedTexImage1D.gl = gl
	gl.glCompressedTexImage1D.__name__ = 'glCompressedTexImage1D'
	gl.glCompressedTexImage1D.errcheck = _errcheck
	gl.glCompressedTexImage1D.restype = GLvoid
	gl.glCompressedTexImage1D.argtypes = (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage1DARB = _getProcAddress('glCompressedTexImage1DARB')
	gl.glCompressedTexImage1DARB.gl = gl
	gl.glCompressedTexImage1DARB.__name__ = 'glCompressedTexImage1DARB'
	gl.glCompressedTexImage1DARB.errcheck = _errcheck
	gl.glCompressedTexImage1DARB.restype = GLvoid
	gl.glCompressedTexImage1DARB.argtypes = (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage2D = _getProcAddress('glCompressedTexImage2D')
	gl.glCompressedTexImage2D.gl = gl
	gl.glCompressedTexImage2D.__name__ = 'glCompressedTexImage2D'
	gl.glCompressedTexImage2D.errcheck = _errcheck
	gl.glCompressedTexImage2D.restype = GLvoid
	gl.glCompressedTexImage2D.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage2DARB = _getProcAddress('glCompressedTexImage2DARB')
	gl.glCompressedTexImage2DARB.gl = gl
	gl.glCompressedTexImage2DARB.__name__ = 'glCompressedTexImage2DARB'
	gl.glCompressedTexImage2DARB.errcheck = _errcheck
	gl.glCompressedTexImage2DARB.restype = GLvoid
	gl.glCompressedTexImage2DARB.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage3D = _getProcAddress('glCompressedTexImage3D')
	gl.glCompressedTexImage3D.gl = gl
	gl.glCompressedTexImage3D.__name__ = 'glCompressedTexImage3D'
	gl.glCompressedTexImage3D.errcheck = _errcheck
	gl.glCompressedTexImage3D.restype = GLvoid
	gl.glCompressedTexImage3D.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage3DARB = _getProcAddress('glCompressedTexImage3DARB')
	gl.glCompressedTexImage3DARB.gl = gl
	gl.glCompressedTexImage3DARB.__name__ = 'glCompressedTexImage3DARB'
	gl.glCompressedTexImage3DARB.errcheck = _errcheck
	gl.glCompressedTexImage3DARB.restype = GLvoid
	gl.glCompressedTexImage3DARB.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexImage3DOES = _getProcAddress('glCompressedTexImage3DOES')
	gl.glCompressedTexImage3DOES.gl = gl
	gl.glCompressedTexImage3DOES.__name__ = 'glCompressedTexImage3DOES'
	gl.glCompressedTexImage3DOES.errcheck = _errcheck
	gl.glCompressedTexImage3DOES.restype = GLvoid
	gl.glCompressedTexImage3DOES.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage1D = _getProcAddress('glCompressedTexSubImage1D')
	gl.glCompressedTexSubImage1D.gl = gl
	gl.glCompressedTexSubImage1D.__name__ = 'glCompressedTexSubImage1D'
	gl.glCompressedTexSubImage1D.errcheck = _errcheck
	gl.glCompressedTexSubImage1D.restype = GLvoid
	gl.glCompressedTexSubImage1D.argtypes = (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage1DARB = _getProcAddress('glCompressedTexSubImage1DARB')
	gl.glCompressedTexSubImage1DARB.gl = gl
	gl.glCompressedTexSubImage1DARB.__name__ = 'glCompressedTexSubImage1DARB'
	gl.glCompressedTexSubImage1DARB.errcheck = _errcheck
	gl.glCompressedTexSubImage1DARB.restype = GLvoid
	gl.glCompressedTexSubImage1DARB.argtypes = (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage2D = _getProcAddress('glCompressedTexSubImage2D')
	gl.glCompressedTexSubImage2D.gl = gl
	gl.glCompressedTexSubImage2D.__name__ = 'glCompressedTexSubImage2D'
	gl.glCompressedTexSubImage2D.errcheck = _errcheck
	gl.glCompressedTexSubImage2D.restype = GLvoid
	gl.glCompressedTexSubImage2D.argtypes = (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage2DARB = _getProcAddress('glCompressedTexSubImage2DARB')
	gl.glCompressedTexSubImage2DARB.gl = gl
	gl.glCompressedTexSubImage2DARB.__name__ = 'glCompressedTexSubImage2DARB'
	gl.glCompressedTexSubImage2DARB.errcheck = _errcheck
	gl.glCompressedTexSubImage2DARB.restype = GLvoid
	gl.glCompressedTexSubImage2DARB.argtypes = (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage3D = _getProcAddress('glCompressedTexSubImage3D')
	gl.glCompressedTexSubImage3D.gl = gl
	gl.glCompressedTexSubImage3D.__name__ = 'glCompressedTexSubImage3D'
	gl.glCompressedTexSubImage3D.errcheck = _errcheck
	gl.glCompressedTexSubImage3D.restype = GLvoid
	gl.glCompressedTexSubImage3D.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage3DARB = _getProcAddress('glCompressedTexSubImage3DARB')
	gl.glCompressedTexSubImage3DARB.gl = gl
	gl.glCompressedTexSubImage3DARB.__name__ = 'glCompressedTexSubImage3DARB'
	gl.glCompressedTexSubImage3DARB.errcheck = _errcheck
	gl.glCompressedTexSubImage3DARB.restype = GLvoid
	gl.glCompressedTexSubImage3DARB.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTexSubImage3DOES = _getProcAddress('glCompressedTexSubImage3DOES')
	gl.glCompressedTexSubImage3DOES.gl = gl
	gl.glCompressedTexSubImage3DOES.__name__ = 'glCompressedTexSubImage3DOES'
	gl.glCompressedTexSubImage3DOES.errcheck = _errcheck
	gl.glCompressedTexSubImage3DOES.restype = GLvoid
	gl.glCompressedTexSubImage3DOES.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureImage1DEXT = _getProcAddress('glCompressedTextureImage1DEXT')
	gl.glCompressedTextureImage1DEXT.gl = gl
	gl.glCompressedTextureImage1DEXT.__name__ = 'glCompressedTextureImage1DEXT'
	gl.glCompressedTextureImage1DEXT.errcheck = _errcheck
	gl.glCompressedTextureImage1DEXT.restype = GLvoid
	gl.glCompressedTextureImage1DEXT.argtypes = (GLuint, GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureImage2DEXT = _getProcAddress('glCompressedTextureImage2DEXT')
	gl.glCompressedTextureImage2DEXT.gl = gl
	gl.glCompressedTextureImage2DEXT.__name__ = 'glCompressedTextureImage2DEXT'
	gl.glCompressedTextureImage2DEXT.errcheck = _errcheck
	gl.glCompressedTextureImage2DEXT.restype = GLvoid
	gl.glCompressedTextureImage2DEXT.argtypes = (GLuint, GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureImage3DEXT = _getProcAddress('glCompressedTextureImage3DEXT')
	gl.glCompressedTextureImage3DEXT.gl = gl
	gl.glCompressedTextureImage3DEXT.__name__ = 'glCompressedTextureImage3DEXT'
	gl.glCompressedTextureImage3DEXT.errcheck = _errcheck
	gl.glCompressedTextureImage3DEXT.restype = GLvoid
	gl.glCompressedTextureImage3DEXT.argtypes = (GLuint, GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureSubImage1D = _getProcAddress('glCompressedTextureSubImage1D')
	gl.glCompressedTextureSubImage1D.gl = gl
	gl.glCompressedTextureSubImage1D.__name__ = 'glCompressedTextureSubImage1D'
	gl.glCompressedTextureSubImage1D.errcheck = _errcheck
	gl.glCompressedTextureSubImage1D.restype = GLvoid
	gl.glCompressedTextureSubImage1D.argtypes = (GLuint, GLint, GLint, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureSubImage1DEXT = _getProcAddress('glCompressedTextureSubImage1DEXT')
	gl.glCompressedTextureSubImage1DEXT.gl = gl
	gl.glCompressedTextureSubImage1DEXT.__name__ = 'glCompressedTextureSubImage1DEXT'
	gl.glCompressedTextureSubImage1DEXT.errcheck = _errcheck
	gl.glCompressedTextureSubImage1DEXT.restype = GLvoid
	gl.glCompressedTextureSubImage1DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureSubImage2D = _getProcAddress('glCompressedTextureSubImage2D')
	gl.glCompressedTextureSubImage2D.gl = gl
	gl.glCompressedTextureSubImage2D.__name__ = 'glCompressedTextureSubImage2D'
	gl.glCompressedTextureSubImage2D.errcheck = _errcheck
	gl.glCompressedTextureSubImage2D.restype = GLvoid
	gl.glCompressedTextureSubImage2D.argtypes = (GLuint, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureSubImage2DEXT = _getProcAddress('glCompressedTextureSubImage2DEXT')
	gl.glCompressedTextureSubImage2DEXT.gl = gl
	gl.glCompressedTextureSubImage2DEXT.__name__ = 'glCompressedTextureSubImage2DEXT'
	gl.glCompressedTextureSubImage2DEXT.errcheck = _errcheck
	gl.glCompressedTextureSubImage2DEXT.restype = GLvoid
	gl.glCompressedTextureSubImage2DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureSubImage3D = _getProcAddress('glCompressedTextureSubImage3D')
	gl.glCompressedTextureSubImage3D.gl = gl
	gl.glCompressedTextureSubImage3D.__name__ = 'glCompressedTextureSubImage3D'
	gl.glCompressedTextureSubImage3D.errcheck = _errcheck
	gl.glCompressedTextureSubImage3D.restype = GLvoid
	gl.glCompressedTextureSubImage3D.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glCompressedTextureSubImage3DEXT = _getProcAddress('glCompressedTextureSubImage3DEXT')
	gl.glCompressedTextureSubImage3DEXT.gl = gl
	gl.glCompressedTextureSubImage3DEXT.__name__ = 'glCompressedTextureSubImage3DEXT'
	gl.glCompressedTextureSubImage3DEXT.errcheck = _errcheck
	gl.glCompressedTextureSubImage3DEXT.restype = GLvoid
	gl.glCompressedTextureSubImage3DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glConvolutionFilter1D = _getProcAddress('glConvolutionFilter1D')
	gl.glConvolutionFilter1D.gl = gl
	gl.glConvolutionFilter1D.__name__ = 'glConvolutionFilter1D'
	gl.glConvolutionFilter1D.errcheck = _errcheck
	gl.glConvolutionFilter1D.restype = GLvoid
	gl.glConvolutionFilter1D.argtypes = (GLenum, GLenum, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glConvolutionFilter1DEXT = _getProcAddress('glConvolutionFilter1DEXT')
	gl.glConvolutionFilter1DEXT.gl = gl
	gl.glConvolutionFilter1DEXT.__name__ = 'glConvolutionFilter1DEXT'
	gl.glConvolutionFilter1DEXT.errcheck = _errcheck
	gl.glConvolutionFilter1DEXT.restype = GLvoid
	gl.glConvolutionFilter1DEXT.argtypes = (GLenum, GLenum, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glConvolutionFilter2D = _getProcAddress('glConvolutionFilter2D')
	gl.glConvolutionFilter2D.gl = gl
	gl.glConvolutionFilter2D.__name__ = 'glConvolutionFilter2D'
	gl.glConvolutionFilter2D.errcheck = _errcheck
	gl.glConvolutionFilter2D.restype = GLvoid
	gl.glConvolutionFilter2D.argtypes = (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glConvolutionFilter2DEXT = _getProcAddress('glConvolutionFilter2DEXT')
	gl.glConvolutionFilter2DEXT.gl = gl
	gl.glConvolutionFilter2DEXT.__name__ = 'glConvolutionFilter2DEXT'
	gl.glConvolutionFilter2DEXT.errcheck = _errcheck
	gl.glConvolutionFilter2DEXT.restype = GLvoid
	gl.glConvolutionFilter2DEXT.argtypes = (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glConvolutionParameterf = _getProcAddress('glConvolutionParameterf')
	gl.glConvolutionParameterf.gl = gl
	gl.glConvolutionParameterf.__name__ = 'glConvolutionParameterf'
	gl.glConvolutionParameterf.errcheck = _errcheck
	gl.glConvolutionParameterf.restype = GLvoid
	gl.glConvolutionParameterf.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glConvolutionParameterfEXT = _getProcAddress('glConvolutionParameterfEXT')
	gl.glConvolutionParameterfEXT.gl = gl
	gl.glConvolutionParameterfEXT.__name__ = 'glConvolutionParameterfEXT'
	gl.glConvolutionParameterfEXT.errcheck = _errcheck
	gl.glConvolutionParameterfEXT.restype = GLvoid
	gl.glConvolutionParameterfEXT.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glConvolutionParameterfv = _getProcAddress('glConvolutionParameterfv')
	gl.glConvolutionParameterfv.gl = gl
	gl.glConvolutionParameterfv.__name__ = 'glConvolutionParameterfv'
	gl.glConvolutionParameterfv.errcheck = _errcheck
	gl.glConvolutionParameterfv.restype = GLvoid
	gl.glConvolutionParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glConvolutionParameterfvEXT = _getProcAddress('glConvolutionParameterfvEXT')
	gl.glConvolutionParameterfvEXT.gl = gl
	gl.glConvolutionParameterfvEXT.__name__ = 'glConvolutionParameterfvEXT'
	gl.glConvolutionParameterfvEXT.errcheck = _errcheck
	gl.glConvolutionParameterfvEXT.restype = GLvoid
	gl.glConvolutionParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glConvolutionParameteri = _getProcAddress('glConvolutionParameteri')
	gl.glConvolutionParameteri.gl = gl
	gl.glConvolutionParameteri.__name__ = 'glConvolutionParameteri'
	gl.glConvolutionParameteri.errcheck = _errcheck
	gl.glConvolutionParameteri.restype = GLvoid
	gl.glConvolutionParameteri.argtypes = (GLenum, GLenum, GLint,)
	gl.glConvolutionParameteriEXT = _getProcAddress('glConvolutionParameteriEXT')
	gl.glConvolutionParameteriEXT.gl = gl
	gl.glConvolutionParameteriEXT.__name__ = 'glConvolutionParameteriEXT'
	gl.glConvolutionParameteriEXT.errcheck = _errcheck
	gl.glConvolutionParameteriEXT.restype = GLvoid
	gl.glConvolutionParameteriEXT.argtypes = (GLenum, GLenum, GLint,)
	gl.glConvolutionParameteriv = _getProcAddress('glConvolutionParameteriv')
	gl.glConvolutionParameteriv.gl = gl
	gl.glConvolutionParameteriv.__name__ = 'glConvolutionParameteriv'
	gl.glConvolutionParameteriv.errcheck = _errcheck
	gl.glConvolutionParameteriv.restype = GLvoid
	gl.glConvolutionParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glConvolutionParameterivEXT = _getProcAddress('glConvolutionParameterivEXT')
	gl.glConvolutionParameterivEXT.gl = gl
	gl.glConvolutionParameterivEXT.__name__ = 'glConvolutionParameterivEXT'
	gl.glConvolutionParameterivEXT.errcheck = _errcheck
	gl.glConvolutionParameterivEXT.restype = GLvoid
	gl.glConvolutionParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glConvolutionParameterxOES = _getProcAddress('glConvolutionParameterxOES')
	gl.glConvolutionParameterxOES.gl = gl
	gl.glConvolutionParameterxOES.__name__ = 'glConvolutionParameterxOES'
	gl.glConvolutionParameterxOES.errcheck = _errcheck
	gl.glConvolutionParameterxOES.restype = GLvoid
	gl.glConvolutionParameterxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glConvolutionParameterxvOES = _getProcAddress('glConvolutionParameterxvOES')
	gl.glConvolutionParameterxvOES.gl = gl
	gl.glConvolutionParameterxvOES.__name__ = 'glConvolutionParameterxvOES'
	gl.glConvolutionParameterxvOES.errcheck = _errcheck
	gl.glConvolutionParameterxvOES.restype = GLvoid
	gl.glConvolutionParameterxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glCopyBufferSubData = _getProcAddress('glCopyBufferSubData')
	gl.glCopyBufferSubData.gl = gl
	gl.glCopyBufferSubData.__name__ = 'glCopyBufferSubData'
	gl.glCopyBufferSubData.errcheck = _errcheck
	gl.glCopyBufferSubData.restype = GLvoid
	gl.glCopyBufferSubData.argtypes = (GLenum, GLenum, GLintptr, GLintptr, GLsizeiptr,)
	gl.glCopyBufferSubDataNV = _getProcAddress('glCopyBufferSubDataNV')
	gl.glCopyBufferSubDataNV.gl = gl
	gl.glCopyBufferSubDataNV.__name__ = 'glCopyBufferSubDataNV'
	gl.glCopyBufferSubDataNV.errcheck = _errcheck
	gl.glCopyBufferSubDataNV.restype = GLvoid
	gl.glCopyBufferSubDataNV.argtypes = (GLenum, GLenum, GLintptr, GLintptr, GLsizeiptr,)
	gl.glCopyColorSubTable = _getProcAddress('glCopyColorSubTable')
	gl.glCopyColorSubTable.gl = gl
	gl.glCopyColorSubTable.__name__ = 'glCopyColorSubTable'
	gl.glCopyColorSubTable.errcheck = _errcheck
	gl.glCopyColorSubTable.restype = GLvoid
	gl.glCopyColorSubTable.argtypes = (GLenum, GLsizei, GLint, GLint, GLsizei,)
	gl.glCopyColorSubTableEXT = _getProcAddress('glCopyColorSubTableEXT')
	gl.glCopyColorSubTableEXT.gl = gl
	gl.glCopyColorSubTableEXT.__name__ = 'glCopyColorSubTableEXT'
	gl.glCopyColorSubTableEXT.errcheck = _errcheck
	gl.glCopyColorSubTableEXT.restype = GLvoid
	gl.glCopyColorSubTableEXT.argtypes = (GLenum, GLsizei, GLint, GLint, GLsizei,)
	gl.glCopyColorTable = _getProcAddress('glCopyColorTable')
	gl.glCopyColorTable.gl = gl
	gl.glCopyColorTable.__name__ = 'glCopyColorTable'
	gl.glCopyColorTable.errcheck = _errcheck
	gl.glCopyColorTable.restype = GLvoid
	gl.glCopyColorTable.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei,)
	gl.glCopyColorTableSGI = _getProcAddress('glCopyColorTableSGI')
	gl.glCopyColorTableSGI.gl = gl
	gl.glCopyColorTableSGI.__name__ = 'glCopyColorTableSGI'
	gl.glCopyColorTableSGI.errcheck = _errcheck
	gl.glCopyColorTableSGI.restype = GLvoid
	gl.glCopyColorTableSGI.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei,)
	gl.glCopyConvolutionFilter1D = _getProcAddress('glCopyConvolutionFilter1D')
	gl.glCopyConvolutionFilter1D.gl = gl
	gl.glCopyConvolutionFilter1D.__name__ = 'glCopyConvolutionFilter1D'
	gl.glCopyConvolutionFilter1D.errcheck = _errcheck
	gl.glCopyConvolutionFilter1D.restype = GLvoid
	gl.glCopyConvolutionFilter1D.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei,)
	gl.glCopyConvolutionFilter1DEXT = _getProcAddress('glCopyConvolutionFilter1DEXT')
	gl.glCopyConvolutionFilter1DEXT.gl = gl
	gl.glCopyConvolutionFilter1DEXT.__name__ = 'glCopyConvolutionFilter1DEXT'
	gl.glCopyConvolutionFilter1DEXT.errcheck = _errcheck
	gl.glCopyConvolutionFilter1DEXT.restype = GLvoid
	gl.glCopyConvolutionFilter1DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei,)
	gl.glCopyConvolutionFilter2D = _getProcAddress('glCopyConvolutionFilter2D')
	gl.glCopyConvolutionFilter2D.gl = gl
	gl.glCopyConvolutionFilter2D.__name__ = 'glCopyConvolutionFilter2D'
	gl.glCopyConvolutionFilter2D.errcheck = _errcheck
	gl.glCopyConvolutionFilter2D.restype = GLvoid
	gl.glCopyConvolutionFilter2D.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyConvolutionFilter2DEXT = _getProcAddress('glCopyConvolutionFilter2DEXT')
	gl.glCopyConvolutionFilter2DEXT.gl = gl
	gl.glCopyConvolutionFilter2DEXT.__name__ = 'glCopyConvolutionFilter2DEXT'
	gl.glCopyConvolutionFilter2DEXT.errcheck = _errcheck
	gl.glCopyConvolutionFilter2DEXT.restype = GLvoid
	gl.glCopyConvolutionFilter2DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyImageSubData = _getProcAddress('glCopyImageSubData')
	gl.glCopyImageSubData.gl = gl
	gl.glCopyImageSubData.__name__ = 'glCopyImageSubData'
	gl.glCopyImageSubData.errcheck = _errcheck
	gl.glCopyImageSubData.restype = GLvoid
	gl.glCopyImageSubData.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLuint, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei,)
	gl.glCopyImageSubDataEXT = _getProcAddress('glCopyImageSubDataEXT')
	gl.glCopyImageSubDataEXT.gl = gl
	gl.glCopyImageSubDataEXT.__name__ = 'glCopyImageSubDataEXT'
	gl.glCopyImageSubDataEXT.errcheck = _errcheck
	gl.glCopyImageSubDataEXT.restype = GLvoid
	gl.glCopyImageSubDataEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLuint, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei,)
	gl.glCopyImageSubDataNV = _getProcAddress('glCopyImageSubDataNV')
	gl.glCopyImageSubDataNV.gl = gl
	gl.glCopyImageSubDataNV.__name__ = 'glCopyImageSubDataNV'
	gl.glCopyImageSubDataNV.errcheck = _errcheck
	gl.glCopyImageSubDataNV.restype = GLvoid
	gl.glCopyImageSubDataNV.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLuint, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei,)
	gl.glCopyMultiTexImage1DEXT = _getProcAddress('glCopyMultiTexImage1DEXT')
	gl.glCopyMultiTexImage1DEXT.gl = gl
	gl.glCopyMultiTexImage1DEXT.__name__ = 'glCopyMultiTexImage1DEXT'
	gl.glCopyMultiTexImage1DEXT.errcheck = _errcheck
	gl.glCopyMultiTexImage1DEXT.restype = GLvoid
	gl.glCopyMultiTexImage1DEXT.argtypes = (GLenum, GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLint,)
	gl.glCopyMultiTexImage2DEXT = _getProcAddress('glCopyMultiTexImage2DEXT')
	gl.glCopyMultiTexImage2DEXT.gl = gl
	gl.glCopyMultiTexImage2DEXT.__name__ = 'glCopyMultiTexImage2DEXT'
	gl.glCopyMultiTexImage2DEXT.errcheck = _errcheck
	gl.glCopyMultiTexImage2DEXT.restype = GLvoid
	gl.glCopyMultiTexImage2DEXT.argtypes = (GLenum, GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLsizei, GLint,)
	gl.glCopyMultiTexSubImage1DEXT = _getProcAddress('glCopyMultiTexSubImage1DEXT')
	gl.glCopyMultiTexSubImage1DEXT.gl = gl
	gl.glCopyMultiTexSubImage1DEXT.__name__ = 'glCopyMultiTexSubImage1DEXT'
	gl.glCopyMultiTexSubImage1DEXT.errcheck = _errcheck
	gl.glCopyMultiTexSubImage1DEXT.restype = GLvoid
	gl.glCopyMultiTexSubImage1DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLint, GLsizei,)
	gl.glCopyMultiTexSubImage2DEXT = _getProcAddress('glCopyMultiTexSubImage2DEXT')
	gl.glCopyMultiTexSubImage2DEXT.gl = gl
	gl.glCopyMultiTexSubImage2DEXT.__name__ = 'glCopyMultiTexSubImage2DEXT'
	gl.glCopyMultiTexSubImage2DEXT.errcheck = _errcheck
	gl.glCopyMultiTexSubImage2DEXT.restype = GLvoid
	gl.glCopyMultiTexSubImage2DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyMultiTexSubImage3DEXT = _getProcAddress('glCopyMultiTexSubImage3DEXT')
	gl.glCopyMultiTexSubImage3DEXT.gl = gl
	gl.glCopyMultiTexSubImage3DEXT.__name__ = 'glCopyMultiTexSubImage3DEXT'
	gl.glCopyMultiTexSubImage3DEXT.errcheck = _errcheck
	gl.glCopyMultiTexSubImage3DEXT.restype = GLvoid
	gl.glCopyMultiTexSubImage3DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyNamedBufferSubData = _getProcAddress('glCopyNamedBufferSubData')
	gl.glCopyNamedBufferSubData.gl = gl
	gl.glCopyNamedBufferSubData.__name__ = 'glCopyNamedBufferSubData'
	gl.glCopyNamedBufferSubData.errcheck = _errcheck
	gl.glCopyNamedBufferSubData.restype = GLvoid
	gl.glCopyNamedBufferSubData.argtypes = (GLuint, GLuint, GLintptr, GLintptr, GLsizei,)
	gl.glCopyPathNV = _getProcAddress('glCopyPathNV')
	gl.glCopyPathNV.gl = gl
	gl.glCopyPathNV.__name__ = 'glCopyPathNV'
	gl.glCopyPathNV.errcheck = _errcheck
	gl.glCopyPathNV.restype = GLvoid
	gl.glCopyPathNV.argtypes = (GLuint, GLuint,)
	gl.glCopyPixels = _getProcAddress('glCopyPixels')
	gl.glCopyPixels.gl = gl
	gl.glCopyPixels.__name__ = 'glCopyPixels'
	gl.glCopyPixels.errcheck = _errcheck
	gl.glCopyPixels.restype = GLvoid
	gl.glCopyPixels.argtypes = (GLint, GLint, GLsizei, GLsizei, GLenum,)
	gl.glCopyTexImage1D = _getProcAddress('glCopyTexImage1D')
	gl.glCopyTexImage1D.gl = gl
	gl.glCopyTexImage1D.__name__ = 'glCopyTexImage1D'
	gl.glCopyTexImage1D.errcheck = _errcheck
	gl.glCopyTexImage1D.restype = GLvoid
	gl.glCopyTexImage1D.argtypes = (GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLint,)
	gl.glCopyTexImage1DEXT = _getProcAddress('glCopyTexImage1DEXT')
	gl.glCopyTexImage1DEXT.gl = gl
	gl.glCopyTexImage1DEXT.__name__ = 'glCopyTexImage1DEXT'
	gl.glCopyTexImage1DEXT.errcheck = _errcheck
	gl.glCopyTexImage1DEXT.restype = GLvoid
	gl.glCopyTexImage1DEXT.argtypes = (GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLint,)
	gl.glCopyTexImage2D = _getProcAddress('glCopyTexImage2D')
	gl.glCopyTexImage2D.gl = gl
	gl.glCopyTexImage2D.__name__ = 'glCopyTexImage2D'
	gl.glCopyTexImage2D.errcheck = _errcheck
	gl.glCopyTexImage2D.restype = GLvoid
	gl.glCopyTexImage2D.argtypes = (GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLsizei, GLint,)
	gl.glCopyTexImage2DEXT = _getProcAddress('glCopyTexImage2DEXT')
	gl.glCopyTexImage2DEXT.gl = gl
	gl.glCopyTexImage2DEXT.__name__ = 'glCopyTexImage2DEXT'
	gl.glCopyTexImage2DEXT.errcheck = _errcheck
	gl.glCopyTexImage2DEXT.restype = GLvoid
	gl.glCopyTexImage2DEXT.argtypes = (GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLsizei, GLint,)
	gl.glCopyTexSubImage1D = _getProcAddress('glCopyTexSubImage1D')
	gl.glCopyTexSubImage1D.gl = gl
	gl.glCopyTexSubImage1D.__name__ = 'glCopyTexSubImage1D'
	gl.glCopyTexSubImage1D.errcheck = _errcheck
	gl.glCopyTexSubImage1D.restype = GLvoid
	gl.glCopyTexSubImage1D.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei,)
	gl.glCopyTexSubImage1DEXT = _getProcAddress('glCopyTexSubImage1DEXT')
	gl.glCopyTexSubImage1DEXT.gl = gl
	gl.glCopyTexSubImage1DEXT.__name__ = 'glCopyTexSubImage1DEXT'
	gl.glCopyTexSubImage1DEXT.errcheck = _errcheck
	gl.glCopyTexSubImage1DEXT.restype = GLvoid
	gl.glCopyTexSubImage1DEXT.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei,)
	gl.glCopyTexSubImage2D = _getProcAddress('glCopyTexSubImage2D')
	gl.glCopyTexSubImage2D.gl = gl
	gl.glCopyTexSubImage2D.__name__ = 'glCopyTexSubImage2D'
	gl.glCopyTexSubImage2D.errcheck = _errcheck
	gl.glCopyTexSubImage2D.restype = GLvoid
	gl.glCopyTexSubImage2D.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTexSubImage2DEXT = _getProcAddress('glCopyTexSubImage2DEXT')
	gl.glCopyTexSubImage2DEXT.gl = gl
	gl.glCopyTexSubImage2DEXT.__name__ = 'glCopyTexSubImage2DEXT'
	gl.glCopyTexSubImage2DEXT.errcheck = _errcheck
	gl.glCopyTexSubImage2DEXT.restype = GLvoid
	gl.glCopyTexSubImage2DEXT.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTexSubImage3D = _getProcAddress('glCopyTexSubImage3D')
	gl.glCopyTexSubImage3D.gl = gl
	gl.glCopyTexSubImage3D.__name__ = 'glCopyTexSubImage3D'
	gl.glCopyTexSubImage3D.errcheck = _errcheck
	gl.glCopyTexSubImage3D.restype = GLvoid
	gl.glCopyTexSubImage3D.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTexSubImage3DEXT = _getProcAddress('glCopyTexSubImage3DEXT')
	gl.glCopyTexSubImage3DEXT.gl = gl
	gl.glCopyTexSubImage3DEXT.__name__ = 'glCopyTexSubImage3DEXT'
	gl.glCopyTexSubImage3DEXT.errcheck = _errcheck
	gl.glCopyTexSubImage3DEXT.restype = GLvoid
	gl.glCopyTexSubImage3DEXT.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTexSubImage3DOES = _getProcAddress('glCopyTexSubImage3DOES')
	gl.glCopyTexSubImage3DOES.gl = gl
	gl.glCopyTexSubImage3DOES.__name__ = 'glCopyTexSubImage3DOES'
	gl.glCopyTexSubImage3DOES.errcheck = _errcheck
	gl.glCopyTexSubImage3DOES.restype = GLvoid
	gl.glCopyTexSubImage3DOES.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTextureImage1DEXT = _getProcAddress('glCopyTextureImage1DEXT')
	gl.glCopyTextureImage1DEXT.gl = gl
	gl.glCopyTextureImage1DEXT.__name__ = 'glCopyTextureImage1DEXT'
	gl.glCopyTextureImage1DEXT.errcheck = _errcheck
	gl.glCopyTextureImage1DEXT.restype = GLvoid
	gl.glCopyTextureImage1DEXT.argtypes = (GLuint, GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLint,)
	gl.glCopyTextureImage2DEXT = _getProcAddress('glCopyTextureImage2DEXT')
	gl.glCopyTextureImage2DEXT.gl = gl
	gl.glCopyTextureImage2DEXT.__name__ = 'glCopyTextureImage2DEXT'
	gl.glCopyTextureImage2DEXT.errcheck = _errcheck
	gl.glCopyTextureImage2DEXT.restype = GLvoid
	gl.glCopyTextureImage2DEXT.argtypes = (GLuint, GLenum, GLint, GLenum, GLint, GLint, GLsizei, GLsizei, GLint,)
	gl.glCopyTextureLevelsAPPLE = _getProcAddress('glCopyTextureLevelsAPPLE')
	gl.glCopyTextureLevelsAPPLE.gl = gl
	gl.glCopyTextureLevelsAPPLE.__name__ = 'glCopyTextureLevelsAPPLE'
	gl.glCopyTextureLevelsAPPLE.errcheck = _errcheck
	gl.glCopyTextureLevelsAPPLE.restype = GLvoid
	gl.glCopyTextureLevelsAPPLE.argtypes = (GLuint, GLuint, GLint, GLsizei,)
	gl.glCopyTextureSubImage1D = _getProcAddress('glCopyTextureSubImage1D')
	gl.glCopyTextureSubImage1D.gl = gl
	gl.glCopyTextureSubImage1D.__name__ = 'glCopyTextureSubImage1D'
	gl.glCopyTextureSubImage1D.errcheck = _errcheck
	gl.glCopyTextureSubImage1D.restype = GLvoid
	gl.glCopyTextureSubImage1D.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei,)
	gl.glCopyTextureSubImage1DEXT = _getProcAddress('glCopyTextureSubImage1DEXT')
	gl.glCopyTextureSubImage1DEXT.gl = gl
	gl.glCopyTextureSubImage1DEXT.__name__ = 'glCopyTextureSubImage1DEXT'
	gl.glCopyTextureSubImage1DEXT.errcheck = _errcheck
	gl.glCopyTextureSubImage1DEXT.restype = GLvoid
	gl.glCopyTextureSubImage1DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLsizei,)
	gl.glCopyTextureSubImage2D = _getProcAddress('glCopyTextureSubImage2D')
	gl.glCopyTextureSubImage2D.gl = gl
	gl.glCopyTextureSubImage2D.__name__ = 'glCopyTextureSubImage2D'
	gl.glCopyTextureSubImage2D.errcheck = _errcheck
	gl.glCopyTextureSubImage2D.restype = GLvoid
	gl.glCopyTextureSubImage2D.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTextureSubImage2DEXT = _getProcAddress('glCopyTextureSubImage2DEXT')
	gl.glCopyTextureSubImage2DEXT.gl = gl
	gl.glCopyTextureSubImage2DEXT.__name__ = 'glCopyTextureSubImage2DEXT'
	gl.glCopyTextureSubImage2DEXT.errcheck = _errcheck
	gl.glCopyTextureSubImage2DEXT.restype = GLvoid
	gl.glCopyTextureSubImage2DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTextureSubImage3D = _getProcAddress('glCopyTextureSubImage3D')
	gl.glCopyTextureSubImage3D.gl = gl
	gl.glCopyTextureSubImage3D.__name__ = 'glCopyTextureSubImage3D'
	gl.glCopyTextureSubImage3D.errcheck = _errcheck
	gl.glCopyTextureSubImage3D.restype = GLvoid
	gl.glCopyTextureSubImage3D.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCopyTextureSubImage3DEXT = _getProcAddress('glCopyTextureSubImage3DEXT')
	gl.glCopyTextureSubImage3DEXT.gl = gl
	gl.glCopyTextureSubImage3DEXT.__name__ = 'glCopyTextureSubImage3DEXT'
	gl.glCopyTextureSubImage3DEXT.errcheck = _errcheck
	gl.glCopyTextureSubImage3DEXT.restype = GLvoid
	gl.glCopyTextureSubImage3DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei,)
	gl.glCoverFillPathInstancedNV = _getProcAddress('glCoverFillPathInstancedNV')
	gl.glCoverFillPathInstancedNV.gl = gl
	gl.glCoverFillPathInstancedNV.__name__ = 'glCoverFillPathInstancedNV'
	gl.glCoverFillPathInstancedNV.errcheck = _errcheck
	gl.glCoverFillPathInstancedNV.restype = GLvoid
	gl.glCoverFillPathInstancedNV.argtypes = (GLsizei, GLenum, POINTER(GLvoid), GLuint, GLenum, GLenum, POINTER(GLfloat),)
	gl.glCoverFillPathNV = _getProcAddress('glCoverFillPathNV')
	gl.glCoverFillPathNV.gl = gl
	gl.glCoverFillPathNV.__name__ = 'glCoverFillPathNV'
	gl.glCoverFillPathNV.errcheck = _errcheck
	gl.glCoverFillPathNV.restype = GLvoid
	gl.glCoverFillPathNV.argtypes = (GLuint, GLenum,)
	gl.glCoverStrokePathInstancedNV = _getProcAddress('glCoverStrokePathInstancedNV')
	gl.glCoverStrokePathInstancedNV.gl = gl
	gl.glCoverStrokePathInstancedNV.__name__ = 'glCoverStrokePathInstancedNV'
	gl.glCoverStrokePathInstancedNV.errcheck = _errcheck
	gl.glCoverStrokePathInstancedNV.restype = GLvoid
	gl.glCoverStrokePathInstancedNV.argtypes = (GLsizei, GLenum, POINTER(GLvoid), GLuint, GLenum, GLenum, POINTER(GLfloat),)
	gl.glCoverStrokePathNV = _getProcAddress('glCoverStrokePathNV')
	gl.glCoverStrokePathNV.gl = gl
	gl.glCoverStrokePathNV.__name__ = 'glCoverStrokePathNV'
	gl.glCoverStrokePathNV.errcheck = _errcheck
	gl.glCoverStrokePathNV.restype = GLvoid
	gl.glCoverStrokePathNV.argtypes = (GLuint, GLenum,)
	gl.glCoverageMaskNV = _getProcAddress('glCoverageMaskNV')
	gl.glCoverageMaskNV.gl = gl
	gl.glCoverageMaskNV.__name__ = 'glCoverageMaskNV'
	gl.glCoverageMaskNV.errcheck = _errcheck
	gl.glCoverageMaskNV.restype = GLvoid
	gl.glCoverageMaskNV.argtypes = (GLboolean,)
	gl.glCoverageOperationNV = _getProcAddress('glCoverageOperationNV')
	gl.glCoverageOperationNV.gl = gl
	gl.glCoverageOperationNV.__name__ = 'glCoverageOperationNV'
	gl.glCoverageOperationNV.errcheck = _errcheck
	gl.glCoverageOperationNV.restype = GLvoid
	gl.glCoverageOperationNV.argtypes = (GLenum,)
	gl.glCreateBuffers = _getProcAddress('glCreateBuffers')
	gl.glCreateBuffers.gl = gl
	gl.glCreateBuffers.__name__ = 'glCreateBuffers'
	gl.glCreateBuffers.errcheck = _errcheck
	gl.glCreateBuffers.restype = GLvoid
	gl.glCreateBuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCreateFramebuffers = _getProcAddress('glCreateFramebuffers')
	gl.glCreateFramebuffers.gl = gl
	gl.glCreateFramebuffers.__name__ = 'glCreateFramebuffers'
	gl.glCreateFramebuffers.errcheck = _errcheck
	gl.glCreateFramebuffers.restype = GLvoid
	gl.glCreateFramebuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCreatePerfQueryINTEL = _getProcAddress('glCreatePerfQueryINTEL')
	gl.glCreatePerfQueryINTEL.gl = gl
	gl.glCreatePerfQueryINTEL.__name__ = 'glCreatePerfQueryINTEL'
	gl.glCreatePerfQueryINTEL.errcheck = _errcheck
	gl.glCreatePerfQueryINTEL.restype = GLvoid
	gl.glCreatePerfQueryINTEL.argtypes = (GLuint, POINTER(GLuint),)
	gl.glCreateProgram = _getProcAddress('glCreateProgram')
	gl.glCreateProgram.gl = gl
	gl.glCreateProgram.__name__ = 'glCreateProgram'
	gl.glCreateProgram.errcheck = _errcheck
	gl.glCreateProgram.restype = GLuint
	gl.glCreateProgram.argtypes = ()
	gl.glCreateProgramObjectARB = _getProcAddress('glCreateProgramObjectARB')
	gl.glCreateProgramObjectARB.gl = gl
	gl.glCreateProgramObjectARB.__name__ = 'glCreateProgramObjectARB'
	gl.glCreateProgramObjectARB.errcheck = _errcheck
	gl.glCreateProgramObjectARB.restype = GLhandleARB
	gl.glCreateProgramObjectARB.argtypes = ()
	gl.glCreateProgramPipelines = _getProcAddress('glCreateProgramPipelines')
	gl.glCreateProgramPipelines.gl = gl
	gl.glCreateProgramPipelines.__name__ = 'glCreateProgramPipelines'
	gl.glCreateProgramPipelines.errcheck = _errcheck
	gl.glCreateProgramPipelines.restype = GLvoid
	gl.glCreateProgramPipelines.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCreateQueries = _getProcAddress('glCreateQueries')
	gl.glCreateQueries.gl = gl
	gl.glCreateQueries.__name__ = 'glCreateQueries'
	gl.glCreateQueries.errcheck = _errcheck
	gl.glCreateQueries.restype = GLvoid
	gl.glCreateQueries.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glCreateRenderbuffers = _getProcAddress('glCreateRenderbuffers')
	gl.glCreateRenderbuffers.gl = gl
	gl.glCreateRenderbuffers.__name__ = 'glCreateRenderbuffers'
	gl.glCreateRenderbuffers.errcheck = _errcheck
	gl.glCreateRenderbuffers.restype = GLvoid
	gl.glCreateRenderbuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCreateSamplers = _getProcAddress('glCreateSamplers')
	gl.glCreateSamplers.gl = gl
	gl.glCreateSamplers.__name__ = 'glCreateSamplers'
	gl.glCreateSamplers.errcheck = _errcheck
	gl.glCreateSamplers.restype = GLvoid
	gl.glCreateSamplers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCreateShader = _getProcAddress('glCreateShader')
	gl.glCreateShader.gl = gl
	gl.glCreateShader.__name__ = 'glCreateShader'
	gl.glCreateShader.errcheck = _errcheck
	gl.glCreateShader.restype = GLuint
	gl.glCreateShader.argtypes = (GLenum,)
	gl.glCreateShaderObjectARB = _getProcAddress('glCreateShaderObjectARB')
	gl.glCreateShaderObjectARB.gl = gl
	gl.glCreateShaderObjectARB.__name__ = 'glCreateShaderObjectARB'
	gl.glCreateShaderObjectARB.errcheck = _errcheck
	gl.glCreateShaderObjectARB.restype = GLhandleARB
	gl.glCreateShaderObjectARB.argtypes = (GLenum,)
	gl.glCreateShaderProgramEXT = _getProcAddress('glCreateShaderProgramEXT')
	gl.glCreateShaderProgramEXT.gl = gl
	gl.glCreateShaderProgramEXT.__name__ = 'glCreateShaderProgramEXT'
	gl.glCreateShaderProgramEXT.errcheck = _errcheck
	gl.glCreateShaderProgramEXT.restype = GLuint
	gl.glCreateShaderProgramEXT.argtypes = (GLenum, POINTER(GLchar),)
	gl.glCreateShaderProgramv = _getProcAddress('glCreateShaderProgramv')
	gl.glCreateShaderProgramv.gl = gl
	gl.glCreateShaderProgramv.__name__ = 'glCreateShaderProgramv'
	gl.glCreateShaderProgramv.errcheck = _errcheck
	gl.glCreateShaderProgramv.restype = GLuint
	gl.glCreateShaderProgramv.argtypes = (GLenum, GLsizei, POINTER(POINTER(GLchar)),)
	gl.glCreateShaderProgramvEXT = _getProcAddress('glCreateShaderProgramvEXT')
	gl.glCreateShaderProgramvEXT.gl = gl
	gl.glCreateShaderProgramvEXT.__name__ = 'glCreateShaderProgramvEXT'
	gl.glCreateShaderProgramvEXT.errcheck = _errcheck
	gl.glCreateShaderProgramvEXT.restype = GLuint
	gl.glCreateShaderProgramvEXT.argtypes = (GLenum, GLsizei, POINTER(POINTER(GLchar)),)
	gl.glCreateSyncFromCLeventARB = _getProcAddress('glCreateSyncFromCLeventARB')
	gl.glCreateSyncFromCLeventARB.gl = gl
	gl.glCreateSyncFromCLeventARB.__name__ = 'glCreateSyncFromCLeventARB'
	gl.glCreateSyncFromCLeventARB.errcheck = _errcheck
	gl.glCreateSyncFromCLeventARB.restype = GLsync
	gl.glCreateSyncFromCLeventARB.argtypes = (POINTER(GLvoid), POINTER(GLvoid), GLbitfield,)
	gl.glCreateTextures = _getProcAddress('glCreateTextures')
	gl.glCreateTextures.gl = gl
	gl.glCreateTextures.__name__ = 'glCreateTextures'
	gl.glCreateTextures.errcheck = _errcheck
	gl.glCreateTextures.restype = GLvoid
	gl.glCreateTextures.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glCreateTransformFeedbacks = _getProcAddress('glCreateTransformFeedbacks')
	gl.glCreateTransformFeedbacks.gl = gl
	gl.glCreateTransformFeedbacks.__name__ = 'glCreateTransformFeedbacks'
	gl.glCreateTransformFeedbacks.errcheck = _errcheck
	gl.glCreateTransformFeedbacks.restype = GLvoid
	gl.glCreateTransformFeedbacks.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCreateVertexArrays = _getProcAddress('glCreateVertexArrays')
	gl.glCreateVertexArrays.gl = gl
	gl.glCreateVertexArrays.__name__ = 'glCreateVertexArrays'
	gl.glCreateVertexArrays.errcheck = _errcheck
	gl.glCreateVertexArrays.restype = GLvoid
	gl.glCreateVertexArrays.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glCullFace = _getProcAddress('glCullFace')
	gl.glCullFace.gl = gl
	gl.glCullFace.__name__ = 'glCullFace'
	gl.glCullFace.errcheck = _errcheck
	gl.glCullFace.restype = GLvoid
	gl.glCullFace.argtypes = (GLenum,)
	gl.glCullParameterdvEXT = _getProcAddress('glCullParameterdvEXT')
	gl.glCullParameterdvEXT.gl = gl
	gl.glCullParameterdvEXT.__name__ = 'glCullParameterdvEXT'
	gl.glCullParameterdvEXT.errcheck = _errcheck
	gl.glCullParameterdvEXT.restype = GLvoid
	gl.glCullParameterdvEXT.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glCullParameterfvEXT = _getProcAddress('glCullParameterfvEXT')
	gl.glCullParameterfvEXT.gl = gl
	gl.glCullParameterfvEXT.__name__ = 'glCullParameterfvEXT'
	gl.glCullParameterfvEXT.errcheck = _errcheck
	gl.glCullParameterfvEXT.restype = GLvoid
	gl.glCullParameterfvEXT.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glCurrentPaletteMatrixARB = _getProcAddress('glCurrentPaletteMatrixARB')
	gl.glCurrentPaletteMatrixARB.gl = gl
	gl.glCurrentPaletteMatrixARB.__name__ = 'glCurrentPaletteMatrixARB'
	gl.glCurrentPaletteMatrixARB.errcheck = _errcheck
	gl.glCurrentPaletteMatrixARB.restype = GLvoid
	gl.glCurrentPaletteMatrixARB.argtypes = (GLint,)
	gl.glCurrentPaletteMatrixOES = _getProcAddress('glCurrentPaletteMatrixOES')
	gl.glCurrentPaletteMatrixOES.gl = gl
	gl.glCurrentPaletteMatrixOES.__name__ = 'glCurrentPaletteMatrixOES'
	gl.glCurrentPaletteMatrixOES.errcheck = _errcheck
	gl.glCurrentPaletteMatrixOES.restype = GLvoid
	gl.glCurrentPaletteMatrixOES.argtypes = (GLuint,)
	gl.glDebugMessageCallback = _getProcAddress('glDebugMessageCallback')
	gl.glDebugMessageCallback.gl = gl
	gl.glDebugMessageCallback.__name__ = 'glDebugMessageCallback'
	gl.glDebugMessageCallback.errcheck = _errcheck
	gl.glDebugMessageCallback.restype = GLvoid
	gl.glDebugMessageCallback.argtypes = (GLDEBUGPROC, POINTER(GLvoid),)
	gl.glDebugMessageCallbackAMD = _getProcAddress('glDebugMessageCallbackAMD')
	gl.glDebugMessageCallbackAMD.gl = gl
	gl.glDebugMessageCallbackAMD.__name__ = 'glDebugMessageCallbackAMD'
	gl.glDebugMessageCallbackAMD.errcheck = _errcheck
	gl.glDebugMessageCallbackAMD.restype = GLvoid
	gl.glDebugMessageCallbackAMD.argtypes = (GLDEBUGPROCAMD, POINTER(GLvoid),)
	gl.glDebugMessageCallbackARB = _getProcAddress('glDebugMessageCallbackARB')
	gl.glDebugMessageCallbackARB.gl = gl
	gl.glDebugMessageCallbackARB.__name__ = 'glDebugMessageCallbackARB'
	gl.glDebugMessageCallbackARB.errcheck = _errcheck
	gl.glDebugMessageCallbackARB.restype = GLvoid
	gl.glDebugMessageCallbackARB.argtypes = (GLDEBUGPROCARB, POINTER(GLvoid),)
	gl.glDebugMessageCallbackKHR = _getProcAddress('glDebugMessageCallbackKHR')
	gl.glDebugMessageCallbackKHR.gl = gl
	gl.glDebugMessageCallbackKHR.__name__ = 'glDebugMessageCallbackKHR'
	gl.glDebugMessageCallbackKHR.errcheck = _errcheck
	gl.glDebugMessageCallbackKHR.restype = GLvoid
	gl.glDebugMessageCallbackKHR.argtypes = (GLDEBUGPROCKHR, POINTER(GLvoid),)
	gl.glDebugMessageControl = _getProcAddress('glDebugMessageControl')
	gl.glDebugMessageControl.gl = gl
	gl.glDebugMessageControl.__name__ = 'glDebugMessageControl'
	gl.glDebugMessageControl.errcheck = _errcheck
	gl.glDebugMessageControl.restype = GLvoid
	gl.glDebugMessageControl.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLuint), GLboolean,)
	gl.glDebugMessageControlARB = _getProcAddress('glDebugMessageControlARB')
	gl.glDebugMessageControlARB.gl = gl
	gl.glDebugMessageControlARB.__name__ = 'glDebugMessageControlARB'
	gl.glDebugMessageControlARB.errcheck = _errcheck
	gl.glDebugMessageControlARB.restype = GLvoid
	gl.glDebugMessageControlARB.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLuint), GLboolean,)
	gl.glDebugMessageControlKHR = _getProcAddress('glDebugMessageControlKHR')
	gl.glDebugMessageControlKHR.gl = gl
	gl.glDebugMessageControlKHR.__name__ = 'glDebugMessageControlKHR'
	gl.glDebugMessageControlKHR.errcheck = _errcheck
	gl.glDebugMessageControlKHR.restype = GLvoid
	gl.glDebugMessageControlKHR.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLuint), GLboolean,)
	gl.glDebugMessageEnableAMD = _getProcAddress('glDebugMessageEnableAMD')
	gl.glDebugMessageEnableAMD.gl = gl
	gl.glDebugMessageEnableAMD.__name__ = 'glDebugMessageEnableAMD'
	gl.glDebugMessageEnableAMD.errcheck = _errcheck
	gl.glDebugMessageEnableAMD.restype = GLvoid
	gl.glDebugMessageEnableAMD.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLuint), GLboolean,)
	gl.glDebugMessageInsert = _getProcAddress('glDebugMessageInsert')
	gl.glDebugMessageInsert.gl = gl
	gl.glDebugMessageInsert.__name__ = 'glDebugMessageInsert'
	gl.glDebugMessageInsert.errcheck = _errcheck
	gl.glDebugMessageInsert.restype = GLvoid
	gl.glDebugMessageInsert.argtypes = (GLenum, GLenum, GLuint, GLenum, GLsizei, POINTER(GLchar),)
	gl.glDebugMessageInsertAMD = _getProcAddress('glDebugMessageInsertAMD')
	gl.glDebugMessageInsertAMD.gl = gl
	gl.glDebugMessageInsertAMD.__name__ = 'glDebugMessageInsertAMD'
	gl.glDebugMessageInsertAMD.errcheck = _errcheck
	gl.glDebugMessageInsertAMD.restype = GLvoid
	gl.glDebugMessageInsertAMD.argtypes = (GLenum, GLenum, GLuint, GLsizei, POINTER(GLchar),)
	gl.glDebugMessageInsertARB = _getProcAddress('glDebugMessageInsertARB')
	gl.glDebugMessageInsertARB.gl = gl
	gl.glDebugMessageInsertARB.__name__ = 'glDebugMessageInsertARB'
	gl.glDebugMessageInsertARB.errcheck = _errcheck
	gl.glDebugMessageInsertARB.restype = GLvoid
	gl.glDebugMessageInsertARB.argtypes = (GLenum, GLenum, GLuint, GLenum, GLsizei, POINTER(GLchar),)
	gl.glDebugMessageInsertKHR = _getProcAddress('glDebugMessageInsertKHR')
	gl.glDebugMessageInsertKHR.gl = gl
	gl.glDebugMessageInsertKHR.__name__ = 'glDebugMessageInsertKHR'
	gl.glDebugMessageInsertKHR.errcheck = _errcheck
	gl.glDebugMessageInsertKHR.restype = GLvoid
	gl.glDebugMessageInsertKHR.argtypes = (GLenum, GLenum, GLuint, GLenum, GLsizei, POINTER(GLchar),)
	gl.glDeformSGIX = _getProcAddress('glDeformSGIX')
	gl.glDeformSGIX.gl = gl
	gl.glDeformSGIX.__name__ = 'glDeformSGIX'
	gl.glDeformSGIX.errcheck = _errcheck
	gl.glDeformSGIX.restype = GLvoid
	gl.glDeformSGIX.argtypes = (GLbitfield,)
	gl.glDeformationMap3dSGIX = _getProcAddress('glDeformationMap3dSGIX')
	gl.glDeformationMap3dSGIX.gl = gl
	gl.glDeformationMap3dSGIX.__name__ = 'glDeformationMap3dSGIX'
	gl.glDeformationMap3dSGIX.errcheck = _errcheck
	gl.glDeformationMap3dSGIX.restype = GLvoid
	gl.glDeformationMap3dSGIX.argtypes = (GLenum, GLdouble, GLdouble, GLint, GLint, GLdouble, GLdouble, GLint, GLint, GLdouble, GLdouble, GLint, GLint, POINTER(GLdouble),)
	gl.glDeformationMap3fSGIX = _getProcAddress('glDeformationMap3fSGIX')
	gl.glDeformationMap3fSGIX.gl = gl
	gl.glDeformationMap3fSGIX.__name__ = 'glDeformationMap3fSGIX'
	gl.glDeformationMap3fSGIX.errcheck = _errcheck
	gl.glDeformationMap3fSGIX.restype = GLvoid
	gl.glDeformationMap3fSGIX.argtypes = (GLenum, GLfloat, GLfloat, GLint, GLint, GLfloat, GLfloat, GLint, GLint, GLfloat, GLfloat, GLint, GLint, POINTER(GLfloat),)
	gl.glDeleteAsyncMarkersSGIX = _getProcAddress('glDeleteAsyncMarkersSGIX')
	gl.glDeleteAsyncMarkersSGIX.gl = gl
	gl.glDeleteAsyncMarkersSGIX.__name__ = 'glDeleteAsyncMarkersSGIX'
	gl.glDeleteAsyncMarkersSGIX.errcheck = _errcheck
	gl.glDeleteAsyncMarkersSGIX.restype = GLvoid
	gl.glDeleteAsyncMarkersSGIX.argtypes = (GLuint, GLsizei,)
	gl.glDeleteBuffers = _getProcAddress('glDeleteBuffers')
	gl.glDeleteBuffers.gl = gl
	gl.glDeleteBuffers.__name__ = 'glDeleteBuffers'
	gl.glDeleteBuffers.errcheck = _errcheck
	gl.glDeleteBuffers.restype = GLvoid
	gl.glDeleteBuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteBuffersARB = _getProcAddress('glDeleteBuffersARB')
	gl.glDeleteBuffersARB.gl = gl
	gl.glDeleteBuffersARB.__name__ = 'glDeleteBuffersARB'
	gl.glDeleteBuffersARB.errcheck = _errcheck
	gl.glDeleteBuffersARB.restype = GLvoid
	gl.glDeleteBuffersARB.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteFencesAPPLE = _getProcAddress('glDeleteFencesAPPLE')
	gl.glDeleteFencesAPPLE.gl = gl
	gl.glDeleteFencesAPPLE.__name__ = 'glDeleteFencesAPPLE'
	gl.glDeleteFencesAPPLE.errcheck = _errcheck
	gl.glDeleteFencesAPPLE.restype = GLvoid
	gl.glDeleteFencesAPPLE.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteFencesNV = _getProcAddress('glDeleteFencesNV')
	gl.glDeleteFencesNV.gl = gl
	gl.glDeleteFencesNV.__name__ = 'glDeleteFencesNV'
	gl.glDeleteFencesNV.errcheck = _errcheck
	gl.glDeleteFencesNV.restype = GLvoid
	gl.glDeleteFencesNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteFragmentShaderATI = _getProcAddress('glDeleteFragmentShaderATI')
	gl.glDeleteFragmentShaderATI.gl = gl
	gl.glDeleteFragmentShaderATI.__name__ = 'glDeleteFragmentShaderATI'
	gl.glDeleteFragmentShaderATI.errcheck = _errcheck
	gl.glDeleteFragmentShaderATI.restype = GLvoid
	gl.glDeleteFragmentShaderATI.argtypes = (GLuint,)
	gl.glDeleteFramebuffers = _getProcAddress('glDeleteFramebuffers')
	gl.glDeleteFramebuffers.gl = gl
	gl.glDeleteFramebuffers.__name__ = 'glDeleteFramebuffers'
	gl.glDeleteFramebuffers.errcheck = _errcheck
	gl.glDeleteFramebuffers.restype = GLvoid
	gl.glDeleteFramebuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteFramebuffersEXT = _getProcAddress('glDeleteFramebuffersEXT')
	gl.glDeleteFramebuffersEXT.gl = gl
	gl.glDeleteFramebuffersEXT.__name__ = 'glDeleteFramebuffersEXT'
	gl.glDeleteFramebuffersEXT.errcheck = _errcheck
	gl.glDeleteFramebuffersEXT.restype = GLvoid
	gl.glDeleteFramebuffersEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteFramebuffersOES = _getProcAddress('glDeleteFramebuffersOES')
	gl.glDeleteFramebuffersOES.gl = gl
	gl.glDeleteFramebuffersOES.__name__ = 'glDeleteFramebuffersOES'
	gl.glDeleteFramebuffersOES.errcheck = _errcheck
	gl.glDeleteFramebuffersOES.restype = GLvoid
	gl.glDeleteFramebuffersOES.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteLists = _getProcAddress('glDeleteLists')
	gl.glDeleteLists.gl = gl
	gl.glDeleteLists.__name__ = 'glDeleteLists'
	gl.glDeleteLists.errcheck = _errcheck
	gl.glDeleteLists.restype = GLvoid
	gl.glDeleteLists.argtypes = (GLuint, GLsizei,)
	gl.glDeleteNamedStringARB = _getProcAddress('glDeleteNamedStringARB')
	gl.glDeleteNamedStringARB.gl = gl
	gl.glDeleteNamedStringARB.__name__ = 'glDeleteNamedStringARB'
	gl.glDeleteNamedStringARB.errcheck = _errcheck
	gl.glDeleteNamedStringARB.restype = GLvoid
	gl.glDeleteNamedStringARB.argtypes = (GLint, POINTER(GLchar),)
	gl.glDeleteNamesAMD = _getProcAddress('glDeleteNamesAMD')
	gl.glDeleteNamesAMD.gl = gl
	gl.glDeleteNamesAMD.__name__ = 'glDeleteNamesAMD'
	gl.glDeleteNamesAMD.errcheck = _errcheck
	gl.glDeleteNamesAMD.restype = GLvoid
	gl.glDeleteNamesAMD.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glDeleteObjectARB = _getProcAddress('glDeleteObjectARB')
	gl.glDeleteObjectARB.gl = gl
	gl.glDeleteObjectARB.__name__ = 'glDeleteObjectARB'
	gl.glDeleteObjectARB.errcheck = _errcheck
	gl.glDeleteObjectARB.restype = GLvoid
	gl.glDeleteObjectARB.argtypes = (GLhandleARB,)
	gl.glDeleteOcclusionQueriesNV = _getProcAddress('glDeleteOcclusionQueriesNV')
	gl.glDeleteOcclusionQueriesNV.gl = gl
	gl.glDeleteOcclusionQueriesNV.__name__ = 'glDeleteOcclusionQueriesNV'
	gl.glDeleteOcclusionQueriesNV.errcheck = _errcheck
	gl.glDeleteOcclusionQueriesNV.restype = GLvoid
	gl.glDeleteOcclusionQueriesNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeletePathsNV = _getProcAddress('glDeletePathsNV')
	gl.glDeletePathsNV.gl = gl
	gl.glDeletePathsNV.__name__ = 'glDeletePathsNV'
	gl.glDeletePathsNV.errcheck = _errcheck
	gl.glDeletePathsNV.restype = GLvoid
	gl.glDeletePathsNV.argtypes = (GLuint, GLsizei,)
	gl.glDeletePerfMonitorsAMD = _getProcAddress('glDeletePerfMonitorsAMD')
	gl.glDeletePerfMonitorsAMD.gl = gl
	gl.glDeletePerfMonitorsAMD.__name__ = 'glDeletePerfMonitorsAMD'
	gl.glDeletePerfMonitorsAMD.errcheck = _errcheck
	gl.glDeletePerfMonitorsAMD.restype = GLvoid
	gl.glDeletePerfMonitorsAMD.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeletePerfQueryINTEL = _getProcAddress('glDeletePerfQueryINTEL')
	gl.glDeletePerfQueryINTEL.gl = gl
	gl.glDeletePerfQueryINTEL.__name__ = 'glDeletePerfQueryINTEL'
	gl.glDeletePerfQueryINTEL.errcheck = _errcheck
	gl.glDeletePerfQueryINTEL.restype = GLvoid
	gl.glDeletePerfQueryINTEL.argtypes = (GLuint,)
	gl.glDeleteProgram = _getProcAddress('glDeleteProgram')
	gl.glDeleteProgram.gl = gl
	gl.glDeleteProgram.__name__ = 'glDeleteProgram'
	gl.glDeleteProgram.errcheck = _errcheck
	gl.glDeleteProgram.restype = GLvoid
	gl.glDeleteProgram.argtypes = (GLuint,)
	gl.glDeleteProgramPipelines = _getProcAddress('glDeleteProgramPipelines')
	gl.glDeleteProgramPipelines.gl = gl
	gl.glDeleteProgramPipelines.__name__ = 'glDeleteProgramPipelines'
	gl.glDeleteProgramPipelines.errcheck = _errcheck
	gl.glDeleteProgramPipelines.restype = GLvoid
	gl.glDeleteProgramPipelines.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteProgramPipelinesEXT = _getProcAddress('glDeleteProgramPipelinesEXT')
	gl.glDeleteProgramPipelinesEXT.gl = gl
	gl.glDeleteProgramPipelinesEXT.__name__ = 'glDeleteProgramPipelinesEXT'
	gl.glDeleteProgramPipelinesEXT.errcheck = _errcheck
	gl.glDeleteProgramPipelinesEXT.restype = GLvoid
	gl.glDeleteProgramPipelinesEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteProgramsARB = _getProcAddress('glDeleteProgramsARB')
	gl.glDeleteProgramsARB.gl = gl
	gl.glDeleteProgramsARB.__name__ = 'glDeleteProgramsARB'
	gl.glDeleteProgramsARB.errcheck = _errcheck
	gl.glDeleteProgramsARB.restype = GLvoid
	gl.glDeleteProgramsARB.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteProgramsNV = _getProcAddress('glDeleteProgramsNV')
	gl.glDeleteProgramsNV.gl = gl
	gl.glDeleteProgramsNV.__name__ = 'glDeleteProgramsNV'
	gl.glDeleteProgramsNV.errcheck = _errcheck
	gl.glDeleteProgramsNV.restype = GLvoid
	gl.glDeleteProgramsNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteQueries = _getProcAddress('glDeleteQueries')
	gl.glDeleteQueries.gl = gl
	gl.glDeleteQueries.__name__ = 'glDeleteQueries'
	gl.glDeleteQueries.errcheck = _errcheck
	gl.glDeleteQueries.restype = GLvoid
	gl.glDeleteQueries.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteQueriesARB = _getProcAddress('glDeleteQueriesARB')
	gl.glDeleteQueriesARB.gl = gl
	gl.glDeleteQueriesARB.__name__ = 'glDeleteQueriesARB'
	gl.glDeleteQueriesARB.errcheck = _errcheck
	gl.glDeleteQueriesARB.restype = GLvoid
	gl.glDeleteQueriesARB.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteQueriesEXT = _getProcAddress('glDeleteQueriesEXT')
	gl.glDeleteQueriesEXT.gl = gl
	gl.glDeleteQueriesEXT.__name__ = 'glDeleteQueriesEXT'
	gl.glDeleteQueriesEXT.errcheck = _errcheck
	gl.glDeleteQueriesEXT.restype = GLvoid
	gl.glDeleteQueriesEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteRenderbuffers = _getProcAddress('glDeleteRenderbuffers')
	gl.glDeleteRenderbuffers.gl = gl
	gl.glDeleteRenderbuffers.__name__ = 'glDeleteRenderbuffers'
	gl.glDeleteRenderbuffers.errcheck = _errcheck
	gl.glDeleteRenderbuffers.restype = GLvoid
	gl.glDeleteRenderbuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteRenderbuffersEXT = _getProcAddress('glDeleteRenderbuffersEXT')
	gl.glDeleteRenderbuffersEXT.gl = gl
	gl.glDeleteRenderbuffersEXT.__name__ = 'glDeleteRenderbuffersEXT'
	gl.glDeleteRenderbuffersEXT.errcheck = _errcheck
	gl.glDeleteRenderbuffersEXT.restype = GLvoid
	gl.glDeleteRenderbuffersEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteRenderbuffersOES = _getProcAddress('glDeleteRenderbuffersOES')
	gl.glDeleteRenderbuffersOES.gl = gl
	gl.glDeleteRenderbuffersOES.__name__ = 'glDeleteRenderbuffersOES'
	gl.glDeleteRenderbuffersOES.errcheck = _errcheck
	gl.glDeleteRenderbuffersOES.restype = GLvoid
	gl.glDeleteRenderbuffersOES.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteSamplers = _getProcAddress('glDeleteSamplers')
	gl.glDeleteSamplers.gl = gl
	gl.glDeleteSamplers.__name__ = 'glDeleteSamplers'
	gl.glDeleteSamplers.errcheck = _errcheck
	gl.glDeleteSamplers.restype = GLvoid
	gl.glDeleteSamplers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteShader = _getProcAddress('glDeleteShader')
	gl.glDeleteShader.gl = gl
	gl.glDeleteShader.__name__ = 'glDeleteShader'
	gl.glDeleteShader.errcheck = _errcheck
	gl.glDeleteShader.restype = GLvoid
	gl.glDeleteShader.argtypes = (GLuint,)
	gl.glDeleteSync = _getProcAddress('glDeleteSync')
	gl.glDeleteSync.gl = gl
	gl.glDeleteSync.__name__ = 'glDeleteSync'
	gl.glDeleteSync.errcheck = _errcheck
	gl.glDeleteSync.restype = GLvoid
	gl.glDeleteSync.argtypes = (GLsync,)
	gl.glDeleteSyncAPPLE = _getProcAddress('glDeleteSyncAPPLE')
	gl.glDeleteSyncAPPLE.gl = gl
	gl.glDeleteSyncAPPLE.__name__ = 'glDeleteSyncAPPLE'
	gl.glDeleteSyncAPPLE.errcheck = _errcheck
	gl.glDeleteSyncAPPLE.restype = GLvoid
	gl.glDeleteSyncAPPLE.argtypes = (GLsync,)
	gl.glDeleteTextures = _getProcAddress('glDeleteTextures')
	gl.glDeleteTextures.gl = gl
	gl.glDeleteTextures.__name__ = 'glDeleteTextures'
	gl.glDeleteTextures.errcheck = _errcheck
	gl.glDeleteTextures.restype = GLvoid
	gl.glDeleteTextures.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteTexturesEXT = _getProcAddress('glDeleteTexturesEXT')
	gl.glDeleteTexturesEXT.gl = gl
	gl.glDeleteTexturesEXT.__name__ = 'glDeleteTexturesEXT'
	gl.glDeleteTexturesEXT.errcheck = _errcheck
	gl.glDeleteTexturesEXT.restype = GLvoid
	gl.glDeleteTexturesEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteTransformFeedbacks = _getProcAddress('glDeleteTransformFeedbacks')
	gl.glDeleteTransformFeedbacks.gl = gl
	gl.glDeleteTransformFeedbacks.__name__ = 'glDeleteTransformFeedbacks'
	gl.glDeleteTransformFeedbacks.errcheck = _errcheck
	gl.glDeleteTransformFeedbacks.restype = GLvoid
	gl.glDeleteTransformFeedbacks.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteTransformFeedbacksNV = _getProcAddress('glDeleteTransformFeedbacksNV')
	gl.glDeleteTransformFeedbacksNV.gl = gl
	gl.glDeleteTransformFeedbacksNV.__name__ = 'glDeleteTransformFeedbacksNV'
	gl.glDeleteTransformFeedbacksNV.errcheck = _errcheck
	gl.glDeleteTransformFeedbacksNV.restype = GLvoid
	gl.glDeleteTransformFeedbacksNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteVertexArrays = _getProcAddress('glDeleteVertexArrays')
	gl.glDeleteVertexArrays.gl = gl
	gl.glDeleteVertexArrays.__name__ = 'glDeleteVertexArrays'
	gl.glDeleteVertexArrays.errcheck = _errcheck
	gl.glDeleteVertexArrays.restype = GLvoid
	gl.glDeleteVertexArrays.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteVertexArraysAPPLE = _getProcAddress('glDeleteVertexArraysAPPLE')
	gl.glDeleteVertexArraysAPPLE.gl = gl
	gl.glDeleteVertexArraysAPPLE.__name__ = 'glDeleteVertexArraysAPPLE'
	gl.glDeleteVertexArraysAPPLE.errcheck = _errcheck
	gl.glDeleteVertexArraysAPPLE.restype = GLvoid
	gl.glDeleteVertexArraysAPPLE.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteVertexArraysOES = _getProcAddress('glDeleteVertexArraysOES')
	gl.glDeleteVertexArraysOES.gl = gl
	gl.glDeleteVertexArraysOES.__name__ = 'glDeleteVertexArraysOES'
	gl.glDeleteVertexArraysOES.errcheck = _errcheck
	gl.glDeleteVertexArraysOES.restype = GLvoid
	gl.glDeleteVertexArraysOES.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glDeleteVertexShaderEXT = _getProcAddress('glDeleteVertexShaderEXT')
	gl.glDeleteVertexShaderEXT.gl = gl
	gl.glDeleteVertexShaderEXT.__name__ = 'glDeleteVertexShaderEXT'
	gl.glDeleteVertexShaderEXT.errcheck = _errcheck
	gl.glDeleteVertexShaderEXT.restype = GLvoid
	gl.glDeleteVertexShaderEXT.argtypes = (GLuint,)
	gl.glDepthBoundsEXT = _getProcAddress('glDepthBoundsEXT')
	gl.glDepthBoundsEXT.gl = gl
	gl.glDepthBoundsEXT.__name__ = 'glDepthBoundsEXT'
	gl.glDepthBoundsEXT.errcheck = _errcheck
	gl.glDepthBoundsEXT.restype = GLvoid
	gl.glDepthBoundsEXT.argtypes = (GLclampd, GLclampd,)
	gl.glDepthBoundsdNV = _getProcAddress('glDepthBoundsdNV')
	gl.glDepthBoundsdNV.gl = gl
	gl.glDepthBoundsdNV.__name__ = 'glDepthBoundsdNV'
	gl.glDepthBoundsdNV.errcheck = _errcheck
	gl.glDepthBoundsdNV.restype = GLvoid
	gl.glDepthBoundsdNV.argtypes = (GLdouble, GLdouble,)
	gl.glDepthFunc = _getProcAddress('glDepthFunc')
	gl.glDepthFunc.gl = gl
	gl.glDepthFunc.__name__ = 'glDepthFunc'
	gl.glDepthFunc.errcheck = _errcheck
	gl.glDepthFunc.restype = GLvoid
	gl.glDepthFunc.argtypes = (GLenum,)
	gl.glDepthMask = _getProcAddress('glDepthMask')
	gl.glDepthMask.gl = gl
	gl.glDepthMask.__name__ = 'glDepthMask'
	gl.glDepthMask.errcheck = _errcheck
	gl.glDepthMask.restype = GLvoid
	gl.glDepthMask.argtypes = (GLboolean,)
	gl.glDepthRange = _getProcAddress('glDepthRange')
	gl.glDepthRange.gl = gl
	gl.glDepthRange.__name__ = 'glDepthRange'
	gl.glDepthRange.errcheck = _errcheck
	gl.glDepthRange.restype = GLvoid
	gl.glDepthRange.argtypes = (GLdouble, GLdouble,)
	gl.glDepthRangeArrayv = _getProcAddress('glDepthRangeArrayv')
	gl.glDepthRangeArrayv.gl = gl
	gl.glDepthRangeArrayv.__name__ = 'glDepthRangeArrayv'
	gl.glDepthRangeArrayv.errcheck = _errcheck
	gl.glDepthRangeArrayv.restype = GLvoid
	gl.glDepthRangeArrayv.argtypes = (GLuint, GLsizei, POINTER(GLdouble),)
	gl.glDepthRangeIndexed = _getProcAddress('glDepthRangeIndexed')
	gl.glDepthRangeIndexed.gl = gl
	gl.glDepthRangeIndexed.__name__ = 'glDepthRangeIndexed'
	gl.glDepthRangeIndexed.errcheck = _errcheck
	gl.glDepthRangeIndexed.restype = GLvoid
	gl.glDepthRangeIndexed.argtypes = (GLuint, GLdouble, GLdouble,)
	gl.glDepthRangedNV = _getProcAddress('glDepthRangedNV')
	gl.glDepthRangedNV.gl = gl
	gl.glDepthRangedNV.__name__ = 'glDepthRangedNV'
	gl.glDepthRangedNV.errcheck = _errcheck
	gl.glDepthRangedNV.restype = GLvoid
	gl.glDepthRangedNV.argtypes = (GLdouble, GLdouble,)
	gl.glDepthRangef = _getProcAddress('glDepthRangef')
	gl.glDepthRangef.gl = gl
	gl.glDepthRangef.__name__ = 'glDepthRangef'
	gl.glDepthRangef.errcheck = _errcheck
	gl.glDepthRangef.restype = GLvoid
	gl.glDepthRangef.argtypes = (GLfloat, GLfloat,)
	gl.glDepthRangefOES = _getProcAddress('glDepthRangefOES')
	gl.glDepthRangefOES.gl = gl
	gl.glDepthRangefOES.__name__ = 'glDepthRangefOES'
	gl.glDepthRangefOES.errcheck = _errcheck
	gl.glDepthRangefOES.restype = GLvoid
	gl.glDepthRangefOES.argtypes = (GLclampf, GLclampf,)
	gl.glDepthRangex = _getProcAddress('glDepthRangex')
	gl.glDepthRangex.gl = gl
	gl.glDepthRangex.__name__ = 'glDepthRangex'
	gl.glDepthRangex.errcheck = _errcheck
	gl.glDepthRangex.restype = GLvoid
	gl.glDepthRangex.argtypes = (GLfixed, GLfixed,)
	gl.glDepthRangexOES = _getProcAddress('glDepthRangexOES')
	gl.glDepthRangexOES.gl = gl
	gl.glDepthRangexOES.__name__ = 'glDepthRangexOES'
	gl.glDepthRangexOES.errcheck = _errcheck
	gl.glDepthRangexOES.restype = GLvoid
	gl.glDepthRangexOES.argtypes = (GLfixed, GLfixed,)
	gl.glDetachObjectARB = _getProcAddress('glDetachObjectARB')
	gl.glDetachObjectARB.gl = gl
	gl.glDetachObjectARB.__name__ = 'glDetachObjectARB'
	gl.glDetachObjectARB.errcheck = _errcheck
	gl.glDetachObjectARB.restype = GLvoid
	gl.glDetachObjectARB.argtypes = (GLhandleARB, GLhandleARB,)
	gl.glDetachShader = _getProcAddress('glDetachShader')
	gl.glDetachShader.gl = gl
	gl.glDetachShader.__name__ = 'glDetachShader'
	gl.glDetachShader.errcheck = _errcheck
	gl.glDetachShader.restype = GLvoid
	gl.glDetachShader.argtypes = (GLuint, GLuint,)
	gl.glDetailTexFuncSGIS = _getProcAddress('glDetailTexFuncSGIS')
	gl.glDetailTexFuncSGIS.gl = gl
	gl.glDetailTexFuncSGIS.__name__ = 'glDetailTexFuncSGIS'
	gl.glDetailTexFuncSGIS.errcheck = _errcheck
	gl.glDetailTexFuncSGIS.restype = GLvoid
	gl.glDetailTexFuncSGIS.argtypes = (GLenum, GLsizei, POINTER(GLfloat),)
	gl.glDisable = _getProcAddress('glDisable')
	gl.glDisable.gl = gl
	gl.glDisable.__name__ = 'glDisable'
	gl.glDisable.errcheck = _errcheck
	gl.glDisable.restype = GLvoid
	gl.glDisable.argtypes = (GLenum,)
	gl.glDisableClientState = _getProcAddress('glDisableClientState')
	gl.glDisableClientState.gl = gl
	gl.glDisableClientState.__name__ = 'glDisableClientState'
	gl.glDisableClientState.errcheck = _errcheck
	gl.glDisableClientState.restype = GLvoid
	gl.glDisableClientState.argtypes = (GLenum,)
	gl.glDisableClientStateIndexedEXT = _getProcAddress('glDisableClientStateIndexedEXT')
	gl.glDisableClientStateIndexedEXT.gl = gl
	gl.glDisableClientStateIndexedEXT.__name__ = 'glDisableClientStateIndexedEXT'
	gl.glDisableClientStateIndexedEXT.errcheck = _errcheck
	gl.glDisableClientStateIndexedEXT.restype = GLvoid
	gl.glDisableClientStateIndexedEXT.argtypes = (GLenum, GLuint,)
	gl.glDisableClientStateiEXT = _getProcAddress('glDisableClientStateiEXT')
	gl.glDisableClientStateiEXT.gl = gl
	gl.glDisableClientStateiEXT.__name__ = 'glDisableClientStateiEXT'
	gl.glDisableClientStateiEXT.errcheck = _errcheck
	gl.glDisableClientStateiEXT.restype = GLvoid
	gl.glDisableClientStateiEXT.argtypes = (GLenum, GLuint,)
	gl.glDisableDriverControlQCOM = _getProcAddress('glDisableDriverControlQCOM')
	gl.glDisableDriverControlQCOM.gl = gl
	gl.glDisableDriverControlQCOM.__name__ = 'glDisableDriverControlQCOM'
	gl.glDisableDriverControlQCOM.errcheck = _errcheck
	gl.glDisableDriverControlQCOM.restype = GLvoid
	gl.glDisableDriverControlQCOM.argtypes = (GLuint,)
	gl.glDisableIndexedEXT = _getProcAddress('glDisableIndexedEXT')
	gl.glDisableIndexedEXT.gl = gl
	gl.glDisableIndexedEXT.__name__ = 'glDisableIndexedEXT'
	gl.glDisableIndexedEXT.errcheck = _errcheck
	gl.glDisableIndexedEXT.restype = GLvoid
	gl.glDisableIndexedEXT.argtypes = (GLenum, GLuint,)
	gl.glDisableVariantClientStateEXT = _getProcAddress('glDisableVariantClientStateEXT')
	gl.glDisableVariantClientStateEXT.gl = gl
	gl.glDisableVariantClientStateEXT.__name__ = 'glDisableVariantClientStateEXT'
	gl.glDisableVariantClientStateEXT.errcheck = _errcheck
	gl.glDisableVariantClientStateEXT.restype = GLvoid
	gl.glDisableVariantClientStateEXT.argtypes = (GLuint,)
	gl.glDisableVertexArrayAttrib = _getProcAddress('glDisableVertexArrayAttrib')
	gl.glDisableVertexArrayAttrib.gl = gl
	gl.glDisableVertexArrayAttrib.__name__ = 'glDisableVertexArrayAttrib'
	gl.glDisableVertexArrayAttrib.errcheck = _errcheck
	gl.glDisableVertexArrayAttrib.restype = GLvoid
	gl.glDisableVertexArrayAttrib.argtypes = (GLuint, GLuint,)
	gl.glDisableVertexArrayAttribEXT = _getProcAddress('glDisableVertexArrayAttribEXT')
	gl.glDisableVertexArrayAttribEXT.gl = gl
	gl.glDisableVertexArrayAttribEXT.__name__ = 'glDisableVertexArrayAttribEXT'
	gl.glDisableVertexArrayAttribEXT.errcheck = _errcheck
	gl.glDisableVertexArrayAttribEXT.restype = GLvoid
	gl.glDisableVertexArrayAttribEXT.argtypes = (GLuint, GLuint,)
	gl.glDisableVertexArrayEXT = _getProcAddress('glDisableVertexArrayEXT')
	gl.glDisableVertexArrayEXT.gl = gl
	gl.glDisableVertexArrayEXT.__name__ = 'glDisableVertexArrayEXT'
	gl.glDisableVertexArrayEXT.errcheck = _errcheck
	gl.glDisableVertexArrayEXT.restype = GLvoid
	gl.glDisableVertexArrayEXT.argtypes = (GLuint, GLenum,)
	gl.glDisableVertexAttribAPPLE = _getProcAddress('glDisableVertexAttribAPPLE')
	gl.glDisableVertexAttribAPPLE.gl = gl
	gl.glDisableVertexAttribAPPLE.__name__ = 'glDisableVertexAttribAPPLE'
	gl.glDisableVertexAttribAPPLE.errcheck = _errcheck
	gl.glDisableVertexAttribAPPLE.restype = GLvoid
	gl.glDisableVertexAttribAPPLE.argtypes = (GLuint, GLenum,)
	gl.glDisableVertexAttribArray = _getProcAddress('glDisableVertexAttribArray')
	gl.glDisableVertexAttribArray.gl = gl
	gl.glDisableVertexAttribArray.__name__ = 'glDisableVertexAttribArray'
	gl.glDisableVertexAttribArray.errcheck = _errcheck
	gl.glDisableVertexAttribArray.restype = GLvoid
	gl.glDisableVertexAttribArray.argtypes = (GLuint,)
	gl.glDisableVertexAttribArrayARB = _getProcAddress('glDisableVertexAttribArrayARB')
	gl.glDisableVertexAttribArrayARB.gl = gl
	gl.glDisableVertexAttribArrayARB.__name__ = 'glDisableVertexAttribArrayARB'
	gl.glDisableVertexAttribArrayARB.errcheck = _errcheck
	gl.glDisableVertexAttribArrayARB.restype = GLvoid
	gl.glDisableVertexAttribArrayARB.argtypes = (GLuint,)
	gl.glDisablei = _getProcAddress('glDisablei')
	gl.glDisablei.gl = gl
	gl.glDisablei.__name__ = 'glDisablei'
	gl.glDisablei.errcheck = _errcheck
	gl.glDisablei.restype = GLvoid
	gl.glDisablei.argtypes = (GLenum, GLuint,)
	gl.glDisableiEXT = _getProcAddress('glDisableiEXT')
	gl.glDisableiEXT.gl = gl
	gl.glDisableiEXT.__name__ = 'glDisableiEXT'
	gl.glDisableiEXT.errcheck = _errcheck
	gl.glDisableiEXT.restype = GLvoid
	gl.glDisableiEXT.argtypes = (GLenum, GLuint,)
	gl.glDiscardFramebufferEXT = _getProcAddress('glDiscardFramebufferEXT')
	gl.glDiscardFramebufferEXT.gl = gl
	gl.glDiscardFramebufferEXT.__name__ = 'glDiscardFramebufferEXT'
	gl.glDiscardFramebufferEXT.errcheck = _errcheck
	gl.glDiscardFramebufferEXT.restype = GLvoid
	gl.glDiscardFramebufferEXT.argtypes = (GLenum, GLsizei, POINTER(GLenum),)
	gl.glDispatchCompute = _getProcAddress('glDispatchCompute')
	gl.glDispatchCompute.gl = gl
	gl.glDispatchCompute.__name__ = 'glDispatchCompute'
	gl.glDispatchCompute.errcheck = _errcheck
	gl.glDispatchCompute.restype = GLvoid
	gl.glDispatchCompute.argtypes = (GLuint, GLuint, GLuint,)
	gl.glDispatchComputeGroupSizeARB = _getProcAddress('glDispatchComputeGroupSizeARB')
	gl.glDispatchComputeGroupSizeARB.gl = gl
	gl.glDispatchComputeGroupSizeARB.__name__ = 'glDispatchComputeGroupSizeARB'
	gl.glDispatchComputeGroupSizeARB.errcheck = _errcheck
	gl.glDispatchComputeGroupSizeARB.restype = GLvoid
	gl.glDispatchComputeGroupSizeARB.argtypes = (GLuint, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glDispatchComputeIndirect = _getProcAddress('glDispatchComputeIndirect')
	gl.glDispatchComputeIndirect.gl = gl
	gl.glDispatchComputeIndirect.__name__ = 'glDispatchComputeIndirect'
	gl.glDispatchComputeIndirect.errcheck = _errcheck
	gl.glDispatchComputeIndirect.restype = GLvoid
	gl.glDispatchComputeIndirect.argtypes = (GLintptr,)
	gl.glDrawArrays = _getProcAddress('glDrawArrays')
	gl.glDrawArrays.gl = gl
	gl.glDrawArrays.__name__ = 'glDrawArrays'
	gl.glDrawArrays.errcheck = _errcheck
	gl.glDrawArrays.restype = GLvoid
	gl.glDrawArrays.argtypes = (GLenum, GLint, GLsizei,)
	gl.glDrawArraysEXT = _getProcAddress('glDrawArraysEXT')
	gl.glDrawArraysEXT.gl = gl
	gl.glDrawArraysEXT.__name__ = 'glDrawArraysEXT'
	gl.glDrawArraysEXT.errcheck = _errcheck
	gl.glDrawArraysEXT.restype = GLvoid
	gl.glDrawArraysEXT.argtypes = (GLenum, GLint, GLsizei,)
	gl.glDrawArraysIndirect = _getProcAddress('glDrawArraysIndirect')
	gl.glDrawArraysIndirect.gl = gl
	gl.glDrawArraysIndirect.__name__ = 'glDrawArraysIndirect'
	gl.glDrawArraysIndirect.errcheck = _errcheck
	gl.glDrawArraysIndirect.restype = GLvoid
	gl.glDrawArraysIndirect.argtypes = (GLenum, POINTER(GLvoid),)
	gl.glDrawArraysInstanced = _getProcAddress('glDrawArraysInstanced')
	gl.glDrawArraysInstanced.gl = gl
	gl.glDrawArraysInstanced.__name__ = 'glDrawArraysInstanced'
	gl.glDrawArraysInstanced.errcheck = _errcheck
	gl.glDrawArraysInstanced.restype = GLvoid
	gl.glDrawArraysInstanced.argtypes = (GLenum, GLint, GLsizei, GLsizei,)
	gl.glDrawArraysInstancedANGLE = _getProcAddress('glDrawArraysInstancedANGLE')
	gl.glDrawArraysInstancedANGLE.gl = gl
	gl.glDrawArraysInstancedANGLE.__name__ = 'glDrawArraysInstancedANGLE'
	gl.glDrawArraysInstancedANGLE.errcheck = _errcheck
	gl.glDrawArraysInstancedANGLE.restype = GLvoid
	gl.glDrawArraysInstancedANGLE.argtypes = (GLenum, GLint, GLsizei, GLsizei,)
	gl.glDrawArraysInstancedARB = _getProcAddress('glDrawArraysInstancedARB')
	gl.glDrawArraysInstancedARB.gl = gl
	gl.glDrawArraysInstancedARB.__name__ = 'glDrawArraysInstancedARB'
	gl.glDrawArraysInstancedARB.errcheck = _errcheck
	gl.glDrawArraysInstancedARB.restype = GLvoid
	gl.glDrawArraysInstancedARB.argtypes = (GLenum, GLint, GLsizei, GLsizei,)
	gl.glDrawArraysInstancedBaseInstance = _getProcAddress('glDrawArraysInstancedBaseInstance')
	gl.glDrawArraysInstancedBaseInstance.gl = gl
	gl.glDrawArraysInstancedBaseInstance.__name__ = 'glDrawArraysInstancedBaseInstance'
	gl.glDrawArraysInstancedBaseInstance.errcheck = _errcheck
	gl.glDrawArraysInstancedBaseInstance.restype = GLvoid
	gl.glDrawArraysInstancedBaseInstance.argtypes = (GLenum, GLint, GLsizei, GLsizei, GLuint,)
	gl.glDrawArraysInstancedEXT = _getProcAddress('glDrawArraysInstancedEXT')
	gl.glDrawArraysInstancedEXT.gl = gl
	gl.glDrawArraysInstancedEXT.__name__ = 'glDrawArraysInstancedEXT'
	gl.glDrawArraysInstancedEXT.errcheck = _errcheck
	gl.glDrawArraysInstancedEXT.restype = GLvoid
	gl.glDrawArraysInstancedEXT.argtypes = (GLenum, GLint, GLsizei, GLsizei,)
	gl.glDrawArraysInstancedNV = _getProcAddress('glDrawArraysInstancedNV')
	gl.glDrawArraysInstancedNV.gl = gl
	gl.glDrawArraysInstancedNV.__name__ = 'glDrawArraysInstancedNV'
	gl.glDrawArraysInstancedNV.errcheck = _errcheck
	gl.glDrawArraysInstancedNV.restype = GLvoid
	gl.glDrawArraysInstancedNV.argtypes = (GLenum, GLint, GLsizei, GLsizei,)
	gl.glDrawBuffer = _getProcAddress('glDrawBuffer')
	gl.glDrawBuffer.gl = gl
	gl.glDrawBuffer.__name__ = 'glDrawBuffer'
	gl.glDrawBuffer.errcheck = _errcheck
	gl.glDrawBuffer.restype = GLvoid
	gl.glDrawBuffer.argtypes = (GLenum,)
	gl.glDrawBuffers = _getProcAddress('glDrawBuffers')
	gl.glDrawBuffers.gl = gl
	gl.glDrawBuffers.__name__ = 'glDrawBuffers'
	gl.glDrawBuffers.errcheck = _errcheck
	gl.glDrawBuffers.restype = GLvoid
	gl.glDrawBuffers.argtypes = (GLsizei, POINTER(GLenum),)
	gl.glDrawBuffersARB = _getProcAddress('glDrawBuffersARB')
	gl.glDrawBuffersARB.gl = gl
	gl.glDrawBuffersARB.__name__ = 'glDrawBuffersARB'
	gl.glDrawBuffersARB.errcheck = _errcheck
	gl.glDrawBuffersARB.restype = GLvoid
	gl.glDrawBuffersARB.argtypes = (GLsizei, POINTER(GLenum),)
	gl.glDrawBuffersATI = _getProcAddress('glDrawBuffersATI')
	gl.glDrawBuffersATI.gl = gl
	gl.glDrawBuffersATI.__name__ = 'glDrawBuffersATI'
	gl.glDrawBuffersATI.errcheck = _errcheck
	gl.glDrawBuffersATI.restype = GLvoid
	gl.glDrawBuffersATI.argtypes = (GLsizei, POINTER(GLenum),)
	gl.glDrawBuffersEXT = _getProcAddress('glDrawBuffersEXT')
	gl.glDrawBuffersEXT.gl = gl
	gl.glDrawBuffersEXT.__name__ = 'glDrawBuffersEXT'
	gl.glDrawBuffersEXT.errcheck = _errcheck
	gl.glDrawBuffersEXT.restype = GLvoid
	gl.glDrawBuffersEXT.argtypes = (GLsizei, POINTER(GLenum),)
	gl.glDrawBuffersIndexedEXT = _getProcAddress('glDrawBuffersIndexedEXT')
	gl.glDrawBuffersIndexedEXT.gl = gl
	gl.glDrawBuffersIndexedEXT.__name__ = 'glDrawBuffersIndexedEXT'
	gl.glDrawBuffersIndexedEXT.errcheck = _errcheck
	gl.glDrawBuffersIndexedEXT.restype = GLvoid
	gl.glDrawBuffersIndexedEXT.argtypes = (GLint, POINTER(GLenum), POINTER(GLint),)
	gl.glDrawBuffersNV = _getProcAddress('glDrawBuffersNV')
	gl.glDrawBuffersNV.gl = gl
	gl.glDrawBuffersNV.__name__ = 'glDrawBuffersNV'
	gl.glDrawBuffersNV.errcheck = _errcheck
	gl.glDrawBuffersNV.restype = GLvoid
	gl.glDrawBuffersNV.argtypes = (GLsizei, POINTER(GLenum),)
	gl.glDrawElementArrayAPPLE = _getProcAddress('glDrawElementArrayAPPLE')
	gl.glDrawElementArrayAPPLE.gl = gl
	gl.glDrawElementArrayAPPLE.__name__ = 'glDrawElementArrayAPPLE'
	gl.glDrawElementArrayAPPLE.errcheck = _errcheck
	gl.glDrawElementArrayAPPLE.restype = GLvoid
	gl.glDrawElementArrayAPPLE.argtypes = (GLenum, GLint, GLsizei,)
	gl.glDrawElementArrayATI = _getProcAddress('glDrawElementArrayATI')
	gl.glDrawElementArrayATI.gl = gl
	gl.glDrawElementArrayATI.__name__ = 'glDrawElementArrayATI'
	gl.glDrawElementArrayATI.errcheck = _errcheck
	gl.glDrawElementArrayATI.restype = GLvoid
	gl.glDrawElementArrayATI.argtypes = (GLenum, GLsizei,)
	gl.glDrawElements = _getProcAddress('glDrawElements')
	gl.glDrawElements.gl = gl
	gl.glDrawElements.__name__ = 'glDrawElements'
	gl.glDrawElements.errcheck = _errcheck
	gl.glDrawElements.restype = GLvoid
	gl.glDrawElements.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid),)
	gl.glDrawElementsBaseVertex = _getProcAddress('glDrawElementsBaseVertex')
	gl.glDrawElementsBaseVertex.gl = gl
	gl.glDrawElementsBaseVertex.__name__ = 'glDrawElementsBaseVertex'
	gl.glDrawElementsBaseVertex.errcheck = _errcheck
	gl.glDrawElementsBaseVertex.restype = GLvoid
	gl.glDrawElementsBaseVertex.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLint,)
	gl.glDrawElementsIndirect = _getProcAddress('glDrawElementsIndirect')
	gl.glDrawElementsIndirect.gl = gl
	gl.glDrawElementsIndirect.__name__ = 'glDrawElementsIndirect'
	gl.glDrawElementsIndirect.errcheck = _errcheck
	gl.glDrawElementsIndirect.restype = GLvoid
	gl.glDrawElementsIndirect.argtypes = (GLenum, GLenum, POINTER(GLvoid),)
	gl.glDrawElementsInstanced = _getProcAddress('glDrawElementsInstanced')
	gl.glDrawElementsInstanced.gl = gl
	gl.glDrawElementsInstanced.__name__ = 'glDrawElementsInstanced'
	gl.glDrawElementsInstanced.errcheck = _errcheck
	gl.glDrawElementsInstanced.restype = GLvoid
	gl.glDrawElementsInstanced.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei,)
	gl.glDrawElementsInstancedANGLE = _getProcAddress('glDrawElementsInstancedANGLE')
	gl.glDrawElementsInstancedANGLE.gl = gl
	gl.glDrawElementsInstancedANGLE.__name__ = 'glDrawElementsInstancedANGLE'
	gl.glDrawElementsInstancedANGLE.errcheck = _errcheck
	gl.glDrawElementsInstancedANGLE.restype = GLvoid
	gl.glDrawElementsInstancedANGLE.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei,)
	gl.glDrawElementsInstancedARB = _getProcAddress('glDrawElementsInstancedARB')
	gl.glDrawElementsInstancedARB.gl = gl
	gl.glDrawElementsInstancedARB.__name__ = 'glDrawElementsInstancedARB'
	gl.glDrawElementsInstancedARB.errcheck = _errcheck
	gl.glDrawElementsInstancedARB.restype = GLvoid
	gl.glDrawElementsInstancedARB.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei,)
	gl.glDrawElementsInstancedBaseInstance = _getProcAddress('glDrawElementsInstancedBaseInstance')
	gl.glDrawElementsInstancedBaseInstance.gl = gl
	gl.glDrawElementsInstancedBaseInstance.__name__ = 'glDrawElementsInstancedBaseInstance'
	gl.glDrawElementsInstancedBaseInstance.errcheck = _errcheck
	gl.glDrawElementsInstancedBaseInstance.restype = GLvoid
	gl.glDrawElementsInstancedBaseInstance.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei, GLuint,)
	gl.glDrawElementsInstancedBaseVertex = _getProcAddress('glDrawElementsInstancedBaseVertex')
	gl.glDrawElementsInstancedBaseVertex.gl = gl
	gl.glDrawElementsInstancedBaseVertex.__name__ = 'glDrawElementsInstancedBaseVertex'
	gl.glDrawElementsInstancedBaseVertex.errcheck = _errcheck
	gl.glDrawElementsInstancedBaseVertex.restype = GLvoid
	gl.glDrawElementsInstancedBaseVertex.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei, GLint,)
	gl.glDrawElementsInstancedBaseVertexBaseInstance = _getProcAddress('glDrawElementsInstancedBaseVertexBaseInstance')
	gl.glDrawElementsInstancedBaseVertexBaseInstance.gl = gl
	gl.glDrawElementsInstancedBaseVertexBaseInstance.__name__ = 'glDrawElementsInstancedBaseVertexBaseInstance'
	gl.glDrawElementsInstancedBaseVertexBaseInstance.errcheck = _errcheck
	gl.glDrawElementsInstancedBaseVertexBaseInstance.restype = GLvoid
	gl.glDrawElementsInstancedBaseVertexBaseInstance.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei, GLint, GLuint,)
	gl.glDrawElementsInstancedEXT = _getProcAddress('glDrawElementsInstancedEXT')
	gl.glDrawElementsInstancedEXT.gl = gl
	gl.glDrawElementsInstancedEXT.__name__ = 'glDrawElementsInstancedEXT'
	gl.glDrawElementsInstancedEXT.errcheck = _errcheck
	gl.glDrawElementsInstancedEXT.restype = GLvoid
	gl.glDrawElementsInstancedEXT.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei,)
	gl.glDrawElementsInstancedNV = _getProcAddress('glDrawElementsInstancedNV')
	gl.glDrawElementsInstancedNV.gl = gl
	gl.glDrawElementsInstancedNV.__name__ = 'glDrawElementsInstancedNV'
	gl.glDrawElementsInstancedNV.errcheck = _errcheck
	gl.glDrawElementsInstancedNV.restype = GLvoid
	gl.glDrawElementsInstancedNV.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLsizei,)
	gl.glDrawMeshArraysSUN = _getProcAddress('glDrawMeshArraysSUN')
	gl.glDrawMeshArraysSUN.gl = gl
	gl.glDrawMeshArraysSUN.__name__ = 'glDrawMeshArraysSUN'
	gl.glDrawMeshArraysSUN.errcheck = _errcheck
	gl.glDrawMeshArraysSUN.restype = GLvoid
	gl.glDrawMeshArraysSUN.argtypes = (GLenum, GLint, GLsizei, GLsizei,)
	gl.glDrawPixels = _getProcAddress('glDrawPixels')
	gl.glDrawPixels.gl = gl
	gl.glDrawPixels.__name__ = 'glDrawPixels'
	gl.glDrawPixels.errcheck = _errcheck
	gl.glDrawPixels.restype = GLvoid
	gl.glDrawPixels.argtypes = (GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glDrawRangeElementArrayAPPLE = _getProcAddress('glDrawRangeElementArrayAPPLE')
	gl.glDrawRangeElementArrayAPPLE.gl = gl
	gl.glDrawRangeElementArrayAPPLE.__name__ = 'glDrawRangeElementArrayAPPLE'
	gl.glDrawRangeElementArrayAPPLE.errcheck = _errcheck
	gl.glDrawRangeElementArrayAPPLE.restype = GLvoid
	gl.glDrawRangeElementArrayAPPLE.argtypes = (GLenum, GLuint, GLuint, GLint, GLsizei,)
	gl.glDrawRangeElementArrayATI = _getProcAddress('glDrawRangeElementArrayATI')
	gl.glDrawRangeElementArrayATI.gl = gl
	gl.glDrawRangeElementArrayATI.__name__ = 'glDrawRangeElementArrayATI'
	gl.glDrawRangeElementArrayATI.errcheck = _errcheck
	gl.glDrawRangeElementArrayATI.restype = GLvoid
	gl.glDrawRangeElementArrayATI.argtypes = (GLenum, GLuint, GLuint, GLsizei,)
	gl.glDrawRangeElements = _getProcAddress('glDrawRangeElements')
	gl.glDrawRangeElements.gl = gl
	gl.glDrawRangeElements.__name__ = 'glDrawRangeElements'
	gl.glDrawRangeElements.errcheck = _errcheck
	gl.glDrawRangeElements.restype = GLvoid
	gl.glDrawRangeElements.argtypes = (GLenum, GLuint, GLuint, GLsizei, GLenum, POINTER(GLvoid),)
	gl.glDrawRangeElementsBaseVertex = _getProcAddress('glDrawRangeElementsBaseVertex')
	gl.glDrawRangeElementsBaseVertex.gl = gl
	gl.glDrawRangeElementsBaseVertex.__name__ = 'glDrawRangeElementsBaseVertex'
	gl.glDrawRangeElementsBaseVertex.errcheck = _errcheck
	gl.glDrawRangeElementsBaseVertex.restype = GLvoid
	gl.glDrawRangeElementsBaseVertex.argtypes = (GLenum, GLuint, GLuint, GLsizei, GLenum, POINTER(GLvoid), GLint,)
	gl.glDrawRangeElementsEXT = _getProcAddress('glDrawRangeElementsEXT')
	gl.glDrawRangeElementsEXT.gl = gl
	gl.glDrawRangeElementsEXT.__name__ = 'glDrawRangeElementsEXT'
	gl.glDrawRangeElementsEXT.errcheck = _errcheck
	gl.glDrawRangeElementsEXT.restype = GLvoid
	gl.glDrawRangeElementsEXT.argtypes = (GLenum, GLuint, GLuint, GLsizei, GLenum, POINTER(GLvoid),)
	gl.glDrawTexfOES = _getProcAddress('glDrawTexfOES')
	gl.glDrawTexfOES.gl = gl
	gl.glDrawTexfOES.__name__ = 'glDrawTexfOES'
	gl.glDrawTexfOES.errcheck = _errcheck
	gl.glDrawTexfOES.restype = GLvoid
	gl.glDrawTexfOES.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glDrawTexfvOES = _getProcAddress('glDrawTexfvOES')
	gl.glDrawTexfvOES.gl = gl
	gl.glDrawTexfvOES.__name__ = 'glDrawTexfvOES'
	gl.glDrawTexfvOES.errcheck = _errcheck
	gl.glDrawTexfvOES.restype = GLvoid
	gl.glDrawTexfvOES.argtypes = (POINTER(GLfloat),)
	gl.glDrawTexiOES = _getProcAddress('glDrawTexiOES')
	gl.glDrawTexiOES.gl = gl
	gl.glDrawTexiOES.__name__ = 'glDrawTexiOES'
	gl.glDrawTexiOES.errcheck = _errcheck
	gl.glDrawTexiOES.restype = GLvoid
	gl.glDrawTexiOES.argtypes = (GLint, GLint, GLint, GLint, GLint,)
	gl.glDrawTexivOES = _getProcAddress('glDrawTexivOES')
	gl.glDrawTexivOES.gl = gl
	gl.glDrawTexivOES.__name__ = 'glDrawTexivOES'
	gl.glDrawTexivOES.errcheck = _errcheck
	gl.glDrawTexivOES.restype = GLvoid
	gl.glDrawTexivOES.argtypes = (POINTER(GLint),)
	gl.glDrawTexsOES = _getProcAddress('glDrawTexsOES')
	gl.glDrawTexsOES.gl = gl
	gl.glDrawTexsOES.__name__ = 'glDrawTexsOES'
	gl.glDrawTexsOES.errcheck = _errcheck
	gl.glDrawTexsOES.restype = GLvoid
	gl.glDrawTexsOES.argtypes = (GLshort, GLshort, GLshort, GLshort, GLshort,)
	gl.glDrawTexsvOES = _getProcAddress('glDrawTexsvOES')
	gl.glDrawTexsvOES.gl = gl
	gl.glDrawTexsvOES.__name__ = 'glDrawTexsvOES'
	gl.glDrawTexsvOES.errcheck = _errcheck
	gl.glDrawTexsvOES.restype = GLvoid
	gl.glDrawTexsvOES.argtypes = (POINTER(GLshort),)
	gl.glDrawTextureNV = _getProcAddress('glDrawTextureNV')
	gl.glDrawTextureNV.gl = gl
	gl.glDrawTextureNV.__name__ = 'glDrawTextureNV'
	gl.glDrawTextureNV.errcheck = _errcheck
	gl.glDrawTextureNV.restype = GLvoid
	gl.glDrawTextureNV.argtypes = (GLuint, GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glDrawTexxOES = _getProcAddress('glDrawTexxOES')
	gl.glDrawTexxOES.gl = gl
	gl.glDrawTexxOES.__name__ = 'glDrawTexxOES'
	gl.glDrawTexxOES.errcheck = _errcheck
	gl.glDrawTexxOES.restype = GLvoid
	gl.glDrawTexxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glDrawTexxvOES = _getProcAddress('glDrawTexxvOES')
	gl.glDrawTexxvOES.gl = gl
	gl.glDrawTexxvOES.__name__ = 'glDrawTexxvOES'
	gl.glDrawTexxvOES.errcheck = _errcheck
	gl.glDrawTexxvOES.restype = GLvoid
	gl.glDrawTexxvOES.argtypes = (POINTER(GLfixed),)
	gl.glDrawTransformFeedback = _getProcAddress('glDrawTransformFeedback')
	gl.glDrawTransformFeedback.gl = gl
	gl.glDrawTransformFeedback.__name__ = 'glDrawTransformFeedback'
	gl.glDrawTransformFeedback.errcheck = _errcheck
	gl.glDrawTransformFeedback.restype = GLvoid
	gl.glDrawTransformFeedback.argtypes = (GLenum, GLuint,)
	gl.glDrawTransformFeedbackInstanced = _getProcAddress('glDrawTransformFeedbackInstanced')
	gl.glDrawTransformFeedbackInstanced.gl = gl
	gl.glDrawTransformFeedbackInstanced.__name__ = 'glDrawTransformFeedbackInstanced'
	gl.glDrawTransformFeedbackInstanced.errcheck = _errcheck
	gl.glDrawTransformFeedbackInstanced.restype = GLvoid
	gl.glDrawTransformFeedbackInstanced.argtypes = (GLenum, GLuint, GLsizei,)
	gl.glDrawTransformFeedbackNV = _getProcAddress('glDrawTransformFeedbackNV')
	gl.glDrawTransformFeedbackNV.gl = gl
	gl.glDrawTransformFeedbackNV.__name__ = 'glDrawTransformFeedbackNV'
	gl.glDrawTransformFeedbackNV.errcheck = _errcheck
	gl.glDrawTransformFeedbackNV.restype = GLvoid
	gl.glDrawTransformFeedbackNV.argtypes = (GLenum, GLuint,)
	gl.glDrawTransformFeedbackStream = _getProcAddress('glDrawTransformFeedbackStream')
	gl.glDrawTransformFeedbackStream.gl = gl
	gl.glDrawTransformFeedbackStream.__name__ = 'glDrawTransformFeedbackStream'
	gl.glDrawTransformFeedbackStream.errcheck = _errcheck
	gl.glDrawTransformFeedbackStream.restype = GLvoid
	gl.glDrawTransformFeedbackStream.argtypes = (GLenum, GLuint, GLuint,)
	gl.glDrawTransformFeedbackStreamInstanced = _getProcAddress('glDrawTransformFeedbackStreamInstanced')
	gl.glDrawTransformFeedbackStreamInstanced.gl = gl
	gl.glDrawTransformFeedbackStreamInstanced.__name__ = 'glDrawTransformFeedbackStreamInstanced'
	gl.glDrawTransformFeedbackStreamInstanced.errcheck = _errcheck
	gl.glDrawTransformFeedbackStreamInstanced.restype = GLvoid
	gl.glDrawTransformFeedbackStreamInstanced.argtypes = (GLenum, GLuint, GLuint, GLsizei,)
	gl.glEGLImageTargetRenderbufferStorageOES = _getProcAddress('glEGLImageTargetRenderbufferStorageOES')
	gl.glEGLImageTargetRenderbufferStorageOES.gl = gl
	gl.glEGLImageTargetRenderbufferStorageOES.__name__ = 'glEGLImageTargetRenderbufferStorageOES'
	gl.glEGLImageTargetRenderbufferStorageOES.errcheck = _errcheck
	gl.glEGLImageTargetRenderbufferStorageOES.restype = GLvoid
	gl.glEGLImageTargetRenderbufferStorageOES.argtypes = (GLenum, GLeglImageOES,)
	gl.glEGLImageTargetTexture2DOES = _getProcAddress('glEGLImageTargetTexture2DOES')
	gl.glEGLImageTargetTexture2DOES.gl = gl
	gl.glEGLImageTargetTexture2DOES.__name__ = 'glEGLImageTargetTexture2DOES'
	gl.glEGLImageTargetTexture2DOES.errcheck = _errcheck
	gl.glEGLImageTargetTexture2DOES.restype = GLvoid
	gl.glEGLImageTargetTexture2DOES.argtypes = (GLenum, GLeglImageOES,)
	gl.glEdgeFlag = _getProcAddress('glEdgeFlag')
	gl.glEdgeFlag.gl = gl
	gl.glEdgeFlag.__name__ = 'glEdgeFlag'
	gl.glEdgeFlag.errcheck = _errcheck
	gl.glEdgeFlag.restype = GLvoid
	gl.glEdgeFlag.argtypes = (GLboolean,)
	gl.glEdgeFlagFormatNV = _getProcAddress('glEdgeFlagFormatNV')
	gl.glEdgeFlagFormatNV.gl = gl
	gl.glEdgeFlagFormatNV.__name__ = 'glEdgeFlagFormatNV'
	gl.glEdgeFlagFormatNV.errcheck = _errcheck
	gl.glEdgeFlagFormatNV.restype = GLvoid
	gl.glEdgeFlagFormatNV.argtypes = (GLsizei,)
	gl.glEdgeFlagPointer = _getProcAddress('glEdgeFlagPointer')
	gl.glEdgeFlagPointer.gl = gl
	gl.glEdgeFlagPointer.__name__ = 'glEdgeFlagPointer'
	gl.glEdgeFlagPointer.errcheck = _errcheck
	gl.glEdgeFlagPointer.restype = GLvoid
	gl.glEdgeFlagPointer.argtypes = (GLsizei, POINTER(GLvoid),)
	gl.glEdgeFlagPointerEXT = _getProcAddress('glEdgeFlagPointerEXT')
	gl.glEdgeFlagPointerEXT.gl = gl
	gl.glEdgeFlagPointerEXT.__name__ = 'glEdgeFlagPointerEXT'
	gl.glEdgeFlagPointerEXT.errcheck = _errcheck
	gl.glEdgeFlagPointerEXT.restype = GLvoid
	gl.glEdgeFlagPointerEXT.argtypes = (GLsizei, GLsizei, POINTER(GLboolean),)
	gl.glEdgeFlagPointerListIBM = _getProcAddress('glEdgeFlagPointerListIBM')
	gl.glEdgeFlagPointerListIBM.gl = gl
	gl.glEdgeFlagPointerListIBM.__name__ = 'glEdgeFlagPointerListIBM'
	gl.glEdgeFlagPointerListIBM.errcheck = _errcheck
	gl.glEdgeFlagPointerListIBM.restype = GLvoid
	gl.glEdgeFlagPointerListIBM.argtypes = (GLint, POINTER(POINTER(GLboolean)), GLint,)
	gl.glEdgeFlagv = _getProcAddress('glEdgeFlagv')
	gl.glEdgeFlagv.gl = gl
	gl.glEdgeFlagv.__name__ = 'glEdgeFlagv'
	gl.glEdgeFlagv.errcheck = _errcheck
	gl.glEdgeFlagv.restype = GLvoid
	gl.glEdgeFlagv.argtypes = (POINTER(GLboolean),)
	gl.glElementPointerAPPLE = _getProcAddress('glElementPointerAPPLE')
	gl.glElementPointerAPPLE.gl = gl
	gl.glElementPointerAPPLE.__name__ = 'glElementPointerAPPLE'
	gl.glElementPointerAPPLE.errcheck = _errcheck
	gl.glElementPointerAPPLE.restype = GLvoid
	gl.glElementPointerAPPLE.argtypes = (GLenum, POINTER(GLvoid),)
	gl.glElementPointerATI = _getProcAddress('glElementPointerATI')
	gl.glElementPointerATI.gl = gl
	gl.glElementPointerATI.__name__ = 'glElementPointerATI'
	gl.glElementPointerATI.errcheck = _errcheck
	gl.glElementPointerATI.restype = GLvoid
	gl.glElementPointerATI.argtypes = (GLenum, POINTER(GLvoid),)
	gl.glEnable = _getProcAddress('glEnable')
	gl.glEnable.gl = gl
	gl.glEnable.__name__ = 'glEnable'
	gl.glEnable.errcheck = _errcheck
	gl.glEnable.restype = GLvoid
	gl.glEnable.argtypes = (GLenum,)
	gl.glEnableClientState = _getProcAddress('glEnableClientState')
	gl.glEnableClientState.gl = gl
	gl.glEnableClientState.__name__ = 'glEnableClientState'
	gl.glEnableClientState.errcheck = _errcheck
	gl.glEnableClientState.restype = GLvoid
	gl.glEnableClientState.argtypes = (GLenum,)
	gl.glEnableClientStateIndexedEXT = _getProcAddress('glEnableClientStateIndexedEXT')
	gl.glEnableClientStateIndexedEXT.gl = gl
	gl.glEnableClientStateIndexedEXT.__name__ = 'glEnableClientStateIndexedEXT'
	gl.glEnableClientStateIndexedEXT.errcheck = _errcheck
	gl.glEnableClientStateIndexedEXT.restype = GLvoid
	gl.glEnableClientStateIndexedEXT.argtypes = (GLenum, GLuint,)
	gl.glEnableClientStateiEXT = _getProcAddress('glEnableClientStateiEXT')
	gl.glEnableClientStateiEXT.gl = gl
	gl.glEnableClientStateiEXT.__name__ = 'glEnableClientStateiEXT'
	gl.glEnableClientStateiEXT.errcheck = _errcheck
	gl.glEnableClientStateiEXT.restype = GLvoid
	gl.glEnableClientStateiEXT.argtypes = (GLenum, GLuint,)
	gl.glEnableDriverControlQCOM = _getProcAddress('glEnableDriverControlQCOM')
	gl.glEnableDriverControlQCOM.gl = gl
	gl.glEnableDriverControlQCOM.__name__ = 'glEnableDriverControlQCOM'
	gl.glEnableDriverControlQCOM.errcheck = _errcheck
	gl.glEnableDriverControlQCOM.restype = GLvoid
	gl.glEnableDriverControlQCOM.argtypes = (GLuint,)
	gl.glEnableIndexedEXT = _getProcAddress('glEnableIndexedEXT')
	gl.glEnableIndexedEXT.gl = gl
	gl.glEnableIndexedEXT.__name__ = 'glEnableIndexedEXT'
	gl.glEnableIndexedEXT.errcheck = _errcheck
	gl.glEnableIndexedEXT.restype = GLvoid
	gl.glEnableIndexedEXT.argtypes = (GLenum, GLuint,)
	gl.glEnableVariantClientStateEXT = _getProcAddress('glEnableVariantClientStateEXT')
	gl.glEnableVariantClientStateEXT.gl = gl
	gl.glEnableVariantClientStateEXT.__name__ = 'glEnableVariantClientStateEXT'
	gl.glEnableVariantClientStateEXT.errcheck = _errcheck
	gl.glEnableVariantClientStateEXT.restype = GLvoid
	gl.glEnableVariantClientStateEXT.argtypes = (GLuint,)
	gl.glEnableVertexArrayAttrib = _getProcAddress('glEnableVertexArrayAttrib')
	gl.glEnableVertexArrayAttrib.gl = gl
	gl.glEnableVertexArrayAttrib.__name__ = 'glEnableVertexArrayAttrib'
	gl.glEnableVertexArrayAttrib.errcheck = _errcheck
	gl.glEnableVertexArrayAttrib.restype = GLvoid
	gl.glEnableVertexArrayAttrib.argtypes = (GLuint, GLuint,)
	gl.glEnableVertexArrayAttribEXT = _getProcAddress('glEnableVertexArrayAttribEXT')
	gl.glEnableVertexArrayAttribEXT.gl = gl
	gl.glEnableVertexArrayAttribEXT.__name__ = 'glEnableVertexArrayAttribEXT'
	gl.glEnableVertexArrayAttribEXT.errcheck = _errcheck
	gl.glEnableVertexArrayAttribEXT.restype = GLvoid
	gl.glEnableVertexArrayAttribEXT.argtypes = (GLuint, GLuint,)
	gl.glEnableVertexArrayEXT = _getProcAddress('glEnableVertexArrayEXT')
	gl.glEnableVertexArrayEXT.gl = gl
	gl.glEnableVertexArrayEXT.__name__ = 'glEnableVertexArrayEXT'
	gl.glEnableVertexArrayEXT.errcheck = _errcheck
	gl.glEnableVertexArrayEXT.restype = GLvoid
	gl.glEnableVertexArrayEXT.argtypes = (GLuint, GLenum,)
	gl.glEnableVertexAttribAPPLE = _getProcAddress('glEnableVertexAttribAPPLE')
	gl.glEnableVertexAttribAPPLE.gl = gl
	gl.glEnableVertexAttribAPPLE.__name__ = 'glEnableVertexAttribAPPLE'
	gl.glEnableVertexAttribAPPLE.errcheck = _errcheck
	gl.glEnableVertexAttribAPPLE.restype = GLvoid
	gl.glEnableVertexAttribAPPLE.argtypes = (GLuint, GLenum,)
	gl.glEnableVertexAttribArray = _getProcAddress('glEnableVertexAttribArray')
	gl.glEnableVertexAttribArray.gl = gl
	gl.glEnableVertexAttribArray.__name__ = 'glEnableVertexAttribArray'
	gl.glEnableVertexAttribArray.errcheck = _errcheck
	gl.glEnableVertexAttribArray.restype = GLvoid
	gl.glEnableVertexAttribArray.argtypes = (GLuint,)
	gl.glEnableVertexAttribArrayARB = _getProcAddress('glEnableVertexAttribArrayARB')
	gl.glEnableVertexAttribArrayARB.gl = gl
	gl.glEnableVertexAttribArrayARB.__name__ = 'glEnableVertexAttribArrayARB'
	gl.glEnableVertexAttribArrayARB.errcheck = _errcheck
	gl.glEnableVertexAttribArrayARB.restype = GLvoid
	gl.glEnableVertexAttribArrayARB.argtypes = (GLuint,)
	gl.glEnablei = _getProcAddress('glEnablei')
	gl.glEnablei.gl = gl
	gl.glEnablei.__name__ = 'glEnablei'
	gl.glEnablei.errcheck = _errcheck
	gl.glEnablei.restype = GLvoid
	gl.glEnablei.argtypes = (GLenum, GLuint,)
	gl.glEnableiEXT = _getProcAddress('glEnableiEXT')
	gl.glEnableiEXT.gl = gl
	gl.glEnableiEXT.__name__ = 'glEnableiEXT'
	gl.glEnableiEXT.errcheck = _errcheck
	gl.glEnableiEXT.restype = GLvoid
	gl.glEnableiEXT.argtypes = (GLenum, GLuint,)
	gl.glEnd = _getProcAddress('glEnd')
	gl.glEnd.gl = gl
	gl.glEnd.__name__ = 'glEnd'
	gl.glEnd.errcheck = _errcheck
	gl.glEnd.restype = GLvoid
	gl.glEnd.argtypes = ()
	gl.glEndConditionalRender = _getProcAddress('glEndConditionalRender')
	gl.glEndConditionalRender.gl = gl
	gl.glEndConditionalRender.__name__ = 'glEndConditionalRender'
	gl.glEndConditionalRender.errcheck = _errcheck
	gl.glEndConditionalRender.restype = GLvoid
	gl.glEndConditionalRender.argtypes = ()
	gl.glEndConditionalRenderNV = _getProcAddress('glEndConditionalRenderNV')
	gl.glEndConditionalRenderNV.gl = gl
	gl.glEndConditionalRenderNV.__name__ = 'glEndConditionalRenderNV'
	gl.glEndConditionalRenderNV.errcheck = _errcheck
	gl.glEndConditionalRenderNV.restype = GLvoid
	gl.glEndConditionalRenderNV.argtypes = ()
	gl.glEndConditionalRenderNVX = _getProcAddress('glEndConditionalRenderNVX')
	gl.glEndConditionalRenderNVX.gl = gl
	gl.glEndConditionalRenderNVX.__name__ = 'glEndConditionalRenderNVX'
	gl.glEndConditionalRenderNVX.errcheck = _errcheck
	gl.glEndConditionalRenderNVX.restype = GLvoid
	gl.glEndConditionalRenderNVX.argtypes = ()
	gl.glEndFragmentShaderATI = _getProcAddress('glEndFragmentShaderATI')
	gl.glEndFragmentShaderATI.gl = gl
	gl.glEndFragmentShaderATI.__name__ = 'glEndFragmentShaderATI'
	gl.glEndFragmentShaderATI.errcheck = _errcheck
	gl.glEndFragmentShaderATI.restype = GLvoid
	gl.glEndFragmentShaderATI.argtypes = ()
	gl.glEndList = _getProcAddress('glEndList')
	gl.glEndList.gl = gl
	gl.glEndList.__name__ = 'glEndList'
	gl.glEndList.errcheck = _errcheck
	gl.glEndList.restype = GLvoid
	gl.glEndList.argtypes = ()
	gl.glEndOcclusionQueryNV = _getProcAddress('glEndOcclusionQueryNV')
	gl.glEndOcclusionQueryNV.gl = gl
	gl.glEndOcclusionQueryNV.__name__ = 'glEndOcclusionQueryNV'
	gl.glEndOcclusionQueryNV.errcheck = _errcheck
	gl.glEndOcclusionQueryNV.restype = GLvoid
	gl.glEndOcclusionQueryNV.argtypes = ()
	gl.glEndPerfMonitorAMD = _getProcAddress('glEndPerfMonitorAMD')
	gl.glEndPerfMonitorAMD.gl = gl
	gl.glEndPerfMonitorAMD.__name__ = 'glEndPerfMonitorAMD'
	gl.glEndPerfMonitorAMD.errcheck = _errcheck
	gl.glEndPerfMonitorAMD.restype = GLvoid
	gl.glEndPerfMonitorAMD.argtypes = (GLuint,)
	gl.glEndPerfQueryINTEL = _getProcAddress('glEndPerfQueryINTEL')
	gl.glEndPerfQueryINTEL.gl = gl
	gl.glEndPerfQueryINTEL.__name__ = 'glEndPerfQueryINTEL'
	gl.glEndPerfQueryINTEL.errcheck = _errcheck
	gl.glEndPerfQueryINTEL.restype = GLvoid
	gl.glEndPerfQueryINTEL.argtypes = (GLuint,)
	gl.glEndQuery = _getProcAddress('glEndQuery')
	gl.glEndQuery.gl = gl
	gl.glEndQuery.__name__ = 'glEndQuery'
	gl.glEndQuery.errcheck = _errcheck
	gl.glEndQuery.restype = GLvoid
	gl.glEndQuery.argtypes = (GLenum,)
	gl.glEndQueryARB = _getProcAddress('glEndQueryARB')
	gl.glEndQueryARB.gl = gl
	gl.glEndQueryARB.__name__ = 'glEndQueryARB'
	gl.glEndQueryARB.errcheck = _errcheck
	gl.glEndQueryARB.restype = GLvoid
	gl.glEndQueryARB.argtypes = (GLenum,)
	gl.glEndQueryEXT = _getProcAddress('glEndQueryEXT')
	gl.glEndQueryEXT.gl = gl
	gl.glEndQueryEXT.__name__ = 'glEndQueryEXT'
	gl.glEndQueryEXT.errcheck = _errcheck
	gl.glEndQueryEXT.restype = GLvoid
	gl.glEndQueryEXT.argtypes = (GLenum,)
	gl.glEndQueryIndexed = _getProcAddress('glEndQueryIndexed')
	gl.glEndQueryIndexed.gl = gl
	gl.glEndQueryIndexed.__name__ = 'glEndQueryIndexed'
	gl.glEndQueryIndexed.errcheck = _errcheck
	gl.glEndQueryIndexed.restype = GLvoid
	gl.glEndQueryIndexed.argtypes = (GLenum, GLuint,)
	gl.glEndTilingQCOM = _getProcAddress('glEndTilingQCOM')
	gl.glEndTilingQCOM.gl = gl
	gl.glEndTilingQCOM.__name__ = 'glEndTilingQCOM'
	gl.glEndTilingQCOM.errcheck = _errcheck
	gl.glEndTilingQCOM.restype = GLvoid
	gl.glEndTilingQCOM.argtypes = (GLbitfield,)
	gl.glEndTransformFeedback = _getProcAddress('glEndTransformFeedback')
	gl.glEndTransformFeedback.gl = gl
	gl.glEndTransformFeedback.__name__ = 'glEndTransformFeedback'
	gl.glEndTransformFeedback.errcheck = _errcheck
	gl.glEndTransformFeedback.restype = GLvoid
	gl.glEndTransformFeedback.argtypes = ()
	gl.glEndTransformFeedbackEXT = _getProcAddress('glEndTransformFeedbackEXT')
	gl.glEndTransformFeedbackEXT.gl = gl
	gl.glEndTransformFeedbackEXT.__name__ = 'glEndTransformFeedbackEXT'
	gl.glEndTransformFeedbackEXT.errcheck = _errcheck
	gl.glEndTransformFeedbackEXT.restype = GLvoid
	gl.glEndTransformFeedbackEXT.argtypes = ()
	gl.glEndTransformFeedbackNV = _getProcAddress('glEndTransformFeedbackNV')
	gl.glEndTransformFeedbackNV.gl = gl
	gl.glEndTransformFeedbackNV.__name__ = 'glEndTransformFeedbackNV'
	gl.glEndTransformFeedbackNV.errcheck = _errcheck
	gl.glEndTransformFeedbackNV.restype = GLvoid
	gl.glEndTransformFeedbackNV.argtypes = ()
	gl.glEndVertexShaderEXT = _getProcAddress('glEndVertexShaderEXT')
	gl.glEndVertexShaderEXT.gl = gl
	gl.glEndVertexShaderEXT.__name__ = 'glEndVertexShaderEXT'
	gl.glEndVertexShaderEXT.errcheck = _errcheck
	gl.glEndVertexShaderEXT.restype = GLvoid
	gl.glEndVertexShaderEXT.argtypes = ()
	gl.glEndVideoCaptureNV = _getProcAddress('glEndVideoCaptureNV')
	gl.glEndVideoCaptureNV.gl = gl
	gl.glEndVideoCaptureNV.__name__ = 'glEndVideoCaptureNV'
	gl.glEndVideoCaptureNV.errcheck = _errcheck
	gl.glEndVideoCaptureNV.restype = GLvoid
	gl.glEndVideoCaptureNV.argtypes = (GLuint,)
	gl.glEvalCoord1d = _getProcAddress('glEvalCoord1d')
	gl.glEvalCoord1d.gl = gl
	gl.glEvalCoord1d.__name__ = 'glEvalCoord1d'
	gl.glEvalCoord1d.errcheck = _errcheck
	gl.glEvalCoord1d.restype = GLvoid
	gl.glEvalCoord1d.argtypes = (GLdouble,)
	gl.glEvalCoord1dv = _getProcAddress('glEvalCoord1dv')
	gl.glEvalCoord1dv.gl = gl
	gl.glEvalCoord1dv.__name__ = 'glEvalCoord1dv'
	gl.glEvalCoord1dv.errcheck = _errcheck
	gl.glEvalCoord1dv.restype = GLvoid
	gl.glEvalCoord1dv.argtypes = (POINTER(GLdouble),)
	gl.glEvalCoord1f = _getProcAddress('glEvalCoord1f')
	gl.glEvalCoord1f.gl = gl
	gl.glEvalCoord1f.__name__ = 'glEvalCoord1f'
	gl.glEvalCoord1f.errcheck = _errcheck
	gl.glEvalCoord1f.restype = GLvoid
	gl.glEvalCoord1f.argtypes = (GLfloat,)
	gl.glEvalCoord1fv = _getProcAddress('glEvalCoord1fv')
	gl.glEvalCoord1fv.gl = gl
	gl.glEvalCoord1fv.__name__ = 'glEvalCoord1fv'
	gl.glEvalCoord1fv.errcheck = _errcheck
	gl.glEvalCoord1fv.restype = GLvoid
	gl.glEvalCoord1fv.argtypes = (POINTER(GLfloat),)
	gl.glEvalCoord1xOES = _getProcAddress('glEvalCoord1xOES')
	gl.glEvalCoord1xOES.gl = gl
	gl.glEvalCoord1xOES.__name__ = 'glEvalCoord1xOES'
	gl.glEvalCoord1xOES.errcheck = _errcheck
	gl.glEvalCoord1xOES.restype = GLvoid
	gl.glEvalCoord1xOES.argtypes = (GLfixed,)
	gl.glEvalCoord1xvOES = _getProcAddress('glEvalCoord1xvOES')
	gl.glEvalCoord1xvOES.gl = gl
	gl.glEvalCoord1xvOES.__name__ = 'glEvalCoord1xvOES'
	gl.glEvalCoord1xvOES.errcheck = _errcheck
	gl.glEvalCoord1xvOES.restype = GLvoid
	gl.glEvalCoord1xvOES.argtypes = (POINTER(GLfixed),)
	gl.glEvalCoord2d = _getProcAddress('glEvalCoord2d')
	gl.glEvalCoord2d.gl = gl
	gl.glEvalCoord2d.__name__ = 'glEvalCoord2d'
	gl.glEvalCoord2d.errcheck = _errcheck
	gl.glEvalCoord2d.restype = GLvoid
	gl.glEvalCoord2d.argtypes = (GLdouble, GLdouble,)
	gl.glEvalCoord2dv = _getProcAddress('glEvalCoord2dv')
	gl.glEvalCoord2dv.gl = gl
	gl.glEvalCoord2dv.__name__ = 'glEvalCoord2dv'
	gl.glEvalCoord2dv.errcheck = _errcheck
	gl.glEvalCoord2dv.restype = GLvoid
	gl.glEvalCoord2dv.argtypes = (POINTER(GLdouble),)
	gl.glEvalCoord2f = _getProcAddress('glEvalCoord2f')
	gl.glEvalCoord2f.gl = gl
	gl.glEvalCoord2f.__name__ = 'glEvalCoord2f'
	gl.glEvalCoord2f.errcheck = _errcheck
	gl.glEvalCoord2f.restype = GLvoid
	gl.glEvalCoord2f.argtypes = (GLfloat, GLfloat,)
	gl.glEvalCoord2fv = _getProcAddress('glEvalCoord2fv')
	gl.glEvalCoord2fv.gl = gl
	gl.glEvalCoord2fv.__name__ = 'glEvalCoord2fv'
	gl.glEvalCoord2fv.errcheck = _errcheck
	gl.glEvalCoord2fv.restype = GLvoid
	gl.glEvalCoord2fv.argtypes = (POINTER(GLfloat),)
	gl.glEvalCoord2xOES = _getProcAddress('glEvalCoord2xOES')
	gl.glEvalCoord2xOES.gl = gl
	gl.glEvalCoord2xOES.__name__ = 'glEvalCoord2xOES'
	gl.glEvalCoord2xOES.errcheck = _errcheck
	gl.glEvalCoord2xOES.restype = GLvoid
	gl.glEvalCoord2xOES.argtypes = (GLfixed, GLfixed,)
	gl.glEvalCoord2xvOES = _getProcAddress('glEvalCoord2xvOES')
	gl.glEvalCoord2xvOES.gl = gl
	gl.glEvalCoord2xvOES.__name__ = 'glEvalCoord2xvOES'
	gl.glEvalCoord2xvOES.errcheck = _errcheck
	gl.glEvalCoord2xvOES.restype = GLvoid
	gl.glEvalCoord2xvOES.argtypes = (POINTER(GLfixed),)
	gl.glEvalMapsNV = _getProcAddress('glEvalMapsNV')
	gl.glEvalMapsNV.gl = gl
	gl.glEvalMapsNV.__name__ = 'glEvalMapsNV'
	gl.glEvalMapsNV.errcheck = _errcheck
	gl.glEvalMapsNV.restype = GLvoid
	gl.glEvalMapsNV.argtypes = (GLenum, GLenum,)
	gl.glEvalMesh1 = _getProcAddress('glEvalMesh1')
	gl.glEvalMesh1.gl = gl
	gl.glEvalMesh1.__name__ = 'glEvalMesh1'
	gl.glEvalMesh1.errcheck = _errcheck
	gl.glEvalMesh1.restype = GLvoid
	gl.glEvalMesh1.argtypes = (GLenum, GLint, GLint,)
	gl.glEvalMesh2 = _getProcAddress('glEvalMesh2')
	gl.glEvalMesh2.gl = gl
	gl.glEvalMesh2.__name__ = 'glEvalMesh2'
	gl.glEvalMesh2.errcheck = _errcheck
	gl.glEvalMesh2.restype = GLvoid
	gl.glEvalMesh2.argtypes = (GLenum, GLint, GLint, GLint, GLint,)
	gl.glEvalPoint1 = _getProcAddress('glEvalPoint1')
	gl.glEvalPoint1.gl = gl
	gl.glEvalPoint1.__name__ = 'glEvalPoint1'
	gl.glEvalPoint1.errcheck = _errcheck
	gl.glEvalPoint1.restype = GLvoid
	gl.glEvalPoint1.argtypes = (GLint,)
	gl.glEvalPoint2 = _getProcAddress('glEvalPoint2')
	gl.glEvalPoint2.gl = gl
	gl.glEvalPoint2.__name__ = 'glEvalPoint2'
	gl.glEvalPoint2.errcheck = _errcheck
	gl.glEvalPoint2.restype = GLvoid
	gl.glEvalPoint2.argtypes = (GLint, GLint,)
	gl.glExecuteProgramNV = _getProcAddress('glExecuteProgramNV')
	gl.glExecuteProgramNV.gl = gl
	gl.glExecuteProgramNV.__name__ = 'glExecuteProgramNV'
	gl.glExecuteProgramNV.errcheck = _errcheck
	gl.glExecuteProgramNV.restype = GLvoid
	gl.glExecuteProgramNV.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glExtGetBufferPointervQCOM = _getProcAddress('glExtGetBufferPointervQCOM')
	gl.glExtGetBufferPointervQCOM.gl = gl
	gl.glExtGetBufferPointervQCOM.__name__ = 'glExtGetBufferPointervQCOM'
	gl.glExtGetBufferPointervQCOM.errcheck = _errcheck
	gl.glExtGetBufferPointervQCOM.restype = GLvoid
	gl.glExtGetBufferPointervQCOM.argtypes = (GLenum, POINTER(POINTER(GLvoid)),)
	gl.glExtGetBuffersQCOM = _getProcAddress('glExtGetBuffersQCOM')
	gl.glExtGetBuffersQCOM.gl = gl
	gl.glExtGetBuffersQCOM.__name__ = 'glExtGetBuffersQCOM'
	gl.glExtGetBuffersQCOM.errcheck = _errcheck
	gl.glExtGetBuffersQCOM.restype = GLvoid
	gl.glExtGetBuffersQCOM.argtypes = (POINTER(GLuint), GLint, POINTER(GLint),)
	gl.glExtGetFramebuffersQCOM = _getProcAddress('glExtGetFramebuffersQCOM')
	gl.glExtGetFramebuffersQCOM.gl = gl
	gl.glExtGetFramebuffersQCOM.__name__ = 'glExtGetFramebuffersQCOM'
	gl.glExtGetFramebuffersQCOM.errcheck = _errcheck
	gl.glExtGetFramebuffersQCOM.restype = GLvoid
	gl.glExtGetFramebuffersQCOM.argtypes = (POINTER(GLuint), GLint, POINTER(GLint),)
	gl.glExtGetProgramBinarySourceQCOM = _getProcAddress('glExtGetProgramBinarySourceQCOM')
	gl.glExtGetProgramBinarySourceQCOM.gl = gl
	gl.glExtGetProgramBinarySourceQCOM.__name__ = 'glExtGetProgramBinarySourceQCOM'
	gl.glExtGetProgramBinarySourceQCOM.errcheck = _errcheck
	gl.glExtGetProgramBinarySourceQCOM.restype = GLvoid
	gl.glExtGetProgramBinarySourceQCOM.argtypes = (GLuint, GLenum, POINTER(GLchar), POINTER(GLint),)
	gl.glExtGetProgramsQCOM = _getProcAddress('glExtGetProgramsQCOM')
	gl.glExtGetProgramsQCOM.gl = gl
	gl.glExtGetProgramsQCOM.__name__ = 'glExtGetProgramsQCOM'
	gl.glExtGetProgramsQCOM.errcheck = _errcheck
	gl.glExtGetProgramsQCOM.restype = GLvoid
	gl.glExtGetProgramsQCOM.argtypes = (POINTER(GLuint), GLint, POINTER(GLint),)
	gl.glExtGetRenderbuffersQCOM = _getProcAddress('glExtGetRenderbuffersQCOM')
	gl.glExtGetRenderbuffersQCOM.gl = gl
	gl.glExtGetRenderbuffersQCOM.__name__ = 'glExtGetRenderbuffersQCOM'
	gl.glExtGetRenderbuffersQCOM.errcheck = _errcheck
	gl.glExtGetRenderbuffersQCOM.restype = GLvoid
	gl.glExtGetRenderbuffersQCOM.argtypes = (POINTER(GLuint), GLint, POINTER(GLint),)
	gl.glExtGetShadersQCOM = _getProcAddress('glExtGetShadersQCOM')
	gl.glExtGetShadersQCOM.gl = gl
	gl.glExtGetShadersQCOM.__name__ = 'glExtGetShadersQCOM'
	gl.glExtGetShadersQCOM.errcheck = _errcheck
	gl.glExtGetShadersQCOM.restype = GLvoid
	gl.glExtGetShadersQCOM.argtypes = (POINTER(GLuint), GLint, POINTER(GLint),)
	gl.glExtGetTexLevelParameterivQCOM = _getProcAddress('glExtGetTexLevelParameterivQCOM')
	gl.glExtGetTexLevelParameterivQCOM.gl = gl
	gl.glExtGetTexLevelParameterivQCOM.__name__ = 'glExtGetTexLevelParameterivQCOM'
	gl.glExtGetTexLevelParameterivQCOM.errcheck = _errcheck
	gl.glExtGetTexLevelParameterivQCOM.restype = GLvoid
	gl.glExtGetTexLevelParameterivQCOM.argtypes = (GLuint, GLenum, GLint, GLenum, POINTER(GLint),)
	gl.glExtGetTexSubImageQCOM = _getProcAddress('glExtGetTexSubImageQCOM')
	gl.glExtGetTexSubImageQCOM.gl = gl
	gl.glExtGetTexSubImageQCOM.__name__ = 'glExtGetTexSubImageQCOM'
	gl.glExtGetTexSubImageQCOM.errcheck = _errcheck
	gl.glExtGetTexSubImageQCOM.restype = GLvoid
	gl.glExtGetTexSubImageQCOM.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glExtGetTexturesQCOM = _getProcAddress('glExtGetTexturesQCOM')
	gl.glExtGetTexturesQCOM.gl = gl
	gl.glExtGetTexturesQCOM.__name__ = 'glExtGetTexturesQCOM'
	gl.glExtGetTexturesQCOM.errcheck = _errcheck
	gl.glExtGetTexturesQCOM.restype = GLvoid
	gl.glExtGetTexturesQCOM.argtypes = (POINTER(GLuint), GLint, POINTER(GLint),)
	gl.glExtIsProgramBinaryQCOM = _getProcAddress('glExtIsProgramBinaryQCOM')
	gl.glExtIsProgramBinaryQCOM.gl = gl
	gl.glExtIsProgramBinaryQCOM.__name__ = 'glExtIsProgramBinaryQCOM'
	gl.glExtIsProgramBinaryQCOM.errcheck = _errcheck
	gl.glExtIsProgramBinaryQCOM.restype = GLboolean
	gl.glExtIsProgramBinaryQCOM.argtypes = (GLuint,)
	gl.glExtTexObjectStateOverrideiQCOM = _getProcAddress('glExtTexObjectStateOverrideiQCOM')
	gl.glExtTexObjectStateOverrideiQCOM.gl = gl
	gl.glExtTexObjectStateOverrideiQCOM.__name__ = 'glExtTexObjectStateOverrideiQCOM'
	gl.glExtTexObjectStateOverrideiQCOM.errcheck = _errcheck
	gl.glExtTexObjectStateOverrideiQCOM.restype = GLvoid
	gl.glExtTexObjectStateOverrideiQCOM.argtypes = (GLenum, GLenum, GLint,)
	gl.glExtractComponentEXT = _getProcAddress('glExtractComponentEXT')
	gl.glExtractComponentEXT.gl = gl
	gl.glExtractComponentEXT.__name__ = 'glExtractComponentEXT'
	gl.glExtractComponentEXT.errcheck = _errcheck
	gl.glExtractComponentEXT.restype = GLvoid
	gl.glExtractComponentEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glFeedbackBuffer = _getProcAddress('glFeedbackBuffer')
	gl.glFeedbackBuffer.gl = gl
	gl.glFeedbackBuffer.__name__ = 'glFeedbackBuffer'
	gl.glFeedbackBuffer.errcheck = _errcheck
	gl.glFeedbackBuffer.restype = GLvoid
	gl.glFeedbackBuffer.argtypes = (GLsizei, GLenum, POINTER(GLfloat),)
	gl.glFeedbackBufferxOES = _getProcAddress('glFeedbackBufferxOES')
	gl.glFeedbackBufferxOES.gl = gl
	gl.glFeedbackBufferxOES.__name__ = 'glFeedbackBufferxOES'
	gl.glFeedbackBufferxOES.errcheck = _errcheck
	gl.glFeedbackBufferxOES.restype = GLvoid
	gl.glFeedbackBufferxOES.argtypes = (GLsizei, GLenum, POINTER(GLfixed),)
	gl.glFenceSync = _getProcAddress('glFenceSync')
	gl.glFenceSync.gl = gl
	gl.glFenceSync.__name__ = 'glFenceSync'
	gl.glFenceSync.errcheck = _errcheck
	gl.glFenceSync.restype = GLsync
	gl.glFenceSync.argtypes = (GLenum, GLbitfield,)
	gl.glFenceSyncAPPLE = _getProcAddress('glFenceSyncAPPLE')
	gl.glFenceSyncAPPLE.gl = gl
	gl.glFenceSyncAPPLE.__name__ = 'glFenceSyncAPPLE'
	gl.glFenceSyncAPPLE.errcheck = _errcheck
	gl.glFenceSyncAPPLE.restype = GLsync
	gl.glFenceSyncAPPLE.argtypes = (GLenum, GLbitfield,)
	gl.glFinalCombinerInputNV = _getProcAddress('glFinalCombinerInputNV')
	gl.glFinalCombinerInputNV.gl = gl
	gl.glFinalCombinerInputNV.__name__ = 'glFinalCombinerInputNV'
	gl.glFinalCombinerInputNV.errcheck = _errcheck
	gl.glFinalCombinerInputNV.restype = GLvoid
	gl.glFinalCombinerInputNV.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glFinish = _getProcAddress('glFinish')
	gl.glFinish.gl = gl
	gl.glFinish.__name__ = 'glFinish'
	gl.glFinish.errcheck = _errcheck
	gl.glFinish.restype = GLvoid
	gl.glFinish.argtypes = ()
	gl.glFinishAsyncSGIX = _getProcAddress('glFinishAsyncSGIX')
	gl.glFinishAsyncSGIX.gl = gl
	gl.glFinishAsyncSGIX.__name__ = 'glFinishAsyncSGIX'
	gl.glFinishAsyncSGIX.errcheck = _errcheck
	gl.glFinishAsyncSGIX.restype = GLint
	gl.glFinishAsyncSGIX.argtypes = (POINTER(GLuint),)
	gl.glFinishFenceAPPLE = _getProcAddress('glFinishFenceAPPLE')
	gl.glFinishFenceAPPLE.gl = gl
	gl.glFinishFenceAPPLE.__name__ = 'glFinishFenceAPPLE'
	gl.glFinishFenceAPPLE.errcheck = _errcheck
	gl.glFinishFenceAPPLE.restype = GLvoid
	gl.glFinishFenceAPPLE.argtypes = (GLuint,)
	gl.glFinishFenceNV = _getProcAddress('glFinishFenceNV')
	gl.glFinishFenceNV.gl = gl
	gl.glFinishFenceNV.__name__ = 'glFinishFenceNV'
	gl.glFinishFenceNV.errcheck = _errcheck
	gl.glFinishFenceNV.restype = GLvoid
	gl.glFinishFenceNV.argtypes = (GLuint,)
	gl.glFinishObjectAPPLE = _getProcAddress('glFinishObjectAPPLE')
	gl.glFinishObjectAPPLE.gl = gl
	gl.glFinishObjectAPPLE.__name__ = 'glFinishObjectAPPLE'
	gl.glFinishObjectAPPLE.errcheck = _errcheck
	gl.glFinishObjectAPPLE.restype = GLvoid
	gl.glFinishObjectAPPLE.argtypes = (GLenum, GLint,)
	gl.glFinishTextureSUNX = _getProcAddress('glFinishTextureSUNX')
	gl.glFinishTextureSUNX.gl = gl
	gl.glFinishTextureSUNX.__name__ = 'glFinishTextureSUNX'
	gl.glFinishTextureSUNX.errcheck = _errcheck
	gl.glFinishTextureSUNX.restype = GLvoid
	gl.glFinishTextureSUNX.argtypes = ()
	gl.glFlush = _getProcAddress('glFlush')
	gl.glFlush.gl = gl
	gl.glFlush.__name__ = 'glFlush'
	gl.glFlush.errcheck = _errcheck
	gl.glFlush.restype = GLvoid
	gl.glFlush.argtypes = ()
	gl.glFlushMappedBufferRange = _getProcAddress('glFlushMappedBufferRange')
	gl.glFlushMappedBufferRange.gl = gl
	gl.glFlushMappedBufferRange.__name__ = 'glFlushMappedBufferRange'
	gl.glFlushMappedBufferRange.errcheck = _errcheck
	gl.glFlushMappedBufferRange.restype = GLvoid
	gl.glFlushMappedBufferRange.argtypes = (GLenum, GLintptr, GLsizeiptr,)
	gl.glFlushMappedBufferRangeAPPLE = _getProcAddress('glFlushMappedBufferRangeAPPLE')
	gl.glFlushMappedBufferRangeAPPLE.gl = gl
	gl.glFlushMappedBufferRangeAPPLE.__name__ = 'glFlushMappedBufferRangeAPPLE'
	gl.glFlushMappedBufferRangeAPPLE.errcheck = _errcheck
	gl.glFlushMappedBufferRangeAPPLE.restype = GLvoid
	gl.glFlushMappedBufferRangeAPPLE.argtypes = (GLenum, GLintptr, GLsizeiptr,)
	gl.glFlushMappedBufferRangeEXT = _getProcAddress('glFlushMappedBufferRangeEXT')
	gl.glFlushMappedBufferRangeEXT.gl = gl
	gl.glFlushMappedBufferRangeEXT.__name__ = 'glFlushMappedBufferRangeEXT'
	gl.glFlushMappedBufferRangeEXT.errcheck = _errcheck
	gl.glFlushMappedBufferRangeEXT.restype = GLvoid
	gl.glFlushMappedBufferRangeEXT.argtypes = (GLenum, GLintptr, GLsizeiptr,)
	gl.glFlushMappedNamedBufferRange = _getProcAddress('glFlushMappedNamedBufferRange')
	gl.glFlushMappedNamedBufferRange.gl = gl
	gl.glFlushMappedNamedBufferRange.__name__ = 'glFlushMappedNamedBufferRange'
	gl.glFlushMappedNamedBufferRange.errcheck = _errcheck
	gl.glFlushMappedNamedBufferRange.restype = GLvoid
	gl.glFlushMappedNamedBufferRange.argtypes = (GLuint, GLintptr, GLsizei,)
	gl.glFlushMappedNamedBufferRangeEXT = _getProcAddress('glFlushMappedNamedBufferRangeEXT')
	gl.glFlushMappedNamedBufferRangeEXT.gl = gl
	gl.glFlushMappedNamedBufferRangeEXT.__name__ = 'glFlushMappedNamedBufferRangeEXT'
	gl.glFlushMappedNamedBufferRangeEXT.errcheck = _errcheck
	gl.glFlushMappedNamedBufferRangeEXT.restype = GLvoid
	gl.glFlushMappedNamedBufferRangeEXT.argtypes = (GLuint, GLintptr, GLsizeiptr,)
	gl.glFlushPixelDataRangeNV = _getProcAddress('glFlushPixelDataRangeNV')
	gl.glFlushPixelDataRangeNV.gl = gl
	gl.glFlushPixelDataRangeNV.__name__ = 'glFlushPixelDataRangeNV'
	gl.glFlushPixelDataRangeNV.errcheck = _errcheck
	gl.glFlushPixelDataRangeNV.restype = GLvoid
	gl.glFlushPixelDataRangeNV.argtypes = (GLenum,)
	gl.glFlushRasterSGIX = _getProcAddress('glFlushRasterSGIX')
	gl.glFlushRasterSGIX.gl = gl
	gl.glFlushRasterSGIX.__name__ = 'glFlushRasterSGIX'
	gl.glFlushRasterSGIX.errcheck = _errcheck
	gl.glFlushRasterSGIX.restype = GLvoid
	gl.glFlushRasterSGIX.argtypes = ()
	gl.glFlushStaticDataIBM = _getProcAddress('glFlushStaticDataIBM')
	gl.glFlushStaticDataIBM.gl = gl
	gl.glFlushStaticDataIBM.__name__ = 'glFlushStaticDataIBM'
	gl.glFlushStaticDataIBM.errcheck = _errcheck
	gl.glFlushStaticDataIBM.restype = GLvoid
	gl.glFlushStaticDataIBM.argtypes = (GLenum,)
	gl.glFlushVertexArrayRangeAPPLE = _getProcAddress('glFlushVertexArrayRangeAPPLE')
	gl.glFlushVertexArrayRangeAPPLE.gl = gl
	gl.glFlushVertexArrayRangeAPPLE.__name__ = 'glFlushVertexArrayRangeAPPLE'
	gl.glFlushVertexArrayRangeAPPLE.errcheck = _errcheck
	gl.glFlushVertexArrayRangeAPPLE.restype = GLvoid
	gl.glFlushVertexArrayRangeAPPLE.argtypes = (GLsizei, POINTER(GLvoid),)
	gl.glFlushVertexArrayRangeNV = _getProcAddress('glFlushVertexArrayRangeNV')
	gl.glFlushVertexArrayRangeNV.gl = gl
	gl.glFlushVertexArrayRangeNV.__name__ = 'glFlushVertexArrayRangeNV'
	gl.glFlushVertexArrayRangeNV.errcheck = _errcheck
	gl.glFlushVertexArrayRangeNV.restype = GLvoid
	gl.glFlushVertexArrayRangeNV.argtypes = ()
	gl.glFogCoordFormatNV = _getProcAddress('glFogCoordFormatNV')
	gl.glFogCoordFormatNV.gl = gl
	gl.glFogCoordFormatNV.__name__ = 'glFogCoordFormatNV'
	gl.glFogCoordFormatNV.errcheck = _errcheck
	gl.glFogCoordFormatNV.restype = GLvoid
	gl.glFogCoordFormatNV.argtypes = (GLenum, GLsizei,)
	gl.glFogCoordPointer = _getProcAddress('glFogCoordPointer')
	gl.glFogCoordPointer.gl = gl
	gl.glFogCoordPointer.__name__ = 'glFogCoordPointer'
	gl.glFogCoordPointer.errcheck = _errcheck
	gl.glFogCoordPointer.restype = GLvoid
	gl.glFogCoordPointer.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glFogCoordPointerEXT = _getProcAddress('glFogCoordPointerEXT')
	gl.glFogCoordPointerEXT.gl = gl
	gl.glFogCoordPointerEXT.__name__ = 'glFogCoordPointerEXT'
	gl.glFogCoordPointerEXT.errcheck = _errcheck
	gl.glFogCoordPointerEXT.restype = GLvoid
	gl.glFogCoordPointerEXT.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glFogCoordPointerListIBM = _getProcAddress('glFogCoordPointerListIBM')
	gl.glFogCoordPointerListIBM.gl = gl
	gl.glFogCoordPointerListIBM.__name__ = 'glFogCoordPointerListIBM'
	gl.glFogCoordPointerListIBM.errcheck = _errcheck
	gl.glFogCoordPointerListIBM.restype = GLvoid
	gl.glFogCoordPointerListIBM.argtypes = (GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glFogCoordd = _getProcAddress('glFogCoordd')
	gl.glFogCoordd.gl = gl
	gl.glFogCoordd.__name__ = 'glFogCoordd'
	gl.glFogCoordd.errcheck = _errcheck
	gl.glFogCoordd.restype = GLvoid
	gl.glFogCoordd.argtypes = (GLdouble,)
	gl.glFogCoorddEXT = _getProcAddress('glFogCoorddEXT')
	gl.glFogCoorddEXT.gl = gl
	gl.glFogCoorddEXT.__name__ = 'glFogCoorddEXT'
	gl.glFogCoorddEXT.errcheck = _errcheck
	gl.glFogCoorddEXT.restype = GLvoid
	gl.glFogCoorddEXT.argtypes = (GLdouble,)
	gl.glFogCoorddv = _getProcAddress('glFogCoorddv')
	gl.glFogCoorddv.gl = gl
	gl.glFogCoorddv.__name__ = 'glFogCoorddv'
	gl.glFogCoorddv.errcheck = _errcheck
	gl.glFogCoorddv.restype = GLvoid
	gl.glFogCoorddv.argtypes = (POINTER(GLdouble),)
	gl.glFogCoorddvEXT = _getProcAddress('glFogCoorddvEXT')
	gl.glFogCoorddvEXT.gl = gl
	gl.glFogCoorddvEXT.__name__ = 'glFogCoorddvEXT'
	gl.glFogCoorddvEXT.errcheck = _errcheck
	gl.glFogCoorddvEXT.restype = GLvoid
	gl.glFogCoorddvEXT.argtypes = (POINTER(GLdouble),)
	gl.glFogCoordf = _getProcAddress('glFogCoordf')
	gl.glFogCoordf.gl = gl
	gl.glFogCoordf.__name__ = 'glFogCoordf'
	gl.glFogCoordf.errcheck = _errcheck
	gl.glFogCoordf.restype = GLvoid
	gl.glFogCoordf.argtypes = (GLfloat,)
	gl.glFogCoordfEXT = _getProcAddress('glFogCoordfEXT')
	gl.glFogCoordfEXT.gl = gl
	gl.glFogCoordfEXT.__name__ = 'glFogCoordfEXT'
	gl.glFogCoordfEXT.errcheck = _errcheck
	gl.glFogCoordfEXT.restype = GLvoid
	gl.glFogCoordfEXT.argtypes = (GLfloat,)
	gl.glFogCoordfv = _getProcAddress('glFogCoordfv')
	gl.glFogCoordfv.gl = gl
	gl.glFogCoordfv.__name__ = 'glFogCoordfv'
	gl.glFogCoordfv.errcheck = _errcheck
	gl.glFogCoordfv.restype = GLvoid
	gl.glFogCoordfv.argtypes = (POINTER(GLfloat),)
	gl.glFogCoordfvEXT = _getProcAddress('glFogCoordfvEXT')
	gl.glFogCoordfvEXT.gl = gl
	gl.glFogCoordfvEXT.__name__ = 'glFogCoordfvEXT'
	gl.glFogCoordfvEXT.errcheck = _errcheck
	gl.glFogCoordfvEXT.restype = GLvoid
	gl.glFogCoordfvEXT.argtypes = (POINTER(GLfloat),)
	gl.glFogCoordhNV = _getProcAddress('glFogCoordhNV')
	gl.glFogCoordhNV.gl = gl
	gl.glFogCoordhNV.__name__ = 'glFogCoordhNV'
	gl.glFogCoordhNV.errcheck = _errcheck
	gl.glFogCoordhNV.restype = GLvoid
	gl.glFogCoordhNV.argtypes = (GLhalfNV,)
	gl.glFogCoordhvNV = _getProcAddress('glFogCoordhvNV')
	gl.glFogCoordhvNV.gl = gl
	gl.glFogCoordhvNV.__name__ = 'glFogCoordhvNV'
	gl.glFogCoordhvNV.errcheck = _errcheck
	gl.glFogCoordhvNV.restype = GLvoid
	gl.glFogCoordhvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glFogFuncSGIS = _getProcAddress('glFogFuncSGIS')
	gl.glFogFuncSGIS.gl = gl
	gl.glFogFuncSGIS.__name__ = 'glFogFuncSGIS'
	gl.glFogFuncSGIS.errcheck = _errcheck
	gl.glFogFuncSGIS.restype = GLvoid
	gl.glFogFuncSGIS.argtypes = (GLsizei, POINTER(GLfloat),)
	gl.glFogf = _getProcAddress('glFogf')
	gl.glFogf.gl = gl
	gl.glFogf.__name__ = 'glFogf'
	gl.glFogf.errcheck = _errcheck
	gl.glFogf.restype = GLvoid
	gl.glFogf.argtypes = (GLenum, GLfloat,)
	gl.glFogfv = _getProcAddress('glFogfv')
	gl.glFogfv.gl = gl
	gl.glFogfv.__name__ = 'glFogfv'
	gl.glFogfv.errcheck = _errcheck
	gl.glFogfv.restype = GLvoid
	gl.glFogfv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glFogi = _getProcAddress('glFogi')
	gl.glFogi.gl = gl
	gl.glFogi.__name__ = 'glFogi'
	gl.glFogi.errcheck = _errcheck
	gl.glFogi.restype = GLvoid
	gl.glFogi.argtypes = (GLenum, GLint,)
	gl.glFogiv = _getProcAddress('glFogiv')
	gl.glFogiv.gl = gl
	gl.glFogiv.__name__ = 'glFogiv'
	gl.glFogiv.errcheck = _errcheck
	gl.glFogiv.restype = GLvoid
	gl.glFogiv.argtypes = (GLenum, POINTER(GLint),)
	gl.glFogx = _getProcAddress('glFogx')
	gl.glFogx.gl = gl
	gl.glFogx.__name__ = 'glFogx'
	gl.glFogx.errcheck = _errcheck
	gl.glFogx.restype = GLvoid
	gl.glFogx.argtypes = (GLenum, GLfixed,)
	gl.glFogxOES = _getProcAddress('glFogxOES')
	gl.glFogxOES.gl = gl
	gl.glFogxOES.__name__ = 'glFogxOES'
	gl.glFogxOES.errcheck = _errcheck
	gl.glFogxOES.restype = GLvoid
	gl.glFogxOES.argtypes = (GLenum, GLfixed,)
	gl.glFogxv = _getProcAddress('glFogxv')
	gl.glFogxv.gl = gl
	gl.glFogxv.__name__ = 'glFogxv'
	gl.glFogxv.errcheck = _errcheck
	gl.glFogxv.restype = GLvoid
	gl.glFogxv.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glFogxvOES = _getProcAddress('glFogxvOES')
	gl.glFogxvOES.gl = gl
	gl.glFogxvOES.__name__ = 'glFogxvOES'
	gl.glFogxvOES.errcheck = _errcheck
	gl.glFogxvOES.restype = GLvoid
	gl.glFogxvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glFragmentColorMaterialSGIX = _getProcAddress('glFragmentColorMaterialSGIX')
	gl.glFragmentColorMaterialSGIX.gl = gl
	gl.glFragmentColorMaterialSGIX.__name__ = 'glFragmentColorMaterialSGIX'
	gl.glFragmentColorMaterialSGIX.errcheck = _errcheck
	gl.glFragmentColorMaterialSGIX.restype = GLvoid
	gl.glFragmentColorMaterialSGIX.argtypes = (GLenum, GLenum,)
	gl.glFragmentLightModelfSGIX = _getProcAddress('glFragmentLightModelfSGIX')
	gl.glFragmentLightModelfSGIX.gl = gl
	gl.glFragmentLightModelfSGIX.__name__ = 'glFragmentLightModelfSGIX'
	gl.glFragmentLightModelfSGIX.errcheck = _errcheck
	gl.glFragmentLightModelfSGIX.restype = GLvoid
	gl.glFragmentLightModelfSGIX.argtypes = (GLenum, GLfloat,)
	gl.glFragmentLightModelfvSGIX = _getProcAddress('glFragmentLightModelfvSGIX')
	gl.glFragmentLightModelfvSGIX.gl = gl
	gl.glFragmentLightModelfvSGIX.__name__ = 'glFragmentLightModelfvSGIX'
	gl.glFragmentLightModelfvSGIX.errcheck = _errcheck
	gl.glFragmentLightModelfvSGIX.restype = GLvoid
	gl.glFragmentLightModelfvSGIX.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glFragmentLightModeliSGIX = _getProcAddress('glFragmentLightModeliSGIX')
	gl.glFragmentLightModeliSGIX.gl = gl
	gl.glFragmentLightModeliSGIX.__name__ = 'glFragmentLightModeliSGIX'
	gl.glFragmentLightModeliSGIX.errcheck = _errcheck
	gl.glFragmentLightModeliSGIX.restype = GLvoid
	gl.glFragmentLightModeliSGIX.argtypes = (GLenum, GLint,)
	gl.glFragmentLightModelivSGIX = _getProcAddress('glFragmentLightModelivSGIX')
	gl.glFragmentLightModelivSGIX.gl = gl
	gl.glFragmentLightModelivSGIX.__name__ = 'glFragmentLightModelivSGIX'
	gl.glFragmentLightModelivSGIX.errcheck = _errcheck
	gl.glFragmentLightModelivSGIX.restype = GLvoid
	gl.glFragmentLightModelivSGIX.argtypes = (GLenum, POINTER(GLint),)
	gl.glFragmentLightfSGIX = _getProcAddress('glFragmentLightfSGIX')
	gl.glFragmentLightfSGIX.gl = gl
	gl.glFragmentLightfSGIX.__name__ = 'glFragmentLightfSGIX'
	gl.glFragmentLightfSGIX.errcheck = _errcheck
	gl.glFragmentLightfSGIX.restype = GLvoid
	gl.glFragmentLightfSGIX.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glFragmentLightfvSGIX = _getProcAddress('glFragmentLightfvSGIX')
	gl.glFragmentLightfvSGIX.gl = gl
	gl.glFragmentLightfvSGIX.__name__ = 'glFragmentLightfvSGIX'
	gl.glFragmentLightfvSGIX.errcheck = _errcheck
	gl.glFragmentLightfvSGIX.restype = GLvoid
	gl.glFragmentLightfvSGIX.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glFragmentLightiSGIX = _getProcAddress('glFragmentLightiSGIX')
	gl.glFragmentLightiSGIX.gl = gl
	gl.glFragmentLightiSGIX.__name__ = 'glFragmentLightiSGIX'
	gl.glFragmentLightiSGIX.errcheck = _errcheck
	gl.glFragmentLightiSGIX.restype = GLvoid
	gl.glFragmentLightiSGIX.argtypes = (GLenum, GLenum, GLint,)
	gl.glFragmentLightivSGIX = _getProcAddress('glFragmentLightivSGIX')
	gl.glFragmentLightivSGIX.gl = gl
	gl.glFragmentLightivSGIX.__name__ = 'glFragmentLightivSGIX'
	gl.glFragmentLightivSGIX.errcheck = _errcheck
	gl.glFragmentLightivSGIX.restype = GLvoid
	gl.glFragmentLightivSGIX.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glFragmentMaterialfSGIX = _getProcAddress('glFragmentMaterialfSGIX')
	gl.glFragmentMaterialfSGIX.gl = gl
	gl.glFragmentMaterialfSGIX.__name__ = 'glFragmentMaterialfSGIX'
	gl.glFragmentMaterialfSGIX.errcheck = _errcheck
	gl.glFragmentMaterialfSGIX.restype = GLvoid
	gl.glFragmentMaterialfSGIX.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glFragmentMaterialfvSGIX = _getProcAddress('glFragmentMaterialfvSGIX')
	gl.glFragmentMaterialfvSGIX.gl = gl
	gl.glFragmentMaterialfvSGIX.__name__ = 'glFragmentMaterialfvSGIX'
	gl.glFragmentMaterialfvSGIX.errcheck = _errcheck
	gl.glFragmentMaterialfvSGIX.restype = GLvoid
	gl.glFragmentMaterialfvSGIX.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glFragmentMaterialiSGIX = _getProcAddress('glFragmentMaterialiSGIX')
	gl.glFragmentMaterialiSGIX.gl = gl
	gl.glFragmentMaterialiSGIX.__name__ = 'glFragmentMaterialiSGIX'
	gl.glFragmentMaterialiSGIX.errcheck = _errcheck
	gl.glFragmentMaterialiSGIX.restype = GLvoid
	gl.glFragmentMaterialiSGIX.argtypes = (GLenum, GLenum, GLint,)
	gl.glFragmentMaterialivSGIX = _getProcAddress('glFragmentMaterialivSGIX')
	gl.glFragmentMaterialivSGIX.gl = gl
	gl.glFragmentMaterialivSGIX.__name__ = 'glFragmentMaterialivSGIX'
	gl.glFragmentMaterialivSGIX.errcheck = _errcheck
	gl.glFragmentMaterialivSGIX.restype = GLvoid
	gl.glFragmentMaterialivSGIX.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glFrameTerminatorGREMEDY = _getProcAddress('glFrameTerminatorGREMEDY')
	gl.glFrameTerminatorGREMEDY.gl = gl
	gl.glFrameTerminatorGREMEDY.__name__ = 'glFrameTerminatorGREMEDY'
	gl.glFrameTerminatorGREMEDY.errcheck = _errcheck
	gl.glFrameTerminatorGREMEDY.restype = GLvoid
	gl.glFrameTerminatorGREMEDY.argtypes = ()
	gl.glFrameZoomSGIX = _getProcAddress('glFrameZoomSGIX')
	gl.glFrameZoomSGIX.gl = gl
	gl.glFrameZoomSGIX.__name__ = 'glFrameZoomSGIX'
	gl.glFrameZoomSGIX.errcheck = _errcheck
	gl.glFrameZoomSGIX.restype = GLvoid
	gl.glFrameZoomSGIX.argtypes = (GLint,)
	gl.glFramebufferDrawBufferEXT = _getProcAddress('glFramebufferDrawBufferEXT')
	gl.glFramebufferDrawBufferEXT.gl = gl
	gl.glFramebufferDrawBufferEXT.__name__ = 'glFramebufferDrawBufferEXT'
	gl.glFramebufferDrawBufferEXT.errcheck = _errcheck
	gl.glFramebufferDrawBufferEXT.restype = GLvoid
	gl.glFramebufferDrawBufferEXT.argtypes = (GLuint, GLenum,)
	gl.glFramebufferDrawBuffersEXT = _getProcAddress('glFramebufferDrawBuffersEXT')
	gl.glFramebufferDrawBuffersEXT.gl = gl
	gl.glFramebufferDrawBuffersEXT.__name__ = 'glFramebufferDrawBuffersEXT'
	gl.glFramebufferDrawBuffersEXT.errcheck = _errcheck
	gl.glFramebufferDrawBuffersEXT.restype = GLvoid
	gl.glFramebufferDrawBuffersEXT.argtypes = (GLuint, GLsizei, POINTER(GLenum),)
	gl.glFramebufferParameteri = _getProcAddress('glFramebufferParameteri')
	gl.glFramebufferParameteri.gl = gl
	gl.glFramebufferParameteri.__name__ = 'glFramebufferParameteri'
	gl.glFramebufferParameteri.errcheck = _errcheck
	gl.glFramebufferParameteri.restype = GLvoid
	gl.glFramebufferParameteri.argtypes = (GLenum, GLenum, GLint,)
	gl.glFramebufferReadBufferEXT = _getProcAddress('glFramebufferReadBufferEXT')
	gl.glFramebufferReadBufferEXT.gl = gl
	gl.glFramebufferReadBufferEXT.__name__ = 'glFramebufferReadBufferEXT'
	gl.glFramebufferReadBufferEXT.errcheck = _errcheck
	gl.glFramebufferReadBufferEXT.restype = GLvoid
	gl.glFramebufferReadBufferEXT.argtypes = (GLuint, GLenum,)
	gl.glFramebufferRenderbuffer = _getProcAddress('glFramebufferRenderbuffer')
	gl.glFramebufferRenderbuffer.gl = gl
	gl.glFramebufferRenderbuffer.__name__ = 'glFramebufferRenderbuffer'
	gl.glFramebufferRenderbuffer.errcheck = _errcheck
	gl.glFramebufferRenderbuffer.restype = GLvoid
	gl.glFramebufferRenderbuffer.argtypes = (GLenum, GLenum, GLenum, GLuint,)
	gl.glFramebufferRenderbufferEXT = _getProcAddress('glFramebufferRenderbufferEXT')
	gl.glFramebufferRenderbufferEXT.gl = gl
	gl.glFramebufferRenderbufferEXT.__name__ = 'glFramebufferRenderbufferEXT'
	gl.glFramebufferRenderbufferEXT.errcheck = _errcheck
	gl.glFramebufferRenderbufferEXT.restype = GLvoid
	gl.glFramebufferRenderbufferEXT.argtypes = (GLenum, GLenum, GLenum, GLuint,)
	gl.glFramebufferRenderbufferOES = _getProcAddress('glFramebufferRenderbufferOES')
	gl.glFramebufferRenderbufferOES.gl = gl
	gl.glFramebufferRenderbufferOES.__name__ = 'glFramebufferRenderbufferOES'
	gl.glFramebufferRenderbufferOES.errcheck = _errcheck
	gl.glFramebufferRenderbufferOES.restype = GLvoid
	gl.glFramebufferRenderbufferOES.argtypes = (GLenum, GLenum, GLenum, GLuint,)
	gl.glFramebufferTexture = _getProcAddress('glFramebufferTexture')
	gl.glFramebufferTexture.gl = gl
	gl.glFramebufferTexture.__name__ = 'glFramebufferTexture'
	gl.glFramebufferTexture.errcheck = _errcheck
	gl.glFramebufferTexture.restype = GLvoid
	gl.glFramebufferTexture.argtypes = (GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTexture1D = _getProcAddress('glFramebufferTexture1D')
	gl.glFramebufferTexture1D.gl = gl
	gl.glFramebufferTexture1D.__name__ = 'glFramebufferTexture1D'
	gl.glFramebufferTexture1D.errcheck = _errcheck
	gl.glFramebufferTexture1D.restype = GLvoid
	gl.glFramebufferTexture1D.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTexture1DEXT = _getProcAddress('glFramebufferTexture1DEXT')
	gl.glFramebufferTexture1DEXT.gl = gl
	gl.glFramebufferTexture1DEXT.__name__ = 'glFramebufferTexture1DEXT'
	gl.glFramebufferTexture1DEXT.errcheck = _errcheck
	gl.glFramebufferTexture1DEXT.restype = GLvoid
	gl.glFramebufferTexture1DEXT.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTexture2D = _getProcAddress('glFramebufferTexture2D')
	gl.glFramebufferTexture2D.gl = gl
	gl.glFramebufferTexture2D.__name__ = 'glFramebufferTexture2D'
	gl.glFramebufferTexture2D.errcheck = _errcheck
	gl.glFramebufferTexture2D.restype = GLvoid
	gl.glFramebufferTexture2D.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTexture2DEXT = _getProcAddress('glFramebufferTexture2DEXT')
	gl.glFramebufferTexture2DEXT.gl = gl
	gl.glFramebufferTexture2DEXT.__name__ = 'glFramebufferTexture2DEXT'
	gl.glFramebufferTexture2DEXT.errcheck = _errcheck
	gl.glFramebufferTexture2DEXT.restype = GLvoid
	gl.glFramebufferTexture2DEXT.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTexture2DMultisampleEXT = _getProcAddress('glFramebufferTexture2DMultisampleEXT')
	gl.glFramebufferTexture2DMultisampleEXT.gl = gl
	gl.glFramebufferTexture2DMultisampleEXT.__name__ = 'glFramebufferTexture2DMultisampleEXT'
	gl.glFramebufferTexture2DMultisampleEXT.errcheck = _errcheck
	gl.glFramebufferTexture2DMultisampleEXT.restype = GLvoid
	gl.glFramebufferTexture2DMultisampleEXT.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint, GLsizei,)
	gl.glFramebufferTexture2DMultisampleIMG = _getProcAddress('glFramebufferTexture2DMultisampleIMG')
	gl.glFramebufferTexture2DMultisampleIMG.gl = gl
	gl.glFramebufferTexture2DMultisampleIMG.__name__ = 'glFramebufferTexture2DMultisampleIMG'
	gl.glFramebufferTexture2DMultisampleIMG.errcheck = _errcheck
	gl.glFramebufferTexture2DMultisampleIMG.restype = GLvoid
	gl.glFramebufferTexture2DMultisampleIMG.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint, GLsizei,)
	gl.glFramebufferTexture2DOES = _getProcAddress('glFramebufferTexture2DOES')
	gl.glFramebufferTexture2DOES.gl = gl
	gl.glFramebufferTexture2DOES.__name__ = 'glFramebufferTexture2DOES'
	gl.glFramebufferTexture2DOES.errcheck = _errcheck
	gl.glFramebufferTexture2DOES.restype = GLvoid
	gl.glFramebufferTexture2DOES.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTexture3D = _getProcAddress('glFramebufferTexture3D')
	gl.glFramebufferTexture3D.gl = gl
	gl.glFramebufferTexture3D.__name__ = 'glFramebufferTexture3D'
	gl.glFramebufferTexture3D.errcheck = _errcheck
	gl.glFramebufferTexture3D.restype = GLvoid
	gl.glFramebufferTexture3D.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glFramebufferTexture3DEXT = _getProcAddress('glFramebufferTexture3DEXT')
	gl.glFramebufferTexture3DEXT.gl = gl
	gl.glFramebufferTexture3DEXT.__name__ = 'glFramebufferTexture3DEXT'
	gl.glFramebufferTexture3DEXT.errcheck = _errcheck
	gl.glFramebufferTexture3DEXT.restype = GLvoid
	gl.glFramebufferTexture3DEXT.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glFramebufferTexture3DOES = _getProcAddress('glFramebufferTexture3DOES')
	gl.glFramebufferTexture3DOES.gl = gl
	gl.glFramebufferTexture3DOES.__name__ = 'glFramebufferTexture3DOES'
	gl.glFramebufferTexture3DOES.errcheck = _errcheck
	gl.glFramebufferTexture3DOES.restype = GLvoid
	gl.glFramebufferTexture3DOES.argtypes = (GLenum, GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glFramebufferTextureARB = _getProcAddress('glFramebufferTextureARB')
	gl.glFramebufferTextureARB.gl = gl
	gl.glFramebufferTextureARB.__name__ = 'glFramebufferTextureARB'
	gl.glFramebufferTextureARB.errcheck = _errcheck
	gl.glFramebufferTextureARB.restype = GLvoid
	gl.glFramebufferTextureARB.argtypes = (GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTextureEXT = _getProcAddress('glFramebufferTextureEXT')
	gl.glFramebufferTextureEXT.gl = gl
	gl.glFramebufferTextureEXT.__name__ = 'glFramebufferTextureEXT'
	gl.glFramebufferTextureEXT.errcheck = _errcheck
	gl.glFramebufferTextureEXT.restype = GLvoid
	gl.glFramebufferTextureEXT.argtypes = (GLenum, GLenum, GLuint, GLint,)
	gl.glFramebufferTextureFaceARB = _getProcAddress('glFramebufferTextureFaceARB')
	gl.glFramebufferTextureFaceARB.gl = gl
	gl.glFramebufferTextureFaceARB.__name__ = 'glFramebufferTextureFaceARB'
	gl.glFramebufferTextureFaceARB.errcheck = _errcheck
	gl.glFramebufferTextureFaceARB.restype = GLvoid
	gl.glFramebufferTextureFaceARB.argtypes = (GLenum, GLenum, GLuint, GLint, GLenum,)
	gl.glFramebufferTextureFaceEXT = _getProcAddress('glFramebufferTextureFaceEXT')
	gl.glFramebufferTextureFaceEXT.gl = gl
	gl.glFramebufferTextureFaceEXT.__name__ = 'glFramebufferTextureFaceEXT'
	gl.glFramebufferTextureFaceEXT.errcheck = _errcheck
	gl.glFramebufferTextureFaceEXT.restype = GLvoid
	gl.glFramebufferTextureFaceEXT.argtypes = (GLenum, GLenum, GLuint, GLint, GLenum,)
	gl.glFramebufferTextureLayer = _getProcAddress('glFramebufferTextureLayer')
	gl.glFramebufferTextureLayer.gl = gl
	gl.glFramebufferTextureLayer.__name__ = 'glFramebufferTextureLayer'
	gl.glFramebufferTextureLayer.errcheck = _errcheck
	gl.glFramebufferTextureLayer.restype = GLvoid
	gl.glFramebufferTextureLayer.argtypes = (GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glFramebufferTextureLayerARB = _getProcAddress('glFramebufferTextureLayerARB')
	gl.glFramebufferTextureLayerARB.gl = gl
	gl.glFramebufferTextureLayerARB.__name__ = 'glFramebufferTextureLayerARB'
	gl.glFramebufferTextureLayerARB.errcheck = _errcheck
	gl.glFramebufferTextureLayerARB.restype = GLvoid
	gl.glFramebufferTextureLayerARB.argtypes = (GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glFramebufferTextureLayerEXT = _getProcAddress('glFramebufferTextureLayerEXT')
	gl.glFramebufferTextureLayerEXT.gl = gl
	gl.glFramebufferTextureLayerEXT.__name__ = 'glFramebufferTextureLayerEXT'
	gl.glFramebufferTextureLayerEXT.errcheck = _errcheck
	gl.glFramebufferTextureLayerEXT.restype = GLvoid
	gl.glFramebufferTextureLayerEXT.argtypes = (GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glFreeObjectBufferATI = _getProcAddress('glFreeObjectBufferATI')
	gl.glFreeObjectBufferATI.gl = gl
	gl.glFreeObjectBufferATI.__name__ = 'glFreeObjectBufferATI'
	gl.glFreeObjectBufferATI.errcheck = _errcheck
	gl.glFreeObjectBufferATI.restype = GLvoid
	gl.glFreeObjectBufferATI.argtypes = (GLuint,)
	gl.glFrontFace = _getProcAddress('glFrontFace')
	gl.glFrontFace.gl = gl
	gl.glFrontFace.__name__ = 'glFrontFace'
	gl.glFrontFace.errcheck = _errcheck
	gl.glFrontFace.restype = GLvoid
	gl.glFrontFace.argtypes = (GLenum,)
	gl.glFrustum = _getProcAddress('glFrustum')
	gl.glFrustum.gl = gl
	gl.glFrustum.__name__ = 'glFrustum'
	gl.glFrustum.errcheck = _errcheck
	gl.glFrustum.restype = GLvoid
	gl.glFrustum.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glFrustumf = _getProcAddress('glFrustumf')
	gl.glFrustumf.gl = gl
	gl.glFrustumf.__name__ = 'glFrustumf'
	gl.glFrustumf.errcheck = _errcheck
	gl.glFrustumf.restype = GLvoid
	gl.glFrustumf.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glFrustumfOES = _getProcAddress('glFrustumfOES')
	gl.glFrustumfOES.gl = gl
	gl.glFrustumfOES.__name__ = 'glFrustumfOES'
	gl.glFrustumfOES.errcheck = _errcheck
	gl.glFrustumfOES.restype = GLvoid
	gl.glFrustumfOES.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glFrustumx = _getProcAddress('glFrustumx')
	gl.glFrustumx.gl = gl
	gl.glFrustumx.__name__ = 'glFrustumx'
	gl.glFrustumx.errcheck = _errcheck
	gl.glFrustumx.restype = GLvoid
	gl.glFrustumx.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glFrustumxOES = _getProcAddress('glFrustumxOES')
	gl.glFrustumxOES.gl = gl
	gl.glFrustumxOES.__name__ = 'glFrustumxOES'
	gl.glFrustumxOES.errcheck = _errcheck
	gl.glFrustumxOES.restype = GLvoid
	gl.glFrustumxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glGenAsyncMarkersSGIX = _getProcAddress('glGenAsyncMarkersSGIX')
	gl.glGenAsyncMarkersSGIX.gl = gl
	gl.glGenAsyncMarkersSGIX.__name__ = 'glGenAsyncMarkersSGIX'
	gl.glGenAsyncMarkersSGIX.errcheck = _errcheck
	gl.glGenAsyncMarkersSGIX.restype = GLuint
	gl.glGenAsyncMarkersSGIX.argtypes = (GLsizei,)
	gl.glGenBuffers = _getProcAddress('glGenBuffers')
	gl.glGenBuffers.gl = gl
	gl.glGenBuffers.__name__ = 'glGenBuffers'
	gl.glGenBuffers.errcheck = _errcheck
	gl.glGenBuffers.restype = GLvoid
	gl.glGenBuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenBuffersARB = _getProcAddress('glGenBuffersARB')
	gl.glGenBuffersARB.gl = gl
	gl.glGenBuffersARB.__name__ = 'glGenBuffersARB'
	gl.glGenBuffersARB.errcheck = _errcheck
	gl.glGenBuffersARB.restype = GLvoid
	gl.glGenBuffersARB.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenFencesAPPLE = _getProcAddress('glGenFencesAPPLE')
	gl.glGenFencesAPPLE.gl = gl
	gl.glGenFencesAPPLE.__name__ = 'glGenFencesAPPLE'
	gl.glGenFencesAPPLE.errcheck = _errcheck
	gl.glGenFencesAPPLE.restype = GLvoid
	gl.glGenFencesAPPLE.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenFencesNV = _getProcAddress('glGenFencesNV')
	gl.glGenFencesNV.gl = gl
	gl.glGenFencesNV.__name__ = 'glGenFencesNV'
	gl.glGenFencesNV.errcheck = _errcheck
	gl.glGenFencesNV.restype = GLvoid
	gl.glGenFencesNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenFragmentShadersATI = _getProcAddress('glGenFragmentShadersATI')
	gl.glGenFragmentShadersATI.gl = gl
	gl.glGenFragmentShadersATI.__name__ = 'glGenFragmentShadersATI'
	gl.glGenFragmentShadersATI.errcheck = _errcheck
	gl.glGenFragmentShadersATI.restype = GLuint
	gl.glGenFragmentShadersATI.argtypes = (GLuint,)
	gl.glGenFramebuffers = _getProcAddress('glGenFramebuffers')
	gl.glGenFramebuffers.gl = gl
	gl.glGenFramebuffers.__name__ = 'glGenFramebuffers'
	gl.glGenFramebuffers.errcheck = _errcheck
	gl.glGenFramebuffers.restype = GLvoid
	gl.glGenFramebuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenFramebuffersEXT = _getProcAddress('glGenFramebuffersEXT')
	gl.glGenFramebuffersEXT.gl = gl
	gl.glGenFramebuffersEXT.__name__ = 'glGenFramebuffersEXT'
	gl.glGenFramebuffersEXT.errcheck = _errcheck
	gl.glGenFramebuffersEXT.restype = GLvoid
	gl.glGenFramebuffersEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenFramebuffersOES = _getProcAddress('glGenFramebuffersOES')
	gl.glGenFramebuffersOES.gl = gl
	gl.glGenFramebuffersOES.__name__ = 'glGenFramebuffersOES'
	gl.glGenFramebuffersOES.errcheck = _errcheck
	gl.glGenFramebuffersOES.restype = GLvoid
	gl.glGenFramebuffersOES.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenLists = _getProcAddress('glGenLists')
	gl.glGenLists.gl = gl
	gl.glGenLists.__name__ = 'glGenLists'
	gl.glGenLists.errcheck = _errcheck
	gl.glGenLists.restype = GLuint
	gl.glGenLists.argtypes = (GLsizei,)
	gl.glGenNamesAMD = _getProcAddress('glGenNamesAMD')
	gl.glGenNamesAMD.gl = gl
	gl.glGenNamesAMD.__name__ = 'glGenNamesAMD'
	gl.glGenNamesAMD.errcheck = _errcheck
	gl.glGenNamesAMD.restype = GLvoid
	gl.glGenNamesAMD.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glGenOcclusionQueriesNV = _getProcAddress('glGenOcclusionQueriesNV')
	gl.glGenOcclusionQueriesNV.gl = gl
	gl.glGenOcclusionQueriesNV.__name__ = 'glGenOcclusionQueriesNV'
	gl.glGenOcclusionQueriesNV.errcheck = _errcheck
	gl.glGenOcclusionQueriesNV.restype = GLvoid
	gl.glGenOcclusionQueriesNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenPathsNV = _getProcAddress('glGenPathsNV')
	gl.glGenPathsNV.gl = gl
	gl.glGenPathsNV.__name__ = 'glGenPathsNV'
	gl.glGenPathsNV.errcheck = _errcheck
	gl.glGenPathsNV.restype = GLuint
	gl.glGenPathsNV.argtypes = (GLsizei,)
	gl.glGenPerfMonitorsAMD = _getProcAddress('glGenPerfMonitorsAMD')
	gl.glGenPerfMonitorsAMD.gl = gl
	gl.glGenPerfMonitorsAMD.__name__ = 'glGenPerfMonitorsAMD'
	gl.glGenPerfMonitorsAMD.errcheck = _errcheck
	gl.glGenPerfMonitorsAMD.restype = GLvoid
	gl.glGenPerfMonitorsAMD.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenProgramPipelines = _getProcAddress('glGenProgramPipelines')
	gl.glGenProgramPipelines.gl = gl
	gl.glGenProgramPipelines.__name__ = 'glGenProgramPipelines'
	gl.glGenProgramPipelines.errcheck = _errcheck
	gl.glGenProgramPipelines.restype = GLvoid
	gl.glGenProgramPipelines.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenProgramPipelinesEXT = _getProcAddress('glGenProgramPipelinesEXT')
	gl.glGenProgramPipelinesEXT.gl = gl
	gl.glGenProgramPipelinesEXT.__name__ = 'glGenProgramPipelinesEXT'
	gl.glGenProgramPipelinesEXT.errcheck = _errcheck
	gl.glGenProgramPipelinesEXT.restype = GLvoid
	gl.glGenProgramPipelinesEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenProgramsARB = _getProcAddress('glGenProgramsARB')
	gl.glGenProgramsARB.gl = gl
	gl.glGenProgramsARB.__name__ = 'glGenProgramsARB'
	gl.glGenProgramsARB.errcheck = _errcheck
	gl.glGenProgramsARB.restype = GLvoid
	gl.glGenProgramsARB.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenProgramsNV = _getProcAddress('glGenProgramsNV')
	gl.glGenProgramsNV.gl = gl
	gl.glGenProgramsNV.__name__ = 'glGenProgramsNV'
	gl.glGenProgramsNV.errcheck = _errcheck
	gl.glGenProgramsNV.restype = GLvoid
	gl.glGenProgramsNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenQueries = _getProcAddress('glGenQueries')
	gl.glGenQueries.gl = gl
	gl.glGenQueries.__name__ = 'glGenQueries'
	gl.glGenQueries.errcheck = _errcheck
	gl.glGenQueries.restype = GLvoid
	gl.glGenQueries.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenQueriesARB = _getProcAddress('glGenQueriesARB')
	gl.glGenQueriesARB.gl = gl
	gl.glGenQueriesARB.__name__ = 'glGenQueriesARB'
	gl.glGenQueriesARB.errcheck = _errcheck
	gl.glGenQueriesARB.restype = GLvoid
	gl.glGenQueriesARB.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenQueriesEXT = _getProcAddress('glGenQueriesEXT')
	gl.glGenQueriesEXT.gl = gl
	gl.glGenQueriesEXT.__name__ = 'glGenQueriesEXT'
	gl.glGenQueriesEXT.errcheck = _errcheck
	gl.glGenQueriesEXT.restype = GLvoid
	gl.glGenQueriesEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenRenderbuffers = _getProcAddress('glGenRenderbuffers')
	gl.glGenRenderbuffers.gl = gl
	gl.glGenRenderbuffers.__name__ = 'glGenRenderbuffers'
	gl.glGenRenderbuffers.errcheck = _errcheck
	gl.glGenRenderbuffers.restype = GLvoid
	gl.glGenRenderbuffers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenRenderbuffersEXT = _getProcAddress('glGenRenderbuffersEXT')
	gl.glGenRenderbuffersEXT.gl = gl
	gl.glGenRenderbuffersEXT.__name__ = 'glGenRenderbuffersEXT'
	gl.glGenRenderbuffersEXT.errcheck = _errcheck
	gl.glGenRenderbuffersEXT.restype = GLvoid
	gl.glGenRenderbuffersEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenRenderbuffersOES = _getProcAddress('glGenRenderbuffersOES')
	gl.glGenRenderbuffersOES.gl = gl
	gl.glGenRenderbuffersOES.__name__ = 'glGenRenderbuffersOES'
	gl.glGenRenderbuffersOES.errcheck = _errcheck
	gl.glGenRenderbuffersOES.restype = GLvoid
	gl.glGenRenderbuffersOES.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenSamplers = _getProcAddress('glGenSamplers')
	gl.glGenSamplers.gl = gl
	gl.glGenSamplers.__name__ = 'glGenSamplers'
	gl.glGenSamplers.errcheck = _errcheck
	gl.glGenSamplers.restype = GLvoid
	gl.glGenSamplers.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenSymbolsEXT = _getProcAddress('glGenSymbolsEXT')
	gl.glGenSymbolsEXT.gl = gl
	gl.glGenSymbolsEXT.__name__ = 'glGenSymbolsEXT'
	gl.glGenSymbolsEXT.errcheck = _errcheck
	gl.glGenSymbolsEXT.restype = GLuint
	gl.glGenSymbolsEXT.argtypes = (GLenum, GLenum, GLenum, GLuint,)
	gl.glGenTextures = _getProcAddress('glGenTextures')
	gl.glGenTextures.gl = gl
	gl.glGenTextures.__name__ = 'glGenTextures'
	gl.glGenTextures.errcheck = _errcheck
	gl.glGenTextures.restype = GLvoid
	gl.glGenTextures.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenTexturesEXT = _getProcAddress('glGenTexturesEXT')
	gl.glGenTexturesEXT.gl = gl
	gl.glGenTexturesEXT.__name__ = 'glGenTexturesEXT'
	gl.glGenTexturesEXT.errcheck = _errcheck
	gl.glGenTexturesEXT.restype = GLvoid
	gl.glGenTexturesEXT.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenTransformFeedbacks = _getProcAddress('glGenTransformFeedbacks')
	gl.glGenTransformFeedbacks.gl = gl
	gl.glGenTransformFeedbacks.__name__ = 'glGenTransformFeedbacks'
	gl.glGenTransformFeedbacks.errcheck = _errcheck
	gl.glGenTransformFeedbacks.restype = GLvoid
	gl.glGenTransformFeedbacks.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenTransformFeedbacksNV = _getProcAddress('glGenTransformFeedbacksNV')
	gl.glGenTransformFeedbacksNV.gl = gl
	gl.glGenTransformFeedbacksNV.__name__ = 'glGenTransformFeedbacksNV'
	gl.glGenTransformFeedbacksNV.errcheck = _errcheck
	gl.glGenTransformFeedbacksNV.restype = GLvoid
	gl.glGenTransformFeedbacksNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenVertexArrays = _getProcAddress('glGenVertexArrays')
	gl.glGenVertexArrays.gl = gl
	gl.glGenVertexArrays.__name__ = 'glGenVertexArrays'
	gl.glGenVertexArrays.errcheck = _errcheck
	gl.glGenVertexArrays.restype = GLvoid
	gl.glGenVertexArrays.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenVertexArraysAPPLE = _getProcAddress('glGenVertexArraysAPPLE')
	gl.glGenVertexArraysAPPLE.gl = gl
	gl.glGenVertexArraysAPPLE.__name__ = 'glGenVertexArraysAPPLE'
	gl.glGenVertexArraysAPPLE.errcheck = _errcheck
	gl.glGenVertexArraysAPPLE.restype = GLvoid
	gl.glGenVertexArraysAPPLE.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenVertexArraysOES = _getProcAddress('glGenVertexArraysOES')
	gl.glGenVertexArraysOES.gl = gl
	gl.glGenVertexArraysOES.__name__ = 'glGenVertexArraysOES'
	gl.glGenVertexArraysOES.errcheck = _errcheck
	gl.glGenVertexArraysOES.restype = GLvoid
	gl.glGenVertexArraysOES.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glGenVertexShadersEXT = _getProcAddress('glGenVertexShadersEXT')
	gl.glGenVertexShadersEXT.gl = gl
	gl.glGenVertexShadersEXT.__name__ = 'glGenVertexShadersEXT'
	gl.glGenVertexShadersEXT.errcheck = _errcheck
	gl.glGenVertexShadersEXT.restype = GLuint
	gl.glGenVertexShadersEXT.argtypes = (GLuint,)
	gl.glGenerateMipmap = _getProcAddress('glGenerateMipmap')
	gl.glGenerateMipmap.gl = gl
	gl.glGenerateMipmap.__name__ = 'glGenerateMipmap'
	gl.glGenerateMipmap.errcheck = _errcheck
	gl.glGenerateMipmap.restype = GLvoid
	gl.glGenerateMipmap.argtypes = (GLenum,)
	gl.glGenerateMipmapEXT = _getProcAddress('glGenerateMipmapEXT')
	gl.glGenerateMipmapEXT.gl = gl
	gl.glGenerateMipmapEXT.__name__ = 'glGenerateMipmapEXT'
	gl.glGenerateMipmapEXT.errcheck = _errcheck
	gl.glGenerateMipmapEXT.restype = GLvoid
	gl.glGenerateMipmapEXT.argtypes = (GLenum,)
	gl.glGenerateMipmapOES = _getProcAddress('glGenerateMipmapOES')
	gl.glGenerateMipmapOES.gl = gl
	gl.glGenerateMipmapOES.__name__ = 'glGenerateMipmapOES'
	gl.glGenerateMipmapOES.errcheck = _errcheck
	gl.glGenerateMipmapOES.restype = GLvoid
	gl.glGenerateMipmapOES.argtypes = (GLenum,)
	gl.glGenerateMultiTexMipmapEXT = _getProcAddress('glGenerateMultiTexMipmapEXT')
	gl.glGenerateMultiTexMipmapEXT.gl = gl
	gl.glGenerateMultiTexMipmapEXT.__name__ = 'glGenerateMultiTexMipmapEXT'
	gl.glGenerateMultiTexMipmapEXT.errcheck = _errcheck
	gl.glGenerateMultiTexMipmapEXT.restype = GLvoid
	gl.glGenerateMultiTexMipmapEXT.argtypes = (GLenum, GLenum,)
	gl.glGenerateTextureMipmap = _getProcAddress('glGenerateTextureMipmap')
	gl.glGenerateTextureMipmap.gl = gl
	gl.glGenerateTextureMipmap.__name__ = 'glGenerateTextureMipmap'
	gl.glGenerateTextureMipmap.errcheck = _errcheck
	gl.glGenerateTextureMipmap.restype = GLvoid
	gl.glGenerateTextureMipmap.argtypes = (GLuint,)
	gl.glGenerateTextureMipmapEXT = _getProcAddress('glGenerateTextureMipmapEXT')
	gl.glGenerateTextureMipmapEXT.gl = gl
	gl.glGenerateTextureMipmapEXT.__name__ = 'glGenerateTextureMipmapEXT'
	gl.glGenerateTextureMipmapEXT.errcheck = _errcheck
	gl.glGenerateTextureMipmapEXT.restype = GLvoid
	gl.glGenerateTextureMipmapEXT.argtypes = (GLuint, GLenum,)
	gl.glGetActiveAtomicCounterBufferiv = _getProcAddress('glGetActiveAtomicCounterBufferiv')
	gl.glGetActiveAtomicCounterBufferiv.gl = gl
	gl.glGetActiveAtomicCounterBufferiv.__name__ = 'glGetActiveAtomicCounterBufferiv'
	gl.glGetActiveAtomicCounterBufferiv.errcheck = _errcheck
	gl.glGetActiveAtomicCounterBufferiv.restype = GLvoid
	gl.glGetActiveAtomicCounterBufferiv.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint),)
	gl.glGetActiveAttrib = _getProcAddress('glGetActiveAttrib')
	gl.glGetActiveAttrib.gl = gl
	gl.glGetActiveAttrib.__name__ = 'glGetActiveAttrib'
	gl.glGetActiveAttrib.errcheck = _errcheck
	gl.glGetActiveAttrib.restype = GLvoid
	gl.glGetActiveAttrib.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLint), POINTER(GLenum), POINTER(GLchar),)
	gl.glGetActiveAttribARB = _getProcAddress('glGetActiveAttribARB')
	gl.glGetActiveAttribARB.gl = gl
	gl.glGetActiveAttribARB.__name__ = 'glGetActiveAttribARB'
	gl.glGetActiveAttribARB.errcheck = _errcheck
	gl.glGetActiveAttribARB.restype = GLvoid
	gl.glGetActiveAttribARB.argtypes = (GLhandleARB, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLint), POINTER(GLenum), POINTER(GLcharARB),)
	gl.glGetActiveSubroutineName = _getProcAddress('glGetActiveSubroutineName')
	gl.glGetActiveSubroutineName.gl = gl
	gl.glGetActiveSubroutineName.__name__ = 'glGetActiveSubroutineName'
	gl.glGetActiveSubroutineName.errcheck = _errcheck
	gl.glGetActiveSubroutineName.restype = GLvoid
	gl.glGetActiveSubroutineName.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetActiveSubroutineUniformName = _getProcAddress('glGetActiveSubroutineUniformName')
	gl.glGetActiveSubroutineUniformName.gl = gl
	gl.glGetActiveSubroutineUniformName.__name__ = 'glGetActiveSubroutineUniformName'
	gl.glGetActiveSubroutineUniformName.errcheck = _errcheck
	gl.glGetActiveSubroutineUniformName.restype = GLvoid
	gl.glGetActiveSubroutineUniformName.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetActiveSubroutineUniformiv = _getProcAddress('glGetActiveSubroutineUniformiv')
	gl.glGetActiveSubroutineUniformiv.gl = gl
	gl.glGetActiveSubroutineUniformiv.__name__ = 'glGetActiveSubroutineUniformiv'
	gl.glGetActiveSubroutineUniformiv.errcheck = _errcheck
	gl.glGetActiveSubroutineUniformiv.restype = GLvoid
	gl.glGetActiveSubroutineUniformiv.argtypes = (GLuint, GLenum, GLuint, GLenum, POINTER(GLint),)
	gl.glGetActiveUniform = _getProcAddress('glGetActiveUniform')
	gl.glGetActiveUniform.gl = gl
	gl.glGetActiveUniform.__name__ = 'glGetActiveUniform'
	gl.glGetActiveUniform.errcheck = _errcheck
	gl.glGetActiveUniform.restype = GLvoid
	gl.glGetActiveUniform.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLint), POINTER(GLenum), POINTER(GLchar),)
	gl.glGetActiveUniformARB = _getProcAddress('glGetActiveUniformARB')
	gl.glGetActiveUniformARB.gl = gl
	gl.glGetActiveUniformARB.__name__ = 'glGetActiveUniformARB'
	gl.glGetActiveUniformARB.errcheck = _errcheck
	gl.glGetActiveUniformARB.restype = GLvoid
	gl.glGetActiveUniformARB.argtypes = (GLhandleARB, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLint), POINTER(GLenum), POINTER(GLcharARB),)
	gl.glGetActiveUniformBlockName = _getProcAddress('glGetActiveUniformBlockName')
	gl.glGetActiveUniformBlockName.gl = gl
	gl.glGetActiveUniformBlockName.__name__ = 'glGetActiveUniformBlockName'
	gl.glGetActiveUniformBlockName.errcheck = _errcheck
	gl.glGetActiveUniformBlockName.restype = GLvoid
	gl.glGetActiveUniformBlockName.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetActiveUniformBlockiv = _getProcAddress('glGetActiveUniformBlockiv')
	gl.glGetActiveUniformBlockiv.gl = gl
	gl.glGetActiveUniformBlockiv.__name__ = 'glGetActiveUniformBlockiv'
	gl.glGetActiveUniformBlockiv.errcheck = _errcheck
	gl.glGetActiveUniformBlockiv.restype = GLvoid
	gl.glGetActiveUniformBlockiv.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint),)
	gl.glGetActiveUniformName = _getProcAddress('glGetActiveUniformName')
	gl.glGetActiveUniformName.gl = gl
	gl.glGetActiveUniformName.__name__ = 'glGetActiveUniformName'
	gl.glGetActiveUniformName.errcheck = _errcheck
	gl.glGetActiveUniformName.restype = GLvoid
	gl.glGetActiveUniformName.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetActiveUniformsiv = _getProcAddress('glGetActiveUniformsiv')
	gl.glGetActiveUniformsiv.gl = gl
	gl.glGetActiveUniformsiv.__name__ = 'glGetActiveUniformsiv'
	gl.glGetActiveUniformsiv.errcheck = _errcheck
	gl.glGetActiveUniformsiv.restype = GLvoid
	gl.glGetActiveUniformsiv.argtypes = (GLuint, GLsizei, POINTER(GLuint), GLenum, POINTER(GLint),)
	gl.glGetActiveVaryingNV = _getProcAddress('glGetActiveVaryingNV')
	gl.glGetActiveVaryingNV.gl = gl
	gl.glGetActiveVaryingNV.__name__ = 'glGetActiveVaryingNV'
	gl.glGetActiveVaryingNV.errcheck = _errcheck
	gl.glGetActiveVaryingNV.restype = GLvoid
	gl.glGetActiveVaryingNV.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLsizei), POINTER(GLenum), POINTER(GLchar),)
	gl.glGetArrayObjectfvATI = _getProcAddress('glGetArrayObjectfvATI')
	gl.glGetArrayObjectfvATI.gl = gl
	gl.glGetArrayObjectfvATI.__name__ = 'glGetArrayObjectfvATI'
	gl.glGetArrayObjectfvATI.errcheck = _errcheck
	gl.glGetArrayObjectfvATI.restype = GLvoid
	gl.glGetArrayObjectfvATI.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetArrayObjectivATI = _getProcAddress('glGetArrayObjectivATI')
	gl.glGetArrayObjectivATI.gl = gl
	gl.glGetArrayObjectivATI.__name__ = 'glGetArrayObjectivATI'
	gl.glGetArrayObjectivATI.errcheck = _errcheck
	gl.glGetArrayObjectivATI.restype = GLvoid
	gl.glGetArrayObjectivATI.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetAttachedObjectsARB = _getProcAddress('glGetAttachedObjectsARB')
	gl.glGetAttachedObjectsARB.gl = gl
	gl.glGetAttachedObjectsARB.__name__ = 'glGetAttachedObjectsARB'
	gl.glGetAttachedObjectsARB.errcheck = _errcheck
	gl.glGetAttachedObjectsARB.restype = GLvoid
	gl.glGetAttachedObjectsARB.argtypes = (GLhandleARB, GLsizei, POINTER(GLsizei), POINTER(GLhandleARB),)
	gl.glGetAttachedShaders = _getProcAddress('glGetAttachedShaders')
	gl.glGetAttachedShaders.gl = gl
	gl.glGetAttachedShaders.__name__ = 'glGetAttachedShaders'
	gl.glGetAttachedShaders.errcheck = _errcheck
	gl.glGetAttachedShaders.restype = GLvoid
	gl.glGetAttachedShaders.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLuint),)
	gl.glGetAttribLocation = _getProcAddress('glGetAttribLocation')
	gl.glGetAttribLocation.gl = gl
	gl.glGetAttribLocation.__name__ = 'glGetAttribLocation'
	gl.glGetAttribLocation.errcheck = _errcheck
	gl.glGetAttribLocation.restype = GLint
	gl.glGetAttribLocation.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetAttribLocationARB = _getProcAddress('glGetAttribLocationARB')
	gl.glGetAttribLocationARB.gl = gl
	gl.glGetAttribLocationARB.__name__ = 'glGetAttribLocationARB'
	gl.glGetAttribLocationARB.errcheck = _errcheck
	gl.glGetAttribLocationARB.restype = GLint
	gl.glGetAttribLocationARB.argtypes = (GLhandleARB, POINTER(GLcharARB),)
	gl.glGetBooleanIndexedvEXT = _getProcAddress('glGetBooleanIndexedvEXT')
	gl.glGetBooleanIndexedvEXT.gl = gl
	gl.glGetBooleanIndexedvEXT.__name__ = 'glGetBooleanIndexedvEXT'
	gl.glGetBooleanIndexedvEXT.errcheck = _errcheck
	gl.glGetBooleanIndexedvEXT.restype = GLvoid
	gl.glGetBooleanIndexedvEXT.argtypes = (GLenum, GLuint, POINTER(GLboolean),)
	gl.glGetBooleani_v = _getProcAddress('glGetBooleani_v')
	gl.glGetBooleani_v.gl = gl
	gl.glGetBooleani_v.__name__ = 'glGetBooleani_v'
	gl.glGetBooleani_v.errcheck = _errcheck
	gl.glGetBooleani_v.restype = GLvoid
	gl.glGetBooleani_v.argtypes = (GLenum, GLuint, POINTER(GLboolean),)
	gl.glGetBooleanv = _getProcAddress('glGetBooleanv')
	gl.glGetBooleanv.gl = gl
	gl.glGetBooleanv.__name__ = 'glGetBooleanv'
	gl.glGetBooleanv.errcheck = _errcheck
	gl.glGetBooleanv.restype = GLvoid
	gl.glGetBooleanv.argtypes = (GLenum, POINTER(GLboolean),)
	gl.glGetBufferParameteri64v = _getProcAddress('glGetBufferParameteri64v')
	gl.glGetBufferParameteri64v.gl = gl
	gl.glGetBufferParameteri64v.__name__ = 'glGetBufferParameteri64v'
	gl.glGetBufferParameteri64v.errcheck = _errcheck
	gl.glGetBufferParameteri64v.restype = GLvoid
	gl.glGetBufferParameteri64v.argtypes = (GLenum, GLenum, POINTER(GLint64),)
	gl.glGetBufferParameteriv = _getProcAddress('glGetBufferParameteriv')
	gl.glGetBufferParameteriv.gl = gl
	gl.glGetBufferParameteriv.__name__ = 'glGetBufferParameteriv'
	gl.glGetBufferParameteriv.errcheck = _errcheck
	gl.glGetBufferParameteriv.restype = GLvoid
	gl.glGetBufferParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetBufferParameterivARB = _getProcAddress('glGetBufferParameterivARB')
	gl.glGetBufferParameterivARB.gl = gl
	gl.glGetBufferParameterivARB.__name__ = 'glGetBufferParameterivARB'
	gl.glGetBufferParameterivARB.errcheck = _errcheck
	gl.glGetBufferParameterivARB.restype = GLvoid
	gl.glGetBufferParameterivARB.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetBufferParameterui64vNV = _getProcAddress('glGetBufferParameterui64vNV')
	gl.glGetBufferParameterui64vNV.gl = gl
	gl.glGetBufferParameterui64vNV.__name__ = 'glGetBufferParameterui64vNV'
	gl.glGetBufferParameterui64vNV.errcheck = _errcheck
	gl.glGetBufferParameterui64vNV.restype = GLvoid
	gl.glGetBufferParameterui64vNV.argtypes = (GLenum, GLenum, POINTER(GLuint64EXT),)
	gl.glGetBufferPointerv = _getProcAddress('glGetBufferPointerv')
	gl.glGetBufferPointerv.gl = gl
	gl.glGetBufferPointerv.__name__ = 'glGetBufferPointerv'
	gl.glGetBufferPointerv.errcheck = _errcheck
	gl.glGetBufferPointerv.restype = GLvoid
	gl.glGetBufferPointerv.argtypes = (GLenum, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetBufferPointervARB = _getProcAddress('glGetBufferPointervARB')
	gl.glGetBufferPointervARB.gl = gl
	gl.glGetBufferPointervARB.__name__ = 'glGetBufferPointervARB'
	gl.glGetBufferPointervARB.errcheck = _errcheck
	gl.glGetBufferPointervARB.restype = GLvoid
	gl.glGetBufferPointervARB.argtypes = (GLenum, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetBufferPointervOES = _getProcAddress('glGetBufferPointervOES')
	gl.glGetBufferPointervOES.gl = gl
	gl.glGetBufferPointervOES.__name__ = 'glGetBufferPointervOES'
	gl.glGetBufferPointervOES.errcheck = _errcheck
	gl.glGetBufferPointervOES.restype = GLvoid
	gl.glGetBufferPointervOES.argtypes = (GLenum, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetBufferSubData = _getProcAddress('glGetBufferSubData')
	gl.glGetBufferSubData.gl = gl
	gl.glGetBufferSubData.__name__ = 'glGetBufferSubData'
	gl.glGetBufferSubData.errcheck = _errcheck
	gl.glGetBufferSubData.restype = GLvoid
	gl.glGetBufferSubData.argtypes = (GLenum, GLintptr, GLsizeiptr, POINTER(GLvoid),)
	gl.glGetBufferSubDataARB = _getProcAddress('glGetBufferSubDataARB')
	gl.glGetBufferSubDataARB.gl = gl
	gl.glGetBufferSubDataARB.__name__ = 'glGetBufferSubDataARB'
	gl.glGetBufferSubDataARB.errcheck = _errcheck
	gl.glGetBufferSubDataARB.restype = GLvoid
	gl.glGetBufferSubDataARB.argtypes = (GLenum, GLintptrARB, GLsizeiptrARB, POINTER(GLvoid),)
	gl.glGetClipPlane = _getProcAddress('glGetClipPlane')
	gl.glGetClipPlane.gl = gl
	gl.glGetClipPlane.__name__ = 'glGetClipPlane'
	gl.glGetClipPlane.errcheck = _errcheck
	gl.glGetClipPlane.restype = GLvoid
	gl.glGetClipPlane.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glGetClipPlanef = _getProcAddress('glGetClipPlanef')
	gl.glGetClipPlanef.gl = gl
	gl.glGetClipPlanef.__name__ = 'glGetClipPlanef'
	gl.glGetClipPlanef.errcheck = _errcheck
	gl.glGetClipPlanef.restype = GLvoid
	gl.glGetClipPlanef.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetClipPlanefOES = _getProcAddress('glGetClipPlanefOES')
	gl.glGetClipPlanefOES.gl = gl
	gl.glGetClipPlanefOES.__name__ = 'glGetClipPlanefOES'
	gl.glGetClipPlanefOES.errcheck = _errcheck
	gl.glGetClipPlanefOES.restype = GLvoid
	gl.glGetClipPlanefOES.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetClipPlanex = _getProcAddress('glGetClipPlanex')
	gl.glGetClipPlanex.gl = gl
	gl.glGetClipPlanex.__name__ = 'glGetClipPlanex'
	gl.glGetClipPlanex.errcheck = _errcheck
	gl.glGetClipPlanex.restype = GLvoid
	gl.glGetClipPlanex.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glGetClipPlanexOES = _getProcAddress('glGetClipPlanexOES')
	gl.glGetClipPlanexOES.gl = gl
	gl.glGetClipPlanexOES.__name__ = 'glGetClipPlanexOES'
	gl.glGetClipPlanexOES.errcheck = _errcheck
	gl.glGetClipPlanexOES.restype = GLvoid
	gl.glGetClipPlanexOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glGetColorTable = _getProcAddress('glGetColorTable')
	gl.glGetColorTable.gl = gl
	gl.glGetColorTable.__name__ = 'glGetColorTable'
	gl.glGetColorTable.errcheck = _errcheck
	gl.glGetColorTable.restype = GLvoid
	gl.glGetColorTable.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetColorTableEXT = _getProcAddress('glGetColorTableEXT')
	gl.glGetColorTableEXT.gl = gl
	gl.glGetColorTableEXT.__name__ = 'glGetColorTableEXT'
	gl.glGetColorTableEXT.errcheck = _errcheck
	gl.glGetColorTableEXT.restype = GLvoid
	gl.glGetColorTableEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetColorTableParameterfv = _getProcAddress('glGetColorTableParameterfv')
	gl.glGetColorTableParameterfv.gl = gl
	gl.glGetColorTableParameterfv.__name__ = 'glGetColorTableParameterfv'
	gl.glGetColorTableParameterfv.errcheck = _errcheck
	gl.glGetColorTableParameterfv.restype = GLvoid
	gl.glGetColorTableParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetColorTableParameterfvEXT = _getProcAddress('glGetColorTableParameterfvEXT')
	gl.glGetColorTableParameterfvEXT.gl = gl
	gl.glGetColorTableParameterfvEXT.__name__ = 'glGetColorTableParameterfvEXT'
	gl.glGetColorTableParameterfvEXT.errcheck = _errcheck
	gl.glGetColorTableParameterfvEXT.restype = GLvoid
	gl.glGetColorTableParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetColorTableParameterfvSGI = _getProcAddress('glGetColorTableParameterfvSGI')
	gl.glGetColorTableParameterfvSGI.gl = gl
	gl.glGetColorTableParameterfvSGI.__name__ = 'glGetColorTableParameterfvSGI'
	gl.glGetColorTableParameterfvSGI.errcheck = _errcheck
	gl.glGetColorTableParameterfvSGI.restype = GLvoid
	gl.glGetColorTableParameterfvSGI.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetColorTableParameteriv = _getProcAddress('glGetColorTableParameteriv')
	gl.glGetColorTableParameteriv.gl = gl
	gl.glGetColorTableParameteriv.__name__ = 'glGetColorTableParameteriv'
	gl.glGetColorTableParameteriv.errcheck = _errcheck
	gl.glGetColorTableParameteriv.restype = GLvoid
	gl.glGetColorTableParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetColorTableParameterivEXT = _getProcAddress('glGetColorTableParameterivEXT')
	gl.glGetColorTableParameterivEXT.gl = gl
	gl.glGetColorTableParameterivEXT.__name__ = 'glGetColorTableParameterivEXT'
	gl.glGetColorTableParameterivEXT.errcheck = _errcheck
	gl.glGetColorTableParameterivEXT.restype = GLvoid
	gl.glGetColorTableParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetColorTableParameterivSGI = _getProcAddress('glGetColorTableParameterivSGI')
	gl.glGetColorTableParameterivSGI.gl = gl
	gl.glGetColorTableParameterivSGI.__name__ = 'glGetColorTableParameterivSGI'
	gl.glGetColorTableParameterivSGI.errcheck = _errcheck
	gl.glGetColorTableParameterivSGI.restype = GLvoid
	gl.glGetColorTableParameterivSGI.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetColorTableSGI = _getProcAddress('glGetColorTableSGI')
	gl.glGetColorTableSGI.gl = gl
	gl.glGetColorTableSGI.__name__ = 'glGetColorTableSGI'
	gl.glGetColorTableSGI.errcheck = _errcheck
	gl.glGetColorTableSGI.restype = GLvoid
	gl.glGetColorTableSGI.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetCombinerInputParameterfvNV = _getProcAddress('glGetCombinerInputParameterfvNV')
	gl.glGetCombinerInputParameterfvNV.gl = gl
	gl.glGetCombinerInputParameterfvNV.__name__ = 'glGetCombinerInputParameterfvNV'
	gl.glGetCombinerInputParameterfvNV.errcheck = _errcheck
	gl.glGetCombinerInputParameterfvNV.restype = GLvoid
	gl.glGetCombinerInputParameterfvNV.argtypes = (GLenum, GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetCombinerInputParameterivNV = _getProcAddress('glGetCombinerInputParameterivNV')
	gl.glGetCombinerInputParameterivNV.gl = gl
	gl.glGetCombinerInputParameterivNV.__name__ = 'glGetCombinerInputParameterivNV'
	gl.glGetCombinerInputParameterivNV.errcheck = _errcheck
	gl.glGetCombinerInputParameterivNV.restype = GLvoid
	gl.glGetCombinerInputParameterivNV.argtypes = (GLenum, GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetCombinerOutputParameterfvNV = _getProcAddress('glGetCombinerOutputParameterfvNV')
	gl.glGetCombinerOutputParameterfvNV.gl = gl
	gl.glGetCombinerOutputParameterfvNV.__name__ = 'glGetCombinerOutputParameterfvNV'
	gl.glGetCombinerOutputParameterfvNV.errcheck = _errcheck
	gl.glGetCombinerOutputParameterfvNV.restype = GLvoid
	gl.glGetCombinerOutputParameterfvNV.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetCombinerOutputParameterivNV = _getProcAddress('glGetCombinerOutputParameterivNV')
	gl.glGetCombinerOutputParameterivNV.gl = gl
	gl.glGetCombinerOutputParameterivNV.__name__ = 'glGetCombinerOutputParameterivNV'
	gl.glGetCombinerOutputParameterivNV.errcheck = _errcheck
	gl.glGetCombinerOutputParameterivNV.restype = GLvoid
	gl.glGetCombinerOutputParameterivNV.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetCombinerStageParameterfvNV = _getProcAddress('glGetCombinerStageParameterfvNV')
	gl.glGetCombinerStageParameterfvNV.gl = gl
	gl.glGetCombinerStageParameterfvNV.__name__ = 'glGetCombinerStageParameterfvNV'
	gl.glGetCombinerStageParameterfvNV.errcheck = _errcheck
	gl.glGetCombinerStageParameterfvNV.restype = GLvoid
	gl.glGetCombinerStageParameterfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetCompressedMultiTexImageEXT = _getProcAddress('glGetCompressedMultiTexImageEXT')
	gl.glGetCompressedMultiTexImageEXT.gl = gl
	gl.glGetCompressedMultiTexImageEXT.__name__ = 'glGetCompressedMultiTexImageEXT'
	gl.glGetCompressedMultiTexImageEXT.errcheck = _errcheck
	gl.glGetCompressedMultiTexImageEXT.restype = GLvoid
	gl.glGetCompressedMultiTexImageEXT.argtypes = (GLenum, GLenum, GLint, POINTER(GLvoid),)
	gl.glGetCompressedTexImage = _getProcAddress('glGetCompressedTexImage')
	gl.glGetCompressedTexImage.gl = gl
	gl.glGetCompressedTexImage.__name__ = 'glGetCompressedTexImage'
	gl.glGetCompressedTexImage.errcheck = _errcheck
	gl.glGetCompressedTexImage.restype = GLvoid
	gl.glGetCompressedTexImage.argtypes = (GLenum, GLint, POINTER(GLvoid),)
	gl.glGetCompressedTexImageARB = _getProcAddress('glGetCompressedTexImageARB')
	gl.glGetCompressedTexImageARB.gl = gl
	gl.glGetCompressedTexImageARB.__name__ = 'glGetCompressedTexImageARB'
	gl.glGetCompressedTexImageARB.errcheck = _errcheck
	gl.glGetCompressedTexImageARB.restype = GLvoid
	gl.glGetCompressedTexImageARB.argtypes = (GLenum, GLint, POINTER(GLvoid),)
	gl.glGetCompressedTextureImage = _getProcAddress('glGetCompressedTextureImage')
	gl.glGetCompressedTextureImage.gl = gl
	gl.glGetCompressedTextureImage.__name__ = 'glGetCompressedTextureImage'
	gl.glGetCompressedTextureImage.errcheck = _errcheck
	gl.glGetCompressedTextureImage.restype = GLvoid
	gl.glGetCompressedTextureImage.argtypes = (GLuint, GLint, GLsizei, POINTER(GLvoid),)
	gl.glGetCompressedTextureImageEXT = _getProcAddress('glGetCompressedTextureImageEXT')
	gl.glGetCompressedTextureImageEXT.gl = gl
	gl.glGetCompressedTextureImageEXT.__name__ = 'glGetCompressedTextureImageEXT'
	gl.glGetCompressedTextureImageEXT.errcheck = _errcheck
	gl.glGetCompressedTextureImageEXT.restype = GLvoid
	gl.glGetCompressedTextureImageEXT.argtypes = (GLuint, GLenum, GLint, POINTER(GLvoid),)
	gl.glGetCompressedTextureSubImage = _getProcAddress('glGetCompressedTextureSubImage')
	gl.glGetCompressedTextureSubImage.gl = gl
	gl.glGetCompressedTextureSubImage.__name__ = 'glGetCompressedTextureSubImage'
	gl.glGetCompressedTextureSubImage.errcheck = _errcheck
	gl.glGetCompressedTextureSubImage.restype = GLvoid
	gl.glGetCompressedTextureSubImage.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLsizei, POINTER(GLvoid),)
	gl.glGetConvolutionFilter = _getProcAddress('glGetConvolutionFilter')
	gl.glGetConvolutionFilter.gl = gl
	gl.glGetConvolutionFilter.__name__ = 'glGetConvolutionFilter'
	gl.glGetConvolutionFilter.errcheck = _errcheck
	gl.glGetConvolutionFilter.restype = GLvoid
	gl.glGetConvolutionFilter.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetConvolutionFilterEXT = _getProcAddress('glGetConvolutionFilterEXT')
	gl.glGetConvolutionFilterEXT.gl = gl
	gl.glGetConvolutionFilterEXT.__name__ = 'glGetConvolutionFilterEXT'
	gl.glGetConvolutionFilterEXT.errcheck = _errcheck
	gl.glGetConvolutionFilterEXT.restype = GLvoid
	gl.glGetConvolutionFilterEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetConvolutionParameterfv = _getProcAddress('glGetConvolutionParameterfv')
	gl.glGetConvolutionParameterfv.gl = gl
	gl.glGetConvolutionParameterfv.__name__ = 'glGetConvolutionParameterfv'
	gl.glGetConvolutionParameterfv.errcheck = _errcheck
	gl.glGetConvolutionParameterfv.restype = GLvoid
	gl.glGetConvolutionParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetConvolutionParameterfvEXT = _getProcAddress('glGetConvolutionParameterfvEXT')
	gl.glGetConvolutionParameterfvEXT.gl = gl
	gl.glGetConvolutionParameterfvEXT.__name__ = 'glGetConvolutionParameterfvEXT'
	gl.glGetConvolutionParameterfvEXT.errcheck = _errcheck
	gl.glGetConvolutionParameterfvEXT.restype = GLvoid
	gl.glGetConvolutionParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetConvolutionParameteriv = _getProcAddress('glGetConvolutionParameteriv')
	gl.glGetConvolutionParameteriv.gl = gl
	gl.glGetConvolutionParameteriv.__name__ = 'glGetConvolutionParameteriv'
	gl.glGetConvolutionParameteriv.errcheck = _errcheck
	gl.glGetConvolutionParameteriv.restype = GLvoid
	gl.glGetConvolutionParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetConvolutionParameterivEXT = _getProcAddress('glGetConvolutionParameterivEXT')
	gl.glGetConvolutionParameterivEXT.gl = gl
	gl.glGetConvolutionParameterivEXT.__name__ = 'glGetConvolutionParameterivEXT'
	gl.glGetConvolutionParameterivEXT.errcheck = _errcheck
	gl.glGetConvolutionParameterivEXT.restype = GLvoid
	gl.glGetConvolutionParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetConvolutionParameterxvOES = _getProcAddress('glGetConvolutionParameterxvOES')
	gl.glGetConvolutionParameterxvOES.gl = gl
	gl.glGetConvolutionParameterxvOES.__name__ = 'glGetConvolutionParameterxvOES'
	gl.glGetConvolutionParameterxvOES.errcheck = _errcheck
	gl.glGetConvolutionParameterxvOES.restype = GLvoid
	gl.glGetConvolutionParameterxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetDebugMessageLog = _getProcAddress('glGetDebugMessageLog')
	gl.glGetDebugMessageLog.gl = gl
	gl.glGetDebugMessageLog.__name__ = 'glGetDebugMessageLog'
	gl.glGetDebugMessageLog.errcheck = _errcheck
	gl.glGetDebugMessageLog.restype = GLuint
	gl.glGetDebugMessageLog.argtypes = (GLuint, GLsizei, POINTER(GLenum), POINTER(GLenum), POINTER(GLuint), POINTER(GLenum), POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetDebugMessageLogAMD = _getProcAddress('glGetDebugMessageLogAMD')
	gl.glGetDebugMessageLogAMD.gl = gl
	gl.glGetDebugMessageLogAMD.__name__ = 'glGetDebugMessageLogAMD'
	gl.glGetDebugMessageLogAMD.errcheck = _errcheck
	gl.glGetDebugMessageLogAMD.restype = GLuint
	gl.glGetDebugMessageLogAMD.argtypes = (GLuint, GLsizei, POINTER(GLenum), POINTER(GLuint), POINTER(GLuint), POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetDebugMessageLogARB = _getProcAddress('glGetDebugMessageLogARB')
	gl.glGetDebugMessageLogARB.gl = gl
	gl.glGetDebugMessageLogARB.__name__ = 'glGetDebugMessageLogARB'
	gl.glGetDebugMessageLogARB.errcheck = _errcheck
	gl.glGetDebugMessageLogARB.restype = GLuint
	gl.glGetDebugMessageLogARB.argtypes = (GLuint, GLsizei, POINTER(GLenum), POINTER(GLenum), POINTER(GLuint), POINTER(GLenum), POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetDebugMessageLogKHR = _getProcAddress('glGetDebugMessageLogKHR')
	gl.glGetDebugMessageLogKHR.gl = gl
	gl.glGetDebugMessageLogKHR.__name__ = 'glGetDebugMessageLogKHR'
	gl.glGetDebugMessageLogKHR.errcheck = _errcheck
	gl.glGetDebugMessageLogKHR.restype = GLuint
	gl.glGetDebugMessageLogKHR.argtypes = (GLuint, GLsizei, POINTER(GLenum), POINTER(GLenum), POINTER(GLuint), POINTER(GLenum), POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetDetailTexFuncSGIS = _getProcAddress('glGetDetailTexFuncSGIS')
	gl.glGetDetailTexFuncSGIS.gl = gl
	gl.glGetDetailTexFuncSGIS.__name__ = 'glGetDetailTexFuncSGIS'
	gl.glGetDetailTexFuncSGIS.errcheck = _errcheck
	gl.glGetDetailTexFuncSGIS.restype = GLvoid
	gl.glGetDetailTexFuncSGIS.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetDoubleIndexedvEXT = _getProcAddress('glGetDoubleIndexedvEXT')
	gl.glGetDoubleIndexedvEXT.gl = gl
	gl.glGetDoubleIndexedvEXT.__name__ = 'glGetDoubleIndexedvEXT'
	gl.glGetDoubleIndexedvEXT.errcheck = _errcheck
	gl.glGetDoubleIndexedvEXT.restype = GLvoid
	gl.glGetDoubleIndexedvEXT.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glGetDoublei_v = _getProcAddress('glGetDoublei_v')
	gl.glGetDoublei_v.gl = gl
	gl.glGetDoublei_v.__name__ = 'glGetDoublei_v'
	gl.glGetDoublei_v.errcheck = _errcheck
	gl.glGetDoublei_v.restype = GLvoid
	gl.glGetDoublei_v.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glGetDoublei_vEXT = _getProcAddress('glGetDoublei_vEXT')
	gl.glGetDoublei_vEXT.gl = gl
	gl.glGetDoublei_vEXT.__name__ = 'glGetDoublei_vEXT'
	gl.glGetDoublei_vEXT.errcheck = _errcheck
	gl.glGetDoublei_vEXT.restype = GLvoid
	gl.glGetDoublei_vEXT.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glGetDoublev = _getProcAddress('glGetDoublev')
	gl.glGetDoublev.gl = gl
	gl.glGetDoublev.__name__ = 'glGetDoublev'
	gl.glGetDoublev.errcheck = _errcheck
	gl.glGetDoublev.restype = GLvoid
	gl.glGetDoublev.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glGetDriverControlStringQCOM = _getProcAddress('glGetDriverControlStringQCOM')
	gl.glGetDriverControlStringQCOM.gl = gl
	gl.glGetDriverControlStringQCOM.__name__ = 'glGetDriverControlStringQCOM'
	gl.glGetDriverControlStringQCOM.errcheck = _errcheck
	gl.glGetDriverControlStringQCOM.restype = GLvoid
	gl.glGetDriverControlStringQCOM.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetDriverControlsQCOM = _getProcAddress('glGetDriverControlsQCOM')
	gl.glGetDriverControlsQCOM.gl = gl
	gl.glGetDriverControlsQCOM.__name__ = 'glGetDriverControlsQCOM'
	gl.glGetDriverControlsQCOM.errcheck = _errcheck
	gl.glGetDriverControlsQCOM.restype = GLvoid
	gl.glGetDriverControlsQCOM.argtypes = (POINTER(GLint), GLsizei, POINTER(GLuint),)
	gl.glGetError = _getProcAddress('glGetError')
	gl.glGetError.gl = gl
	gl.glGetError.__name__ = 'glGetError'
	gl.glGetError.restype = GLenum
	gl.glGetError.argtypes = ()
	gl.glGetFenceivNV = _getProcAddress('glGetFenceivNV')
	gl.glGetFenceivNV.gl = gl
	gl.glGetFenceivNV.__name__ = 'glGetFenceivNV'
	gl.glGetFenceivNV.errcheck = _errcheck
	gl.glGetFenceivNV.restype = GLvoid
	gl.glGetFenceivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetFinalCombinerInputParameterfvNV = _getProcAddress('glGetFinalCombinerInputParameterfvNV')
	gl.glGetFinalCombinerInputParameterfvNV.gl = gl
	gl.glGetFinalCombinerInputParameterfvNV.__name__ = 'glGetFinalCombinerInputParameterfvNV'
	gl.glGetFinalCombinerInputParameterfvNV.errcheck = _errcheck
	gl.glGetFinalCombinerInputParameterfvNV.restype = GLvoid
	gl.glGetFinalCombinerInputParameterfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetFinalCombinerInputParameterivNV = _getProcAddress('glGetFinalCombinerInputParameterivNV')
	gl.glGetFinalCombinerInputParameterivNV.gl = gl
	gl.glGetFinalCombinerInputParameterivNV.__name__ = 'glGetFinalCombinerInputParameterivNV'
	gl.glGetFinalCombinerInputParameterivNV.errcheck = _errcheck
	gl.glGetFinalCombinerInputParameterivNV.restype = GLvoid
	gl.glGetFinalCombinerInputParameterivNV.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetFirstPerfQueryIdINTEL = _getProcAddress('glGetFirstPerfQueryIdINTEL')
	gl.glGetFirstPerfQueryIdINTEL.gl = gl
	gl.glGetFirstPerfQueryIdINTEL.__name__ = 'glGetFirstPerfQueryIdINTEL'
	gl.glGetFirstPerfQueryIdINTEL.errcheck = _errcheck
	gl.glGetFirstPerfQueryIdINTEL.restype = GLvoid
	gl.glGetFirstPerfQueryIdINTEL.argtypes = (POINTER(GLuint),)
	gl.glGetFixedv = _getProcAddress('glGetFixedv')
	gl.glGetFixedv.gl = gl
	gl.glGetFixedv.__name__ = 'glGetFixedv'
	gl.glGetFixedv.errcheck = _errcheck
	gl.glGetFixedv.restype = GLvoid
	gl.glGetFixedv.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glGetFixedvOES = _getProcAddress('glGetFixedvOES')
	gl.glGetFixedvOES.gl = gl
	gl.glGetFixedvOES.__name__ = 'glGetFixedvOES'
	gl.glGetFixedvOES.errcheck = _errcheck
	gl.glGetFixedvOES.restype = GLvoid
	gl.glGetFixedvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glGetFloatIndexedvEXT = _getProcAddress('glGetFloatIndexedvEXT')
	gl.glGetFloatIndexedvEXT.gl = gl
	gl.glGetFloatIndexedvEXT.__name__ = 'glGetFloatIndexedvEXT'
	gl.glGetFloatIndexedvEXT.errcheck = _errcheck
	gl.glGetFloatIndexedvEXT.restype = GLvoid
	gl.glGetFloatIndexedvEXT.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetFloati_v = _getProcAddress('glGetFloati_v')
	gl.glGetFloati_v.gl = gl
	gl.glGetFloati_v.__name__ = 'glGetFloati_v'
	gl.glGetFloati_v.errcheck = _errcheck
	gl.glGetFloati_v.restype = GLvoid
	gl.glGetFloati_v.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetFloati_vEXT = _getProcAddress('glGetFloati_vEXT')
	gl.glGetFloati_vEXT.gl = gl
	gl.glGetFloati_vEXT.__name__ = 'glGetFloati_vEXT'
	gl.glGetFloati_vEXT.errcheck = _errcheck
	gl.glGetFloati_vEXT.restype = GLvoid
	gl.glGetFloati_vEXT.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetFloatv = _getProcAddress('glGetFloatv')
	gl.glGetFloatv.gl = gl
	gl.glGetFloatv.__name__ = 'glGetFloatv'
	gl.glGetFloatv.errcheck = _errcheck
	gl.glGetFloatv.restype = GLvoid
	gl.glGetFloatv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetFogFuncSGIS = _getProcAddress('glGetFogFuncSGIS')
	gl.glGetFogFuncSGIS.gl = gl
	gl.glGetFogFuncSGIS.__name__ = 'glGetFogFuncSGIS'
	gl.glGetFogFuncSGIS.errcheck = _errcheck
	gl.glGetFogFuncSGIS.restype = GLvoid
	gl.glGetFogFuncSGIS.argtypes = (POINTER(GLfloat),)
	gl.glGetFragDataIndex = _getProcAddress('glGetFragDataIndex')
	gl.glGetFragDataIndex.gl = gl
	gl.glGetFragDataIndex.__name__ = 'glGetFragDataIndex'
	gl.glGetFragDataIndex.errcheck = _errcheck
	gl.glGetFragDataIndex.restype = GLint
	gl.glGetFragDataIndex.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetFragDataLocation = _getProcAddress('glGetFragDataLocation')
	gl.glGetFragDataLocation.gl = gl
	gl.glGetFragDataLocation.__name__ = 'glGetFragDataLocation'
	gl.glGetFragDataLocation.errcheck = _errcheck
	gl.glGetFragDataLocation.restype = GLint
	gl.glGetFragDataLocation.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetFragDataLocationEXT = _getProcAddress('glGetFragDataLocationEXT')
	gl.glGetFragDataLocationEXT.gl = gl
	gl.glGetFragDataLocationEXT.__name__ = 'glGetFragDataLocationEXT'
	gl.glGetFragDataLocationEXT.errcheck = _errcheck
	gl.glGetFragDataLocationEXT.restype = GLint
	gl.glGetFragDataLocationEXT.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetFragmentLightfvSGIX = _getProcAddress('glGetFragmentLightfvSGIX')
	gl.glGetFragmentLightfvSGIX.gl = gl
	gl.glGetFragmentLightfvSGIX.__name__ = 'glGetFragmentLightfvSGIX'
	gl.glGetFragmentLightfvSGIX.errcheck = _errcheck
	gl.glGetFragmentLightfvSGIX.restype = GLvoid
	gl.glGetFragmentLightfvSGIX.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetFragmentLightivSGIX = _getProcAddress('glGetFragmentLightivSGIX')
	gl.glGetFragmentLightivSGIX.gl = gl
	gl.glGetFragmentLightivSGIX.__name__ = 'glGetFragmentLightivSGIX'
	gl.glGetFragmentLightivSGIX.errcheck = _errcheck
	gl.glGetFragmentLightivSGIX.restype = GLvoid
	gl.glGetFragmentLightivSGIX.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetFragmentMaterialfvSGIX = _getProcAddress('glGetFragmentMaterialfvSGIX')
	gl.glGetFragmentMaterialfvSGIX.gl = gl
	gl.glGetFragmentMaterialfvSGIX.__name__ = 'glGetFragmentMaterialfvSGIX'
	gl.glGetFragmentMaterialfvSGIX.errcheck = _errcheck
	gl.glGetFragmentMaterialfvSGIX.restype = GLvoid
	gl.glGetFragmentMaterialfvSGIX.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetFragmentMaterialivSGIX = _getProcAddress('glGetFragmentMaterialivSGIX')
	gl.glGetFragmentMaterialivSGIX.gl = gl
	gl.glGetFragmentMaterialivSGIX.__name__ = 'glGetFragmentMaterialivSGIX'
	gl.glGetFragmentMaterialivSGIX.errcheck = _errcheck
	gl.glGetFragmentMaterialivSGIX.restype = GLvoid
	gl.glGetFragmentMaterialivSGIX.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetFramebufferAttachmentParameteriv = _getProcAddress('glGetFramebufferAttachmentParameteriv')
	gl.glGetFramebufferAttachmentParameteriv.gl = gl
	gl.glGetFramebufferAttachmentParameteriv.__name__ = 'glGetFramebufferAttachmentParameteriv'
	gl.glGetFramebufferAttachmentParameteriv.errcheck = _errcheck
	gl.glGetFramebufferAttachmentParameteriv.restype = GLvoid
	gl.glGetFramebufferAttachmentParameteriv.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetFramebufferAttachmentParameterivEXT = _getProcAddress('glGetFramebufferAttachmentParameterivEXT')
	gl.glGetFramebufferAttachmentParameterivEXT.gl = gl
	gl.glGetFramebufferAttachmentParameterivEXT.__name__ = 'glGetFramebufferAttachmentParameterivEXT'
	gl.glGetFramebufferAttachmentParameterivEXT.errcheck = _errcheck
	gl.glGetFramebufferAttachmentParameterivEXT.restype = GLvoid
	gl.glGetFramebufferAttachmentParameterivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetFramebufferAttachmentParameterivOES = _getProcAddress('glGetFramebufferAttachmentParameterivOES')
	gl.glGetFramebufferAttachmentParameterivOES.gl = gl
	gl.glGetFramebufferAttachmentParameterivOES.__name__ = 'glGetFramebufferAttachmentParameterivOES'
	gl.glGetFramebufferAttachmentParameterivOES.errcheck = _errcheck
	gl.glGetFramebufferAttachmentParameterivOES.restype = GLvoid
	gl.glGetFramebufferAttachmentParameterivOES.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetFramebufferParameteriv = _getProcAddress('glGetFramebufferParameteriv')
	gl.glGetFramebufferParameteriv.gl = gl
	gl.glGetFramebufferParameteriv.__name__ = 'glGetFramebufferParameteriv'
	gl.glGetFramebufferParameteriv.errcheck = _errcheck
	gl.glGetFramebufferParameteriv.restype = GLvoid
	gl.glGetFramebufferParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetFramebufferParameterivEXT = _getProcAddress('glGetFramebufferParameterivEXT')
	gl.glGetFramebufferParameterivEXT.gl = gl
	gl.glGetFramebufferParameterivEXT.__name__ = 'glGetFramebufferParameterivEXT'
	gl.glGetFramebufferParameterivEXT.errcheck = _errcheck
	gl.glGetFramebufferParameterivEXT.restype = GLvoid
	gl.glGetFramebufferParameterivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetGraphicsResetStatus = _getProcAddress('glGetGraphicsResetStatus')
	gl.glGetGraphicsResetStatus.gl = gl
	gl.glGetGraphicsResetStatus.__name__ = 'glGetGraphicsResetStatus'
	gl.glGetGraphicsResetStatus.errcheck = _errcheck
	gl.glGetGraphicsResetStatus.restype = GLenum
	gl.glGetGraphicsResetStatus.argtypes = ()
	gl.glGetGraphicsResetStatusARB = _getProcAddress('glGetGraphicsResetStatusARB')
	gl.glGetGraphicsResetStatusARB.gl = gl
	gl.glGetGraphicsResetStatusARB.__name__ = 'glGetGraphicsResetStatusARB'
	gl.glGetGraphicsResetStatusARB.errcheck = _errcheck
	gl.glGetGraphicsResetStatusARB.restype = GLenum
	gl.glGetGraphicsResetStatusARB.argtypes = ()
	gl.glGetGraphicsResetStatusEXT = _getProcAddress('glGetGraphicsResetStatusEXT')
	gl.glGetGraphicsResetStatusEXT.gl = gl
	gl.glGetGraphicsResetStatusEXT.__name__ = 'glGetGraphicsResetStatusEXT'
	gl.glGetGraphicsResetStatusEXT.errcheck = _errcheck
	gl.glGetGraphicsResetStatusEXT.restype = GLenum
	gl.glGetGraphicsResetStatusEXT.argtypes = ()
	gl.glGetGraphicsResetStatusKHR = _getProcAddress('glGetGraphicsResetStatusKHR')
	gl.glGetGraphicsResetStatusKHR.gl = gl
	gl.glGetGraphicsResetStatusKHR.__name__ = 'glGetGraphicsResetStatusKHR'
	gl.glGetGraphicsResetStatusKHR.errcheck = _errcheck
	gl.glGetGraphicsResetStatusKHR.restype = GLenum
	gl.glGetGraphicsResetStatusKHR.argtypes = ()
	gl.glGetHandleARB = _getProcAddress('glGetHandleARB')
	gl.glGetHandleARB.gl = gl
	gl.glGetHandleARB.__name__ = 'glGetHandleARB'
	gl.glGetHandleARB.errcheck = _errcheck
	gl.glGetHandleARB.restype = GLhandleARB
	gl.glGetHandleARB.argtypes = (GLenum,)
	gl.glGetHistogram = _getProcAddress('glGetHistogram')
	gl.glGetHistogram.gl = gl
	gl.glGetHistogram.__name__ = 'glGetHistogram'
	gl.glGetHistogram.errcheck = _errcheck
	gl.glGetHistogram.restype = GLvoid
	gl.glGetHistogram.argtypes = (GLenum, GLboolean, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetHistogramEXT = _getProcAddress('glGetHistogramEXT')
	gl.glGetHistogramEXT.gl = gl
	gl.glGetHistogramEXT.__name__ = 'glGetHistogramEXT'
	gl.glGetHistogramEXT.errcheck = _errcheck
	gl.glGetHistogramEXT.restype = GLvoid
	gl.glGetHistogramEXT.argtypes = (GLenum, GLboolean, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetHistogramParameterfv = _getProcAddress('glGetHistogramParameterfv')
	gl.glGetHistogramParameterfv.gl = gl
	gl.glGetHistogramParameterfv.__name__ = 'glGetHistogramParameterfv'
	gl.glGetHistogramParameterfv.errcheck = _errcheck
	gl.glGetHistogramParameterfv.restype = GLvoid
	gl.glGetHistogramParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetHistogramParameterfvEXT = _getProcAddress('glGetHistogramParameterfvEXT')
	gl.glGetHistogramParameterfvEXT.gl = gl
	gl.glGetHistogramParameterfvEXT.__name__ = 'glGetHistogramParameterfvEXT'
	gl.glGetHistogramParameterfvEXT.errcheck = _errcheck
	gl.glGetHistogramParameterfvEXT.restype = GLvoid
	gl.glGetHistogramParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetHistogramParameteriv = _getProcAddress('glGetHistogramParameteriv')
	gl.glGetHistogramParameteriv.gl = gl
	gl.glGetHistogramParameteriv.__name__ = 'glGetHistogramParameteriv'
	gl.glGetHistogramParameteriv.errcheck = _errcheck
	gl.glGetHistogramParameteriv.restype = GLvoid
	gl.glGetHistogramParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetHistogramParameterivEXT = _getProcAddress('glGetHistogramParameterivEXT')
	gl.glGetHistogramParameterivEXT.gl = gl
	gl.glGetHistogramParameterivEXT.__name__ = 'glGetHistogramParameterivEXT'
	gl.glGetHistogramParameterivEXT.errcheck = _errcheck
	gl.glGetHistogramParameterivEXT.restype = GLvoid
	gl.glGetHistogramParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetHistogramParameterxvOES = _getProcAddress('glGetHistogramParameterxvOES')
	gl.glGetHistogramParameterxvOES.gl = gl
	gl.glGetHistogramParameterxvOES.__name__ = 'glGetHistogramParameterxvOES'
	gl.glGetHistogramParameterxvOES.errcheck = _errcheck
	gl.glGetHistogramParameterxvOES.restype = GLvoid
	gl.glGetHistogramParameterxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetImageHandleARB = _getProcAddress('glGetImageHandleARB')
	gl.glGetImageHandleARB.gl = gl
	gl.glGetImageHandleARB.__name__ = 'glGetImageHandleARB'
	gl.glGetImageHandleARB.errcheck = _errcheck
	gl.glGetImageHandleARB.restype = GLuint64
	gl.glGetImageHandleARB.argtypes = (GLuint, GLint, GLboolean, GLint, GLenum,)
	gl.glGetImageHandleNV = _getProcAddress('glGetImageHandleNV')
	gl.glGetImageHandleNV.gl = gl
	gl.glGetImageHandleNV.__name__ = 'glGetImageHandleNV'
	gl.glGetImageHandleNV.errcheck = _errcheck
	gl.glGetImageHandleNV.restype = GLuint64
	gl.glGetImageHandleNV.argtypes = (GLuint, GLint, GLboolean, GLint, GLenum,)
	gl.glGetImageTransformParameterfvHP = _getProcAddress('glGetImageTransformParameterfvHP')
	gl.glGetImageTransformParameterfvHP.gl = gl
	gl.glGetImageTransformParameterfvHP.__name__ = 'glGetImageTransformParameterfvHP'
	gl.glGetImageTransformParameterfvHP.errcheck = _errcheck
	gl.glGetImageTransformParameterfvHP.restype = GLvoid
	gl.glGetImageTransformParameterfvHP.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetImageTransformParameterivHP = _getProcAddress('glGetImageTransformParameterivHP')
	gl.glGetImageTransformParameterivHP.gl = gl
	gl.glGetImageTransformParameterivHP.__name__ = 'glGetImageTransformParameterivHP'
	gl.glGetImageTransformParameterivHP.errcheck = _errcheck
	gl.glGetImageTransformParameterivHP.restype = GLvoid
	gl.glGetImageTransformParameterivHP.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetInfoLogARB = _getProcAddress('glGetInfoLogARB')
	gl.glGetInfoLogARB.gl = gl
	gl.glGetInfoLogARB.__name__ = 'glGetInfoLogARB'
	gl.glGetInfoLogARB.errcheck = _errcheck
	gl.glGetInfoLogARB.restype = GLvoid
	gl.glGetInfoLogARB.argtypes = (GLhandleARB, GLsizei, POINTER(GLsizei), POINTER(GLcharARB),)
	gl.glGetInstrumentsSGIX = _getProcAddress('glGetInstrumentsSGIX')
	gl.glGetInstrumentsSGIX.gl = gl
	gl.glGetInstrumentsSGIX.__name__ = 'glGetInstrumentsSGIX'
	gl.glGetInstrumentsSGIX.errcheck = _errcheck
	gl.glGetInstrumentsSGIX.restype = GLint
	gl.glGetInstrumentsSGIX.argtypes = ()
	gl.glGetInteger64i_v = _getProcAddress('glGetInteger64i_v')
	gl.glGetInteger64i_v.gl = gl
	gl.glGetInteger64i_v.__name__ = 'glGetInteger64i_v'
	gl.glGetInteger64i_v.errcheck = _errcheck
	gl.glGetInteger64i_v.restype = GLvoid
	gl.glGetInteger64i_v.argtypes = (GLenum, GLuint, POINTER(GLint64),)
	gl.glGetInteger64v = _getProcAddress('glGetInteger64v')
	gl.glGetInteger64v.gl = gl
	gl.glGetInteger64v.__name__ = 'glGetInteger64v'
	gl.glGetInteger64v.errcheck = _errcheck
	gl.glGetInteger64v.restype = GLvoid
	gl.glGetInteger64v.argtypes = (GLenum, POINTER(GLint64),)
	gl.glGetInteger64vAPPLE = _getProcAddress('glGetInteger64vAPPLE')
	gl.glGetInteger64vAPPLE.gl = gl
	gl.glGetInteger64vAPPLE.__name__ = 'glGetInteger64vAPPLE'
	gl.glGetInteger64vAPPLE.errcheck = _errcheck
	gl.glGetInteger64vAPPLE.restype = GLvoid
	gl.glGetInteger64vAPPLE.argtypes = (GLenum, POINTER(GLint64),)
	gl.glGetIntegerIndexedvEXT = _getProcAddress('glGetIntegerIndexedvEXT')
	gl.glGetIntegerIndexedvEXT.gl = gl
	gl.glGetIntegerIndexedvEXT.__name__ = 'glGetIntegerIndexedvEXT'
	gl.glGetIntegerIndexedvEXT.errcheck = _errcheck
	gl.glGetIntegerIndexedvEXT.restype = GLvoid
	gl.glGetIntegerIndexedvEXT.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glGetIntegeri_v = _getProcAddress('glGetIntegeri_v')
	gl.glGetIntegeri_v.gl = gl
	gl.glGetIntegeri_v.__name__ = 'glGetIntegeri_v'
	gl.glGetIntegeri_v.errcheck = _errcheck
	gl.glGetIntegeri_v.restype = GLvoid
	gl.glGetIntegeri_v.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glGetIntegeri_vEXT = _getProcAddress('glGetIntegeri_vEXT')
	gl.glGetIntegeri_vEXT.gl = gl
	gl.glGetIntegeri_vEXT.__name__ = 'glGetIntegeri_vEXT'
	gl.glGetIntegeri_vEXT.errcheck = _errcheck
	gl.glGetIntegeri_vEXT.restype = GLvoid
	gl.glGetIntegeri_vEXT.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glGetIntegerui64i_vNV = _getProcAddress('glGetIntegerui64i_vNV')
	gl.glGetIntegerui64i_vNV.gl = gl
	gl.glGetIntegerui64i_vNV.__name__ = 'glGetIntegerui64i_vNV'
	gl.glGetIntegerui64i_vNV.errcheck = _errcheck
	gl.glGetIntegerui64i_vNV.restype = GLvoid
	gl.glGetIntegerui64i_vNV.argtypes = (GLenum, GLuint, POINTER(GLuint64EXT),)
	gl.glGetIntegerui64vNV = _getProcAddress('glGetIntegerui64vNV')
	gl.glGetIntegerui64vNV.gl = gl
	gl.glGetIntegerui64vNV.__name__ = 'glGetIntegerui64vNV'
	gl.glGetIntegerui64vNV.errcheck = _errcheck
	gl.glGetIntegerui64vNV.restype = GLvoid
	gl.glGetIntegerui64vNV.argtypes = (GLenum, POINTER(GLuint64EXT),)
	gl.glGetIntegerv = _getProcAddress('glGetIntegerv')
	gl.glGetIntegerv.gl = gl
	gl.glGetIntegerv.__name__ = 'glGetIntegerv'
	gl.glGetIntegerv.errcheck = _errcheck
	gl.glGetIntegerv.restype = GLvoid
	gl.glGetIntegerv.argtypes = (GLenum, POINTER(GLint),)
	gl.glGetInternalformati64v = _getProcAddress('glGetInternalformati64v')
	gl.glGetInternalformati64v.gl = gl
	gl.glGetInternalformati64v.__name__ = 'glGetInternalformati64v'
	gl.glGetInternalformati64v.errcheck = _errcheck
	gl.glGetInternalformati64v.restype = GLvoid
	gl.glGetInternalformati64v.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLint64),)
	gl.glGetInternalformativ = _getProcAddress('glGetInternalformativ')
	gl.glGetInternalformativ.gl = gl
	gl.glGetInternalformativ.__name__ = 'glGetInternalformativ'
	gl.glGetInternalformativ.errcheck = _errcheck
	gl.glGetInternalformativ.restype = GLvoid
	gl.glGetInternalformativ.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLint),)
	gl.glGetInvariantBooleanvEXT = _getProcAddress('glGetInvariantBooleanvEXT')
	gl.glGetInvariantBooleanvEXT.gl = gl
	gl.glGetInvariantBooleanvEXT.__name__ = 'glGetInvariantBooleanvEXT'
	gl.glGetInvariantBooleanvEXT.errcheck = _errcheck
	gl.glGetInvariantBooleanvEXT.restype = GLvoid
	gl.glGetInvariantBooleanvEXT.argtypes = (GLuint, GLenum, POINTER(GLboolean),)
	gl.glGetInvariantFloatvEXT = _getProcAddress('glGetInvariantFloatvEXT')
	gl.glGetInvariantFloatvEXT.gl = gl
	gl.glGetInvariantFloatvEXT.__name__ = 'glGetInvariantFloatvEXT'
	gl.glGetInvariantFloatvEXT.errcheck = _errcheck
	gl.glGetInvariantFloatvEXT.restype = GLvoid
	gl.glGetInvariantFloatvEXT.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetInvariantIntegervEXT = _getProcAddress('glGetInvariantIntegervEXT')
	gl.glGetInvariantIntegervEXT.gl = gl
	gl.glGetInvariantIntegervEXT.__name__ = 'glGetInvariantIntegervEXT'
	gl.glGetInvariantIntegervEXT.errcheck = _errcheck
	gl.glGetInvariantIntegervEXT.restype = GLvoid
	gl.glGetInvariantIntegervEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetLightfv = _getProcAddress('glGetLightfv')
	gl.glGetLightfv.gl = gl
	gl.glGetLightfv.__name__ = 'glGetLightfv'
	gl.glGetLightfv.errcheck = _errcheck
	gl.glGetLightfv.restype = GLvoid
	gl.glGetLightfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetLightiv = _getProcAddress('glGetLightiv')
	gl.glGetLightiv.gl = gl
	gl.glGetLightiv.__name__ = 'glGetLightiv'
	gl.glGetLightiv.errcheck = _errcheck
	gl.glGetLightiv.restype = GLvoid
	gl.glGetLightiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetLightxOES = _getProcAddress('glGetLightxOES')
	gl.glGetLightxOES.gl = gl
	gl.glGetLightxOES.__name__ = 'glGetLightxOES'
	gl.glGetLightxOES.errcheck = _errcheck
	gl.glGetLightxOES.restype = GLvoid
	gl.glGetLightxOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetLightxv = _getProcAddress('glGetLightxv')
	gl.glGetLightxv.gl = gl
	gl.glGetLightxv.__name__ = 'glGetLightxv'
	gl.glGetLightxv.errcheck = _errcheck
	gl.glGetLightxv.restype = GLvoid
	gl.glGetLightxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetLightxvOES = _getProcAddress('glGetLightxvOES')
	gl.glGetLightxvOES.gl = gl
	gl.glGetLightxvOES.__name__ = 'glGetLightxvOES'
	gl.glGetLightxvOES.errcheck = _errcheck
	gl.glGetLightxvOES.restype = GLvoid
	gl.glGetLightxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetListParameterfvSGIX = _getProcAddress('glGetListParameterfvSGIX')
	gl.glGetListParameterfvSGIX.gl = gl
	gl.glGetListParameterfvSGIX.__name__ = 'glGetListParameterfvSGIX'
	gl.glGetListParameterfvSGIX.errcheck = _errcheck
	gl.glGetListParameterfvSGIX.restype = GLvoid
	gl.glGetListParameterfvSGIX.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetListParameterivSGIX = _getProcAddress('glGetListParameterivSGIX')
	gl.glGetListParameterivSGIX.gl = gl
	gl.glGetListParameterivSGIX.__name__ = 'glGetListParameterivSGIX'
	gl.glGetListParameterivSGIX.errcheck = _errcheck
	gl.glGetListParameterivSGIX.restype = GLvoid
	gl.glGetListParameterivSGIX.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetLocalConstantBooleanvEXT = _getProcAddress('glGetLocalConstantBooleanvEXT')
	gl.glGetLocalConstantBooleanvEXT.gl = gl
	gl.glGetLocalConstantBooleanvEXT.__name__ = 'glGetLocalConstantBooleanvEXT'
	gl.glGetLocalConstantBooleanvEXT.errcheck = _errcheck
	gl.glGetLocalConstantBooleanvEXT.restype = GLvoid
	gl.glGetLocalConstantBooleanvEXT.argtypes = (GLuint, GLenum, POINTER(GLboolean),)
	gl.glGetLocalConstantFloatvEXT = _getProcAddress('glGetLocalConstantFloatvEXT')
	gl.glGetLocalConstantFloatvEXT.gl = gl
	gl.glGetLocalConstantFloatvEXT.__name__ = 'glGetLocalConstantFloatvEXT'
	gl.glGetLocalConstantFloatvEXT.errcheck = _errcheck
	gl.glGetLocalConstantFloatvEXT.restype = GLvoid
	gl.glGetLocalConstantFloatvEXT.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetLocalConstantIntegervEXT = _getProcAddress('glGetLocalConstantIntegervEXT')
	gl.glGetLocalConstantIntegervEXT.gl = gl
	gl.glGetLocalConstantIntegervEXT.__name__ = 'glGetLocalConstantIntegervEXT'
	gl.glGetLocalConstantIntegervEXT.errcheck = _errcheck
	gl.glGetLocalConstantIntegervEXT.restype = GLvoid
	gl.glGetLocalConstantIntegervEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetMapAttribParameterfvNV = _getProcAddress('glGetMapAttribParameterfvNV')
	gl.glGetMapAttribParameterfvNV.gl = gl
	gl.glGetMapAttribParameterfvNV.__name__ = 'glGetMapAttribParameterfvNV'
	gl.glGetMapAttribParameterfvNV.errcheck = _errcheck
	gl.glGetMapAttribParameterfvNV.restype = GLvoid
	gl.glGetMapAttribParameterfvNV.argtypes = (GLenum, GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetMapAttribParameterivNV = _getProcAddress('glGetMapAttribParameterivNV')
	gl.glGetMapAttribParameterivNV.gl = gl
	gl.glGetMapAttribParameterivNV.__name__ = 'glGetMapAttribParameterivNV'
	gl.glGetMapAttribParameterivNV.errcheck = _errcheck
	gl.glGetMapAttribParameterivNV.restype = GLvoid
	gl.glGetMapAttribParameterivNV.argtypes = (GLenum, GLuint, GLenum, POINTER(GLint),)
	gl.glGetMapControlPointsNV = _getProcAddress('glGetMapControlPointsNV')
	gl.glGetMapControlPointsNV.gl = gl
	gl.glGetMapControlPointsNV.__name__ = 'glGetMapControlPointsNV'
	gl.glGetMapControlPointsNV.errcheck = _errcheck
	gl.glGetMapControlPointsNV.restype = GLvoid
	gl.glGetMapControlPointsNV.argtypes = (GLenum, GLuint, GLenum, GLsizei, GLsizei, GLboolean, POINTER(GLvoid),)
	gl.glGetMapParameterfvNV = _getProcAddress('glGetMapParameterfvNV')
	gl.glGetMapParameterfvNV.gl = gl
	gl.glGetMapParameterfvNV.__name__ = 'glGetMapParameterfvNV'
	gl.glGetMapParameterfvNV.errcheck = _errcheck
	gl.glGetMapParameterfvNV.restype = GLvoid
	gl.glGetMapParameterfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMapParameterivNV = _getProcAddress('glGetMapParameterivNV')
	gl.glGetMapParameterivNV.gl = gl
	gl.glGetMapParameterivNV.__name__ = 'glGetMapParameterivNV'
	gl.glGetMapParameterivNV.errcheck = _errcheck
	gl.glGetMapParameterivNV.restype = GLvoid
	gl.glGetMapParameterivNV.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetMapdv = _getProcAddress('glGetMapdv')
	gl.glGetMapdv.gl = gl
	gl.glGetMapdv.__name__ = 'glGetMapdv'
	gl.glGetMapdv.errcheck = _errcheck
	gl.glGetMapdv.restype = GLvoid
	gl.glGetMapdv.argtypes = (GLenum, GLenum, POINTER(GLdouble),)
	gl.glGetMapfv = _getProcAddress('glGetMapfv')
	gl.glGetMapfv.gl = gl
	gl.glGetMapfv.__name__ = 'glGetMapfv'
	gl.glGetMapfv.errcheck = _errcheck
	gl.glGetMapfv.restype = GLvoid
	gl.glGetMapfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMapiv = _getProcAddress('glGetMapiv')
	gl.glGetMapiv.gl = gl
	gl.glGetMapiv.__name__ = 'glGetMapiv'
	gl.glGetMapiv.errcheck = _errcheck
	gl.glGetMapiv.restype = GLvoid
	gl.glGetMapiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetMapxvOES = _getProcAddress('glGetMapxvOES')
	gl.glGetMapxvOES.gl = gl
	gl.glGetMapxvOES.__name__ = 'glGetMapxvOES'
	gl.glGetMapxvOES.errcheck = _errcheck
	gl.glGetMapxvOES.restype = GLvoid
	gl.glGetMapxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetMaterialfv = _getProcAddress('glGetMaterialfv')
	gl.glGetMaterialfv.gl = gl
	gl.glGetMaterialfv.__name__ = 'glGetMaterialfv'
	gl.glGetMaterialfv.errcheck = _errcheck
	gl.glGetMaterialfv.restype = GLvoid
	gl.glGetMaterialfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMaterialiv = _getProcAddress('glGetMaterialiv')
	gl.glGetMaterialiv.gl = gl
	gl.glGetMaterialiv.__name__ = 'glGetMaterialiv'
	gl.glGetMaterialiv.errcheck = _errcheck
	gl.glGetMaterialiv.restype = GLvoid
	gl.glGetMaterialiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetMaterialxOES = _getProcAddress('glGetMaterialxOES')
	gl.glGetMaterialxOES.gl = gl
	gl.glGetMaterialxOES.__name__ = 'glGetMaterialxOES'
	gl.glGetMaterialxOES.errcheck = _errcheck
	gl.glGetMaterialxOES.restype = GLvoid
	gl.glGetMaterialxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glGetMaterialxv = _getProcAddress('glGetMaterialxv')
	gl.glGetMaterialxv.gl = gl
	gl.glGetMaterialxv.__name__ = 'glGetMaterialxv'
	gl.glGetMaterialxv.errcheck = _errcheck
	gl.glGetMaterialxv.restype = GLvoid
	gl.glGetMaterialxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetMaterialxvOES = _getProcAddress('glGetMaterialxvOES')
	gl.glGetMaterialxvOES.gl = gl
	gl.glGetMaterialxvOES.__name__ = 'glGetMaterialxvOES'
	gl.glGetMaterialxvOES.errcheck = _errcheck
	gl.glGetMaterialxvOES.restype = GLvoid
	gl.glGetMaterialxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetMinmax = _getProcAddress('glGetMinmax')
	gl.glGetMinmax.gl = gl
	gl.glGetMinmax.__name__ = 'glGetMinmax'
	gl.glGetMinmax.errcheck = _errcheck
	gl.glGetMinmax.restype = GLvoid
	gl.glGetMinmax.argtypes = (GLenum, GLboolean, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetMinmaxEXT = _getProcAddress('glGetMinmaxEXT')
	gl.glGetMinmaxEXT.gl = gl
	gl.glGetMinmaxEXT.__name__ = 'glGetMinmaxEXT'
	gl.glGetMinmaxEXT.errcheck = _errcheck
	gl.glGetMinmaxEXT.restype = GLvoid
	gl.glGetMinmaxEXT.argtypes = (GLenum, GLboolean, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetMinmaxParameterfv = _getProcAddress('glGetMinmaxParameterfv')
	gl.glGetMinmaxParameterfv.gl = gl
	gl.glGetMinmaxParameterfv.__name__ = 'glGetMinmaxParameterfv'
	gl.glGetMinmaxParameterfv.errcheck = _errcheck
	gl.glGetMinmaxParameterfv.restype = GLvoid
	gl.glGetMinmaxParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMinmaxParameterfvEXT = _getProcAddress('glGetMinmaxParameterfvEXT')
	gl.glGetMinmaxParameterfvEXT.gl = gl
	gl.glGetMinmaxParameterfvEXT.__name__ = 'glGetMinmaxParameterfvEXT'
	gl.glGetMinmaxParameterfvEXT.errcheck = _errcheck
	gl.glGetMinmaxParameterfvEXT.restype = GLvoid
	gl.glGetMinmaxParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMinmaxParameteriv = _getProcAddress('glGetMinmaxParameteriv')
	gl.glGetMinmaxParameteriv.gl = gl
	gl.glGetMinmaxParameteriv.__name__ = 'glGetMinmaxParameteriv'
	gl.glGetMinmaxParameteriv.errcheck = _errcheck
	gl.glGetMinmaxParameteriv.restype = GLvoid
	gl.glGetMinmaxParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetMinmaxParameterivEXT = _getProcAddress('glGetMinmaxParameterivEXT')
	gl.glGetMinmaxParameterivEXT.gl = gl
	gl.glGetMinmaxParameterivEXT.__name__ = 'glGetMinmaxParameterivEXT'
	gl.glGetMinmaxParameterivEXT.errcheck = _errcheck
	gl.glGetMinmaxParameterivEXT.restype = GLvoid
	gl.glGetMinmaxParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetMultiTexEnvfvEXT = _getProcAddress('glGetMultiTexEnvfvEXT')
	gl.glGetMultiTexEnvfvEXT.gl = gl
	gl.glGetMultiTexEnvfvEXT.__name__ = 'glGetMultiTexEnvfvEXT'
	gl.glGetMultiTexEnvfvEXT.errcheck = _errcheck
	gl.glGetMultiTexEnvfvEXT.restype = GLvoid
	gl.glGetMultiTexEnvfvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMultiTexEnvivEXT = _getProcAddress('glGetMultiTexEnvivEXT')
	gl.glGetMultiTexEnvivEXT.gl = gl
	gl.glGetMultiTexEnvivEXT.__name__ = 'glGetMultiTexEnvivEXT'
	gl.glGetMultiTexEnvivEXT.errcheck = _errcheck
	gl.glGetMultiTexEnvivEXT.restype = GLvoid
	gl.glGetMultiTexEnvivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetMultiTexGendvEXT = _getProcAddress('glGetMultiTexGendvEXT')
	gl.glGetMultiTexGendvEXT.gl = gl
	gl.glGetMultiTexGendvEXT.__name__ = 'glGetMultiTexGendvEXT'
	gl.glGetMultiTexGendvEXT.errcheck = _errcheck
	gl.glGetMultiTexGendvEXT.restype = GLvoid
	gl.glGetMultiTexGendvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLdouble),)
	gl.glGetMultiTexGenfvEXT = _getProcAddress('glGetMultiTexGenfvEXT')
	gl.glGetMultiTexGenfvEXT.gl = gl
	gl.glGetMultiTexGenfvEXT.__name__ = 'glGetMultiTexGenfvEXT'
	gl.glGetMultiTexGenfvEXT.errcheck = _errcheck
	gl.glGetMultiTexGenfvEXT.restype = GLvoid
	gl.glGetMultiTexGenfvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMultiTexGenivEXT = _getProcAddress('glGetMultiTexGenivEXT')
	gl.glGetMultiTexGenivEXT.gl = gl
	gl.glGetMultiTexGenivEXT.__name__ = 'glGetMultiTexGenivEXT'
	gl.glGetMultiTexGenivEXT.errcheck = _errcheck
	gl.glGetMultiTexGenivEXT.restype = GLvoid
	gl.glGetMultiTexGenivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetMultiTexImageEXT = _getProcAddress('glGetMultiTexImageEXT')
	gl.glGetMultiTexImageEXT.gl = gl
	gl.glGetMultiTexImageEXT.__name__ = 'glGetMultiTexImageEXT'
	gl.glGetMultiTexImageEXT.errcheck = _errcheck
	gl.glGetMultiTexImageEXT.restype = GLvoid
	gl.glGetMultiTexImageEXT.argtypes = (GLenum, GLenum, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetMultiTexLevelParameterfvEXT = _getProcAddress('glGetMultiTexLevelParameterfvEXT')
	gl.glGetMultiTexLevelParameterfvEXT.gl = gl
	gl.glGetMultiTexLevelParameterfvEXT.__name__ = 'glGetMultiTexLevelParameterfvEXT'
	gl.glGetMultiTexLevelParameterfvEXT.errcheck = _errcheck
	gl.glGetMultiTexLevelParameterfvEXT.restype = GLvoid
	gl.glGetMultiTexLevelParameterfvEXT.argtypes = (GLenum, GLenum, GLint, GLenum, POINTER(GLfloat),)
	gl.glGetMultiTexLevelParameterivEXT = _getProcAddress('glGetMultiTexLevelParameterivEXT')
	gl.glGetMultiTexLevelParameterivEXT.gl = gl
	gl.glGetMultiTexLevelParameterivEXT.__name__ = 'glGetMultiTexLevelParameterivEXT'
	gl.glGetMultiTexLevelParameterivEXT.errcheck = _errcheck
	gl.glGetMultiTexLevelParameterivEXT.restype = GLvoid
	gl.glGetMultiTexLevelParameterivEXT.argtypes = (GLenum, GLenum, GLint, GLenum, POINTER(GLint),)
	gl.glGetMultiTexParameterIivEXT = _getProcAddress('glGetMultiTexParameterIivEXT')
	gl.glGetMultiTexParameterIivEXT.gl = gl
	gl.glGetMultiTexParameterIivEXT.__name__ = 'glGetMultiTexParameterIivEXT'
	gl.glGetMultiTexParameterIivEXT.errcheck = _errcheck
	gl.glGetMultiTexParameterIivEXT.restype = GLvoid
	gl.glGetMultiTexParameterIivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetMultiTexParameterIuivEXT = _getProcAddress('glGetMultiTexParameterIuivEXT')
	gl.glGetMultiTexParameterIuivEXT.gl = gl
	gl.glGetMultiTexParameterIuivEXT.__name__ = 'glGetMultiTexParameterIuivEXT'
	gl.glGetMultiTexParameterIuivEXT.errcheck = _errcheck
	gl.glGetMultiTexParameterIuivEXT.restype = GLvoid
	gl.glGetMultiTexParameterIuivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLuint),)
	gl.glGetMultiTexParameterfvEXT = _getProcAddress('glGetMultiTexParameterfvEXT')
	gl.glGetMultiTexParameterfvEXT.gl = gl
	gl.glGetMultiTexParameterfvEXT.__name__ = 'glGetMultiTexParameterfvEXT'
	gl.glGetMultiTexParameterfvEXT.errcheck = _errcheck
	gl.glGetMultiTexParameterfvEXT.restype = GLvoid
	gl.glGetMultiTexParameterfvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetMultiTexParameterivEXT = _getProcAddress('glGetMultiTexParameterivEXT')
	gl.glGetMultiTexParameterivEXT.gl = gl
	gl.glGetMultiTexParameterivEXT.__name__ = 'glGetMultiTexParameterivEXT'
	gl.glGetMultiTexParameterivEXT.errcheck = _errcheck
	gl.glGetMultiTexParameterivEXT.restype = GLvoid
	gl.glGetMultiTexParameterivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glGetMultisamplefv = _getProcAddress('glGetMultisamplefv')
	gl.glGetMultisamplefv.gl = gl
	gl.glGetMultisamplefv.__name__ = 'glGetMultisamplefv'
	gl.glGetMultisamplefv.errcheck = _errcheck
	gl.glGetMultisamplefv.restype = GLvoid
	gl.glGetMultisamplefv.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetMultisamplefvNV = _getProcAddress('glGetMultisamplefvNV')
	gl.glGetMultisamplefvNV.gl = gl
	gl.glGetMultisamplefvNV.__name__ = 'glGetMultisamplefvNV'
	gl.glGetMultisamplefvNV.errcheck = _errcheck
	gl.glGetMultisamplefvNV.restype = GLvoid
	gl.glGetMultisamplefvNV.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetNamedBufferParameteri64v = _getProcAddress('glGetNamedBufferParameteri64v')
	gl.glGetNamedBufferParameteri64v.gl = gl
	gl.glGetNamedBufferParameteri64v.__name__ = 'glGetNamedBufferParameteri64v'
	gl.glGetNamedBufferParameteri64v.errcheck = _errcheck
	gl.glGetNamedBufferParameteri64v.restype = GLvoid
	gl.glGetNamedBufferParameteri64v.argtypes = (GLuint, GLenum, POINTER(GLint64),)
	gl.glGetNamedBufferParameteriv = _getProcAddress('glGetNamedBufferParameteriv')
	gl.glGetNamedBufferParameteriv.gl = gl
	gl.glGetNamedBufferParameteriv.__name__ = 'glGetNamedBufferParameteriv'
	gl.glGetNamedBufferParameteriv.errcheck = _errcheck
	gl.glGetNamedBufferParameteriv.restype = GLvoid
	gl.glGetNamedBufferParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetNamedBufferParameterivEXT = _getProcAddress('glGetNamedBufferParameterivEXT')
	gl.glGetNamedBufferParameterivEXT.gl = gl
	gl.glGetNamedBufferParameterivEXT.__name__ = 'glGetNamedBufferParameterivEXT'
	gl.glGetNamedBufferParameterivEXT.errcheck = _errcheck
	gl.glGetNamedBufferParameterivEXT.restype = GLvoid
	gl.glGetNamedBufferParameterivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetNamedBufferParameterui64vNV = _getProcAddress('glGetNamedBufferParameterui64vNV')
	gl.glGetNamedBufferParameterui64vNV.gl = gl
	gl.glGetNamedBufferParameterui64vNV.__name__ = 'glGetNamedBufferParameterui64vNV'
	gl.glGetNamedBufferParameterui64vNV.errcheck = _errcheck
	gl.glGetNamedBufferParameterui64vNV.restype = GLvoid
	gl.glGetNamedBufferParameterui64vNV.argtypes = (GLuint, GLenum, POINTER(GLuint64EXT),)
	gl.glGetNamedBufferPointerv = _getProcAddress('glGetNamedBufferPointerv')
	gl.glGetNamedBufferPointerv.gl = gl
	gl.glGetNamedBufferPointerv.__name__ = 'glGetNamedBufferPointerv'
	gl.glGetNamedBufferPointerv.errcheck = _errcheck
	gl.glGetNamedBufferPointerv.restype = GLvoid
	gl.glGetNamedBufferPointerv.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetNamedBufferPointervEXT = _getProcAddress('glGetNamedBufferPointervEXT')
	gl.glGetNamedBufferPointervEXT.gl = gl
	gl.glGetNamedBufferPointervEXT.__name__ = 'glGetNamedBufferPointervEXT'
	gl.glGetNamedBufferPointervEXT.errcheck = _errcheck
	gl.glGetNamedBufferPointervEXT.restype = GLvoid
	gl.glGetNamedBufferPointervEXT.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetNamedBufferSubData = _getProcAddress('glGetNamedBufferSubData')
	gl.glGetNamedBufferSubData.gl = gl
	gl.glGetNamedBufferSubData.__name__ = 'glGetNamedBufferSubData'
	gl.glGetNamedBufferSubData.errcheck = _errcheck
	gl.glGetNamedBufferSubData.restype = GLvoid
	gl.glGetNamedBufferSubData.argtypes = (GLuint, GLintptr, GLsizei, POINTER(GLvoid),)
	gl.glGetNamedBufferSubDataEXT = _getProcAddress('glGetNamedBufferSubDataEXT')
	gl.glGetNamedBufferSubDataEXT.gl = gl
	gl.glGetNamedBufferSubDataEXT.__name__ = 'glGetNamedBufferSubDataEXT'
	gl.glGetNamedBufferSubDataEXT.errcheck = _errcheck
	gl.glGetNamedBufferSubDataEXT.restype = GLvoid
	gl.glGetNamedBufferSubDataEXT.argtypes = (GLuint, GLintptr, GLsizeiptr, POINTER(GLvoid),)
	gl.glGetNamedFramebufferAttachmentParameteriv = _getProcAddress('glGetNamedFramebufferAttachmentParameteriv')
	gl.glGetNamedFramebufferAttachmentParameteriv.gl = gl
	gl.glGetNamedFramebufferAttachmentParameteriv.__name__ = 'glGetNamedFramebufferAttachmentParameteriv'
	gl.glGetNamedFramebufferAttachmentParameteriv.errcheck = _errcheck
	gl.glGetNamedFramebufferAttachmentParameteriv.restype = GLvoid
	gl.glGetNamedFramebufferAttachmentParameteriv.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetNamedFramebufferAttachmentParameterivEXT = _getProcAddress('glGetNamedFramebufferAttachmentParameterivEXT')
	gl.glGetNamedFramebufferAttachmentParameterivEXT.gl = gl
	gl.glGetNamedFramebufferAttachmentParameterivEXT.__name__ = 'glGetNamedFramebufferAttachmentParameterivEXT'
	gl.glGetNamedFramebufferAttachmentParameterivEXT.errcheck = _errcheck
	gl.glGetNamedFramebufferAttachmentParameterivEXT.restype = GLvoid
	gl.glGetNamedFramebufferAttachmentParameterivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetNamedFramebufferParameteriv = _getProcAddress('glGetNamedFramebufferParameteriv')
	gl.glGetNamedFramebufferParameteriv.gl = gl
	gl.glGetNamedFramebufferParameteriv.__name__ = 'glGetNamedFramebufferParameteriv'
	gl.glGetNamedFramebufferParameteriv.errcheck = _errcheck
	gl.glGetNamedFramebufferParameteriv.restype = GLvoid
	gl.glGetNamedFramebufferParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetNamedFramebufferParameterivEXT = _getProcAddress('glGetNamedFramebufferParameterivEXT')
	gl.glGetNamedFramebufferParameterivEXT.gl = gl
	gl.glGetNamedFramebufferParameterivEXT.__name__ = 'glGetNamedFramebufferParameterivEXT'
	gl.glGetNamedFramebufferParameterivEXT.errcheck = _errcheck
	gl.glGetNamedFramebufferParameterivEXT.restype = GLvoid
	gl.glGetNamedFramebufferParameterivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetNamedProgramLocalParameterIivEXT = _getProcAddress('glGetNamedProgramLocalParameterIivEXT')
	gl.glGetNamedProgramLocalParameterIivEXT.gl = gl
	gl.glGetNamedProgramLocalParameterIivEXT.__name__ = 'glGetNamedProgramLocalParameterIivEXT'
	gl.glGetNamedProgramLocalParameterIivEXT.errcheck = _errcheck
	gl.glGetNamedProgramLocalParameterIivEXT.restype = GLvoid
	gl.glGetNamedProgramLocalParameterIivEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLint),)
	gl.glGetNamedProgramLocalParameterIuivEXT = _getProcAddress('glGetNamedProgramLocalParameterIuivEXT')
	gl.glGetNamedProgramLocalParameterIuivEXT.gl = gl
	gl.glGetNamedProgramLocalParameterIuivEXT.__name__ = 'glGetNamedProgramLocalParameterIuivEXT'
	gl.glGetNamedProgramLocalParameterIuivEXT.errcheck = _errcheck
	gl.glGetNamedProgramLocalParameterIuivEXT.restype = GLvoid
	gl.glGetNamedProgramLocalParameterIuivEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLuint),)
	gl.glGetNamedProgramLocalParameterdvEXT = _getProcAddress('glGetNamedProgramLocalParameterdvEXT')
	gl.glGetNamedProgramLocalParameterdvEXT.gl = gl
	gl.glGetNamedProgramLocalParameterdvEXT.__name__ = 'glGetNamedProgramLocalParameterdvEXT'
	gl.glGetNamedProgramLocalParameterdvEXT.errcheck = _errcheck
	gl.glGetNamedProgramLocalParameterdvEXT.restype = GLvoid
	gl.glGetNamedProgramLocalParameterdvEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLdouble),)
	gl.glGetNamedProgramLocalParameterfvEXT = _getProcAddress('glGetNamedProgramLocalParameterfvEXT')
	gl.glGetNamedProgramLocalParameterfvEXT.gl = gl
	gl.glGetNamedProgramLocalParameterfvEXT.__name__ = 'glGetNamedProgramLocalParameterfvEXT'
	gl.glGetNamedProgramLocalParameterfvEXT.errcheck = _errcheck
	gl.glGetNamedProgramLocalParameterfvEXT.restype = GLvoid
	gl.glGetNamedProgramLocalParameterfvEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetNamedProgramStringEXT = _getProcAddress('glGetNamedProgramStringEXT')
	gl.glGetNamedProgramStringEXT.gl = gl
	gl.glGetNamedProgramStringEXT.__name__ = 'glGetNamedProgramStringEXT'
	gl.glGetNamedProgramStringEXT.errcheck = _errcheck
	gl.glGetNamedProgramStringEXT.restype = GLvoid
	gl.glGetNamedProgramStringEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetNamedProgramivEXT = _getProcAddress('glGetNamedProgramivEXT')
	gl.glGetNamedProgramivEXT.gl = gl
	gl.glGetNamedProgramivEXT.__name__ = 'glGetNamedProgramivEXT'
	gl.glGetNamedProgramivEXT.errcheck = _errcheck
	gl.glGetNamedProgramivEXT.restype = GLvoid
	gl.glGetNamedProgramivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetNamedRenderbufferParameteriv = _getProcAddress('glGetNamedRenderbufferParameteriv')
	gl.glGetNamedRenderbufferParameteriv.gl = gl
	gl.glGetNamedRenderbufferParameteriv.__name__ = 'glGetNamedRenderbufferParameteriv'
	gl.glGetNamedRenderbufferParameteriv.errcheck = _errcheck
	gl.glGetNamedRenderbufferParameteriv.restype = GLvoid
	gl.glGetNamedRenderbufferParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetNamedRenderbufferParameterivEXT = _getProcAddress('glGetNamedRenderbufferParameterivEXT')
	gl.glGetNamedRenderbufferParameterivEXT.gl = gl
	gl.glGetNamedRenderbufferParameterivEXT.__name__ = 'glGetNamedRenderbufferParameterivEXT'
	gl.glGetNamedRenderbufferParameterivEXT.errcheck = _errcheck
	gl.glGetNamedRenderbufferParameterivEXT.restype = GLvoid
	gl.glGetNamedRenderbufferParameterivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetNamedStringARB = _getProcAddress('glGetNamedStringARB')
	gl.glGetNamedStringARB.gl = gl
	gl.glGetNamedStringARB.__name__ = 'glGetNamedStringARB'
	gl.glGetNamedStringARB.errcheck = _errcheck
	gl.glGetNamedStringARB.restype = GLvoid
	gl.glGetNamedStringARB.argtypes = (GLint, POINTER(GLchar), GLsizei, POINTER(GLint), POINTER(GLchar),)
	gl.glGetNamedStringivARB = _getProcAddress('glGetNamedStringivARB')
	gl.glGetNamedStringivARB.gl = gl
	gl.glGetNamedStringivARB.__name__ = 'glGetNamedStringivARB'
	gl.glGetNamedStringivARB.errcheck = _errcheck
	gl.glGetNamedStringivARB.restype = GLvoid
	gl.glGetNamedStringivARB.argtypes = (GLint, POINTER(GLchar), GLenum, POINTER(GLint),)
	gl.glGetNextPerfQueryIdINTEL = _getProcAddress('glGetNextPerfQueryIdINTEL')
	gl.glGetNextPerfQueryIdINTEL.gl = gl
	gl.glGetNextPerfQueryIdINTEL.__name__ = 'glGetNextPerfQueryIdINTEL'
	gl.glGetNextPerfQueryIdINTEL.errcheck = _errcheck
	gl.glGetNextPerfQueryIdINTEL.restype = GLvoid
	gl.glGetNextPerfQueryIdINTEL.argtypes = (GLuint, POINTER(GLuint),)
	gl.glGetObjectBufferfvATI = _getProcAddress('glGetObjectBufferfvATI')
	gl.glGetObjectBufferfvATI.gl = gl
	gl.glGetObjectBufferfvATI.__name__ = 'glGetObjectBufferfvATI'
	gl.glGetObjectBufferfvATI.errcheck = _errcheck
	gl.glGetObjectBufferfvATI.restype = GLvoid
	gl.glGetObjectBufferfvATI.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetObjectBufferivATI = _getProcAddress('glGetObjectBufferivATI')
	gl.glGetObjectBufferivATI.gl = gl
	gl.glGetObjectBufferivATI.__name__ = 'glGetObjectBufferivATI'
	gl.glGetObjectBufferivATI.errcheck = _errcheck
	gl.glGetObjectBufferivATI.restype = GLvoid
	gl.glGetObjectBufferivATI.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetObjectLabel = _getProcAddress('glGetObjectLabel')
	gl.glGetObjectLabel.gl = gl
	gl.glGetObjectLabel.__name__ = 'glGetObjectLabel'
	gl.glGetObjectLabel.errcheck = _errcheck
	gl.glGetObjectLabel.restype = GLvoid
	gl.glGetObjectLabel.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetObjectLabelEXT = _getProcAddress('glGetObjectLabelEXT')
	gl.glGetObjectLabelEXT.gl = gl
	gl.glGetObjectLabelEXT.__name__ = 'glGetObjectLabelEXT'
	gl.glGetObjectLabelEXT.errcheck = _errcheck
	gl.glGetObjectLabelEXT.restype = GLvoid
	gl.glGetObjectLabelEXT.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetObjectLabelKHR = _getProcAddress('glGetObjectLabelKHR')
	gl.glGetObjectLabelKHR.gl = gl
	gl.glGetObjectLabelKHR.__name__ = 'glGetObjectLabelKHR'
	gl.glGetObjectLabelKHR.errcheck = _errcheck
	gl.glGetObjectLabelKHR.restype = GLvoid
	gl.glGetObjectLabelKHR.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetObjectParameterfvARB = _getProcAddress('glGetObjectParameterfvARB')
	gl.glGetObjectParameterfvARB.gl = gl
	gl.glGetObjectParameterfvARB.__name__ = 'glGetObjectParameterfvARB'
	gl.glGetObjectParameterfvARB.errcheck = _errcheck
	gl.glGetObjectParameterfvARB.restype = GLvoid
	gl.glGetObjectParameterfvARB.argtypes = (GLhandleARB, GLenum, POINTER(GLfloat),)
	gl.glGetObjectParameterivAPPLE = _getProcAddress('glGetObjectParameterivAPPLE')
	gl.glGetObjectParameterivAPPLE.gl = gl
	gl.glGetObjectParameterivAPPLE.__name__ = 'glGetObjectParameterivAPPLE'
	gl.glGetObjectParameterivAPPLE.errcheck = _errcheck
	gl.glGetObjectParameterivAPPLE.restype = GLvoid
	gl.glGetObjectParameterivAPPLE.argtypes = (GLenum, GLuint, GLenum, POINTER(GLint),)
	gl.glGetObjectParameterivARB = _getProcAddress('glGetObjectParameterivARB')
	gl.glGetObjectParameterivARB.gl = gl
	gl.glGetObjectParameterivARB.__name__ = 'glGetObjectParameterivARB'
	gl.glGetObjectParameterivARB.errcheck = _errcheck
	gl.glGetObjectParameterivARB.restype = GLvoid
	gl.glGetObjectParameterivARB.argtypes = (GLhandleARB, GLenum, POINTER(GLint),)
	gl.glGetObjectPtrLabel = _getProcAddress('glGetObjectPtrLabel')
	gl.glGetObjectPtrLabel.gl = gl
	gl.glGetObjectPtrLabel.__name__ = 'glGetObjectPtrLabel'
	gl.glGetObjectPtrLabel.errcheck = _errcheck
	gl.glGetObjectPtrLabel.restype = GLvoid
	gl.glGetObjectPtrLabel.argtypes = (POINTER(GLvoid), GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetObjectPtrLabelKHR = _getProcAddress('glGetObjectPtrLabelKHR')
	gl.glGetObjectPtrLabelKHR.gl = gl
	gl.glGetObjectPtrLabelKHR.__name__ = 'glGetObjectPtrLabelKHR'
	gl.glGetObjectPtrLabelKHR.errcheck = _errcheck
	gl.glGetObjectPtrLabelKHR.restype = GLvoid
	gl.glGetObjectPtrLabelKHR.argtypes = (POINTER(GLvoid), GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetOcclusionQueryivNV = _getProcAddress('glGetOcclusionQueryivNV')
	gl.glGetOcclusionQueryivNV.gl = gl
	gl.glGetOcclusionQueryivNV.__name__ = 'glGetOcclusionQueryivNV'
	gl.glGetOcclusionQueryivNV.errcheck = _errcheck
	gl.glGetOcclusionQueryivNV.restype = GLvoid
	gl.glGetOcclusionQueryivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetOcclusionQueryuivNV = _getProcAddress('glGetOcclusionQueryuivNV')
	gl.glGetOcclusionQueryuivNV.gl = gl
	gl.glGetOcclusionQueryuivNV.__name__ = 'glGetOcclusionQueryuivNV'
	gl.glGetOcclusionQueryuivNV.errcheck = _errcheck
	gl.glGetOcclusionQueryuivNV.restype = GLvoid
	gl.glGetOcclusionQueryuivNV.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetPathColorGenfvNV = _getProcAddress('glGetPathColorGenfvNV')
	gl.glGetPathColorGenfvNV.gl = gl
	gl.glGetPathColorGenfvNV.__name__ = 'glGetPathColorGenfvNV'
	gl.glGetPathColorGenfvNV.errcheck = _errcheck
	gl.glGetPathColorGenfvNV.restype = GLvoid
	gl.glGetPathColorGenfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetPathColorGenivNV = _getProcAddress('glGetPathColorGenivNV')
	gl.glGetPathColorGenivNV.gl = gl
	gl.glGetPathColorGenivNV.__name__ = 'glGetPathColorGenivNV'
	gl.glGetPathColorGenivNV.errcheck = _errcheck
	gl.glGetPathColorGenivNV.restype = GLvoid
	gl.glGetPathColorGenivNV.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetPathCommandsNV = _getProcAddress('glGetPathCommandsNV')
	gl.glGetPathCommandsNV.gl = gl
	gl.glGetPathCommandsNV.__name__ = 'glGetPathCommandsNV'
	gl.glGetPathCommandsNV.errcheck = _errcheck
	gl.glGetPathCommandsNV.restype = GLvoid
	gl.glGetPathCommandsNV.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glGetPathCoordsNV = _getProcAddress('glGetPathCoordsNV')
	gl.glGetPathCoordsNV.gl = gl
	gl.glGetPathCoordsNV.__name__ = 'glGetPathCoordsNV'
	gl.glGetPathCoordsNV.errcheck = _errcheck
	gl.glGetPathCoordsNV.restype = GLvoid
	gl.glGetPathCoordsNV.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glGetPathDashArrayNV = _getProcAddress('glGetPathDashArrayNV')
	gl.glGetPathDashArrayNV.gl = gl
	gl.glGetPathDashArrayNV.__name__ = 'glGetPathDashArrayNV'
	gl.glGetPathDashArrayNV.errcheck = _errcheck
	gl.glGetPathDashArrayNV.restype = GLvoid
	gl.glGetPathDashArrayNV.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glGetPathLengthNV = _getProcAddress('glGetPathLengthNV')
	gl.glGetPathLengthNV.gl = gl
	gl.glGetPathLengthNV.__name__ = 'glGetPathLengthNV'
	gl.glGetPathLengthNV.errcheck = _errcheck
	gl.glGetPathLengthNV.restype = GLfloat
	gl.glGetPathLengthNV.argtypes = (GLuint, GLsizei, GLsizei,)
	gl.glGetPathMetricRangeNV = _getProcAddress('glGetPathMetricRangeNV')
	gl.glGetPathMetricRangeNV.gl = gl
	gl.glGetPathMetricRangeNV.__name__ = 'glGetPathMetricRangeNV'
	gl.glGetPathMetricRangeNV.errcheck = _errcheck
	gl.glGetPathMetricRangeNV.restype = GLvoid
	gl.glGetPathMetricRangeNV.argtypes = (GLbitfield, GLuint, GLsizei, GLsizei, POINTER(GLfloat),)
	gl.glGetPathMetricsNV = _getProcAddress('glGetPathMetricsNV')
	gl.glGetPathMetricsNV.gl = gl
	gl.glGetPathMetricsNV.__name__ = 'glGetPathMetricsNV'
	gl.glGetPathMetricsNV.errcheck = _errcheck
	gl.glGetPathMetricsNV.restype = GLvoid
	gl.glGetPathMetricsNV.argtypes = (GLbitfield, GLsizei, GLenum, POINTER(GLvoid), GLuint, GLsizei, POINTER(GLfloat),)
	gl.glGetPathParameterfvNV = _getProcAddress('glGetPathParameterfvNV')
	gl.glGetPathParameterfvNV.gl = gl
	gl.glGetPathParameterfvNV.__name__ = 'glGetPathParameterfvNV'
	gl.glGetPathParameterfvNV.errcheck = _errcheck
	gl.glGetPathParameterfvNV.restype = GLvoid
	gl.glGetPathParameterfvNV.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetPathParameterivNV = _getProcAddress('glGetPathParameterivNV')
	gl.glGetPathParameterivNV.gl = gl
	gl.glGetPathParameterivNV.__name__ = 'glGetPathParameterivNV'
	gl.glGetPathParameterivNV.errcheck = _errcheck
	gl.glGetPathParameterivNV.restype = GLvoid
	gl.glGetPathParameterivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetPathSpacingNV = _getProcAddress('glGetPathSpacingNV')
	gl.glGetPathSpacingNV.gl = gl
	gl.glGetPathSpacingNV.__name__ = 'glGetPathSpacingNV'
	gl.glGetPathSpacingNV.errcheck = _errcheck
	gl.glGetPathSpacingNV.restype = GLvoid
	gl.glGetPathSpacingNV.argtypes = (GLenum, GLsizei, GLenum, POINTER(GLvoid), GLuint, GLfloat, GLfloat, GLenum, POINTER(GLfloat),)
	gl.glGetPathTexGenfvNV = _getProcAddress('glGetPathTexGenfvNV')
	gl.glGetPathTexGenfvNV.gl = gl
	gl.glGetPathTexGenfvNV.__name__ = 'glGetPathTexGenfvNV'
	gl.glGetPathTexGenfvNV.errcheck = _errcheck
	gl.glGetPathTexGenfvNV.restype = GLvoid
	gl.glGetPathTexGenfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetPathTexGenivNV = _getProcAddress('glGetPathTexGenivNV')
	gl.glGetPathTexGenivNV.gl = gl
	gl.glGetPathTexGenivNV.__name__ = 'glGetPathTexGenivNV'
	gl.glGetPathTexGenivNV.errcheck = _errcheck
	gl.glGetPathTexGenivNV.restype = GLvoid
	gl.glGetPathTexGenivNV.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetPerfCounterInfoINTEL = _getProcAddress('glGetPerfCounterInfoINTEL')
	gl.glGetPerfCounterInfoINTEL.gl = gl
	gl.glGetPerfCounterInfoINTEL.__name__ = 'glGetPerfCounterInfoINTEL'
	gl.glGetPerfCounterInfoINTEL.errcheck = _errcheck
	gl.glGetPerfCounterInfoINTEL.restype = GLvoid
	gl.glGetPerfCounterInfoINTEL.argtypes = (GLuint, GLuint, GLuint, POINTER(GLchar), GLuint, POINTER(GLchar), POINTER(GLuint), POINTER(GLuint), POINTER(GLuint), POINTER(GLuint), POINTER(GLuint64),)
	gl.glGetPerfMonitorCounterDataAMD = _getProcAddress('glGetPerfMonitorCounterDataAMD')
	gl.glGetPerfMonitorCounterDataAMD.gl = gl
	gl.glGetPerfMonitorCounterDataAMD.__name__ = 'glGetPerfMonitorCounterDataAMD'
	gl.glGetPerfMonitorCounterDataAMD.errcheck = _errcheck
	gl.glGetPerfMonitorCounterDataAMD.restype = GLvoid
	gl.glGetPerfMonitorCounterDataAMD.argtypes = (GLuint, GLenum, GLsizei, POINTER(GLuint), POINTER(GLint),)
	gl.glGetPerfMonitorCounterInfoAMD = _getProcAddress('glGetPerfMonitorCounterInfoAMD')
	gl.glGetPerfMonitorCounterInfoAMD.gl = gl
	gl.glGetPerfMonitorCounterInfoAMD.__name__ = 'glGetPerfMonitorCounterInfoAMD'
	gl.glGetPerfMonitorCounterInfoAMD.errcheck = _errcheck
	gl.glGetPerfMonitorCounterInfoAMD.restype = GLvoid
	gl.glGetPerfMonitorCounterInfoAMD.argtypes = (GLuint, GLuint, GLenum, POINTER(GLvoid),)
	gl.glGetPerfMonitorCounterStringAMD = _getProcAddress('glGetPerfMonitorCounterStringAMD')
	gl.glGetPerfMonitorCounterStringAMD.gl = gl
	gl.glGetPerfMonitorCounterStringAMD.__name__ = 'glGetPerfMonitorCounterStringAMD'
	gl.glGetPerfMonitorCounterStringAMD.errcheck = _errcheck
	gl.glGetPerfMonitorCounterStringAMD.restype = GLvoid
	gl.glGetPerfMonitorCounterStringAMD.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetPerfMonitorCountersAMD = _getProcAddress('glGetPerfMonitorCountersAMD')
	gl.glGetPerfMonitorCountersAMD.gl = gl
	gl.glGetPerfMonitorCountersAMD.__name__ = 'glGetPerfMonitorCountersAMD'
	gl.glGetPerfMonitorCountersAMD.errcheck = _errcheck
	gl.glGetPerfMonitorCountersAMD.restype = GLvoid
	gl.glGetPerfMonitorCountersAMD.argtypes = (GLuint, POINTER(GLint), POINTER(GLint), GLsizei, POINTER(GLuint),)
	gl.glGetPerfMonitorGroupStringAMD = _getProcAddress('glGetPerfMonitorGroupStringAMD')
	gl.glGetPerfMonitorGroupStringAMD.gl = gl
	gl.glGetPerfMonitorGroupStringAMD.__name__ = 'glGetPerfMonitorGroupStringAMD'
	gl.glGetPerfMonitorGroupStringAMD.errcheck = _errcheck
	gl.glGetPerfMonitorGroupStringAMD.restype = GLvoid
	gl.glGetPerfMonitorGroupStringAMD.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetPerfMonitorGroupsAMD = _getProcAddress('glGetPerfMonitorGroupsAMD')
	gl.glGetPerfMonitorGroupsAMD.gl = gl
	gl.glGetPerfMonitorGroupsAMD.__name__ = 'glGetPerfMonitorGroupsAMD'
	gl.glGetPerfMonitorGroupsAMD.errcheck = _errcheck
	gl.glGetPerfMonitorGroupsAMD.restype = GLvoid
	gl.glGetPerfMonitorGroupsAMD.argtypes = (POINTER(GLint), GLsizei, POINTER(GLuint),)
	gl.glGetPerfQueryDataINTEL = _getProcAddress('glGetPerfQueryDataINTEL')
	gl.glGetPerfQueryDataINTEL.gl = gl
	gl.glGetPerfQueryDataINTEL.__name__ = 'glGetPerfQueryDataINTEL'
	gl.glGetPerfQueryDataINTEL.errcheck = _errcheck
	gl.glGetPerfQueryDataINTEL.restype = GLvoid
	gl.glGetPerfQueryDataINTEL.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLvoid), POINTER(GLuint),)
	gl.glGetPerfQueryIdByNameINTEL = _getProcAddress('glGetPerfQueryIdByNameINTEL')
	gl.glGetPerfQueryIdByNameINTEL.gl = gl
	gl.glGetPerfQueryIdByNameINTEL.__name__ = 'glGetPerfQueryIdByNameINTEL'
	gl.glGetPerfQueryIdByNameINTEL.errcheck = _errcheck
	gl.glGetPerfQueryIdByNameINTEL.restype = GLvoid
	gl.glGetPerfQueryIdByNameINTEL.argtypes = (POINTER(GLchar), POINTER(GLuint),)
	gl.glGetPerfQueryInfoINTEL = _getProcAddress('glGetPerfQueryInfoINTEL')
	gl.glGetPerfQueryInfoINTEL.gl = gl
	gl.glGetPerfQueryInfoINTEL.__name__ = 'glGetPerfQueryInfoINTEL'
	gl.glGetPerfQueryInfoINTEL.errcheck = _errcheck
	gl.glGetPerfQueryInfoINTEL.restype = GLvoid
	gl.glGetPerfQueryInfoINTEL.argtypes = (GLuint, GLuint, POINTER(GLchar), POINTER(GLuint), POINTER(GLuint), POINTER(GLuint), POINTER(GLuint),)
	gl.glGetPixelMapfv = _getProcAddress('glGetPixelMapfv')
	gl.glGetPixelMapfv.gl = gl
	gl.glGetPixelMapfv.__name__ = 'glGetPixelMapfv'
	gl.glGetPixelMapfv.errcheck = _errcheck
	gl.glGetPixelMapfv.restype = GLvoid
	gl.glGetPixelMapfv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetPixelMapuiv = _getProcAddress('glGetPixelMapuiv')
	gl.glGetPixelMapuiv.gl = gl
	gl.glGetPixelMapuiv.__name__ = 'glGetPixelMapuiv'
	gl.glGetPixelMapuiv.errcheck = _errcheck
	gl.glGetPixelMapuiv.restype = GLvoid
	gl.glGetPixelMapuiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glGetPixelMapusv = _getProcAddress('glGetPixelMapusv')
	gl.glGetPixelMapusv.gl = gl
	gl.glGetPixelMapusv.__name__ = 'glGetPixelMapusv'
	gl.glGetPixelMapusv.errcheck = _errcheck
	gl.glGetPixelMapusv.restype = GLvoid
	gl.glGetPixelMapusv.argtypes = (GLenum, POINTER(GLushort),)
	gl.glGetPixelMapxv = _getProcAddress('glGetPixelMapxv')
	gl.glGetPixelMapxv.gl = gl
	gl.glGetPixelMapxv.__name__ = 'glGetPixelMapxv'
	gl.glGetPixelMapxv.errcheck = _errcheck
	gl.glGetPixelMapxv.restype = GLvoid
	gl.glGetPixelMapxv.argtypes = (GLenum, GLint, POINTER(GLfixed),)
	gl.glGetPixelTexGenParameterfvSGIS = _getProcAddress('glGetPixelTexGenParameterfvSGIS')
	gl.glGetPixelTexGenParameterfvSGIS.gl = gl
	gl.glGetPixelTexGenParameterfvSGIS.__name__ = 'glGetPixelTexGenParameterfvSGIS'
	gl.glGetPixelTexGenParameterfvSGIS.errcheck = _errcheck
	gl.glGetPixelTexGenParameterfvSGIS.restype = GLvoid
	gl.glGetPixelTexGenParameterfvSGIS.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetPixelTexGenParameterivSGIS = _getProcAddress('glGetPixelTexGenParameterivSGIS')
	gl.glGetPixelTexGenParameterivSGIS.gl = gl
	gl.glGetPixelTexGenParameterivSGIS.__name__ = 'glGetPixelTexGenParameterivSGIS'
	gl.glGetPixelTexGenParameterivSGIS.errcheck = _errcheck
	gl.glGetPixelTexGenParameterivSGIS.restype = GLvoid
	gl.glGetPixelTexGenParameterivSGIS.argtypes = (GLenum, POINTER(GLint),)
	gl.glGetPixelTransformParameterfvEXT = _getProcAddress('glGetPixelTransformParameterfvEXT')
	gl.glGetPixelTransformParameterfvEXT.gl = gl
	gl.glGetPixelTransformParameterfvEXT.__name__ = 'glGetPixelTransformParameterfvEXT'
	gl.glGetPixelTransformParameterfvEXT.errcheck = _errcheck
	gl.glGetPixelTransformParameterfvEXT.restype = GLvoid
	gl.glGetPixelTransformParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetPixelTransformParameterivEXT = _getProcAddress('glGetPixelTransformParameterivEXT')
	gl.glGetPixelTransformParameterivEXT.gl = gl
	gl.glGetPixelTransformParameterivEXT.__name__ = 'glGetPixelTransformParameterivEXT'
	gl.glGetPixelTransformParameterivEXT.errcheck = _errcheck
	gl.glGetPixelTransformParameterivEXT.restype = GLvoid
	gl.glGetPixelTransformParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetPointerIndexedvEXT = _getProcAddress('glGetPointerIndexedvEXT')
	gl.glGetPointerIndexedvEXT.gl = gl
	gl.glGetPointerIndexedvEXT.__name__ = 'glGetPointerIndexedvEXT'
	gl.glGetPointerIndexedvEXT.errcheck = _errcheck
	gl.glGetPointerIndexedvEXT.restype = GLvoid
	gl.glGetPointerIndexedvEXT.argtypes = (GLenum, GLuint, POINTER(POINTER(GLvoid)),)
	gl.glGetPointeri_vEXT = _getProcAddress('glGetPointeri_vEXT')
	gl.glGetPointeri_vEXT.gl = gl
	gl.glGetPointeri_vEXT.__name__ = 'glGetPointeri_vEXT'
	gl.glGetPointeri_vEXT.errcheck = _errcheck
	gl.glGetPointeri_vEXT.restype = GLvoid
	gl.glGetPointeri_vEXT.argtypes = (GLenum, GLuint, POINTER(POINTER(GLvoid)),)
	gl.glGetPointerv = _getProcAddress('glGetPointerv')
	gl.glGetPointerv.gl = gl
	gl.glGetPointerv.__name__ = 'glGetPointerv'
	gl.glGetPointerv.errcheck = _errcheck
	gl.glGetPointerv.restype = GLvoid
	gl.glGetPointerv.argtypes = (GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetPointervEXT = _getProcAddress('glGetPointervEXT')
	gl.glGetPointervEXT.gl = gl
	gl.glGetPointervEXT.__name__ = 'glGetPointervEXT'
	gl.glGetPointervEXT.errcheck = _errcheck
	gl.glGetPointervEXT.restype = GLvoid
	gl.glGetPointervEXT.argtypes = (GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetPointervKHR = _getProcAddress('glGetPointervKHR')
	gl.glGetPointervKHR.gl = gl
	gl.glGetPointervKHR.__name__ = 'glGetPointervKHR'
	gl.glGetPointervKHR.errcheck = _errcheck
	gl.glGetPointervKHR.restype = GLvoid
	gl.glGetPointervKHR.argtypes = (GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetPolygonStipple = _getProcAddress('glGetPolygonStipple')
	gl.glGetPolygonStipple.gl = gl
	gl.glGetPolygonStipple.__name__ = 'glGetPolygonStipple'
	gl.glGetPolygonStipple.errcheck = _errcheck
	gl.glGetPolygonStipple.restype = GLvoid
	gl.glGetPolygonStipple.argtypes = (POINTER(GLubyte),)
	gl.glGetProgramBinary = _getProcAddress('glGetProgramBinary')
	gl.glGetProgramBinary.gl = gl
	gl.glGetProgramBinary.__name__ = 'glGetProgramBinary'
	gl.glGetProgramBinary.errcheck = _errcheck
	gl.glGetProgramBinary.restype = GLvoid
	gl.glGetProgramBinary.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLenum), POINTER(GLvoid),)
	gl.glGetProgramBinaryOES = _getProcAddress('glGetProgramBinaryOES')
	gl.glGetProgramBinaryOES.gl = gl
	gl.glGetProgramBinaryOES.__name__ = 'glGetProgramBinaryOES'
	gl.glGetProgramBinaryOES.errcheck = _errcheck
	gl.glGetProgramBinaryOES.restype = GLvoid
	gl.glGetProgramBinaryOES.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLenum), POINTER(GLvoid),)
	gl.glGetProgramEnvParameterIivNV = _getProcAddress('glGetProgramEnvParameterIivNV')
	gl.glGetProgramEnvParameterIivNV.gl = gl
	gl.glGetProgramEnvParameterIivNV.__name__ = 'glGetProgramEnvParameterIivNV'
	gl.glGetProgramEnvParameterIivNV.errcheck = _errcheck
	gl.glGetProgramEnvParameterIivNV.restype = GLvoid
	gl.glGetProgramEnvParameterIivNV.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glGetProgramEnvParameterIuivNV = _getProcAddress('glGetProgramEnvParameterIuivNV')
	gl.glGetProgramEnvParameterIuivNV.gl = gl
	gl.glGetProgramEnvParameterIuivNV.__name__ = 'glGetProgramEnvParameterIuivNV'
	gl.glGetProgramEnvParameterIuivNV.errcheck = _errcheck
	gl.glGetProgramEnvParameterIuivNV.restype = GLvoid
	gl.glGetProgramEnvParameterIuivNV.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glGetProgramEnvParameterdvARB = _getProcAddress('glGetProgramEnvParameterdvARB')
	gl.glGetProgramEnvParameterdvARB.gl = gl
	gl.glGetProgramEnvParameterdvARB.__name__ = 'glGetProgramEnvParameterdvARB'
	gl.glGetProgramEnvParameterdvARB.errcheck = _errcheck
	gl.glGetProgramEnvParameterdvARB.restype = GLvoid
	gl.glGetProgramEnvParameterdvARB.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glGetProgramEnvParameterfvARB = _getProcAddress('glGetProgramEnvParameterfvARB')
	gl.glGetProgramEnvParameterfvARB.gl = gl
	gl.glGetProgramEnvParameterfvARB.__name__ = 'glGetProgramEnvParameterfvARB'
	gl.glGetProgramEnvParameterfvARB.errcheck = _errcheck
	gl.glGetProgramEnvParameterfvARB.restype = GLvoid
	gl.glGetProgramEnvParameterfvARB.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetProgramInfoLog = _getProcAddress('glGetProgramInfoLog')
	gl.glGetProgramInfoLog.gl = gl
	gl.glGetProgramInfoLog.__name__ = 'glGetProgramInfoLog'
	gl.glGetProgramInfoLog.errcheck = _errcheck
	gl.glGetProgramInfoLog.restype = GLvoid
	gl.glGetProgramInfoLog.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetProgramInterfaceiv = _getProcAddress('glGetProgramInterfaceiv')
	gl.glGetProgramInterfaceiv.gl = gl
	gl.glGetProgramInterfaceiv.__name__ = 'glGetProgramInterfaceiv'
	gl.glGetProgramInterfaceiv.errcheck = _errcheck
	gl.glGetProgramInterfaceiv.restype = GLvoid
	gl.glGetProgramInterfaceiv.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetProgramLocalParameterIivNV = _getProcAddress('glGetProgramLocalParameterIivNV')
	gl.glGetProgramLocalParameterIivNV.gl = gl
	gl.glGetProgramLocalParameterIivNV.__name__ = 'glGetProgramLocalParameterIivNV'
	gl.glGetProgramLocalParameterIivNV.errcheck = _errcheck
	gl.glGetProgramLocalParameterIivNV.restype = GLvoid
	gl.glGetProgramLocalParameterIivNV.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glGetProgramLocalParameterIuivNV = _getProcAddress('glGetProgramLocalParameterIuivNV')
	gl.glGetProgramLocalParameterIuivNV.gl = gl
	gl.glGetProgramLocalParameterIuivNV.__name__ = 'glGetProgramLocalParameterIuivNV'
	gl.glGetProgramLocalParameterIuivNV.errcheck = _errcheck
	gl.glGetProgramLocalParameterIuivNV.restype = GLvoid
	gl.glGetProgramLocalParameterIuivNV.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glGetProgramLocalParameterdvARB = _getProcAddress('glGetProgramLocalParameterdvARB')
	gl.glGetProgramLocalParameterdvARB.gl = gl
	gl.glGetProgramLocalParameterdvARB.__name__ = 'glGetProgramLocalParameterdvARB'
	gl.glGetProgramLocalParameterdvARB.errcheck = _errcheck
	gl.glGetProgramLocalParameterdvARB.restype = GLvoid
	gl.glGetProgramLocalParameterdvARB.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glGetProgramLocalParameterfvARB = _getProcAddress('glGetProgramLocalParameterfvARB')
	gl.glGetProgramLocalParameterfvARB.gl = gl
	gl.glGetProgramLocalParameterfvARB.__name__ = 'glGetProgramLocalParameterfvARB'
	gl.glGetProgramLocalParameterfvARB.errcheck = _errcheck
	gl.glGetProgramLocalParameterfvARB.restype = GLvoid
	gl.glGetProgramLocalParameterfvARB.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glGetProgramNamedParameterdvNV = _getProcAddress('glGetProgramNamedParameterdvNV')
	gl.glGetProgramNamedParameterdvNV.gl = gl
	gl.glGetProgramNamedParameterdvNV.__name__ = 'glGetProgramNamedParameterdvNV'
	gl.glGetProgramNamedParameterdvNV.errcheck = _errcheck
	gl.glGetProgramNamedParameterdvNV.restype = GLvoid
	gl.glGetProgramNamedParameterdvNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), POINTER(GLdouble),)
	gl.glGetProgramNamedParameterfvNV = _getProcAddress('glGetProgramNamedParameterfvNV')
	gl.glGetProgramNamedParameterfvNV.gl = gl
	gl.glGetProgramNamedParameterfvNV.__name__ = 'glGetProgramNamedParameterfvNV'
	gl.glGetProgramNamedParameterfvNV.errcheck = _errcheck
	gl.glGetProgramNamedParameterfvNV.restype = GLvoid
	gl.glGetProgramNamedParameterfvNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), POINTER(GLfloat),)
	gl.glGetProgramParameterdvNV = _getProcAddress('glGetProgramParameterdvNV')
	gl.glGetProgramParameterdvNV.gl = gl
	gl.glGetProgramParameterdvNV.__name__ = 'glGetProgramParameterdvNV'
	gl.glGetProgramParameterdvNV.errcheck = _errcheck
	gl.glGetProgramParameterdvNV.restype = GLvoid
	gl.glGetProgramParameterdvNV.argtypes = (GLenum, GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetProgramParameterfvNV = _getProcAddress('glGetProgramParameterfvNV')
	gl.glGetProgramParameterfvNV.gl = gl
	gl.glGetProgramParameterfvNV.__name__ = 'glGetProgramParameterfvNV'
	gl.glGetProgramParameterfvNV.errcheck = _errcheck
	gl.glGetProgramParameterfvNV.restype = GLvoid
	gl.glGetProgramParameterfvNV.argtypes = (GLenum, GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetProgramPipelineInfoLog = _getProcAddress('glGetProgramPipelineInfoLog')
	gl.glGetProgramPipelineInfoLog.gl = gl
	gl.glGetProgramPipelineInfoLog.__name__ = 'glGetProgramPipelineInfoLog'
	gl.glGetProgramPipelineInfoLog.errcheck = _errcheck
	gl.glGetProgramPipelineInfoLog.restype = GLvoid
	gl.glGetProgramPipelineInfoLog.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetProgramPipelineInfoLogEXT = _getProcAddress('glGetProgramPipelineInfoLogEXT')
	gl.glGetProgramPipelineInfoLogEXT.gl = gl
	gl.glGetProgramPipelineInfoLogEXT.__name__ = 'glGetProgramPipelineInfoLogEXT'
	gl.glGetProgramPipelineInfoLogEXT.errcheck = _errcheck
	gl.glGetProgramPipelineInfoLogEXT.restype = GLvoid
	gl.glGetProgramPipelineInfoLogEXT.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetProgramPipelineiv = _getProcAddress('glGetProgramPipelineiv')
	gl.glGetProgramPipelineiv.gl = gl
	gl.glGetProgramPipelineiv.__name__ = 'glGetProgramPipelineiv'
	gl.glGetProgramPipelineiv.errcheck = _errcheck
	gl.glGetProgramPipelineiv.restype = GLvoid
	gl.glGetProgramPipelineiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetProgramPipelineivEXT = _getProcAddress('glGetProgramPipelineivEXT')
	gl.glGetProgramPipelineivEXT.gl = gl
	gl.glGetProgramPipelineivEXT.__name__ = 'glGetProgramPipelineivEXT'
	gl.glGetProgramPipelineivEXT.errcheck = _errcheck
	gl.glGetProgramPipelineivEXT.restype = GLvoid
	gl.glGetProgramPipelineivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetProgramResourceIndex = _getProcAddress('glGetProgramResourceIndex')
	gl.glGetProgramResourceIndex.gl = gl
	gl.glGetProgramResourceIndex.__name__ = 'glGetProgramResourceIndex'
	gl.glGetProgramResourceIndex.errcheck = _errcheck
	gl.glGetProgramResourceIndex.restype = GLuint
	gl.glGetProgramResourceIndex.argtypes = (GLuint, GLenum, POINTER(GLchar),)
	gl.glGetProgramResourceLocation = _getProcAddress('glGetProgramResourceLocation')
	gl.glGetProgramResourceLocation.gl = gl
	gl.glGetProgramResourceLocation.__name__ = 'glGetProgramResourceLocation'
	gl.glGetProgramResourceLocation.errcheck = _errcheck
	gl.glGetProgramResourceLocation.restype = GLint
	gl.glGetProgramResourceLocation.argtypes = (GLuint, GLenum, POINTER(GLchar),)
	gl.glGetProgramResourceLocationIndex = _getProcAddress('glGetProgramResourceLocationIndex')
	gl.glGetProgramResourceLocationIndex.gl = gl
	gl.glGetProgramResourceLocationIndex.__name__ = 'glGetProgramResourceLocationIndex'
	gl.glGetProgramResourceLocationIndex.errcheck = _errcheck
	gl.glGetProgramResourceLocationIndex.restype = GLint
	gl.glGetProgramResourceLocationIndex.argtypes = (GLuint, GLenum, POINTER(GLchar),)
	gl.glGetProgramResourceName = _getProcAddress('glGetProgramResourceName')
	gl.glGetProgramResourceName.gl = gl
	gl.glGetProgramResourceName.__name__ = 'glGetProgramResourceName'
	gl.glGetProgramResourceName.errcheck = _errcheck
	gl.glGetProgramResourceName.restype = GLvoid
	gl.glGetProgramResourceName.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetProgramResourcefvNV = _getProcAddress('glGetProgramResourcefvNV')
	gl.glGetProgramResourcefvNV.gl = gl
	gl.glGetProgramResourcefvNV.__name__ = 'glGetProgramResourcefvNV'
	gl.glGetProgramResourcefvNV.errcheck = _errcheck
	gl.glGetProgramResourcefvNV.restype = GLvoid
	gl.glGetProgramResourcefvNV.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLenum), GLsizei, POINTER(GLsizei), POINTER(GLfloat),)
	gl.glGetProgramResourceiv = _getProcAddress('glGetProgramResourceiv')
	gl.glGetProgramResourceiv.gl = gl
	gl.glGetProgramResourceiv.__name__ = 'glGetProgramResourceiv'
	gl.glGetProgramResourceiv.errcheck = _errcheck
	gl.glGetProgramResourceiv.restype = GLvoid
	gl.glGetProgramResourceiv.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLenum), GLsizei, POINTER(GLsizei), POINTER(GLint),)
	gl.glGetProgramStageiv = _getProcAddress('glGetProgramStageiv')
	gl.glGetProgramStageiv.gl = gl
	gl.glGetProgramStageiv.__name__ = 'glGetProgramStageiv'
	gl.glGetProgramStageiv.errcheck = _errcheck
	gl.glGetProgramStageiv.restype = GLvoid
	gl.glGetProgramStageiv.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetProgramStringARB = _getProcAddress('glGetProgramStringARB')
	gl.glGetProgramStringARB.gl = gl
	gl.glGetProgramStringARB.__name__ = 'glGetProgramStringARB'
	gl.glGetProgramStringARB.errcheck = _errcheck
	gl.glGetProgramStringARB.restype = GLvoid
	gl.glGetProgramStringARB.argtypes = (GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetProgramStringNV = _getProcAddress('glGetProgramStringNV')
	gl.glGetProgramStringNV.gl = gl
	gl.glGetProgramStringNV.__name__ = 'glGetProgramStringNV'
	gl.glGetProgramStringNV.errcheck = _errcheck
	gl.glGetProgramStringNV.restype = GLvoid
	gl.glGetProgramStringNV.argtypes = (GLuint, GLenum, POINTER(GLubyte),)
	gl.glGetProgramSubroutineParameteruivNV = _getProcAddress('glGetProgramSubroutineParameteruivNV')
	gl.glGetProgramSubroutineParameteruivNV.gl = gl
	gl.glGetProgramSubroutineParameteruivNV.__name__ = 'glGetProgramSubroutineParameteruivNV'
	gl.glGetProgramSubroutineParameteruivNV.errcheck = _errcheck
	gl.glGetProgramSubroutineParameteruivNV.restype = GLvoid
	gl.glGetProgramSubroutineParameteruivNV.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glGetProgramiv = _getProcAddress('glGetProgramiv')
	gl.glGetProgramiv.gl = gl
	gl.glGetProgramiv.__name__ = 'glGetProgramiv'
	gl.glGetProgramiv.errcheck = _errcheck
	gl.glGetProgramiv.restype = GLvoid
	gl.glGetProgramiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetProgramivARB = _getProcAddress('glGetProgramivARB')
	gl.glGetProgramivARB.gl = gl
	gl.glGetProgramivARB.__name__ = 'glGetProgramivARB'
	gl.glGetProgramivARB.errcheck = _errcheck
	gl.glGetProgramivARB.restype = GLvoid
	gl.glGetProgramivARB.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetProgramivNV = _getProcAddress('glGetProgramivNV')
	gl.glGetProgramivNV.gl = gl
	gl.glGetProgramivNV.__name__ = 'glGetProgramivNV'
	gl.glGetProgramivNV.errcheck = _errcheck
	gl.glGetProgramivNV.restype = GLvoid
	gl.glGetProgramivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetQueryBufferObjecti64v = _getProcAddress('glGetQueryBufferObjecti64v')
	gl.glGetQueryBufferObjecti64v.gl = gl
	gl.glGetQueryBufferObjecti64v.__name__ = 'glGetQueryBufferObjecti64v'
	gl.glGetQueryBufferObjecti64v.errcheck = _errcheck
	gl.glGetQueryBufferObjecti64v.restype = GLvoid
	gl.glGetQueryBufferObjecti64v.argtypes = (GLuint, GLuint, GLenum, GLintptr,)
	gl.glGetQueryBufferObjectiv = _getProcAddress('glGetQueryBufferObjectiv')
	gl.glGetQueryBufferObjectiv.gl = gl
	gl.glGetQueryBufferObjectiv.__name__ = 'glGetQueryBufferObjectiv'
	gl.glGetQueryBufferObjectiv.errcheck = _errcheck
	gl.glGetQueryBufferObjectiv.restype = GLvoid
	gl.glGetQueryBufferObjectiv.argtypes = (GLuint, GLuint, GLenum, GLintptr,)
	gl.glGetQueryBufferObjectui64v = _getProcAddress('glGetQueryBufferObjectui64v')
	gl.glGetQueryBufferObjectui64v.gl = gl
	gl.glGetQueryBufferObjectui64v.__name__ = 'glGetQueryBufferObjectui64v'
	gl.glGetQueryBufferObjectui64v.errcheck = _errcheck
	gl.glGetQueryBufferObjectui64v.restype = GLvoid
	gl.glGetQueryBufferObjectui64v.argtypes = (GLuint, GLuint, GLenum, GLintptr,)
	gl.glGetQueryBufferObjectuiv = _getProcAddress('glGetQueryBufferObjectuiv')
	gl.glGetQueryBufferObjectuiv.gl = gl
	gl.glGetQueryBufferObjectuiv.__name__ = 'glGetQueryBufferObjectuiv'
	gl.glGetQueryBufferObjectuiv.errcheck = _errcheck
	gl.glGetQueryBufferObjectuiv.restype = GLvoid
	gl.glGetQueryBufferObjectuiv.argtypes = (GLuint, GLuint, GLenum, GLintptr,)
	gl.glGetQueryIndexediv = _getProcAddress('glGetQueryIndexediv')
	gl.glGetQueryIndexediv.gl = gl
	gl.glGetQueryIndexediv.__name__ = 'glGetQueryIndexediv'
	gl.glGetQueryIndexediv.errcheck = _errcheck
	gl.glGetQueryIndexediv.restype = GLvoid
	gl.glGetQueryIndexediv.argtypes = (GLenum, GLuint, GLenum, POINTER(GLint),)
	gl.glGetQueryObjecti64v = _getProcAddress('glGetQueryObjecti64v')
	gl.glGetQueryObjecti64v.gl = gl
	gl.glGetQueryObjecti64v.__name__ = 'glGetQueryObjecti64v'
	gl.glGetQueryObjecti64v.errcheck = _errcheck
	gl.glGetQueryObjecti64v.restype = GLvoid
	gl.glGetQueryObjecti64v.argtypes = (GLuint, GLenum, POINTER(GLint64),)
	gl.glGetQueryObjecti64vEXT = _getProcAddress('glGetQueryObjecti64vEXT')
	gl.glGetQueryObjecti64vEXT.gl = gl
	gl.glGetQueryObjecti64vEXT.__name__ = 'glGetQueryObjecti64vEXT'
	gl.glGetQueryObjecti64vEXT.errcheck = _errcheck
	gl.glGetQueryObjecti64vEXT.restype = GLvoid
	gl.glGetQueryObjecti64vEXT.argtypes = (GLuint, GLenum, POINTER(GLint64),)
	gl.glGetQueryObjectiv = _getProcAddress('glGetQueryObjectiv')
	gl.glGetQueryObjectiv.gl = gl
	gl.glGetQueryObjectiv.__name__ = 'glGetQueryObjectiv'
	gl.glGetQueryObjectiv.errcheck = _errcheck
	gl.glGetQueryObjectiv.restype = GLvoid
	gl.glGetQueryObjectiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetQueryObjectivARB = _getProcAddress('glGetQueryObjectivARB')
	gl.glGetQueryObjectivARB.gl = gl
	gl.glGetQueryObjectivARB.__name__ = 'glGetQueryObjectivARB'
	gl.glGetQueryObjectivARB.errcheck = _errcheck
	gl.glGetQueryObjectivARB.restype = GLvoid
	gl.glGetQueryObjectivARB.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetQueryObjectivEXT = _getProcAddress('glGetQueryObjectivEXT')
	gl.glGetQueryObjectivEXT.gl = gl
	gl.glGetQueryObjectivEXT.__name__ = 'glGetQueryObjectivEXT'
	gl.glGetQueryObjectivEXT.errcheck = _errcheck
	gl.glGetQueryObjectivEXT.restype = GLvoid
	gl.glGetQueryObjectivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetQueryObjectui64v = _getProcAddress('glGetQueryObjectui64v')
	gl.glGetQueryObjectui64v.gl = gl
	gl.glGetQueryObjectui64v.__name__ = 'glGetQueryObjectui64v'
	gl.glGetQueryObjectui64v.errcheck = _errcheck
	gl.glGetQueryObjectui64v.restype = GLvoid
	gl.glGetQueryObjectui64v.argtypes = (GLuint, GLenum, POINTER(GLuint64),)
	gl.glGetQueryObjectui64vEXT = _getProcAddress('glGetQueryObjectui64vEXT')
	gl.glGetQueryObjectui64vEXT.gl = gl
	gl.glGetQueryObjectui64vEXT.__name__ = 'glGetQueryObjectui64vEXT'
	gl.glGetQueryObjectui64vEXT.errcheck = _errcheck
	gl.glGetQueryObjectui64vEXT.restype = GLvoid
	gl.glGetQueryObjectui64vEXT.argtypes = (GLuint, GLenum, POINTER(GLuint64),)
	gl.glGetQueryObjectuiv = _getProcAddress('glGetQueryObjectuiv')
	gl.glGetQueryObjectuiv.gl = gl
	gl.glGetQueryObjectuiv.__name__ = 'glGetQueryObjectuiv'
	gl.glGetQueryObjectuiv.errcheck = _errcheck
	gl.glGetQueryObjectuiv.restype = GLvoid
	gl.glGetQueryObjectuiv.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetQueryObjectuivARB = _getProcAddress('glGetQueryObjectuivARB')
	gl.glGetQueryObjectuivARB.gl = gl
	gl.glGetQueryObjectuivARB.__name__ = 'glGetQueryObjectuivARB'
	gl.glGetQueryObjectuivARB.errcheck = _errcheck
	gl.glGetQueryObjectuivARB.restype = GLvoid
	gl.glGetQueryObjectuivARB.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetQueryObjectuivEXT = _getProcAddress('glGetQueryObjectuivEXT')
	gl.glGetQueryObjectuivEXT.gl = gl
	gl.glGetQueryObjectuivEXT.__name__ = 'glGetQueryObjectuivEXT'
	gl.glGetQueryObjectuivEXT.errcheck = _errcheck
	gl.glGetQueryObjectuivEXT.restype = GLvoid
	gl.glGetQueryObjectuivEXT.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetQueryiv = _getProcAddress('glGetQueryiv')
	gl.glGetQueryiv.gl = gl
	gl.glGetQueryiv.__name__ = 'glGetQueryiv'
	gl.glGetQueryiv.errcheck = _errcheck
	gl.glGetQueryiv.restype = GLvoid
	gl.glGetQueryiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetQueryivARB = _getProcAddress('glGetQueryivARB')
	gl.glGetQueryivARB.gl = gl
	gl.glGetQueryivARB.__name__ = 'glGetQueryivARB'
	gl.glGetQueryivARB.errcheck = _errcheck
	gl.glGetQueryivARB.restype = GLvoid
	gl.glGetQueryivARB.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetQueryivEXT = _getProcAddress('glGetQueryivEXT')
	gl.glGetQueryivEXT.gl = gl
	gl.glGetQueryivEXT.__name__ = 'glGetQueryivEXT'
	gl.glGetQueryivEXT.errcheck = _errcheck
	gl.glGetQueryivEXT.restype = GLvoid
	gl.glGetQueryivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetRenderbufferParameteriv = _getProcAddress('glGetRenderbufferParameteriv')
	gl.glGetRenderbufferParameteriv.gl = gl
	gl.glGetRenderbufferParameteriv.__name__ = 'glGetRenderbufferParameteriv'
	gl.glGetRenderbufferParameteriv.errcheck = _errcheck
	gl.glGetRenderbufferParameteriv.restype = GLvoid
	gl.glGetRenderbufferParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetRenderbufferParameterivEXT = _getProcAddress('glGetRenderbufferParameterivEXT')
	gl.glGetRenderbufferParameterivEXT.gl = gl
	gl.glGetRenderbufferParameterivEXT.__name__ = 'glGetRenderbufferParameterivEXT'
	gl.glGetRenderbufferParameterivEXT.errcheck = _errcheck
	gl.glGetRenderbufferParameterivEXT.restype = GLvoid
	gl.glGetRenderbufferParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetRenderbufferParameterivOES = _getProcAddress('glGetRenderbufferParameterivOES')
	gl.glGetRenderbufferParameterivOES.gl = gl
	gl.glGetRenderbufferParameterivOES.__name__ = 'glGetRenderbufferParameterivOES'
	gl.glGetRenderbufferParameterivOES.errcheck = _errcheck
	gl.glGetRenderbufferParameterivOES.restype = GLvoid
	gl.glGetRenderbufferParameterivOES.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetSamplerParameterIiv = _getProcAddress('glGetSamplerParameterIiv')
	gl.glGetSamplerParameterIiv.gl = gl
	gl.glGetSamplerParameterIiv.__name__ = 'glGetSamplerParameterIiv'
	gl.glGetSamplerParameterIiv.errcheck = _errcheck
	gl.glGetSamplerParameterIiv.restype = GLvoid
	gl.glGetSamplerParameterIiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetSamplerParameterIivEXT = _getProcAddress('glGetSamplerParameterIivEXT')
	gl.glGetSamplerParameterIivEXT.gl = gl
	gl.glGetSamplerParameterIivEXT.__name__ = 'glGetSamplerParameterIivEXT'
	gl.glGetSamplerParameterIivEXT.errcheck = _errcheck
	gl.glGetSamplerParameterIivEXT.restype = GLvoid
	gl.glGetSamplerParameterIivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetSamplerParameterIuiv = _getProcAddress('glGetSamplerParameterIuiv')
	gl.glGetSamplerParameterIuiv.gl = gl
	gl.glGetSamplerParameterIuiv.__name__ = 'glGetSamplerParameterIuiv'
	gl.glGetSamplerParameterIuiv.errcheck = _errcheck
	gl.glGetSamplerParameterIuiv.restype = GLvoid
	gl.glGetSamplerParameterIuiv.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetSamplerParameterIuivEXT = _getProcAddress('glGetSamplerParameterIuivEXT')
	gl.glGetSamplerParameterIuivEXT.gl = gl
	gl.glGetSamplerParameterIuivEXT.__name__ = 'glGetSamplerParameterIuivEXT'
	gl.glGetSamplerParameterIuivEXT.errcheck = _errcheck
	gl.glGetSamplerParameterIuivEXT.restype = GLvoid
	gl.glGetSamplerParameterIuivEXT.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetSamplerParameterfv = _getProcAddress('glGetSamplerParameterfv')
	gl.glGetSamplerParameterfv.gl = gl
	gl.glGetSamplerParameterfv.__name__ = 'glGetSamplerParameterfv'
	gl.glGetSamplerParameterfv.errcheck = _errcheck
	gl.glGetSamplerParameterfv.restype = GLvoid
	gl.glGetSamplerParameterfv.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetSamplerParameteriv = _getProcAddress('glGetSamplerParameteriv')
	gl.glGetSamplerParameteriv.gl = gl
	gl.glGetSamplerParameteriv.__name__ = 'glGetSamplerParameteriv'
	gl.glGetSamplerParameteriv.errcheck = _errcheck
	gl.glGetSamplerParameteriv.restype = GLvoid
	gl.glGetSamplerParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetSeparableFilter = _getProcAddress('glGetSeparableFilter')
	gl.glGetSeparableFilter.gl = gl
	gl.glGetSeparableFilter.__name__ = 'glGetSeparableFilter'
	gl.glGetSeparableFilter.errcheck = _errcheck
	gl.glGetSeparableFilter.restype = GLvoid
	gl.glGetSeparableFilter.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid), POINTER(GLvoid), POINTER(GLvoid),)
	gl.glGetSeparableFilterEXT = _getProcAddress('glGetSeparableFilterEXT')
	gl.glGetSeparableFilterEXT.gl = gl
	gl.glGetSeparableFilterEXT.__name__ = 'glGetSeparableFilterEXT'
	gl.glGetSeparableFilterEXT.errcheck = _errcheck
	gl.glGetSeparableFilterEXT.restype = GLvoid
	gl.glGetSeparableFilterEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLvoid), POINTER(GLvoid), POINTER(GLvoid),)
	gl.glGetShaderInfoLog = _getProcAddress('glGetShaderInfoLog')
	gl.glGetShaderInfoLog.gl = gl
	gl.glGetShaderInfoLog.__name__ = 'glGetShaderInfoLog'
	gl.glGetShaderInfoLog.errcheck = _errcheck
	gl.glGetShaderInfoLog.restype = GLvoid
	gl.glGetShaderInfoLog.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetShaderPrecisionFormat = _getProcAddress('glGetShaderPrecisionFormat')
	gl.glGetShaderPrecisionFormat.gl = gl
	gl.glGetShaderPrecisionFormat.__name__ = 'glGetShaderPrecisionFormat'
	gl.glGetShaderPrecisionFormat.errcheck = _errcheck
	gl.glGetShaderPrecisionFormat.restype = GLvoid
	gl.glGetShaderPrecisionFormat.argtypes = (GLenum, GLenum, POINTER(GLint), POINTER(GLint),)
	gl.glGetShaderSource = _getProcAddress('glGetShaderSource')
	gl.glGetShaderSource.gl = gl
	gl.glGetShaderSource.__name__ = 'glGetShaderSource'
	gl.glGetShaderSource.errcheck = _errcheck
	gl.glGetShaderSource.restype = GLvoid
	gl.glGetShaderSource.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetShaderSourceARB = _getProcAddress('glGetShaderSourceARB')
	gl.glGetShaderSourceARB.gl = gl
	gl.glGetShaderSourceARB.__name__ = 'glGetShaderSourceARB'
	gl.glGetShaderSourceARB.errcheck = _errcheck
	gl.glGetShaderSourceARB.restype = GLvoid
	gl.glGetShaderSourceARB.argtypes = (GLhandleARB, GLsizei, POINTER(GLsizei), POINTER(GLcharARB),)
	gl.glGetShaderiv = _getProcAddress('glGetShaderiv')
	gl.glGetShaderiv.gl = gl
	gl.glGetShaderiv.__name__ = 'glGetShaderiv'
	gl.glGetShaderiv.errcheck = _errcheck
	gl.glGetShaderiv.restype = GLvoid
	gl.glGetShaderiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetSharpenTexFuncSGIS = _getProcAddress('glGetSharpenTexFuncSGIS')
	gl.glGetSharpenTexFuncSGIS.gl = gl
	gl.glGetSharpenTexFuncSGIS.__name__ = 'glGetSharpenTexFuncSGIS'
	gl.glGetSharpenTexFuncSGIS.errcheck = _errcheck
	gl.glGetSharpenTexFuncSGIS.restype = GLvoid
	gl.glGetSharpenTexFuncSGIS.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetString = _getProcAddress('glGetString')
	gl.glGetString.gl = gl
	gl.glGetString.__name__ = 'glGetString'
	gl.glGetString.errcheck = _errcheck
	gl.glGetString.restype = POINTER(GLubyte)
	gl.glGetString.argtypes = (GLenum,)
	gl.glGetStringi = _getProcAddress('glGetStringi')
	gl.glGetStringi.gl = gl
	gl.glGetStringi.__name__ = 'glGetStringi'
	gl.glGetStringi.errcheck = _errcheck
	gl.glGetStringi.restype = POINTER(GLubyte)
	gl.glGetStringi.argtypes = (GLenum, GLuint,)
	gl.glGetSubroutineIndex = _getProcAddress('glGetSubroutineIndex')
	gl.glGetSubroutineIndex.gl = gl
	gl.glGetSubroutineIndex.__name__ = 'glGetSubroutineIndex'
	gl.glGetSubroutineIndex.errcheck = _errcheck
	gl.glGetSubroutineIndex.restype = GLuint
	gl.glGetSubroutineIndex.argtypes = (GLuint, GLenum, POINTER(GLchar),)
	gl.glGetSubroutineUniformLocation = _getProcAddress('glGetSubroutineUniformLocation')
	gl.glGetSubroutineUniformLocation.gl = gl
	gl.glGetSubroutineUniformLocation.__name__ = 'glGetSubroutineUniformLocation'
	gl.glGetSubroutineUniformLocation.errcheck = _errcheck
	gl.glGetSubroutineUniformLocation.restype = GLint
	gl.glGetSubroutineUniformLocation.argtypes = (GLuint, GLenum, POINTER(GLchar),)
	gl.glGetSynciv = _getProcAddress('glGetSynciv')
	gl.glGetSynciv.gl = gl
	gl.glGetSynciv.__name__ = 'glGetSynciv'
	gl.glGetSynciv.errcheck = _errcheck
	gl.glGetSynciv.restype = GLvoid
	gl.glGetSynciv.argtypes = (GLsync, GLenum, GLsizei, POINTER(GLsizei), POINTER(GLint),)
	gl.glGetSyncivAPPLE = _getProcAddress('glGetSyncivAPPLE')
	gl.glGetSyncivAPPLE.gl = gl
	gl.glGetSyncivAPPLE.__name__ = 'glGetSyncivAPPLE'
	gl.glGetSyncivAPPLE.errcheck = _errcheck
	gl.glGetSyncivAPPLE.restype = GLvoid
	gl.glGetSyncivAPPLE.argtypes = (GLsync, GLenum, GLsizei, POINTER(GLsizei), POINTER(GLint),)
	gl.glGetTexBumpParameterfvATI = _getProcAddress('glGetTexBumpParameterfvATI')
	gl.glGetTexBumpParameterfvATI.gl = gl
	gl.glGetTexBumpParameterfvATI.__name__ = 'glGetTexBumpParameterfvATI'
	gl.glGetTexBumpParameterfvATI.errcheck = _errcheck
	gl.glGetTexBumpParameterfvATI.restype = GLvoid
	gl.glGetTexBumpParameterfvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glGetTexBumpParameterivATI = _getProcAddress('glGetTexBumpParameterivATI')
	gl.glGetTexBumpParameterivATI.gl = gl
	gl.glGetTexBumpParameterivATI.__name__ = 'glGetTexBumpParameterivATI'
	gl.glGetTexBumpParameterivATI.errcheck = _errcheck
	gl.glGetTexBumpParameterivATI.restype = GLvoid
	gl.glGetTexBumpParameterivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glGetTexEnvfv = _getProcAddress('glGetTexEnvfv')
	gl.glGetTexEnvfv.gl = gl
	gl.glGetTexEnvfv.__name__ = 'glGetTexEnvfv'
	gl.glGetTexEnvfv.errcheck = _errcheck
	gl.glGetTexEnvfv.restype = GLvoid
	gl.glGetTexEnvfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetTexEnviv = _getProcAddress('glGetTexEnviv')
	gl.glGetTexEnviv.gl = gl
	gl.glGetTexEnviv.__name__ = 'glGetTexEnviv'
	gl.glGetTexEnviv.errcheck = _errcheck
	gl.glGetTexEnviv.restype = GLvoid
	gl.glGetTexEnviv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetTexEnvxv = _getProcAddress('glGetTexEnvxv')
	gl.glGetTexEnvxv.gl = gl
	gl.glGetTexEnvxv.__name__ = 'glGetTexEnvxv'
	gl.glGetTexEnvxv.errcheck = _errcheck
	gl.glGetTexEnvxv.restype = GLvoid
	gl.glGetTexEnvxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetTexEnvxvOES = _getProcAddress('glGetTexEnvxvOES')
	gl.glGetTexEnvxvOES.gl = gl
	gl.glGetTexEnvxvOES.__name__ = 'glGetTexEnvxvOES'
	gl.glGetTexEnvxvOES.errcheck = _errcheck
	gl.glGetTexEnvxvOES.restype = GLvoid
	gl.glGetTexEnvxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetTexFilterFuncSGIS = _getProcAddress('glGetTexFilterFuncSGIS')
	gl.glGetTexFilterFuncSGIS.gl = gl
	gl.glGetTexFilterFuncSGIS.__name__ = 'glGetTexFilterFuncSGIS'
	gl.glGetTexFilterFuncSGIS.errcheck = _errcheck
	gl.glGetTexFilterFuncSGIS.restype = GLvoid
	gl.glGetTexFilterFuncSGIS.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetTexGendv = _getProcAddress('glGetTexGendv')
	gl.glGetTexGendv.gl = gl
	gl.glGetTexGendv.__name__ = 'glGetTexGendv'
	gl.glGetTexGendv.errcheck = _errcheck
	gl.glGetTexGendv.restype = GLvoid
	gl.glGetTexGendv.argtypes = (GLenum, GLenum, POINTER(GLdouble),)
	gl.glGetTexGenfv = _getProcAddress('glGetTexGenfv')
	gl.glGetTexGenfv.gl = gl
	gl.glGetTexGenfv.__name__ = 'glGetTexGenfv'
	gl.glGetTexGenfv.errcheck = _errcheck
	gl.glGetTexGenfv.restype = GLvoid
	gl.glGetTexGenfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetTexGenfvOES = _getProcAddress('glGetTexGenfvOES')
	gl.glGetTexGenfvOES.gl = gl
	gl.glGetTexGenfvOES.__name__ = 'glGetTexGenfvOES'
	gl.glGetTexGenfvOES.errcheck = _errcheck
	gl.glGetTexGenfvOES.restype = GLvoid
	gl.glGetTexGenfvOES.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetTexGeniv = _getProcAddress('glGetTexGeniv')
	gl.glGetTexGeniv.gl = gl
	gl.glGetTexGeniv.__name__ = 'glGetTexGeniv'
	gl.glGetTexGeniv.errcheck = _errcheck
	gl.glGetTexGeniv.restype = GLvoid
	gl.glGetTexGeniv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetTexGenivOES = _getProcAddress('glGetTexGenivOES')
	gl.glGetTexGenivOES.gl = gl
	gl.glGetTexGenivOES.__name__ = 'glGetTexGenivOES'
	gl.glGetTexGenivOES.errcheck = _errcheck
	gl.glGetTexGenivOES.restype = GLvoid
	gl.glGetTexGenivOES.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetTexGenxvOES = _getProcAddress('glGetTexGenxvOES')
	gl.glGetTexGenxvOES.gl = gl
	gl.glGetTexGenxvOES.__name__ = 'glGetTexGenxvOES'
	gl.glGetTexGenxvOES.errcheck = _errcheck
	gl.glGetTexGenxvOES.restype = GLvoid
	gl.glGetTexGenxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetTexImage = _getProcAddress('glGetTexImage')
	gl.glGetTexImage.gl = gl
	gl.glGetTexImage.__name__ = 'glGetTexImage'
	gl.glGetTexImage.errcheck = _errcheck
	gl.glGetTexImage.restype = GLvoid
	gl.glGetTexImage.argtypes = (GLenum, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetTexLevelParameterfv = _getProcAddress('glGetTexLevelParameterfv')
	gl.glGetTexLevelParameterfv.gl = gl
	gl.glGetTexLevelParameterfv.__name__ = 'glGetTexLevelParameterfv'
	gl.glGetTexLevelParameterfv.errcheck = _errcheck
	gl.glGetTexLevelParameterfv.restype = GLvoid
	gl.glGetTexLevelParameterfv.argtypes = (GLenum, GLint, GLenum, POINTER(GLfloat),)
	gl.glGetTexLevelParameteriv = _getProcAddress('glGetTexLevelParameteriv')
	gl.glGetTexLevelParameteriv.gl = gl
	gl.glGetTexLevelParameteriv.__name__ = 'glGetTexLevelParameteriv'
	gl.glGetTexLevelParameteriv.errcheck = _errcheck
	gl.glGetTexLevelParameteriv.restype = GLvoid
	gl.glGetTexLevelParameteriv.argtypes = (GLenum, GLint, GLenum, POINTER(GLint),)
	gl.glGetTexLevelParameterxvOES = _getProcAddress('glGetTexLevelParameterxvOES')
	gl.glGetTexLevelParameterxvOES.gl = gl
	gl.glGetTexLevelParameterxvOES.__name__ = 'glGetTexLevelParameterxvOES'
	gl.glGetTexLevelParameterxvOES.errcheck = _errcheck
	gl.glGetTexLevelParameterxvOES.restype = GLvoid
	gl.glGetTexLevelParameterxvOES.argtypes = (GLenum, GLint, GLenum, POINTER(GLfixed),)
	gl.glGetTexParameterIiv = _getProcAddress('glGetTexParameterIiv')
	gl.glGetTexParameterIiv.gl = gl
	gl.glGetTexParameterIiv.__name__ = 'glGetTexParameterIiv'
	gl.glGetTexParameterIiv.errcheck = _errcheck
	gl.glGetTexParameterIiv.restype = GLvoid
	gl.glGetTexParameterIiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetTexParameterIivEXT = _getProcAddress('glGetTexParameterIivEXT')
	gl.glGetTexParameterIivEXT.gl = gl
	gl.glGetTexParameterIivEXT.__name__ = 'glGetTexParameterIivEXT'
	gl.glGetTexParameterIivEXT.errcheck = _errcheck
	gl.glGetTexParameterIivEXT.restype = GLvoid
	gl.glGetTexParameterIivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetTexParameterIuiv = _getProcAddress('glGetTexParameterIuiv')
	gl.glGetTexParameterIuiv.gl = gl
	gl.glGetTexParameterIuiv.__name__ = 'glGetTexParameterIuiv'
	gl.glGetTexParameterIuiv.errcheck = _errcheck
	gl.glGetTexParameterIuiv.restype = GLvoid
	gl.glGetTexParameterIuiv.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glGetTexParameterIuivEXT = _getProcAddress('glGetTexParameterIuivEXT')
	gl.glGetTexParameterIuivEXT.gl = gl
	gl.glGetTexParameterIuivEXT.__name__ = 'glGetTexParameterIuivEXT'
	gl.glGetTexParameterIuivEXT.errcheck = _errcheck
	gl.glGetTexParameterIuivEXT.restype = GLvoid
	gl.glGetTexParameterIuivEXT.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glGetTexParameterPointervAPPLE = _getProcAddress('glGetTexParameterPointervAPPLE')
	gl.glGetTexParameterPointervAPPLE.gl = gl
	gl.glGetTexParameterPointervAPPLE.__name__ = 'glGetTexParameterPointervAPPLE'
	gl.glGetTexParameterPointervAPPLE.errcheck = _errcheck
	gl.glGetTexParameterPointervAPPLE.restype = GLvoid
	gl.glGetTexParameterPointervAPPLE.argtypes = (GLenum, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetTexParameterfv = _getProcAddress('glGetTexParameterfv')
	gl.glGetTexParameterfv.gl = gl
	gl.glGetTexParameterfv.__name__ = 'glGetTexParameterfv'
	gl.glGetTexParameterfv.errcheck = _errcheck
	gl.glGetTexParameterfv.restype = GLvoid
	gl.glGetTexParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetTexParameteriv = _getProcAddress('glGetTexParameteriv')
	gl.glGetTexParameteriv.gl = gl
	gl.glGetTexParameteriv.__name__ = 'glGetTexParameteriv'
	gl.glGetTexParameteriv.errcheck = _errcheck
	gl.glGetTexParameteriv.restype = GLvoid
	gl.glGetTexParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glGetTexParameterxv = _getProcAddress('glGetTexParameterxv')
	gl.glGetTexParameterxv.gl = gl
	gl.glGetTexParameterxv.__name__ = 'glGetTexParameterxv'
	gl.glGetTexParameterxv.errcheck = _errcheck
	gl.glGetTexParameterxv.restype = GLvoid
	gl.glGetTexParameterxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetTexParameterxvOES = _getProcAddress('glGetTexParameterxvOES')
	gl.glGetTexParameterxvOES.gl = gl
	gl.glGetTexParameterxvOES.__name__ = 'glGetTexParameterxvOES'
	gl.glGetTexParameterxvOES.errcheck = _errcheck
	gl.glGetTexParameterxvOES.restype = GLvoid
	gl.glGetTexParameterxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glGetTextureHandleARB = _getProcAddress('glGetTextureHandleARB')
	gl.glGetTextureHandleARB.gl = gl
	gl.glGetTextureHandleARB.__name__ = 'glGetTextureHandleARB'
	gl.glGetTextureHandleARB.errcheck = _errcheck
	gl.glGetTextureHandleARB.restype = GLuint64
	gl.glGetTextureHandleARB.argtypes = (GLuint,)
	gl.glGetTextureHandleNV = _getProcAddress('glGetTextureHandleNV')
	gl.glGetTextureHandleNV.gl = gl
	gl.glGetTextureHandleNV.__name__ = 'glGetTextureHandleNV'
	gl.glGetTextureHandleNV.errcheck = _errcheck
	gl.glGetTextureHandleNV.restype = GLuint64
	gl.glGetTextureHandleNV.argtypes = (GLuint,)
	gl.glGetTextureImage = _getProcAddress('glGetTextureImage')
	gl.glGetTextureImage.gl = gl
	gl.glGetTextureImage.__name__ = 'glGetTextureImage'
	gl.glGetTextureImage.errcheck = _errcheck
	gl.glGetTextureImage.restype = GLvoid
	gl.glGetTextureImage.argtypes = (GLuint, GLint, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetTextureImageEXT = _getProcAddress('glGetTextureImageEXT')
	gl.glGetTextureImageEXT.gl = gl
	gl.glGetTextureImageEXT.__name__ = 'glGetTextureImageEXT'
	gl.glGetTextureImageEXT.errcheck = _errcheck
	gl.glGetTextureImageEXT.restype = GLvoid
	gl.glGetTextureImageEXT.argtypes = (GLuint, GLenum, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glGetTextureLevelParameterfv = _getProcAddress('glGetTextureLevelParameterfv')
	gl.glGetTextureLevelParameterfv.gl = gl
	gl.glGetTextureLevelParameterfv.__name__ = 'glGetTextureLevelParameterfv'
	gl.glGetTextureLevelParameterfv.errcheck = _errcheck
	gl.glGetTextureLevelParameterfv.restype = GLvoid
	gl.glGetTextureLevelParameterfv.argtypes = (GLuint, GLint, GLenum, POINTER(GLfloat),)
	gl.glGetTextureLevelParameterfvEXT = _getProcAddress('glGetTextureLevelParameterfvEXT')
	gl.glGetTextureLevelParameterfvEXT.gl = gl
	gl.glGetTextureLevelParameterfvEXT.__name__ = 'glGetTextureLevelParameterfvEXT'
	gl.glGetTextureLevelParameterfvEXT.errcheck = _errcheck
	gl.glGetTextureLevelParameterfvEXT.restype = GLvoid
	gl.glGetTextureLevelParameterfvEXT.argtypes = (GLuint, GLenum, GLint, GLenum, POINTER(GLfloat),)
	gl.glGetTextureLevelParameteriv = _getProcAddress('glGetTextureLevelParameteriv')
	gl.glGetTextureLevelParameteriv.gl = gl
	gl.glGetTextureLevelParameteriv.__name__ = 'glGetTextureLevelParameteriv'
	gl.glGetTextureLevelParameteriv.errcheck = _errcheck
	gl.glGetTextureLevelParameteriv.restype = GLvoid
	gl.glGetTextureLevelParameteriv.argtypes = (GLuint, GLint, GLenum, POINTER(GLint),)
	gl.glGetTextureLevelParameterivEXT = _getProcAddress('glGetTextureLevelParameterivEXT')
	gl.glGetTextureLevelParameterivEXT.gl = gl
	gl.glGetTextureLevelParameterivEXT.__name__ = 'glGetTextureLevelParameterivEXT'
	gl.glGetTextureLevelParameterivEXT.errcheck = _errcheck
	gl.glGetTextureLevelParameterivEXT.restype = GLvoid
	gl.glGetTextureLevelParameterivEXT.argtypes = (GLuint, GLenum, GLint, GLenum, POINTER(GLint),)
	gl.glGetTextureParameterIiv = _getProcAddress('glGetTextureParameterIiv')
	gl.glGetTextureParameterIiv.gl = gl
	gl.glGetTextureParameterIiv.__name__ = 'glGetTextureParameterIiv'
	gl.glGetTextureParameterIiv.errcheck = _errcheck
	gl.glGetTextureParameterIiv.restype = GLvoid
	gl.glGetTextureParameterIiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetTextureParameterIivEXT = _getProcAddress('glGetTextureParameterIivEXT')
	gl.glGetTextureParameterIivEXT.gl = gl
	gl.glGetTextureParameterIivEXT.__name__ = 'glGetTextureParameterIivEXT'
	gl.glGetTextureParameterIivEXT.errcheck = _errcheck
	gl.glGetTextureParameterIivEXT.restype = GLvoid
	gl.glGetTextureParameterIivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetTextureParameterIuiv = _getProcAddress('glGetTextureParameterIuiv')
	gl.glGetTextureParameterIuiv.gl = gl
	gl.glGetTextureParameterIuiv.__name__ = 'glGetTextureParameterIuiv'
	gl.glGetTextureParameterIuiv.errcheck = _errcheck
	gl.glGetTextureParameterIuiv.restype = GLvoid
	gl.glGetTextureParameterIuiv.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetTextureParameterIuivEXT = _getProcAddress('glGetTextureParameterIuivEXT')
	gl.glGetTextureParameterIuivEXT.gl = gl
	gl.glGetTextureParameterIuivEXT.__name__ = 'glGetTextureParameterIuivEXT'
	gl.glGetTextureParameterIuivEXT.errcheck = _errcheck
	gl.glGetTextureParameterIuivEXT.restype = GLvoid
	gl.glGetTextureParameterIuivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLuint),)
	gl.glGetTextureParameterfv = _getProcAddress('glGetTextureParameterfv')
	gl.glGetTextureParameterfv.gl = gl
	gl.glGetTextureParameterfv.__name__ = 'glGetTextureParameterfv'
	gl.glGetTextureParameterfv.errcheck = _errcheck
	gl.glGetTextureParameterfv.restype = GLvoid
	gl.glGetTextureParameterfv.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetTextureParameterfvEXT = _getProcAddress('glGetTextureParameterfvEXT')
	gl.glGetTextureParameterfvEXT.gl = gl
	gl.glGetTextureParameterfvEXT.__name__ = 'glGetTextureParameterfvEXT'
	gl.glGetTextureParameterfvEXT.errcheck = _errcheck
	gl.glGetTextureParameterfvEXT.restype = GLvoid
	gl.glGetTextureParameterfvEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLfloat),)
	gl.glGetTextureParameteriv = _getProcAddress('glGetTextureParameteriv')
	gl.glGetTextureParameteriv.gl = gl
	gl.glGetTextureParameteriv.__name__ = 'glGetTextureParameteriv'
	gl.glGetTextureParameteriv.errcheck = _errcheck
	gl.glGetTextureParameteriv.restype = GLvoid
	gl.glGetTextureParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetTextureParameterivEXT = _getProcAddress('glGetTextureParameterivEXT')
	gl.glGetTextureParameterivEXT.gl = gl
	gl.glGetTextureParameterivEXT.__name__ = 'glGetTextureParameterivEXT'
	gl.glGetTextureParameterivEXT.errcheck = _errcheck
	gl.glGetTextureParameterivEXT.restype = GLvoid
	gl.glGetTextureParameterivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glGetTextureSamplerHandleARB = _getProcAddress('glGetTextureSamplerHandleARB')
	gl.glGetTextureSamplerHandleARB.gl = gl
	gl.glGetTextureSamplerHandleARB.__name__ = 'glGetTextureSamplerHandleARB'
	gl.glGetTextureSamplerHandleARB.errcheck = _errcheck
	gl.glGetTextureSamplerHandleARB.restype = GLuint64
	gl.glGetTextureSamplerHandleARB.argtypes = (GLuint, GLuint,)
	gl.glGetTextureSamplerHandleNV = _getProcAddress('glGetTextureSamplerHandleNV')
	gl.glGetTextureSamplerHandleNV.gl = gl
	gl.glGetTextureSamplerHandleNV.__name__ = 'glGetTextureSamplerHandleNV'
	gl.glGetTextureSamplerHandleNV.errcheck = _errcheck
	gl.glGetTextureSamplerHandleNV.restype = GLuint64
	gl.glGetTextureSamplerHandleNV.argtypes = (GLuint, GLuint,)
	gl.glGetTextureSubImage = _getProcAddress('glGetTextureSubImage')
	gl.glGetTextureSubImage.gl = gl
	gl.glGetTextureSubImage.__name__ = 'glGetTextureSubImage'
	gl.glGetTextureSubImage.errcheck = _errcheck
	gl.glGetTextureSubImage.restype = GLvoid
	gl.glGetTextureSubImage.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetTrackMatrixivNV = _getProcAddress('glGetTrackMatrixivNV')
	gl.glGetTrackMatrixivNV.gl = gl
	gl.glGetTrackMatrixivNV.__name__ = 'glGetTrackMatrixivNV'
	gl.glGetTrackMatrixivNV.errcheck = _errcheck
	gl.glGetTrackMatrixivNV.restype = GLvoid
	gl.glGetTrackMatrixivNV.argtypes = (GLenum, GLuint, GLenum, POINTER(GLint),)
	gl.glGetTransformFeedbackVarying = _getProcAddress('glGetTransformFeedbackVarying')
	gl.glGetTransformFeedbackVarying.gl = gl
	gl.glGetTransformFeedbackVarying.__name__ = 'glGetTransformFeedbackVarying'
	gl.glGetTransformFeedbackVarying.errcheck = _errcheck
	gl.glGetTransformFeedbackVarying.restype = GLvoid
	gl.glGetTransformFeedbackVarying.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLsizei), POINTER(GLenum), POINTER(GLchar),)
	gl.glGetTransformFeedbackVaryingEXT = _getProcAddress('glGetTransformFeedbackVaryingEXT')
	gl.glGetTransformFeedbackVaryingEXT.gl = gl
	gl.glGetTransformFeedbackVaryingEXT.__name__ = 'glGetTransformFeedbackVaryingEXT'
	gl.glGetTransformFeedbackVaryingEXT.errcheck = _errcheck
	gl.glGetTransformFeedbackVaryingEXT.restype = GLvoid
	gl.glGetTransformFeedbackVaryingEXT.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLsizei), POINTER(GLsizei), POINTER(GLenum), POINTER(GLchar),)
	gl.glGetTransformFeedbackVaryingNV = _getProcAddress('glGetTransformFeedbackVaryingNV')
	gl.glGetTransformFeedbackVaryingNV.gl = gl
	gl.glGetTransformFeedbackVaryingNV.__name__ = 'glGetTransformFeedbackVaryingNV'
	gl.glGetTransformFeedbackVaryingNV.errcheck = _errcheck
	gl.glGetTransformFeedbackVaryingNV.restype = GLvoid
	gl.glGetTransformFeedbackVaryingNV.argtypes = (GLuint, GLuint, POINTER(GLint),)
	gl.glGetTransformFeedbacki64_v = _getProcAddress('glGetTransformFeedbacki64_v')
	gl.glGetTransformFeedbacki64_v.gl = gl
	gl.glGetTransformFeedbacki64_v.__name__ = 'glGetTransformFeedbacki64_v'
	gl.glGetTransformFeedbacki64_v.errcheck = _errcheck
	gl.glGetTransformFeedbacki64_v.restype = GLvoid
	gl.glGetTransformFeedbacki64_v.argtypes = (GLuint, GLenum, GLuint, POINTER(GLint64),)
	gl.glGetTransformFeedbacki_v = _getProcAddress('glGetTransformFeedbacki_v')
	gl.glGetTransformFeedbacki_v.gl = gl
	gl.glGetTransformFeedbacki_v.__name__ = 'glGetTransformFeedbacki_v'
	gl.glGetTransformFeedbacki_v.errcheck = _errcheck
	gl.glGetTransformFeedbacki_v.restype = GLvoid
	gl.glGetTransformFeedbacki_v.argtypes = (GLuint, GLenum, GLuint, POINTER(GLint),)
	gl.glGetTransformFeedbackiv = _getProcAddress('glGetTransformFeedbackiv')
	gl.glGetTransformFeedbackiv.gl = gl
	gl.glGetTransformFeedbackiv.__name__ = 'glGetTransformFeedbackiv'
	gl.glGetTransformFeedbackiv.errcheck = _errcheck
	gl.glGetTransformFeedbackiv.restype = GLvoid
	gl.glGetTransformFeedbackiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetTranslatedShaderSourceANGLE = _getProcAddress('glGetTranslatedShaderSourceANGLE')
	gl.glGetTranslatedShaderSourceANGLE.gl = gl
	gl.glGetTranslatedShaderSourceANGLE.__name__ = 'glGetTranslatedShaderSourceANGLE'
	gl.glGetTranslatedShaderSourceANGLE.errcheck = _errcheck
	gl.glGetTranslatedShaderSourceANGLE.restype = GLvoid
	gl.glGetTranslatedShaderSourceANGLE.argtypes = (GLuint, GLsizei, POINTER(GLsizei), POINTER(GLchar),)
	gl.glGetUniformBlockIndex = _getProcAddress('glGetUniformBlockIndex')
	gl.glGetUniformBlockIndex.gl = gl
	gl.glGetUniformBlockIndex.__name__ = 'glGetUniformBlockIndex'
	gl.glGetUniformBlockIndex.errcheck = _errcheck
	gl.glGetUniformBlockIndex.restype = GLuint
	gl.glGetUniformBlockIndex.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetUniformBufferSizeEXT = _getProcAddress('glGetUniformBufferSizeEXT')
	gl.glGetUniformBufferSizeEXT.gl = gl
	gl.glGetUniformBufferSizeEXT.__name__ = 'glGetUniformBufferSizeEXT'
	gl.glGetUniformBufferSizeEXT.errcheck = _errcheck
	gl.glGetUniformBufferSizeEXT.restype = GLint
	gl.glGetUniformBufferSizeEXT.argtypes = (GLuint, GLint,)
	gl.glGetUniformIndices = _getProcAddress('glGetUniformIndices')
	gl.glGetUniformIndices.gl = gl
	gl.glGetUniformIndices.__name__ = 'glGetUniformIndices'
	gl.glGetUniformIndices.errcheck = _errcheck
	gl.glGetUniformIndices.restype = GLvoid
	gl.glGetUniformIndices.argtypes = (GLuint, GLsizei, POINTER(POINTER(GLchar)), POINTER(GLuint),)
	gl.glGetUniformLocation = _getProcAddress('glGetUniformLocation')
	gl.glGetUniformLocation.gl = gl
	gl.glGetUniformLocation.__name__ = 'glGetUniformLocation'
	gl.glGetUniformLocation.errcheck = _errcheck
	gl.glGetUniformLocation.restype = GLint
	gl.glGetUniformLocation.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetUniformLocationARB = _getProcAddress('glGetUniformLocationARB')
	gl.glGetUniformLocationARB.gl = gl
	gl.glGetUniformLocationARB.__name__ = 'glGetUniformLocationARB'
	gl.glGetUniformLocationARB.errcheck = _errcheck
	gl.glGetUniformLocationARB.restype = GLint
	gl.glGetUniformLocationARB.argtypes = (GLhandleARB, POINTER(GLcharARB),)
	gl.glGetUniformOffsetEXT = _getProcAddress('glGetUniformOffsetEXT')
	gl.glGetUniformOffsetEXT.gl = gl
	gl.glGetUniformOffsetEXT.__name__ = 'glGetUniformOffsetEXT'
	gl.glGetUniformOffsetEXT.errcheck = _errcheck
	gl.glGetUniformOffsetEXT.restype = GLintptr
	gl.glGetUniformOffsetEXT.argtypes = (GLuint, GLint,)
	gl.glGetUniformSubroutineuiv = _getProcAddress('glGetUniformSubroutineuiv')
	gl.glGetUniformSubroutineuiv.gl = gl
	gl.glGetUniformSubroutineuiv.__name__ = 'glGetUniformSubroutineuiv'
	gl.glGetUniformSubroutineuiv.errcheck = _errcheck
	gl.glGetUniformSubroutineuiv.restype = GLvoid
	gl.glGetUniformSubroutineuiv.argtypes = (GLenum, GLint, POINTER(GLuint),)
	gl.glGetUniformdv = _getProcAddress('glGetUniformdv')
	gl.glGetUniformdv.gl = gl
	gl.glGetUniformdv.__name__ = 'glGetUniformdv'
	gl.glGetUniformdv.errcheck = _errcheck
	gl.glGetUniformdv.restype = GLvoid
	gl.glGetUniformdv.argtypes = (GLuint, GLint, POINTER(GLdouble),)
	gl.glGetUniformfv = _getProcAddress('glGetUniformfv')
	gl.glGetUniformfv.gl = gl
	gl.glGetUniformfv.__name__ = 'glGetUniformfv'
	gl.glGetUniformfv.errcheck = _errcheck
	gl.glGetUniformfv.restype = GLvoid
	gl.glGetUniformfv.argtypes = (GLuint, GLint, POINTER(GLfloat),)
	gl.glGetUniformfvARB = _getProcAddress('glGetUniformfvARB')
	gl.glGetUniformfvARB.gl = gl
	gl.glGetUniformfvARB.__name__ = 'glGetUniformfvARB'
	gl.glGetUniformfvARB.errcheck = _errcheck
	gl.glGetUniformfvARB.restype = GLvoid
	gl.glGetUniformfvARB.argtypes = (GLhandleARB, GLint, POINTER(GLfloat),)
	gl.glGetUniformi64vNV = _getProcAddress('glGetUniformi64vNV')
	gl.glGetUniformi64vNV.gl = gl
	gl.glGetUniformi64vNV.__name__ = 'glGetUniformi64vNV'
	gl.glGetUniformi64vNV.errcheck = _errcheck
	gl.glGetUniformi64vNV.restype = GLvoid
	gl.glGetUniformi64vNV.argtypes = (GLuint, GLint, POINTER(GLint64EXT),)
	gl.glGetUniformiv = _getProcAddress('glGetUniformiv')
	gl.glGetUniformiv.gl = gl
	gl.glGetUniformiv.__name__ = 'glGetUniformiv'
	gl.glGetUniformiv.errcheck = _errcheck
	gl.glGetUniformiv.restype = GLvoid
	gl.glGetUniformiv.argtypes = (GLuint, GLint, POINTER(GLint),)
	gl.glGetUniformivARB = _getProcAddress('glGetUniformivARB')
	gl.glGetUniformivARB.gl = gl
	gl.glGetUniformivARB.__name__ = 'glGetUniformivARB'
	gl.glGetUniformivARB.errcheck = _errcheck
	gl.glGetUniformivARB.restype = GLvoid
	gl.glGetUniformivARB.argtypes = (GLhandleARB, GLint, POINTER(GLint),)
	gl.glGetUniformui64vNV = _getProcAddress('glGetUniformui64vNV')
	gl.glGetUniformui64vNV.gl = gl
	gl.glGetUniformui64vNV.__name__ = 'glGetUniformui64vNV'
	gl.glGetUniformui64vNV.errcheck = _errcheck
	gl.glGetUniformui64vNV.restype = GLvoid
	gl.glGetUniformui64vNV.argtypes = (GLuint, GLint, POINTER(GLuint64EXT),)
	gl.glGetUniformuiv = _getProcAddress('glGetUniformuiv')
	gl.glGetUniformuiv.gl = gl
	gl.glGetUniformuiv.__name__ = 'glGetUniformuiv'
	gl.glGetUniformuiv.errcheck = _errcheck
	gl.glGetUniformuiv.restype = GLvoid
	gl.glGetUniformuiv.argtypes = (GLuint, GLint, POINTER(GLuint),)
	gl.glGetUniformuivEXT = _getProcAddress('glGetUniformuivEXT')
	gl.glGetUniformuivEXT.gl = gl
	gl.glGetUniformuivEXT.__name__ = 'glGetUniformuivEXT'
	gl.glGetUniformuivEXT.errcheck = _errcheck
	gl.glGetUniformuivEXT.restype = GLvoid
	gl.glGetUniformuivEXT.argtypes = (GLuint, GLint, POINTER(GLuint),)
	gl.glGetVariantArrayObjectfvATI = _getProcAddress('glGetVariantArrayObjectfvATI')
	gl.glGetVariantArrayObjectfvATI.gl = gl
	gl.glGetVariantArrayObjectfvATI.__name__ = 'glGetVariantArrayObjectfvATI'
	gl.glGetVariantArrayObjectfvATI.errcheck = _errcheck
	gl.glGetVariantArrayObjectfvATI.restype = GLvoid
	gl.glGetVariantArrayObjectfvATI.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVariantArrayObjectivATI = _getProcAddress('glGetVariantArrayObjectivATI')
	gl.glGetVariantArrayObjectivATI.gl = gl
	gl.glGetVariantArrayObjectivATI.__name__ = 'glGetVariantArrayObjectivATI'
	gl.glGetVariantArrayObjectivATI.errcheck = _errcheck
	gl.glGetVariantArrayObjectivATI.restype = GLvoid
	gl.glGetVariantArrayObjectivATI.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVariantBooleanvEXT = _getProcAddress('glGetVariantBooleanvEXT')
	gl.glGetVariantBooleanvEXT.gl = gl
	gl.glGetVariantBooleanvEXT.__name__ = 'glGetVariantBooleanvEXT'
	gl.glGetVariantBooleanvEXT.errcheck = _errcheck
	gl.glGetVariantBooleanvEXT.restype = GLvoid
	gl.glGetVariantBooleanvEXT.argtypes = (GLuint, GLenum, POINTER(GLboolean),)
	gl.glGetVariantFloatvEXT = _getProcAddress('glGetVariantFloatvEXT')
	gl.glGetVariantFloatvEXT.gl = gl
	gl.glGetVariantFloatvEXT.__name__ = 'glGetVariantFloatvEXT'
	gl.glGetVariantFloatvEXT.errcheck = _errcheck
	gl.glGetVariantFloatvEXT.restype = GLvoid
	gl.glGetVariantFloatvEXT.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVariantIntegervEXT = _getProcAddress('glGetVariantIntegervEXT')
	gl.glGetVariantIntegervEXT.gl = gl
	gl.glGetVariantIntegervEXT.__name__ = 'glGetVariantIntegervEXT'
	gl.glGetVariantIntegervEXT.errcheck = _errcheck
	gl.glGetVariantIntegervEXT.restype = GLvoid
	gl.glGetVariantIntegervEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVariantPointervEXT = _getProcAddress('glGetVariantPointervEXT')
	gl.glGetVariantPointervEXT.gl = gl
	gl.glGetVariantPointervEXT.__name__ = 'glGetVariantPointervEXT'
	gl.glGetVariantPointervEXT.errcheck = _errcheck
	gl.glGetVariantPointervEXT.restype = GLvoid
	gl.glGetVariantPointervEXT.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetVaryingLocationNV = _getProcAddress('glGetVaryingLocationNV')
	gl.glGetVaryingLocationNV.gl = gl
	gl.glGetVaryingLocationNV.__name__ = 'glGetVaryingLocationNV'
	gl.glGetVaryingLocationNV.errcheck = _errcheck
	gl.glGetVaryingLocationNV.restype = GLint
	gl.glGetVaryingLocationNV.argtypes = (GLuint, POINTER(GLchar),)
	gl.glGetVertexArrayIndexed64iv = _getProcAddress('glGetVertexArrayIndexed64iv')
	gl.glGetVertexArrayIndexed64iv.gl = gl
	gl.glGetVertexArrayIndexed64iv.__name__ = 'glGetVertexArrayIndexed64iv'
	gl.glGetVertexArrayIndexed64iv.errcheck = _errcheck
	gl.glGetVertexArrayIndexed64iv.restype = GLvoid
	gl.glGetVertexArrayIndexed64iv.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint64),)
	gl.glGetVertexArrayIndexediv = _getProcAddress('glGetVertexArrayIndexediv')
	gl.glGetVertexArrayIndexediv.gl = gl
	gl.glGetVertexArrayIndexediv.__name__ = 'glGetVertexArrayIndexediv'
	gl.glGetVertexArrayIndexediv.errcheck = _errcheck
	gl.glGetVertexArrayIndexediv.restype = GLvoid
	gl.glGetVertexArrayIndexediv.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexArrayIntegeri_vEXT = _getProcAddress('glGetVertexArrayIntegeri_vEXT')
	gl.glGetVertexArrayIntegeri_vEXT.gl = gl
	gl.glGetVertexArrayIntegeri_vEXT.__name__ = 'glGetVertexArrayIntegeri_vEXT'
	gl.glGetVertexArrayIntegeri_vEXT.errcheck = _errcheck
	gl.glGetVertexArrayIntegeri_vEXT.restype = GLvoid
	gl.glGetVertexArrayIntegeri_vEXT.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexArrayIntegervEXT = _getProcAddress('glGetVertexArrayIntegervEXT')
	gl.glGetVertexArrayIntegervEXT.gl = gl
	gl.glGetVertexArrayIntegervEXT.__name__ = 'glGetVertexArrayIntegervEXT'
	gl.glGetVertexArrayIntegervEXT.errcheck = _errcheck
	gl.glGetVertexArrayIntegervEXT.restype = GLvoid
	gl.glGetVertexArrayIntegervEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexArrayPointeri_vEXT = _getProcAddress('glGetVertexArrayPointeri_vEXT')
	gl.glGetVertexArrayPointeri_vEXT.gl = gl
	gl.glGetVertexArrayPointeri_vEXT.__name__ = 'glGetVertexArrayPointeri_vEXT'
	gl.glGetVertexArrayPointeri_vEXT.errcheck = _errcheck
	gl.glGetVertexArrayPointeri_vEXT.restype = GLvoid
	gl.glGetVertexArrayPointeri_vEXT.argtypes = (GLuint, GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetVertexArrayPointervEXT = _getProcAddress('glGetVertexArrayPointervEXT')
	gl.glGetVertexArrayPointervEXT.gl = gl
	gl.glGetVertexArrayPointervEXT.__name__ = 'glGetVertexArrayPointervEXT'
	gl.glGetVertexArrayPointervEXT.errcheck = _errcheck
	gl.glGetVertexArrayPointervEXT.restype = GLvoid
	gl.glGetVertexArrayPointervEXT.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetVertexArrayiv = _getProcAddress('glGetVertexArrayiv')
	gl.glGetVertexArrayiv.gl = gl
	gl.glGetVertexArrayiv.__name__ = 'glGetVertexArrayiv'
	gl.glGetVertexArrayiv.errcheck = _errcheck
	gl.glGetVertexArrayiv.restype = GLvoid
	gl.glGetVertexArrayiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexAttribArrayObjectfvATI = _getProcAddress('glGetVertexAttribArrayObjectfvATI')
	gl.glGetVertexAttribArrayObjectfvATI.gl = gl
	gl.glGetVertexAttribArrayObjectfvATI.__name__ = 'glGetVertexAttribArrayObjectfvATI'
	gl.glGetVertexAttribArrayObjectfvATI.errcheck = _errcheck
	gl.glGetVertexAttribArrayObjectfvATI.restype = GLvoid
	gl.glGetVertexAttribArrayObjectfvATI.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVertexAttribArrayObjectivATI = _getProcAddress('glGetVertexAttribArrayObjectivATI')
	gl.glGetVertexAttribArrayObjectivATI.gl = gl
	gl.glGetVertexAttribArrayObjectivATI.__name__ = 'glGetVertexAttribArrayObjectivATI'
	gl.glGetVertexAttribArrayObjectivATI.errcheck = _errcheck
	gl.glGetVertexAttribArrayObjectivATI.restype = GLvoid
	gl.glGetVertexAttribArrayObjectivATI.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexAttribIiv = _getProcAddress('glGetVertexAttribIiv')
	gl.glGetVertexAttribIiv.gl = gl
	gl.glGetVertexAttribIiv.__name__ = 'glGetVertexAttribIiv'
	gl.glGetVertexAttribIiv.errcheck = _errcheck
	gl.glGetVertexAttribIiv.restype = GLvoid
	gl.glGetVertexAttribIiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexAttribIivEXT = _getProcAddress('glGetVertexAttribIivEXT')
	gl.glGetVertexAttribIivEXT.gl = gl
	gl.glGetVertexAttribIivEXT.__name__ = 'glGetVertexAttribIivEXT'
	gl.glGetVertexAttribIivEXT.errcheck = _errcheck
	gl.glGetVertexAttribIivEXT.restype = GLvoid
	gl.glGetVertexAttribIivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexAttribIuiv = _getProcAddress('glGetVertexAttribIuiv')
	gl.glGetVertexAttribIuiv.gl = gl
	gl.glGetVertexAttribIuiv.__name__ = 'glGetVertexAttribIuiv'
	gl.glGetVertexAttribIuiv.errcheck = _errcheck
	gl.glGetVertexAttribIuiv.restype = GLvoid
	gl.glGetVertexAttribIuiv.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetVertexAttribIuivEXT = _getProcAddress('glGetVertexAttribIuivEXT')
	gl.glGetVertexAttribIuivEXT.gl = gl
	gl.glGetVertexAttribIuivEXT.__name__ = 'glGetVertexAttribIuivEXT'
	gl.glGetVertexAttribIuivEXT.errcheck = _errcheck
	gl.glGetVertexAttribIuivEXT.restype = GLvoid
	gl.glGetVertexAttribIuivEXT.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetVertexAttribLdv = _getProcAddress('glGetVertexAttribLdv')
	gl.glGetVertexAttribLdv.gl = gl
	gl.glGetVertexAttribLdv.__name__ = 'glGetVertexAttribLdv'
	gl.glGetVertexAttribLdv.errcheck = _errcheck
	gl.glGetVertexAttribLdv.restype = GLvoid
	gl.glGetVertexAttribLdv.argtypes = (GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetVertexAttribLdvEXT = _getProcAddress('glGetVertexAttribLdvEXT')
	gl.glGetVertexAttribLdvEXT.gl = gl
	gl.glGetVertexAttribLdvEXT.__name__ = 'glGetVertexAttribLdvEXT'
	gl.glGetVertexAttribLdvEXT.errcheck = _errcheck
	gl.glGetVertexAttribLdvEXT.restype = GLvoid
	gl.glGetVertexAttribLdvEXT.argtypes = (GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetVertexAttribLi64vNV = _getProcAddress('glGetVertexAttribLi64vNV')
	gl.glGetVertexAttribLi64vNV.gl = gl
	gl.glGetVertexAttribLi64vNV.__name__ = 'glGetVertexAttribLi64vNV'
	gl.glGetVertexAttribLi64vNV.errcheck = _errcheck
	gl.glGetVertexAttribLi64vNV.restype = GLvoid
	gl.glGetVertexAttribLi64vNV.argtypes = (GLuint, GLenum, POINTER(GLint64EXT),)
	gl.glGetVertexAttribLui64vARB = _getProcAddress('glGetVertexAttribLui64vARB')
	gl.glGetVertexAttribLui64vARB.gl = gl
	gl.glGetVertexAttribLui64vARB.__name__ = 'glGetVertexAttribLui64vARB'
	gl.glGetVertexAttribLui64vARB.errcheck = _errcheck
	gl.glGetVertexAttribLui64vARB.restype = GLvoid
	gl.glGetVertexAttribLui64vARB.argtypes = (GLuint, GLenum, POINTER(GLuint64EXT),)
	gl.glGetVertexAttribLui64vNV = _getProcAddress('glGetVertexAttribLui64vNV')
	gl.glGetVertexAttribLui64vNV.gl = gl
	gl.glGetVertexAttribLui64vNV.__name__ = 'glGetVertexAttribLui64vNV'
	gl.glGetVertexAttribLui64vNV.errcheck = _errcheck
	gl.glGetVertexAttribLui64vNV.restype = GLvoid
	gl.glGetVertexAttribLui64vNV.argtypes = (GLuint, GLenum, POINTER(GLuint64EXT),)
	gl.glGetVertexAttribPointerv = _getProcAddress('glGetVertexAttribPointerv')
	gl.glGetVertexAttribPointerv.gl = gl
	gl.glGetVertexAttribPointerv.__name__ = 'glGetVertexAttribPointerv'
	gl.glGetVertexAttribPointerv.errcheck = _errcheck
	gl.glGetVertexAttribPointerv.restype = GLvoid
	gl.glGetVertexAttribPointerv.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetVertexAttribPointervARB = _getProcAddress('glGetVertexAttribPointervARB')
	gl.glGetVertexAttribPointervARB.gl = gl
	gl.glGetVertexAttribPointervARB.__name__ = 'glGetVertexAttribPointervARB'
	gl.glGetVertexAttribPointervARB.errcheck = _errcheck
	gl.glGetVertexAttribPointervARB.restype = GLvoid
	gl.glGetVertexAttribPointervARB.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetVertexAttribPointervNV = _getProcAddress('glGetVertexAttribPointervNV')
	gl.glGetVertexAttribPointervNV.gl = gl
	gl.glGetVertexAttribPointervNV.__name__ = 'glGetVertexAttribPointervNV'
	gl.glGetVertexAttribPointervNV.errcheck = _errcheck
	gl.glGetVertexAttribPointervNV.restype = GLvoid
	gl.glGetVertexAttribPointervNV.argtypes = (GLuint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glGetVertexAttribdv = _getProcAddress('glGetVertexAttribdv')
	gl.glGetVertexAttribdv.gl = gl
	gl.glGetVertexAttribdv.__name__ = 'glGetVertexAttribdv'
	gl.glGetVertexAttribdv.errcheck = _errcheck
	gl.glGetVertexAttribdv.restype = GLvoid
	gl.glGetVertexAttribdv.argtypes = (GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetVertexAttribdvARB = _getProcAddress('glGetVertexAttribdvARB')
	gl.glGetVertexAttribdvARB.gl = gl
	gl.glGetVertexAttribdvARB.__name__ = 'glGetVertexAttribdvARB'
	gl.glGetVertexAttribdvARB.errcheck = _errcheck
	gl.glGetVertexAttribdvARB.restype = GLvoid
	gl.glGetVertexAttribdvARB.argtypes = (GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetVertexAttribdvNV = _getProcAddress('glGetVertexAttribdvNV')
	gl.glGetVertexAttribdvNV.gl = gl
	gl.glGetVertexAttribdvNV.__name__ = 'glGetVertexAttribdvNV'
	gl.glGetVertexAttribdvNV.errcheck = _errcheck
	gl.glGetVertexAttribdvNV.restype = GLvoid
	gl.glGetVertexAttribdvNV.argtypes = (GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetVertexAttribfv = _getProcAddress('glGetVertexAttribfv')
	gl.glGetVertexAttribfv.gl = gl
	gl.glGetVertexAttribfv.__name__ = 'glGetVertexAttribfv'
	gl.glGetVertexAttribfv.errcheck = _errcheck
	gl.glGetVertexAttribfv.restype = GLvoid
	gl.glGetVertexAttribfv.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVertexAttribfvARB = _getProcAddress('glGetVertexAttribfvARB')
	gl.glGetVertexAttribfvARB.gl = gl
	gl.glGetVertexAttribfvARB.__name__ = 'glGetVertexAttribfvARB'
	gl.glGetVertexAttribfvARB.errcheck = _errcheck
	gl.glGetVertexAttribfvARB.restype = GLvoid
	gl.glGetVertexAttribfvARB.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVertexAttribfvNV = _getProcAddress('glGetVertexAttribfvNV')
	gl.glGetVertexAttribfvNV.gl = gl
	gl.glGetVertexAttribfvNV.__name__ = 'glGetVertexAttribfvNV'
	gl.glGetVertexAttribfvNV.errcheck = _errcheck
	gl.glGetVertexAttribfvNV.restype = GLvoid
	gl.glGetVertexAttribfvNV.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVertexAttribiv = _getProcAddress('glGetVertexAttribiv')
	gl.glGetVertexAttribiv.gl = gl
	gl.glGetVertexAttribiv.__name__ = 'glGetVertexAttribiv'
	gl.glGetVertexAttribiv.errcheck = _errcheck
	gl.glGetVertexAttribiv.restype = GLvoid
	gl.glGetVertexAttribiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexAttribivARB = _getProcAddress('glGetVertexAttribivARB')
	gl.glGetVertexAttribivARB.gl = gl
	gl.glGetVertexAttribivARB.__name__ = 'glGetVertexAttribivARB'
	gl.glGetVertexAttribivARB.errcheck = _errcheck
	gl.glGetVertexAttribivARB.restype = GLvoid
	gl.glGetVertexAttribivARB.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVertexAttribivNV = _getProcAddress('glGetVertexAttribivNV')
	gl.glGetVertexAttribivNV.gl = gl
	gl.glGetVertexAttribivNV.__name__ = 'glGetVertexAttribivNV'
	gl.glGetVertexAttribivNV.errcheck = _errcheck
	gl.glGetVertexAttribivNV.restype = GLvoid
	gl.glGetVertexAttribivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVideoCaptureStreamdvNV = _getProcAddress('glGetVideoCaptureStreamdvNV')
	gl.glGetVideoCaptureStreamdvNV.gl = gl
	gl.glGetVideoCaptureStreamdvNV.__name__ = 'glGetVideoCaptureStreamdvNV'
	gl.glGetVideoCaptureStreamdvNV.errcheck = _errcheck
	gl.glGetVideoCaptureStreamdvNV.restype = GLvoid
	gl.glGetVideoCaptureStreamdvNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLdouble),)
	gl.glGetVideoCaptureStreamfvNV = _getProcAddress('glGetVideoCaptureStreamfvNV')
	gl.glGetVideoCaptureStreamfvNV.gl = gl
	gl.glGetVideoCaptureStreamfvNV.__name__ = 'glGetVideoCaptureStreamfvNV'
	gl.glGetVideoCaptureStreamfvNV.errcheck = _errcheck
	gl.glGetVideoCaptureStreamfvNV.restype = GLvoid
	gl.glGetVideoCaptureStreamfvNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLfloat),)
	gl.glGetVideoCaptureStreamivNV = _getProcAddress('glGetVideoCaptureStreamivNV')
	gl.glGetVideoCaptureStreamivNV.gl = gl
	gl.glGetVideoCaptureStreamivNV.__name__ = 'glGetVideoCaptureStreamivNV'
	gl.glGetVideoCaptureStreamivNV.errcheck = _errcheck
	gl.glGetVideoCaptureStreamivNV.restype = GLvoid
	gl.glGetVideoCaptureStreamivNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint),)
	gl.glGetVideoCaptureivNV = _getProcAddress('glGetVideoCaptureivNV')
	gl.glGetVideoCaptureivNV.gl = gl
	gl.glGetVideoCaptureivNV.__name__ = 'glGetVideoCaptureivNV'
	gl.glGetVideoCaptureivNV.errcheck = _errcheck
	gl.glGetVideoCaptureivNV.restype = GLvoid
	gl.glGetVideoCaptureivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVideoi64vNV = _getProcAddress('glGetVideoi64vNV')
	gl.glGetVideoi64vNV.gl = gl
	gl.glGetVideoi64vNV.__name__ = 'glGetVideoi64vNV'
	gl.glGetVideoi64vNV.errcheck = _errcheck
	gl.glGetVideoi64vNV.restype = GLvoid
	gl.glGetVideoi64vNV.argtypes = (GLuint, GLenum, POINTER(GLint64EXT),)
	gl.glGetVideoivNV = _getProcAddress('glGetVideoivNV')
	gl.glGetVideoivNV.gl = gl
	gl.glGetVideoivNV.__name__ = 'glGetVideoivNV'
	gl.glGetVideoivNV.errcheck = _errcheck
	gl.glGetVideoivNV.restype = GLvoid
	gl.glGetVideoivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glGetVideoui64vNV = _getProcAddress('glGetVideoui64vNV')
	gl.glGetVideoui64vNV.gl = gl
	gl.glGetVideoui64vNV.__name__ = 'glGetVideoui64vNV'
	gl.glGetVideoui64vNV.errcheck = _errcheck
	gl.glGetVideoui64vNV.restype = GLvoid
	gl.glGetVideoui64vNV.argtypes = (GLuint, GLenum, POINTER(GLuint64EXT),)
	gl.glGetVideouivNV = _getProcAddress('glGetVideouivNV')
	gl.glGetVideouivNV.gl = gl
	gl.glGetVideouivNV.__name__ = 'glGetVideouivNV'
	gl.glGetVideouivNV.errcheck = _errcheck
	gl.glGetVideouivNV.restype = GLvoid
	gl.glGetVideouivNV.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glGetnColorTable = _getProcAddress('glGetnColorTable')
	gl.glGetnColorTable.gl = gl
	gl.glGetnColorTable.__name__ = 'glGetnColorTable'
	gl.glGetnColorTable.errcheck = _errcheck
	gl.glGetnColorTable.restype = GLvoid
	gl.glGetnColorTable.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnColorTableARB = _getProcAddress('glGetnColorTableARB')
	gl.glGetnColorTableARB.gl = gl
	gl.glGetnColorTableARB.__name__ = 'glGetnColorTableARB'
	gl.glGetnColorTableARB.errcheck = _errcheck
	gl.glGetnColorTableARB.restype = GLvoid
	gl.glGetnColorTableARB.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnCompressedTexImage = _getProcAddress('glGetnCompressedTexImage')
	gl.glGetnCompressedTexImage.gl = gl
	gl.glGetnCompressedTexImage.__name__ = 'glGetnCompressedTexImage'
	gl.glGetnCompressedTexImage.errcheck = _errcheck
	gl.glGetnCompressedTexImage.restype = GLvoid
	gl.glGetnCompressedTexImage.argtypes = (GLenum, GLint, GLsizei, POINTER(GLvoid),)
	gl.glGetnCompressedTexImageARB = _getProcAddress('glGetnCompressedTexImageARB')
	gl.glGetnCompressedTexImageARB.gl = gl
	gl.glGetnCompressedTexImageARB.__name__ = 'glGetnCompressedTexImageARB'
	gl.glGetnCompressedTexImageARB.errcheck = _errcheck
	gl.glGetnCompressedTexImageARB.restype = GLvoid
	gl.glGetnCompressedTexImageARB.argtypes = (GLenum, GLint, GLsizei, POINTER(GLvoid),)
	gl.glGetnConvolutionFilter = _getProcAddress('glGetnConvolutionFilter')
	gl.glGetnConvolutionFilter.gl = gl
	gl.glGetnConvolutionFilter.__name__ = 'glGetnConvolutionFilter'
	gl.glGetnConvolutionFilter.errcheck = _errcheck
	gl.glGetnConvolutionFilter.restype = GLvoid
	gl.glGetnConvolutionFilter.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnConvolutionFilterARB = _getProcAddress('glGetnConvolutionFilterARB')
	gl.glGetnConvolutionFilterARB.gl = gl
	gl.glGetnConvolutionFilterARB.__name__ = 'glGetnConvolutionFilterARB'
	gl.glGetnConvolutionFilterARB.errcheck = _errcheck
	gl.glGetnConvolutionFilterARB.restype = GLvoid
	gl.glGetnConvolutionFilterARB.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnHistogram = _getProcAddress('glGetnHistogram')
	gl.glGetnHistogram.gl = gl
	gl.glGetnHistogram.__name__ = 'glGetnHistogram'
	gl.glGetnHistogram.errcheck = _errcheck
	gl.glGetnHistogram.restype = GLvoid
	gl.glGetnHistogram.argtypes = (GLenum, GLboolean, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnHistogramARB = _getProcAddress('glGetnHistogramARB')
	gl.glGetnHistogramARB.gl = gl
	gl.glGetnHistogramARB.__name__ = 'glGetnHistogramARB'
	gl.glGetnHistogramARB.errcheck = _errcheck
	gl.glGetnHistogramARB.restype = GLvoid
	gl.glGetnHistogramARB.argtypes = (GLenum, GLboolean, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnMapdv = _getProcAddress('glGetnMapdv')
	gl.glGetnMapdv.gl = gl
	gl.glGetnMapdv.__name__ = 'glGetnMapdv'
	gl.glGetnMapdv.errcheck = _errcheck
	gl.glGetnMapdv.restype = GLvoid
	gl.glGetnMapdv.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLdouble),)
	gl.glGetnMapdvARB = _getProcAddress('glGetnMapdvARB')
	gl.glGetnMapdvARB.gl = gl
	gl.glGetnMapdvARB.__name__ = 'glGetnMapdvARB'
	gl.glGetnMapdvARB.errcheck = _errcheck
	gl.glGetnMapdvARB.restype = GLvoid
	gl.glGetnMapdvARB.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLdouble),)
	gl.glGetnMapfv = _getProcAddress('glGetnMapfv')
	gl.glGetnMapfv.gl = gl
	gl.glGetnMapfv.__name__ = 'glGetnMapfv'
	gl.glGetnMapfv.errcheck = _errcheck
	gl.glGetnMapfv.restype = GLvoid
	gl.glGetnMapfv.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLfloat),)
	gl.glGetnMapfvARB = _getProcAddress('glGetnMapfvARB')
	gl.glGetnMapfvARB.gl = gl
	gl.glGetnMapfvARB.__name__ = 'glGetnMapfvARB'
	gl.glGetnMapfvARB.errcheck = _errcheck
	gl.glGetnMapfvARB.restype = GLvoid
	gl.glGetnMapfvARB.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLfloat),)
	gl.glGetnMapiv = _getProcAddress('glGetnMapiv')
	gl.glGetnMapiv.gl = gl
	gl.glGetnMapiv.__name__ = 'glGetnMapiv'
	gl.glGetnMapiv.errcheck = _errcheck
	gl.glGetnMapiv.restype = GLvoid
	gl.glGetnMapiv.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLint),)
	gl.glGetnMapivARB = _getProcAddress('glGetnMapivARB')
	gl.glGetnMapivARB.gl = gl
	gl.glGetnMapivARB.__name__ = 'glGetnMapivARB'
	gl.glGetnMapivARB.errcheck = _errcheck
	gl.glGetnMapivARB.restype = GLvoid
	gl.glGetnMapivARB.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLint),)
	gl.glGetnMinmax = _getProcAddress('glGetnMinmax')
	gl.glGetnMinmax.gl = gl
	gl.glGetnMinmax.__name__ = 'glGetnMinmax'
	gl.glGetnMinmax.errcheck = _errcheck
	gl.glGetnMinmax.restype = GLvoid
	gl.glGetnMinmax.argtypes = (GLenum, GLboolean, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnMinmaxARB = _getProcAddress('glGetnMinmaxARB')
	gl.glGetnMinmaxARB.gl = gl
	gl.glGetnMinmaxARB.__name__ = 'glGetnMinmaxARB'
	gl.glGetnMinmaxARB.errcheck = _errcheck
	gl.glGetnMinmaxARB.restype = GLvoid
	gl.glGetnMinmaxARB.argtypes = (GLenum, GLboolean, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnPixelMapfv = _getProcAddress('glGetnPixelMapfv')
	gl.glGetnPixelMapfv.gl = gl
	gl.glGetnPixelMapfv.__name__ = 'glGetnPixelMapfv'
	gl.glGetnPixelMapfv.errcheck = _errcheck
	gl.glGetnPixelMapfv.restype = GLvoid
	gl.glGetnPixelMapfv.argtypes = (GLenum, GLsizei, POINTER(GLfloat),)
	gl.glGetnPixelMapfvARB = _getProcAddress('glGetnPixelMapfvARB')
	gl.glGetnPixelMapfvARB.gl = gl
	gl.glGetnPixelMapfvARB.__name__ = 'glGetnPixelMapfvARB'
	gl.glGetnPixelMapfvARB.errcheck = _errcheck
	gl.glGetnPixelMapfvARB.restype = GLvoid
	gl.glGetnPixelMapfvARB.argtypes = (GLenum, GLsizei, POINTER(GLfloat),)
	gl.glGetnPixelMapuiv = _getProcAddress('glGetnPixelMapuiv')
	gl.glGetnPixelMapuiv.gl = gl
	gl.glGetnPixelMapuiv.__name__ = 'glGetnPixelMapuiv'
	gl.glGetnPixelMapuiv.errcheck = _errcheck
	gl.glGetnPixelMapuiv.restype = GLvoid
	gl.glGetnPixelMapuiv.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glGetnPixelMapuivARB = _getProcAddress('glGetnPixelMapuivARB')
	gl.glGetnPixelMapuivARB.gl = gl
	gl.glGetnPixelMapuivARB.__name__ = 'glGetnPixelMapuivARB'
	gl.glGetnPixelMapuivARB.errcheck = _errcheck
	gl.glGetnPixelMapuivARB.restype = GLvoid
	gl.glGetnPixelMapuivARB.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glGetnPixelMapusv = _getProcAddress('glGetnPixelMapusv')
	gl.glGetnPixelMapusv.gl = gl
	gl.glGetnPixelMapusv.__name__ = 'glGetnPixelMapusv'
	gl.glGetnPixelMapusv.errcheck = _errcheck
	gl.glGetnPixelMapusv.restype = GLvoid
	gl.glGetnPixelMapusv.argtypes = (GLenum, GLsizei, POINTER(GLushort),)
	gl.glGetnPixelMapusvARB = _getProcAddress('glGetnPixelMapusvARB')
	gl.glGetnPixelMapusvARB.gl = gl
	gl.glGetnPixelMapusvARB.__name__ = 'glGetnPixelMapusvARB'
	gl.glGetnPixelMapusvARB.errcheck = _errcheck
	gl.glGetnPixelMapusvARB.restype = GLvoid
	gl.glGetnPixelMapusvARB.argtypes = (GLenum, GLsizei, POINTER(GLushort),)
	gl.glGetnPolygonStipple = _getProcAddress('glGetnPolygonStipple')
	gl.glGetnPolygonStipple.gl = gl
	gl.glGetnPolygonStipple.__name__ = 'glGetnPolygonStipple'
	gl.glGetnPolygonStipple.errcheck = _errcheck
	gl.glGetnPolygonStipple.restype = GLvoid
	gl.glGetnPolygonStipple.argtypes = (GLsizei, POINTER(GLubyte),)
	gl.glGetnPolygonStippleARB = _getProcAddress('glGetnPolygonStippleARB')
	gl.glGetnPolygonStippleARB.gl = gl
	gl.glGetnPolygonStippleARB.__name__ = 'glGetnPolygonStippleARB'
	gl.glGetnPolygonStippleARB.errcheck = _errcheck
	gl.glGetnPolygonStippleARB.restype = GLvoid
	gl.glGetnPolygonStippleARB.argtypes = (GLsizei, POINTER(GLubyte),)
	gl.glGetnSeparableFilter = _getProcAddress('glGetnSeparableFilter')
	gl.glGetnSeparableFilter.gl = gl
	gl.glGetnSeparableFilter.__name__ = 'glGetnSeparableFilter'
	gl.glGetnSeparableFilter.errcheck = _errcheck
	gl.glGetnSeparableFilter.restype = GLvoid
	gl.glGetnSeparableFilter.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLvoid), GLsizei, POINTER(GLvoid), POINTER(GLvoid),)
	gl.glGetnSeparableFilterARB = _getProcAddress('glGetnSeparableFilterARB')
	gl.glGetnSeparableFilterARB.gl = gl
	gl.glGetnSeparableFilterARB.__name__ = 'glGetnSeparableFilterARB'
	gl.glGetnSeparableFilterARB.errcheck = _errcheck
	gl.glGetnSeparableFilterARB.restype = GLvoid
	gl.glGetnSeparableFilterARB.argtypes = (GLenum, GLenum, GLenum, GLsizei, POINTER(GLvoid), GLsizei, POINTER(GLvoid), POINTER(GLvoid),)
	gl.glGetnTexImage = _getProcAddress('glGetnTexImage')
	gl.glGetnTexImage.gl = gl
	gl.glGetnTexImage.__name__ = 'glGetnTexImage'
	gl.glGetnTexImage.errcheck = _errcheck
	gl.glGetnTexImage.restype = GLvoid
	gl.glGetnTexImage.argtypes = (GLenum, GLint, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnTexImageARB = _getProcAddress('glGetnTexImageARB')
	gl.glGetnTexImageARB.gl = gl
	gl.glGetnTexImageARB.__name__ = 'glGetnTexImageARB'
	gl.glGetnTexImageARB.errcheck = _errcheck
	gl.glGetnTexImageARB.restype = GLvoid
	gl.glGetnTexImageARB.argtypes = (GLenum, GLint, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glGetnUniformdv = _getProcAddress('glGetnUniformdv')
	gl.glGetnUniformdv.gl = gl
	gl.glGetnUniformdv.__name__ = 'glGetnUniformdv'
	gl.glGetnUniformdv.errcheck = _errcheck
	gl.glGetnUniformdv.restype = GLvoid
	gl.glGetnUniformdv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glGetnUniformdvARB = _getProcAddress('glGetnUniformdvARB')
	gl.glGetnUniformdvARB.gl = gl
	gl.glGetnUniformdvARB.__name__ = 'glGetnUniformdvARB'
	gl.glGetnUniformdvARB.errcheck = _errcheck
	gl.glGetnUniformdvARB.restype = GLvoid
	gl.glGetnUniformdvARB.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glGetnUniformfv = _getProcAddress('glGetnUniformfv')
	gl.glGetnUniformfv.gl = gl
	gl.glGetnUniformfv.__name__ = 'glGetnUniformfv'
	gl.glGetnUniformfv.errcheck = _errcheck
	gl.glGetnUniformfv.restype = GLvoid
	gl.glGetnUniformfv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glGetnUniformfvARB = _getProcAddress('glGetnUniformfvARB')
	gl.glGetnUniformfvARB.gl = gl
	gl.glGetnUniformfvARB.__name__ = 'glGetnUniformfvARB'
	gl.glGetnUniformfvARB.errcheck = _errcheck
	gl.glGetnUniformfvARB.restype = GLvoid
	gl.glGetnUniformfvARB.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glGetnUniformfvEXT = _getProcAddress('glGetnUniformfvEXT')
	gl.glGetnUniformfvEXT.gl = gl
	gl.glGetnUniformfvEXT.__name__ = 'glGetnUniformfvEXT'
	gl.glGetnUniformfvEXT.errcheck = _errcheck
	gl.glGetnUniformfvEXT.restype = GLvoid
	gl.glGetnUniformfvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glGetnUniformfvKHR = _getProcAddress('glGetnUniformfvKHR')
	gl.glGetnUniformfvKHR.gl = gl
	gl.glGetnUniformfvKHR.__name__ = 'glGetnUniformfvKHR'
	gl.glGetnUniformfvKHR.errcheck = _errcheck
	gl.glGetnUniformfvKHR.restype = GLvoid
	gl.glGetnUniformfvKHR.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glGetnUniformiv = _getProcAddress('glGetnUniformiv')
	gl.glGetnUniformiv.gl = gl
	gl.glGetnUniformiv.__name__ = 'glGetnUniformiv'
	gl.glGetnUniformiv.errcheck = _errcheck
	gl.glGetnUniformiv.restype = GLvoid
	gl.glGetnUniformiv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glGetnUniformivARB = _getProcAddress('glGetnUniformivARB')
	gl.glGetnUniformivARB.gl = gl
	gl.glGetnUniformivARB.__name__ = 'glGetnUniformivARB'
	gl.glGetnUniformivARB.errcheck = _errcheck
	gl.glGetnUniformivARB.restype = GLvoid
	gl.glGetnUniformivARB.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glGetnUniformivEXT = _getProcAddress('glGetnUniformivEXT')
	gl.glGetnUniformivEXT.gl = gl
	gl.glGetnUniformivEXT.__name__ = 'glGetnUniformivEXT'
	gl.glGetnUniformivEXT.errcheck = _errcheck
	gl.glGetnUniformivEXT.restype = GLvoid
	gl.glGetnUniformivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glGetnUniformivKHR = _getProcAddress('glGetnUniformivKHR')
	gl.glGetnUniformivKHR.gl = gl
	gl.glGetnUniformivKHR.__name__ = 'glGetnUniformivKHR'
	gl.glGetnUniformivKHR.errcheck = _errcheck
	gl.glGetnUniformivKHR.restype = GLvoid
	gl.glGetnUniformivKHR.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glGetnUniformuiv = _getProcAddress('glGetnUniformuiv')
	gl.glGetnUniformuiv.gl = gl
	gl.glGetnUniformuiv.__name__ = 'glGetnUniformuiv'
	gl.glGetnUniformuiv.errcheck = _errcheck
	gl.glGetnUniformuiv.restype = GLvoid
	gl.glGetnUniformuiv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glGetnUniformuivARB = _getProcAddress('glGetnUniformuivARB')
	gl.glGetnUniformuivARB.gl = gl
	gl.glGetnUniformuivARB.__name__ = 'glGetnUniformuivARB'
	gl.glGetnUniformuivARB.errcheck = _errcheck
	gl.glGetnUniformuivARB.restype = GLvoid
	gl.glGetnUniformuivARB.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glGetnUniformuivKHR = _getProcAddress('glGetnUniformuivKHR')
	gl.glGetnUniformuivKHR.gl = gl
	gl.glGetnUniformuivKHR.__name__ = 'glGetnUniformuivKHR'
	gl.glGetnUniformuivKHR.errcheck = _errcheck
	gl.glGetnUniformuivKHR.restype = GLvoid
	gl.glGetnUniformuivKHR.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glGlobalAlphaFactorbSUN = _getProcAddress('glGlobalAlphaFactorbSUN')
	gl.glGlobalAlphaFactorbSUN.gl = gl
	gl.glGlobalAlphaFactorbSUN.__name__ = 'glGlobalAlphaFactorbSUN'
	gl.glGlobalAlphaFactorbSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactorbSUN.restype = GLvoid
	gl.glGlobalAlphaFactorbSUN.argtypes = (GLbyte,)
	gl.glGlobalAlphaFactordSUN = _getProcAddress('glGlobalAlphaFactordSUN')
	gl.glGlobalAlphaFactordSUN.gl = gl
	gl.glGlobalAlphaFactordSUN.__name__ = 'glGlobalAlphaFactordSUN'
	gl.glGlobalAlphaFactordSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactordSUN.restype = GLvoid
	gl.glGlobalAlphaFactordSUN.argtypes = (GLdouble,)
	gl.glGlobalAlphaFactorfSUN = _getProcAddress('glGlobalAlphaFactorfSUN')
	gl.glGlobalAlphaFactorfSUN.gl = gl
	gl.glGlobalAlphaFactorfSUN.__name__ = 'glGlobalAlphaFactorfSUN'
	gl.glGlobalAlphaFactorfSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactorfSUN.restype = GLvoid
	gl.glGlobalAlphaFactorfSUN.argtypes = (GLfloat,)
	gl.glGlobalAlphaFactoriSUN = _getProcAddress('glGlobalAlphaFactoriSUN')
	gl.glGlobalAlphaFactoriSUN.gl = gl
	gl.glGlobalAlphaFactoriSUN.__name__ = 'glGlobalAlphaFactoriSUN'
	gl.glGlobalAlphaFactoriSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactoriSUN.restype = GLvoid
	gl.glGlobalAlphaFactoriSUN.argtypes = (GLint,)
	gl.glGlobalAlphaFactorsSUN = _getProcAddress('glGlobalAlphaFactorsSUN')
	gl.glGlobalAlphaFactorsSUN.gl = gl
	gl.glGlobalAlphaFactorsSUN.__name__ = 'glGlobalAlphaFactorsSUN'
	gl.glGlobalAlphaFactorsSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactorsSUN.restype = GLvoid
	gl.glGlobalAlphaFactorsSUN.argtypes = (GLshort,)
	gl.glGlobalAlphaFactorubSUN = _getProcAddress('glGlobalAlphaFactorubSUN')
	gl.glGlobalAlphaFactorubSUN.gl = gl
	gl.glGlobalAlphaFactorubSUN.__name__ = 'glGlobalAlphaFactorubSUN'
	gl.glGlobalAlphaFactorubSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactorubSUN.restype = GLvoid
	gl.glGlobalAlphaFactorubSUN.argtypes = (GLubyte,)
	gl.glGlobalAlphaFactoruiSUN = _getProcAddress('glGlobalAlphaFactoruiSUN')
	gl.glGlobalAlphaFactoruiSUN.gl = gl
	gl.glGlobalAlphaFactoruiSUN.__name__ = 'glGlobalAlphaFactoruiSUN'
	gl.glGlobalAlphaFactoruiSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactoruiSUN.restype = GLvoid
	gl.glGlobalAlphaFactoruiSUN.argtypes = (GLuint,)
	gl.glGlobalAlphaFactorusSUN = _getProcAddress('glGlobalAlphaFactorusSUN')
	gl.glGlobalAlphaFactorusSUN.gl = gl
	gl.glGlobalAlphaFactorusSUN.__name__ = 'glGlobalAlphaFactorusSUN'
	gl.glGlobalAlphaFactorusSUN.errcheck = _errcheck
	gl.glGlobalAlphaFactorusSUN.restype = GLvoid
	gl.glGlobalAlphaFactorusSUN.argtypes = (GLushort,)
	gl.glHint = _getProcAddress('glHint')
	gl.glHint.gl = gl
	gl.glHint.__name__ = 'glHint'
	gl.glHint.errcheck = _errcheck
	gl.glHint.restype = GLvoid
	gl.glHint.argtypes = (GLenum, GLenum,)
	gl.glHintPGI = _getProcAddress('glHintPGI')
	gl.glHintPGI.gl = gl
	gl.glHintPGI.__name__ = 'glHintPGI'
	gl.glHintPGI.errcheck = _errcheck
	gl.glHintPGI.restype = GLvoid
	gl.glHintPGI.argtypes = (GLenum, GLint,)
	gl.glHistogram = _getProcAddress('glHistogram')
	gl.glHistogram.gl = gl
	gl.glHistogram.__name__ = 'glHistogram'
	gl.glHistogram.errcheck = _errcheck
	gl.glHistogram.restype = GLvoid
	gl.glHistogram.argtypes = (GLenum, GLsizei, GLenum, GLboolean,)
	gl.glHistogramEXT = _getProcAddress('glHistogramEXT')
	gl.glHistogramEXT.gl = gl
	gl.glHistogramEXT.__name__ = 'glHistogramEXT'
	gl.glHistogramEXT.errcheck = _errcheck
	gl.glHistogramEXT.restype = GLvoid
	gl.glHistogramEXT.argtypes = (GLenum, GLsizei, GLenum, GLboolean,)
	gl.glIglooInterfaceSGIX = _getProcAddress('glIglooInterfaceSGIX')
	gl.glIglooInterfaceSGIX.gl = gl
	gl.glIglooInterfaceSGIX.__name__ = 'glIglooInterfaceSGIX'
	gl.glIglooInterfaceSGIX.errcheck = _errcheck
	gl.glIglooInterfaceSGIX.restype = GLvoid
	gl.glIglooInterfaceSGIX.argtypes = (GLenum, POINTER(GLvoid),)
	gl.glImageTransformParameterfHP = _getProcAddress('glImageTransformParameterfHP')
	gl.glImageTransformParameterfHP.gl = gl
	gl.glImageTransformParameterfHP.__name__ = 'glImageTransformParameterfHP'
	gl.glImageTransformParameterfHP.errcheck = _errcheck
	gl.glImageTransformParameterfHP.restype = GLvoid
	gl.glImageTransformParameterfHP.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glImageTransformParameterfvHP = _getProcAddress('glImageTransformParameterfvHP')
	gl.glImageTransformParameterfvHP.gl = gl
	gl.glImageTransformParameterfvHP.__name__ = 'glImageTransformParameterfvHP'
	gl.glImageTransformParameterfvHP.errcheck = _errcheck
	gl.glImageTransformParameterfvHP.restype = GLvoid
	gl.glImageTransformParameterfvHP.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glImageTransformParameteriHP = _getProcAddress('glImageTransformParameteriHP')
	gl.glImageTransformParameteriHP.gl = gl
	gl.glImageTransformParameteriHP.__name__ = 'glImageTransformParameteriHP'
	gl.glImageTransformParameteriHP.errcheck = _errcheck
	gl.glImageTransformParameteriHP.restype = GLvoid
	gl.glImageTransformParameteriHP.argtypes = (GLenum, GLenum, GLint,)
	gl.glImageTransformParameterivHP = _getProcAddress('glImageTransformParameterivHP')
	gl.glImageTransformParameterivHP.gl = gl
	gl.glImageTransformParameterivHP.__name__ = 'glImageTransformParameterivHP'
	gl.glImageTransformParameterivHP.errcheck = _errcheck
	gl.glImageTransformParameterivHP.restype = GLvoid
	gl.glImageTransformParameterivHP.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glImportSyncEXT = _getProcAddress('glImportSyncEXT')
	gl.glImportSyncEXT.gl = gl
	gl.glImportSyncEXT.__name__ = 'glImportSyncEXT'
	gl.glImportSyncEXT.errcheck = _errcheck
	gl.glImportSyncEXT.restype = GLsync
	gl.glImportSyncEXT.argtypes = (GLenum, GLintptr, GLbitfield,)
	gl.glIndexFormatNV = _getProcAddress('glIndexFormatNV')
	gl.glIndexFormatNV.gl = gl
	gl.glIndexFormatNV.__name__ = 'glIndexFormatNV'
	gl.glIndexFormatNV.errcheck = _errcheck
	gl.glIndexFormatNV.restype = GLvoid
	gl.glIndexFormatNV.argtypes = (GLenum, GLsizei,)
	gl.glIndexFuncEXT = _getProcAddress('glIndexFuncEXT')
	gl.glIndexFuncEXT.gl = gl
	gl.glIndexFuncEXT.__name__ = 'glIndexFuncEXT'
	gl.glIndexFuncEXT.errcheck = _errcheck
	gl.glIndexFuncEXT.restype = GLvoid
	gl.glIndexFuncEXT.argtypes = (GLenum, GLclampf,)
	gl.glIndexMask = _getProcAddress('glIndexMask')
	gl.glIndexMask.gl = gl
	gl.glIndexMask.__name__ = 'glIndexMask'
	gl.glIndexMask.errcheck = _errcheck
	gl.glIndexMask.restype = GLvoid
	gl.glIndexMask.argtypes = (GLuint,)
	gl.glIndexMaterialEXT = _getProcAddress('glIndexMaterialEXT')
	gl.glIndexMaterialEXT.gl = gl
	gl.glIndexMaterialEXT.__name__ = 'glIndexMaterialEXT'
	gl.glIndexMaterialEXT.errcheck = _errcheck
	gl.glIndexMaterialEXT.restype = GLvoid
	gl.glIndexMaterialEXT.argtypes = (GLenum, GLenum,)
	gl.glIndexPointer = _getProcAddress('glIndexPointer')
	gl.glIndexPointer.gl = gl
	gl.glIndexPointer.__name__ = 'glIndexPointer'
	gl.glIndexPointer.errcheck = _errcheck
	gl.glIndexPointer.restype = GLvoid
	gl.glIndexPointer.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glIndexPointerEXT = _getProcAddress('glIndexPointerEXT')
	gl.glIndexPointerEXT.gl = gl
	gl.glIndexPointerEXT.__name__ = 'glIndexPointerEXT'
	gl.glIndexPointerEXT.errcheck = _errcheck
	gl.glIndexPointerEXT.restype = GLvoid
	gl.glIndexPointerEXT.argtypes = (GLenum, GLsizei, GLsizei, POINTER(GLvoid),)
	gl.glIndexPointerListIBM = _getProcAddress('glIndexPointerListIBM')
	gl.glIndexPointerListIBM.gl = gl
	gl.glIndexPointerListIBM.__name__ = 'glIndexPointerListIBM'
	gl.glIndexPointerListIBM.errcheck = _errcheck
	gl.glIndexPointerListIBM.restype = GLvoid
	gl.glIndexPointerListIBM.argtypes = (GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glIndexd = _getProcAddress('glIndexd')
	gl.glIndexd.gl = gl
	gl.glIndexd.__name__ = 'glIndexd'
	gl.glIndexd.errcheck = _errcheck
	gl.glIndexd.restype = GLvoid
	gl.glIndexd.argtypes = (GLdouble,)
	gl.glIndexdv = _getProcAddress('glIndexdv')
	gl.glIndexdv.gl = gl
	gl.glIndexdv.__name__ = 'glIndexdv'
	gl.glIndexdv.errcheck = _errcheck
	gl.glIndexdv.restype = GLvoid
	gl.glIndexdv.argtypes = (POINTER(GLdouble),)
	gl.glIndexf = _getProcAddress('glIndexf')
	gl.glIndexf.gl = gl
	gl.glIndexf.__name__ = 'glIndexf'
	gl.glIndexf.errcheck = _errcheck
	gl.glIndexf.restype = GLvoid
	gl.glIndexf.argtypes = (GLfloat,)
	gl.glIndexfv = _getProcAddress('glIndexfv')
	gl.glIndexfv.gl = gl
	gl.glIndexfv.__name__ = 'glIndexfv'
	gl.glIndexfv.errcheck = _errcheck
	gl.glIndexfv.restype = GLvoid
	gl.glIndexfv.argtypes = (POINTER(GLfloat),)
	gl.glIndexi = _getProcAddress('glIndexi')
	gl.glIndexi.gl = gl
	gl.glIndexi.__name__ = 'glIndexi'
	gl.glIndexi.errcheck = _errcheck
	gl.glIndexi.restype = GLvoid
	gl.glIndexi.argtypes = (GLint,)
	gl.glIndexiv = _getProcAddress('glIndexiv')
	gl.glIndexiv.gl = gl
	gl.glIndexiv.__name__ = 'glIndexiv'
	gl.glIndexiv.errcheck = _errcheck
	gl.glIndexiv.restype = GLvoid
	gl.glIndexiv.argtypes = (POINTER(GLint),)
	gl.glIndexs = _getProcAddress('glIndexs')
	gl.glIndexs.gl = gl
	gl.glIndexs.__name__ = 'glIndexs'
	gl.glIndexs.errcheck = _errcheck
	gl.glIndexs.restype = GLvoid
	gl.glIndexs.argtypes = (GLshort,)
	gl.glIndexsv = _getProcAddress('glIndexsv')
	gl.glIndexsv.gl = gl
	gl.glIndexsv.__name__ = 'glIndexsv'
	gl.glIndexsv.errcheck = _errcheck
	gl.glIndexsv.restype = GLvoid
	gl.glIndexsv.argtypes = (POINTER(GLshort),)
	gl.glIndexub = _getProcAddress('glIndexub')
	gl.glIndexub.gl = gl
	gl.glIndexub.__name__ = 'glIndexub'
	gl.glIndexub.errcheck = _errcheck
	gl.glIndexub.restype = GLvoid
	gl.glIndexub.argtypes = (GLubyte,)
	gl.glIndexubv = _getProcAddress('glIndexubv')
	gl.glIndexubv.gl = gl
	gl.glIndexubv.__name__ = 'glIndexubv'
	gl.glIndexubv.errcheck = _errcheck
	gl.glIndexubv.restype = GLvoid
	gl.glIndexubv.argtypes = (POINTER(GLubyte),)
	gl.glIndexxOES = _getProcAddress('glIndexxOES')
	gl.glIndexxOES.gl = gl
	gl.glIndexxOES.__name__ = 'glIndexxOES'
	gl.glIndexxOES.errcheck = _errcheck
	gl.glIndexxOES.restype = GLvoid
	gl.glIndexxOES.argtypes = (GLfixed,)
	gl.glIndexxvOES = _getProcAddress('glIndexxvOES')
	gl.glIndexxvOES.gl = gl
	gl.glIndexxvOES.__name__ = 'glIndexxvOES'
	gl.glIndexxvOES.errcheck = _errcheck
	gl.glIndexxvOES.restype = GLvoid
	gl.glIndexxvOES.argtypes = (POINTER(GLfixed),)
	gl.glInitNames = _getProcAddress('glInitNames')
	gl.glInitNames.gl = gl
	gl.glInitNames.__name__ = 'glInitNames'
	gl.glInitNames.errcheck = _errcheck
	gl.glInitNames.restype = GLvoid
	gl.glInitNames.argtypes = ()
	gl.glInsertComponentEXT = _getProcAddress('glInsertComponentEXT')
	gl.glInsertComponentEXT.gl = gl
	gl.glInsertComponentEXT.__name__ = 'glInsertComponentEXT'
	gl.glInsertComponentEXT.errcheck = _errcheck
	gl.glInsertComponentEXT.restype = GLvoid
	gl.glInsertComponentEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glInsertEventMarkerEXT = _getProcAddress('glInsertEventMarkerEXT')
	gl.glInsertEventMarkerEXT.gl = gl
	gl.glInsertEventMarkerEXT.__name__ = 'glInsertEventMarkerEXT'
	gl.glInsertEventMarkerEXT.errcheck = _errcheck
	gl.glInsertEventMarkerEXT.restype = GLvoid
	gl.glInsertEventMarkerEXT.argtypes = (GLsizei, POINTER(GLchar),)
	gl.glInstrumentsBufferSGIX = _getProcAddress('glInstrumentsBufferSGIX')
	gl.glInstrumentsBufferSGIX.gl = gl
	gl.glInstrumentsBufferSGIX.__name__ = 'glInstrumentsBufferSGIX'
	gl.glInstrumentsBufferSGIX.errcheck = _errcheck
	gl.glInstrumentsBufferSGIX.restype = GLvoid
	gl.glInstrumentsBufferSGIX.argtypes = (GLsizei, POINTER(GLint),)
	gl.glInterleavedArrays = _getProcAddress('glInterleavedArrays')
	gl.glInterleavedArrays.gl = gl
	gl.glInterleavedArrays.__name__ = 'glInterleavedArrays'
	gl.glInterleavedArrays.errcheck = _errcheck
	gl.glInterleavedArrays.restype = GLvoid
	gl.glInterleavedArrays.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glInterpolatePathsNV = _getProcAddress('glInterpolatePathsNV')
	gl.glInterpolatePathsNV.gl = gl
	gl.glInterpolatePathsNV.__name__ = 'glInterpolatePathsNV'
	gl.glInterpolatePathsNV.errcheck = _errcheck
	gl.glInterpolatePathsNV.restype = GLvoid
	gl.glInterpolatePathsNV.argtypes = (GLuint, GLuint, GLuint, GLfloat,)
	gl.glInvalidateBufferData = _getProcAddress('glInvalidateBufferData')
	gl.glInvalidateBufferData.gl = gl
	gl.glInvalidateBufferData.__name__ = 'glInvalidateBufferData'
	gl.glInvalidateBufferData.errcheck = _errcheck
	gl.glInvalidateBufferData.restype = GLvoid
	gl.glInvalidateBufferData.argtypes = (GLuint,)
	gl.glInvalidateBufferSubData = _getProcAddress('glInvalidateBufferSubData')
	gl.glInvalidateBufferSubData.gl = gl
	gl.glInvalidateBufferSubData.__name__ = 'glInvalidateBufferSubData'
	gl.glInvalidateBufferSubData.errcheck = _errcheck
	gl.glInvalidateBufferSubData.restype = GLvoid
	gl.glInvalidateBufferSubData.argtypes = (GLuint, GLintptr, GLsizeiptr,)
	gl.glInvalidateFramebuffer = _getProcAddress('glInvalidateFramebuffer')
	gl.glInvalidateFramebuffer.gl = gl
	gl.glInvalidateFramebuffer.__name__ = 'glInvalidateFramebuffer'
	gl.glInvalidateFramebuffer.errcheck = _errcheck
	gl.glInvalidateFramebuffer.restype = GLvoid
	gl.glInvalidateFramebuffer.argtypes = (GLenum, GLsizei, POINTER(GLenum),)
	gl.glInvalidateNamedFramebufferData = _getProcAddress('glInvalidateNamedFramebufferData')
	gl.glInvalidateNamedFramebufferData.gl = gl
	gl.glInvalidateNamedFramebufferData.__name__ = 'glInvalidateNamedFramebufferData'
	gl.glInvalidateNamedFramebufferData.errcheck = _errcheck
	gl.glInvalidateNamedFramebufferData.restype = GLvoid
	gl.glInvalidateNamedFramebufferData.argtypes = (GLuint, GLsizei, POINTER(GLenum),)
	gl.glInvalidateNamedFramebufferSubData = _getProcAddress('glInvalidateNamedFramebufferSubData')
	gl.glInvalidateNamedFramebufferSubData.gl = gl
	gl.glInvalidateNamedFramebufferSubData.__name__ = 'glInvalidateNamedFramebufferSubData'
	gl.glInvalidateNamedFramebufferSubData.errcheck = _errcheck
	gl.glInvalidateNamedFramebufferSubData.restype = GLvoid
	gl.glInvalidateNamedFramebufferSubData.argtypes = (GLuint, GLsizei, POINTER(GLenum), GLint, GLint, GLsizei, GLsizei,)
	gl.glInvalidateSubFramebuffer = _getProcAddress('glInvalidateSubFramebuffer')
	gl.glInvalidateSubFramebuffer.gl = gl
	gl.glInvalidateSubFramebuffer.__name__ = 'glInvalidateSubFramebuffer'
	gl.glInvalidateSubFramebuffer.errcheck = _errcheck
	gl.glInvalidateSubFramebuffer.restype = GLvoid
	gl.glInvalidateSubFramebuffer.argtypes = (GLenum, GLsizei, POINTER(GLenum), GLint, GLint, GLsizei, GLsizei,)
	gl.glInvalidateTexImage = _getProcAddress('glInvalidateTexImage')
	gl.glInvalidateTexImage.gl = gl
	gl.glInvalidateTexImage.__name__ = 'glInvalidateTexImage'
	gl.glInvalidateTexImage.errcheck = _errcheck
	gl.glInvalidateTexImage.restype = GLvoid
	gl.glInvalidateTexImage.argtypes = (GLuint, GLint,)
	gl.glInvalidateTexSubImage = _getProcAddress('glInvalidateTexSubImage')
	gl.glInvalidateTexSubImage.gl = gl
	gl.glInvalidateTexSubImage.__name__ = 'glInvalidateTexSubImage'
	gl.glInvalidateTexSubImage.errcheck = _errcheck
	gl.glInvalidateTexSubImage.restype = GLvoid
	gl.glInvalidateTexSubImage.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei,)
	gl.glIsAsyncMarkerSGIX = _getProcAddress('glIsAsyncMarkerSGIX')
	gl.glIsAsyncMarkerSGIX.gl = gl
	gl.glIsAsyncMarkerSGIX.__name__ = 'glIsAsyncMarkerSGIX'
	gl.glIsAsyncMarkerSGIX.errcheck = _errcheck
	gl.glIsAsyncMarkerSGIX.restype = GLboolean
	gl.glIsAsyncMarkerSGIX.argtypes = (GLuint,)
	gl.glIsBuffer = _getProcAddress('glIsBuffer')
	gl.glIsBuffer.gl = gl
	gl.glIsBuffer.__name__ = 'glIsBuffer'
	gl.glIsBuffer.errcheck = _errcheck
	gl.glIsBuffer.restype = GLboolean
	gl.glIsBuffer.argtypes = (GLuint,)
	gl.glIsBufferARB = _getProcAddress('glIsBufferARB')
	gl.glIsBufferARB.gl = gl
	gl.glIsBufferARB.__name__ = 'glIsBufferARB'
	gl.glIsBufferARB.errcheck = _errcheck
	gl.glIsBufferARB.restype = GLboolean
	gl.glIsBufferARB.argtypes = (GLuint,)
	gl.glIsBufferResidentNV = _getProcAddress('glIsBufferResidentNV')
	gl.glIsBufferResidentNV.gl = gl
	gl.glIsBufferResidentNV.__name__ = 'glIsBufferResidentNV'
	gl.glIsBufferResidentNV.errcheck = _errcheck
	gl.glIsBufferResidentNV.restype = GLboolean
	gl.glIsBufferResidentNV.argtypes = (GLenum,)
	gl.glIsEnabled = _getProcAddress('glIsEnabled')
	gl.glIsEnabled.gl = gl
	gl.glIsEnabled.__name__ = 'glIsEnabled'
	gl.glIsEnabled.errcheck = _errcheck
	gl.glIsEnabled.restype = GLboolean
	gl.glIsEnabled.argtypes = (GLenum,)
	gl.glIsEnabledIndexedEXT = _getProcAddress('glIsEnabledIndexedEXT')
	gl.glIsEnabledIndexedEXT.gl = gl
	gl.glIsEnabledIndexedEXT.__name__ = 'glIsEnabledIndexedEXT'
	gl.glIsEnabledIndexedEXT.errcheck = _errcheck
	gl.glIsEnabledIndexedEXT.restype = GLboolean
	gl.glIsEnabledIndexedEXT.argtypes = (GLenum, GLuint,)
	gl.glIsEnabledi = _getProcAddress('glIsEnabledi')
	gl.glIsEnabledi.gl = gl
	gl.glIsEnabledi.__name__ = 'glIsEnabledi'
	gl.glIsEnabledi.errcheck = _errcheck
	gl.glIsEnabledi.restype = GLboolean
	gl.glIsEnabledi.argtypes = (GLenum, GLuint,)
	gl.glIsEnablediEXT = _getProcAddress('glIsEnablediEXT')
	gl.glIsEnablediEXT.gl = gl
	gl.glIsEnablediEXT.__name__ = 'glIsEnablediEXT'
	gl.glIsEnablediEXT.errcheck = _errcheck
	gl.glIsEnablediEXT.restype = GLboolean
	gl.glIsEnablediEXT.argtypes = (GLenum, GLuint,)
	gl.glIsFenceAPPLE = _getProcAddress('glIsFenceAPPLE')
	gl.glIsFenceAPPLE.gl = gl
	gl.glIsFenceAPPLE.__name__ = 'glIsFenceAPPLE'
	gl.glIsFenceAPPLE.errcheck = _errcheck
	gl.glIsFenceAPPLE.restype = GLboolean
	gl.glIsFenceAPPLE.argtypes = (GLuint,)
	gl.glIsFenceNV = _getProcAddress('glIsFenceNV')
	gl.glIsFenceNV.gl = gl
	gl.glIsFenceNV.__name__ = 'glIsFenceNV'
	gl.glIsFenceNV.errcheck = _errcheck
	gl.glIsFenceNV.restype = GLboolean
	gl.glIsFenceNV.argtypes = (GLuint,)
	gl.glIsFramebuffer = _getProcAddress('glIsFramebuffer')
	gl.glIsFramebuffer.gl = gl
	gl.glIsFramebuffer.__name__ = 'glIsFramebuffer'
	gl.glIsFramebuffer.errcheck = _errcheck
	gl.glIsFramebuffer.restype = GLboolean
	gl.glIsFramebuffer.argtypes = (GLuint,)
	gl.glIsFramebufferEXT = _getProcAddress('glIsFramebufferEXT')
	gl.glIsFramebufferEXT.gl = gl
	gl.glIsFramebufferEXT.__name__ = 'glIsFramebufferEXT'
	gl.glIsFramebufferEXT.errcheck = _errcheck
	gl.glIsFramebufferEXT.restype = GLboolean
	gl.glIsFramebufferEXT.argtypes = (GLuint,)
	gl.glIsFramebufferOES = _getProcAddress('glIsFramebufferOES')
	gl.glIsFramebufferOES.gl = gl
	gl.glIsFramebufferOES.__name__ = 'glIsFramebufferOES'
	gl.glIsFramebufferOES.errcheck = _errcheck
	gl.glIsFramebufferOES.restype = GLboolean
	gl.glIsFramebufferOES.argtypes = (GLuint,)
	gl.glIsImageHandleResidentARB = _getProcAddress('glIsImageHandleResidentARB')
	gl.glIsImageHandleResidentARB.gl = gl
	gl.glIsImageHandleResidentARB.__name__ = 'glIsImageHandleResidentARB'
	gl.glIsImageHandleResidentARB.errcheck = _errcheck
	gl.glIsImageHandleResidentARB.restype = GLboolean
	gl.glIsImageHandleResidentARB.argtypes = (GLuint64,)
	gl.glIsImageHandleResidentNV = _getProcAddress('glIsImageHandleResidentNV')
	gl.glIsImageHandleResidentNV.gl = gl
	gl.glIsImageHandleResidentNV.__name__ = 'glIsImageHandleResidentNV'
	gl.glIsImageHandleResidentNV.errcheck = _errcheck
	gl.glIsImageHandleResidentNV.restype = GLboolean
	gl.glIsImageHandleResidentNV.argtypes = (GLuint64,)
	gl.glIsList = _getProcAddress('glIsList')
	gl.glIsList.gl = gl
	gl.glIsList.__name__ = 'glIsList'
	gl.glIsList.errcheck = _errcheck
	gl.glIsList.restype = GLboolean
	gl.glIsList.argtypes = (GLuint,)
	gl.glIsNameAMD = _getProcAddress('glIsNameAMD')
	gl.glIsNameAMD.gl = gl
	gl.glIsNameAMD.__name__ = 'glIsNameAMD'
	gl.glIsNameAMD.errcheck = _errcheck
	gl.glIsNameAMD.restype = GLboolean
	gl.glIsNameAMD.argtypes = (GLenum, GLuint,)
	gl.glIsNamedBufferResidentNV = _getProcAddress('glIsNamedBufferResidentNV')
	gl.glIsNamedBufferResidentNV.gl = gl
	gl.glIsNamedBufferResidentNV.__name__ = 'glIsNamedBufferResidentNV'
	gl.glIsNamedBufferResidentNV.errcheck = _errcheck
	gl.glIsNamedBufferResidentNV.restype = GLboolean
	gl.glIsNamedBufferResidentNV.argtypes = (GLuint,)
	gl.glIsNamedStringARB = _getProcAddress('glIsNamedStringARB')
	gl.glIsNamedStringARB.gl = gl
	gl.glIsNamedStringARB.__name__ = 'glIsNamedStringARB'
	gl.glIsNamedStringARB.errcheck = _errcheck
	gl.glIsNamedStringARB.restype = GLboolean
	gl.glIsNamedStringARB.argtypes = (GLint, POINTER(GLchar),)
	gl.glIsObjectBufferATI = _getProcAddress('glIsObjectBufferATI')
	gl.glIsObjectBufferATI.gl = gl
	gl.glIsObjectBufferATI.__name__ = 'glIsObjectBufferATI'
	gl.glIsObjectBufferATI.errcheck = _errcheck
	gl.glIsObjectBufferATI.restype = GLboolean
	gl.glIsObjectBufferATI.argtypes = (GLuint,)
	gl.glIsOcclusionQueryNV = _getProcAddress('glIsOcclusionQueryNV')
	gl.glIsOcclusionQueryNV.gl = gl
	gl.glIsOcclusionQueryNV.__name__ = 'glIsOcclusionQueryNV'
	gl.glIsOcclusionQueryNV.errcheck = _errcheck
	gl.glIsOcclusionQueryNV.restype = GLboolean
	gl.glIsOcclusionQueryNV.argtypes = (GLuint,)
	gl.glIsPathNV = _getProcAddress('glIsPathNV')
	gl.glIsPathNV.gl = gl
	gl.glIsPathNV.__name__ = 'glIsPathNV'
	gl.glIsPathNV.errcheck = _errcheck
	gl.glIsPathNV.restype = GLboolean
	gl.glIsPathNV.argtypes = (GLuint,)
	gl.glIsPointInFillPathNV = _getProcAddress('glIsPointInFillPathNV')
	gl.glIsPointInFillPathNV.gl = gl
	gl.glIsPointInFillPathNV.__name__ = 'glIsPointInFillPathNV'
	gl.glIsPointInFillPathNV.errcheck = _errcheck
	gl.glIsPointInFillPathNV.restype = GLboolean
	gl.glIsPointInFillPathNV.argtypes = (GLuint, GLuint, GLfloat, GLfloat,)
	gl.glIsPointInStrokePathNV = _getProcAddress('glIsPointInStrokePathNV')
	gl.glIsPointInStrokePathNV.gl = gl
	gl.glIsPointInStrokePathNV.__name__ = 'glIsPointInStrokePathNV'
	gl.glIsPointInStrokePathNV.errcheck = _errcheck
	gl.glIsPointInStrokePathNV.restype = GLboolean
	gl.glIsPointInStrokePathNV.argtypes = (GLuint, GLfloat, GLfloat,)
	gl.glIsProgram = _getProcAddress('glIsProgram')
	gl.glIsProgram.gl = gl
	gl.glIsProgram.__name__ = 'glIsProgram'
	gl.glIsProgram.errcheck = _errcheck
	gl.glIsProgram.restype = GLboolean
	gl.glIsProgram.argtypes = (GLuint,)
	gl.glIsProgramARB = _getProcAddress('glIsProgramARB')
	gl.glIsProgramARB.gl = gl
	gl.glIsProgramARB.__name__ = 'glIsProgramARB'
	gl.glIsProgramARB.errcheck = _errcheck
	gl.glIsProgramARB.restype = GLboolean
	gl.glIsProgramARB.argtypes = (GLuint,)
	gl.glIsProgramNV = _getProcAddress('glIsProgramNV')
	gl.glIsProgramNV.gl = gl
	gl.glIsProgramNV.__name__ = 'glIsProgramNV'
	gl.glIsProgramNV.errcheck = _errcheck
	gl.glIsProgramNV.restype = GLboolean
	gl.glIsProgramNV.argtypes = (GLuint,)
	gl.glIsProgramPipeline = _getProcAddress('glIsProgramPipeline')
	gl.glIsProgramPipeline.gl = gl
	gl.glIsProgramPipeline.__name__ = 'glIsProgramPipeline'
	gl.glIsProgramPipeline.errcheck = _errcheck
	gl.glIsProgramPipeline.restype = GLboolean
	gl.glIsProgramPipeline.argtypes = (GLuint,)
	gl.glIsProgramPipelineEXT = _getProcAddress('glIsProgramPipelineEXT')
	gl.glIsProgramPipelineEXT.gl = gl
	gl.glIsProgramPipelineEXT.__name__ = 'glIsProgramPipelineEXT'
	gl.glIsProgramPipelineEXT.errcheck = _errcheck
	gl.glIsProgramPipelineEXT.restype = GLboolean
	gl.glIsProgramPipelineEXT.argtypes = (GLuint,)
	gl.glIsQuery = _getProcAddress('glIsQuery')
	gl.glIsQuery.gl = gl
	gl.glIsQuery.__name__ = 'glIsQuery'
	gl.glIsQuery.errcheck = _errcheck
	gl.glIsQuery.restype = GLboolean
	gl.glIsQuery.argtypes = (GLuint,)
	gl.glIsQueryARB = _getProcAddress('glIsQueryARB')
	gl.glIsQueryARB.gl = gl
	gl.glIsQueryARB.__name__ = 'glIsQueryARB'
	gl.glIsQueryARB.errcheck = _errcheck
	gl.glIsQueryARB.restype = GLboolean
	gl.glIsQueryARB.argtypes = (GLuint,)
	gl.glIsQueryEXT = _getProcAddress('glIsQueryEXT')
	gl.glIsQueryEXT.gl = gl
	gl.glIsQueryEXT.__name__ = 'glIsQueryEXT'
	gl.glIsQueryEXT.errcheck = _errcheck
	gl.glIsQueryEXT.restype = GLboolean
	gl.glIsQueryEXT.argtypes = (GLuint,)
	gl.glIsRenderbuffer = _getProcAddress('glIsRenderbuffer')
	gl.glIsRenderbuffer.gl = gl
	gl.glIsRenderbuffer.__name__ = 'glIsRenderbuffer'
	gl.glIsRenderbuffer.errcheck = _errcheck
	gl.glIsRenderbuffer.restype = GLboolean
	gl.glIsRenderbuffer.argtypes = (GLuint,)
	gl.glIsRenderbufferEXT = _getProcAddress('glIsRenderbufferEXT')
	gl.glIsRenderbufferEXT.gl = gl
	gl.glIsRenderbufferEXT.__name__ = 'glIsRenderbufferEXT'
	gl.glIsRenderbufferEXT.errcheck = _errcheck
	gl.glIsRenderbufferEXT.restype = GLboolean
	gl.glIsRenderbufferEXT.argtypes = (GLuint,)
	gl.glIsRenderbufferOES = _getProcAddress('glIsRenderbufferOES')
	gl.glIsRenderbufferOES.gl = gl
	gl.glIsRenderbufferOES.__name__ = 'glIsRenderbufferOES'
	gl.glIsRenderbufferOES.errcheck = _errcheck
	gl.glIsRenderbufferOES.restype = GLboolean
	gl.glIsRenderbufferOES.argtypes = (GLuint,)
	gl.glIsSampler = _getProcAddress('glIsSampler')
	gl.glIsSampler.gl = gl
	gl.glIsSampler.__name__ = 'glIsSampler'
	gl.glIsSampler.errcheck = _errcheck
	gl.glIsSampler.restype = GLboolean
	gl.glIsSampler.argtypes = (GLuint,)
	gl.glIsShader = _getProcAddress('glIsShader')
	gl.glIsShader.gl = gl
	gl.glIsShader.__name__ = 'glIsShader'
	gl.glIsShader.errcheck = _errcheck
	gl.glIsShader.restype = GLboolean
	gl.glIsShader.argtypes = (GLuint,)
	gl.glIsSync = _getProcAddress('glIsSync')
	gl.glIsSync.gl = gl
	gl.glIsSync.__name__ = 'glIsSync'
	gl.glIsSync.errcheck = _errcheck
	gl.glIsSync.restype = GLboolean
	gl.glIsSync.argtypes = (GLsync,)
	gl.glIsSyncAPPLE = _getProcAddress('glIsSyncAPPLE')
	gl.glIsSyncAPPLE.gl = gl
	gl.glIsSyncAPPLE.__name__ = 'glIsSyncAPPLE'
	gl.glIsSyncAPPLE.errcheck = _errcheck
	gl.glIsSyncAPPLE.restype = GLboolean
	gl.glIsSyncAPPLE.argtypes = (GLsync,)
	gl.glIsTexture = _getProcAddress('glIsTexture')
	gl.glIsTexture.gl = gl
	gl.glIsTexture.__name__ = 'glIsTexture'
	gl.glIsTexture.errcheck = _errcheck
	gl.glIsTexture.restype = GLboolean
	gl.glIsTexture.argtypes = (GLuint,)
	gl.glIsTextureEXT = _getProcAddress('glIsTextureEXT')
	gl.glIsTextureEXT.gl = gl
	gl.glIsTextureEXT.__name__ = 'glIsTextureEXT'
	gl.glIsTextureEXT.errcheck = _errcheck
	gl.glIsTextureEXT.restype = GLboolean
	gl.glIsTextureEXT.argtypes = (GLuint,)
	gl.glIsTextureHandleResidentARB = _getProcAddress('glIsTextureHandleResidentARB')
	gl.glIsTextureHandleResidentARB.gl = gl
	gl.glIsTextureHandleResidentARB.__name__ = 'glIsTextureHandleResidentARB'
	gl.glIsTextureHandleResidentARB.errcheck = _errcheck
	gl.glIsTextureHandleResidentARB.restype = GLboolean
	gl.glIsTextureHandleResidentARB.argtypes = (GLuint64,)
	gl.glIsTextureHandleResidentNV = _getProcAddress('glIsTextureHandleResidentNV')
	gl.glIsTextureHandleResidentNV.gl = gl
	gl.glIsTextureHandleResidentNV.__name__ = 'glIsTextureHandleResidentNV'
	gl.glIsTextureHandleResidentNV.errcheck = _errcheck
	gl.glIsTextureHandleResidentNV.restype = GLboolean
	gl.glIsTextureHandleResidentNV.argtypes = (GLuint64,)
	gl.glIsTransformFeedback = _getProcAddress('glIsTransformFeedback')
	gl.glIsTransformFeedback.gl = gl
	gl.glIsTransformFeedback.__name__ = 'glIsTransformFeedback'
	gl.glIsTransformFeedback.errcheck = _errcheck
	gl.glIsTransformFeedback.restype = GLboolean
	gl.glIsTransformFeedback.argtypes = (GLuint,)
	gl.glIsTransformFeedbackNV = _getProcAddress('glIsTransformFeedbackNV')
	gl.glIsTransformFeedbackNV.gl = gl
	gl.glIsTransformFeedbackNV.__name__ = 'glIsTransformFeedbackNV'
	gl.glIsTransformFeedbackNV.errcheck = _errcheck
	gl.glIsTransformFeedbackNV.restype = GLboolean
	gl.glIsTransformFeedbackNV.argtypes = (GLuint,)
	gl.glIsVariantEnabledEXT = _getProcAddress('glIsVariantEnabledEXT')
	gl.glIsVariantEnabledEXT.gl = gl
	gl.glIsVariantEnabledEXT.__name__ = 'glIsVariantEnabledEXT'
	gl.glIsVariantEnabledEXT.errcheck = _errcheck
	gl.glIsVariantEnabledEXT.restype = GLboolean
	gl.glIsVariantEnabledEXT.argtypes = (GLuint, GLenum,)
	gl.glIsVertexArray = _getProcAddress('glIsVertexArray')
	gl.glIsVertexArray.gl = gl
	gl.glIsVertexArray.__name__ = 'glIsVertexArray'
	gl.glIsVertexArray.errcheck = _errcheck
	gl.glIsVertexArray.restype = GLboolean
	gl.glIsVertexArray.argtypes = (GLuint,)
	gl.glIsVertexArrayAPPLE = _getProcAddress('glIsVertexArrayAPPLE')
	gl.glIsVertexArrayAPPLE.gl = gl
	gl.glIsVertexArrayAPPLE.__name__ = 'glIsVertexArrayAPPLE'
	gl.glIsVertexArrayAPPLE.errcheck = _errcheck
	gl.glIsVertexArrayAPPLE.restype = GLboolean
	gl.glIsVertexArrayAPPLE.argtypes = (GLuint,)
	gl.glIsVertexArrayOES = _getProcAddress('glIsVertexArrayOES')
	gl.glIsVertexArrayOES.gl = gl
	gl.glIsVertexArrayOES.__name__ = 'glIsVertexArrayOES'
	gl.glIsVertexArrayOES.errcheck = _errcheck
	gl.glIsVertexArrayOES.restype = GLboolean
	gl.glIsVertexArrayOES.argtypes = (GLuint,)
	gl.glIsVertexAttribEnabledAPPLE = _getProcAddress('glIsVertexAttribEnabledAPPLE')
	gl.glIsVertexAttribEnabledAPPLE.gl = gl
	gl.glIsVertexAttribEnabledAPPLE.__name__ = 'glIsVertexAttribEnabledAPPLE'
	gl.glIsVertexAttribEnabledAPPLE.errcheck = _errcheck
	gl.glIsVertexAttribEnabledAPPLE.restype = GLboolean
	gl.glIsVertexAttribEnabledAPPLE.argtypes = (GLuint, GLenum,)
	gl.glLabelObjectEXT = _getProcAddress('glLabelObjectEXT')
	gl.glLabelObjectEXT.gl = gl
	gl.glLabelObjectEXT.__name__ = 'glLabelObjectEXT'
	gl.glLabelObjectEXT.errcheck = _errcheck
	gl.glLabelObjectEXT.restype = GLvoid
	gl.glLabelObjectEXT.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLchar),)
	gl.glLightEnviSGIX = _getProcAddress('glLightEnviSGIX')
	gl.glLightEnviSGIX.gl = gl
	gl.glLightEnviSGIX.__name__ = 'glLightEnviSGIX'
	gl.glLightEnviSGIX.errcheck = _errcheck
	gl.glLightEnviSGIX.restype = GLvoid
	gl.glLightEnviSGIX.argtypes = (GLenum, GLint,)
	gl.glLightModelf = _getProcAddress('glLightModelf')
	gl.glLightModelf.gl = gl
	gl.glLightModelf.__name__ = 'glLightModelf'
	gl.glLightModelf.errcheck = _errcheck
	gl.glLightModelf.restype = GLvoid
	gl.glLightModelf.argtypes = (GLenum, GLfloat,)
	gl.glLightModelfv = _getProcAddress('glLightModelfv')
	gl.glLightModelfv.gl = gl
	gl.glLightModelfv.__name__ = 'glLightModelfv'
	gl.glLightModelfv.errcheck = _errcheck
	gl.glLightModelfv.restype = GLvoid
	gl.glLightModelfv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glLightModeli = _getProcAddress('glLightModeli')
	gl.glLightModeli.gl = gl
	gl.glLightModeli.__name__ = 'glLightModeli'
	gl.glLightModeli.errcheck = _errcheck
	gl.glLightModeli.restype = GLvoid
	gl.glLightModeli.argtypes = (GLenum, GLint,)
	gl.glLightModeliv = _getProcAddress('glLightModeliv')
	gl.glLightModeliv.gl = gl
	gl.glLightModeliv.__name__ = 'glLightModeliv'
	gl.glLightModeliv.errcheck = _errcheck
	gl.glLightModeliv.restype = GLvoid
	gl.glLightModeliv.argtypes = (GLenum, POINTER(GLint),)
	gl.glLightModelx = _getProcAddress('glLightModelx')
	gl.glLightModelx.gl = gl
	gl.glLightModelx.__name__ = 'glLightModelx'
	gl.glLightModelx.errcheck = _errcheck
	gl.glLightModelx.restype = GLvoid
	gl.glLightModelx.argtypes = (GLenum, GLfixed,)
	gl.glLightModelxOES = _getProcAddress('glLightModelxOES')
	gl.glLightModelxOES.gl = gl
	gl.glLightModelxOES.__name__ = 'glLightModelxOES'
	gl.glLightModelxOES.errcheck = _errcheck
	gl.glLightModelxOES.restype = GLvoid
	gl.glLightModelxOES.argtypes = (GLenum, GLfixed,)
	gl.glLightModelxv = _getProcAddress('glLightModelxv')
	gl.glLightModelxv.gl = gl
	gl.glLightModelxv.__name__ = 'glLightModelxv'
	gl.glLightModelxv.errcheck = _errcheck
	gl.glLightModelxv.restype = GLvoid
	gl.glLightModelxv.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glLightModelxvOES = _getProcAddress('glLightModelxvOES')
	gl.glLightModelxvOES.gl = gl
	gl.glLightModelxvOES.__name__ = 'glLightModelxvOES'
	gl.glLightModelxvOES.errcheck = _errcheck
	gl.glLightModelxvOES.restype = GLvoid
	gl.glLightModelxvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glLightf = _getProcAddress('glLightf')
	gl.glLightf.gl = gl
	gl.glLightf.__name__ = 'glLightf'
	gl.glLightf.errcheck = _errcheck
	gl.glLightf.restype = GLvoid
	gl.glLightf.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glLightfv = _getProcAddress('glLightfv')
	gl.glLightfv.gl = gl
	gl.glLightfv.__name__ = 'glLightfv'
	gl.glLightfv.errcheck = _errcheck
	gl.glLightfv.restype = GLvoid
	gl.glLightfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glLighti = _getProcAddress('glLighti')
	gl.glLighti.gl = gl
	gl.glLighti.__name__ = 'glLighti'
	gl.glLighti.errcheck = _errcheck
	gl.glLighti.restype = GLvoid
	gl.glLighti.argtypes = (GLenum, GLenum, GLint,)
	gl.glLightiv = _getProcAddress('glLightiv')
	gl.glLightiv.gl = gl
	gl.glLightiv.__name__ = 'glLightiv'
	gl.glLightiv.errcheck = _errcheck
	gl.glLightiv.restype = GLvoid
	gl.glLightiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glLightx = _getProcAddress('glLightx')
	gl.glLightx.gl = gl
	gl.glLightx.__name__ = 'glLightx'
	gl.glLightx.errcheck = _errcheck
	gl.glLightx.restype = GLvoid
	gl.glLightx.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glLightxOES = _getProcAddress('glLightxOES')
	gl.glLightxOES.gl = gl
	gl.glLightxOES.__name__ = 'glLightxOES'
	gl.glLightxOES.errcheck = _errcheck
	gl.glLightxOES.restype = GLvoid
	gl.glLightxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glLightxv = _getProcAddress('glLightxv')
	gl.glLightxv.gl = gl
	gl.glLightxv.__name__ = 'glLightxv'
	gl.glLightxv.errcheck = _errcheck
	gl.glLightxv.restype = GLvoid
	gl.glLightxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glLightxvOES = _getProcAddress('glLightxvOES')
	gl.glLightxvOES.gl = gl
	gl.glLightxvOES.__name__ = 'glLightxvOES'
	gl.glLightxvOES.errcheck = _errcheck
	gl.glLightxvOES.restype = GLvoid
	gl.glLightxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glLineStipple = _getProcAddress('glLineStipple')
	gl.glLineStipple.gl = gl
	gl.glLineStipple.__name__ = 'glLineStipple'
	gl.glLineStipple.errcheck = _errcheck
	gl.glLineStipple.restype = GLvoid
	gl.glLineStipple.argtypes = (GLint, GLushort,)
	gl.glLineWidth = _getProcAddress('glLineWidth')
	gl.glLineWidth.gl = gl
	gl.glLineWidth.__name__ = 'glLineWidth'
	gl.glLineWidth.errcheck = _errcheck
	gl.glLineWidth.restype = GLvoid
	gl.glLineWidth.argtypes = (GLfloat,)
	gl.glLineWidthx = _getProcAddress('glLineWidthx')
	gl.glLineWidthx.gl = gl
	gl.glLineWidthx.__name__ = 'glLineWidthx'
	gl.glLineWidthx.errcheck = _errcheck
	gl.glLineWidthx.restype = GLvoid
	gl.glLineWidthx.argtypes = (GLfixed,)
	gl.glLineWidthxOES = _getProcAddress('glLineWidthxOES')
	gl.glLineWidthxOES.gl = gl
	gl.glLineWidthxOES.__name__ = 'glLineWidthxOES'
	gl.glLineWidthxOES.errcheck = _errcheck
	gl.glLineWidthxOES.restype = GLvoid
	gl.glLineWidthxOES.argtypes = (GLfixed,)
	gl.glLinkProgram = _getProcAddress('glLinkProgram')
	gl.glLinkProgram.gl = gl
	gl.glLinkProgram.__name__ = 'glLinkProgram'
	gl.glLinkProgram.errcheck = _errcheck
	gl.glLinkProgram.restype = GLvoid
	gl.glLinkProgram.argtypes = (GLuint,)
	gl.glLinkProgramARB = _getProcAddress('glLinkProgramARB')
	gl.glLinkProgramARB.gl = gl
	gl.glLinkProgramARB.__name__ = 'glLinkProgramARB'
	gl.glLinkProgramARB.errcheck = _errcheck
	gl.glLinkProgramARB.restype = GLvoid
	gl.glLinkProgramARB.argtypes = (GLhandleARB,)
	gl.glListBase = _getProcAddress('glListBase')
	gl.glListBase.gl = gl
	gl.glListBase.__name__ = 'glListBase'
	gl.glListBase.errcheck = _errcheck
	gl.glListBase.restype = GLvoid
	gl.glListBase.argtypes = (GLuint,)
	gl.glListParameterfSGIX = _getProcAddress('glListParameterfSGIX')
	gl.glListParameterfSGIX.gl = gl
	gl.glListParameterfSGIX.__name__ = 'glListParameterfSGIX'
	gl.glListParameterfSGIX.errcheck = _errcheck
	gl.glListParameterfSGIX.restype = GLvoid
	gl.glListParameterfSGIX.argtypes = (GLuint, GLenum, GLfloat,)
	gl.glListParameterfvSGIX = _getProcAddress('glListParameterfvSGIX')
	gl.glListParameterfvSGIX.gl = gl
	gl.glListParameterfvSGIX.__name__ = 'glListParameterfvSGIX'
	gl.glListParameterfvSGIX.errcheck = _errcheck
	gl.glListParameterfvSGIX.restype = GLvoid
	gl.glListParameterfvSGIX.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glListParameteriSGIX = _getProcAddress('glListParameteriSGIX')
	gl.glListParameteriSGIX.gl = gl
	gl.glListParameteriSGIX.__name__ = 'glListParameteriSGIX'
	gl.glListParameteriSGIX.errcheck = _errcheck
	gl.glListParameteriSGIX.restype = GLvoid
	gl.glListParameteriSGIX.argtypes = (GLuint, GLenum, GLint,)
	gl.glListParameterivSGIX = _getProcAddress('glListParameterivSGIX')
	gl.glListParameterivSGIX.gl = gl
	gl.glListParameterivSGIX.__name__ = 'glListParameterivSGIX'
	gl.glListParameterivSGIX.errcheck = _errcheck
	gl.glListParameterivSGIX.restype = GLvoid
	gl.glListParameterivSGIX.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glLoadIdentity = _getProcAddress('glLoadIdentity')
	gl.glLoadIdentity.gl = gl
	gl.glLoadIdentity.__name__ = 'glLoadIdentity'
	gl.glLoadIdentity.errcheck = _errcheck
	gl.glLoadIdentity.restype = GLvoid
	gl.glLoadIdentity.argtypes = ()
	gl.glLoadIdentityDeformationMapSGIX = _getProcAddress('glLoadIdentityDeformationMapSGIX')
	gl.glLoadIdentityDeformationMapSGIX.gl = gl
	gl.glLoadIdentityDeformationMapSGIX.__name__ = 'glLoadIdentityDeformationMapSGIX'
	gl.glLoadIdentityDeformationMapSGIX.errcheck = _errcheck
	gl.glLoadIdentityDeformationMapSGIX.restype = GLvoid
	gl.glLoadIdentityDeformationMapSGIX.argtypes = (GLbitfield,)
	gl.glLoadMatrixd = _getProcAddress('glLoadMatrixd')
	gl.glLoadMatrixd.gl = gl
	gl.glLoadMatrixd.__name__ = 'glLoadMatrixd'
	gl.glLoadMatrixd.errcheck = _errcheck
	gl.glLoadMatrixd.restype = GLvoid
	gl.glLoadMatrixd.argtypes = (POINTER(GLdouble),)
	gl.glLoadMatrixf = _getProcAddress('glLoadMatrixf')
	gl.glLoadMatrixf.gl = gl
	gl.glLoadMatrixf.__name__ = 'glLoadMatrixf'
	gl.glLoadMatrixf.errcheck = _errcheck
	gl.glLoadMatrixf.restype = GLvoid
	gl.glLoadMatrixf.argtypes = (POINTER(GLfloat),)
	gl.glLoadMatrixx = _getProcAddress('glLoadMatrixx')
	gl.glLoadMatrixx.gl = gl
	gl.glLoadMatrixx.__name__ = 'glLoadMatrixx'
	gl.glLoadMatrixx.errcheck = _errcheck
	gl.glLoadMatrixx.restype = GLvoid
	gl.glLoadMatrixx.argtypes = (POINTER(GLfixed),)
	gl.glLoadMatrixxOES = _getProcAddress('glLoadMatrixxOES')
	gl.glLoadMatrixxOES.gl = gl
	gl.glLoadMatrixxOES.__name__ = 'glLoadMatrixxOES'
	gl.glLoadMatrixxOES.errcheck = _errcheck
	gl.glLoadMatrixxOES.restype = GLvoid
	gl.glLoadMatrixxOES.argtypes = (POINTER(GLfixed),)
	gl.glLoadName = _getProcAddress('glLoadName')
	gl.glLoadName.gl = gl
	gl.glLoadName.__name__ = 'glLoadName'
	gl.glLoadName.errcheck = _errcheck
	gl.glLoadName.restype = GLvoid
	gl.glLoadName.argtypes = (GLuint,)
	gl.glLoadPaletteFromModelViewMatrixOES = _getProcAddress('glLoadPaletteFromModelViewMatrixOES')
	gl.glLoadPaletteFromModelViewMatrixOES.gl = gl
	gl.glLoadPaletteFromModelViewMatrixOES.__name__ = 'glLoadPaletteFromModelViewMatrixOES'
	gl.glLoadPaletteFromModelViewMatrixOES.errcheck = _errcheck
	gl.glLoadPaletteFromModelViewMatrixOES.restype = GLvoid
	gl.glLoadPaletteFromModelViewMatrixOES.argtypes = ()
	gl.glLoadProgramNV = _getProcAddress('glLoadProgramNV')
	gl.glLoadProgramNV.gl = gl
	gl.glLoadProgramNV.__name__ = 'glLoadProgramNV'
	gl.glLoadProgramNV.errcheck = _errcheck
	gl.glLoadProgramNV.restype = GLvoid
	gl.glLoadProgramNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLubyte),)
	gl.glLoadTransposeMatrixd = _getProcAddress('glLoadTransposeMatrixd')
	gl.glLoadTransposeMatrixd.gl = gl
	gl.glLoadTransposeMatrixd.__name__ = 'glLoadTransposeMatrixd'
	gl.glLoadTransposeMatrixd.errcheck = _errcheck
	gl.glLoadTransposeMatrixd.restype = GLvoid
	gl.glLoadTransposeMatrixd.argtypes = (POINTER(GLdouble),)
	gl.glLoadTransposeMatrixdARB = _getProcAddress('glLoadTransposeMatrixdARB')
	gl.glLoadTransposeMatrixdARB.gl = gl
	gl.glLoadTransposeMatrixdARB.__name__ = 'glLoadTransposeMatrixdARB'
	gl.glLoadTransposeMatrixdARB.errcheck = _errcheck
	gl.glLoadTransposeMatrixdARB.restype = GLvoid
	gl.glLoadTransposeMatrixdARB.argtypes = (POINTER(GLdouble),)
	gl.glLoadTransposeMatrixf = _getProcAddress('glLoadTransposeMatrixf')
	gl.glLoadTransposeMatrixf.gl = gl
	gl.glLoadTransposeMatrixf.__name__ = 'glLoadTransposeMatrixf'
	gl.glLoadTransposeMatrixf.errcheck = _errcheck
	gl.glLoadTransposeMatrixf.restype = GLvoid
	gl.glLoadTransposeMatrixf.argtypes = (POINTER(GLfloat),)
	gl.glLoadTransposeMatrixfARB = _getProcAddress('glLoadTransposeMatrixfARB')
	gl.glLoadTransposeMatrixfARB.gl = gl
	gl.glLoadTransposeMatrixfARB.__name__ = 'glLoadTransposeMatrixfARB'
	gl.glLoadTransposeMatrixfARB.errcheck = _errcheck
	gl.glLoadTransposeMatrixfARB.restype = GLvoid
	gl.glLoadTransposeMatrixfARB.argtypes = (POINTER(GLfloat),)
	gl.glLoadTransposeMatrixxOES = _getProcAddress('glLoadTransposeMatrixxOES')
	gl.glLoadTransposeMatrixxOES.gl = gl
	gl.glLoadTransposeMatrixxOES.__name__ = 'glLoadTransposeMatrixxOES'
	gl.glLoadTransposeMatrixxOES.errcheck = _errcheck
	gl.glLoadTransposeMatrixxOES.restype = GLvoid
	gl.glLoadTransposeMatrixxOES.argtypes = (POINTER(GLfixed),)
	gl.glLockArraysEXT = _getProcAddress('glLockArraysEXT')
	gl.glLockArraysEXT.gl = gl
	gl.glLockArraysEXT.__name__ = 'glLockArraysEXT'
	gl.glLockArraysEXT.errcheck = _errcheck
	gl.glLockArraysEXT.restype = GLvoid
	gl.glLockArraysEXT.argtypes = (GLint, GLsizei,)
	gl.glLogicOp = _getProcAddress('glLogicOp')
	gl.glLogicOp.gl = gl
	gl.glLogicOp.__name__ = 'glLogicOp'
	gl.glLogicOp.errcheck = _errcheck
	gl.glLogicOp.restype = GLvoid
	gl.glLogicOp.argtypes = (GLenum,)
	gl.glMakeBufferNonResidentNV = _getProcAddress('glMakeBufferNonResidentNV')
	gl.glMakeBufferNonResidentNV.gl = gl
	gl.glMakeBufferNonResidentNV.__name__ = 'glMakeBufferNonResidentNV'
	gl.glMakeBufferNonResidentNV.errcheck = _errcheck
	gl.glMakeBufferNonResidentNV.restype = GLvoid
	gl.glMakeBufferNonResidentNV.argtypes = (GLenum,)
	gl.glMakeBufferResidentNV = _getProcAddress('glMakeBufferResidentNV')
	gl.glMakeBufferResidentNV.gl = gl
	gl.glMakeBufferResidentNV.__name__ = 'glMakeBufferResidentNV'
	gl.glMakeBufferResidentNV.errcheck = _errcheck
	gl.glMakeBufferResidentNV.restype = GLvoid
	gl.glMakeBufferResidentNV.argtypes = (GLenum, GLenum,)
	gl.glMakeImageHandleNonResidentARB = _getProcAddress('glMakeImageHandleNonResidentARB')
	gl.glMakeImageHandleNonResidentARB.gl = gl
	gl.glMakeImageHandleNonResidentARB.__name__ = 'glMakeImageHandleNonResidentARB'
	gl.glMakeImageHandleNonResidentARB.errcheck = _errcheck
	gl.glMakeImageHandleNonResidentARB.restype = GLvoid
	gl.glMakeImageHandleNonResidentARB.argtypes = (GLuint64,)
	gl.glMakeImageHandleNonResidentNV = _getProcAddress('glMakeImageHandleNonResidentNV')
	gl.glMakeImageHandleNonResidentNV.gl = gl
	gl.glMakeImageHandleNonResidentNV.__name__ = 'glMakeImageHandleNonResidentNV'
	gl.glMakeImageHandleNonResidentNV.errcheck = _errcheck
	gl.glMakeImageHandleNonResidentNV.restype = GLvoid
	gl.glMakeImageHandleNonResidentNV.argtypes = (GLuint64,)
	gl.glMakeImageHandleResidentARB = _getProcAddress('glMakeImageHandleResidentARB')
	gl.glMakeImageHandleResidentARB.gl = gl
	gl.glMakeImageHandleResidentARB.__name__ = 'glMakeImageHandleResidentARB'
	gl.glMakeImageHandleResidentARB.errcheck = _errcheck
	gl.glMakeImageHandleResidentARB.restype = GLvoid
	gl.glMakeImageHandleResidentARB.argtypes = (GLuint64, GLenum,)
	gl.glMakeImageHandleResidentNV = _getProcAddress('glMakeImageHandleResidentNV')
	gl.glMakeImageHandleResidentNV.gl = gl
	gl.glMakeImageHandleResidentNV.__name__ = 'glMakeImageHandleResidentNV'
	gl.glMakeImageHandleResidentNV.errcheck = _errcheck
	gl.glMakeImageHandleResidentNV.restype = GLvoid
	gl.glMakeImageHandleResidentNV.argtypes = (GLuint64, GLenum,)
	gl.glMakeNamedBufferNonResidentNV = _getProcAddress('glMakeNamedBufferNonResidentNV')
	gl.glMakeNamedBufferNonResidentNV.gl = gl
	gl.glMakeNamedBufferNonResidentNV.__name__ = 'glMakeNamedBufferNonResidentNV'
	gl.glMakeNamedBufferNonResidentNV.errcheck = _errcheck
	gl.glMakeNamedBufferNonResidentNV.restype = GLvoid
	gl.glMakeNamedBufferNonResidentNV.argtypes = (GLuint,)
	gl.glMakeNamedBufferResidentNV = _getProcAddress('glMakeNamedBufferResidentNV')
	gl.glMakeNamedBufferResidentNV.gl = gl
	gl.glMakeNamedBufferResidentNV.__name__ = 'glMakeNamedBufferResidentNV'
	gl.glMakeNamedBufferResidentNV.errcheck = _errcheck
	gl.glMakeNamedBufferResidentNV.restype = GLvoid
	gl.glMakeNamedBufferResidentNV.argtypes = (GLuint, GLenum,)
	gl.glMakeTextureHandleNonResidentARB = _getProcAddress('glMakeTextureHandleNonResidentARB')
	gl.glMakeTextureHandleNonResidentARB.gl = gl
	gl.glMakeTextureHandleNonResidentARB.__name__ = 'glMakeTextureHandleNonResidentARB'
	gl.glMakeTextureHandleNonResidentARB.errcheck = _errcheck
	gl.glMakeTextureHandleNonResidentARB.restype = GLvoid
	gl.glMakeTextureHandleNonResidentARB.argtypes = (GLuint64,)
	gl.glMakeTextureHandleNonResidentNV = _getProcAddress('glMakeTextureHandleNonResidentNV')
	gl.glMakeTextureHandleNonResidentNV.gl = gl
	gl.glMakeTextureHandleNonResidentNV.__name__ = 'glMakeTextureHandleNonResidentNV'
	gl.glMakeTextureHandleNonResidentNV.errcheck = _errcheck
	gl.glMakeTextureHandleNonResidentNV.restype = GLvoid
	gl.glMakeTextureHandleNonResidentNV.argtypes = (GLuint64,)
	gl.glMakeTextureHandleResidentARB = _getProcAddress('glMakeTextureHandleResidentARB')
	gl.glMakeTextureHandleResidentARB.gl = gl
	gl.glMakeTextureHandleResidentARB.__name__ = 'glMakeTextureHandleResidentARB'
	gl.glMakeTextureHandleResidentARB.errcheck = _errcheck
	gl.glMakeTextureHandleResidentARB.restype = GLvoid
	gl.glMakeTextureHandleResidentARB.argtypes = (GLuint64,)
	gl.glMakeTextureHandleResidentNV = _getProcAddress('glMakeTextureHandleResidentNV')
	gl.glMakeTextureHandleResidentNV.gl = gl
	gl.glMakeTextureHandleResidentNV.__name__ = 'glMakeTextureHandleResidentNV'
	gl.glMakeTextureHandleResidentNV.errcheck = _errcheck
	gl.glMakeTextureHandleResidentNV.restype = GLvoid
	gl.glMakeTextureHandleResidentNV.argtypes = (GLuint64,)
	gl.glMap1d = _getProcAddress('glMap1d')
	gl.glMap1d.gl = gl
	gl.glMap1d.__name__ = 'glMap1d'
	gl.glMap1d.errcheck = _errcheck
	gl.glMap1d.restype = GLvoid
	gl.glMap1d.argtypes = (GLenum, GLdouble, GLdouble, GLint, GLint, POINTER(GLdouble),)
	gl.glMap1f = _getProcAddress('glMap1f')
	gl.glMap1f.gl = gl
	gl.glMap1f.__name__ = 'glMap1f'
	gl.glMap1f.errcheck = _errcheck
	gl.glMap1f.restype = GLvoid
	gl.glMap1f.argtypes = (GLenum, GLfloat, GLfloat, GLint, GLint, POINTER(GLfloat),)
	gl.glMap1xOES = _getProcAddress('glMap1xOES')
	gl.glMap1xOES.gl = gl
	gl.glMap1xOES.__name__ = 'glMap1xOES'
	gl.glMap1xOES.errcheck = _errcheck
	gl.glMap1xOES.restype = GLvoid
	gl.glMap1xOES.argtypes = (GLenum, GLfixed, GLfixed, GLint, GLint, GLfixed,)
	gl.glMap2d = _getProcAddress('glMap2d')
	gl.glMap2d.gl = gl
	gl.glMap2d.__name__ = 'glMap2d'
	gl.glMap2d.errcheck = _errcheck
	gl.glMap2d.restype = GLvoid
	gl.glMap2d.argtypes = (GLenum, GLdouble, GLdouble, GLint, GLint, GLdouble, GLdouble, GLint, GLint, POINTER(GLdouble),)
	gl.glMap2f = _getProcAddress('glMap2f')
	gl.glMap2f.gl = gl
	gl.glMap2f.__name__ = 'glMap2f'
	gl.glMap2f.errcheck = _errcheck
	gl.glMap2f.restype = GLvoid
	gl.glMap2f.argtypes = (GLenum, GLfloat, GLfloat, GLint, GLint, GLfloat, GLfloat, GLint, GLint, POINTER(GLfloat),)
	gl.glMap2xOES = _getProcAddress('glMap2xOES')
	gl.glMap2xOES.gl = gl
	gl.glMap2xOES.__name__ = 'glMap2xOES'
	gl.glMap2xOES.errcheck = _errcheck
	gl.glMap2xOES.restype = GLvoid
	gl.glMap2xOES.argtypes = (GLenum, GLfixed, GLfixed, GLint, GLint, GLfixed, GLfixed, GLint, GLint, GLfixed,)
	gl.glMapBuffer = _getProcAddress('glMapBuffer')
	gl.glMapBuffer.gl = gl
	gl.glMapBuffer.__name__ = 'glMapBuffer'
	gl.glMapBuffer.errcheck = _errcheck
	gl.glMapBuffer.restype = POINTER(GLvoid)
	gl.glMapBuffer.argtypes = (GLenum, GLenum,)
	gl.glMapBufferARB = _getProcAddress('glMapBufferARB')
	gl.glMapBufferARB.gl = gl
	gl.glMapBufferARB.__name__ = 'glMapBufferARB'
	gl.glMapBufferARB.errcheck = _errcheck
	gl.glMapBufferARB.restype = POINTER(GLvoid)
	gl.glMapBufferARB.argtypes = (GLenum, GLenum,)
	gl.glMapBufferOES = _getProcAddress('glMapBufferOES')
	gl.glMapBufferOES.gl = gl
	gl.glMapBufferOES.__name__ = 'glMapBufferOES'
	gl.glMapBufferOES.errcheck = _errcheck
	gl.glMapBufferOES.restype = POINTER(GLvoid)
	gl.glMapBufferOES.argtypes = (GLenum, GLenum,)
	gl.glMapBufferRange = _getProcAddress('glMapBufferRange')
	gl.glMapBufferRange.gl = gl
	gl.glMapBufferRange.__name__ = 'glMapBufferRange'
	gl.glMapBufferRange.errcheck = _errcheck
	gl.glMapBufferRange.restype = POINTER(GLvoid)
	gl.glMapBufferRange.argtypes = (GLenum, GLintptr, GLsizeiptr, GLbitfield,)
	gl.glMapBufferRangeEXT = _getProcAddress('glMapBufferRangeEXT')
	gl.glMapBufferRangeEXT.gl = gl
	gl.glMapBufferRangeEXT.__name__ = 'glMapBufferRangeEXT'
	gl.glMapBufferRangeEXT.errcheck = _errcheck
	gl.glMapBufferRangeEXT.restype = POINTER(GLvoid)
	gl.glMapBufferRangeEXT.argtypes = (GLenum, GLintptr, GLsizeiptr, GLbitfield,)
	gl.glMapControlPointsNV = _getProcAddress('glMapControlPointsNV')
	gl.glMapControlPointsNV.gl = gl
	gl.glMapControlPointsNV.__name__ = 'glMapControlPointsNV'
	gl.glMapControlPointsNV.errcheck = _errcheck
	gl.glMapControlPointsNV.restype = GLvoid
	gl.glMapControlPointsNV.argtypes = (GLenum, GLuint, GLenum, GLsizei, GLsizei, GLint, GLint, GLboolean, POINTER(GLvoid),)
	gl.glMapGrid1d = _getProcAddress('glMapGrid1d')
	gl.glMapGrid1d.gl = gl
	gl.glMapGrid1d.__name__ = 'glMapGrid1d'
	gl.glMapGrid1d.errcheck = _errcheck
	gl.glMapGrid1d.restype = GLvoid
	gl.glMapGrid1d.argtypes = (GLint, GLdouble, GLdouble,)
	gl.glMapGrid1f = _getProcAddress('glMapGrid1f')
	gl.glMapGrid1f.gl = gl
	gl.glMapGrid1f.__name__ = 'glMapGrid1f'
	gl.glMapGrid1f.errcheck = _errcheck
	gl.glMapGrid1f.restype = GLvoid
	gl.glMapGrid1f.argtypes = (GLint, GLfloat, GLfloat,)
	gl.glMapGrid1xOES = _getProcAddress('glMapGrid1xOES')
	gl.glMapGrid1xOES.gl = gl
	gl.glMapGrid1xOES.__name__ = 'glMapGrid1xOES'
	gl.glMapGrid1xOES.errcheck = _errcheck
	gl.glMapGrid1xOES.restype = GLvoid
	gl.glMapGrid1xOES.argtypes = (GLint, GLfixed, GLfixed,)
	gl.glMapGrid2d = _getProcAddress('glMapGrid2d')
	gl.glMapGrid2d.gl = gl
	gl.glMapGrid2d.__name__ = 'glMapGrid2d'
	gl.glMapGrid2d.errcheck = _errcheck
	gl.glMapGrid2d.restype = GLvoid
	gl.glMapGrid2d.argtypes = (GLint, GLdouble, GLdouble, GLint, GLdouble, GLdouble,)
	gl.glMapGrid2f = _getProcAddress('glMapGrid2f')
	gl.glMapGrid2f.gl = gl
	gl.glMapGrid2f.__name__ = 'glMapGrid2f'
	gl.glMapGrid2f.errcheck = _errcheck
	gl.glMapGrid2f.restype = GLvoid
	gl.glMapGrid2f.argtypes = (GLint, GLfloat, GLfloat, GLint, GLfloat, GLfloat,)
	gl.glMapGrid2xOES = _getProcAddress('glMapGrid2xOES')
	gl.glMapGrid2xOES.gl = gl
	gl.glMapGrid2xOES.__name__ = 'glMapGrid2xOES'
	gl.glMapGrid2xOES.errcheck = _errcheck
	gl.glMapGrid2xOES.restype = GLvoid
	gl.glMapGrid2xOES.argtypes = (GLint, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glMapNamedBuffer = _getProcAddress('glMapNamedBuffer')
	gl.glMapNamedBuffer.gl = gl
	gl.glMapNamedBuffer.__name__ = 'glMapNamedBuffer'
	gl.glMapNamedBuffer.errcheck = _errcheck
	gl.glMapNamedBuffer.restype = POINTER(GLvoid)
	gl.glMapNamedBuffer.argtypes = (GLuint, GLenum,)
	gl.glMapNamedBufferEXT = _getProcAddress('glMapNamedBufferEXT')
	gl.glMapNamedBufferEXT.gl = gl
	gl.glMapNamedBufferEXT.__name__ = 'glMapNamedBufferEXT'
	gl.glMapNamedBufferEXT.errcheck = _errcheck
	gl.glMapNamedBufferEXT.restype = POINTER(GLvoid)
	gl.glMapNamedBufferEXT.argtypes = (GLuint, GLenum,)
	gl.glMapNamedBufferRange = _getProcAddress('glMapNamedBufferRange')
	gl.glMapNamedBufferRange.gl = gl
	gl.glMapNamedBufferRange.__name__ = 'glMapNamedBufferRange'
	gl.glMapNamedBufferRange.errcheck = _errcheck
	gl.glMapNamedBufferRange.restype = POINTER(GLvoid)
	gl.glMapNamedBufferRange.argtypes = (GLuint, GLintptr, GLsizei, GLbitfield,)
	gl.glMapNamedBufferRangeEXT = _getProcAddress('glMapNamedBufferRangeEXT')
	gl.glMapNamedBufferRangeEXT.gl = gl
	gl.glMapNamedBufferRangeEXT.__name__ = 'glMapNamedBufferRangeEXT'
	gl.glMapNamedBufferRangeEXT.errcheck = _errcheck
	gl.glMapNamedBufferRangeEXT.restype = POINTER(GLvoid)
	gl.glMapNamedBufferRangeEXT.argtypes = (GLuint, GLintptr, GLsizeiptr, GLbitfield,)
	gl.glMapObjectBufferATI = _getProcAddress('glMapObjectBufferATI')
	gl.glMapObjectBufferATI.gl = gl
	gl.glMapObjectBufferATI.__name__ = 'glMapObjectBufferATI'
	gl.glMapObjectBufferATI.errcheck = _errcheck
	gl.glMapObjectBufferATI.restype = POINTER(GLvoid)
	gl.glMapObjectBufferATI.argtypes = (GLuint,)
	gl.glMapParameterfvNV = _getProcAddress('glMapParameterfvNV')
	gl.glMapParameterfvNV.gl = gl
	gl.glMapParameterfvNV.__name__ = 'glMapParameterfvNV'
	gl.glMapParameterfvNV.errcheck = _errcheck
	gl.glMapParameterfvNV.restype = GLvoid
	gl.glMapParameterfvNV.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glMapParameterivNV = _getProcAddress('glMapParameterivNV')
	gl.glMapParameterivNV.gl = gl
	gl.glMapParameterivNV.__name__ = 'glMapParameterivNV'
	gl.glMapParameterivNV.errcheck = _errcheck
	gl.glMapParameterivNV.restype = GLvoid
	gl.glMapParameterivNV.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glMapTexture2DINTEL = _getProcAddress('glMapTexture2DINTEL')
	gl.glMapTexture2DINTEL.gl = gl
	gl.glMapTexture2DINTEL.__name__ = 'glMapTexture2DINTEL'
	gl.glMapTexture2DINTEL.errcheck = _errcheck
	gl.glMapTexture2DINTEL.restype = POINTER(GLvoid)
	gl.glMapTexture2DINTEL.argtypes = (GLuint, GLint, GLbitfield, POINTER(GLint), POINTER(GLenum),)
	gl.glMapVertexAttrib1dAPPLE = _getProcAddress('glMapVertexAttrib1dAPPLE')
	gl.glMapVertexAttrib1dAPPLE.gl = gl
	gl.glMapVertexAttrib1dAPPLE.__name__ = 'glMapVertexAttrib1dAPPLE'
	gl.glMapVertexAttrib1dAPPLE.errcheck = _errcheck
	gl.glMapVertexAttrib1dAPPLE.restype = GLvoid
	gl.glMapVertexAttrib1dAPPLE.argtypes = (GLuint, GLuint, GLdouble, GLdouble, GLint, GLint, POINTER(GLdouble),)
	gl.glMapVertexAttrib1fAPPLE = _getProcAddress('glMapVertexAttrib1fAPPLE')
	gl.glMapVertexAttrib1fAPPLE.gl = gl
	gl.glMapVertexAttrib1fAPPLE.__name__ = 'glMapVertexAttrib1fAPPLE'
	gl.glMapVertexAttrib1fAPPLE.errcheck = _errcheck
	gl.glMapVertexAttrib1fAPPLE.restype = GLvoid
	gl.glMapVertexAttrib1fAPPLE.argtypes = (GLuint, GLuint, GLfloat, GLfloat, GLint, GLint, POINTER(GLfloat),)
	gl.glMapVertexAttrib2dAPPLE = _getProcAddress('glMapVertexAttrib2dAPPLE')
	gl.glMapVertexAttrib2dAPPLE.gl = gl
	gl.glMapVertexAttrib2dAPPLE.__name__ = 'glMapVertexAttrib2dAPPLE'
	gl.glMapVertexAttrib2dAPPLE.errcheck = _errcheck
	gl.glMapVertexAttrib2dAPPLE.restype = GLvoid
	gl.glMapVertexAttrib2dAPPLE.argtypes = (GLuint, GLuint, GLdouble, GLdouble, GLint, GLint, GLdouble, GLdouble, GLint, GLint, POINTER(GLdouble),)
	gl.glMapVertexAttrib2fAPPLE = _getProcAddress('glMapVertexAttrib2fAPPLE')
	gl.glMapVertexAttrib2fAPPLE.gl = gl
	gl.glMapVertexAttrib2fAPPLE.__name__ = 'glMapVertexAttrib2fAPPLE'
	gl.glMapVertexAttrib2fAPPLE.errcheck = _errcheck
	gl.glMapVertexAttrib2fAPPLE.restype = GLvoid
	gl.glMapVertexAttrib2fAPPLE.argtypes = (GLuint, GLuint, GLfloat, GLfloat, GLint, GLint, GLfloat, GLfloat, GLint, GLint, POINTER(GLfloat),)
	gl.glMaterialf = _getProcAddress('glMaterialf')
	gl.glMaterialf.gl = gl
	gl.glMaterialf.__name__ = 'glMaterialf'
	gl.glMaterialf.errcheck = _errcheck
	gl.glMaterialf.restype = GLvoid
	gl.glMaterialf.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glMaterialfv = _getProcAddress('glMaterialfv')
	gl.glMaterialfv.gl = gl
	gl.glMaterialfv.__name__ = 'glMaterialfv'
	gl.glMaterialfv.errcheck = _errcheck
	gl.glMaterialfv.restype = GLvoid
	gl.glMaterialfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glMateriali = _getProcAddress('glMateriali')
	gl.glMateriali.gl = gl
	gl.glMateriali.__name__ = 'glMateriali'
	gl.glMateriali.errcheck = _errcheck
	gl.glMateriali.restype = GLvoid
	gl.glMateriali.argtypes = (GLenum, GLenum, GLint,)
	gl.glMaterialiv = _getProcAddress('glMaterialiv')
	gl.glMaterialiv.gl = gl
	gl.glMaterialiv.__name__ = 'glMaterialiv'
	gl.glMaterialiv.errcheck = _errcheck
	gl.glMaterialiv.restype = GLvoid
	gl.glMaterialiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glMaterialx = _getProcAddress('glMaterialx')
	gl.glMaterialx.gl = gl
	gl.glMaterialx.__name__ = 'glMaterialx'
	gl.glMaterialx.errcheck = _errcheck
	gl.glMaterialx.restype = GLvoid
	gl.glMaterialx.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glMaterialxOES = _getProcAddress('glMaterialxOES')
	gl.glMaterialxOES.gl = gl
	gl.glMaterialxOES.__name__ = 'glMaterialxOES'
	gl.glMaterialxOES.errcheck = _errcheck
	gl.glMaterialxOES.restype = GLvoid
	gl.glMaterialxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glMaterialxv = _getProcAddress('glMaterialxv')
	gl.glMaterialxv.gl = gl
	gl.glMaterialxv.__name__ = 'glMaterialxv'
	gl.glMaterialxv.errcheck = _errcheck
	gl.glMaterialxv.restype = GLvoid
	gl.glMaterialxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glMaterialxvOES = _getProcAddress('glMaterialxvOES')
	gl.glMaterialxvOES.gl = gl
	gl.glMaterialxvOES.__name__ = 'glMaterialxvOES'
	gl.glMaterialxvOES.errcheck = _errcheck
	gl.glMaterialxvOES.restype = GLvoid
	gl.glMaterialxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glMatrixFrustumEXT = _getProcAddress('glMatrixFrustumEXT')
	gl.glMatrixFrustumEXT.gl = gl
	gl.glMatrixFrustumEXT.__name__ = 'glMatrixFrustumEXT'
	gl.glMatrixFrustumEXT.errcheck = _errcheck
	gl.glMatrixFrustumEXT.restype = GLvoid
	gl.glMatrixFrustumEXT.argtypes = (GLenum, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glMatrixIndexPointerARB = _getProcAddress('glMatrixIndexPointerARB')
	gl.glMatrixIndexPointerARB.gl = gl
	gl.glMatrixIndexPointerARB.__name__ = 'glMatrixIndexPointerARB'
	gl.glMatrixIndexPointerARB.errcheck = _errcheck
	gl.glMatrixIndexPointerARB.restype = GLvoid
	gl.glMatrixIndexPointerARB.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glMatrixIndexPointerOES = _getProcAddress('glMatrixIndexPointerOES')
	gl.glMatrixIndexPointerOES.gl = gl
	gl.glMatrixIndexPointerOES.__name__ = 'glMatrixIndexPointerOES'
	gl.glMatrixIndexPointerOES.errcheck = _errcheck
	gl.glMatrixIndexPointerOES.restype = GLvoid
	gl.glMatrixIndexPointerOES.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glMatrixIndexubvARB = _getProcAddress('glMatrixIndexubvARB')
	gl.glMatrixIndexubvARB.gl = gl
	gl.glMatrixIndexubvARB.__name__ = 'glMatrixIndexubvARB'
	gl.glMatrixIndexubvARB.errcheck = _errcheck
	gl.glMatrixIndexubvARB.restype = GLvoid
	gl.glMatrixIndexubvARB.argtypes = (GLint, POINTER(GLubyte),)
	gl.glMatrixIndexuivARB = _getProcAddress('glMatrixIndexuivARB')
	gl.glMatrixIndexuivARB.gl = gl
	gl.glMatrixIndexuivARB.__name__ = 'glMatrixIndexuivARB'
	gl.glMatrixIndexuivARB.errcheck = _errcheck
	gl.glMatrixIndexuivARB.restype = GLvoid
	gl.glMatrixIndexuivARB.argtypes = (GLint, POINTER(GLuint),)
	gl.glMatrixIndexusvARB = _getProcAddress('glMatrixIndexusvARB')
	gl.glMatrixIndexusvARB.gl = gl
	gl.glMatrixIndexusvARB.__name__ = 'glMatrixIndexusvARB'
	gl.glMatrixIndexusvARB.errcheck = _errcheck
	gl.glMatrixIndexusvARB.restype = GLvoid
	gl.glMatrixIndexusvARB.argtypes = (GLint, POINTER(GLushort),)
	gl.glMatrixLoad3x2fNV = _getProcAddress('glMatrixLoad3x2fNV')
	gl.glMatrixLoad3x2fNV.gl = gl
	gl.glMatrixLoad3x2fNV.__name__ = 'glMatrixLoad3x2fNV'
	gl.glMatrixLoad3x2fNV.errcheck = _errcheck
	gl.glMatrixLoad3x2fNV.restype = GLvoid
	gl.glMatrixLoad3x2fNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixLoad3x3fNV = _getProcAddress('glMatrixLoad3x3fNV')
	gl.glMatrixLoad3x3fNV.gl = gl
	gl.glMatrixLoad3x3fNV.__name__ = 'glMatrixLoad3x3fNV'
	gl.glMatrixLoad3x3fNV.errcheck = _errcheck
	gl.glMatrixLoad3x3fNV.restype = GLvoid
	gl.glMatrixLoad3x3fNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixLoadIdentityEXT = _getProcAddress('glMatrixLoadIdentityEXT')
	gl.glMatrixLoadIdentityEXT.gl = gl
	gl.glMatrixLoadIdentityEXT.__name__ = 'glMatrixLoadIdentityEXT'
	gl.glMatrixLoadIdentityEXT.errcheck = _errcheck
	gl.glMatrixLoadIdentityEXT.restype = GLvoid
	gl.glMatrixLoadIdentityEXT.argtypes = (GLenum,)
	gl.glMatrixLoadTranspose3x3fNV = _getProcAddress('glMatrixLoadTranspose3x3fNV')
	gl.glMatrixLoadTranspose3x3fNV.gl = gl
	gl.glMatrixLoadTranspose3x3fNV.__name__ = 'glMatrixLoadTranspose3x3fNV'
	gl.glMatrixLoadTranspose3x3fNV.errcheck = _errcheck
	gl.glMatrixLoadTranspose3x3fNV.restype = GLvoid
	gl.glMatrixLoadTranspose3x3fNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixLoadTransposedEXT = _getProcAddress('glMatrixLoadTransposedEXT')
	gl.glMatrixLoadTransposedEXT.gl = gl
	gl.glMatrixLoadTransposedEXT.__name__ = 'glMatrixLoadTransposedEXT'
	gl.glMatrixLoadTransposedEXT.errcheck = _errcheck
	gl.glMatrixLoadTransposedEXT.restype = GLvoid
	gl.glMatrixLoadTransposedEXT.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMatrixLoadTransposefEXT = _getProcAddress('glMatrixLoadTransposefEXT')
	gl.glMatrixLoadTransposefEXT.gl = gl
	gl.glMatrixLoadTransposefEXT.__name__ = 'glMatrixLoadTransposefEXT'
	gl.glMatrixLoadTransposefEXT.errcheck = _errcheck
	gl.glMatrixLoadTransposefEXT.restype = GLvoid
	gl.glMatrixLoadTransposefEXT.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixLoaddEXT = _getProcAddress('glMatrixLoaddEXT')
	gl.glMatrixLoaddEXT.gl = gl
	gl.glMatrixLoaddEXT.__name__ = 'glMatrixLoaddEXT'
	gl.glMatrixLoaddEXT.errcheck = _errcheck
	gl.glMatrixLoaddEXT.restype = GLvoid
	gl.glMatrixLoaddEXT.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMatrixLoadfEXT = _getProcAddress('glMatrixLoadfEXT')
	gl.glMatrixLoadfEXT.gl = gl
	gl.glMatrixLoadfEXT.__name__ = 'glMatrixLoadfEXT'
	gl.glMatrixLoadfEXT.errcheck = _errcheck
	gl.glMatrixLoadfEXT.restype = GLvoid
	gl.glMatrixLoadfEXT.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixMode = _getProcAddress('glMatrixMode')
	gl.glMatrixMode.gl = gl
	gl.glMatrixMode.__name__ = 'glMatrixMode'
	gl.glMatrixMode.errcheck = _errcheck
	gl.glMatrixMode.restype = GLvoid
	gl.glMatrixMode.argtypes = (GLenum,)
	gl.glMatrixMult3x2fNV = _getProcAddress('glMatrixMult3x2fNV')
	gl.glMatrixMult3x2fNV.gl = gl
	gl.glMatrixMult3x2fNV.__name__ = 'glMatrixMult3x2fNV'
	gl.glMatrixMult3x2fNV.errcheck = _errcheck
	gl.glMatrixMult3x2fNV.restype = GLvoid
	gl.glMatrixMult3x2fNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixMult3x3fNV = _getProcAddress('glMatrixMult3x3fNV')
	gl.glMatrixMult3x3fNV.gl = gl
	gl.glMatrixMult3x3fNV.__name__ = 'glMatrixMult3x3fNV'
	gl.glMatrixMult3x3fNV.errcheck = _errcheck
	gl.glMatrixMult3x3fNV.restype = GLvoid
	gl.glMatrixMult3x3fNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixMultTranspose3x3fNV = _getProcAddress('glMatrixMultTranspose3x3fNV')
	gl.glMatrixMultTranspose3x3fNV.gl = gl
	gl.glMatrixMultTranspose3x3fNV.__name__ = 'glMatrixMultTranspose3x3fNV'
	gl.glMatrixMultTranspose3x3fNV.errcheck = _errcheck
	gl.glMatrixMultTranspose3x3fNV.restype = GLvoid
	gl.glMatrixMultTranspose3x3fNV.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixMultTransposedEXT = _getProcAddress('glMatrixMultTransposedEXT')
	gl.glMatrixMultTransposedEXT.gl = gl
	gl.glMatrixMultTransposedEXT.__name__ = 'glMatrixMultTransposedEXT'
	gl.glMatrixMultTransposedEXT.errcheck = _errcheck
	gl.glMatrixMultTransposedEXT.restype = GLvoid
	gl.glMatrixMultTransposedEXT.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMatrixMultTransposefEXT = _getProcAddress('glMatrixMultTransposefEXT')
	gl.glMatrixMultTransposefEXT.gl = gl
	gl.glMatrixMultTransposefEXT.__name__ = 'glMatrixMultTransposefEXT'
	gl.glMatrixMultTransposefEXT.errcheck = _errcheck
	gl.glMatrixMultTransposefEXT.restype = GLvoid
	gl.glMatrixMultTransposefEXT.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixMultdEXT = _getProcAddress('glMatrixMultdEXT')
	gl.glMatrixMultdEXT.gl = gl
	gl.glMatrixMultdEXT.__name__ = 'glMatrixMultdEXT'
	gl.glMatrixMultdEXT.errcheck = _errcheck
	gl.glMatrixMultdEXT.restype = GLvoid
	gl.glMatrixMultdEXT.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMatrixMultfEXT = _getProcAddress('glMatrixMultfEXT')
	gl.glMatrixMultfEXT.gl = gl
	gl.glMatrixMultfEXT.__name__ = 'glMatrixMultfEXT'
	gl.glMatrixMultfEXT.errcheck = _errcheck
	gl.glMatrixMultfEXT.restype = GLvoid
	gl.glMatrixMultfEXT.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMatrixOrthoEXT = _getProcAddress('glMatrixOrthoEXT')
	gl.glMatrixOrthoEXT.gl = gl
	gl.glMatrixOrthoEXT.__name__ = 'glMatrixOrthoEXT'
	gl.glMatrixOrthoEXT.errcheck = _errcheck
	gl.glMatrixOrthoEXT.restype = GLvoid
	gl.glMatrixOrthoEXT.argtypes = (GLenum, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glMatrixPopEXT = _getProcAddress('glMatrixPopEXT')
	gl.glMatrixPopEXT.gl = gl
	gl.glMatrixPopEXT.__name__ = 'glMatrixPopEXT'
	gl.glMatrixPopEXT.errcheck = _errcheck
	gl.glMatrixPopEXT.restype = GLvoid
	gl.glMatrixPopEXT.argtypes = (GLenum,)
	gl.glMatrixPushEXT = _getProcAddress('glMatrixPushEXT')
	gl.glMatrixPushEXT.gl = gl
	gl.glMatrixPushEXT.__name__ = 'glMatrixPushEXT'
	gl.glMatrixPushEXT.errcheck = _errcheck
	gl.glMatrixPushEXT.restype = GLvoid
	gl.glMatrixPushEXT.argtypes = (GLenum,)
	gl.glMatrixRotatedEXT = _getProcAddress('glMatrixRotatedEXT')
	gl.glMatrixRotatedEXT.gl = gl
	gl.glMatrixRotatedEXT.__name__ = 'glMatrixRotatedEXT'
	gl.glMatrixRotatedEXT.errcheck = _errcheck
	gl.glMatrixRotatedEXT.restype = GLvoid
	gl.glMatrixRotatedEXT.argtypes = (GLenum, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glMatrixRotatefEXT = _getProcAddress('glMatrixRotatefEXT')
	gl.glMatrixRotatefEXT.gl = gl
	gl.glMatrixRotatefEXT.__name__ = 'glMatrixRotatefEXT'
	gl.glMatrixRotatefEXT.errcheck = _errcheck
	gl.glMatrixRotatefEXT.restype = GLvoid
	gl.glMatrixRotatefEXT.argtypes = (GLenum, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glMatrixScaledEXT = _getProcAddress('glMatrixScaledEXT')
	gl.glMatrixScaledEXT.gl = gl
	gl.glMatrixScaledEXT.__name__ = 'glMatrixScaledEXT'
	gl.glMatrixScaledEXT.errcheck = _errcheck
	gl.glMatrixScaledEXT.restype = GLvoid
	gl.glMatrixScaledEXT.argtypes = (GLenum, GLdouble, GLdouble, GLdouble,)
	gl.glMatrixScalefEXT = _getProcAddress('glMatrixScalefEXT')
	gl.glMatrixScalefEXT.gl = gl
	gl.glMatrixScalefEXT.__name__ = 'glMatrixScalefEXT'
	gl.glMatrixScalefEXT.errcheck = _errcheck
	gl.glMatrixScalefEXT.restype = GLvoid
	gl.glMatrixScalefEXT.argtypes = (GLenum, GLfloat, GLfloat, GLfloat,)
	gl.glMatrixTranslatedEXT = _getProcAddress('glMatrixTranslatedEXT')
	gl.glMatrixTranslatedEXT.gl = gl
	gl.glMatrixTranslatedEXT.__name__ = 'glMatrixTranslatedEXT'
	gl.glMatrixTranslatedEXT.errcheck = _errcheck
	gl.glMatrixTranslatedEXT.restype = GLvoid
	gl.glMatrixTranslatedEXT.argtypes = (GLenum, GLdouble, GLdouble, GLdouble,)
	gl.glMatrixTranslatefEXT = _getProcAddress('glMatrixTranslatefEXT')
	gl.glMatrixTranslatefEXT.gl = gl
	gl.glMatrixTranslatefEXT.__name__ = 'glMatrixTranslatefEXT'
	gl.glMatrixTranslatefEXT.errcheck = _errcheck
	gl.glMatrixTranslatefEXT.restype = GLvoid
	gl.glMatrixTranslatefEXT.argtypes = (GLenum, GLfloat, GLfloat, GLfloat,)
	gl.glMemoryBarrier = _getProcAddress('glMemoryBarrier')
	gl.glMemoryBarrier.gl = gl
	gl.glMemoryBarrier.__name__ = 'glMemoryBarrier'
	gl.glMemoryBarrier.errcheck = _errcheck
	gl.glMemoryBarrier.restype = GLvoid
	gl.glMemoryBarrier.argtypes = (GLbitfield,)
	gl.glMemoryBarrierByRegion = _getProcAddress('glMemoryBarrierByRegion')
	gl.glMemoryBarrierByRegion.gl = gl
	gl.glMemoryBarrierByRegion.__name__ = 'glMemoryBarrierByRegion'
	gl.glMemoryBarrierByRegion.errcheck = _errcheck
	gl.glMemoryBarrierByRegion.restype = GLvoid
	gl.glMemoryBarrierByRegion.argtypes = (GLbitfield,)
	gl.glMemoryBarrierEXT = _getProcAddress('glMemoryBarrierEXT')
	gl.glMemoryBarrierEXT.gl = gl
	gl.glMemoryBarrierEXT.__name__ = 'glMemoryBarrierEXT'
	gl.glMemoryBarrierEXT.errcheck = _errcheck
	gl.glMemoryBarrierEXT.restype = GLvoid
	gl.glMemoryBarrierEXT.argtypes = (GLbitfield,)
	gl.glMinSampleShading = _getProcAddress('glMinSampleShading')
	gl.glMinSampleShading.gl = gl
	gl.glMinSampleShading.__name__ = 'glMinSampleShading'
	gl.glMinSampleShading.errcheck = _errcheck
	gl.glMinSampleShading.restype = GLvoid
	gl.glMinSampleShading.argtypes = (GLfloat,)
	gl.glMinSampleShadingARB = _getProcAddress('glMinSampleShadingARB')
	gl.glMinSampleShadingARB.gl = gl
	gl.glMinSampleShadingARB.__name__ = 'glMinSampleShadingARB'
	gl.glMinSampleShadingARB.errcheck = _errcheck
	gl.glMinSampleShadingARB.restype = GLvoid
	gl.glMinSampleShadingARB.argtypes = (GLfloat,)
	gl.glMinSampleShadingOES = _getProcAddress('glMinSampleShadingOES')
	gl.glMinSampleShadingOES.gl = gl
	gl.glMinSampleShadingOES.__name__ = 'glMinSampleShadingOES'
	gl.glMinSampleShadingOES.errcheck = _errcheck
	gl.glMinSampleShadingOES.restype = GLvoid
	gl.glMinSampleShadingOES.argtypes = (GLfloat,)
	gl.glMinmax = _getProcAddress('glMinmax')
	gl.glMinmax.gl = gl
	gl.glMinmax.__name__ = 'glMinmax'
	gl.glMinmax.errcheck = _errcheck
	gl.glMinmax.restype = GLvoid
	gl.glMinmax.argtypes = (GLenum, GLenum, GLboolean,)
	gl.glMinmaxEXT = _getProcAddress('glMinmaxEXT')
	gl.glMinmaxEXT.gl = gl
	gl.glMinmaxEXT.__name__ = 'glMinmaxEXT'
	gl.glMinmaxEXT.errcheck = _errcheck
	gl.glMinmaxEXT.restype = GLvoid
	gl.glMinmaxEXT.argtypes = (GLenum, GLenum, GLboolean,)
	gl.glMultMatrixd = _getProcAddress('glMultMatrixd')
	gl.glMultMatrixd.gl = gl
	gl.glMultMatrixd.__name__ = 'glMultMatrixd'
	gl.glMultMatrixd.errcheck = _errcheck
	gl.glMultMatrixd.restype = GLvoid
	gl.glMultMatrixd.argtypes = (POINTER(GLdouble),)
	gl.glMultMatrixf = _getProcAddress('glMultMatrixf')
	gl.glMultMatrixf.gl = gl
	gl.glMultMatrixf.__name__ = 'glMultMatrixf'
	gl.glMultMatrixf.errcheck = _errcheck
	gl.glMultMatrixf.restype = GLvoid
	gl.glMultMatrixf.argtypes = (POINTER(GLfloat),)
	gl.glMultMatrixx = _getProcAddress('glMultMatrixx')
	gl.glMultMatrixx.gl = gl
	gl.glMultMatrixx.__name__ = 'glMultMatrixx'
	gl.glMultMatrixx.errcheck = _errcheck
	gl.glMultMatrixx.restype = GLvoid
	gl.glMultMatrixx.argtypes = (POINTER(GLfixed),)
	gl.glMultMatrixxOES = _getProcAddress('glMultMatrixxOES')
	gl.glMultMatrixxOES.gl = gl
	gl.glMultMatrixxOES.__name__ = 'glMultMatrixxOES'
	gl.glMultMatrixxOES.errcheck = _errcheck
	gl.glMultMatrixxOES.restype = GLvoid
	gl.glMultMatrixxOES.argtypes = (POINTER(GLfixed),)
	gl.glMultTransposeMatrixd = _getProcAddress('glMultTransposeMatrixd')
	gl.glMultTransposeMatrixd.gl = gl
	gl.glMultTransposeMatrixd.__name__ = 'glMultTransposeMatrixd'
	gl.glMultTransposeMatrixd.errcheck = _errcheck
	gl.glMultTransposeMatrixd.restype = GLvoid
	gl.glMultTransposeMatrixd.argtypes = (POINTER(GLdouble),)
	gl.glMultTransposeMatrixdARB = _getProcAddress('glMultTransposeMatrixdARB')
	gl.glMultTransposeMatrixdARB.gl = gl
	gl.glMultTransposeMatrixdARB.__name__ = 'glMultTransposeMatrixdARB'
	gl.glMultTransposeMatrixdARB.errcheck = _errcheck
	gl.glMultTransposeMatrixdARB.restype = GLvoid
	gl.glMultTransposeMatrixdARB.argtypes = (POINTER(GLdouble),)
	gl.glMultTransposeMatrixf = _getProcAddress('glMultTransposeMatrixf')
	gl.glMultTransposeMatrixf.gl = gl
	gl.glMultTransposeMatrixf.__name__ = 'glMultTransposeMatrixf'
	gl.glMultTransposeMatrixf.errcheck = _errcheck
	gl.glMultTransposeMatrixf.restype = GLvoid
	gl.glMultTransposeMatrixf.argtypes = (POINTER(GLfloat),)
	gl.glMultTransposeMatrixfARB = _getProcAddress('glMultTransposeMatrixfARB')
	gl.glMultTransposeMatrixfARB.gl = gl
	gl.glMultTransposeMatrixfARB.__name__ = 'glMultTransposeMatrixfARB'
	gl.glMultTransposeMatrixfARB.errcheck = _errcheck
	gl.glMultTransposeMatrixfARB.restype = GLvoid
	gl.glMultTransposeMatrixfARB.argtypes = (POINTER(GLfloat),)
	gl.glMultTransposeMatrixxOES = _getProcAddress('glMultTransposeMatrixxOES')
	gl.glMultTransposeMatrixxOES.gl = gl
	gl.glMultTransposeMatrixxOES.__name__ = 'glMultTransposeMatrixxOES'
	gl.glMultTransposeMatrixxOES.errcheck = _errcheck
	gl.glMultTransposeMatrixxOES.restype = GLvoid
	gl.glMultTransposeMatrixxOES.argtypes = (POINTER(GLfixed),)
	gl.glMultiDrawArrays = _getProcAddress('glMultiDrawArrays')
	gl.glMultiDrawArrays.gl = gl
	gl.glMultiDrawArrays.__name__ = 'glMultiDrawArrays'
	gl.glMultiDrawArrays.errcheck = _errcheck
	gl.glMultiDrawArrays.restype = GLvoid
	gl.glMultiDrawArrays.argtypes = (GLenum, POINTER(GLint), POINTER(GLsizei), GLsizei,)
	gl.glMultiDrawArraysEXT = _getProcAddress('glMultiDrawArraysEXT')
	gl.glMultiDrawArraysEXT.gl = gl
	gl.glMultiDrawArraysEXT.__name__ = 'glMultiDrawArraysEXT'
	gl.glMultiDrawArraysEXT.errcheck = _errcheck
	gl.glMultiDrawArraysEXT.restype = GLvoid
	gl.glMultiDrawArraysEXT.argtypes = (GLenum, POINTER(GLint), POINTER(GLsizei), GLsizei,)
	gl.glMultiDrawArraysIndirect = _getProcAddress('glMultiDrawArraysIndirect')
	gl.glMultiDrawArraysIndirect.gl = gl
	gl.glMultiDrawArraysIndirect.__name__ = 'glMultiDrawArraysIndirect'
	gl.glMultiDrawArraysIndirect.errcheck = _errcheck
	gl.glMultiDrawArraysIndirect.restype = GLvoid
	gl.glMultiDrawArraysIndirect.argtypes = (GLenum, POINTER(GLvoid), GLsizei, GLsizei,)
	gl.glMultiDrawArraysIndirectAMD = _getProcAddress('glMultiDrawArraysIndirectAMD')
	gl.glMultiDrawArraysIndirectAMD.gl = gl
	gl.glMultiDrawArraysIndirectAMD.__name__ = 'glMultiDrawArraysIndirectAMD'
	gl.glMultiDrawArraysIndirectAMD.errcheck = _errcheck
	gl.glMultiDrawArraysIndirectAMD.restype = GLvoid
	gl.glMultiDrawArraysIndirectAMD.argtypes = (GLenum, POINTER(GLvoid), GLsizei, GLsizei,)
	gl.glMultiDrawArraysIndirectBindlessCountNV = _getProcAddress('glMultiDrawArraysIndirectBindlessCountNV')
	gl.glMultiDrawArraysIndirectBindlessCountNV.gl = gl
	gl.glMultiDrawArraysIndirectBindlessCountNV.__name__ = 'glMultiDrawArraysIndirectBindlessCountNV'
	gl.glMultiDrawArraysIndirectBindlessCountNV.errcheck = _errcheck
	gl.glMultiDrawArraysIndirectBindlessCountNV.restype = GLvoid
	gl.glMultiDrawArraysIndirectBindlessCountNV.argtypes = (GLenum, POINTER(GLvoid), GLsizei, GLsizei, GLsizei, GLint,)
	gl.glMultiDrawArraysIndirectBindlessNV = _getProcAddress('glMultiDrawArraysIndirectBindlessNV')
	gl.glMultiDrawArraysIndirectBindlessNV.gl = gl
	gl.glMultiDrawArraysIndirectBindlessNV.__name__ = 'glMultiDrawArraysIndirectBindlessNV'
	gl.glMultiDrawArraysIndirectBindlessNV.errcheck = _errcheck
	gl.glMultiDrawArraysIndirectBindlessNV.restype = GLvoid
	gl.glMultiDrawArraysIndirectBindlessNV.argtypes = (GLenum, POINTER(GLvoid), GLsizei, GLsizei, GLint,)
	gl.glMultiDrawArraysIndirectCountARB = _getProcAddress('glMultiDrawArraysIndirectCountARB')
	gl.glMultiDrawArraysIndirectCountARB.gl = gl
	gl.glMultiDrawArraysIndirectCountARB.__name__ = 'glMultiDrawArraysIndirectCountARB'
	gl.glMultiDrawArraysIndirectCountARB.errcheck = _errcheck
	gl.glMultiDrawArraysIndirectCountARB.restype = GLvoid
	gl.glMultiDrawArraysIndirectCountARB.argtypes = (GLenum, GLintptr, GLintptr, GLsizei, GLsizei,)
	gl.glMultiDrawElementArrayAPPLE = _getProcAddress('glMultiDrawElementArrayAPPLE')
	gl.glMultiDrawElementArrayAPPLE.gl = gl
	gl.glMultiDrawElementArrayAPPLE.__name__ = 'glMultiDrawElementArrayAPPLE'
	gl.glMultiDrawElementArrayAPPLE.errcheck = _errcheck
	gl.glMultiDrawElementArrayAPPLE.restype = GLvoid
	gl.glMultiDrawElementArrayAPPLE.argtypes = (GLenum, POINTER(GLint), POINTER(GLsizei), GLsizei,)
	gl.glMultiDrawElements = _getProcAddress('glMultiDrawElements')
	gl.glMultiDrawElements.gl = gl
	gl.glMultiDrawElements.__name__ = 'glMultiDrawElements'
	gl.glMultiDrawElements.errcheck = _errcheck
	gl.glMultiDrawElements.restype = GLvoid
	gl.glMultiDrawElements.argtypes = (GLenum, POINTER(GLsizei), GLenum, POINTER(POINTER(GLvoid)), GLsizei,)
	gl.glMultiDrawElementsBaseVertex = _getProcAddress('glMultiDrawElementsBaseVertex')
	gl.glMultiDrawElementsBaseVertex.gl = gl
	gl.glMultiDrawElementsBaseVertex.__name__ = 'glMultiDrawElementsBaseVertex'
	gl.glMultiDrawElementsBaseVertex.errcheck = _errcheck
	gl.glMultiDrawElementsBaseVertex.restype = GLvoid
	gl.glMultiDrawElementsBaseVertex.argtypes = (GLenum, POINTER(GLsizei), GLenum, POINTER(POINTER(GLvoid)), GLsizei, POINTER(GLint),)
	gl.glMultiDrawElementsEXT = _getProcAddress('glMultiDrawElementsEXT')
	gl.glMultiDrawElementsEXT.gl = gl
	gl.glMultiDrawElementsEXT.__name__ = 'glMultiDrawElementsEXT'
	gl.glMultiDrawElementsEXT.errcheck = _errcheck
	gl.glMultiDrawElementsEXT.restype = GLvoid
	gl.glMultiDrawElementsEXT.argtypes = (GLenum, POINTER(GLsizei), GLenum, POINTER(POINTER(GLvoid)), GLsizei,)
	gl.glMultiDrawElementsIndirect = _getProcAddress('glMultiDrawElementsIndirect')
	gl.glMultiDrawElementsIndirect.gl = gl
	gl.glMultiDrawElementsIndirect.__name__ = 'glMultiDrawElementsIndirect'
	gl.glMultiDrawElementsIndirect.errcheck = _errcheck
	gl.glMultiDrawElementsIndirect.restype = GLvoid
	gl.glMultiDrawElementsIndirect.argtypes = (GLenum, GLenum, POINTER(GLvoid), GLsizei, GLsizei,)
	gl.glMultiDrawElementsIndirectAMD = _getProcAddress('glMultiDrawElementsIndirectAMD')
	gl.glMultiDrawElementsIndirectAMD.gl = gl
	gl.glMultiDrawElementsIndirectAMD.__name__ = 'glMultiDrawElementsIndirectAMD'
	gl.glMultiDrawElementsIndirectAMD.errcheck = _errcheck
	gl.glMultiDrawElementsIndirectAMD.restype = GLvoid
	gl.glMultiDrawElementsIndirectAMD.argtypes = (GLenum, GLenum, POINTER(GLvoid), GLsizei, GLsizei,)
	gl.glMultiDrawElementsIndirectBindlessCountNV = _getProcAddress('glMultiDrawElementsIndirectBindlessCountNV')
	gl.glMultiDrawElementsIndirectBindlessCountNV.gl = gl
	gl.glMultiDrawElementsIndirectBindlessCountNV.__name__ = 'glMultiDrawElementsIndirectBindlessCountNV'
	gl.glMultiDrawElementsIndirectBindlessCountNV.errcheck = _errcheck
	gl.glMultiDrawElementsIndirectBindlessCountNV.restype = GLvoid
	gl.glMultiDrawElementsIndirectBindlessCountNV.argtypes = (GLenum, GLenum, POINTER(GLvoid), GLsizei, GLsizei, GLsizei, GLint,)
	gl.glMultiDrawElementsIndirectBindlessNV = _getProcAddress('glMultiDrawElementsIndirectBindlessNV')
	gl.glMultiDrawElementsIndirectBindlessNV.gl = gl
	gl.glMultiDrawElementsIndirectBindlessNV.__name__ = 'glMultiDrawElementsIndirectBindlessNV'
	gl.glMultiDrawElementsIndirectBindlessNV.errcheck = _errcheck
	gl.glMultiDrawElementsIndirectBindlessNV.restype = GLvoid
	gl.glMultiDrawElementsIndirectBindlessNV.argtypes = (GLenum, GLenum, POINTER(GLvoid), GLsizei, GLsizei, GLint,)
	gl.glMultiDrawElementsIndirectCountARB = _getProcAddress('glMultiDrawElementsIndirectCountARB')
	gl.glMultiDrawElementsIndirectCountARB.gl = gl
	gl.glMultiDrawElementsIndirectCountARB.__name__ = 'glMultiDrawElementsIndirectCountARB'
	gl.glMultiDrawElementsIndirectCountARB.errcheck = _errcheck
	gl.glMultiDrawElementsIndirectCountARB.restype = GLvoid
	gl.glMultiDrawElementsIndirectCountARB.argtypes = (GLenum, GLenum, GLintptr, GLintptr, GLsizei, GLsizei,)
	gl.glMultiDrawRangeElementArrayAPPLE = _getProcAddress('glMultiDrawRangeElementArrayAPPLE')
	gl.glMultiDrawRangeElementArrayAPPLE.gl = gl
	gl.glMultiDrawRangeElementArrayAPPLE.__name__ = 'glMultiDrawRangeElementArrayAPPLE'
	gl.glMultiDrawRangeElementArrayAPPLE.errcheck = _errcheck
	gl.glMultiDrawRangeElementArrayAPPLE.restype = GLvoid
	gl.glMultiDrawRangeElementArrayAPPLE.argtypes = (GLenum, GLuint, GLuint, POINTER(GLint), POINTER(GLsizei), GLsizei,)
	gl.glMultiModeDrawArraysIBM = _getProcAddress('glMultiModeDrawArraysIBM')
	gl.glMultiModeDrawArraysIBM.gl = gl
	gl.glMultiModeDrawArraysIBM.__name__ = 'glMultiModeDrawArraysIBM'
	gl.glMultiModeDrawArraysIBM.errcheck = _errcheck
	gl.glMultiModeDrawArraysIBM.restype = GLvoid
	gl.glMultiModeDrawArraysIBM.argtypes = (POINTER(GLenum), POINTER(GLint), POINTER(GLsizei), GLsizei, GLint,)
	gl.glMultiModeDrawElementsIBM = _getProcAddress('glMultiModeDrawElementsIBM')
	gl.glMultiModeDrawElementsIBM.gl = gl
	gl.glMultiModeDrawElementsIBM.__name__ = 'glMultiModeDrawElementsIBM'
	gl.glMultiModeDrawElementsIBM.errcheck = _errcheck
	gl.glMultiModeDrawElementsIBM.restype = GLvoid
	gl.glMultiModeDrawElementsIBM.argtypes = (POINTER(GLenum), POINTER(GLsizei), GLenum, POINTER(POINTER(GLvoid)), GLsizei, GLint,)
	gl.glMultiTexBufferEXT = _getProcAddress('glMultiTexBufferEXT')
	gl.glMultiTexBufferEXT.gl = gl
	gl.glMultiTexBufferEXT.__name__ = 'glMultiTexBufferEXT'
	gl.glMultiTexBufferEXT.errcheck = _errcheck
	gl.glMultiTexBufferEXT.restype = GLvoid
	gl.glMultiTexBufferEXT.argtypes = (GLenum, GLenum, GLenum, GLuint,)
	gl.glMultiTexCoord1bOES = _getProcAddress('glMultiTexCoord1bOES')
	gl.glMultiTexCoord1bOES.gl = gl
	gl.glMultiTexCoord1bOES.__name__ = 'glMultiTexCoord1bOES'
	gl.glMultiTexCoord1bOES.errcheck = _errcheck
	gl.glMultiTexCoord1bOES.restype = GLvoid
	gl.glMultiTexCoord1bOES.argtypes = (GLenum, GLbyte,)
	gl.glMultiTexCoord1bvOES = _getProcAddress('glMultiTexCoord1bvOES')
	gl.glMultiTexCoord1bvOES.gl = gl
	gl.glMultiTexCoord1bvOES.__name__ = 'glMultiTexCoord1bvOES'
	gl.glMultiTexCoord1bvOES.errcheck = _errcheck
	gl.glMultiTexCoord1bvOES.restype = GLvoid
	gl.glMultiTexCoord1bvOES.argtypes = (GLenum, POINTER(GLbyte),)
	gl.glMultiTexCoord1d = _getProcAddress('glMultiTexCoord1d')
	gl.glMultiTexCoord1d.gl = gl
	gl.glMultiTexCoord1d.__name__ = 'glMultiTexCoord1d'
	gl.glMultiTexCoord1d.errcheck = _errcheck
	gl.glMultiTexCoord1d.restype = GLvoid
	gl.glMultiTexCoord1d.argtypes = (GLenum, GLdouble,)
	gl.glMultiTexCoord1dARB = _getProcAddress('glMultiTexCoord1dARB')
	gl.glMultiTexCoord1dARB.gl = gl
	gl.glMultiTexCoord1dARB.__name__ = 'glMultiTexCoord1dARB'
	gl.glMultiTexCoord1dARB.errcheck = _errcheck
	gl.glMultiTexCoord1dARB.restype = GLvoid
	gl.glMultiTexCoord1dARB.argtypes = (GLenum, GLdouble,)
	gl.glMultiTexCoord1dv = _getProcAddress('glMultiTexCoord1dv')
	gl.glMultiTexCoord1dv.gl = gl
	gl.glMultiTexCoord1dv.__name__ = 'glMultiTexCoord1dv'
	gl.glMultiTexCoord1dv.errcheck = _errcheck
	gl.glMultiTexCoord1dv.restype = GLvoid
	gl.glMultiTexCoord1dv.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord1dvARB = _getProcAddress('glMultiTexCoord1dvARB')
	gl.glMultiTexCoord1dvARB.gl = gl
	gl.glMultiTexCoord1dvARB.__name__ = 'glMultiTexCoord1dvARB'
	gl.glMultiTexCoord1dvARB.errcheck = _errcheck
	gl.glMultiTexCoord1dvARB.restype = GLvoid
	gl.glMultiTexCoord1dvARB.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord1f = _getProcAddress('glMultiTexCoord1f')
	gl.glMultiTexCoord1f.gl = gl
	gl.glMultiTexCoord1f.__name__ = 'glMultiTexCoord1f'
	gl.glMultiTexCoord1f.errcheck = _errcheck
	gl.glMultiTexCoord1f.restype = GLvoid
	gl.glMultiTexCoord1f.argtypes = (GLenum, GLfloat,)
	gl.glMultiTexCoord1fARB = _getProcAddress('glMultiTexCoord1fARB')
	gl.glMultiTexCoord1fARB.gl = gl
	gl.glMultiTexCoord1fARB.__name__ = 'glMultiTexCoord1fARB'
	gl.glMultiTexCoord1fARB.errcheck = _errcheck
	gl.glMultiTexCoord1fARB.restype = GLvoid
	gl.glMultiTexCoord1fARB.argtypes = (GLenum, GLfloat,)
	gl.glMultiTexCoord1fv = _getProcAddress('glMultiTexCoord1fv')
	gl.glMultiTexCoord1fv.gl = gl
	gl.glMultiTexCoord1fv.__name__ = 'glMultiTexCoord1fv'
	gl.glMultiTexCoord1fv.errcheck = _errcheck
	gl.glMultiTexCoord1fv.restype = GLvoid
	gl.glMultiTexCoord1fv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord1fvARB = _getProcAddress('glMultiTexCoord1fvARB')
	gl.glMultiTexCoord1fvARB.gl = gl
	gl.glMultiTexCoord1fvARB.__name__ = 'glMultiTexCoord1fvARB'
	gl.glMultiTexCoord1fvARB.errcheck = _errcheck
	gl.glMultiTexCoord1fvARB.restype = GLvoid
	gl.glMultiTexCoord1fvARB.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord1hNV = _getProcAddress('glMultiTexCoord1hNV')
	gl.glMultiTexCoord1hNV.gl = gl
	gl.glMultiTexCoord1hNV.__name__ = 'glMultiTexCoord1hNV'
	gl.glMultiTexCoord1hNV.errcheck = _errcheck
	gl.glMultiTexCoord1hNV.restype = GLvoid
	gl.glMultiTexCoord1hNV.argtypes = (GLenum, GLhalfNV,)
	gl.glMultiTexCoord1hvNV = _getProcAddress('glMultiTexCoord1hvNV')
	gl.glMultiTexCoord1hvNV.gl = gl
	gl.glMultiTexCoord1hvNV.__name__ = 'glMultiTexCoord1hvNV'
	gl.glMultiTexCoord1hvNV.errcheck = _errcheck
	gl.glMultiTexCoord1hvNV.restype = GLvoid
	gl.glMultiTexCoord1hvNV.argtypes = (GLenum, POINTER(GLhalfNV),)
	gl.glMultiTexCoord1i = _getProcAddress('glMultiTexCoord1i')
	gl.glMultiTexCoord1i.gl = gl
	gl.glMultiTexCoord1i.__name__ = 'glMultiTexCoord1i'
	gl.glMultiTexCoord1i.errcheck = _errcheck
	gl.glMultiTexCoord1i.restype = GLvoid
	gl.glMultiTexCoord1i.argtypes = (GLenum, GLint,)
	gl.glMultiTexCoord1iARB = _getProcAddress('glMultiTexCoord1iARB')
	gl.glMultiTexCoord1iARB.gl = gl
	gl.glMultiTexCoord1iARB.__name__ = 'glMultiTexCoord1iARB'
	gl.glMultiTexCoord1iARB.errcheck = _errcheck
	gl.glMultiTexCoord1iARB.restype = GLvoid
	gl.glMultiTexCoord1iARB.argtypes = (GLenum, GLint,)
	gl.glMultiTexCoord1iv = _getProcAddress('glMultiTexCoord1iv')
	gl.glMultiTexCoord1iv.gl = gl
	gl.glMultiTexCoord1iv.__name__ = 'glMultiTexCoord1iv'
	gl.glMultiTexCoord1iv.errcheck = _errcheck
	gl.glMultiTexCoord1iv.restype = GLvoid
	gl.glMultiTexCoord1iv.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord1ivARB = _getProcAddress('glMultiTexCoord1ivARB')
	gl.glMultiTexCoord1ivARB.gl = gl
	gl.glMultiTexCoord1ivARB.__name__ = 'glMultiTexCoord1ivARB'
	gl.glMultiTexCoord1ivARB.errcheck = _errcheck
	gl.glMultiTexCoord1ivARB.restype = GLvoid
	gl.glMultiTexCoord1ivARB.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord1s = _getProcAddress('glMultiTexCoord1s')
	gl.glMultiTexCoord1s.gl = gl
	gl.glMultiTexCoord1s.__name__ = 'glMultiTexCoord1s'
	gl.glMultiTexCoord1s.errcheck = _errcheck
	gl.glMultiTexCoord1s.restype = GLvoid
	gl.glMultiTexCoord1s.argtypes = (GLenum, GLshort,)
	gl.glMultiTexCoord1sARB = _getProcAddress('glMultiTexCoord1sARB')
	gl.glMultiTexCoord1sARB.gl = gl
	gl.glMultiTexCoord1sARB.__name__ = 'glMultiTexCoord1sARB'
	gl.glMultiTexCoord1sARB.errcheck = _errcheck
	gl.glMultiTexCoord1sARB.restype = GLvoid
	gl.glMultiTexCoord1sARB.argtypes = (GLenum, GLshort,)
	gl.glMultiTexCoord1sv = _getProcAddress('glMultiTexCoord1sv')
	gl.glMultiTexCoord1sv.gl = gl
	gl.glMultiTexCoord1sv.__name__ = 'glMultiTexCoord1sv'
	gl.glMultiTexCoord1sv.errcheck = _errcheck
	gl.glMultiTexCoord1sv.restype = GLvoid
	gl.glMultiTexCoord1sv.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord1svARB = _getProcAddress('glMultiTexCoord1svARB')
	gl.glMultiTexCoord1svARB.gl = gl
	gl.glMultiTexCoord1svARB.__name__ = 'glMultiTexCoord1svARB'
	gl.glMultiTexCoord1svARB.errcheck = _errcheck
	gl.glMultiTexCoord1svARB.restype = GLvoid
	gl.glMultiTexCoord1svARB.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord1xOES = _getProcAddress('glMultiTexCoord1xOES')
	gl.glMultiTexCoord1xOES.gl = gl
	gl.glMultiTexCoord1xOES.__name__ = 'glMultiTexCoord1xOES'
	gl.glMultiTexCoord1xOES.errcheck = _errcheck
	gl.glMultiTexCoord1xOES.restype = GLvoid
	gl.glMultiTexCoord1xOES.argtypes = (GLenum, GLfixed,)
	gl.glMultiTexCoord1xvOES = _getProcAddress('glMultiTexCoord1xvOES')
	gl.glMultiTexCoord1xvOES.gl = gl
	gl.glMultiTexCoord1xvOES.__name__ = 'glMultiTexCoord1xvOES'
	gl.glMultiTexCoord1xvOES.errcheck = _errcheck
	gl.glMultiTexCoord1xvOES.restype = GLvoid
	gl.glMultiTexCoord1xvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glMultiTexCoord2bOES = _getProcAddress('glMultiTexCoord2bOES')
	gl.glMultiTexCoord2bOES.gl = gl
	gl.glMultiTexCoord2bOES.__name__ = 'glMultiTexCoord2bOES'
	gl.glMultiTexCoord2bOES.errcheck = _errcheck
	gl.glMultiTexCoord2bOES.restype = GLvoid
	gl.glMultiTexCoord2bOES.argtypes = (GLenum, GLbyte, GLbyte,)
	gl.glMultiTexCoord2bvOES = _getProcAddress('glMultiTexCoord2bvOES')
	gl.glMultiTexCoord2bvOES.gl = gl
	gl.glMultiTexCoord2bvOES.__name__ = 'glMultiTexCoord2bvOES'
	gl.glMultiTexCoord2bvOES.errcheck = _errcheck
	gl.glMultiTexCoord2bvOES.restype = GLvoid
	gl.glMultiTexCoord2bvOES.argtypes = (GLenum, POINTER(GLbyte),)
	gl.glMultiTexCoord2d = _getProcAddress('glMultiTexCoord2d')
	gl.glMultiTexCoord2d.gl = gl
	gl.glMultiTexCoord2d.__name__ = 'glMultiTexCoord2d'
	gl.glMultiTexCoord2d.errcheck = _errcheck
	gl.glMultiTexCoord2d.restype = GLvoid
	gl.glMultiTexCoord2d.argtypes = (GLenum, GLdouble, GLdouble,)
	gl.glMultiTexCoord2dARB = _getProcAddress('glMultiTexCoord2dARB')
	gl.glMultiTexCoord2dARB.gl = gl
	gl.glMultiTexCoord2dARB.__name__ = 'glMultiTexCoord2dARB'
	gl.glMultiTexCoord2dARB.errcheck = _errcheck
	gl.glMultiTexCoord2dARB.restype = GLvoid
	gl.glMultiTexCoord2dARB.argtypes = (GLenum, GLdouble, GLdouble,)
	gl.glMultiTexCoord2dv = _getProcAddress('glMultiTexCoord2dv')
	gl.glMultiTexCoord2dv.gl = gl
	gl.glMultiTexCoord2dv.__name__ = 'glMultiTexCoord2dv'
	gl.glMultiTexCoord2dv.errcheck = _errcheck
	gl.glMultiTexCoord2dv.restype = GLvoid
	gl.glMultiTexCoord2dv.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord2dvARB = _getProcAddress('glMultiTexCoord2dvARB')
	gl.glMultiTexCoord2dvARB.gl = gl
	gl.glMultiTexCoord2dvARB.__name__ = 'glMultiTexCoord2dvARB'
	gl.glMultiTexCoord2dvARB.errcheck = _errcheck
	gl.glMultiTexCoord2dvARB.restype = GLvoid
	gl.glMultiTexCoord2dvARB.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord2f = _getProcAddress('glMultiTexCoord2f')
	gl.glMultiTexCoord2f.gl = gl
	gl.glMultiTexCoord2f.__name__ = 'glMultiTexCoord2f'
	gl.glMultiTexCoord2f.errcheck = _errcheck
	gl.glMultiTexCoord2f.restype = GLvoid
	gl.glMultiTexCoord2f.argtypes = (GLenum, GLfloat, GLfloat,)
	gl.glMultiTexCoord2fARB = _getProcAddress('glMultiTexCoord2fARB')
	gl.glMultiTexCoord2fARB.gl = gl
	gl.glMultiTexCoord2fARB.__name__ = 'glMultiTexCoord2fARB'
	gl.glMultiTexCoord2fARB.errcheck = _errcheck
	gl.glMultiTexCoord2fARB.restype = GLvoid
	gl.glMultiTexCoord2fARB.argtypes = (GLenum, GLfloat, GLfloat,)
	gl.glMultiTexCoord2fv = _getProcAddress('glMultiTexCoord2fv')
	gl.glMultiTexCoord2fv.gl = gl
	gl.glMultiTexCoord2fv.__name__ = 'glMultiTexCoord2fv'
	gl.glMultiTexCoord2fv.errcheck = _errcheck
	gl.glMultiTexCoord2fv.restype = GLvoid
	gl.glMultiTexCoord2fv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord2fvARB = _getProcAddress('glMultiTexCoord2fvARB')
	gl.glMultiTexCoord2fvARB.gl = gl
	gl.glMultiTexCoord2fvARB.__name__ = 'glMultiTexCoord2fvARB'
	gl.glMultiTexCoord2fvARB.errcheck = _errcheck
	gl.glMultiTexCoord2fvARB.restype = GLvoid
	gl.glMultiTexCoord2fvARB.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord2hNV = _getProcAddress('glMultiTexCoord2hNV')
	gl.glMultiTexCoord2hNV.gl = gl
	gl.glMultiTexCoord2hNV.__name__ = 'glMultiTexCoord2hNV'
	gl.glMultiTexCoord2hNV.errcheck = _errcheck
	gl.glMultiTexCoord2hNV.restype = GLvoid
	gl.glMultiTexCoord2hNV.argtypes = (GLenum, GLhalfNV, GLhalfNV,)
	gl.glMultiTexCoord2hvNV = _getProcAddress('glMultiTexCoord2hvNV')
	gl.glMultiTexCoord2hvNV.gl = gl
	gl.glMultiTexCoord2hvNV.__name__ = 'glMultiTexCoord2hvNV'
	gl.glMultiTexCoord2hvNV.errcheck = _errcheck
	gl.glMultiTexCoord2hvNV.restype = GLvoid
	gl.glMultiTexCoord2hvNV.argtypes = (GLenum, POINTER(GLhalfNV),)
	gl.glMultiTexCoord2i = _getProcAddress('glMultiTexCoord2i')
	gl.glMultiTexCoord2i.gl = gl
	gl.glMultiTexCoord2i.__name__ = 'glMultiTexCoord2i'
	gl.glMultiTexCoord2i.errcheck = _errcheck
	gl.glMultiTexCoord2i.restype = GLvoid
	gl.glMultiTexCoord2i.argtypes = (GLenum, GLint, GLint,)
	gl.glMultiTexCoord2iARB = _getProcAddress('glMultiTexCoord2iARB')
	gl.glMultiTexCoord2iARB.gl = gl
	gl.glMultiTexCoord2iARB.__name__ = 'glMultiTexCoord2iARB'
	gl.glMultiTexCoord2iARB.errcheck = _errcheck
	gl.glMultiTexCoord2iARB.restype = GLvoid
	gl.glMultiTexCoord2iARB.argtypes = (GLenum, GLint, GLint,)
	gl.glMultiTexCoord2iv = _getProcAddress('glMultiTexCoord2iv')
	gl.glMultiTexCoord2iv.gl = gl
	gl.glMultiTexCoord2iv.__name__ = 'glMultiTexCoord2iv'
	gl.glMultiTexCoord2iv.errcheck = _errcheck
	gl.glMultiTexCoord2iv.restype = GLvoid
	gl.glMultiTexCoord2iv.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord2ivARB = _getProcAddress('glMultiTexCoord2ivARB')
	gl.glMultiTexCoord2ivARB.gl = gl
	gl.glMultiTexCoord2ivARB.__name__ = 'glMultiTexCoord2ivARB'
	gl.glMultiTexCoord2ivARB.errcheck = _errcheck
	gl.glMultiTexCoord2ivARB.restype = GLvoid
	gl.glMultiTexCoord2ivARB.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord2s = _getProcAddress('glMultiTexCoord2s')
	gl.glMultiTexCoord2s.gl = gl
	gl.glMultiTexCoord2s.__name__ = 'glMultiTexCoord2s'
	gl.glMultiTexCoord2s.errcheck = _errcheck
	gl.glMultiTexCoord2s.restype = GLvoid
	gl.glMultiTexCoord2s.argtypes = (GLenum, GLshort, GLshort,)
	gl.glMultiTexCoord2sARB = _getProcAddress('glMultiTexCoord2sARB')
	gl.glMultiTexCoord2sARB.gl = gl
	gl.glMultiTexCoord2sARB.__name__ = 'glMultiTexCoord2sARB'
	gl.glMultiTexCoord2sARB.errcheck = _errcheck
	gl.glMultiTexCoord2sARB.restype = GLvoid
	gl.glMultiTexCoord2sARB.argtypes = (GLenum, GLshort, GLshort,)
	gl.glMultiTexCoord2sv = _getProcAddress('glMultiTexCoord2sv')
	gl.glMultiTexCoord2sv.gl = gl
	gl.glMultiTexCoord2sv.__name__ = 'glMultiTexCoord2sv'
	gl.glMultiTexCoord2sv.errcheck = _errcheck
	gl.glMultiTexCoord2sv.restype = GLvoid
	gl.glMultiTexCoord2sv.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord2svARB = _getProcAddress('glMultiTexCoord2svARB')
	gl.glMultiTexCoord2svARB.gl = gl
	gl.glMultiTexCoord2svARB.__name__ = 'glMultiTexCoord2svARB'
	gl.glMultiTexCoord2svARB.errcheck = _errcheck
	gl.glMultiTexCoord2svARB.restype = GLvoid
	gl.glMultiTexCoord2svARB.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord2xOES = _getProcAddress('glMultiTexCoord2xOES')
	gl.glMultiTexCoord2xOES.gl = gl
	gl.glMultiTexCoord2xOES.__name__ = 'glMultiTexCoord2xOES'
	gl.glMultiTexCoord2xOES.errcheck = _errcheck
	gl.glMultiTexCoord2xOES.restype = GLvoid
	gl.glMultiTexCoord2xOES.argtypes = (GLenum, GLfixed, GLfixed,)
	gl.glMultiTexCoord2xvOES = _getProcAddress('glMultiTexCoord2xvOES')
	gl.glMultiTexCoord2xvOES.gl = gl
	gl.glMultiTexCoord2xvOES.__name__ = 'glMultiTexCoord2xvOES'
	gl.glMultiTexCoord2xvOES.errcheck = _errcheck
	gl.glMultiTexCoord2xvOES.restype = GLvoid
	gl.glMultiTexCoord2xvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glMultiTexCoord3bOES = _getProcAddress('glMultiTexCoord3bOES')
	gl.glMultiTexCoord3bOES.gl = gl
	gl.glMultiTexCoord3bOES.__name__ = 'glMultiTexCoord3bOES'
	gl.glMultiTexCoord3bOES.errcheck = _errcheck
	gl.glMultiTexCoord3bOES.restype = GLvoid
	gl.glMultiTexCoord3bOES.argtypes = (GLenum, GLbyte, GLbyte, GLbyte,)
	gl.glMultiTexCoord3bvOES = _getProcAddress('glMultiTexCoord3bvOES')
	gl.glMultiTexCoord3bvOES.gl = gl
	gl.glMultiTexCoord3bvOES.__name__ = 'glMultiTexCoord3bvOES'
	gl.glMultiTexCoord3bvOES.errcheck = _errcheck
	gl.glMultiTexCoord3bvOES.restype = GLvoid
	gl.glMultiTexCoord3bvOES.argtypes = (GLenum, POINTER(GLbyte),)
	gl.glMultiTexCoord3d = _getProcAddress('glMultiTexCoord3d')
	gl.glMultiTexCoord3d.gl = gl
	gl.glMultiTexCoord3d.__name__ = 'glMultiTexCoord3d'
	gl.glMultiTexCoord3d.errcheck = _errcheck
	gl.glMultiTexCoord3d.restype = GLvoid
	gl.glMultiTexCoord3d.argtypes = (GLenum, GLdouble, GLdouble, GLdouble,)
	gl.glMultiTexCoord3dARB = _getProcAddress('glMultiTexCoord3dARB')
	gl.glMultiTexCoord3dARB.gl = gl
	gl.glMultiTexCoord3dARB.__name__ = 'glMultiTexCoord3dARB'
	gl.glMultiTexCoord3dARB.errcheck = _errcheck
	gl.glMultiTexCoord3dARB.restype = GLvoid
	gl.glMultiTexCoord3dARB.argtypes = (GLenum, GLdouble, GLdouble, GLdouble,)
	gl.glMultiTexCoord3dv = _getProcAddress('glMultiTexCoord3dv')
	gl.glMultiTexCoord3dv.gl = gl
	gl.glMultiTexCoord3dv.__name__ = 'glMultiTexCoord3dv'
	gl.glMultiTexCoord3dv.errcheck = _errcheck
	gl.glMultiTexCoord3dv.restype = GLvoid
	gl.glMultiTexCoord3dv.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord3dvARB = _getProcAddress('glMultiTexCoord3dvARB')
	gl.glMultiTexCoord3dvARB.gl = gl
	gl.glMultiTexCoord3dvARB.__name__ = 'glMultiTexCoord3dvARB'
	gl.glMultiTexCoord3dvARB.errcheck = _errcheck
	gl.glMultiTexCoord3dvARB.restype = GLvoid
	gl.glMultiTexCoord3dvARB.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord3f = _getProcAddress('glMultiTexCoord3f')
	gl.glMultiTexCoord3f.gl = gl
	gl.glMultiTexCoord3f.__name__ = 'glMultiTexCoord3f'
	gl.glMultiTexCoord3f.errcheck = _errcheck
	gl.glMultiTexCoord3f.restype = GLvoid
	gl.glMultiTexCoord3f.argtypes = (GLenum, GLfloat, GLfloat, GLfloat,)
	gl.glMultiTexCoord3fARB = _getProcAddress('glMultiTexCoord3fARB')
	gl.glMultiTexCoord3fARB.gl = gl
	gl.glMultiTexCoord3fARB.__name__ = 'glMultiTexCoord3fARB'
	gl.glMultiTexCoord3fARB.errcheck = _errcheck
	gl.glMultiTexCoord3fARB.restype = GLvoid
	gl.glMultiTexCoord3fARB.argtypes = (GLenum, GLfloat, GLfloat, GLfloat,)
	gl.glMultiTexCoord3fv = _getProcAddress('glMultiTexCoord3fv')
	gl.glMultiTexCoord3fv.gl = gl
	gl.glMultiTexCoord3fv.__name__ = 'glMultiTexCoord3fv'
	gl.glMultiTexCoord3fv.errcheck = _errcheck
	gl.glMultiTexCoord3fv.restype = GLvoid
	gl.glMultiTexCoord3fv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord3fvARB = _getProcAddress('glMultiTexCoord3fvARB')
	gl.glMultiTexCoord3fvARB.gl = gl
	gl.glMultiTexCoord3fvARB.__name__ = 'glMultiTexCoord3fvARB'
	gl.glMultiTexCoord3fvARB.errcheck = _errcheck
	gl.glMultiTexCoord3fvARB.restype = GLvoid
	gl.glMultiTexCoord3fvARB.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord3hNV = _getProcAddress('glMultiTexCoord3hNV')
	gl.glMultiTexCoord3hNV.gl = gl
	gl.glMultiTexCoord3hNV.__name__ = 'glMultiTexCoord3hNV'
	gl.glMultiTexCoord3hNV.errcheck = _errcheck
	gl.glMultiTexCoord3hNV.restype = GLvoid
	gl.glMultiTexCoord3hNV.argtypes = (GLenum, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glMultiTexCoord3hvNV = _getProcAddress('glMultiTexCoord3hvNV')
	gl.glMultiTexCoord3hvNV.gl = gl
	gl.glMultiTexCoord3hvNV.__name__ = 'glMultiTexCoord3hvNV'
	gl.glMultiTexCoord3hvNV.errcheck = _errcheck
	gl.glMultiTexCoord3hvNV.restype = GLvoid
	gl.glMultiTexCoord3hvNV.argtypes = (GLenum, POINTER(GLhalfNV),)
	gl.glMultiTexCoord3i = _getProcAddress('glMultiTexCoord3i')
	gl.glMultiTexCoord3i.gl = gl
	gl.glMultiTexCoord3i.__name__ = 'glMultiTexCoord3i'
	gl.glMultiTexCoord3i.errcheck = _errcheck
	gl.glMultiTexCoord3i.restype = GLvoid
	gl.glMultiTexCoord3i.argtypes = (GLenum, GLint, GLint, GLint,)
	gl.glMultiTexCoord3iARB = _getProcAddress('glMultiTexCoord3iARB')
	gl.glMultiTexCoord3iARB.gl = gl
	gl.glMultiTexCoord3iARB.__name__ = 'glMultiTexCoord3iARB'
	gl.glMultiTexCoord3iARB.errcheck = _errcheck
	gl.glMultiTexCoord3iARB.restype = GLvoid
	gl.glMultiTexCoord3iARB.argtypes = (GLenum, GLint, GLint, GLint,)
	gl.glMultiTexCoord3iv = _getProcAddress('glMultiTexCoord3iv')
	gl.glMultiTexCoord3iv.gl = gl
	gl.glMultiTexCoord3iv.__name__ = 'glMultiTexCoord3iv'
	gl.glMultiTexCoord3iv.errcheck = _errcheck
	gl.glMultiTexCoord3iv.restype = GLvoid
	gl.glMultiTexCoord3iv.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord3ivARB = _getProcAddress('glMultiTexCoord3ivARB')
	gl.glMultiTexCoord3ivARB.gl = gl
	gl.glMultiTexCoord3ivARB.__name__ = 'glMultiTexCoord3ivARB'
	gl.glMultiTexCoord3ivARB.errcheck = _errcheck
	gl.glMultiTexCoord3ivARB.restype = GLvoid
	gl.glMultiTexCoord3ivARB.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord3s = _getProcAddress('glMultiTexCoord3s')
	gl.glMultiTexCoord3s.gl = gl
	gl.glMultiTexCoord3s.__name__ = 'glMultiTexCoord3s'
	gl.glMultiTexCoord3s.errcheck = _errcheck
	gl.glMultiTexCoord3s.restype = GLvoid
	gl.glMultiTexCoord3s.argtypes = (GLenum, GLshort, GLshort, GLshort,)
	gl.glMultiTexCoord3sARB = _getProcAddress('glMultiTexCoord3sARB')
	gl.glMultiTexCoord3sARB.gl = gl
	gl.glMultiTexCoord3sARB.__name__ = 'glMultiTexCoord3sARB'
	gl.glMultiTexCoord3sARB.errcheck = _errcheck
	gl.glMultiTexCoord3sARB.restype = GLvoid
	gl.glMultiTexCoord3sARB.argtypes = (GLenum, GLshort, GLshort, GLshort,)
	gl.glMultiTexCoord3sv = _getProcAddress('glMultiTexCoord3sv')
	gl.glMultiTexCoord3sv.gl = gl
	gl.glMultiTexCoord3sv.__name__ = 'glMultiTexCoord3sv'
	gl.glMultiTexCoord3sv.errcheck = _errcheck
	gl.glMultiTexCoord3sv.restype = GLvoid
	gl.glMultiTexCoord3sv.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord3svARB = _getProcAddress('glMultiTexCoord3svARB')
	gl.glMultiTexCoord3svARB.gl = gl
	gl.glMultiTexCoord3svARB.__name__ = 'glMultiTexCoord3svARB'
	gl.glMultiTexCoord3svARB.errcheck = _errcheck
	gl.glMultiTexCoord3svARB.restype = GLvoid
	gl.glMultiTexCoord3svARB.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord3xOES = _getProcAddress('glMultiTexCoord3xOES')
	gl.glMultiTexCoord3xOES.gl = gl
	gl.glMultiTexCoord3xOES.__name__ = 'glMultiTexCoord3xOES'
	gl.glMultiTexCoord3xOES.errcheck = _errcheck
	gl.glMultiTexCoord3xOES.restype = GLvoid
	gl.glMultiTexCoord3xOES.argtypes = (GLenum, GLfixed, GLfixed, GLfixed,)
	gl.glMultiTexCoord3xvOES = _getProcAddress('glMultiTexCoord3xvOES')
	gl.glMultiTexCoord3xvOES.gl = gl
	gl.glMultiTexCoord3xvOES.__name__ = 'glMultiTexCoord3xvOES'
	gl.glMultiTexCoord3xvOES.errcheck = _errcheck
	gl.glMultiTexCoord3xvOES.restype = GLvoid
	gl.glMultiTexCoord3xvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glMultiTexCoord4bOES = _getProcAddress('glMultiTexCoord4bOES')
	gl.glMultiTexCoord4bOES.gl = gl
	gl.glMultiTexCoord4bOES.__name__ = 'glMultiTexCoord4bOES'
	gl.glMultiTexCoord4bOES.errcheck = _errcheck
	gl.glMultiTexCoord4bOES.restype = GLvoid
	gl.glMultiTexCoord4bOES.argtypes = (GLenum, GLbyte, GLbyte, GLbyte, GLbyte,)
	gl.glMultiTexCoord4bvOES = _getProcAddress('glMultiTexCoord4bvOES')
	gl.glMultiTexCoord4bvOES.gl = gl
	gl.glMultiTexCoord4bvOES.__name__ = 'glMultiTexCoord4bvOES'
	gl.glMultiTexCoord4bvOES.errcheck = _errcheck
	gl.glMultiTexCoord4bvOES.restype = GLvoid
	gl.glMultiTexCoord4bvOES.argtypes = (GLenum, POINTER(GLbyte),)
	gl.glMultiTexCoord4d = _getProcAddress('glMultiTexCoord4d')
	gl.glMultiTexCoord4d.gl = gl
	gl.glMultiTexCoord4d.__name__ = 'glMultiTexCoord4d'
	gl.glMultiTexCoord4d.errcheck = _errcheck
	gl.glMultiTexCoord4d.restype = GLvoid
	gl.glMultiTexCoord4d.argtypes = (GLenum, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glMultiTexCoord4dARB = _getProcAddress('glMultiTexCoord4dARB')
	gl.glMultiTexCoord4dARB.gl = gl
	gl.glMultiTexCoord4dARB.__name__ = 'glMultiTexCoord4dARB'
	gl.glMultiTexCoord4dARB.errcheck = _errcheck
	gl.glMultiTexCoord4dARB.restype = GLvoid
	gl.glMultiTexCoord4dARB.argtypes = (GLenum, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glMultiTexCoord4dv = _getProcAddress('glMultiTexCoord4dv')
	gl.glMultiTexCoord4dv.gl = gl
	gl.glMultiTexCoord4dv.__name__ = 'glMultiTexCoord4dv'
	gl.glMultiTexCoord4dv.errcheck = _errcheck
	gl.glMultiTexCoord4dv.restype = GLvoid
	gl.glMultiTexCoord4dv.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord4dvARB = _getProcAddress('glMultiTexCoord4dvARB')
	gl.glMultiTexCoord4dvARB.gl = gl
	gl.glMultiTexCoord4dvARB.__name__ = 'glMultiTexCoord4dvARB'
	gl.glMultiTexCoord4dvARB.errcheck = _errcheck
	gl.glMultiTexCoord4dvARB.restype = GLvoid
	gl.glMultiTexCoord4dvARB.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glMultiTexCoord4f = _getProcAddress('glMultiTexCoord4f')
	gl.glMultiTexCoord4f.gl = gl
	gl.glMultiTexCoord4f.__name__ = 'glMultiTexCoord4f'
	gl.glMultiTexCoord4f.errcheck = _errcheck
	gl.glMultiTexCoord4f.restype = GLvoid
	gl.glMultiTexCoord4f.argtypes = (GLenum, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glMultiTexCoord4fARB = _getProcAddress('glMultiTexCoord4fARB')
	gl.glMultiTexCoord4fARB.gl = gl
	gl.glMultiTexCoord4fARB.__name__ = 'glMultiTexCoord4fARB'
	gl.glMultiTexCoord4fARB.errcheck = _errcheck
	gl.glMultiTexCoord4fARB.restype = GLvoid
	gl.glMultiTexCoord4fARB.argtypes = (GLenum, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glMultiTexCoord4fv = _getProcAddress('glMultiTexCoord4fv')
	gl.glMultiTexCoord4fv.gl = gl
	gl.glMultiTexCoord4fv.__name__ = 'glMultiTexCoord4fv'
	gl.glMultiTexCoord4fv.errcheck = _errcheck
	gl.glMultiTexCoord4fv.restype = GLvoid
	gl.glMultiTexCoord4fv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord4fvARB = _getProcAddress('glMultiTexCoord4fvARB')
	gl.glMultiTexCoord4fvARB.gl = gl
	gl.glMultiTexCoord4fvARB.__name__ = 'glMultiTexCoord4fvARB'
	gl.glMultiTexCoord4fvARB.errcheck = _errcheck
	gl.glMultiTexCoord4fvARB.restype = GLvoid
	gl.glMultiTexCoord4fvARB.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glMultiTexCoord4hNV = _getProcAddress('glMultiTexCoord4hNV')
	gl.glMultiTexCoord4hNV.gl = gl
	gl.glMultiTexCoord4hNV.__name__ = 'glMultiTexCoord4hNV'
	gl.glMultiTexCoord4hNV.errcheck = _errcheck
	gl.glMultiTexCoord4hNV.restype = GLvoid
	gl.glMultiTexCoord4hNV.argtypes = (GLenum, GLhalfNV, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glMultiTexCoord4hvNV = _getProcAddress('glMultiTexCoord4hvNV')
	gl.glMultiTexCoord4hvNV.gl = gl
	gl.glMultiTexCoord4hvNV.__name__ = 'glMultiTexCoord4hvNV'
	gl.glMultiTexCoord4hvNV.errcheck = _errcheck
	gl.glMultiTexCoord4hvNV.restype = GLvoid
	gl.glMultiTexCoord4hvNV.argtypes = (GLenum, POINTER(GLhalfNV),)
	gl.glMultiTexCoord4i = _getProcAddress('glMultiTexCoord4i')
	gl.glMultiTexCoord4i.gl = gl
	gl.glMultiTexCoord4i.__name__ = 'glMultiTexCoord4i'
	gl.glMultiTexCoord4i.errcheck = _errcheck
	gl.glMultiTexCoord4i.restype = GLvoid
	gl.glMultiTexCoord4i.argtypes = (GLenum, GLint, GLint, GLint, GLint,)
	gl.glMultiTexCoord4iARB = _getProcAddress('glMultiTexCoord4iARB')
	gl.glMultiTexCoord4iARB.gl = gl
	gl.glMultiTexCoord4iARB.__name__ = 'glMultiTexCoord4iARB'
	gl.glMultiTexCoord4iARB.errcheck = _errcheck
	gl.glMultiTexCoord4iARB.restype = GLvoid
	gl.glMultiTexCoord4iARB.argtypes = (GLenum, GLint, GLint, GLint, GLint,)
	gl.glMultiTexCoord4iv = _getProcAddress('glMultiTexCoord4iv')
	gl.glMultiTexCoord4iv.gl = gl
	gl.glMultiTexCoord4iv.__name__ = 'glMultiTexCoord4iv'
	gl.glMultiTexCoord4iv.errcheck = _errcheck
	gl.glMultiTexCoord4iv.restype = GLvoid
	gl.glMultiTexCoord4iv.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord4ivARB = _getProcAddress('glMultiTexCoord4ivARB')
	gl.glMultiTexCoord4ivARB.gl = gl
	gl.glMultiTexCoord4ivARB.__name__ = 'glMultiTexCoord4ivARB'
	gl.glMultiTexCoord4ivARB.errcheck = _errcheck
	gl.glMultiTexCoord4ivARB.restype = GLvoid
	gl.glMultiTexCoord4ivARB.argtypes = (GLenum, POINTER(GLint),)
	gl.glMultiTexCoord4s = _getProcAddress('glMultiTexCoord4s')
	gl.glMultiTexCoord4s.gl = gl
	gl.glMultiTexCoord4s.__name__ = 'glMultiTexCoord4s'
	gl.glMultiTexCoord4s.errcheck = _errcheck
	gl.glMultiTexCoord4s.restype = GLvoid
	gl.glMultiTexCoord4s.argtypes = (GLenum, GLshort, GLshort, GLshort, GLshort,)
	gl.glMultiTexCoord4sARB = _getProcAddress('glMultiTexCoord4sARB')
	gl.glMultiTexCoord4sARB.gl = gl
	gl.glMultiTexCoord4sARB.__name__ = 'glMultiTexCoord4sARB'
	gl.glMultiTexCoord4sARB.errcheck = _errcheck
	gl.glMultiTexCoord4sARB.restype = GLvoid
	gl.glMultiTexCoord4sARB.argtypes = (GLenum, GLshort, GLshort, GLshort, GLshort,)
	gl.glMultiTexCoord4sv = _getProcAddress('glMultiTexCoord4sv')
	gl.glMultiTexCoord4sv.gl = gl
	gl.glMultiTexCoord4sv.__name__ = 'glMultiTexCoord4sv'
	gl.glMultiTexCoord4sv.errcheck = _errcheck
	gl.glMultiTexCoord4sv.restype = GLvoid
	gl.glMultiTexCoord4sv.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord4svARB = _getProcAddress('glMultiTexCoord4svARB')
	gl.glMultiTexCoord4svARB.gl = gl
	gl.glMultiTexCoord4svARB.__name__ = 'glMultiTexCoord4svARB'
	gl.glMultiTexCoord4svARB.errcheck = _errcheck
	gl.glMultiTexCoord4svARB.restype = GLvoid
	gl.glMultiTexCoord4svARB.argtypes = (GLenum, POINTER(GLshort),)
	gl.glMultiTexCoord4x = _getProcAddress('glMultiTexCoord4x')
	gl.glMultiTexCoord4x.gl = gl
	gl.glMultiTexCoord4x.__name__ = 'glMultiTexCoord4x'
	gl.glMultiTexCoord4x.errcheck = _errcheck
	gl.glMultiTexCoord4x.restype = GLvoid
	gl.glMultiTexCoord4x.argtypes = (GLenum, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glMultiTexCoord4xOES = _getProcAddress('glMultiTexCoord4xOES')
	gl.glMultiTexCoord4xOES.gl = gl
	gl.glMultiTexCoord4xOES.__name__ = 'glMultiTexCoord4xOES'
	gl.glMultiTexCoord4xOES.errcheck = _errcheck
	gl.glMultiTexCoord4xOES.restype = GLvoid
	gl.glMultiTexCoord4xOES.argtypes = (GLenum, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glMultiTexCoord4xvOES = _getProcAddress('glMultiTexCoord4xvOES')
	gl.glMultiTexCoord4xvOES.gl = gl
	gl.glMultiTexCoord4xvOES.__name__ = 'glMultiTexCoord4xvOES'
	gl.glMultiTexCoord4xvOES.errcheck = _errcheck
	gl.glMultiTexCoord4xvOES.restype = GLvoid
	gl.glMultiTexCoord4xvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glMultiTexCoordP1ui = _getProcAddress('glMultiTexCoordP1ui')
	gl.glMultiTexCoordP1ui.gl = gl
	gl.glMultiTexCoordP1ui.__name__ = 'glMultiTexCoordP1ui'
	gl.glMultiTexCoordP1ui.errcheck = _errcheck
	gl.glMultiTexCoordP1ui.restype = GLvoid
	gl.glMultiTexCoordP1ui.argtypes = (GLenum, GLenum, GLuint,)
	gl.glMultiTexCoordP1uiv = _getProcAddress('glMultiTexCoordP1uiv')
	gl.glMultiTexCoordP1uiv.gl = gl
	gl.glMultiTexCoordP1uiv.__name__ = 'glMultiTexCoordP1uiv'
	gl.glMultiTexCoordP1uiv.errcheck = _errcheck
	gl.glMultiTexCoordP1uiv.restype = GLvoid
	gl.glMultiTexCoordP1uiv.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glMultiTexCoordP2ui = _getProcAddress('glMultiTexCoordP2ui')
	gl.glMultiTexCoordP2ui.gl = gl
	gl.glMultiTexCoordP2ui.__name__ = 'glMultiTexCoordP2ui'
	gl.glMultiTexCoordP2ui.errcheck = _errcheck
	gl.glMultiTexCoordP2ui.restype = GLvoid
	gl.glMultiTexCoordP2ui.argtypes = (GLenum, GLenum, GLuint,)
	gl.glMultiTexCoordP2uiv = _getProcAddress('glMultiTexCoordP2uiv')
	gl.glMultiTexCoordP2uiv.gl = gl
	gl.glMultiTexCoordP2uiv.__name__ = 'glMultiTexCoordP2uiv'
	gl.glMultiTexCoordP2uiv.errcheck = _errcheck
	gl.glMultiTexCoordP2uiv.restype = GLvoid
	gl.glMultiTexCoordP2uiv.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glMultiTexCoordP3ui = _getProcAddress('glMultiTexCoordP3ui')
	gl.glMultiTexCoordP3ui.gl = gl
	gl.glMultiTexCoordP3ui.__name__ = 'glMultiTexCoordP3ui'
	gl.glMultiTexCoordP3ui.errcheck = _errcheck
	gl.glMultiTexCoordP3ui.restype = GLvoid
	gl.glMultiTexCoordP3ui.argtypes = (GLenum, GLenum, GLuint,)
	gl.glMultiTexCoordP3uiv = _getProcAddress('glMultiTexCoordP3uiv')
	gl.glMultiTexCoordP3uiv.gl = gl
	gl.glMultiTexCoordP3uiv.__name__ = 'glMultiTexCoordP3uiv'
	gl.glMultiTexCoordP3uiv.errcheck = _errcheck
	gl.glMultiTexCoordP3uiv.restype = GLvoid
	gl.glMultiTexCoordP3uiv.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glMultiTexCoordP4ui = _getProcAddress('glMultiTexCoordP4ui')
	gl.glMultiTexCoordP4ui.gl = gl
	gl.glMultiTexCoordP4ui.__name__ = 'glMultiTexCoordP4ui'
	gl.glMultiTexCoordP4ui.errcheck = _errcheck
	gl.glMultiTexCoordP4ui.restype = GLvoid
	gl.glMultiTexCoordP4ui.argtypes = (GLenum, GLenum, GLuint,)
	gl.glMultiTexCoordP4uiv = _getProcAddress('glMultiTexCoordP4uiv')
	gl.glMultiTexCoordP4uiv.gl = gl
	gl.glMultiTexCoordP4uiv.__name__ = 'glMultiTexCoordP4uiv'
	gl.glMultiTexCoordP4uiv.errcheck = _errcheck
	gl.glMultiTexCoordP4uiv.restype = GLvoid
	gl.glMultiTexCoordP4uiv.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glMultiTexCoordPointerEXT = _getProcAddress('glMultiTexCoordPointerEXT')
	gl.glMultiTexCoordPointerEXT.gl = gl
	gl.glMultiTexCoordPointerEXT.__name__ = 'glMultiTexCoordPointerEXT'
	gl.glMultiTexCoordPointerEXT.errcheck = _errcheck
	gl.glMultiTexCoordPointerEXT.restype = GLvoid
	gl.glMultiTexCoordPointerEXT.argtypes = (GLenum, GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glMultiTexEnvfEXT = _getProcAddress('glMultiTexEnvfEXT')
	gl.glMultiTexEnvfEXT.gl = gl
	gl.glMultiTexEnvfEXT.__name__ = 'glMultiTexEnvfEXT'
	gl.glMultiTexEnvfEXT.errcheck = _errcheck
	gl.glMultiTexEnvfEXT.restype = GLvoid
	gl.glMultiTexEnvfEXT.argtypes = (GLenum, GLenum, GLenum, GLfloat,)
	gl.glMultiTexEnvfvEXT = _getProcAddress('glMultiTexEnvfvEXT')
	gl.glMultiTexEnvfvEXT.gl = gl
	gl.glMultiTexEnvfvEXT.__name__ = 'glMultiTexEnvfvEXT'
	gl.glMultiTexEnvfvEXT.errcheck = _errcheck
	gl.glMultiTexEnvfvEXT.restype = GLvoid
	gl.glMultiTexEnvfvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glMultiTexEnviEXT = _getProcAddress('glMultiTexEnviEXT')
	gl.glMultiTexEnviEXT.gl = gl
	gl.glMultiTexEnviEXT.__name__ = 'glMultiTexEnviEXT'
	gl.glMultiTexEnviEXT.errcheck = _errcheck
	gl.glMultiTexEnviEXT.restype = GLvoid
	gl.glMultiTexEnviEXT.argtypes = (GLenum, GLenum, GLenum, GLint,)
	gl.glMultiTexEnvivEXT = _getProcAddress('glMultiTexEnvivEXT')
	gl.glMultiTexEnvivEXT.gl = gl
	gl.glMultiTexEnvivEXT.__name__ = 'glMultiTexEnvivEXT'
	gl.glMultiTexEnvivEXT.errcheck = _errcheck
	gl.glMultiTexEnvivEXT.restype = GLvoid
	gl.glMultiTexEnvivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glMultiTexGendEXT = _getProcAddress('glMultiTexGendEXT')
	gl.glMultiTexGendEXT.gl = gl
	gl.glMultiTexGendEXT.__name__ = 'glMultiTexGendEXT'
	gl.glMultiTexGendEXT.errcheck = _errcheck
	gl.glMultiTexGendEXT.restype = GLvoid
	gl.glMultiTexGendEXT.argtypes = (GLenum, GLenum, GLenum, GLdouble,)
	gl.glMultiTexGendvEXT = _getProcAddress('glMultiTexGendvEXT')
	gl.glMultiTexGendvEXT.gl = gl
	gl.glMultiTexGendvEXT.__name__ = 'glMultiTexGendvEXT'
	gl.glMultiTexGendvEXT.errcheck = _errcheck
	gl.glMultiTexGendvEXT.restype = GLvoid
	gl.glMultiTexGendvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLdouble),)
	gl.glMultiTexGenfEXT = _getProcAddress('glMultiTexGenfEXT')
	gl.glMultiTexGenfEXT.gl = gl
	gl.glMultiTexGenfEXT.__name__ = 'glMultiTexGenfEXT'
	gl.glMultiTexGenfEXT.errcheck = _errcheck
	gl.glMultiTexGenfEXT.restype = GLvoid
	gl.glMultiTexGenfEXT.argtypes = (GLenum, GLenum, GLenum, GLfloat,)
	gl.glMultiTexGenfvEXT = _getProcAddress('glMultiTexGenfvEXT')
	gl.glMultiTexGenfvEXT.gl = gl
	gl.glMultiTexGenfvEXT.__name__ = 'glMultiTexGenfvEXT'
	gl.glMultiTexGenfvEXT.errcheck = _errcheck
	gl.glMultiTexGenfvEXT.restype = GLvoid
	gl.glMultiTexGenfvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glMultiTexGeniEXT = _getProcAddress('glMultiTexGeniEXT')
	gl.glMultiTexGeniEXT.gl = gl
	gl.glMultiTexGeniEXT.__name__ = 'glMultiTexGeniEXT'
	gl.glMultiTexGeniEXT.errcheck = _errcheck
	gl.glMultiTexGeniEXT.restype = GLvoid
	gl.glMultiTexGeniEXT.argtypes = (GLenum, GLenum, GLenum, GLint,)
	gl.glMultiTexGenivEXT = _getProcAddress('glMultiTexGenivEXT')
	gl.glMultiTexGenivEXT.gl = gl
	gl.glMultiTexGenivEXT.__name__ = 'glMultiTexGenivEXT'
	gl.glMultiTexGenivEXT.errcheck = _errcheck
	gl.glMultiTexGenivEXT.restype = GLvoid
	gl.glMultiTexGenivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glMultiTexImage1DEXT = _getProcAddress('glMultiTexImage1DEXT')
	gl.glMultiTexImage1DEXT.gl = gl
	gl.glMultiTexImage1DEXT.__name__ = 'glMultiTexImage1DEXT'
	gl.glMultiTexImage1DEXT.errcheck = _errcheck
	gl.glMultiTexImage1DEXT.restype = GLvoid
	gl.glMultiTexImage1DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glMultiTexImage2DEXT = _getProcAddress('glMultiTexImage2DEXT')
	gl.glMultiTexImage2DEXT.gl = gl
	gl.glMultiTexImage2DEXT.__name__ = 'glMultiTexImage2DEXT'
	gl.glMultiTexImage2DEXT.errcheck = _errcheck
	gl.glMultiTexImage2DEXT.restype = GLvoid
	gl.glMultiTexImage2DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glMultiTexImage3DEXT = _getProcAddress('glMultiTexImage3DEXT')
	gl.glMultiTexImage3DEXT.gl = gl
	gl.glMultiTexImage3DEXT.__name__ = 'glMultiTexImage3DEXT'
	gl.glMultiTexImage3DEXT.errcheck = _errcheck
	gl.glMultiTexImage3DEXT.restype = GLvoid
	gl.glMultiTexImage3DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glMultiTexParameterIivEXT = _getProcAddress('glMultiTexParameterIivEXT')
	gl.glMultiTexParameterIivEXT.gl = gl
	gl.glMultiTexParameterIivEXT.__name__ = 'glMultiTexParameterIivEXT'
	gl.glMultiTexParameterIivEXT.errcheck = _errcheck
	gl.glMultiTexParameterIivEXT.restype = GLvoid
	gl.glMultiTexParameterIivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glMultiTexParameterIuivEXT = _getProcAddress('glMultiTexParameterIuivEXT')
	gl.glMultiTexParameterIuivEXT.gl = gl
	gl.glMultiTexParameterIuivEXT.__name__ = 'glMultiTexParameterIuivEXT'
	gl.glMultiTexParameterIuivEXT.errcheck = _errcheck
	gl.glMultiTexParameterIuivEXT.restype = GLvoid
	gl.glMultiTexParameterIuivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLuint),)
	gl.glMultiTexParameterfEXT = _getProcAddress('glMultiTexParameterfEXT')
	gl.glMultiTexParameterfEXT.gl = gl
	gl.glMultiTexParameterfEXT.__name__ = 'glMultiTexParameterfEXT'
	gl.glMultiTexParameterfEXT.errcheck = _errcheck
	gl.glMultiTexParameterfEXT.restype = GLvoid
	gl.glMultiTexParameterfEXT.argtypes = (GLenum, GLenum, GLenum, GLfloat,)
	gl.glMultiTexParameterfvEXT = _getProcAddress('glMultiTexParameterfvEXT')
	gl.glMultiTexParameterfvEXT.gl = gl
	gl.glMultiTexParameterfvEXT.__name__ = 'glMultiTexParameterfvEXT'
	gl.glMultiTexParameterfvEXT.errcheck = _errcheck
	gl.glMultiTexParameterfvEXT.restype = GLvoid
	gl.glMultiTexParameterfvEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glMultiTexParameteriEXT = _getProcAddress('glMultiTexParameteriEXT')
	gl.glMultiTexParameteriEXT.gl = gl
	gl.glMultiTexParameteriEXT.__name__ = 'glMultiTexParameteriEXT'
	gl.glMultiTexParameteriEXT.errcheck = _errcheck
	gl.glMultiTexParameteriEXT.restype = GLvoid
	gl.glMultiTexParameteriEXT.argtypes = (GLenum, GLenum, GLenum, GLint,)
	gl.glMultiTexParameterivEXT = _getProcAddress('glMultiTexParameterivEXT')
	gl.glMultiTexParameterivEXT.gl = gl
	gl.glMultiTexParameterivEXT.__name__ = 'glMultiTexParameterivEXT'
	gl.glMultiTexParameterivEXT.errcheck = _errcheck
	gl.glMultiTexParameterivEXT.restype = GLvoid
	gl.glMultiTexParameterivEXT.argtypes = (GLenum, GLenum, GLenum, POINTER(GLint),)
	gl.glMultiTexRenderbufferEXT = _getProcAddress('glMultiTexRenderbufferEXT')
	gl.glMultiTexRenderbufferEXT.gl = gl
	gl.glMultiTexRenderbufferEXT.__name__ = 'glMultiTexRenderbufferEXT'
	gl.glMultiTexRenderbufferEXT.errcheck = _errcheck
	gl.glMultiTexRenderbufferEXT.restype = GLvoid
	gl.glMultiTexRenderbufferEXT.argtypes = (GLenum, GLenum, GLuint,)
	gl.glMultiTexSubImage1DEXT = _getProcAddress('glMultiTexSubImage1DEXT')
	gl.glMultiTexSubImage1DEXT.gl = gl
	gl.glMultiTexSubImage1DEXT.__name__ = 'glMultiTexSubImage1DEXT'
	gl.glMultiTexSubImage1DEXT.errcheck = _errcheck
	gl.glMultiTexSubImage1DEXT.restype = GLvoid
	gl.glMultiTexSubImage1DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glMultiTexSubImage2DEXT = _getProcAddress('glMultiTexSubImage2DEXT')
	gl.glMultiTexSubImage2DEXT.gl = gl
	gl.glMultiTexSubImage2DEXT.__name__ = 'glMultiTexSubImage2DEXT'
	gl.glMultiTexSubImage2DEXT.errcheck = _errcheck
	gl.glMultiTexSubImage2DEXT.restype = GLvoid
	gl.glMultiTexSubImage2DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glMultiTexSubImage3DEXT = _getProcAddress('glMultiTexSubImage3DEXT')
	gl.glMultiTexSubImage3DEXT.gl = gl
	gl.glMultiTexSubImage3DEXT.__name__ = 'glMultiTexSubImage3DEXT'
	gl.glMultiTexSubImage3DEXT.errcheck = _errcheck
	gl.glMultiTexSubImage3DEXT.restype = GLvoid
	gl.glMultiTexSubImage3DEXT.argtypes = (GLenum, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glNamedBufferData = _getProcAddress('glNamedBufferData')
	gl.glNamedBufferData.gl = gl
	gl.glNamedBufferData.__name__ = 'glNamedBufferData'
	gl.glNamedBufferData.errcheck = _errcheck
	gl.glNamedBufferData.restype = GLvoid
	gl.glNamedBufferData.argtypes = (GLuint, GLsizei, POINTER(GLvoid), GLenum,)
	gl.glNamedBufferDataEXT = _getProcAddress('glNamedBufferDataEXT')
	gl.glNamedBufferDataEXT.gl = gl
	gl.glNamedBufferDataEXT.__name__ = 'glNamedBufferDataEXT'
	gl.glNamedBufferDataEXT.errcheck = _errcheck
	gl.glNamedBufferDataEXT.restype = GLvoid
	gl.glNamedBufferDataEXT.argtypes = (GLuint, GLsizeiptr, POINTER(GLvoid), GLenum,)
	gl.glNamedBufferPageCommitmentARB = _getProcAddress('glNamedBufferPageCommitmentARB')
	gl.glNamedBufferPageCommitmentARB.gl = gl
	gl.glNamedBufferPageCommitmentARB.__name__ = 'glNamedBufferPageCommitmentARB'
	gl.glNamedBufferPageCommitmentARB.errcheck = _errcheck
	gl.glNamedBufferPageCommitmentARB.restype = GLvoid
	gl.glNamedBufferPageCommitmentARB.argtypes = (GLuint, GLintptr, GLsizeiptr, GLboolean,)
	gl.glNamedBufferPageCommitmentEXT = _getProcAddress('glNamedBufferPageCommitmentEXT')
	gl.glNamedBufferPageCommitmentEXT.gl = gl
	gl.glNamedBufferPageCommitmentEXT.__name__ = 'glNamedBufferPageCommitmentEXT'
	gl.glNamedBufferPageCommitmentEXT.errcheck = _errcheck
	gl.glNamedBufferPageCommitmentEXT.restype = GLvoid
	gl.glNamedBufferPageCommitmentEXT.argtypes = (GLuint, GLintptr, GLsizeiptr, GLboolean,)
	gl.glNamedBufferStorage = _getProcAddress('glNamedBufferStorage')
	gl.glNamedBufferStorage.gl = gl
	gl.glNamedBufferStorage.__name__ = 'glNamedBufferStorage'
	gl.glNamedBufferStorage.errcheck = _errcheck
	gl.glNamedBufferStorage.restype = GLvoid
	gl.glNamedBufferStorage.argtypes = (GLuint, GLsizei, POINTER(GLvoid), GLbitfield,)
	gl.glNamedBufferStorageEXT = _getProcAddress('glNamedBufferStorageEXT')
	gl.glNamedBufferStorageEXT.gl = gl
	gl.glNamedBufferStorageEXT.__name__ = 'glNamedBufferStorageEXT'
	gl.glNamedBufferStorageEXT.errcheck = _errcheck
	gl.glNamedBufferStorageEXT.restype = GLvoid
	gl.glNamedBufferStorageEXT.argtypes = (GLuint, GLsizeiptr, POINTER(GLvoid), GLbitfield,)
	gl.glNamedBufferSubData = _getProcAddress('glNamedBufferSubData')
	gl.glNamedBufferSubData.gl = gl
	gl.glNamedBufferSubData.__name__ = 'glNamedBufferSubData'
	gl.glNamedBufferSubData.errcheck = _errcheck
	gl.glNamedBufferSubData.restype = GLvoid
	gl.glNamedBufferSubData.argtypes = (GLuint, GLintptr, GLsizei, POINTER(GLvoid),)
	gl.glNamedBufferSubDataEXT = _getProcAddress('glNamedBufferSubDataEXT')
	gl.glNamedBufferSubDataEXT.gl = gl
	gl.glNamedBufferSubDataEXT.__name__ = 'glNamedBufferSubDataEXT'
	gl.glNamedBufferSubDataEXT.errcheck = _errcheck
	gl.glNamedBufferSubDataEXT.restype = GLvoid
	gl.glNamedBufferSubDataEXT.argtypes = (GLuint, GLintptr, GLsizeiptr, POINTER(GLvoid),)
	gl.glNamedCopyBufferSubDataEXT = _getProcAddress('glNamedCopyBufferSubDataEXT')
	gl.glNamedCopyBufferSubDataEXT.gl = gl
	gl.glNamedCopyBufferSubDataEXT.__name__ = 'glNamedCopyBufferSubDataEXT'
	gl.glNamedCopyBufferSubDataEXT.errcheck = _errcheck
	gl.glNamedCopyBufferSubDataEXT.restype = GLvoid
	gl.glNamedCopyBufferSubDataEXT.argtypes = (GLuint, GLuint, GLintptr, GLintptr, GLsizeiptr,)
	gl.glNamedFramebufferDrawBuffer = _getProcAddress('glNamedFramebufferDrawBuffer')
	gl.glNamedFramebufferDrawBuffer.gl = gl
	gl.glNamedFramebufferDrawBuffer.__name__ = 'glNamedFramebufferDrawBuffer'
	gl.glNamedFramebufferDrawBuffer.errcheck = _errcheck
	gl.glNamedFramebufferDrawBuffer.restype = GLvoid
	gl.glNamedFramebufferDrawBuffer.argtypes = (GLuint, GLenum,)
	gl.glNamedFramebufferDrawBuffers = _getProcAddress('glNamedFramebufferDrawBuffers')
	gl.glNamedFramebufferDrawBuffers.gl = gl
	gl.glNamedFramebufferDrawBuffers.__name__ = 'glNamedFramebufferDrawBuffers'
	gl.glNamedFramebufferDrawBuffers.errcheck = _errcheck
	gl.glNamedFramebufferDrawBuffers.restype = GLvoid
	gl.glNamedFramebufferDrawBuffers.argtypes = (GLuint, GLsizei, POINTER(GLenum),)
	gl.glNamedFramebufferParameteri = _getProcAddress('glNamedFramebufferParameteri')
	gl.glNamedFramebufferParameteri.gl = gl
	gl.glNamedFramebufferParameteri.__name__ = 'glNamedFramebufferParameteri'
	gl.glNamedFramebufferParameteri.errcheck = _errcheck
	gl.glNamedFramebufferParameteri.restype = GLvoid
	gl.glNamedFramebufferParameteri.argtypes = (GLuint, GLenum, GLint,)
	gl.glNamedFramebufferParameteriEXT = _getProcAddress('glNamedFramebufferParameteriEXT')
	gl.glNamedFramebufferParameteriEXT.gl = gl
	gl.glNamedFramebufferParameteriEXT.__name__ = 'glNamedFramebufferParameteriEXT'
	gl.glNamedFramebufferParameteriEXT.errcheck = _errcheck
	gl.glNamedFramebufferParameteriEXT.restype = GLvoid
	gl.glNamedFramebufferParameteriEXT.argtypes = (GLuint, GLenum, GLint,)
	gl.glNamedFramebufferReadBuffer = _getProcAddress('glNamedFramebufferReadBuffer')
	gl.glNamedFramebufferReadBuffer.gl = gl
	gl.glNamedFramebufferReadBuffer.__name__ = 'glNamedFramebufferReadBuffer'
	gl.glNamedFramebufferReadBuffer.errcheck = _errcheck
	gl.glNamedFramebufferReadBuffer.restype = GLvoid
	gl.glNamedFramebufferReadBuffer.argtypes = (GLuint, GLenum,)
	gl.glNamedFramebufferRenderbuffer = _getProcAddress('glNamedFramebufferRenderbuffer')
	gl.glNamedFramebufferRenderbuffer.gl = gl
	gl.glNamedFramebufferRenderbuffer.__name__ = 'glNamedFramebufferRenderbuffer'
	gl.glNamedFramebufferRenderbuffer.errcheck = _errcheck
	gl.glNamedFramebufferRenderbuffer.restype = GLvoid
	gl.glNamedFramebufferRenderbuffer.argtypes = (GLuint, GLenum, GLenum, GLuint,)
	gl.glNamedFramebufferRenderbufferEXT = _getProcAddress('glNamedFramebufferRenderbufferEXT')
	gl.glNamedFramebufferRenderbufferEXT.gl = gl
	gl.glNamedFramebufferRenderbufferEXT.__name__ = 'glNamedFramebufferRenderbufferEXT'
	gl.glNamedFramebufferRenderbufferEXT.errcheck = _errcheck
	gl.glNamedFramebufferRenderbufferEXT.restype = GLvoid
	gl.glNamedFramebufferRenderbufferEXT.argtypes = (GLuint, GLenum, GLenum, GLuint,)
	gl.glNamedFramebufferTexture = _getProcAddress('glNamedFramebufferTexture')
	gl.glNamedFramebufferTexture.gl = gl
	gl.glNamedFramebufferTexture.__name__ = 'glNamedFramebufferTexture'
	gl.glNamedFramebufferTexture.errcheck = _errcheck
	gl.glNamedFramebufferTexture.restype = GLvoid
	gl.glNamedFramebufferTexture.argtypes = (GLuint, GLenum, GLuint, GLint,)
	gl.glNamedFramebufferTexture1DEXT = _getProcAddress('glNamedFramebufferTexture1DEXT')
	gl.glNamedFramebufferTexture1DEXT.gl = gl
	gl.glNamedFramebufferTexture1DEXT.__name__ = 'glNamedFramebufferTexture1DEXT'
	gl.glNamedFramebufferTexture1DEXT.errcheck = _errcheck
	gl.glNamedFramebufferTexture1DEXT.restype = GLvoid
	gl.glNamedFramebufferTexture1DEXT.argtypes = (GLuint, GLenum, GLenum, GLuint, GLint,)
	gl.glNamedFramebufferTexture2DEXT = _getProcAddress('glNamedFramebufferTexture2DEXT')
	gl.glNamedFramebufferTexture2DEXT.gl = gl
	gl.glNamedFramebufferTexture2DEXT.__name__ = 'glNamedFramebufferTexture2DEXT'
	gl.glNamedFramebufferTexture2DEXT.errcheck = _errcheck
	gl.glNamedFramebufferTexture2DEXT.restype = GLvoid
	gl.glNamedFramebufferTexture2DEXT.argtypes = (GLuint, GLenum, GLenum, GLuint, GLint,)
	gl.glNamedFramebufferTexture3DEXT = _getProcAddress('glNamedFramebufferTexture3DEXT')
	gl.glNamedFramebufferTexture3DEXT.gl = gl
	gl.glNamedFramebufferTexture3DEXT.__name__ = 'glNamedFramebufferTexture3DEXT'
	gl.glNamedFramebufferTexture3DEXT.errcheck = _errcheck
	gl.glNamedFramebufferTexture3DEXT.restype = GLvoid
	gl.glNamedFramebufferTexture3DEXT.argtypes = (GLuint, GLenum, GLenum, GLuint, GLint, GLint,)
	gl.glNamedFramebufferTextureEXT = _getProcAddress('glNamedFramebufferTextureEXT')
	gl.glNamedFramebufferTextureEXT.gl = gl
	gl.glNamedFramebufferTextureEXT.__name__ = 'glNamedFramebufferTextureEXT'
	gl.glNamedFramebufferTextureEXT.errcheck = _errcheck
	gl.glNamedFramebufferTextureEXT.restype = GLvoid
	gl.glNamedFramebufferTextureEXT.argtypes = (GLuint, GLenum, GLuint, GLint,)
	gl.glNamedFramebufferTextureFaceEXT = _getProcAddress('glNamedFramebufferTextureFaceEXT')
	gl.glNamedFramebufferTextureFaceEXT.gl = gl
	gl.glNamedFramebufferTextureFaceEXT.__name__ = 'glNamedFramebufferTextureFaceEXT'
	gl.glNamedFramebufferTextureFaceEXT.errcheck = _errcheck
	gl.glNamedFramebufferTextureFaceEXT.restype = GLvoid
	gl.glNamedFramebufferTextureFaceEXT.argtypes = (GLuint, GLenum, GLuint, GLint, GLenum,)
	gl.glNamedFramebufferTextureLayer = _getProcAddress('glNamedFramebufferTextureLayer')
	gl.glNamedFramebufferTextureLayer.gl = gl
	gl.glNamedFramebufferTextureLayer.__name__ = 'glNamedFramebufferTextureLayer'
	gl.glNamedFramebufferTextureLayer.errcheck = _errcheck
	gl.glNamedFramebufferTextureLayer.restype = GLvoid
	gl.glNamedFramebufferTextureLayer.argtypes = (GLuint, GLenum, GLuint, GLint, GLint,)
	gl.glNamedFramebufferTextureLayerEXT = _getProcAddress('glNamedFramebufferTextureLayerEXT')
	gl.glNamedFramebufferTextureLayerEXT.gl = gl
	gl.glNamedFramebufferTextureLayerEXT.__name__ = 'glNamedFramebufferTextureLayerEXT'
	gl.glNamedFramebufferTextureLayerEXT.errcheck = _errcheck
	gl.glNamedFramebufferTextureLayerEXT.restype = GLvoid
	gl.glNamedFramebufferTextureLayerEXT.argtypes = (GLuint, GLenum, GLuint, GLint, GLint,)
	gl.glNamedProgramLocalParameter4dEXT = _getProcAddress('glNamedProgramLocalParameter4dEXT')
	gl.glNamedProgramLocalParameter4dEXT.gl = gl
	gl.glNamedProgramLocalParameter4dEXT.__name__ = 'glNamedProgramLocalParameter4dEXT'
	gl.glNamedProgramLocalParameter4dEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameter4dEXT.restype = GLvoid
	gl.glNamedProgramLocalParameter4dEXT.argtypes = (GLuint, GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glNamedProgramLocalParameter4dvEXT = _getProcAddress('glNamedProgramLocalParameter4dvEXT')
	gl.glNamedProgramLocalParameter4dvEXT.gl = gl
	gl.glNamedProgramLocalParameter4dvEXT.__name__ = 'glNamedProgramLocalParameter4dvEXT'
	gl.glNamedProgramLocalParameter4dvEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameter4dvEXT.restype = GLvoid
	gl.glNamedProgramLocalParameter4dvEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLdouble),)
	gl.glNamedProgramLocalParameter4fEXT = _getProcAddress('glNamedProgramLocalParameter4fEXT')
	gl.glNamedProgramLocalParameter4fEXT.gl = gl
	gl.glNamedProgramLocalParameter4fEXT.__name__ = 'glNamedProgramLocalParameter4fEXT'
	gl.glNamedProgramLocalParameter4fEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameter4fEXT.restype = GLvoid
	gl.glNamedProgramLocalParameter4fEXT.argtypes = (GLuint, GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glNamedProgramLocalParameter4fvEXT = _getProcAddress('glNamedProgramLocalParameter4fvEXT')
	gl.glNamedProgramLocalParameter4fvEXT.gl = gl
	gl.glNamedProgramLocalParameter4fvEXT.__name__ = 'glNamedProgramLocalParameter4fvEXT'
	gl.glNamedProgramLocalParameter4fvEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameter4fvEXT.restype = GLvoid
	gl.glNamedProgramLocalParameter4fvEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLfloat),)
	gl.glNamedProgramLocalParameterI4iEXT = _getProcAddress('glNamedProgramLocalParameterI4iEXT')
	gl.glNamedProgramLocalParameterI4iEXT.gl = gl
	gl.glNamedProgramLocalParameterI4iEXT.__name__ = 'glNamedProgramLocalParameterI4iEXT'
	gl.glNamedProgramLocalParameterI4iEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameterI4iEXT.restype = GLvoid
	gl.glNamedProgramLocalParameterI4iEXT.argtypes = (GLuint, GLenum, GLuint, GLint, GLint, GLint, GLint,)
	gl.glNamedProgramLocalParameterI4ivEXT = _getProcAddress('glNamedProgramLocalParameterI4ivEXT')
	gl.glNamedProgramLocalParameterI4ivEXT.gl = gl
	gl.glNamedProgramLocalParameterI4ivEXT.__name__ = 'glNamedProgramLocalParameterI4ivEXT'
	gl.glNamedProgramLocalParameterI4ivEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameterI4ivEXT.restype = GLvoid
	gl.glNamedProgramLocalParameterI4ivEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLint),)
	gl.glNamedProgramLocalParameterI4uiEXT = _getProcAddress('glNamedProgramLocalParameterI4uiEXT')
	gl.glNamedProgramLocalParameterI4uiEXT.gl = gl
	gl.glNamedProgramLocalParameterI4uiEXT.__name__ = 'glNamedProgramLocalParameterI4uiEXT'
	gl.glNamedProgramLocalParameterI4uiEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameterI4uiEXT.restype = GLvoid
	gl.glNamedProgramLocalParameterI4uiEXT.argtypes = (GLuint, GLenum, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glNamedProgramLocalParameterI4uivEXT = _getProcAddress('glNamedProgramLocalParameterI4uivEXT')
	gl.glNamedProgramLocalParameterI4uivEXT.gl = gl
	gl.glNamedProgramLocalParameterI4uivEXT.__name__ = 'glNamedProgramLocalParameterI4uivEXT'
	gl.glNamedProgramLocalParameterI4uivEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameterI4uivEXT.restype = GLvoid
	gl.glNamedProgramLocalParameterI4uivEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLuint),)
	gl.glNamedProgramLocalParameters4fvEXT = _getProcAddress('glNamedProgramLocalParameters4fvEXT')
	gl.glNamedProgramLocalParameters4fvEXT.gl = gl
	gl.glNamedProgramLocalParameters4fvEXT.__name__ = 'glNamedProgramLocalParameters4fvEXT'
	gl.glNamedProgramLocalParameters4fvEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParameters4fvEXT.restype = GLvoid
	gl.glNamedProgramLocalParameters4fvEXT.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glNamedProgramLocalParametersI4ivEXT = _getProcAddress('glNamedProgramLocalParametersI4ivEXT')
	gl.glNamedProgramLocalParametersI4ivEXT.gl = gl
	gl.glNamedProgramLocalParametersI4ivEXT.__name__ = 'glNamedProgramLocalParametersI4ivEXT'
	gl.glNamedProgramLocalParametersI4ivEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParametersI4ivEXT.restype = GLvoid
	gl.glNamedProgramLocalParametersI4ivEXT.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLint),)
	gl.glNamedProgramLocalParametersI4uivEXT = _getProcAddress('glNamedProgramLocalParametersI4uivEXT')
	gl.glNamedProgramLocalParametersI4uivEXT.gl = gl
	gl.glNamedProgramLocalParametersI4uivEXT.__name__ = 'glNamedProgramLocalParametersI4uivEXT'
	gl.glNamedProgramLocalParametersI4uivEXT.errcheck = _errcheck
	gl.glNamedProgramLocalParametersI4uivEXT.restype = GLvoid
	gl.glNamedProgramLocalParametersI4uivEXT.argtypes = (GLuint, GLenum, GLuint, GLsizei, POINTER(GLuint),)
	gl.glNamedProgramStringEXT = _getProcAddress('glNamedProgramStringEXT')
	gl.glNamedProgramStringEXT.gl = gl
	gl.glNamedProgramStringEXT.__name__ = 'glNamedProgramStringEXT'
	gl.glNamedProgramStringEXT.errcheck = _errcheck
	gl.glNamedProgramStringEXT.restype = GLvoid
	gl.glNamedProgramStringEXT.argtypes = (GLuint, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glNamedRenderbufferStorage = _getProcAddress('glNamedRenderbufferStorage')
	gl.glNamedRenderbufferStorage.gl = gl
	gl.glNamedRenderbufferStorage.__name__ = 'glNamedRenderbufferStorage'
	gl.glNamedRenderbufferStorage.errcheck = _errcheck
	gl.glNamedRenderbufferStorage.restype = GLvoid
	gl.glNamedRenderbufferStorage.argtypes = (GLuint, GLenum, GLsizei, GLsizei,)
	gl.glNamedRenderbufferStorageEXT = _getProcAddress('glNamedRenderbufferStorageEXT')
	gl.glNamedRenderbufferStorageEXT.gl = gl
	gl.glNamedRenderbufferStorageEXT.__name__ = 'glNamedRenderbufferStorageEXT'
	gl.glNamedRenderbufferStorageEXT.errcheck = _errcheck
	gl.glNamedRenderbufferStorageEXT.restype = GLvoid
	gl.glNamedRenderbufferStorageEXT.argtypes = (GLuint, GLenum, GLsizei, GLsizei,)
	gl.glNamedRenderbufferStorageMultisample = _getProcAddress('glNamedRenderbufferStorageMultisample')
	gl.glNamedRenderbufferStorageMultisample.gl = gl
	gl.glNamedRenderbufferStorageMultisample.__name__ = 'glNamedRenderbufferStorageMultisample'
	gl.glNamedRenderbufferStorageMultisample.errcheck = _errcheck
	gl.glNamedRenderbufferStorageMultisample.restype = GLvoid
	gl.glNamedRenderbufferStorageMultisample.argtypes = (GLuint, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glNamedRenderbufferStorageMultisampleCoverageEXT = _getProcAddress('glNamedRenderbufferStorageMultisampleCoverageEXT')
	gl.glNamedRenderbufferStorageMultisampleCoverageEXT.gl = gl
	gl.glNamedRenderbufferStorageMultisampleCoverageEXT.__name__ = 'glNamedRenderbufferStorageMultisampleCoverageEXT'
	gl.glNamedRenderbufferStorageMultisampleCoverageEXT.errcheck = _errcheck
	gl.glNamedRenderbufferStorageMultisampleCoverageEXT.restype = GLvoid
	gl.glNamedRenderbufferStorageMultisampleCoverageEXT.argtypes = (GLuint, GLsizei, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glNamedRenderbufferStorageMultisampleEXT = _getProcAddress('glNamedRenderbufferStorageMultisampleEXT')
	gl.glNamedRenderbufferStorageMultisampleEXT.gl = gl
	gl.glNamedRenderbufferStorageMultisampleEXT.__name__ = 'glNamedRenderbufferStorageMultisampleEXT'
	gl.glNamedRenderbufferStorageMultisampleEXT.errcheck = _errcheck
	gl.glNamedRenderbufferStorageMultisampleEXT.restype = GLvoid
	gl.glNamedRenderbufferStorageMultisampleEXT.argtypes = (GLuint, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glNamedStringARB = _getProcAddress('glNamedStringARB')
	gl.glNamedStringARB.gl = gl
	gl.glNamedStringARB.__name__ = 'glNamedStringARB'
	gl.glNamedStringARB.errcheck = _errcheck
	gl.glNamedStringARB.restype = GLvoid
	gl.glNamedStringARB.argtypes = (GLenum, GLint, POINTER(GLchar), GLint, POINTER(GLchar),)
	gl.glNewList = _getProcAddress('glNewList')
	gl.glNewList.gl = gl
	gl.glNewList.__name__ = 'glNewList'
	gl.glNewList.errcheck = _errcheck
	gl.glNewList.restype = GLvoid
	gl.glNewList.argtypes = (GLuint, GLenum,)
	gl.glNewObjectBufferATI = _getProcAddress('glNewObjectBufferATI')
	gl.glNewObjectBufferATI.gl = gl
	gl.glNewObjectBufferATI.__name__ = 'glNewObjectBufferATI'
	gl.glNewObjectBufferATI.errcheck = _errcheck
	gl.glNewObjectBufferATI.restype = GLuint
	gl.glNewObjectBufferATI.argtypes = (GLsizei, POINTER(GLvoid), GLenum,)
	gl.glNormal3b = _getProcAddress('glNormal3b')
	gl.glNormal3b.gl = gl
	gl.glNormal3b.__name__ = 'glNormal3b'
	gl.glNormal3b.errcheck = _errcheck
	gl.glNormal3b.restype = GLvoid
	gl.glNormal3b.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glNormal3bv = _getProcAddress('glNormal3bv')
	gl.glNormal3bv.gl = gl
	gl.glNormal3bv.__name__ = 'glNormal3bv'
	gl.glNormal3bv.errcheck = _errcheck
	gl.glNormal3bv.restype = GLvoid
	gl.glNormal3bv.argtypes = (POINTER(GLbyte),)
	gl.glNormal3d = _getProcAddress('glNormal3d')
	gl.glNormal3d.gl = gl
	gl.glNormal3d.__name__ = 'glNormal3d'
	gl.glNormal3d.errcheck = _errcheck
	gl.glNormal3d.restype = GLvoid
	gl.glNormal3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glNormal3dv = _getProcAddress('glNormal3dv')
	gl.glNormal3dv.gl = gl
	gl.glNormal3dv.__name__ = 'glNormal3dv'
	gl.glNormal3dv.errcheck = _errcheck
	gl.glNormal3dv.restype = GLvoid
	gl.glNormal3dv.argtypes = (POINTER(GLdouble),)
	gl.glNormal3f = _getProcAddress('glNormal3f')
	gl.glNormal3f.gl = gl
	gl.glNormal3f.__name__ = 'glNormal3f'
	gl.glNormal3f.errcheck = _errcheck
	gl.glNormal3f.restype = GLvoid
	gl.glNormal3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glNormal3fVertex3fSUN = _getProcAddress('glNormal3fVertex3fSUN')
	gl.glNormal3fVertex3fSUN.gl = gl
	gl.glNormal3fVertex3fSUN.__name__ = 'glNormal3fVertex3fSUN'
	gl.glNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glNormal3fVertex3fSUN.restype = GLvoid
	gl.glNormal3fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glNormal3fVertex3fvSUN = _getProcAddress('glNormal3fVertex3fvSUN')
	gl.glNormal3fVertex3fvSUN.gl = gl
	gl.glNormal3fVertex3fvSUN.__name__ = 'glNormal3fVertex3fvSUN'
	gl.glNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glNormal3fVertex3fvSUN.restype = GLvoid
	gl.glNormal3fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat),)
	gl.glNormal3fv = _getProcAddress('glNormal3fv')
	gl.glNormal3fv.gl = gl
	gl.glNormal3fv.__name__ = 'glNormal3fv'
	gl.glNormal3fv.errcheck = _errcheck
	gl.glNormal3fv.restype = GLvoid
	gl.glNormal3fv.argtypes = (POINTER(GLfloat),)
	gl.glNormal3hNV = _getProcAddress('glNormal3hNV')
	gl.glNormal3hNV.gl = gl
	gl.glNormal3hNV.__name__ = 'glNormal3hNV'
	gl.glNormal3hNV.errcheck = _errcheck
	gl.glNormal3hNV.restype = GLvoid
	gl.glNormal3hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glNormal3hvNV = _getProcAddress('glNormal3hvNV')
	gl.glNormal3hvNV.gl = gl
	gl.glNormal3hvNV.__name__ = 'glNormal3hvNV'
	gl.glNormal3hvNV.errcheck = _errcheck
	gl.glNormal3hvNV.restype = GLvoid
	gl.glNormal3hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glNormal3i = _getProcAddress('glNormal3i')
	gl.glNormal3i.gl = gl
	gl.glNormal3i.__name__ = 'glNormal3i'
	gl.glNormal3i.errcheck = _errcheck
	gl.glNormal3i.restype = GLvoid
	gl.glNormal3i.argtypes = (GLint, GLint, GLint,)
	gl.glNormal3iv = _getProcAddress('glNormal3iv')
	gl.glNormal3iv.gl = gl
	gl.glNormal3iv.__name__ = 'glNormal3iv'
	gl.glNormal3iv.errcheck = _errcheck
	gl.glNormal3iv.restype = GLvoid
	gl.glNormal3iv.argtypes = (POINTER(GLint),)
	gl.glNormal3s = _getProcAddress('glNormal3s')
	gl.glNormal3s.gl = gl
	gl.glNormal3s.__name__ = 'glNormal3s'
	gl.glNormal3s.errcheck = _errcheck
	gl.glNormal3s.restype = GLvoid
	gl.glNormal3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glNormal3sv = _getProcAddress('glNormal3sv')
	gl.glNormal3sv.gl = gl
	gl.glNormal3sv.__name__ = 'glNormal3sv'
	gl.glNormal3sv.errcheck = _errcheck
	gl.glNormal3sv.restype = GLvoid
	gl.glNormal3sv.argtypes = (POINTER(GLshort),)
	gl.glNormal3x = _getProcAddress('glNormal3x')
	gl.glNormal3x.gl = gl
	gl.glNormal3x.__name__ = 'glNormal3x'
	gl.glNormal3x.errcheck = _errcheck
	gl.glNormal3x.restype = GLvoid
	gl.glNormal3x.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glNormal3xOES = _getProcAddress('glNormal3xOES')
	gl.glNormal3xOES.gl = gl
	gl.glNormal3xOES.__name__ = 'glNormal3xOES'
	gl.glNormal3xOES.errcheck = _errcheck
	gl.glNormal3xOES.restype = GLvoid
	gl.glNormal3xOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glNormal3xvOES = _getProcAddress('glNormal3xvOES')
	gl.glNormal3xvOES.gl = gl
	gl.glNormal3xvOES.__name__ = 'glNormal3xvOES'
	gl.glNormal3xvOES.errcheck = _errcheck
	gl.glNormal3xvOES.restype = GLvoid
	gl.glNormal3xvOES.argtypes = (POINTER(GLfixed),)
	gl.glNormalFormatNV = _getProcAddress('glNormalFormatNV')
	gl.glNormalFormatNV.gl = gl
	gl.glNormalFormatNV.__name__ = 'glNormalFormatNV'
	gl.glNormalFormatNV.errcheck = _errcheck
	gl.glNormalFormatNV.restype = GLvoid
	gl.glNormalFormatNV.argtypes = (GLenum, GLsizei,)
	gl.glNormalP3ui = _getProcAddress('glNormalP3ui')
	gl.glNormalP3ui.gl = gl
	gl.glNormalP3ui.__name__ = 'glNormalP3ui'
	gl.glNormalP3ui.errcheck = _errcheck
	gl.glNormalP3ui.restype = GLvoid
	gl.glNormalP3ui.argtypes = (GLenum, GLuint,)
	gl.glNormalP3uiv = _getProcAddress('glNormalP3uiv')
	gl.glNormalP3uiv.gl = gl
	gl.glNormalP3uiv.__name__ = 'glNormalP3uiv'
	gl.glNormalP3uiv.errcheck = _errcheck
	gl.glNormalP3uiv.restype = GLvoid
	gl.glNormalP3uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glNormalPointer = _getProcAddress('glNormalPointer')
	gl.glNormalPointer.gl = gl
	gl.glNormalPointer.__name__ = 'glNormalPointer'
	gl.glNormalPointer.errcheck = _errcheck
	gl.glNormalPointer.restype = GLvoid
	gl.glNormalPointer.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glNormalPointerEXT = _getProcAddress('glNormalPointerEXT')
	gl.glNormalPointerEXT.gl = gl
	gl.glNormalPointerEXT.__name__ = 'glNormalPointerEXT'
	gl.glNormalPointerEXT.errcheck = _errcheck
	gl.glNormalPointerEXT.restype = GLvoid
	gl.glNormalPointerEXT.argtypes = (GLenum, GLsizei, GLsizei, POINTER(GLvoid),)
	gl.glNormalPointerListIBM = _getProcAddress('glNormalPointerListIBM')
	gl.glNormalPointerListIBM.gl = gl
	gl.glNormalPointerListIBM.__name__ = 'glNormalPointerListIBM'
	gl.glNormalPointerListIBM.errcheck = _errcheck
	gl.glNormalPointerListIBM.restype = GLvoid
	gl.glNormalPointerListIBM.argtypes = (GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glNormalPointervINTEL = _getProcAddress('glNormalPointervINTEL')
	gl.glNormalPointervINTEL.gl = gl
	gl.glNormalPointervINTEL.__name__ = 'glNormalPointervINTEL'
	gl.glNormalPointervINTEL.errcheck = _errcheck
	gl.glNormalPointervINTEL.restype = GLvoid
	gl.glNormalPointervINTEL.argtypes = (GLenum, POINTER(POINTER(GLvoid)),)
	gl.glNormalStream3bATI = _getProcAddress('glNormalStream3bATI')
	gl.glNormalStream3bATI.gl = gl
	gl.glNormalStream3bATI.__name__ = 'glNormalStream3bATI'
	gl.glNormalStream3bATI.errcheck = _errcheck
	gl.glNormalStream3bATI.restype = GLvoid
	gl.glNormalStream3bATI.argtypes = (GLenum, GLbyte, GLbyte, GLbyte,)
	gl.glNormalStream3bvATI = _getProcAddress('glNormalStream3bvATI')
	gl.glNormalStream3bvATI.gl = gl
	gl.glNormalStream3bvATI.__name__ = 'glNormalStream3bvATI'
	gl.glNormalStream3bvATI.errcheck = _errcheck
	gl.glNormalStream3bvATI.restype = GLvoid
	gl.glNormalStream3bvATI.argtypes = (GLenum, POINTER(GLbyte),)
	gl.glNormalStream3dATI = _getProcAddress('glNormalStream3dATI')
	gl.glNormalStream3dATI.gl = gl
	gl.glNormalStream3dATI.__name__ = 'glNormalStream3dATI'
	gl.glNormalStream3dATI.errcheck = _errcheck
	gl.glNormalStream3dATI.restype = GLvoid
	gl.glNormalStream3dATI.argtypes = (GLenum, GLdouble, GLdouble, GLdouble,)
	gl.glNormalStream3dvATI = _getProcAddress('glNormalStream3dvATI')
	gl.glNormalStream3dvATI.gl = gl
	gl.glNormalStream3dvATI.__name__ = 'glNormalStream3dvATI'
	gl.glNormalStream3dvATI.errcheck = _errcheck
	gl.glNormalStream3dvATI.restype = GLvoid
	gl.glNormalStream3dvATI.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glNormalStream3fATI = _getProcAddress('glNormalStream3fATI')
	gl.glNormalStream3fATI.gl = gl
	gl.glNormalStream3fATI.__name__ = 'glNormalStream3fATI'
	gl.glNormalStream3fATI.errcheck = _errcheck
	gl.glNormalStream3fATI.restype = GLvoid
	gl.glNormalStream3fATI.argtypes = (GLenum, GLfloat, GLfloat, GLfloat,)
	gl.glNormalStream3fvATI = _getProcAddress('glNormalStream3fvATI')
	gl.glNormalStream3fvATI.gl = gl
	gl.glNormalStream3fvATI.__name__ = 'glNormalStream3fvATI'
	gl.glNormalStream3fvATI.errcheck = _errcheck
	gl.glNormalStream3fvATI.restype = GLvoid
	gl.glNormalStream3fvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glNormalStream3iATI = _getProcAddress('glNormalStream3iATI')
	gl.glNormalStream3iATI.gl = gl
	gl.glNormalStream3iATI.__name__ = 'glNormalStream3iATI'
	gl.glNormalStream3iATI.errcheck = _errcheck
	gl.glNormalStream3iATI.restype = GLvoid
	gl.glNormalStream3iATI.argtypes = (GLenum, GLint, GLint, GLint,)
	gl.glNormalStream3ivATI = _getProcAddress('glNormalStream3ivATI')
	gl.glNormalStream3ivATI.gl = gl
	gl.glNormalStream3ivATI.__name__ = 'glNormalStream3ivATI'
	gl.glNormalStream3ivATI.errcheck = _errcheck
	gl.glNormalStream3ivATI.restype = GLvoid
	gl.glNormalStream3ivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glNormalStream3sATI = _getProcAddress('glNormalStream3sATI')
	gl.glNormalStream3sATI.gl = gl
	gl.glNormalStream3sATI.__name__ = 'glNormalStream3sATI'
	gl.glNormalStream3sATI.errcheck = _errcheck
	gl.glNormalStream3sATI.restype = GLvoid
	gl.glNormalStream3sATI.argtypes = (GLenum, GLshort, GLshort, GLshort,)
	gl.glNormalStream3svATI = _getProcAddress('glNormalStream3svATI')
	gl.glNormalStream3svATI.gl = gl
	gl.glNormalStream3svATI.__name__ = 'glNormalStream3svATI'
	gl.glNormalStream3svATI.errcheck = _errcheck
	gl.glNormalStream3svATI.restype = GLvoid
	gl.glNormalStream3svATI.argtypes = (GLenum, POINTER(GLshort),)
	gl.glObjectLabel = _getProcAddress('glObjectLabel')
	gl.glObjectLabel.gl = gl
	gl.glObjectLabel.__name__ = 'glObjectLabel'
	gl.glObjectLabel.errcheck = _errcheck
	gl.glObjectLabel.restype = GLvoid
	gl.glObjectLabel.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLchar),)
	gl.glObjectLabelKHR = _getProcAddress('glObjectLabelKHR')
	gl.glObjectLabelKHR.gl = gl
	gl.glObjectLabelKHR.__name__ = 'glObjectLabelKHR'
	gl.glObjectLabelKHR.errcheck = _errcheck
	gl.glObjectLabelKHR.restype = GLvoid
	gl.glObjectLabelKHR.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLchar),)
	gl.glObjectPtrLabel = _getProcAddress('glObjectPtrLabel')
	gl.glObjectPtrLabel.gl = gl
	gl.glObjectPtrLabel.__name__ = 'glObjectPtrLabel'
	gl.glObjectPtrLabel.errcheck = _errcheck
	gl.glObjectPtrLabel.restype = GLvoid
	gl.glObjectPtrLabel.argtypes = (POINTER(GLvoid), GLsizei, POINTER(GLchar),)
	gl.glObjectPtrLabelKHR = _getProcAddress('glObjectPtrLabelKHR')
	gl.glObjectPtrLabelKHR.gl = gl
	gl.glObjectPtrLabelKHR.__name__ = 'glObjectPtrLabelKHR'
	gl.glObjectPtrLabelKHR.errcheck = _errcheck
	gl.glObjectPtrLabelKHR.restype = GLvoid
	gl.glObjectPtrLabelKHR.argtypes = (POINTER(GLvoid), GLsizei, POINTER(GLchar),)
	gl.glObjectPurgeableAPPLE = _getProcAddress('glObjectPurgeableAPPLE')
	gl.glObjectPurgeableAPPLE.gl = gl
	gl.glObjectPurgeableAPPLE.__name__ = 'glObjectPurgeableAPPLE'
	gl.glObjectPurgeableAPPLE.errcheck = _errcheck
	gl.glObjectPurgeableAPPLE.restype = GLenum
	gl.glObjectPurgeableAPPLE.argtypes = (GLenum, GLuint, GLenum,)
	gl.glObjectUnpurgeableAPPLE = _getProcAddress('glObjectUnpurgeableAPPLE')
	gl.glObjectUnpurgeableAPPLE.gl = gl
	gl.glObjectUnpurgeableAPPLE.__name__ = 'glObjectUnpurgeableAPPLE'
	gl.glObjectUnpurgeableAPPLE.errcheck = _errcheck
	gl.glObjectUnpurgeableAPPLE.restype = GLenum
	gl.glObjectUnpurgeableAPPLE.argtypes = (GLenum, GLuint, GLenum,)
	gl.glOrtho = _getProcAddress('glOrtho')
	gl.glOrtho.gl = gl
	gl.glOrtho.__name__ = 'glOrtho'
	gl.glOrtho.errcheck = _errcheck
	gl.glOrtho.restype = GLvoid
	gl.glOrtho.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glOrthof = _getProcAddress('glOrthof')
	gl.glOrthof.gl = gl
	gl.glOrthof.__name__ = 'glOrthof'
	gl.glOrthof.errcheck = _errcheck
	gl.glOrthof.restype = GLvoid
	gl.glOrthof.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glOrthofOES = _getProcAddress('glOrthofOES')
	gl.glOrthofOES.gl = gl
	gl.glOrthofOES.__name__ = 'glOrthofOES'
	gl.glOrthofOES.errcheck = _errcheck
	gl.glOrthofOES.restype = GLvoid
	gl.glOrthofOES.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glOrthox = _getProcAddress('glOrthox')
	gl.glOrthox.gl = gl
	gl.glOrthox.__name__ = 'glOrthox'
	gl.glOrthox.errcheck = _errcheck
	gl.glOrthox.restype = GLvoid
	gl.glOrthox.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glOrthoxOES = _getProcAddress('glOrthoxOES')
	gl.glOrthoxOES.gl = gl
	gl.glOrthoxOES.__name__ = 'glOrthoxOES'
	gl.glOrthoxOES.errcheck = _errcheck
	gl.glOrthoxOES.restype = GLvoid
	gl.glOrthoxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glPNTrianglesfATI = _getProcAddress('glPNTrianglesfATI')
	gl.glPNTrianglesfATI.gl = gl
	gl.glPNTrianglesfATI.__name__ = 'glPNTrianglesfATI'
	gl.glPNTrianglesfATI.errcheck = _errcheck
	gl.glPNTrianglesfATI.restype = GLvoid
	gl.glPNTrianglesfATI.argtypes = (GLenum, GLfloat,)
	gl.glPNTrianglesiATI = _getProcAddress('glPNTrianglesiATI')
	gl.glPNTrianglesiATI.gl = gl
	gl.glPNTrianglesiATI.__name__ = 'glPNTrianglesiATI'
	gl.glPNTrianglesiATI.errcheck = _errcheck
	gl.glPNTrianglesiATI.restype = GLvoid
	gl.glPNTrianglesiATI.argtypes = (GLenum, GLint,)
	gl.glPassTexCoordATI = _getProcAddress('glPassTexCoordATI')
	gl.glPassTexCoordATI.gl = gl
	gl.glPassTexCoordATI.__name__ = 'glPassTexCoordATI'
	gl.glPassTexCoordATI.errcheck = _errcheck
	gl.glPassTexCoordATI.restype = GLvoid
	gl.glPassTexCoordATI.argtypes = (GLuint, GLuint, GLenum,)
	gl.glPassThrough = _getProcAddress('glPassThrough')
	gl.glPassThrough.gl = gl
	gl.glPassThrough.__name__ = 'glPassThrough'
	gl.glPassThrough.errcheck = _errcheck
	gl.glPassThrough.restype = GLvoid
	gl.glPassThrough.argtypes = (GLfloat,)
	gl.glPassThroughxOES = _getProcAddress('glPassThroughxOES')
	gl.glPassThroughxOES.gl = gl
	gl.glPassThroughxOES.__name__ = 'glPassThroughxOES'
	gl.glPassThroughxOES.errcheck = _errcheck
	gl.glPassThroughxOES.restype = GLvoid
	gl.glPassThroughxOES.argtypes = (GLfixed,)
	gl.glPatchParameterfv = _getProcAddress('glPatchParameterfv')
	gl.glPatchParameterfv.gl = gl
	gl.glPatchParameterfv.__name__ = 'glPatchParameterfv'
	gl.glPatchParameterfv.errcheck = _errcheck
	gl.glPatchParameterfv.restype = GLvoid
	gl.glPatchParameterfv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glPatchParameteri = _getProcAddress('glPatchParameteri')
	gl.glPatchParameteri.gl = gl
	gl.glPatchParameteri.__name__ = 'glPatchParameteri'
	gl.glPatchParameteri.errcheck = _errcheck
	gl.glPatchParameteri.restype = GLvoid
	gl.glPatchParameteri.argtypes = (GLenum, GLint,)
	gl.glPatchParameteriEXT = _getProcAddress('glPatchParameteriEXT')
	gl.glPatchParameteriEXT.gl = gl
	gl.glPatchParameteriEXT.__name__ = 'glPatchParameteriEXT'
	gl.glPatchParameteriEXT.errcheck = _errcheck
	gl.glPatchParameteriEXT.restype = GLvoid
	gl.glPatchParameteriEXT.argtypes = (GLenum, GLint,)
	gl.glPathColorGenNV = _getProcAddress('glPathColorGenNV')
	gl.glPathColorGenNV.gl = gl
	gl.glPathColorGenNV.__name__ = 'glPathColorGenNV'
	gl.glPathColorGenNV.errcheck = _errcheck
	gl.glPathColorGenNV.restype = GLvoid
	gl.glPathColorGenNV.argtypes = (GLenum, GLenum, GLenum, POINTER(GLfloat),)
	gl.glPathCommandsNV = _getProcAddress('glPathCommandsNV')
	gl.glPathCommandsNV.gl = gl
	gl.glPathCommandsNV.__name__ = 'glPathCommandsNV'
	gl.glPathCommandsNV.errcheck = _errcheck
	gl.glPathCommandsNV.restype = GLvoid
	gl.glPathCommandsNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), GLsizei, GLenum, POINTER(GLvoid),)
	gl.glPathCoordsNV = _getProcAddress('glPathCoordsNV')
	gl.glPathCoordsNV.gl = gl
	gl.glPathCoordsNV.__name__ = 'glPathCoordsNV'
	gl.glPathCoordsNV.errcheck = _errcheck
	gl.glPathCoordsNV.restype = GLvoid
	gl.glPathCoordsNV.argtypes = (GLuint, GLsizei, GLenum, POINTER(GLvoid),)
	gl.glPathCoverDepthFuncNV = _getProcAddress('glPathCoverDepthFuncNV')
	gl.glPathCoverDepthFuncNV.gl = gl
	gl.glPathCoverDepthFuncNV.__name__ = 'glPathCoverDepthFuncNV'
	gl.glPathCoverDepthFuncNV.errcheck = _errcheck
	gl.glPathCoverDepthFuncNV.restype = GLvoid
	gl.glPathCoverDepthFuncNV.argtypes = (GLenum,)
	gl.glPathDashArrayNV = _getProcAddress('glPathDashArrayNV')
	gl.glPathDashArrayNV.gl = gl
	gl.glPathDashArrayNV.__name__ = 'glPathDashArrayNV'
	gl.glPathDashArrayNV.errcheck = _errcheck
	gl.glPathDashArrayNV.restype = GLvoid
	gl.glPathDashArrayNV.argtypes = (GLuint, GLsizei, POINTER(GLfloat),)
	gl.glPathFogGenNV = _getProcAddress('glPathFogGenNV')
	gl.glPathFogGenNV.gl = gl
	gl.glPathFogGenNV.__name__ = 'glPathFogGenNV'
	gl.glPathFogGenNV.errcheck = _errcheck
	gl.glPathFogGenNV.restype = GLvoid
	gl.glPathFogGenNV.argtypes = (GLenum,)
	gl.glPathGlyphIndexArrayNV = _getProcAddress('glPathGlyphIndexArrayNV')
	gl.glPathGlyphIndexArrayNV.gl = gl
	gl.glPathGlyphIndexArrayNV.__name__ = 'glPathGlyphIndexArrayNV'
	gl.glPathGlyphIndexArrayNV.errcheck = _errcheck
	gl.glPathGlyphIndexArrayNV.restype = GLenum
	gl.glPathGlyphIndexArrayNV.argtypes = (GLuint, GLenum, POINTER(GLvoid), GLbitfield, GLuint, GLsizei, GLuint, GLfloat,)
	gl.glPathGlyphIndexRangeNV = _getProcAddress('glPathGlyphIndexRangeNV')
	gl.glPathGlyphIndexRangeNV.gl = gl
	gl.glPathGlyphIndexRangeNV.__name__ = 'glPathGlyphIndexRangeNV'
	gl.glPathGlyphIndexRangeNV.errcheck = _errcheck
	gl.glPathGlyphIndexRangeNV.restype = GLenum
	gl.glPathGlyphIndexRangeNV.argtypes = (GLenum, POINTER(GLvoid), GLbitfield, GLuint, GLfloat, GLuint,)
	gl.glPathGlyphRangeNV = _getProcAddress('glPathGlyphRangeNV')
	gl.glPathGlyphRangeNV.gl = gl
	gl.glPathGlyphRangeNV.__name__ = 'glPathGlyphRangeNV'
	gl.glPathGlyphRangeNV.errcheck = _errcheck
	gl.glPathGlyphRangeNV.restype = GLvoid
	gl.glPathGlyphRangeNV.argtypes = (GLuint, GLenum, POINTER(GLvoid), GLbitfield, GLuint, GLsizei, GLenum, GLuint, GLfloat,)
	gl.glPathGlyphsNV = _getProcAddress('glPathGlyphsNV')
	gl.glPathGlyphsNV.gl = gl
	gl.glPathGlyphsNV.__name__ = 'glPathGlyphsNV'
	gl.glPathGlyphsNV.errcheck = _errcheck
	gl.glPathGlyphsNV.restype = GLvoid
	gl.glPathGlyphsNV.argtypes = (GLuint, GLenum, POINTER(GLvoid), GLbitfield, GLsizei, GLenum, POINTER(GLvoid), GLenum, GLuint, GLfloat,)
	gl.glPathMemoryGlyphIndexArrayNV = _getProcAddress('glPathMemoryGlyphIndexArrayNV')
	gl.glPathMemoryGlyphIndexArrayNV.gl = gl
	gl.glPathMemoryGlyphIndexArrayNV.__name__ = 'glPathMemoryGlyphIndexArrayNV'
	gl.glPathMemoryGlyphIndexArrayNV.errcheck = _errcheck
	gl.glPathMemoryGlyphIndexArrayNV.restype = GLenum
	gl.glPathMemoryGlyphIndexArrayNV.argtypes = (GLuint, GLenum, GLsizeiptr, POINTER(GLvoid), GLsizei, GLuint, GLsizei, GLuint, GLfloat,)
	gl.glPathParameterfNV = _getProcAddress('glPathParameterfNV')
	gl.glPathParameterfNV.gl = gl
	gl.glPathParameterfNV.__name__ = 'glPathParameterfNV'
	gl.glPathParameterfNV.errcheck = _errcheck
	gl.glPathParameterfNV.restype = GLvoid
	gl.glPathParameterfNV.argtypes = (GLuint, GLenum, GLfloat,)
	gl.glPathParameterfvNV = _getProcAddress('glPathParameterfvNV')
	gl.glPathParameterfvNV.gl = gl
	gl.glPathParameterfvNV.__name__ = 'glPathParameterfvNV'
	gl.glPathParameterfvNV.errcheck = _errcheck
	gl.glPathParameterfvNV.restype = GLvoid
	gl.glPathParameterfvNV.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glPathParameteriNV = _getProcAddress('glPathParameteriNV')
	gl.glPathParameteriNV.gl = gl
	gl.glPathParameteriNV.__name__ = 'glPathParameteriNV'
	gl.glPathParameteriNV.errcheck = _errcheck
	gl.glPathParameteriNV.restype = GLvoid
	gl.glPathParameteriNV.argtypes = (GLuint, GLenum, GLint,)
	gl.glPathParameterivNV = _getProcAddress('glPathParameterivNV')
	gl.glPathParameterivNV.gl = gl
	gl.glPathParameterivNV.__name__ = 'glPathParameterivNV'
	gl.glPathParameterivNV.errcheck = _errcheck
	gl.glPathParameterivNV.restype = GLvoid
	gl.glPathParameterivNV.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glPathStencilDepthOffsetNV = _getProcAddress('glPathStencilDepthOffsetNV')
	gl.glPathStencilDepthOffsetNV.gl = gl
	gl.glPathStencilDepthOffsetNV.__name__ = 'glPathStencilDepthOffsetNV'
	gl.glPathStencilDepthOffsetNV.errcheck = _errcheck
	gl.glPathStencilDepthOffsetNV.restype = GLvoid
	gl.glPathStencilDepthOffsetNV.argtypes = (GLfloat, GLfloat,)
	gl.glPathStencilFuncNV = _getProcAddress('glPathStencilFuncNV')
	gl.glPathStencilFuncNV.gl = gl
	gl.glPathStencilFuncNV.__name__ = 'glPathStencilFuncNV'
	gl.glPathStencilFuncNV.errcheck = _errcheck
	gl.glPathStencilFuncNV.restype = GLvoid
	gl.glPathStencilFuncNV.argtypes = (GLenum, GLint, GLuint,)
	gl.glPathStringNV = _getProcAddress('glPathStringNV')
	gl.glPathStringNV.gl = gl
	gl.glPathStringNV.__name__ = 'glPathStringNV'
	gl.glPathStringNV.errcheck = _errcheck
	gl.glPathStringNV.restype = GLvoid
	gl.glPathStringNV.argtypes = (GLuint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glPathSubCommandsNV = _getProcAddress('glPathSubCommandsNV')
	gl.glPathSubCommandsNV.gl = gl
	gl.glPathSubCommandsNV.__name__ = 'glPathSubCommandsNV'
	gl.glPathSubCommandsNV.errcheck = _errcheck
	gl.glPathSubCommandsNV.restype = GLvoid
	gl.glPathSubCommandsNV.argtypes = (GLuint, GLsizei, GLsizei, GLsizei, POINTER(GLubyte), GLsizei, GLenum, POINTER(GLvoid),)
	gl.glPathSubCoordsNV = _getProcAddress('glPathSubCoordsNV')
	gl.glPathSubCoordsNV.gl = gl
	gl.glPathSubCoordsNV.__name__ = 'glPathSubCoordsNV'
	gl.glPathSubCoordsNV.errcheck = _errcheck
	gl.glPathSubCoordsNV.restype = GLvoid
	gl.glPathSubCoordsNV.argtypes = (GLuint, GLsizei, GLsizei, GLenum, POINTER(GLvoid),)
	gl.glPathTexGenNV = _getProcAddress('glPathTexGenNV')
	gl.glPathTexGenNV.gl = gl
	gl.glPathTexGenNV.__name__ = 'glPathTexGenNV'
	gl.glPathTexGenNV.errcheck = _errcheck
	gl.glPathTexGenNV.restype = GLvoid
	gl.glPathTexGenNV.argtypes = (GLenum, GLenum, GLint, POINTER(GLfloat),)
	gl.glPauseTransformFeedback = _getProcAddress('glPauseTransformFeedback')
	gl.glPauseTransformFeedback.gl = gl
	gl.glPauseTransformFeedback.__name__ = 'glPauseTransformFeedback'
	gl.glPauseTransformFeedback.errcheck = _errcheck
	gl.glPauseTransformFeedback.restype = GLvoid
	gl.glPauseTransformFeedback.argtypes = ()
	gl.glPauseTransformFeedbackNV = _getProcAddress('glPauseTransformFeedbackNV')
	gl.glPauseTransformFeedbackNV.gl = gl
	gl.glPauseTransformFeedbackNV.__name__ = 'glPauseTransformFeedbackNV'
	gl.glPauseTransformFeedbackNV.errcheck = _errcheck
	gl.glPauseTransformFeedbackNV.restype = GLvoid
	gl.glPauseTransformFeedbackNV.argtypes = ()
	gl.glPixelDataRangeNV = _getProcAddress('glPixelDataRangeNV')
	gl.glPixelDataRangeNV.gl = gl
	gl.glPixelDataRangeNV.__name__ = 'glPixelDataRangeNV'
	gl.glPixelDataRangeNV.errcheck = _errcheck
	gl.glPixelDataRangeNV.restype = GLvoid
	gl.glPixelDataRangeNV.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glPixelMapfv = _getProcAddress('glPixelMapfv')
	gl.glPixelMapfv.gl = gl
	gl.glPixelMapfv.__name__ = 'glPixelMapfv'
	gl.glPixelMapfv.errcheck = _errcheck
	gl.glPixelMapfv.restype = GLvoid
	gl.glPixelMapfv.argtypes = (GLenum, GLsizei, POINTER(GLfloat),)
	gl.glPixelMapuiv = _getProcAddress('glPixelMapuiv')
	gl.glPixelMapuiv.gl = gl
	gl.glPixelMapuiv.__name__ = 'glPixelMapuiv'
	gl.glPixelMapuiv.errcheck = _errcheck
	gl.glPixelMapuiv.restype = GLvoid
	gl.glPixelMapuiv.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glPixelMapusv = _getProcAddress('glPixelMapusv')
	gl.glPixelMapusv.gl = gl
	gl.glPixelMapusv.__name__ = 'glPixelMapusv'
	gl.glPixelMapusv.errcheck = _errcheck
	gl.glPixelMapusv.restype = GLvoid
	gl.glPixelMapusv.argtypes = (GLenum, GLsizei, POINTER(GLushort),)
	gl.glPixelMapx = _getProcAddress('glPixelMapx')
	gl.glPixelMapx.gl = gl
	gl.glPixelMapx.__name__ = 'glPixelMapx'
	gl.glPixelMapx.errcheck = _errcheck
	gl.glPixelMapx.restype = GLvoid
	gl.glPixelMapx.argtypes = (GLenum, GLint, POINTER(GLfixed),)
	gl.glPixelStoref = _getProcAddress('glPixelStoref')
	gl.glPixelStoref.gl = gl
	gl.glPixelStoref.__name__ = 'glPixelStoref'
	gl.glPixelStoref.errcheck = _errcheck
	gl.glPixelStoref.restype = GLvoid
	gl.glPixelStoref.argtypes = (GLenum, GLfloat,)
	gl.glPixelStorei = _getProcAddress('glPixelStorei')
	gl.glPixelStorei.gl = gl
	gl.glPixelStorei.__name__ = 'glPixelStorei'
	gl.glPixelStorei.errcheck = _errcheck
	gl.glPixelStorei.restype = GLvoid
	gl.glPixelStorei.argtypes = (GLenum, GLint,)
	gl.glPixelStorex = _getProcAddress('glPixelStorex')
	gl.glPixelStorex.gl = gl
	gl.glPixelStorex.__name__ = 'glPixelStorex'
	gl.glPixelStorex.errcheck = _errcheck
	gl.glPixelStorex.restype = GLvoid
	gl.glPixelStorex.argtypes = (GLenum, GLfixed,)
	gl.glPixelTexGenParameterfSGIS = _getProcAddress('glPixelTexGenParameterfSGIS')
	gl.glPixelTexGenParameterfSGIS.gl = gl
	gl.glPixelTexGenParameterfSGIS.__name__ = 'glPixelTexGenParameterfSGIS'
	gl.glPixelTexGenParameterfSGIS.errcheck = _errcheck
	gl.glPixelTexGenParameterfSGIS.restype = GLvoid
	gl.glPixelTexGenParameterfSGIS.argtypes = (GLenum, GLfloat,)
	gl.glPixelTexGenParameterfvSGIS = _getProcAddress('glPixelTexGenParameterfvSGIS')
	gl.glPixelTexGenParameterfvSGIS.gl = gl
	gl.glPixelTexGenParameterfvSGIS.__name__ = 'glPixelTexGenParameterfvSGIS'
	gl.glPixelTexGenParameterfvSGIS.errcheck = _errcheck
	gl.glPixelTexGenParameterfvSGIS.restype = GLvoid
	gl.glPixelTexGenParameterfvSGIS.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glPixelTexGenParameteriSGIS = _getProcAddress('glPixelTexGenParameteriSGIS')
	gl.glPixelTexGenParameteriSGIS.gl = gl
	gl.glPixelTexGenParameteriSGIS.__name__ = 'glPixelTexGenParameteriSGIS'
	gl.glPixelTexGenParameteriSGIS.errcheck = _errcheck
	gl.glPixelTexGenParameteriSGIS.restype = GLvoid
	gl.glPixelTexGenParameteriSGIS.argtypes = (GLenum, GLint,)
	gl.glPixelTexGenParameterivSGIS = _getProcAddress('glPixelTexGenParameterivSGIS')
	gl.glPixelTexGenParameterivSGIS.gl = gl
	gl.glPixelTexGenParameterivSGIS.__name__ = 'glPixelTexGenParameterivSGIS'
	gl.glPixelTexGenParameterivSGIS.errcheck = _errcheck
	gl.glPixelTexGenParameterivSGIS.restype = GLvoid
	gl.glPixelTexGenParameterivSGIS.argtypes = (GLenum, POINTER(GLint),)
	gl.glPixelTexGenSGIX = _getProcAddress('glPixelTexGenSGIX')
	gl.glPixelTexGenSGIX.gl = gl
	gl.glPixelTexGenSGIX.__name__ = 'glPixelTexGenSGIX'
	gl.glPixelTexGenSGIX.errcheck = _errcheck
	gl.glPixelTexGenSGIX.restype = GLvoid
	gl.glPixelTexGenSGIX.argtypes = (GLenum,)
	gl.glPixelTransferf = _getProcAddress('glPixelTransferf')
	gl.glPixelTransferf.gl = gl
	gl.glPixelTransferf.__name__ = 'glPixelTransferf'
	gl.glPixelTransferf.errcheck = _errcheck
	gl.glPixelTransferf.restype = GLvoid
	gl.glPixelTransferf.argtypes = (GLenum, GLfloat,)
	gl.glPixelTransferi = _getProcAddress('glPixelTransferi')
	gl.glPixelTransferi.gl = gl
	gl.glPixelTransferi.__name__ = 'glPixelTransferi'
	gl.glPixelTransferi.errcheck = _errcheck
	gl.glPixelTransferi.restype = GLvoid
	gl.glPixelTransferi.argtypes = (GLenum, GLint,)
	gl.glPixelTransferxOES = _getProcAddress('glPixelTransferxOES')
	gl.glPixelTransferxOES.gl = gl
	gl.glPixelTransferxOES.__name__ = 'glPixelTransferxOES'
	gl.glPixelTransferxOES.errcheck = _errcheck
	gl.glPixelTransferxOES.restype = GLvoid
	gl.glPixelTransferxOES.argtypes = (GLenum, GLfixed,)
	gl.glPixelTransformParameterfEXT = _getProcAddress('glPixelTransformParameterfEXT')
	gl.glPixelTransformParameterfEXT.gl = gl
	gl.glPixelTransformParameterfEXT.__name__ = 'glPixelTransformParameterfEXT'
	gl.glPixelTransformParameterfEXT.errcheck = _errcheck
	gl.glPixelTransformParameterfEXT.restype = GLvoid
	gl.glPixelTransformParameterfEXT.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glPixelTransformParameterfvEXT = _getProcAddress('glPixelTransformParameterfvEXT')
	gl.glPixelTransformParameterfvEXT.gl = gl
	gl.glPixelTransformParameterfvEXT.__name__ = 'glPixelTransformParameterfvEXT'
	gl.glPixelTransformParameterfvEXT.errcheck = _errcheck
	gl.glPixelTransformParameterfvEXT.restype = GLvoid
	gl.glPixelTransformParameterfvEXT.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glPixelTransformParameteriEXT = _getProcAddress('glPixelTransformParameteriEXT')
	gl.glPixelTransformParameteriEXT.gl = gl
	gl.glPixelTransformParameteriEXT.__name__ = 'glPixelTransformParameteriEXT'
	gl.glPixelTransformParameteriEXT.errcheck = _errcheck
	gl.glPixelTransformParameteriEXT.restype = GLvoid
	gl.glPixelTransformParameteriEXT.argtypes = (GLenum, GLenum, GLint,)
	gl.glPixelTransformParameterivEXT = _getProcAddress('glPixelTransformParameterivEXT')
	gl.glPixelTransformParameterivEXT.gl = gl
	gl.glPixelTransformParameterivEXT.__name__ = 'glPixelTransformParameterivEXT'
	gl.glPixelTransformParameterivEXT.errcheck = _errcheck
	gl.glPixelTransformParameterivEXT.restype = GLvoid
	gl.glPixelTransformParameterivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glPixelZoom = _getProcAddress('glPixelZoom')
	gl.glPixelZoom.gl = gl
	gl.glPixelZoom.__name__ = 'glPixelZoom'
	gl.glPixelZoom.errcheck = _errcheck
	gl.glPixelZoom.restype = GLvoid
	gl.glPixelZoom.argtypes = (GLfloat, GLfloat,)
	gl.glPixelZoomxOES = _getProcAddress('glPixelZoomxOES')
	gl.glPixelZoomxOES.gl = gl
	gl.glPixelZoomxOES.__name__ = 'glPixelZoomxOES'
	gl.glPixelZoomxOES.errcheck = _errcheck
	gl.glPixelZoomxOES.restype = GLvoid
	gl.glPixelZoomxOES.argtypes = (GLfixed, GLfixed,)
	gl.glPointAlongPathNV = _getProcAddress('glPointAlongPathNV')
	gl.glPointAlongPathNV.gl = gl
	gl.glPointAlongPathNV.__name__ = 'glPointAlongPathNV'
	gl.glPointAlongPathNV.errcheck = _errcheck
	gl.glPointAlongPathNV.restype = GLboolean
	gl.glPointAlongPathNV.argtypes = (GLuint, GLsizei, GLsizei, GLfloat, POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glPointParameterf = _getProcAddress('glPointParameterf')
	gl.glPointParameterf.gl = gl
	gl.glPointParameterf.__name__ = 'glPointParameterf'
	gl.glPointParameterf.errcheck = _errcheck
	gl.glPointParameterf.restype = GLvoid
	gl.glPointParameterf.argtypes = (GLenum, GLfloat,)
	gl.glPointParameterfARB = _getProcAddress('glPointParameterfARB')
	gl.glPointParameterfARB.gl = gl
	gl.glPointParameterfARB.__name__ = 'glPointParameterfARB'
	gl.glPointParameterfARB.errcheck = _errcheck
	gl.glPointParameterfARB.restype = GLvoid
	gl.glPointParameterfARB.argtypes = (GLenum, GLfloat,)
	gl.glPointParameterfEXT = _getProcAddress('glPointParameterfEXT')
	gl.glPointParameterfEXT.gl = gl
	gl.glPointParameterfEXT.__name__ = 'glPointParameterfEXT'
	gl.glPointParameterfEXT.errcheck = _errcheck
	gl.glPointParameterfEXT.restype = GLvoid
	gl.glPointParameterfEXT.argtypes = (GLenum, GLfloat,)
	gl.glPointParameterfSGIS = _getProcAddress('glPointParameterfSGIS')
	gl.glPointParameterfSGIS.gl = gl
	gl.glPointParameterfSGIS.__name__ = 'glPointParameterfSGIS'
	gl.glPointParameterfSGIS.errcheck = _errcheck
	gl.glPointParameterfSGIS.restype = GLvoid
	gl.glPointParameterfSGIS.argtypes = (GLenum, GLfloat,)
	gl.glPointParameterfv = _getProcAddress('glPointParameterfv')
	gl.glPointParameterfv.gl = gl
	gl.glPointParameterfv.__name__ = 'glPointParameterfv'
	gl.glPointParameterfv.errcheck = _errcheck
	gl.glPointParameterfv.restype = GLvoid
	gl.glPointParameterfv.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glPointParameterfvARB = _getProcAddress('glPointParameterfvARB')
	gl.glPointParameterfvARB.gl = gl
	gl.glPointParameterfvARB.__name__ = 'glPointParameterfvARB'
	gl.glPointParameterfvARB.errcheck = _errcheck
	gl.glPointParameterfvARB.restype = GLvoid
	gl.glPointParameterfvARB.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glPointParameterfvEXT = _getProcAddress('glPointParameterfvEXT')
	gl.glPointParameterfvEXT.gl = gl
	gl.glPointParameterfvEXT.__name__ = 'glPointParameterfvEXT'
	gl.glPointParameterfvEXT.errcheck = _errcheck
	gl.glPointParameterfvEXT.restype = GLvoid
	gl.glPointParameterfvEXT.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glPointParameterfvSGIS = _getProcAddress('glPointParameterfvSGIS')
	gl.glPointParameterfvSGIS.gl = gl
	gl.glPointParameterfvSGIS.__name__ = 'glPointParameterfvSGIS'
	gl.glPointParameterfvSGIS.errcheck = _errcheck
	gl.glPointParameterfvSGIS.restype = GLvoid
	gl.glPointParameterfvSGIS.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glPointParameteri = _getProcAddress('glPointParameteri')
	gl.glPointParameteri.gl = gl
	gl.glPointParameteri.__name__ = 'glPointParameteri'
	gl.glPointParameteri.errcheck = _errcheck
	gl.glPointParameteri.restype = GLvoid
	gl.glPointParameteri.argtypes = (GLenum, GLint,)
	gl.glPointParameteriNV = _getProcAddress('glPointParameteriNV')
	gl.glPointParameteriNV.gl = gl
	gl.glPointParameteriNV.__name__ = 'glPointParameteriNV'
	gl.glPointParameteriNV.errcheck = _errcheck
	gl.glPointParameteriNV.restype = GLvoid
	gl.glPointParameteriNV.argtypes = (GLenum, GLint,)
	gl.glPointParameteriv = _getProcAddress('glPointParameteriv')
	gl.glPointParameteriv.gl = gl
	gl.glPointParameteriv.__name__ = 'glPointParameteriv'
	gl.glPointParameteriv.errcheck = _errcheck
	gl.glPointParameteriv.restype = GLvoid
	gl.glPointParameteriv.argtypes = (GLenum, POINTER(GLint),)
	gl.glPointParameterivNV = _getProcAddress('glPointParameterivNV')
	gl.glPointParameterivNV.gl = gl
	gl.glPointParameterivNV.__name__ = 'glPointParameterivNV'
	gl.glPointParameterivNV.errcheck = _errcheck
	gl.glPointParameterivNV.restype = GLvoid
	gl.glPointParameterivNV.argtypes = (GLenum, POINTER(GLint),)
	gl.glPointParameterx = _getProcAddress('glPointParameterx')
	gl.glPointParameterx.gl = gl
	gl.glPointParameterx.__name__ = 'glPointParameterx'
	gl.glPointParameterx.errcheck = _errcheck
	gl.glPointParameterx.restype = GLvoid
	gl.glPointParameterx.argtypes = (GLenum, GLfixed,)
	gl.glPointParameterxOES = _getProcAddress('glPointParameterxOES')
	gl.glPointParameterxOES.gl = gl
	gl.glPointParameterxOES.__name__ = 'glPointParameterxOES'
	gl.glPointParameterxOES.errcheck = _errcheck
	gl.glPointParameterxOES.restype = GLvoid
	gl.glPointParameterxOES.argtypes = (GLenum, GLfixed,)
	gl.glPointParameterxv = _getProcAddress('glPointParameterxv')
	gl.glPointParameterxv.gl = gl
	gl.glPointParameterxv.__name__ = 'glPointParameterxv'
	gl.glPointParameterxv.errcheck = _errcheck
	gl.glPointParameterxv.restype = GLvoid
	gl.glPointParameterxv.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glPointParameterxvOES = _getProcAddress('glPointParameterxvOES')
	gl.glPointParameterxvOES.gl = gl
	gl.glPointParameterxvOES.__name__ = 'glPointParameterxvOES'
	gl.glPointParameterxvOES.errcheck = _errcheck
	gl.glPointParameterxvOES.restype = GLvoid
	gl.glPointParameterxvOES.argtypes = (GLenum, POINTER(GLfixed),)
	gl.glPointSize = _getProcAddress('glPointSize')
	gl.glPointSize.gl = gl
	gl.glPointSize.__name__ = 'glPointSize'
	gl.glPointSize.errcheck = _errcheck
	gl.glPointSize.restype = GLvoid
	gl.glPointSize.argtypes = (GLfloat,)
	gl.glPointSizePointerOES = _getProcAddress('glPointSizePointerOES')
	gl.glPointSizePointerOES.gl = gl
	gl.glPointSizePointerOES.__name__ = 'glPointSizePointerOES'
	gl.glPointSizePointerOES.errcheck = _errcheck
	gl.glPointSizePointerOES.restype = GLvoid
	gl.glPointSizePointerOES.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glPointSizex = _getProcAddress('glPointSizex')
	gl.glPointSizex.gl = gl
	gl.glPointSizex.__name__ = 'glPointSizex'
	gl.glPointSizex.errcheck = _errcheck
	gl.glPointSizex.restype = GLvoid
	gl.glPointSizex.argtypes = (GLfixed,)
	gl.glPointSizexOES = _getProcAddress('glPointSizexOES')
	gl.glPointSizexOES.gl = gl
	gl.glPointSizexOES.__name__ = 'glPointSizexOES'
	gl.glPointSizexOES.errcheck = _errcheck
	gl.glPointSizexOES.restype = GLvoid
	gl.glPointSizexOES.argtypes = (GLfixed,)
	gl.glPollAsyncSGIX = _getProcAddress('glPollAsyncSGIX')
	gl.glPollAsyncSGIX.gl = gl
	gl.glPollAsyncSGIX.__name__ = 'glPollAsyncSGIX'
	gl.glPollAsyncSGIX.errcheck = _errcheck
	gl.glPollAsyncSGIX.restype = GLint
	gl.glPollAsyncSGIX.argtypes = (POINTER(GLuint),)
	gl.glPollInstrumentsSGIX = _getProcAddress('glPollInstrumentsSGIX')
	gl.glPollInstrumentsSGIX.gl = gl
	gl.glPollInstrumentsSGIX.__name__ = 'glPollInstrumentsSGIX'
	gl.glPollInstrumentsSGIX.errcheck = _errcheck
	gl.glPollInstrumentsSGIX.restype = GLint
	gl.glPollInstrumentsSGIX.argtypes = (POINTER(GLint),)
	gl.glPolygonMode = _getProcAddress('glPolygonMode')
	gl.glPolygonMode.gl = gl
	gl.glPolygonMode.__name__ = 'glPolygonMode'
	gl.glPolygonMode.errcheck = _errcheck
	gl.glPolygonMode.restype = GLvoid
	gl.glPolygonMode.argtypes = (GLenum, GLenum,)
	gl.glPolygonOffset = _getProcAddress('glPolygonOffset')
	gl.glPolygonOffset.gl = gl
	gl.glPolygonOffset.__name__ = 'glPolygonOffset'
	gl.glPolygonOffset.errcheck = _errcheck
	gl.glPolygonOffset.restype = GLvoid
	gl.glPolygonOffset.argtypes = (GLfloat, GLfloat,)
	gl.glPolygonOffsetEXT = _getProcAddress('glPolygonOffsetEXT')
	gl.glPolygonOffsetEXT.gl = gl
	gl.glPolygonOffsetEXT.__name__ = 'glPolygonOffsetEXT'
	gl.glPolygonOffsetEXT.errcheck = _errcheck
	gl.glPolygonOffsetEXT.restype = GLvoid
	gl.glPolygonOffsetEXT.argtypes = (GLfloat, GLfloat,)
	gl.glPolygonOffsetx = _getProcAddress('glPolygonOffsetx')
	gl.glPolygonOffsetx.gl = gl
	gl.glPolygonOffsetx.__name__ = 'glPolygonOffsetx'
	gl.glPolygonOffsetx.errcheck = _errcheck
	gl.glPolygonOffsetx.restype = GLvoid
	gl.glPolygonOffsetx.argtypes = (GLfixed, GLfixed,)
	gl.glPolygonOffsetxOES = _getProcAddress('glPolygonOffsetxOES')
	gl.glPolygonOffsetxOES.gl = gl
	gl.glPolygonOffsetxOES.__name__ = 'glPolygonOffsetxOES'
	gl.glPolygonOffsetxOES.errcheck = _errcheck
	gl.glPolygonOffsetxOES.restype = GLvoid
	gl.glPolygonOffsetxOES.argtypes = (GLfixed, GLfixed,)
	gl.glPolygonOffsetClampEXT = _getProcAddress('glPolygonOffsetClampEXT')
	gl.glPolygonOffsetClampEXT.gl = gl
	gl.glPolygonOffsetClampEXT.__name__ = 'glPolygonOffsetClampEXT'
	gl.glPolygonOffsetClampEXT.errcheck = _errcheck
	gl.glPolygonOffsetClampEXT.restype = GLvoid
	gl.glPolygonOffsetClampEXT.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glPolygonStipple = _getProcAddress('glPolygonStipple')
	gl.glPolygonStipple.gl = gl
	gl.glPolygonStipple.__name__ = 'glPolygonStipple'
	gl.glPolygonStipple.errcheck = _errcheck
	gl.glPolygonStipple.restype = GLvoid
	gl.glPolygonStipple.argtypes = (POINTER(GLubyte),)
	gl.glPopAttrib = _getProcAddress('glPopAttrib')
	gl.glPopAttrib.gl = gl
	gl.glPopAttrib.__name__ = 'glPopAttrib'
	gl.glPopAttrib.errcheck = _errcheck
	gl.glPopAttrib.restype = GLvoid
	gl.glPopAttrib.argtypes = ()
	gl.glPopClientAttrib = _getProcAddress('glPopClientAttrib')
	gl.glPopClientAttrib.gl = gl
	gl.glPopClientAttrib.__name__ = 'glPopClientAttrib'
	gl.glPopClientAttrib.errcheck = _errcheck
	gl.glPopClientAttrib.restype = GLvoid
	gl.glPopClientAttrib.argtypes = ()
	gl.glPopDebugGroup = _getProcAddress('glPopDebugGroup')
	gl.glPopDebugGroup.gl = gl
	gl.glPopDebugGroup.__name__ = 'glPopDebugGroup'
	gl.glPopDebugGroup.errcheck = _errcheck
	gl.glPopDebugGroup.restype = GLvoid
	gl.glPopDebugGroup.argtypes = ()
	gl.glPopDebugGroupKHR = _getProcAddress('glPopDebugGroupKHR')
	gl.glPopDebugGroupKHR.gl = gl
	gl.glPopDebugGroupKHR.__name__ = 'glPopDebugGroupKHR'
	gl.glPopDebugGroupKHR.errcheck = _errcheck
	gl.glPopDebugGroupKHR.restype = GLvoid
	gl.glPopDebugGroupKHR.argtypes = ()
	gl.glPopGroupMarkerEXT = _getProcAddress('glPopGroupMarkerEXT')
	gl.glPopGroupMarkerEXT.gl = gl
	gl.glPopGroupMarkerEXT.__name__ = 'glPopGroupMarkerEXT'
	gl.glPopGroupMarkerEXT.errcheck = _errcheck
	gl.glPopGroupMarkerEXT.restype = GLvoid
	gl.glPopGroupMarkerEXT.argtypes = ()
	gl.glPopMatrix = _getProcAddress('glPopMatrix')
	gl.glPopMatrix.gl = gl
	gl.glPopMatrix.__name__ = 'glPopMatrix'
	gl.glPopMatrix.errcheck = _errcheck
	gl.glPopMatrix.restype = GLvoid
	gl.glPopMatrix.argtypes = ()
	gl.glPopName = _getProcAddress('glPopName')
	gl.glPopName.gl = gl
	gl.glPopName.__name__ = 'glPopName'
	gl.glPopName.errcheck = _errcheck
	gl.glPopName.restype = GLvoid
	gl.glPopName.argtypes = ()
	gl.glPresentFrameDualFillNV = _getProcAddress('glPresentFrameDualFillNV')
	gl.glPresentFrameDualFillNV.gl = gl
	gl.glPresentFrameDualFillNV.__name__ = 'glPresentFrameDualFillNV'
	gl.glPresentFrameDualFillNV.errcheck = _errcheck
	gl.glPresentFrameDualFillNV.restype = GLvoid
	gl.glPresentFrameDualFillNV.argtypes = (GLuint, GLuint64EXT, GLuint, GLuint, GLenum, GLenum, GLuint, GLenum, GLuint, GLenum, GLuint, GLenum, GLuint,)
	gl.glPresentFrameKeyedNV = _getProcAddress('glPresentFrameKeyedNV')
	gl.glPresentFrameKeyedNV.gl = gl
	gl.glPresentFrameKeyedNV.__name__ = 'glPresentFrameKeyedNV'
	gl.glPresentFrameKeyedNV.errcheck = _errcheck
	gl.glPresentFrameKeyedNV.restype = GLvoid
	gl.glPresentFrameKeyedNV.argtypes = (GLuint, GLuint64EXT, GLuint, GLuint, GLenum, GLenum, GLuint, GLuint, GLenum, GLuint, GLuint,)
	gl.glPrimitiveBoundingBoxEXT = _getProcAddress('glPrimitiveBoundingBoxEXT')
	gl.glPrimitiveBoundingBoxEXT.gl = gl
	gl.glPrimitiveBoundingBoxEXT.__name__ = 'glPrimitiveBoundingBoxEXT'
	gl.glPrimitiveBoundingBoxEXT.errcheck = _errcheck
	gl.glPrimitiveBoundingBoxEXT.restype = GLvoid
	gl.glPrimitiveBoundingBoxEXT.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glPrimitiveRestartIndex = _getProcAddress('glPrimitiveRestartIndex')
	gl.glPrimitiveRestartIndex.gl = gl
	gl.glPrimitiveRestartIndex.__name__ = 'glPrimitiveRestartIndex'
	gl.glPrimitiveRestartIndex.errcheck = _errcheck
	gl.glPrimitiveRestartIndex.restype = GLvoid
	gl.glPrimitiveRestartIndex.argtypes = (GLuint,)
	gl.glPrimitiveRestartIndexNV = _getProcAddress('glPrimitiveRestartIndexNV')
	gl.glPrimitiveRestartIndexNV.gl = gl
	gl.glPrimitiveRestartIndexNV.__name__ = 'glPrimitiveRestartIndexNV'
	gl.glPrimitiveRestartIndexNV.errcheck = _errcheck
	gl.glPrimitiveRestartIndexNV.restype = GLvoid
	gl.glPrimitiveRestartIndexNV.argtypes = (GLuint,)
	gl.glPrimitiveRestartNV = _getProcAddress('glPrimitiveRestartNV')
	gl.glPrimitiveRestartNV.gl = gl
	gl.glPrimitiveRestartNV.__name__ = 'glPrimitiveRestartNV'
	gl.glPrimitiveRestartNV.errcheck = _errcheck
	gl.glPrimitiveRestartNV.restype = GLvoid
	gl.glPrimitiveRestartNV.argtypes = ()
	gl.glPrioritizeTextures = _getProcAddress('glPrioritizeTextures')
	gl.glPrioritizeTextures.gl = gl
	gl.glPrioritizeTextures.__name__ = 'glPrioritizeTextures'
	gl.glPrioritizeTextures.errcheck = _errcheck
	gl.glPrioritizeTextures.restype = GLvoid
	gl.glPrioritizeTextures.argtypes = (GLsizei, POINTER(GLuint), POINTER(GLfloat),)
	gl.glPrioritizeTexturesEXT = _getProcAddress('glPrioritizeTexturesEXT')
	gl.glPrioritizeTexturesEXT.gl = gl
	gl.glPrioritizeTexturesEXT.__name__ = 'glPrioritizeTexturesEXT'
	gl.glPrioritizeTexturesEXT.errcheck = _errcheck
	gl.glPrioritizeTexturesEXT.restype = GLvoid
	gl.glPrioritizeTexturesEXT.argtypes = (GLsizei, POINTER(GLuint), POINTER(GLclampf),)
	gl.glPrioritizeTexturesxOES = _getProcAddress('glPrioritizeTexturesxOES')
	gl.glPrioritizeTexturesxOES.gl = gl
	gl.glPrioritizeTexturesxOES.__name__ = 'glPrioritizeTexturesxOES'
	gl.glPrioritizeTexturesxOES.errcheck = _errcheck
	gl.glPrioritizeTexturesxOES.restype = GLvoid
	gl.glPrioritizeTexturesxOES.argtypes = (GLsizei, POINTER(GLuint), POINTER(GLfixed),)
	gl.glProgramBinary = _getProcAddress('glProgramBinary')
	gl.glProgramBinary.gl = gl
	gl.glProgramBinary.__name__ = 'glProgramBinary'
	gl.glProgramBinary.errcheck = _errcheck
	gl.glProgramBinary.restype = GLvoid
	gl.glProgramBinary.argtypes = (GLuint, GLenum, POINTER(GLvoid), GLsizei,)
	gl.glProgramBinaryOES = _getProcAddress('glProgramBinaryOES')
	gl.glProgramBinaryOES.gl = gl
	gl.glProgramBinaryOES.__name__ = 'glProgramBinaryOES'
	gl.glProgramBinaryOES.errcheck = _errcheck
	gl.glProgramBinaryOES.restype = GLvoid
	gl.glProgramBinaryOES.argtypes = (GLuint, GLenum, POINTER(GLvoid), GLint,)
	gl.glProgramBufferParametersIivNV = _getProcAddress('glProgramBufferParametersIivNV')
	gl.glProgramBufferParametersIivNV.gl = gl
	gl.glProgramBufferParametersIivNV.__name__ = 'glProgramBufferParametersIivNV'
	gl.glProgramBufferParametersIivNV.errcheck = _errcheck
	gl.glProgramBufferParametersIivNV.restype = GLvoid
	gl.glProgramBufferParametersIivNV.argtypes = (GLenum, GLuint, GLuint, GLsizei, POINTER(GLint),)
	gl.glProgramBufferParametersIuivNV = _getProcAddress('glProgramBufferParametersIuivNV')
	gl.glProgramBufferParametersIuivNV.gl = gl
	gl.glProgramBufferParametersIuivNV.__name__ = 'glProgramBufferParametersIuivNV'
	gl.glProgramBufferParametersIuivNV.errcheck = _errcheck
	gl.glProgramBufferParametersIuivNV.restype = GLvoid
	gl.glProgramBufferParametersIuivNV.argtypes = (GLenum, GLuint, GLuint, GLsizei, POINTER(GLuint),)
	gl.glProgramBufferParametersfvNV = _getProcAddress('glProgramBufferParametersfvNV')
	gl.glProgramBufferParametersfvNV.gl = gl
	gl.glProgramBufferParametersfvNV.__name__ = 'glProgramBufferParametersfvNV'
	gl.glProgramBufferParametersfvNV.errcheck = _errcheck
	gl.glProgramBufferParametersfvNV.restype = GLvoid
	gl.glProgramBufferParametersfvNV.argtypes = (GLenum, GLuint, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glProgramEnvParameter4dARB = _getProcAddress('glProgramEnvParameter4dARB')
	gl.glProgramEnvParameter4dARB.gl = gl
	gl.glProgramEnvParameter4dARB.__name__ = 'glProgramEnvParameter4dARB'
	gl.glProgramEnvParameter4dARB.errcheck = _errcheck
	gl.glProgramEnvParameter4dARB.restype = GLvoid
	gl.glProgramEnvParameter4dARB.argtypes = (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glProgramEnvParameter4dvARB = _getProcAddress('glProgramEnvParameter4dvARB')
	gl.glProgramEnvParameter4dvARB.gl = gl
	gl.glProgramEnvParameter4dvARB.__name__ = 'glProgramEnvParameter4dvARB'
	gl.glProgramEnvParameter4dvARB.errcheck = _errcheck
	gl.glProgramEnvParameter4dvARB.restype = GLvoid
	gl.glProgramEnvParameter4dvARB.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glProgramEnvParameter4fARB = _getProcAddress('glProgramEnvParameter4fARB')
	gl.glProgramEnvParameter4fARB.gl = gl
	gl.glProgramEnvParameter4fARB.__name__ = 'glProgramEnvParameter4fARB'
	gl.glProgramEnvParameter4fARB.errcheck = _errcheck
	gl.glProgramEnvParameter4fARB.restype = GLvoid
	gl.glProgramEnvParameter4fARB.argtypes = (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glProgramEnvParameter4fvARB = _getProcAddress('glProgramEnvParameter4fvARB')
	gl.glProgramEnvParameter4fvARB.gl = gl
	gl.glProgramEnvParameter4fvARB.__name__ = 'glProgramEnvParameter4fvARB'
	gl.glProgramEnvParameter4fvARB.errcheck = _errcheck
	gl.glProgramEnvParameter4fvARB.restype = GLvoid
	gl.glProgramEnvParameter4fvARB.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glProgramEnvParameterI4iNV = _getProcAddress('glProgramEnvParameterI4iNV')
	gl.glProgramEnvParameterI4iNV.gl = gl
	gl.glProgramEnvParameterI4iNV.__name__ = 'glProgramEnvParameterI4iNV'
	gl.glProgramEnvParameterI4iNV.errcheck = _errcheck
	gl.glProgramEnvParameterI4iNV.restype = GLvoid
	gl.glProgramEnvParameterI4iNV.argtypes = (GLenum, GLuint, GLint, GLint, GLint, GLint,)
	gl.glProgramEnvParameterI4ivNV = _getProcAddress('glProgramEnvParameterI4ivNV')
	gl.glProgramEnvParameterI4ivNV.gl = gl
	gl.glProgramEnvParameterI4ivNV.__name__ = 'glProgramEnvParameterI4ivNV'
	gl.glProgramEnvParameterI4ivNV.errcheck = _errcheck
	gl.glProgramEnvParameterI4ivNV.restype = GLvoid
	gl.glProgramEnvParameterI4ivNV.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glProgramEnvParameterI4uiNV = _getProcAddress('glProgramEnvParameterI4uiNV')
	gl.glProgramEnvParameterI4uiNV.gl = gl
	gl.glProgramEnvParameterI4uiNV.__name__ = 'glProgramEnvParameterI4uiNV'
	gl.glProgramEnvParameterI4uiNV.errcheck = _errcheck
	gl.glProgramEnvParameterI4uiNV.restype = GLvoid
	gl.glProgramEnvParameterI4uiNV.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glProgramEnvParameterI4uivNV = _getProcAddress('glProgramEnvParameterI4uivNV')
	gl.glProgramEnvParameterI4uivNV.gl = gl
	gl.glProgramEnvParameterI4uivNV.__name__ = 'glProgramEnvParameterI4uivNV'
	gl.glProgramEnvParameterI4uivNV.errcheck = _errcheck
	gl.glProgramEnvParameterI4uivNV.restype = GLvoid
	gl.glProgramEnvParameterI4uivNV.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glProgramEnvParameters4fvEXT = _getProcAddress('glProgramEnvParameters4fvEXT')
	gl.glProgramEnvParameters4fvEXT.gl = gl
	gl.glProgramEnvParameters4fvEXT.__name__ = 'glProgramEnvParameters4fvEXT'
	gl.glProgramEnvParameters4fvEXT.errcheck = _errcheck
	gl.glProgramEnvParameters4fvEXT.restype = GLvoid
	gl.glProgramEnvParameters4fvEXT.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glProgramEnvParametersI4ivNV = _getProcAddress('glProgramEnvParametersI4ivNV')
	gl.glProgramEnvParametersI4ivNV.gl = gl
	gl.glProgramEnvParametersI4ivNV.__name__ = 'glProgramEnvParametersI4ivNV'
	gl.glProgramEnvParametersI4ivNV.errcheck = _errcheck
	gl.glProgramEnvParametersI4ivNV.restype = GLvoid
	gl.glProgramEnvParametersI4ivNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLint),)
	gl.glProgramEnvParametersI4uivNV = _getProcAddress('glProgramEnvParametersI4uivNV')
	gl.glProgramEnvParametersI4uivNV.gl = gl
	gl.glProgramEnvParametersI4uivNV.__name__ = 'glProgramEnvParametersI4uivNV'
	gl.glProgramEnvParametersI4uivNV.errcheck = _errcheck
	gl.glProgramEnvParametersI4uivNV.restype = GLvoid
	gl.glProgramEnvParametersI4uivNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLuint),)
	gl.glProgramLocalParameter4dARB = _getProcAddress('glProgramLocalParameter4dARB')
	gl.glProgramLocalParameter4dARB.gl = gl
	gl.glProgramLocalParameter4dARB.__name__ = 'glProgramLocalParameter4dARB'
	gl.glProgramLocalParameter4dARB.errcheck = _errcheck
	gl.glProgramLocalParameter4dARB.restype = GLvoid
	gl.glProgramLocalParameter4dARB.argtypes = (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glProgramLocalParameter4dvARB = _getProcAddress('glProgramLocalParameter4dvARB')
	gl.glProgramLocalParameter4dvARB.gl = gl
	gl.glProgramLocalParameter4dvARB.__name__ = 'glProgramLocalParameter4dvARB'
	gl.glProgramLocalParameter4dvARB.errcheck = _errcheck
	gl.glProgramLocalParameter4dvARB.restype = GLvoid
	gl.glProgramLocalParameter4dvARB.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glProgramLocalParameter4fARB = _getProcAddress('glProgramLocalParameter4fARB')
	gl.glProgramLocalParameter4fARB.gl = gl
	gl.glProgramLocalParameter4fARB.__name__ = 'glProgramLocalParameter4fARB'
	gl.glProgramLocalParameter4fARB.errcheck = _errcheck
	gl.glProgramLocalParameter4fARB.restype = GLvoid
	gl.glProgramLocalParameter4fARB.argtypes = (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glProgramLocalParameter4fvARB = _getProcAddress('glProgramLocalParameter4fvARB')
	gl.glProgramLocalParameter4fvARB.gl = gl
	gl.glProgramLocalParameter4fvARB.__name__ = 'glProgramLocalParameter4fvARB'
	gl.glProgramLocalParameter4fvARB.errcheck = _errcheck
	gl.glProgramLocalParameter4fvARB.restype = GLvoid
	gl.glProgramLocalParameter4fvARB.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glProgramLocalParameterI4iNV = _getProcAddress('glProgramLocalParameterI4iNV')
	gl.glProgramLocalParameterI4iNV.gl = gl
	gl.glProgramLocalParameterI4iNV.__name__ = 'glProgramLocalParameterI4iNV'
	gl.glProgramLocalParameterI4iNV.errcheck = _errcheck
	gl.glProgramLocalParameterI4iNV.restype = GLvoid
	gl.glProgramLocalParameterI4iNV.argtypes = (GLenum, GLuint, GLint, GLint, GLint, GLint,)
	gl.glProgramLocalParameterI4ivNV = _getProcAddress('glProgramLocalParameterI4ivNV')
	gl.glProgramLocalParameterI4ivNV.gl = gl
	gl.glProgramLocalParameterI4ivNV.__name__ = 'glProgramLocalParameterI4ivNV'
	gl.glProgramLocalParameterI4ivNV.errcheck = _errcheck
	gl.glProgramLocalParameterI4ivNV.restype = GLvoid
	gl.glProgramLocalParameterI4ivNV.argtypes = (GLenum, GLuint, POINTER(GLint),)
	gl.glProgramLocalParameterI4uiNV = _getProcAddress('glProgramLocalParameterI4uiNV')
	gl.glProgramLocalParameterI4uiNV.gl = gl
	gl.glProgramLocalParameterI4uiNV.__name__ = 'glProgramLocalParameterI4uiNV'
	gl.glProgramLocalParameterI4uiNV.errcheck = _errcheck
	gl.glProgramLocalParameterI4uiNV.restype = GLvoid
	gl.glProgramLocalParameterI4uiNV.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glProgramLocalParameterI4uivNV = _getProcAddress('glProgramLocalParameterI4uivNV')
	gl.glProgramLocalParameterI4uivNV.gl = gl
	gl.glProgramLocalParameterI4uivNV.__name__ = 'glProgramLocalParameterI4uivNV'
	gl.glProgramLocalParameterI4uivNV.errcheck = _errcheck
	gl.glProgramLocalParameterI4uivNV.restype = GLvoid
	gl.glProgramLocalParameterI4uivNV.argtypes = (GLenum, GLuint, POINTER(GLuint),)
	gl.glProgramLocalParameters4fvEXT = _getProcAddress('glProgramLocalParameters4fvEXT')
	gl.glProgramLocalParameters4fvEXT.gl = gl
	gl.glProgramLocalParameters4fvEXT.__name__ = 'glProgramLocalParameters4fvEXT'
	gl.glProgramLocalParameters4fvEXT.errcheck = _errcheck
	gl.glProgramLocalParameters4fvEXT.restype = GLvoid
	gl.glProgramLocalParameters4fvEXT.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glProgramLocalParametersI4ivNV = _getProcAddress('glProgramLocalParametersI4ivNV')
	gl.glProgramLocalParametersI4ivNV.gl = gl
	gl.glProgramLocalParametersI4ivNV.__name__ = 'glProgramLocalParametersI4ivNV'
	gl.glProgramLocalParametersI4ivNV.errcheck = _errcheck
	gl.glProgramLocalParametersI4ivNV.restype = GLvoid
	gl.glProgramLocalParametersI4ivNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLint),)
	gl.glProgramLocalParametersI4uivNV = _getProcAddress('glProgramLocalParametersI4uivNV')
	gl.glProgramLocalParametersI4uivNV.gl = gl
	gl.glProgramLocalParametersI4uivNV.__name__ = 'glProgramLocalParametersI4uivNV'
	gl.glProgramLocalParametersI4uivNV.errcheck = _errcheck
	gl.glProgramLocalParametersI4uivNV.restype = GLvoid
	gl.glProgramLocalParametersI4uivNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLuint),)
	gl.glProgramNamedParameter4dNV = _getProcAddress('glProgramNamedParameter4dNV')
	gl.glProgramNamedParameter4dNV.gl = gl
	gl.glProgramNamedParameter4dNV.__name__ = 'glProgramNamedParameter4dNV'
	gl.glProgramNamedParameter4dNV.errcheck = _errcheck
	gl.glProgramNamedParameter4dNV.restype = GLvoid
	gl.glProgramNamedParameter4dNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glProgramNamedParameter4dvNV = _getProcAddress('glProgramNamedParameter4dvNV')
	gl.glProgramNamedParameter4dvNV.gl = gl
	gl.glProgramNamedParameter4dvNV.__name__ = 'glProgramNamedParameter4dvNV'
	gl.glProgramNamedParameter4dvNV.errcheck = _errcheck
	gl.glProgramNamedParameter4dvNV.restype = GLvoid
	gl.glProgramNamedParameter4dvNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), POINTER(GLdouble),)
	gl.glProgramNamedParameter4fNV = _getProcAddress('glProgramNamedParameter4fNV')
	gl.glProgramNamedParameter4fNV.gl = gl
	gl.glProgramNamedParameter4fNV.__name__ = 'glProgramNamedParameter4fNV'
	gl.glProgramNamedParameter4fNV.errcheck = _errcheck
	gl.glProgramNamedParameter4fNV.restype = GLvoid
	gl.glProgramNamedParameter4fNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glProgramNamedParameter4fvNV = _getProcAddress('glProgramNamedParameter4fvNV')
	gl.glProgramNamedParameter4fvNV.gl = gl
	gl.glProgramNamedParameter4fvNV.__name__ = 'glProgramNamedParameter4fvNV'
	gl.glProgramNamedParameter4fvNV.errcheck = _errcheck
	gl.glProgramNamedParameter4fvNV.restype = GLvoid
	gl.glProgramNamedParameter4fvNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte), POINTER(GLfloat),)
	gl.glProgramParameter4dNV = _getProcAddress('glProgramParameter4dNV')
	gl.glProgramParameter4dNV.gl = gl
	gl.glProgramParameter4dNV.__name__ = 'glProgramParameter4dNV'
	gl.glProgramParameter4dNV.errcheck = _errcheck
	gl.glProgramParameter4dNV.restype = GLvoid
	gl.glProgramParameter4dNV.argtypes = (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glProgramParameter4dvNV = _getProcAddress('glProgramParameter4dvNV')
	gl.glProgramParameter4dvNV.gl = gl
	gl.glProgramParameter4dvNV.__name__ = 'glProgramParameter4dvNV'
	gl.glProgramParameter4dvNV.errcheck = _errcheck
	gl.glProgramParameter4dvNV.restype = GLvoid
	gl.glProgramParameter4dvNV.argtypes = (GLenum, GLuint, POINTER(GLdouble),)
	gl.glProgramParameter4fNV = _getProcAddress('glProgramParameter4fNV')
	gl.glProgramParameter4fNV.gl = gl
	gl.glProgramParameter4fNV.__name__ = 'glProgramParameter4fNV'
	gl.glProgramParameter4fNV.errcheck = _errcheck
	gl.glProgramParameter4fNV.restype = GLvoid
	gl.glProgramParameter4fNV.argtypes = (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glProgramParameter4fvNV = _getProcAddress('glProgramParameter4fvNV')
	gl.glProgramParameter4fvNV.gl = gl
	gl.glProgramParameter4fvNV.__name__ = 'glProgramParameter4fvNV'
	gl.glProgramParameter4fvNV.errcheck = _errcheck
	gl.glProgramParameter4fvNV.restype = GLvoid
	gl.glProgramParameter4fvNV.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glProgramParameteri = _getProcAddress('glProgramParameteri')
	gl.glProgramParameteri.gl = gl
	gl.glProgramParameteri.__name__ = 'glProgramParameteri'
	gl.glProgramParameteri.errcheck = _errcheck
	gl.glProgramParameteri.restype = GLvoid
	gl.glProgramParameteri.argtypes = (GLuint, GLenum, GLint,)
	gl.glProgramParameteriARB = _getProcAddress('glProgramParameteriARB')
	gl.glProgramParameteriARB.gl = gl
	gl.glProgramParameteriARB.__name__ = 'glProgramParameteriARB'
	gl.glProgramParameteriARB.errcheck = _errcheck
	gl.glProgramParameteriARB.restype = GLvoid
	gl.glProgramParameteriARB.argtypes = (GLuint, GLenum, GLint,)
	gl.glProgramParameteriEXT = _getProcAddress('glProgramParameteriEXT')
	gl.glProgramParameteriEXT.gl = gl
	gl.glProgramParameteriEXT.__name__ = 'glProgramParameteriEXT'
	gl.glProgramParameteriEXT.errcheck = _errcheck
	gl.glProgramParameteriEXT.restype = GLvoid
	gl.glProgramParameteriEXT.argtypes = (GLuint, GLenum, GLint,)
	gl.glProgramParameters4dvNV = _getProcAddress('glProgramParameters4dvNV')
	gl.glProgramParameters4dvNV.gl = gl
	gl.glProgramParameters4dvNV.__name__ = 'glProgramParameters4dvNV'
	gl.glProgramParameters4dvNV.errcheck = _errcheck
	gl.glProgramParameters4dvNV.restype = GLvoid
	gl.glProgramParameters4dvNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLdouble),)
	gl.glProgramParameters4fvNV = _getProcAddress('glProgramParameters4fvNV')
	gl.glProgramParameters4fvNV.gl = gl
	gl.glProgramParameters4fvNV.__name__ = 'glProgramParameters4fvNV'
	gl.glProgramParameters4fvNV.errcheck = _errcheck
	gl.glProgramParameters4fvNV.restype = GLvoid
	gl.glProgramParameters4fvNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glProgramPathFragmentInputGenNV = _getProcAddress('glProgramPathFragmentInputGenNV')
	gl.glProgramPathFragmentInputGenNV.gl = gl
	gl.glProgramPathFragmentInputGenNV.__name__ = 'glProgramPathFragmentInputGenNV'
	gl.glProgramPathFragmentInputGenNV.errcheck = _errcheck
	gl.glProgramPathFragmentInputGenNV.restype = GLvoid
	gl.glProgramPathFragmentInputGenNV.argtypes = (GLuint, GLint, GLenum, GLint, POINTER(GLfloat),)
	gl.glProgramStringARB = _getProcAddress('glProgramStringARB')
	gl.glProgramStringARB.gl = gl
	gl.glProgramStringARB.__name__ = 'glProgramStringARB'
	gl.glProgramStringARB.errcheck = _errcheck
	gl.glProgramStringARB.restype = GLvoid
	gl.glProgramStringARB.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glProgramSubroutineParametersuivNV = _getProcAddress('glProgramSubroutineParametersuivNV')
	gl.glProgramSubroutineParametersuivNV.gl = gl
	gl.glProgramSubroutineParametersuivNV.__name__ = 'glProgramSubroutineParametersuivNV'
	gl.glProgramSubroutineParametersuivNV.errcheck = _errcheck
	gl.glProgramSubroutineParametersuivNV.restype = GLvoid
	gl.glProgramSubroutineParametersuivNV.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform1d = _getProcAddress('glProgramUniform1d')
	gl.glProgramUniform1d.gl = gl
	gl.glProgramUniform1d.__name__ = 'glProgramUniform1d'
	gl.glProgramUniform1d.errcheck = _errcheck
	gl.glProgramUniform1d.restype = GLvoid
	gl.glProgramUniform1d.argtypes = (GLuint, GLint, GLdouble,)
	gl.glProgramUniform1dEXT = _getProcAddress('glProgramUniform1dEXT')
	gl.glProgramUniform1dEXT.gl = gl
	gl.glProgramUniform1dEXT.__name__ = 'glProgramUniform1dEXT'
	gl.glProgramUniform1dEXT.errcheck = _errcheck
	gl.glProgramUniform1dEXT.restype = GLvoid
	gl.glProgramUniform1dEXT.argtypes = (GLuint, GLint, GLdouble,)
	gl.glProgramUniform1dv = _getProcAddress('glProgramUniform1dv')
	gl.glProgramUniform1dv.gl = gl
	gl.glProgramUniform1dv.__name__ = 'glProgramUniform1dv'
	gl.glProgramUniform1dv.errcheck = _errcheck
	gl.glProgramUniform1dv.restype = GLvoid
	gl.glProgramUniform1dv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform1dvEXT = _getProcAddress('glProgramUniform1dvEXT')
	gl.glProgramUniform1dvEXT.gl = gl
	gl.glProgramUniform1dvEXT.__name__ = 'glProgramUniform1dvEXT'
	gl.glProgramUniform1dvEXT.errcheck = _errcheck
	gl.glProgramUniform1dvEXT.restype = GLvoid
	gl.glProgramUniform1dvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform1f = _getProcAddress('glProgramUniform1f')
	gl.glProgramUniform1f.gl = gl
	gl.glProgramUniform1f.__name__ = 'glProgramUniform1f'
	gl.glProgramUniform1f.errcheck = _errcheck
	gl.glProgramUniform1f.restype = GLvoid
	gl.glProgramUniform1f.argtypes = (GLuint, GLint, GLfloat,)
	gl.glProgramUniform1fEXT = _getProcAddress('glProgramUniform1fEXT')
	gl.glProgramUniform1fEXT.gl = gl
	gl.glProgramUniform1fEXT.__name__ = 'glProgramUniform1fEXT'
	gl.glProgramUniform1fEXT.errcheck = _errcheck
	gl.glProgramUniform1fEXT.restype = GLvoid
	gl.glProgramUniform1fEXT.argtypes = (GLuint, GLint, GLfloat,)
	gl.glProgramUniform1fv = _getProcAddress('glProgramUniform1fv')
	gl.glProgramUniform1fv.gl = gl
	gl.glProgramUniform1fv.__name__ = 'glProgramUniform1fv'
	gl.glProgramUniform1fv.errcheck = _errcheck
	gl.glProgramUniform1fv.restype = GLvoid
	gl.glProgramUniform1fv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform1fvEXT = _getProcAddress('glProgramUniform1fvEXT')
	gl.glProgramUniform1fvEXT.gl = gl
	gl.glProgramUniform1fvEXT.__name__ = 'glProgramUniform1fvEXT'
	gl.glProgramUniform1fvEXT.errcheck = _errcheck
	gl.glProgramUniform1fvEXT.restype = GLvoid
	gl.glProgramUniform1fvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform1i = _getProcAddress('glProgramUniform1i')
	gl.glProgramUniform1i.gl = gl
	gl.glProgramUniform1i.__name__ = 'glProgramUniform1i'
	gl.glProgramUniform1i.errcheck = _errcheck
	gl.glProgramUniform1i.restype = GLvoid
	gl.glProgramUniform1i.argtypes = (GLuint, GLint, GLint,)
	gl.glProgramUniform1i64NV = _getProcAddress('glProgramUniform1i64NV')
	gl.glProgramUniform1i64NV.gl = gl
	gl.glProgramUniform1i64NV.__name__ = 'glProgramUniform1i64NV'
	gl.glProgramUniform1i64NV.errcheck = _errcheck
	gl.glProgramUniform1i64NV.restype = GLvoid
	gl.glProgramUniform1i64NV.argtypes = (GLuint, GLint, GLint64EXT,)
	gl.glProgramUniform1i64vNV = _getProcAddress('glProgramUniform1i64vNV')
	gl.glProgramUniform1i64vNV.gl = gl
	gl.glProgramUniform1i64vNV.__name__ = 'glProgramUniform1i64vNV'
	gl.glProgramUniform1i64vNV.errcheck = _errcheck
	gl.glProgramUniform1i64vNV.restype = GLvoid
	gl.glProgramUniform1i64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glProgramUniform1iEXT = _getProcAddress('glProgramUniform1iEXT')
	gl.glProgramUniform1iEXT.gl = gl
	gl.glProgramUniform1iEXT.__name__ = 'glProgramUniform1iEXT'
	gl.glProgramUniform1iEXT.errcheck = _errcheck
	gl.glProgramUniform1iEXT.restype = GLvoid
	gl.glProgramUniform1iEXT.argtypes = (GLuint, GLint, GLint,)
	gl.glProgramUniform1iv = _getProcAddress('glProgramUniform1iv')
	gl.glProgramUniform1iv.gl = gl
	gl.glProgramUniform1iv.__name__ = 'glProgramUniform1iv'
	gl.glProgramUniform1iv.errcheck = _errcheck
	gl.glProgramUniform1iv.restype = GLvoid
	gl.glProgramUniform1iv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform1ivEXT = _getProcAddress('glProgramUniform1ivEXT')
	gl.glProgramUniform1ivEXT.gl = gl
	gl.glProgramUniform1ivEXT.__name__ = 'glProgramUniform1ivEXT'
	gl.glProgramUniform1ivEXT.errcheck = _errcheck
	gl.glProgramUniform1ivEXT.restype = GLvoid
	gl.glProgramUniform1ivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform1ui = _getProcAddress('glProgramUniform1ui')
	gl.glProgramUniform1ui.gl = gl
	gl.glProgramUniform1ui.__name__ = 'glProgramUniform1ui'
	gl.glProgramUniform1ui.errcheck = _errcheck
	gl.glProgramUniform1ui.restype = GLvoid
	gl.glProgramUniform1ui.argtypes = (GLuint, GLint, GLuint,)
	gl.glProgramUniform1ui64NV = _getProcAddress('glProgramUniform1ui64NV')
	gl.glProgramUniform1ui64NV.gl = gl
	gl.glProgramUniform1ui64NV.__name__ = 'glProgramUniform1ui64NV'
	gl.glProgramUniform1ui64NV.errcheck = _errcheck
	gl.glProgramUniform1ui64NV.restype = GLvoid
	gl.glProgramUniform1ui64NV.argtypes = (GLuint, GLint, GLuint64EXT,)
	gl.glProgramUniform1ui64vNV = _getProcAddress('glProgramUniform1ui64vNV')
	gl.glProgramUniform1ui64vNV.gl = gl
	gl.glProgramUniform1ui64vNV.__name__ = 'glProgramUniform1ui64vNV'
	gl.glProgramUniform1ui64vNV.errcheck = _errcheck
	gl.glProgramUniform1ui64vNV.restype = GLvoid
	gl.glProgramUniform1ui64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glProgramUniform1uiEXT = _getProcAddress('glProgramUniform1uiEXT')
	gl.glProgramUniform1uiEXT.gl = gl
	gl.glProgramUniform1uiEXT.__name__ = 'glProgramUniform1uiEXT'
	gl.glProgramUniform1uiEXT.errcheck = _errcheck
	gl.glProgramUniform1uiEXT.restype = GLvoid
	gl.glProgramUniform1uiEXT.argtypes = (GLuint, GLint, GLuint,)
	gl.glProgramUniform1uiv = _getProcAddress('glProgramUniform1uiv')
	gl.glProgramUniform1uiv.gl = gl
	gl.glProgramUniform1uiv.__name__ = 'glProgramUniform1uiv'
	gl.glProgramUniform1uiv.errcheck = _errcheck
	gl.glProgramUniform1uiv.restype = GLvoid
	gl.glProgramUniform1uiv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform1uivEXT = _getProcAddress('glProgramUniform1uivEXT')
	gl.glProgramUniform1uivEXT.gl = gl
	gl.glProgramUniform1uivEXT.__name__ = 'glProgramUniform1uivEXT'
	gl.glProgramUniform1uivEXT.errcheck = _errcheck
	gl.glProgramUniform1uivEXT.restype = GLvoid
	gl.glProgramUniform1uivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform2d = _getProcAddress('glProgramUniform2d')
	gl.glProgramUniform2d.gl = gl
	gl.glProgramUniform2d.__name__ = 'glProgramUniform2d'
	gl.glProgramUniform2d.errcheck = _errcheck
	gl.glProgramUniform2d.restype = GLvoid
	gl.glProgramUniform2d.argtypes = (GLuint, GLint, GLdouble, GLdouble,)
	gl.glProgramUniform2dEXT = _getProcAddress('glProgramUniform2dEXT')
	gl.glProgramUniform2dEXT.gl = gl
	gl.glProgramUniform2dEXT.__name__ = 'glProgramUniform2dEXT'
	gl.glProgramUniform2dEXT.errcheck = _errcheck
	gl.glProgramUniform2dEXT.restype = GLvoid
	gl.glProgramUniform2dEXT.argtypes = (GLuint, GLint, GLdouble, GLdouble,)
	gl.glProgramUniform2dv = _getProcAddress('glProgramUniform2dv')
	gl.glProgramUniform2dv.gl = gl
	gl.glProgramUniform2dv.__name__ = 'glProgramUniform2dv'
	gl.glProgramUniform2dv.errcheck = _errcheck
	gl.glProgramUniform2dv.restype = GLvoid
	gl.glProgramUniform2dv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform2dvEXT = _getProcAddress('glProgramUniform2dvEXT')
	gl.glProgramUniform2dvEXT.gl = gl
	gl.glProgramUniform2dvEXT.__name__ = 'glProgramUniform2dvEXT'
	gl.glProgramUniform2dvEXT.errcheck = _errcheck
	gl.glProgramUniform2dvEXT.restype = GLvoid
	gl.glProgramUniform2dvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform2f = _getProcAddress('glProgramUniform2f')
	gl.glProgramUniform2f.gl = gl
	gl.glProgramUniform2f.__name__ = 'glProgramUniform2f'
	gl.glProgramUniform2f.errcheck = _errcheck
	gl.glProgramUniform2f.restype = GLvoid
	gl.glProgramUniform2f.argtypes = (GLuint, GLint, GLfloat, GLfloat,)
	gl.glProgramUniform2fEXT = _getProcAddress('glProgramUniform2fEXT')
	gl.glProgramUniform2fEXT.gl = gl
	gl.glProgramUniform2fEXT.__name__ = 'glProgramUniform2fEXT'
	gl.glProgramUniform2fEXT.errcheck = _errcheck
	gl.glProgramUniform2fEXT.restype = GLvoid
	gl.glProgramUniform2fEXT.argtypes = (GLuint, GLint, GLfloat, GLfloat,)
	gl.glProgramUniform2fv = _getProcAddress('glProgramUniform2fv')
	gl.glProgramUniform2fv.gl = gl
	gl.glProgramUniform2fv.__name__ = 'glProgramUniform2fv'
	gl.glProgramUniform2fv.errcheck = _errcheck
	gl.glProgramUniform2fv.restype = GLvoid
	gl.glProgramUniform2fv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform2fvEXT = _getProcAddress('glProgramUniform2fvEXT')
	gl.glProgramUniform2fvEXT.gl = gl
	gl.glProgramUniform2fvEXT.__name__ = 'glProgramUniform2fvEXT'
	gl.glProgramUniform2fvEXT.errcheck = _errcheck
	gl.glProgramUniform2fvEXT.restype = GLvoid
	gl.glProgramUniform2fvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform2i = _getProcAddress('glProgramUniform2i')
	gl.glProgramUniform2i.gl = gl
	gl.glProgramUniform2i.__name__ = 'glProgramUniform2i'
	gl.glProgramUniform2i.errcheck = _errcheck
	gl.glProgramUniform2i.restype = GLvoid
	gl.glProgramUniform2i.argtypes = (GLuint, GLint, GLint, GLint,)
	gl.glProgramUniform2i64NV = _getProcAddress('glProgramUniform2i64NV')
	gl.glProgramUniform2i64NV.gl = gl
	gl.glProgramUniform2i64NV.__name__ = 'glProgramUniform2i64NV'
	gl.glProgramUniform2i64NV.errcheck = _errcheck
	gl.glProgramUniform2i64NV.restype = GLvoid
	gl.glProgramUniform2i64NV.argtypes = (GLuint, GLint, GLint64EXT, GLint64EXT,)
	gl.glProgramUniform2i64vNV = _getProcAddress('glProgramUniform2i64vNV')
	gl.glProgramUniform2i64vNV.gl = gl
	gl.glProgramUniform2i64vNV.__name__ = 'glProgramUniform2i64vNV'
	gl.glProgramUniform2i64vNV.errcheck = _errcheck
	gl.glProgramUniform2i64vNV.restype = GLvoid
	gl.glProgramUniform2i64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glProgramUniform2iEXT = _getProcAddress('glProgramUniform2iEXT')
	gl.glProgramUniform2iEXT.gl = gl
	gl.glProgramUniform2iEXT.__name__ = 'glProgramUniform2iEXT'
	gl.glProgramUniform2iEXT.errcheck = _errcheck
	gl.glProgramUniform2iEXT.restype = GLvoid
	gl.glProgramUniform2iEXT.argtypes = (GLuint, GLint, GLint, GLint,)
	gl.glProgramUniform2iv = _getProcAddress('glProgramUniform2iv')
	gl.glProgramUniform2iv.gl = gl
	gl.glProgramUniform2iv.__name__ = 'glProgramUniform2iv'
	gl.glProgramUniform2iv.errcheck = _errcheck
	gl.glProgramUniform2iv.restype = GLvoid
	gl.glProgramUniform2iv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform2ivEXT = _getProcAddress('glProgramUniform2ivEXT')
	gl.glProgramUniform2ivEXT.gl = gl
	gl.glProgramUniform2ivEXT.__name__ = 'glProgramUniform2ivEXT'
	gl.glProgramUniform2ivEXT.errcheck = _errcheck
	gl.glProgramUniform2ivEXT.restype = GLvoid
	gl.glProgramUniform2ivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform2ui = _getProcAddress('glProgramUniform2ui')
	gl.glProgramUniform2ui.gl = gl
	gl.glProgramUniform2ui.__name__ = 'glProgramUniform2ui'
	gl.glProgramUniform2ui.errcheck = _errcheck
	gl.glProgramUniform2ui.restype = GLvoid
	gl.glProgramUniform2ui.argtypes = (GLuint, GLint, GLuint, GLuint,)
	gl.glProgramUniform2ui64NV = _getProcAddress('glProgramUniform2ui64NV')
	gl.glProgramUniform2ui64NV.gl = gl
	gl.glProgramUniform2ui64NV.__name__ = 'glProgramUniform2ui64NV'
	gl.glProgramUniform2ui64NV.errcheck = _errcheck
	gl.glProgramUniform2ui64NV.restype = GLvoid
	gl.glProgramUniform2ui64NV.argtypes = (GLuint, GLint, GLuint64EXT, GLuint64EXT,)
	gl.glProgramUniform2ui64vNV = _getProcAddress('glProgramUniform2ui64vNV')
	gl.glProgramUniform2ui64vNV.gl = gl
	gl.glProgramUniform2ui64vNV.__name__ = 'glProgramUniform2ui64vNV'
	gl.glProgramUniform2ui64vNV.errcheck = _errcheck
	gl.glProgramUniform2ui64vNV.restype = GLvoid
	gl.glProgramUniform2ui64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glProgramUniform2uiEXT = _getProcAddress('glProgramUniform2uiEXT')
	gl.glProgramUniform2uiEXT.gl = gl
	gl.glProgramUniform2uiEXT.__name__ = 'glProgramUniform2uiEXT'
	gl.glProgramUniform2uiEXT.errcheck = _errcheck
	gl.glProgramUniform2uiEXT.restype = GLvoid
	gl.glProgramUniform2uiEXT.argtypes = (GLuint, GLint, GLuint, GLuint,)
	gl.glProgramUniform2uiv = _getProcAddress('glProgramUniform2uiv')
	gl.glProgramUniform2uiv.gl = gl
	gl.glProgramUniform2uiv.__name__ = 'glProgramUniform2uiv'
	gl.glProgramUniform2uiv.errcheck = _errcheck
	gl.glProgramUniform2uiv.restype = GLvoid
	gl.glProgramUniform2uiv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform2uivEXT = _getProcAddress('glProgramUniform2uivEXT')
	gl.glProgramUniform2uivEXT.gl = gl
	gl.glProgramUniform2uivEXT.__name__ = 'glProgramUniform2uivEXT'
	gl.glProgramUniform2uivEXT.errcheck = _errcheck
	gl.glProgramUniform2uivEXT.restype = GLvoid
	gl.glProgramUniform2uivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform3d = _getProcAddress('glProgramUniform3d')
	gl.glProgramUniform3d.gl = gl
	gl.glProgramUniform3d.__name__ = 'glProgramUniform3d'
	gl.glProgramUniform3d.errcheck = _errcheck
	gl.glProgramUniform3d.restype = GLvoid
	gl.glProgramUniform3d.argtypes = (GLuint, GLint, GLdouble, GLdouble, GLdouble,)
	gl.glProgramUniform3dEXT = _getProcAddress('glProgramUniform3dEXT')
	gl.glProgramUniform3dEXT.gl = gl
	gl.glProgramUniform3dEXT.__name__ = 'glProgramUniform3dEXT'
	gl.glProgramUniform3dEXT.errcheck = _errcheck
	gl.glProgramUniform3dEXT.restype = GLvoid
	gl.glProgramUniform3dEXT.argtypes = (GLuint, GLint, GLdouble, GLdouble, GLdouble,)
	gl.glProgramUniform3dv = _getProcAddress('glProgramUniform3dv')
	gl.glProgramUniform3dv.gl = gl
	gl.glProgramUniform3dv.__name__ = 'glProgramUniform3dv'
	gl.glProgramUniform3dv.errcheck = _errcheck
	gl.glProgramUniform3dv.restype = GLvoid
	gl.glProgramUniform3dv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform3dvEXT = _getProcAddress('glProgramUniform3dvEXT')
	gl.glProgramUniform3dvEXT.gl = gl
	gl.glProgramUniform3dvEXT.__name__ = 'glProgramUniform3dvEXT'
	gl.glProgramUniform3dvEXT.errcheck = _errcheck
	gl.glProgramUniform3dvEXT.restype = GLvoid
	gl.glProgramUniform3dvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform3f = _getProcAddress('glProgramUniform3f')
	gl.glProgramUniform3f.gl = gl
	gl.glProgramUniform3f.__name__ = 'glProgramUniform3f'
	gl.glProgramUniform3f.errcheck = _errcheck
	gl.glProgramUniform3f.restype = GLvoid
	gl.glProgramUniform3f.argtypes = (GLuint, GLint, GLfloat, GLfloat, GLfloat,)
	gl.glProgramUniform3fEXT = _getProcAddress('glProgramUniform3fEXT')
	gl.glProgramUniform3fEXT.gl = gl
	gl.glProgramUniform3fEXT.__name__ = 'glProgramUniform3fEXT'
	gl.glProgramUniform3fEXT.errcheck = _errcheck
	gl.glProgramUniform3fEXT.restype = GLvoid
	gl.glProgramUniform3fEXT.argtypes = (GLuint, GLint, GLfloat, GLfloat, GLfloat,)
	gl.glProgramUniform3fv = _getProcAddress('glProgramUniform3fv')
	gl.glProgramUniform3fv.gl = gl
	gl.glProgramUniform3fv.__name__ = 'glProgramUniform3fv'
	gl.glProgramUniform3fv.errcheck = _errcheck
	gl.glProgramUniform3fv.restype = GLvoid
	gl.glProgramUniform3fv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform3fvEXT = _getProcAddress('glProgramUniform3fvEXT')
	gl.glProgramUniform3fvEXT.gl = gl
	gl.glProgramUniform3fvEXT.__name__ = 'glProgramUniform3fvEXT'
	gl.glProgramUniform3fvEXT.errcheck = _errcheck
	gl.glProgramUniform3fvEXT.restype = GLvoid
	gl.glProgramUniform3fvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform3i = _getProcAddress('glProgramUniform3i')
	gl.glProgramUniform3i.gl = gl
	gl.glProgramUniform3i.__name__ = 'glProgramUniform3i'
	gl.glProgramUniform3i.errcheck = _errcheck
	gl.glProgramUniform3i.restype = GLvoid
	gl.glProgramUniform3i.argtypes = (GLuint, GLint, GLint, GLint, GLint,)
	gl.glProgramUniform3i64NV = _getProcAddress('glProgramUniform3i64NV')
	gl.glProgramUniform3i64NV.gl = gl
	gl.glProgramUniform3i64NV.__name__ = 'glProgramUniform3i64NV'
	gl.glProgramUniform3i64NV.errcheck = _errcheck
	gl.glProgramUniform3i64NV.restype = GLvoid
	gl.glProgramUniform3i64NV.argtypes = (GLuint, GLint, GLint64EXT, GLint64EXT, GLint64EXT,)
	gl.glProgramUniform3i64vNV = _getProcAddress('glProgramUniform3i64vNV')
	gl.glProgramUniform3i64vNV.gl = gl
	gl.glProgramUniform3i64vNV.__name__ = 'glProgramUniform3i64vNV'
	gl.glProgramUniform3i64vNV.errcheck = _errcheck
	gl.glProgramUniform3i64vNV.restype = GLvoid
	gl.glProgramUniform3i64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glProgramUniform3iEXT = _getProcAddress('glProgramUniform3iEXT')
	gl.glProgramUniform3iEXT.gl = gl
	gl.glProgramUniform3iEXT.__name__ = 'glProgramUniform3iEXT'
	gl.glProgramUniform3iEXT.errcheck = _errcheck
	gl.glProgramUniform3iEXT.restype = GLvoid
	gl.glProgramUniform3iEXT.argtypes = (GLuint, GLint, GLint, GLint, GLint,)
	gl.glProgramUniform3iv = _getProcAddress('glProgramUniform3iv')
	gl.glProgramUniform3iv.gl = gl
	gl.glProgramUniform3iv.__name__ = 'glProgramUniform3iv'
	gl.glProgramUniform3iv.errcheck = _errcheck
	gl.glProgramUniform3iv.restype = GLvoid
	gl.glProgramUniform3iv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform3ivEXT = _getProcAddress('glProgramUniform3ivEXT')
	gl.glProgramUniform3ivEXT.gl = gl
	gl.glProgramUniform3ivEXT.__name__ = 'glProgramUniform3ivEXT'
	gl.glProgramUniform3ivEXT.errcheck = _errcheck
	gl.glProgramUniform3ivEXT.restype = GLvoid
	gl.glProgramUniform3ivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform3ui = _getProcAddress('glProgramUniform3ui')
	gl.glProgramUniform3ui.gl = gl
	gl.glProgramUniform3ui.__name__ = 'glProgramUniform3ui'
	gl.glProgramUniform3ui.errcheck = _errcheck
	gl.glProgramUniform3ui.restype = GLvoid
	gl.glProgramUniform3ui.argtypes = (GLuint, GLint, GLuint, GLuint, GLuint,)
	gl.glProgramUniform3ui64NV = _getProcAddress('glProgramUniform3ui64NV')
	gl.glProgramUniform3ui64NV.gl = gl
	gl.glProgramUniform3ui64NV.__name__ = 'glProgramUniform3ui64NV'
	gl.glProgramUniform3ui64NV.errcheck = _errcheck
	gl.glProgramUniform3ui64NV.restype = GLvoid
	gl.glProgramUniform3ui64NV.argtypes = (GLuint, GLint, GLuint64EXT, GLuint64EXT, GLuint64EXT,)
	gl.glProgramUniform3ui64vNV = _getProcAddress('glProgramUniform3ui64vNV')
	gl.glProgramUniform3ui64vNV.gl = gl
	gl.glProgramUniform3ui64vNV.__name__ = 'glProgramUniform3ui64vNV'
	gl.glProgramUniform3ui64vNV.errcheck = _errcheck
	gl.glProgramUniform3ui64vNV.restype = GLvoid
	gl.glProgramUniform3ui64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glProgramUniform3uiEXT = _getProcAddress('glProgramUniform3uiEXT')
	gl.glProgramUniform3uiEXT.gl = gl
	gl.glProgramUniform3uiEXT.__name__ = 'glProgramUniform3uiEXT'
	gl.glProgramUniform3uiEXT.errcheck = _errcheck
	gl.glProgramUniform3uiEXT.restype = GLvoid
	gl.glProgramUniform3uiEXT.argtypes = (GLuint, GLint, GLuint, GLuint, GLuint,)
	gl.glProgramUniform3uiv = _getProcAddress('glProgramUniform3uiv')
	gl.glProgramUniform3uiv.gl = gl
	gl.glProgramUniform3uiv.__name__ = 'glProgramUniform3uiv'
	gl.glProgramUniform3uiv.errcheck = _errcheck
	gl.glProgramUniform3uiv.restype = GLvoid
	gl.glProgramUniform3uiv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform3uivEXT = _getProcAddress('glProgramUniform3uivEXT')
	gl.glProgramUniform3uivEXT.gl = gl
	gl.glProgramUniform3uivEXT.__name__ = 'glProgramUniform3uivEXT'
	gl.glProgramUniform3uivEXT.errcheck = _errcheck
	gl.glProgramUniform3uivEXT.restype = GLvoid
	gl.glProgramUniform3uivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform4d = _getProcAddress('glProgramUniform4d')
	gl.glProgramUniform4d.gl = gl
	gl.glProgramUniform4d.__name__ = 'glProgramUniform4d'
	gl.glProgramUniform4d.errcheck = _errcheck
	gl.glProgramUniform4d.restype = GLvoid
	gl.glProgramUniform4d.argtypes = (GLuint, GLint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glProgramUniform4dEXT = _getProcAddress('glProgramUniform4dEXT')
	gl.glProgramUniform4dEXT.gl = gl
	gl.glProgramUniform4dEXT.__name__ = 'glProgramUniform4dEXT'
	gl.glProgramUniform4dEXT.errcheck = _errcheck
	gl.glProgramUniform4dEXT.restype = GLvoid
	gl.glProgramUniform4dEXT.argtypes = (GLuint, GLint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glProgramUniform4dv = _getProcAddress('glProgramUniform4dv')
	gl.glProgramUniform4dv.gl = gl
	gl.glProgramUniform4dv.__name__ = 'glProgramUniform4dv'
	gl.glProgramUniform4dv.errcheck = _errcheck
	gl.glProgramUniform4dv.restype = GLvoid
	gl.glProgramUniform4dv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform4dvEXT = _getProcAddress('glProgramUniform4dvEXT')
	gl.glProgramUniform4dvEXT.gl = gl
	gl.glProgramUniform4dvEXT.__name__ = 'glProgramUniform4dvEXT'
	gl.glProgramUniform4dvEXT.errcheck = _errcheck
	gl.glProgramUniform4dvEXT.restype = GLvoid
	gl.glProgramUniform4dvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLdouble),)
	gl.glProgramUniform4f = _getProcAddress('glProgramUniform4f')
	gl.glProgramUniform4f.gl = gl
	gl.glProgramUniform4f.__name__ = 'glProgramUniform4f'
	gl.glProgramUniform4f.errcheck = _errcheck
	gl.glProgramUniform4f.restype = GLvoid
	gl.glProgramUniform4f.argtypes = (GLuint, GLint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glProgramUniform4fEXT = _getProcAddress('glProgramUniform4fEXT')
	gl.glProgramUniform4fEXT.gl = gl
	gl.glProgramUniform4fEXT.__name__ = 'glProgramUniform4fEXT'
	gl.glProgramUniform4fEXT.errcheck = _errcheck
	gl.glProgramUniform4fEXT.restype = GLvoid
	gl.glProgramUniform4fEXT.argtypes = (GLuint, GLint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glProgramUniform4fv = _getProcAddress('glProgramUniform4fv')
	gl.glProgramUniform4fv.gl = gl
	gl.glProgramUniform4fv.__name__ = 'glProgramUniform4fv'
	gl.glProgramUniform4fv.errcheck = _errcheck
	gl.glProgramUniform4fv.restype = GLvoid
	gl.glProgramUniform4fv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform4fvEXT = _getProcAddress('glProgramUniform4fvEXT')
	gl.glProgramUniform4fvEXT.gl = gl
	gl.glProgramUniform4fvEXT.__name__ = 'glProgramUniform4fvEXT'
	gl.glProgramUniform4fvEXT.errcheck = _errcheck
	gl.glProgramUniform4fvEXT.restype = GLvoid
	gl.glProgramUniform4fvEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLfloat),)
	gl.glProgramUniform4i = _getProcAddress('glProgramUniform4i')
	gl.glProgramUniform4i.gl = gl
	gl.glProgramUniform4i.__name__ = 'glProgramUniform4i'
	gl.glProgramUniform4i.errcheck = _errcheck
	gl.glProgramUniform4i.restype = GLvoid
	gl.glProgramUniform4i.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLint,)
	gl.glProgramUniform4i64NV = _getProcAddress('glProgramUniform4i64NV')
	gl.glProgramUniform4i64NV.gl = gl
	gl.glProgramUniform4i64NV.__name__ = 'glProgramUniform4i64NV'
	gl.glProgramUniform4i64NV.errcheck = _errcheck
	gl.glProgramUniform4i64NV.restype = GLvoid
	gl.glProgramUniform4i64NV.argtypes = (GLuint, GLint, GLint64EXT, GLint64EXT, GLint64EXT, GLint64EXT,)
	gl.glProgramUniform4i64vNV = _getProcAddress('glProgramUniform4i64vNV')
	gl.glProgramUniform4i64vNV.gl = gl
	gl.glProgramUniform4i64vNV.__name__ = 'glProgramUniform4i64vNV'
	gl.glProgramUniform4i64vNV.errcheck = _errcheck
	gl.glProgramUniform4i64vNV.restype = GLvoid
	gl.glProgramUniform4i64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glProgramUniform4iEXT = _getProcAddress('glProgramUniform4iEXT')
	gl.glProgramUniform4iEXT.gl = gl
	gl.glProgramUniform4iEXT.__name__ = 'glProgramUniform4iEXT'
	gl.glProgramUniform4iEXT.errcheck = _errcheck
	gl.glProgramUniform4iEXT.restype = GLvoid
	gl.glProgramUniform4iEXT.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLint,)
	gl.glProgramUniform4iv = _getProcAddress('glProgramUniform4iv')
	gl.glProgramUniform4iv.gl = gl
	gl.glProgramUniform4iv.__name__ = 'glProgramUniform4iv'
	gl.glProgramUniform4iv.errcheck = _errcheck
	gl.glProgramUniform4iv.restype = GLvoid
	gl.glProgramUniform4iv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform4ivEXT = _getProcAddress('glProgramUniform4ivEXT')
	gl.glProgramUniform4ivEXT.gl = gl
	gl.glProgramUniform4ivEXT.__name__ = 'glProgramUniform4ivEXT'
	gl.glProgramUniform4ivEXT.errcheck = _errcheck
	gl.glProgramUniform4ivEXT.restype = GLvoid
	gl.glProgramUniform4ivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLint),)
	gl.glProgramUniform4ui = _getProcAddress('glProgramUniform4ui')
	gl.glProgramUniform4ui.gl = gl
	gl.glProgramUniform4ui.__name__ = 'glProgramUniform4ui'
	gl.glProgramUniform4ui.errcheck = _errcheck
	gl.glProgramUniform4ui.restype = GLvoid
	gl.glProgramUniform4ui.argtypes = (GLuint, GLint, GLuint, GLuint, GLuint, GLuint,)
	gl.glProgramUniform4ui64NV = _getProcAddress('glProgramUniform4ui64NV')
	gl.glProgramUniform4ui64NV.gl = gl
	gl.glProgramUniform4ui64NV.__name__ = 'glProgramUniform4ui64NV'
	gl.glProgramUniform4ui64NV.errcheck = _errcheck
	gl.glProgramUniform4ui64NV.restype = GLvoid
	gl.glProgramUniform4ui64NV.argtypes = (GLuint, GLint, GLuint64EXT, GLuint64EXT, GLuint64EXT, GLuint64EXT,)
	gl.glProgramUniform4ui64vNV = _getProcAddress('glProgramUniform4ui64vNV')
	gl.glProgramUniform4ui64vNV.gl = gl
	gl.glProgramUniform4ui64vNV.__name__ = 'glProgramUniform4ui64vNV'
	gl.glProgramUniform4ui64vNV.errcheck = _errcheck
	gl.glProgramUniform4ui64vNV.restype = GLvoid
	gl.glProgramUniform4ui64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glProgramUniform4uiEXT = _getProcAddress('glProgramUniform4uiEXT')
	gl.glProgramUniform4uiEXT.gl = gl
	gl.glProgramUniform4uiEXT.__name__ = 'glProgramUniform4uiEXT'
	gl.glProgramUniform4uiEXT.errcheck = _errcheck
	gl.glProgramUniform4uiEXT.restype = GLvoid
	gl.glProgramUniform4uiEXT.argtypes = (GLuint, GLint, GLuint, GLuint, GLuint, GLuint,)
	gl.glProgramUniform4uiv = _getProcAddress('glProgramUniform4uiv')
	gl.glProgramUniform4uiv.gl = gl
	gl.glProgramUniform4uiv.__name__ = 'glProgramUniform4uiv'
	gl.glProgramUniform4uiv.errcheck = _errcheck
	gl.glProgramUniform4uiv.restype = GLvoid
	gl.glProgramUniform4uiv.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniform4uivEXT = _getProcAddress('glProgramUniform4uivEXT')
	gl.glProgramUniform4uivEXT.gl = gl
	gl.glProgramUniform4uivEXT.__name__ = 'glProgramUniform4uivEXT'
	gl.glProgramUniform4uivEXT.errcheck = _errcheck
	gl.glProgramUniform4uivEXT.restype = GLvoid
	gl.glProgramUniform4uivEXT.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint),)
	gl.glProgramUniformHandleui64ARB = _getProcAddress('glProgramUniformHandleui64ARB')
	gl.glProgramUniformHandleui64ARB.gl = gl
	gl.glProgramUniformHandleui64ARB.__name__ = 'glProgramUniformHandleui64ARB'
	gl.glProgramUniformHandleui64ARB.errcheck = _errcheck
	gl.glProgramUniformHandleui64ARB.restype = GLvoid
	gl.glProgramUniformHandleui64ARB.argtypes = (GLuint, GLint, GLuint64,)
	gl.glProgramUniformHandleui64NV = _getProcAddress('glProgramUniformHandleui64NV')
	gl.glProgramUniformHandleui64NV.gl = gl
	gl.glProgramUniformHandleui64NV.__name__ = 'glProgramUniformHandleui64NV'
	gl.glProgramUniformHandleui64NV.errcheck = _errcheck
	gl.glProgramUniformHandleui64NV.restype = GLvoid
	gl.glProgramUniformHandleui64NV.argtypes = (GLuint, GLint, GLuint64,)
	gl.glProgramUniformHandleui64vARB = _getProcAddress('glProgramUniformHandleui64vARB')
	gl.glProgramUniformHandleui64vARB.gl = gl
	gl.glProgramUniformHandleui64vARB.__name__ = 'glProgramUniformHandleui64vARB'
	gl.glProgramUniformHandleui64vARB.errcheck = _errcheck
	gl.glProgramUniformHandleui64vARB.restype = GLvoid
	gl.glProgramUniformHandleui64vARB.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64),)
	gl.glProgramUniformHandleui64vNV = _getProcAddress('glProgramUniformHandleui64vNV')
	gl.glProgramUniformHandleui64vNV.gl = gl
	gl.glProgramUniformHandleui64vNV.__name__ = 'glProgramUniformHandleui64vNV'
	gl.glProgramUniformHandleui64vNV.errcheck = _errcheck
	gl.glProgramUniformHandleui64vNV.restype = GLvoid
	gl.glProgramUniformHandleui64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64),)
	gl.glProgramUniformMatrix2dv = _getProcAddress('glProgramUniformMatrix2dv')
	gl.glProgramUniformMatrix2dv.gl = gl
	gl.glProgramUniformMatrix2dv.__name__ = 'glProgramUniformMatrix2dv'
	gl.glProgramUniformMatrix2dv.errcheck = _errcheck
	gl.glProgramUniformMatrix2dv.restype = GLvoid
	gl.glProgramUniformMatrix2dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix2dvEXT = _getProcAddress('glProgramUniformMatrix2dvEXT')
	gl.glProgramUniformMatrix2dvEXT.gl = gl
	gl.glProgramUniformMatrix2dvEXT.__name__ = 'glProgramUniformMatrix2dvEXT'
	gl.glProgramUniformMatrix2dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix2dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix2dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix2fv = _getProcAddress('glProgramUniformMatrix2fv')
	gl.glProgramUniformMatrix2fv.gl = gl
	gl.glProgramUniformMatrix2fv.__name__ = 'glProgramUniformMatrix2fv'
	gl.glProgramUniformMatrix2fv.errcheck = _errcheck
	gl.glProgramUniformMatrix2fv.restype = GLvoid
	gl.glProgramUniformMatrix2fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix2fvEXT = _getProcAddress('glProgramUniformMatrix2fvEXT')
	gl.glProgramUniformMatrix2fvEXT.gl = gl
	gl.glProgramUniformMatrix2fvEXT.__name__ = 'glProgramUniformMatrix2fvEXT'
	gl.glProgramUniformMatrix2fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix2fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix2fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix2x3dv = _getProcAddress('glProgramUniformMatrix2x3dv')
	gl.glProgramUniformMatrix2x3dv.gl = gl
	gl.glProgramUniformMatrix2x3dv.__name__ = 'glProgramUniformMatrix2x3dv'
	gl.glProgramUniformMatrix2x3dv.errcheck = _errcheck
	gl.glProgramUniformMatrix2x3dv.restype = GLvoid
	gl.glProgramUniformMatrix2x3dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix2x3dvEXT = _getProcAddress('glProgramUniformMatrix2x3dvEXT')
	gl.glProgramUniformMatrix2x3dvEXT.gl = gl
	gl.glProgramUniformMatrix2x3dvEXT.__name__ = 'glProgramUniformMatrix2x3dvEXT'
	gl.glProgramUniformMatrix2x3dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix2x3dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix2x3dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix2x3fv = _getProcAddress('glProgramUniformMatrix2x3fv')
	gl.glProgramUniformMatrix2x3fv.gl = gl
	gl.glProgramUniformMatrix2x3fv.__name__ = 'glProgramUniformMatrix2x3fv'
	gl.glProgramUniformMatrix2x3fv.errcheck = _errcheck
	gl.glProgramUniformMatrix2x3fv.restype = GLvoid
	gl.glProgramUniformMatrix2x3fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix2x3fvEXT = _getProcAddress('glProgramUniformMatrix2x3fvEXT')
	gl.glProgramUniformMatrix2x3fvEXT.gl = gl
	gl.glProgramUniformMatrix2x3fvEXT.__name__ = 'glProgramUniformMatrix2x3fvEXT'
	gl.glProgramUniformMatrix2x3fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix2x3fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix2x3fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix2x4dv = _getProcAddress('glProgramUniformMatrix2x4dv')
	gl.glProgramUniformMatrix2x4dv.gl = gl
	gl.glProgramUniformMatrix2x4dv.__name__ = 'glProgramUniformMatrix2x4dv'
	gl.glProgramUniformMatrix2x4dv.errcheck = _errcheck
	gl.glProgramUniformMatrix2x4dv.restype = GLvoid
	gl.glProgramUniformMatrix2x4dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix2x4dvEXT = _getProcAddress('glProgramUniformMatrix2x4dvEXT')
	gl.glProgramUniformMatrix2x4dvEXT.gl = gl
	gl.glProgramUniformMatrix2x4dvEXT.__name__ = 'glProgramUniformMatrix2x4dvEXT'
	gl.glProgramUniformMatrix2x4dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix2x4dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix2x4dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix2x4fv = _getProcAddress('glProgramUniformMatrix2x4fv')
	gl.glProgramUniformMatrix2x4fv.gl = gl
	gl.glProgramUniformMatrix2x4fv.__name__ = 'glProgramUniformMatrix2x4fv'
	gl.glProgramUniformMatrix2x4fv.errcheck = _errcheck
	gl.glProgramUniformMatrix2x4fv.restype = GLvoid
	gl.glProgramUniformMatrix2x4fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix2x4fvEXT = _getProcAddress('glProgramUniformMatrix2x4fvEXT')
	gl.glProgramUniformMatrix2x4fvEXT.gl = gl
	gl.glProgramUniformMatrix2x4fvEXT.__name__ = 'glProgramUniformMatrix2x4fvEXT'
	gl.glProgramUniformMatrix2x4fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix2x4fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix2x4fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix3dv = _getProcAddress('glProgramUniformMatrix3dv')
	gl.glProgramUniformMatrix3dv.gl = gl
	gl.glProgramUniformMatrix3dv.__name__ = 'glProgramUniformMatrix3dv'
	gl.glProgramUniformMatrix3dv.errcheck = _errcheck
	gl.glProgramUniformMatrix3dv.restype = GLvoid
	gl.glProgramUniformMatrix3dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix3dvEXT = _getProcAddress('glProgramUniformMatrix3dvEXT')
	gl.glProgramUniformMatrix3dvEXT.gl = gl
	gl.glProgramUniformMatrix3dvEXT.__name__ = 'glProgramUniformMatrix3dvEXT'
	gl.glProgramUniformMatrix3dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix3dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix3dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix3fv = _getProcAddress('glProgramUniformMatrix3fv')
	gl.glProgramUniformMatrix3fv.gl = gl
	gl.glProgramUniformMatrix3fv.__name__ = 'glProgramUniformMatrix3fv'
	gl.glProgramUniformMatrix3fv.errcheck = _errcheck
	gl.glProgramUniformMatrix3fv.restype = GLvoid
	gl.glProgramUniformMatrix3fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix3fvEXT = _getProcAddress('glProgramUniformMatrix3fvEXT')
	gl.glProgramUniformMatrix3fvEXT.gl = gl
	gl.glProgramUniformMatrix3fvEXT.__name__ = 'glProgramUniformMatrix3fvEXT'
	gl.glProgramUniformMatrix3fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix3fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix3fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix3x2dv = _getProcAddress('glProgramUniformMatrix3x2dv')
	gl.glProgramUniformMatrix3x2dv.gl = gl
	gl.glProgramUniformMatrix3x2dv.__name__ = 'glProgramUniformMatrix3x2dv'
	gl.glProgramUniformMatrix3x2dv.errcheck = _errcheck
	gl.glProgramUniformMatrix3x2dv.restype = GLvoid
	gl.glProgramUniformMatrix3x2dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix3x2dvEXT = _getProcAddress('glProgramUniformMatrix3x2dvEXT')
	gl.glProgramUniformMatrix3x2dvEXT.gl = gl
	gl.glProgramUniformMatrix3x2dvEXT.__name__ = 'glProgramUniformMatrix3x2dvEXT'
	gl.glProgramUniformMatrix3x2dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix3x2dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix3x2dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix3x2fv = _getProcAddress('glProgramUniformMatrix3x2fv')
	gl.glProgramUniformMatrix3x2fv.gl = gl
	gl.glProgramUniformMatrix3x2fv.__name__ = 'glProgramUniformMatrix3x2fv'
	gl.glProgramUniformMatrix3x2fv.errcheck = _errcheck
	gl.glProgramUniformMatrix3x2fv.restype = GLvoid
	gl.glProgramUniformMatrix3x2fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix3x2fvEXT = _getProcAddress('glProgramUniformMatrix3x2fvEXT')
	gl.glProgramUniformMatrix3x2fvEXT.gl = gl
	gl.glProgramUniformMatrix3x2fvEXT.__name__ = 'glProgramUniformMatrix3x2fvEXT'
	gl.glProgramUniformMatrix3x2fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix3x2fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix3x2fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix3x4dv = _getProcAddress('glProgramUniformMatrix3x4dv')
	gl.glProgramUniformMatrix3x4dv.gl = gl
	gl.glProgramUniformMatrix3x4dv.__name__ = 'glProgramUniformMatrix3x4dv'
	gl.glProgramUniformMatrix3x4dv.errcheck = _errcheck
	gl.glProgramUniformMatrix3x4dv.restype = GLvoid
	gl.glProgramUniformMatrix3x4dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix3x4dvEXT = _getProcAddress('glProgramUniformMatrix3x4dvEXT')
	gl.glProgramUniformMatrix3x4dvEXT.gl = gl
	gl.glProgramUniformMatrix3x4dvEXT.__name__ = 'glProgramUniformMatrix3x4dvEXT'
	gl.glProgramUniformMatrix3x4dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix3x4dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix3x4dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix3x4fv = _getProcAddress('glProgramUniformMatrix3x4fv')
	gl.glProgramUniformMatrix3x4fv.gl = gl
	gl.glProgramUniformMatrix3x4fv.__name__ = 'glProgramUniformMatrix3x4fv'
	gl.glProgramUniformMatrix3x4fv.errcheck = _errcheck
	gl.glProgramUniformMatrix3x4fv.restype = GLvoid
	gl.glProgramUniformMatrix3x4fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix3x4fvEXT = _getProcAddress('glProgramUniformMatrix3x4fvEXT')
	gl.glProgramUniformMatrix3x4fvEXT.gl = gl
	gl.glProgramUniformMatrix3x4fvEXT.__name__ = 'glProgramUniformMatrix3x4fvEXT'
	gl.glProgramUniformMatrix3x4fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix3x4fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix3x4fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix4dv = _getProcAddress('glProgramUniformMatrix4dv')
	gl.glProgramUniformMatrix4dv.gl = gl
	gl.glProgramUniformMatrix4dv.__name__ = 'glProgramUniformMatrix4dv'
	gl.glProgramUniformMatrix4dv.errcheck = _errcheck
	gl.glProgramUniformMatrix4dv.restype = GLvoid
	gl.glProgramUniformMatrix4dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix4dvEXT = _getProcAddress('glProgramUniformMatrix4dvEXT')
	gl.glProgramUniformMatrix4dvEXT.gl = gl
	gl.glProgramUniformMatrix4dvEXT.__name__ = 'glProgramUniformMatrix4dvEXT'
	gl.glProgramUniformMatrix4dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix4dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix4dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix4fv = _getProcAddress('glProgramUniformMatrix4fv')
	gl.glProgramUniformMatrix4fv.gl = gl
	gl.glProgramUniformMatrix4fv.__name__ = 'glProgramUniformMatrix4fv'
	gl.glProgramUniformMatrix4fv.errcheck = _errcheck
	gl.glProgramUniformMatrix4fv.restype = GLvoid
	gl.glProgramUniformMatrix4fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix4fvEXT = _getProcAddress('glProgramUniformMatrix4fvEXT')
	gl.glProgramUniformMatrix4fvEXT.gl = gl
	gl.glProgramUniformMatrix4fvEXT.__name__ = 'glProgramUniformMatrix4fvEXT'
	gl.glProgramUniformMatrix4fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix4fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix4fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix4x2dv = _getProcAddress('glProgramUniformMatrix4x2dv')
	gl.glProgramUniformMatrix4x2dv.gl = gl
	gl.glProgramUniformMatrix4x2dv.__name__ = 'glProgramUniformMatrix4x2dv'
	gl.glProgramUniformMatrix4x2dv.errcheck = _errcheck
	gl.glProgramUniformMatrix4x2dv.restype = GLvoid
	gl.glProgramUniformMatrix4x2dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix4x2dvEXT = _getProcAddress('glProgramUniformMatrix4x2dvEXT')
	gl.glProgramUniformMatrix4x2dvEXT.gl = gl
	gl.glProgramUniformMatrix4x2dvEXT.__name__ = 'glProgramUniformMatrix4x2dvEXT'
	gl.glProgramUniformMatrix4x2dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix4x2dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix4x2dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix4x2fv = _getProcAddress('glProgramUniformMatrix4x2fv')
	gl.glProgramUniformMatrix4x2fv.gl = gl
	gl.glProgramUniformMatrix4x2fv.__name__ = 'glProgramUniformMatrix4x2fv'
	gl.glProgramUniformMatrix4x2fv.errcheck = _errcheck
	gl.glProgramUniformMatrix4x2fv.restype = GLvoid
	gl.glProgramUniformMatrix4x2fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix4x2fvEXT = _getProcAddress('glProgramUniformMatrix4x2fvEXT')
	gl.glProgramUniformMatrix4x2fvEXT.gl = gl
	gl.glProgramUniformMatrix4x2fvEXT.__name__ = 'glProgramUniformMatrix4x2fvEXT'
	gl.glProgramUniformMatrix4x2fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix4x2fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix4x2fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix4x3dv = _getProcAddress('glProgramUniformMatrix4x3dv')
	gl.glProgramUniformMatrix4x3dv.gl = gl
	gl.glProgramUniformMatrix4x3dv.__name__ = 'glProgramUniformMatrix4x3dv'
	gl.glProgramUniformMatrix4x3dv.errcheck = _errcheck
	gl.glProgramUniformMatrix4x3dv.restype = GLvoid
	gl.glProgramUniformMatrix4x3dv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix4x3dvEXT = _getProcAddress('glProgramUniformMatrix4x3dvEXT')
	gl.glProgramUniformMatrix4x3dvEXT.gl = gl
	gl.glProgramUniformMatrix4x3dvEXT.__name__ = 'glProgramUniformMatrix4x3dvEXT'
	gl.glProgramUniformMatrix4x3dvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix4x3dvEXT.restype = GLvoid
	gl.glProgramUniformMatrix4x3dvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glProgramUniformMatrix4x3fv = _getProcAddress('glProgramUniformMatrix4x3fv')
	gl.glProgramUniformMatrix4x3fv.gl = gl
	gl.glProgramUniformMatrix4x3fv.__name__ = 'glProgramUniformMatrix4x3fv'
	gl.glProgramUniformMatrix4x3fv.errcheck = _errcheck
	gl.glProgramUniformMatrix4x3fv.restype = GLvoid
	gl.glProgramUniformMatrix4x3fv.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformMatrix4x3fvEXT = _getProcAddress('glProgramUniformMatrix4x3fvEXT')
	gl.glProgramUniformMatrix4x3fvEXT.gl = gl
	gl.glProgramUniformMatrix4x3fvEXT.__name__ = 'glProgramUniformMatrix4x3fvEXT'
	gl.glProgramUniformMatrix4x3fvEXT.errcheck = _errcheck
	gl.glProgramUniformMatrix4x3fvEXT.restype = GLvoid
	gl.glProgramUniformMatrix4x3fvEXT.argtypes = (GLuint, GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glProgramUniformui64NV = _getProcAddress('glProgramUniformui64NV')
	gl.glProgramUniformui64NV.gl = gl
	gl.glProgramUniformui64NV.__name__ = 'glProgramUniformui64NV'
	gl.glProgramUniformui64NV.errcheck = _errcheck
	gl.glProgramUniformui64NV.restype = GLvoid
	gl.glProgramUniformui64NV.argtypes = (GLuint, GLint, GLuint64EXT,)
	gl.glProgramUniformui64vNV = _getProcAddress('glProgramUniformui64vNV')
	gl.glProgramUniformui64vNV.gl = gl
	gl.glProgramUniformui64vNV.__name__ = 'glProgramUniformui64vNV'
	gl.glProgramUniformui64vNV.errcheck = _errcheck
	gl.glProgramUniformui64vNV.restype = GLvoid
	gl.glProgramUniformui64vNV.argtypes = (GLuint, GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glProgramVertexLimitNV = _getProcAddress('glProgramVertexLimitNV')
	gl.glProgramVertexLimitNV.gl = gl
	gl.glProgramVertexLimitNV.__name__ = 'glProgramVertexLimitNV'
	gl.glProgramVertexLimitNV.errcheck = _errcheck
	gl.glProgramVertexLimitNV.restype = GLvoid
	gl.glProgramVertexLimitNV.argtypes = (GLenum, GLint,)
	gl.glProvokingVertex = _getProcAddress('glProvokingVertex')
	gl.glProvokingVertex.gl = gl
	gl.glProvokingVertex.__name__ = 'glProvokingVertex'
	gl.glProvokingVertex.errcheck = _errcheck
	gl.glProvokingVertex.restype = GLvoid
	gl.glProvokingVertex.argtypes = (GLenum,)
	gl.glProvokingVertexEXT = _getProcAddress('glProvokingVertexEXT')
	gl.glProvokingVertexEXT.gl = gl
	gl.glProvokingVertexEXT.__name__ = 'glProvokingVertexEXT'
	gl.glProvokingVertexEXT.errcheck = _errcheck
	gl.glProvokingVertexEXT.restype = GLvoid
	gl.glProvokingVertexEXT.argtypes = (GLenum,)
	gl.glPushAttrib = _getProcAddress('glPushAttrib')
	gl.glPushAttrib.gl = gl
	gl.glPushAttrib.__name__ = 'glPushAttrib'
	gl.glPushAttrib.errcheck = _errcheck
	gl.glPushAttrib.restype = GLvoid
	gl.glPushAttrib.argtypes = (GLbitfield,)
	gl.glPushClientAttrib = _getProcAddress('glPushClientAttrib')
	gl.glPushClientAttrib.gl = gl
	gl.glPushClientAttrib.__name__ = 'glPushClientAttrib'
	gl.glPushClientAttrib.errcheck = _errcheck
	gl.glPushClientAttrib.restype = GLvoid
	gl.glPushClientAttrib.argtypes = (GLbitfield,)
	gl.glPushClientAttribDefaultEXT = _getProcAddress('glPushClientAttribDefaultEXT')
	gl.glPushClientAttribDefaultEXT.gl = gl
	gl.glPushClientAttribDefaultEXT.__name__ = 'glPushClientAttribDefaultEXT'
	gl.glPushClientAttribDefaultEXT.errcheck = _errcheck
	gl.glPushClientAttribDefaultEXT.restype = GLvoid
	gl.glPushClientAttribDefaultEXT.argtypes = (GLbitfield,)
	gl.glPushDebugGroup = _getProcAddress('glPushDebugGroup')
	gl.glPushDebugGroup.gl = gl
	gl.glPushDebugGroup.__name__ = 'glPushDebugGroup'
	gl.glPushDebugGroup.errcheck = _errcheck
	gl.glPushDebugGroup.restype = GLvoid
	gl.glPushDebugGroup.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLchar),)
	gl.glPushDebugGroupKHR = _getProcAddress('glPushDebugGroupKHR')
	gl.glPushDebugGroupKHR.gl = gl
	gl.glPushDebugGroupKHR.__name__ = 'glPushDebugGroupKHR'
	gl.glPushDebugGroupKHR.errcheck = _errcheck
	gl.glPushDebugGroupKHR.restype = GLvoid
	gl.glPushDebugGroupKHR.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLchar),)
	gl.glPushGroupMarkerEXT = _getProcAddress('glPushGroupMarkerEXT')
	gl.glPushGroupMarkerEXT.gl = gl
	gl.glPushGroupMarkerEXT.__name__ = 'glPushGroupMarkerEXT'
	gl.glPushGroupMarkerEXT.errcheck = _errcheck
	gl.glPushGroupMarkerEXT.restype = GLvoid
	gl.glPushGroupMarkerEXT.argtypes = (GLsizei, POINTER(GLchar),)
	gl.glPushMatrix = _getProcAddress('glPushMatrix')
	gl.glPushMatrix.gl = gl
	gl.glPushMatrix.__name__ = 'glPushMatrix'
	gl.glPushMatrix.errcheck = _errcheck
	gl.glPushMatrix.restype = GLvoid
	gl.glPushMatrix.argtypes = ()
	gl.glPushName = _getProcAddress('glPushName')
	gl.glPushName.gl = gl
	gl.glPushName.__name__ = 'glPushName'
	gl.glPushName.errcheck = _errcheck
	gl.glPushName.restype = GLvoid
	gl.glPushName.argtypes = (GLuint,)
	gl.glQueryCounter = _getProcAddress('glQueryCounter')
	gl.glQueryCounter.gl = gl
	gl.glQueryCounter.__name__ = 'glQueryCounter'
	gl.glQueryCounter.errcheck = _errcheck
	gl.glQueryCounter.restype = GLvoid
	gl.glQueryCounter.argtypes = (GLuint, GLenum,)
	gl.glQueryCounterEXT = _getProcAddress('glQueryCounterEXT')
	gl.glQueryCounterEXT.gl = gl
	gl.glQueryCounterEXT.__name__ = 'glQueryCounterEXT'
	gl.glQueryCounterEXT.errcheck = _errcheck
	gl.glQueryCounterEXT.restype = GLvoid
	gl.glQueryCounterEXT.argtypes = (GLuint, GLenum,)
	gl.glQueryMatrixxOES = _getProcAddress('glQueryMatrixxOES')
	gl.glQueryMatrixxOES.gl = gl
	gl.glQueryMatrixxOES.__name__ = 'glQueryMatrixxOES'
	gl.glQueryMatrixxOES.errcheck = _errcheck
	gl.glQueryMatrixxOES.restype = GLbitfield
	gl.glQueryMatrixxOES.argtypes = (POINTER(GLfixed), POINTER(GLint),)
	gl.glQueryObjectParameteruiAMD = _getProcAddress('glQueryObjectParameteruiAMD')
	gl.glQueryObjectParameteruiAMD.gl = gl
	gl.glQueryObjectParameteruiAMD.__name__ = 'glQueryObjectParameteruiAMD'
	gl.glQueryObjectParameteruiAMD.errcheck = _errcheck
	gl.glQueryObjectParameteruiAMD.restype = GLvoid
	gl.glQueryObjectParameteruiAMD.argtypes = (GLenum, GLuint, GLenum, GLuint,)
	gl.glRasterPos2d = _getProcAddress('glRasterPos2d')
	gl.glRasterPos2d.gl = gl
	gl.glRasterPos2d.__name__ = 'glRasterPos2d'
	gl.glRasterPos2d.errcheck = _errcheck
	gl.glRasterPos2d.restype = GLvoid
	gl.glRasterPos2d.argtypes = (GLdouble, GLdouble,)
	gl.glRasterPos2dv = _getProcAddress('glRasterPos2dv')
	gl.glRasterPos2dv.gl = gl
	gl.glRasterPos2dv.__name__ = 'glRasterPos2dv'
	gl.glRasterPos2dv.errcheck = _errcheck
	gl.glRasterPos2dv.restype = GLvoid
	gl.glRasterPos2dv.argtypes = (POINTER(GLdouble),)
	gl.glRasterPos2f = _getProcAddress('glRasterPos2f')
	gl.glRasterPos2f.gl = gl
	gl.glRasterPos2f.__name__ = 'glRasterPos2f'
	gl.glRasterPos2f.errcheck = _errcheck
	gl.glRasterPos2f.restype = GLvoid
	gl.glRasterPos2f.argtypes = (GLfloat, GLfloat,)
	gl.glRasterPos2fv = _getProcAddress('glRasterPos2fv')
	gl.glRasterPos2fv.gl = gl
	gl.glRasterPos2fv.__name__ = 'glRasterPos2fv'
	gl.glRasterPos2fv.errcheck = _errcheck
	gl.glRasterPos2fv.restype = GLvoid
	gl.glRasterPos2fv.argtypes = (POINTER(GLfloat),)
	gl.glRasterPos2i = _getProcAddress('glRasterPos2i')
	gl.glRasterPos2i.gl = gl
	gl.glRasterPos2i.__name__ = 'glRasterPos2i'
	gl.glRasterPos2i.errcheck = _errcheck
	gl.glRasterPos2i.restype = GLvoid
	gl.glRasterPos2i.argtypes = (GLint, GLint,)
	gl.glRasterPos2iv = _getProcAddress('glRasterPos2iv')
	gl.glRasterPos2iv.gl = gl
	gl.glRasterPos2iv.__name__ = 'glRasterPos2iv'
	gl.glRasterPos2iv.errcheck = _errcheck
	gl.glRasterPos2iv.restype = GLvoid
	gl.glRasterPos2iv.argtypes = (POINTER(GLint),)
	gl.glRasterPos2s = _getProcAddress('glRasterPos2s')
	gl.glRasterPos2s.gl = gl
	gl.glRasterPos2s.__name__ = 'glRasterPos2s'
	gl.glRasterPos2s.errcheck = _errcheck
	gl.glRasterPos2s.restype = GLvoid
	gl.glRasterPos2s.argtypes = (GLshort, GLshort,)
	gl.glRasterPos2sv = _getProcAddress('glRasterPos2sv')
	gl.glRasterPos2sv.gl = gl
	gl.glRasterPos2sv.__name__ = 'glRasterPos2sv'
	gl.glRasterPos2sv.errcheck = _errcheck
	gl.glRasterPos2sv.restype = GLvoid
	gl.glRasterPos2sv.argtypes = (POINTER(GLshort),)
	gl.glRasterPos2xOES = _getProcAddress('glRasterPos2xOES')
	gl.glRasterPos2xOES.gl = gl
	gl.glRasterPos2xOES.__name__ = 'glRasterPos2xOES'
	gl.glRasterPos2xOES.errcheck = _errcheck
	gl.glRasterPos2xOES.restype = GLvoid
	gl.glRasterPos2xOES.argtypes = (GLfixed, GLfixed,)
	gl.glRasterPos2xvOES = _getProcAddress('glRasterPos2xvOES')
	gl.glRasterPos2xvOES.gl = gl
	gl.glRasterPos2xvOES.__name__ = 'glRasterPos2xvOES'
	gl.glRasterPos2xvOES.errcheck = _errcheck
	gl.glRasterPos2xvOES.restype = GLvoid
	gl.glRasterPos2xvOES.argtypes = (POINTER(GLfixed),)
	gl.glRasterPos3d = _getProcAddress('glRasterPos3d')
	gl.glRasterPos3d.gl = gl
	gl.glRasterPos3d.__name__ = 'glRasterPos3d'
	gl.glRasterPos3d.errcheck = _errcheck
	gl.glRasterPos3d.restype = GLvoid
	gl.glRasterPos3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glRasterPos3dv = _getProcAddress('glRasterPos3dv')
	gl.glRasterPos3dv.gl = gl
	gl.glRasterPos3dv.__name__ = 'glRasterPos3dv'
	gl.glRasterPos3dv.errcheck = _errcheck
	gl.glRasterPos3dv.restype = GLvoid
	gl.glRasterPos3dv.argtypes = (POINTER(GLdouble),)
	gl.glRasterPos3f = _getProcAddress('glRasterPos3f')
	gl.glRasterPos3f.gl = gl
	gl.glRasterPos3f.__name__ = 'glRasterPos3f'
	gl.glRasterPos3f.errcheck = _errcheck
	gl.glRasterPos3f.restype = GLvoid
	gl.glRasterPos3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glRasterPos3fv = _getProcAddress('glRasterPos3fv')
	gl.glRasterPos3fv.gl = gl
	gl.glRasterPos3fv.__name__ = 'glRasterPos3fv'
	gl.glRasterPos3fv.errcheck = _errcheck
	gl.glRasterPos3fv.restype = GLvoid
	gl.glRasterPos3fv.argtypes = (POINTER(GLfloat),)
	gl.glRasterPos3i = _getProcAddress('glRasterPos3i')
	gl.glRasterPos3i.gl = gl
	gl.glRasterPos3i.__name__ = 'glRasterPos3i'
	gl.glRasterPos3i.errcheck = _errcheck
	gl.glRasterPos3i.restype = GLvoid
	gl.glRasterPos3i.argtypes = (GLint, GLint, GLint,)
	gl.glRasterPos3iv = _getProcAddress('glRasterPos3iv')
	gl.glRasterPos3iv.gl = gl
	gl.glRasterPos3iv.__name__ = 'glRasterPos3iv'
	gl.glRasterPos3iv.errcheck = _errcheck
	gl.glRasterPos3iv.restype = GLvoid
	gl.glRasterPos3iv.argtypes = (POINTER(GLint),)
	gl.glRasterPos3s = _getProcAddress('glRasterPos3s')
	gl.glRasterPos3s.gl = gl
	gl.glRasterPos3s.__name__ = 'glRasterPos3s'
	gl.glRasterPos3s.errcheck = _errcheck
	gl.glRasterPos3s.restype = GLvoid
	gl.glRasterPos3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glRasterPos3sv = _getProcAddress('glRasterPos3sv')
	gl.glRasterPos3sv.gl = gl
	gl.glRasterPos3sv.__name__ = 'glRasterPos3sv'
	gl.glRasterPos3sv.errcheck = _errcheck
	gl.glRasterPos3sv.restype = GLvoid
	gl.glRasterPos3sv.argtypes = (POINTER(GLshort),)
	gl.glRasterPos3xOES = _getProcAddress('glRasterPos3xOES')
	gl.glRasterPos3xOES.gl = gl
	gl.glRasterPos3xOES.__name__ = 'glRasterPos3xOES'
	gl.glRasterPos3xOES.errcheck = _errcheck
	gl.glRasterPos3xOES.restype = GLvoid
	gl.glRasterPos3xOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glRasterPos3xvOES = _getProcAddress('glRasterPos3xvOES')
	gl.glRasterPos3xvOES.gl = gl
	gl.glRasterPos3xvOES.__name__ = 'glRasterPos3xvOES'
	gl.glRasterPos3xvOES.errcheck = _errcheck
	gl.glRasterPos3xvOES.restype = GLvoid
	gl.glRasterPos3xvOES.argtypes = (POINTER(GLfixed),)
	gl.glRasterPos4d = _getProcAddress('glRasterPos4d')
	gl.glRasterPos4d.gl = gl
	gl.glRasterPos4d.__name__ = 'glRasterPos4d'
	gl.glRasterPos4d.errcheck = _errcheck
	gl.glRasterPos4d.restype = GLvoid
	gl.glRasterPos4d.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glRasterPos4dv = _getProcAddress('glRasterPos4dv')
	gl.glRasterPos4dv.gl = gl
	gl.glRasterPos4dv.__name__ = 'glRasterPos4dv'
	gl.glRasterPos4dv.errcheck = _errcheck
	gl.glRasterPos4dv.restype = GLvoid
	gl.glRasterPos4dv.argtypes = (POINTER(GLdouble),)
	gl.glRasterPos4f = _getProcAddress('glRasterPos4f')
	gl.glRasterPos4f.gl = gl
	gl.glRasterPos4f.__name__ = 'glRasterPos4f'
	gl.glRasterPos4f.errcheck = _errcheck
	gl.glRasterPos4f.restype = GLvoid
	gl.glRasterPos4f.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glRasterPos4fv = _getProcAddress('glRasterPos4fv')
	gl.glRasterPos4fv.gl = gl
	gl.glRasterPos4fv.__name__ = 'glRasterPos4fv'
	gl.glRasterPos4fv.errcheck = _errcheck
	gl.glRasterPos4fv.restype = GLvoid
	gl.glRasterPos4fv.argtypes = (POINTER(GLfloat),)
	gl.glRasterPos4i = _getProcAddress('glRasterPos4i')
	gl.glRasterPos4i.gl = gl
	gl.glRasterPos4i.__name__ = 'glRasterPos4i'
	gl.glRasterPos4i.errcheck = _errcheck
	gl.glRasterPos4i.restype = GLvoid
	gl.glRasterPos4i.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glRasterPos4iv = _getProcAddress('glRasterPos4iv')
	gl.glRasterPos4iv.gl = gl
	gl.glRasterPos4iv.__name__ = 'glRasterPos4iv'
	gl.glRasterPos4iv.errcheck = _errcheck
	gl.glRasterPos4iv.restype = GLvoid
	gl.glRasterPos4iv.argtypes = (POINTER(GLint),)
	gl.glRasterPos4s = _getProcAddress('glRasterPos4s')
	gl.glRasterPos4s.gl = gl
	gl.glRasterPos4s.__name__ = 'glRasterPos4s'
	gl.glRasterPos4s.errcheck = _errcheck
	gl.glRasterPos4s.restype = GLvoid
	gl.glRasterPos4s.argtypes = (GLshort, GLshort, GLshort, GLshort,)
	gl.glRasterPos4sv = _getProcAddress('glRasterPos4sv')
	gl.glRasterPos4sv.gl = gl
	gl.glRasterPos4sv.__name__ = 'glRasterPos4sv'
	gl.glRasterPos4sv.errcheck = _errcheck
	gl.glRasterPos4sv.restype = GLvoid
	gl.glRasterPos4sv.argtypes = (POINTER(GLshort),)
	gl.glRasterPos4xOES = _getProcAddress('glRasterPos4xOES')
	gl.glRasterPos4xOES.gl = gl
	gl.glRasterPos4xOES.__name__ = 'glRasterPos4xOES'
	gl.glRasterPos4xOES.errcheck = _errcheck
	gl.glRasterPos4xOES.restype = GLvoid
	gl.glRasterPos4xOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glRasterPos4xvOES = _getProcAddress('glRasterPos4xvOES')
	gl.glRasterPos4xvOES.gl = gl
	gl.glRasterPos4xvOES.__name__ = 'glRasterPos4xvOES'
	gl.glRasterPos4xvOES.errcheck = _errcheck
	gl.glRasterPos4xvOES.restype = GLvoid
	gl.glRasterPos4xvOES.argtypes = (POINTER(GLfixed),)
	gl.glReadBuffer = _getProcAddress('glReadBuffer')
	gl.glReadBuffer.gl = gl
	gl.glReadBuffer.__name__ = 'glReadBuffer'
	gl.glReadBuffer.errcheck = _errcheck
	gl.glReadBuffer.restype = GLvoid
	gl.glReadBuffer.argtypes = (GLenum,)
	gl.glReadBufferIndexedEXT = _getProcAddress('glReadBufferIndexedEXT')
	gl.glReadBufferIndexedEXT.gl = gl
	gl.glReadBufferIndexedEXT.__name__ = 'glReadBufferIndexedEXT'
	gl.glReadBufferIndexedEXT.errcheck = _errcheck
	gl.glReadBufferIndexedEXT.restype = GLvoid
	gl.glReadBufferIndexedEXT.argtypes = (GLenum, GLint,)
	gl.glReadBufferNV = _getProcAddress('glReadBufferNV')
	gl.glReadBufferNV.gl = gl
	gl.glReadBufferNV.__name__ = 'glReadBufferNV'
	gl.glReadBufferNV.errcheck = _errcheck
	gl.glReadBufferNV.restype = GLvoid
	gl.glReadBufferNV.argtypes = (GLenum,)
	gl.glReadInstrumentsSGIX = _getProcAddress('glReadInstrumentsSGIX')
	gl.glReadInstrumentsSGIX.gl = gl
	gl.glReadInstrumentsSGIX.__name__ = 'glReadInstrumentsSGIX'
	gl.glReadInstrumentsSGIX.errcheck = _errcheck
	gl.glReadInstrumentsSGIX.restype = GLvoid
	gl.glReadInstrumentsSGIX.argtypes = (GLint,)
	gl.glReadPixels = _getProcAddress('glReadPixels')
	gl.glReadPixels.gl = gl
	gl.glReadPixels.__name__ = 'glReadPixels'
	gl.glReadPixels.errcheck = _errcheck
	gl.glReadPixels.restype = GLvoid
	gl.glReadPixels.argtypes = (GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glReadnPixels = _getProcAddress('glReadnPixels')
	gl.glReadnPixels.gl = gl
	gl.glReadnPixels.__name__ = 'glReadnPixels'
	gl.glReadnPixels.errcheck = _errcheck
	gl.glReadnPixels.restype = GLvoid
	gl.glReadnPixels.argtypes = (GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glReadnPixelsARB = _getProcAddress('glReadnPixelsARB')
	gl.glReadnPixelsARB.gl = gl
	gl.glReadnPixelsARB.__name__ = 'glReadnPixelsARB'
	gl.glReadnPixelsARB.errcheck = _errcheck
	gl.glReadnPixelsARB.restype = GLvoid
	gl.glReadnPixelsARB.argtypes = (GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glReadnPixelsEXT = _getProcAddress('glReadnPixelsEXT')
	gl.glReadnPixelsEXT.gl = gl
	gl.glReadnPixelsEXT.__name__ = 'glReadnPixelsEXT'
	gl.glReadnPixelsEXT.errcheck = _errcheck
	gl.glReadnPixelsEXT.restype = GLvoid
	gl.glReadnPixelsEXT.argtypes = (GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glReadnPixelsKHR = _getProcAddress('glReadnPixelsKHR')
	gl.glReadnPixelsKHR.gl = gl
	gl.glReadnPixelsKHR.__name__ = 'glReadnPixelsKHR'
	gl.glReadnPixelsKHR.errcheck = _errcheck
	gl.glReadnPixelsKHR.restype = GLvoid
	gl.glReadnPixelsKHR.argtypes = (GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glRectd = _getProcAddress('glRectd')
	gl.glRectd.gl = gl
	gl.glRectd.__name__ = 'glRectd'
	gl.glRectd.errcheck = _errcheck
	gl.glRectd.restype = GLvoid
	gl.glRectd.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glRectdv = _getProcAddress('glRectdv')
	gl.glRectdv.gl = gl
	gl.glRectdv.__name__ = 'glRectdv'
	gl.glRectdv.errcheck = _errcheck
	gl.glRectdv.restype = GLvoid
	gl.glRectdv.argtypes = (POINTER(GLdouble), POINTER(GLdouble),)
	gl.glRectf = _getProcAddress('glRectf')
	gl.glRectf.gl = gl
	gl.glRectf.__name__ = 'glRectf'
	gl.glRectf.errcheck = _errcheck
	gl.glRectf.restype = GLvoid
	gl.glRectf.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glRectfv = _getProcAddress('glRectfv')
	gl.glRectfv.gl = gl
	gl.glRectfv.__name__ = 'glRectfv'
	gl.glRectfv.errcheck = _errcheck
	gl.glRectfv.restype = GLvoid
	gl.glRectfv.argtypes = (POINTER(GLfloat), POINTER(GLfloat),)
	gl.glRecti = _getProcAddress('glRecti')
	gl.glRecti.gl = gl
	gl.glRecti.__name__ = 'glRecti'
	gl.glRecti.errcheck = _errcheck
	gl.glRecti.restype = GLvoid
	gl.glRecti.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glRectiv = _getProcAddress('glRectiv')
	gl.glRectiv.gl = gl
	gl.glRectiv.__name__ = 'glRectiv'
	gl.glRectiv.errcheck = _errcheck
	gl.glRectiv.restype = GLvoid
	gl.glRectiv.argtypes = (POINTER(GLint), POINTER(GLint),)
	gl.glRects = _getProcAddress('glRects')
	gl.glRects.gl = gl
	gl.glRects.__name__ = 'glRects'
	gl.glRects.errcheck = _errcheck
	gl.glRects.restype = GLvoid
	gl.glRects.argtypes = (GLshort, GLshort, GLshort, GLshort,)
	gl.glRectsv = _getProcAddress('glRectsv')
	gl.glRectsv.gl = gl
	gl.glRectsv.__name__ = 'glRectsv'
	gl.glRectsv.errcheck = _errcheck
	gl.glRectsv.restype = GLvoid
	gl.glRectsv.argtypes = (POINTER(GLshort), POINTER(GLshort),)
	gl.glRectxOES = _getProcAddress('glRectxOES')
	gl.glRectxOES.gl = gl
	gl.glRectxOES.__name__ = 'glRectxOES'
	gl.glRectxOES.errcheck = _errcheck
	gl.glRectxOES.restype = GLvoid
	gl.glRectxOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glRectxvOES = _getProcAddress('glRectxvOES')
	gl.glRectxvOES.gl = gl
	gl.glRectxvOES.__name__ = 'glRectxvOES'
	gl.glRectxvOES.errcheck = _errcheck
	gl.glRectxvOES.restype = GLvoid
	gl.glRectxvOES.argtypes = (POINTER(GLfixed), POINTER(GLfixed),)
	gl.glReferencePlaneSGIX = _getProcAddress('glReferencePlaneSGIX')
	gl.glReferencePlaneSGIX.gl = gl
	gl.glReferencePlaneSGIX.__name__ = 'glReferencePlaneSGIX'
	gl.glReferencePlaneSGIX.errcheck = _errcheck
	gl.glReferencePlaneSGIX.restype = GLvoid
	gl.glReferencePlaneSGIX.argtypes = (POINTER(GLdouble),)
	gl.glReleaseShaderCompiler = _getProcAddress('glReleaseShaderCompiler')
	gl.glReleaseShaderCompiler.gl = gl
	gl.glReleaseShaderCompiler.__name__ = 'glReleaseShaderCompiler'
	gl.glReleaseShaderCompiler.errcheck = _errcheck
	gl.glReleaseShaderCompiler.restype = GLvoid
	gl.glReleaseShaderCompiler.argtypes = ()
	gl.glRenderMode = _getProcAddress('glRenderMode')
	gl.glRenderMode.gl = gl
	gl.glRenderMode.__name__ = 'glRenderMode'
	gl.glRenderMode.errcheck = _errcheck
	gl.glRenderMode.restype = GLint
	gl.glRenderMode.argtypes = (GLenum,)
	gl.glRenderbufferStorage = _getProcAddress('glRenderbufferStorage')
	gl.glRenderbufferStorage.gl = gl
	gl.glRenderbufferStorage.__name__ = 'glRenderbufferStorage'
	gl.glRenderbufferStorage.errcheck = _errcheck
	gl.glRenderbufferStorage.restype = GLvoid
	gl.glRenderbufferStorage.argtypes = (GLenum, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageEXT = _getProcAddress('glRenderbufferStorageEXT')
	gl.glRenderbufferStorageEXT.gl = gl
	gl.glRenderbufferStorageEXT.__name__ = 'glRenderbufferStorageEXT'
	gl.glRenderbufferStorageEXT.errcheck = _errcheck
	gl.glRenderbufferStorageEXT.restype = GLvoid
	gl.glRenderbufferStorageEXT.argtypes = (GLenum, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisample = _getProcAddress('glRenderbufferStorageMultisample')
	gl.glRenderbufferStorageMultisample.gl = gl
	gl.glRenderbufferStorageMultisample.__name__ = 'glRenderbufferStorageMultisample'
	gl.glRenderbufferStorageMultisample.errcheck = _errcheck
	gl.glRenderbufferStorageMultisample.restype = GLvoid
	gl.glRenderbufferStorageMultisample.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisampleANGLE = _getProcAddress('glRenderbufferStorageMultisampleANGLE')
	gl.glRenderbufferStorageMultisampleANGLE.gl = gl
	gl.glRenderbufferStorageMultisampleANGLE.__name__ = 'glRenderbufferStorageMultisampleANGLE'
	gl.glRenderbufferStorageMultisampleANGLE.errcheck = _errcheck
	gl.glRenderbufferStorageMultisampleANGLE.restype = GLvoid
	gl.glRenderbufferStorageMultisampleANGLE.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisampleAPPLE = _getProcAddress('glRenderbufferStorageMultisampleAPPLE')
	gl.glRenderbufferStorageMultisampleAPPLE.gl = gl
	gl.glRenderbufferStorageMultisampleAPPLE.__name__ = 'glRenderbufferStorageMultisampleAPPLE'
	gl.glRenderbufferStorageMultisampleAPPLE.errcheck = _errcheck
	gl.glRenderbufferStorageMultisampleAPPLE.restype = GLvoid
	gl.glRenderbufferStorageMultisampleAPPLE.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisampleCoverageNV = _getProcAddress('glRenderbufferStorageMultisampleCoverageNV')
	gl.glRenderbufferStorageMultisampleCoverageNV.gl = gl
	gl.glRenderbufferStorageMultisampleCoverageNV.__name__ = 'glRenderbufferStorageMultisampleCoverageNV'
	gl.glRenderbufferStorageMultisampleCoverageNV.errcheck = _errcheck
	gl.glRenderbufferStorageMultisampleCoverageNV.restype = GLvoid
	gl.glRenderbufferStorageMultisampleCoverageNV.argtypes = (GLenum, GLsizei, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisampleEXT = _getProcAddress('glRenderbufferStorageMultisampleEXT')
	gl.glRenderbufferStorageMultisampleEXT.gl = gl
	gl.glRenderbufferStorageMultisampleEXT.__name__ = 'glRenderbufferStorageMultisampleEXT'
	gl.glRenderbufferStorageMultisampleEXT.errcheck = _errcheck
	gl.glRenderbufferStorageMultisampleEXT.restype = GLvoid
	gl.glRenderbufferStorageMultisampleEXT.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisampleIMG = _getProcAddress('glRenderbufferStorageMultisampleIMG')
	gl.glRenderbufferStorageMultisampleIMG.gl = gl
	gl.glRenderbufferStorageMultisampleIMG.__name__ = 'glRenderbufferStorageMultisampleIMG'
	gl.glRenderbufferStorageMultisampleIMG.errcheck = _errcheck
	gl.glRenderbufferStorageMultisampleIMG.restype = GLvoid
	gl.glRenderbufferStorageMultisampleIMG.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageMultisampleNV = _getProcAddress('glRenderbufferStorageMultisampleNV')
	gl.glRenderbufferStorageMultisampleNV.gl = gl
	gl.glRenderbufferStorageMultisampleNV.__name__ = 'glRenderbufferStorageMultisampleNV'
	gl.glRenderbufferStorageMultisampleNV.errcheck = _errcheck
	gl.glRenderbufferStorageMultisampleNV.restype = GLvoid
	gl.glRenderbufferStorageMultisampleNV.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glRenderbufferStorageOES = _getProcAddress('glRenderbufferStorageOES')
	gl.glRenderbufferStorageOES.gl = gl
	gl.glRenderbufferStorageOES.__name__ = 'glRenderbufferStorageOES'
	gl.glRenderbufferStorageOES.errcheck = _errcheck
	gl.glRenderbufferStorageOES.restype = GLvoid
	gl.glRenderbufferStorageOES.argtypes = (GLenum, GLenum, GLsizei, GLsizei,)
	gl.glReplacementCodePointerSUN = _getProcAddress('glReplacementCodePointerSUN')
	gl.glReplacementCodePointerSUN.gl = gl
	gl.glReplacementCodePointerSUN.__name__ = 'glReplacementCodePointerSUN'
	gl.glReplacementCodePointerSUN.errcheck = _errcheck
	gl.glReplacementCodePointerSUN.restype = GLvoid
	gl.glReplacementCodePointerSUN.argtypes = (GLenum, GLsizei, POINTER(POINTER(GLvoid)),)
	gl.glReplacementCodeubSUN = _getProcAddress('glReplacementCodeubSUN')
	gl.glReplacementCodeubSUN.gl = gl
	gl.glReplacementCodeubSUN.__name__ = 'glReplacementCodeubSUN'
	gl.glReplacementCodeubSUN.errcheck = _errcheck
	gl.glReplacementCodeubSUN.restype = GLvoid
	gl.glReplacementCodeubSUN.argtypes = (GLubyte,)
	gl.glReplacementCodeubvSUN = _getProcAddress('glReplacementCodeubvSUN')
	gl.glReplacementCodeubvSUN.gl = gl
	gl.glReplacementCodeubvSUN.__name__ = 'glReplacementCodeubvSUN'
	gl.glReplacementCodeubvSUN.errcheck = _errcheck
	gl.glReplacementCodeubvSUN.restype = GLvoid
	gl.glReplacementCodeubvSUN.argtypes = (POINTER(GLubyte),)
	gl.glReplacementCodeuiColor3fVertex3fSUN = _getProcAddress('glReplacementCodeuiColor3fVertex3fSUN')
	gl.glReplacementCodeuiColor3fVertex3fSUN.gl = gl
	gl.glReplacementCodeuiColor3fVertex3fSUN.__name__ = 'glReplacementCodeuiColor3fVertex3fSUN'
	gl.glReplacementCodeuiColor3fVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiColor3fVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiColor3fVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiColor3fVertex3fvSUN = _getProcAddress('glReplacementCodeuiColor3fVertex3fvSUN')
	gl.glReplacementCodeuiColor3fVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiColor3fVertex3fvSUN.__name__ = 'glReplacementCodeuiColor3fVertex3fvSUN'
	gl.glReplacementCodeuiColor3fVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiColor3fVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiColor3fVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glReplacementCodeuiColor4fNormal3fVertex3fSUN = _getProcAddress('glReplacementCodeuiColor4fNormal3fVertex3fSUN')
	gl.glReplacementCodeuiColor4fNormal3fVertex3fSUN.gl = gl
	gl.glReplacementCodeuiColor4fNormal3fVertex3fSUN.__name__ = 'glReplacementCodeuiColor4fNormal3fVertex3fSUN'
	gl.glReplacementCodeuiColor4fNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiColor4fNormal3fVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiColor4fNormal3fVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiColor4fNormal3fVertex3fvSUN = _getProcAddress('glReplacementCodeuiColor4fNormal3fVertex3fvSUN')
	gl.glReplacementCodeuiColor4fNormal3fVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiColor4fNormal3fVertex3fvSUN.__name__ = 'glReplacementCodeuiColor4fNormal3fVertex3fvSUN'
	gl.glReplacementCodeuiColor4fNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiColor4fNormal3fVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiColor4fNormal3fVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glReplacementCodeuiColor4ubVertex3fSUN = _getProcAddress('glReplacementCodeuiColor4ubVertex3fSUN')
	gl.glReplacementCodeuiColor4ubVertex3fSUN.gl = gl
	gl.glReplacementCodeuiColor4ubVertex3fSUN.__name__ = 'glReplacementCodeuiColor4ubVertex3fSUN'
	gl.glReplacementCodeuiColor4ubVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiColor4ubVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiColor4ubVertex3fSUN.argtypes = (GLuint, GLubyte, GLubyte, GLubyte, GLubyte, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiColor4ubVertex3fvSUN = _getProcAddress('glReplacementCodeuiColor4ubVertex3fvSUN')
	gl.glReplacementCodeuiColor4ubVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiColor4ubVertex3fvSUN.__name__ = 'glReplacementCodeuiColor4ubVertex3fvSUN'
	gl.glReplacementCodeuiColor4ubVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiColor4ubVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiColor4ubVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLubyte), POINTER(GLfloat),)
	gl.glReplacementCodeuiNormal3fVertex3fSUN = _getProcAddress('glReplacementCodeuiNormal3fVertex3fSUN')
	gl.glReplacementCodeuiNormal3fVertex3fSUN.gl = gl
	gl.glReplacementCodeuiNormal3fVertex3fSUN.__name__ = 'glReplacementCodeuiNormal3fVertex3fSUN'
	gl.glReplacementCodeuiNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiNormal3fVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiNormal3fVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiNormal3fVertex3fvSUN = _getProcAddress('glReplacementCodeuiNormal3fVertex3fvSUN')
	gl.glReplacementCodeuiNormal3fVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiNormal3fVertex3fvSUN.__name__ = 'glReplacementCodeuiNormal3fVertex3fvSUN'
	gl.glReplacementCodeuiNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiNormal3fVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiNormal3fVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glReplacementCodeuiSUN = _getProcAddress('glReplacementCodeuiSUN')
	gl.glReplacementCodeuiSUN.gl = gl
	gl.glReplacementCodeuiSUN.__name__ = 'glReplacementCodeuiSUN'
	gl.glReplacementCodeuiSUN.errcheck = _errcheck
	gl.glReplacementCodeuiSUN.restype = GLvoid
	gl.glReplacementCodeuiSUN.argtypes = (GLuint,)
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN = _getProcAddress('glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN')
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN.gl = gl
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN.__name__ = 'glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN'
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN = _getProcAddress('glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN')
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN.__name__ = 'glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN'
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiTexCoord2fColor4fNormal3fVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN = _getProcAddress('glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN')
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN.gl = gl
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN.__name__ = 'glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN'
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN = _getProcAddress('glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN')
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN.__name__ = 'glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN'
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiTexCoord2fNormal3fVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glReplacementCodeuiTexCoord2fVertex3fSUN = _getProcAddress('glReplacementCodeuiTexCoord2fVertex3fSUN')
	gl.glReplacementCodeuiTexCoord2fVertex3fSUN.gl = gl
	gl.glReplacementCodeuiTexCoord2fVertex3fSUN.__name__ = 'glReplacementCodeuiTexCoord2fVertex3fSUN'
	gl.glReplacementCodeuiTexCoord2fVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiTexCoord2fVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiTexCoord2fVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiTexCoord2fVertex3fvSUN = _getProcAddress('glReplacementCodeuiTexCoord2fVertex3fvSUN')
	gl.glReplacementCodeuiTexCoord2fVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiTexCoord2fVertex3fvSUN.__name__ = 'glReplacementCodeuiTexCoord2fVertex3fvSUN'
	gl.glReplacementCodeuiTexCoord2fVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiTexCoord2fVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiTexCoord2fVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glReplacementCodeuiVertex3fSUN = _getProcAddress('glReplacementCodeuiVertex3fSUN')
	gl.glReplacementCodeuiVertex3fSUN.gl = gl
	gl.glReplacementCodeuiVertex3fSUN.__name__ = 'glReplacementCodeuiVertex3fSUN'
	gl.glReplacementCodeuiVertex3fSUN.errcheck = _errcheck
	gl.glReplacementCodeuiVertex3fSUN.restype = GLvoid
	gl.glReplacementCodeuiVertex3fSUN.argtypes = (GLuint, GLfloat, GLfloat, GLfloat,)
	gl.glReplacementCodeuiVertex3fvSUN = _getProcAddress('glReplacementCodeuiVertex3fvSUN')
	gl.glReplacementCodeuiVertex3fvSUN.gl = gl
	gl.glReplacementCodeuiVertex3fvSUN.__name__ = 'glReplacementCodeuiVertex3fvSUN'
	gl.glReplacementCodeuiVertex3fvSUN.errcheck = _errcheck
	gl.glReplacementCodeuiVertex3fvSUN.restype = GLvoid
	gl.glReplacementCodeuiVertex3fvSUN.argtypes = (POINTER(GLuint), POINTER(GLfloat),)
	gl.glReplacementCodeuivSUN = _getProcAddress('glReplacementCodeuivSUN')
	gl.glReplacementCodeuivSUN.gl = gl
	gl.glReplacementCodeuivSUN.__name__ = 'glReplacementCodeuivSUN'
	gl.glReplacementCodeuivSUN.errcheck = _errcheck
	gl.glReplacementCodeuivSUN.restype = GLvoid
	gl.glReplacementCodeuivSUN.argtypes = (POINTER(GLuint),)
	gl.glReplacementCodeusSUN = _getProcAddress('glReplacementCodeusSUN')
	gl.glReplacementCodeusSUN.gl = gl
	gl.glReplacementCodeusSUN.__name__ = 'glReplacementCodeusSUN'
	gl.glReplacementCodeusSUN.errcheck = _errcheck
	gl.glReplacementCodeusSUN.restype = GLvoid
	gl.glReplacementCodeusSUN.argtypes = (GLushort,)
	gl.glReplacementCodeusvSUN = _getProcAddress('glReplacementCodeusvSUN')
	gl.glReplacementCodeusvSUN.gl = gl
	gl.glReplacementCodeusvSUN.__name__ = 'glReplacementCodeusvSUN'
	gl.glReplacementCodeusvSUN.errcheck = _errcheck
	gl.glReplacementCodeusvSUN.restype = GLvoid
	gl.glReplacementCodeusvSUN.argtypes = (POINTER(GLushort),)
	gl.glRequestResidentProgramsNV = _getProcAddress('glRequestResidentProgramsNV')
	gl.glRequestResidentProgramsNV.gl = gl
	gl.glRequestResidentProgramsNV.__name__ = 'glRequestResidentProgramsNV'
	gl.glRequestResidentProgramsNV.errcheck = _errcheck
	gl.glRequestResidentProgramsNV.restype = GLvoid
	gl.glRequestResidentProgramsNV.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glResetHistogram = _getProcAddress('glResetHistogram')
	gl.glResetHistogram.gl = gl
	gl.glResetHistogram.__name__ = 'glResetHistogram'
	gl.glResetHistogram.errcheck = _errcheck
	gl.glResetHistogram.restype = GLvoid
	gl.glResetHistogram.argtypes = (GLenum,)
	gl.glResetHistogramEXT = _getProcAddress('glResetHistogramEXT')
	gl.glResetHistogramEXT.gl = gl
	gl.glResetHistogramEXT.__name__ = 'glResetHistogramEXT'
	gl.glResetHistogramEXT.errcheck = _errcheck
	gl.glResetHistogramEXT.restype = GLvoid
	gl.glResetHistogramEXT.argtypes = (GLenum,)
	gl.glResetMinmax = _getProcAddress('glResetMinmax')
	gl.glResetMinmax.gl = gl
	gl.glResetMinmax.__name__ = 'glResetMinmax'
	gl.glResetMinmax.errcheck = _errcheck
	gl.glResetMinmax.restype = GLvoid
	gl.glResetMinmax.argtypes = (GLenum,)
	gl.glResetMinmaxEXT = _getProcAddress('glResetMinmaxEXT')
	gl.glResetMinmaxEXT.gl = gl
	gl.glResetMinmaxEXT.__name__ = 'glResetMinmaxEXT'
	gl.glResetMinmaxEXT.errcheck = _errcheck
	gl.glResetMinmaxEXT.restype = GLvoid
	gl.glResetMinmaxEXT.argtypes = (GLenum,)
	gl.glResizeBuffersMESA = _getProcAddress('glResizeBuffersMESA')
	gl.glResizeBuffersMESA.gl = gl
	gl.glResizeBuffersMESA.__name__ = 'glResizeBuffersMESA'
	gl.glResizeBuffersMESA.errcheck = _errcheck
	gl.glResizeBuffersMESA.restype = GLvoid
	gl.glResizeBuffersMESA.argtypes = ()
	gl.glResolveMultisampleFramebufferAPPLE = _getProcAddress('glResolveMultisampleFramebufferAPPLE')
	gl.glResolveMultisampleFramebufferAPPLE.gl = gl
	gl.glResolveMultisampleFramebufferAPPLE.__name__ = 'glResolveMultisampleFramebufferAPPLE'
	gl.glResolveMultisampleFramebufferAPPLE.errcheck = _errcheck
	gl.glResolveMultisampleFramebufferAPPLE.restype = GLvoid
	gl.glResolveMultisampleFramebufferAPPLE.argtypes = ()
	gl.glResumeTransformFeedback = _getProcAddress('glResumeTransformFeedback')
	gl.glResumeTransformFeedback.gl = gl
	gl.glResumeTransformFeedback.__name__ = 'glResumeTransformFeedback'
	gl.glResumeTransformFeedback.errcheck = _errcheck
	gl.glResumeTransformFeedback.restype = GLvoid
	gl.glResumeTransformFeedback.argtypes = ()
	gl.glResumeTransformFeedbackNV = _getProcAddress('glResumeTransformFeedbackNV')
	gl.glResumeTransformFeedbackNV.gl = gl
	gl.glResumeTransformFeedbackNV.__name__ = 'glResumeTransformFeedbackNV'
	gl.glResumeTransformFeedbackNV.errcheck = _errcheck
	gl.glResumeTransformFeedbackNV.restype = GLvoid
	gl.glResumeTransformFeedbackNV.argtypes = ()
	gl.glRotated = _getProcAddress('glRotated')
	gl.glRotated.gl = gl
	gl.glRotated.__name__ = 'glRotated'
	gl.glRotated.errcheck = _errcheck
	gl.glRotated.restype = GLvoid
	gl.glRotated.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glRotatef = _getProcAddress('glRotatef')
	gl.glRotatef.gl = gl
	gl.glRotatef.__name__ = 'glRotatef'
	gl.glRotatef.errcheck = _errcheck
	gl.glRotatef.restype = GLvoid
	gl.glRotatef.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glRotatex = _getProcAddress('glRotatex')
	gl.glRotatex.gl = gl
	gl.glRotatex.__name__ = 'glRotatex'
	gl.glRotatex.errcheck = _errcheck
	gl.glRotatex.restype = GLvoid
	gl.glRotatex.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glRotatexOES = _getProcAddress('glRotatexOES')
	gl.glRotatexOES.gl = gl
	gl.glRotatexOES.__name__ = 'glRotatexOES'
	gl.glRotatexOES.errcheck = _errcheck
	gl.glRotatexOES.restype = GLvoid
	gl.glRotatexOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glSampleCoverage = _getProcAddress('glSampleCoverage')
	gl.glSampleCoverage.gl = gl
	gl.glSampleCoverage.__name__ = 'glSampleCoverage'
	gl.glSampleCoverage.errcheck = _errcheck
	gl.glSampleCoverage.restype = GLvoid
	gl.glSampleCoverage.argtypes = (GLfloat, GLboolean,)
	gl.glSampleCoverageARB = _getProcAddress('glSampleCoverageARB')
	gl.glSampleCoverageARB.gl = gl
	gl.glSampleCoverageARB.__name__ = 'glSampleCoverageARB'
	gl.glSampleCoverageARB.errcheck = _errcheck
	gl.glSampleCoverageARB.restype = GLvoid
	gl.glSampleCoverageARB.argtypes = (GLfloat, GLboolean,)
	gl.glSampleCoverageOES = _getProcAddress('glSampleCoverageOES')
	gl.glSampleCoverageOES.gl = gl
	gl.glSampleCoverageOES.__name__ = 'glSampleCoverageOES'
	gl.glSampleCoverageOES.errcheck = _errcheck
	gl.glSampleCoverageOES.restype = GLvoid
	gl.glSampleCoverageOES.argtypes = (GLfixed, GLboolean,)
	gl.glSampleCoveragex = _getProcAddress('glSampleCoveragex')
	gl.glSampleCoveragex.gl = gl
	gl.glSampleCoveragex.__name__ = 'glSampleCoveragex'
	gl.glSampleCoveragex.errcheck = _errcheck
	gl.glSampleCoveragex.restype = GLvoid
	gl.glSampleCoveragex.argtypes = (GLclampx, GLboolean,)
	gl.glSampleCoveragexOES = _getProcAddress('glSampleCoveragexOES')
	gl.glSampleCoveragexOES.gl = gl
	gl.glSampleCoveragexOES.__name__ = 'glSampleCoveragexOES'
	gl.glSampleCoveragexOES.errcheck = _errcheck
	gl.glSampleCoveragexOES.restype = GLvoid
	gl.glSampleCoveragexOES.argtypes = (GLclampx, GLboolean,)
	gl.glSampleMapATI = _getProcAddress('glSampleMapATI')
	gl.glSampleMapATI.gl = gl
	gl.glSampleMapATI.__name__ = 'glSampleMapATI'
	gl.glSampleMapATI.errcheck = _errcheck
	gl.glSampleMapATI.restype = GLvoid
	gl.glSampleMapATI.argtypes = (GLuint, GLuint, GLenum,)
	gl.glSampleMaskEXT = _getProcAddress('glSampleMaskEXT')
	gl.glSampleMaskEXT.gl = gl
	gl.glSampleMaskEXT.__name__ = 'glSampleMaskEXT'
	gl.glSampleMaskEXT.errcheck = _errcheck
	gl.glSampleMaskEXT.restype = GLvoid
	gl.glSampleMaskEXT.argtypes = (GLclampf, GLboolean,)
	gl.glSampleMaskIndexedNV = _getProcAddress('glSampleMaskIndexedNV')
	gl.glSampleMaskIndexedNV.gl = gl
	gl.glSampleMaskIndexedNV.__name__ = 'glSampleMaskIndexedNV'
	gl.glSampleMaskIndexedNV.errcheck = _errcheck
	gl.glSampleMaskIndexedNV.restype = GLvoid
	gl.glSampleMaskIndexedNV.argtypes = (GLuint, GLbitfield,)
	gl.glSampleMaskSGIS = _getProcAddress('glSampleMaskSGIS')
	gl.glSampleMaskSGIS.gl = gl
	gl.glSampleMaskSGIS.__name__ = 'glSampleMaskSGIS'
	gl.glSampleMaskSGIS.errcheck = _errcheck
	gl.glSampleMaskSGIS.restype = GLvoid
	gl.glSampleMaskSGIS.argtypes = (GLclampf, GLboolean,)
	gl.glSampleMaski = _getProcAddress('glSampleMaski')
	gl.glSampleMaski.gl = gl
	gl.glSampleMaski.__name__ = 'glSampleMaski'
	gl.glSampleMaski.errcheck = _errcheck
	gl.glSampleMaski.restype = GLvoid
	gl.glSampleMaski.argtypes = (GLuint, GLbitfield,)
	gl.glSamplePatternEXT = _getProcAddress('glSamplePatternEXT')
	gl.glSamplePatternEXT.gl = gl
	gl.glSamplePatternEXT.__name__ = 'glSamplePatternEXT'
	gl.glSamplePatternEXT.errcheck = _errcheck
	gl.glSamplePatternEXT.restype = GLvoid
	gl.glSamplePatternEXT.argtypes = (GLenum,)
	gl.glSamplePatternSGIS = _getProcAddress('glSamplePatternSGIS')
	gl.glSamplePatternSGIS.gl = gl
	gl.glSamplePatternSGIS.__name__ = 'glSamplePatternSGIS'
	gl.glSamplePatternSGIS.errcheck = _errcheck
	gl.glSamplePatternSGIS.restype = GLvoid
	gl.glSamplePatternSGIS.argtypes = (GLenum,)
	gl.glSamplerParameterIiv = _getProcAddress('glSamplerParameterIiv')
	gl.glSamplerParameterIiv.gl = gl
	gl.glSamplerParameterIiv.__name__ = 'glSamplerParameterIiv'
	gl.glSamplerParameterIiv.errcheck = _errcheck
	gl.glSamplerParameterIiv.restype = GLvoid
	gl.glSamplerParameterIiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glSamplerParameterIivEXT = _getProcAddress('glSamplerParameterIivEXT')
	gl.glSamplerParameterIivEXT.gl = gl
	gl.glSamplerParameterIivEXT.__name__ = 'glSamplerParameterIivEXT'
	gl.glSamplerParameterIivEXT.errcheck = _errcheck
	gl.glSamplerParameterIivEXT.restype = GLvoid
	gl.glSamplerParameterIivEXT.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glSamplerParameterIuiv = _getProcAddress('glSamplerParameterIuiv')
	gl.glSamplerParameterIuiv.gl = gl
	gl.glSamplerParameterIuiv.__name__ = 'glSamplerParameterIuiv'
	gl.glSamplerParameterIuiv.errcheck = _errcheck
	gl.glSamplerParameterIuiv.restype = GLvoid
	gl.glSamplerParameterIuiv.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glSamplerParameterIuivEXT = _getProcAddress('glSamplerParameterIuivEXT')
	gl.glSamplerParameterIuivEXT.gl = gl
	gl.glSamplerParameterIuivEXT.__name__ = 'glSamplerParameterIuivEXT'
	gl.glSamplerParameterIuivEXT.errcheck = _errcheck
	gl.glSamplerParameterIuivEXT.restype = GLvoid
	gl.glSamplerParameterIuivEXT.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glSamplerParameterf = _getProcAddress('glSamplerParameterf')
	gl.glSamplerParameterf.gl = gl
	gl.glSamplerParameterf.__name__ = 'glSamplerParameterf'
	gl.glSamplerParameterf.errcheck = _errcheck
	gl.glSamplerParameterf.restype = GLvoid
	gl.glSamplerParameterf.argtypes = (GLuint, GLenum, GLfloat,)
	gl.glSamplerParameterfv = _getProcAddress('glSamplerParameterfv')
	gl.glSamplerParameterfv.gl = gl
	gl.glSamplerParameterfv.__name__ = 'glSamplerParameterfv'
	gl.glSamplerParameterfv.errcheck = _errcheck
	gl.glSamplerParameterfv.restype = GLvoid
	gl.glSamplerParameterfv.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glSamplerParameteri = _getProcAddress('glSamplerParameteri')
	gl.glSamplerParameteri.gl = gl
	gl.glSamplerParameteri.__name__ = 'glSamplerParameteri'
	gl.glSamplerParameteri.errcheck = _errcheck
	gl.glSamplerParameteri.restype = GLvoid
	gl.glSamplerParameteri.argtypes = (GLuint, GLenum, GLint,)
	gl.glSamplerParameteriv = _getProcAddress('glSamplerParameteriv')
	gl.glSamplerParameteriv.gl = gl
	gl.glSamplerParameteriv.__name__ = 'glSamplerParameteriv'
	gl.glSamplerParameteriv.errcheck = _errcheck
	gl.glSamplerParameteriv.restype = GLvoid
	gl.glSamplerParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glScaled = _getProcAddress('glScaled')
	gl.glScaled.gl = gl
	gl.glScaled.__name__ = 'glScaled'
	gl.glScaled.errcheck = _errcheck
	gl.glScaled.restype = GLvoid
	gl.glScaled.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glScalef = _getProcAddress('glScalef')
	gl.glScalef.gl = gl
	gl.glScalef.__name__ = 'glScalef'
	gl.glScalef.errcheck = _errcheck
	gl.glScalef.restype = GLvoid
	gl.glScalef.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glScalex = _getProcAddress('glScalex')
	gl.glScalex.gl = gl
	gl.glScalex.__name__ = 'glScalex'
	gl.glScalex.errcheck = _errcheck
	gl.glScalex.restype = GLvoid
	gl.glScalex.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glScalexOES = _getProcAddress('glScalexOES')
	gl.glScalexOES.gl = gl
	gl.glScalexOES.__name__ = 'glScalexOES'
	gl.glScalexOES.errcheck = _errcheck
	gl.glScalexOES.restype = GLvoid
	gl.glScalexOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glScissor = _getProcAddress('glScissor')
	gl.glScissor.gl = gl
	gl.glScissor.__name__ = 'glScissor'
	gl.glScissor.errcheck = _errcheck
	gl.glScissor.restype = GLvoid
	gl.glScissor.argtypes = (GLint, GLint, GLsizei, GLsizei,)
	gl.glScissorArrayv = _getProcAddress('glScissorArrayv')
	gl.glScissorArrayv.gl = gl
	gl.glScissorArrayv.__name__ = 'glScissorArrayv'
	gl.glScissorArrayv.errcheck = _errcheck
	gl.glScissorArrayv.restype = GLvoid
	gl.glScissorArrayv.argtypes = (GLuint, GLsizei, POINTER(GLint),)
	gl.glScissorIndexed = _getProcAddress('glScissorIndexed')
	gl.glScissorIndexed.gl = gl
	gl.glScissorIndexed.__name__ = 'glScissorIndexed'
	gl.glScissorIndexed.errcheck = _errcheck
	gl.glScissorIndexed.restype = GLvoid
	gl.glScissorIndexed.argtypes = (GLuint, GLint, GLint, GLsizei, GLsizei,)
	gl.glScissorIndexedv = _getProcAddress('glScissorIndexedv')
	gl.glScissorIndexedv.gl = gl
	gl.glScissorIndexedv.__name__ = 'glScissorIndexedv'
	gl.glScissorIndexedv.errcheck = _errcheck
	gl.glScissorIndexedv.restype = GLvoid
	gl.glScissorIndexedv.argtypes = (GLuint, POINTER(GLint),)
	gl.glSecondaryColor3b = _getProcAddress('glSecondaryColor3b')
	gl.glSecondaryColor3b.gl = gl
	gl.glSecondaryColor3b.__name__ = 'glSecondaryColor3b'
	gl.glSecondaryColor3b.errcheck = _errcheck
	gl.glSecondaryColor3b.restype = GLvoid
	gl.glSecondaryColor3b.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glSecondaryColor3bEXT = _getProcAddress('glSecondaryColor3bEXT')
	gl.glSecondaryColor3bEXT.gl = gl
	gl.glSecondaryColor3bEXT.__name__ = 'glSecondaryColor3bEXT'
	gl.glSecondaryColor3bEXT.errcheck = _errcheck
	gl.glSecondaryColor3bEXT.restype = GLvoid
	gl.glSecondaryColor3bEXT.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glSecondaryColor3bv = _getProcAddress('glSecondaryColor3bv')
	gl.glSecondaryColor3bv.gl = gl
	gl.glSecondaryColor3bv.__name__ = 'glSecondaryColor3bv'
	gl.glSecondaryColor3bv.errcheck = _errcheck
	gl.glSecondaryColor3bv.restype = GLvoid
	gl.glSecondaryColor3bv.argtypes = (POINTER(GLbyte),)
	gl.glSecondaryColor3bvEXT = _getProcAddress('glSecondaryColor3bvEXT')
	gl.glSecondaryColor3bvEXT.gl = gl
	gl.glSecondaryColor3bvEXT.__name__ = 'glSecondaryColor3bvEXT'
	gl.glSecondaryColor3bvEXT.errcheck = _errcheck
	gl.glSecondaryColor3bvEXT.restype = GLvoid
	gl.glSecondaryColor3bvEXT.argtypes = (POINTER(GLbyte),)
	gl.glSecondaryColor3d = _getProcAddress('glSecondaryColor3d')
	gl.glSecondaryColor3d.gl = gl
	gl.glSecondaryColor3d.__name__ = 'glSecondaryColor3d'
	gl.glSecondaryColor3d.errcheck = _errcheck
	gl.glSecondaryColor3d.restype = GLvoid
	gl.glSecondaryColor3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glSecondaryColor3dEXT = _getProcAddress('glSecondaryColor3dEXT')
	gl.glSecondaryColor3dEXT.gl = gl
	gl.glSecondaryColor3dEXT.__name__ = 'glSecondaryColor3dEXT'
	gl.glSecondaryColor3dEXT.errcheck = _errcheck
	gl.glSecondaryColor3dEXT.restype = GLvoid
	gl.glSecondaryColor3dEXT.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glSecondaryColor3dv = _getProcAddress('glSecondaryColor3dv')
	gl.glSecondaryColor3dv.gl = gl
	gl.glSecondaryColor3dv.__name__ = 'glSecondaryColor3dv'
	gl.glSecondaryColor3dv.errcheck = _errcheck
	gl.glSecondaryColor3dv.restype = GLvoid
	gl.glSecondaryColor3dv.argtypes = (POINTER(GLdouble),)
	gl.glSecondaryColor3dvEXT = _getProcAddress('glSecondaryColor3dvEXT')
	gl.glSecondaryColor3dvEXT.gl = gl
	gl.glSecondaryColor3dvEXT.__name__ = 'glSecondaryColor3dvEXT'
	gl.glSecondaryColor3dvEXT.errcheck = _errcheck
	gl.glSecondaryColor3dvEXT.restype = GLvoid
	gl.glSecondaryColor3dvEXT.argtypes = (POINTER(GLdouble),)
	gl.glSecondaryColor3f = _getProcAddress('glSecondaryColor3f')
	gl.glSecondaryColor3f.gl = gl
	gl.glSecondaryColor3f.__name__ = 'glSecondaryColor3f'
	gl.glSecondaryColor3f.errcheck = _errcheck
	gl.glSecondaryColor3f.restype = GLvoid
	gl.glSecondaryColor3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glSecondaryColor3fEXT = _getProcAddress('glSecondaryColor3fEXT')
	gl.glSecondaryColor3fEXT.gl = gl
	gl.glSecondaryColor3fEXT.__name__ = 'glSecondaryColor3fEXT'
	gl.glSecondaryColor3fEXT.errcheck = _errcheck
	gl.glSecondaryColor3fEXT.restype = GLvoid
	gl.glSecondaryColor3fEXT.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glSecondaryColor3fv = _getProcAddress('glSecondaryColor3fv')
	gl.glSecondaryColor3fv.gl = gl
	gl.glSecondaryColor3fv.__name__ = 'glSecondaryColor3fv'
	gl.glSecondaryColor3fv.errcheck = _errcheck
	gl.glSecondaryColor3fv.restype = GLvoid
	gl.glSecondaryColor3fv.argtypes = (POINTER(GLfloat),)
	gl.glSecondaryColor3fvEXT = _getProcAddress('glSecondaryColor3fvEXT')
	gl.glSecondaryColor3fvEXT.gl = gl
	gl.glSecondaryColor3fvEXT.__name__ = 'glSecondaryColor3fvEXT'
	gl.glSecondaryColor3fvEXT.errcheck = _errcheck
	gl.glSecondaryColor3fvEXT.restype = GLvoid
	gl.glSecondaryColor3fvEXT.argtypes = (POINTER(GLfloat),)
	gl.glSecondaryColor3hNV = _getProcAddress('glSecondaryColor3hNV')
	gl.glSecondaryColor3hNV.gl = gl
	gl.glSecondaryColor3hNV.__name__ = 'glSecondaryColor3hNV'
	gl.glSecondaryColor3hNV.errcheck = _errcheck
	gl.glSecondaryColor3hNV.restype = GLvoid
	gl.glSecondaryColor3hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glSecondaryColor3hvNV = _getProcAddress('glSecondaryColor3hvNV')
	gl.glSecondaryColor3hvNV.gl = gl
	gl.glSecondaryColor3hvNV.__name__ = 'glSecondaryColor3hvNV'
	gl.glSecondaryColor3hvNV.errcheck = _errcheck
	gl.glSecondaryColor3hvNV.restype = GLvoid
	gl.glSecondaryColor3hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glSecondaryColor3i = _getProcAddress('glSecondaryColor3i')
	gl.glSecondaryColor3i.gl = gl
	gl.glSecondaryColor3i.__name__ = 'glSecondaryColor3i'
	gl.glSecondaryColor3i.errcheck = _errcheck
	gl.glSecondaryColor3i.restype = GLvoid
	gl.glSecondaryColor3i.argtypes = (GLint, GLint, GLint,)
	gl.glSecondaryColor3iEXT = _getProcAddress('glSecondaryColor3iEXT')
	gl.glSecondaryColor3iEXT.gl = gl
	gl.glSecondaryColor3iEXT.__name__ = 'glSecondaryColor3iEXT'
	gl.glSecondaryColor3iEXT.errcheck = _errcheck
	gl.glSecondaryColor3iEXT.restype = GLvoid
	gl.glSecondaryColor3iEXT.argtypes = (GLint, GLint, GLint,)
	gl.glSecondaryColor3iv = _getProcAddress('glSecondaryColor3iv')
	gl.glSecondaryColor3iv.gl = gl
	gl.glSecondaryColor3iv.__name__ = 'glSecondaryColor3iv'
	gl.glSecondaryColor3iv.errcheck = _errcheck
	gl.glSecondaryColor3iv.restype = GLvoid
	gl.glSecondaryColor3iv.argtypes = (POINTER(GLint),)
	gl.glSecondaryColor3ivEXT = _getProcAddress('glSecondaryColor3ivEXT')
	gl.glSecondaryColor3ivEXT.gl = gl
	gl.glSecondaryColor3ivEXT.__name__ = 'glSecondaryColor3ivEXT'
	gl.glSecondaryColor3ivEXT.errcheck = _errcheck
	gl.glSecondaryColor3ivEXT.restype = GLvoid
	gl.glSecondaryColor3ivEXT.argtypes = (POINTER(GLint),)
	gl.glSecondaryColor3s = _getProcAddress('glSecondaryColor3s')
	gl.glSecondaryColor3s.gl = gl
	gl.glSecondaryColor3s.__name__ = 'glSecondaryColor3s'
	gl.glSecondaryColor3s.errcheck = _errcheck
	gl.glSecondaryColor3s.restype = GLvoid
	gl.glSecondaryColor3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glSecondaryColor3sEXT = _getProcAddress('glSecondaryColor3sEXT')
	gl.glSecondaryColor3sEXT.gl = gl
	gl.glSecondaryColor3sEXT.__name__ = 'glSecondaryColor3sEXT'
	gl.glSecondaryColor3sEXT.errcheck = _errcheck
	gl.glSecondaryColor3sEXT.restype = GLvoid
	gl.glSecondaryColor3sEXT.argtypes = (GLshort, GLshort, GLshort,)
	gl.glSecondaryColor3sv = _getProcAddress('glSecondaryColor3sv')
	gl.glSecondaryColor3sv.gl = gl
	gl.glSecondaryColor3sv.__name__ = 'glSecondaryColor3sv'
	gl.glSecondaryColor3sv.errcheck = _errcheck
	gl.glSecondaryColor3sv.restype = GLvoid
	gl.glSecondaryColor3sv.argtypes = (POINTER(GLshort),)
	gl.glSecondaryColor3svEXT = _getProcAddress('glSecondaryColor3svEXT')
	gl.glSecondaryColor3svEXT.gl = gl
	gl.glSecondaryColor3svEXT.__name__ = 'glSecondaryColor3svEXT'
	gl.glSecondaryColor3svEXT.errcheck = _errcheck
	gl.glSecondaryColor3svEXT.restype = GLvoid
	gl.glSecondaryColor3svEXT.argtypes = (POINTER(GLshort),)
	gl.glSecondaryColor3ub = _getProcAddress('glSecondaryColor3ub')
	gl.glSecondaryColor3ub.gl = gl
	gl.glSecondaryColor3ub.__name__ = 'glSecondaryColor3ub'
	gl.glSecondaryColor3ub.errcheck = _errcheck
	gl.glSecondaryColor3ub.restype = GLvoid
	gl.glSecondaryColor3ub.argtypes = (GLubyte, GLubyte, GLubyte,)
	gl.glSecondaryColor3ubEXT = _getProcAddress('glSecondaryColor3ubEXT')
	gl.glSecondaryColor3ubEXT.gl = gl
	gl.glSecondaryColor3ubEXT.__name__ = 'glSecondaryColor3ubEXT'
	gl.glSecondaryColor3ubEXT.errcheck = _errcheck
	gl.glSecondaryColor3ubEXT.restype = GLvoid
	gl.glSecondaryColor3ubEXT.argtypes = (GLubyte, GLubyte, GLubyte,)
	gl.glSecondaryColor3ubv = _getProcAddress('glSecondaryColor3ubv')
	gl.glSecondaryColor3ubv.gl = gl
	gl.glSecondaryColor3ubv.__name__ = 'glSecondaryColor3ubv'
	gl.glSecondaryColor3ubv.errcheck = _errcheck
	gl.glSecondaryColor3ubv.restype = GLvoid
	gl.glSecondaryColor3ubv.argtypes = (POINTER(GLubyte),)
	gl.glSecondaryColor3ubvEXT = _getProcAddress('glSecondaryColor3ubvEXT')
	gl.glSecondaryColor3ubvEXT.gl = gl
	gl.glSecondaryColor3ubvEXT.__name__ = 'glSecondaryColor3ubvEXT'
	gl.glSecondaryColor3ubvEXT.errcheck = _errcheck
	gl.glSecondaryColor3ubvEXT.restype = GLvoid
	gl.glSecondaryColor3ubvEXT.argtypes = (POINTER(GLubyte),)
	gl.glSecondaryColor3ui = _getProcAddress('glSecondaryColor3ui')
	gl.glSecondaryColor3ui.gl = gl
	gl.glSecondaryColor3ui.__name__ = 'glSecondaryColor3ui'
	gl.glSecondaryColor3ui.errcheck = _errcheck
	gl.glSecondaryColor3ui.restype = GLvoid
	gl.glSecondaryColor3ui.argtypes = (GLuint, GLuint, GLuint,)
	gl.glSecondaryColor3uiEXT = _getProcAddress('glSecondaryColor3uiEXT')
	gl.glSecondaryColor3uiEXT.gl = gl
	gl.glSecondaryColor3uiEXT.__name__ = 'glSecondaryColor3uiEXT'
	gl.glSecondaryColor3uiEXT.errcheck = _errcheck
	gl.glSecondaryColor3uiEXT.restype = GLvoid
	gl.glSecondaryColor3uiEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glSecondaryColor3uiv = _getProcAddress('glSecondaryColor3uiv')
	gl.glSecondaryColor3uiv.gl = gl
	gl.glSecondaryColor3uiv.__name__ = 'glSecondaryColor3uiv'
	gl.glSecondaryColor3uiv.errcheck = _errcheck
	gl.glSecondaryColor3uiv.restype = GLvoid
	gl.glSecondaryColor3uiv.argtypes = (POINTER(GLuint),)
	gl.glSecondaryColor3uivEXT = _getProcAddress('glSecondaryColor3uivEXT')
	gl.glSecondaryColor3uivEXT.gl = gl
	gl.glSecondaryColor3uivEXT.__name__ = 'glSecondaryColor3uivEXT'
	gl.glSecondaryColor3uivEXT.errcheck = _errcheck
	gl.glSecondaryColor3uivEXT.restype = GLvoid
	gl.glSecondaryColor3uivEXT.argtypes = (POINTER(GLuint),)
	gl.glSecondaryColor3us = _getProcAddress('glSecondaryColor3us')
	gl.glSecondaryColor3us.gl = gl
	gl.glSecondaryColor3us.__name__ = 'glSecondaryColor3us'
	gl.glSecondaryColor3us.errcheck = _errcheck
	gl.glSecondaryColor3us.restype = GLvoid
	gl.glSecondaryColor3us.argtypes = (GLushort, GLushort, GLushort,)
	gl.glSecondaryColor3usEXT = _getProcAddress('glSecondaryColor3usEXT')
	gl.glSecondaryColor3usEXT.gl = gl
	gl.glSecondaryColor3usEXT.__name__ = 'glSecondaryColor3usEXT'
	gl.glSecondaryColor3usEXT.errcheck = _errcheck
	gl.glSecondaryColor3usEXT.restype = GLvoid
	gl.glSecondaryColor3usEXT.argtypes = (GLushort, GLushort, GLushort,)
	gl.glSecondaryColor3usv = _getProcAddress('glSecondaryColor3usv')
	gl.glSecondaryColor3usv.gl = gl
	gl.glSecondaryColor3usv.__name__ = 'glSecondaryColor3usv'
	gl.glSecondaryColor3usv.errcheck = _errcheck
	gl.glSecondaryColor3usv.restype = GLvoid
	gl.glSecondaryColor3usv.argtypes = (POINTER(GLushort),)
	gl.glSecondaryColor3usvEXT = _getProcAddress('glSecondaryColor3usvEXT')
	gl.glSecondaryColor3usvEXT.gl = gl
	gl.glSecondaryColor3usvEXT.__name__ = 'glSecondaryColor3usvEXT'
	gl.glSecondaryColor3usvEXT.errcheck = _errcheck
	gl.glSecondaryColor3usvEXT.restype = GLvoid
	gl.glSecondaryColor3usvEXT.argtypes = (POINTER(GLushort),)
	gl.glSecondaryColorFormatNV = _getProcAddress('glSecondaryColorFormatNV')
	gl.glSecondaryColorFormatNV.gl = gl
	gl.glSecondaryColorFormatNV.__name__ = 'glSecondaryColorFormatNV'
	gl.glSecondaryColorFormatNV.errcheck = _errcheck
	gl.glSecondaryColorFormatNV.restype = GLvoid
	gl.glSecondaryColorFormatNV.argtypes = (GLint, GLenum, GLsizei,)
	gl.glSecondaryColorP3ui = _getProcAddress('glSecondaryColorP3ui')
	gl.glSecondaryColorP3ui.gl = gl
	gl.glSecondaryColorP3ui.__name__ = 'glSecondaryColorP3ui'
	gl.glSecondaryColorP3ui.errcheck = _errcheck
	gl.glSecondaryColorP3ui.restype = GLvoid
	gl.glSecondaryColorP3ui.argtypes = (GLenum, GLuint,)
	gl.glSecondaryColorP3uiv = _getProcAddress('glSecondaryColorP3uiv')
	gl.glSecondaryColorP3uiv.gl = gl
	gl.glSecondaryColorP3uiv.__name__ = 'glSecondaryColorP3uiv'
	gl.glSecondaryColorP3uiv.errcheck = _errcheck
	gl.glSecondaryColorP3uiv.restype = GLvoid
	gl.glSecondaryColorP3uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glSecondaryColorPointer = _getProcAddress('glSecondaryColorPointer')
	gl.glSecondaryColorPointer.gl = gl
	gl.glSecondaryColorPointer.__name__ = 'glSecondaryColorPointer'
	gl.glSecondaryColorPointer.errcheck = _errcheck
	gl.glSecondaryColorPointer.restype = GLvoid
	gl.glSecondaryColorPointer.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glSecondaryColorPointerEXT = _getProcAddress('glSecondaryColorPointerEXT')
	gl.glSecondaryColorPointerEXT.gl = gl
	gl.glSecondaryColorPointerEXT.__name__ = 'glSecondaryColorPointerEXT'
	gl.glSecondaryColorPointerEXT.errcheck = _errcheck
	gl.glSecondaryColorPointerEXT.restype = GLvoid
	gl.glSecondaryColorPointerEXT.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glSecondaryColorPointerListIBM = _getProcAddress('glSecondaryColorPointerListIBM')
	gl.glSecondaryColorPointerListIBM.gl = gl
	gl.glSecondaryColorPointerListIBM.__name__ = 'glSecondaryColorPointerListIBM'
	gl.glSecondaryColorPointerListIBM.errcheck = _errcheck
	gl.glSecondaryColorPointerListIBM.restype = GLvoid
	gl.glSecondaryColorPointerListIBM.argtypes = (GLint, GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glSelectBuffer = _getProcAddress('glSelectBuffer')
	gl.glSelectBuffer.gl = gl
	gl.glSelectBuffer.__name__ = 'glSelectBuffer'
	gl.glSelectBuffer.errcheck = _errcheck
	gl.glSelectBuffer.restype = GLvoid
	gl.glSelectBuffer.argtypes = (GLsizei, POINTER(GLuint),)
	gl.glSelectPerfMonitorCountersAMD = _getProcAddress('glSelectPerfMonitorCountersAMD')
	gl.glSelectPerfMonitorCountersAMD.gl = gl
	gl.glSelectPerfMonitorCountersAMD.__name__ = 'glSelectPerfMonitorCountersAMD'
	gl.glSelectPerfMonitorCountersAMD.errcheck = _errcheck
	gl.glSelectPerfMonitorCountersAMD.restype = GLvoid
	gl.glSelectPerfMonitorCountersAMD.argtypes = (GLuint, GLboolean, GLuint, GLint, POINTER(GLuint),)
	gl.glSeparableFilter2D = _getProcAddress('glSeparableFilter2D')
	gl.glSeparableFilter2D.gl = gl
	gl.glSeparableFilter2D.__name__ = 'glSeparableFilter2D'
	gl.glSeparableFilter2D.errcheck = _errcheck
	gl.glSeparableFilter2D.restype = GLvoid
	gl.glSeparableFilter2D.argtypes = (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid), POINTER(GLvoid),)
	gl.glSeparableFilter2DEXT = _getProcAddress('glSeparableFilter2DEXT')
	gl.glSeparableFilter2DEXT.gl = gl
	gl.glSeparableFilter2DEXT.__name__ = 'glSeparableFilter2DEXT'
	gl.glSeparableFilter2DEXT.errcheck = _errcheck
	gl.glSeparableFilter2DEXT.restype = GLvoid
	gl.glSeparableFilter2DEXT.argtypes = (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid), POINTER(GLvoid),)
	gl.glSetFenceAPPLE = _getProcAddress('glSetFenceAPPLE')
	gl.glSetFenceAPPLE.gl = gl
	gl.glSetFenceAPPLE.__name__ = 'glSetFenceAPPLE'
	gl.glSetFenceAPPLE.errcheck = _errcheck
	gl.glSetFenceAPPLE.restype = GLvoid
	gl.glSetFenceAPPLE.argtypes = (GLuint,)
	gl.glSetFenceNV = _getProcAddress('glSetFenceNV')
	gl.glSetFenceNV.gl = gl
	gl.glSetFenceNV.__name__ = 'glSetFenceNV'
	gl.glSetFenceNV.errcheck = _errcheck
	gl.glSetFenceNV.restype = GLvoid
	gl.glSetFenceNV.argtypes = (GLuint, GLenum,)
	gl.glSetFragmentShaderConstantATI = _getProcAddress('glSetFragmentShaderConstantATI')
	gl.glSetFragmentShaderConstantATI.gl = gl
	gl.glSetFragmentShaderConstantATI.__name__ = 'glSetFragmentShaderConstantATI'
	gl.glSetFragmentShaderConstantATI.errcheck = _errcheck
	gl.glSetFragmentShaderConstantATI.restype = GLvoid
	gl.glSetFragmentShaderConstantATI.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glSetInvariantEXT = _getProcAddress('glSetInvariantEXT')
	gl.glSetInvariantEXT.gl = gl
	gl.glSetInvariantEXT.__name__ = 'glSetInvariantEXT'
	gl.glSetInvariantEXT.errcheck = _errcheck
	gl.glSetInvariantEXT.restype = GLvoid
	gl.glSetInvariantEXT.argtypes = (GLuint, GLenum, POINTER(GLvoid),)
	gl.glSetLocalConstantEXT = _getProcAddress('glSetLocalConstantEXT')
	gl.glSetLocalConstantEXT.gl = gl
	gl.glSetLocalConstantEXT.__name__ = 'glSetLocalConstantEXT'
	gl.glSetLocalConstantEXT.errcheck = _errcheck
	gl.glSetLocalConstantEXT.restype = GLvoid
	gl.glSetLocalConstantEXT.argtypes = (GLuint, GLenum, POINTER(GLvoid),)
	gl.glSetMultisamplefvAMD = _getProcAddress('glSetMultisamplefvAMD')
	gl.glSetMultisamplefvAMD.gl = gl
	gl.glSetMultisamplefvAMD.__name__ = 'glSetMultisamplefvAMD'
	gl.glSetMultisamplefvAMD.errcheck = _errcheck
	gl.glSetMultisamplefvAMD.restype = GLvoid
	gl.glSetMultisamplefvAMD.argtypes = (GLenum, GLuint, POINTER(GLfloat),)
	gl.glShadeModel = _getProcAddress('glShadeModel')
	gl.glShadeModel.gl = gl
	gl.glShadeModel.__name__ = 'glShadeModel'
	gl.glShadeModel.errcheck = _errcheck
	gl.glShadeModel.restype = GLvoid
	gl.glShadeModel.argtypes = (GLenum,)
	gl.glShaderBinary = _getProcAddress('glShaderBinary')
	gl.glShaderBinary.gl = gl
	gl.glShaderBinary.__name__ = 'glShaderBinary'
	gl.glShaderBinary.errcheck = _errcheck
	gl.glShaderBinary.restype = GLvoid
	gl.glShaderBinary.argtypes = (GLsizei, POINTER(GLuint), GLenum, POINTER(GLvoid), GLsizei,)
	gl.glShaderOp1EXT = _getProcAddress('glShaderOp1EXT')
	gl.glShaderOp1EXT.gl = gl
	gl.glShaderOp1EXT.__name__ = 'glShaderOp1EXT'
	gl.glShaderOp1EXT.errcheck = _errcheck
	gl.glShaderOp1EXT.restype = GLvoid
	gl.glShaderOp1EXT.argtypes = (GLenum, GLuint, GLuint,)
	gl.glShaderOp2EXT = _getProcAddress('glShaderOp2EXT')
	gl.glShaderOp2EXT.gl = gl
	gl.glShaderOp2EXT.__name__ = 'glShaderOp2EXT'
	gl.glShaderOp2EXT.errcheck = _errcheck
	gl.glShaderOp2EXT.restype = GLvoid
	gl.glShaderOp2EXT.argtypes = (GLenum, GLuint, GLuint, GLuint,)
	gl.glShaderOp3EXT = _getProcAddress('glShaderOp3EXT')
	gl.glShaderOp3EXT.gl = gl
	gl.glShaderOp3EXT.__name__ = 'glShaderOp3EXT'
	gl.glShaderOp3EXT.errcheck = _errcheck
	gl.glShaderOp3EXT.restype = GLvoid
	gl.glShaderOp3EXT.argtypes = (GLenum, GLuint, GLuint, GLuint, GLuint,)
	gl.glShaderSource = _getProcAddress('glShaderSource')
	gl.glShaderSource.gl = gl
	gl.glShaderSource.__name__ = 'glShaderSource'
	gl.glShaderSource.errcheck = _errcheck
	gl.glShaderSource.restype = GLvoid
	gl.glShaderSource.argtypes = (GLuint, GLsizei, POINTER(POINTER(GLchar)), POINTER(GLint),)
	gl.glShaderSourceARB = _getProcAddress('glShaderSourceARB')
	gl.glShaderSourceARB.gl = gl
	gl.glShaderSourceARB.__name__ = 'glShaderSourceARB'
	gl.glShaderSourceARB.errcheck = _errcheck
	gl.glShaderSourceARB.restype = GLvoid
	gl.glShaderSourceARB.argtypes = (GLhandleARB, GLsizei, POINTER(POINTER(GLcharARB)), POINTER(GLint),)
	gl.glShaderStorageBlockBinding = _getProcAddress('glShaderStorageBlockBinding')
	gl.glShaderStorageBlockBinding.gl = gl
	gl.glShaderStorageBlockBinding.__name__ = 'glShaderStorageBlockBinding'
	gl.glShaderStorageBlockBinding.errcheck = _errcheck
	gl.glShaderStorageBlockBinding.restype = GLvoid
	gl.glShaderStorageBlockBinding.argtypes = (GLuint, GLuint, GLuint,)
	gl.glSharpenTexFuncSGIS = _getProcAddress('glSharpenTexFuncSGIS')
	gl.glSharpenTexFuncSGIS.gl = gl
	gl.glSharpenTexFuncSGIS.__name__ = 'glSharpenTexFuncSGIS'
	gl.glSharpenTexFuncSGIS.errcheck = _errcheck
	gl.glSharpenTexFuncSGIS.restype = GLvoid
	gl.glSharpenTexFuncSGIS.argtypes = (GLenum, GLsizei, POINTER(GLfloat),)
	gl.glSpriteParameterfSGIX = _getProcAddress('glSpriteParameterfSGIX')
	gl.glSpriteParameterfSGIX.gl = gl
	gl.glSpriteParameterfSGIX.__name__ = 'glSpriteParameterfSGIX'
	gl.glSpriteParameterfSGIX.errcheck = _errcheck
	gl.glSpriteParameterfSGIX.restype = GLvoid
	gl.glSpriteParameterfSGIX.argtypes = (GLenum, GLfloat,)
	gl.glSpriteParameterfvSGIX = _getProcAddress('glSpriteParameterfvSGIX')
	gl.glSpriteParameterfvSGIX.gl = gl
	gl.glSpriteParameterfvSGIX.__name__ = 'glSpriteParameterfvSGIX'
	gl.glSpriteParameterfvSGIX.errcheck = _errcheck
	gl.glSpriteParameterfvSGIX.restype = GLvoid
	gl.glSpriteParameterfvSGIX.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glSpriteParameteriSGIX = _getProcAddress('glSpriteParameteriSGIX')
	gl.glSpriteParameteriSGIX.gl = gl
	gl.glSpriteParameteriSGIX.__name__ = 'glSpriteParameteriSGIX'
	gl.glSpriteParameteriSGIX.errcheck = _errcheck
	gl.glSpriteParameteriSGIX.restype = GLvoid
	gl.glSpriteParameteriSGIX.argtypes = (GLenum, GLint,)
	gl.glSpriteParameterivSGIX = _getProcAddress('glSpriteParameterivSGIX')
	gl.glSpriteParameterivSGIX.gl = gl
	gl.glSpriteParameterivSGIX.__name__ = 'glSpriteParameterivSGIX'
	gl.glSpriteParameterivSGIX.errcheck = _errcheck
	gl.glSpriteParameterivSGIX.restype = GLvoid
	gl.glSpriteParameterivSGIX.argtypes = (GLenum, POINTER(GLint),)
	gl.glStartInstrumentsSGIX = _getProcAddress('glStartInstrumentsSGIX')
	gl.glStartInstrumentsSGIX.gl = gl
	gl.glStartInstrumentsSGIX.__name__ = 'glStartInstrumentsSGIX'
	gl.glStartInstrumentsSGIX.errcheck = _errcheck
	gl.glStartInstrumentsSGIX.restype = GLvoid
	gl.glStartInstrumentsSGIX.argtypes = ()
	gl.glStartTilingQCOM = _getProcAddress('glStartTilingQCOM')
	gl.glStartTilingQCOM.gl = gl
	gl.glStartTilingQCOM.__name__ = 'glStartTilingQCOM'
	gl.glStartTilingQCOM.errcheck = _errcheck
	gl.glStartTilingQCOM.restype = GLvoid
	gl.glStartTilingQCOM.argtypes = (GLuint, GLuint, GLuint, GLuint, GLbitfield,)
	gl.glStencilClearTagEXT = _getProcAddress('glStencilClearTagEXT')
	gl.glStencilClearTagEXT.gl = gl
	gl.glStencilClearTagEXT.__name__ = 'glStencilClearTagEXT'
	gl.glStencilClearTagEXT.errcheck = _errcheck
	gl.glStencilClearTagEXT.restype = GLvoid
	gl.glStencilClearTagEXT.argtypes = (GLsizei, GLuint,)
	gl.glStencilFillPathInstancedNV = _getProcAddress('glStencilFillPathInstancedNV')
	gl.glStencilFillPathInstancedNV.gl = gl
	gl.glStencilFillPathInstancedNV.__name__ = 'glStencilFillPathInstancedNV'
	gl.glStencilFillPathInstancedNV.errcheck = _errcheck
	gl.glStencilFillPathInstancedNV.restype = GLvoid
	gl.glStencilFillPathInstancedNV.argtypes = (GLsizei, GLenum, POINTER(GLvoid), GLuint, GLenum, GLuint, GLenum, POINTER(GLfloat),)
	gl.glStencilFillPathNV = _getProcAddress('glStencilFillPathNV')
	gl.glStencilFillPathNV.gl = gl
	gl.glStencilFillPathNV.__name__ = 'glStencilFillPathNV'
	gl.glStencilFillPathNV.errcheck = _errcheck
	gl.glStencilFillPathNV.restype = GLvoid
	gl.glStencilFillPathNV.argtypes = (GLuint, GLenum, GLuint,)
	gl.glStencilFunc = _getProcAddress('glStencilFunc')
	gl.glStencilFunc.gl = gl
	gl.glStencilFunc.__name__ = 'glStencilFunc'
	gl.glStencilFunc.errcheck = _errcheck
	gl.glStencilFunc.restype = GLvoid
	gl.glStencilFunc.argtypes = (GLenum, GLint, GLuint,)
	gl.glStencilFuncSeparate = _getProcAddress('glStencilFuncSeparate')
	gl.glStencilFuncSeparate.gl = gl
	gl.glStencilFuncSeparate.__name__ = 'glStencilFuncSeparate'
	gl.glStencilFuncSeparate.errcheck = _errcheck
	gl.glStencilFuncSeparate.restype = GLvoid
	gl.glStencilFuncSeparate.argtypes = (GLenum, GLenum, GLint, GLuint,)
	gl.glStencilFuncSeparateATI = _getProcAddress('glStencilFuncSeparateATI')
	gl.glStencilFuncSeparateATI.gl = gl
	gl.glStencilFuncSeparateATI.__name__ = 'glStencilFuncSeparateATI'
	gl.glStencilFuncSeparateATI.errcheck = _errcheck
	gl.glStencilFuncSeparateATI.restype = GLvoid
	gl.glStencilFuncSeparateATI.argtypes = (GLenum, GLenum, GLint, GLuint,)
	gl.glStencilMask = _getProcAddress('glStencilMask')
	gl.glStencilMask.gl = gl
	gl.glStencilMask.__name__ = 'glStencilMask'
	gl.glStencilMask.errcheck = _errcheck
	gl.glStencilMask.restype = GLvoid
	gl.glStencilMask.argtypes = (GLuint,)
	gl.glStencilMaskSeparate = _getProcAddress('glStencilMaskSeparate')
	gl.glStencilMaskSeparate.gl = gl
	gl.glStencilMaskSeparate.__name__ = 'glStencilMaskSeparate'
	gl.glStencilMaskSeparate.errcheck = _errcheck
	gl.glStencilMaskSeparate.restype = GLvoid
	gl.glStencilMaskSeparate.argtypes = (GLenum, GLuint,)
	gl.glStencilOp = _getProcAddress('glStencilOp')
	gl.glStencilOp.gl = gl
	gl.glStencilOp.__name__ = 'glStencilOp'
	gl.glStencilOp.errcheck = _errcheck
	gl.glStencilOp.restype = GLvoid
	gl.glStencilOp.argtypes = (GLenum, GLenum, GLenum,)
	gl.glStencilOpSeparate = _getProcAddress('glStencilOpSeparate')
	gl.glStencilOpSeparate.gl = gl
	gl.glStencilOpSeparate.__name__ = 'glStencilOpSeparate'
	gl.glStencilOpSeparate.errcheck = _errcheck
	gl.glStencilOpSeparate.restype = GLvoid
	gl.glStencilOpSeparate.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glStencilOpSeparateATI = _getProcAddress('glStencilOpSeparateATI')
	gl.glStencilOpSeparateATI.gl = gl
	gl.glStencilOpSeparateATI.__name__ = 'glStencilOpSeparateATI'
	gl.glStencilOpSeparateATI.errcheck = _errcheck
	gl.glStencilOpSeparateATI.restype = GLvoid
	gl.glStencilOpSeparateATI.argtypes = (GLenum, GLenum, GLenum, GLenum,)
	gl.glStencilOpValueAMD = _getProcAddress('glStencilOpValueAMD')
	gl.glStencilOpValueAMD.gl = gl
	gl.glStencilOpValueAMD.__name__ = 'glStencilOpValueAMD'
	gl.glStencilOpValueAMD.errcheck = _errcheck
	gl.glStencilOpValueAMD.restype = GLvoid
	gl.glStencilOpValueAMD.argtypes = (GLenum, GLuint,)
	gl.glStencilStrokePathInstancedNV = _getProcAddress('glStencilStrokePathInstancedNV')
	gl.glStencilStrokePathInstancedNV.gl = gl
	gl.glStencilStrokePathInstancedNV.__name__ = 'glStencilStrokePathInstancedNV'
	gl.glStencilStrokePathInstancedNV.errcheck = _errcheck
	gl.glStencilStrokePathInstancedNV.restype = GLvoid
	gl.glStencilStrokePathInstancedNV.argtypes = (GLsizei, GLenum, POINTER(GLvoid), GLuint, GLint, GLuint, GLenum, POINTER(GLfloat),)
	gl.glStencilStrokePathNV = _getProcAddress('glStencilStrokePathNV')
	gl.glStencilStrokePathNV.gl = gl
	gl.glStencilStrokePathNV.__name__ = 'glStencilStrokePathNV'
	gl.glStencilStrokePathNV.errcheck = _errcheck
	gl.glStencilStrokePathNV.restype = GLvoid
	gl.glStencilStrokePathNV.argtypes = (GLuint, GLint, GLuint,)
	gl.glStencilThenCoverFillPathInstancedNV = _getProcAddress('glStencilThenCoverFillPathInstancedNV')
	gl.glStencilThenCoverFillPathInstancedNV.gl = gl
	gl.glStencilThenCoverFillPathInstancedNV.__name__ = 'glStencilThenCoverFillPathInstancedNV'
	gl.glStencilThenCoverFillPathInstancedNV.errcheck = _errcheck
	gl.glStencilThenCoverFillPathInstancedNV.restype = GLvoid
	gl.glStencilThenCoverFillPathInstancedNV.argtypes = (GLsizei, GLenum, POINTER(GLvoid), GLuint, GLenum, GLuint, GLenum, GLenum, POINTER(GLfloat),)
	gl.glStencilThenCoverFillPathNV = _getProcAddress('glStencilThenCoverFillPathNV')
	gl.glStencilThenCoverFillPathNV.gl = gl
	gl.glStencilThenCoverFillPathNV.__name__ = 'glStencilThenCoverFillPathNV'
	gl.glStencilThenCoverFillPathNV.errcheck = _errcheck
	gl.glStencilThenCoverFillPathNV.restype = GLvoid
	gl.glStencilThenCoverFillPathNV.argtypes = (GLuint, GLenum, GLuint, GLenum,)
	gl.glStencilThenCoverStrokePathInstancedNV = _getProcAddress('glStencilThenCoverStrokePathInstancedNV')
	gl.glStencilThenCoverStrokePathInstancedNV.gl = gl
	gl.glStencilThenCoverStrokePathInstancedNV.__name__ = 'glStencilThenCoverStrokePathInstancedNV'
	gl.glStencilThenCoverStrokePathInstancedNV.errcheck = _errcheck
	gl.glStencilThenCoverStrokePathInstancedNV.restype = GLvoid
	gl.glStencilThenCoverStrokePathInstancedNV.argtypes = (GLsizei, GLenum, POINTER(GLvoid), GLuint, GLint, GLuint, GLenum, GLenum, POINTER(GLfloat),)
	gl.glStencilThenCoverStrokePathNV = _getProcAddress('glStencilThenCoverStrokePathNV')
	gl.glStencilThenCoverStrokePathNV.gl = gl
	gl.glStencilThenCoverStrokePathNV.__name__ = 'glStencilThenCoverStrokePathNV'
	gl.glStencilThenCoverStrokePathNV.errcheck = _errcheck
	gl.glStencilThenCoverStrokePathNV.restype = GLvoid
	gl.glStencilThenCoverStrokePathNV.argtypes = (GLuint, GLint, GLuint, GLenum,)
	gl.glStopInstrumentsSGIX = _getProcAddress('glStopInstrumentsSGIX')
	gl.glStopInstrumentsSGIX.gl = gl
	gl.glStopInstrumentsSGIX.__name__ = 'glStopInstrumentsSGIX'
	gl.glStopInstrumentsSGIX.errcheck = _errcheck
	gl.glStopInstrumentsSGIX.restype = GLvoid
	gl.glStopInstrumentsSGIX.argtypes = (GLint,)
	gl.glStringMarkerGREMEDY = _getProcAddress('glStringMarkerGREMEDY')
	gl.glStringMarkerGREMEDY.gl = gl
	gl.glStringMarkerGREMEDY.__name__ = 'glStringMarkerGREMEDY'
	gl.glStringMarkerGREMEDY.errcheck = _errcheck
	gl.glStringMarkerGREMEDY.restype = GLvoid
	gl.glStringMarkerGREMEDY.argtypes = (GLsizei, POINTER(GLvoid),)
	gl.glSwizzleEXT = _getProcAddress('glSwizzleEXT')
	gl.glSwizzleEXT.gl = gl
	gl.glSwizzleEXT.__name__ = 'glSwizzleEXT'
	gl.glSwizzleEXT.errcheck = _errcheck
	gl.glSwizzleEXT.restype = GLvoid
	gl.glSwizzleEXT.argtypes = (GLuint, GLuint, GLenum, GLenum, GLenum, GLenum,)
	gl.glSyncTextureINTEL = _getProcAddress('glSyncTextureINTEL')
	gl.glSyncTextureINTEL.gl = gl
	gl.glSyncTextureINTEL.__name__ = 'glSyncTextureINTEL'
	gl.glSyncTextureINTEL.errcheck = _errcheck
	gl.glSyncTextureINTEL.restype = GLvoid
	gl.glSyncTextureINTEL.argtypes = (GLuint,)
	gl.glTagSampleBufferSGIX = _getProcAddress('glTagSampleBufferSGIX')
	gl.glTagSampleBufferSGIX.gl = gl
	gl.glTagSampleBufferSGIX.__name__ = 'glTagSampleBufferSGIX'
	gl.glTagSampleBufferSGIX.errcheck = _errcheck
	gl.glTagSampleBufferSGIX.restype = GLvoid
	gl.glTagSampleBufferSGIX.argtypes = ()
	gl.glTangent3bEXT = _getProcAddress('glTangent3bEXT')
	gl.glTangent3bEXT.gl = gl
	gl.glTangent3bEXT.__name__ = 'glTangent3bEXT'
	gl.glTangent3bEXT.errcheck = _errcheck
	gl.glTangent3bEXT.restype = GLvoid
	gl.glTangent3bEXT.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glTangent3bvEXT = _getProcAddress('glTangent3bvEXT')
	gl.glTangent3bvEXT.gl = gl
	gl.glTangent3bvEXT.__name__ = 'glTangent3bvEXT'
	gl.glTangent3bvEXT.errcheck = _errcheck
	gl.glTangent3bvEXT.restype = GLvoid
	gl.glTangent3bvEXT.argtypes = (POINTER(GLbyte),)
	gl.glTangent3dEXT = _getProcAddress('glTangent3dEXT')
	gl.glTangent3dEXT.gl = gl
	gl.glTangent3dEXT.__name__ = 'glTangent3dEXT'
	gl.glTangent3dEXT.errcheck = _errcheck
	gl.glTangent3dEXT.restype = GLvoid
	gl.glTangent3dEXT.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glTangent3dvEXT = _getProcAddress('glTangent3dvEXT')
	gl.glTangent3dvEXT.gl = gl
	gl.glTangent3dvEXT.__name__ = 'glTangent3dvEXT'
	gl.glTangent3dvEXT.errcheck = _errcheck
	gl.glTangent3dvEXT.restype = GLvoid
	gl.glTangent3dvEXT.argtypes = (POINTER(GLdouble),)
	gl.glTangent3fEXT = _getProcAddress('glTangent3fEXT')
	gl.glTangent3fEXT.gl = gl
	gl.glTangent3fEXT.__name__ = 'glTangent3fEXT'
	gl.glTangent3fEXT.errcheck = _errcheck
	gl.glTangent3fEXT.restype = GLvoid
	gl.glTangent3fEXT.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glTangent3fvEXT = _getProcAddress('glTangent3fvEXT')
	gl.glTangent3fvEXT.gl = gl
	gl.glTangent3fvEXT.__name__ = 'glTangent3fvEXT'
	gl.glTangent3fvEXT.errcheck = _errcheck
	gl.glTangent3fvEXT.restype = GLvoid
	gl.glTangent3fvEXT.argtypes = (POINTER(GLfloat),)
	gl.glTangent3iEXT = _getProcAddress('glTangent3iEXT')
	gl.glTangent3iEXT.gl = gl
	gl.glTangent3iEXT.__name__ = 'glTangent3iEXT'
	gl.glTangent3iEXT.errcheck = _errcheck
	gl.glTangent3iEXT.restype = GLvoid
	gl.glTangent3iEXT.argtypes = (GLint, GLint, GLint,)
	gl.glTangent3ivEXT = _getProcAddress('glTangent3ivEXT')
	gl.glTangent3ivEXT.gl = gl
	gl.glTangent3ivEXT.__name__ = 'glTangent3ivEXT'
	gl.glTangent3ivEXT.errcheck = _errcheck
	gl.glTangent3ivEXT.restype = GLvoid
	gl.glTangent3ivEXT.argtypes = (POINTER(GLint),)
	gl.glTangent3sEXT = _getProcAddress('glTangent3sEXT')
	gl.glTangent3sEXT.gl = gl
	gl.glTangent3sEXT.__name__ = 'glTangent3sEXT'
	gl.glTangent3sEXT.errcheck = _errcheck
	gl.glTangent3sEXT.restype = GLvoid
	gl.glTangent3sEXT.argtypes = (GLshort, GLshort, GLshort,)
	gl.glTangent3svEXT = _getProcAddress('glTangent3svEXT')
	gl.glTangent3svEXT.gl = gl
	gl.glTangent3svEXT.__name__ = 'glTangent3svEXT'
	gl.glTangent3svEXT.errcheck = _errcheck
	gl.glTangent3svEXT.restype = GLvoid
	gl.glTangent3svEXT.argtypes = (POINTER(GLshort),)
	gl.glTangentPointerEXT = _getProcAddress('glTangentPointerEXT')
	gl.glTangentPointerEXT.gl = gl
	gl.glTangentPointerEXT.__name__ = 'glTangentPointerEXT'
	gl.glTangentPointerEXT.errcheck = _errcheck
	gl.glTangentPointerEXT.restype = GLvoid
	gl.glTangentPointerEXT.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glTbufferMask3DFX = _getProcAddress('glTbufferMask3DFX')
	gl.glTbufferMask3DFX.gl = gl
	gl.glTbufferMask3DFX.__name__ = 'glTbufferMask3DFX'
	gl.glTbufferMask3DFX.errcheck = _errcheck
	gl.glTbufferMask3DFX.restype = GLvoid
	gl.glTbufferMask3DFX.argtypes = (GLuint,)
	gl.glTessellationFactorAMD = _getProcAddress('glTessellationFactorAMD')
	gl.glTessellationFactorAMD.gl = gl
	gl.glTessellationFactorAMD.__name__ = 'glTessellationFactorAMD'
	gl.glTessellationFactorAMD.errcheck = _errcheck
	gl.glTessellationFactorAMD.restype = GLvoid
	gl.glTessellationFactorAMD.argtypes = (GLfloat,)
	gl.glTessellationModeAMD = _getProcAddress('glTessellationModeAMD')
	gl.glTessellationModeAMD.gl = gl
	gl.glTessellationModeAMD.__name__ = 'glTessellationModeAMD'
	gl.glTessellationModeAMD.errcheck = _errcheck
	gl.glTessellationModeAMD.restype = GLvoid
	gl.glTessellationModeAMD.argtypes = (GLenum,)
	gl.glTestFenceAPPLE = _getProcAddress('glTestFenceAPPLE')
	gl.glTestFenceAPPLE.gl = gl
	gl.glTestFenceAPPLE.__name__ = 'glTestFenceAPPLE'
	gl.glTestFenceAPPLE.errcheck = _errcheck
	gl.glTestFenceAPPLE.restype = GLboolean
	gl.glTestFenceAPPLE.argtypes = (GLuint,)
	gl.glTestFenceNV = _getProcAddress('glTestFenceNV')
	gl.glTestFenceNV.gl = gl
	gl.glTestFenceNV.__name__ = 'glTestFenceNV'
	gl.glTestFenceNV.errcheck = _errcheck
	gl.glTestFenceNV.restype = GLboolean
	gl.glTestFenceNV.argtypes = (GLuint,)
	gl.glTestObjectAPPLE = _getProcAddress('glTestObjectAPPLE')
	gl.glTestObjectAPPLE.gl = gl
	gl.glTestObjectAPPLE.__name__ = 'glTestObjectAPPLE'
	gl.glTestObjectAPPLE.errcheck = _errcheck
	gl.glTestObjectAPPLE.restype = GLboolean
	gl.glTestObjectAPPLE.argtypes = (GLenum, GLuint,)
	gl.glTexBuffer = _getProcAddress('glTexBuffer')
	gl.glTexBuffer.gl = gl
	gl.glTexBuffer.__name__ = 'glTexBuffer'
	gl.glTexBuffer.errcheck = _errcheck
	gl.glTexBuffer.restype = GLvoid
	gl.glTexBuffer.argtypes = (GLenum, GLenum, GLuint,)
	gl.glTexBufferARB = _getProcAddress('glTexBufferARB')
	gl.glTexBufferARB.gl = gl
	gl.glTexBufferARB.__name__ = 'glTexBufferARB'
	gl.glTexBufferARB.errcheck = _errcheck
	gl.glTexBufferARB.restype = GLvoid
	gl.glTexBufferARB.argtypes = (GLenum, GLenum, GLuint,)
	gl.glTexBufferEXT = _getProcAddress('glTexBufferEXT')
	gl.glTexBufferEXT.gl = gl
	gl.glTexBufferEXT.__name__ = 'glTexBufferEXT'
	gl.glTexBufferEXT.errcheck = _errcheck
	gl.glTexBufferEXT.restype = GLvoid
	gl.glTexBufferEXT.argtypes = (GLenum, GLenum, GLuint,)
	gl.glTexBufferRange = _getProcAddress('glTexBufferRange')
	gl.glTexBufferRange.gl = gl
	gl.glTexBufferRange.__name__ = 'glTexBufferRange'
	gl.glTexBufferRange.errcheck = _errcheck
	gl.glTexBufferRange.restype = GLvoid
	gl.glTexBufferRange.argtypes = (GLenum, GLenum, GLuint, GLintptr, GLsizeiptr,)
	gl.glTexBufferRangeEXT = _getProcAddress('glTexBufferRangeEXT')
	gl.glTexBufferRangeEXT.gl = gl
	gl.glTexBufferRangeEXT.__name__ = 'glTexBufferRangeEXT'
	gl.glTexBufferRangeEXT.errcheck = _errcheck
	gl.glTexBufferRangeEXT.restype = GLvoid
	gl.glTexBufferRangeEXT.argtypes = (GLenum, GLenum, GLuint, GLintptr, GLsizeiptr,)
	gl.glTexBumpParameterfvATI = _getProcAddress('glTexBumpParameterfvATI')
	gl.glTexBumpParameterfvATI.gl = gl
	gl.glTexBumpParameterfvATI.__name__ = 'glTexBumpParameterfvATI'
	gl.glTexBumpParameterfvATI.errcheck = _errcheck
	gl.glTexBumpParameterfvATI.restype = GLvoid
	gl.glTexBumpParameterfvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glTexBumpParameterivATI = _getProcAddress('glTexBumpParameterivATI')
	gl.glTexBumpParameterivATI.gl = gl
	gl.glTexBumpParameterivATI.__name__ = 'glTexBumpParameterivATI'
	gl.glTexBumpParameterivATI.errcheck = _errcheck
	gl.glTexBumpParameterivATI.restype = GLvoid
	gl.glTexBumpParameterivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glTexCoord1bOES = _getProcAddress('glTexCoord1bOES')
	gl.glTexCoord1bOES.gl = gl
	gl.glTexCoord1bOES.__name__ = 'glTexCoord1bOES'
	gl.glTexCoord1bOES.errcheck = _errcheck
	gl.glTexCoord1bOES.restype = GLvoid
	gl.glTexCoord1bOES.argtypes = (GLbyte,)
	gl.glTexCoord1bvOES = _getProcAddress('glTexCoord1bvOES')
	gl.glTexCoord1bvOES.gl = gl
	gl.glTexCoord1bvOES.__name__ = 'glTexCoord1bvOES'
	gl.glTexCoord1bvOES.errcheck = _errcheck
	gl.glTexCoord1bvOES.restype = GLvoid
	gl.glTexCoord1bvOES.argtypes = (POINTER(GLbyte),)
	gl.glTexCoord1d = _getProcAddress('glTexCoord1d')
	gl.glTexCoord1d.gl = gl
	gl.glTexCoord1d.__name__ = 'glTexCoord1d'
	gl.glTexCoord1d.errcheck = _errcheck
	gl.glTexCoord1d.restype = GLvoid
	gl.glTexCoord1d.argtypes = (GLdouble,)
	gl.glTexCoord1dv = _getProcAddress('glTexCoord1dv')
	gl.glTexCoord1dv.gl = gl
	gl.glTexCoord1dv.__name__ = 'glTexCoord1dv'
	gl.glTexCoord1dv.errcheck = _errcheck
	gl.glTexCoord1dv.restype = GLvoid
	gl.glTexCoord1dv.argtypes = (POINTER(GLdouble),)
	gl.glTexCoord1f = _getProcAddress('glTexCoord1f')
	gl.glTexCoord1f.gl = gl
	gl.glTexCoord1f.__name__ = 'glTexCoord1f'
	gl.glTexCoord1f.errcheck = _errcheck
	gl.glTexCoord1f.restype = GLvoid
	gl.glTexCoord1f.argtypes = (GLfloat,)
	gl.glTexCoord1fv = _getProcAddress('glTexCoord1fv')
	gl.glTexCoord1fv.gl = gl
	gl.glTexCoord1fv.__name__ = 'glTexCoord1fv'
	gl.glTexCoord1fv.errcheck = _errcheck
	gl.glTexCoord1fv.restype = GLvoid
	gl.glTexCoord1fv.argtypes = (POINTER(GLfloat),)
	gl.glTexCoord1hNV = _getProcAddress('glTexCoord1hNV')
	gl.glTexCoord1hNV.gl = gl
	gl.glTexCoord1hNV.__name__ = 'glTexCoord1hNV'
	gl.glTexCoord1hNV.errcheck = _errcheck
	gl.glTexCoord1hNV.restype = GLvoid
	gl.glTexCoord1hNV.argtypes = (GLhalfNV,)
	gl.glTexCoord1hvNV = _getProcAddress('glTexCoord1hvNV')
	gl.glTexCoord1hvNV.gl = gl
	gl.glTexCoord1hvNV.__name__ = 'glTexCoord1hvNV'
	gl.glTexCoord1hvNV.errcheck = _errcheck
	gl.glTexCoord1hvNV.restype = GLvoid
	gl.glTexCoord1hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glTexCoord1i = _getProcAddress('glTexCoord1i')
	gl.glTexCoord1i.gl = gl
	gl.glTexCoord1i.__name__ = 'glTexCoord1i'
	gl.glTexCoord1i.errcheck = _errcheck
	gl.glTexCoord1i.restype = GLvoid
	gl.glTexCoord1i.argtypes = (GLint,)
	gl.glTexCoord1iv = _getProcAddress('glTexCoord1iv')
	gl.glTexCoord1iv.gl = gl
	gl.glTexCoord1iv.__name__ = 'glTexCoord1iv'
	gl.glTexCoord1iv.errcheck = _errcheck
	gl.glTexCoord1iv.restype = GLvoid
	gl.glTexCoord1iv.argtypes = (POINTER(GLint),)
	gl.glTexCoord1s = _getProcAddress('glTexCoord1s')
	gl.glTexCoord1s.gl = gl
	gl.glTexCoord1s.__name__ = 'glTexCoord1s'
	gl.glTexCoord1s.errcheck = _errcheck
	gl.glTexCoord1s.restype = GLvoid
	gl.glTexCoord1s.argtypes = (GLshort,)
	gl.glTexCoord1sv = _getProcAddress('glTexCoord1sv')
	gl.glTexCoord1sv.gl = gl
	gl.glTexCoord1sv.__name__ = 'glTexCoord1sv'
	gl.glTexCoord1sv.errcheck = _errcheck
	gl.glTexCoord1sv.restype = GLvoid
	gl.glTexCoord1sv.argtypes = (POINTER(GLshort),)
	gl.glTexCoord1xOES = _getProcAddress('glTexCoord1xOES')
	gl.glTexCoord1xOES.gl = gl
	gl.glTexCoord1xOES.__name__ = 'glTexCoord1xOES'
	gl.glTexCoord1xOES.errcheck = _errcheck
	gl.glTexCoord1xOES.restype = GLvoid
	gl.glTexCoord1xOES.argtypes = (GLfixed,)
	gl.glTexCoord1xvOES = _getProcAddress('glTexCoord1xvOES')
	gl.glTexCoord1xvOES.gl = gl
	gl.glTexCoord1xvOES.__name__ = 'glTexCoord1xvOES'
	gl.glTexCoord1xvOES.errcheck = _errcheck
	gl.glTexCoord1xvOES.restype = GLvoid
	gl.glTexCoord1xvOES.argtypes = (POINTER(GLfixed),)
	gl.glTexCoord2bOES = _getProcAddress('glTexCoord2bOES')
	gl.glTexCoord2bOES.gl = gl
	gl.glTexCoord2bOES.__name__ = 'glTexCoord2bOES'
	gl.glTexCoord2bOES.errcheck = _errcheck
	gl.glTexCoord2bOES.restype = GLvoid
	gl.glTexCoord2bOES.argtypes = (GLbyte, GLbyte,)
	gl.glTexCoord2bvOES = _getProcAddress('glTexCoord2bvOES')
	gl.glTexCoord2bvOES.gl = gl
	gl.glTexCoord2bvOES.__name__ = 'glTexCoord2bvOES'
	gl.glTexCoord2bvOES.errcheck = _errcheck
	gl.glTexCoord2bvOES.restype = GLvoid
	gl.glTexCoord2bvOES.argtypes = (POINTER(GLbyte),)
	gl.glTexCoord2d = _getProcAddress('glTexCoord2d')
	gl.glTexCoord2d.gl = gl
	gl.glTexCoord2d.__name__ = 'glTexCoord2d'
	gl.glTexCoord2d.errcheck = _errcheck
	gl.glTexCoord2d.restype = GLvoid
	gl.glTexCoord2d.argtypes = (GLdouble, GLdouble,)
	gl.glTexCoord2dv = _getProcAddress('glTexCoord2dv')
	gl.glTexCoord2dv.gl = gl
	gl.glTexCoord2dv.__name__ = 'glTexCoord2dv'
	gl.glTexCoord2dv.errcheck = _errcheck
	gl.glTexCoord2dv.restype = GLvoid
	gl.glTexCoord2dv.argtypes = (POINTER(GLdouble),)
	gl.glTexCoord2f = _getProcAddress('glTexCoord2f')
	gl.glTexCoord2f.gl = gl
	gl.glTexCoord2f.__name__ = 'glTexCoord2f'
	gl.glTexCoord2f.errcheck = _errcheck
	gl.glTexCoord2f.restype = GLvoid
	gl.glTexCoord2f.argtypes = (GLfloat, GLfloat,)
	gl.glTexCoord2fColor3fVertex3fSUN = _getProcAddress('glTexCoord2fColor3fVertex3fSUN')
	gl.glTexCoord2fColor3fVertex3fSUN.gl = gl
	gl.glTexCoord2fColor3fVertex3fSUN.__name__ = 'glTexCoord2fColor3fVertex3fSUN'
	gl.glTexCoord2fColor3fVertex3fSUN.errcheck = _errcheck
	gl.glTexCoord2fColor3fVertex3fSUN.restype = GLvoid
	gl.glTexCoord2fColor3fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord2fColor3fVertex3fvSUN = _getProcAddress('glTexCoord2fColor3fVertex3fvSUN')
	gl.glTexCoord2fColor3fVertex3fvSUN.gl = gl
	gl.glTexCoord2fColor3fVertex3fvSUN.__name__ = 'glTexCoord2fColor3fVertex3fvSUN'
	gl.glTexCoord2fColor3fVertex3fvSUN.errcheck = _errcheck
	gl.glTexCoord2fColor3fVertex3fvSUN.restype = GLvoid
	gl.glTexCoord2fColor3fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glTexCoord2fColor4fNormal3fVertex3fSUN = _getProcAddress('glTexCoord2fColor4fNormal3fVertex3fSUN')
	gl.glTexCoord2fColor4fNormal3fVertex3fSUN.gl = gl
	gl.glTexCoord2fColor4fNormal3fVertex3fSUN.__name__ = 'glTexCoord2fColor4fNormal3fVertex3fSUN'
	gl.glTexCoord2fColor4fNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glTexCoord2fColor4fNormal3fVertex3fSUN.restype = GLvoid
	gl.glTexCoord2fColor4fNormal3fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord2fColor4fNormal3fVertex3fvSUN = _getProcAddress('glTexCoord2fColor4fNormal3fVertex3fvSUN')
	gl.glTexCoord2fColor4fNormal3fVertex3fvSUN.gl = gl
	gl.glTexCoord2fColor4fNormal3fVertex3fvSUN.__name__ = 'glTexCoord2fColor4fNormal3fVertex3fvSUN'
	gl.glTexCoord2fColor4fNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glTexCoord2fColor4fNormal3fVertex3fvSUN.restype = GLvoid
	gl.glTexCoord2fColor4fNormal3fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glTexCoord2fColor4ubVertex3fSUN = _getProcAddress('glTexCoord2fColor4ubVertex3fSUN')
	gl.glTexCoord2fColor4ubVertex3fSUN.gl = gl
	gl.glTexCoord2fColor4ubVertex3fSUN.__name__ = 'glTexCoord2fColor4ubVertex3fSUN'
	gl.glTexCoord2fColor4ubVertex3fSUN.errcheck = _errcheck
	gl.glTexCoord2fColor4ubVertex3fSUN.restype = GLvoid
	gl.glTexCoord2fColor4ubVertex3fSUN.argtypes = (GLfloat, GLfloat, GLubyte, GLubyte, GLubyte, GLubyte, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord2fColor4ubVertex3fvSUN = _getProcAddress('glTexCoord2fColor4ubVertex3fvSUN')
	gl.glTexCoord2fColor4ubVertex3fvSUN.gl = gl
	gl.glTexCoord2fColor4ubVertex3fvSUN.__name__ = 'glTexCoord2fColor4ubVertex3fvSUN'
	gl.glTexCoord2fColor4ubVertex3fvSUN.errcheck = _errcheck
	gl.glTexCoord2fColor4ubVertex3fvSUN.restype = GLvoid
	gl.glTexCoord2fColor4ubVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLubyte), POINTER(GLfloat),)
	gl.glTexCoord2fNormal3fVertex3fSUN = _getProcAddress('glTexCoord2fNormal3fVertex3fSUN')
	gl.glTexCoord2fNormal3fVertex3fSUN.gl = gl
	gl.glTexCoord2fNormal3fVertex3fSUN.__name__ = 'glTexCoord2fNormal3fVertex3fSUN'
	gl.glTexCoord2fNormal3fVertex3fSUN.errcheck = _errcheck
	gl.glTexCoord2fNormal3fVertex3fSUN.restype = GLvoid
	gl.glTexCoord2fNormal3fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord2fNormal3fVertex3fvSUN = _getProcAddress('glTexCoord2fNormal3fVertex3fvSUN')
	gl.glTexCoord2fNormal3fVertex3fvSUN.gl = gl
	gl.glTexCoord2fNormal3fVertex3fvSUN.__name__ = 'glTexCoord2fNormal3fVertex3fvSUN'
	gl.glTexCoord2fNormal3fVertex3fvSUN.errcheck = _errcheck
	gl.glTexCoord2fNormal3fVertex3fvSUN.restype = GLvoid
	gl.glTexCoord2fNormal3fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glTexCoord2fVertex3fSUN = _getProcAddress('glTexCoord2fVertex3fSUN')
	gl.glTexCoord2fVertex3fSUN.gl = gl
	gl.glTexCoord2fVertex3fSUN.__name__ = 'glTexCoord2fVertex3fSUN'
	gl.glTexCoord2fVertex3fSUN.errcheck = _errcheck
	gl.glTexCoord2fVertex3fSUN.restype = GLvoid
	gl.glTexCoord2fVertex3fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord2fVertex3fvSUN = _getProcAddress('glTexCoord2fVertex3fvSUN')
	gl.glTexCoord2fVertex3fvSUN.gl = gl
	gl.glTexCoord2fVertex3fvSUN.__name__ = 'glTexCoord2fVertex3fvSUN'
	gl.glTexCoord2fVertex3fvSUN.errcheck = _errcheck
	gl.glTexCoord2fVertex3fvSUN.restype = GLvoid
	gl.glTexCoord2fVertex3fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat),)
	gl.glTexCoord2fv = _getProcAddress('glTexCoord2fv')
	gl.glTexCoord2fv.gl = gl
	gl.glTexCoord2fv.__name__ = 'glTexCoord2fv'
	gl.glTexCoord2fv.errcheck = _errcheck
	gl.glTexCoord2fv.restype = GLvoid
	gl.glTexCoord2fv.argtypes = (POINTER(GLfloat),)
	gl.glTexCoord2hNV = _getProcAddress('glTexCoord2hNV')
	gl.glTexCoord2hNV.gl = gl
	gl.glTexCoord2hNV.__name__ = 'glTexCoord2hNV'
	gl.glTexCoord2hNV.errcheck = _errcheck
	gl.glTexCoord2hNV.restype = GLvoid
	gl.glTexCoord2hNV.argtypes = (GLhalfNV, GLhalfNV,)
	gl.glTexCoord2hvNV = _getProcAddress('glTexCoord2hvNV')
	gl.glTexCoord2hvNV.gl = gl
	gl.glTexCoord2hvNV.__name__ = 'glTexCoord2hvNV'
	gl.glTexCoord2hvNV.errcheck = _errcheck
	gl.glTexCoord2hvNV.restype = GLvoid
	gl.glTexCoord2hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glTexCoord2i = _getProcAddress('glTexCoord2i')
	gl.glTexCoord2i.gl = gl
	gl.glTexCoord2i.__name__ = 'glTexCoord2i'
	gl.glTexCoord2i.errcheck = _errcheck
	gl.glTexCoord2i.restype = GLvoid
	gl.glTexCoord2i.argtypes = (GLint, GLint,)
	gl.glTexCoord2iv = _getProcAddress('glTexCoord2iv')
	gl.glTexCoord2iv.gl = gl
	gl.glTexCoord2iv.__name__ = 'glTexCoord2iv'
	gl.glTexCoord2iv.errcheck = _errcheck
	gl.glTexCoord2iv.restype = GLvoid
	gl.glTexCoord2iv.argtypes = (POINTER(GLint),)
	gl.glTexCoord2s = _getProcAddress('glTexCoord2s')
	gl.glTexCoord2s.gl = gl
	gl.glTexCoord2s.__name__ = 'glTexCoord2s'
	gl.glTexCoord2s.errcheck = _errcheck
	gl.glTexCoord2s.restype = GLvoid
	gl.glTexCoord2s.argtypes = (GLshort, GLshort,)
	gl.glTexCoord2sv = _getProcAddress('glTexCoord2sv')
	gl.glTexCoord2sv.gl = gl
	gl.glTexCoord2sv.__name__ = 'glTexCoord2sv'
	gl.glTexCoord2sv.errcheck = _errcheck
	gl.glTexCoord2sv.restype = GLvoid
	gl.glTexCoord2sv.argtypes = (POINTER(GLshort),)
	gl.glTexCoord2xOES = _getProcAddress('glTexCoord2xOES')
	gl.glTexCoord2xOES.gl = gl
	gl.glTexCoord2xOES.__name__ = 'glTexCoord2xOES'
	gl.glTexCoord2xOES.errcheck = _errcheck
	gl.glTexCoord2xOES.restype = GLvoid
	gl.glTexCoord2xOES.argtypes = (GLfixed, GLfixed,)
	gl.glTexCoord2xvOES = _getProcAddress('glTexCoord2xvOES')
	gl.glTexCoord2xvOES.gl = gl
	gl.glTexCoord2xvOES.__name__ = 'glTexCoord2xvOES'
	gl.glTexCoord2xvOES.errcheck = _errcheck
	gl.glTexCoord2xvOES.restype = GLvoid
	gl.glTexCoord2xvOES.argtypes = (POINTER(GLfixed),)
	gl.glTexCoord3bOES = _getProcAddress('glTexCoord3bOES')
	gl.glTexCoord3bOES.gl = gl
	gl.glTexCoord3bOES.__name__ = 'glTexCoord3bOES'
	gl.glTexCoord3bOES.errcheck = _errcheck
	gl.glTexCoord3bOES.restype = GLvoid
	gl.glTexCoord3bOES.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glTexCoord3bvOES = _getProcAddress('glTexCoord3bvOES')
	gl.glTexCoord3bvOES.gl = gl
	gl.glTexCoord3bvOES.__name__ = 'glTexCoord3bvOES'
	gl.glTexCoord3bvOES.errcheck = _errcheck
	gl.glTexCoord3bvOES.restype = GLvoid
	gl.glTexCoord3bvOES.argtypes = (POINTER(GLbyte),)
	gl.glTexCoord3d = _getProcAddress('glTexCoord3d')
	gl.glTexCoord3d.gl = gl
	gl.glTexCoord3d.__name__ = 'glTexCoord3d'
	gl.glTexCoord3d.errcheck = _errcheck
	gl.glTexCoord3d.restype = GLvoid
	gl.glTexCoord3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glTexCoord3dv = _getProcAddress('glTexCoord3dv')
	gl.glTexCoord3dv.gl = gl
	gl.glTexCoord3dv.__name__ = 'glTexCoord3dv'
	gl.glTexCoord3dv.errcheck = _errcheck
	gl.glTexCoord3dv.restype = GLvoid
	gl.glTexCoord3dv.argtypes = (POINTER(GLdouble),)
	gl.glTexCoord3f = _getProcAddress('glTexCoord3f')
	gl.glTexCoord3f.gl = gl
	gl.glTexCoord3f.__name__ = 'glTexCoord3f'
	gl.glTexCoord3f.errcheck = _errcheck
	gl.glTexCoord3f.restype = GLvoid
	gl.glTexCoord3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord3fv = _getProcAddress('glTexCoord3fv')
	gl.glTexCoord3fv.gl = gl
	gl.glTexCoord3fv.__name__ = 'glTexCoord3fv'
	gl.glTexCoord3fv.errcheck = _errcheck
	gl.glTexCoord3fv.restype = GLvoid
	gl.glTexCoord3fv.argtypes = (POINTER(GLfloat),)
	gl.glTexCoord3hNV = _getProcAddress('glTexCoord3hNV')
	gl.glTexCoord3hNV.gl = gl
	gl.glTexCoord3hNV.__name__ = 'glTexCoord3hNV'
	gl.glTexCoord3hNV.errcheck = _errcheck
	gl.glTexCoord3hNV.restype = GLvoid
	gl.glTexCoord3hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glTexCoord3hvNV = _getProcAddress('glTexCoord3hvNV')
	gl.glTexCoord3hvNV.gl = gl
	gl.glTexCoord3hvNV.__name__ = 'glTexCoord3hvNV'
	gl.glTexCoord3hvNV.errcheck = _errcheck
	gl.glTexCoord3hvNV.restype = GLvoid
	gl.glTexCoord3hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glTexCoord3i = _getProcAddress('glTexCoord3i')
	gl.glTexCoord3i.gl = gl
	gl.glTexCoord3i.__name__ = 'glTexCoord3i'
	gl.glTexCoord3i.errcheck = _errcheck
	gl.glTexCoord3i.restype = GLvoid
	gl.glTexCoord3i.argtypes = (GLint, GLint, GLint,)
	gl.glTexCoord3iv = _getProcAddress('glTexCoord3iv')
	gl.glTexCoord3iv.gl = gl
	gl.glTexCoord3iv.__name__ = 'glTexCoord3iv'
	gl.glTexCoord3iv.errcheck = _errcheck
	gl.glTexCoord3iv.restype = GLvoid
	gl.glTexCoord3iv.argtypes = (POINTER(GLint),)
	gl.glTexCoord3s = _getProcAddress('glTexCoord3s')
	gl.glTexCoord3s.gl = gl
	gl.glTexCoord3s.__name__ = 'glTexCoord3s'
	gl.glTexCoord3s.errcheck = _errcheck
	gl.glTexCoord3s.restype = GLvoid
	gl.glTexCoord3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glTexCoord3sv = _getProcAddress('glTexCoord3sv')
	gl.glTexCoord3sv.gl = gl
	gl.glTexCoord3sv.__name__ = 'glTexCoord3sv'
	gl.glTexCoord3sv.errcheck = _errcheck
	gl.glTexCoord3sv.restype = GLvoid
	gl.glTexCoord3sv.argtypes = (POINTER(GLshort),)
	gl.glTexCoord3xOES = _getProcAddress('glTexCoord3xOES')
	gl.glTexCoord3xOES.gl = gl
	gl.glTexCoord3xOES.__name__ = 'glTexCoord3xOES'
	gl.glTexCoord3xOES.errcheck = _errcheck
	gl.glTexCoord3xOES.restype = GLvoid
	gl.glTexCoord3xOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glTexCoord3xvOES = _getProcAddress('glTexCoord3xvOES')
	gl.glTexCoord3xvOES.gl = gl
	gl.glTexCoord3xvOES.__name__ = 'glTexCoord3xvOES'
	gl.glTexCoord3xvOES.errcheck = _errcheck
	gl.glTexCoord3xvOES.restype = GLvoid
	gl.glTexCoord3xvOES.argtypes = (POINTER(GLfixed),)
	gl.glTexCoord4bOES = _getProcAddress('glTexCoord4bOES')
	gl.glTexCoord4bOES.gl = gl
	gl.glTexCoord4bOES.__name__ = 'glTexCoord4bOES'
	gl.glTexCoord4bOES.errcheck = _errcheck
	gl.glTexCoord4bOES.restype = GLvoid
	gl.glTexCoord4bOES.argtypes = (GLbyte, GLbyte, GLbyte, GLbyte,)
	gl.glTexCoord4bvOES = _getProcAddress('glTexCoord4bvOES')
	gl.glTexCoord4bvOES.gl = gl
	gl.glTexCoord4bvOES.__name__ = 'glTexCoord4bvOES'
	gl.glTexCoord4bvOES.errcheck = _errcheck
	gl.glTexCoord4bvOES.restype = GLvoid
	gl.glTexCoord4bvOES.argtypes = (POINTER(GLbyte),)
	gl.glTexCoord4d = _getProcAddress('glTexCoord4d')
	gl.glTexCoord4d.gl = gl
	gl.glTexCoord4d.__name__ = 'glTexCoord4d'
	gl.glTexCoord4d.errcheck = _errcheck
	gl.glTexCoord4d.restype = GLvoid
	gl.glTexCoord4d.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glTexCoord4dv = _getProcAddress('glTexCoord4dv')
	gl.glTexCoord4dv.gl = gl
	gl.glTexCoord4dv.__name__ = 'glTexCoord4dv'
	gl.glTexCoord4dv.errcheck = _errcheck
	gl.glTexCoord4dv.restype = GLvoid
	gl.glTexCoord4dv.argtypes = (POINTER(GLdouble),)
	gl.glTexCoord4f = _getProcAddress('glTexCoord4f')
	gl.glTexCoord4f.gl = gl
	gl.glTexCoord4f.__name__ = 'glTexCoord4f'
	gl.glTexCoord4f.errcheck = _errcheck
	gl.glTexCoord4f.restype = GLvoid
	gl.glTexCoord4f.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord4fColor4fNormal3fVertex4fSUN = _getProcAddress('glTexCoord4fColor4fNormal3fVertex4fSUN')
	gl.glTexCoord4fColor4fNormal3fVertex4fSUN.gl = gl
	gl.glTexCoord4fColor4fNormal3fVertex4fSUN.__name__ = 'glTexCoord4fColor4fNormal3fVertex4fSUN'
	gl.glTexCoord4fColor4fNormal3fVertex4fSUN.errcheck = _errcheck
	gl.glTexCoord4fColor4fNormal3fVertex4fSUN.restype = GLvoid
	gl.glTexCoord4fColor4fNormal3fVertex4fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord4fColor4fNormal3fVertex4fvSUN = _getProcAddress('glTexCoord4fColor4fNormal3fVertex4fvSUN')
	gl.glTexCoord4fColor4fNormal3fVertex4fvSUN.gl = gl
	gl.glTexCoord4fColor4fNormal3fVertex4fvSUN.__name__ = 'glTexCoord4fColor4fNormal3fVertex4fvSUN'
	gl.glTexCoord4fColor4fNormal3fVertex4fvSUN.errcheck = _errcheck
	gl.glTexCoord4fColor4fNormal3fVertex4fvSUN.restype = GLvoid
	gl.glTexCoord4fColor4fNormal3fVertex4fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat), POINTER(GLfloat),)
	gl.glTexCoord4fVertex4fSUN = _getProcAddress('glTexCoord4fVertex4fSUN')
	gl.glTexCoord4fVertex4fSUN.gl = gl
	gl.glTexCoord4fVertex4fSUN.__name__ = 'glTexCoord4fVertex4fSUN'
	gl.glTexCoord4fVertex4fSUN.errcheck = _errcheck
	gl.glTexCoord4fVertex4fSUN.restype = GLvoid
	gl.glTexCoord4fVertex4fSUN.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glTexCoord4fVertex4fvSUN = _getProcAddress('glTexCoord4fVertex4fvSUN')
	gl.glTexCoord4fVertex4fvSUN.gl = gl
	gl.glTexCoord4fVertex4fvSUN.__name__ = 'glTexCoord4fVertex4fvSUN'
	gl.glTexCoord4fVertex4fvSUN.errcheck = _errcheck
	gl.glTexCoord4fVertex4fvSUN.restype = GLvoid
	gl.glTexCoord4fVertex4fvSUN.argtypes = (POINTER(GLfloat), POINTER(GLfloat),)
	gl.glTexCoord4fv = _getProcAddress('glTexCoord4fv')
	gl.glTexCoord4fv.gl = gl
	gl.glTexCoord4fv.__name__ = 'glTexCoord4fv'
	gl.glTexCoord4fv.errcheck = _errcheck
	gl.glTexCoord4fv.restype = GLvoid
	gl.glTexCoord4fv.argtypes = (POINTER(GLfloat),)
	gl.glTexCoord4hNV = _getProcAddress('glTexCoord4hNV')
	gl.glTexCoord4hNV.gl = gl
	gl.glTexCoord4hNV.__name__ = 'glTexCoord4hNV'
	gl.glTexCoord4hNV.errcheck = _errcheck
	gl.glTexCoord4hNV.restype = GLvoid
	gl.glTexCoord4hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glTexCoord4hvNV = _getProcAddress('glTexCoord4hvNV')
	gl.glTexCoord4hvNV.gl = gl
	gl.glTexCoord4hvNV.__name__ = 'glTexCoord4hvNV'
	gl.glTexCoord4hvNV.errcheck = _errcheck
	gl.glTexCoord4hvNV.restype = GLvoid
	gl.glTexCoord4hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glTexCoord4i = _getProcAddress('glTexCoord4i')
	gl.glTexCoord4i.gl = gl
	gl.glTexCoord4i.__name__ = 'glTexCoord4i'
	gl.glTexCoord4i.errcheck = _errcheck
	gl.glTexCoord4i.restype = GLvoid
	gl.glTexCoord4i.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glTexCoord4iv = _getProcAddress('glTexCoord4iv')
	gl.glTexCoord4iv.gl = gl
	gl.glTexCoord4iv.__name__ = 'glTexCoord4iv'
	gl.glTexCoord4iv.errcheck = _errcheck
	gl.glTexCoord4iv.restype = GLvoid
	gl.glTexCoord4iv.argtypes = (POINTER(GLint),)
	gl.glTexCoord4s = _getProcAddress('glTexCoord4s')
	gl.glTexCoord4s.gl = gl
	gl.glTexCoord4s.__name__ = 'glTexCoord4s'
	gl.glTexCoord4s.errcheck = _errcheck
	gl.glTexCoord4s.restype = GLvoid
	gl.glTexCoord4s.argtypes = (GLshort, GLshort, GLshort, GLshort,)
	gl.glTexCoord4sv = _getProcAddress('glTexCoord4sv')
	gl.glTexCoord4sv.gl = gl
	gl.glTexCoord4sv.__name__ = 'glTexCoord4sv'
	gl.glTexCoord4sv.errcheck = _errcheck
	gl.glTexCoord4sv.restype = GLvoid
	gl.glTexCoord4sv.argtypes = (POINTER(GLshort),)
	gl.glTexCoord4xOES = _getProcAddress('glTexCoord4xOES')
	gl.glTexCoord4xOES.gl = gl
	gl.glTexCoord4xOES.__name__ = 'glTexCoord4xOES'
	gl.glTexCoord4xOES.errcheck = _errcheck
	gl.glTexCoord4xOES.restype = GLvoid
	gl.glTexCoord4xOES.argtypes = (GLfixed, GLfixed, GLfixed, GLfixed,)
	gl.glTexCoord4xvOES = _getProcAddress('glTexCoord4xvOES')
	gl.glTexCoord4xvOES.gl = gl
	gl.glTexCoord4xvOES.__name__ = 'glTexCoord4xvOES'
	gl.glTexCoord4xvOES.errcheck = _errcheck
	gl.glTexCoord4xvOES.restype = GLvoid
	gl.glTexCoord4xvOES.argtypes = (POINTER(GLfixed),)
	gl.glTexCoordFormatNV = _getProcAddress('glTexCoordFormatNV')
	gl.glTexCoordFormatNV.gl = gl
	gl.glTexCoordFormatNV.__name__ = 'glTexCoordFormatNV'
	gl.glTexCoordFormatNV.errcheck = _errcheck
	gl.glTexCoordFormatNV.restype = GLvoid
	gl.glTexCoordFormatNV.argtypes = (GLint, GLenum, GLsizei,)
	gl.glTexCoordP1ui = _getProcAddress('glTexCoordP1ui')
	gl.glTexCoordP1ui.gl = gl
	gl.glTexCoordP1ui.__name__ = 'glTexCoordP1ui'
	gl.glTexCoordP1ui.errcheck = _errcheck
	gl.glTexCoordP1ui.restype = GLvoid
	gl.glTexCoordP1ui.argtypes = (GLenum, GLuint,)
	gl.glTexCoordP1uiv = _getProcAddress('glTexCoordP1uiv')
	gl.glTexCoordP1uiv.gl = gl
	gl.glTexCoordP1uiv.__name__ = 'glTexCoordP1uiv'
	gl.glTexCoordP1uiv.errcheck = _errcheck
	gl.glTexCoordP1uiv.restype = GLvoid
	gl.glTexCoordP1uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glTexCoordP2ui = _getProcAddress('glTexCoordP2ui')
	gl.glTexCoordP2ui.gl = gl
	gl.glTexCoordP2ui.__name__ = 'glTexCoordP2ui'
	gl.glTexCoordP2ui.errcheck = _errcheck
	gl.glTexCoordP2ui.restype = GLvoid
	gl.glTexCoordP2ui.argtypes = (GLenum, GLuint,)
	gl.glTexCoordP2uiv = _getProcAddress('glTexCoordP2uiv')
	gl.glTexCoordP2uiv.gl = gl
	gl.glTexCoordP2uiv.__name__ = 'glTexCoordP2uiv'
	gl.glTexCoordP2uiv.errcheck = _errcheck
	gl.glTexCoordP2uiv.restype = GLvoid
	gl.glTexCoordP2uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glTexCoordP3ui = _getProcAddress('glTexCoordP3ui')
	gl.glTexCoordP3ui.gl = gl
	gl.glTexCoordP3ui.__name__ = 'glTexCoordP3ui'
	gl.glTexCoordP3ui.errcheck = _errcheck
	gl.glTexCoordP3ui.restype = GLvoid
	gl.glTexCoordP3ui.argtypes = (GLenum, GLuint,)
	gl.glTexCoordP3uiv = _getProcAddress('glTexCoordP3uiv')
	gl.glTexCoordP3uiv.gl = gl
	gl.glTexCoordP3uiv.__name__ = 'glTexCoordP3uiv'
	gl.glTexCoordP3uiv.errcheck = _errcheck
	gl.glTexCoordP3uiv.restype = GLvoid
	gl.glTexCoordP3uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glTexCoordP4ui = _getProcAddress('glTexCoordP4ui')
	gl.glTexCoordP4ui.gl = gl
	gl.glTexCoordP4ui.__name__ = 'glTexCoordP4ui'
	gl.glTexCoordP4ui.errcheck = _errcheck
	gl.glTexCoordP4ui.restype = GLvoid
	gl.glTexCoordP4ui.argtypes = (GLenum, GLuint,)
	gl.glTexCoordP4uiv = _getProcAddress('glTexCoordP4uiv')
	gl.glTexCoordP4uiv.gl = gl
	gl.glTexCoordP4uiv.__name__ = 'glTexCoordP4uiv'
	gl.glTexCoordP4uiv.errcheck = _errcheck
	gl.glTexCoordP4uiv.restype = GLvoid
	gl.glTexCoordP4uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glTexCoordPointer = _getProcAddress('glTexCoordPointer')
	gl.glTexCoordPointer.gl = gl
	gl.glTexCoordPointer.__name__ = 'glTexCoordPointer'
	gl.glTexCoordPointer.errcheck = _errcheck
	gl.glTexCoordPointer.restype = GLvoid
	gl.glTexCoordPointer.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glTexCoordPointerEXT = _getProcAddress('glTexCoordPointerEXT')
	gl.glTexCoordPointerEXT.gl = gl
	gl.glTexCoordPointerEXT.__name__ = 'glTexCoordPointerEXT'
	gl.glTexCoordPointerEXT.errcheck = _errcheck
	gl.glTexCoordPointerEXT.restype = GLvoid
	gl.glTexCoordPointerEXT.argtypes = (GLint, GLenum, GLsizei, GLsizei, POINTER(GLvoid),)
	gl.glTexCoordPointerListIBM = _getProcAddress('glTexCoordPointerListIBM')
	gl.glTexCoordPointerListIBM.gl = gl
	gl.glTexCoordPointerListIBM.__name__ = 'glTexCoordPointerListIBM'
	gl.glTexCoordPointerListIBM.errcheck = _errcheck
	gl.glTexCoordPointerListIBM.restype = GLvoid
	gl.glTexCoordPointerListIBM.argtypes = (GLint, GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glTexCoordPointervINTEL = _getProcAddress('glTexCoordPointervINTEL')
	gl.glTexCoordPointervINTEL.gl = gl
	gl.glTexCoordPointervINTEL.__name__ = 'glTexCoordPointervINTEL'
	gl.glTexCoordPointervINTEL.errcheck = _errcheck
	gl.glTexCoordPointervINTEL.restype = GLvoid
	gl.glTexCoordPointervINTEL.argtypes = (GLint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glTexEnvf = _getProcAddress('glTexEnvf')
	gl.glTexEnvf.gl = gl
	gl.glTexEnvf.__name__ = 'glTexEnvf'
	gl.glTexEnvf.errcheck = _errcheck
	gl.glTexEnvf.restype = GLvoid
	gl.glTexEnvf.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glTexEnvfv = _getProcAddress('glTexEnvfv')
	gl.glTexEnvfv.gl = gl
	gl.glTexEnvfv.__name__ = 'glTexEnvfv'
	gl.glTexEnvfv.errcheck = _errcheck
	gl.glTexEnvfv.restype = GLvoid
	gl.glTexEnvfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glTexEnvi = _getProcAddress('glTexEnvi')
	gl.glTexEnvi.gl = gl
	gl.glTexEnvi.__name__ = 'glTexEnvi'
	gl.glTexEnvi.errcheck = _errcheck
	gl.glTexEnvi.restype = GLvoid
	gl.glTexEnvi.argtypes = (GLenum, GLenum, GLint,)
	gl.glTexEnviv = _getProcAddress('glTexEnviv')
	gl.glTexEnviv.gl = gl
	gl.glTexEnviv.__name__ = 'glTexEnviv'
	gl.glTexEnviv.errcheck = _errcheck
	gl.glTexEnviv.restype = GLvoid
	gl.glTexEnviv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glTexEnvx = _getProcAddress('glTexEnvx')
	gl.glTexEnvx.gl = gl
	gl.glTexEnvx.__name__ = 'glTexEnvx'
	gl.glTexEnvx.errcheck = _errcheck
	gl.glTexEnvx.restype = GLvoid
	gl.glTexEnvx.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glTexEnvxOES = _getProcAddress('glTexEnvxOES')
	gl.glTexEnvxOES.gl = gl
	gl.glTexEnvxOES.__name__ = 'glTexEnvxOES'
	gl.glTexEnvxOES.errcheck = _errcheck
	gl.glTexEnvxOES.restype = GLvoid
	gl.glTexEnvxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glTexEnvxv = _getProcAddress('glTexEnvxv')
	gl.glTexEnvxv.gl = gl
	gl.glTexEnvxv.__name__ = 'glTexEnvxv'
	gl.glTexEnvxv.errcheck = _errcheck
	gl.glTexEnvxv.restype = GLvoid
	gl.glTexEnvxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glTexEnvxvOES = _getProcAddress('glTexEnvxvOES')
	gl.glTexEnvxvOES.gl = gl
	gl.glTexEnvxvOES.__name__ = 'glTexEnvxvOES'
	gl.glTexEnvxvOES.errcheck = _errcheck
	gl.glTexEnvxvOES.restype = GLvoid
	gl.glTexEnvxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glTexFilterFuncSGIS = _getProcAddress('glTexFilterFuncSGIS')
	gl.glTexFilterFuncSGIS.gl = gl
	gl.glTexFilterFuncSGIS.__name__ = 'glTexFilterFuncSGIS'
	gl.glTexFilterFuncSGIS.errcheck = _errcheck
	gl.glTexFilterFuncSGIS.restype = GLvoid
	gl.glTexFilterFuncSGIS.argtypes = (GLenum, GLenum, GLsizei, POINTER(GLfloat),)
	gl.glTexGend = _getProcAddress('glTexGend')
	gl.glTexGend.gl = gl
	gl.glTexGend.__name__ = 'glTexGend'
	gl.glTexGend.errcheck = _errcheck
	gl.glTexGend.restype = GLvoid
	gl.glTexGend.argtypes = (GLenum, GLenum, GLdouble,)
	gl.glTexGendv = _getProcAddress('glTexGendv')
	gl.glTexGendv.gl = gl
	gl.glTexGendv.__name__ = 'glTexGendv'
	gl.glTexGendv.errcheck = _errcheck
	gl.glTexGendv.restype = GLvoid
	gl.glTexGendv.argtypes = (GLenum, GLenum, POINTER(GLdouble),)
	gl.glTexGenf = _getProcAddress('glTexGenf')
	gl.glTexGenf.gl = gl
	gl.glTexGenf.__name__ = 'glTexGenf'
	gl.glTexGenf.errcheck = _errcheck
	gl.glTexGenf.restype = GLvoid
	gl.glTexGenf.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glTexGenfOES = _getProcAddress('glTexGenfOES')
	gl.glTexGenfOES.gl = gl
	gl.glTexGenfOES.__name__ = 'glTexGenfOES'
	gl.glTexGenfOES.errcheck = _errcheck
	gl.glTexGenfOES.restype = GLvoid
	gl.glTexGenfOES.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glTexGenfv = _getProcAddress('glTexGenfv')
	gl.glTexGenfv.gl = gl
	gl.glTexGenfv.__name__ = 'glTexGenfv'
	gl.glTexGenfv.errcheck = _errcheck
	gl.glTexGenfv.restype = GLvoid
	gl.glTexGenfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glTexGenfvOES = _getProcAddress('glTexGenfvOES')
	gl.glTexGenfvOES.gl = gl
	gl.glTexGenfvOES.__name__ = 'glTexGenfvOES'
	gl.glTexGenfvOES.errcheck = _errcheck
	gl.glTexGenfvOES.restype = GLvoid
	gl.glTexGenfvOES.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glTexGeni = _getProcAddress('glTexGeni')
	gl.glTexGeni.gl = gl
	gl.glTexGeni.__name__ = 'glTexGeni'
	gl.glTexGeni.errcheck = _errcheck
	gl.glTexGeni.restype = GLvoid
	gl.glTexGeni.argtypes = (GLenum, GLenum, GLint,)
	gl.glTexGeniOES = _getProcAddress('glTexGeniOES')
	gl.glTexGeniOES.gl = gl
	gl.glTexGeniOES.__name__ = 'glTexGeniOES'
	gl.glTexGeniOES.errcheck = _errcheck
	gl.glTexGeniOES.restype = GLvoid
	gl.glTexGeniOES.argtypes = (GLenum, GLenum, GLint,)
	gl.glTexGeniv = _getProcAddress('glTexGeniv')
	gl.glTexGeniv.gl = gl
	gl.glTexGeniv.__name__ = 'glTexGeniv'
	gl.glTexGeniv.errcheck = _errcheck
	gl.glTexGeniv.restype = GLvoid
	gl.glTexGeniv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glTexGenivOES = _getProcAddress('glTexGenivOES')
	gl.glTexGenivOES.gl = gl
	gl.glTexGenivOES.__name__ = 'glTexGenivOES'
	gl.glTexGenivOES.errcheck = _errcheck
	gl.glTexGenivOES.restype = GLvoid
	gl.glTexGenivOES.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glTexGenxOES = _getProcAddress('glTexGenxOES')
	gl.glTexGenxOES.gl = gl
	gl.glTexGenxOES.__name__ = 'glTexGenxOES'
	gl.glTexGenxOES.errcheck = _errcheck
	gl.glTexGenxOES.restype = GLvoid
	gl.glTexGenxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glTexGenxvOES = _getProcAddress('glTexGenxvOES')
	gl.glTexGenxvOES.gl = gl
	gl.glTexGenxvOES.__name__ = 'glTexGenxvOES'
	gl.glTexGenxvOES.errcheck = _errcheck
	gl.glTexGenxvOES.restype = GLvoid
	gl.glTexGenxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glTexImage1D = _getProcAddress('glTexImage1D')
	gl.glTexImage1D.gl = gl
	gl.glTexImage1D.__name__ = 'glTexImage1D'
	gl.glTexImage1D.errcheck = _errcheck
	gl.glTexImage1D.restype = GLvoid
	gl.glTexImage1D.argtypes = (GLenum, GLint, GLint, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexImage2D = _getProcAddress('glTexImage2D')
	gl.glTexImage2D.gl = gl
	gl.glTexImage2D.__name__ = 'glTexImage2D'
	gl.glTexImage2D.errcheck = _errcheck
	gl.glTexImage2D.restype = GLvoid
	gl.glTexImage2D.argtypes = (GLenum, GLint, GLint, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexImage2DMultisample = _getProcAddress('glTexImage2DMultisample')
	gl.glTexImage2DMultisample.gl = gl
	gl.glTexImage2DMultisample.__name__ = 'glTexImage2DMultisample'
	gl.glTexImage2DMultisample.errcheck = _errcheck
	gl.glTexImage2DMultisample.restype = GLvoid
	gl.glTexImage2DMultisample.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLboolean,)
	gl.glTexImage2DMultisampleCoverageNV = _getProcAddress('glTexImage2DMultisampleCoverageNV')
	gl.glTexImage2DMultisampleCoverageNV.gl = gl
	gl.glTexImage2DMultisampleCoverageNV.__name__ = 'glTexImage2DMultisampleCoverageNV'
	gl.glTexImage2DMultisampleCoverageNV.errcheck = _errcheck
	gl.glTexImage2DMultisampleCoverageNV.restype = GLvoid
	gl.glTexImage2DMultisampleCoverageNV.argtypes = (GLenum, GLsizei, GLsizei, GLint, GLsizei, GLsizei, GLboolean,)
	gl.glTexImage3D = _getProcAddress('glTexImage3D')
	gl.glTexImage3D.gl = gl
	gl.glTexImage3D.__name__ = 'glTexImage3D'
	gl.glTexImage3D.errcheck = _errcheck
	gl.glTexImage3D.restype = GLvoid
	gl.glTexImage3D.argtypes = (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexImage3DEXT = _getProcAddress('glTexImage3DEXT')
	gl.glTexImage3DEXT.gl = gl
	gl.glTexImage3DEXT.__name__ = 'glTexImage3DEXT'
	gl.glTexImage3DEXT.errcheck = _errcheck
	gl.glTexImage3DEXT.restype = GLvoid
	gl.glTexImage3DEXT.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexImage3DMultisample = _getProcAddress('glTexImage3DMultisample')
	gl.glTexImage3DMultisample.gl = gl
	gl.glTexImage3DMultisample.__name__ = 'glTexImage3DMultisample'
	gl.glTexImage3DMultisample.errcheck = _errcheck
	gl.glTexImage3DMultisample.restype = GLvoid
	gl.glTexImage3DMultisample.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTexImage3DMultisampleCoverageNV = _getProcAddress('glTexImage3DMultisampleCoverageNV')
	gl.glTexImage3DMultisampleCoverageNV.gl = gl
	gl.glTexImage3DMultisampleCoverageNV.__name__ = 'glTexImage3DMultisampleCoverageNV'
	gl.glTexImage3DMultisampleCoverageNV.errcheck = _errcheck
	gl.glTexImage3DMultisampleCoverageNV.restype = GLvoid
	gl.glTexImage3DMultisampleCoverageNV.argtypes = (GLenum, GLsizei, GLsizei, GLint, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTexImage3DOES = _getProcAddress('glTexImage3DOES')
	gl.glTexImage3DOES.gl = gl
	gl.glTexImage3DOES.__name__ = 'glTexImage3DOES'
	gl.glTexImage3DOES.errcheck = _errcheck
	gl.glTexImage3DOES.restype = GLvoid
	gl.glTexImage3DOES.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexImage4DSGIS = _getProcAddress('glTexImage4DSGIS')
	gl.glTexImage4DSGIS.gl = gl
	gl.glTexImage4DSGIS.__name__ = 'glTexImage4DSGIS'
	gl.glTexImage4DSGIS.errcheck = _errcheck
	gl.glTexImage4DSGIS.restype = GLvoid
	gl.glTexImage4DSGIS.argtypes = (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexPageCommitmentARB = _getProcAddress('glTexPageCommitmentARB')
	gl.glTexPageCommitmentARB.gl = gl
	gl.glTexPageCommitmentARB.__name__ = 'glTexPageCommitmentARB'
	gl.glTexPageCommitmentARB.errcheck = _errcheck
	gl.glTexPageCommitmentARB.restype = GLvoid
	gl.glTexPageCommitmentARB.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTexParameterIiv = _getProcAddress('glTexParameterIiv')
	gl.glTexParameterIiv.gl = gl
	gl.glTexParameterIiv.__name__ = 'glTexParameterIiv'
	gl.glTexParameterIiv.errcheck = _errcheck
	gl.glTexParameterIiv.restype = GLvoid
	gl.glTexParameterIiv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glTexParameterIivEXT = _getProcAddress('glTexParameterIivEXT')
	gl.glTexParameterIivEXT.gl = gl
	gl.glTexParameterIivEXT.__name__ = 'glTexParameterIivEXT'
	gl.glTexParameterIivEXT.errcheck = _errcheck
	gl.glTexParameterIivEXT.restype = GLvoid
	gl.glTexParameterIivEXT.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glTexParameterIuiv = _getProcAddress('glTexParameterIuiv')
	gl.glTexParameterIuiv.gl = gl
	gl.glTexParameterIuiv.__name__ = 'glTexParameterIuiv'
	gl.glTexParameterIuiv.errcheck = _errcheck
	gl.glTexParameterIuiv.restype = GLvoid
	gl.glTexParameterIuiv.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glTexParameterIuivEXT = _getProcAddress('glTexParameterIuivEXT')
	gl.glTexParameterIuivEXT.gl = gl
	gl.glTexParameterIuivEXT.__name__ = 'glTexParameterIuivEXT'
	gl.glTexParameterIuivEXT.errcheck = _errcheck
	gl.glTexParameterIuivEXT.restype = GLvoid
	gl.glTexParameterIuivEXT.argtypes = (GLenum, GLenum, POINTER(GLuint),)
	gl.glTexParameterf = _getProcAddress('glTexParameterf')
	gl.glTexParameterf.gl = gl
	gl.glTexParameterf.__name__ = 'glTexParameterf'
	gl.glTexParameterf.errcheck = _errcheck
	gl.glTexParameterf.restype = GLvoid
	gl.glTexParameterf.argtypes = (GLenum, GLenum, GLfloat,)
	gl.glTexParameterfv = _getProcAddress('glTexParameterfv')
	gl.glTexParameterfv.gl = gl
	gl.glTexParameterfv.__name__ = 'glTexParameterfv'
	gl.glTexParameterfv.errcheck = _errcheck
	gl.glTexParameterfv.restype = GLvoid
	gl.glTexParameterfv.argtypes = (GLenum, GLenum, POINTER(GLfloat),)
	gl.glTexParameteri = _getProcAddress('glTexParameteri')
	gl.glTexParameteri.gl = gl
	gl.glTexParameteri.__name__ = 'glTexParameteri'
	gl.glTexParameteri.errcheck = _errcheck
	gl.glTexParameteri.restype = GLvoid
	gl.glTexParameteri.argtypes = (GLenum, GLenum, GLint,)
	gl.glTexParameteriv = _getProcAddress('glTexParameteriv')
	gl.glTexParameteriv.gl = gl
	gl.glTexParameteriv.__name__ = 'glTexParameteriv'
	gl.glTexParameteriv.errcheck = _errcheck
	gl.glTexParameteriv.restype = GLvoid
	gl.glTexParameteriv.argtypes = (GLenum, GLenum, POINTER(GLint),)
	gl.glTexParameterx = _getProcAddress('glTexParameterx')
	gl.glTexParameterx.gl = gl
	gl.glTexParameterx.__name__ = 'glTexParameterx'
	gl.glTexParameterx.errcheck = _errcheck
	gl.glTexParameterx.restype = GLvoid
	gl.glTexParameterx.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glTexParameterxOES = _getProcAddress('glTexParameterxOES')
	gl.glTexParameterxOES.gl = gl
	gl.glTexParameterxOES.__name__ = 'glTexParameterxOES'
	gl.glTexParameterxOES.errcheck = _errcheck
	gl.glTexParameterxOES.restype = GLvoid
	gl.glTexParameterxOES.argtypes = (GLenum, GLenum, GLfixed,)
	gl.glTexParameterxv = _getProcAddress('glTexParameterxv')
	gl.glTexParameterxv.gl = gl
	gl.glTexParameterxv.__name__ = 'glTexParameterxv'
	gl.glTexParameterxv.errcheck = _errcheck
	gl.glTexParameterxv.restype = GLvoid
	gl.glTexParameterxv.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glTexParameterxvOES = _getProcAddress('glTexParameterxvOES')
	gl.glTexParameterxvOES.gl = gl
	gl.glTexParameterxvOES.__name__ = 'glTexParameterxvOES'
	gl.glTexParameterxvOES.errcheck = _errcheck
	gl.glTexParameterxvOES.restype = GLvoid
	gl.glTexParameterxvOES.argtypes = (GLenum, GLenum, POINTER(GLfixed),)
	gl.glTexRenderbufferNV = _getProcAddress('glTexRenderbufferNV')
	gl.glTexRenderbufferNV.gl = gl
	gl.glTexRenderbufferNV.__name__ = 'glTexRenderbufferNV'
	gl.glTexRenderbufferNV.errcheck = _errcheck
	gl.glTexRenderbufferNV.restype = GLvoid
	gl.glTexRenderbufferNV.argtypes = (GLenum, GLuint,)
	gl.glTexStorage1D = _getProcAddress('glTexStorage1D')
	gl.glTexStorage1D.gl = gl
	gl.glTexStorage1D.__name__ = 'glTexStorage1D'
	gl.glTexStorage1D.errcheck = _errcheck
	gl.glTexStorage1D.restype = GLvoid
	gl.glTexStorage1D.argtypes = (GLenum, GLsizei, GLenum, GLsizei,)
	gl.glTexStorage1DEXT = _getProcAddress('glTexStorage1DEXT')
	gl.glTexStorage1DEXT.gl = gl
	gl.glTexStorage1DEXT.__name__ = 'glTexStorage1DEXT'
	gl.glTexStorage1DEXT.errcheck = _errcheck
	gl.glTexStorage1DEXT.restype = GLvoid
	gl.glTexStorage1DEXT.argtypes = (GLenum, GLsizei, GLenum, GLsizei,)
	gl.glTexStorage2D = _getProcAddress('glTexStorage2D')
	gl.glTexStorage2D.gl = gl
	gl.glTexStorage2D.__name__ = 'glTexStorage2D'
	gl.glTexStorage2D.errcheck = _errcheck
	gl.glTexStorage2D.restype = GLvoid
	gl.glTexStorage2D.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glTexStorage2DEXT = _getProcAddress('glTexStorage2DEXT')
	gl.glTexStorage2DEXT.gl = gl
	gl.glTexStorage2DEXT.__name__ = 'glTexStorage2DEXT'
	gl.glTexStorage2DEXT.errcheck = _errcheck
	gl.glTexStorage2DEXT.restype = GLvoid
	gl.glTexStorage2DEXT.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glTexStorage2DMultisample = _getProcAddress('glTexStorage2DMultisample')
	gl.glTexStorage2DMultisample.gl = gl
	gl.glTexStorage2DMultisample.__name__ = 'glTexStorage2DMultisample'
	gl.glTexStorage2DMultisample.errcheck = _errcheck
	gl.glTexStorage2DMultisample.restype = GLvoid
	gl.glTexStorage2DMultisample.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLboolean,)
	gl.glTexStorage3D = _getProcAddress('glTexStorage3D')
	gl.glTexStorage3D.gl = gl
	gl.glTexStorage3D.__name__ = 'glTexStorage3D'
	gl.glTexStorage3D.errcheck = _errcheck
	gl.glTexStorage3D.restype = GLvoid
	gl.glTexStorage3D.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei,)
	gl.glTexStorage3DEXT = _getProcAddress('glTexStorage3DEXT')
	gl.glTexStorage3DEXT.gl = gl
	gl.glTexStorage3DEXT.__name__ = 'glTexStorage3DEXT'
	gl.glTexStorage3DEXT.errcheck = _errcheck
	gl.glTexStorage3DEXT.restype = GLvoid
	gl.glTexStorage3DEXT.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei,)
	gl.glTexStorage3DMultisample = _getProcAddress('glTexStorage3DMultisample')
	gl.glTexStorage3DMultisample.gl = gl
	gl.glTexStorage3DMultisample.__name__ = 'glTexStorage3DMultisample'
	gl.glTexStorage3DMultisample.errcheck = _errcheck
	gl.glTexStorage3DMultisample.restype = GLvoid
	gl.glTexStorage3DMultisample.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTexStorage3DMultisampleOES = _getProcAddress('glTexStorage3DMultisampleOES')
	gl.glTexStorage3DMultisampleOES.gl = gl
	gl.glTexStorage3DMultisampleOES.__name__ = 'glTexStorage3DMultisampleOES'
	gl.glTexStorage3DMultisampleOES.errcheck = _errcheck
	gl.glTexStorage3DMultisampleOES.restype = GLvoid
	gl.glTexStorage3DMultisampleOES.argtypes = (GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTexStorageSparseAMD = _getProcAddress('glTexStorageSparseAMD')
	gl.glTexStorageSparseAMD.gl = gl
	gl.glTexStorageSparseAMD.__name__ = 'glTexStorageSparseAMD'
	gl.glTexStorageSparseAMD.errcheck = _errcheck
	gl.glTexStorageSparseAMD.restype = GLvoid
	gl.glTexStorageSparseAMD.argtypes = (GLenum, GLenum, GLsizei, GLsizei, GLsizei, GLsizei, GLbitfield,)
	gl.glTexSubImage1D = _getProcAddress('glTexSubImage1D')
	gl.glTexSubImage1D.gl = gl
	gl.glTexSubImage1D.__name__ = 'glTexSubImage1D'
	gl.glTexSubImage1D.errcheck = _errcheck
	gl.glTexSubImage1D.restype = GLvoid
	gl.glTexSubImage1D.argtypes = (GLenum, GLint, GLint, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage1DEXT = _getProcAddress('glTexSubImage1DEXT')
	gl.glTexSubImage1DEXT.gl = gl
	gl.glTexSubImage1DEXT.__name__ = 'glTexSubImage1DEXT'
	gl.glTexSubImage1DEXT.errcheck = _errcheck
	gl.glTexSubImage1DEXT.restype = GLvoid
	gl.glTexSubImage1DEXT.argtypes = (GLenum, GLint, GLint, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage2D = _getProcAddress('glTexSubImage2D')
	gl.glTexSubImage2D.gl = gl
	gl.glTexSubImage2D.__name__ = 'glTexSubImage2D'
	gl.glTexSubImage2D.errcheck = _errcheck
	gl.glTexSubImage2D.restype = GLvoid
	gl.glTexSubImage2D.argtypes = (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage2DEXT = _getProcAddress('glTexSubImage2DEXT')
	gl.glTexSubImage2DEXT.gl = gl
	gl.glTexSubImage2DEXT.__name__ = 'glTexSubImage2DEXT'
	gl.glTexSubImage2DEXT.errcheck = _errcheck
	gl.glTexSubImage2DEXT.restype = GLvoid
	gl.glTexSubImage2DEXT.argtypes = (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage3D = _getProcAddress('glTexSubImage3D')
	gl.glTexSubImage3D.gl = gl
	gl.glTexSubImage3D.__name__ = 'glTexSubImage3D'
	gl.glTexSubImage3D.errcheck = _errcheck
	gl.glTexSubImage3D.restype = GLvoid
	gl.glTexSubImage3D.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage3DEXT = _getProcAddress('glTexSubImage3DEXT')
	gl.glTexSubImage3DEXT.gl = gl
	gl.glTexSubImage3DEXT.__name__ = 'glTexSubImage3DEXT'
	gl.glTexSubImage3DEXT.errcheck = _errcheck
	gl.glTexSubImage3DEXT.restype = GLvoid
	gl.glTexSubImage3DEXT.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage3DOES = _getProcAddress('glTexSubImage3DOES')
	gl.glTexSubImage3DOES.gl = gl
	gl.glTexSubImage3DOES.__name__ = 'glTexSubImage3DOES'
	gl.glTexSubImage3DOES.errcheck = _errcheck
	gl.glTexSubImage3DOES.restype = GLvoid
	gl.glTexSubImage3DOES.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTexSubImage4DSGIS = _getProcAddress('glTexSubImage4DSGIS')
	gl.glTexSubImage4DSGIS.gl = gl
	gl.glTexSubImage4DSGIS.__name__ = 'glTexSubImage4DSGIS'
	gl.glTexSubImage4DSGIS.errcheck = _errcheck
	gl.glTexSubImage4DSGIS.restype = GLvoid
	gl.glTexSubImage4DSGIS.argtypes = (GLenum, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureBarrier = _getProcAddress('glTextureBarrier')
	gl.glTextureBarrier.gl = gl
	gl.glTextureBarrier.__name__ = 'glTextureBarrier'
	gl.glTextureBarrier.errcheck = _errcheck
	gl.glTextureBarrier.restype = GLvoid
	gl.glTextureBarrier.argtypes = ()
	gl.glTextureBarrierNV = _getProcAddress('glTextureBarrierNV')
	gl.glTextureBarrierNV.gl = gl
	gl.glTextureBarrierNV.__name__ = 'glTextureBarrierNV'
	gl.glTextureBarrierNV.errcheck = _errcheck
	gl.glTextureBarrierNV.restype = GLvoid
	gl.glTextureBarrierNV.argtypes = ()
	gl.glTextureBuffer = _getProcAddress('glTextureBuffer')
	gl.glTextureBuffer.gl = gl
	gl.glTextureBuffer.__name__ = 'glTextureBuffer'
	gl.glTextureBuffer.errcheck = _errcheck
	gl.glTextureBuffer.restype = GLvoid
	gl.glTextureBuffer.argtypes = (GLuint, GLenum, GLuint,)
	gl.glTextureBufferEXT = _getProcAddress('glTextureBufferEXT')
	gl.glTextureBufferEXT.gl = gl
	gl.glTextureBufferEXT.__name__ = 'glTextureBufferEXT'
	gl.glTextureBufferEXT.errcheck = _errcheck
	gl.glTextureBufferEXT.restype = GLvoid
	gl.glTextureBufferEXT.argtypes = (GLuint, GLenum, GLenum, GLuint,)
	gl.glTextureBufferRange = _getProcAddress('glTextureBufferRange')
	gl.glTextureBufferRange.gl = gl
	gl.glTextureBufferRange.__name__ = 'glTextureBufferRange'
	gl.glTextureBufferRange.errcheck = _errcheck
	gl.glTextureBufferRange.restype = GLvoid
	gl.glTextureBufferRange.argtypes = (GLuint, GLenum, GLuint, GLintptr, GLsizei,)
	gl.glTextureBufferRangeEXT = _getProcAddress('glTextureBufferRangeEXT')
	gl.glTextureBufferRangeEXT.gl = gl
	gl.glTextureBufferRangeEXT.__name__ = 'glTextureBufferRangeEXT'
	gl.glTextureBufferRangeEXT.errcheck = _errcheck
	gl.glTextureBufferRangeEXT.restype = GLvoid
	gl.glTextureBufferRangeEXT.argtypes = (GLuint, GLenum, GLenum, GLuint, GLintptr, GLsizeiptr,)
	gl.glTextureColorMaskSGIS = _getProcAddress('glTextureColorMaskSGIS')
	gl.glTextureColorMaskSGIS.gl = gl
	gl.glTextureColorMaskSGIS.__name__ = 'glTextureColorMaskSGIS'
	gl.glTextureColorMaskSGIS.errcheck = _errcheck
	gl.glTextureColorMaskSGIS.restype = GLvoid
	gl.glTextureColorMaskSGIS.argtypes = (GLboolean, GLboolean, GLboolean, GLboolean,)
	gl.glTextureImage1DEXT = _getProcAddress('glTextureImage1DEXT')
	gl.glTextureImage1DEXT.gl = gl
	gl.glTextureImage1DEXT.__name__ = 'glTextureImage1DEXT'
	gl.glTextureImage1DEXT.errcheck = _errcheck
	gl.glTextureImage1DEXT.restype = GLvoid
	gl.glTextureImage1DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureImage2DEXT = _getProcAddress('glTextureImage2DEXT')
	gl.glTextureImage2DEXT.gl = gl
	gl.glTextureImage2DEXT.__name__ = 'glTextureImage2DEXT'
	gl.glTextureImage2DEXT.errcheck = _errcheck
	gl.glTextureImage2DEXT.restype = GLvoid
	gl.glTextureImage2DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureImage2DMultisampleCoverageNV = _getProcAddress('glTextureImage2DMultisampleCoverageNV')
	gl.glTextureImage2DMultisampleCoverageNV.gl = gl
	gl.glTextureImage2DMultisampleCoverageNV.__name__ = 'glTextureImage2DMultisampleCoverageNV'
	gl.glTextureImage2DMultisampleCoverageNV.errcheck = _errcheck
	gl.glTextureImage2DMultisampleCoverageNV.restype = GLvoid
	gl.glTextureImage2DMultisampleCoverageNV.argtypes = (GLuint, GLenum, GLsizei, GLsizei, GLint, GLsizei, GLsizei, GLboolean,)
	gl.glTextureImage2DMultisampleNV = _getProcAddress('glTextureImage2DMultisampleNV')
	gl.glTextureImage2DMultisampleNV.gl = gl
	gl.glTextureImage2DMultisampleNV.__name__ = 'glTextureImage2DMultisampleNV'
	gl.glTextureImage2DMultisampleNV.errcheck = _errcheck
	gl.glTextureImage2DMultisampleNV.restype = GLvoid
	gl.glTextureImage2DMultisampleNV.argtypes = (GLuint, GLenum, GLsizei, GLint, GLsizei, GLsizei, GLboolean,)
	gl.glTextureImage3DEXT = _getProcAddress('glTextureImage3DEXT')
	gl.glTextureImage3DEXT.gl = gl
	gl.glTextureImage3DEXT.__name__ = 'glTextureImage3DEXT'
	gl.glTextureImage3DEXT.errcheck = _errcheck
	gl.glTextureImage3DEXT.restype = GLvoid
	gl.glTextureImage3DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureImage3DMultisampleCoverageNV = _getProcAddress('glTextureImage3DMultisampleCoverageNV')
	gl.glTextureImage3DMultisampleCoverageNV.gl = gl
	gl.glTextureImage3DMultisampleCoverageNV.__name__ = 'glTextureImage3DMultisampleCoverageNV'
	gl.glTextureImage3DMultisampleCoverageNV.errcheck = _errcheck
	gl.glTextureImage3DMultisampleCoverageNV.restype = GLvoid
	gl.glTextureImage3DMultisampleCoverageNV.argtypes = (GLuint, GLenum, GLsizei, GLsizei, GLint, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTextureImage3DMultisampleNV = _getProcAddress('glTextureImage3DMultisampleNV')
	gl.glTextureImage3DMultisampleNV.gl = gl
	gl.glTextureImage3DMultisampleNV.__name__ = 'glTextureImage3DMultisampleNV'
	gl.glTextureImage3DMultisampleNV.errcheck = _errcheck
	gl.glTextureImage3DMultisampleNV.restype = GLvoid
	gl.glTextureImage3DMultisampleNV.argtypes = (GLuint, GLenum, GLsizei, GLint, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTextureLightEXT = _getProcAddress('glTextureLightEXT')
	gl.glTextureLightEXT.gl = gl
	gl.glTextureLightEXT.__name__ = 'glTextureLightEXT'
	gl.glTextureLightEXT.errcheck = _errcheck
	gl.glTextureLightEXT.restype = GLvoid
	gl.glTextureLightEXT.argtypes = (GLenum,)
	gl.glTextureMaterialEXT = _getProcAddress('glTextureMaterialEXT')
	gl.glTextureMaterialEXT.gl = gl
	gl.glTextureMaterialEXT.__name__ = 'glTextureMaterialEXT'
	gl.glTextureMaterialEXT.errcheck = _errcheck
	gl.glTextureMaterialEXT.restype = GLvoid
	gl.glTextureMaterialEXT.argtypes = (GLenum, GLenum,)
	gl.glTextureNormalEXT = _getProcAddress('glTextureNormalEXT')
	gl.glTextureNormalEXT.gl = gl
	gl.glTextureNormalEXT.__name__ = 'glTextureNormalEXT'
	gl.glTextureNormalEXT.errcheck = _errcheck
	gl.glTextureNormalEXT.restype = GLvoid
	gl.glTextureNormalEXT.argtypes = (GLenum,)
	gl.glTexturePageCommitmentEXT = _getProcAddress('glTexturePageCommitmentEXT')
	gl.glTexturePageCommitmentEXT.gl = gl
	gl.glTexturePageCommitmentEXT.__name__ = 'glTexturePageCommitmentEXT'
	gl.glTexturePageCommitmentEXT.errcheck = _errcheck
	gl.glTexturePageCommitmentEXT.restype = GLvoid
	gl.glTexturePageCommitmentEXT.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTextureParameterIiv = _getProcAddress('glTextureParameterIiv')
	gl.glTextureParameterIiv.gl = gl
	gl.glTextureParameterIiv.__name__ = 'glTextureParameterIiv'
	gl.glTextureParameterIiv.errcheck = _errcheck
	gl.glTextureParameterIiv.restype = GLvoid
	gl.glTextureParameterIiv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glTextureParameterIivEXT = _getProcAddress('glTextureParameterIivEXT')
	gl.glTextureParameterIivEXT.gl = gl
	gl.glTextureParameterIivEXT.__name__ = 'glTextureParameterIivEXT'
	gl.glTextureParameterIivEXT.errcheck = _errcheck
	gl.glTextureParameterIivEXT.restype = GLvoid
	gl.glTextureParameterIivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glTextureParameterIuiv = _getProcAddress('glTextureParameterIuiv')
	gl.glTextureParameterIuiv.gl = gl
	gl.glTextureParameterIuiv.__name__ = 'glTextureParameterIuiv'
	gl.glTextureParameterIuiv.errcheck = _errcheck
	gl.glTextureParameterIuiv.restype = GLvoid
	gl.glTextureParameterIuiv.argtypes = (GLuint, GLenum, POINTER(GLuint),)
	gl.glTextureParameterIuivEXT = _getProcAddress('glTextureParameterIuivEXT')
	gl.glTextureParameterIuivEXT.gl = gl
	gl.glTextureParameterIuivEXT.__name__ = 'glTextureParameterIuivEXT'
	gl.glTextureParameterIuivEXT.errcheck = _errcheck
	gl.glTextureParameterIuivEXT.restype = GLvoid
	gl.glTextureParameterIuivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLuint),)
	gl.glTextureParameterf = _getProcAddress('glTextureParameterf')
	gl.glTextureParameterf.gl = gl
	gl.glTextureParameterf.__name__ = 'glTextureParameterf'
	gl.glTextureParameterf.errcheck = _errcheck
	gl.glTextureParameterf.restype = GLvoid
	gl.glTextureParameterf.argtypes = (GLuint, GLenum, GLfloat,)
	gl.glTextureParameterfEXT = _getProcAddress('glTextureParameterfEXT')
	gl.glTextureParameterfEXT.gl = gl
	gl.glTextureParameterfEXT.__name__ = 'glTextureParameterfEXT'
	gl.glTextureParameterfEXT.errcheck = _errcheck
	gl.glTextureParameterfEXT.restype = GLvoid
	gl.glTextureParameterfEXT.argtypes = (GLuint, GLenum, GLenum, GLfloat,)
	gl.glTextureParameterfv = _getProcAddress('glTextureParameterfv')
	gl.glTextureParameterfv.gl = gl
	gl.glTextureParameterfv.__name__ = 'glTextureParameterfv'
	gl.glTextureParameterfv.errcheck = _errcheck
	gl.glTextureParameterfv.restype = GLvoid
	gl.glTextureParameterfv.argtypes = (GLuint, GLenum, POINTER(GLfloat),)
	gl.glTextureParameterfvEXT = _getProcAddress('glTextureParameterfvEXT')
	gl.glTextureParameterfvEXT.gl = gl
	gl.glTextureParameterfvEXT.__name__ = 'glTextureParameterfvEXT'
	gl.glTextureParameterfvEXT.errcheck = _errcheck
	gl.glTextureParameterfvEXT.restype = GLvoid
	gl.glTextureParameterfvEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLfloat),)
	gl.glTextureParameteri = _getProcAddress('glTextureParameteri')
	gl.glTextureParameteri.gl = gl
	gl.glTextureParameteri.__name__ = 'glTextureParameteri'
	gl.glTextureParameteri.errcheck = _errcheck
	gl.glTextureParameteri.restype = GLvoid
	gl.glTextureParameteri.argtypes = (GLuint, GLenum, GLint,)
	gl.glTextureParameteriEXT = _getProcAddress('glTextureParameteriEXT')
	gl.glTextureParameteriEXT.gl = gl
	gl.glTextureParameteriEXT.__name__ = 'glTextureParameteriEXT'
	gl.glTextureParameteriEXT.errcheck = _errcheck
	gl.glTextureParameteriEXT.restype = GLvoid
	gl.glTextureParameteriEXT.argtypes = (GLuint, GLenum, GLenum, GLint,)
	gl.glTextureParameteriv = _getProcAddress('glTextureParameteriv')
	gl.glTextureParameteriv.gl = gl
	gl.glTextureParameteriv.__name__ = 'glTextureParameteriv'
	gl.glTextureParameteriv.errcheck = _errcheck
	gl.glTextureParameteriv.restype = GLvoid
	gl.glTextureParameteriv.argtypes = (GLuint, GLenum, POINTER(GLint),)
	gl.glTextureParameterivEXT = _getProcAddress('glTextureParameterivEXT')
	gl.glTextureParameterivEXT.gl = gl
	gl.glTextureParameterivEXT.__name__ = 'glTextureParameterivEXT'
	gl.glTextureParameterivEXT.errcheck = _errcheck
	gl.glTextureParameterivEXT.restype = GLvoid
	gl.glTextureParameterivEXT.argtypes = (GLuint, GLenum, GLenum, POINTER(GLint),)
	gl.glTextureRangeAPPLE = _getProcAddress('glTextureRangeAPPLE')
	gl.glTextureRangeAPPLE.gl = gl
	gl.glTextureRangeAPPLE.__name__ = 'glTextureRangeAPPLE'
	gl.glTextureRangeAPPLE.errcheck = _errcheck
	gl.glTextureRangeAPPLE.restype = GLvoid
	gl.glTextureRangeAPPLE.argtypes = (GLenum, GLsizei, POINTER(GLvoid),)
	gl.glTextureRenderbufferEXT = _getProcAddress('glTextureRenderbufferEXT')
	gl.glTextureRenderbufferEXT.gl = gl
	gl.glTextureRenderbufferEXT.__name__ = 'glTextureRenderbufferEXT'
	gl.glTextureRenderbufferEXT.errcheck = _errcheck
	gl.glTextureRenderbufferEXT.restype = GLvoid
	gl.glTextureRenderbufferEXT.argtypes = (GLuint, GLenum, GLuint,)
	gl.glTextureStorage1D = _getProcAddress('glTextureStorage1D')
	gl.glTextureStorage1D.gl = gl
	gl.glTextureStorage1D.__name__ = 'glTextureStorage1D'
	gl.glTextureStorage1D.errcheck = _errcheck
	gl.glTextureStorage1D.restype = GLvoid
	gl.glTextureStorage1D.argtypes = (GLuint, GLsizei, GLenum, GLsizei,)
	gl.glTextureStorage1DEXT = _getProcAddress('glTextureStorage1DEXT')
	gl.glTextureStorage1DEXT.gl = gl
	gl.glTextureStorage1DEXT.__name__ = 'glTextureStorage1DEXT'
	gl.glTextureStorage1DEXT.errcheck = _errcheck
	gl.glTextureStorage1DEXT.restype = GLvoid
	gl.glTextureStorage1DEXT.argtypes = (GLuint, GLenum, GLsizei, GLenum, GLsizei,)
	gl.glTextureStorage2D = _getProcAddress('glTextureStorage2D')
	gl.glTextureStorage2D.gl = gl
	gl.glTextureStorage2D.__name__ = 'glTextureStorage2D'
	gl.glTextureStorage2D.errcheck = _errcheck
	gl.glTextureStorage2D.restype = GLvoid
	gl.glTextureStorage2D.argtypes = (GLuint, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glTextureStorage2DEXT = _getProcAddress('glTextureStorage2DEXT')
	gl.glTextureStorage2DEXT.gl = gl
	gl.glTextureStorage2DEXT.__name__ = 'glTextureStorage2DEXT'
	gl.glTextureStorage2DEXT.errcheck = _errcheck
	gl.glTextureStorage2DEXT.restype = GLvoid
	gl.glTextureStorage2DEXT.argtypes = (GLuint, GLenum, GLsizei, GLenum, GLsizei, GLsizei,)
	gl.glTextureStorage2DMultisample = _getProcAddress('glTextureStorage2DMultisample')
	gl.glTextureStorage2DMultisample.gl = gl
	gl.glTextureStorage2DMultisample.__name__ = 'glTextureStorage2DMultisample'
	gl.glTextureStorage2DMultisample.errcheck = _errcheck
	gl.glTextureStorage2DMultisample.restype = GLvoid
	gl.glTextureStorage2DMultisample.argtypes = (GLuint, GLsizei, GLenum, GLsizei, GLsizei, GLboolean,)
	gl.glTextureStorage2DMultisampleEXT = _getProcAddress('glTextureStorage2DMultisampleEXT')
	gl.glTextureStorage2DMultisampleEXT.gl = gl
	gl.glTextureStorage2DMultisampleEXT.__name__ = 'glTextureStorage2DMultisampleEXT'
	gl.glTextureStorage2DMultisampleEXT.errcheck = _errcheck
	gl.glTextureStorage2DMultisampleEXT.restype = GLvoid
	gl.glTextureStorage2DMultisampleEXT.argtypes = (GLuint, GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLboolean,)
	gl.glTextureStorage3D = _getProcAddress('glTextureStorage3D')
	gl.glTextureStorage3D.gl = gl
	gl.glTextureStorage3D.__name__ = 'glTextureStorage3D'
	gl.glTextureStorage3D.errcheck = _errcheck
	gl.glTextureStorage3D.restype = GLvoid
	gl.glTextureStorage3D.argtypes = (GLuint, GLsizei, GLenum, GLsizei, GLsizei, GLsizei,)
	gl.glTextureStorage3DEXT = _getProcAddress('glTextureStorage3DEXT')
	gl.glTextureStorage3DEXT.gl = gl
	gl.glTextureStorage3DEXT.__name__ = 'glTextureStorage3DEXT'
	gl.glTextureStorage3DEXT.errcheck = _errcheck
	gl.glTextureStorage3DEXT.restype = GLvoid
	gl.glTextureStorage3DEXT.argtypes = (GLuint, GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei,)
	gl.glTextureStorage3DMultisample = _getProcAddress('glTextureStorage3DMultisample')
	gl.glTextureStorage3DMultisample.gl = gl
	gl.glTextureStorage3DMultisample.__name__ = 'glTextureStorage3DMultisample'
	gl.glTextureStorage3DMultisample.errcheck = _errcheck
	gl.glTextureStorage3DMultisample.restype = GLvoid
	gl.glTextureStorage3DMultisample.argtypes = (GLuint, GLsizei, GLenum, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTextureStorage3DMultisampleEXT = _getProcAddress('glTextureStorage3DMultisampleEXT')
	gl.glTextureStorage3DMultisampleEXT.gl = gl
	gl.glTextureStorage3DMultisampleEXT.__name__ = 'glTextureStorage3DMultisampleEXT'
	gl.glTextureStorage3DMultisampleEXT.errcheck = _errcheck
	gl.glTextureStorage3DMultisampleEXT.restype = GLvoid
	gl.glTextureStorage3DMultisampleEXT.argtypes = (GLuint, GLenum, GLsizei, GLenum, GLsizei, GLsizei, GLsizei, GLboolean,)
	gl.glTextureStorageSparseAMD = _getProcAddress('glTextureStorageSparseAMD')
	gl.glTextureStorageSparseAMD.gl = gl
	gl.glTextureStorageSparseAMD.__name__ = 'glTextureStorageSparseAMD'
	gl.glTextureStorageSparseAMD.errcheck = _errcheck
	gl.glTextureStorageSparseAMD.restype = GLvoid
	gl.glTextureStorageSparseAMD.argtypes = (GLuint, GLenum, GLenum, GLsizei, GLsizei, GLsizei, GLsizei, GLbitfield,)
	gl.glTextureSubImage1D = _getProcAddress('glTextureSubImage1D')
	gl.glTextureSubImage1D.gl = gl
	gl.glTextureSubImage1D.__name__ = 'glTextureSubImage1D'
	gl.glTextureSubImage1D.errcheck = _errcheck
	gl.glTextureSubImage1D.restype = GLvoid
	gl.glTextureSubImage1D.argtypes = (GLuint, GLint, GLint, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureSubImage1DEXT = _getProcAddress('glTextureSubImage1DEXT')
	gl.glTextureSubImage1DEXT.gl = gl
	gl.glTextureSubImage1DEXT.__name__ = 'glTextureSubImage1DEXT'
	gl.glTextureSubImage1DEXT.errcheck = _errcheck
	gl.glTextureSubImage1DEXT.restype = GLvoid
	gl.glTextureSubImage1DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureSubImage2D = _getProcAddress('glTextureSubImage2D')
	gl.glTextureSubImage2D.gl = gl
	gl.glTextureSubImage2D.__name__ = 'glTextureSubImage2D'
	gl.glTextureSubImage2D.errcheck = _errcheck
	gl.glTextureSubImage2D.restype = GLvoid
	gl.glTextureSubImage2D.argtypes = (GLuint, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureSubImage2DEXT = _getProcAddress('glTextureSubImage2DEXT')
	gl.glTextureSubImage2DEXT.gl = gl
	gl.glTextureSubImage2DEXT.__name__ = 'glTextureSubImage2DEXT'
	gl.glTextureSubImage2DEXT.errcheck = _errcheck
	gl.glTextureSubImage2DEXT.restype = GLvoid
	gl.glTextureSubImage2DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureSubImage3D = _getProcAddress('glTextureSubImage3D')
	gl.glTextureSubImage3D.gl = gl
	gl.glTextureSubImage3D.__name__ = 'glTextureSubImage3D'
	gl.glTextureSubImage3D.errcheck = _errcheck
	gl.glTextureSubImage3D.restype = GLvoid
	gl.glTextureSubImage3D.argtypes = (GLuint, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureSubImage3DEXT = _getProcAddress('glTextureSubImage3DEXT')
	gl.glTextureSubImage3DEXT.gl = gl
	gl.glTextureSubImage3DEXT.__name__ = 'glTextureSubImage3DEXT'
	gl.glTextureSubImage3DEXT.errcheck = _errcheck
	gl.glTextureSubImage3DEXT.restype = GLvoid
	gl.glTextureSubImage3DEXT.argtypes = (GLuint, GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, POINTER(GLvoid),)
	gl.glTextureView = _getProcAddress('glTextureView')
	gl.glTextureView.gl = gl
	gl.glTextureView.__name__ = 'glTextureView'
	gl.glTextureView.errcheck = _errcheck
	gl.glTextureView.restype = GLvoid
	gl.glTextureView.argtypes = (GLuint, GLenum, GLuint, GLenum, GLuint, GLuint, GLuint, GLuint,)
	gl.glTextureViewEXT = _getProcAddress('glTextureViewEXT')
	gl.glTextureViewEXT.gl = gl
	gl.glTextureViewEXT.__name__ = 'glTextureViewEXT'
	gl.glTextureViewEXT.errcheck = _errcheck
	gl.glTextureViewEXT.restype = GLvoid
	gl.glTextureViewEXT.argtypes = (GLuint, GLenum, GLuint, GLenum, GLuint, GLuint, GLuint, GLuint,)
	gl.glTrackMatrixNV = _getProcAddress('glTrackMatrixNV')
	gl.glTrackMatrixNV.gl = gl
	gl.glTrackMatrixNV.__name__ = 'glTrackMatrixNV'
	gl.glTrackMatrixNV.errcheck = _errcheck
	gl.glTrackMatrixNV.restype = GLvoid
	gl.glTrackMatrixNV.argtypes = (GLenum, GLuint, GLenum, GLenum,)
	gl.glTransformFeedbackAttribsNV = _getProcAddress('glTransformFeedbackAttribsNV')
	gl.glTransformFeedbackAttribsNV.gl = gl
	gl.glTransformFeedbackAttribsNV.__name__ = 'glTransformFeedbackAttribsNV'
	gl.glTransformFeedbackAttribsNV.errcheck = _errcheck
	gl.glTransformFeedbackAttribsNV.restype = GLvoid
	gl.glTransformFeedbackAttribsNV.argtypes = (GLsizei, POINTER(GLint), GLenum,)
	gl.glTransformFeedbackBufferBase = _getProcAddress('glTransformFeedbackBufferBase')
	gl.glTransformFeedbackBufferBase.gl = gl
	gl.glTransformFeedbackBufferBase.__name__ = 'glTransformFeedbackBufferBase'
	gl.glTransformFeedbackBufferBase.errcheck = _errcheck
	gl.glTransformFeedbackBufferBase.restype = GLvoid
	gl.glTransformFeedbackBufferBase.argtypes = (GLuint, GLuint, GLuint,)
	gl.glTransformFeedbackBufferRange = _getProcAddress('glTransformFeedbackBufferRange')
	gl.glTransformFeedbackBufferRange.gl = gl
	gl.glTransformFeedbackBufferRange.__name__ = 'glTransformFeedbackBufferRange'
	gl.glTransformFeedbackBufferRange.errcheck = _errcheck
	gl.glTransformFeedbackBufferRange.restype = GLvoid
	gl.glTransformFeedbackBufferRange.argtypes = (GLuint, GLuint, GLuint, GLintptr, GLsizei,)
	gl.glTransformFeedbackStreamAttribsNV = _getProcAddress('glTransformFeedbackStreamAttribsNV')
	gl.glTransformFeedbackStreamAttribsNV.gl = gl
	gl.glTransformFeedbackStreamAttribsNV.__name__ = 'glTransformFeedbackStreamAttribsNV'
	gl.glTransformFeedbackStreamAttribsNV.errcheck = _errcheck
	gl.glTransformFeedbackStreamAttribsNV.restype = GLvoid
	gl.glTransformFeedbackStreamAttribsNV.argtypes = (GLsizei, POINTER(GLint), GLsizei, POINTER(GLint), GLenum,)
	gl.glTransformFeedbackVaryings = _getProcAddress('glTransformFeedbackVaryings')
	gl.glTransformFeedbackVaryings.gl = gl
	gl.glTransformFeedbackVaryings.__name__ = 'glTransformFeedbackVaryings'
	gl.glTransformFeedbackVaryings.errcheck = _errcheck
	gl.glTransformFeedbackVaryings.restype = GLvoid
	gl.glTransformFeedbackVaryings.argtypes = (GLuint, GLsizei, POINTER(POINTER(GLchar)), GLenum,)
	gl.glTransformFeedbackVaryingsEXT = _getProcAddress('glTransformFeedbackVaryingsEXT')
	gl.glTransformFeedbackVaryingsEXT.gl = gl
	gl.glTransformFeedbackVaryingsEXT.__name__ = 'glTransformFeedbackVaryingsEXT'
	gl.glTransformFeedbackVaryingsEXT.errcheck = _errcheck
	gl.glTransformFeedbackVaryingsEXT.restype = GLvoid
	gl.glTransformFeedbackVaryingsEXT.argtypes = (GLuint, GLsizei, POINTER(POINTER(GLchar)), GLenum,)
	gl.glTransformFeedbackVaryingsNV = _getProcAddress('glTransformFeedbackVaryingsNV')
	gl.glTransformFeedbackVaryingsNV.gl = gl
	gl.glTransformFeedbackVaryingsNV.__name__ = 'glTransformFeedbackVaryingsNV'
	gl.glTransformFeedbackVaryingsNV.errcheck = _errcheck
	gl.glTransformFeedbackVaryingsNV.restype = GLvoid
	gl.glTransformFeedbackVaryingsNV.argtypes = (GLuint, GLsizei, POINTER(GLint), GLenum,)
	gl.glTransformPathNV = _getProcAddress('glTransformPathNV')
	gl.glTransformPathNV.gl = gl
	gl.glTransformPathNV.__name__ = 'glTransformPathNV'
	gl.glTransformPathNV.errcheck = _errcheck
	gl.glTransformPathNV.restype = GLvoid
	gl.glTransformPathNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLfloat),)
	gl.glTranslated = _getProcAddress('glTranslated')
	gl.glTranslated.gl = gl
	gl.glTranslated.__name__ = 'glTranslated'
	gl.glTranslated.errcheck = _errcheck
	gl.glTranslated.restype = GLvoid
	gl.glTranslated.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glTranslatef = _getProcAddress('glTranslatef')
	gl.glTranslatef.gl = gl
	gl.glTranslatef.__name__ = 'glTranslatef'
	gl.glTranslatef.errcheck = _errcheck
	gl.glTranslatef.restype = GLvoid
	gl.glTranslatef.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glTranslatex = _getProcAddress('glTranslatex')
	gl.glTranslatex.gl = gl
	gl.glTranslatex.__name__ = 'glTranslatex'
	gl.glTranslatex.errcheck = _errcheck
	gl.glTranslatex.restype = GLvoid
	gl.glTranslatex.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glTranslatexOES = _getProcAddress('glTranslatexOES')
	gl.glTranslatexOES.gl = gl
	gl.glTranslatexOES.__name__ = 'glTranslatexOES'
	gl.glTranslatexOES.errcheck = _errcheck
	gl.glTranslatexOES.restype = GLvoid
	gl.glTranslatexOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glUniform1d = _getProcAddress('glUniform1d')
	gl.glUniform1d.gl = gl
	gl.glUniform1d.__name__ = 'glUniform1d'
	gl.glUniform1d.errcheck = _errcheck
	gl.glUniform1d.restype = GLvoid
	gl.glUniform1d.argtypes = (GLint, GLdouble,)
	gl.glUniform1dv = _getProcAddress('glUniform1dv')
	gl.glUniform1dv.gl = gl
	gl.glUniform1dv.__name__ = 'glUniform1dv'
	gl.glUniform1dv.errcheck = _errcheck
	gl.glUniform1dv.restype = GLvoid
	gl.glUniform1dv.argtypes = (GLint, GLsizei, POINTER(GLdouble),)
	gl.glUniform1f = _getProcAddress('glUniform1f')
	gl.glUniform1f.gl = gl
	gl.glUniform1f.__name__ = 'glUniform1f'
	gl.glUniform1f.errcheck = _errcheck
	gl.glUniform1f.restype = GLvoid
	gl.glUniform1f.argtypes = (GLint, GLfloat,)
	gl.glUniform1fARB = _getProcAddress('glUniform1fARB')
	gl.glUniform1fARB.gl = gl
	gl.glUniform1fARB.__name__ = 'glUniform1fARB'
	gl.glUniform1fARB.errcheck = _errcheck
	gl.glUniform1fARB.restype = GLvoid
	gl.glUniform1fARB.argtypes = (GLint, GLfloat,)
	gl.glUniform1fv = _getProcAddress('glUniform1fv')
	gl.glUniform1fv.gl = gl
	gl.glUniform1fv.__name__ = 'glUniform1fv'
	gl.glUniform1fv.errcheck = _errcheck
	gl.glUniform1fv.restype = GLvoid
	gl.glUniform1fv.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform1fvARB = _getProcAddress('glUniform1fvARB')
	gl.glUniform1fvARB.gl = gl
	gl.glUniform1fvARB.__name__ = 'glUniform1fvARB'
	gl.glUniform1fvARB.errcheck = _errcheck
	gl.glUniform1fvARB.restype = GLvoid
	gl.glUniform1fvARB.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform1i = _getProcAddress('glUniform1i')
	gl.glUniform1i.gl = gl
	gl.glUniform1i.__name__ = 'glUniform1i'
	gl.glUniform1i.errcheck = _errcheck
	gl.glUniform1i.restype = GLvoid
	gl.glUniform1i.argtypes = (GLint, GLint,)
	gl.glUniform1i64NV = _getProcAddress('glUniform1i64NV')
	gl.glUniform1i64NV.gl = gl
	gl.glUniform1i64NV.__name__ = 'glUniform1i64NV'
	gl.glUniform1i64NV.errcheck = _errcheck
	gl.glUniform1i64NV.restype = GLvoid
	gl.glUniform1i64NV.argtypes = (GLint, GLint64EXT,)
	gl.glUniform1i64vNV = _getProcAddress('glUniform1i64vNV')
	gl.glUniform1i64vNV.gl = gl
	gl.glUniform1i64vNV.__name__ = 'glUniform1i64vNV'
	gl.glUniform1i64vNV.errcheck = _errcheck
	gl.glUniform1i64vNV.restype = GLvoid
	gl.glUniform1i64vNV.argtypes = (GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glUniform1iARB = _getProcAddress('glUniform1iARB')
	gl.glUniform1iARB.gl = gl
	gl.glUniform1iARB.__name__ = 'glUniform1iARB'
	gl.glUniform1iARB.errcheck = _errcheck
	gl.glUniform1iARB.restype = GLvoid
	gl.glUniform1iARB.argtypes = (GLint, GLint,)
	gl.glUniform1iv = _getProcAddress('glUniform1iv')
	gl.glUniform1iv.gl = gl
	gl.glUniform1iv.__name__ = 'glUniform1iv'
	gl.glUniform1iv.errcheck = _errcheck
	gl.glUniform1iv.restype = GLvoid
	gl.glUniform1iv.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform1ivARB = _getProcAddress('glUniform1ivARB')
	gl.glUniform1ivARB.gl = gl
	gl.glUniform1ivARB.__name__ = 'glUniform1ivARB'
	gl.glUniform1ivARB.errcheck = _errcheck
	gl.glUniform1ivARB.restype = GLvoid
	gl.glUniform1ivARB.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform1ui = _getProcAddress('glUniform1ui')
	gl.glUniform1ui.gl = gl
	gl.glUniform1ui.__name__ = 'glUniform1ui'
	gl.glUniform1ui.errcheck = _errcheck
	gl.glUniform1ui.restype = GLvoid
	gl.glUniform1ui.argtypes = (GLint, GLuint,)
	gl.glUniform1ui64NV = _getProcAddress('glUniform1ui64NV')
	gl.glUniform1ui64NV.gl = gl
	gl.glUniform1ui64NV.__name__ = 'glUniform1ui64NV'
	gl.glUniform1ui64NV.errcheck = _errcheck
	gl.glUniform1ui64NV.restype = GLvoid
	gl.glUniform1ui64NV.argtypes = (GLint, GLuint64EXT,)
	gl.glUniform1ui64vNV = _getProcAddress('glUniform1ui64vNV')
	gl.glUniform1ui64vNV.gl = gl
	gl.glUniform1ui64vNV.__name__ = 'glUniform1ui64vNV'
	gl.glUniform1ui64vNV.errcheck = _errcheck
	gl.glUniform1ui64vNV.restype = GLvoid
	gl.glUniform1ui64vNV.argtypes = (GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glUniform1uiEXT = _getProcAddress('glUniform1uiEXT')
	gl.glUniform1uiEXT.gl = gl
	gl.glUniform1uiEXT.__name__ = 'glUniform1uiEXT'
	gl.glUniform1uiEXT.errcheck = _errcheck
	gl.glUniform1uiEXT.restype = GLvoid
	gl.glUniform1uiEXT.argtypes = (GLint, GLuint,)
	gl.glUniform1uiv = _getProcAddress('glUniform1uiv')
	gl.glUniform1uiv.gl = gl
	gl.glUniform1uiv.__name__ = 'glUniform1uiv'
	gl.glUniform1uiv.errcheck = _errcheck
	gl.glUniform1uiv.restype = GLvoid
	gl.glUniform1uiv.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform1uivEXT = _getProcAddress('glUniform1uivEXT')
	gl.glUniform1uivEXT.gl = gl
	gl.glUniform1uivEXT.__name__ = 'glUniform1uivEXT'
	gl.glUniform1uivEXT.errcheck = _errcheck
	gl.glUniform1uivEXT.restype = GLvoid
	gl.glUniform1uivEXT.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform2d = _getProcAddress('glUniform2d')
	gl.glUniform2d.gl = gl
	gl.glUniform2d.__name__ = 'glUniform2d'
	gl.glUniform2d.errcheck = _errcheck
	gl.glUniform2d.restype = GLvoid
	gl.glUniform2d.argtypes = (GLint, GLdouble, GLdouble,)
	gl.glUniform2dv = _getProcAddress('glUniform2dv')
	gl.glUniform2dv.gl = gl
	gl.glUniform2dv.__name__ = 'glUniform2dv'
	gl.glUniform2dv.errcheck = _errcheck
	gl.glUniform2dv.restype = GLvoid
	gl.glUniform2dv.argtypes = (GLint, GLsizei, POINTER(GLdouble),)
	gl.glUniform2f = _getProcAddress('glUniform2f')
	gl.glUniform2f.gl = gl
	gl.glUniform2f.__name__ = 'glUniform2f'
	gl.glUniform2f.errcheck = _errcheck
	gl.glUniform2f.restype = GLvoid
	gl.glUniform2f.argtypes = (GLint, GLfloat, GLfloat,)
	gl.glUniform2fARB = _getProcAddress('glUniform2fARB')
	gl.glUniform2fARB.gl = gl
	gl.glUniform2fARB.__name__ = 'glUniform2fARB'
	gl.glUniform2fARB.errcheck = _errcheck
	gl.glUniform2fARB.restype = GLvoid
	gl.glUniform2fARB.argtypes = (GLint, GLfloat, GLfloat,)
	gl.glUniform2fv = _getProcAddress('glUniform2fv')
	gl.glUniform2fv.gl = gl
	gl.glUniform2fv.__name__ = 'glUniform2fv'
	gl.glUniform2fv.errcheck = _errcheck
	gl.glUniform2fv.restype = GLvoid
	gl.glUniform2fv.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform2fvARB = _getProcAddress('glUniform2fvARB')
	gl.glUniform2fvARB.gl = gl
	gl.glUniform2fvARB.__name__ = 'glUniform2fvARB'
	gl.glUniform2fvARB.errcheck = _errcheck
	gl.glUniform2fvARB.restype = GLvoid
	gl.glUniform2fvARB.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform2i = _getProcAddress('glUniform2i')
	gl.glUniform2i.gl = gl
	gl.glUniform2i.__name__ = 'glUniform2i'
	gl.glUniform2i.errcheck = _errcheck
	gl.glUniform2i.restype = GLvoid
	gl.glUniform2i.argtypes = (GLint, GLint, GLint,)
	gl.glUniform2i64NV = _getProcAddress('glUniform2i64NV')
	gl.glUniform2i64NV.gl = gl
	gl.glUniform2i64NV.__name__ = 'glUniform2i64NV'
	gl.glUniform2i64NV.errcheck = _errcheck
	gl.glUniform2i64NV.restype = GLvoid
	gl.glUniform2i64NV.argtypes = (GLint, GLint64EXT, GLint64EXT,)
	gl.glUniform2i64vNV = _getProcAddress('glUniform2i64vNV')
	gl.glUniform2i64vNV.gl = gl
	gl.glUniform2i64vNV.__name__ = 'glUniform2i64vNV'
	gl.glUniform2i64vNV.errcheck = _errcheck
	gl.glUniform2i64vNV.restype = GLvoid
	gl.glUniform2i64vNV.argtypes = (GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glUniform2iARB = _getProcAddress('glUniform2iARB')
	gl.glUniform2iARB.gl = gl
	gl.glUniform2iARB.__name__ = 'glUniform2iARB'
	gl.glUniform2iARB.errcheck = _errcheck
	gl.glUniform2iARB.restype = GLvoid
	gl.glUniform2iARB.argtypes = (GLint, GLint, GLint,)
	gl.glUniform2iv = _getProcAddress('glUniform2iv')
	gl.glUniform2iv.gl = gl
	gl.glUniform2iv.__name__ = 'glUniform2iv'
	gl.glUniform2iv.errcheck = _errcheck
	gl.glUniform2iv.restype = GLvoid
	gl.glUniform2iv.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform2ivARB = _getProcAddress('glUniform2ivARB')
	gl.glUniform2ivARB.gl = gl
	gl.glUniform2ivARB.__name__ = 'glUniform2ivARB'
	gl.glUniform2ivARB.errcheck = _errcheck
	gl.glUniform2ivARB.restype = GLvoid
	gl.glUniform2ivARB.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform2ui = _getProcAddress('glUniform2ui')
	gl.glUniform2ui.gl = gl
	gl.glUniform2ui.__name__ = 'glUniform2ui'
	gl.glUniform2ui.errcheck = _errcheck
	gl.glUniform2ui.restype = GLvoid
	gl.glUniform2ui.argtypes = (GLint, GLuint, GLuint,)
	gl.glUniform2ui64NV = _getProcAddress('glUniform2ui64NV')
	gl.glUniform2ui64NV.gl = gl
	gl.glUniform2ui64NV.__name__ = 'glUniform2ui64NV'
	gl.glUniform2ui64NV.errcheck = _errcheck
	gl.glUniform2ui64NV.restype = GLvoid
	gl.glUniform2ui64NV.argtypes = (GLint, GLuint64EXT, GLuint64EXT,)
	gl.glUniform2ui64vNV = _getProcAddress('glUniform2ui64vNV')
	gl.glUniform2ui64vNV.gl = gl
	gl.glUniform2ui64vNV.__name__ = 'glUniform2ui64vNV'
	gl.glUniform2ui64vNV.errcheck = _errcheck
	gl.glUniform2ui64vNV.restype = GLvoid
	gl.glUniform2ui64vNV.argtypes = (GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glUniform2uiEXT = _getProcAddress('glUniform2uiEXT')
	gl.glUniform2uiEXT.gl = gl
	gl.glUniform2uiEXT.__name__ = 'glUniform2uiEXT'
	gl.glUniform2uiEXT.errcheck = _errcheck
	gl.glUniform2uiEXT.restype = GLvoid
	gl.glUniform2uiEXT.argtypes = (GLint, GLuint, GLuint,)
	gl.glUniform2uiv = _getProcAddress('glUniform2uiv')
	gl.glUniform2uiv.gl = gl
	gl.glUniform2uiv.__name__ = 'glUniform2uiv'
	gl.glUniform2uiv.errcheck = _errcheck
	gl.glUniform2uiv.restype = GLvoid
	gl.glUniform2uiv.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform2uivEXT = _getProcAddress('glUniform2uivEXT')
	gl.glUniform2uivEXT.gl = gl
	gl.glUniform2uivEXT.__name__ = 'glUniform2uivEXT'
	gl.glUniform2uivEXT.errcheck = _errcheck
	gl.glUniform2uivEXT.restype = GLvoid
	gl.glUniform2uivEXT.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform3d = _getProcAddress('glUniform3d')
	gl.glUniform3d.gl = gl
	gl.glUniform3d.__name__ = 'glUniform3d'
	gl.glUniform3d.errcheck = _errcheck
	gl.glUniform3d.restype = GLvoid
	gl.glUniform3d.argtypes = (GLint, GLdouble, GLdouble, GLdouble,)
	gl.glUniform3dv = _getProcAddress('glUniform3dv')
	gl.glUniform3dv.gl = gl
	gl.glUniform3dv.__name__ = 'glUniform3dv'
	gl.glUniform3dv.errcheck = _errcheck
	gl.glUniform3dv.restype = GLvoid
	gl.glUniform3dv.argtypes = (GLint, GLsizei, POINTER(GLdouble),)
	gl.glUniform3f = _getProcAddress('glUniform3f')
	gl.glUniform3f.gl = gl
	gl.glUniform3f.__name__ = 'glUniform3f'
	gl.glUniform3f.errcheck = _errcheck
	gl.glUniform3f.restype = GLvoid
	gl.glUniform3f.argtypes = (GLint, GLfloat, GLfloat, GLfloat,)
	gl.glUniform3fARB = _getProcAddress('glUniform3fARB')
	gl.glUniform3fARB.gl = gl
	gl.glUniform3fARB.__name__ = 'glUniform3fARB'
	gl.glUniform3fARB.errcheck = _errcheck
	gl.glUniform3fARB.restype = GLvoid
	gl.glUniform3fARB.argtypes = (GLint, GLfloat, GLfloat, GLfloat,)
	gl.glUniform3fv = _getProcAddress('glUniform3fv')
	gl.glUniform3fv.gl = gl
	gl.glUniform3fv.__name__ = 'glUniform3fv'
	gl.glUniform3fv.errcheck = _errcheck
	gl.glUniform3fv.restype = GLvoid
	gl.glUniform3fv.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform3fvARB = _getProcAddress('glUniform3fvARB')
	gl.glUniform3fvARB.gl = gl
	gl.glUniform3fvARB.__name__ = 'glUniform3fvARB'
	gl.glUniform3fvARB.errcheck = _errcheck
	gl.glUniform3fvARB.restype = GLvoid
	gl.glUniform3fvARB.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform3i = _getProcAddress('glUniform3i')
	gl.glUniform3i.gl = gl
	gl.glUniform3i.__name__ = 'glUniform3i'
	gl.glUniform3i.errcheck = _errcheck
	gl.glUniform3i.restype = GLvoid
	gl.glUniform3i.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glUniform3i64NV = _getProcAddress('glUniform3i64NV')
	gl.glUniform3i64NV.gl = gl
	gl.glUniform3i64NV.__name__ = 'glUniform3i64NV'
	gl.glUniform3i64NV.errcheck = _errcheck
	gl.glUniform3i64NV.restype = GLvoid
	gl.glUniform3i64NV.argtypes = (GLint, GLint64EXT, GLint64EXT, GLint64EXT,)
	gl.glUniform3i64vNV = _getProcAddress('glUniform3i64vNV')
	gl.glUniform3i64vNV.gl = gl
	gl.glUniform3i64vNV.__name__ = 'glUniform3i64vNV'
	gl.glUniform3i64vNV.errcheck = _errcheck
	gl.glUniform3i64vNV.restype = GLvoid
	gl.glUniform3i64vNV.argtypes = (GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glUniform3iARB = _getProcAddress('glUniform3iARB')
	gl.glUniform3iARB.gl = gl
	gl.glUniform3iARB.__name__ = 'glUniform3iARB'
	gl.glUniform3iARB.errcheck = _errcheck
	gl.glUniform3iARB.restype = GLvoid
	gl.glUniform3iARB.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glUniform3iv = _getProcAddress('glUniform3iv')
	gl.glUniform3iv.gl = gl
	gl.glUniform3iv.__name__ = 'glUniform3iv'
	gl.glUniform3iv.errcheck = _errcheck
	gl.glUniform3iv.restype = GLvoid
	gl.glUniform3iv.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform3ivARB = _getProcAddress('glUniform3ivARB')
	gl.glUniform3ivARB.gl = gl
	gl.glUniform3ivARB.__name__ = 'glUniform3ivARB'
	gl.glUniform3ivARB.errcheck = _errcheck
	gl.glUniform3ivARB.restype = GLvoid
	gl.glUniform3ivARB.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform3ui = _getProcAddress('glUniform3ui')
	gl.glUniform3ui.gl = gl
	gl.glUniform3ui.__name__ = 'glUniform3ui'
	gl.glUniform3ui.errcheck = _errcheck
	gl.glUniform3ui.restype = GLvoid
	gl.glUniform3ui.argtypes = (GLint, GLuint, GLuint, GLuint,)
	gl.glUniform3ui64NV = _getProcAddress('glUniform3ui64NV')
	gl.glUniform3ui64NV.gl = gl
	gl.glUniform3ui64NV.__name__ = 'glUniform3ui64NV'
	gl.glUniform3ui64NV.errcheck = _errcheck
	gl.glUniform3ui64NV.restype = GLvoid
	gl.glUniform3ui64NV.argtypes = (GLint, GLuint64EXT, GLuint64EXT, GLuint64EXT,)
	gl.glUniform3ui64vNV = _getProcAddress('glUniform3ui64vNV')
	gl.glUniform3ui64vNV.gl = gl
	gl.glUniform3ui64vNV.__name__ = 'glUniform3ui64vNV'
	gl.glUniform3ui64vNV.errcheck = _errcheck
	gl.glUniform3ui64vNV.restype = GLvoid
	gl.glUniform3ui64vNV.argtypes = (GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glUniform3uiEXT = _getProcAddress('glUniform3uiEXT')
	gl.glUniform3uiEXT.gl = gl
	gl.glUniform3uiEXT.__name__ = 'glUniform3uiEXT'
	gl.glUniform3uiEXT.errcheck = _errcheck
	gl.glUniform3uiEXT.restype = GLvoid
	gl.glUniform3uiEXT.argtypes = (GLint, GLuint, GLuint, GLuint,)
	gl.glUniform3uiv = _getProcAddress('glUniform3uiv')
	gl.glUniform3uiv.gl = gl
	gl.glUniform3uiv.__name__ = 'glUniform3uiv'
	gl.glUniform3uiv.errcheck = _errcheck
	gl.glUniform3uiv.restype = GLvoid
	gl.glUniform3uiv.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform3uivEXT = _getProcAddress('glUniform3uivEXT')
	gl.glUniform3uivEXT.gl = gl
	gl.glUniform3uivEXT.__name__ = 'glUniform3uivEXT'
	gl.glUniform3uivEXT.errcheck = _errcheck
	gl.glUniform3uivEXT.restype = GLvoid
	gl.glUniform3uivEXT.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform4d = _getProcAddress('glUniform4d')
	gl.glUniform4d.gl = gl
	gl.glUniform4d.__name__ = 'glUniform4d'
	gl.glUniform4d.errcheck = _errcheck
	gl.glUniform4d.restype = GLvoid
	gl.glUniform4d.argtypes = (GLint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glUniform4dv = _getProcAddress('glUniform4dv')
	gl.glUniform4dv.gl = gl
	gl.glUniform4dv.__name__ = 'glUniform4dv'
	gl.glUniform4dv.errcheck = _errcheck
	gl.glUniform4dv.restype = GLvoid
	gl.glUniform4dv.argtypes = (GLint, GLsizei, POINTER(GLdouble),)
	gl.glUniform4f = _getProcAddress('glUniform4f')
	gl.glUniform4f.gl = gl
	gl.glUniform4f.__name__ = 'glUniform4f'
	gl.glUniform4f.errcheck = _errcheck
	gl.glUniform4f.restype = GLvoid
	gl.glUniform4f.argtypes = (GLint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glUniform4fARB = _getProcAddress('glUniform4fARB')
	gl.glUniform4fARB.gl = gl
	gl.glUniform4fARB.__name__ = 'glUniform4fARB'
	gl.glUniform4fARB.errcheck = _errcheck
	gl.glUniform4fARB.restype = GLvoid
	gl.glUniform4fARB.argtypes = (GLint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glUniform4fv = _getProcAddress('glUniform4fv')
	gl.glUniform4fv.gl = gl
	gl.glUniform4fv.__name__ = 'glUniform4fv'
	gl.glUniform4fv.errcheck = _errcheck
	gl.glUniform4fv.restype = GLvoid
	gl.glUniform4fv.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform4fvARB = _getProcAddress('glUniform4fvARB')
	gl.glUniform4fvARB.gl = gl
	gl.glUniform4fvARB.__name__ = 'glUniform4fvARB'
	gl.glUniform4fvARB.errcheck = _errcheck
	gl.glUniform4fvARB.restype = GLvoid
	gl.glUniform4fvARB.argtypes = (GLint, GLsizei, POINTER(GLfloat),)
	gl.glUniform4i = _getProcAddress('glUniform4i')
	gl.glUniform4i.gl = gl
	gl.glUniform4i.__name__ = 'glUniform4i'
	gl.glUniform4i.errcheck = _errcheck
	gl.glUniform4i.restype = GLvoid
	gl.glUniform4i.argtypes = (GLint, GLint, GLint, GLint, GLint,)
	gl.glUniform4i64NV = _getProcAddress('glUniform4i64NV')
	gl.glUniform4i64NV.gl = gl
	gl.glUniform4i64NV.__name__ = 'glUniform4i64NV'
	gl.glUniform4i64NV.errcheck = _errcheck
	gl.glUniform4i64NV.restype = GLvoid
	gl.glUniform4i64NV.argtypes = (GLint, GLint64EXT, GLint64EXT, GLint64EXT, GLint64EXT,)
	gl.glUniform4i64vNV = _getProcAddress('glUniform4i64vNV')
	gl.glUniform4i64vNV.gl = gl
	gl.glUniform4i64vNV.__name__ = 'glUniform4i64vNV'
	gl.glUniform4i64vNV.errcheck = _errcheck
	gl.glUniform4i64vNV.restype = GLvoid
	gl.glUniform4i64vNV.argtypes = (GLint, GLsizei, POINTER(GLint64EXT),)
	gl.glUniform4iARB = _getProcAddress('glUniform4iARB')
	gl.glUniform4iARB.gl = gl
	gl.glUniform4iARB.__name__ = 'glUniform4iARB'
	gl.glUniform4iARB.errcheck = _errcheck
	gl.glUniform4iARB.restype = GLvoid
	gl.glUniform4iARB.argtypes = (GLint, GLint, GLint, GLint, GLint,)
	gl.glUniform4iv = _getProcAddress('glUniform4iv')
	gl.glUniform4iv.gl = gl
	gl.glUniform4iv.__name__ = 'glUniform4iv'
	gl.glUniform4iv.errcheck = _errcheck
	gl.glUniform4iv.restype = GLvoid
	gl.glUniform4iv.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform4ivARB = _getProcAddress('glUniform4ivARB')
	gl.glUniform4ivARB.gl = gl
	gl.glUniform4ivARB.__name__ = 'glUniform4ivARB'
	gl.glUniform4ivARB.errcheck = _errcheck
	gl.glUniform4ivARB.restype = GLvoid
	gl.glUniform4ivARB.argtypes = (GLint, GLsizei, POINTER(GLint),)
	gl.glUniform4ui = _getProcAddress('glUniform4ui')
	gl.glUniform4ui.gl = gl
	gl.glUniform4ui.__name__ = 'glUniform4ui'
	gl.glUniform4ui.errcheck = _errcheck
	gl.glUniform4ui.restype = GLvoid
	gl.glUniform4ui.argtypes = (GLint, GLuint, GLuint, GLuint, GLuint,)
	gl.glUniform4ui64NV = _getProcAddress('glUniform4ui64NV')
	gl.glUniform4ui64NV.gl = gl
	gl.glUniform4ui64NV.__name__ = 'glUniform4ui64NV'
	gl.glUniform4ui64NV.errcheck = _errcheck
	gl.glUniform4ui64NV.restype = GLvoid
	gl.glUniform4ui64NV.argtypes = (GLint, GLuint64EXT, GLuint64EXT, GLuint64EXT, GLuint64EXT,)
	gl.glUniform4ui64vNV = _getProcAddress('glUniform4ui64vNV')
	gl.glUniform4ui64vNV.gl = gl
	gl.glUniform4ui64vNV.__name__ = 'glUniform4ui64vNV'
	gl.glUniform4ui64vNV.errcheck = _errcheck
	gl.glUniform4ui64vNV.restype = GLvoid
	gl.glUniform4ui64vNV.argtypes = (GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glUniform4uiEXT = _getProcAddress('glUniform4uiEXT')
	gl.glUniform4uiEXT.gl = gl
	gl.glUniform4uiEXT.__name__ = 'glUniform4uiEXT'
	gl.glUniform4uiEXT.errcheck = _errcheck
	gl.glUniform4uiEXT.restype = GLvoid
	gl.glUniform4uiEXT.argtypes = (GLint, GLuint, GLuint, GLuint, GLuint,)
	gl.glUniform4uiv = _getProcAddress('glUniform4uiv')
	gl.glUniform4uiv.gl = gl
	gl.glUniform4uiv.__name__ = 'glUniform4uiv'
	gl.glUniform4uiv.errcheck = _errcheck
	gl.glUniform4uiv.restype = GLvoid
	gl.glUniform4uiv.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniform4uivEXT = _getProcAddress('glUniform4uivEXT')
	gl.glUniform4uivEXT.gl = gl
	gl.glUniform4uivEXT.__name__ = 'glUniform4uivEXT'
	gl.glUniform4uivEXT.errcheck = _errcheck
	gl.glUniform4uivEXT.restype = GLvoid
	gl.glUniform4uivEXT.argtypes = (GLint, GLsizei, POINTER(GLuint),)
	gl.glUniformBlockBinding = _getProcAddress('glUniformBlockBinding')
	gl.glUniformBlockBinding.gl = gl
	gl.glUniformBlockBinding.__name__ = 'glUniformBlockBinding'
	gl.glUniformBlockBinding.errcheck = _errcheck
	gl.glUniformBlockBinding.restype = GLvoid
	gl.glUniformBlockBinding.argtypes = (GLuint, GLuint, GLuint,)
	gl.glUniformBufferEXT = _getProcAddress('glUniformBufferEXT')
	gl.glUniformBufferEXT.gl = gl
	gl.glUniformBufferEXT.__name__ = 'glUniformBufferEXT'
	gl.glUniformBufferEXT.errcheck = _errcheck
	gl.glUniformBufferEXT.restype = GLvoid
	gl.glUniformBufferEXT.argtypes = (GLuint, GLint, GLuint,)
	gl.glUniformHandleui64ARB = _getProcAddress('glUniformHandleui64ARB')
	gl.glUniformHandleui64ARB.gl = gl
	gl.glUniformHandleui64ARB.__name__ = 'glUniformHandleui64ARB'
	gl.glUniformHandleui64ARB.errcheck = _errcheck
	gl.glUniformHandleui64ARB.restype = GLvoid
	gl.glUniformHandleui64ARB.argtypes = (GLint, GLuint64,)
	gl.glUniformHandleui64NV = _getProcAddress('glUniformHandleui64NV')
	gl.glUniformHandleui64NV.gl = gl
	gl.glUniformHandleui64NV.__name__ = 'glUniformHandleui64NV'
	gl.glUniformHandleui64NV.errcheck = _errcheck
	gl.glUniformHandleui64NV.restype = GLvoid
	gl.glUniformHandleui64NV.argtypes = (GLint, GLuint64,)
	gl.glUniformHandleui64vARB = _getProcAddress('glUniformHandleui64vARB')
	gl.glUniformHandleui64vARB.gl = gl
	gl.glUniformHandleui64vARB.__name__ = 'glUniformHandleui64vARB'
	gl.glUniformHandleui64vARB.errcheck = _errcheck
	gl.glUniformHandleui64vARB.restype = GLvoid
	gl.glUniformHandleui64vARB.argtypes = (GLint, GLsizei, POINTER(GLuint64),)
	gl.glUniformHandleui64vNV = _getProcAddress('glUniformHandleui64vNV')
	gl.glUniformHandleui64vNV.gl = gl
	gl.glUniformHandleui64vNV.__name__ = 'glUniformHandleui64vNV'
	gl.glUniformHandleui64vNV.errcheck = _errcheck
	gl.glUniformHandleui64vNV.restype = GLvoid
	gl.glUniformHandleui64vNV.argtypes = (GLint, GLsizei, POINTER(GLuint64),)
	gl.glUniformMatrix2dv = _getProcAddress('glUniformMatrix2dv')
	gl.glUniformMatrix2dv.gl = gl
	gl.glUniformMatrix2dv.__name__ = 'glUniformMatrix2dv'
	gl.glUniformMatrix2dv.errcheck = _errcheck
	gl.glUniformMatrix2dv.restype = GLvoid
	gl.glUniformMatrix2dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix2fv = _getProcAddress('glUniformMatrix2fv')
	gl.glUniformMatrix2fv.gl = gl
	gl.glUniformMatrix2fv.__name__ = 'glUniformMatrix2fv'
	gl.glUniformMatrix2fv.errcheck = _errcheck
	gl.glUniformMatrix2fv.restype = GLvoid
	gl.glUniformMatrix2fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix2fvARB = _getProcAddress('glUniformMatrix2fvARB')
	gl.glUniformMatrix2fvARB.gl = gl
	gl.glUniformMatrix2fvARB.__name__ = 'glUniformMatrix2fvARB'
	gl.glUniformMatrix2fvARB.errcheck = _errcheck
	gl.glUniformMatrix2fvARB.restype = GLvoid
	gl.glUniformMatrix2fvARB.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix2x3dv = _getProcAddress('glUniformMatrix2x3dv')
	gl.glUniformMatrix2x3dv.gl = gl
	gl.glUniformMatrix2x3dv.__name__ = 'glUniformMatrix2x3dv'
	gl.glUniformMatrix2x3dv.errcheck = _errcheck
	gl.glUniformMatrix2x3dv.restype = GLvoid
	gl.glUniformMatrix2x3dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix2x3fv = _getProcAddress('glUniformMatrix2x3fv')
	gl.glUniformMatrix2x3fv.gl = gl
	gl.glUniformMatrix2x3fv.__name__ = 'glUniformMatrix2x3fv'
	gl.glUniformMatrix2x3fv.errcheck = _errcheck
	gl.glUniformMatrix2x3fv.restype = GLvoid
	gl.glUniformMatrix2x3fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix2x3fvNV = _getProcAddress('glUniformMatrix2x3fvNV')
	gl.glUniformMatrix2x3fvNV.gl = gl
	gl.glUniformMatrix2x3fvNV.__name__ = 'glUniformMatrix2x3fvNV'
	gl.glUniformMatrix2x3fvNV.errcheck = _errcheck
	gl.glUniformMatrix2x3fvNV.restype = GLvoid
	gl.glUniformMatrix2x3fvNV.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix2x4dv = _getProcAddress('glUniformMatrix2x4dv')
	gl.glUniformMatrix2x4dv.gl = gl
	gl.glUniformMatrix2x4dv.__name__ = 'glUniformMatrix2x4dv'
	gl.glUniformMatrix2x4dv.errcheck = _errcheck
	gl.glUniformMatrix2x4dv.restype = GLvoid
	gl.glUniformMatrix2x4dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix2x4fv = _getProcAddress('glUniformMatrix2x4fv')
	gl.glUniformMatrix2x4fv.gl = gl
	gl.glUniformMatrix2x4fv.__name__ = 'glUniformMatrix2x4fv'
	gl.glUniformMatrix2x4fv.errcheck = _errcheck
	gl.glUniformMatrix2x4fv.restype = GLvoid
	gl.glUniformMatrix2x4fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix2x4fvNV = _getProcAddress('glUniformMatrix2x4fvNV')
	gl.glUniformMatrix2x4fvNV.gl = gl
	gl.glUniformMatrix2x4fvNV.__name__ = 'glUniformMatrix2x4fvNV'
	gl.glUniformMatrix2x4fvNV.errcheck = _errcheck
	gl.glUniformMatrix2x4fvNV.restype = GLvoid
	gl.glUniformMatrix2x4fvNV.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix3dv = _getProcAddress('glUniformMatrix3dv')
	gl.glUniformMatrix3dv.gl = gl
	gl.glUniformMatrix3dv.__name__ = 'glUniformMatrix3dv'
	gl.glUniformMatrix3dv.errcheck = _errcheck
	gl.glUniformMatrix3dv.restype = GLvoid
	gl.glUniformMatrix3dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix3fv = _getProcAddress('glUniformMatrix3fv')
	gl.glUniformMatrix3fv.gl = gl
	gl.glUniformMatrix3fv.__name__ = 'glUniformMatrix3fv'
	gl.glUniformMatrix3fv.errcheck = _errcheck
	gl.glUniformMatrix3fv.restype = GLvoid
	gl.glUniformMatrix3fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix3fvARB = _getProcAddress('glUniformMatrix3fvARB')
	gl.glUniformMatrix3fvARB.gl = gl
	gl.glUniformMatrix3fvARB.__name__ = 'glUniformMatrix3fvARB'
	gl.glUniformMatrix3fvARB.errcheck = _errcheck
	gl.glUniformMatrix3fvARB.restype = GLvoid
	gl.glUniformMatrix3fvARB.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix3x2dv = _getProcAddress('glUniformMatrix3x2dv')
	gl.glUniformMatrix3x2dv.gl = gl
	gl.glUniformMatrix3x2dv.__name__ = 'glUniformMatrix3x2dv'
	gl.glUniformMatrix3x2dv.errcheck = _errcheck
	gl.glUniformMatrix3x2dv.restype = GLvoid
	gl.glUniformMatrix3x2dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix3x2fv = _getProcAddress('glUniformMatrix3x2fv')
	gl.glUniformMatrix3x2fv.gl = gl
	gl.glUniformMatrix3x2fv.__name__ = 'glUniformMatrix3x2fv'
	gl.glUniformMatrix3x2fv.errcheck = _errcheck
	gl.glUniformMatrix3x2fv.restype = GLvoid
	gl.glUniformMatrix3x2fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix3x2fvNV = _getProcAddress('glUniformMatrix3x2fvNV')
	gl.glUniformMatrix3x2fvNV.gl = gl
	gl.glUniformMatrix3x2fvNV.__name__ = 'glUniformMatrix3x2fvNV'
	gl.glUniformMatrix3x2fvNV.errcheck = _errcheck
	gl.glUniformMatrix3x2fvNV.restype = GLvoid
	gl.glUniformMatrix3x2fvNV.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix3x4dv = _getProcAddress('glUniformMatrix3x4dv')
	gl.glUniformMatrix3x4dv.gl = gl
	gl.glUniformMatrix3x4dv.__name__ = 'glUniformMatrix3x4dv'
	gl.glUniformMatrix3x4dv.errcheck = _errcheck
	gl.glUniformMatrix3x4dv.restype = GLvoid
	gl.glUniformMatrix3x4dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix3x4fv = _getProcAddress('glUniformMatrix3x4fv')
	gl.glUniformMatrix3x4fv.gl = gl
	gl.glUniformMatrix3x4fv.__name__ = 'glUniformMatrix3x4fv'
	gl.glUniformMatrix3x4fv.errcheck = _errcheck
	gl.glUniformMatrix3x4fv.restype = GLvoid
	gl.glUniformMatrix3x4fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix3x4fvNV = _getProcAddress('glUniformMatrix3x4fvNV')
	gl.glUniformMatrix3x4fvNV.gl = gl
	gl.glUniformMatrix3x4fvNV.__name__ = 'glUniformMatrix3x4fvNV'
	gl.glUniformMatrix3x4fvNV.errcheck = _errcheck
	gl.glUniformMatrix3x4fvNV.restype = GLvoid
	gl.glUniformMatrix3x4fvNV.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix4dv = _getProcAddress('glUniformMatrix4dv')
	gl.glUniformMatrix4dv.gl = gl
	gl.glUniformMatrix4dv.__name__ = 'glUniformMatrix4dv'
	gl.glUniformMatrix4dv.errcheck = _errcheck
	gl.glUniformMatrix4dv.restype = GLvoid
	gl.glUniformMatrix4dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix4fv = _getProcAddress('glUniformMatrix4fv')
	gl.glUniformMatrix4fv.gl = gl
	gl.glUniformMatrix4fv.__name__ = 'glUniformMatrix4fv'
	gl.glUniformMatrix4fv.errcheck = _errcheck
	gl.glUniformMatrix4fv.restype = GLvoid
	gl.glUniformMatrix4fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix4fvARB = _getProcAddress('glUniformMatrix4fvARB')
	gl.glUniformMatrix4fvARB.gl = gl
	gl.glUniformMatrix4fvARB.__name__ = 'glUniformMatrix4fvARB'
	gl.glUniformMatrix4fvARB.errcheck = _errcheck
	gl.glUniformMatrix4fvARB.restype = GLvoid
	gl.glUniformMatrix4fvARB.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix4x2dv = _getProcAddress('glUniformMatrix4x2dv')
	gl.glUniformMatrix4x2dv.gl = gl
	gl.glUniformMatrix4x2dv.__name__ = 'glUniformMatrix4x2dv'
	gl.glUniformMatrix4x2dv.errcheck = _errcheck
	gl.glUniformMatrix4x2dv.restype = GLvoid
	gl.glUniformMatrix4x2dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix4x2fv = _getProcAddress('glUniformMatrix4x2fv')
	gl.glUniformMatrix4x2fv.gl = gl
	gl.glUniformMatrix4x2fv.__name__ = 'glUniformMatrix4x2fv'
	gl.glUniformMatrix4x2fv.errcheck = _errcheck
	gl.glUniformMatrix4x2fv.restype = GLvoid
	gl.glUniformMatrix4x2fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix4x2fvNV = _getProcAddress('glUniformMatrix4x2fvNV')
	gl.glUniformMatrix4x2fvNV.gl = gl
	gl.glUniformMatrix4x2fvNV.__name__ = 'glUniformMatrix4x2fvNV'
	gl.glUniformMatrix4x2fvNV.errcheck = _errcheck
	gl.glUniformMatrix4x2fvNV.restype = GLvoid
	gl.glUniformMatrix4x2fvNV.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix4x3dv = _getProcAddress('glUniformMatrix4x3dv')
	gl.glUniformMatrix4x3dv.gl = gl
	gl.glUniformMatrix4x3dv.__name__ = 'glUniformMatrix4x3dv'
	gl.glUniformMatrix4x3dv.errcheck = _errcheck
	gl.glUniformMatrix4x3dv.restype = GLvoid
	gl.glUniformMatrix4x3dv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLdouble),)
	gl.glUniformMatrix4x3fv = _getProcAddress('glUniformMatrix4x3fv')
	gl.glUniformMatrix4x3fv.gl = gl
	gl.glUniformMatrix4x3fv.__name__ = 'glUniformMatrix4x3fv'
	gl.glUniformMatrix4x3fv.errcheck = _errcheck
	gl.glUniformMatrix4x3fv.restype = GLvoid
	gl.glUniformMatrix4x3fv.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformMatrix4x3fvNV = _getProcAddress('glUniformMatrix4x3fvNV')
	gl.glUniformMatrix4x3fvNV.gl = gl
	gl.glUniformMatrix4x3fvNV.__name__ = 'glUniformMatrix4x3fvNV'
	gl.glUniformMatrix4x3fvNV.errcheck = _errcheck
	gl.glUniformMatrix4x3fvNV.restype = GLvoid
	gl.glUniformMatrix4x3fvNV.argtypes = (GLint, GLsizei, GLboolean, POINTER(GLfloat),)
	gl.glUniformSubroutinesuiv = _getProcAddress('glUniformSubroutinesuiv')
	gl.glUniformSubroutinesuiv.gl = gl
	gl.glUniformSubroutinesuiv.__name__ = 'glUniformSubroutinesuiv'
	gl.glUniformSubroutinesuiv.errcheck = _errcheck
	gl.glUniformSubroutinesuiv.restype = GLvoid
	gl.glUniformSubroutinesuiv.argtypes = (GLenum, GLsizei, POINTER(GLuint),)
	gl.glUniformui64NV = _getProcAddress('glUniformui64NV')
	gl.glUniformui64NV.gl = gl
	gl.glUniformui64NV.__name__ = 'glUniformui64NV'
	gl.glUniformui64NV.errcheck = _errcheck
	gl.glUniformui64NV.restype = GLvoid
	gl.glUniformui64NV.argtypes = (GLint, GLuint64EXT,)
	gl.glUniformui64vNV = _getProcAddress('glUniformui64vNV')
	gl.glUniformui64vNV.gl = gl
	gl.glUniformui64vNV.__name__ = 'glUniformui64vNV'
	gl.glUniformui64vNV.errcheck = _errcheck
	gl.glUniformui64vNV.restype = GLvoid
	gl.glUniformui64vNV.argtypes = (GLint, GLsizei, POINTER(GLuint64EXT),)
	gl.glUnlockArraysEXT = _getProcAddress('glUnlockArraysEXT')
	gl.glUnlockArraysEXT.gl = gl
	gl.glUnlockArraysEXT.__name__ = 'glUnlockArraysEXT'
	gl.glUnlockArraysEXT.errcheck = _errcheck
	gl.glUnlockArraysEXT.restype = GLvoid
	gl.glUnlockArraysEXT.argtypes = ()
	gl.glUnmapBuffer = _getProcAddress('glUnmapBuffer')
	gl.glUnmapBuffer.gl = gl
	gl.glUnmapBuffer.__name__ = 'glUnmapBuffer'
	gl.glUnmapBuffer.errcheck = _errcheck
	gl.glUnmapBuffer.restype = GLboolean
	gl.glUnmapBuffer.argtypes = (GLenum,)
	gl.glUnmapBufferARB = _getProcAddress('glUnmapBufferARB')
	gl.glUnmapBufferARB.gl = gl
	gl.glUnmapBufferARB.__name__ = 'glUnmapBufferARB'
	gl.glUnmapBufferARB.errcheck = _errcheck
	gl.glUnmapBufferARB.restype = GLboolean
	gl.glUnmapBufferARB.argtypes = (GLenum,)
	gl.glUnmapBufferOES = _getProcAddress('glUnmapBufferOES')
	gl.glUnmapBufferOES.gl = gl
	gl.glUnmapBufferOES.__name__ = 'glUnmapBufferOES'
	gl.glUnmapBufferOES.errcheck = _errcheck
	gl.glUnmapBufferOES.restype = GLboolean
	gl.glUnmapBufferOES.argtypes = (GLenum,)
	gl.glUnmapNamedBuffer = _getProcAddress('glUnmapNamedBuffer')
	gl.glUnmapNamedBuffer.gl = gl
	gl.glUnmapNamedBuffer.__name__ = 'glUnmapNamedBuffer'
	gl.glUnmapNamedBuffer.errcheck = _errcheck
	gl.glUnmapNamedBuffer.restype = GLboolean
	gl.glUnmapNamedBuffer.argtypes = (GLuint,)
	gl.glUnmapNamedBufferEXT = _getProcAddress('glUnmapNamedBufferEXT')
	gl.glUnmapNamedBufferEXT.gl = gl
	gl.glUnmapNamedBufferEXT.__name__ = 'glUnmapNamedBufferEXT'
	gl.glUnmapNamedBufferEXT.errcheck = _errcheck
	gl.glUnmapNamedBufferEXT.restype = GLboolean
	gl.glUnmapNamedBufferEXT.argtypes = (GLuint,)
	gl.glUnmapObjectBufferATI = _getProcAddress('glUnmapObjectBufferATI')
	gl.glUnmapObjectBufferATI.gl = gl
	gl.glUnmapObjectBufferATI.__name__ = 'glUnmapObjectBufferATI'
	gl.glUnmapObjectBufferATI.errcheck = _errcheck
	gl.glUnmapObjectBufferATI.restype = GLvoid
	gl.glUnmapObjectBufferATI.argtypes = (GLuint,)
	gl.glUnmapTexture2DINTEL = _getProcAddress('glUnmapTexture2DINTEL')
	gl.glUnmapTexture2DINTEL.gl = gl
	gl.glUnmapTexture2DINTEL.__name__ = 'glUnmapTexture2DINTEL'
	gl.glUnmapTexture2DINTEL.errcheck = _errcheck
	gl.glUnmapTexture2DINTEL.restype = GLvoid
	gl.glUnmapTexture2DINTEL.argtypes = (GLuint, GLint,)
	gl.glUpdateObjectBufferATI = _getProcAddress('glUpdateObjectBufferATI')
	gl.glUpdateObjectBufferATI.gl = gl
	gl.glUpdateObjectBufferATI.__name__ = 'glUpdateObjectBufferATI'
	gl.glUpdateObjectBufferATI.errcheck = _errcheck
	gl.glUpdateObjectBufferATI.restype = GLvoid
	gl.glUpdateObjectBufferATI.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLvoid), GLenum,)
	gl.glUseProgram = _getProcAddress('glUseProgram')
	gl.glUseProgram.gl = gl
	gl.glUseProgram.__name__ = 'glUseProgram'
	gl.glUseProgram.errcheck = _errcheck
	gl.glUseProgram.restype = GLvoid
	gl.glUseProgram.argtypes = (GLuint,)
	gl.glUseProgramObjectARB = _getProcAddress('glUseProgramObjectARB')
	gl.glUseProgramObjectARB.gl = gl
	gl.glUseProgramObjectARB.__name__ = 'glUseProgramObjectARB'
	gl.glUseProgramObjectARB.errcheck = _errcheck
	gl.glUseProgramObjectARB.restype = GLvoid
	gl.glUseProgramObjectARB.argtypes = (GLhandleARB,)
	gl.glUseProgramStages = _getProcAddress('glUseProgramStages')
	gl.glUseProgramStages.gl = gl
	gl.glUseProgramStages.__name__ = 'glUseProgramStages'
	gl.glUseProgramStages.errcheck = _errcheck
	gl.glUseProgramStages.restype = GLvoid
	gl.glUseProgramStages.argtypes = (GLuint, GLbitfield, GLuint,)
	gl.glUseProgramStagesEXT = _getProcAddress('glUseProgramStagesEXT')
	gl.glUseProgramStagesEXT.gl = gl
	gl.glUseProgramStagesEXT.__name__ = 'glUseProgramStagesEXT'
	gl.glUseProgramStagesEXT.errcheck = _errcheck
	gl.glUseProgramStagesEXT.restype = GLvoid
	gl.glUseProgramStagesEXT.argtypes = (GLuint, GLbitfield, GLuint,)
	gl.glUseShaderProgramEXT = _getProcAddress('glUseShaderProgramEXT')
	gl.glUseShaderProgramEXT.gl = gl
	gl.glUseShaderProgramEXT.__name__ = 'glUseShaderProgramEXT'
	gl.glUseShaderProgramEXT.errcheck = _errcheck
	gl.glUseShaderProgramEXT.restype = GLvoid
	gl.glUseShaderProgramEXT.argtypes = (GLenum, GLuint,)
	gl.glVDPAUFiniNV = _getProcAddress('glVDPAUFiniNV')
	gl.glVDPAUFiniNV.gl = gl
	gl.glVDPAUFiniNV.__name__ = 'glVDPAUFiniNV'
	gl.glVDPAUFiniNV.errcheck = _errcheck
	gl.glVDPAUFiniNV.restype = GLvoid
	gl.glVDPAUFiniNV.argtypes = ()
	gl.glVDPAUGetSurfaceivNV = _getProcAddress('glVDPAUGetSurfaceivNV')
	gl.glVDPAUGetSurfaceivNV.gl = gl
	gl.glVDPAUGetSurfaceivNV.__name__ = 'glVDPAUGetSurfaceivNV'
	gl.glVDPAUGetSurfaceivNV.errcheck = _errcheck
	gl.glVDPAUGetSurfaceivNV.restype = GLvoid
	gl.glVDPAUGetSurfaceivNV.argtypes = (GLvdpauSurfaceNV, GLenum, GLsizei, POINTER(GLsizei), POINTER(GLint),)
	gl.glVDPAUInitNV = _getProcAddress('glVDPAUInitNV')
	gl.glVDPAUInitNV.gl = gl
	gl.glVDPAUInitNV.__name__ = 'glVDPAUInitNV'
	gl.glVDPAUInitNV.errcheck = _errcheck
	gl.glVDPAUInitNV.restype = GLvoid
	gl.glVDPAUInitNV.argtypes = (POINTER(GLvoid), POINTER(GLvoid),)
	gl.glVDPAUIsSurfaceNV = _getProcAddress('glVDPAUIsSurfaceNV')
	gl.glVDPAUIsSurfaceNV.gl = gl
	gl.glVDPAUIsSurfaceNV.__name__ = 'glVDPAUIsSurfaceNV'
	gl.glVDPAUIsSurfaceNV.errcheck = _errcheck
	gl.glVDPAUIsSurfaceNV.restype = GLboolean
	gl.glVDPAUIsSurfaceNV.argtypes = (GLvdpauSurfaceNV,)
	gl.glVDPAUMapSurfacesNV = _getProcAddress('glVDPAUMapSurfacesNV')
	gl.glVDPAUMapSurfacesNV.gl = gl
	gl.glVDPAUMapSurfacesNV.__name__ = 'glVDPAUMapSurfacesNV'
	gl.glVDPAUMapSurfacesNV.errcheck = _errcheck
	gl.glVDPAUMapSurfacesNV.restype = GLvoid
	gl.glVDPAUMapSurfacesNV.argtypes = (GLsizei, POINTER(GLvdpauSurfaceNV),)
	gl.glVDPAURegisterOutputSurfaceNV = _getProcAddress('glVDPAURegisterOutputSurfaceNV')
	gl.glVDPAURegisterOutputSurfaceNV.gl = gl
	gl.glVDPAURegisterOutputSurfaceNV.__name__ = 'glVDPAURegisterOutputSurfaceNV'
	gl.glVDPAURegisterOutputSurfaceNV.errcheck = _errcheck
	gl.glVDPAURegisterOutputSurfaceNV.restype = GLvdpauSurfaceNV
	gl.glVDPAURegisterOutputSurfaceNV.argtypes = (POINTER(GLvoid), GLenum, GLsizei, POINTER(GLuint),)
	gl.glVDPAURegisterVideoSurfaceNV = _getProcAddress('glVDPAURegisterVideoSurfaceNV')
	gl.glVDPAURegisterVideoSurfaceNV.gl = gl
	gl.glVDPAURegisterVideoSurfaceNV.__name__ = 'glVDPAURegisterVideoSurfaceNV'
	gl.glVDPAURegisterVideoSurfaceNV.errcheck = _errcheck
	gl.glVDPAURegisterVideoSurfaceNV.restype = GLvdpauSurfaceNV
	gl.glVDPAURegisterVideoSurfaceNV.argtypes = (POINTER(GLvoid), GLenum, GLsizei, POINTER(GLuint),)
	gl.glVDPAUSurfaceAccessNV = _getProcAddress('glVDPAUSurfaceAccessNV')
	gl.glVDPAUSurfaceAccessNV.gl = gl
	gl.glVDPAUSurfaceAccessNV.__name__ = 'glVDPAUSurfaceAccessNV'
	gl.glVDPAUSurfaceAccessNV.errcheck = _errcheck
	gl.glVDPAUSurfaceAccessNV.restype = GLvoid
	gl.glVDPAUSurfaceAccessNV.argtypes = (GLvdpauSurfaceNV, GLenum,)
	gl.glVDPAUUnmapSurfacesNV = _getProcAddress('glVDPAUUnmapSurfacesNV')
	gl.glVDPAUUnmapSurfacesNV.gl = gl
	gl.glVDPAUUnmapSurfacesNV.__name__ = 'glVDPAUUnmapSurfacesNV'
	gl.glVDPAUUnmapSurfacesNV.errcheck = _errcheck
	gl.glVDPAUUnmapSurfacesNV.restype = GLvoid
	gl.glVDPAUUnmapSurfacesNV.argtypes = (GLsizei, POINTER(GLvdpauSurfaceNV),)
	gl.glVDPAUUnregisterSurfaceNV = _getProcAddress('glVDPAUUnregisterSurfaceNV')
	gl.glVDPAUUnregisterSurfaceNV.gl = gl
	gl.glVDPAUUnregisterSurfaceNV.__name__ = 'glVDPAUUnregisterSurfaceNV'
	gl.glVDPAUUnregisterSurfaceNV.errcheck = _errcheck
	gl.glVDPAUUnregisterSurfaceNV.restype = GLvoid
	gl.glVDPAUUnregisterSurfaceNV.argtypes = (GLvdpauSurfaceNV,)
	gl.glValidateProgram = _getProcAddress('glValidateProgram')
	gl.glValidateProgram.gl = gl
	gl.glValidateProgram.__name__ = 'glValidateProgram'
	gl.glValidateProgram.errcheck = _errcheck
	gl.glValidateProgram.restype = GLvoid
	gl.glValidateProgram.argtypes = (GLuint,)
	gl.glValidateProgramARB = _getProcAddress('glValidateProgramARB')
	gl.glValidateProgramARB.gl = gl
	gl.glValidateProgramARB.__name__ = 'glValidateProgramARB'
	gl.glValidateProgramARB.errcheck = _errcheck
	gl.glValidateProgramARB.restype = GLvoid
	gl.glValidateProgramARB.argtypes = (GLhandleARB,)
	gl.glValidateProgramPipeline = _getProcAddress('glValidateProgramPipeline')
	gl.glValidateProgramPipeline.gl = gl
	gl.glValidateProgramPipeline.__name__ = 'glValidateProgramPipeline'
	gl.glValidateProgramPipeline.errcheck = _errcheck
	gl.glValidateProgramPipeline.restype = GLvoid
	gl.glValidateProgramPipeline.argtypes = (GLuint,)
	gl.glValidateProgramPipelineEXT = _getProcAddress('glValidateProgramPipelineEXT')
	gl.glValidateProgramPipelineEXT.gl = gl
	gl.glValidateProgramPipelineEXT.__name__ = 'glValidateProgramPipelineEXT'
	gl.glValidateProgramPipelineEXT.errcheck = _errcheck
	gl.glValidateProgramPipelineEXT.restype = GLvoid
	gl.glValidateProgramPipelineEXT.argtypes = (GLuint,)
	gl.glVariantArrayObjectATI = _getProcAddress('glVariantArrayObjectATI')
	gl.glVariantArrayObjectATI.gl = gl
	gl.glVariantArrayObjectATI.__name__ = 'glVariantArrayObjectATI'
	gl.glVariantArrayObjectATI.errcheck = _errcheck
	gl.glVariantArrayObjectATI.restype = GLvoid
	gl.glVariantArrayObjectATI.argtypes = (GLuint, GLenum, GLsizei, GLuint, GLuint,)
	gl.glVariantPointerEXT = _getProcAddress('glVariantPointerEXT')
	gl.glVariantPointerEXT.gl = gl
	gl.glVariantPointerEXT.__name__ = 'glVariantPointerEXT'
	gl.glVariantPointerEXT.errcheck = _errcheck
	gl.glVariantPointerEXT.restype = GLvoid
	gl.glVariantPointerEXT.argtypes = (GLuint, GLenum, GLuint, POINTER(GLvoid),)
	gl.glVariantbvEXT = _getProcAddress('glVariantbvEXT')
	gl.glVariantbvEXT.gl = gl
	gl.glVariantbvEXT.__name__ = 'glVariantbvEXT'
	gl.glVariantbvEXT.errcheck = _errcheck
	gl.glVariantbvEXT.restype = GLvoid
	gl.glVariantbvEXT.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVariantdvEXT = _getProcAddress('glVariantdvEXT')
	gl.glVariantdvEXT.gl = gl
	gl.glVariantdvEXT.__name__ = 'glVariantdvEXT'
	gl.glVariantdvEXT.errcheck = _errcheck
	gl.glVariantdvEXT.restype = GLvoid
	gl.glVariantdvEXT.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVariantfvEXT = _getProcAddress('glVariantfvEXT')
	gl.glVariantfvEXT.gl = gl
	gl.glVariantfvEXT.__name__ = 'glVariantfvEXT'
	gl.glVariantfvEXT.errcheck = _errcheck
	gl.glVariantfvEXT.restype = GLvoid
	gl.glVariantfvEXT.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVariantivEXT = _getProcAddress('glVariantivEXT')
	gl.glVariantivEXT.gl = gl
	gl.glVariantivEXT.__name__ = 'glVariantivEXT'
	gl.glVariantivEXT.errcheck = _errcheck
	gl.glVariantivEXT.restype = GLvoid
	gl.glVariantivEXT.argtypes = (GLuint, POINTER(GLint),)
	gl.glVariantsvEXT = _getProcAddress('glVariantsvEXT')
	gl.glVariantsvEXT.gl = gl
	gl.glVariantsvEXT.__name__ = 'glVariantsvEXT'
	gl.glVariantsvEXT.errcheck = _errcheck
	gl.glVariantsvEXT.restype = GLvoid
	gl.glVariantsvEXT.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVariantubvEXT = _getProcAddress('glVariantubvEXT')
	gl.glVariantubvEXT.gl = gl
	gl.glVariantubvEXT.__name__ = 'glVariantubvEXT'
	gl.glVariantubvEXT.errcheck = _errcheck
	gl.glVariantubvEXT.restype = GLvoid
	gl.glVariantubvEXT.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVariantuivEXT = _getProcAddress('glVariantuivEXT')
	gl.glVariantuivEXT.gl = gl
	gl.glVariantuivEXT.__name__ = 'glVariantuivEXT'
	gl.glVariantuivEXT.errcheck = _errcheck
	gl.glVariantuivEXT.restype = GLvoid
	gl.glVariantuivEXT.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVariantusvEXT = _getProcAddress('glVariantusvEXT')
	gl.glVariantusvEXT.gl = gl
	gl.glVariantusvEXT.__name__ = 'glVariantusvEXT'
	gl.glVariantusvEXT.errcheck = _errcheck
	gl.glVariantusvEXT.restype = GLvoid
	gl.glVariantusvEXT.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertex2bOES = _getProcAddress('glVertex2bOES')
	gl.glVertex2bOES.gl = gl
	gl.glVertex2bOES.__name__ = 'glVertex2bOES'
	gl.glVertex2bOES.errcheck = _errcheck
	gl.glVertex2bOES.restype = GLvoid
	gl.glVertex2bOES.argtypes = (GLbyte, GLbyte,)
	gl.glVertex2bvOES = _getProcAddress('glVertex2bvOES')
	gl.glVertex2bvOES.gl = gl
	gl.glVertex2bvOES.__name__ = 'glVertex2bvOES'
	gl.glVertex2bvOES.errcheck = _errcheck
	gl.glVertex2bvOES.restype = GLvoid
	gl.glVertex2bvOES.argtypes = (POINTER(GLbyte),)
	gl.glVertex2d = _getProcAddress('glVertex2d')
	gl.glVertex2d.gl = gl
	gl.glVertex2d.__name__ = 'glVertex2d'
	gl.glVertex2d.errcheck = _errcheck
	gl.glVertex2d.restype = GLvoid
	gl.glVertex2d.argtypes = (GLdouble, GLdouble,)
	gl.glVertex2dv = _getProcAddress('glVertex2dv')
	gl.glVertex2dv.gl = gl
	gl.glVertex2dv.__name__ = 'glVertex2dv'
	gl.glVertex2dv.errcheck = _errcheck
	gl.glVertex2dv.restype = GLvoid
	gl.glVertex2dv.argtypes = (POINTER(GLdouble),)
	gl.glVertex2f = _getProcAddress('glVertex2f')
	gl.glVertex2f.gl = gl
	gl.glVertex2f.__name__ = 'glVertex2f'
	gl.glVertex2f.errcheck = _errcheck
	gl.glVertex2f.restype = GLvoid
	gl.glVertex2f.argtypes = (GLfloat, GLfloat,)
	gl.glVertex2fv = _getProcAddress('glVertex2fv')
	gl.glVertex2fv.gl = gl
	gl.glVertex2fv.__name__ = 'glVertex2fv'
	gl.glVertex2fv.errcheck = _errcheck
	gl.glVertex2fv.restype = GLvoid
	gl.glVertex2fv.argtypes = (POINTER(GLfloat),)
	gl.glVertex2hNV = _getProcAddress('glVertex2hNV')
	gl.glVertex2hNV.gl = gl
	gl.glVertex2hNV.__name__ = 'glVertex2hNV'
	gl.glVertex2hNV.errcheck = _errcheck
	gl.glVertex2hNV.restype = GLvoid
	gl.glVertex2hNV.argtypes = (GLhalfNV, GLhalfNV,)
	gl.glVertex2hvNV = _getProcAddress('glVertex2hvNV')
	gl.glVertex2hvNV.gl = gl
	gl.glVertex2hvNV.__name__ = 'glVertex2hvNV'
	gl.glVertex2hvNV.errcheck = _errcheck
	gl.glVertex2hvNV.restype = GLvoid
	gl.glVertex2hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glVertex2i = _getProcAddress('glVertex2i')
	gl.glVertex2i.gl = gl
	gl.glVertex2i.__name__ = 'glVertex2i'
	gl.glVertex2i.errcheck = _errcheck
	gl.glVertex2i.restype = GLvoid
	gl.glVertex2i.argtypes = (GLint, GLint,)
	gl.glVertex2iv = _getProcAddress('glVertex2iv')
	gl.glVertex2iv.gl = gl
	gl.glVertex2iv.__name__ = 'glVertex2iv'
	gl.glVertex2iv.errcheck = _errcheck
	gl.glVertex2iv.restype = GLvoid
	gl.glVertex2iv.argtypes = (POINTER(GLint),)
	gl.glVertex2s = _getProcAddress('glVertex2s')
	gl.glVertex2s.gl = gl
	gl.glVertex2s.__name__ = 'glVertex2s'
	gl.glVertex2s.errcheck = _errcheck
	gl.glVertex2s.restype = GLvoid
	gl.glVertex2s.argtypes = (GLshort, GLshort,)
	gl.glVertex2sv = _getProcAddress('glVertex2sv')
	gl.glVertex2sv.gl = gl
	gl.glVertex2sv.__name__ = 'glVertex2sv'
	gl.glVertex2sv.errcheck = _errcheck
	gl.glVertex2sv.restype = GLvoid
	gl.glVertex2sv.argtypes = (POINTER(GLshort),)
	gl.glVertex2xOES = _getProcAddress('glVertex2xOES')
	gl.glVertex2xOES.gl = gl
	gl.glVertex2xOES.__name__ = 'glVertex2xOES'
	gl.glVertex2xOES.errcheck = _errcheck
	gl.glVertex2xOES.restype = GLvoid
	gl.glVertex2xOES.argtypes = (GLfixed,)
	gl.glVertex2xvOES = _getProcAddress('glVertex2xvOES')
	gl.glVertex2xvOES.gl = gl
	gl.glVertex2xvOES.__name__ = 'glVertex2xvOES'
	gl.glVertex2xvOES.errcheck = _errcheck
	gl.glVertex2xvOES.restype = GLvoid
	gl.glVertex2xvOES.argtypes = (POINTER(GLfixed),)
	gl.glVertex3bOES = _getProcAddress('glVertex3bOES')
	gl.glVertex3bOES.gl = gl
	gl.glVertex3bOES.__name__ = 'glVertex3bOES'
	gl.glVertex3bOES.errcheck = _errcheck
	gl.glVertex3bOES.restype = GLvoid
	gl.glVertex3bOES.argtypes = (GLbyte, GLbyte, GLbyte,)
	gl.glVertex3bvOES = _getProcAddress('glVertex3bvOES')
	gl.glVertex3bvOES.gl = gl
	gl.glVertex3bvOES.__name__ = 'glVertex3bvOES'
	gl.glVertex3bvOES.errcheck = _errcheck
	gl.glVertex3bvOES.restype = GLvoid
	gl.glVertex3bvOES.argtypes = (POINTER(GLbyte),)
	gl.glVertex3d = _getProcAddress('glVertex3d')
	gl.glVertex3d.gl = gl
	gl.glVertex3d.__name__ = 'glVertex3d'
	gl.glVertex3d.errcheck = _errcheck
	gl.glVertex3d.restype = GLvoid
	gl.glVertex3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glVertex3dv = _getProcAddress('glVertex3dv')
	gl.glVertex3dv.gl = gl
	gl.glVertex3dv.__name__ = 'glVertex3dv'
	gl.glVertex3dv.errcheck = _errcheck
	gl.glVertex3dv.restype = GLvoid
	gl.glVertex3dv.argtypes = (POINTER(GLdouble),)
	gl.glVertex3f = _getProcAddress('glVertex3f')
	gl.glVertex3f.gl = gl
	gl.glVertex3f.__name__ = 'glVertex3f'
	gl.glVertex3f.errcheck = _errcheck
	gl.glVertex3f.restype = GLvoid
	gl.glVertex3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glVertex3fv = _getProcAddress('glVertex3fv')
	gl.glVertex3fv.gl = gl
	gl.glVertex3fv.__name__ = 'glVertex3fv'
	gl.glVertex3fv.errcheck = _errcheck
	gl.glVertex3fv.restype = GLvoid
	gl.glVertex3fv.argtypes = (POINTER(GLfloat),)
	gl.glVertex3hNV = _getProcAddress('glVertex3hNV')
	gl.glVertex3hNV.gl = gl
	gl.glVertex3hNV.__name__ = 'glVertex3hNV'
	gl.glVertex3hNV.errcheck = _errcheck
	gl.glVertex3hNV.restype = GLvoid
	gl.glVertex3hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glVertex3hvNV = _getProcAddress('glVertex3hvNV')
	gl.glVertex3hvNV.gl = gl
	gl.glVertex3hvNV.__name__ = 'glVertex3hvNV'
	gl.glVertex3hvNV.errcheck = _errcheck
	gl.glVertex3hvNV.restype = GLvoid
	gl.glVertex3hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glVertex3i = _getProcAddress('glVertex3i')
	gl.glVertex3i.gl = gl
	gl.glVertex3i.__name__ = 'glVertex3i'
	gl.glVertex3i.errcheck = _errcheck
	gl.glVertex3i.restype = GLvoid
	gl.glVertex3i.argtypes = (GLint, GLint, GLint,)
	gl.glVertex3iv = _getProcAddress('glVertex3iv')
	gl.glVertex3iv.gl = gl
	gl.glVertex3iv.__name__ = 'glVertex3iv'
	gl.glVertex3iv.errcheck = _errcheck
	gl.glVertex3iv.restype = GLvoid
	gl.glVertex3iv.argtypes = (POINTER(GLint),)
	gl.glVertex3s = _getProcAddress('glVertex3s')
	gl.glVertex3s.gl = gl
	gl.glVertex3s.__name__ = 'glVertex3s'
	gl.glVertex3s.errcheck = _errcheck
	gl.glVertex3s.restype = GLvoid
	gl.glVertex3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glVertex3sv = _getProcAddress('glVertex3sv')
	gl.glVertex3sv.gl = gl
	gl.glVertex3sv.__name__ = 'glVertex3sv'
	gl.glVertex3sv.errcheck = _errcheck
	gl.glVertex3sv.restype = GLvoid
	gl.glVertex3sv.argtypes = (POINTER(GLshort),)
	gl.glVertex3xOES = _getProcAddress('glVertex3xOES')
	gl.glVertex3xOES.gl = gl
	gl.glVertex3xOES.__name__ = 'glVertex3xOES'
	gl.glVertex3xOES.errcheck = _errcheck
	gl.glVertex3xOES.restype = GLvoid
	gl.glVertex3xOES.argtypes = (GLfixed, GLfixed,)
	gl.glVertex3xvOES = _getProcAddress('glVertex3xvOES')
	gl.glVertex3xvOES.gl = gl
	gl.glVertex3xvOES.__name__ = 'glVertex3xvOES'
	gl.glVertex3xvOES.errcheck = _errcheck
	gl.glVertex3xvOES.restype = GLvoid
	gl.glVertex3xvOES.argtypes = (POINTER(GLfixed),)
	gl.glVertex4bOES = _getProcAddress('glVertex4bOES')
	gl.glVertex4bOES.gl = gl
	gl.glVertex4bOES.__name__ = 'glVertex4bOES'
	gl.glVertex4bOES.errcheck = _errcheck
	gl.glVertex4bOES.restype = GLvoid
	gl.glVertex4bOES.argtypes = (GLbyte, GLbyte, GLbyte, GLbyte,)
	gl.glVertex4bvOES = _getProcAddress('glVertex4bvOES')
	gl.glVertex4bvOES.gl = gl
	gl.glVertex4bvOES.__name__ = 'glVertex4bvOES'
	gl.glVertex4bvOES.errcheck = _errcheck
	gl.glVertex4bvOES.restype = GLvoid
	gl.glVertex4bvOES.argtypes = (POINTER(GLbyte),)
	gl.glVertex4d = _getProcAddress('glVertex4d')
	gl.glVertex4d.gl = gl
	gl.glVertex4d.__name__ = 'glVertex4d'
	gl.glVertex4d.errcheck = _errcheck
	gl.glVertex4d.restype = GLvoid
	gl.glVertex4d.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertex4dv = _getProcAddress('glVertex4dv')
	gl.glVertex4dv.gl = gl
	gl.glVertex4dv.__name__ = 'glVertex4dv'
	gl.glVertex4dv.errcheck = _errcheck
	gl.glVertex4dv.restype = GLvoid
	gl.glVertex4dv.argtypes = (POINTER(GLdouble),)
	gl.glVertex4f = _getProcAddress('glVertex4f')
	gl.glVertex4f.gl = gl
	gl.glVertex4f.__name__ = 'glVertex4f'
	gl.glVertex4f.errcheck = _errcheck
	gl.glVertex4f.restype = GLvoid
	gl.glVertex4f.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glVertex4fv = _getProcAddress('glVertex4fv')
	gl.glVertex4fv.gl = gl
	gl.glVertex4fv.__name__ = 'glVertex4fv'
	gl.glVertex4fv.errcheck = _errcheck
	gl.glVertex4fv.restype = GLvoid
	gl.glVertex4fv.argtypes = (POINTER(GLfloat),)
	gl.glVertex4hNV = _getProcAddress('glVertex4hNV')
	gl.glVertex4hNV.gl = gl
	gl.glVertex4hNV.__name__ = 'glVertex4hNV'
	gl.glVertex4hNV.errcheck = _errcheck
	gl.glVertex4hNV.restype = GLvoid
	gl.glVertex4hNV.argtypes = (GLhalfNV, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glVertex4hvNV = _getProcAddress('glVertex4hvNV')
	gl.glVertex4hvNV.gl = gl
	gl.glVertex4hvNV.__name__ = 'glVertex4hvNV'
	gl.glVertex4hvNV.errcheck = _errcheck
	gl.glVertex4hvNV.restype = GLvoid
	gl.glVertex4hvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glVertex4i = _getProcAddress('glVertex4i')
	gl.glVertex4i.gl = gl
	gl.glVertex4i.__name__ = 'glVertex4i'
	gl.glVertex4i.errcheck = _errcheck
	gl.glVertex4i.restype = GLvoid
	gl.glVertex4i.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glVertex4iv = _getProcAddress('glVertex4iv')
	gl.glVertex4iv.gl = gl
	gl.glVertex4iv.__name__ = 'glVertex4iv'
	gl.glVertex4iv.errcheck = _errcheck
	gl.glVertex4iv.restype = GLvoid
	gl.glVertex4iv.argtypes = (POINTER(GLint),)
	gl.glVertex4s = _getProcAddress('glVertex4s')
	gl.glVertex4s.gl = gl
	gl.glVertex4s.__name__ = 'glVertex4s'
	gl.glVertex4s.errcheck = _errcheck
	gl.glVertex4s.restype = GLvoid
	gl.glVertex4s.argtypes = (GLshort, GLshort, GLshort, GLshort,)
	gl.glVertex4sv = _getProcAddress('glVertex4sv')
	gl.glVertex4sv.gl = gl
	gl.glVertex4sv.__name__ = 'glVertex4sv'
	gl.glVertex4sv.errcheck = _errcheck
	gl.glVertex4sv.restype = GLvoid
	gl.glVertex4sv.argtypes = (POINTER(GLshort),)
	gl.glVertex4xOES = _getProcAddress('glVertex4xOES')
	gl.glVertex4xOES.gl = gl
	gl.glVertex4xOES.__name__ = 'glVertex4xOES'
	gl.glVertex4xOES.errcheck = _errcheck
	gl.glVertex4xOES.restype = GLvoid
	gl.glVertex4xOES.argtypes = (GLfixed, GLfixed, GLfixed,)
	gl.glVertex4xvOES = _getProcAddress('glVertex4xvOES')
	gl.glVertex4xvOES.gl = gl
	gl.glVertex4xvOES.__name__ = 'glVertex4xvOES'
	gl.glVertex4xvOES.errcheck = _errcheck
	gl.glVertex4xvOES.restype = GLvoid
	gl.glVertex4xvOES.argtypes = (POINTER(GLfixed),)
	gl.glVertexArrayAttribBinding = _getProcAddress('glVertexArrayAttribBinding')
	gl.glVertexArrayAttribBinding.gl = gl
	gl.glVertexArrayAttribBinding.__name__ = 'glVertexArrayAttribBinding'
	gl.glVertexArrayAttribBinding.errcheck = _errcheck
	gl.glVertexArrayAttribBinding.restype = GLvoid
	gl.glVertexArrayAttribBinding.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexArrayAttribFormat = _getProcAddress('glVertexArrayAttribFormat')
	gl.glVertexArrayAttribFormat.gl = gl
	gl.glVertexArrayAttribFormat.__name__ = 'glVertexArrayAttribFormat'
	gl.glVertexArrayAttribFormat.errcheck = _errcheck
	gl.glVertexArrayAttribFormat.restype = GLvoid
	gl.glVertexArrayAttribFormat.argtypes = (GLuint, GLuint, GLint, GLenum, GLboolean, GLuint,)
	gl.glVertexArrayAttribIFormat = _getProcAddress('glVertexArrayAttribIFormat')
	gl.glVertexArrayAttribIFormat.gl = gl
	gl.glVertexArrayAttribIFormat.__name__ = 'glVertexArrayAttribIFormat'
	gl.glVertexArrayAttribIFormat.errcheck = _errcheck
	gl.glVertexArrayAttribIFormat.restype = GLvoid
	gl.glVertexArrayAttribIFormat.argtypes = (GLuint, GLuint, GLint, GLenum, GLuint,)
	gl.glVertexArrayAttribLFormat = _getProcAddress('glVertexArrayAttribLFormat')
	gl.glVertexArrayAttribLFormat.gl = gl
	gl.glVertexArrayAttribLFormat.__name__ = 'glVertexArrayAttribLFormat'
	gl.glVertexArrayAttribLFormat.errcheck = _errcheck
	gl.glVertexArrayAttribLFormat.restype = GLvoid
	gl.glVertexArrayAttribLFormat.argtypes = (GLuint, GLuint, GLint, GLenum, GLuint,)
	gl.glVertexArrayBindVertexBufferEXT = _getProcAddress('glVertexArrayBindVertexBufferEXT')
	gl.glVertexArrayBindVertexBufferEXT.gl = gl
	gl.glVertexArrayBindVertexBufferEXT.__name__ = 'glVertexArrayBindVertexBufferEXT'
	gl.glVertexArrayBindVertexBufferEXT.errcheck = _errcheck
	gl.glVertexArrayBindVertexBufferEXT.restype = GLvoid
	gl.glVertexArrayBindVertexBufferEXT.argtypes = (GLuint, GLuint, GLuint, GLintptr, GLsizei,)
	gl.glVertexArrayBindingDivisor = _getProcAddress('glVertexArrayBindingDivisor')
	gl.glVertexArrayBindingDivisor.gl = gl
	gl.glVertexArrayBindingDivisor.__name__ = 'glVertexArrayBindingDivisor'
	gl.glVertexArrayBindingDivisor.errcheck = _errcheck
	gl.glVertexArrayBindingDivisor.restype = GLvoid
	gl.glVertexArrayBindingDivisor.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexArrayColorOffsetEXT = _getProcAddress('glVertexArrayColorOffsetEXT')
	gl.glVertexArrayColorOffsetEXT.gl = gl
	gl.glVertexArrayColorOffsetEXT.__name__ = 'glVertexArrayColorOffsetEXT'
	gl.glVertexArrayColorOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayColorOffsetEXT.restype = GLvoid
	gl.glVertexArrayColorOffsetEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayEdgeFlagOffsetEXT = _getProcAddress('glVertexArrayEdgeFlagOffsetEXT')
	gl.glVertexArrayEdgeFlagOffsetEXT.gl = gl
	gl.glVertexArrayEdgeFlagOffsetEXT.__name__ = 'glVertexArrayEdgeFlagOffsetEXT'
	gl.glVertexArrayEdgeFlagOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayEdgeFlagOffsetEXT.restype = GLvoid
	gl.glVertexArrayEdgeFlagOffsetEXT.argtypes = (GLuint, GLuint, GLsizei, GLintptr,)
	gl.glVertexArrayElementBuffer = _getProcAddress('glVertexArrayElementBuffer')
	gl.glVertexArrayElementBuffer.gl = gl
	gl.glVertexArrayElementBuffer.__name__ = 'glVertexArrayElementBuffer'
	gl.glVertexArrayElementBuffer.errcheck = _errcheck
	gl.glVertexArrayElementBuffer.restype = GLvoid
	gl.glVertexArrayElementBuffer.argtypes = (GLuint, GLuint,)
	gl.glVertexArrayFogCoordOffsetEXT = _getProcAddress('glVertexArrayFogCoordOffsetEXT')
	gl.glVertexArrayFogCoordOffsetEXT.gl = gl
	gl.glVertexArrayFogCoordOffsetEXT.__name__ = 'glVertexArrayFogCoordOffsetEXT'
	gl.glVertexArrayFogCoordOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayFogCoordOffsetEXT.restype = GLvoid
	gl.glVertexArrayFogCoordOffsetEXT.argtypes = (GLuint, GLuint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayIndexOffsetEXT = _getProcAddress('glVertexArrayIndexOffsetEXT')
	gl.glVertexArrayIndexOffsetEXT.gl = gl
	gl.glVertexArrayIndexOffsetEXT.__name__ = 'glVertexArrayIndexOffsetEXT'
	gl.glVertexArrayIndexOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayIndexOffsetEXT.restype = GLvoid
	gl.glVertexArrayIndexOffsetEXT.argtypes = (GLuint, GLuint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayMultiTexCoordOffsetEXT = _getProcAddress('glVertexArrayMultiTexCoordOffsetEXT')
	gl.glVertexArrayMultiTexCoordOffsetEXT.gl = gl
	gl.glVertexArrayMultiTexCoordOffsetEXT.__name__ = 'glVertexArrayMultiTexCoordOffsetEXT'
	gl.glVertexArrayMultiTexCoordOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayMultiTexCoordOffsetEXT.restype = GLvoid
	gl.glVertexArrayMultiTexCoordOffsetEXT.argtypes = (GLuint, GLuint, GLenum, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayNormalOffsetEXT = _getProcAddress('glVertexArrayNormalOffsetEXT')
	gl.glVertexArrayNormalOffsetEXT.gl = gl
	gl.glVertexArrayNormalOffsetEXT.__name__ = 'glVertexArrayNormalOffsetEXT'
	gl.glVertexArrayNormalOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayNormalOffsetEXT.restype = GLvoid
	gl.glVertexArrayNormalOffsetEXT.argtypes = (GLuint, GLuint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayParameteriAPPLE = _getProcAddress('glVertexArrayParameteriAPPLE')
	gl.glVertexArrayParameteriAPPLE.gl = gl
	gl.glVertexArrayParameteriAPPLE.__name__ = 'glVertexArrayParameteriAPPLE'
	gl.glVertexArrayParameteriAPPLE.errcheck = _errcheck
	gl.glVertexArrayParameteriAPPLE.restype = GLvoid
	gl.glVertexArrayParameteriAPPLE.argtypes = (GLenum, GLint,)
	gl.glVertexArrayRangeAPPLE = _getProcAddress('glVertexArrayRangeAPPLE')
	gl.glVertexArrayRangeAPPLE.gl = gl
	gl.glVertexArrayRangeAPPLE.__name__ = 'glVertexArrayRangeAPPLE'
	gl.glVertexArrayRangeAPPLE.errcheck = _errcheck
	gl.glVertexArrayRangeAPPLE.restype = GLvoid
	gl.glVertexArrayRangeAPPLE.argtypes = (GLsizei, POINTER(GLvoid),)
	gl.glVertexArrayRangeNV = _getProcAddress('glVertexArrayRangeNV')
	gl.glVertexArrayRangeNV.gl = gl
	gl.glVertexArrayRangeNV.__name__ = 'glVertexArrayRangeNV'
	gl.glVertexArrayRangeNV.errcheck = _errcheck
	gl.glVertexArrayRangeNV.restype = GLvoid
	gl.glVertexArrayRangeNV.argtypes = (GLsizei, POINTER(GLvoid),)
	gl.glVertexArraySecondaryColorOffsetEXT = _getProcAddress('glVertexArraySecondaryColorOffsetEXT')
	gl.glVertexArraySecondaryColorOffsetEXT.gl = gl
	gl.glVertexArraySecondaryColorOffsetEXT.__name__ = 'glVertexArraySecondaryColorOffsetEXT'
	gl.glVertexArraySecondaryColorOffsetEXT.errcheck = _errcheck
	gl.glVertexArraySecondaryColorOffsetEXT.restype = GLvoid
	gl.glVertexArraySecondaryColorOffsetEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayTexCoordOffsetEXT = _getProcAddress('glVertexArrayTexCoordOffsetEXT')
	gl.glVertexArrayTexCoordOffsetEXT.gl = gl
	gl.glVertexArrayTexCoordOffsetEXT.__name__ = 'glVertexArrayTexCoordOffsetEXT'
	gl.glVertexArrayTexCoordOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayTexCoordOffsetEXT.restype = GLvoid
	gl.glVertexArrayTexCoordOffsetEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayVertexAttribBindingEXT = _getProcAddress('glVertexArrayVertexAttribBindingEXT')
	gl.glVertexArrayVertexAttribBindingEXT.gl = gl
	gl.glVertexArrayVertexAttribBindingEXT.__name__ = 'glVertexArrayVertexAttribBindingEXT'
	gl.glVertexArrayVertexAttribBindingEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribBindingEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribBindingEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexArrayVertexAttribDivisorEXT = _getProcAddress('glVertexArrayVertexAttribDivisorEXT')
	gl.glVertexArrayVertexAttribDivisorEXT.gl = gl
	gl.glVertexArrayVertexAttribDivisorEXT.__name__ = 'glVertexArrayVertexAttribDivisorEXT'
	gl.glVertexArrayVertexAttribDivisorEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribDivisorEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribDivisorEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexArrayVertexAttribFormatEXT = _getProcAddress('glVertexArrayVertexAttribFormatEXT')
	gl.glVertexArrayVertexAttribFormatEXT.gl = gl
	gl.glVertexArrayVertexAttribFormatEXT.__name__ = 'glVertexArrayVertexAttribFormatEXT'
	gl.glVertexArrayVertexAttribFormatEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribFormatEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribFormatEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLboolean, GLuint,)
	gl.glVertexArrayVertexAttribIFormatEXT = _getProcAddress('glVertexArrayVertexAttribIFormatEXT')
	gl.glVertexArrayVertexAttribIFormatEXT.gl = gl
	gl.glVertexArrayVertexAttribIFormatEXT.__name__ = 'glVertexArrayVertexAttribIFormatEXT'
	gl.glVertexArrayVertexAttribIFormatEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribIFormatEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribIFormatEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLuint,)
	gl.glVertexArrayVertexAttribIOffsetEXT = _getProcAddress('glVertexArrayVertexAttribIOffsetEXT')
	gl.glVertexArrayVertexAttribIOffsetEXT.gl = gl
	gl.glVertexArrayVertexAttribIOffsetEXT.__name__ = 'glVertexArrayVertexAttribIOffsetEXT'
	gl.glVertexArrayVertexAttribIOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribIOffsetEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribIOffsetEXT.argtypes = (GLuint, GLuint, GLuint, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayVertexAttribLFormatEXT = _getProcAddress('glVertexArrayVertexAttribLFormatEXT')
	gl.glVertexArrayVertexAttribLFormatEXT.gl = gl
	gl.glVertexArrayVertexAttribLFormatEXT.__name__ = 'glVertexArrayVertexAttribLFormatEXT'
	gl.glVertexArrayVertexAttribLFormatEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribLFormatEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribLFormatEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLuint,)
	gl.glVertexArrayVertexAttribLOffsetEXT = _getProcAddress('glVertexArrayVertexAttribLOffsetEXT')
	gl.glVertexArrayVertexAttribLOffsetEXT.gl = gl
	gl.glVertexArrayVertexAttribLOffsetEXT.__name__ = 'glVertexArrayVertexAttribLOffsetEXT'
	gl.glVertexArrayVertexAttribLOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribLOffsetEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribLOffsetEXT.argtypes = (GLuint, GLuint, GLuint, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexArrayVertexAttribOffsetEXT = _getProcAddress('glVertexArrayVertexAttribOffsetEXT')
	gl.glVertexArrayVertexAttribOffsetEXT.gl = gl
	gl.glVertexArrayVertexAttribOffsetEXT.__name__ = 'glVertexArrayVertexAttribOffsetEXT'
	gl.glVertexArrayVertexAttribOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayVertexAttribOffsetEXT.restype = GLvoid
	gl.glVertexArrayVertexAttribOffsetEXT.argtypes = (GLuint, GLuint, GLuint, GLint, GLenum, GLboolean, GLsizei, GLintptr,)
	gl.glVertexArrayVertexBindingDivisorEXT = _getProcAddress('glVertexArrayVertexBindingDivisorEXT')
	gl.glVertexArrayVertexBindingDivisorEXT.gl = gl
	gl.glVertexArrayVertexBindingDivisorEXT.__name__ = 'glVertexArrayVertexBindingDivisorEXT'
	gl.glVertexArrayVertexBindingDivisorEXT.errcheck = _errcheck
	gl.glVertexArrayVertexBindingDivisorEXT.restype = GLvoid
	gl.glVertexArrayVertexBindingDivisorEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexArrayVertexBuffer = _getProcAddress('glVertexArrayVertexBuffer')
	gl.glVertexArrayVertexBuffer.gl = gl
	gl.glVertexArrayVertexBuffer.__name__ = 'glVertexArrayVertexBuffer'
	gl.glVertexArrayVertexBuffer.errcheck = _errcheck
	gl.glVertexArrayVertexBuffer.restype = GLvoid
	gl.glVertexArrayVertexBuffer.argtypes = (GLuint, GLuint, GLuint, GLintptr, GLsizei,)
	gl.glVertexArrayVertexBuffers = _getProcAddress('glVertexArrayVertexBuffers')
	gl.glVertexArrayVertexBuffers.gl = gl
	gl.glVertexArrayVertexBuffers.__name__ = 'glVertexArrayVertexBuffers'
	gl.glVertexArrayVertexBuffers.errcheck = _errcheck
	gl.glVertexArrayVertexBuffers.restype = GLvoid
	gl.glVertexArrayVertexBuffers.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLuint), POINTER(GLintptr), POINTER(GLsizei),)
	gl.glVertexArrayVertexOffsetEXT = _getProcAddress('glVertexArrayVertexOffsetEXT')
	gl.glVertexArrayVertexOffsetEXT.gl = gl
	gl.glVertexArrayVertexOffsetEXT.__name__ = 'glVertexArrayVertexOffsetEXT'
	gl.glVertexArrayVertexOffsetEXT.errcheck = _errcheck
	gl.glVertexArrayVertexOffsetEXT.restype = GLvoid
	gl.glVertexArrayVertexOffsetEXT.argtypes = (GLuint, GLuint, GLint, GLenum, GLsizei, GLintptr,)
	gl.glVertexAttrib1d = _getProcAddress('glVertexAttrib1d')
	gl.glVertexAttrib1d.gl = gl
	gl.glVertexAttrib1d.__name__ = 'glVertexAttrib1d'
	gl.glVertexAttrib1d.errcheck = _errcheck
	gl.glVertexAttrib1d.restype = GLvoid
	gl.glVertexAttrib1d.argtypes = (GLuint, GLdouble,)
	gl.glVertexAttrib1dARB = _getProcAddress('glVertexAttrib1dARB')
	gl.glVertexAttrib1dARB.gl = gl
	gl.glVertexAttrib1dARB.__name__ = 'glVertexAttrib1dARB'
	gl.glVertexAttrib1dARB.errcheck = _errcheck
	gl.glVertexAttrib1dARB.restype = GLvoid
	gl.glVertexAttrib1dARB.argtypes = (GLuint, GLdouble,)
	gl.glVertexAttrib1dNV = _getProcAddress('glVertexAttrib1dNV')
	gl.glVertexAttrib1dNV.gl = gl
	gl.glVertexAttrib1dNV.__name__ = 'glVertexAttrib1dNV'
	gl.glVertexAttrib1dNV.errcheck = _errcheck
	gl.glVertexAttrib1dNV.restype = GLvoid
	gl.glVertexAttrib1dNV.argtypes = (GLuint, GLdouble,)
	gl.glVertexAttrib1dv = _getProcAddress('glVertexAttrib1dv')
	gl.glVertexAttrib1dv.gl = gl
	gl.glVertexAttrib1dv.__name__ = 'glVertexAttrib1dv'
	gl.glVertexAttrib1dv.errcheck = _errcheck
	gl.glVertexAttrib1dv.restype = GLvoid
	gl.glVertexAttrib1dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib1dvARB = _getProcAddress('glVertexAttrib1dvARB')
	gl.glVertexAttrib1dvARB.gl = gl
	gl.glVertexAttrib1dvARB.__name__ = 'glVertexAttrib1dvARB'
	gl.glVertexAttrib1dvARB.errcheck = _errcheck
	gl.glVertexAttrib1dvARB.restype = GLvoid
	gl.glVertexAttrib1dvARB.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib1dvNV = _getProcAddress('glVertexAttrib1dvNV')
	gl.glVertexAttrib1dvNV.gl = gl
	gl.glVertexAttrib1dvNV.__name__ = 'glVertexAttrib1dvNV'
	gl.glVertexAttrib1dvNV.errcheck = _errcheck
	gl.glVertexAttrib1dvNV.restype = GLvoid
	gl.glVertexAttrib1dvNV.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib1f = _getProcAddress('glVertexAttrib1f')
	gl.glVertexAttrib1f.gl = gl
	gl.glVertexAttrib1f.__name__ = 'glVertexAttrib1f'
	gl.glVertexAttrib1f.errcheck = _errcheck
	gl.glVertexAttrib1f.restype = GLvoid
	gl.glVertexAttrib1f.argtypes = (GLuint, GLfloat,)
	gl.glVertexAttrib1fARB = _getProcAddress('glVertexAttrib1fARB')
	gl.glVertexAttrib1fARB.gl = gl
	gl.glVertexAttrib1fARB.__name__ = 'glVertexAttrib1fARB'
	gl.glVertexAttrib1fARB.errcheck = _errcheck
	gl.glVertexAttrib1fARB.restype = GLvoid
	gl.glVertexAttrib1fARB.argtypes = (GLuint, GLfloat,)
	gl.glVertexAttrib1fNV = _getProcAddress('glVertexAttrib1fNV')
	gl.glVertexAttrib1fNV.gl = gl
	gl.glVertexAttrib1fNV.__name__ = 'glVertexAttrib1fNV'
	gl.glVertexAttrib1fNV.errcheck = _errcheck
	gl.glVertexAttrib1fNV.restype = GLvoid
	gl.glVertexAttrib1fNV.argtypes = (GLuint, GLfloat,)
	gl.glVertexAttrib1fv = _getProcAddress('glVertexAttrib1fv')
	gl.glVertexAttrib1fv.gl = gl
	gl.glVertexAttrib1fv.__name__ = 'glVertexAttrib1fv'
	gl.glVertexAttrib1fv.errcheck = _errcheck
	gl.glVertexAttrib1fv.restype = GLvoid
	gl.glVertexAttrib1fv.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib1fvARB = _getProcAddress('glVertexAttrib1fvARB')
	gl.glVertexAttrib1fvARB.gl = gl
	gl.glVertexAttrib1fvARB.__name__ = 'glVertexAttrib1fvARB'
	gl.glVertexAttrib1fvARB.errcheck = _errcheck
	gl.glVertexAttrib1fvARB.restype = GLvoid
	gl.glVertexAttrib1fvARB.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib1fvNV = _getProcAddress('glVertexAttrib1fvNV')
	gl.glVertexAttrib1fvNV.gl = gl
	gl.glVertexAttrib1fvNV.__name__ = 'glVertexAttrib1fvNV'
	gl.glVertexAttrib1fvNV.errcheck = _errcheck
	gl.glVertexAttrib1fvNV.restype = GLvoid
	gl.glVertexAttrib1fvNV.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib1hNV = _getProcAddress('glVertexAttrib1hNV')
	gl.glVertexAttrib1hNV.gl = gl
	gl.glVertexAttrib1hNV.__name__ = 'glVertexAttrib1hNV'
	gl.glVertexAttrib1hNV.errcheck = _errcheck
	gl.glVertexAttrib1hNV.restype = GLvoid
	gl.glVertexAttrib1hNV.argtypes = (GLuint, GLhalfNV,)
	gl.glVertexAttrib1hvNV = _getProcAddress('glVertexAttrib1hvNV')
	gl.glVertexAttrib1hvNV.gl = gl
	gl.glVertexAttrib1hvNV.__name__ = 'glVertexAttrib1hvNV'
	gl.glVertexAttrib1hvNV.errcheck = _errcheck
	gl.glVertexAttrib1hvNV.restype = GLvoid
	gl.glVertexAttrib1hvNV.argtypes = (GLuint, POINTER(GLhalfNV),)
	gl.glVertexAttrib1s = _getProcAddress('glVertexAttrib1s')
	gl.glVertexAttrib1s.gl = gl
	gl.glVertexAttrib1s.__name__ = 'glVertexAttrib1s'
	gl.glVertexAttrib1s.errcheck = _errcheck
	gl.glVertexAttrib1s.restype = GLvoid
	gl.glVertexAttrib1s.argtypes = (GLuint, GLshort,)
	gl.glVertexAttrib1sARB = _getProcAddress('glVertexAttrib1sARB')
	gl.glVertexAttrib1sARB.gl = gl
	gl.glVertexAttrib1sARB.__name__ = 'glVertexAttrib1sARB'
	gl.glVertexAttrib1sARB.errcheck = _errcheck
	gl.glVertexAttrib1sARB.restype = GLvoid
	gl.glVertexAttrib1sARB.argtypes = (GLuint, GLshort,)
	gl.glVertexAttrib1sNV = _getProcAddress('glVertexAttrib1sNV')
	gl.glVertexAttrib1sNV.gl = gl
	gl.glVertexAttrib1sNV.__name__ = 'glVertexAttrib1sNV'
	gl.glVertexAttrib1sNV.errcheck = _errcheck
	gl.glVertexAttrib1sNV.restype = GLvoid
	gl.glVertexAttrib1sNV.argtypes = (GLuint, GLshort,)
	gl.glVertexAttrib1sv = _getProcAddress('glVertexAttrib1sv')
	gl.glVertexAttrib1sv.gl = gl
	gl.glVertexAttrib1sv.__name__ = 'glVertexAttrib1sv'
	gl.glVertexAttrib1sv.errcheck = _errcheck
	gl.glVertexAttrib1sv.restype = GLvoid
	gl.glVertexAttrib1sv.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib1svARB = _getProcAddress('glVertexAttrib1svARB')
	gl.glVertexAttrib1svARB.gl = gl
	gl.glVertexAttrib1svARB.__name__ = 'glVertexAttrib1svARB'
	gl.glVertexAttrib1svARB.errcheck = _errcheck
	gl.glVertexAttrib1svARB.restype = GLvoid
	gl.glVertexAttrib1svARB.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib1svNV = _getProcAddress('glVertexAttrib1svNV')
	gl.glVertexAttrib1svNV.gl = gl
	gl.glVertexAttrib1svNV.__name__ = 'glVertexAttrib1svNV'
	gl.glVertexAttrib1svNV.errcheck = _errcheck
	gl.glVertexAttrib1svNV.restype = GLvoid
	gl.glVertexAttrib1svNV.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib2d = _getProcAddress('glVertexAttrib2d')
	gl.glVertexAttrib2d.gl = gl
	gl.glVertexAttrib2d.__name__ = 'glVertexAttrib2d'
	gl.glVertexAttrib2d.errcheck = _errcheck
	gl.glVertexAttrib2d.restype = GLvoid
	gl.glVertexAttrib2d.argtypes = (GLuint, GLdouble, GLdouble,)
	gl.glVertexAttrib2dARB = _getProcAddress('glVertexAttrib2dARB')
	gl.glVertexAttrib2dARB.gl = gl
	gl.glVertexAttrib2dARB.__name__ = 'glVertexAttrib2dARB'
	gl.glVertexAttrib2dARB.errcheck = _errcheck
	gl.glVertexAttrib2dARB.restype = GLvoid
	gl.glVertexAttrib2dARB.argtypes = (GLuint, GLdouble, GLdouble,)
	gl.glVertexAttrib2dNV = _getProcAddress('glVertexAttrib2dNV')
	gl.glVertexAttrib2dNV.gl = gl
	gl.glVertexAttrib2dNV.__name__ = 'glVertexAttrib2dNV'
	gl.glVertexAttrib2dNV.errcheck = _errcheck
	gl.glVertexAttrib2dNV.restype = GLvoid
	gl.glVertexAttrib2dNV.argtypes = (GLuint, GLdouble, GLdouble,)
	gl.glVertexAttrib2dv = _getProcAddress('glVertexAttrib2dv')
	gl.glVertexAttrib2dv.gl = gl
	gl.glVertexAttrib2dv.__name__ = 'glVertexAttrib2dv'
	gl.glVertexAttrib2dv.errcheck = _errcheck
	gl.glVertexAttrib2dv.restype = GLvoid
	gl.glVertexAttrib2dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib2dvARB = _getProcAddress('glVertexAttrib2dvARB')
	gl.glVertexAttrib2dvARB.gl = gl
	gl.glVertexAttrib2dvARB.__name__ = 'glVertexAttrib2dvARB'
	gl.glVertexAttrib2dvARB.errcheck = _errcheck
	gl.glVertexAttrib2dvARB.restype = GLvoid
	gl.glVertexAttrib2dvARB.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib2dvNV = _getProcAddress('glVertexAttrib2dvNV')
	gl.glVertexAttrib2dvNV.gl = gl
	gl.glVertexAttrib2dvNV.__name__ = 'glVertexAttrib2dvNV'
	gl.glVertexAttrib2dvNV.errcheck = _errcheck
	gl.glVertexAttrib2dvNV.restype = GLvoid
	gl.glVertexAttrib2dvNV.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib2f = _getProcAddress('glVertexAttrib2f')
	gl.glVertexAttrib2f.gl = gl
	gl.glVertexAttrib2f.__name__ = 'glVertexAttrib2f'
	gl.glVertexAttrib2f.errcheck = _errcheck
	gl.glVertexAttrib2f.restype = GLvoid
	gl.glVertexAttrib2f.argtypes = (GLuint, GLfloat, GLfloat,)
	gl.glVertexAttrib2fARB = _getProcAddress('glVertexAttrib2fARB')
	gl.glVertexAttrib2fARB.gl = gl
	gl.glVertexAttrib2fARB.__name__ = 'glVertexAttrib2fARB'
	gl.glVertexAttrib2fARB.errcheck = _errcheck
	gl.glVertexAttrib2fARB.restype = GLvoid
	gl.glVertexAttrib2fARB.argtypes = (GLuint, GLfloat, GLfloat,)
	gl.glVertexAttrib2fNV = _getProcAddress('glVertexAttrib2fNV')
	gl.glVertexAttrib2fNV.gl = gl
	gl.glVertexAttrib2fNV.__name__ = 'glVertexAttrib2fNV'
	gl.glVertexAttrib2fNV.errcheck = _errcheck
	gl.glVertexAttrib2fNV.restype = GLvoid
	gl.glVertexAttrib2fNV.argtypes = (GLuint, GLfloat, GLfloat,)
	gl.glVertexAttrib2fv = _getProcAddress('glVertexAttrib2fv')
	gl.glVertexAttrib2fv.gl = gl
	gl.glVertexAttrib2fv.__name__ = 'glVertexAttrib2fv'
	gl.glVertexAttrib2fv.errcheck = _errcheck
	gl.glVertexAttrib2fv.restype = GLvoid
	gl.glVertexAttrib2fv.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib2fvARB = _getProcAddress('glVertexAttrib2fvARB')
	gl.glVertexAttrib2fvARB.gl = gl
	gl.glVertexAttrib2fvARB.__name__ = 'glVertexAttrib2fvARB'
	gl.glVertexAttrib2fvARB.errcheck = _errcheck
	gl.glVertexAttrib2fvARB.restype = GLvoid
	gl.glVertexAttrib2fvARB.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib2fvNV = _getProcAddress('glVertexAttrib2fvNV')
	gl.glVertexAttrib2fvNV.gl = gl
	gl.glVertexAttrib2fvNV.__name__ = 'glVertexAttrib2fvNV'
	gl.glVertexAttrib2fvNV.errcheck = _errcheck
	gl.glVertexAttrib2fvNV.restype = GLvoid
	gl.glVertexAttrib2fvNV.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib2hNV = _getProcAddress('glVertexAttrib2hNV')
	gl.glVertexAttrib2hNV.gl = gl
	gl.glVertexAttrib2hNV.__name__ = 'glVertexAttrib2hNV'
	gl.glVertexAttrib2hNV.errcheck = _errcheck
	gl.glVertexAttrib2hNV.restype = GLvoid
	gl.glVertexAttrib2hNV.argtypes = (GLuint, GLhalfNV, GLhalfNV,)
	gl.glVertexAttrib2hvNV = _getProcAddress('glVertexAttrib2hvNV')
	gl.glVertexAttrib2hvNV.gl = gl
	gl.glVertexAttrib2hvNV.__name__ = 'glVertexAttrib2hvNV'
	gl.glVertexAttrib2hvNV.errcheck = _errcheck
	gl.glVertexAttrib2hvNV.restype = GLvoid
	gl.glVertexAttrib2hvNV.argtypes = (GLuint, POINTER(GLhalfNV),)
	gl.glVertexAttrib2s = _getProcAddress('glVertexAttrib2s')
	gl.glVertexAttrib2s.gl = gl
	gl.glVertexAttrib2s.__name__ = 'glVertexAttrib2s'
	gl.glVertexAttrib2s.errcheck = _errcheck
	gl.glVertexAttrib2s.restype = GLvoid
	gl.glVertexAttrib2s.argtypes = (GLuint, GLshort, GLshort,)
	gl.glVertexAttrib2sARB = _getProcAddress('glVertexAttrib2sARB')
	gl.glVertexAttrib2sARB.gl = gl
	gl.glVertexAttrib2sARB.__name__ = 'glVertexAttrib2sARB'
	gl.glVertexAttrib2sARB.errcheck = _errcheck
	gl.glVertexAttrib2sARB.restype = GLvoid
	gl.glVertexAttrib2sARB.argtypes = (GLuint, GLshort, GLshort,)
	gl.glVertexAttrib2sNV = _getProcAddress('glVertexAttrib2sNV')
	gl.glVertexAttrib2sNV.gl = gl
	gl.glVertexAttrib2sNV.__name__ = 'glVertexAttrib2sNV'
	gl.glVertexAttrib2sNV.errcheck = _errcheck
	gl.glVertexAttrib2sNV.restype = GLvoid
	gl.glVertexAttrib2sNV.argtypes = (GLuint, GLshort, GLshort,)
	gl.glVertexAttrib2sv = _getProcAddress('glVertexAttrib2sv')
	gl.glVertexAttrib2sv.gl = gl
	gl.glVertexAttrib2sv.__name__ = 'glVertexAttrib2sv'
	gl.glVertexAttrib2sv.errcheck = _errcheck
	gl.glVertexAttrib2sv.restype = GLvoid
	gl.glVertexAttrib2sv.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib2svARB = _getProcAddress('glVertexAttrib2svARB')
	gl.glVertexAttrib2svARB.gl = gl
	gl.glVertexAttrib2svARB.__name__ = 'glVertexAttrib2svARB'
	gl.glVertexAttrib2svARB.errcheck = _errcheck
	gl.glVertexAttrib2svARB.restype = GLvoid
	gl.glVertexAttrib2svARB.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib2svNV = _getProcAddress('glVertexAttrib2svNV')
	gl.glVertexAttrib2svNV.gl = gl
	gl.glVertexAttrib2svNV.__name__ = 'glVertexAttrib2svNV'
	gl.glVertexAttrib2svNV.errcheck = _errcheck
	gl.glVertexAttrib2svNV.restype = GLvoid
	gl.glVertexAttrib2svNV.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib3d = _getProcAddress('glVertexAttrib3d')
	gl.glVertexAttrib3d.gl = gl
	gl.glVertexAttrib3d.__name__ = 'glVertexAttrib3d'
	gl.glVertexAttrib3d.errcheck = _errcheck
	gl.glVertexAttrib3d.restype = GLvoid
	gl.glVertexAttrib3d.argtypes = (GLuint, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttrib3dARB = _getProcAddress('glVertexAttrib3dARB')
	gl.glVertexAttrib3dARB.gl = gl
	gl.glVertexAttrib3dARB.__name__ = 'glVertexAttrib3dARB'
	gl.glVertexAttrib3dARB.errcheck = _errcheck
	gl.glVertexAttrib3dARB.restype = GLvoid
	gl.glVertexAttrib3dARB.argtypes = (GLuint, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttrib3dNV = _getProcAddress('glVertexAttrib3dNV')
	gl.glVertexAttrib3dNV.gl = gl
	gl.glVertexAttrib3dNV.__name__ = 'glVertexAttrib3dNV'
	gl.glVertexAttrib3dNV.errcheck = _errcheck
	gl.glVertexAttrib3dNV.restype = GLvoid
	gl.glVertexAttrib3dNV.argtypes = (GLuint, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttrib3dv = _getProcAddress('glVertexAttrib3dv')
	gl.glVertexAttrib3dv.gl = gl
	gl.glVertexAttrib3dv.__name__ = 'glVertexAttrib3dv'
	gl.glVertexAttrib3dv.errcheck = _errcheck
	gl.glVertexAttrib3dv.restype = GLvoid
	gl.glVertexAttrib3dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib3dvARB = _getProcAddress('glVertexAttrib3dvARB')
	gl.glVertexAttrib3dvARB.gl = gl
	gl.glVertexAttrib3dvARB.__name__ = 'glVertexAttrib3dvARB'
	gl.glVertexAttrib3dvARB.errcheck = _errcheck
	gl.glVertexAttrib3dvARB.restype = GLvoid
	gl.glVertexAttrib3dvARB.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib3dvNV = _getProcAddress('glVertexAttrib3dvNV')
	gl.glVertexAttrib3dvNV.gl = gl
	gl.glVertexAttrib3dvNV.__name__ = 'glVertexAttrib3dvNV'
	gl.glVertexAttrib3dvNV.errcheck = _errcheck
	gl.glVertexAttrib3dvNV.restype = GLvoid
	gl.glVertexAttrib3dvNV.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib3f = _getProcAddress('glVertexAttrib3f')
	gl.glVertexAttrib3f.gl = gl
	gl.glVertexAttrib3f.__name__ = 'glVertexAttrib3f'
	gl.glVertexAttrib3f.errcheck = _errcheck
	gl.glVertexAttrib3f.restype = GLvoid
	gl.glVertexAttrib3f.argtypes = (GLuint, GLfloat, GLfloat, GLfloat,)
	gl.glVertexAttrib3fARB = _getProcAddress('glVertexAttrib3fARB')
	gl.glVertexAttrib3fARB.gl = gl
	gl.glVertexAttrib3fARB.__name__ = 'glVertexAttrib3fARB'
	gl.glVertexAttrib3fARB.errcheck = _errcheck
	gl.glVertexAttrib3fARB.restype = GLvoid
	gl.glVertexAttrib3fARB.argtypes = (GLuint, GLfloat, GLfloat, GLfloat,)
	gl.glVertexAttrib3fNV = _getProcAddress('glVertexAttrib3fNV')
	gl.glVertexAttrib3fNV.gl = gl
	gl.glVertexAttrib3fNV.__name__ = 'glVertexAttrib3fNV'
	gl.glVertexAttrib3fNV.errcheck = _errcheck
	gl.glVertexAttrib3fNV.restype = GLvoid
	gl.glVertexAttrib3fNV.argtypes = (GLuint, GLfloat, GLfloat, GLfloat,)
	gl.glVertexAttrib3fv = _getProcAddress('glVertexAttrib3fv')
	gl.glVertexAttrib3fv.gl = gl
	gl.glVertexAttrib3fv.__name__ = 'glVertexAttrib3fv'
	gl.glVertexAttrib3fv.errcheck = _errcheck
	gl.glVertexAttrib3fv.restype = GLvoid
	gl.glVertexAttrib3fv.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib3fvARB = _getProcAddress('glVertexAttrib3fvARB')
	gl.glVertexAttrib3fvARB.gl = gl
	gl.glVertexAttrib3fvARB.__name__ = 'glVertexAttrib3fvARB'
	gl.glVertexAttrib3fvARB.errcheck = _errcheck
	gl.glVertexAttrib3fvARB.restype = GLvoid
	gl.glVertexAttrib3fvARB.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib3fvNV = _getProcAddress('glVertexAttrib3fvNV')
	gl.glVertexAttrib3fvNV.gl = gl
	gl.glVertexAttrib3fvNV.__name__ = 'glVertexAttrib3fvNV'
	gl.glVertexAttrib3fvNV.errcheck = _errcheck
	gl.glVertexAttrib3fvNV.restype = GLvoid
	gl.glVertexAttrib3fvNV.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib3hNV = _getProcAddress('glVertexAttrib3hNV')
	gl.glVertexAttrib3hNV.gl = gl
	gl.glVertexAttrib3hNV.__name__ = 'glVertexAttrib3hNV'
	gl.glVertexAttrib3hNV.errcheck = _errcheck
	gl.glVertexAttrib3hNV.restype = GLvoid
	gl.glVertexAttrib3hNV.argtypes = (GLuint, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glVertexAttrib3hvNV = _getProcAddress('glVertexAttrib3hvNV')
	gl.glVertexAttrib3hvNV.gl = gl
	gl.glVertexAttrib3hvNV.__name__ = 'glVertexAttrib3hvNV'
	gl.glVertexAttrib3hvNV.errcheck = _errcheck
	gl.glVertexAttrib3hvNV.restype = GLvoid
	gl.glVertexAttrib3hvNV.argtypes = (GLuint, POINTER(GLhalfNV),)
	gl.glVertexAttrib3s = _getProcAddress('glVertexAttrib3s')
	gl.glVertexAttrib3s.gl = gl
	gl.glVertexAttrib3s.__name__ = 'glVertexAttrib3s'
	gl.glVertexAttrib3s.errcheck = _errcheck
	gl.glVertexAttrib3s.restype = GLvoid
	gl.glVertexAttrib3s.argtypes = (GLuint, GLshort, GLshort, GLshort,)
	gl.glVertexAttrib3sARB = _getProcAddress('glVertexAttrib3sARB')
	gl.glVertexAttrib3sARB.gl = gl
	gl.glVertexAttrib3sARB.__name__ = 'glVertexAttrib3sARB'
	gl.glVertexAttrib3sARB.errcheck = _errcheck
	gl.glVertexAttrib3sARB.restype = GLvoid
	gl.glVertexAttrib3sARB.argtypes = (GLuint, GLshort, GLshort, GLshort,)
	gl.glVertexAttrib3sNV = _getProcAddress('glVertexAttrib3sNV')
	gl.glVertexAttrib3sNV.gl = gl
	gl.glVertexAttrib3sNV.__name__ = 'glVertexAttrib3sNV'
	gl.glVertexAttrib3sNV.errcheck = _errcheck
	gl.glVertexAttrib3sNV.restype = GLvoid
	gl.glVertexAttrib3sNV.argtypes = (GLuint, GLshort, GLshort, GLshort,)
	gl.glVertexAttrib3sv = _getProcAddress('glVertexAttrib3sv')
	gl.glVertexAttrib3sv.gl = gl
	gl.glVertexAttrib3sv.__name__ = 'glVertexAttrib3sv'
	gl.glVertexAttrib3sv.errcheck = _errcheck
	gl.glVertexAttrib3sv.restype = GLvoid
	gl.glVertexAttrib3sv.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib3svARB = _getProcAddress('glVertexAttrib3svARB')
	gl.glVertexAttrib3svARB.gl = gl
	gl.glVertexAttrib3svARB.__name__ = 'glVertexAttrib3svARB'
	gl.glVertexAttrib3svARB.errcheck = _errcheck
	gl.glVertexAttrib3svARB.restype = GLvoid
	gl.glVertexAttrib3svARB.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib3svNV = _getProcAddress('glVertexAttrib3svNV')
	gl.glVertexAttrib3svNV.gl = gl
	gl.glVertexAttrib3svNV.__name__ = 'glVertexAttrib3svNV'
	gl.glVertexAttrib3svNV.errcheck = _errcheck
	gl.glVertexAttrib3svNV.restype = GLvoid
	gl.glVertexAttrib3svNV.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib4Nbv = _getProcAddress('glVertexAttrib4Nbv')
	gl.glVertexAttrib4Nbv.gl = gl
	gl.glVertexAttrib4Nbv.__name__ = 'glVertexAttrib4Nbv'
	gl.glVertexAttrib4Nbv.errcheck = _errcheck
	gl.glVertexAttrib4Nbv.restype = GLvoid
	gl.glVertexAttrib4Nbv.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVertexAttrib4NbvARB = _getProcAddress('glVertexAttrib4NbvARB')
	gl.glVertexAttrib4NbvARB.gl = gl
	gl.glVertexAttrib4NbvARB.__name__ = 'glVertexAttrib4NbvARB'
	gl.glVertexAttrib4NbvARB.errcheck = _errcheck
	gl.glVertexAttrib4NbvARB.restype = GLvoid
	gl.glVertexAttrib4NbvARB.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVertexAttrib4Niv = _getProcAddress('glVertexAttrib4Niv')
	gl.glVertexAttrib4Niv.gl = gl
	gl.glVertexAttrib4Niv.__name__ = 'glVertexAttrib4Niv'
	gl.glVertexAttrib4Niv.errcheck = _errcheck
	gl.glVertexAttrib4Niv.restype = GLvoid
	gl.glVertexAttrib4Niv.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttrib4NivARB = _getProcAddress('glVertexAttrib4NivARB')
	gl.glVertexAttrib4NivARB.gl = gl
	gl.glVertexAttrib4NivARB.__name__ = 'glVertexAttrib4NivARB'
	gl.glVertexAttrib4NivARB.errcheck = _errcheck
	gl.glVertexAttrib4NivARB.restype = GLvoid
	gl.glVertexAttrib4NivARB.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttrib4Nsv = _getProcAddress('glVertexAttrib4Nsv')
	gl.glVertexAttrib4Nsv.gl = gl
	gl.glVertexAttrib4Nsv.__name__ = 'glVertexAttrib4Nsv'
	gl.glVertexAttrib4Nsv.errcheck = _errcheck
	gl.glVertexAttrib4Nsv.restype = GLvoid
	gl.glVertexAttrib4Nsv.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib4NsvARB = _getProcAddress('glVertexAttrib4NsvARB')
	gl.glVertexAttrib4NsvARB.gl = gl
	gl.glVertexAttrib4NsvARB.__name__ = 'glVertexAttrib4NsvARB'
	gl.glVertexAttrib4NsvARB.errcheck = _errcheck
	gl.glVertexAttrib4NsvARB.restype = GLvoid
	gl.glVertexAttrib4NsvARB.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib4Nub = _getProcAddress('glVertexAttrib4Nub')
	gl.glVertexAttrib4Nub.gl = gl
	gl.glVertexAttrib4Nub.__name__ = 'glVertexAttrib4Nub'
	gl.glVertexAttrib4Nub.errcheck = _errcheck
	gl.glVertexAttrib4Nub.restype = GLvoid
	gl.glVertexAttrib4Nub.argtypes = (GLuint, GLubyte, GLubyte, GLubyte, GLubyte,)
	gl.glVertexAttrib4NubARB = _getProcAddress('glVertexAttrib4NubARB')
	gl.glVertexAttrib4NubARB.gl = gl
	gl.glVertexAttrib4NubARB.__name__ = 'glVertexAttrib4NubARB'
	gl.glVertexAttrib4NubARB.errcheck = _errcheck
	gl.glVertexAttrib4NubARB.restype = GLvoid
	gl.glVertexAttrib4NubARB.argtypes = (GLuint, GLubyte, GLubyte, GLubyte, GLubyte,)
	gl.glVertexAttrib4Nubv = _getProcAddress('glVertexAttrib4Nubv')
	gl.glVertexAttrib4Nubv.gl = gl
	gl.glVertexAttrib4Nubv.__name__ = 'glVertexAttrib4Nubv'
	gl.glVertexAttrib4Nubv.errcheck = _errcheck
	gl.glVertexAttrib4Nubv.restype = GLvoid
	gl.glVertexAttrib4Nubv.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttrib4NubvARB = _getProcAddress('glVertexAttrib4NubvARB')
	gl.glVertexAttrib4NubvARB.gl = gl
	gl.glVertexAttrib4NubvARB.__name__ = 'glVertexAttrib4NubvARB'
	gl.glVertexAttrib4NubvARB.errcheck = _errcheck
	gl.glVertexAttrib4NubvARB.restype = GLvoid
	gl.glVertexAttrib4NubvARB.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttrib4Nuiv = _getProcAddress('glVertexAttrib4Nuiv')
	gl.glVertexAttrib4Nuiv.gl = gl
	gl.glVertexAttrib4Nuiv.__name__ = 'glVertexAttrib4Nuiv'
	gl.glVertexAttrib4Nuiv.errcheck = _errcheck
	gl.glVertexAttrib4Nuiv.restype = GLvoid
	gl.glVertexAttrib4Nuiv.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttrib4NuivARB = _getProcAddress('glVertexAttrib4NuivARB')
	gl.glVertexAttrib4NuivARB.gl = gl
	gl.glVertexAttrib4NuivARB.__name__ = 'glVertexAttrib4NuivARB'
	gl.glVertexAttrib4NuivARB.errcheck = _errcheck
	gl.glVertexAttrib4NuivARB.restype = GLvoid
	gl.glVertexAttrib4NuivARB.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttrib4Nusv = _getProcAddress('glVertexAttrib4Nusv')
	gl.glVertexAttrib4Nusv.gl = gl
	gl.glVertexAttrib4Nusv.__name__ = 'glVertexAttrib4Nusv'
	gl.glVertexAttrib4Nusv.errcheck = _errcheck
	gl.glVertexAttrib4Nusv.restype = GLvoid
	gl.glVertexAttrib4Nusv.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertexAttrib4NusvARB = _getProcAddress('glVertexAttrib4NusvARB')
	gl.glVertexAttrib4NusvARB.gl = gl
	gl.glVertexAttrib4NusvARB.__name__ = 'glVertexAttrib4NusvARB'
	gl.glVertexAttrib4NusvARB.errcheck = _errcheck
	gl.glVertexAttrib4NusvARB.restype = GLvoid
	gl.glVertexAttrib4NusvARB.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertexAttrib4bv = _getProcAddress('glVertexAttrib4bv')
	gl.glVertexAttrib4bv.gl = gl
	gl.glVertexAttrib4bv.__name__ = 'glVertexAttrib4bv'
	gl.glVertexAttrib4bv.errcheck = _errcheck
	gl.glVertexAttrib4bv.restype = GLvoid
	gl.glVertexAttrib4bv.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVertexAttrib4bvARB = _getProcAddress('glVertexAttrib4bvARB')
	gl.glVertexAttrib4bvARB.gl = gl
	gl.glVertexAttrib4bvARB.__name__ = 'glVertexAttrib4bvARB'
	gl.glVertexAttrib4bvARB.errcheck = _errcheck
	gl.glVertexAttrib4bvARB.restype = GLvoid
	gl.glVertexAttrib4bvARB.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVertexAttrib4d = _getProcAddress('glVertexAttrib4d')
	gl.glVertexAttrib4d.gl = gl
	gl.glVertexAttrib4d.__name__ = 'glVertexAttrib4d'
	gl.glVertexAttrib4d.errcheck = _errcheck
	gl.glVertexAttrib4d.restype = GLvoid
	gl.glVertexAttrib4d.argtypes = (GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttrib4dARB = _getProcAddress('glVertexAttrib4dARB')
	gl.glVertexAttrib4dARB.gl = gl
	gl.glVertexAttrib4dARB.__name__ = 'glVertexAttrib4dARB'
	gl.glVertexAttrib4dARB.errcheck = _errcheck
	gl.glVertexAttrib4dARB.restype = GLvoid
	gl.glVertexAttrib4dARB.argtypes = (GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttrib4dNV = _getProcAddress('glVertexAttrib4dNV')
	gl.glVertexAttrib4dNV.gl = gl
	gl.glVertexAttrib4dNV.__name__ = 'glVertexAttrib4dNV'
	gl.glVertexAttrib4dNV.errcheck = _errcheck
	gl.glVertexAttrib4dNV.restype = GLvoid
	gl.glVertexAttrib4dNV.argtypes = (GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttrib4dv = _getProcAddress('glVertexAttrib4dv')
	gl.glVertexAttrib4dv.gl = gl
	gl.glVertexAttrib4dv.__name__ = 'glVertexAttrib4dv'
	gl.glVertexAttrib4dv.errcheck = _errcheck
	gl.glVertexAttrib4dv.restype = GLvoid
	gl.glVertexAttrib4dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib4dvARB = _getProcAddress('glVertexAttrib4dvARB')
	gl.glVertexAttrib4dvARB.gl = gl
	gl.glVertexAttrib4dvARB.__name__ = 'glVertexAttrib4dvARB'
	gl.glVertexAttrib4dvARB.errcheck = _errcheck
	gl.glVertexAttrib4dvARB.restype = GLvoid
	gl.glVertexAttrib4dvARB.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib4dvNV = _getProcAddress('glVertexAttrib4dvNV')
	gl.glVertexAttrib4dvNV.gl = gl
	gl.glVertexAttrib4dvNV.__name__ = 'glVertexAttrib4dvNV'
	gl.glVertexAttrib4dvNV.errcheck = _errcheck
	gl.glVertexAttrib4dvNV.restype = GLvoid
	gl.glVertexAttrib4dvNV.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttrib4f = _getProcAddress('glVertexAttrib4f')
	gl.glVertexAttrib4f.gl = gl
	gl.glVertexAttrib4f.__name__ = 'glVertexAttrib4f'
	gl.glVertexAttrib4f.errcheck = _errcheck
	gl.glVertexAttrib4f.restype = GLvoid
	gl.glVertexAttrib4f.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glVertexAttrib4fARB = _getProcAddress('glVertexAttrib4fARB')
	gl.glVertexAttrib4fARB.gl = gl
	gl.glVertexAttrib4fARB.__name__ = 'glVertexAttrib4fARB'
	gl.glVertexAttrib4fARB.errcheck = _errcheck
	gl.glVertexAttrib4fARB.restype = GLvoid
	gl.glVertexAttrib4fARB.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glVertexAttrib4fNV = _getProcAddress('glVertexAttrib4fNV')
	gl.glVertexAttrib4fNV.gl = gl
	gl.glVertexAttrib4fNV.__name__ = 'glVertexAttrib4fNV'
	gl.glVertexAttrib4fNV.errcheck = _errcheck
	gl.glVertexAttrib4fNV.restype = GLvoid
	gl.glVertexAttrib4fNV.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glVertexAttrib4fv = _getProcAddress('glVertexAttrib4fv')
	gl.glVertexAttrib4fv.gl = gl
	gl.glVertexAttrib4fv.__name__ = 'glVertexAttrib4fv'
	gl.glVertexAttrib4fv.errcheck = _errcheck
	gl.glVertexAttrib4fv.restype = GLvoid
	gl.glVertexAttrib4fv.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib4fvARB = _getProcAddress('glVertexAttrib4fvARB')
	gl.glVertexAttrib4fvARB.gl = gl
	gl.glVertexAttrib4fvARB.__name__ = 'glVertexAttrib4fvARB'
	gl.glVertexAttrib4fvARB.errcheck = _errcheck
	gl.glVertexAttrib4fvARB.restype = GLvoid
	gl.glVertexAttrib4fvARB.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib4fvNV = _getProcAddress('glVertexAttrib4fvNV')
	gl.glVertexAttrib4fvNV.gl = gl
	gl.glVertexAttrib4fvNV.__name__ = 'glVertexAttrib4fvNV'
	gl.glVertexAttrib4fvNV.errcheck = _errcheck
	gl.glVertexAttrib4fvNV.restype = GLvoid
	gl.glVertexAttrib4fvNV.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glVertexAttrib4hNV = _getProcAddress('glVertexAttrib4hNV')
	gl.glVertexAttrib4hNV.gl = gl
	gl.glVertexAttrib4hNV.__name__ = 'glVertexAttrib4hNV'
	gl.glVertexAttrib4hNV.errcheck = _errcheck
	gl.glVertexAttrib4hNV.restype = GLvoid
	gl.glVertexAttrib4hNV.argtypes = (GLuint, GLhalfNV, GLhalfNV, GLhalfNV, GLhalfNV,)
	gl.glVertexAttrib4hvNV = _getProcAddress('glVertexAttrib4hvNV')
	gl.glVertexAttrib4hvNV.gl = gl
	gl.glVertexAttrib4hvNV.__name__ = 'glVertexAttrib4hvNV'
	gl.glVertexAttrib4hvNV.errcheck = _errcheck
	gl.glVertexAttrib4hvNV.restype = GLvoid
	gl.glVertexAttrib4hvNV.argtypes = (GLuint, POINTER(GLhalfNV),)
	gl.glVertexAttrib4iv = _getProcAddress('glVertexAttrib4iv')
	gl.glVertexAttrib4iv.gl = gl
	gl.glVertexAttrib4iv.__name__ = 'glVertexAttrib4iv'
	gl.glVertexAttrib4iv.errcheck = _errcheck
	gl.glVertexAttrib4iv.restype = GLvoid
	gl.glVertexAttrib4iv.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttrib4ivARB = _getProcAddress('glVertexAttrib4ivARB')
	gl.glVertexAttrib4ivARB.gl = gl
	gl.glVertexAttrib4ivARB.__name__ = 'glVertexAttrib4ivARB'
	gl.glVertexAttrib4ivARB.errcheck = _errcheck
	gl.glVertexAttrib4ivARB.restype = GLvoid
	gl.glVertexAttrib4ivARB.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttrib4s = _getProcAddress('glVertexAttrib4s')
	gl.glVertexAttrib4s.gl = gl
	gl.glVertexAttrib4s.__name__ = 'glVertexAttrib4s'
	gl.glVertexAttrib4s.errcheck = _errcheck
	gl.glVertexAttrib4s.restype = GLvoid
	gl.glVertexAttrib4s.argtypes = (GLuint, GLshort, GLshort, GLshort, GLshort,)
	gl.glVertexAttrib4sARB = _getProcAddress('glVertexAttrib4sARB')
	gl.glVertexAttrib4sARB.gl = gl
	gl.glVertexAttrib4sARB.__name__ = 'glVertexAttrib4sARB'
	gl.glVertexAttrib4sARB.errcheck = _errcheck
	gl.glVertexAttrib4sARB.restype = GLvoid
	gl.glVertexAttrib4sARB.argtypes = (GLuint, GLshort, GLshort, GLshort, GLshort,)
	gl.glVertexAttrib4sNV = _getProcAddress('glVertexAttrib4sNV')
	gl.glVertexAttrib4sNV.gl = gl
	gl.glVertexAttrib4sNV.__name__ = 'glVertexAttrib4sNV'
	gl.glVertexAttrib4sNV.errcheck = _errcheck
	gl.glVertexAttrib4sNV.restype = GLvoid
	gl.glVertexAttrib4sNV.argtypes = (GLuint, GLshort, GLshort, GLshort, GLshort,)
	gl.glVertexAttrib4sv = _getProcAddress('glVertexAttrib4sv')
	gl.glVertexAttrib4sv.gl = gl
	gl.glVertexAttrib4sv.__name__ = 'glVertexAttrib4sv'
	gl.glVertexAttrib4sv.errcheck = _errcheck
	gl.glVertexAttrib4sv.restype = GLvoid
	gl.glVertexAttrib4sv.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib4svARB = _getProcAddress('glVertexAttrib4svARB')
	gl.glVertexAttrib4svARB.gl = gl
	gl.glVertexAttrib4svARB.__name__ = 'glVertexAttrib4svARB'
	gl.glVertexAttrib4svARB.errcheck = _errcheck
	gl.glVertexAttrib4svARB.restype = GLvoid
	gl.glVertexAttrib4svARB.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib4svNV = _getProcAddress('glVertexAttrib4svNV')
	gl.glVertexAttrib4svNV.gl = gl
	gl.glVertexAttrib4svNV.__name__ = 'glVertexAttrib4svNV'
	gl.glVertexAttrib4svNV.errcheck = _errcheck
	gl.glVertexAttrib4svNV.restype = GLvoid
	gl.glVertexAttrib4svNV.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttrib4ubNV = _getProcAddress('glVertexAttrib4ubNV')
	gl.glVertexAttrib4ubNV.gl = gl
	gl.glVertexAttrib4ubNV.__name__ = 'glVertexAttrib4ubNV'
	gl.glVertexAttrib4ubNV.errcheck = _errcheck
	gl.glVertexAttrib4ubNV.restype = GLvoid
	gl.glVertexAttrib4ubNV.argtypes = (GLuint, GLubyte, GLubyte, GLubyte, GLubyte,)
	gl.glVertexAttrib4ubv = _getProcAddress('glVertexAttrib4ubv')
	gl.glVertexAttrib4ubv.gl = gl
	gl.glVertexAttrib4ubv.__name__ = 'glVertexAttrib4ubv'
	gl.glVertexAttrib4ubv.errcheck = _errcheck
	gl.glVertexAttrib4ubv.restype = GLvoid
	gl.glVertexAttrib4ubv.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttrib4ubvARB = _getProcAddress('glVertexAttrib4ubvARB')
	gl.glVertexAttrib4ubvARB.gl = gl
	gl.glVertexAttrib4ubvARB.__name__ = 'glVertexAttrib4ubvARB'
	gl.glVertexAttrib4ubvARB.errcheck = _errcheck
	gl.glVertexAttrib4ubvARB.restype = GLvoid
	gl.glVertexAttrib4ubvARB.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttrib4ubvNV = _getProcAddress('glVertexAttrib4ubvNV')
	gl.glVertexAttrib4ubvNV.gl = gl
	gl.glVertexAttrib4ubvNV.__name__ = 'glVertexAttrib4ubvNV'
	gl.glVertexAttrib4ubvNV.errcheck = _errcheck
	gl.glVertexAttrib4ubvNV.restype = GLvoid
	gl.glVertexAttrib4ubvNV.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttrib4uiv = _getProcAddress('glVertexAttrib4uiv')
	gl.glVertexAttrib4uiv.gl = gl
	gl.glVertexAttrib4uiv.__name__ = 'glVertexAttrib4uiv'
	gl.glVertexAttrib4uiv.errcheck = _errcheck
	gl.glVertexAttrib4uiv.restype = GLvoid
	gl.glVertexAttrib4uiv.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttrib4uivARB = _getProcAddress('glVertexAttrib4uivARB')
	gl.glVertexAttrib4uivARB.gl = gl
	gl.glVertexAttrib4uivARB.__name__ = 'glVertexAttrib4uivARB'
	gl.glVertexAttrib4uivARB.errcheck = _errcheck
	gl.glVertexAttrib4uivARB.restype = GLvoid
	gl.glVertexAttrib4uivARB.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttrib4usv = _getProcAddress('glVertexAttrib4usv')
	gl.glVertexAttrib4usv.gl = gl
	gl.glVertexAttrib4usv.__name__ = 'glVertexAttrib4usv'
	gl.glVertexAttrib4usv.errcheck = _errcheck
	gl.glVertexAttrib4usv.restype = GLvoid
	gl.glVertexAttrib4usv.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertexAttrib4usvARB = _getProcAddress('glVertexAttrib4usvARB')
	gl.glVertexAttrib4usvARB.gl = gl
	gl.glVertexAttrib4usvARB.__name__ = 'glVertexAttrib4usvARB'
	gl.glVertexAttrib4usvARB.errcheck = _errcheck
	gl.glVertexAttrib4usvARB.restype = GLvoid
	gl.glVertexAttrib4usvARB.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertexAttribArrayObjectATI = _getProcAddress('glVertexAttribArrayObjectATI')
	gl.glVertexAttribArrayObjectATI.gl = gl
	gl.glVertexAttribArrayObjectATI.__name__ = 'glVertexAttribArrayObjectATI'
	gl.glVertexAttribArrayObjectATI.errcheck = _errcheck
	gl.glVertexAttribArrayObjectATI.restype = GLvoid
	gl.glVertexAttribArrayObjectATI.argtypes = (GLuint, GLint, GLenum, GLboolean, GLsizei, GLuint, GLuint,)
	gl.glVertexAttribBinding = _getProcAddress('glVertexAttribBinding')
	gl.glVertexAttribBinding.gl = gl
	gl.glVertexAttribBinding.__name__ = 'glVertexAttribBinding'
	gl.glVertexAttribBinding.errcheck = _errcheck
	gl.glVertexAttribBinding.restype = GLvoid
	gl.glVertexAttribBinding.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribDivisor = _getProcAddress('glVertexAttribDivisor')
	gl.glVertexAttribDivisor.gl = gl
	gl.glVertexAttribDivisor.__name__ = 'glVertexAttribDivisor'
	gl.glVertexAttribDivisor.errcheck = _errcheck
	gl.glVertexAttribDivisor.restype = GLvoid
	gl.glVertexAttribDivisor.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribDivisorANGLE = _getProcAddress('glVertexAttribDivisorANGLE')
	gl.glVertexAttribDivisorANGLE.gl = gl
	gl.glVertexAttribDivisorANGLE.__name__ = 'glVertexAttribDivisorANGLE'
	gl.glVertexAttribDivisorANGLE.errcheck = _errcheck
	gl.glVertexAttribDivisorANGLE.restype = GLvoid
	gl.glVertexAttribDivisorANGLE.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribDivisorARB = _getProcAddress('glVertexAttribDivisorARB')
	gl.glVertexAttribDivisorARB.gl = gl
	gl.glVertexAttribDivisorARB.__name__ = 'glVertexAttribDivisorARB'
	gl.glVertexAttribDivisorARB.errcheck = _errcheck
	gl.glVertexAttribDivisorARB.restype = GLvoid
	gl.glVertexAttribDivisorARB.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribDivisorEXT = _getProcAddress('glVertexAttribDivisorEXT')
	gl.glVertexAttribDivisorEXT.gl = gl
	gl.glVertexAttribDivisorEXT.__name__ = 'glVertexAttribDivisorEXT'
	gl.glVertexAttribDivisorEXT.errcheck = _errcheck
	gl.glVertexAttribDivisorEXT.restype = GLvoid
	gl.glVertexAttribDivisorEXT.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribDivisorNV = _getProcAddress('glVertexAttribDivisorNV')
	gl.glVertexAttribDivisorNV.gl = gl
	gl.glVertexAttribDivisorNV.__name__ = 'glVertexAttribDivisorNV'
	gl.glVertexAttribDivisorNV.errcheck = _errcheck
	gl.glVertexAttribDivisorNV.restype = GLvoid
	gl.glVertexAttribDivisorNV.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribFormat = _getProcAddress('glVertexAttribFormat')
	gl.glVertexAttribFormat.gl = gl
	gl.glVertexAttribFormat.__name__ = 'glVertexAttribFormat'
	gl.glVertexAttribFormat.errcheck = _errcheck
	gl.glVertexAttribFormat.restype = GLvoid
	gl.glVertexAttribFormat.argtypes = (GLuint, GLint, GLenum, GLboolean, GLuint,)
	gl.glVertexAttribFormatNV = _getProcAddress('glVertexAttribFormatNV')
	gl.glVertexAttribFormatNV.gl = gl
	gl.glVertexAttribFormatNV.__name__ = 'glVertexAttribFormatNV'
	gl.glVertexAttribFormatNV.errcheck = _errcheck
	gl.glVertexAttribFormatNV.restype = GLvoid
	gl.glVertexAttribFormatNV.argtypes = (GLuint, GLint, GLenum, GLboolean, GLsizei,)
	gl.glVertexAttribI1i = _getProcAddress('glVertexAttribI1i')
	gl.glVertexAttribI1i.gl = gl
	gl.glVertexAttribI1i.__name__ = 'glVertexAttribI1i'
	gl.glVertexAttribI1i.errcheck = _errcheck
	gl.glVertexAttribI1i.restype = GLvoid
	gl.glVertexAttribI1i.argtypes = (GLuint, GLint,)
	gl.glVertexAttribI1iEXT = _getProcAddress('glVertexAttribI1iEXT')
	gl.glVertexAttribI1iEXT.gl = gl
	gl.glVertexAttribI1iEXT.__name__ = 'glVertexAttribI1iEXT'
	gl.glVertexAttribI1iEXT.errcheck = _errcheck
	gl.glVertexAttribI1iEXT.restype = GLvoid
	gl.glVertexAttribI1iEXT.argtypes = (GLuint, GLint,)
	gl.glVertexAttribI1iv = _getProcAddress('glVertexAttribI1iv')
	gl.glVertexAttribI1iv.gl = gl
	gl.glVertexAttribI1iv.__name__ = 'glVertexAttribI1iv'
	gl.glVertexAttribI1iv.errcheck = _errcheck
	gl.glVertexAttribI1iv.restype = GLvoid
	gl.glVertexAttribI1iv.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI1ivEXT = _getProcAddress('glVertexAttribI1ivEXT')
	gl.glVertexAttribI1ivEXT.gl = gl
	gl.glVertexAttribI1ivEXT.__name__ = 'glVertexAttribI1ivEXT'
	gl.glVertexAttribI1ivEXT.errcheck = _errcheck
	gl.glVertexAttribI1ivEXT.restype = GLvoid
	gl.glVertexAttribI1ivEXT.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI1ui = _getProcAddress('glVertexAttribI1ui')
	gl.glVertexAttribI1ui.gl = gl
	gl.glVertexAttribI1ui.__name__ = 'glVertexAttribI1ui'
	gl.glVertexAttribI1ui.errcheck = _errcheck
	gl.glVertexAttribI1ui.restype = GLvoid
	gl.glVertexAttribI1ui.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribI1uiEXT = _getProcAddress('glVertexAttribI1uiEXT')
	gl.glVertexAttribI1uiEXT.gl = gl
	gl.glVertexAttribI1uiEXT.__name__ = 'glVertexAttribI1uiEXT'
	gl.glVertexAttribI1uiEXT.errcheck = _errcheck
	gl.glVertexAttribI1uiEXT.restype = GLvoid
	gl.glVertexAttribI1uiEXT.argtypes = (GLuint, GLuint,)
	gl.glVertexAttribI1uiv = _getProcAddress('glVertexAttribI1uiv')
	gl.glVertexAttribI1uiv.gl = gl
	gl.glVertexAttribI1uiv.__name__ = 'glVertexAttribI1uiv'
	gl.glVertexAttribI1uiv.errcheck = _errcheck
	gl.glVertexAttribI1uiv.restype = GLvoid
	gl.glVertexAttribI1uiv.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI1uivEXT = _getProcAddress('glVertexAttribI1uivEXT')
	gl.glVertexAttribI1uivEXT.gl = gl
	gl.glVertexAttribI1uivEXT.__name__ = 'glVertexAttribI1uivEXT'
	gl.glVertexAttribI1uivEXT.errcheck = _errcheck
	gl.glVertexAttribI1uivEXT.restype = GLvoid
	gl.glVertexAttribI1uivEXT.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI2i = _getProcAddress('glVertexAttribI2i')
	gl.glVertexAttribI2i.gl = gl
	gl.glVertexAttribI2i.__name__ = 'glVertexAttribI2i'
	gl.glVertexAttribI2i.errcheck = _errcheck
	gl.glVertexAttribI2i.restype = GLvoid
	gl.glVertexAttribI2i.argtypes = (GLuint, GLint, GLint,)
	gl.glVertexAttribI2iEXT = _getProcAddress('glVertexAttribI2iEXT')
	gl.glVertexAttribI2iEXT.gl = gl
	gl.glVertexAttribI2iEXT.__name__ = 'glVertexAttribI2iEXT'
	gl.glVertexAttribI2iEXT.errcheck = _errcheck
	gl.glVertexAttribI2iEXT.restype = GLvoid
	gl.glVertexAttribI2iEXT.argtypes = (GLuint, GLint, GLint,)
	gl.glVertexAttribI2iv = _getProcAddress('glVertexAttribI2iv')
	gl.glVertexAttribI2iv.gl = gl
	gl.glVertexAttribI2iv.__name__ = 'glVertexAttribI2iv'
	gl.glVertexAttribI2iv.errcheck = _errcheck
	gl.glVertexAttribI2iv.restype = GLvoid
	gl.glVertexAttribI2iv.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI2ivEXT = _getProcAddress('glVertexAttribI2ivEXT')
	gl.glVertexAttribI2ivEXT.gl = gl
	gl.glVertexAttribI2ivEXT.__name__ = 'glVertexAttribI2ivEXT'
	gl.glVertexAttribI2ivEXT.errcheck = _errcheck
	gl.glVertexAttribI2ivEXT.restype = GLvoid
	gl.glVertexAttribI2ivEXT.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI2ui = _getProcAddress('glVertexAttribI2ui')
	gl.glVertexAttribI2ui.gl = gl
	gl.glVertexAttribI2ui.__name__ = 'glVertexAttribI2ui'
	gl.glVertexAttribI2ui.errcheck = _errcheck
	gl.glVertexAttribI2ui.restype = GLvoid
	gl.glVertexAttribI2ui.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexAttribI2uiEXT = _getProcAddress('glVertexAttribI2uiEXT')
	gl.glVertexAttribI2uiEXT.gl = gl
	gl.glVertexAttribI2uiEXT.__name__ = 'glVertexAttribI2uiEXT'
	gl.glVertexAttribI2uiEXT.errcheck = _errcheck
	gl.glVertexAttribI2uiEXT.restype = GLvoid
	gl.glVertexAttribI2uiEXT.argtypes = (GLuint, GLuint, GLuint,)
	gl.glVertexAttribI2uiv = _getProcAddress('glVertexAttribI2uiv')
	gl.glVertexAttribI2uiv.gl = gl
	gl.glVertexAttribI2uiv.__name__ = 'glVertexAttribI2uiv'
	gl.glVertexAttribI2uiv.errcheck = _errcheck
	gl.glVertexAttribI2uiv.restype = GLvoid
	gl.glVertexAttribI2uiv.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI2uivEXT = _getProcAddress('glVertexAttribI2uivEXT')
	gl.glVertexAttribI2uivEXT.gl = gl
	gl.glVertexAttribI2uivEXT.__name__ = 'glVertexAttribI2uivEXT'
	gl.glVertexAttribI2uivEXT.errcheck = _errcheck
	gl.glVertexAttribI2uivEXT.restype = GLvoid
	gl.glVertexAttribI2uivEXT.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI3i = _getProcAddress('glVertexAttribI3i')
	gl.glVertexAttribI3i.gl = gl
	gl.glVertexAttribI3i.__name__ = 'glVertexAttribI3i'
	gl.glVertexAttribI3i.errcheck = _errcheck
	gl.glVertexAttribI3i.restype = GLvoid
	gl.glVertexAttribI3i.argtypes = (GLuint, GLint, GLint, GLint,)
	gl.glVertexAttribI3iEXT = _getProcAddress('glVertexAttribI3iEXT')
	gl.glVertexAttribI3iEXT.gl = gl
	gl.glVertexAttribI3iEXT.__name__ = 'glVertexAttribI3iEXT'
	gl.glVertexAttribI3iEXT.errcheck = _errcheck
	gl.glVertexAttribI3iEXT.restype = GLvoid
	gl.glVertexAttribI3iEXT.argtypes = (GLuint, GLint, GLint, GLint,)
	gl.glVertexAttribI3iv = _getProcAddress('glVertexAttribI3iv')
	gl.glVertexAttribI3iv.gl = gl
	gl.glVertexAttribI3iv.__name__ = 'glVertexAttribI3iv'
	gl.glVertexAttribI3iv.errcheck = _errcheck
	gl.glVertexAttribI3iv.restype = GLvoid
	gl.glVertexAttribI3iv.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI3ivEXT = _getProcAddress('glVertexAttribI3ivEXT')
	gl.glVertexAttribI3ivEXT.gl = gl
	gl.glVertexAttribI3ivEXT.__name__ = 'glVertexAttribI3ivEXT'
	gl.glVertexAttribI3ivEXT.errcheck = _errcheck
	gl.glVertexAttribI3ivEXT.restype = GLvoid
	gl.glVertexAttribI3ivEXT.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI3ui = _getProcAddress('glVertexAttribI3ui')
	gl.glVertexAttribI3ui.gl = gl
	gl.glVertexAttribI3ui.__name__ = 'glVertexAttribI3ui'
	gl.glVertexAttribI3ui.errcheck = _errcheck
	gl.glVertexAttribI3ui.restype = GLvoid
	gl.glVertexAttribI3ui.argtypes = (GLuint, GLuint, GLuint, GLuint,)
	gl.glVertexAttribI3uiEXT = _getProcAddress('glVertexAttribI3uiEXT')
	gl.glVertexAttribI3uiEXT.gl = gl
	gl.glVertexAttribI3uiEXT.__name__ = 'glVertexAttribI3uiEXT'
	gl.glVertexAttribI3uiEXT.errcheck = _errcheck
	gl.glVertexAttribI3uiEXT.restype = GLvoid
	gl.glVertexAttribI3uiEXT.argtypes = (GLuint, GLuint, GLuint, GLuint,)
	gl.glVertexAttribI3uiv = _getProcAddress('glVertexAttribI3uiv')
	gl.glVertexAttribI3uiv.gl = gl
	gl.glVertexAttribI3uiv.__name__ = 'glVertexAttribI3uiv'
	gl.glVertexAttribI3uiv.errcheck = _errcheck
	gl.glVertexAttribI3uiv.restype = GLvoid
	gl.glVertexAttribI3uiv.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI3uivEXT = _getProcAddress('glVertexAttribI3uivEXT')
	gl.glVertexAttribI3uivEXT.gl = gl
	gl.glVertexAttribI3uivEXT.__name__ = 'glVertexAttribI3uivEXT'
	gl.glVertexAttribI3uivEXT.errcheck = _errcheck
	gl.glVertexAttribI3uivEXT.restype = GLvoid
	gl.glVertexAttribI3uivEXT.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI4bv = _getProcAddress('glVertexAttribI4bv')
	gl.glVertexAttribI4bv.gl = gl
	gl.glVertexAttribI4bv.__name__ = 'glVertexAttribI4bv'
	gl.glVertexAttribI4bv.errcheck = _errcheck
	gl.glVertexAttribI4bv.restype = GLvoid
	gl.glVertexAttribI4bv.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVertexAttribI4bvEXT = _getProcAddress('glVertexAttribI4bvEXT')
	gl.glVertexAttribI4bvEXT.gl = gl
	gl.glVertexAttribI4bvEXT.__name__ = 'glVertexAttribI4bvEXT'
	gl.glVertexAttribI4bvEXT.errcheck = _errcheck
	gl.glVertexAttribI4bvEXT.restype = GLvoid
	gl.glVertexAttribI4bvEXT.argtypes = (GLuint, POINTER(GLbyte),)
	gl.glVertexAttribI4i = _getProcAddress('glVertexAttribI4i')
	gl.glVertexAttribI4i.gl = gl
	gl.glVertexAttribI4i.__name__ = 'glVertexAttribI4i'
	gl.glVertexAttribI4i.errcheck = _errcheck
	gl.glVertexAttribI4i.restype = GLvoid
	gl.glVertexAttribI4i.argtypes = (GLuint, GLint, GLint, GLint, GLint,)
	gl.glVertexAttribI4iEXT = _getProcAddress('glVertexAttribI4iEXT')
	gl.glVertexAttribI4iEXT.gl = gl
	gl.glVertexAttribI4iEXT.__name__ = 'glVertexAttribI4iEXT'
	gl.glVertexAttribI4iEXT.errcheck = _errcheck
	gl.glVertexAttribI4iEXT.restype = GLvoid
	gl.glVertexAttribI4iEXT.argtypes = (GLuint, GLint, GLint, GLint, GLint,)
	gl.glVertexAttribI4iv = _getProcAddress('glVertexAttribI4iv')
	gl.glVertexAttribI4iv.gl = gl
	gl.glVertexAttribI4iv.__name__ = 'glVertexAttribI4iv'
	gl.glVertexAttribI4iv.errcheck = _errcheck
	gl.glVertexAttribI4iv.restype = GLvoid
	gl.glVertexAttribI4iv.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI4ivEXT = _getProcAddress('glVertexAttribI4ivEXT')
	gl.glVertexAttribI4ivEXT.gl = gl
	gl.glVertexAttribI4ivEXT.__name__ = 'glVertexAttribI4ivEXT'
	gl.glVertexAttribI4ivEXT.errcheck = _errcheck
	gl.glVertexAttribI4ivEXT.restype = GLvoid
	gl.glVertexAttribI4ivEXT.argtypes = (GLuint, POINTER(GLint),)
	gl.glVertexAttribI4sv = _getProcAddress('glVertexAttribI4sv')
	gl.glVertexAttribI4sv.gl = gl
	gl.glVertexAttribI4sv.__name__ = 'glVertexAttribI4sv'
	gl.glVertexAttribI4sv.errcheck = _errcheck
	gl.glVertexAttribI4sv.restype = GLvoid
	gl.glVertexAttribI4sv.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttribI4svEXT = _getProcAddress('glVertexAttribI4svEXT')
	gl.glVertexAttribI4svEXT.gl = gl
	gl.glVertexAttribI4svEXT.__name__ = 'glVertexAttribI4svEXT'
	gl.glVertexAttribI4svEXT.errcheck = _errcheck
	gl.glVertexAttribI4svEXT.restype = GLvoid
	gl.glVertexAttribI4svEXT.argtypes = (GLuint, POINTER(GLshort),)
	gl.glVertexAttribI4ubv = _getProcAddress('glVertexAttribI4ubv')
	gl.glVertexAttribI4ubv.gl = gl
	gl.glVertexAttribI4ubv.__name__ = 'glVertexAttribI4ubv'
	gl.glVertexAttribI4ubv.errcheck = _errcheck
	gl.glVertexAttribI4ubv.restype = GLvoid
	gl.glVertexAttribI4ubv.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttribI4ubvEXT = _getProcAddress('glVertexAttribI4ubvEXT')
	gl.glVertexAttribI4ubvEXT.gl = gl
	gl.glVertexAttribI4ubvEXT.__name__ = 'glVertexAttribI4ubvEXT'
	gl.glVertexAttribI4ubvEXT.errcheck = _errcheck
	gl.glVertexAttribI4ubvEXT.restype = GLvoid
	gl.glVertexAttribI4ubvEXT.argtypes = (GLuint, POINTER(GLubyte),)
	gl.glVertexAttribI4ui = _getProcAddress('glVertexAttribI4ui')
	gl.glVertexAttribI4ui.gl = gl
	gl.glVertexAttribI4ui.__name__ = 'glVertexAttribI4ui'
	gl.glVertexAttribI4ui.errcheck = _errcheck
	gl.glVertexAttribI4ui.restype = GLvoid
	gl.glVertexAttribI4ui.argtypes = (GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glVertexAttribI4uiEXT = _getProcAddress('glVertexAttribI4uiEXT')
	gl.glVertexAttribI4uiEXT.gl = gl
	gl.glVertexAttribI4uiEXT.__name__ = 'glVertexAttribI4uiEXT'
	gl.glVertexAttribI4uiEXT.errcheck = _errcheck
	gl.glVertexAttribI4uiEXT.restype = GLvoid
	gl.glVertexAttribI4uiEXT.argtypes = (GLuint, GLuint, GLuint, GLuint, GLuint,)
	gl.glVertexAttribI4uiv = _getProcAddress('glVertexAttribI4uiv')
	gl.glVertexAttribI4uiv.gl = gl
	gl.glVertexAttribI4uiv.__name__ = 'glVertexAttribI4uiv'
	gl.glVertexAttribI4uiv.errcheck = _errcheck
	gl.glVertexAttribI4uiv.restype = GLvoid
	gl.glVertexAttribI4uiv.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI4uivEXT = _getProcAddress('glVertexAttribI4uivEXT')
	gl.glVertexAttribI4uivEXT.gl = gl
	gl.glVertexAttribI4uivEXT.__name__ = 'glVertexAttribI4uivEXT'
	gl.glVertexAttribI4uivEXT.errcheck = _errcheck
	gl.glVertexAttribI4uivEXT.restype = GLvoid
	gl.glVertexAttribI4uivEXT.argtypes = (GLuint, POINTER(GLuint),)
	gl.glVertexAttribI4usv = _getProcAddress('glVertexAttribI4usv')
	gl.glVertexAttribI4usv.gl = gl
	gl.glVertexAttribI4usv.__name__ = 'glVertexAttribI4usv'
	gl.glVertexAttribI4usv.errcheck = _errcheck
	gl.glVertexAttribI4usv.restype = GLvoid
	gl.glVertexAttribI4usv.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertexAttribI4usvEXT = _getProcAddress('glVertexAttribI4usvEXT')
	gl.glVertexAttribI4usvEXT.gl = gl
	gl.glVertexAttribI4usvEXT.__name__ = 'glVertexAttribI4usvEXT'
	gl.glVertexAttribI4usvEXT.errcheck = _errcheck
	gl.glVertexAttribI4usvEXT.restype = GLvoid
	gl.glVertexAttribI4usvEXT.argtypes = (GLuint, POINTER(GLushort),)
	gl.glVertexAttribIFormat = _getProcAddress('glVertexAttribIFormat')
	gl.glVertexAttribIFormat.gl = gl
	gl.glVertexAttribIFormat.__name__ = 'glVertexAttribIFormat'
	gl.glVertexAttribIFormat.errcheck = _errcheck
	gl.glVertexAttribIFormat.restype = GLvoid
	gl.glVertexAttribIFormat.argtypes = (GLuint, GLint, GLenum, GLuint,)
	gl.glVertexAttribIFormatNV = _getProcAddress('glVertexAttribIFormatNV')
	gl.glVertexAttribIFormatNV.gl = gl
	gl.glVertexAttribIFormatNV.__name__ = 'glVertexAttribIFormatNV'
	gl.glVertexAttribIFormatNV.errcheck = _errcheck
	gl.glVertexAttribIFormatNV.restype = GLvoid
	gl.glVertexAttribIFormatNV.argtypes = (GLuint, GLint, GLenum, GLsizei,)
	gl.glVertexAttribIPointer = _getProcAddress('glVertexAttribIPointer')
	gl.glVertexAttribIPointer.gl = gl
	gl.glVertexAttribIPointer.__name__ = 'glVertexAttribIPointer'
	gl.glVertexAttribIPointer.errcheck = _errcheck
	gl.glVertexAttribIPointer.restype = GLvoid
	gl.glVertexAttribIPointer.argtypes = (GLuint, GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribIPointerEXT = _getProcAddress('glVertexAttribIPointerEXT')
	gl.glVertexAttribIPointerEXT.gl = gl
	gl.glVertexAttribIPointerEXT.__name__ = 'glVertexAttribIPointerEXT'
	gl.glVertexAttribIPointerEXT.errcheck = _errcheck
	gl.glVertexAttribIPointerEXT.restype = GLvoid
	gl.glVertexAttribIPointerEXT.argtypes = (GLuint, GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribL1d = _getProcAddress('glVertexAttribL1d')
	gl.glVertexAttribL1d.gl = gl
	gl.glVertexAttribL1d.__name__ = 'glVertexAttribL1d'
	gl.glVertexAttribL1d.errcheck = _errcheck
	gl.glVertexAttribL1d.restype = GLvoid
	gl.glVertexAttribL1d.argtypes = (GLuint, GLdouble,)
	gl.glVertexAttribL1dEXT = _getProcAddress('glVertexAttribL1dEXT')
	gl.glVertexAttribL1dEXT.gl = gl
	gl.glVertexAttribL1dEXT.__name__ = 'glVertexAttribL1dEXT'
	gl.glVertexAttribL1dEXT.errcheck = _errcheck
	gl.glVertexAttribL1dEXT.restype = GLvoid
	gl.glVertexAttribL1dEXT.argtypes = (GLuint, GLdouble,)
	gl.glVertexAttribL1dv = _getProcAddress('glVertexAttribL1dv')
	gl.glVertexAttribL1dv.gl = gl
	gl.glVertexAttribL1dv.__name__ = 'glVertexAttribL1dv'
	gl.glVertexAttribL1dv.errcheck = _errcheck
	gl.glVertexAttribL1dv.restype = GLvoid
	gl.glVertexAttribL1dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL1dvEXT = _getProcAddress('glVertexAttribL1dvEXT')
	gl.glVertexAttribL1dvEXT.gl = gl
	gl.glVertexAttribL1dvEXT.__name__ = 'glVertexAttribL1dvEXT'
	gl.glVertexAttribL1dvEXT.errcheck = _errcheck
	gl.glVertexAttribL1dvEXT.restype = GLvoid
	gl.glVertexAttribL1dvEXT.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL1i64NV = _getProcAddress('glVertexAttribL1i64NV')
	gl.glVertexAttribL1i64NV.gl = gl
	gl.glVertexAttribL1i64NV.__name__ = 'glVertexAttribL1i64NV'
	gl.glVertexAttribL1i64NV.errcheck = _errcheck
	gl.glVertexAttribL1i64NV.restype = GLvoid
	gl.glVertexAttribL1i64NV.argtypes = (GLuint, GLint64EXT,)
	gl.glVertexAttribL1i64vNV = _getProcAddress('glVertexAttribL1i64vNV')
	gl.glVertexAttribL1i64vNV.gl = gl
	gl.glVertexAttribL1i64vNV.__name__ = 'glVertexAttribL1i64vNV'
	gl.glVertexAttribL1i64vNV.errcheck = _errcheck
	gl.glVertexAttribL1i64vNV.restype = GLvoid
	gl.glVertexAttribL1i64vNV.argtypes = (GLuint, POINTER(GLint64EXT),)
	gl.glVertexAttribL1ui64ARB = _getProcAddress('glVertexAttribL1ui64ARB')
	gl.glVertexAttribL1ui64ARB.gl = gl
	gl.glVertexAttribL1ui64ARB.__name__ = 'glVertexAttribL1ui64ARB'
	gl.glVertexAttribL1ui64ARB.errcheck = _errcheck
	gl.glVertexAttribL1ui64ARB.restype = GLvoid
	gl.glVertexAttribL1ui64ARB.argtypes = (GLuint, GLuint64EXT,)
	gl.glVertexAttribL1ui64NV = _getProcAddress('glVertexAttribL1ui64NV')
	gl.glVertexAttribL1ui64NV.gl = gl
	gl.glVertexAttribL1ui64NV.__name__ = 'glVertexAttribL1ui64NV'
	gl.glVertexAttribL1ui64NV.errcheck = _errcheck
	gl.glVertexAttribL1ui64NV.restype = GLvoid
	gl.glVertexAttribL1ui64NV.argtypes = (GLuint, GLuint64EXT,)
	gl.glVertexAttribL1ui64vARB = _getProcAddress('glVertexAttribL1ui64vARB')
	gl.glVertexAttribL1ui64vARB.gl = gl
	gl.glVertexAttribL1ui64vARB.__name__ = 'glVertexAttribL1ui64vARB'
	gl.glVertexAttribL1ui64vARB.errcheck = _errcheck
	gl.glVertexAttribL1ui64vARB.restype = GLvoid
	gl.glVertexAttribL1ui64vARB.argtypes = (GLuint, POINTER(GLuint64EXT),)
	gl.glVertexAttribL1ui64vNV = _getProcAddress('glVertexAttribL1ui64vNV')
	gl.glVertexAttribL1ui64vNV.gl = gl
	gl.glVertexAttribL1ui64vNV.__name__ = 'glVertexAttribL1ui64vNV'
	gl.glVertexAttribL1ui64vNV.errcheck = _errcheck
	gl.glVertexAttribL1ui64vNV.restype = GLvoid
	gl.glVertexAttribL1ui64vNV.argtypes = (GLuint, POINTER(GLuint64EXT),)
	gl.glVertexAttribL2d = _getProcAddress('glVertexAttribL2d')
	gl.glVertexAttribL2d.gl = gl
	gl.glVertexAttribL2d.__name__ = 'glVertexAttribL2d'
	gl.glVertexAttribL2d.errcheck = _errcheck
	gl.glVertexAttribL2d.restype = GLvoid
	gl.glVertexAttribL2d.argtypes = (GLuint, GLdouble, GLdouble,)
	gl.glVertexAttribL2dEXT = _getProcAddress('glVertexAttribL2dEXT')
	gl.glVertexAttribL2dEXT.gl = gl
	gl.glVertexAttribL2dEXT.__name__ = 'glVertexAttribL2dEXT'
	gl.glVertexAttribL2dEXT.errcheck = _errcheck
	gl.glVertexAttribL2dEXT.restype = GLvoid
	gl.glVertexAttribL2dEXT.argtypes = (GLuint, GLdouble, GLdouble,)
	gl.glVertexAttribL2dv = _getProcAddress('glVertexAttribL2dv')
	gl.glVertexAttribL2dv.gl = gl
	gl.glVertexAttribL2dv.__name__ = 'glVertexAttribL2dv'
	gl.glVertexAttribL2dv.errcheck = _errcheck
	gl.glVertexAttribL2dv.restype = GLvoid
	gl.glVertexAttribL2dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL2dvEXT = _getProcAddress('glVertexAttribL2dvEXT')
	gl.glVertexAttribL2dvEXT.gl = gl
	gl.glVertexAttribL2dvEXT.__name__ = 'glVertexAttribL2dvEXT'
	gl.glVertexAttribL2dvEXT.errcheck = _errcheck
	gl.glVertexAttribL2dvEXT.restype = GLvoid
	gl.glVertexAttribL2dvEXT.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL2i64NV = _getProcAddress('glVertexAttribL2i64NV')
	gl.glVertexAttribL2i64NV.gl = gl
	gl.glVertexAttribL2i64NV.__name__ = 'glVertexAttribL2i64NV'
	gl.glVertexAttribL2i64NV.errcheck = _errcheck
	gl.glVertexAttribL2i64NV.restype = GLvoid
	gl.glVertexAttribL2i64NV.argtypes = (GLuint, GLint64EXT, GLint64EXT,)
	gl.glVertexAttribL2i64vNV = _getProcAddress('glVertexAttribL2i64vNV')
	gl.glVertexAttribL2i64vNV.gl = gl
	gl.glVertexAttribL2i64vNV.__name__ = 'glVertexAttribL2i64vNV'
	gl.glVertexAttribL2i64vNV.errcheck = _errcheck
	gl.glVertexAttribL2i64vNV.restype = GLvoid
	gl.glVertexAttribL2i64vNV.argtypes = (GLuint, POINTER(GLint64EXT),)
	gl.glVertexAttribL2ui64NV = _getProcAddress('glVertexAttribL2ui64NV')
	gl.glVertexAttribL2ui64NV.gl = gl
	gl.glVertexAttribL2ui64NV.__name__ = 'glVertexAttribL2ui64NV'
	gl.glVertexAttribL2ui64NV.errcheck = _errcheck
	gl.glVertexAttribL2ui64NV.restype = GLvoid
	gl.glVertexAttribL2ui64NV.argtypes = (GLuint, GLuint64EXT, GLuint64EXT,)
	gl.glVertexAttribL2ui64vNV = _getProcAddress('glVertexAttribL2ui64vNV')
	gl.glVertexAttribL2ui64vNV.gl = gl
	gl.glVertexAttribL2ui64vNV.__name__ = 'glVertexAttribL2ui64vNV'
	gl.glVertexAttribL2ui64vNV.errcheck = _errcheck
	gl.glVertexAttribL2ui64vNV.restype = GLvoid
	gl.glVertexAttribL2ui64vNV.argtypes = (GLuint, POINTER(GLuint64EXT),)
	gl.glVertexAttribL3d = _getProcAddress('glVertexAttribL3d')
	gl.glVertexAttribL3d.gl = gl
	gl.glVertexAttribL3d.__name__ = 'glVertexAttribL3d'
	gl.glVertexAttribL3d.errcheck = _errcheck
	gl.glVertexAttribL3d.restype = GLvoid
	gl.glVertexAttribL3d.argtypes = (GLuint, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttribL3dEXT = _getProcAddress('glVertexAttribL3dEXT')
	gl.glVertexAttribL3dEXT.gl = gl
	gl.glVertexAttribL3dEXT.__name__ = 'glVertexAttribL3dEXT'
	gl.glVertexAttribL3dEXT.errcheck = _errcheck
	gl.glVertexAttribL3dEXT.restype = GLvoid
	gl.glVertexAttribL3dEXT.argtypes = (GLuint, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttribL3dv = _getProcAddress('glVertexAttribL3dv')
	gl.glVertexAttribL3dv.gl = gl
	gl.glVertexAttribL3dv.__name__ = 'glVertexAttribL3dv'
	gl.glVertexAttribL3dv.errcheck = _errcheck
	gl.glVertexAttribL3dv.restype = GLvoid
	gl.glVertexAttribL3dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL3dvEXT = _getProcAddress('glVertexAttribL3dvEXT')
	gl.glVertexAttribL3dvEXT.gl = gl
	gl.glVertexAttribL3dvEXT.__name__ = 'glVertexAttribL3dvEXT'
	gl.glVertexAttribL3dvEXT.errcheck = _errcheck
	gl.glVertexAttribL3dvEXT.restype = GLvoid
	gl.glVertexAttribL3dvEXT.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL3i64NV = _getProcAddress('glVertexAttribL3i64NV')
	gl.glVertexAttribL3i64NV.gl = gl
	gl.glVertexAttribL3i64NV.__name__ = 'glVertexAttribL3i64NV'
	gl.glVertexAttribL3i64NV.errcheck = _errcheck
	gl.glVertexAttribL3i64NV.restype = GLvoid
	gl.glVertexAttribL3i64NV.argtypes = (GLuint, GLint64EXT, GLint64EXT, GLint64EXT,)
	gl.glVertexAttribL3i64vNV = _getProcAddress('glVertexAttribL3i64vNV')
	gl.glVertexAttribL3i64vNV.gl = gl
	gl.glVertexAttribL3i64vNV.__name__ = 'glVertexAttribL3i64vNV'
	gl.glVertexAttribL3i64vNV.errcheck = _errcheck
	gl.glVertexAttribL3i64vNV.restype = GLvoid
	gl.glVertexAttribL3i64vNV.argtypes = (GLuint, POINTER(GLint64EXT),)
	gl.glVertexAttribL3ui64NV = _getProcAddress('glVertexAttribL3ui64NV')
	gl.glVertexAttribL3ui64NV.gl = gl
	gl.glVertexAttribL3ui64NV.__name__ = 'glVertexAttribL3ui64NV'
	gl.glVertexAttribL3ui64NV.errcheck = _errcheck
	gl.glVertexAttribL3ui64NV.restype = GLvoid
	gl.glVertexAttribL3ui64NV.argtypes = (GLuint, GLuint64EXT, GLuint64EXT, GLuint64EXT,)
	gl.glVertexAttribL3ui64vNV = _getProcAddress('glVertexAttribL3ui64vNV')
	gl.glVertexAttribL3ui64vNV.gl = gl
	gl.glVertexAttribL3ui64vNV.__name__ = 'glVertexAttribL3ui64vNV'
	gl.glVertexAttribL3ui64vNV.errcheck = _errcheck
	gl.glVertexAttribL3ui64vNV.restype = GLvoid
	gl.glVertexAttribL3ui64vNV.argtypes = (GLuint, POINTER(GLuint64EXT),)
	gl.glVertexAttribL4d = _getProcAddress('glVertexAttribL4d')
	gl.glVertexAttribL4d.gl = gl
	gl.glVertexAttribL4d.__name__ = 'glVertexAttribL4d'
	gl.glVertexAttribL4d.errcheck = _errcheck
	gl.glVertexAttribL4d.restype = GLvoid
	gl.glVertexAttribL4d.argtypes = (GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttribL4dEXT = _getProcAddress('glVertexAttribL4dEXT')
	gl.glVertexAttribL4dEXT.gl = gl
	gl.glVertexAttribL4dEXT.__name__ = 'glVertexAttribL4dEXT'
	gl.glVertexAttribL4dEXT.errcheck = _errcheck
	gl.glVertexAttribL4dEXT.restype = GLvoid
	gl.glVertexAttribL4dEXT.argtypes = (GLuint, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertexAttribL4dv = _getProcAddress('glVertexAttribL4dv')
	gl.glVertexAttribL4dv.gl = gl
	gl.glVertexAttribL4dv.__name__ = 'glVertexAttribL4dv'
	gl.glVertexAttribL4dv.errcheck = _errcheck
	gl.glVertexAttribL4dv.restype = GLvoid
	gl.glVertexAttribL4dv.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL4dvEXT = _getProcAddress('glVertexAttribL4dvEXT')
	gl.glVertexAttribL4dvEXT.gl = gl
	gl.glVertexAttribL4dvEXT.__name__ = 'glVertexAttribL4dvEXT'
	gl.glVertexAttribL4dvEXT.errcheck = _errcheck
	gl.glVertexAttribL4dvEXT.restype = GLvoid
	gl.glVertexAttribL4dvEXT.argtypes = (GLuint, POINTER(GLdouble),)
	gl.glVertexAttribL4i64NV = _getProcAddress('glVertexAttribL4i64NV')
	gl.glVertexAttribL4i64NV.gl = gl
	gl.glVertexAttribL4i64NV.__name__ = 'glVertexAttribL4i64NV'
	gl.glVertexAttribL4i64NV.errcheck = _errcheck
	gl.glVertexAttribL4i64NV.restype = GLvoid
	gl.glVertexAttribL4i64NV.argtypes = (GLuint, GLint64EXT, GLint64EXT, GLint64EXT, GLint64EXT,)
	gl.glVertexAttribL4i64vNV = _getProcAddress('glVertexAttribL4i64vNV')
	gl.glVertexAttribL4i64vNV.gl = gl
	gl.glVertexAttribL4i64vNV.__name__ = 'glVertexAttribL4i64vNV'
	gl.glVertexAttribL4i64vNV.errcheck = _errcheck
	gl.glVertexAttribL4i64vNV.restype = GLvoid
	gl.glVertexAttribL4i64vNV.argtypes = (GLuint, POINTER(GLint64EXT),)
	gl.glVertexAttribL4ui64NV = _getProcAddress('glVertexAttribL4ui64NV')
	gl.glVertexAttribL4ui64NV.gl = gl
	gl.glVertexAttribL4ui64NV.__name__ = 'glVertexAttribL4ui64NV'
	gl.glVertexAttribL4ui64NV.errcheck = _errcheck
	gl.glVertexAttribL4ui64NV.restype = GLvoid
	gl.glVertexAttribL4ui64NV.argtypes = (GLuint, GLuint64EXT, GLuint64EXT, GLuint64EXT, GLuint64EXT,)
	gl.glVertexAttribL4ui64vNV = _getProcAddress('glVertexAttribL4ui64vNV')
	gl.glVertexAttribL4ui64vNV.gl = gl
	gl.glVertexAttribL4ui64vNV.__name__ = 'glVertexAttribL4ui64vNV'
	gl.glVertexAttribL4ui64vNV.errcheck = _errcheck
	gl.glVertexAttribL4ui64vNV.restype = GLvoid
	gl.glVertexAttribL4ui64vNV.argtypes = (GLuint, POINTER(GLuint64EXT),)
	gl.glVertexAttribLFormat = _getProcAddress('glVertexAttribLFormat')
	gl.glVertexAttribLFormat.gl = gl
	gl.glVertexAttribLFormat.__name__ = 'glVertexAttribLFormat'
	gl.glVertexAttribLFormat.errcheck = _errcheck
	gl.glVertexAttribLFormat.restype = GLvoid
	gl.glVertexAttribLFormat.argtypes = (GLuint, GLint, GLenum, GLuint,)
	gl.glVertexAttribLFormatNV = _getProcAddress('glVertexAttribLFormatNV')
	gl.glVertexAttribLFormatNV.gl = gl
	gl.glVertexAttribLFormatNV.__name__ = 'glVertexAttribLFormatNV'
	gl.glVertexAttribLFormatNV.errcheck = _errcheck
	gl.glVertexAttribLFormatNV.restype = GLvoid
	gl.glVertexAttribLFormatNV.argtypes = (GLuint, GLint, GLenum, GLsizei,)
	gl.glVertexAttribLPointer = _getProcAddress('glVertexAttribLPointer')
	gl.glVertexAttribLPointer.gl = gl
	gl.glVertexAttribLPointer.__name__ = 'glVertexAttribLPointer'
	gl.glVertexAttribLPointer.errcheck = _errcheck
	gl.glVertexAttribLPointer.restype = GLvoid
	gl.glVertexAttribLPointer.argtypes = (GLuint, GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribLPointerEXT = _getProcAddress('glVertexAttribLPointerEXT')
	gl.glVertexAttribLPointerEXT.gl = gl
	gl.glVertexAttribLPointerEXT.__name__ = 'glVertexAttribLPointerEXT'
	gl.glVertexAttribLPointerEXT.errcheck = _errcheck
	gl.glVertexAttribLPointerEXT.restype = GLvoid
	gl.glVertexAttribLPointerEXT.argtypes = (GLuint, GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribP1ui = _getProcAddress('glVertexAttribP1ui')
	gl.glVertexAttribP1ui.gl = gl
	gl.glVertexAttribP1ui.__name__ = 'glVertexAttribP1ui'
	gl.glVertexAttribP1ui.errcheck = _errcheck
	gl.glVertexAttribP1ui.restype = GLvoid
	gl.glVertexAttribP1ui.argtypes = (GLuint, GLenum, GLboolean, GLuint,)
	gl.glVertexAttribP1uiv = _getProcAddress('glVertexAttribP1uiv')
	gl.glVertexAttribP1uiv.gl = gl
	gl.glVertexAttribP1uiv.__name__ = 'glVertexAttribP1uiv'
	gl.glVertexAttribP1uiv.errcheck = _errcheck
	gl.glVertexAttribP1uiv.restype = GLvoid
	gl.glVertexAttribP1uiv.argtypes = (GLuint, GLenum, GLboolean, POINTER(GLuint),)
	gl.glVertexAttribP2ui = _getProcAddress('glVertexAttribP2ui')
	gl.glVertexAttribP2ui.gl = gl
	gl.glVertexAttribP2ui.__name__ = 'glVertexAttribP2ui'
	gl.glVertexAttribP2ui.errcheck = _errcheck
	gl.glVertexAttribP2ui.restype = GLvoid
	gl.glVertexAttribP2ui.argtypes = (GLuint, GLenum, GLboolean, GLuint,)
	gl.glVertexAttribP2uiv = _getProcAddress('glVertexAttribP2uiv')
	gl.glVertexAttribP2uiv.gl = gl
	gl.glVertexAttribP2uiv.__name__ = 'glVertexAttribP2uiv'
	gl.glVertexAttribP2uiv.errcheck = _errcheck
	gl.glVertexAttribP2uiv.restype = GLvoid
	gl.glVertexAttribP2uiv.argtypes = (GLuint, GLenum, GLboolean, POINTER(GLuint),)
	gl.glVertexAttribP3ui = _getProcAddress('glVertexAttribP3ui')
	gl.glVertexAttribP3ui.gl = gl
	gl.glVertexAttribP3ui.__name__ = 'glVertexAttribP3ui'
	gl.glVertexAttribP3ui.errcheck = _errcheck
	gl.glVertexAttribP3ui.restype = GLvoid
	gl.glVertexAttribP3ui.argtypes = (GLuint, GLenum, GLboolean, GLuint,)
	gl.glVertexAttribP3uiv = _getProcAddress('glVertexAttribP3uiv')
	gl.glVertexAttribP3uiv.gl = gl
	gl.glVertexAttribP3uiv.__name__ = 'glVertexAttribP3uiv'
	gl.glVertexAttribP3uiv.errcheck = _errcheck
	gl.glVertexAttribP3uiv.restype = GLvoid
	gl.glVertexAttribP3uiv.argtypes = (GLuint, GLenum, GLboolean, POINTER(GLuint),)
	gl.glVertexAttribP4ui = _getProcAddress('glVertexAttribP4ui')
	gl.glVertexAttribP4ui.gl = gl
	gl.glVertexAttribP4ui.__name__ = 'glVertexAttribP4ui'
	gl.glVertexAttribP4ui.errcheck = _errcheck
	gl.glVertexAttribP4ui.restype = GLvoid
	gl.glVertexAttribP4ui.argtypes = (GLuint, GLenum, GLboolean, GLuint,)
	gl.glVertexAttribP4uiv = _getProcAddress('glVertexAttribP4uiv')
	gl.glVertexAttribP4uiv.gl = gl
	gl.glVertexAttribP4uiv.__name__ = 'glVertexAttribP4uiv'
	gl.glVertexAttribP4uiv.errcheck = _errcheck
	gl.glVertexAttribP4uiv.restype = GLvoid
	gl.glVertexAttribP4uiv.argtypes = (GLuint, GLenum, GLboolean, POINTER(GLuint),)
	gl.glVertexAttribParameteriAMD = _getProcAddress('glVertexAttribParameteriAMD')
	gl.glVertexAttribParameteriAMD.gl = gl
	gl.glVertexAttribParameteriAMD.__name__ = 'glVertexAttribParameteriAMD'
	gl.glVertexAttribParameteriAMD.errcheck = _errcheck
	gl.glVertexAttribParameteriAMD.restype = GLvoid
	gl.glVertexAttribParameteriAMD.argtypes = (GLuint, GLenum, GLint,)
	gl.glVertexAttribPointer = _getProcAddress('glVertexAttribPointer')
	gl.glVertexAttribPointer.gl = gl
	gl.glVertexAttribPointer.__name__ = 'glVertexAttribPointer'
	gl.glVertexAttribPointer.errcheck = _errcheck
	gl.glVertexAttribPointer.restype = GLvoid
	gl.glVertexAttribPointer.argtypes = (GLuint, GLint, GLenum, GLboolean, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribPointerARB = _getProcAddress('glVertexAttribPointerARB')
	gl.glVertexAttribPointerARB.gl = gl
	gl.glVertexAttribPointerARB.__name__ = 'glVertexAttribPointerARB'
	gl.glVertexAttribPointerARB.errcheck = _errcheck
	gl.glVertexAttribPointerARB.restype = GLvoid
	gl.glVertexAttribPointerARB.argtypes = (GLuint, GLint, GLenum, GLboolean, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribPointerNV = _getProcAddress('glVertexAttribPointerNV')
	gl.glVertexAttribPointerNV.gl = gl
	gl.glVertexAttribPointerNV.__name__ = 'glVertexAttribPointerNV'
	gl.glVertexAttribPointerNV.errcheck = _errcheck
	gl.glVertexAttribPointerNV.restype = GLvoid
	gl.glVertexAttribPointerNV.argtypes = (GLuint, GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexAttribs1dvNV = _getProcAddress('glVertexAttribs1dvNV')
	gl.glVertexAttribs1dvNV.gl = gl
	gl.glVertexAttribs1dvNV.__name__ = 'glVertexAttribs1dvNV'
	gl.glVertexAttribs1dvNV.errcheck = _errcheck
	gl.glVertexAttribs1dvNV.restype = GLvoid
	gl.glVertexAttribs1dvNV.argtypes = (GLuint, GLsizei, POINTER(GLdouble),)
	gl.glVertexAttribs1fvNV = _getProcAddress('glVertexAttribs1fvNV')
	gl.glVertexAttribs1fvNV.gl = gl
	gl.glVertexAttribs1fvNV.__name__ = 'glVertexAttribs1fvNV'
	gl.glVertexAttribs1fvNV.errcheck = _errcheck
	gl.glVertexAttribs1fvNV.restype = GLvoid
	gl.glVertexAttribs1fvNV.argtypes = (GLuint, GLsizei, POINTER(GLfloat),)
	gl.glVertexAttribs1hvNV = _getProcAddress('glVertexAttribs1hvNV')
	gl.glVertexAttribs1hvNV.gl = gl
	gl.glVertexAttribs1hvNV.__name__ = 'glVertexAttribs1hvNV'
	gl.glVertexAttribs1hvNV.errcheck = _errcheck
	gl.glVertexAttribs1hvNV.restype = GLvoid
	gl.glVertexAttribs1hvNV.argtypes = (GLuint, GLsizei, POINTER(GLhalfNV),)
	gl.glVertexAttribs1svNV = _getProcAddress('glVertexAttribs1svNV')
	gl.glVertexAttribs1svNV.gl = gl
	gl.glVertexAttribs1svNV.__name__ = 'glVertexAttribs1svNV'
	gl.glVertexAttribs1svNV.errcheck = _errcheck
	gl.glVertexAttribs1svNV.restype = GLvoid
	gl.glVertexAttribs1svNV.argtypes = (GLuint, GLsizei, POINTER(GLshort),)
	gl.glVertexAttribs2dvNV = _getProcAddress('glVertexAttribs2dvNV')
	gl.glVertexAttribs2dvNV.gl = gl
	gl.glVertexAttribs2dvNV.__name__ = 'glVertexAttribs2dvNV'
	gl.glVertexAttribs2dvNV.errcheck = _errcheck
	gl.glVertexAttribs2dvNV.restype = GLvoid
	gl.glVertexAttribs2dvNV.argtypes = (GLuint, GLsizei, POINTER(GLdouble),)
	gl.glVertexAttribs2fvNV = _getProcAddress('glVertexAttribs2fvNV')
	gl.glVertexAttribs2fvNV.gl = gl
	gl.glVertexAttribs2fvNV.__name__ = 'glVertexAttribs2fvNV'
	gl.glVertexAttribs2fvNV.errcheck = _errcheck
	gl.glVertexAttribs2fvNV.restype = GLvoid
	gl.glVertexAttribs2fvNV.argtypes = (GLuint, GLsizei, POINTER(GLfloat),)
	gl.glVertexAttribs2hvNV = _getProcAddress('glVertexAttribs2hvNV')
	gl.glVertexAttribs2hvNV.gl = gl
	gl.glVertexAttribs2hvNV.__name__ = 'glVertexAttribs2hvNV'
	gl.glVertexAttribs2hvNV.errcheck = _errcheck
	gl.glVertexAttribs2hvNV.restype = GLvoid
	gl.glVertexAttribs2hvNV.argtypes = (GLuint, GLsizei, POINTER(GLhalfNV),)
	gl.glVertexAttribs2svNV = _getProcAddress('glVertexAttribs2svNV')
	gl.glVertexAttribs2svNV.gl = gl
	gl.glVertexAttribs2svNV.__name__ = 'glVertexAttribs2svNV'
	gl.glVertexAttribs2svNV.errcheck = _errcheck
	gl.glVertexAttribs2svNV.restype = GLvoid
	gl.glVertexAttribs2svNV.argtypes = (GLuint, GLsizei, POINTER(GLshort),)
	gl.glVertexAttribs3dvNV = _getProcAddress('glVertexAttribs3dvNV')
	gl.glVertexAttribs3dvNV.gl = gl
	gl.glVertexAttribs3dvNV.__name__ = 'glVertexAttribs3dvNV'
	gl.glVertexAttribs3dvNV.errcheck = _errcheck
	gl.glVertexAttribs3dvNV.restype = GLvoid
	gl.glVertexAttribs3dvNV.argtypes = (GLuint, GLsizei, POINTER(GLdouble),)
	gl.glVertexAttribs3fvNV = _getProcAddress('glVertexAttribs3fvNV')
	gl.glVertexAttribs3fvNV.gl = gl
	gl.glVertexAttribs3fvNV.__name__ = 'glVertexAttribs3fvNV'
	gl.glVertexAttribs3fvNV.errcheck = _errcheck
	gl.glVertexAttribs3fvNV.restype = GLvoid
	gl.glVertexAttribs3fvNV.argtypes = (GLuint, GLsizei, POINTER(GLfloat),)
	gl.glVertexAttribs3hvNV = _getProcAddress('glVertexAttribs3hvNV')
	gl.glVertexAttribs3hvNV.gl = gl
	gl.glVertexAttribs3hvNV.__name__ = 'glVertexAttribs3hvNV'
	gl.glVertexAttribs3hvNV.errcheck = _errcheck
	gl.glVertexAttribs3hvNV.restype = GLvoid
	gl.glVertexAttribs3hvNV.argtypes = (GLuint, GLsizei, POINTER(GLhalfNV),)
	gl.glVertexAttribs3svNV = _getProcAddress('glVertexAttribs3svNV')
	gl.glVertexAttribs3svNV.gl = gl
	gl.glVertexAttribs3svNV.__name__ = 'glVertexAttribs3svNV'
	gl.glVertexAttribs3svNV.errcheck = _errcheck
	gl.glVertexAttribs3svNV.restype = GLvoid
	gl.glVertexAttribs3svNV.argtypes = (GLuint, GLsizei, POINTER(GLshort),)
	gl.glVertexAttribs4dvNV = _getProcAddress('glVertexAttribs4dvNV')
	gl.glVertexAttribs4dvNV.gl = gl
	gl.glVertexAttribs4dvNV.__name__ = 'glVertexAttribs4dvNV'
	gl.glVertexAttribs4dvNV.errcheck = _errcheck
	gl.glVertexAttribs4dvNV.restype = GLvoid
	gl.glVertexAttribs4dvNV.argtypes = (GLuint, GLsizei, POINTER(GLdouble),)
	gl.glVertexAttribs4fvNV = _getProcAddress('glVertexAttribs4fvNV')
	gl.glVertexAttribs4fvNV.gl = gl
	gl.glVertexAttribs4fvNV.__name__ = 'glVertexAttribs4fvNV'
	gl.glVertexAttribs4fvNV.errcheck = _errcheck
	gl.glVertexAttribs4fvNV.restype = GLvoid
	gl.glVertexAttribs4fvNV.argtypes = (GLuint, GLsizei, POINTER(GLfloat),)
	gl.glVertexAttribs4hvNV = _getProcAddress('glVertexAttribs4hvNV')
	gl.glVertexAttribs4hvNV.gl = gl
	gl.glVertexAttribs4hvNV.__name__ = 'glVertexAttribs4hvNV'
	gl.glVertexAttribs4hvNV.errcheck = _errcheck
	gl.glVertexAttribs4hvNV.restype = GLvoid
	gl.glVertexAttribs4hvNV.argtypes = (GLuint, GLsizei, POINTER(GLhalfNV),)
	gl.glVertexAttribs4svNV = _getProcAddress('glVertexAttribs4svNV')
	gl.glVertexAttribs4svNV.gl = gl
	gl.glVertexAttribs4svNV.__name__ = 'glVertexAttribs4svNV'
	gl.glVertexAttribs4svNV.errcheck = _errcheck
	gl.glVertexAttribs4svNV.restype = GLvoid
	gl.glVertexAttribs4svNV.argtypes = (GLuint, GLsizei, POINTER(GLshort),)
	gl.glVertexAttribs4ubvNV = _getProcAddress('glVertexAttribs4ubvNV')
	gl.glVertexAttribs4ubvNV.gl = gl
	gl.glVertexAttribs4ubvNV.__name__ = 'glVertexAttribs4ubvNV'
	gl.glVertexAttribs4ubvNV.errcheck = _errcheck
	gl.glVertexAttribs4ubvNV.restype = GLvoid
	gl.glVertexAttribs4ubvNV.argtypes = (GLuint, GLsizei, POINTER(GLubyte),)
	gl.glVertexBindingDivisor = _getProcAddress('glVertexBindingDivisor')
	gl.glVertexBindingDivisor.gl = gl
	gl.glVertexBindingDivisor.__name__ = 'glVertexBindingDivisor'
	gl.glVertexBindingDivisor.errcheck = _errcheck
	gl.glVertexBindingDivisor.restype = GLvoid
	gl.glVertexBindingDivisor.argtypes = (GLuint, GLuint,)
	gl.glVertexBlendARB = _getProcAddress('glVertexBlendARB')
	gl.glVertexBlendARB.gl = gl
	gl.glVertexBlendARB.__name__ = 'glVertexBlendARB'
	gl.glVertexBlendARB.errcheck = _errcheck
	gl.glVertexBlendARB.restype = GLvoid
	gl.glVertexBlendARB.argtypes = (GLint,)
	gl.glVertexBlendEnvfATI = _getProcAddress('glVertexBlendEnvfATI')
	gl.glVertexBlendEnvfATI.gl = gl
	gl.glVertexBlendEnvfATI.__name__ = 'glVertexBlendEnvfATI'
	gl.glVertexBlendEnvfATI.errcheck = _errcheck
	gl.glVertexBlendEnvfATI.restype = GLvoid
	gl.glVertexBlendEnvfATI.argtypes = (GLenum, GLfloat,)
	gl.glVertexBlendEnviATI = _getProcAddress('glVertexBlendEnviATI')
	gl.glVertexBlendEnviATI.gl = gl
	gl.glVertexBlendEnviATI.__name__ = 'glVertexBlendEnviATI'
	gl.glVertexBlendEnviATI.errcheck = _errcheck
	gl.glVertexBlendEnviATI.restype = GLvoid
	gl.glVertexBlendEnviATI.argtypes = (GLenum, GLint,)
	gl.glVertexFormatNV = _getProcAddress('glVertexFormatNV')
	gl.glVertexFormatNV.gl = gl
	gl.glVertexFormatNV.__name__ = 'glVertexFormatNV'
	gl.glVertexFormatNV.errcheck = _errcheck
	gl.glVertexFormatNV.restype = GLvoid
	gl.glVertexFormatNV.argtypes = (GLint, GLenum, GLsizei,)
	gl.glVertexP2ui = _getProcAddress('glVertexP2ui')
	gl.glVertexP2ui.gl = gl
	gl.glVertexP2ui.__name__ = 'glVertexP2ui'
	gl.glVertexP2ui.errcheck = _errcheck
	gl.glVertexP2ui.restype = GLvoid
	gl.glVertexP2ui.argtypes = (GLenum, GLuint,)
	gl.glVertexP2uiv = _getProcAddress('glVertexP2uiv')
	gl.glVertexP2uiv.gl = gl
	gl.glVertexP2uiv.__name__ = 'glVertexP2uiv'
	gl.glVertexP2uiv.errcheck = _errcheck
	gl.glVertexP2uiv.restype = GLvoid
	gl.glVertexP2uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glVertexP3ui = _getProcAddress('glVertexP3ui')
	gl.glVertexP3ui.gl = gl
	gl.glVertexP3ui.__name__ = 'glVertexP3ui'
	gl.glVertexP3ui.errcheck = _errcheck
	gl.glVertexP3ui.restype = GLvoid
	gl.glVertexP3ui.argtypes = (GLenum, GLuint,)
	gl.glVertexP3uiv = _getProcAddress('glVertexP3uiv')
	gl.glVertexP3uiv.gl = gl
	gl.glVertexP3uiv.__name__ = 'glVertexP3uiv'
	gl.glVertexP3uiv.errcheck = _errcheck
	gl.glVertexP3uiv.restype = GLvoid
	gl.glVertexP3uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glVertexP4ui = _getProcAddress('glVertexP4ui')
	gl.glVertexP4ui.gl = gl
	gl.glVertexP4ui.__name__ = 'glVertexP4ui'
	gl.glVertexP4ui.errcheck = _errcheck
	gl.glVertexP4ui.restype = GLvoid
	gl.glVertexP4ui.argtypes = (GLenum, GLuint,)
	gl.glVertexP4uiv = _getProcAddress('glVertexP4uiv')
	gl.glVertexP4uiv.gl = gl
	gl.glVertexP4uiv.__name__ = 'glVertexP4uiv'
	gl.glVertexP4uiv.errcheck = _errcheck
	gl.glVertexP4uiv.restype = GLvoid
	gl.glVertexP4uiv.argtypes = (GLenum, POINTER(GLuint),)
	gl.glVertexPointer = _getProcAddress('glVertexPointer')
	gl.glVertexPointer.gl = gl
	gl.glVertexPointer.__name__ = 'glVertexPointer'
	gl.glVertexPointer.errcheck = _errcheck
	gl.glVertexPointer.restype = GLvoid
	gl.glVertexPointer.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexPointerEXT = _getProcAddress('glVertexPointerEXT')
	gl.glVertexPointerEXT.gl = gl
	gl.glVertexPointerEXT.__name__ = 'glVertexPointerEXT'
	gl.glVertexPointerEXT.errcheck = _errcheck
	gl.glVertexPointerEXT.restype = GLvoid
	gl.glVertexPointerEXT.argtypes = (GLint, GLenum, GLsizei, GLsizei, POINTER(GLvoid),)
	gl.glVertexPointerListIBM = _getProcAddress('glVertexPointerListIBM')
	gl.glVertexPointerListIBM.gl = gl
	gl.glVertexPointerListIBM.__name__ = 'glVertexPointerListIBM'
	gl.glVertexPointerListIBM.errcheck = _errcheck
	gl.glVertexPointerListIBM.restype = GLvoid
	gl.glVertexPointerListIBM.argtypes = (GLint, GLenum, GLint, POINTER(POINTER(GLvoid)), GLint,)
	gl.glVertexPointervINTEL = _getProcAddress('glVertexPointervINTEL')
	gl.glVertexPointervINTEL.gl = gl
	gl.glVertexPointervINTEL.__name__ = 'glVertexPointervINTEL'
	gl.glVertexPointervINTEL.errcheck = _errcheck
	gl.glVertexPointervINTEL.restype = GLvoid
	gl.glVertexPointervINTEL.argtypes = (GLint, GLenum, POINTER(POINTER(GLvoid)),)
	gl.glVertexStream1dATI = _getProcAddress('glVertexStream1dATI')
	gl.glVertexStream1dATI.gl = gl
	gl.glVertexStream1dATI.__name__ = 'glVertexStream1dATI'
	gl.glVertexStream1dATI.errcheck = _errcheck
	gl.glVertexStream1dATI.restype = GLvoid
	gl.glVertexStream1dATI.argtypes = (GLenum, GLdouble,)
	gl.glVertexStream1dvATI = _getProcAddress('glVertexStream1dvATI')
	gl.glVertexStream1dvATI.gl = gl
	gl.glVertexStream1dvATI.__name__ = 'glVertexStream1dvATI'
	gl.glVertexStream1dvATI.errcheck = _errcheck
	gl.glVertexStream1dvATI.restype = GLvoid
	gl.glVertexStream1dvATI.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glVertexStream1fATI = _getProcAddress('glVertexStream1fATI')
	gl.glVertexStream1fATI.gl = gl
	gl.glVertexStream1fATI.__name__ = 'glVertexStream1fATI'
	gl.glVertexStream1fATI.errcheck = _errcheck
	gl.glVertexStream1fATI.restype = GLvoid
	gl.glVertexStream1fATI.argtypes = (GLenum, GLfloat,)
	gl.glVertexStream1fvATI = _getProcAddress('glVertexStream1fvATI')
	gl.glVertexStream1fvATI.gl = gl
	gl.glVertexStream1fvATI.__name__ = 'glVertexStream1fvATI'
	gl.glVertexStream1fvATI.errcheck = _errcheck
	gl.glVertexStream1fvATI.restype = GLvoid
	gl.glVertexStream1fvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glVertexStream1iATI = _getProcAddress('glVertexStream1iATI')
	gl.glVertexStream1iATI.gl = gl
	gl.glVertexStream1iATI.__name__ = 'glVertexStream1iATI'
	gl.glVertexStream1iATI.errcheck = _errcheck
	gl.glVertexStream1iATI.restype = GLvoid
	gl.glVertexStream1iATI.argtypes = (GLenum, GLint,)
	gl.glVertexStream1ivATI = _getProcAddress('glVertexStream1ivATI')
	gl.glVertexStream1ivATI.gl = gl
	gl.glVertexStream1ivATI.__name__ = 'glVertexStream1ivATI'
	gl.glVertexStream1ivATI.errcheck = _errcheck
	gl.glVertexStream1ivATI.restype = GLvoid
	gl.glVertexStream1ivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glVertexStream1sATI = _getProcAddress('glVertexStream1sATI')
	gl.glVertexStream1sATI.gl = gl
	gl.glVertexStream1sATI.__name__ = 'glVertexStream1sATI'
	gl.glVertexStream1sATI.errcheck = _errcheck
	gl.glVertexStream1sATI.restype = GLvoid
	gl.glVertexStream1sATI.argtypes = (GLenum, GLshort,)
	gl.glVertexStream1svATI = _getProcAddress('glVertexStream1svATI')
	gl.glVertexStream1svATI.gl = gl
	gl.glVertexStream1svATI.__name__ = 'glVertexStream1svATI'
	gl.glVertexStream1svATI.errcheck = _errcheck
	gl.glVertexStream1svATI.restype = GLvoid
	gl.glVertexStream1svATI.argtypes = (GLenum, POINTER(GLshort),)
	gl.glVertexStream2dATI = _getProcAddress('glVertexStream2dATI')
	gl.glVertexStream2dATI.gl = gl
	gl.glVertexStream2dATI.__name__ = 'glVertexStream2dATI'
	gl.glVertexStream2dATI.errcheck = _errcheck
	gl.glVertexStream2dATI.restype = GLvoid
	gl.glVertexStream2dATI.argtypes = (GLenum, GLdouble, GLdouble,)
	gl.glVertexStream2dvATI = _getProcAddress('glVertexStream2dvATI')
	gl.glVertexStream2dvATI.gl = gl
	gl.glVertexStream2dvATI.__name__ = 'glVertexStream2dvATI'
	gl.glVertexStream2dvATI.errcheck = _errcheck
	gl.glVertexStream2dvATI.restype = GLvoid
	gl.glVertexStream2dvATI.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glVertexStream2fATI = _getProcAddress('glVertexStream2fATI')
	gl.glVertexStream2fATI.gl = gl
	gl.glVertexStream2fATI.__name__ = 'glVertexStream2fATI'
	gl.glVertexStream2fATI.errcheck = _errcheck
	gl.glVertexStream2fATI.restype = GLvoid
	gl.glVertexStream2fATI.argtypes = (GLenum, GLfloat, GLfloat,)
	gl.glVertexStream2fvATI = _getProcAddress('glVertexStream2fvATI')
	gl.glVertexStream2fvATI.gl = gl
	gl.glVertexStream2fvATI.__name__ = 'glVertexStream2fvATI'
	gl.glVertexStream2fvATI.errcheck = _errcheck
	gl.glVertexStream2fvATI.restype = GLvoid
	gl.glVertexStream2fvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glVertexStream2iATI = _getProcAddress('glVertexStream2iATI')
	gl.glVertexStream2iATI.gl = gl
	gl.glVertexStream2iATI.__name__ = 'glVertexStream2iATI'
	gl.glVertexStream2iATI.errcheck = _errcheck
	gl.glVertexStream2iATI.restype = GLvoid
	gl.glVertexStream2iATI.argtypes = (GLenum, GLint, GLint,)
	gl.glVertexStream2ivATI = _getProcAddress('glVertexStream2ivATI')
	gl.glVertexStream2ivATI.gl = gl
	gl.glVertexStream2ivATI.__name__ = 'glVertexStream2ivATI'
	gl.glVertexStream2ivATI.errcheck = _errcheck
	gl.glVertexStream2ivATI.restype = GLvoid
	gl.glVertexStream2ivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glVertexStream2sATI = _getProcAddress('glVertexStream2sATI')
	gl.glVertexStream2sATI.gl = gl
	gl.glVertexStream2sATI.__name__ = 'glVertexStream2sATI'
	gl.glVertexStream2sATI.errcheck = _errcheck
	gl.glVertexStream2sATI.restype = GLvoid
	gl.glVertexStream2sATI.argtypes = (GLenum, GLshort, GLshort,)
	gl.glVertexStream2svATI = _getProcAddress('glVertexStream2svATI')
	gl.glVertexStream2svATI.gl = gl
	gl.glVertexStream2svATI.__name__ = 'glVertexStream2svATI'
	gl.glVertexStream2svATI.errcheck = _errcheck
	gl.glVertexStream2svATI.restype = GLvoid
	gl.glVertexStream2svATI.argtypes = (GLenum, POINTER(GLshort),)
	gl.glVertexStream3dATI = _getProcAddress('glVertexStream3dATI')
	gl.glVertexStream3dATI.gl = gl
	gl.glVertexStream3dATI.__name__ = 'glVertexStream3dATI'
	gl.glVertexStream3dATI.errcheck = _errcheck
	gl.glVertexStream3dATI.restype = GLvoid
	gl.glVertexStream3dATI.argtypes = (GLenum, GLdouble, GLdouble, GLdouble,)
	gl.glVertexStream3dvATI = _getProcAddress('glVertexStream3dvATI')
	gl.glVertexStream3dvATI.gl = gl
	gl.glVertexStream3dvATI.__name__ = 'glVertexStream3dvATI'
	gl.glVertexStream3dvATI.errcheck = _errcheck
	gl.glVertexStream3dvATI.restype = GLvoid
	gl.glVertexStream3dvATI.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glVertexStream3fATI = _getProcAddress('glVertexStream3fATI')
	gl.glVertexStream3fATI.gl = gl
	gl.glVertexStream3fATI.__name__ = 'glVertexStream3fATI'
	gl.glVertexStream3fATI.errcheck = _errcheck
	gl.glVertexStream3fATI.restype = GLvoid
	gl.glVertexStream3fATI.argtypes = (GLenum, GLfloat, GLfloat, GLfloat,)
	gl.glVertexStream3fvATI = _getProcAddress('glVertexStream3fvATI')
	gl.glVertexStream3fvATI.gl = gl
	gl.glVertexStream3fvATI.__name__ = 'glVertexStream3fvATI'
	gl.glVertexStream3fvATI.errcheck = _errcheck
	gl.glVertexStream3fvATI.restype = GLvoid
	gl.glVertexStream3fvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glVertexStream3iATI = _getProcAddress('glVertexStream3iATI')
	gl.glVertexStream3iATI.gl = gl
	gl.glVertexStream3iATI.__name__ = 'glVertexStream3iATI'
	gl.glVertexStream3iATI.errcheck = _errcheck
	gl.glVertexStream3iATI.restype = GLvoid
	gl.glVertexStream3iATI.argtypes = (GLenum, GLint, GLint, GLint,)
	gl.glVertexStream3ivATI = _getProcAddress('glVertexStream3ivATI')
	gl.glVertexStream3ivATI.gl = gl
	gl.glVertexStream3ivATI.__name__ = 'glVertexStream3ivATI'
	gl.glVertexStream3ivATI.errcheck = _errcheck
	gl.glVertexStream3ivATI.restype = GLvoid
	gl.glVertexStream3ivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glVertexStream3sATI = _getProcAddress('glVertexStream3sATI')
	gl.glVertexStream3sATI.gl = gl
	gl.glVertexStream3sATI.__name__ = 'glVertexStream3sATI'
	gl.glVertexStream3sATI.errcheck = _errcheck
	gl.glVertexStream3sATI.restype = GLvoid
	gl.glVertexStream3sATI.argtypes = (GLenum, GLshort, GLshort, GLshort,)
	gl.glVertexStream3svATI = _getProcAddress('glVertexStream3svATI')
	gl.glVertexStream3svATI.gl = gl
	gl.glVertexStream3svATI.__name__ = 'glVertexStream3svATI'
	gl.glVertexStream3svATI.errcheck = _errcheck
	gl.glVertexStream3svATI.restype = GLvoid
	gl.glVertexStream3svATI.argtypes = (GLenum, POINTER(GLshort),)
	gl.glVertexStream4dATI = _getProcAddress('glVertexStream4dATI')
	gl.glVertexStream4dATI.gl = gl
	gl.glVertexStream4dATI.__name__ = 'glVertexStream4dATI'
	gl.glVertexStream4dATI.errcheck = _errcheck
	gl.glVertexStream4dATI.restype = GLvoid
	gl.glVertexStream4dATI.argtypes = (GLenum, GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glVertexStream4dvATI = _getProcAddress('glVertexStream4dvATI')
	gl.glVertexStream4dvATI.gl = gl
	gl.glVertexStream4dvATI.__name__ = 'glVertexStream4dvATI'
	gl.glVertexStream4dvATI.errcheck = _errcheck
	gl.glVertexStream4dvATI.restype = GLvoid
	gl.glVertexStream4dvATI.argtypes = (GLenum, POINTER(GLdouble),)
	gl.glVertexStream4fATI = _getProcAddress('glVertexStream4fATI')
	gl.glVertexStream4fATI.gl = gl
	gl.glVertexStream4fATI.__name__ = 'glVertexStream4fATI'
	gl.glVertexStream4fATI.errcheck = _errcheck
	gl.glVertexStream4fATI.restype = GLvoid
	gl.glVertexStream4fATI.argtypes = (GLenum, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glVertexStream4fvATI = _getProcAddress('glVertexStream4fvATI')
	gl.glVertexStream4fvATI.gl = gl
	gl.glVertexStream4fvATI.__name__ = 'glVertexStream4fvATI'
	gl.glVertexStream4fvATI.errcheck = _errcheck
	gl.glVertexStream4fvATI.restype = GLvoid
	gl.glVertexStream4fvATI.argtypes = (GLenum, POINTER(GLfloat),)
	gl.glVertexStream4iATI = _getProcAddress('glVertexStream4iATI')
	gl.glVertexStream4iATI.gl = gl
	gl.glVertexStream4iATI.__name__ = 'glVertexStream4iATI'
	gl.glVertexStream4iATI.errcheck = _errcheck
	gl.glVertexStream4iATI.restype = GLvoid
	gl.glVertexStream4iATI.argtypes = (GLenum, GLint, GLint, GLint, GLint,)
	gl.glVertexStream4ivATI = _getProcAddress('glVertexStream4ivATI')
	gl.glVertexStream4ivATI.gl = gl
	gl.glVertexStream4ivATI.__name__ = 'glVertexStream4ivATI'
	gl.glVertexStream4ivATI.errcheck = _errcheck
	gl.glVertexStream4ivATI.restype = GLvoid
	gl.glVertexStream4ivATI.argtypes = (GLenum, POINTER(GLint),)
	gl.glVertexStream4sATI = _getProcAddress('glVertexStream4sATI')
	gl.glVertexStream4sATI.gl = gl
	gl.glVertexStream4sATI.__name__ = 'glVertexStream4sATI'
	gl.glVertexStream4sATI.errcheck = _errcheck
	gl.glVertexStream4sATI.restype = GLvoid
	gl.glVertexStream4sATI.argtypes = (GLenum, GLshort, GLshort, GLshort, GLshort,)
	gl.glVertexStream4svATI = _getProcAddress('glVertexStream4svATI')
	gl.glVertexStream4svATI.gl = gl
	gl.glVertexStream4svATI.__name__ = 'glVertexStream4svATI'
	gl.glVertexStream4svATI.errcheck = _errcheck
	gl.glVertexStream4svATI.restype = GLvoid
	gl.glVertexStream4svATI.argtypes = (GLenum, POINTER(GLshort),)
	gl.glVertexWeightPointerEXT = _getProcAddress('glVertexWeightPointerEXT')
	gl.glVertexWeightPointerEXT.gl = gl
	gl.glVertexWeightPointerEXT.__name__ = 'glVertexWeightPointerEXT'
	gl.glVertexWeightPointerEXT.errcheck = _errcheck
	gl.glVertexWeightPointerEXT.restype = GLvoid
	gl.glVertexWeightPointerEXT.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glVertexWeightfEXT = _getProcAddress('glVertexWeightfEXT')
	gl.glVertexWeightfEXT.gl = gl
	gl.glVertexWeightfEXT.__name__ = 'glVertexWeightfEXT'
	gl.glVertexWeightfEXT.errcheck = _errcheck
	gl.glVertexWeightfEXT.restype = GLvoid
	gl.glVertexWeightfEXT.argtypes = (GLfloat,)
	gl.glVertexWeightfvEXT = _getProcAddress('glVertexWeightfvEXT')
	gl.glVertexWeightfvEXT.gl = gl
	gl.glVertexWeightfvEXT.__name__ = 'glVertexWeightfvEXT'
	gl.glVertexWeightfvEXT.errcheck = _errcheck
	gl.glVertexWeightfvEXT.restype = GLvoid
	gl.glVertexWeightfvEXT.argtypes = (POINTER(GLfloat),)
	gl.glVertexWeighthNV = _getProcAddress('glVertexWeighthNV')
	gl.glVertexWeighthNV.gl = gl
	gl.glVertexWeighthNV.__name__ = 'glVertexWeighthNV'
	gl.glVertexWeighthNV.errcheck = _errcheck
	gl.glVertexWeighthNV.restype = GLvoid
	gl.glVertexWeighthNV.argtypes = (GLhalfNV,)
	gl.glVertexWeighthvNV = _getProcAddress('glVertexWeighthvNV')
	gl.glVertexWeighthvNV.gl = gl
	gl.glVertexWeighthvNV.__name__ = 'glVertexWeighthvNV'
	gl.glVertexWeighthvNV.errcheck = _errcheck
	gl.glVertexWeighthvNV.restype = GLvoid
	gl.glVertexWeighthvNV.argtypes = (POINTER(GLhalfNV),)
	gl.glVideoCaptureNV = _getProcAddress('glVideoCaptureNV')
	gl.glVideoCaptureNV.gl = gl
	gl.glVideoCaptureNV.__name__ = 'glVideoCaptureNV'
	gl.glVideoCaptureNV.errcheck = _errcheck
	gl.glVideoCaptureNV.restype = GLenum
	gl.glVideoCaptureNV.argtypes = (GLuint, POINTER(GLuint), POINTER(GLuint64EXT),)
	gl.glVideoCaptureStreamParameterdvNV = _getProcAddress('glVideoCaptureStreamParameterdvNV')
	gl.glVideoCaptureStreamParameterdvNV.gl = gl
	gl.glVideoCaptureStreamParameterdvNV.__name__ = 'glVideoCaptureStreamParameterdvNV'
	gl.glVideoCaptureStreamParameterdvNV.errcheck = _errcheck
	gl.glVideoCaptureStreamParameterdvNV.restype = GLvoid
	gl.glVideoCaptureStreamParameterdvNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLdouble),)
	gl.glVideoCaptureStreamParameterfvNV = _getProcAddress('glVideoCaptureStreamParameterfvNV')
	gl.glVideoCaptureStreamParameterfvNV.gl = gl
	gl.glVideoCaptureStreamParameterfvNV.__name__ = 'glVideoCaptureStreamParameterfvNV'
	gl.glVideoCaptureStreamParameterfvNV.errcheck = _errcheck
	gl.glVideoCaptureStreamParameterfvNV.restype = GLvoid
	gl.glVideoCaptureStreamParameterfvNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLfloat),)
	gl.glVideoCaptureStreamParameterivNV = _getProcAddress('glVideoCaptureStreamParameterivNV')
	gl.glVideoCaptureStreamParameterivNV.gl = gl
	gl.glVideoCaptureStreamParameterivNV.__name__ = 'glVideoCaptureStreamParameterivNV'
	gl.glVideoCaptureStreamParameterivNV.errcheck = _errcheck
	gl.glVideoCaptureStreamParameterivNV.restype = GLvoid
	gl.glVideoCaptureStreamParameterivNV.argtypes = (GLuint, GLuint, GLenum, POINTER(GLint),)
	gl.glViewport = _getProcAddress('glViewport')
	gl.glViewport.gl = gl
	gl.glViewport.__name__ = 'glViewport'
	gl.glViewport.errcheck = _errcheck
	gl.glViewport.restype = GLvoid
	gl.glViewport.argtypes = (GLint, GLint, GLsizei, GLsizei,)
	gl.glViewportArrayv = _getProcAddress('glViewportArrayv')
	gl.glViewportArrayv.gl = gl
	gl.glViewportArrayv.__name__ = 'glViewportArrayv'
	gl.glViewportArrayv.errcheck = _errcheck
	gl.glViewportArrayv.restype = GLvoid
	gl.glViewportArrayv.argtypes = (GLuint, GLsizei, POINTER(GLfloat),)
	gl.glViewportIndexedf = _getProcAddress('glViewportIndexedf')
	gl.glViewportIndexedf.gl = gl
	gl.glViewportIndexedf.__name__ = 'glViewportIndexedf'
	gl.glViewportIndexedf.errcheck = _errcheck
	gl.glViewportIndexedf.restype = GLvoid
	gl.glViewportIndexedf.argtypes = (GLuint, GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glViewportIndexedfv = _getProcAddress('glViewportIndexedfv')
	gl.glViewportIndexedfv.gl = gl
	gl.glViewportIndexedfv.__name__ = 'glViewportIndexedfv'
	gl.glViewportIndexedfv.errcheck = _errcheck
	gl.glViewportIndexedfv.restype = GLvoid
	gl.glViewportIndexedfv.argtypes = (GLuint, POINTER(GLfloat),)
	gl.glWaitSync = _getProcAddress('glWaitSync')
	gl.glWaitSync.gl = gl
	gl.glWaitSync.__name__ = 'glWaitSync'
	gl.glWaitSync.errcheck = _errcheck
	gl.glWaitSync.restype = GLvoid
	gl.glWaitSync.argtypes = (GLsync, GLbitfield, GLuint64,)
	gl.glWaitSyncAPPLE = _getProcAddress('glWaitSyncAPPLE')
	gl.glWaitSyncAPPLE.gl = gl
	gl.glWaitSyncAPPLE.__name__ = 'glWaitSyncAPPLE'
	gl.glWaitSyncAPPLE.errcheck = _errcheck
	gl.glWaitSyncAPPLE.restype = GLvoid
	gl.glWaitSyncAPPLE.argtypes = (GLsync, GLbitfield, GLuint64,)
	gl.glWeightPathsNV = _getProcAddress('glWeightPathsNV')
	gl.glWeightPathsNV.gl = gl
	gl.glWeightPathsNV.__name__ = 'glWeightPathsNV'
	gl.glWeightPathsNV.errcheck = _errcheck
	gl.glWeightPathsNV.restype = GLvoid
	gl.glWeightPathsNV.argtypes = (GLuint, GLsizei, POINTER(GLuint), POINTER(GLfloat),)
	gl.glWeightPointerARB = _getProcAddress('glWeightPointerARB')
	gl.glWeightPointerARB.gl = gl
	gl.glWeightPointerARB.__name__ = 'glWeightPointerARB'
	gl.glWeightPointerARB.errcheck = _errcheck
	gl.glWeightPointerARB.restype = GLvoid
	gl.glWeightPointerARB.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glWeightPointerOES = _getProcAddress('glWeightPointerOES')
	gl.glWeightPointerOES.gl = gl
	gl.glWeightPointerOES.__name__ = 'glWeightPointerOES'
	gl.glWeightPointerOES.errcheck = _errcheck
	gl.glWeightPointerOES.restype = GLvoid
	gl.glWeightPointerOES.argtypes = (GLint, GLenum, GLsizei, POINTER(GLvoid),)
	gl.glWeightbvARB = _getProcAddress('glWeightbvARB')
	gl.glWeightbvARB.gl = gl
	gl.glWeightbvARB.__name__ = 'glWeightbvARB'
	gl.glWeightbvARB.errcheck = _errcheck
	gl.glWeightbvARB.restype = GLvoid
	gl.glWeightbvARB.argtypes = (GLint, POINTER(GLbyte),)
	gl.glWeightdvARB = _getProcAddress('glWeightdvARB')
	gl.glWeightdvARB.gl = gl
	gl.glWeightdvARB.__name__ = 'glWeightdvARB'
	gl.glWeightdvARB.errcheck = _errcheck
	gl.glWeightdvARB.restype = GLvoid
	gl.glWeightdvARB.argtypes = (GLint, POINTER(GLdouble),)
	gl.glWeightfvARB = _getProcAddress('glWeightfvARB')
	gl.glWeightfvARB.gl = gl
	gl.glWeightfvARB.__name__ = 'glWeightfvARB'
	gl.glWeightfvARB.errcheck = _errcheck
	gl.glWeightfvARB.restype = GLvoid
	gl.glWeightfvARB.argtypes = (GLint, POINTER(GLfloat),)
	gl.glWeightivARB = _getProcAddress('glWeightivARB')
	gl.glWeightivARB.gl = gl
	gl.glWeightivARB.__name__ = 'glWeightivARB'
	gl.glWeightivARB.errcheck = _errcheck
	gl.glWeightivARB.restype = GLvoid
	gl.glWeightivARB.argtypes = (GLint, POINTER(GLint),)
	gl.glWeightsvARB = _getProcAddress('glWeightsvARB')
	gl.glWeightsvARB.gl = gl
	gl.glWeightsvARB.__name__ = 'glWeightsvARB'
	gl.glWeightsvARB.errcheck = _errcheck
	gl.glWeightsvARB.restype = GLvoid
	gl.glWeightsvARB.argtypes = (GLint, POINTER(GLshort),)
	gl.glWeightubvARB = _getProcAddress('glWeightubvARB')
	gl.glWeightubvARB.gl = gl
	gl.glWeightubvARB.__name__ = 'glWeightubvARB'
	gl.glWeightubvARB.errcheck = _errcheck
	gl.glWeightubvARB.restype = GLvoid
	gl.glWeightubvARB.argtypes = (GLint, POINTER(GLubyte),)
	gl.glWeightuivARB = _getProcAddress('glWeightuivARB')
	gl.glWeightuivARB.gl = gl
	gl.glWeightuivARB.__name__ = 'glWeightuivARB'
	gl.glWeightuivARB.errcheck = _errcheck
	gl.glWeightuivARB.restype = GLvoid
	gl.glWeightuivARB.argtypes = (GLint, POINTER(GLuint),)
	gl.glWeightusvARB = _getProcAddress('glWeightusvARB')
	gl.glWeightusvARB.gl = gl
	gl.glWeightusvARB.__name__ = 'glWeightusvARB'
	gl.glWeightusvARB.errcheck = _errcheck
	gl.glWeightusvARB.restype = GLvoid
	gl.glWeightusvARB.argtypes = (GLint, POINTER(GLushort),)
	gl.glWindowPos2d = _getProcAddress('glWindowPos2d')
	gl.glWindowPos2d.gl = gl
	gl.glWindowPos2d.__name__ = 'glWindowPos2d'
	gl.glWindowPos2d.errcheck = _errcheck
	gl.glWindowPos2d.restype = GLvoid
	gl.glWindowPos2d.argtypes = (GLdouble, GLdouble,)
	gl.glWindowPos2dARB = _getProcAddress('glWindowPos2dARB')
	gl.glWindowPos2dARB.gl = gl
	gl.glWindowPos2dARB.__name__ = 'glWindowPos2dARB'
	gl.glWindowPos2dARB.errcheck = _errcheck
	gl.glWindowPos2dARB.restype = GLvoid
	gl.glWindowPos2dARB.argtypes = (GLdouble, GLdouble,)
	gl.glWindowPos2dMESA = _getProcAddress('glWindowPos2dMESA')
	gl.glWindowPos2dMESA.gl = gl
	gl.glWindowPos2dMESA.__name__ = 'glWindowPos2dMESA'
	gl.glWindowPos2dMESA.errcheck = _errcheck
	gl.glWindowPos2dMESA.restype = GLvoid
	gl.glWindowPos2dMESA.argtypes = (GLdouble, GLdouble,)
	gl.glWindowPos2dv = _getProcAddress('glWindowPos2dv')
	gl.glWindowPos2dv.gl = gl
	gl.glWindowPos2dv.__name__ = 'glWindowPos2dv'
	gl.glWindowPos2dv.errcheck = _errcheck
	gl.glWindowPos2dv.restype = GLvoid
	gl.glWindowPos2dv.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos2dvARB = _getProcAddress('glWindowPos2dvARB')
	gl.glWindowPos2dvARB.gl = gl
	gl.glWindowPos2dvARB.__name__ = 'glWindowPos2dvARB'
	gl.glWindowPos2dvARB.errcheck = _errcheck
	gl.glWindowPos2dvARB.restype = GLvoid
	gl.glWindowPos2dvARB.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos2dvMESA = _getProcAddress('glWindowPos2dvMESA')
	gl.glWindowPos2dvMESA.gl = gl
	gl.glWindowPos2dvMESA.__name__ = 'glWindowPos2dvMESA'
	gl.glWindowPos2dvMESA.errcheck = _errcheck
	gl.glWindowPos2dvMESA.restype = GLvoid
	gl.glWindowPos2dvMESA.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos2f = _getProcAddress('glWindowPos2f')
	gl.glWindowPos2f.gl = gl
	gl.glWindowPos2f.__name__ = 'glWindowPos2f'
	gl.glWindowPos2f.errcheck = _errcheck
	gl.glWindowPos2f.restype = GLvoid
	gl.glWindowPos2f.argtypes = (GLfloat, GLfloat,)
	gl.glWindowPos2fARB = _getProcAddress('glWindowPos2fARB')
	gl.glWindowPos2fARB.gl = gl
	gl.glWindowPos2fARB.__name__ = 'glWindowPos2fARB'
	gl.glWindowPos2fARB.errcheck = _errcheck
	gl.glWindowPos2fARB.restype = GLvoid
	gl.glWindowPos2fARB.argtypes = (GLfloat, GLfloat,)
	gl.glWindowPos2fMESA = _getProcAddress('glWindowPos2fMESA')
	gl.glWindowPos2fMESA.gl = gl
	gl.glWindowPos2fMESA.__name__ = 'glWindowPos2fMESA'
	gl.glWindowPos2fMESA.errcheck = _errcheck
	gl.glWindowPos2fMESA.restype = GLvoid
	gl.glWindowPos2fMESA.argtypes = (GLfloat, GLfloat,)
	gl.glWindowPos2fv = _getProcAddress('glWindowPos2fv')
	gl.glWindowPos2fv.gl = gl
	gl.glWindowPos2fv.__name__ = 'glWindowPos2fv'
	gl.glWindowPos2fv.errcheck = _errcheck
	gl.glWindowPos2fv.restype = GLvoid
	gl.glWindowPos2fv.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos2fvARB = _getProcAddress('glWindowPos2fvARB')
	gl.glWindowPos2fvARB.gl = gl
	gl.glWindowPos2fvARB.__name__ = 'glWindowPos2fvARB'
	gl.glWindowPos2fvARB.errcheck = _errcheck
	gl.glWindowPos2fvARB.restype = GLvoid
	gl.glWindowPos2fvARB.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos2fvMESA = _getProcAddress('glWindowPos2fvMESA')
	gl.glWindowPos2fvMESA.gl = gl
	gl.glWindowPos2fvMESA.__name__ = 'glWindowPos2fvMESA'
	gl.glWindowPos2fvMESA.errcheck = _errcheck
	gl.glWindowPos2fvMESA.restype = GLvoid
	gl.glWindowPos2fvMESA.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos2i = _getProcAddress('glWindowPos2i')
	gl.glWindowPos2i.gl = gl
	gl.glWindowPos2i.__name__ = 'glWindowPos2i'
	gl.glWindowPos2i.errcheck = _errcheck
	gl.glWindowPos2i.restype = GLvoid
	gl.glWindowPos2i.argtypes = (GLint, GLint,)
	gl.glWindowPos2iARB = _getProcAddress('glWindowPos2iARB')
	gl.glWindowPos2iARB.gl = gl
	gl.glWindowPos2iARB.__name__ = 'glWindowPos2iARB'
	gl.glWindowPos2iARB.errcheck = _errcheck
	gl.glWindowPos2iARB.restype = GLvoid
	gl.glWindowPos2iARB.argtypes = (GLint, GLint,)
	gl.glWindowPos2iMESA = _getProcAddress('glWindowPos2iMESA')
	gl.glWindowPos2iMESA.gl = gl
	gl.glWindowPos2iMESA.__name__ = 'glWindowPos2iMESA'
	gl.glWindowPos2iMESA.errcheck = _errcheck
	gl.glWindowPos2iMESA.restype = GLvoid
	gl.glWindowPos2iMESA.argtypes = (GLint, GLint,)
	gl.glWindowPos2iv = _getProcAddress('glWindowPos2iv')
	gl.glWindowPos2iv.gl = gl
	gl.glWindowPos2iv.__name__ = 'glWindowPos2iv'
	gl.glWindowPos2iv.errcheck = _errcheck
	gl.glWindowPos2iv.restype = GLvoid
	gl.glWindowPos2iv.argtypes = (POINTER(GLint),)
	gl.glWindowPos2ivARB = _getProcAddress('glWindowPos2ivARB')
	gl.glWindowPos2ivARB.gl = gl
	gl.glWindowPos2ivARB.__name__ = 'glWindowPos2ivARB'
	gl.glWindowPos2ivARB.errcheck = _errcheck
	gl.glWindowPos2ivARB.restype = GLvoid
	gl.glWindowPos2ivARB.argtypes = (POINTER(GLint),)
	gl.glWindowPos2ivMESA = _getProcAddress('glWindowPos2ivMESA')
	gl.glWindowPos2ivMESA.gl = gl
	gl.glWindowPos2ivMESA.__name__ = 'glWindowPos2ivMESA'
	gl.glWindowPos2ivMESA.errcheck = _errcheck
	gl.glWindowPos2ivMESA.restype = GLvoid
	gl.glWindowPos2ivMESA.argtypes = (POINTER(GLint),)
	gl.glWindowPos2s = _getProcAddress('glWindowPos2s')
	gl.glWindowPos2s.gl = gl
	gl.glWindowPos2s.__name__ = 'glWindowPos2s'
	gl.glWindowPos2s.errcheck = _errcheck
	gl.glWindowPos2s.restype = GLvoid
	gl.glWindowPos2s.argtypes = (GLshort, GLshort,)
	gl.glWindowPos2sARB = _getProcAddress('glWindowPos2sARB')
	gl.glWindowPos2sARB.gl = gl
	gl.glWindowPos2sARB.__name__ = 'glWindowPos2sARB'
	gl.glWindowPos2sARB.errcheck = _errcheck
	gl.glWindowPos2sARB.restype = GLvoid
	gl.glWindowPos2sARB.argtypes = (GLshort, GLshort,)
	gl.glWindowPos2sMESA = _getProcAddress('glWindowPos2sMESA')
	gl.glWindowPos2sMESA.gl = gl
	gl.glWindowPos2sMESA.__name__ = 'glWindowPos2sMESA'
	gl.glWindowPos2sMESA.errcheck = _errcheck
	gl.glWindowPos2sMESA.restype = GLvoid
	gl.glWindowPos2sMESA.argtypes = (GLshort, GLshort,)
	gl.glWindowPos2sv = _getProcAddress('glWindowPos2sv')
	gl.glWindowPos2sv.gl = gl
	gl.glWindowPos2sv.__name__ = 'glWindowPos2sv'
	gl.glWindowPos2sv.errcheck = _errcheck
	gl.glWindowPos2sv.restype = GLvoid
	gl.glWindowPos2sv.argtypes = (POINTER(GLshort),)
	gl.glWindowPos2svARB = _getProcAddress('glWindowPos2svARB')
	gl.glWindowPos2svARB.gl = gl
	gl.glWindowPos2svARB.__name__ = 'glWindowPos2svARB'
	gl.glWindowPos2svARB.errcheck = _errcheck
	gl.glWindowPos2svARB.restype = GLvoid
	gl.glWindowPos2svARB.argtypes = (POINTER(GLshort),)
	gl.glWindowPos2svMESA = _getProcAddress('glWindowPos2svMESA')
	gl.glWindowPos2svMESA.gl = gl
	gl.glWindowPos2svMESA.__name__ = 'glWindowPos2svMESA'
	gl.glWindowPos2svMESA.errcheck = _errcheck
	gl.glWindowPos2svMESA.restype = GLvoid
	gl.glWindowPos2svMESA.argtypes = (POINTER(GLshort),)
	gl.glWindowPos3d = _getProcAddress('glWindowPos3d')
	gl.glWindowPos3d.gl = gl
	gl.glWindowPos3d.__name__ = 'glWindowPos3d'
	gl.glWindowPos3d.errcheck = _errcheck
	gl.glWindowPos3d.restype = GLvoid
	gl.glWindowPos3d.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glWindowPos3dARB = _getProcAddress('glWindowPos3dARB')
	gl.glWindowPos3dARB.gl = gl
	gl.glWindowPos3dARB.__name__ = 'glWindowPos3dARB'
	gl.glWindowPos3dARB.errcheck = _errcheck
	gl.glWindowPos3dARB.restype = GLvoid
	gl.glWindowPos3dARB.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glWindowPos3dMESA = _getProcAddress('glWindowPos3dMESA')
	gl.glWindowPos3dMESA.gl = gl
	gl.glWindowPos3dMESA.__name__ = 'glWindowPos3dMESA'
	gl.glWindowPos3dMESA.errcheck = _errcheck
	gl.glWindowPos3dMESA.restype = GLvoid
	gl.glWindowPos3dMESA.argtypes = (GLdouble, GLdouble, GLdouble,)
	gl.glWindowPos3dv = _getProcAddress('glWindowPos3dv')
	gl.glWindowPos3dv.gl = gl
	gl.glWindowPos3dv.__name__ = 'glWindowPos3dv'
	gl.glWindowPos3dv.errcheck = _errcheck
	gl.glWindowPos3dv.restype = GLvoid
	gl.glWindowPos3dv.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos3dvARB = _getProcAddress('glWindowPos3dvARB')
	gl.glWindowPos3dvARB.gl = gl
	gl.glWindowPos3dvARB.__name__ = 'glWindowPos3dvARB'
	gl.glWindowPos3dvARB.errcheck = _errcheck
	gl.glWindowPos3dvARB.restype = GLvoid
	gl.glWindowPos3dvARB.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos3dvMESA = _getProcAddress('glWindowPos3dvMESA')
	gl.glWindowPos3dvMESA.gl = gl
	gl.glWindowPos3dvMESA.__name__ = 'glWindowPos3dvMESA'
	gl.glWindowPos3dvMESA.errcheck = _errcheck
	gl.glWindowPos3dvMESA.restype = GLvoid
	gl.glWindowPos3dvMESA.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos3f = _getProcAddress('glWindowPos3f')
	gl.glWindowPos3f.gl = gl
	gl.glWindowPos3f.__name__ = 'glWindowPos3f'
	gl.glWindowPos3f.errcheck = _errcheck
	gl.glWindowPos3f.restype = GLvoid
	gl.glWindowPos3f.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glWindowPos3fARB = _getProcAddress('glWindowPos3fARB')
	gl.glWindowPos3fARB.gl = gl
	gl.glWindowPos3fARB.__name__ = 'glWindowPos3fARB'
	gl.glWindowPos3fARB.errcheck = _errcheck
	gl.glWindowPos3fARB.restype = GLvoid
	gl.glWindowPos3fARB.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glWindowPos3fMESA = _getProcAddress('glWindowPos3fMESA')
	gl.glWindowPos3fMESA.gl = gl
	gl.glWindowPos3fMESA.__name__ = 'glWindowPos3fMESA'
	gl.glWindowPos3fMESA.errcheck = _errcheck
	gl.glWindowPos3fMESA.restype = GLvoid
	gl.glWindowPos3fMESA.argtypes = (GLfloat, GLfloat, GLfloat,)
	gl.glWindowPos3fv = _getProcAddress('glWindowPos3fv')
	gl.glWindowPos3fv.gl = gl
	gl.glWindowPos3fv.__name__ = 'glWindowPos3fv'
	gl.glWindowPos3fv.errcheck = _errcheck
	gl.glWindowPos3fv.restype = GLvoid
	gl.glWindowPos3fv.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos3fvARB = _getProcAddress('glWindowPos3fvARB')
	gl.glWindowPos3fvARB.gl = gl
	gl.glWindowPos3fvARB.__name__ = 'glWindowPos3fvARB'
	gl.glWindowPos3fvARB.errcheck = _errcheck
	gl.glWindowPos3fvARB.restype = GLvoid
	gl.glWindowPos3fvARB.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos3fvMESA = _getProcAddress('glWindowPos3fvMESA')
	gl.glWindowPos3fvMESA.gl = gl
	gl.glWindowPos3fvMESA.__name__ = 'glWindowPos3fvMESA'
	gl.glWindowPos3fvMESA.errcheck = _errcheck
	gl.glWindowPos3fvMESA.restype = GLvoid
	gl.glWindowPos3fvMESA.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos3i = _getProcAddress('glWindowPos3i')
	gl.glWindowPos3i.gl = gl
	gl.glWindowPos3i.__name__ = 'glWindowPos3i'
	gl.glWindowPos3i.errcheck = _errcheck
	gl.glWindowPos3i.restype = GLvoid
	gl.glWindowPos3i.argtypes = (GLint, GLint, GLint,)
	gl.glWindowPos3iARB = _getProcAddress('glWindowPos3iARB')
	gl.glWindowPos3iARB.gl = gl
	gl.glWindowPos3iARB.__name__ = 'glWindowPos3iARB'
	gl.glWindowPos3iARB.errcheck = _errcheck
	gl.glWindowPos3iARB.restype = GLvoid
	gl.glWindowPos3iARB.argtypes = (GLint, GLint, GLint,)
	gl.glWindowPos3iMESA = _getProcAddress('glWindowPos3iMESA')
	gl.glWindowPos3iMESA.gl = gl
	gl.glWindowPos3iMESA.__name__ = 'glWindowPos3iMESA'
	gl.glWindowPos3iMESA.errcheck = _errcheck
	gl.glWindowPos3iMESA.restype = GLvoid
	gl.glWindowPos3iMESA.argtypes = (GLint, GLint, GLint,)
	gl.glWindowPos3iv = _getProcAddress('glWindowPos3iv')
	gl.glWindowPos3iv.gl = gl
	gl.glWindowPos3iv.__name__ = 'glWindowPos3iv'
	gl.glWindowPos3iv.errcheck = _errcheck
	gl.glWindowPos3iv.restype = GLvoid
	gl.glWindowPos3iv.argtypes = (POINTER(GLint),)
	gl.glWindowPos3ivARB = _getProcAddress('glWindowPos3ivARB')
	gl.glWindowPos3ivARB.gl = gl
	gl.glWindowPos3ivARB.__name__ = 'glWindowPos3ivARB'
	gl.glWindowPos3ivARB.errcheck = _errcheck
	gl.glWindowPos3ivARB.restype = GLvoid
	gl.glWindowPos3ivARB.argtypes = (POINTER(GLint),)
	gl.glWindowPos3ivMESA = _getProcAddress('glWindowPos3ivMESA')
	gl.glWindowPos3ivMESA.gl = gl
	gl.glWindowPos3ivMESA.__name__ = 'glWindowPos3ivMESA'
	gl.glWindowPos3ivMESA.errcheck = _errcheck
	gl.glWindowPos3ivMESA.restype = GLvoid
	gl.glWindowPos3ivMESA.argtypes = (POINTER(GLint),)
	gl.glWindowPos3s = _getProcAddress('glWindowPos3s')
	gl.glWindowPos3s.gl = gl
	gl.glWindowPos3s.__name__ = 'glWindowPos3s'
	gl.glWindowPos3s.errcheck = _errcheck
	gl.glWindowPos3s.restype = GLvoid
	gl.glWindowPos3s.argtypes = (GLshort, GLshort, GLshort,)
	gl.glWindowPos3sARB = _getProcAddress('glWindowPos3sARB')
	gl.glWindowPos3sARB.gl = gl
	gl.glWindowPos3sARB.__name__ = 'glWindowPos3sARB'
	gl.glWindowPos3sARB.errcheck = _errcheck
	gl.glWindowPos3sARB.restype = GLvoid
	gl.glWindowPos3sARB.argtypes = (GLshort, GLshort, GLshort,)
	gl.glWindowPos3sMESA = _getProcAddress('glWindowPos3sMESA')
	gl.glWindowPos3sMESA.gl = gl
	gl.glWindowPos3sMESA.__name__ = 'glWindowPos3sMESA'
	gl.glWindowPos3sMESA.errcheck = _errcheck
	gl.glWindowPos3sMESA.restype = GLvoid
	gl.glWindowPos3sMESA.argtypes = (GLshort, GLshort, GLshort,)
	gl.glWindowPos3sv = _getProcAddress('glWindowPos3sv')
	gl.glWindowPos3sv.gl = gl
	gl.glWindowPos3sv.__name__ = 'glWindowPos3sv'
	gl.glWindowPos3sv.errcheck = _errcheck
	gl.glWindowPos3sv.restype = GLvoid
	gl.glWindowPos3sv.argtypes = (POINTER(GLshort),)
	gl.glWindowPos3svARB = _getProcAddress('glWindowPos3svARB')
	gl.glWindowPos3svARB.gl = gl
	gl.glWindowPos3svARB.__name__ = 'glWindowPos3svARB'
	gl.glWindowPos3svARB.errcheck = _errcheck
	gl.glWindowPos3svARB.restype = GLvoid
	gl.glWindowPos3svARB.argtypes = (POINTER(GLshort),)
	gl.glWindowPos3svMESA = _getProcAddress('glWindowPos3svMESA')
	gl.glWindowPos3svMESA.gl = gl
	gl.glWindowPos3svMESA.__name__ = 'glWindowPos3svMESA'
	gl.glWindowPos3svMESA.errcheck = _errcheck
	gl.glWindowPos3svMESA.restype = GLvoid
	gl.glWindowPos3svMESA.argtypes = (POINTER(GLshort),)
	gl.glWindowPos4dMESA = _getProcAddress('glWindowPos4dMESA')
	gl.glWindowPos4dMESA.gl = gl
	gl.glWindowPos4dMESA.__name__ = 'glWindowPos4dMESA'
	gl.glWindowPos4dMESA.errcheck = _errcheck
	gl.glWindowPos4dMESA.restype = GLvoid
	gl.glWindowPos4dMESA.argtypes = (GLdouble, GLdouble, GLdouble, GLdouble,)
	gl.glWindowPos4dvMESA = _getProcAddress('glWindowPos4dvMESA')
	gl.glWindowPos4dvMESA.gl = gl
	gl.glWindowPos4dvMESA.__name__ = 'glWindowPos4dvMESA'
	gl.glWindowPos4dvMESA.errcheck = _errcheck
	gl.glWindowPos4dvMESA.restype = GLvoid
	gl.glWindowPos4dvMESA.argtypes = (POINTER(GLdouble),)
	gl.glWindowPos4fMESA = _getProcAddress('glWindowPos4fMESA')
	gl.glWindowPos4fMESA.gl = gl
	gl.glWindowPos4fMESA.__name__ = 'glWindowPos4fMESA'
	gl.glWindowPos4fMESA.errcheck = _errcheck
	gl.glWindowPos4fMESA.restype = GLvoid
	gl.glWindowPos4fMESA.argtypes = (GLfloat, GLfloat, GLfloat, GLfloat,)
	gl.glWindowPos4fvMESA = _getProcAddress('glWindowPos4fvMESA')
	gl.glWindowPos4fvMESA.gl = gl
	gl.glWindowPos4fvMESA.__name__ = 'glWindowPos4fvMESA'
	gl.glWindowPos4fvMESA.errcheck = _errcheck
	gl.glWindowPos4fvMESA.restype = GLvoid
	gl.glWindowPos4fvMESA.argtypes = (POINTER(GLfloat),)
	gl.glWindowPos4iMESA = _getProcAddress('glWindowPos4iMESA')
	gl.glWindowPos4iMESA.gl = gl
	gl.glWindowPos4iMESA.__name__ = 'glWindowPos4iMESA'
	gl.glWindowPos4iMESA.errcheck = _errcheck
	gl.glWindowPos4iMESA.restype = GLvoid
	gl.glWindowPos4iMESA.argtypes = (GLint, GLint, GLint, GLint,)
	gl.glWindowPos4ivMESA = _getProcAddress('glWindowPos4ivMESA')
	gl.glWindowPos4ivMESA.gl = gl
	gl.glWindowPos4ivMESA.__name__ = 'glWindowPos4ivMESA'
	gl.glWindowPos4ivMESA.errcheck = _errcheck
	gl.glWindowPos4ivMESA.restype = GLvoid
	gl.glWindowPos4ivMESA.argtypes = (POINTER(GLint),)
	gl.glWindowPos4sMESA = _getProcAddress('glWindowPos4sMESA')
	gl.glWindowPos4sMESA.gl = gl
	gl.glWindowPos4sMESA.__name__ = 'glWindowPos4sMESA'
	gl.glWindowPos4sMESA.errcheck = _errcheck
	gl.glWindowPos4sMESA.restype = GLvoid
	gl.glWindowPos4sMESA.argtypes = (GLshort, GLshort, GLshort, GLshort,)
	gl.glWindowPos4svMESA = _getProcAddress('glWindowPos4svMESA')
	gl.glWindowPos4svMESA.gl = gl
	gl.glWindowPos4svMESA.__name__ = 'glWindowPos4svMESA'
	gl.glWindowPos4svMESA.errcheck = _errcheck
	gl.glWindowPos4svMESA.restype = GLvoid
	gl.glWindowPos4svMESA.argtypes = (POINTER(GLshort),)
	gl.glWriteMaskEXT = _getProcAddress('glWriteMaskEXT')
	gl.glWriteMaskEXT.gl = gl
	gl.glWriteMaskEXT.__name__ = 'glWriteMaskEXT'
	gl.glWriteMaskEXT.errcheck = _errcheck
	gl.glWriteMaskEXT.restype = GLvoid
	gl.glWriteMaskEXT.argtypes = (GLuint, GLuint, GLenum, GLenum, GLenum, GLenum,)
	gl.glCoverageModulationNV = _getProcAddress('glCoverageModulationNV')
	gl.glCoverageModulationNV.gl = gl
	gl.glCoverageModulationNV.__name__ = 'glCoverageModulationNV'
	gl.glCoverageModulationNV.errcheck = _errcheck
	gl.glCoverageModulationNV.restype = GLvoid
	gl.glCoverageModulationNV.argtypes = (GLenum,)
	gl.glCoverageModulationTableNV = _getProcAddress('glCoverageModulationTableNV')
	gl.glCoverageModulationTableNV.gl = gl
	gl.glCoverageModulationTableNV.__name__ = 'glCoverageModulationTableNV'
	gl.glCoverageModulationTableNV.errcheck = _errcheck
	gl.glCoverageModulationTableNV.restype = GLvoid
	gl.glCoverageModulationTableNV.argtypes = (GLsizei, POINTER(GLfloat),)
	gl.glFragmentCoverageColorNV = _getProcAddress('glFragmentCoverageColorNV')
	gl.glFragmentCoverageColorNV.gl = gl
	gl.glFragmentCoverageColorNV.__name__ = 'glFragmentCoverageColorNV'
	gl.glFragmentCoverageColorNV.errcheck = _errcheck
	gl.glFragmentCoverageColorNV.restype = GLvoid
	gl.glFragmentCoverageColorNV.argtypes = (GLuint,)
	gl.glFramebufferSampleLocationsfvNV = _getProcAddress('glFramebufferSampleLocationsfvNV')
	gl.glFramebufferSampleLocationsfvNV.gl = gl
	gl.glFramebufferSampleLocationsfvNV.__name__ = 'glFramebufferSampleLocationsfvNV'
	gl.glFramebufferSampleLocationsfvNV.errcheck = _errcheck
	gl.glFramebufferSampleLocationsfvNV.restype = GLvoid
	gl.glFramebufferSampleLocationsfvNV.argtypes = (GLenum, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glGetCoverageModulationTableNV = _getProcAddress('glGetCoverageModulationTableNV')
	gl.glGetCoverageModulationTableNV.gl = gl
	gl.glGetCoverageModulationTableNV.__name__ = 'glGetCoverageModulationTableNV'
	gl.glGetCoverageModulationTableNV.errcheck = _errcheck
	gl.glGetCoverageModulationTableNV.restype = GLvoid
	gl.glGetCoverageModulationTableNV.argtypes = (GLsizei, POINTER(GLfloat),)
	gl.glNamedFramebufferSampleLocationsfvNV = _getProcAddress('glNamedFramebufferSampleLocationsfvNV')
	gl.glNamedFramebufferSampleLocationsfvNV.gl = gl
	gl.glNamedFramebufferSampleLocationsfvNV.__name__ = 'glNamedFramebufferSampleLocationsfvNV'
	gl.glNamedFramebufferSampleLocationsfvNV.errcheck = _errcheck
	gl.glNamedFramebufferSampleLocationsfvNV.restype = GLvoid
	gl.glNamedFramebufferSampleLocationsfvNV.argtypes = (GLuint, GLuint, GLsizei, POINTER(GLfloat),)
	gl.glRasterSamplesEXT = _getProcAddress('glRasterSamplesEXT')
	gl.glRasterSamplesEXT.gl = gl
	gl.glRasterSamplesEXT.__name__ = 'glRasterSamplesEXT'
	gl.glRasterSamplesEXT.errcheck = _errcheck
	gl.glRasterSamplesEXT.restype = GLvoid
	gl.glRasterSamplesEXT.argtypes = (GLuint, GLboolean,)
	gl.glResolveDepthValuesNV = _getProcAddress('glResolveDepthValuesNV')
	gl.glResolveDepthValuesNV.gl = gl
	gl.glResolveDepthValuesNV.__name__ = 'glResolveDepthValuesNV'
	gl.glResolveDepthValuesNV.errcheck = _errcheck
	gl.glResolveDepthValuesNV.restype = GLvoid
	gl.glResolveDepthValuesNV.argtypes = ()
	gl.glSubpixelPrecisionBiasNV = _getProcAddress('glSubpixelPrecisionBiasNV')
	gl.glSubpixelPrecisionBiasNV.gl = gl
	gl.glSubpixelPrecisionBiasNV.__name__ = 'glSubpixelPrecisionBiasNV'
	gl.glSubpixelPrecisionBiasNV.errcheck = _errcheck
	gl.glSubpixelPrecisionBiasNV.restype = GLvoid
	gl.glSubpixelPrecisionBiasNV.argtypes = (GLuint, GLuint,)

	print 'PyGLoo: initialized API entry points'

	return gl
# }

